/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for internal device driver structures
*
****************************************************************************/

#ifndef __SNAP_AGP_DRIVER_H
#define __SNAP_AGP_DRIVER_H

#include "snap/agp/common.h"

/*--------------------- Macros and type definitions -----------------------*/

#pragma pack(1)

/* Strings embedded in the driver files */

#define SCITECH_STR         "SciTech Software, Inc."
#define COPYRIGHT_STR       SCITECH_COPYRIGHT_MSG_SHORT
#define CONFIG_SIGNATURE    "ofnIgifnoc_VRD"
#define CERTIFY_SIGNATURE   "ofnIyfitrec_VRD"

/****************************************************************************
REMARKS:
Structure used to track allocated AGP physical contexts

MEMBERS:
memoryBase      - Physical starting address of memory range
startPage       - Starting page in aperture for this context
numPages        - Number of reserved pages in memory range
committedPages  - Number of committed pages in memory range
types           - Flags the type of memory range
memory          - Array of pointers to committed memory pages
physMemory      - Array of physical memory addresses for committed memory
****************************************************************************/
typedef struct {
    N_physAddr      memoryBase;
    long            startPage;
    ulong           numPages;
    ulong           committedPages;
    ulong           type;
    void            **memory;
    N_physAddr      *physMemory;
    } AGP_physCtx;

/****************************************************************************
REMARKS:
Structure for block list of AGP physical context structures.

MEMBERS:
bl          - Memory pool containg objects
freeList    - Start of free list for pool
****************************************************************************/
typedef struct {
    N_blockList     bl;
    AGP_physCtx     *freeList;
    } AGP_physCtxList;

/****************************************************************************
REMARKS:
Structure containing vectors to chipset functions implemented for all cards
****************************************************************************/
typedef struct {
    N_int32         (*ReserveMemory)(AGP_physCtx *ctx);
    N_int32         (*ReleaseMemory)(AGP_physCtx *ctx);
    N_int32         (*MapMemory)(AGP_physCtx *ctx,N_int32 numPages,N_int32 startOffset);
    N_int32         (*UnMapMemory)(AGP_physCtx *ctx,N_int32 numPages,N_int32 startOffset);
    } AGP_chipsetFuncs;

/* Pack the main device context structure to 4 byte boundaries for
 * performance.
 */

#pragma pack(4)

/****************************************************************************
REMARKS:
Structure defining the main device context structure for the device driver.
This structure contains all variable declarations that are used by the
core generic device driver code. Private chipset specific variables should
not be declared in here by declared in a chipset specific structure.

HEADER:
graphics/driver.h

MEMBERS:
AGPDevice       - AGP device controlled by AGP bridge (ie: graphics card)
AGPBridge       - AGP bridge PCI device info
hostBridge      - Host bridge PCI device info
chipID          - Internal AGP chipset id
revID           - Internal AGP chipset revision id
deviceIndex     - Device index that this driver is controlling
shared          - True if the device driver is in shared memory
globalOpt       - Global configuration options
apertureBase    - Current AGP aperture base address
apertureSize    - Current AGP aperture size id
****************************************************************************/
typedef struct {
    /*-----------------------------------------------------------------------
     * Detected chipset configuration
     *---------------------------------------------------------------------*/
    PCIDeviceInfo       AGPDevice;
    PCIDeviceInfo       AGPBridge;
    PCIDeviceInfo       hostBridge;
    long                chipID;
    long                revID;
    N_int32             deviceIndex;
    N_int32             fakeAGPDriver;

    /*-----------------------------------------------------------------------
     * Current configuration information
     *---------------------------------------------------------------------*/
    AGP_globalOptions   globalOpt;
    N_physAddr          apertureBase;
    N_int32             apertureSize;

    /*-----------------------------------------------------------------------
     * Internal working variables
     *---------------------------------------------------------------------*/
    AGP_physCtxList     physList;

    /*-----------------------------------------------------------------------
     * Tables of public interface functions
     *---------------------------------------------------------------------*/
    AGP_driverFuncs     driver;

    /*-----------------------------------------------------------------------
     * Tables of internal helper functions
     *---------------------------------------------------------------------*/
    AGP_chipsetFuncs    chip;

    /*-----------------------------------------------------------------------
     * Public SNAP header block
     *---------------------------------------------------------------------*/
    AGP_devCtx          dc;
    } AGP_driver;

#pragma pack()

// Macros to delay for micro and milli seconds

#define usDelay(d)                  _AGP_drvImports.usDelay(d)
#define N_log                       _AGP_drvImports.N_log

// Macros to access buffer management callbacks

#define N_createBlockList(l,os)     _AGP_drvImports.N_createBlockList(l,os)
#define N_resizeBlockList(l)        _AGP_drvImports.N_resizeBlockList(l)
#define N_freeBlockList(l)          _AGP_drvImports.N_freeBlockList(l)
#define N_buildFreeList(l)          _AGP_drvImports.N_buildFreeList(l)

// Macros to access linked list callbacks

#define N_initList(l)               _AGP_drvImports.N_initList(l)
#define N_destroyList(l,f)          _AGP_drvImports.N_destroyList(l,f)
#define N_addAfter(l,n,a)           _AGP_drvImports.N_addAfter(l,n,a)
#define N_addToHead(l,n)            _AGP_drvImports.N_addToHead(l,n)
#define N_addToTail(l,n)            _AGP_drvImports.N_addToTail(l,n)
#define N_removeNext(l,n)           _AGP_drvImports.N_removeNext(l,n)
#define N_firstNode(l)              _AGP_drvImports.N_firstNode(l)
#define N_nextNode(p)               _AGP_drvImports.N_nextNode(p)
#define N_findNode(l,n)             _AGP_drvImports.N_findNode(l,n)

// Macros to read and write PCI registers

#define PCI_readRegB(index) \
    _AGP_drvImports.PCI_accessReg(index,0,PCI_READ_BYTE,&DV.PCIConfig)

#define PCI_readRegW(index) \
    _AGP_drvImports.PCI_accessReg(index,0,PCI_READ_WORD,&DV.PCIConfig)

#define PCI_readRegL(index) \
    _AGP_drvImports.PCI_accessReg(index,0,PCI_READ_DWORD,&DV.PCIConfig)

#define PCI_writeRegB(index,value)  \
    _AGP_drvImports.PCI_accessReg(index,value,PCI_WRITE_BYTE,&DV.PCIConfig)

#define PCI_writeRegW(index,value)  \
    _AGP_drvImports.PCI_accessReg(index,value,PCI_WRITE_WORD,&DV.PCIConfig)

#define PCI_writeRegL(index,value)  \
    _AGP_drvImports.PCI_accessReg(index,value,PCI_WRITE_DWORD,&DV.PCIConfig)

// Macros to read and write PCI registers (with passed in config space)

#define PCI_readRegExtB(info,index) \
    _AGP_drvImports.PCI_accessReg(index,0,PCI_READ_BYTE,info)

#define PCI_readRegExtW(info,index) \
    _AGP_drvImports.PCI_accessReg(index,0,PCI_READ_WORD,info)

#define PCI_readRegExtL(info,index) \
    _AGP_drvImports.PCI_accessReg(index,0,PCI_READ_DWORD,info)

#define PCI_writeRegExtB(info,index,value)  \
    _AGP_drvImports.PCI_accessReg(index,value,PCI_WRITE_BYTE,info)

#define PCI_writeRegExtW(info,index,value)  \
    _AGP_drvImports.PCI_accessReg(index,value,PCI_WRITE_WORD,info)

#define PCI_writeRegExtL(info,index,value)  \
    _AGP_drvImports.PCI_accessReg(index,value,PCI_WRITE_DWORD,info)

// Macros to read, modify, write PCI registers

#define PCI_setRegB(index,mask) \
    PCI_writeRegB(index,PCI_readRegB(index) | mask)

#define PCI_setRegW(index,mask) \
    PCI_writeRegW(index,PCI_readRegW(index) | mask)

#define PCI_setRegL(index,mask) \
    PCI_writeRegL(index,PCI_readRegL(index) | mask)

#define PCI_clrRegB(index,mask) \
    PCI_writeRegB(index,PCI_readRegB(index) & ~mask)

#define PCI_clrRegW(index,mask) \
    PCI_writeRegW(index,PCI_readRegW(index) & ~mask)

#define PCI_clrRegL(index,mask) \
    PCI_writeRegL(index,PCI_readRegL(index) & ~mask)

#define PCI_modRegB(index,mask,value)   \
    PCI_writeRegB(index,(PCI_readRegB(index) & ~mask) | ((value)&mask))

#define PCI_modRegW(index,mask,value)   \
    PCI_writeRegW(index,(PCI_readRegW(index) & ~mask) | ((value)&mask))

#define PCI_modRegL(index,mask,value)   \
    PCI_writeRegL(index,(PCI_readRegL(index) & ~mask) | ((value)&mask))

// Macros to read, modify, write PCI registers (passed in config)

#define PCI_setRegExtB(info,index,mask) \
    PCI_writeRegExtB(info,index,PCI_readRegExtB(info,index) | mask)

#define PCI_setRegExtW(info,index,mask) \
    PCI_writeRegExtW(info,index,PCI_readRegExtW(info,index) | mask)

#define PCI_setRegExtL(info,index,mask) \
    PCI_writeRegExtL(info,index,PCI_readRegExtL(info,index) | mask)

#define PCI_clrRegExtB(info,index,mask) \
    PCI_writeRegExtB(info,index,PCI_readRegExtB(info,index) & ~mask)

#define PCI_clrRegExtW(info,index,mask) \
    PCI_writeRegExtW(info,index,PCI_readRegExtW(info,index) & ~mask)

#define PCI_clrRegExtL(info,index,mask) \
    PCI_writeRegExtL(info,index,PCI_readRegExtL(info,index) & ~mask)

#define PCI_modRegExtB(info,index,mask,value)   \
    PCI_writeRegExtB(info,index,(PCI_readRegExtB(info,index) & ~mask) | ((value)&mask))

#define PCI_modRegExtW(info,index,mask,value)   \
    PCI_writeRegExtW(info,index,(PCI_readRegExtW(info,index) & ~mask) | ((value)&mask))

#define PCI_modRegExtL(info,index,mask,value)   \
    PCI_writeRegExtL(info,index,(PCI_readRegExtL(info,index) & ~mask) | ((value)&mask))

// Macro to initialise public API function structures

#define INIT_STRUCTURE(p)       \
    memset(&p,0xFF,sizeof(p));  \
    p.dwSize = sizeof(p);

/*---------------------------- Global Variables ---------------------------*/

/* Global variables declared in the chipset detect.c/config.c modules */

extern char                 *AGP_manufacturerName;
extern char                 *AGP_chipsetNames[];
extern AGP_configInfo       DRV_agpConfigInfo;
extern AGP_certifyChipInfo  DRV_agpCertifyChipInfo[];
extern AGP_certifyInfo      DRV_agpCertifyInfo;

/* Imports to OS specific functions */

extern N_imports    _VARAPI _N_imports;
extern PM_imports   _VARAPI _PM_imports;

/* Global driver device context structure */

extern AGP_driver           AGP;

/* Path to configuration directory */

extern char                 AGP_path[256];

/* Macro to access the SNAP device context in the global DC */

#define DC  (AGP.dc)

/* Externals to all imported function structures */

extern DRV_imports  _VARAPI _AGP_drvImports;
extern N_imports    _VARAPI _N_imports;
extern AGP_imports  _VARAPI _AGP_imports;

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

/* Generic functions in getfunc.c */

void        NAPI DRV_agpGetConfigInfo(AGP_configInfo *info);
void        NAPI DRV_agpGetCertifyInfo(AGP_certifyInfo *info);

/* Generic detection code in gdetect.c */

N_int32     NAPI DRV_agpInitDriver(void);
ibool       NAPI DRV_agpQueryFunctions(N_uint32 id,void *funcs);
void        NAPI DRV_agpUnloadDriver(void);

/* Chipset specific code in detect.c */

ibool       DRV_agpChipDetect(void);

/* Chipset specific code in driver.c */

N_int32     DRV_agpConfigure(void);
void        DRV_agpShutdown(void);

/* Generic API functions in gen_api.c */

N_int32     NAPI DRV_agpGetApertureSize(void);
N_physAddr  NAPI DRV_agpGetApertureBase(void);
N_int32     NAPI DRV_agpReservePhysical(N_uint32 numPages,N_int32 flags,void **PhysContext,N_physAddr *physAddr);
N_int32     NAPI DRV_agpReleasePhysical(void *PhysContext);
N_int32     NAPI DRV_agpCommitPhysical(void *PhysContext,N_uint32 numPages,N_uint32 startOffset,N_physAddr *physAddr);
N_int32     NAPI DRV_agpFreePhysical(void *PhysContext,N_uint32 numPages,N_uint32 startOffset);

/* Generic helper functions in gen_help.c */

uchar       DRV_findAGPCaps(PCIDeviceInfo *info,PCIAGPCapability *caps);
ibool       DRV_getAGPCaps(PCIDeviceInfo *info,PCIAGPCapability *caps);
ibool       DRV_setAGPCaps(PCIDeviceInfo *info,PCIAGPCapability *caps);
AGP_physCtx *DRV_newAGPPhysContext(void);
void        DRV_freeAGPPhysContext(AGP_physCtx *r);

/* TODO: Macros to lock and unlock exclusive access. I think these are only necessary
 *       for SMP machines, so we may need this down the track!
 */

#define LOCK_MASTER()
#define UNLOCK_MASTER()
#define LOCK_TARGET()
#define UNLOCK_TARGET()

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __SNAP_AGP_DRIVER_H */

