/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for the SNAP block list manager
*
****************************************************************************/

#ifndef __BLOCKLST_H
#define __BLOCKLST_H

#include "snap/graphics.h"
#include "snap/io.h"
#ifndef __WIN32_VXD__
#include <string.h>
#endif

/*---------------------- Macros and type definitions ----------------------*/

#pragma pack(4)

// Macro to determine the size of blocks allocated on the block list.
// We allocate the blocks in large chunks to avoid memory management
// overheads, and make the blocks multiples of a memory page.

#define BLOCK_SIZE          4096

// Macro to get the to the next block on the freelist

#define FREELIST_NEXT(p)    (*(void**)(p))

// Return a pointer to the user space given the address of the header of
// a node.

#define LST_USERSPACE(h)    ((void*)((node_t*)(h) + 1))

// Return a pointer to the header of a node, given the address of the
// user space.

#define LST_HEADER(n)       ((node_t*)(n) - 1)

// Return a pointer to the user space of the list's head node. This user
// space does not actually exist, but it is useful to be able to address
// it to enable insertion at the start of the list.

#define LST_HEAD(l)         LST_USERSPACE((l)->head)

// Determine if a list is empty

#define LST_EMPTY(l)        ((l)->count == 0)

/****************************************************************************
REMARKS:
Structure for the list header structure for tracking objects.

MEMBERS:
next    - Pointer to next object in the list
****************************************************************************/
typedef struct node_t {
    struct node_t   *next;
    } node_t;

/****************************************************************************
REMARKS:
Structure for the the linked list itself.

MEMBERS:
count   - Number of elements currently in list
head    - Pointer to head element of list
z       - Pointer to last node of list
hz      - Space for head and z nodes
****************************************************************************/
typedef struct {
    int     count;
    node_t  *head;
    node_t  *z;
    node_t  hz[2];
    } list_t;

/****************************************************************************
REMARKS:
Structure for the header block for blocks on the block list.

MEMBERS:
next    - Pointer to next block in list
****************************************************************************/
typedef struct block_t {
    struct block_t  *next;
    } block_t;

/****************************************************************************
REMARKS:
Structure for the header block for blocks on the block list.

MEMBERS:
curBlock    - List of memory pool blocks
blockCount  - Number of pool blocks allocated
max         - Maximum objects in block
objectSize  - Size of objects in block
****************************************************************************/
typedef struct {
    block_t *curBlock;
    int     blockCount;
    int     max;
    int     objectSize;
    } blockList_t;

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

/* Internal linked list management functions */

void        LM_initList(list_t *l);
void        LM_destroyList(list_t *l,void (*freeNode)(void *node));
void        LM_addAfter(list_t *l,void *node,void *after);
void        LM_addToHead(list_t *l,void *node);
void        LM_addToTail(list_t *l,void *node);
void *      LM_removeNext(list_t *l,void *node);
void *      LM_firstNode(list_t *l);
void *      LM_nextNode(void *prev);
void *      LM_findNode(list_t *l,void *n);

/* Internal block list management routines */

ibool       LM_createBlockList(blockList_t *l,int objectSize);
ibool       LM_resizeBlockList(blockList_t *l);
void        LM_freeBlockList(blockList_t *l);
void *      LM_buildFreeList(blockList_t *l);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __BLOCKLST_H */

