/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file describing the monitor configuration files for
*               SciTech Display Doctor. These files contain the monitor
*               operation limit information as well as complete CRTC timing
*               information for the product.
*
****************************************************************************/

#ifndef __SNAP_GRAPHICS_CRTC_H
#define __SNAP_GRAPHICS_CRTC_H

#include "snap/graphics.h"
#include "snap/ddc.h"

/*---------------------- Macros and type definitions ----------------------*/

#pragma pack(1)

#define CDBX_SIG_LEN        15
#define CDBX_NAME           "crtc%d%d.dat"
#define CDBX_SIGNATURE      "crtcnew.dat"
#define gaDefaultRefresh    0x80

/* Error codes returned by database routines */

typedef enum {
    CDBX_ok,                            /* Everything is cool           */
    CDBX_fileNotFound,                  /* Database file not found      */
    CDBX_corrupt,                       /* Database is corrupted        */
    CDBX_outOfMemory,                   /* Not enough memory to load DB */
    CDBX_notFound,                      /* Entry was not found in DB    */
    CDBX_IOError,                       /* Fatal IO error               */
    CDBX_invalidDB,                     /* Database handle is invalid   */
    CDBX_duplicateRecord                /* Record already exists        */
    } CDBX_errCodes;

typedef struct {
    char    signature[CDBX_SIG_LEN];    /* Signature                    */
    long    entries;                    /* Number of entries in database*/
    char    majorVersion;               /* Major version of data        */
    char    minorVersion;               /* Minor version of data        */
    } CDBX_header;

typedef struct {
    ushort      xRes;
    ushort      yRes;
    GA_CRTCInfo crtc;
    } CDBX_record;

typedef struct {
    CDBX_header h;                      /* Header record                */
    long        cursor;                 /* Current record index         */
    ibool       shared;                 /* True if the data is shared   */
    CDBX_record records[1];             /* Array of records             */
    } CDBX;

#pragma pack()

typedef void (NAPIP GA_createCRTCTimings_t)(int device,int head,const char *SNAPPath,PM_imports *pmImp);
typedef N_uint32 (NAPIP GA_GetClosestPixelClock_t)(N_int32 outputHead,N_int32 xRes,N_int32 yRes,N_int32 bitsPerPixel,N_uint32 pixelClock);

/*------------------------ Function Prototypes ----------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

/* Routines for reading/writing CRTC info from the CRTCxx.DAT database */

ibool   CDBX_init(int maxDevices,int maxHeads);
ibool   CDBX_open(int device,int numHeads);
ibool   CDBX_openHead(int device,int head);
void    CDBX_close(int device);
int     CDBX_getVersion(int device,int head,char *major,char *minor);
int     CDBX_isFirst(int device,int head);
int     CDBX_isLast(int device,int head);
int     CDBX_isEmpty(int device,int head);
int     CDBX_first(int device,int head,CDBX_record *rec);
int     CDBX_last(int device,int head,CDBX_record *rec);
int     CDBX_next(int device,int head,CDBX_record *rec);
int     CDBX_prev(int device,int head,CDBX_record *rec);
int     CDBX_insert(int device,int head,CDBX_record *rec,ibool updateDuplicate);
int     CDBX_update(int device,int head,CDBX_record *rec);
int     CDBX_flush(int device,int head);
ibool	CDBX_write(char *filename,void *data,int size);
ibool	CDBX_read(char *filename,void *data,int size);
int     CDBX_getErrCode(void);
char *  CDBX_getErrorMsg(void);

/* Utility function to return the address and size of the CRTC tables */

void    NAPI CDBX_getCRTCTable(int device,int head,void **data,int *size);

/* Utility function to find the timings for a specific resolution and refresh */

ibool   NAPI CDBX_getRefreshList(int device,int head,int attributes,int xRes,int yRes,int bitsPerPixel,int tripleScanLowRes,int doubleScan512,GA_monitor *monitor,N_uint32 maxClock,N_uint8 maxRefresh,int size,N_int32 *refreshRateList,GA_GetClosestPixelClock_t findClosestClock);
ibool   NAPI CDBX_getCRTCTimings(int device,int head,int attributes,int xRes,int yRes,int tripleScanLowRes,int doubleScan512,int refresh,GA_monitor *monitor,N_uint32 maxClock,N_uint8 maxRefresh,CDBX_record *crtc);

/* Utilities functions to change the configured refresh rates */

void    NAPI CDBX_setGlobalRefresh(int device,int head,int refresh);
ibool   NAPI CDBX_setCRTCTimings(int device,int head,int xRes,int yRes,GA_CRTCInfo *crtc,ibool addNewTiming,ibool makeDefault);

/* Obsolete functions for backwards binary compatibility */

ibool   NAPI CDBX_getRefreshListV1(int device,int attributes,int xRes,int yRes,int tripleScanLowRes,int doubleScan512,GA_monitor *monitor,N_uint32 maxClock,N_uint8 maxRefresh,int size,N_int32 *refreshRateList);
ibool   NAPI CDBX_getRefreshListV2(int device,int attributes,int xRes,int yRes,int bitsPerPixel,int tripleScanLowRes,int doubleScan512,GA_monitor *monitor,N_uint32 maxClock,N_uint8 maxRefresh,int size,N_int32 *refreshRateList,GA_GetClosestPixelClock_t findClosestClock);
ibool   NAPI CDBX_getCRTCTimingsV2(int device,int attributes,int xRes,int yRes,int tripleScanLowRes,int doubleScan512,int refresh,GA_monitor *monitor,N_uint32 maxClock,N_uint8 maxRefresh,CDBX_record *crtc);
void    NAPI CDBX_setGlobalRefreshV2(int device,int refresh);
void    NAPI CDBX_setCRTCTimingsV1(int device,int xRes,int yRes,GA_CRTCInfo *crtc);
ibool   NAPI CDBX_setCRTCTimingsV2(int device,int xRes,int yRes,GA_CRTCInfo *crtc,ibool addNewTiming,ibool makeDefault);

#ifdef __cplusplus
}
#endif

#endif  /* __SNAP_GRAPHICS_CRTC_H */

