/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for internal device driver structures. This
*               header file is only ever included for actual device
*               driver code itself, never by the device neutral loader
*               code.
*
****************************************************************************/

#ifndef __SNAP_GRAPHICS_DRIVER_H
#define __SNAP_GRAPHICS_DRIVER_H

#include "snap/graphics.h"
#include "snap/ddc.h"
#include "snap/common/io.h"
#include "snap/common/internal.h"
#include "snap/graphics/common.h"
#include "snap/graphics/version.h"
#include "snap/graphics/regdefs.h"
#include "snap/graphics/vga.h"
#include "snap/graphics/accel.h"
#include "snap/graphics/refresh.h"
#ifndef __WIN32_VXD__
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#endif

/*--------------------- Macros and type definitions -----------------------*/

#pragma pack(1)

/* Strings embedded in the driver files */

#define SCITECH_STR         "SciTech Software, Inc."
#define COPYRIGHT_STR       SCITECH_COPYRIGHT_MSG_SHORT
#define CONFIG_SIGNATURE    "ofnIgifnoc_VRD"
#define CERTIFY_SIGNATURE   "ofnIyfitrec_VRD"

/****************************************************************************
REMARKS:
Enumeration defining a number of internal constants used by the SNAP
drivers.

HEADER:
graphics/driver.h

MEMBERS:
GA_MAX_MODES            - Maximum number of display modes supported
GA_MAX_CLOCKS           - Maximum number of discrete pixel clocks
GA_PIXEL_FORMATS        - Maximum number of pixel formats (1 per RGB format)
****************************************************************************/
typedef enum {
    GA_MAX_MODES            = 255,
    GA_MAX_CLOCKS           = 128,
    GA_PIXEL_FORMATS        = 4,
    GA_MAX_HB_WIDTH         = (0x24 * 8),
    GA_MAX_VB_WIDTH         = 0x100,
    } GA_constants;

/****************************************************************************
REMARKS:
Enumeration defining the mode flags used to define the type of graphics
mode and the capabilities of this mode.

HEADER:
graphics/driver.h

MEMBERS:
gaVGA               - Mode is a VGA core compatible mode
gaSVGA              - Mode is an extended VGA compatible mode
gaNonVGA            - Mode in NonVGA
gaAccVGA            - Accelerated VGA compatible mode available
gaAccNonVGA         - Accelerated NonVGA mode available
gaLinearOnly        - Mode is a linear buffer only
gaBankedOnly        - Mode is a banked only
gaStereo            - Mode supports hardware stereo page flipping
gaHWCursor          - Mode supports hardware cursor
gaVideo             - Mode supports hardware video overlays
ga3D                - Mode supports hardware 3D acceleration
gaDoubleScan        - Mode is double scanned
gaTripleScan        - Mode is triple scanned
gaText              - Mode is a text mode
gaMultiplex         - Pixel multiplexing is enabled for this mode
gaInterleaved       - Controller is running in interleaved mode
gaSVGA4             - Mode is an extended VGA 4bpp compatible mode
gaUseBIOS           - The mode is set by the BIOS
gaPanningMode       - The mode must be panned for this output device
gaNoInterlaced      - Interlaced should be disabled for this mode
gaCLK_ScaleNone     - Pixel clock is not scaled
gaCLK_Scale05       - Pixel clock is scaled by 0.5 from normalised value
gaCLK_Scale15       - Pixel clock is scaled by 1.5 from normalised value
gaCLK_Scale20       - Pixel clock is scaled by 2 from normalised value
gaCLK_Scale30       - Pixel clock is scaled by 3 from normalised value
gaCLK_ScaleMask     - Mask for clock scale bits
gaCLK_ScaleShift    - Shift factor to get clock scale
gaCRTC_ScaleNone    - Horizontal CRTC is not scaled
gaCRTC_Scale05      - Horizontal CRTC is scaled by 0.5 from normalised value
gaCRTC_Scale15      - Horizontal CRTC is scaled by 1.5 from normalised value
gaCRTC_Scale20      - Horizontal CRTC is scaled by 2 from normalised value
gaCRTC_Scale30      - Horizontal CRTC is scaled by 3 from normalised value
gaCRTC_Scale40      - Horizontal CRTC is scaled by 4 from normalised value
gaCRTC_ScaleMask    - Mask for Horizontal CRTC scale bits
gaCRTC_ScaleShift   - Shift factor to get Horizontal CRTC scale
****************************************************************************/
typedef enum {
    gaVGA               = 0x00000001,
    gaSVGA              = 0x00000002,
    gaNonVGA            = 0x00000004,
    gaAccVGA            = 0x00000008,
    gaAccNonVGA         = 0x00000010,
    gaLinearOnly        = 0x00000020,
    gaBankedOnly        = 0x00000040,
    gaStereo            = 0x00000080,
    gaHWCursor          = 0x00000100,
    gaVideo             = 0x00000200,
    ga3D                = 0x00000400,
    gaIsDoubleScan      = 0x00000800,
    gaIsTripleScan      = 0x00001000,
    gaText              = 0x00002000,
    gaMultiplex         = 0x00004000,
    gaInterleaved       = 0x00008000,
    gaSVGA4             = 0x00010000,
    gaUseBIOS           = 0x00020000,
    gaPanningMode       = 0x00040000,
    gaNoInterlaced      = 0x00080000,
    gaCLK_ScaleNone     = 0x00000000,
    gaCLK_Scale05       = 0x00100000,
    gaCLK_Scale15       = 0x00200000,
    gaCLK_Scale20       = 0x00400000,
    gaCLK_Scale30       = 0x00800000,
    gaCLK_ScaleMask     = 0x00F00000,
    gaCLK_ScaleShift    = 20,
    gaCRTC_ScaleNone    = 0x00000000,
    gaCRTC_Scale05      = 0x01000000,
    gaCRTC_Scale15      = 0x02000000,
    gaCRTC_Scale20      = 0x04000000,
    gaCRTC_Scale30      = 0x08000000,
    gaCRTC_Scale40      = 0x10000000,
    gaCRTC_ScaleMask    = 0x1F000000,
    gaCRTC_ScaleShift   = 24,
    } GA_modeFlags;

/* Flags for common flag combinations */

#define gaVGAMode           (gaVGA | gaSVGA | gaAccVGA)
#define gaNonVGAMode        (gaNonVGA | gaAccNonVGA)
#define gaNoAccMode         (gaVGA | gaSVGA | gaNonVGA)
#define gaAccMode           (gaAccVGA | gaAccNonVGA)
#define gaAllModes          (gaVGAMode | gaAccMode)

/****************************************************************************
REMARKS:
Enumeration defining the different scale flags that we support. These values
are the unshift versions of the scale flags stored in the flags for the
mode entry.

HEADER:
graphics/driver.h

MEMBERS:
gaScaleNone - Not scaled
gaScale05   - Scaled by 0.5 from normalised value
gaScale15   - Scaled by 1.5 from normalised value
gaScale20   - Scaled by 2 from normalised value
gaScale30   - Scaled by 3 from normalised value
gaScale40   - Scaled by 4 from normalised value
****************************************************************************/
typedef enum {
    gaScaleNone     = 0x00,
    gaScale05       = 0x01,
    gaScale15       = 0x02,
    gaScale20       = 0x04,
    gaScale30       = 0x08,
    gaScale40       = 0x10,
    } GA_scaleFlags;

/****************************************************************************
REMARKS:
Clock programming flags for the GA_PLLParams flags member. These flags
inform the generic clock computation code to favour certain values during
the calculations.

HEADER:
graphics/driver.h

MEMBERS:
gaPLL_tableDriven       - Clock programming is table driven rather than a PLL
gaPLL_favorMinN         - Clock programming should favor minimum N values
gaPLL_favorMaxN         - Clock programming should favor maximum N values
gaPLL_favorMinM         - Clock programming should favor minimum M values
gaPLL_favorMaxM         - Clock programming should favor maximum M values
gaPLL_favorMinVCO       - Clock programming should favor minimum VCO values
gaPLL_favorMaxVCO       - Clock programming should favor maximum VCO values
gaPLL_clockHalve8Bit    - Clock supports halving in 8bpp modes
gaPLL_clockHalve16Bit   - Clock supports halving in 8bpp modes
gaPLL_clockHalve24Bit   - Clock supports halving in 8bpp modes
gaPLL_clockHalve32Bit   - Clock supports halving in 8bpp modes
gaPLL_clockQuarter8Bit  - Clock supports quartering in 8bpp modes
gaPLL_clockQuarter16Bit - Clock supports quartering in 8bpp modes
gaPLL_clockQuarter24Bit - Clock supports quartering in 8bpp modes
gaPLL_clockQuarter32Bit - Clock supports quartering in 8bpp modes
gaPLL_hotFrequencies    - Hot frequency list should be used to map out clocks
****************************************************************************/
typedef enum {
    gaPLL_tableDriven       = 0x40000000,
    gaPLL_favorMinN         = 0x00000001,
    gaPLL_favorMaxN         = 0x00000000,
    gaPLL_favorMinM         = 0x00000002,
    gaPLL_favorMaxM         = 0x00000000,
    gaPLL_favorMinVCO       = 0x00000004,
    gaPLL_favorMaxVCO       = 0x00000000,
    gaPLL_clockHalve8Bit    = 0x00000010,
    gaPLL_clockHalve16Bit   = 0x00000020,
    gaPLL_clockHalve24Bit   = 0x00000040,
    gaPLL_clockHalve32Bit   = 0x00000080,
    gaPLL_clockQuarter8Bit  = 0x00000100,
    gaPLL_clockQuarter16Bit = 0x00000200,
    gaPLL_clockQuarter24Bit = 0x00000400,
    gaPLL_clockQuarter32Bit = 0x00000800,
    gaPLL_hotFrequencies    = 0x00001000,
    } GA_PLLFlags;

#define gaPLL_clockHalve        (gaPLL_clockHalve8Bit | gaPLL_clockHalve16Bit | gaPLL_clockHalve24Bit | gaPLL_clockHalve32Bit)
#define gaPLL_clockQuarter      (gaPLL_clockQuarter8Bit | gaPLL_clockQuarter16Bit | gaPLL_clockQuarter24Bit | gaPLL_clockQuarter32Bit)

/****************************************************************************
REMARKS:
Structure defining the parameters to our generic pixel clock programming.
The following formulas are using to find the PLL programming parameters
that are used to program the pixel clock in the chip:

    Fout = (M+k1) * Fref * k3
           ------------------
              (N+k2) * 2^R

    Fvco = (M+k1) * Fref * k3
           ------------------
                 (N+k2)

Note: If the pixel clock is table driven, the list of hot frequencies
      becomes the list of available pixel clocks.

HEADER:
graphics/driver.h

MEMBERS:
flags       - PLL programming flags as defined in the GA_PLLFlags enumeration
fRef        - PLL Reference clock frequency as a 16.16 fixed point number
minFreq     - Minimum pixel clock frequency supported by DAC (default 12.59)
minVCO      - Minimum VCO frequency as a 16.16 fixed point value
maxVCO      - Maximum VCO frequency as a 16.16 fixed point value
minM        - Minimum M value for the above equation (0-255)
maxM        - Maximum M value for the above equation (0-255)
minN        - Minimum N value for the above equation (0-255)
maxN        - Maximum N value for the above equation (0-255)
minR        - Minimum R value for the above equation (0-5)
maxR        - Maximum R value for the above equation (0-5)
k1          - Constant k1 from the above equation
k2          - Constant k2 from the above equation
k3          - Constant k3 from the above equation
numClocks   - Number of clocks in the hot frequency list
hotFreq     - List of up to 128 hot frequencies to avoid programming
****************************************************************************/
typedef struct {
    ulong           flags;
    N_fix32         fRef;
    N_fix32         minFreq;
    N_fix32         minVCO;
    N_fix32         maxVCO;
    long            minM;
    long            maxM;
    long            minN;
    long            maxN;
    long            minR;
    long            maxR;
    long            k1;
    long            k2;
    long            k3;
    int             numClocks;
    N_fix32         hotFreq[GA_MAX_CLOCKS+1];
    } GA_PLLParams;

#define validClocks hotFreq
#define CLOSE_CLOCK N_FIXCONST(0.02)

/****************************************************************************
REMARKS:
Structure to collect all the chipset specific data structures together as
a single block.

HEADER:
graphics/driver.h
****************************************************************************/
typedef struct {
    ulong           *vgaChipsetProfile;
    uchar           *vgaChipsetData;
    ulong           *POSTchipsetProfile;
    ulong           *chipsetProfile;
    ulong           *colorDepthProfile;
    ulong           *modeProfile;
    uchar           *chipsetData;
    uchar           *colorDepth4Data;
    uchar           *colorDepth8Data;
    uchar           *colorDepth15Data;
    uchar           *colorDepth16Data;
    uchar           *colorDepth24Data;
    uchar           *colorDepth32Data;
    uchar           *SVGA4Data;
    uchar           *SVGAData;
    } GA_chipEntry;

/****************************************************************************
REMARKS:
Structure to collect all the RAMDAC specific data structures together as
a single block.

HEADER:
graphics/driver.h
****************************************************************************/
typedef struct {
    uchar           *VGAData;
    uchar           *colorDepth4Data;
    uchar           *colorDepth8Data;
    uchar           *colorDepth15Data;
    uchar           *colorDepth16Data;
    uchar           *colorDepth24Data;
    uchar           *colorDepth32Data;
    } GA_DACEntry;

/****************************************************************************
REMARKS:
Structure to describe the text mode character cell height for each display
mode.

HEADER:
graphics/driver.h
****************************************************************************/
typedef struct {
    ushort          mode;
    uchar           charHeight;
    } GA_textEntry;

/****************************************************************************
REMARKS:
Structure containing vectors to chipset functions implemented for all cards
{secret}
****************************************************************************/
typedef struct {
    void        (*getDefaultOptions)(void);
    void        (*getPendingOptions)(void);
    void        (*unlockRegisters)(void);
    void        (*unlockCRTCRegisters)(void);
    void        (*setDPMS)(uchar state);
    void        (*lockCRTCRegisters)(void);
    void        (*lockRegisters)(void);
    void        (*blankDisplay)(void);
    void        (*unblankDisplay)(void);
    void        (*setBIOSMode)(void);
    void        (*getBIOSRefreshList)(GA_modeEntry *mode,N_int32 *refreshList);
    void        (*preModeSet)(void);
    void        (*setChipsetRegisters)(void);
    void        (*setChipsetRegistersExt)(int initAccel);
    void        (*setMemoryClock)(void);
    void        (*setEngineClock)(void);
    void        (*setCRTCRegisters)(GA_CRTCInfo *crtc);
    void        (*selectClock)(int select,int divide);
    void        (*setPixelClock)(N_fix32 clock);
    N_int32     (*roundBytesPerLine)(N_int32 bytesPerLine,N_int32 bitsPerPixel);
    N_int32     (*setBytesPerLine)(N_int32 bytesPerLine);
    void        (*enableLinear)(void);
    void        (*disableLinear)(void);
    void        (*clearMemory)(ibool entireFramebuffer);
    void        (*setupAccelerator)(void);
    void        (*setOptions)(void);
    void        (*postModeSet)(void);
    ushort      (*setDACPage)(int page);
    void        (*clearDACPage)(void);
    int         (*getDACPort)(int index);
    uchar       (*rdDACReg)(uchar index);
    void        (*wrDACReg)(uchar index,uchar value);
    uchar       (*rdDACIdx)(uchar index);
    void        (*wrDACIdx)(uchar index,uchar value);
    ulong       (*getTextModeFlags)(GA_mode *mode);
    ulong       (*getModeFlags)(GA_mode *mode);
    ulong       (*getModeFlagsExt)(GA_mode *mode,N_int32 outputDevice,N_int32 outputHead,N_uint16 *physicalXRes,N_uint16 *physicalYRes);
    void        (*get2DAccelInfo)(GA_modeInfo *gaInfo,GA_modeEntry *entry,N_int32 virtualX,N_int32 virtualY);
    void        (*getVideoAccelInfo)(GA_modeInfo *gaInfo,GA_modeEntry *entry);
    void        (*get3DAccelInfo)(GA_modeInfo *gaInfo,GA_modeEntry *entry);
    GA_chipEntry* (*getChipData)(void);
    void        (*configChipsetFlags)(void);
    void        (*initCursor)(void);
    void        (*unloadDriver)(void);
    void        (*preModeReset)(void);
    void        (*postModeReset)(void);
    void        (*saveCursorState)(void);
    void        (*restoreCursorState)(void);
    } GA_chipsetFuncs;

/****************************************************************************
REMARKS:
Structure containing vectors to chipset functions implemented for VGA
compatible hardware only.
{secret}
****************************************************************************/
typedef struct {
    void        (*getBlankTimings)(VGA_hCRTC *h,VGA_vCRTC *v,GA_CRTCInfo *crtc);
    void        (*lockDownTimings)(VGA_hCRTC *h,VGA_vCRTC *v,GA_CRTCInfo *crtc);
    } GA_VGAFuncs;

/****************************************************************************
REMARKS:
Structure containing vectors to PLL clock specific helper functions
{secret}
****************************************************************************/
typedef struct {
    N_fix32     (*calculateClock)(GA_PLLParams *params,N_fix32 clock,int *M,int *N,int *R);
    int         (*setPLLPixelClock)(N_fix32 clock);
    void        (*setPLLMemoryClock)(N_fix32 clock);
    void        (*setPLLEngineClock)(N_fix32 clock);
    void        (*adjustMaxPixelClock)(GA_modeInfo *gaInfo,GA_modeEntry *entry);
    N_fix32     (*calculateClockExt)(N_int32 outputHead,GA_PLLParams *params,N_fix32 clock,int *M,int *N,int *R);
    } GA_PLLFuncs;

/****************************************************************************
REMARKS:
Structure containing vectors to RAMDAC specific help functions
{secret}
****************************************************************************/
typedef struct {
    void        (*setDACRegisters)(void);
    ibool       (*set8BitDAC)(ibool enable);
    } GA_DACFuncs;

/* Pack the main device context structure to 4 byte boundaries for
 * performance.
 */

#pragma pack(4)

/****************************************************************************
REMARKS:
Structure defining the main device context structure for the device driver.
This structure contains all variable declarations that are used by the
core generic device driver code. Private chipset specific variables should
not be declared in here by declared in a chipset specific structure.

HEADER:
graphics/driver.h

MEMBERS:
dc                  - Public SNAP Graphics Architecture device context
PCIConfig           - PCI device configuration space information
chipID              - Internal SuperVGA chipset id
revID               - Internal chipset revision id
dacID               - Current RAMDAC id
clockID             - Current clock chip id
mfrID               - Board manufacturer id if appropriate (0 otherwise)
biosID              - BIOS vendor id if appropriate (0 otherwise)
sharedIRQ           - True if our IRQ is shared with another device
VGAMemory           - Amount of memory for VGA modes (Kb)
VGA4Memory          - Amount of memory for VGA 4 bpp modes (Kb)
NVMemory            - Amount of memory for NonVGA modes (Kb)
maxBytesPerLineTxt  - Maximum bytes per line value for VGA text modes
maxBytesPerLine4    - Maximum bytes per line value for VGA 4bpp modes
maxBytesPerLine8    - Maximum bytes per line value for VGA 8bpp modes
maxBytesPerLineNV8  - Maximum bytes per line value for NonVGA modes
maxBytesPerLineNV16 - Maximum bytes per line value for NonVGA modes
maxBytesPerLineNV24 - Maximum bytes per line value for NonVGA modes
maxBytesPerLineNV32 - Maximum bytes per line value for NonVGA modes
maxAccelY           - Maximum Y coordinated for drawing functions (-1 if no restriction)
maxAccelLinearMem   - Maximum addressable memory for linear drawing functions (Kb)
resScanLines        - Number of scanlines to reserve before maxAccelY
reserveMemory       - Amount of memory to reserve (internal to driver)
defaultModeFlags    - Default values for mode flags in tables
HBlankEndMask       - Mask for VGA horizontal blank end
HSyncEndMask        - Mask for VGA horizontal sync end
VBlankEndMask       - Mask for VGA vertical blank end
VSyncEndMask        - Mask for VGA vertical sync end
dacIndex            - DAC index register for indexed DAC accesses
dacData             - DAC data register for indexed DAC accesses
pixelClockMaxVGATxt - Maximum normalised pixel clock for VGA text modes
pixelClockMaxVGA4   - Maximum normalised pixel clock for VGA 4bpp modes
pixelClockMaxVGA8   - Maximum normalised pixel clock for VGA 8bpp modes
pixelClockMax8      - Maximum normalised pixel clock for 8bpp modes
pixelClockMax16     - Maximum normalised pixel clock for 16bpp modes
pixelClockMax24     - Maximum normalised pixel clock for 24bpp modes
pixelClockMax32     - Maximum normalised pixel clock for 32bpp modes
pixelClock          - Pixel clock progamming parameters
PLLLoopDivide       - Pixel clock PLL loop divide factor
memoryClock         - Memory clock programming parameters
memoryClockVGA      - Memory clock frequency to program for VGA modes
engineClock         - Engine clock programming parameters
engineClockVGA      - Engine clock frequency to program for VGA modes
cntDACWidth         - Current RAMDAC PEL width (6 or 8)
cntBusWidth         - Current RAMDAC bus width
cntPixMux           - Current pixel multiplexing factor for RAMDAC/CRTC
cntMemoryClock      - Current memory clock frequency
cntEngineClock      - Current engine clock frequency
cntMode             - Current display mode structure
cntModeInfo         - GA_modeInfo structure for current mode
cntMonitor          - Current configured monitor information
opt                 - Current set of configuration options
AvailableModes      - List of all available display mode numbers
modeTable           - Table of all available display mode information
pixelFormats        - Table of pixel formats for RGB modes
clocks              - Table of discrete pixel clocks (if any)
maxClocks           - Maximum number of discrete pixel clocks
****************************************************************************/
typedef struct {
    /*-----------------------------------------------------------------------
     * Detected chipset configuration
     *---------------------------------------------------------------------*/
    PCIDeviceInfo       PCIConfig;
    long                chipID;
    long                revID;
    long                dacID;
    long                clockID;
    long                mfrID;
    long                biosID;
    long                sharedIRQ;
    ulong               DPMSCaps;
    ulong               SCICaps;
    int                 VGAMemory;
    int                 VGA4Memory;
    int                 VGA8Memory;
    int                 VGA16Memory;
    int                 VGA24Memory;
    int                 VGA32Memory;
    int                 NVMemory;
    int                 NV8Memory;
    int                 NV16Memory;
    int                 NV24Memory;
    int                 NV32Memory;
    void                *HWLinearMem;
    ulong               HWLinearSize;
    ulong               HWLinearBasePtr;
    ulong               maxBytesPerLineTxt;
    ulong               maxBytesPerLine4;
    ulong               maxBytesPerLine8;
    ulong               maxBytesPerLine24;
    ulong               maxBytesPerLineNV8;
    ulong               maxBytesPerLineNV16;
    ulong               maxBytesPerLineNV24;
    ulong               maxBytesPerLineNV32;
    ulong               maxAccelY;
    ulong               maxAccelLinearMem;
    ulong               resScanLines;
    ulong               reserveMemory;
    ulong               defaultModeFlags;
    ulong               HBlankEndMask;
    ulong               HSyncEndMask;
    ulong               VBlankEndMask;
    ulong               VSyncEndMask;
    ushort              *HCRTCRegsDesc;
    ushort              *VCRTCRegsDesc;
    REF2D_driver        *ref2d;

    /*-----------------------------------------------------------------------
     * Detected RAMDAC configuration
     *---------------------------------------------------------------------*/
    ulong               dacIndex;
    ulong               dacData;
    ulong               *dacProfile;
    GA_DACEntry         *dacModeData;

    /*-----------------------------------------------------------------------
     * Detected PLL Clock configuration.
     *---------------------------------------------------------------------*/
    N_fix32             pixelClockMaxTxt;
    N_fix32             pixelClockMax4;
    N_fix32             pixelClockMax8;
    N_fix32             pixelClockMax16;
    N_fix32             pixelClockMax24;
    N_fix32             pixelClockMax32;
    N_fix32             pixelClockMaxNV8;
    N_fix32             pixelClockMaxNV16;
    N_fix32             pixelClockMaxNV24;
    N_fix32             pixelClockMaxNV32;
    GA_PLLParams        pixelClock;
    ulong               PLLLoopDivide;
    GA_PLLParams        memoryClock;
    N_fix32             memoryClockVGA;
    GA_PLLParams        engineClock;
    N_fix32             engineClockVGA;

    /*-----------------------------------------------------------------------
     * Current configuration information
     *---------------------------------------------------------------------*/
    ulong               cntBank;
    ulong               cntDisplayStart;
    ulong               cntDACWidth;
    ulong               cntBusWidth;
    ulong               cntPixMux;
    ulong               cntMaxMem;
    N_int32             cntDPMSState;
    N_fix32             cntMemoryClock;
    N_fix32             cntEngineClock;
    N_fix32             cntPixelClock;
    GA_sharedInfo       *s;
    N_int32             cntVirtualX;
    N_int32             cntVirtualY;
    N_uint32            cntFlags;
    VGA_hCRTC           hCRTC;
    VGA_vCRTC           vCRTC;
    GA_monitorInfo      cntMonitor;
    GA_options          opt;
    GA_options          licensed;
    GA_globalOptions    globalOpt;
    GA_modeProfile      prof;
    N_int32             LCDUseBIOS;
    N_int32             outputDevice[GA_MAX_HEADS];

    /*-----------------------------------------------------------------------
     * Internal lookup tables
     *---------------------------------------------------------------------*/
    N_uint16            AvailableModes[GA_MAX_MODES+1];
    GA_pixelFormat      *pf24;
    GA_pixelFormat      *pf32;

    /*-----------------------------------------------------------------------
     * Tables of public interface functions
     *---------------------------------------------------------------------*/
    GA_initFuncs        init;
    GA_driverFuncs      driver;
    GA_cursorFuncs      cursor;
    GA_videoFuncs       video;
    GA_DPMSFuncs        dpms;
    GA_SCIFuncs         sci;
    GA_VBEFuncs         vbe;
    GA_bufferFuncs      buf;
    GA_clipperFuncs     clip;
    GA_2DStateFuncs     state2d;
    GA_2DRenderFuncs    draw_8bpp;
    GA_2DRenderFuncs    draw_16bpp;
    GA_2DRenderFuncs    draw_24bpp;
    GA_2DRenderFuncs    draw_32bpp;
    GA_2DRenderFuncs    drawSW;

    /*-----------------------------------------------------------------------
     * Tables of internal helper functions
     *---------------------------------------------------------------------*/
    GA_chipsetFuncs     chip;
    GA_VGAFuncs         vga;
    GA_PLLFuncs         pll;
    GA_DACFuncs         dac;

    /*-----------------------------------------------------------------------
     * Public SNAP header block
     *---------------------------------------------------------------------*/
    GA_devCtx           dc;
    } GA_driver;

/****************************************************************************
REMARKS:
Structure defining the functions needed for emulating ROP3 functions in
hardware using multiple passes with two way ROP's.
****************************************************************************/
typedef struct {
    void (*SaveMix)(void);
    void (*RestoreMix)(void);
    void (*BltFromTemp)(int left,int width,int dstLeft,int dstTop,int mix);
    void (*BltToTemp)(int left,int top,int width,int dstLeft,int mix);
    void (*BltToTempLin)(N_int32 srcOfs,N_int32 srcPitch,int left,int top,int width,int dstLeft,int mix);
    void (*BltToTempSys)(void *srcAddr,N_int32 srcPitch,int left,int top,int width,int dstLeft,int mix);
    void (*DrawPattToTemp)(int dstLeft,int dstTop,int width);
    void (*DrawColorPattToTemp)(int dstLeft,int dstTop,int width);
    } GA_ROP3EmuFuncs;

#pragma pack()

// TODO: Need to change all of these to support multi-head!!

// Macros to access shared data structures internally

#define cntModeNum  s->cntModeNum
#define cntMode     s->cntMode
#define cntModeInfo s->cntModeInfo
#define cntCRTC     s->cntCRTC

// Common DPMS Power Management flag combinations

#define DPMSNormal      (DPMS_standby | DPMS_suspend | DPMS_off)

// Common SCI capabilities flag combinations

#define SCINormal       (SCI_writeSCL | SCI_writeSDA | SCI_readSCL | SCI_readSDA)
#define SCIMinimal      (SCI_writeSCL | SCI_writeSDA | SCI_readSDA)

// Macros to define the scanline indices for the FatEllipseList functions

#define I_LEFTL     0
#define I_LEFTR     1
#define I_RIGHTL    2
#define I_RIGHTR    3

// Macros to set up function pointers

#if defined(__IBMC__) || defined(__IBMCPP__)
#define     CSTUB_VECTOR        (__c_code_ptr_t)DRV_stub
#define     STUB_VECTOR         (__code_ptr_t)DRV_stub
#define     ISTUB_VECTOR        (__icode_ptr_t)DRV_stub
#define     SSTUB_VECTOR        (__scode_ptr_t)DRV_stub
#define     LSTUB_VECTOR        (__lcode_ptr_t)DRV_stub
#define     UISTUB_VECTOR       (__uicode_ptr_t)DRV_stub
#define     USSTUB_VECTOR       (__uscode_ptr_t)DRV_stub
#define     ULSTUB_VECTOR       (__ulcode_ptr_t)DRV_stub
#else
#define     CSTUB_VECTOR        (void*)DRV_stub
#define     STUB_VECTOR         (void*)DRV_stub
#define     ISTUB_VECTOR        (void*)DRV_stub
#define     SSTUB_VECTOR        (void*)DRV_stub
#define     LSTUB_VECTOR        (void*)DRV_stub
#define     UISTUB_VECTOR       (void*)DRV_stub
#define     USSTUB_VECTOR       (void*)DRV_stub
#define     ULSTUB_VECTOR       (void*)DRV_stub
#endif

// Macros to check if function pointers are valid

#define IS_VALID_PTR(p) \
    (p && p != NOT_SUPPORTED && p != NOT_IMPLEMENTED && p != NEED_SPECS)

#define IS_CVALID_PTR(p)    \
    (p && p != CNOT_SUPPORTED && p != CNOT_IMPLEMENTED && p != CNEED_SPECS)

// Macro to find the length of a Win16 function

#define Win16Len(name)  ((ulong)name##_End - (ulong)name)

// Macros to set the drawing functions

#define Set2DDraw8(prefix,name)             \
    DV.draw_8bpp.name = prefix##_##name##8

#define Set2DDraw16(prefix,name)            \
    DV.draw_16bpp.name = prefix##_##name##16

#define Set2DDraw24(prefix,name)            \
    DV.draw_24bpp.name = prefix##_##name##24

#define Set2DDraw32(prefix,name)            \
    DV.draw_32bpp.name = prefix##_##name##32

#define Set2DDraw(prefix,name)              \
    DV.draw_8bpp.name = prefix##_##name;    \
    DV.draw_16bpp.name = prefix##_##name;   \
    DV.draw_24bpp.name = prefix##_##name;   \
    DV.draw_32bpp.name = prefix##_##name

#define NotSupported2DDraw8(name)           \
    DV.draw_8bpp.name = CNOT_SUPPORTED

#define NotSupported2DDraw16(name)          \
    DV.draw_16bpp.name = CNOT_SUPPORTED

#define NotSupported2DDraw24(name)          \
    DV.draw_24bpp.name = CNOT_SUPPORTED

#define NotSupported2DDraw32(name)          \
    DV.draw_32bpp.name = CNOT_SUPPORTED

#define NotSupported2DDraw(name)            \
    DV.draw_8bpp.name = CNOT_SUPPORTED;      \
    DV.draw_16bpp.name = CNOT_SUPPORTED;     \
    DV.draw_24bpp.name = CNOT_SUPPORTED;     \
    DV.draw_32bpp.name = CNOT_SUPPORTED

#define NeedSpecs2DDraw8(name)           \
    DV.draw_8bpp.name = CNEED_SPECS

#define NeedSpecs2DDraw16(name)          \
    DV.draw_16bpp.name = CNEED_SPECS

#define NeedSpecs2DDraw24(name)          \
    DV.draw_24bpp.name = CNEED_SPECS

#define NeedSpecs2DDraw32(name)          \
    DV.draw_32bpp.name = CNEED_SPECS

#define NeedSpecs2DDraw(name)            \
    DV.draw_8bpp.name = CNEED_SPECS;      \
    DV.draw_16bpp.name = CNEED_SPECS;     \
    DV.draw_24bpp.name = CNEED_SPECS;     \
    DV.draw_32bpp.name = CNEED_SPECS

#define NotImplemented2DDraw8(name)          \
    DV.draw_8bpp.name = CNOT_IMPLEMENTED

#define NotImplemented2DDraw16(name)             \
    DV.draw_16bpp.name = CNOT_IMPLEMENTED

#define NotImplemented2DDraw24(name)             \
    DV.draw_24bpp.name = CNOT_IMPLEMENTED

#define NotImplemented2DDraw32(name)             \
    DV.draw_32bpp.name = CNOT_IMPLEMENTED

#define NotImplemented2DDraw(name)           \
    DV.draw_8bpp.name = CNOT_IMPLEMENTED;      \
    DV.draw_16bpp.name = CNOT_IMPLEMENTED;     \
    DV.draw_24bpp.name = CNOT_IMPLEMENTED;     \
    DV.draw_32bpp.name = CNOT_IMPLEMENTED

// Macros to delay for micro and milli seconds

#define usDelay(d)  _DRV_imports.usDelay(d)
#define msDelay(d)  _DRV_imports.msDelay(d)
#define N_log       _DRV_imports.N_log

// Macros to read and write PCI registers

#define PCI_readRegB(index) \
    _DRV_imports.PCI_accessReg(index,0,PCI_READ_BYTE,&DV.PCIConfig)

#define PCI_readRegW(index) \
    _DRV_imports.PCI_accessReg(index,0,PCI_READ_WORD,&DV.PCIConfig)

#define PCI_readRegL(index) \
    _DRV_imports.PCI_accessReg(index,0,PCI_READ_DWORD,&DV.PCIConfig)

#define PCI_writeRegB(index,value)  \
    _DRV_imports.PCI_accessReg(index,value,PCI_WRITE_BYTE,&DV.PCIConfig)

#define PCI_writeRegW(index,value)  \
    _DRV_imports.PCI_accessReg(index,value,PCI_WRITE_WORD,&DV.PCIConfig)

#define PCI_writeRegL(index,value)  \
    _DRV_imports.PCI_accessReg(index,value,PCI_WRITE_DWORD,&DV.PCIConfig)

// Macros to read and write PCI registers (with passed in config space)

#define PCI_readRegExtB(info,index) \
    _DRV_imports.PCI_accessReg(index,0,PCI_READ_BYTE,info)

#define PCI_readRegExtW(info,index) \
    _DRV_imports.PCI_accessReg(index,0,PCI_READ_WORD,info)

#define PCI_readRegExtL(info,index) \
    _DRV_imports.PCI_accessReg(index,0,PCI_READ_DWORD,info)

#define PCI_writeRegExtB(info,index,value)  \
    _DRV_imports.PCI_accessReg(index,value,PCI_WRITE_BYTE,info)

#define PCI_writeRegExtW(info,index,value)  \
    _DRV_imports.PCI_accessReg(index,value,PCI_WRITE_WORD,info)

#define PCI_writeRegExtL(info,index,value)  \
    _DRV_imports.PCI_accessReg(index,value,PCI_WRITE_DWORD,info)

// Macros to read, modify, write PCI registers

#define PCI_setRegB(index,mask) \
    PCI_writeRegB(index,PCI_readRegB(index) | mask)

#define PCI_setRegW(index,mask) \
    PCI_writeRegW(index,PCI_readRegW(index) | mask)

#define PCI_setRegL(index,mask) \
    PCI_writeRegL(index,PCI_readRegL(index) | mask)

#define PCI_clrRegB(index,mask) \
    PCI_writeRegB(index,PCI_readRegB(index) & ~mask)

#define PCI_clrRegW(index,mask) \
    PCI_writeRegW(index,PCI_readRegW(index) & ~mask)

#define PCI_clrRegL(index,mask) \
    PCI_writeRegL(index,PCI_readRegL(index) & ~mask)

#define PCI_modRegB(index,mask,value)   \
    PCI_writeRegB(index,(PCI_readRegB(index) & ~mask) | ((value)&mask))

#define PCI_modRegW(index,mask,value)   \
    PCI_writeRegW(index,(PCI_readRegW(index) & ~mask) | ((value)&mask))

#define PCI_modRegL(index,mask,value)   \
    PCI_writeRegL(index,(PCI_readRegL(index) & ~mask) | ((value)&mask))

// Macros to read, modify, write PCI registers (passed in config)

#define PCI_setRegExtB(info,index,mask) \
    PCI_writeRegExtB(info,index,PCI_readRegExtB(info,index) | mask)

#define PCI_setRegExtW(info,index,mask) \
    PCI_writeRegExtW(info,index,PCI_readRegExtW(info,index) | mask)

#define PCI_setRegExtL(info,index,mask) \
    PCI_writeRegExtL(info,index,PCI_readRegExtL(info,index) | mask)

#define PCI_clrRegExtB(info,index,mask) \
    PCI_writeRegExtB(info,index,PCI_readRegExtB(info,index) & ~mask)

#define PCI_clrRegExtW(info,index,mask) \
    PCI_writeRegExtW(info,index,PCI_readRegExtW(info,index) & ~mask)

#define PCI_clrRegExtL(info,index,mask) \
    PCI_writeRegExtL(info,index,PCI_readRegExtL(info,index) & ~mask)

#define PCI_modRegExtB(info,index,mask,value)   \
    PCI_writeRegExtB(info,index,(PCI_readRegExtB(info,index) & ~mask) | ((value)&mask))

#define PCI_modRegExtW(info,index,mask,value)   \
    PCI_writeRegExtW(info,index,(PCI_readRegExtW(info,index) & ~mask) | ((value)&mask))

#define PCI_modRegExtL(info,index,mask,value)   \
    PCI_writeRegExtL(info,index,(PCI_readRegExtL(info,index) & ~mask) | ((value)&mask))

// Macro to access the complete PCI device list
#ifdef DETECT_ONLY
#define PCI_getPCIList(l,c) \
    _DETECT_imports.PCI_getPCIList(l,c)
#else
#define PCI_getPCIList(l,c) \
    _DRV_imports.PCI_getPCIList(l,c)
#endif

// Macros to read, modify, write RAMDAC registers

#define modDACReg(index,mask,value)             \
    DV.chip.wrDACReg(index,(DV.chip.rdDACReg(index) & ~mask) | ((value)&mask))

#define setDACReg(index,mask)                   \
    DV.chip.wrDACReg(index,DV.chip.rdDACReg(index) | mask)

#define clrDACReg(index,mask)                   \
    DV.chip.wrDACReg(index,DV.chip.rdDACReg(index) & ~mask)

// Macros to read, modify, write double indexed RAMDAC registers

#define modDACIdx(index,mask,value)             \
    DV.chip.wrDACIdx(index,(DV.chip.rdDACIdx(index) & ~mask) | ((value)&mask))

#define setDACIdx(index,mask)                   \
    DV.chip.wrDACIdx(index,DV.chip.rdDACIdx(index) | mask)

#define clrDACIdx(index,mask)                   \
    DV.chip.wrDACIdx(index,DV.chip.rdDACIdx(index) & ~mask)

/* Macros to make it easier to call important device driver functions */

#define DRV_mapBIOSImage(d)                 _DRV_imports.mapBIOSImage(d)
#define DRV_unmapBIOSImage(d)               _DRV_imports.unmapBIOSImage(d)
#define DDC_quickDetect(s,c)                _DRV_imports.DDC_quickDetect(s,c)
#define DDC_getFlatPanelResolution(s,c,x,y) _DRV_imports.DDC_getFlatPanelResolution(s,c,x,y)

/*---------------------------- Global Variables ---------------------------*/

/* Global variables declared in the chipset detect.c/config.c modules */

extern char         *DRV_manufacturerName;
extern char         *DRV_chipsetNames[];
extern char         *DRV_chipsetFilenames[];
extern char         *DRV_DACNames[];
extern char         *DRV_clockNames[];
extern ulong        *DRV_regProfile[];
extern ulong        *DRV_dacProfile[];
extern GA_chipEntry *DRV_chipData[];
extern GA_chipEntry *DRV_chipDataLCD[];
extern GA_chipEntry *DRV_chipDataPAL[];
extern GA_chipEntry *DRV_chipDataNTSC[];
extern GA_pixelFormat DRV_pf15;
extern GA_pixelFormat DRV_pf16;
extern GA_pixelFormat DRV_pf24_RGB;
extern GA_pixelFormat DRV_pf24_BGR;
extern GA_pixelFormat DRV_pf32_ARGB;
extern GA_pixelFormat DRV_pf32_ABGR;
extern GA_pixelFormat DRV_pf32_RGBA;
extern GA_pixelFormat DRV_pf32_BGRA;
extern GA_configInfo DRV_configInfo;
extern GA_certifyChipInfo   DRV_certifyChipInfo[];
extern GA_certifyInfo       DRV_certifyInfo;

/* Global variables for VGA mode register information */

extern ulong        VGA_chipsetProfile[];
extern ulong        VGA_modeProfile[];
extern uchar        VGA_chipsetData[];
extern uchar        VGA_TextData[];
extern uchar        VGA_640x480xVGA4Data[];
extern uchar        VGA_320x200xVGA8Data[];

/* Lookup table to mirror 8-bit values */

extern uchar        DRV_mirrorTable[];

/* Imports to OS specific functions */

extern N_imports    _VARAPI _N_imports;
extern PM_imports   _VARAPI _PM_imports;

/* Global driver device context structure */

extern GA_driver    DV;
extern GA_monitor   monitor[GA_MAX_HEADS];

/* Path to configuration directory */

extern char         DRV_path[256];
extern int          DRV_loadInGUI;

/* Macro to access the SNAP device context in the global DC */

#define DC  (DV.dc)

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

/* Generic functions in gen_api.c */

void        NAPI DRV_getConfigInfo(GA_configInfo *info);
void        NAPI DRV_getCertifyInfo(GA_certifyInfo *info);
void        NAPI DRV_getOptions(GA_options *opt);
void        NAPI DRV_setOptions(GA_options *opt);
void        NAPI DRV_setModeProfile(GA_modeProfile *prof);

/* Generic detection code in gdetect.c */

N_int32     NAPI DRV_initDriver(void);
ibool       NAPI DRV_queryFunctions(N_uint32 id,N_int32 safetyLevel,void *funcs);
void        NAPI DRV_unloadDriver(void);

/* Chipset specific code in detect.c */

ibool       CHIP_wakeup(int loadInGUI);
ibool       PNP_detect(void);
ibool       CHIP_detect(void);

/* Chipset specfic code in config.c */

N_int32     DRV_configure(int loadInGUI);

/* Functions to initialise the clock and ramdac modules */

void        CLOCK_init(void);
void        RAMDAC_init(void);

/* Functions to initialise the 2D acceleration module */

void        ACCEL2D_init(void);

/* Function to initialise the TV and flat panel interface */

void        TV_init(void);
void        LCDPANEL_init(void);

/* Mode init API functions in gen_api.c */

GA_chipEntry *DRV_getChipData(void);
GA_modeEntry *DRV_validateMode(N_int32 mode,N_int32 outputDevice,N_int32 outputHead,N_uint16 *physicalXRes,N_uint16 *physicalYRes);
N_fix32     DRV_findClosestDiscreteClock(N_fix32 clock,int bits,int *select,int *divide);
N_int32     NAPI DRV_GetVideoModeInfo(N_uint32 mode,GA_modeInfo *modeInfo);
N_int32     NAPI DRV_SetVideoMode(N_uint32 mode,N_int32 *virtualX,N_int32 *virtualY,N_int32 *bytesPerLine,N_int32 *maxMem,N_int32 refreshRate,GA_CRTCInfo *crtc);
N_uint32    NAPI DRV_GetVideoMode(void);
void        NAPI DRV_GetCurrentVideoModeInfo(GA_modeInfo *modeInfo);
N_int32     NAPI DRV_GetCustomVideoModeInfo(N_int32 xRes,N_int32 yRes,N_int32 virtualX,N_int32 virtualY,N_int32 bitsPerPixel,GA_modeInfo *modeInfo);
N_int32     NAPI DRV_SetCustomVideoMode(N_int32 xRes,N_int32 yRes,N_int32 bitsPerPixel,N_uint32 flags,N_int32 *virtualX,N_int32 *virtualY,N_int32 *bytesPerLine,N_int32 *maxMem,GA_CRTCInfo *crtc);
N_uint32    NAPI DRV_GetClosestPixelClockV1(N_int32 xRes,N_int32 yRes,N_int32 bitsPerPixel,N_uint32 pixelClock);
N_uint32    NAPI DRV_GetClosestPixelClock(N_int32 outputHead,N_int32 xRes,N_int32 yRes,N_int32 bitsPerPixel,N_uint32 pixelClock);
void        NAPI DRV_GetCRTCTimings(GA_CRTCInfo *crtc);
void        NAPI DRV_SetCRTCTimings(GA_CRTCInfo *crtc);
void        NAPI DRV_SaveCRTCTimings(GA_CRTCInfo *crtc);
void        NAPI DRV_SetGlobalRefresh(N_int32 refresh,N_int32 outputHead);
N_int32     NAPI DRV_SetDrawBuffer(GA_buffer *drawBuf);
N_int32     NAPI DRV_SetDisplayOutput(N_int32 device);
N_int32     NAPI DRV_GetDisplayOutput(void);
N_int32     NAPI DRV_SetDisplayOutputExt(N_int32 device,N_int32 outputHead);
N_int32     NAPI DRV_GetDisplayOutputExt(N_int32 outputHead);
void        NAPI DRV_SetSoftwareRenderFuncs(GA_2DRenderFuncs *softwareFuncs);
void        NAPI DRV_SetRef2dPointer(REF2D_driver *ref2d);
N_uint32    DRV_scaleValue(N_uint32 value,int scale);
N_uint32    DRV_unscaleValue(N_uint32 value,int scale);
N_uint32    DRV_scaleClock(N_uint32 value,GA_modeEntry *entry);
N_uint32    DRV_unscaleClock(N_uint32 value,GA_modeEntry *entry);
N_uint32    DRV_scaleCRTC(N_uint32 value);
N_uint32    DRV_unscaleCRTC(N_uint32 value);
void        NAPI DRV_SetDisplayStartXY(N_int32 x,N_int32 y,N_int32 waitVRT);
N_int32     NAPI DRV_GetCurrentRefreshRate(void);
N_int32     NAPI DRV_PollForDisplaySwitch(void);
void        NAPI DRV_PerformDisplaySwitch(void);
N_int32     NAPI DRV_SwitchPhysicalResolution(N_int32 physicalXResolution,N_int32 physicalYResolution,N_int32 refreshRate);
N_int32     NAPI DRV_GetVideoModeInfoExt(N_uint32 mode,GA_modeInfo *modeInfo,N_int32 outputDevice,N_int32 outputHead);
N_int32     NAPI DRV_GetCustomVideoModeInfoExt(N_int32 xRes,N_int32 yRes,N_int32 virtualX,N_int32 virtualY,N_int32 bitsPerPixel,GA_modeInfo *modeInfo,N_int32 outputDevice,N_int32 outputHead);
void        DRV_LoadTextFont(void);
void        DRV_LoadTextFontExt(int height);
N_int32     NAPI DRV_GetNumberOfHeads(void);
N_int32     NAPI DRV_SetActiveHead(N_int32 headIndex);
N_int32     NAPI DRV_GetActiveHead(void);

/* Generic VGA API functions in gen_vga.c */

N_int32     NAPI VGA_GetDisplayStartStatus(void);
N_int32     NAPI VGA_IsVSync(void);
void        NAPI VGA_WaitVSync(void);
void        NAPI VGA_SetVSyncWidth(N_int32 width);
N_int32     NAPI VGA_GetVSyncWidth(void);
void        NAPI VGA_SetPaletteData(GA_palette *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI VGA_GetPaletteData(GA_palette *pal,N_int32 num,N_int32 index);
void        NAPI VGA_SetGammaCorrectData(GA_palette *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI VGA_GetGammaCorrectData(GA_palette *pal,N_int32 num,N_int32 index);
void        VGA_delay(void);

/* Generic device support helper code in gen_help.c */

void        DRV_stub(void);
int         DRV_findLinearMem(int maxMB);
int         DRV_getStateSize(ulong *profile,ibool comments);
uchar *     DRV_saveState(ulong *profile,uchar *buffer,ibool comments);
uchar *     DRV_restoreState(ulong *profile,uchar *buffer,ibool comments);
int         DRV_getDACStateSize(ulong *profile,ibool comments);
uchar *     DRV_saveDACState(ulong *profile,uchar *buffer,ibool comments);
uchar *     DRV_restoreDACState(ulong *profile,uchar *buffer,ibool comments);
void        NAPI DRV_getUniqueFilename(char *filename,int type);
void        NAPI DRV_getMonitorInfo(GA_monitor *monitor,N_int32 outputHead);
void        NAPI DRV_setMonitorInfo(GA_monitor *monitor,N_int32 outputHead);
int         DRV_savePOSTRegisters(void);
int         DRV_restorePOSTRegisters(void);
void        DRV_writeMultiRegisters(uchar *h,ushort *profile);
void        DRV_setChipsetRegisters(void);
void        DRV_setDACRegisters(void);
void        DRV_clearMemory(ibool entireFramebuffer);
N_fix32     DRV_calculateClock(GA_PLLParams *params,N_fix32 clock,int *M,int *N,int *R);
N_int32 *   NAPI DRV_GetEmptyMixTable(void);
ulong       DRV_getTextModeFlags(GA_mode *mode);
void        NAPI DRV_SetPaletteDataExt(GA_paletteExt *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI DRV_GetPaletteDataExt(GA_paletteExt *pal,N_int32 num,N_int32 index);
void        NAPI DRV_SetGammaCorrectDataExt(GA_paletteExt *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI DRV_GetGammaCorrectDataExt(GA_paletteExt *pal,N_int32 num,N_int32 index);
void        DRV_configChipsetFlags(void);
void        DRV_adjustMaxPixelClock(GA_modeInfo *gaInfo,GA_modeEntry *entry);
void        DRV_setChipsetRegistersExt(int initAccel);

/* Generic DPMS and DDC helper functions in gen_ddc.c */

int         NAPI DRV_DPMSdetect(N_int16 *supportedStates);
void        NAPI DRV_DPMSsetState(N_int32 state);
int         NAPI DRV_SCIdetect(N_uint8 *capabilities,N_int32 *numMonitorPorts);

/* Generic VGA helper functions in gen_vga.c */

void        VGA_init(void);
void        VGA_unlockCRTCRegisters(void);
void        VGA_lockCRTCRegisters(void);
void        VGA_blankDisplay(void);
void        VGA_unblankDisplay(void);
void        VGA_preModeSet(void);
void        VGA_setCRTCRegisters(GA_CRTCInfo *crtc);
N_int32     VGA_roundBytesPerLine(N_int32 bytesPerLine,N_int32 bitsPerPixel);
N_int32     VGA_setBytesPerLine(N_int32 bytesPerLine);
void        VGA_postModeSet(void);
void        VGA_dacToPel(void);
uchar       VGA_triggerDac(void);
uchar       VGA_dacToCommand(void);
void        VGA_dacMode(uchar value);
void        VGA_dacModeMasked(uchar value,uchar mask);
int         VGA_getDACPort(int index);
void        VGA_getBlankTimings(VGA_hCRTC *h,VGA_vCRTC *v,GA_CRTCInfo *crtc);
void        VGA_lockDownTimings(VGA_hCRTC *h,VGA_vCRTC *v,GA_CRTCInfo *crtc);
ibool       VGA_set8BitDAC(ibool enable);
void        VGA_loadFont(uchar *data,int cellHeight);

/* Discrete clock detection functions */

#ifdef INCLUDE_DISCRETECLOCKS
void        ICS9161CLK_init(void);
int         ICS9161_setPLLPixelClock(N_fix32 clock);
void        DISCRETECLK_init(void);
void        DRV_detectDiscreteClocks(int maxClocks);
void        DRV_analyseDiscreteClocks(void);
#endif

/* Generic linear buffer detetion code in find_lin.c */

ulong       DRV_getMainMemSize(void);
void        DRV_saveMode(void);
void        DRV_writeLinearTestString(void);
ibool       DRV_checkLinearBuffer(ulong addr);
ulong       DRV_findLinearBuffer(ulong maxAddr);
void        DRV_restoreMode(void);

/* CMOS register functions in _cmos.asm */

uchar       _ASMAPI DRV_readCMOS(int index);

/* Generic VBE/Core emulation helper functions */

ibool       NAPI DRVVBE_SetBytesPerLine(int bytesPerLine,int *newBytes);
ibool       NAPI DRVVBE_Set8BitDAC(ibool enable);
void        NAPI DRVVBE_SetPaletteData(GA_palette *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI DRVVBE_GetPaletteData(GA_palette *pal,N_int32 num,N_int32 index);

/* Generic driver helper functions in ports.c */

uchar       DRV_rdDACReg(uchar index);
void        DRV_wrDACReg(uchar index,uchar value);
uchar       DRV_rdDACIdx(uchar index);
void        DRV_wrDACIdx(uchar index,uchar value);

/* Generic driver ellipse functions in ellipse.c */

void        NAPI DRV_DrawEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI DRV_DrawPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI DRV_DrawColorPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI DRV_DrawFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI DRV_DrawPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI DRV_DrawColorPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);

/* Generic ROP3 functions in rop3emu.c */

void        DRV_SetupROP3Emulator(GA_ROP3EmuFuncs *emuFuncs);
void        NAPI DRV_BitBltPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI DRV_BitBltPattLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI DRV_BitBltPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI DRV_BitBltColorPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI DRV_BitBltColorPattLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI DRV_BitBltColorPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __SNAP_GRAPHICS_DRIVER_H */

