/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for defining the strcutures for declaring
*               register tables.
*
****************************************************************************/

#ifndef __SNAP_GRAPHICS_REGDEFS_H
#define __SNAP_GRAPHICS_REGDEFS_H

/*--------------------- Macros and type definitions -----------------------*/

#pragma pack(1)

/****************************************************************************
REMARKS:
Enumeration defining the different types of register width.

HEADER:
graphics/regdef.h

MEMBERS:
REG_SZ_BYTE     - Register is byte width (8-bits)
REG_SZ_WORD     - Register is word width (16-bits)
REG_SZ_DWORD    - Register is dword width (32-bits)
****************************************************************************/
typedef enum {
    REG_SZ_BYTE,
    REG_SZ_WORD,
    REG_SZ_DWORD,
    } GA_RegisterAccessType;

/****************************************************************************
REMARKS:
Enumeration defining the masks to access different parts of a register
definition. The register definition bits are defined as follows:

    B31     : 0 = Memory Mapped, 1 = I/O
    B30-29  : 00 = BYTE, 01 = WORD, 10 = DWORD

    B28-0   MEM: 29 bit (512Mb) offset for register

    B25     I/O: 0 = Fixed I/O, 1 = Relocateable I/O
    B24     I/O: 0 = Indexed I/O, 1 = Direct I/O
    B23-16  I/O: Index for register (if B24 = 0)
    B15-0   I/O: 16-bit I/O port address or offset

HEADER:
graphics/regdef.h

MEMBERS:
REG_MMIO_OFFSET_MASK    - Mask to extract MMIO register offset
REG_IO_PORT_MASK        - Mask to extract IO register port
REG_IO_INDEX_MASK       - Mask to extract IO register index
REG_IO_INDEX_SHIFT      - Shift to extract IO register index
REG_IO_IS_INDEXED       - Mask to determine if register is indexed
REG_IO_IS_RELOC         - Mask to determine if register is relocateble
REG_TYPE_MASK           - Mask to extract the access width
REG_TYPE_SHIFT          - Shift to extract the access width
REG_IS_MMIO             - Mask to determine if register is memory mappped
****************************************************************************/
#define REG_MMIO_OFFSET_MASK    0x1FFFFFFFUL
#define REG_IO_PORT_MASK        0x0000FFFFUL
#define REG_IO_INDEX_MASK       0x00FF0000UL
#define REG_IO_INDEX_SHIFT      16
#define REG_IO_IS_INDEXED       0x01000000UL
#define REG_IO_IS_RELOC         0x02000000UL
#define REG_TYPE_MASK           0x70000000UL
#define REG_TYPE_SHIFT          29
#define REG_IS_MMIO             0x80000000UL

/* Macros to extract components and test conditions for register defs */

#define regMMIOOffset(v)        ((v) & REG_MMIO_OFFSET_MASK)
#define regIOPort(v)            ((v) & REG_IO_PORT_MASK)
#define regIOIndex(v)           (((v) & REG_IO_INDEX_MASK) >> REG_IO_INDEX_SHIFT)
#define regIsIndexed(v)         ((v) & REG_IO_IS_INDEXED)
#define regIsReloc(v)           ((v) & REG_IO_IS_RELOC)
#define regWidth(v)             (((v) & REG_TYPE_MASK) >> REG_TYPE_SHIFT)
#define regIsMMIO(v)            ((v) & REG_IS_MMIO)
#define regIsReference(v)       (regIsMMIO(v) && (regMMIOOffset(v) == REG_MMIO_OFFSET_MASK))

/* Macros to declare MMIO registers */

#define _MMIO_B(offset)         ((offset) | (REG_SZ_BYTE << 29) | REG_IS_MMIO)
#define _MMIO_W(offset)         ((offset) | (REG_SZ_WORD << 29) | REG_IS_MMIO)
#define _MMIO_D(offset)         ((offset) | (REG_SZ_DWORD << 29) | REG_IS_MMIO)
#define _MMIO_IB(port,index)    ((port) | (REG_SZ_BYTE << 29) | (index << 16) | REG_IO_IS_INDEXED | REG_IS_MMIO)
#define _MMIO_IW(port,index)    ((port) | (REG_SZ_WORD << 29) | (index << 16) | REG_IO_IS_INDEXED | REG_IS_MMIO)
#define _MMIO_ID(port,index)    ((port) | (REG_SZ_DWORD << 29)| (index << 16) | REG_IO_IS_INDEXED | REG_IS_MMIO)

/* Macros to declare fixed I/O registers */

#define _FIXIO_I(port,index)    ((port) | (REG_SZ_BYTE << 29) | (index << 16) | REG_IO_IS_INDEXED)
#define _FIXIO_IW(port,index)   ((port) | (REG_SZ_WORD << 29) | (index << 16) | REG_IO_IS_INDEXED)
#define _FIXIO_ID(port,index)   ((port) | (REG_SZ_DWORD << 29) | (index << 16) | REG_IO_IS_INDEXED)
#define _FIXIO_B(port)          ((port) | (REG_SZ_BYTE << 29))
#define _FIXIO_W(port)          ((port) | (REG_SZ_WORD << 29))
#define _FIXIO_D(port)          ((port) | (REG_SZ_DWORD << 29))

/* Macros to declare relocateable I/O registers */

#define _RELIO_I(port,index)    (_FIXIO_I(port,index) | REG_IO_IS_RELOC)
#define _RELIO_IW(port,index)   (_FIXIO_IW(port,index) | REG_IO_IS_RELOC)
#define _RELIO_ID(port,index)   (_FIXIO_ID(port,index) | REG_IO_IS_RELOC)
#define _RELIO_B(port)          (_FIXIO_B(port) | REG_IO_IS_RELOC)
#define _RELIO_W(port)          (_FIXIO_W(port) | REG_IO_IS_RELOC)
#define _RELIO_D(port)          (_FIXIO_D(port) | REG_IO_IS_RELOC)

/* Macro to declare a reference to another register table by symbol. We do
 * this by inserting a dummy MMIO regiser definition with a huge offset
 * that we can test against. The next entry in the table is the address
 * of another chipset table to include register definitions from.
 */

#define REG_REFERENCE(name)     _MMIO_B(REG_MMIO_OFFSET_MASK),(ulong)(name)
#define REG_TERMINATOR          0xFFFFFFFFL

/* Macros to declare data in the mode data tables */

#ifdef __BIG_ENDIAN__
#define	_BYTE(v)				(v)
#define	_WORD(v)				(uchar)(v >> 8),(uchar)(v)
#define	_DWORD(v)				(uchar)(v >> 24),(uchar)(v >> 16),(uchar)(v >> 8),(uchar)(v)
#else
#define	_BYTE(v)				(v)
#define	_WORD(v)				(uchar)(v),(uchar)(v >> 8)
#define	_DWORD(v)				(uchar)(v),(uchar)(v >> 8),(uchar)(v >> 16),(uchar)(v >> 24)
#endif
#define	DATA_TERMINATOR			0xFF

#ifdef  __ANALYSE__

/* Macros to declare registers with comment strings */

#define MMIO_IB(port,index,desc)    _MMIO_IB(port,index),(ulong)desc
#define MMIO_IW(port,index,desc)    _MMIO_IW(port,index),(ulong)desc
#define MMIO_ID(port,index,desc)    _MMIO_ID(port,index),(ulong)desc
#define MMIO_B(offset,desc)         _MMIO_B(offset),(ulong)desc
#define MMIO_W(offset,desc)         _MMIO_W(offset),(ulong)desc
#define MMIO_D(offset,desc)         _MMIO_D(offset),(ulong)desc
#define FIXIO_I(port,index,desc)    _FIXIO_I(port,index),(ulong)desc
#define FIXIO_IW(port,index,desc)   _FIXIO_IW(port,index),(ulong)desc
#define FIXIO_ID(port,index,desc)   _FIXIO_ID(port,index),(ulong)desc
#define FIXIO_B(port,desc)          _FIXIO_B(port),(ulong)desc
#define FIXIO_W(port,desc)          _FIXIO_W(port),(ulong)desc
#define FIXIO_D(port,desc)          _FIXIO_D(port),(ulong)desc
#define RELIO_I(port,index,desc)    _RELIO_I(port,index),(ulong)desc
#define RELIO_IW(port,index,desc)   _RELIO_IW(port,index),(ulong)desc
#define RELIO_ID(port,index,desc)   _RELIO_ID(port,index),(ulong)desc
#define RELIO_B(port,desc)          _RELIO_B(port),(ulong)desc
#define RELIO_W(port,desc)          _RELIO_W(port),(ulong)desc
#define RELIO_D(port,desc)          _RELIO_D(port),(ulong)desc

#else

/* Macros to declare registers without comment strings */

#define MMIO_I(port,index,desc)     _MMIO_I(port,index)
#define MMIO_IW(port,index,desc)    _MMIO_IW(port,index)
#define MMIO_ID(port,index,desc)    _MMIO_ID(port,index)
#define MMIO_B(offset,desc)         _MMIO_B(offset)
#define MMIO_W(offset,desc)         _MMIO_W(offset)
#define MMIO_D(offset,desc)         _MMIO_D(offset)
#define FIXIO_I(port,index,desc)    _FIXIO_I(port,index)
#define FIXIO_IW(port,index,desc)   _FIXIO_IW(port,index)
#define FIXIO_ID(port,index,desc)   _FIXIO_ID(port,index)
#define FIXIO_B(port,desc)          _FIXIO_B(port)
#define FIXIO_W(port,desc)          _FIXIO_W(port)
#define FIXIO_D(port,desc)          _FIXIO_D(port)
#define RELIO_I(port,index,desc)    _RELIO_I(port,index)
#define RELIO_IW(port,index,desc)   _RELIO_IW(port,index)
#define RELIO_ID(port,index,desc)   _RELIO_ID(port,index)
#define RELIO_B(port,desc)          _RELIO_B(port)
#define RELIO_W(port,desc)          _RELIO_W(port)
#define RELIO_D(port,desc)          _RELIO_D(port)

#endif  // __ANALYSE__

/****************************************************************************
REMARKS:
Enumeration defining the masks to access different parts of a RAMDAC register
definition. The register definition bits are defined as follows:

        B15     : 0 = Indexed, 1 = Double Indexed
        B14-0   : Index of register

HEADER:
graphics/regdef.h

MEMBERS:
DAC_INDEX_MASK          - Mask to extract the RAMDAC register index
DAC_IS_DOUBLE_INDEXED   - Set to indicate it is double indexed
****************************************************************************/
#define DAC_INDEX_MASK          0x7FFF
#define DAC_IS_DOUBLE_INDEXED   0x8000

#define dacIndex(v)         ((v) & DAC_INDEX_MASK)
#define dacIsDbIndex(v)     ((v) & DAC_IS_DOUBLE_INDEXED)
#define dacIsReference(v)   ((v) == 0xFFFFFFFEL)

#ifdef  __ANALYSE__

/* Macros to declare RAMDAC registers with comment strings */

#define DAC_I(index,desc)   (ulong)(index),(ulong)desc
#define DAC_D(index,desc)   (ulong)((index) | DAC_IS_DOUBLE_INDEXED),(ulong)desc

#else

/* Macros to declare RAMDAC registers without comment strings */

#define DAC_I(index,desc)   (ulong)(index)
#define DAC_D(index,desc)   (ulong)((index) | DAC_IS_DOUBLE_INDEXED)

#endif

#define DAC_REFERENCE(name) (ulong)0xFFFFFFFEL,(ulong)(name)
#define DAC_TERMINATOR      0xFFFFFFFFL

#pragma pack()

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __SNAP_GRAPHICS_REGDEFS_H */

