/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Function prototypes for generic VGA and driver helper
*               functions.
*
****************************************************************************/

#ifndef __SNAP_GRAPHICS_VGA_H
#define __SNAP_GRAPHICS_VGA_H

/*--------------------- Macros and type definitions -----------------------*/

/****************************************************************************
REMARKS:
Structure to store the VGA normalised horiztonal CRTC timings. The difference
is that we include the VGA blank timings and horizontal display values
in the tables, allowing the chipset specific code to override the default
values that are programmed into the hardware by the VGA compatible CRTC
programming code.

HEADER:
graphics/vga.h

MEMBERS:
hTotal          - Horizontal total
hDisp           - Horizontal displayed
hBlankStart     - Horizontal blank start
hBlankEnd       - Horizontal blank end
hSyncStart      - Horizontal sync start
hSyncEnd        - Horizontal sync end
****************************************************************************/
typedef struct {
    ushort      hTotal;
    ushort      hDisp;
    ushort      hBlankStart;
    ushort      hBlankEnd;
    ushort      hSyncStart;
    ushort      hSyncEnd;
    } VGA_hCRTC;

/****************************************************************************
REMARKS:
Structure to store the VGA normalised vertical CRTC timings. The difference
is that we include the VGA blank timings and horizontal display values
in the tables, allowing the chipset specific code to override the default
values that are programmed into the hardware by the VGA compatible CRTC
programming code.

HEADER:
graphics/vga.h

MEMBERS:
vTotal          - Vertical total
vDisp           - Vertical displayed
vBlankStart     - Vertical blank start
vBlankEnd       - Vertical blank end
vSyncStart      - Vertical sync start
vSyncEnd        - Vertical sync end
****************************************************************************/
typedef struct {
    ushort      vTotal;
    ushort      vDisp;
    ushort      vBlankStart;
    ushort      vBlankEnd;
    ushort      vSyncStart;
    ushort      vSyncEnd;
    } VGA_vCRTC;

// Macros for declaring flags on VGA CRTC register descriptions

#define vgaInterlace    0x8000  /* Flag for interlace on hSyncStart     */
#define vgaHSyncMask    0x7FFF  /* Mask for horizontal sync start       */

#define vgaSyncVNeg     0x8000  /* V sync negative on vSyncStart        */
#define vgaSyncHNeg     0x4000  /* H sync negative on vSyncStart        */
#define vgaDblScan      0x2000  /* Flag for double scan on vSyncStart   */
#define vgaTrpScan      0x1000  /* Flag for triple scan on vSyncStart   */
#define vgaSyncPolMask  0xC000  /* Sync polarities mask                 */
#define vgaVSyncMask    0x0FFF  /* Mask for vertical sync start         */

// Standard RAMDAC register definitions

#define dacSTDwrInx     0
#define dacSTDpelData   1
#define dacSTDpelMask   2
#define dacSTDrdInx     3

// Macros used to encode the extended register decoding tables. This used
// primarily for accessing CRTC register values, but it can be used for
// any other purpose that requires collecting bits from multiple registers.

#define RG_NEWBYTE_CMD  0x0000
#define RG_NEWWORD_CMD  0x1000
#define RG_DPORT_CMD    0x2000
#define RG_INDEX_CMD    0x3000
#define RG_IPORT_CMD    0x4000
#define RG_MASK_CMD     0x5000
#define RG_STARTBIT_CMD 0x6000
#define RG_FINISH_CMD   0x7000
#define RG_CNTBYTE_CMD  0x8000
#define RG_CNTWORD_CMD  0x9000
#define RG_CMD_MASK     0xF000

#define RG_NEWBYTE(offset)      (RG_NEWBYTE_CMD | offset)
#define RG_NEWWORD(offset)      (RG_NEWWORD_CMD | offset)
#define RG_CNTBYTE(offset)      (RG_CNTBYTE_CMD | offset)
#define RG_CNTWORD(offset)      (RG_CNTWORD_CMD | offset)
#define RG_DPORT(port)          (RG_DPORT_CMD), port
#define RG_INDEX(index)         (RG_INDEX_CMD | index)
#define RG_IPORT(port,index)    (RG_IPORT_CMD | index), port
#define RG_MASK(mask)           (RG_MASK_CMD | mask)
#define RG_STARTBIT(startbit)   (RG_STARTBIT_CMD | startbit)
#define RG_FINISH               RG_FINISH_CMD

#define RG_TOTAL        RG_NEWWORD(0)
#define RG_DISPLAY      RG_NEWWORD(2)
#define RG_BLANK_START  RG_NEWWORD(4)
#define RG_BLANK_END    RG_NEWWORD(6)
#define RG_SYNC_START   RG_NEWWORD(8)
#define RG_SYNC_END     RG_NEWWORD(10)

//---------------------------------------------------------------------------
// Register profile for horizontal Standard VGA CRTC registers.
//---------------------------------------------------------------------------

#define VGA_HCRTC_REGS_DESC     \
    RG_TOTAL,                   \
    RG_IPORT    (0x3D4, 0x00),  \
    RG_MASK     (B11111111),    \
                                \
    RG_DISPLAY,                 \
    RG_INDEX    (0x01),         \
    RG_MASK     (B11111111),    \
                                \
    RG_BLANK_START,             \
    RG_INDEX    (0x02),         \
    RG_MASK     (B11111111),    \
                                \
    RG_BLANK_END,               \
    RG_INDEX    (0x03),         \
    RG_MASK     (B00011111),    \
    RG_INDEX    (0x05),         \
    RG_MASK     (B10000000),    \
                                \
    RG_SYNC_START,              \
    RG_INDEX    (0x04),         \
    RG_MASK     (B11111111),    \
                                \
    RG_SYNC_END,                \
    RG_INDEX    (0x05),         \
    RG_MASK     (B00011111)

//---------------------------------------------------------------------------
// Register profile for vertical Standard VGA CRTC registers.
//---------------------------------------------------------------------------

#define VGA_VCRTC_REGS_BASIC    \
    RG_TOTAL,                   \
    RG_IPORT    (0x3D4, 0x06),  \
    RG_MASK     (B11111111),    \
    RG_INDEX    (0x07),         \
    RG_MASK     (B00000001),    \
    RG_MASK     (B00100000),    \
                                \
    RG_DISPLAY,                 \
    RG_INDEX    (0x12),         \
    RG_MASK     (B11111111),    \
    RG_INDEX    (0x07),         \
    RG_MASK     (B00000010),    \
    RG_MASK     (B01000000),    \
                                \
    RG_BLANK_START,             \
    RG_INDEX    (0x15),         \
    RG_MASK     (B11111111),    \
    RG_INDEX    (0x07),         \
    RG_MASK     (B00001000),    \
    RG_INDEX    (0x09),         \
    RG_MASK     (B00100000),    \
                                \
    RG_BLANK_END,               \
    RG_INDEX    (0x16),         \
    RG_MASK     (B11111111),    \
                                \
    RG_SYNC_START,              \
    RG_INDEX    (0x10),         \
    RG_MASK     (B11111111),    \
    RG_INDEX    (0x07),         \
    RG_MASK     (B00000100),    \
    RG_MASK     (B10000000),    \
                                \
    RG_SYNC_END,                \
    RG_INDEX    (0x11),         \
    RG_MASK     (B00001111)

#define VGA_VCRTC_REGS_DESC     \
    VGA_VCRTC_REGS_BASIC,       \
                                \
    RG_SYNC_START,              \
    RG_STARTBIT (13),           \
    RG_INDEX    (0x09),         \
    RG_MASK     (B00000001),    \
    RG_STARTBIT (12),           \
    RG_INDEX    (0x09),         \
    RG_MASK     (B00000010)

//---------------------------------------------------------------------------
// Register profile for alternate vertical Standard VGA CRTC registers. This
// is basically the same as the above version but it implements the double
// scan bit using an alternate method to the code above (triple scan is
// not supported with this method).
//---------------------------------------------------------------------------

#define VGA_VCRTC_REGS_DESC_ALT \
    VGA_VCRTC_REGS_BASIC,       \
                                \
    RG_CNTWORD  (8),            \
    RG_STARTBIT (13),           \
    RG_INDEX    (0x09),         \
    RG_MASK     (B10000000)

// Macro to wait until not in the vertical retrace period (ie: display
// enable is asserted). We skip this step if the waitVRT flag is
// set to -1 since (which assumes we are already just before the
// vertical retrace period).

#define WaitDEVGA()                 \
    if (waitVRT != -1) {            \
        while (inpb(0x3DA) & 8)     \
            ;                       \
        }

// Macro to clear the VGA vertical retrace latch. This should enable
// triple buffering support for all VGA compatible cards provided the
// vertical retrace is enabled.

#define ClearVERTLatch()                    \
{                                           \
    int v;                                  \
    if (!DV.sharedIRQ) {                    \
        outpb(0x3D4,0x11);                  \
        v = inpb(0x3D5);                    \
        outpb(0x3D5,(v & 0xEF) | 0x20);     \
        outpb(0x3D5,(v & 0xDF) | 0x10);     \
        }                                   \
}

// Macro to wait until the start of the vertical retrace period

#define WaitVERTVGA()                       \
    {                                       \
        while (!(inpb(0x3DA) & 8))          \
            ;                               \
    }

// Macro to set the low VGA compatible start address bits. This version
// is for VGA text display modes and takes a byte offset in display memory
// but clears the PEL panning register.

#define SetLowStartAddress0(offset)         \
{                                           \
    inpb(0x3DA);                            \
    outpb(0x3C0,0x13 | 0x20);               \
    outpb(0x3C0,0x08);                      \
    wrinx(0x3D4,0x0D,(uchar)offset);        \
    wrinx(0x3D4,0x0C,(uchar)(offset >> 8)); \
}

// Macro to set the low VGA compatible start address bits, including the
// PEL panning register. This version is for 16-color display modes and
// text modes and takes a pixel offset in display memory.

#define SetLowStartAddress4(offset)         \
{                                           \
    inpb(0x3DA);                            \
    outpb(0x3C0,0x13 | 0x20);               \
    outpb(0x3C0,offset & 0x7);              \
    offset >>= 3;                           \
    wrinx(0x3D4,0x0D,(uchar)offset);        \
    wrinx(0x3D4,0x0C,(uchar)(offset >> 8)); \
}

// This is the same as the above version, but does not program the PEL
// panning register.

#define SetLowStartAddressNoPel(offset)     \
{                                           \
    offset >>= 3;                           \
    wrinx(0x3D4,0x0D,(uchar)offset);        \
    wrinx(0x3D4,0x0C,(uchar)(offset >> 8)); \
}

// Macro to set the low VGA compatible start address bits, including the
// PEL panning register. This version is for 256-color and above
// display modes and and takes a byte offset in display memory.

#define SetLowStartAddress8(offset)         \
{                                           \
    inpb(0x3DA);                            \
    outpb(0x3C0,0x13 | 0x20);               \
    outpb(0x3C0,(offset & 0x3) << 1);       \
    offset >>= 2;                           \
    wrinx(0x3D4,0x0D,(uchar)offset);        \
    wrinx(0x3D4,0x0C,(uchar)(offset >> 8)); \
}

// This is the same as the above version, but does not program the PEL
// panning register.

#define SetLowStartAddress8NoPel(offset)    \
{                                           \
    offset >>= 2;                           \
    wrinx(0x3D4,0x0D,(uchar)offset);        \
    wrinx(0x3D4,0x0C,(uchar)(offset >> 8)); \
}

/*---------------------------- Global Variables ---------------------------*/

extern ulong    VGA_regProfile[];           // VGA register profile
extern ushort   VGA_HCRTCRegsDesc[];        // VGA HCRTC register profile
extern ushort   VGA_VCRTCRegsDesc[];        // VGA VCRTC register profile
extern ushort   VGA_VCRTCRegsDescAlt[];     // Alternate VCRTC register profile

#endif /* __SNAP_GRAPHICS_VGA_H */

