/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Generic device detection module.
*
****************************************************************************/

#include "snap/agp/driver.h"

/*--------------------------- Global variables ----------------------------*/

char                AGP_path[256];
AGP_driver          AGP;                // Global device driver structure
N_imports _VARAPI   _N_imports;         // Imports to OS specific functions
PM_imports _VARAPI  _PM_imports;        // Imports to PM library functions
DRV_imports _VARAPI _AGP_drvImports;    // Imports to AGP specific functions

AGP_configInfo      DRV_agpConfigInfo = {
    sizeof(AGP_configInfo),
    CONFIG_SIGNATURE,
    CONFIG_SIGNATURE,
    DRIVER_VERSION,
    DRIVER_DATE,
    false,
    "",
    0,
    };

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
PARAMETERS:
mfrName         - Place to store the manufacturer name
chipsetNames    - Place to store the list of chipset names

REMARKS:
Returns the chipset driver manufacturer name and list of supported chipsets.
****************************************************************************/
void _CEXPORT AGP_getChipsetNames(
    char **mfr,
    char ***chipsetNames)
{
    *mfr = AGP_manufacturerName;
    *chipsetNames = AGP_chipsetNames;
}

/****************************************************************************
PARAMETERS:
info    - Place to return the certification information

REMARKS:
Returns the chipset driver certification information structures.
****************************************************************************/
void _CEXPORT AGP_getCertifyInfo(
    AGP_certifyInfo *info)
{
    COPY_STRUCTURE(info,&DRV_agpCertifyInfo);
}

/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler for non-imported N_imports.
****************************************************************************/
static void _N_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported N_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler for non-imported AGP_imports.
****************************************************************************/
static void _DRV_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported DRV_imports import function called! Please re-compile!\n");
}

/****************************************************************************
PARAMETERS:
path        - Path to Binary Portable DLL

RETURNS:
Pointer to exported function list

REMARKS:
Does generic Plug and Play device detection and fills in the device resource
allocation information for the device so that the driver loader can
allocate the necessary OS resources for the device before calling the
real detection code.
****************************************************************************/
AGP_devCtx * _CEXPORT AGP_initDriver(
    const char *path,
    N_int32 deviceIndex,
    PCIDeviceInfo *AGPDevice,
    PCIDeviceInfo *AGPBridge,
    PCIDeviceInfo *hostBridge,
    N_imports *nImp,
    PM_imports *pmImp,
    DRV_imports *drvImp,
    AGP_globalOptions *globalOpt)
{
    int     i,max;
    ulong   *p;

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_N_imports)/sizeof(AGP_initLibrary_t);
    for (i = 0,p = (ulong*)&_N_imports; i < max; i++)
        *p++ = (ulong)_N_fatalErrorHandler;
    max = sizeof(_PM_imports)/sizeof(AGP_initLibrary_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;
    max = sizeof(_AGP_drvImports)/sizeof(AGP_initLibrary_t);
    for (i = 0,p = (ulong*)&_AGP_drvImports; i < max; i++)
        *p++ = (ulong)_DRV_fatalErrorHandler;

    // Now copy all our imported functions
    memcpy(&_N_imports,nImp,MIN(sizeof(_N_imports),nImp->dwSize));
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    memcpy(&_AGP_drvImports,drvImp,MIN(sizeof(_AGP_drvImports),drvImp->dwSize));

    // Initialise default information
    memset(&AGP,0,sizeof(AGP));
    strcpy(AGP_path,path);
    AGP.deviceIndex = deviceIndex;
    if (AGPDevice) memcpy(&AGP.AGPDevice,AGPDevice,sizeof(AGP.AGPDevice));
    if (AGPBridge) memcpy(&AGP.AGPBridge,AGPBridge,sizeof(AGP.AGPBridge));
    if (hostBridge) memcpy(&AGP.hostBridge,hostBridge,sizeof(AGP.hostBridge));
    strcpy(DC.Signature,AGP_SIGNATURE);
    DC.Version = AGP_VERSION;
    DC.DriverRev = (AGP_RELEASE_MAJOR << 8) | AGP_RELEASE_MINOR;
    strcpy(DC.OemVendorName,SCITECH_STR);
    strcpy(DC.OemCopyright,COPYRIGHT_STR);
    AGP.globalOpt = *globalOpt;

    // Now detect the install chipset configuration
    if (!DRV_agpChipDetect())
        return NULL;

    // Fill in the driver init function pointers
    DC.InitDriver = DRV_agpInitDriver;
    DC.QueryFunctions = DRV_agpQueryFunctions;
    DC.UnloadDriver = DRV_agpUnloadDriver;
    return &DC;
}

/****************************************************************************
DESCRIPTION:
Detects that the device is installed and configures the driver.

RETURNS:
Status value for the completion of driver initialisation process

REMARKS:
Does generic device detection and fills in the generic components of the
SNAP device context header information after calling the chipset
specific detection code.
****************************************************************************/
N_int32 NAPI DRV_agpInitDriver(void)
{
    int                 status,i;
    char                *chipName;
    AGP_certifyChipInfo *certified;
    PCIAGPCapability    masterCap;
    PCIAGPCapability    targetCap;
    ulong               dataRate;
    ulong               fastWrite;
    ulong               SBAEnable;
    int                 maxAGPRate = AGP.globalOpt.maxAGPRate;
    ibool               enableFastWrite = AGP.globalOpt.bEnableFastWrite;

    // Check if chipset is certified, fail if not!
    certified = NULL;
    if (DRV_agpCertifyInfo.NumCertifiedChips) {
        certified = DRV_agpCertifyInfo.CertifiedCards;
        chipName = AGP_chipsetNames[AGP.chipID];
        for (i = 0; i < DRV_agpCertifyInfo.NumCertifiedChips; i++,certified++) {
            if (strcmp(chipName,certified->ChipsetName) == 0) {
                if (certified->CertifyVersion >= AGP.globalOpt.wCertifiedVersion) {
                    // This chipset is certified, so allow driver to load
                    break;
                    }
                }
            }

        // Check if the chipset is certified, and bail if not and the
        // options do not allow certified drivers to run.
        if (i == DRV_agpCertifyInfo.NumCertifiedChips) {
            if (!AGP.globalOpt.bAllowNonCertified)
                return nNotCertified;
            certified = NULL;
            }
        }

    // Set all API functions to NOT_IMPLEMENTED by default
    INIT_STRUCTURE(AGP.driver);

    // Configure all the generic API functions for the driver.
    AGP.driver.GetConfigInfo        = DRV_agpGetConfigInfo;
    AGP.driver.GetCertifyInfo       = DRV_agpGetCertifyInfo;
    AGP.driver.GetApertureSize      = DRV_agpGetApertureSize;
    AGP.driver.GetApertureBase       = DRV_agpGetApertureBase;
    AGP.driver.ReservePhysical      = DRV_agpReservePhysical;
    AGP.driver.ReleasePhysical       = DRV_agpReleasePhysical;
    AGP.driver.CommitPhysical        = DRV_agpCommitPhysical;
    AGP.driver.FreePhysical          = DRV_agpFreePhysical;

    // Now configure the driver based on all the detected information and
    // any options information loaded from disk
    if ((status = DRV_agpConfigure()) != nOK)
        return status;

    // Initialise all generic memory block lists
    if (!N_createBlockList(&AGP.physList.bl,sizeof(AGP_physCtx)))
        return nOutOfMemory;
    AGP.physList.freeList = N_buildFreeList(&AGP.physList.bl);

    // Get the master and target AGP capabilities. Note that some AGP drivers
    // (like the Intel i810/815) are not real AGP drivers, but instead special
    // ones that control the internal graphics device directly. Hence all the
    // AGP capabilities features are not present on these chipsets and we have
    // to skip this setup code.
    if (!AGP.fakeAGPDriver) {
        if (!DRV_getAGPCaps(&AGP.AGPDevice,&masterCap))
            return nNotDetected;
        if (!DRV_getAGPCaps(&AGP.hostBridge,&targetCap))
            return nNotDetected;

        // Some broken cards (Matrox Millenium II "AGP") report no valid
        // supported transfer rates. These are not really AGP cards. They
        // have an AGP Capabilities structure that reports no capabilities.
        if (masterCap.AGPStatus.rate == 0)
            return nNotDetected;

        // Determine the greatest common denominator for data rate
        dataRate = targetCap.AGPStatus.rate & masterCap.AGPStatus.rate;
        if (targetCap.AGPStatus.agp3mode) {
            // AGP 3.0 and later devices, where maximum data rate is 8x
            if (maxAGPRate < 4)
                maxAGPRate = 4;
            if (dataRate & 2)
                dataRate = MIN(8,maxAGPRate) >> 2;
            else if (dataRate & 1)
                dataRate = MIN(4,maxAGPRate) >> 2;
            }
        else {
            // AGP 2.0 and ealier devices, where maximum data rate is 4x
            if (maxAGPRate > 4)
                maxAGPRate = 4;
            if (dataRate & 4)
                dataRate = MIN(4,maxAGPRate);
            else if (dataRate & 2)
                dataRate = MIN(2,maxAGPRate);
            else if (dataRate & 1)
                dataRate = MIN(1,maxAGPRate);
            }

        // Enable side band addressing if both master and target support it
        SBAEnable = (targetCap.AGPStatus.sideBandAddressing & masterCap.AGPStatus.sideBandAddressing);

        // Enable FastWrite if both master and target support it
        fastWrite = enableFastWrite && ((targetCap.AGPStatus.fastWrite & masterCap.AGPStatus.fastWrite));

        // Enable the Target device first
        targetCap.AGPCommand.rate = dataRate;
        targetCap.AGPCommand.AGPEnable = 1;
        targetCap.AGPCommand.SBAEnable = SBAEnable;
        targetCap.AGPCommand.fastWriteEnable = fastWrite;
        targetCap.AGPCommand.fourGBEnable = 0;
        if (!DRV_setAGPCaps(&AGP.AGPDevice,&targetCap))
            return nNotDetected;

        // Now enable the Master device
        masterCap.AGPCommand.rate = dataRate;
        masterCap.AGPCommand.AGPEnable = 1;
        masterCap.AGPCommand.SBAEnable = SBAEnable;
        masterCap.AGPCommand.fastWriteEnable = fastWrite;
        masterCap.AGPCommand.fourGBEnable = 0;
        masterCap.AGPCommand.requestQueueDepth = targetCap.AGPStatus.requestQueueDepthMaximum;
        if (!DRV_setAGPCaps(&AGP.hostBridge,&masterCap))
            return nNotDetected;
        }

    // TODO: Somewhere in here we have to determine the 'favoured' memory
    //       ranges for the AGP bus. This can be used to allocate physical
    //       memory that responds faster to AGP accesses than other areas
    //       of memory. See the Windows NT DDK.

    // Construct the detected configuration info block
    DRV_agpConfigInfo.dwSize = sizeof(DRV_agpConfigInfo);
    strcpy(DRV_agpConfigInfo.ManufacturerName,AGP_manufacturerName);
    strcpy(DRV_agpConfigInfo.ChipsetName,AGP_chipsetNames[AGP.chipID]);
    if (certified) {
        DRV_agpConfigInfo.Certified = true;
        strcpy(DRV_agpConfigInfo.CertifiedDate,certified->CertifiedDate);
        DRV_agpConfigInfo.CertifyVersion = certified->CertifyVersion;
        }

    // Success!
    return nOK;
}

/****************************************************************************
REMARKS:
Frees all memory resources and does any necessary memory unloading
****************************************************************************/
void NAPI DRV_agpUnloadDriver(void)
{
    // TODO: We need to automatically clean up *all* memory mappings owned by
    //       the AGP driver when it shuts down!

    // Shut down the chipset specific driver
    DRV_agpShutdown();

    // Free all generic block lists
    N_freeBlockList(&AGP.physList.bl);
}

