/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the generic SNAP AGP driver
*               API functions based on chipset specific support modules.
*
****************************************************************************/

#include "snap/agp/driver.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
RETURNS:
Current AGP aperture size

REMARKS:
Returns the current AGP aperture size
****************************************************************************/
N_int32 NAPI DRV_agpGetApertureSize(void)
{
    return AGP.apertureSize;
}

/****************************************************************************
RETURNS:
Current physical base address for AGP aperture

REMARKS:
Returns the current physical base address for AGP aperture
****************************************************************************/
N_physAddr  NAPI DRV_agpGetApertureBase(void)
{
    return AGP.apertureBase;
}

/****************************************************************************
PARAMETERS:
numPages    - Number of memory pages that should be reserved
type        - Type of memory to allocate
physContext - Returns the physical context handle for the mapping
physAddr    - Returns the physical address for the mapping

RETURNS:
Error code defined by N_errorType, nOK on success.

REMARKS:
This function reserves a range of physical memory addresses on the system
bus which the AGP controller will respond to. If this function succeeds,
the AGP controller can respond to the reserved physical address range on
the bus. However you must first call CommitPhysical to cause this memory
to actually be committed for use before it can be accessed.
****************************************************************************/
N_int32 NAPI DRV_agpReservePhysical(
    N_uint32 numPages,
    N_int32 type,
    void **physContext,
    N_physAddr *physAddr)
{
    N_int32     status,linSize,physSize,reserved = false;
    AGP_physCtx *ctx;

    // Allocate a new context to hold the reserved memory range
    if ((ctx = DRV_newAGPPhysContext()) == NULL) {
        status = nOutOfMemory;
        goto Error;
        }

    // Create arrays large enough to track all committed memory
    linSize = numPages * sizeof(*ctx->memory);
    physSize = numPages * sizeof(*ctx->physMemory);
    ctx->memory = malloc(linSize);
    ctx->physMemory = malloc(physSize);
    if (ctx->memory == NULL || ctx->physMemory == NULL) {
        status = nOutOfMemory;
        goto Error;
        }
    memset(ctx->memory,0,linSize);
    memset(ctx->physMemory,0,physSize);

    // Fill in the memory ctx details, and try to reserve it in the GART
    ctx->numPages = numPages;
    ctx->committedPages = 0;
    ctx->type = type;
    LOCK_MASTER();
    status = AGP.chip.ReserveMemory(ctx);
    UNLOCK_MASTER();
    if (status != nOK)
        goto Error;

    // Attempt to write combine the requested region if write combined
    reserved = true;
    if (ctx->type == agpWriteCombine) {
        if (PM_enableWriteCombine(ctx->memoryBase,(ctx->numPages * PM_PAGE_SIZE),PM_MTRR_WRCOMB) != PM_MTRR_ERR_OK) {
            status = nMemMapError;
            goto Error;
            }
        }

    // Return the mapped memory ctx information to the caller
    *physContext = ctx;
    *physAddr = ctx->memoryBase;
    return nOK;

    // Common error handling clean up code
Error:
    if (ctx) {
        if (reserved) {
            LOCK_MASTER();
            AGP.chip.ReleaseMemory(ctx);
            UNLOCK_MASTER();
            }
        if (ctx->memory) free(ctx->memory);
        if (ctx->physMemory) free(ctx->physMemory);
        DRV_freeAGPPhysContext(ctx);
        }
    return status;
}

/****************************************************************************
PARAMETERS:
physContext - Physical AGP context to release

RETURNS:
Error code defined by N_errorType, nOK on success.

REMARKS:
This function releases a range of physical memory addresses on the system
bus which the AGP controller will respond to. All committed memory for
the physical address range covered by the context will be released.
****************************************************************************/
N_int32 NAPI DRV_agpReleasePhysical(
    void *physContext)
{
    N_int32 status;
    AGP_physCtx *ctx = (AGP_physCtx*)physContext;

    // Make sure the range is empty
    if (ctx->committedPages != 0)
        return nInvalidParameter;

    // Release all the memory for the range
    LOCK_MASTER();
    status = AGP.chip.ReleaseMemory(ctx);
    UNLOCK_MASTER();

    // Finally free all memory blocks and return status
    free(ctx->memory);
    free(ctx->physMemory);
    DRV_freeAGPPhysContext(ctx);
    return status;
}

/****************************************************************************
PARAMETERS:
physContext - Physical AGP context to commit memory for
numPages    - Number of pages to be committed
startOffset - Offset in pages into the reserved physical context
physAddr    - Returns the physical address of the committed memory

RETURNS:
Error code defined by N_errorType, nOK on success.

REMARKS:
This function commits into the specified physical context that was previously
reserved by a call to ReservePhysical. You can use the startOffset and
numPages parameters to only commit portions of the reserved memory range at
a time.
****************************************************************************/
N_int32 NAPI DRV_agpCommitPhysical(
    void *physContext,
    N_uint32 numPages,
    N_uint32 startOffset,
    N_physAddr *physAddr)
{
    AGP_physCtx *ctx = (AGP_physCtx*)physContext;
    int         i;
    N_int32     status;
    N_int32     startPage = ctx->startPage + startOffset;
    void        **mem;
    N_physAddr  *phys;

    // Allocate enough pages for the committed memory range
    mem = &ctx->memory[startPage];
    phys = &ctx->physMemory[startPage];
    for (i = 0; i < numPages; i++,mem++,phys++) {
        if ((*mem = PM_allocPage(true)) == NULL)
            goto Error;
        if ((*phys = PM_getPhysicalAddr(*mem)) == PM_BAD_PHYS_ADDRESS)
            goto Error;
        }

    // Now map the memory into the GART
    LOCK_MASTER();
    status = AGP.chip.MapMemory(ctx,numPages,startOffset);
    UNLOCK_MASTER();
    if (status != nOK)
        goto Error;

    // Indicate that we have committed the memory and return status
    ctx->committedPages += numPages;
    *physAddr = ctx->memoryBase + (startPage * PM_PAGE_SIZE);
    return nOK;

Error:
    // Free the allocated pages
    mem = &ctx->memory[startPage];
    phys = &ctx->physMemory[startPage];
    for (i = 0; i < ctx->numPages; i++,mem++,phys++) {
        if (*mem) PM_freePage(*mem);
        *mem = NULL;
        *phys = 0;
        }
    return status;
}

/****************************************************************************
PARAMETERS:
physContext - Physical AGP context to free memory for
numPages    - Number of pages to be freed
startOffset - Offset in pages into the reserved physical context

RETURNS:
Error code defined by N_errorType, nOK on success.

REMARKS:
This function frees memory previously committed by the CommitPhysical
function. Note that you can free a portion of a memory range that was
previously committed if you wish.
****************************************************************************/
N_int32 NAPI DRV_agpFreePhysical(
    void *physContext,
    N_uint32 numPages,
    N_uint32 startOffset)
{
    AGP_physCtx *ctx = (AGP_physCtx*)physContext;
    int         i;
    N_int32     status;
    N_int32     startPage = ctx->startPage + startOffset;
    void        **mem;
    N_physAddr  *phys;

    // Unmap the memory from the GART
    LOCK_MASTER();
    status = AGP.chip.UnMapMemory(ctx,numPages,startOffset);
    UNLOCK_MASTER();
    if (status != nOK)
        return status;

    // Free the committed memory pages
    mem = &ctx->memory[startPage];
    phys = &ctx->physMemory[startPage];
    for (i = 0; i < ctx->numPages; i++,mem++,phys++) {
        if (*mem) PM_freePage(*mem);
        *mem = NULL;
        *phys = 0;
        }

    // Indicate that we have unmapped the memory and return status
    ctx->committedPages -= numPages;
    return nOK;
}
