/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the geneic helper functions for AGP
*               drivers.
*
****************************************************************************/

#include "snap/agp/driver.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
PARAMETERS:
info    - PCI device information to search for AGP capabilities block

RETURNS:
Offset in config space to start of AGP capabilities block, 0 if none found.
****************************************************************************/
uchar DRV_findAGPCaps(
    PCIDeviceInfo *info,
    PCIAGPCapability *caps)
{
    uchar capPtr;

    // Check the Status register to see if this device supports capability
    // lists. If not, it is not an AGP-compliant device.
    if ((info->Status & PCI_statusCapabilitiesList) == 0)
        return 0;

    // The device supports capability lists, find the AGP capabilities
    if ((info->HeaderType & (~PCI_multiFunctionType)) == PCI_bridgeType)
        capPtr = info->u.type1.CapabilitiesPointer;
    else
        capPtr = info->u.type0.CapabilitiesPointer;

    // Search for the AGP capabilities list
    while (capPtr != 0) {
        // Read the start of the capabilities block at this offset
        PCI_readRegBlock(info,capPtr,&caps->h,sizeof(caps->h));
        if (caps->h.capsID == PCI_capsAGP) {
            // Found the AGP capabilities block
            return capPtr;
            }
        else {
            // Some other capability, so keep looking
            capPtr = caps->h.next;
            }
        }

    // If we get here, then no AGP capabilities list was found!
    return 0;
}

/****************************************************************************
PARAMETERS:
info    - PCI device information to get for AGP capabilities block for
caps    - Place to store the returned master device capabilities structure

RETURNS:
True on success, false on failure
****************************************************************************/
ibool DRV_getAGPCaps(
    PCIDeviceInfo *info,
    PCIAGPCapability *caps)
{
    uchar capPtr;

    // Find the header block for the AGP capabilities list
    if ((capPtr = DRV_findAGPCaps(info,caps)) == 0)
        return false;

    // Now read the rest of the AGP capabilities list
    PCI_readRegBlock(info,capPtr+sizeof(caps->h),&caps->h+1,sizeof(*caps)-sizeof(caps->h));
    return true;
}

/****************************************************************************
PARAMETERS:
info    - PCI device information to get for AGP capabilities block for
caps    - Place to store the returned master device capabilities structure

RETURNS:
True on success, false on failure
****************************************************************************/
ibool DRV_setAGPCaps(
    PCIDeviceInfo *info,
    PCIAGPCapability *caps)
{
    uchar capPtr;

    // Find the header block for the AGP capabilities list
    if ((capPtr = DRV_findAGPCaps(info,caps)) == 0)
        return false;

    // Now write the AGP command register
    PCI_writeRegBlock(info,
        capPtr+PM_FIELD_OFFSET(PCIAGPCapability,AGPCommand),
        &caps->AGPCommand,sizeof(caps->AGPCommand));
    return true;
}

/****************************************************************************
RETURNS:
Pointer to new AGP_physCtx structure

REMARKS:
Allocates space for a new AGP_physCtx structure from the free list. If
there are no more buffers, we attempt to re-size the buffer pool,
returning NULL if we run out of memory.
****************************************************************************/
AGP_physCtx * DRV_newAGPPhysContext(void)
{
    AGP_physCtx *r;

    if (AGP.physList.freeList == NULL) {
        if (!N_resizeBlockList(&AGP.physList.bl))
            return NULL;
        AGP.physList.freeList = N_buildFreeList(&AGP.physList.bl);
        }
    r = AGP.physList.freeList;
    AGP.physList.freeList = N_FREELIST_NEXT(r);
    return r;
}

/****************************************************************************
PARAMETERS:
buffer  - Pointer to the buffer to free

REMARKS:
Frees the buffer by putting it onto the buffer free list.
****************************************************************************/
void DRV_freeAGPPhysContext(
    AGP_physCtx *r)
{
    N_FREELIST_NEXT(r) = AGP.physList.freeList;
    AGP.physList.freeList = r;
}

