/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module for returning function groups from the driver.
*
****************************************************************************/

#include "snap/agp/driver.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
PARAMETERS:
funcs       - Pointer to the function list to make valid
size        - Size of the function list in bytes

RETURNS:
True if the functions are valid, false if not.

REMARKS:
This function validates the function list and returns a NULL pointer if
the function list is empty and does not contain any functions. Internally
in our drivers we list functions that have not yet been implemented on
a particular hardware device with a value of 0xFFFFFFFF for the function
pointer. Unless the safetyLevel is set to -1, we convert those pointers
to NULL pointers so the user code will not try to call them.

However our test programs can request the blocks of function pointers
with a safetyLevel of -1 and get a list of all functions that are
implemented, not supported and simply not implemented yet. This will help
our engineers know which functions need to be implemented or at least
mapped our for specific hardware devices.
****************************************************************************/
static ibool MakeValid(
    ulong *funcs,
    int size)
{
    int     i,empty = true;

    for (i = 0; i < size/4; i++) {
        if (funcs[i] == 0xFFFFFFFF || funcs[i] == 0xFFFFFFFE || funcs[i] == 0xFFFFFFFD)
            funcs[i] = 0;
        if (funcs[i] != 0)
            empty = false;
        }
    return !empty;
}

#define RETURN_VALID(p) \
{ \
    COPY_STRUCTURE2(funcs,&(p),dwSize); \
    return MakeValid(((ulong*)funcs)+1,dwSize-sizeof(ulong)); \
}

/****************************************************************************
PARAMETERS:
id          - Identifier for the function group to get pointers for
funcs       - Pointer to the function block to fill in

RETURNS:
True if the functions are supported, false if not.

REMARKS:
This function is the main function that is used by the graphics application
code to get a block of function pointers for a specified function group.
The function groups are defined AGP_funcGroupsType enumeration, and breaks
up the functions in the device driver API into groups of logically similar
functions.
****************************************************************************/
ibool NAPI DRV_agpQueryFunctions(
    N_uint32 id,
    void *funcs)
{
    ulong   dwSize = *((ulong*)funcs);

    memset(funcs,0,dwSize);
    switch (id) {
        case AGP_GET_DRIVERFUNCS:
            RETURN_VALID(AGP.driver);
        }
    return false;
}

/****************************************************************************
PARAMETERS:
info    - Place to copy the configuration information

REMARKS:
Returns the configuration information for the chipset driver.
****************************************************************************/
void NAPI DRV_agpGetConfigInfo(
    AGP_configInfo *info)
{
    COPY_STRUCTURE(info,&DRV_agpConfigInfo);
}

/****************************************************************************
PARAMETERS:
info    - Place to copy the configuration information

REMARKS:
Returns the configuration information for the chipset driver.
****************************************************************************/
void NAPI DRV_agpGetCertifyInfo(
    AGP_certifyInfo *info)
{
    COPY_STRUCTURE(info,&DRV_agpCertifyInfo);
}

