/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Block memory management routines for SNAP drivers. These
*               functions are used to manage blocks of objects of fixed
*               sizes for efficient memory management using a free list.
*
****************************************************************************/

#include "snap/common/mem.h"

/*------------------------- Implementation --------------------------------*/

/****************************************************************************
PARAMETERS:
l           - Block list to allocate
objectSize  - Size of each object in bytes

RETURNS:
True on success, false on failure to allocate block list.

REMARKS:
Creates the initial block list by allocating and building the block
list free list pool. Block lists are arbitrary blocks of objects,
form the building block for memory management for all the region functions.
Whenever the block list needs to be resized, we always resize it in
increments of 'N_BLOCK_SIZE'.
****************************************************************************/
ibool NAPI N_createBlockList(
    N_blockList *l,
    int objectSize)
{
    int max = (N_BLOCK_SIZE - sizeof(N_block)) / objectSize;

    if ((l->curBlock = malloc(N_BLOCK_SIZE)) == NULL)
        return false;
    l->curBlock->next   = NULL;
    l->max              = max;
    l->objectSize       = objectSize;
    l->blockCount       = 1;
    return true;
}

/****************************************************************************
PARAMETERS:
l   - Block list to resize

RETURNS:
True on success, false on failure to allocate block list.

REMARKS:
Resizes the block list if full by allocating another pool block and linking
it onto the list.
****************************************************************************/
ibool NAPI N_resizeBlockList(
    N_blockList *l)
{
    N_block     *temp;

    if ((temp = malloc(N_BLOCK_SIZE)) == NULL)
        return false;
    temp->next = l->curBlock;
    l->curBlock = temp;
    l->blockCount++;
    return true;
}

/****************************************************************************
PARAMETERS:
l   - Block list to free

REMARKS:
Frees all of the pool blocks in the list
****************************************************************************/
void NAPI N_freeBlockList(
    N_blockList *l)
{
    while (l->blockCount) {
        N_block *temp = l->curBlock;
        l->curBlock = temp->next;
        free(temp);
        l->blockCount--;
        }
}

/****************************************************************************
PARAMETERS:
l   - Block list to build free list for

REMARKS:
Builds the free list of object in the first block in the list, assuming
the size of the blocks specified in the blockList.
****************************************************************************/
void * NAPI N_buildFreeList(
    N_blockList *l)
{
    int     i,max = l->max,size = l->objectSize;
    void    *freeList;
    char    *p,*next;

    p = freeList = (l->curBlock+1);
    for (i = 0; i < max; i++,p = next)
        N_FREELIST_NEXT(p) = next = p+size;
    N_FREELIST_NEXT(p-size) = NULL;
    return freeList;
}

