/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Linked list management routines for SNAP drivers.
*
* ***************************************************************************/

#include "snap/common/list.h"

/*------------------------- Implementation --------------------------------*/

/****************************************************************************
PARAMETERS:
l   - Pointer to list to initialise

REMARKS:
Initialises a linked list for use.
****************************************************************************/
void NAPI N_initList(
    N_list *l)
{
    l->count = 0;
    l->head = &(l->hz[0]);
    l->z = &(l->hz[1]);
    l->head->next = l->z->next = l->z;
    l->z->prev = l->head->prev = l->head;
}

/****************************************************************************
PARAMETERS:
l           - List to kill
freeNode    - Pointer to user routine to free a node

REMARKS:
Kills the list l, by deleting all of the elements contained within the list
one by one and then deleting the list itself. Note that we call the user
supplied routine freeNode)() to free each list node. This allows the user
program to perform any extra processing needed to kill each node (if each
node contains pointers to other items on the heap for example).

Do *NOT* pass free() as the free node function! The node pointer passed to
the free function will be the *user space* pointer and must be translated
to the list header pointer before calling the actual free function. Ie:

    void MyFreeNode(
        void *node)
    {
        if (node)
            free(LST_HEADER(node));
    }

****************************************************************************/
void NAPI N_destroyList(
    N_list *l,
    void (*freeNode)(
        void *node))
{
    N_node  *n,*p;

    n = l->head->next;
    while (n != l->z) {         /* Free all nodes in list               */
        p = n;
        n = n->next;
        (*freeNode)(LST_USERSPACE(p));
        }
}

/****************************************************************************
PARAMETERS:
l       - List to insert node into
node    - Pointer to user space of node to insert
after   - Pointer to user space of node to insert node after

REMARKS:
Inserts a new node into the list after the node 'after'. To insert a new
node at the beginning of the list, user the macro LST_HEAD in place of
'after'. ie:

    N_addAfter(mylist,node,LST_HEAD(mylist));

****************************************************************************/
void NAPI N_addAfter(
    N_list *l,
    void *node,
    void *after)
{
    N_node  *n = LST_HEADER(node),*a = LST_HEADER(after);

    n->next = a->next;
    a->next = n;
	n->prev = a;
	n->next->prev = n;
    l->count++;
}

/****************************************************************************
PARAMETERS:
l       - List to insert node into
node    - Pointer to user space of node to insert
before  - Pointer to user space of node to insert node before

REMARKS:
Inserts a new node into the list before the node 'before'.
****************************************************************************/
void NAPI N_addBefore(
    N_list *l,
    void *node,
    void *before)
{
    N_node  *n = LST_HEADER(node),*b = LST_HEADER(before);

    n->next = b;
    b->prev->next = n;
    n->prev = b->prev;
    b->prev = n;
    l->count++;
}

/****************************************************************************
PARAMETERS:
l       - List to insert node into
node    - Pointer to user space of node to insert

REMARKS:
Inserts a new node at the head of the list.
****************************************************************************/
void NAPI N_addToHead(
    N_list *l,
    void *node)
{
    N_addAfter(l,node,LST_HEAD(l));
}

/****************************************************************************
PARAMETERS:
l       - List to insert node into
node    - Pointer to user space of node to insert

REMARKS:
Inserts a new node at the end of the list.
****************************************************************************/
void NAPI N_addToTail(
    N_list *l,
    void *node)
{
    N_addAfter(l,node,LST_TAIL(l));
}

/****************************************************************************
PARAMETERS:
l       - List to remove node from.
node    - Node to remove from list

RETURNS:
Pointer to the removed node's userspace.

REMARKS:
Removes the node from the list l.
****************************************************************************/
void * NAPI N_remove(
    N_list *l,
    void *node)
{
    N_node  *n = LST_HEADER(node);

    n->next->prev = n->prev;
    n->prev->next = n->next;
    l->count--;
    return node;
}

/****************************************************************************
PARAMETERS:
l       - List to remove node from.
node    - Node to remove the next node from

RETURNS:
Pointer to the removed node's userspace.

REMARKS:
Removes the node AFTER 'node' from the list l.
****************************************************************************/
void * NAPI N_removeNext(
    N_list *l,
    void *node)
{
    N_node  *n = LST_HEADER(node);

    node = LST_USERSPACE(n->next);
	n->next->next->prev = n;
	n->next = n->next->next;
    l->count--;
    return node;
}

/****************************************************************************
PARAMETERS:
l       - List to obtain first node from

RETURNS:
Pointer to first node in list, NULL if list is empty.

REMARKS:
Returns a pointer to the user space of the first node in the list. If the
list is empty, we return NULL.
****************************************************************************/
void * NAPI N_firstNode(
    N_list *l)
{
    N_node  *n;

    n = l->head->next;
    return (n == l->z ? NULL : LST_USERSPACE(n));
}

/****************************************************************************
PARAMETERS:
l       - List to obtain last node from

RETURNS:
Pointer to last node in list, NULL if list is empty.

REMARKS:
Returns a pointer to the user space of the last node in the list. If the
list is empty, we return NULL.
****************************************************************************/
void * NAPI N_lastNode(
    N_list *l)
{
    N_node  *n;

    n = l->z->prev;
    return (n == l->head ? NULL : LST_USERSPACE(n));
}

/****************************************************************************
PARAMETERS:
prev    - Previous node in list to obtain next node from

RETURNS:
Pointer to the next node in the list, NULL at end of list.

REMARKS:
Returns a pointer to the user space of the next node in the list given a
pointer to the user space of the previous node. If we have reached the end
of the list, we return NULL. The end of the list is detected when the next
pointer of a node points back to itself, as does the dummy last node's next
pointer. This enables us to detect the end of the list without needed access
to the list data structure itself.

Note:   We do no checking to ensure that 'prev' is NOT a NULL pointer.
****************************************************************************/
void * NAPI N_nextNode(
    void *prev)
{
    N_node  *n = LST_HEADER(prev);

    n = n->next;
    return (n == n->next ? NULL : LST_USERSPACE(n));
}

/****************************************************************************
PARAMETERS:
next    - Next node in list to obtain previous node from

RETURNS:
Pointer to the previous node in the list, NULL at end of list.

REMARKS:
Returns a pointer to the user space of the previous node in the list given a
pointer to the user space of the next node. If we have reached the start
of the list, we return NULL.

Note:   We do no checking to ensure that 'next' is NOT a NULL pointer.
****************************************************************************/
void * NAPI N_prevNode(
    void *next)
{
    N_node  *n = LST_HEADER(next);

    n = n->prev;
    return (n == n->prev ? NULL : LST_USERSPACE(n));
}

