/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  SciTech SNAP internal functions common to all device
*               drivers.
*
****************************************************************************/

#include "snap/common/internal.h"
#include "clib/peloader.h"
#include "clib/os/os.h"
#ifndef __WIN32_VXD__
#include <ctype.h>
#include <stdarg.h>
#endif

/*------------------------- Global Variables ------------------------------*/

extern char     SNAPDriverPath[PM_MAX_PATH];
extern char     SNAPConfigPath[PM_MAX_PATH];
extern char     SNAPLogName[];
int             inited = false;
int             status = nOK;
int             disableLogging = false;
static char     flatname[PM_MAX_PATH];
static long     numFiles = 0;
static FLAT_DIR *flatDirectory;

/*-------------------------- Implementation -------------------------------*/

/****************************************************************************
PARAMETERS:
str - String to convert

REMARKS:
Simple function to decrypt a string in the directory table.
****************************************************************************/
static void mungeStr(
    char *str,
    int len)
{
    while (len--)
        str[len] ^= (char)(0xFF - (1 << ((len+5)&7)));
}

/****************************************************************************
DESCRIPTION:
Initialises the flat file handling library for the given Binary Portable DLL.

HEADER:
snap/internal.h

PARAMETERS:
bpdname - Name of Binary Portable DLL to initialise

RETURNS:
SNAP generic file handle for opened file, 0 on error.

REMARKS:
This function initialises the flat file handling library given a pointer
to the binary portable DLL that the flat file is bound to.
{secret}
****************************************************************************/
ibool FLAT_init(
    const char *bpdname)
{
    FILE        *f;
    int         i,size;
    FLAT_DIR    *dir;

    // Open the flat file handle
    strcpy(flatname,bpdname);
    if ((f = fopen(flatname,"rb")) == NULL)
        return false;
    if ((size = PE_getFileSize(f,0)) == -1)
        goto Error;
    fseek(f,size,SEEK_SET);
    if (fread(&numFiles,1,sizeof(numFiles),f) != sizeof(numFiles))
        goto Error;
    size = numFiles * sizeof(FLAT_DIR);
    if ((flatDirectory = malloc(size)) == NULL)
        goto Error;
    if (fread(flatDirectory,1,size,f) != size)
        goto Error;
    for (dir = flatDirectory,i = 0; i < numFiles; i++,dir++)
        mungeStr(dir->filename,sizeof(dir->filename));
    fclose(f);
    return inited = true;

Error:
    free(flatDirectory);
    flatDirectory = NULL;
    fclose(f);
    return false;
}

/****************************************************************************
RETURNS:
Pointer to the flat directory structure

REMARKS:
External function for access to the flat directory.
****************************************************************************/
FLAT_DIR * FLAT_openDirectory(
    int *count)
{
    *count = numFiles;
    return flatDirectory;
}

/****************************************************************************
REMARKS:
External function to close the flat directory.
****************************************************************************/
void FLAT_closeDirectory(void)
{
    /* do nothing */
}

/****************************************************************************
REMARKS:
Returns the length of a file on disk in bytes.
****************************************************************************/
static long GetFileSize(
    FILE *f)
{
    long size,oldpos = ftell(f);
    fseek(f,0,SEEK_END);
    size = ftell(f);
    fseek(f,oldpos,SEEK_SET);
    return size;
}

/****************************************************************************
DESCRIPTION:
Opens a file for binary read operations.

HEADER:
snap/internal.h

PARAMETERS:
filename    - Name of file to open

RETURNS:
SNAP generic file handle for opened file, 0 on error.

REMARKS:
This function open a file for reading given the filename. If an error occurs
this function simply returns 0.
{secret}
****************************************************************************/
N_FILE * FLAT_openFile(
    const char *filename)
{
    int         i;
    N_FILE      *f = NULL;
    FLAT_DIR    *dir;
    char        path[PM_MAX_PATH];

    // Try to load the file externally first
    if ((f = calloc(1,sizeof(N_FILE))) == NULL)
        return NULL;
    strcpy(path,SNAPDriverPath);
    strcat(path,filename);
    if ((f->f = fopen(path,"rb")) != NULL) {
        f->start = 0;
        f->offset = 0;
        f->length = GetFileSize(f->f);
        return f;
        }

    // Try to get it from the flat file if that fails
    dir = flatDirectory;
    for (i = 0; i < numFiles; i++,dir++) {
        if (strcmp(dir->filename,filename) == 0) {
            if ((f->f = fopen(flatname,"rb")) != NULL) {
                f->start = dir->start;
                f->offset = 0;
                f->length = dir->length;
                }
            break;
            }
        }
    if (!f->f) {
        free(f);
        f = NULL;
        }
    return f;
}

/****************************************************************************
DESCRIPTION:
Reads a block of binary data from a file opened with GA_openFile.

HEADER:
snap/internal.h

PARAMETERS:
f       - SNAP file handle to read data from
buf     - Buffer to read data into
size    - Size of buffer to read data into

RETURNS:
Number of bytes actually read from the file (<= size).

REMARKS:
This function reads up to size bytes from the file, and returns the actual
number of bytes read.
{secret}
****************************************************************************/
long FLAT_readFile(
    N_FILE *f,
    void *buf,
    long size)
{
    long bytesRead;

    fseek(f->f,f->start + f->offset,SEEK_SET);
    bytesRead = fread(buf,1,size,f->f);
    f->offset += bytesRead;
    return bytesRead;
}

/****************************************************************************
DESCRIPTION:
Close a file previously opened with GA_openFile.

HEADER:
snap/internal.h

PARAMETERS:
f       - SNAP file handle to close

REMARKS:
This function simply close the open file handle.
{secret}
****************************************************************************/
void FLAT_closeFile(
    N_FILE *f)
{
    if (f) {
        fclose(f->f);
        free(f);
        }
}

/****************************************************************************
RETURNS:
Path to the SNAP log file.

REMARKS:
Logs a text string entry to the log file with the specified format. Note
that the string returned is temporary and should be copied to a different
buffer by the calling program immediately after calling this function.
{secret}
****************************************************************************/
const char * NAPI N_getLogName(void)
{
    static char filename[PM_MAX_PATH];
    strcpy(filename,SNAPConfigPath);
    strcat(filename,SNAPLogName);
    return filename;
}

/****************************************************************************
PARAMETERS:
fmt - Format string
... - Remainder of parameters

REMARKS:
Logs a text string entry to the log file with the specified format.
{secret}
****************************************************************************/
void NAPI N_log(
    const char *fmt,
    ...)
{
    char            path[PM_MAX_PATH];
    va_list         argptr;
    FILE            *f;
    static ibool    firstTime = true;

    if (disableLogging)
        return;
    if (firstTime) {
        if ((f = fopen(N_getLogName(),"w")) == NULL) {
            // Try creating the config directory if this fails, since it
            // may not have been created automatically.
            strcpy(path,SNAPConfigPath);
            path[strlen(path)-1] = 0;
            PM_mkdir(path);
            if ((f = fopen(N_getLogName(),"w")) != NULL)
                fclose(f);
            }
        else
            fclose(f);
        firstTime = false;
        }
#ifndef __WIN32_VXD__
    if ((f = fopen(N_getLogName(),"a+")) != NULL) {
        va_start(argptr, fmt);
        vfprintf(f, fmt, argptr);
        va_end(argptr);
        fclose(f);
        }
#endif
}

/****************************************************************************
REMARKS:
Detects and logs the CPU processor information to the SNAP log file.
****************************************************************************/
void N_logCPUInfo(void)
{
    _log("%s\n\n", CPU_getProcessorName());
}

#define isNewLine(p)        ((p) == '\r' || (p) == '\n')
#define isEOF(p,data,size)  (((p) - data) >= size)
#define skipNewLine(p)      while (isNewLine(*p)) p++;

/****************************************************************************
DESCRIPTION:
Finds the name of the driver to load from the driver directory table

HEADER:
graphics/common.h

PARAMETERS:
info        - PCI configuration space for the device
filename    - Place to store the filename for the driver to load

RETURNS:
True on success, false on error.

REMARKS:
Searches the driver directory table on disk to find the appropriate driver
filename to load from disk for the passed in PCI configuration information.
{secret}
****************************************************************************/
ibool _PCI_findDriverName(
    PCIDeviceInfo *info,
    char *filename,
    char *tabFile)
{
    N_FILE  *f;
    int     size,haveDriver;
    char    *data,*p,*vId,*dId,*name;
    ushort  vendorID,deviceID;
    ibool   found = false;

    // Open the driver directory table from the flat file directory
    if ((f = FLAT_openFile(tabFile)) == NULL)
        return false;
    size = f->length;
    if ((data = malloc(size)) == NULL) {
        FLAT_closeFile(f);
        return false;
        }
    FLAT_readFile(f,data,size);
    FLAT_closeFile(f);

    // Now search the table for the appropriate driver name
    p = data;
    while (!isEOF(p,data,size)) {
        // Skip comments and blank lines
        if (*p == ';') {
            while (!isNewLine(*p) && !isEOF(p,data,size))
                p++;
            }
        if (isEOF(p,data,size))
            break;
        if (isNewLine(*p)) {
            skipNewLine(p);
            continue;
            }

        // Now get the VBE/AF flag from table
        haveDriver = (*p == '0');
        p += 2;

        // Now parse the vendorID from table
        vId = p;
        while (*p != ':' && !isNewLine(*p) && !isEOF(p,data,size))
            p++;
        if (isNewLine(*p) || isEOF(p,data,size))
            break;
        *p++ = 0;

        // Now parse the deviceID from table
        dId = p;
        while (!isspace(*p) && !isNewLine(*p) && !isEOF(p,data,size))
            p++;
        if (isNewLine(*p) || isEOF(p,data,size))
            break;
        *p++ = 0;

        // Now parse the driver name from table
        while (isspace(*p) && !isNewLine(*p) && !isEOF(p,data,size))
            p++;
        if (isNewLine(*p) || isEOF(p,data,size))
            break;
        name = p;
        while (!isspace(*p) && !isNewLine(*p) && !isEOF(p,data,size))
            p++;
        if (isNewLine(*p) || isEOF(p,data,size))
            break;
        *p++ = 0;

        // Ignore everything to the end of the line
        while (!isNewLine(*p) && !isEOF(p,data,size))
            p++;
        sscanf(vId,"%hx", &vendorID);
        sscanf(dId,"%hx", &deviceID);
        if (vendorID == info->VendorID && deviceID == info->DeviceID) {
            // We found it so finish the full driver name and break out
            // of the loop.
            strcpy(filename,name);
            found = haveDriver;
            break;
            }
        }
    free(data);
    return found;
}

