/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANCI C
* Environment:  Any
*
* Description:  Module implementing all the CRTC database access routines
*               used by Display Doctor.
*
****************************************************************************/

#include "loader.h"
#include "snap/graphics/gaver.h"

#ifdef __PPC__
#define CDBX_makeLE(p)     _CDBX_swap(p,FALSE)
#define CDBX_makeNative(p) _CDBX_swap(p,TRUE)
#else
#define CDBX_makeLE(p)
#define CDBX_makeNative(p)
#endif

/*------------------------- Global Variables ------------------------------*/

#define MKCRTC_NAME     "mkcrtc.bpd"

static int      CDBX_errCode;
static CDBX     ***db;
static size_t   **dbSize;
static int      MaxDevices = -1;
static int      MaxHeads = -1;

/*------------------------- Implementation --------------------------------*/

/****************************************************************************
RETURNS:
Previous error code.
{secret}
****************************************************************************/
int CDBX_getErrCode(void)
{
    return CDBX_errCode;
}

/****************************************************************************
RETURNS:
String error message
{secret}
****************************************************************************/
char * CDBX_getErrorMsg(void)
{
    static char *msg[] = {
        "Everything is cool",
        "Database file not found",
        "Database is corrupted",
        "Not enough memory to load database",
        "Record was not found in database",
        "Fatal IO error",
        "Database handle is invalid",
        };
    return msg[CDBX_errCode];
}

/*------------------------- Generic Database Code -------------------------*/

/****************************************************************************
REMARKS:
Initialise the database internals. Here we allocate enough space to hold
the CRTC database structures for all graphics cards in the system with the
maximum number of heads for each device.
****************************************************************************/
ibool CDBX_init(
    int maxDevices,
    int maxHeads)
{
    int i;

    if (MaxDevices == -1) {
        MaxDevices = maxDevices;
        MaxHeads = maxHeads;
        if ((db = calloc(MaxDevices,sizeof(*db))) == NULL)
            return false;
        if ((dbSize = calloc(MaxDevices,sizeof(*dbSize))) == NULL)
            return false;
        for (i = 0; i < MaxDevices; i++) {
            if ((db[i] = calloc(MaxHeads,sizeof(**db))) == NULL)
                return false;
            if ((dbSize[i] = calloc(MaxHeads,sizeof(**dbSize))) == NULL)
                return false;
            }
        }
    return true;
}

#ifndef BUILD_MKCRTC
/****************************************************************************
PARAMETERS:
deviceIndex - Index of device to generate timings for

REMARKS:
Internal function to generate the default CRTC timings for the device.
****************************************************************************/
static ibool GenerateCRTCTimings(
    int device,
    int head)
{
    PE_MODULE               *hGenCRTC;
    GA_createCRTCTimings_t  GA_createCRTCTimings;
    ulong                   size;

    if ((hGenCRTC = GA_loadLibrary(MKCRTC_NAME,&size,false)) == NULL)
        return false;
    if ((GA_createCRTCTimings = (GA_createCRTCTimings_t)PE_getProcAddress(hGenCRTC,"_GA_createCRTCTimings")) == NULL)
        return false;
    GA_createCRTCTimings(device,head,SNAPConfigPath,&_PM_imports);
    PE_freeLibrary(hGenCRTC);
    return true;
}
#endif

/****************************************************************************
REMARKS:
Opens the database file on disk for reading or updating.
****************************************************************************/
static FILE *CDBX_openDB(
    int device,
    int head,
    int update)
{
    FILE    *f;
    char    filename[PM_MAX_PATH],buf[PM_MAX_PATH];

    CDBX_errCode = CDBX_ok;
    sprintf(buf,CDBX_NAME,head,device);
    strcpy(filename,SNAPConfigPath);
    strcat(filename,buf);
    if ((f = fopen(filename,update ? "wb" : "rb")) == NULL) {
#ifndef BUILD_MKCRTC
        if (!GenerateCRTCTimings(device,head)) {
            CDBX_errCode = CDBX_IOError;
            return NULL;
            }
        if ((f = fopen(filename,update ? "wb" : "rb")) == NULL)
#endif
            {
            CDBX_errCode = CDBX_IOError;
            return NULL;
            }
        }
    return f;
}

/****************************************************************************
REMARKS:
Closes the database and free's the memory used. We close all heads associated
with that device.
****************************************************************************/
void CDBX_close(
    int device)
{
    int head;

    for (head = 0; head < MaxHeads; head++) {
        if (db[device][head]) {
            free(db[device][head]);
            db[device][head] = NULL;
            }
        }
}

/****************************************************************************
REMARKS:
Converts to and from little-endian format
****************************************************************************/
#ifdef __PPC__
static void _CDBX_swap(
    CDBX *pdev,
    ibool isLE)
{
    int       i;

    if (isLE)
        PM_convLE_32(pdev->h.entries);
    if (pdev->h.entries > 0x100000)
        PM_fatalError("CDBX_swap() given wrong endian flag");
    for (i = 0; i < pdev->h.entries; i++) {
        PM_convLE_16(pdev->records[i].xRes);
        PM_convLE_16(pdev->records[i].yRes);
        PM_convLE_16(pdev->records[i].crtc.HorizontalTotal);
        PM_convLE_16(pdev->records[i].crtc.HorizontalSyncStart);
        PM_convLE_16(pdev->records[i].crtc.HorizontalSyncEnd);
        PM_convLE_16(pdev->records[i].crtc.VerticalTotal);
        PM_convLE_16(pdev->records[i].crtc.VerticalSyncStart);
        PM_convLE_16(pdev->records[i].crtc.VerticalSyncEnd);
        PM_convLE_32(pdev->records[i].crtc.PixelClock);
        PM_convLE_16(pdev->records[i].crtc.RefreshRate);
        }
    if (!isLE)
        PM_convLE_32(pdev->h.entries);
}
#endif

/****************************************************************************
REMARKS:
Attempts to open the database file. This may fail if the file was not found
or the file was not a valid database file (check the CDBX_errCode for the
error).
****************************************************************************/
ibool CDBX_openHead(
    int device,
    int head)
{
    FILE        *f;
    CDBX_header h;

    if ((device >= MaxDevices) || (head >= MaxHeads))
        return false;
    db[device][head] = NULL;
    CDBX_errCode = CDBX_ok;
    if ((f = CDBX_openDB(device,head,false)) == NULL)
        goto Error;
    fread(&h,1,sizeof(h),f);
    if (strcmp(h.signature,CDBX_SIGNATURE) != 0) {
        CDBX_errCode = CDBX_corrupt;
        goto Error;
        }
    fseek(f,0,SEEK_SET);
    dbSize[device][head] = sizeof(CDBX) + (PM_getLE_32(h.entries)-1) * sizeof(CDBX_record);
    if ((db[device][head] = malloc(dbSize[device][head])) == NULL) {
        CDBX_errCode = CDBX_outOfMemory;
        goto Error;
        }
    if (fread(db[device][head],1,dbSize[device][head],f) != dbSize[device][head]) {
        CDBX_errCode = CDBX_IOError;
        goto Error;
        }
    fclose(f);
    CDBX_makeNative(db[device][head]);
    db[device][head]->cursor = 0;
    return true;

Error:
    if (f) fclose(f);
    CDBX_close(device);
    return false;
}

/****************************************************************************
REMARKS:
Attempts to open the database file. This may fail if the file was not found
or the file was not a valid database file (check the CDBX_errCode for the
error). Note that this version will open all heads for the device, up to
the number of heads passed in 'numHeads'.
****************************************************************************/
ibool CDBX_open(
    int device,
    int numHeads)
{
    int head;

    if ((device >= MaxDevices) || (numHeads >= MaxHeads))
        return false;
    for (head = 0; head < numHeads; head++) {
        if (!CDBX_openHead(device,head))
            return false;
        }
    return true;
}

/****************************************************************************
REMARKS:
Do parameter checks for the head and device values. Note that if the
current head is set to clone mode, we always use the CRTC database for
the primary head.
****************************************************************************/
static ibool CDBX_checkDeviceHead(
    int device,
    int *head)
{
    if (*head == gaActiveHeadClone)
        *head = gaActiveHeadPrimary;
    if ((device < 0) || (*head < 0))
        return false;
    if ((device > MaxDevices) || (*head > MaxHeads))
        return false;
    return true;
}

/****************************************************************************
REMARKS:
Returns the version of the SNAP release that created the CRTC database.
****************************************************************************/
int CDBX_getVersion(
    int device,
    int head,
    char *major,
    char *minor)
{
    if (!CDBX_checkDeviceHead(device,&head))
        return CDBX_errCode = CDBX_invalidDB;
    if (db[device][head]) {
        *major = db[device][head]->h.majorVersion;
        *minor = db[device][head]->h.minorVersion;
        return CDBX_errCode = CDBX_ok;
        }
    return CDBX_errCode = CDBX_fileNotFound;
}

/****************************************************************************
REMARKS:
Function to compare to CRTC database entries for sorting purposes.
****************************************************************************/
static int CDBX_cmp(
    const void *_a,
    const void *_b)
{
    CDBX_record *a = (CDBX_record*)_a;
    CDBX_record *b = (CDBX_record*)_b;

    if (a->xRes == b->xRes) {
        if (a->yRes == b->yRes) {
            if (a->crtc.RefreshRate == b->crtc.RefreshRate) {
                ulong aflags = a->crtc.Flags & ~gaDefaultRefresh;
                ulong bflags = b->crtc.Flags & ~gaDefaultRefresh;
                if (aflags == bflags)
                    return 0;
                return aflags - bflags;
                }
            else
                return a->crtc.RefreshRate - b->crtc.RefreshRate;
            }
        else
            return a->yRes - b->yRes;
        }
    else
        return a->xRes - b->xRes;
}

/****************************************************************************
REMARKS:
Moves to the first record in the database.
****************************************************************************/
int CDBX_first(
    int device,
    int head,
    CDBX_record *rec)
{
    if (!CDBX_checkDeviceHead(device,&head))
        return CDBX_errCode = CDBX_invalidDB;
    if (db[device][head]->h.entries) {
        db[device][head]->cursor = 0;
        memcpy(rec,&db[device][head]->records[db[device][head]->cursor],sizeof(CDBX_record));
        return (CDBX_errCode = CDBX_ok);
        }
    return (CDBX_errCode = CDBX_notFound);
}

/****************************************************************************
REMARKS:
Moves to the last record in the database.
****************************************************************************/
int CDBX_last(
    int device,
    int head,
    CDBX_record *rec)
{
    if (!CDBX_checkDeviceHead(device,&head))
        return CDBX_errCode = CDBX_invalidDB;
    if (db[device][head]->h.entries) {
        db[device][head]->cursor = (db[device][head]->h.entries-1);
        memcpy(rec,&db[device][head]->records[db[device][head]->cursor],sizeof(CDBX_record));
        return (CDBX_errCode = CDBX_ok);
        }
    return (CDBX_errCode = CDBX_notFound);
}

/****************************************************************************
REMARKS:
Moves to the next record in the database.
****************************************************************************/
int CDBX_next(
    int device,
    int head,
    CDBX_record *rec)
{
    if (!CDBX_checkDeviceHead(device,&head))
        return CDBX_errCode = CDBX_invalidDB;
    if (db[device][head]->cursor < db[device][head]->h.entries-1) {
        db[device][head]->cursor++;
        memcpy(rec,&db[device][head]->records[db[device][head]->cursor],sizeof(CDBX_record));
        return (CDBX_errCode = CDBX_ok);
        }
    return (CDBX_errCode = CDBX_notFound);
}

/****************************************************************************
REMARKS:
Moves to the next record in the database.
****************************************************************************/
int CDBX_prev(
    int device,
    int head,
    CDBX_record *rec)
{
    if (!CDBX_checkDeviceHead(device,&head))
        return CDBX_errCode = CDBX_invalidDB;
    if (db[device][head]->cursor > 0) {
        db[device][head]->cursor--;
        memcpy(rec,&db[device][head]->records[db[device][head]->cursor],sizeof(CDBX_record));
        return (CDBX_errCode = CDBX_ok);
        }
    return (CDBX_errCode = CDBX_notFound);
}

/****************************************************************************
REMARKS:
Inserts the specified record into the database. Because the database's are
always very small, we simply do a complete copy of the record array,
inserting the new record along the way. If the record already exists, we
update the existing record with the new data.
****************************************************************************/
int CDBX_insert(
    int device,
    int head,
    CDBX_record *rec,
    ibool updateDuplicate)
{
    if (!CDBX_checkDeviceHead(device,&head))
        return CDBX_errCode = CDBX_invalidDB;
    if (db[device][head]) {
        CDBX        *newdb;
        char        *p,*newp,*maxp,*maxnewp;
        int         val;

        /* Allocate space for new records */
        dbSize[device][head] = sizeof(CDBX) + (db[device][head]->h.entries) * sizeof(CDBX_record);
        if ((newdb = malloc(dbSize[device][head])) == NULL)
            return CDBX_outOfMemory;
        memcpy(newdb,db[device][head],sizeof(CDBX) - sizeof(CDBX_record));
        p = (char*)db[device][head]->records;
        maxp = (char*)&db[device][head]->records[db[device][head]->h.entries];
        newp = (char*)newdb->records;
        maxnewp = (char*)&newdb->records[db[device][head]->h.entries+1];

        /* Copy all records prior to new record */
        if (p) {
            while ((p < maxp) && ((val = CDBX_cmp(p,rec)) <= 0)) {
                /* Check if the record already exists and replace if so */
                if (val == 0) {
                    free(newdb);
                    if (updateDuplicate) {
                        memcpy(p,rec,sizeof(CDBX_record));
                        return (CDBX_errCode = CDBX_ok);
                        }
                    return (CDBX_errCode = CDBX_duplicateRecord);
                    }
                memcpy(newp,p,sizeof(CDBX_record));
                p += sizeof(CDBX_record);
                newp += sizeof(CDBX_record);
                }
            }

        /* Add new record */
        memcpy(newp,rec,sizeof(CDBX_record));
        newdb->cursor = (newp - (char*)newdb->records) / sizeof(CDBX_record);
        newp += sizeof(CDBX_record);

        /* Copy all remaining records */
        if (p) {
            while (newp < maxnewp) {
                memcpy(newp,p,sizeof(CDBX_record));
                p += sizeof(CDBX_record);
                newp += sizeof(CDBX_record);
                }
            }

        if (db[device][head])
            free(db[device][head]);
        db[device][head] = newdb;
        db[device][head]->h.entries++;
        return (CDBX_errCode = CDBX_ok);
        }
    return (CDBX_errCode = CDBX_invalidDB);
}

/****************************************************************************
REMARKS:
Updates the current record with the new values in the passed in record.
****************************************************************************/
int CDBX_update(
    int device,
    int head,
    CDBX_record *rec)
{
    if (!CDBX_checkDeviceHead(device,&head))
        return CDBX_errCode = CDBX_invalidDB;
    if (db[device][head]->cursor <= db[device][head]->h.entries-1) {
        memcpy(&db[device][head]->records[db[device][head]->cursor],rec,sizeof(CDBX_record));
        return (CDBX_errCode = CDBX_ok);
        }
    return (CDBX_errCode = CDBX_notFound);
}

/****************************************************************************
REMARKS:
Flushes the database records to disk after they have been updated.
****************************************************************************/
int CDBX_flush(
    int device,
    int head)
{
    int     count;
    FILE    *f = NULL;

    if (!CDBX_checkDeviceHead(device,&head))
        return CDBX_errCode = CDBX_invalidDB;
    CDBX_errCode = CDBX_invalidDB;
    if (db[device][head]) {
        if ((f = CDBX_openDB(device,head,true)) == NULL)
            return CDBX_errCode;
        db[device][head]->h.majorVersion = GA_RELEASE_MAJOR;
        db[device][head]->h.minorVersion = GA_RELEASE_MINOR;
        CDBX_makeLE(db[device][head]);
        count = fwrite(db[device][head],1,dbSize[device][head],f);
        CDBX_makeNative(db[device][head]);
        fclose(f);
        return ((count != dbSize[device][head]) ? (CDBX_errCode = CDBX_IOError) : (CDBX_errCode = CDBX_ok));
        }
    return (CDBX_invalidDB);
}

/****************************************************************************
REMARKS:
Reads the given database records from disk.
****************************************************************************/
ibool CDBX_read(
    char *filename,
    void *data,
    int size)
{
    FILE    *f;
    int     count;

    if ((f = fopen(filename,"rb")) == NULL)
        return false;
    if ((count = fread(data,1,size,f)) == size)
        CDBX_makeNative((CDBX *)data);
    fclose(f);
    return (count == size);
}

/****************************************************************************
REMARKS:
Writes the given database records to disk.
****************************************************************************/
ibool CDBX_write(
    char *filename,
    void *data,
    int size)
{
    FILE    *f;
    int     count;

    if ((f = fopen(filename,"wb")) == NULL)
        return false;
    CDBX_makeLE((CDBX*)data);
    count = fwrite(data,1,size,f);
    CDBX_makeNative((CDBX *) data);
    fclose(f);
    return (count == size);
}

/****************************************************************************
REMARKS:
Function to build the list of valid refresh rates for a particular display
mode.
****************************************************************************/
ibool NAPI CDBX_getRefreshList(
    int device,
    int head,
    int attributes,
    int xRes,
    int yRes,
    int bitsPerPixel,
    int tripleScanLowRes,
    int doubleScan512,
    GA_monitor *monitor,
    N_uint32 maxClock,
    N_uint8 maxRefresh,
    int size,
    N_int32 *refreshRateList,
    GA_GetClosestPixelClock_t findClosestClock)
{
    CDBX_record rec;
    int         count = 0,refresh,lastRefresh = -1;

    // See revision #31 for the broken code to fix up discrete clock
    // refresh rates. We took this out because it was causing problems
    // with disappearing modes.
    if (!CDBX_checkDeviceHead(device,&head))
        return false;
    if (CDBX_first(device,head,&rec) == CDBX_ok) {
        do {
            if (!(attributes & gaHaveDoubleScan) && (rec.crtc.Flags & gaDoubleScan))
                continue;
            if (rec.xRes == xRes && rec.yRes == yRes) {
                if (count < size) {
                    if (monitor->maxResolution != 0xFF) {
                        // Filter refresh rates based on monitor information
                        int hFreq = (rec.crtc.PixelClock / rec.crtc.HorizontalTotal) / 1000;
                        int totalPels = rec.crtc.VerticalTotal * rec.crtc.HorizontalTotal;
                        int vFreq = (rec.crtc.PixelClock + totalPels / 2) / totalPels;
                        if (hFreq < monitor->minHScan || hFreq > monitor->maxHScan)
                            continue;
                        if (vFreq < monitor->minVScan || vFreq > monitor->maxVScan)
                            continue;
                        }
                    if (rec.crtc.PixelClock > maxClock)
                        continue;
                    if (rec.xRes == 320 && (rec.yRes == 200 || rec.yRes == 240)) {
                        if (tripleScanLowRes) {
                            if (rec.crtc.Flags & gaDoubleScan)
                                continue;
                            }
                        else {
                            if (rec.crtc.Flags & gaTripleScan)
                                continue;
                            }
                        }
                    if (rec.xRes == 512 && rec.yRes == 384) {
                        if (doubleScan512) {
                            if (!(rec.crtc.Flags & gaDoubleScan))
                                continue;
                            }
                        else {
                            if (rec.crtc.Flags & gaDoubleScan)
                                continue;
                            }
                        }
                    lastRefresh = refresh = rec.crtc.RefreshRate / 100;
                    if (rec.crtc.Flags & gaInterlaced) {
                        if (!(attributes & gaHaveInterlaced))
                            continue;
                        refresh = -refresh;
                        if (maxRefresh != 0xFF && (refresh/2) > maxRefresh)
                            continue;
                        }
                    else {
                        if (maxRefresh != 0xFF && refresh > maxRefresh)
                            continue;
                        }
                    *refreshRateList++ = refresh;
                    count++;
                    }
                }
            } while (CDBX_next(device,head,&rec) == CDBX_ok);
        }
    *refreshRateList = -1;
    return count > 0;
}

/****************************************************************************
REMARKS:
Function to find the CRTC timings for a particular display mode and refresh.
****************************************************************************/
ibool NAPI CDBX_getCRTCTimings(
    int device,
    int head,
    int attributes,
    int xRes,
    int yRes,
    int tripleScanLowRes,
    int doubleScan512,
    int refresh,
    GA_monitor *monitor,
    N_uint32 maxClock,
    N_uint8 maxRefresh,
    CDBX_record *crtc)
{
    CDBX_record bestFit;
    int         bestRefresh = -1,newRefresh,matchRefresh,isInterlaced = false;
    int         maxBestRefresh = refresh ? refresh : 255;
    ibool       firstRefresh = true;

    if (!CDBX_checkDeviceHead(device,&head))
        return false;
    matchRefresh = refresh * 100;
    if (refresh < 0) {
        // Handle requests for interlaced refresh rates
        refresh = -refresh/2;
        matchRefresh = -matchRefresh;
        isInterlaced = true;
        }
    if (CDBX_first(device,head,crtc) == CDBX_ok) {
        do {
            if (!(attributes & gaHaveDoubleScan) && (crtc->crtc.Flags & gaDoubleScan))
                continue;
            if (crtc->xRes == xRes && crtc->yRes == yRes) {
                if (monitor->maxResolution != 0xFF) {
                    // Filter refresh rates based on monitor information
                    int hFreq = (crtc->crtc.PixelClock / crtc->crtc.HorizontalTotal) / 1000;
                    int totalPels = crtc->crtc.VerticalTotal * crtc->crtc.HorizontalTotal;
                    int vFreq = (crtc->crtc.PixelClock + totalPels / 2) / totalPels;
                    if (hFreq < monitor->minHScan || hFreq > monitor->maxHScan)
                        continue;
                    if (vFreq < monitor->minVScan || vFreq > monitor->maxVScan)
                        continue;
                    }
                if (crtc->crtc.PixelClock > maxClock)
                    continue;
                if (crtc->xRes == 320 && (crtc->yRes == 200 || crtc->yRes == 240)) {
                    if (tripleScanLowRes) {
                        if (crtc->crtc.Flags & gaDoubleScan)
                            continue;
                        }
                    else {
                        if (crtc->crtc.Flags & gaTripleScan)
                            continue;
                        }
                    }
                if (crtc->xRes == 512 && crtc->yRes == 384) {
                    if (doubleScan512) {
                        if (!(crtc->crtc.Flags & gaDoubleScan))
                            continue;
                        }
                    else {
                        if (crtc->crtc.Flags & gaDoubleScan)
                            continue;
                        }
                    }
                newRefresh = crtc->crtc.RefreshRate / 100;
                if (crtc->crtc.Flags & gaInterlaced)
                    newRefresh = newRefresh / 2;
                if (maxRefresh != 0xFF && newRefresh > maxRefresh)
                    continue;
                if (firstRefresh || (bestRefresh < newRefresh && newRefresh <= maxBestRefresh)) {
                    bestFit = *crtc;
                    bestRefresh = newRefresh;
                    firstRefresh = false;
                    }
                if (refresh == 0) {
                    if (crtc->crtc.Flags & gaDefaultRefresh) {
                        if (crtc->crtc.Flags & gaInterlaced && !(attributes & gaHaveInterlaced))
                            return false;
                        return true;
                        }
                    }
                else if (crtc->crtc.RefreshRate == matchRefresh && ((crtc->crtc.Flags & gaInterlaced) == isInterlaced)) {
                    if (crtc->crtc.Flags & gaInterlaced && !(attributes & gaHaveInterlaced))
                        return false;
                    return true;
                    }
                }
            } while (CDBX_next(device,head,crtc) == CDBX_ok);
        }

    // If we get here and we were searching for the default refresh, then
    // we return the best fit value if one was found.
    if (bestRefresh != -1) {
        *crtc = bestFit;
        return true;
        }
    return false;
}

/****************************************************************************
REMARKS:
Function to set the global refresh rate for all display modes to be <=
the passed in refresh rate.
****************************************************************************/
void NAPI CDBX_setGlobalRefresh(
    int device,
    int head,
    int refresh)
{
    CDBX_record rec;
    int         xRes = -1,yRes = -1;

    // Now set the default refresh flag for the highest refresh useable
    // by that display mode, or the current global refresh whichever is
    // lower. We process the CRTC database in reverse order so that we
    // are processing the highest refresh rates first.
    //
    // Note that we only do this for 640x480 and above display modes, not
    // low resolution modes.
    if (!CDBX_checkDeviceHead(device,&head))
        return;
    refresh *= 100;
    if (CDBX_last(device,head,&rec) == CDBX_ok) {
        do {
            rec.crtc.Flags &= ~gaDefaultRefresh;
            if (!(rec.crtc.Flags & gaInterlaced)) {
                // We skip all interlaced modes as possible defaults
                if ((xRes != rec.xRes || yRes != rec.yRes) && rec.crtc.RefreshRate <= refresh) {
                    // Mark this as the default for this display mode
                    xRes = rec.xRes;
                    yRes = rec.yRes;
                    rec.crtc.Flags |= gaDefaultRefresh;
                    }
                }
            if (rec.xRes >= 640 && rec.yRes >= 480) {
                if (CDBX_update(device,head,&rec) != CDBX_ok)
                    return;
                }
            } while (CDBX_prev(device,head,&rec) == CDBX_ok);
        }
}

/****************************************************************************
REMARKS:
Function to set the refresh rate for the specific display mode to the passed
in CRTC parameters.
****************************************************************************/
ibool NAPI CDBX_setCRTCTimings(
    int device,
    int head,
    int xRes,
    int yRes,
    GA_CRTCInfo *crtc,
    ibool addNewTiming,
    ibool makeDefault)
{
    CDBX_record rec,newRec;

    // Clear the default refresh rate flag for the display mode if we
    // are changing the default refresh rate
    if (!CDBX_checkDeviceHead(device,&head))
        return false;
    if (makeDefault && CDBX_first(device,head,&rec) == CDBX_ok) {
        do {
            if (rec.xRes == xRes && rec.yRes == yRes) {
                rec.crtc.Flags &= ~gaDefaultRefresh;
                if (CDBX_update(device,head,&rec) != CDBX_ok)
                    return false;
                }
            } while (CDBX_next(device,head,&rec) == CDBX_ok);
        }

    // Overwrite any existing entry for this refresh rate, or add a new
    // entry to the database if one does not exist. Note that we round up
    // the refresh rate to the nearest 1Hz increment.
    newRec.xRes = xRes;
    newRec.yRes = yRes;
    newRec.crtc = *crtc;
    newRec.crtc.RefreshRate = ((newRec.crtc.RefreshRate + 50) / 100) * 100;
    if (makeDefault)
        newRec.crtc.Flags |= gaDefaultRefresh;
    else
        newRec.crtc.Flags &= ~gaDefaultRefresh;
    if (addNewTiming) {
        // Insert the new record, adding it if it doesn't exist
        return (CDBX_insert(device,head,&newRec,true) == CDBX_ok);
        }
    else if (CDBX_first(device,head,&rec) == CDBX_ok) {
        // Update the existing record only if it exists
        do {
            if (rec.xRes == xRes && rec.yRes == yRes && rec.crtc.RefreshRate == newRec.crtc.RefreshRate) {
                return (CDBX_update(device,head,&newRec) == CDBX_ok);
                }
            } while (CDBX_next(device,head,&rec) == CDBX_ok);
        }
    return false;
}

/****************************************************************************
REMARKS:
Function to return the address and size of the CRTC tables to be saved to
disk for the CRTC database.
****************************************************************************/
void NAPI CDBX_getCRTCTable(
    int device,
    int head,
    void **data,
    int *size)
{
    if (!CDBX_checkDeviceHead(device,&head))
        return;
    *data = db[device][head];
    *size = dbSize[device][head];
}

/****************************************************************************
REMARKS:
Function to set the refresh rate for the specific display mode to the passed
in CRTC parameters.
****************************************************************************/
void NAPI CDBX_setCRTCTimingsV1(
    int device,
    int xRes,
    int yRes,
    GA_CRTCInfo *crtc)
{
    CDBX_setCRTCTimings(device,0,xRes,yRes,crtc,true,true);
}

/****************************************************************************
REMARKS:
Backwards compatibility function.
****************************************************************************/
ibool NAPI CDBX_getRefreshListV1(
    int device,
    int attributes,
    int xRes,
    int yRes,
    int tripleScanLowRes,
    int doubleScan512,
    GA_monitor *monitor,
    N_uint32 maxClock,
    N_uint8 maxRefresh,
    int size,
    N_int32 *refreshRateList)
{
    return CDBX_getRefreshList(device,0,attributes,xRes,yRes,-1,tripleScanLowRes,
        doubleScan512,monitor,maxClock,maxRefresh,size,refreshRateList,NULL);
}

/****************************************************************************
REMARKS:
Function to build the list of valid refresh rates for a particular display
mode.
****************************************************************************/
ibool NAPI CDBX_getRefreshListV2(
    int device,
    int attributes,
    int xRes,
    int yRes,
    int bitsPerPixel,
    int tripleScanLowRes,
    int doubleScan512,
    GA_monitor *monitor,
    N_uint32 maxClock,
    N_uint8 maxRefresh,
    int size,
    N_int32 *refreshRateList,
    GA_GetClosestPixelClock_t findClosestClock)
{
    return CDBX_getRefreshList(device,0,attributes,xRes,yRes,bitsPerPixel,
        tripleScanLowRes,doubleScan512,monitor,maxClock,maxRefresh,size,
        refreshRateList,findClosestClock);
}

/****************************************************************************
REMARKS:
Function to find the CRTC timings for a particular display mode and refresh.
****************************************************************************/
ibool NAPI CDBX_getCRTCTimingsV2(
    int device,
    int attributes,
    int xRes,
    int yRes,
    int tripleScanLowRes,
    int doubleScan512,
    int refresh,
    GA_monitor *monitor,
    N_uint32 maxClock,
    N_uint8 maxRefresh,
    CDBX_record *crtc)
{
    return CDBX_getCRTCTimings(device,0,attributes,xRes,yRes,tripleScanLowRes,
        doubleScan512,refresh,monitor,maxClock,maxRefresh,crtc);
}

/****************************************************************************
REMARKS:
Function to set the global refresh rate for all display modes to be <=
the passed in refresh rate.
****************************************************************************/
void NAPI CDBX_setGlobalRefreshV2(
    int device,
    int refresh)
{
    CDBX_setGlobalRefresh(device,0,refresh);
}

/****************************************************************************
REMARKS:
Function to set the refresh rate for the specific display mode to the passed
in CRTC parameters.
****************************************************************************/
ibool NAPI CDBX_setCRTCTimingsV2(
    int device,
    int xRes,
    int yRes,
    GA_CRTCInfo *crtc,
    ibool addNewTiming,
    ibool makeDefault)
{
    return CDBX_setCRTCTimings(device,0,xRes,yRes,crtc,addNewTiming,makeDefault);
}

