/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  Module to implement DDC functionality based on the
*               low level SCI Serial Control Interface functions.
*
****************************************************************************/

#include "loader.h"

/*--------------------------- Global variables ----------------------------*/

#define MAX_CONTROLS        40
#define MAX_MULTI_CONTROL   40
#define MAX_CAPS            1024

static int      numControls,capsLen,haveGetMultiple,haveSetMultiple;
static uchar    availableControls[MAX_CONTROLS];
static long     maxValues[MAX_CONTROLS];
static long     defaultValues[MAX_CONTROLS];
static char     caps[MAX_CAPS];
static uchar    commonControls[] = {
    MCS_degauss,
    MCS_brightness,
    MCS_contrast,
    MCS_selectColorPreset,
    MCS_redVideoGain,
    MCS_greenVideoGain,
    MCS_blueVideoGain,
    MCS_focus,
    MCS_horizontalPosition,
    MCS_horizontalSize,
    MCS_horizontalPincushion,
    MCS_horizontalPincushionBalance,
    MCS_horizontalMisconvergence,
    MCS_verticalPosition,
    MCS_verticalSize,
    MCS_verticalMisconvergence,
    MCS_parallelogramDistortion,
    MCS_trapezoidalDistortion,
    MCS_tilt,
    MCS_topCornerDistortionControl,
    MCS_topCornerDistortionBalance,
    MCS_bottomCornerDistortionControl,
    MCS_bottomCornerDistortionBalance,
    MCS_hue,
    MCS_saturation,
    MCS_colorCurveAdjust,
    MCS_horizontalMoire,
    MCS_verticalMoire,
    MCS_audioSpeakerVolume,
    MCS_microphoneSpeakerVolume,
    MCS_onScreenDisplayEnable,
    MCS_languageSelect,
    MCS_redVideoBlackLevel,
    MCS_greenVideoBlackLevel,
    MCS_blueVideoBlackLevel,
    0xFF,
    };

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
DESCRIPTION:
Enables the DDC/CI communications channel to the monitor.

PARAMETERS:
id  - Place to store the id string

RETURNS:
0 on success, 1 on a communications error
****************************************************************************/
static int EnableDDC2Bi(
    ibool enable)
{
    int     status,retry = 0;
    uchar   msg[256];

    // Now enable the DDC/CI interface
    do {
        msg[0] = 0xF5;
        msg[1] = enable ? 0x01 : 0x00;
        if ((status = SendDDC2BiMsg(msg, 0x82)) != 0)
            continue;
        } while (status && ++retry <= 2);
    return status;
}

/****************************************************************************
RETURNS:
True on success, false on failure to read string.

REMARKS:
Read the DDC/CI capabilities string from disk
****************************************************************************/
static int ReadCapsString(
    char *PNPID,
    char *caps,
    int *capsLen)
{
    FILE    *f;
    int     status = false;
    char    filename[256];

    // Find the name of the capabilities string file
    strcpy(filename,SNAPConfigPath);
    strcat(filename,PNPID);
    strcat(filename,".ddc");
    if ((f = fopen(filename,"rb")) == NULL)
        return false;
    if (fread(filename,1,8,f) != 8)
        goto Error;
    if (strcmp(filename,PNPID) != 0)
        goto Error;
    if (fread(capsLen,1,sizeof(*capsLen),f) != sizeof(*capsLen))
        goto Error;
    if ((int)fread(caps,1,*capsLen,f) != *capsLen)
        goto Error;
    status = true;
Error:
    fclose(f);
    return status;
}

/****************************************************************************
RETURNS:
True on success, false on failure to read string.

REMARKS:
Write the DDC/CI capabilities string to disk
****************************************************************************/
static int WriteCapsString(
    char *PNPID,
    char *caps,
    int *capsLen)
{
    FILE    *f;
    int     status = false;
    char    filename[256];

    // Find the name of the capabilities string file
    strcpy(filename,SNAPConfigPath);
    strcat(filename,PNPID);
    strcat(filename,".ddc");
    if ((f = fopen(filename,"wb")) == NULL)
        return false;
    if (fwrite(PNPID,1,8,f) != 8)
        goto Error;
    if (fwrite(capsLen,1,sizeof(*capsLen),f) != sizeof(*capsLen))
        goto Error;
    if ((int)fwrite(caps,1,*capsLen,f) != *capsLen)
        goto Error;
    status = true;
Error:
    fclose(f);
    return status;
}

/****************************************************************************
PARAMETERS:
index       - Index into control array
controlCode - MCCS control code (MCS_controlsType)
max         - Place to store the maximum value for the control
value       - Place to store the current control value

RETURNS:
True if the function succeeded, false if it failed.

REMARKS:
This internal function reads the current value and maximum value for a
single control. We use this internally when we need it.
****************************************************************************/
static ibool _MCS_getControl(
    int index,
    uchar controlCode,
    ushort *value,
    ushort *max)
{
    int     status,retry = 0;
    uchar   msg[256],length;

    // Maximum value has not yet been read, so read it
    do {
        msg[0] = 0x01;
        msg[1] = controlCode;
        if ((status = SendDDC2BiMsg(msg, 0x82)) != 0)
            continue;
        if ((status = GetDDC2BiMsg(msg, &length)) != 0)
            continue;
        if ((length & 0x7F) < 8 || msg[1] != 0)
            return false;
        *max = (ushort)msg[4] * 256 + msg[5];
        *value = (ushort)msg[6] * 256 + msg[7];
        if (index != -1) {
            maxValues[index] = *max;
            if (defaultValues[index] == -1)
                defaultValues[index] = *value;
            }
        } while (status && ++retry <= 2);
    return (status == 0);
}

/****************************************************************************
DESCRIPTION:
Checks to see if DDC/CI communication is available and opens it

HEADER:
snap/ddc.h

PARAMETERS:
dc      - SNAP device driver to use for communications
channel - I2C channel to use to control the monitor (0 for primary monitor)

RETURNS:
One of the DDC_errCode error return values.

REMARKS:
This function initializes the DDC communications module. After checking
that the I2C interface is working, this function will attempt to
initialize the DDC communication channel and verify that DDC2B communication
is possible. If DDC2B is not available, this function returns a value of
ddcNoCommunication (ie: cannot communicate with slave). The most likely cause
of this failure condition is that there is no DDC capable monitor attached to
the graphics device.

Once DDC2B functionality has been detected, this function attempts to
read the capabilities string from the monitor using the DDC/CI protocol. If
a DDC/CI compatible monitor is successfully found, this function
returns ddcOk. If the capabilities string could not be read, this function
returns ddcNotAvailable, which generally indicates that the attached monitor
is not DDC/CI capable. This function also parses the monitor capabilities
string to determine what features the monitor supports and initializes
the device driver accordingly.

Note:   After you have finished with communications over the DDC/CI
        interface, you /must/ call MCS_end to close the communications
        channel.

SEE ALSO:
MCS_end
****************************************************************************/
int NAPI MCS_beginExt(
    GA_devCtx *dc,
    N_int32 channel)
{
    int             i,status,control;
    ushort          value,max;
    char            *p,buf[10];
    GA_initFuncs    init;
    GA_options      opt;
    uchar           edid[128];
    EDID_record     rec;
    GA_SCIFuncs     sci;
    N_uint8         SCI_caps;
    N_int32         SCI_ports;

    // Check for certified SNAP driver
    if (_GA_checkDDCCertification(dc) != nOK)
        return ddcNotAvailable;

    // Check if we have support for DDC/CI enabled
    init.dwSize = sizeof(init);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&init))
        return ddcNotAvailable;
    opt.dwSize = sizeof(opt);
    init.GetOptions(&opt);
    if (!opt.bDDCCI)
        return ddcNotAvailable;

    // Make sure the low level SCI functions are available. It is possible
    // for DDC_init to succeed if it is using the BIOS.
    sci.dwSize = sizeof(sci);
    if (!GA_queryFunctions(dc,GA_GET_SCIFUNCS,&sci))
        return ddcNotAvailable;
    if (sci.SCIdetect(&SCI_caps,&SCI_ports) == 0)
        return ddcNotAvailable;

    // Initialise the low level DDC interface
    if ((status = DDC_init(dc)) != ddcOk)
        return status;

    // Read the EDID from the monitor so we can detect what it is
    DDC_readEDID(0xA0,edid,sizeof(edid),0,0);
    if (!EDID_parse(edid,&rec,false))
        return ddcNoCommunication;

    // Initialise the low level DDC/CI interface
    if (EnableDDC2Bi(true) != 0)
        return ddcNotAvailable;

    // Attempt to read the capabilities string from disk, as we cache it
    // to avoid problems with buggy DDC/CI compatible monitors.
    if (!ReadCapsString(rec.PNPID,caps,&capsLen)) {
        // Read capabilities string and determine supported controls
        if ((capsLen = MCS_getCapabilitiesString(caps,sizeof(caps))) == -1) {
            EnableDDC2Bi(false);
            return ddcNotAvailable;
            }
        }

    // Determine supported controls from the capabilities string
    for (p = caps; (p - caps) < capsLen; p++) {
        if (strncmp(p,"vcp(",4) == 0)
            break;
        }
    if ((p - caps) == capsLen) {
        // If we get here, it means that the monitor firmware is buggy and
        // does not list any of the supported controls in the capabilities
        // string. To work around this, we go through a list of common
        // VCP controls and test to see if they are supported or not.
        numControls = 0;
        for (i = 0; commonControls[i] != 0xFF; i++) {
            control = commonControls[i];
            if (_MCS_getControl(-1,control,&value,&max)) {
                availableControls[numControls] = control;
                maxValues[numControls] = -1;
                defaultValues[numControls] = -1;
                numControls++;
                }
            }
        if (numControls == 0)
            return ddcNotAvailable;

        // Since we now know what controls are supported, build a valid
        // vcp() string and attach it to the capabilities string so that
        // we can parse it next time we read the string from disk.
        strcat(caps,"vcp(");
        for (i = 0; i < numControls; i++) {
            sprintf(buf,"%02X",availableControls[i]);
            strcat(caps,buf);
            }
        strcat(caps,")");
        capsLen = strlen(caps);
        }
    else {
        // We have a valid vcp() string, so parse it for the list of
        // supported controls.
        p += 4;
        numControls = 0;
        while (*p != ')' && ((p - caps) < capsLen)) {
            if (numControls == MAX_CONTROLS)
                PM_fatalError("INTERNAL: Too many VCP controls for DDC driver!");
            sscanf(p,"%2x",&control);
            availableControls[numControls] = control;
            maxValues[numControls] = -1;
            defaultValues[numControls] = -1;
            numControls++;
            if (p[2] == ' ')
                p += 3;
            else
                p += 2;
            }
        }

    // Write the capabilities string to disk to cache it for next time
    WriteCapsString(rec.PNPID,caps,&capsLen);

    // Determine if the monitor supports commands 0x0D and 0x0F which
    // is the new mutiple set/get commands.
    haveGetMultiple = false;
    haveSetMultiple = false;
    for (p = caps; (p - caps) < capsLen; p++) {
        if (strncmp(p,"cmds(",5) == 0)
            break;
        }
    if ((p - caps) != capsLen) {
        // We have a valid cmds() string, so parse it for the extended cmds
        p += 5;
        while (*p != ')' && ((p - caps) < capsLen)) {
            sscanf(p,"%2x",&control);
            if (control == 0x0D)
                haveGetMultiple = true;
            if (control == 0x0F)
                haveSetMultiple = true;
            if (p[2] == ' ')
                p += 3;
            else
                p += 2;
            }
        }
    return ddcOk;
}

/****************************************************************************
DESCRIPTION:
Checks to see if DDC/CI communication is available and opens it

HEADER:
snap/ddc.h

PARAMETERS:
dc  - SNAP device driver to use for communications

RETURNS:
0 if DDC available, 1 if not available, -1 if unable to communicate via DDC

REMARKS:
Obsolete. Use MCS_beginExt instead.

SEE ALSO:
MCS_beginExt
****************************************************************************/
int NAPI MCS_begin(
    GA_devCtx *dc)
{
    return MCS_beginExt(dc,SCI_channelMonitorPrimary);
}

/****************************************************************************
DESCRIPTION:
Closes the DDC/CI communications channel

HEADER:
snap/ddc.h

REMARKS:
This function closes the DDC/CI communications channel and must be called
after you have finished DDC/CI communications.

SEE ALSO:
MCS_begin
****************************************************************************/
void NAPI MCS_end(void)
{
    // Disable the low level DDC/CI interface
    EnableDDC2Bi(false);
}

/****************************************************************************
DESCRIPTION:
Reads the DDC/CI capabilities string from the attached monitor.

HEADER:
snap/ddc.h

PARAMETERS:
data    - Buffer to read capabilities string into
maxlen  - Maximum length of the buffer to recieve data

RETURNS:
Length of the capabilities string read from monitor, or -1 on error.

REMARKS:
This function reads the DDC/CI capabilities string from the attached
monitor and returns it in the buffer passed in the data parameter. This
function will only read up to maxlen characters from the monitor, and it
will return the actual number of characters read from the monitor. For a
description of what the capabilities string contains, please refer to the
VESA DDC/CI specification or the ACCESS.bus specification.

Note:   This function removes the binary EDID data that may be present in
        the capabilities string before returning it, so that the string
        may be viewed entirely as an ASCII string. The EDID data can be
        easily read with the DDC_readEDID function so it is superfluous
        to include the EDID in the capabilities string.
****************************************************************************/
int NAPI MCS_getCapabilitiesString(
    char *data,
    N_int32 maxlen)
{
    int     status,retry,numCaps;
    char    *p,*start;
    uchar   msg[512],length;
    ushort  value,max;

    numCaps = 0;
    do {
        retry = 0;
        do {
            msg[0] = 0xF3;
            msg[1] = (uchar)(numCaps >> 8);
            msg[2] = (uchar)(numCaps);
            if ((status = SendDDC2BiMsg(msg, 0x83)) != 0)
                continue;
            if ((status = GetDDC2BiMsg(msg, &length)) != 0)
                continue;
            if (msg[0] != 0xE3)
                return -1;
            length = (length & 0x7F) - 3;
            if (length == 0 || (numCaps + length) >= maxlen) {
                // Delete the useless binary EDID information from the string
                if ((start = strstr(data,"edid bin(")) != NULL) {
                    for (p = start; *p != ')' && (p - data) < numCaps; p++)
                        ;
                    length = numCaps-(p-data);
                    memmove(start,p,length);
                    numCaps -= (p-start);
                    data[numCaps] = 0;
                    }

                // For some reason the Radius monitors screw things up when the
                // capabilities string is read, but if we set any control value
                // it clears the condition.
                if (_MCS_getControl(-1,MCS_horizontalPosition,&value,&max))
                    MCS_setControlValue(MCS_horizontalPosition,value);
                return numCaps;
                }
            memcpy(&data[numCaps],&msg[3],length);
            } while (status && ++retry <= 2);
        numCaps += length;
        } while (retry < 2);
    return -1;
}

/****************************************************************************
DESCRIPTION:
Checks to see if an MCCS control is supported

HEADER:
snap/ddc.h

PARAMETERS:
controlCode - MCCS control code to check (MCS_controlsType)

RETURNS:
True if the control is supported, false if not.

REMARKS:
This function verifies that the specified MCCS control is supported by the
attached monitor. If the control is suported this function returns
true, otherwise it returns false.

SEE ALSO:
MCS_enableControl, MCS_getControlValue, MCS_setControlValue,
MCS_resetControl, MCS_saveCurrentSettings
****************************************************************************/
ibool NAPI MCS_isControlSupported(
    uchar controlCode)
{
    int i;

    // Search the list of supported controls to see if control exists
    for (i = 0; i < numControls; i++) {
        if (availableControls[i] == controlCode)
            return true;
        }
    return false;
}

/****************************************************************************
DESCRIPTION:
Enables or disables a specific MCCS control

HEADER:
snap/ddc.h

PARAMETERS:
controlCode - MCCS control code (MCS_controlsType)
enable      - 1 to enable, 0 to disable the control

RETURNS:
True if the function succeeded, false if it failed.

REMARKS:
This function enables or disables a specified MCCS control for the
attached monitor. This intention of this function is for controls that
have an enabled or disabled state, such as a microphone input, S-Video
camera input or other such controls. Disabling the control effectively
turns off the input, and enabling it turns it on again. Note that most
controls like audio volume controls can be fully controlled via the
regular MCS_setControlValue function.

SEE ALSO:
MCS_isControlSupported, MCS_getControlValue, MCS_setControlValue,
MCS_resetControl, MCS_saveCurrentSettings
****************************************************************************/
ibool NAPI MCS_enableControl(
    uchar controlCode,
    N_int32 enable)
{
    int     status,retry = 0;
    uchar   msg[256];

    do {
        msg[0] = enable ? 0x0B : 0x0A;
        msg[1] = controlCode;
        if ((status = SendDDC2BiMsg(msg, 0x82)) != 0)
            continue;
        } while (status && ++retry <= 2);
    return (status == 0);
}

/****************************************************************************
DESCRIPTION:
Returns the maximum value for an MCCS control

HEADER:
snap/ddc.h

PARAMETERS:
controlCode - MCCS control code (MCS_controlsType)
max         - Place to store the maximum value for the control

RETURNS:
True if the function succeeded, false if it failed.

REMARKS:
This function obtains the maximum value for a specified MCCS control for the
attached monitor. Continuous controls may accept any value between zero and
the maximum value, specific to each control. Non-continuous controls can
only accept specific values. The maximum value returned for non-continuous
controls represents the maximum index for the acceptable values for the
control. A value of 0 for non-continous controls represents that no value
has been selected.

Note:   This function internally reads the maximum value from the
        attached monitor only once for each control between an MCS_begin and
        MCS_end block. If you call it again after the maximum value for
        a control has been read, it simply returns an internal cached
        value for maximum performance.

SEE ALSO:
MCS_isControlSupported, MCS_enableControl, MCS_getControlValue,
MCS_setControlValue, MCS_resetControl, MCS_saveCurrentSettings
****************************************************************************/
ibool NAPI MCS_getControlMax(
    uchar controlCode,
    ushort *max)
{
    int     index;
    ushort  value;

    // Find the control in the list of available controls
    for (index = 0; index < numControls; index++) {
        if (availableControls[index] == controlCode)
            break;
        }
    if (index == numControls)
        return false;

    // Check if the value has already been read and cached
    if (maxValues[index] != -1) {
        *max = maxValues[index];
        return true;
        }

    // Maximum value has not yet been read, so read it
    return _MCS_getControl(index,controlCode,&value,max);
}

/****************************************************************************
DESCRIPTION:
Returns the current value for an MCCS control

HEADER:
snap/ddc.h

PARAMETERS:
controlCode - MCCS control code (MCS_controlsType)
value       - Place to store the current value for the control

RETURNS:
True if the function succeeded, false if it failed.

REMARKS:
This function obtains the current value for a specified MCCS control for the
attached monitor. Continuous controls may accept any value between zero and
the maximum value, specific to each control. Non-continuous controls can
only accept specific values. A value of 0 for non-continous controls
represents that no value has been selected.

Note:   If you need to retrieve the values of multiple controls at a time,
        you should use the more efficient MCS_getControlValues function
        which can read multiple control values directly from the monitor
        in a single packet.

SEE ALSO:
MCS_isControlSupported, MCS_enableControl, MCS_getControlValues,
MCS_setControlValue, MCS_setControlValues, MCS_resetControl,
MCS_saveCurrentSettings
****************************************************************************/
ibool NAPI MCS_getControlValue(
    uchar controlCode,
    ushort *value)
{
    int     index;
    ushort  max;

    // Find the control in the list of available controls
    for (index = 0; index < numControls; index++) {
        if (availableControls[index] == controlCode)
            break;
        }
    if (index == numControls)
        return false;

    // Now read the value of the control
    return _MCS_getControl(index,controlCode,value,&max);
}

/****************************************************************************
DESCRIPTION:
Returns the current value for a list of MCCS controls

HEADER:
snap/ddc.h

PARAMETERS:
numControls     - Number of control values to read (40 max)
controlCodes    - Array of MCCS control codes (MCS_controlsType)
values          - Array to store the current value for each control

RETURNS:
True if the function succeeded, false if it failed.

REMARKS:
This function gets the current value for multiple MCCS controls at a time.
This function is similar to MCS_getControlValue, but is much faster for
reading the value of multiple controls due to the reduced traffic over the
I2C bus.

Note:   This function can only read up to 40 controls at a time. If you
        need to read more than 40 controls, you will need to call this
        function mutiple times.

SEE ALSO:
MCS_isControlSupported, MCS_enableControl, MCS_getControlValue,
MCS_setControlValue, MCS_setControlValues, MCS_resetControl,
MCS_saveCurrentSettings
****************************************************************************/
ibool NAPI MCS_getControlValues(
    N_int32 numControls,
    uchar *controlCodes,
    ushort *values)
{
    int     i,status,retry = 0;
    uchar   msg[256],length;
    ibool   success = true;

    // Check that we don't go over our maximum number of controls
    if (numControls > MAX_MULTI_CONTROL)
        return false;

    // Handle backwards compatibility with DDC/CI 1.0 hardware
    if (!haveGetMultiple) {
        for (i = 0; i < numControls; i++) {
            if (!MCS_getControlValue(controlCodes[i],&values[i])) {
                values[i] = 0xFFFF;
                success = false;
                }
            }
        return success;
        }

    // Now read the value of the controls
    do {
        msg[0] = 0x0D;
        msg[1] = numControls;
        for (i = 0; i < numControls; i++)
            msg[i+2] = controlCodes[i];
        if ((status = SendDDC2BiMsg(msg, 0x82+numControls)) != 0)
            continue;
        if ((status = GetDDC2BiMsg(msg, &length)) != 0)
            continue;
        if ((length & 0x7F) != (2+numControls*3) || msg[2] != 0x0E || msg[1] != 0)
            return false;
        for (i = 0; i < numControls; i++)
            values[i] = (ushort)msg[i*3+3] * 256 + msg[i*3+4];
        } while (status && ++retry <= 2);
    return (status == 0);
}

/****************************************************************************
DESCRIPTION:
Sets the value of an MCCS control to the specified value

HEADER:
snap/ddc.h

PARAMETERS:
controlCode - MCCS control code (MCS_controlsType)
value       - New value for the control

RETURNS:
True if the function succeeded, false if it failed.

REMARKS:
This function sets the current value for a specified MCCS control to the
specified value. Continuous controls may accept any value between zero and
the maximum value, specific to each control. Non-continuous controls can
only accept specific values. The maximum value returned for non-continuous
controls represents the maximum index for the acceptable values for the
control. A value of 0 for non-continous controls represents that no value
has been selected.

Note:   If you need to set the values of multiple controls at a time,
        you should use the more efficient MCS_setControlValues function
        which can set up to 40 control values at the same time.

SEE ALSO:
MCS_isControlSupported, MCS_enableControl, MCS_getControlValue,
MCS_resetControl, MCS_saveCurrentSettings
****************************************************************************/
ibool NAPI MCS_setControlValue(
    uchar controlCode,
    ushort value)
{
    int     status,retry = 0;
    uchar   msg[256];

    do {
        msg[0] = 0x03;
        msg[1] = controlCode;
        msg[2] = (uchar)(value >> 8);
        msg[3] = (uchar)(value);
        if ((status = SendDDC2BiMsg(msg, 0x84)) != 0)
            continue;
        } while (status && ++retry <= 2);
    return (status == 0);
}

/****************************************************************************
DESCRIPTION:
Sets the value of multiple MCCS controls at a time

HEADER:
snap/ddc.h

PARAMETERS:
numControls     - Number of control values to program (40 max)
controlCodes    - Array of MCCS control codes (MCS_controlsType)
values          - Array of new values for the control

RETURNS:
True if the function succeeded, false if it failed.

REMARKS:
This function sets the current value for multiple MCCS controls at a time.
This function is similar to MCS_setControlValue, but is much faster for
setting the value of multiple controls due to reduced traffic over the
I2C bus.

Note:   This function can only program up to 40 controls at a time. If you
        need to program more than 40 controls, you will need to call this
        function multiple times.

SEE ALSO:
MCS_isControlSupported, MCS_enableControl, MCS_getControlValue,
MCS_getControlValues, MCS_setControlValue, MCS_resetControl,
MCS_saveCurrentSettings
****************************************************************************/
ibool NAPI MCS_setControlValues(
    N_int32 numControls,
    uchar *controlCodes,
    ushort *values)
{
    int     i,status,retry = 0;
    uchar   msg[256];

    // Check that we don't go over our maximum number of controls
    if (numControls > MAX_MULTI_CONTROL)
        return false;

    // Handle backwards compatibility with DDC/CI 1.0 hardware
    if (!haveSetMultiple) {
        for (i = 0; i < numControls; i++) {
            if (!MCS_setControlValue(controlCodes[i],values[i]))
                return false;
            }
        return true;
        }

    // Set multiple values at a time over the bus
    do {
        msg[0] = 0x0F;
        msg[1] = numControls;
        for (i = 0; i < numControls; i++) {
            msg[i*3+2] = controlCodes[i];
            msg[i*3+3] = (uchar)(values[i] >> 8);
            msg[i*3+4] = (uchar)(values[i]);
            }
        if ((status = SendDDC2BiMsg(msg, 0x82+numControls*3)) != 0)
            continue;
        } while (status && ++retry <= 2);
    return (status == 0);
}

/****************************************************************************
DESCRIPTION:
Resets the value of an MCCS control to the factory default setting

HEADER:
snap/ddc.h

PARAMETERS:
controlCode - MCCS control code (MCS_controlsType)

RETURNS:
True if the function succeeded, false if it failed.

REMARKS:
This function resets the value for a specified MCCS control to the
factory default setting. This can be used to reset values to reasonable
settings if the user has completely tweaked them out of control.

SEE ALSO:
MCS_isControlSupported, MCS_enableControl, MCS_getControlValue,
MCS_setControlValue, MCS_saveCurrentSettings
****************************************************************************/
ibool NAPI MCS_resetControl(
    uchar controlCode)
{
    int     index,status,retry = 0;
    uchar   msg[256],length;

    // Find the control in the list of available controls
    for (index = 0; index < numControls; index++) {
        if (availableControls[index] == controlCode)
            break;
        }
    if (index == numControls)
        return false;

    do {
        msg[0] = 0x09;
        msg[1] = controlCode;
        if ((status = SendDDC2BiMsg(msg, 0x82)) != 0)
            continue;
        if ((status = GetDDC2BiMsg(msg, &length)) != 0)
            continue;
        } while (status && ++retry <= 2);
    if (status != 0) {
        if (defaultValues[index] != -1 && controlCode != MCS_degauss)
            MCS_setControlValue(controlCode,defaultValues[index]);
        return true;
        }
    return true;
}

/****************************************************************************
DESCRIPTION:
Saves the current settings in the monitor NVRAM

HEADER:
snap/ddc.h

RETURNS:
True if the function succeeded, false if it failed.

REMARKS:
This function saves the current settings for all controls in the non-volatile
RAM (NVRAM) of the attached monitor. This essentially tells the monitor to
remember all the current settings for the current display mode in the NVRAM
in the monitor.

SEE ALSO:
MCS_isControlSupported, MCS_enableControl, MCS_getControlValue,
MCS_setControlValue, MCS_resetControl
****************************************************************************/
ibool NAPI MCS_saveCurrentSettings(void)
{
    int     status,retry = 0;
    uchar   msg[256];

    do {
        msg[0] = 0x0C;
        if ((status = SendDDC2BiMsg(msg, 0x81)) != 0)
            continue;
        } while (status && ++retry <= 2);
    return (status == 0);
}

/****************************************************************************
DESCRIPTION:
Reads the current horizontal and vertical frequency from the monitor

HEADER:
snap/ddc.h

PARAMETERS:
flags   - Place to store the mode flags
hFreq   - Place to store the horizontal frequency (kHz * 100)
vFreq   - Place to store the vertical freqency (Hz * 100)

RETURNS:
True if the function succeeded, false if it failed.

REMARKS:
This function reads the horizontal and vertical frequencies for the
current display mode from the monitor. The horizontal frequency is
reported in units of kHz * 100 (ie: a value of 3150 is 31.5 Khz). The
vertical frequency is reported in units of Hz * 100 (ie: a a value of 7500 is
75 Hz).

This function also reads the sync polarities and returns them in
the flags parameter. The values returned in the flags parameter are defined
by the MCS_polarityFlagsType enumeration.

Note:   It is /highly/ recommended that all monitor vendors implement this
        DDC/CI function, since this function contains useful feedback
        information to the user and graphics device drivers.
****************************************************************************/
ibool NAPI MCS_getTimingReport(
    uchar *flags,
    ushort *hFreq,
    ushort *vFreq)
{
    int     status,retry = 0;
    uchar   msg[256],length;

    do {
        msg[0] = 0x07;
        if ((status = SendDDC2BiMsg(msg, 0x81)) != 0)
            continue;
        if ((status = GetDDC2BiMsg(msg, &length)) != 0)
            continue;
        if ((length & 0x7F) < 6) {
            status = 1;
            continue;
            }
        *flags = msg[1];
        *hFreq = (ushort)msg[2] * 256 + msg[3];
        *vFreq = (ushort)msg[4] * 256 + msg[5];
        } while (status && ++retry <= 2);
    return (status == 0);
}

/****************************************************************************
DESCRIPTION:
Reads the self test report from the attached monitor

HEADER:
snap/ddc.h

PARAMETERS:
flags   - Place to store the mode flags
data    - Place to store the data read from the display
length  - Place to store the length of the data read

RETURNS:
True if the function succeeded, false if it failed.

REMARKS:
This function gets the monitor to perform a self test operation and report
the results back to the host. The information reported back is monitor
vendor specific, so this function is generally used by monitor vendors
to read self test information from monitors in the field for diagnostics
and reporting.
****************************************************************************/
ibool NAPI MCS_getSelfTestReport(
    uchar *flags,
    uchar *data,
    uchar *length)
{
    int     status,retry = 0;
    uchar   msg[256],len;

    do {
        msg[0] = 0xB1;
        if ((status = SendDDC2BiMsg(msg, 0x81)) != 0)
            continue;
        if ((status = GetDDC2BiMsg(data, &len)) != 0)
            continue;
        if ((len & 0x7F) < 2) {
            status = 1;
            continue;
            }
        *length = len & 0x7F;
        *flags = data[1];
        } while (status && ++retry <= 2);
    return (status == 0);
}
