/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  Module to implement DDC functionality based on the
*               low level SCI Serial Control Interface functions.
*
****************************************************************************/

#include "loader.h"
#ifndef __WIN32_VXD__
#include <stdarg.h>
#endif

/*------------------------- Global Variables ------------------------------*/

static GA_SCIFuncs  sci;
static N_uint8      SCI_caps;
static N_int32      SCI_ports;
static int          delayIndex;
static int          CommError;
static int          channel;
static ushort       DDC_version;
#ifndef DEBUG_NEW_DRIVER
static uchar        DDC_transferTime,DDC_attributes;
#endif
#ifdef  DEBUG_DDC
static int          indent;
#define DBG(d)      d
#else
#define DBG(d)
#endif

/* 40 millisecond delay after sending messages. The spec says 40
 * milliseconds, however 40 does not always work in practice but 50 seems
 * to be pretty good for all monitors. We also have code that will speed
 * this up once we have initialised the hardware the first time.
 */

int DDC2BI_sendDelay = 40;
int DDC2BI_receiveDelay = 40;

/* Array of delay times in microseconds for inter-byte processing on slave */

static int delayTimes[] = {
    75,125,250,-1,
    };

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
DESCRIPTION:
Wait for the short period of time to ensure that we never clock faster than
100kHz, per I2C spec. During one full clock period we call this routine
three times:

    SCL LOW, SCL HIGH, SCL LOW.

Therefore, a clock period of 12us (3 x 4us) results in a clock frequency
throttle of 83.3kHz.  That is, of course, assuming an infinitely fast CPU
so that there is no instruction overhead.
****************************************************************************/
#define Wait4us()   _OS_delay(4)

/****************************************************************************
DESCRIPTION:
Opens the VBE/SCI Serial Communications Interface.

PARAMETERS:
channel - Monitor port to use for all subsequent communications
****************************************************************************/
static void sci_begin(void)
{
    sci.SCIbegin(channel);
}

/****************************************************************************
DESCRIPTION:
Closes the VBE/SCI Serial Communications Interface
****************************************************************************/
static void sci_end(void)
{
    sci.SCIend(channel);
}

#ifdef  DEBUG_DDC
static FILE *ddclog = NULL;

/****************************************************************************
DESCRIPTION:
Internal function to indent the text for the DDC debug log file.
****************************************************************************/
static const char *Indent(
    int inc)
{
    static char buf[80];
    int i;
    for (i = 0; i < indent*2; i++)
        buf[i] = ' ';
    buf[i] = 0;
    indent += inc;
    return buf;
}

/****************************************************************************
PARAMETERS:
fmt - Format string
... - Remainder of parameters

REMARKS:
Logs a text string entry to the log file with the specified format.
****************************************************************************/
static void DDC_log(
    const char *fmt,
    ...)
{
    va_list         argptr;
    static ibool    firstTime = true;

    if (firstTime) {
        ddclog = fopen("ddc.log","w");
        firstTime = false;
        }
    if (ddclog == NULL)
        ddclog = fopen("ddc.log","a+");
    if (ddclog) {
        va_start(argptr, fmt);
        vfprintf(ddclog, fmt, argptr);
        va_end(argptr);
        }
}

/****************************************************************************
REMARKS:
Ends logging for the debug log file.
****************************************************************************/
static void DDC_endLog(void)
{
    fclose(ddclog);
    ddclog = NULL;
}

/****************************************************************************
REMARKS:
Toggles the I2C line with debug information logging
****************************************************************************/
static void scl(int bit)
{
#ifdef  LOG_SCL_SDA
    DDC_log("%sSCL <- %d\n", Indent(0), (bit) == 0 ? 0 : 1);
#endif
    sci.SCIwriteSCL(channel,(bit) == 0 ? 0 : 1);
}

/****************************************************************************
REMARKS:
Toggles the I2C line with debug information logging
****************************************************************************/
static void sda(int bit)
{
#ifdef  LOG_SCL_SDA
    DDC_log("%sSDA <- %d\n", Indent(0), (bit) == 0 ? 0 : 1);
#endif
    sci.SCIwriteSDA(channel,(bit) == 0 ? 0 : 1);
}

/****************************************************************************
REMARKS:
Reads the I2C line with debug information logging
****************************************************************************/
static int read_scl(void)
{
    int bit = (sci.SCIreadSCL(channel) == 0 ? 0 : 1);
#ifdef  LOG_SCL_SDA
    DDC_log("%sSCL -> %d\n", Indent(0), bit);
#endif
    return bit;
}

/****************************************************************************
REMARKS:
Reads the I2C line with debug information logging
****************************************************************************/
static int read_sda(void)
{
    int bit = (sci.SCIreadSDA(channel) == 0 ? 0 : 1);
#ifdef  LOG_SCL_SDA
    DDC_log("%sSDA -> %d\n", Indent(0), bit);
#endif
    return bit;
}
#else
#define scl(bit)    sci.SCIwriteSCL(channel,(bit) == 0 ? 0 : 1)
#define sda(bit)    sci.SCIwriteSDA(channel,(bit) == 0 ? 0 : 1)
#define read_scl()  (sci.SCIreadSCL(channel) == 0 ? 0 : 1)
#define read_sda()  (sci.SCIreadSDA(channel) == 0 ? 0 : 1)
#endif

/****************************************************************************
DESCRIPTION:
Initialise the time delay. If we have clock stretching ability (ie: we can
read from SCL), then we can start out with no delay.
****************************************************************************/
static void InitTimeDelay(void)
{
    delayIndex = 0;
DBG(DDC_log("InitTimeDelay to %d us\n",delayTimes[delayIndex]);)
}

/****************************************************************************
DESCRIPTION:
Wait for a short period of time between processing bytes, to allow the
slave to catch up. If we have clock stretching, this is not needed, however
if we don't have clock stretching we need this (and we bump up the delay
until we get a transfer).
****************************************************************************/
static void IssueTimeDelay(void)
{
    if (!(SCI_caps & SCI_readSCL))
        _OS_delay(delayTimes[delayIndex]);
}

/****************************************************************************
DESCRIPTION:
Bump the current time delay to slow down DDC communications

RETURNS:
True if new time delay selected, false if max time delay reached
****************************************************************************/
static ibool BumpTimeDelay(void)
{
    if (SCI_caps & SCI_readSCL)
        return false;
    if (delayTimes[++delayIndex] == -1) {
DBG(DDC_log("BumpTimeDelay timed out!\n");)
        delayIndex--;
        return false;
        }
DBG(DDC_log("BumpTimeDelay to %d us\n",delayTimes[delayIndex]);)
    return true;
}

/****************************************************************************
DESCRIPTION:
Wait for the I2C SCL Clock Line to become high.

REMARKS:
If the graphics controller supports reading of the SCL clock line, this
function effectively implements clock stretching whereby the I2C slave will
hold the clock line low to speed throttle the I2C host. Hence we wait in
here until the clock line goes high to allow for this.

If the graphics controller does not support reading of the SCL clock line,
this function simply returns after the wait period and we have to use
timer based speed throttling.

Note:   We also check for timeout conditions when waiting for the SCL
        line to go high during clock stretching, and will return false if
        we time out waiting for SCL to go high. If this happens, then we
        return false and the remaining code should bail out and send an
        error recovery signal.

RETURNS:
True if SCL returned high, false on wait for SCL timeout.
****************************************************************************/
static ibool WaitSCLHigh(void)
{
    int i;

    if (SCI_caps & SCI_readSCL) {
        for (i = 0; i < 500; i++) {
            if (read_scl() == 1)
                break;
            Wait4us();
            }
DBG(    if (i == 500) DDC_log("WaitSCLHigh timed out!\n");)
        return (read_scl() == 1);
        }
    Wait4us();
    return true;
}

/****************************************************************************
DESCRIPTION:
Send a start sequence to the I2C bus

RETURNS:
0 on success, 1 on a communications error
****************************************************************************/
static int SendStart(void)
{
DBG(DDC_log("%sSendStart()\n",Indent(1));)
    /* Setup for the pending start sequence */
    sda(1);
    Wait4us();
    if (read_sda() == 0)
        return (CommError = 1);
    Wait4us();
    scl(1);
    if (!WaitSCLHigh())
        return (CommError = 1);

    /* The real start sequence */
    sda(0);
    Wait4us();
    scl(0);
    Wait4us();
DBG(Indent(-1);)
    return 0;
}

/****************************************************************************
DESCRIPTION:
Send a stop sequence to the I2C bus

RETURNS:
0 on success, 1 on a communications error
****************************************************************************/
static int SendStop(void)
{
DBG(DDC_log("%sSendStop()\n",Indent(1));)
    scl(0);
    Wait4us();
    sda(0);
    Wait4us();
    scl(1);
    if (!WaitSCLHigh())
        return (CommError = 1);
    sda(1);
    Wait4us();
DBG(Indent(-1);)
    return 0;
}

/****************************************************************************
DESCRIPTION:
Send a bit of data to the I2C bus

PARAMETERS:
bit - Bit of data to send to I2C bus (0 or 1)

RETURNS:
0 on success, 1 on a communications error
****************************************************************************/
static int SendBit(
    int bit)
{
#ifdef  LOG_BITS
DBG(DDC_log("%sSendBit(%d)\n",Indent(1), (bit) == 0 ? 0 : 1);)
#endif
    scl(0);
    Wait4us();
    sda(bit);
    Wait4us();
    scl(1);
    if (!WaitSCLHigh())
        return (CommError = 1);
    scl(0);
    Wait4us();
DBG(Indent(-1);)
    return 0;
}

/****************************************************************************
DESCRIPTION:
Read a single bit of data from the I2C bus

RETURNS:
Bit of data read from I2C bus (0 or 1)
****************************************************************************/
static uchar ReadBit(void)
{
    uchar bit;

#ifdef  LOG_BITS
DBG(DDC_log("%sReadBit...\n",Indent(1));)
#endif
    scl(0);
    Wait4us();
    sda(1);
    Wait4us();
    scl(1);
    if (!WaitSCLHigh())
        return (CommError = 1);
    bit = read_sda();
    scl(0);
    Wait4us();
#ifdef  LOG_BITS
DBG(DDC_log("%s = %d\n",Indent(-1), bit);)
#endif
    return bit;
}

/****************************************************************************
DESCRIPTION:
Send a byte of data to the I2C bus

PARAMETERS:
byte    - Byte of data to send to I2C bus

RETURNS:
0 on success, 1 on a communications error
****************************************************************************/
static int SendByte(
    uchar byte)
{
DBG(DDC_log("%sSendByte(0x%02X)\n",Indent(1),byte);)
    SendBit(byte&128);
    SendBit(byte&64 );
    SendBit(byte&32 );
    SendBit(byte&16 );
    SendBit(byte&8  );
    SendBit(byte&4  );
    SendBit(byte&2  );
    SendBit(byte&1  );
    if (CommError || ReadBit())
        return (CommError = 1);
    IssueTimeDelay();
DBG(Indent(-1);)
    return 0;
}

/****************************************************************************
DESCRIPTION:
Read a byte of data via the I2C bus

PARAMETERS:
byte    - Place to store the byte of data read from I2C bus
ack     - 0 to ACK the slave for the last byte sent, 1 to NACK it

RETURNS:
0 on success, 1 on a communications error
****************************************************************************/
static uchar ReadByte(
    uchar *byte,
    int ack)
{
DBG(DDC_log("%sReadByte...\n",Indent(1));)
    *byte  = ReadBit() << 7;
    *byte |= ReadBit() << 6;
    *byte |= ReadBit() << 5;
    *byte |= ReadBit() << 4;
    *byte |= ReadBit() << 3;
    *byte |= ReadBit() << 2;
    *byte |= ReadBit() << 1;
    *byte |= ReadBit();
    if (CommError || SendBit(ack))
        return (CommError = 1);
    IssueTimeDelay();
DBG(DDC_log("%s = 0x%02X\n",Indent(-1), *byte);)
    return 0;
}

/****************************************************************************
DESCRIPTION:
Write a block of data via the I2C bus

PARAMETERS:
data    - Data to write to the I2C bus
length  - Number of bytes to write

RETURNS:
0 on success, 1 on a communications error
****************************************************************************/
static int WriteBlock(
    uchar *data,
    int length)
{
    int i;

    for (i = 0; i < length; i++) {
        if (SendByte(data[i]))
            return (CommError = 1);
        }
    return 0;
}

/****************************************************************************
DESCRIPTION:
Read a block of data via the I2C bus

PARAMETERS:
data    - Place to store the data read from I2C bus
length  - Number of bytes to read

RETURNS:
0 on success, 1 on a communications error
****************************************************************************/
static int ReadBlock(
    uchar *data,
    int length)
{
    int i;

    for (i = 0; i < length; i++) {
        if (ReadByte(&data[i],(i == length-1) ? 1 : 0))
            return (CommError = 1);
        }
    return 0;
}

/****************************************************************************
DESCRIPTION:
Send an error recovery signal to the I2C bus, to recover after an error
during communications. This is important to ensure that we reset the I2C
bus if we are communicating too fast or something goes wrong, so that
we avoid situations of accidentally writing data to the EDID blocks.
****************************************************************************/
static void ErrorRecovery(void)
{
    int i;

    /* Send up to a maximum of 9 stop signals */
DBG(DDC_log("ErrorRecovery()\n");)
    for (i = 0; i < 9; i++) {
        if (SCI_caps & SCI_readSCL) {
            if ((read_sda() == 1) & (read_scl() == 1))
                break;
            }
        SendStop();
        }

    /* Clear the communications error flag */
    CommError = 0;
}

/****************************************************************************
DESCRIPTION:
Send the slave starting address to begin writing data

PARAMETERS:
slaveAddress    - Slave DDC address to write data to
address         - Offset within block of data to write

RETURNS:
0 on success, 1 on a communications error
****************************************************************************/
static int SendWriteAddress(
    int slaveAddr,
    int address)
{
DBG(DDC_log("%sSendWriteAddress(0x%02X,%d)\n",Indent(1),slaveAddr,address);)
    if (SendStart())
        return (CommError = 1); // Recieved NACK, bail out
    if (SendByte(slaveAddr))
        return (CommError = 1); // Recieved NACK, bail out
    if (SendByte(address))
        return (CommError = 1); // Recieved NACK, bail out
DBG(Indent(-1);)
    return 0;
}

/****************************************************************************
DESCRIPTION:
Send the slave starting address to begin reading data

PARAMETERS:
slaveAddress    - Slave DDC address to read data from
address         - Offset within block of data to read

RETURNS:
0 on success, 1 on a communications error
****************************************************************************/
static int SendReadAddress(
    int slaveAddr,
    int address)
{
DBG(DDC_log("%sSendReadAddress(0x%02X,%d)\n",Indent(1),slaveAddr,address);)
    if (SendWriteAddress(slaveAddr,address))
        return (CommError = 1); // Recieved NACK, bail out
    if (SendStart())
        return (CommError = 1); // Recieved NACK, bail out
    if (SendByte(slaveAddr+1))
        return (CommError = 1); // Recieved NACK, bail out
DBG(Indent(-1);)
    return 0;
}

/****************************************************************************
DESCRIPTION:
Waits for the specified number of milliseconds

PARAMETERS:
msecs   - Number of milliseconds to delay
****************************************************************************/
static void WaitMS(
    int msecs)
{
    _OS_delay((msecs / 2) * 1000);
    _OS_delay((msecs / 2) * 1000);
}

/****************************************************************************
DESCRIPTION:
Sends a DDC2Bi message and delay 40 milliseconds after sending

PARAMETERS:
data    - Data to send
length  - Length of the data to send

RETURNS:
0 on success, 1 on a communications error
{secret}
****************************************************************************/
int SendDDC2BiMsg(
    uchar *data,
    uchar length)
{
    int     i,l;
    uchar   checksum,abMsg[256];

    // Send data in access bus format
DBG(DDC_log("\nSendDDC2BiMsg\n");)
    sci_begin();
    abMsg[0] = 0x51;
    abMsg[1] = length;
    checksum = 0x6E ^ 0x51 ^ length;
    l = length & 0x7F;
    for (i = 0; i < l; i++) {
        abMsg[i+2] = data[i];
        checksum ^= data[i];
        }
    abMsg[l+2] = checksum;
    ErrorRecovery();            // Recover from any errors
    if (SendStart())
        goto Error;             // Recieved NACK, bail out
    if (SendByte(0x6E))
        goto Error;             // Recieved NACK, bail out
    if (WriteBlock(abMsg,l+3))
        goto Error;
    if (SendStop())
        goto Error;
    sci_end();
    WaitMS(DDC2BI_sendDelay);
DBG(DDC_log("\nEnd SendDDC2BiMsg - Success!\n");
    DDC_endLog();)
    return 0;

Error:
    ErrorRecovery();
    sci_end();
DBG(DDC_log("\nEnd SendDDC2BiMsg - Failed!\n");
    DDC_endLog();)
    return 1;
}

/****************************************************************************
DESCRIPTION:
Recieves a DDC2Bi message and delay 40 milliseconds after sending

PARAMETERS:
data    - Data to send
length  - Place to store the length of the data recieved

RETURNS:
0 on success, 1 on a communications error
{secret}
****************************************************************************/
int GetDDC2BiMsg(
    uchar *data,
    uchar *length)
{
    int     i,l;
    uchar   header,checksum;

    // Receive data in DDC2Bi mode
DBG(DDC_log("\nGetDDC2BiMsg\n");)
    sci_begin();
    ErrorRecovery();            // Recover from any errors
    if (SendStart())
        goto Error;             // Recieved NACK, bail out
    if (SendByte(0x6F))
        goto Error;             // Recieved NACK, bail out
    if (ReadByte(&header,0) || header != 0x6E)
        goto Error;
    if (ReadByte(length,0))
        goto Error;
    l = (*length & 0x7F)+1;     // Add checksum to read
    if (ReadBlock(data,l))
        goto Error;
    if (SendStop())
        goto Error;

    // Compute the checksum
    checksum = 0x6E ^ 0x50 ^ *length;
    for (i = 0; i < l; i++)
        checksum ^= data[i];
    if (checksum != 0)
        goto Error;
    sci_end();
    WaitMS(DDC2BI_receiveDelay);
DBG(DDC_log("\nEnd GetDDC2BiMsg - Success!\n");
    DDC_endLog();)
    return 0;

Error:
    ErrorRecovery();
    sci_end();
DBG(DDC_log("\nEnd GetDDC2BiMsg - Failed!\n");
    DDC_endLog();)
    return 1;
}

/****************************************************************************
DESCRIPTION:
Checks for valid DDC2B communications functionality

RETURNS:
True if DDC2B is available, false if not

REMARKS:
This function verifies whether a monitor is capable of DDC2B protocol
by requesting an EDID block and checking that the monitor slave has ACK'd
us at least once during the request.

We do this by issuing two command cycles to the monitor slave.  The first
command cycle is a write of A0h and a write of 00h (to designate starting
address of the EDID block). The second command cycle is a write of A1h and
then a read (of the first EDID byte).  Check the ACK or NACK for each byte
written (three in all).  If there is at least one ACK, then the slave is
judged to be DDC2B compliant.

This function also handles bumping of the time delay for host time based
DDC communications where the graphics controller does not support clock
stretching.
****************************************************************************/
static ibool CheckDDC2B(void)
{
    int     found;
    uchar   h[8],hgood[8] = {0x00,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0x00};

    // Set monitor port and initialise time delay
    InitTimeDelay();

    // Attempt DDC2B communications
DBG(DDC_log("\nCheckDDC2B()\n");)
    sci_begin();
    ErrorRecovery();
    while (1) {
        // Send starting address to slave for reading (0xA0 EDID block)
        found = true;
        if (SendReadAddress(0xA0,0))
            found = false;
        else {
            // Read the EDID header bytes
            if (ReadBlock(h,sizeof(h)))
                found = false;
            SendStop();
            if (memcmp(h,hgood,sizeof(h)) != 0)
                found = false;
            }
        if (found)
            break;

        // We did not recieve any ACK's, so send an error recovery
        // signal and increase the time delay if we dont have clock
        // stretching support. Note that even if we have the ability
        // to read the SCL line, we still bump the time delay as we don't
        // know if the slave I2C monitor interface implements clock
        // stretching correctly or not.
        ErrorRecovery();
        if (!BumpTimeDelay())
            break;
        }
    ErrorRecovery();
    sci_end();
DBG(DDC_log("\nEnd CheckDDC2B()\n");
    DDC_endLog();)
    return found;
}

/****************************************************************************
PARAMETERS:
sciFuncs    - Pointer to SCI functions to use for detection
iChannel    - DDC channel to use for communications (0 for primary monitor)

RETURNS:
True if a DDC monitor attached to channel, false if not.

REMARKS:
This internal function is called by device drivers to do a quick check on
a DDC port to see if a DDC monitor is present. If one is present this
function will return true. If not this function will return false. The
primary purpose of this function is to allow device drivers access to
a function that will allow them to automatically determine which port
the primary and secondary displays are attached to.
{secret}
****************************************************************************/
int NAPI _DDC_quickDetect(
    GA_SCIFuncs *sciFuncs,
    N_int32 iChannel)
{
    sci = *sciFuncs;
    if (sci.SCIdetect(&SCI_caps,&SCI_ports) != 0) {
        channel = iChannel;
        return CheckDDC2B();
        }
    return false;
}

/****************************************************************************
PARAMETERS:
sciFuncs    - Pointer to SCI functions to use for detection
iChannel    - DDC channel to use for communications (0 for primary monitor)
xRes        - Place to store the LCD panel X resolution
yRes        - Place to store the LCD panel Y resolution

RETURNS:
True if panel found and resolution successfully parsed, false if not.

REMARKS:
This internal function is called by device drivers to do DDC to the
DVI/DFP flat panel to determine the resolution of the panel if there is
no other way to detect the panel size from the BIOS etc.
{secret}
****************************************************************************/
int NAPI _DDC_getFlatPanelResolution(
    GA_SCIFuncs *sciFuncs,
    N_int32 iChannel,
    N_int16 *xRes,
    N_int16 *yRes)
{
    uchar       edid[128];
    EDID_record edidRec;

    sci = *sciFuncs;
    if (sci.SCIdetect(&SCI_caps,&SCI_ports) == 0)
        return false;
    channel = iChannel;
    if (!DDC_readEDID(0xA0,edid,sizeof(edid),0,iChannel))
        return false;
    if (!EDID_parse(edid,&edidRec,false))
        return false;
    if (!(edidRec.flags & EDID_Level_Digital))
        return false;
    if (edidRec.numDetailedTimings == 0)
        return false;
    *xRes = edidRec.detailedTimings[0].hActive;
    *yRes = edidRec.detailedTimings[0].vActive;
    return true;
}

/****************************************************************************
DESCRIPTION:
Checks to see if DDC communication is available.

HEADER:
snap/ddc.h

PARAMETERS:
dc          - SNAP device driver to use for communications
iChannel    - DDC channel to use for communications (0 for primary monitor)

RETURNS:
One of the DDC_errCode error return values.

REMARKS:
This function initializes the DDC communications module. After checking
that the I2C interface is working, this function will then attempt to
initialize the DDC communication channel and verify that DDC2B communication
is possible. If DDC2B is not available, this function will return a value of
ddcNoCommunication (ie: cannot communicate with slave). The most likely cause
of this failure condition is that there is no DDC capable monitor attached
to the graphics device.

Note:   If this function returns a value of ddcNotAvailable, it means the
        installed graphics device does not support DDC comunications.

SEE ALSO:
DDC_readEDID, DDC_initExt
****************************************************************************/
int NAPI DDC_initExt(
    GA_devCtx *dc,
    N_int32 iChannel)
{
#ifndef DEBUG_NEW_DRIVER
    RMREGS      regs;
    RMSREGS     sregs;
#endif
#ifdef DEBUG_DDC
    indent = 0;
#endif

    // Check for certified SNAP driver
    if (_GA_checkDDCCertification(dc) != nOK)
        return ddcNotAvailable;

    // Check for presence of the Serial Control Interface functions
    DDC_version = 0;
    sci.dwSize = sizeof(sci);
    if (!GA_queryFunctions(dc,GA_GET_SCIFUNCS,&sci))
        return ddcNotAvailable;
    if (sci.SCIdetect(&SCI_caps,&SCI_ports) != 0) {
        // Now attempt to verify the presence of DDC2B communications. This
        // function also initialised the time delay used to speed throttle
        // the DDC communications, if the host graphics controller does not
        // support clock stretching (ie: we can't read from the SCL clock line).
        channel = iChannel;
        if (CheckDDC2B())
            return ddcOk;
        if (CheckDDC2B())
            return ddcOk;
        return ddcNoCommunication;
        }
#ifndef DEBUG_NEW_DRIVER
    // If we don't have the SCI functions, check we can do this
    // via the BIOS for the graphics card (if available).
    if (!(_IS_OS(_OS_OS2) && PM_haveBIOSAccess()) && !(dc->WorkArounds & gaNoDDCBIOS) && iChannel == SCI_channelMonitorPrimary) {
        regs.x.ax = 0x4F15;
        regs.h.bl = 0;
        regs.x.cx = 0;
        regs.x.di = 0;
        sregs.es = 0;
        PM_int86x(0x10, &regs, &regs, &sregs);
        if (regs.x.ax != 0x004F)
            return ddcNotAvailable;
        DDC_version = regs.x.cx;
        DDC_attributes = regs.h.bl;
        DDC_transferTime = regs.h.bh;
        if (DDC_version == 0 && (DDC_attributes & 0x3) != 0)
            DDC_version = 0x100;
        if (DDC_version == 0 || DDC_attributes == 0)
            return ddcNotAvailable;
        return ddcOk;
        }
#endif
    return ddcNotAvailable;
}

/****************************************************************************
DESCRIPTION:
Checks to see if DDC communication is available.

HEADER:
snap/ddc.h

PARAMETERS:
dc      - SNAP device driver to use for communications

RETURNS:
One of the DDC_errCode error return values.

REMARKS:
Obsolete. Use DDC_initExt() instead.

SEE ALSO:
DDC_initExt
****************************************************************************/
int NAPI DDC_init(
    GA_devCtx *dc)
{
    return DDC_initExt(dc,SCI_channelMonitorPrimary);
}

/****************************************************************************
DESCRIPTION:
Attempts to read the EDID information block from the monitor.

HEADER:
snap/ddc.h

PARAMETERS:
slaveAddr   - Slave address to read the EDID data from
edid        - Place to store the EDID information read
length      - Number of bytes of EDID data to read
blockNumber - EDID block number to read (generally 0)
iChannel    - DDC channel to use for communications (0 for primary monitor)

RETURNS:
True on success, false for invalid checksum or communications error.

REMARKS:
This function attempts to read the EDID information from the DDC2B slave.
This function also does a checksum on the incoming EDID data, and if the
checksum fails will return false. The slave address that is passed to this
function should be 0xA0 to read the regular 128 byte EDID for the DDC 2.0
specification.

For DDC 3.0 you can pass in values of 0xA2 and 0xA6 for the Plug and Display
and FPDI 256 byte extended EDID blocks respectively.

SEE ALSO:
DDC_initExt, EDID_parse
****************************************************************************/
ibool NAPI DDC_readEDID(
    N_int32 slaveAddr,
    uchar *edid,
    N_int32 length,
    N_int32 blockNumber,
    N_int32 iChannel)
{
    RMREGS  regs;
    RMSREGS sregs;
    uchar * rmPtr;
    uint    rseg,roff,rlen;
    uchar   checksum;
    int     i,startAddress = blockNumber * length;

    channel = iChannel;
    if (DDC_version && iChannel == SCI_channelMonitorPrimary) {
        if (slaveAddr != 0xA0 || length != 128)
            return false;
        if ((rmPtr = PM_getVESABuf(&rlen, &rseg, &roff)) == NULL)
            return false;
        regs.x.ax = 0x4F15;
        regs.h.bl = 1;
        regs.x.cx = 0;
        regs.x.dx = 0;
        regs.x.di = (ushort)roff;
        sregs.es = (ushort)rseg;
#if defined(DDC_BIOS_IOTRACE) && defined(DEBUG_X86EMU)
        if (BE_loaded)
            BE_setDebugFlags(DEBUG_IO_TRACE_F);
#endif
        PM_int86x(0x10, &regs, &regs, &sregs);
#if defined(DDC_BIOS_IOTRACE) && defined(DEBUG_X86EMU)
        if (BE_loaded)
            BE_setDebugFlags(0);
#endif
        memcpy(edid, rmPtr, length);
        return true;
        }
    else {
        // Send starting address to slave for read operation
DBG(    DDC_log("\nDDC_readEDID()\n");)
        sci_begin();
        ErrorRecovery();
        if (SendReadAddress(slaveAddr,startAddress))
            goto Error;

        // Read the EDID bytes
        if (ReadBlock(edid,length))
            goto Error;
        for (i = checksum = 0; i < length; i++)
            checksum += edid[i];

        // Send stop to slave
        SendStop();
        ErrorRecovery();
        sci_end();
DBG(    DDC_log("\nEnd DDC_readEDID() - Success!\n");
        DDC_endLog();)
        return (checksum == 0);

Error:
        ErrorRecovery();
        sci_end();
DBG(    DDC_log("\nEnd DDC_readEDID() - Failed!\n");
        DDC_endLog();)
        return false;
        }
}

/****************************************************************************
DESCRIPTION:
Attempts to write the EDID information block to the monitor.

HEADER:
snap/ddc.h

PARAMETERS:
dc          - SNAP device driver to use for communications
slaveAddr   - Slave address to write the EDID data to (ie: 0xA0 for EDID 2)
edid        - Place to store the EDID information read
length      - Number of bytes of EDID data to read
blockNumber - EDID block number to write (generally 0)
iChannel    - DDC channel to use for communications (0 for primary monitor)

RETURNS:
True on success, false on error.

REMARKS:
This function attempts to write the EDID information to the monitor. Most
monitors allow writing of the EDID data to the monitor so that they can
be configured during factory testing and setup. This feature is used to
reprogram the EDID information in a monitor in the field.
Note that in order to re-program the EDID information in the monitor, you
need to ensure that the VSYNC is held high for the duration of writes to
the EDID, as this is the write protection used by the monitor.

SEE ALSO:
DDC_initExt, DDC_readEDID
****************************************************************************/
ibool NAPI DDC_writeEDID(
    GA_devCtx *dc,
    N_int32 slaveAddr,
    uchar *edid,
    N_int32 length,
    N_int32 blockNumber,
    N_int32 iChannel)
{
    GA_driverFuncs  drv;
    int             i,j,k,l,valid,startAddress = blockNumber * length;

    // Initialise DDC communications, and don't bother the check for
    // a valid EDID header (because when an EDID is destroyed, the header
    // is the first to go!).
    channel = iChannel;
    drv.dwSize = sizeof(drv);
    if (!GA_queryFunctions(dc,GA_GET_DRIVERFUNCS,&drv))
        return false;
    sci.dwSize = sizeof(sci);
    if (!GA_queryFunctions(dc,GA_GET_SCIFUNCS,&sci))
        return false;
    if (sci.SCIdetect(&SCI_caps,&SCI_ports) == 0)
        return false;

    // Loop around up to 8 times writing the EDID and verifying that it
    // wrote correctly. Sometimes some bytes get dropped during the
    // writing process.
    sci_begin();
    valid = false;
    for (l = 0; l < 8 && !valid; l++) {
        // Write entire EDID block in 8 byte chunks during the vertical
        // retrace period.
        for (i = 0; i < length; i += 8) {
            // We do the entire sequence multiple times, since it appears that
            // sometimes it does not succeed on the first few attempts.
            for (k = 0; k < 4; k++) {
                drv.WaitVSync();
                if (SendWriteAddress(slaveAddr,startAddress + i))
                    goto Error;

                // Write 8 bytes of EDID data during VSync
                for (j = 0; j < 8; j++) {
                    drv.WaitVSync();
                    if (SendByte(edid[i+j]/*,(j == 7) ? 1 : 0)*/))
                        break;
                    }

                // Send stop to slave
                SendStop();
                }
            }

        // Now read the EDID back and compare it
        if (SendReadAddress(slaveAddr,startAddress))
            goto Error;
        valid = true;
        for (i = 0; i < length; i++) {
            uchar byte;
            if (ReadByte(&byte,(i == length-1) ? 1 : 0))
                goto Error;
            if (byte != edid[i])
                valid = false;
            }

        // Send stop to slave
        SendStop();
        }
    sci_end();
    return true;

Error:
    ErrorRecovery();
    sci_end();
    return false;
}

