/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Hardware specific device detection code.
*
****************************************************************************/

#include "internal.h"

/*--------------------------- Global variables ----------------------------*/

char *DRV_chipsetFilenames[] = {
    "banshee",
    "voodoo3",
    "voodoo4",
    "voodoo5",
    NULL
    };

/* Names of RAMDAC id's. */

char *DRV_DACNames[] = {
    "3dfx Internal 24 bit DAC",
    };

/* Names of chip revision id's. */

char *DRV_clockNames[] = {
    "3dfx Internal Clock",
    };

#include "crtc.c"
#include "../vgadata.c"
#include "data/banshee.c"
#include "data/voodoo3.c"
#include "data/voodoo4.c"
#include "data/voodoo5.c"

GA_chipEntry *DRV_chipData[] = {
    &BANSHEE_chipData,
    &VOODOO3_chipData,
    &VOODOO4_chipData,
    &VOODOO5_chipData,
    };

static ibool hasBeenPOSTed;

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Wakes up the graphics controller if running as a secondary controller.
If the BIOS has POST'ed the card (which happens under Windows 98 with
multi controllers, and under Win95 and Linux with our multi-controller
support), this code is skipped. Otherwise we need to wake up the chip
and progrm the BIOS default values into a number of registers as defined
in CIRRUSPOSTchipsetProfile. The data that we need to program is loaded
dynamically from disk
****************************************************************************/
ibool CHIP_wakeup(int loadInGUI)
{
    if (!loadInGUI && VGA_NOT_ACTIVE()) {
        // Must be running as secondary without the chip having already
        // been woken up. So wake up the chip and remember that we need
        // to POST it later on.
        outpb(0x3C3,0x01);      // Enable VGA
        outpb(0x3C2,0x01);      // Decode 0x3D4 and memory
        outpb(0x3C6,0xFF);      // Unmask DAC
        hasBeenPOSTed = false;
        }
    else {
        // Card has alread been POSTed so we can skip writing the POST
        // register values later on.
        hasBeenPOSTed = true;
        }
    return true;
}

/****************************************************************************
REMARKS:
Configure the global driver device context with the appropriate information
for the detected hardware.
****************************************************************************/
N_int32 DRV_configure(int loadInGUI)
{
    int init0,chips,chipSize,banks;

    //------------------------------------------------------------------
    // Detect the amount of memory on board.
    //------------------------------------------------------------------
    if (DV.chipID >= TDFX_VOODOO4) {
        // TODO: I hacked this for the moment for the Voodoo4!
        init0 = inpd(TDFX_DRAMINIT0);
        if (DV.chipID == TDFX_VOODOO4)
            chips = (init0 & SST_SGRAM_NUM_CHIPSETS) ? 4 : 2;
        else
            chips = (init0 & SST_SGRAM_NUM_CHIPSETS) ? 8 : 4;
        chipSize = 1 << ((init0 & 0x38000000) >> 28);
        banks = (init0 & 0x40000000) ? 4 : 2;
        DC.TotalMemory = chips * chipSize * banks * 1024;
        }
    else {
        if (inpd(TDFX_DRAMINIT1) & SST_MCTL_TYPE_SDRAM)
            DC.TotalMemory = 16384;
        else {
            init0 = inpd(TDFX_DRAMINIT0);
            if (init0 & SST_SGRAM_NUM_CHIPSETS)
                chips = 8;
            else
                chips = 4;
            if ((init0 & SST_SGRAM_TYPE) == SST_SGRAM_TYPE_8MBIT)
                chipSize = 8;
            else
                chipSize = 16;
            DC.TotalMemory = (chips * chipSize) * 128;
            }
        }

    //------------------------------------------------------------------
    // Detect the installed RAMDAC and pixel clock
    //------------------------------------------------------------------
    DV.dacID = DAC_TDFX;
    DV.clockID = CLK_TDFX;
    DV.chip.setPixelClock = TDFX_setPixelClock;

    //------------------------------------------------------------------
    // Configure the capabilities information
    //------------------------------------------------------------------
    DC.Attributes = gaHaveDisplayStart |
                    gaHaveLinearBuffer |
                    gaHaveAccel2D |
                    gaHaveHWCursor |
                    gaHave8BitDAC |
                    gaHaveDoubleScan |
                    gaHaveTripleScan;

    // Integer line drawing in 24bpp produces some artifacts, but it works
    // fine in all other color depths. However we don't have a way to
    // specify that lines are only broken in 24bpp modes, so we set this
    // flag anyway to get past conformance.
    DC.WorkArounds = gaBrokenLines;

    //------------------------------------------------------------------
    // Configure the amount of addressable display memory. The memory
    // layout is as follows:
    //
    //  +------------------------------------+ <- DC.LinearBase
    //  |                                    |
    //  |        X/Y User Video Memory       |
    //  |                                    |
    //  +------------------------------------+ <-- Total-1k (DV.maxAccelLinearMem)
    //  | 1024 bytes - hardware cursor       |
    //  +------------------------------------+ <- DC.TotalMemory * 1024
    //
    //------------------------------------------------------------------
    if (DV.chipID >= TDFX_VOODOO4) {
        if (DC.TotalMemory > 16384) {
            // There appears to be a 4096 byte dead zone at the top of the
            // 32Meg address space.  It interferes with the HW cursor as well
            // as scanline system blits.
            DV.maxAccelLinearMem = DV.NVMemory = DV.VGAMemory = DV.VGA4Memory = DC.TotalMemory - 5;
            }
        else
            DV.maxAccelLinearMem = DV.NVMemory = DV.VGAMemory = DV.VGA4Memory = DC.TotalMemory - 1;
        }
    else
        DV.maxAccelLinearMem = DV.NVMemory = DV.VGAMemory = DV.VGA4Memory = DC.TotalMemory - 1;
    if (DC.TotalMemory > 1024)
        DV.VGA4Memory = 256;
    DV.resScanLines = 2;
    DV.maxAccelY = -1;

    //------------------------------------------------------------------
    // Configure the maximum scanline widths for each color depth
    //------------------------------------------------------------------
    DV.maxBytesPerLineTxt       = TDFX_MAX_OFFSET << 1;
    DV.maxBytesPerLine4         = TDFX_MAX_OFFSET << 1;
    DV.maxBytesPerLine8         = TDFX_MAX_OFFSET << 3;
    DV.maxBytesPerLine24        = TDFX_MAX_OFFSET << 3;

    //------------------------------------------------------------------
    // Configure VGA blank/sync timing masks and CRTC register tables
    //------------------------------------------------------------------
    DV.HBlankEndMask            = 0x7F;
    DV.HSyncEndMask             = 0x3F;
    DV.VBlankEndMask            = 0x1FF;
    DV.VSyncEndMask             = 0x0F;
    DV.HCRTCRegsDesc            = TDFX_HCRTCRegsDesc;
    DV.VCRTCRegsDesc            = TDFX_VCRTCRegsDesc;

    //------------------------------------------------------------------
    // Configure the DPMS and DDC capabilities and support functions
    //------------------------------------------------------------------
    DV.DPMSCaps                 = DPMSNormal;
    DV.SCICaps                  = SCINormal;
    DV.chip.setDPMS             = TDFX_setDPMS;
    DV.sci.SCIbegin             = TDFX_SCIbegin;
    DV.sci.SCIwriteSCL          = TDFX_SCIwriteSCL;
    DV.sci.SCIwriteSDA          = TDFX_SCIwriteSDA;
    DV.sci.SCIreadSCL           = TDFX_SCIreadSCL;
    DV.sci.SCIreadSDA           = TDFX_SCIreadSDA;
    DV.sci.SCIend               = TDFX_SCIend;

    //------------------------------------------------------------------
    // Configure the chipset specific support functions
    //------------------------------------------------------------------
    DV.chip.setChipsetRegisters = TDFX_setChipsetRegisters;
    DV.chip.unlockRegisters     = TDFX_unlockRegisters;
    DV.chip.lockRegisters       = TDFX_lockRegisters;
    DV.chip.get2DAccelInfo      = TDFX_get2DAccelInfo;
    DV.chip.getVideoAccelInfo   = TDFX_getVideoAccelInfo;
    DV.chip.get3DAccelInfo      = TDFX_get3DAccelInfo;
    DV.vga.lockDownTimings      = TDFX_lockDownTimings;
    DV.chip.setCRTCRegisters    = TDFX_setCRTCRegisters;
    DV.chip.setupAccelerator    = TDFX_setupAccelerator;
    DV.chip.setBytesPerLine     = TDFX_setBytesPerLine;
    DV.chip.enableLinear        = TDFX_enableLinear;
    DV.chip.disableLinear       = TDFX_disableLinear;
    DV.chip.getTextModeFlags    = TDFX_getTextModeFlags;
    DV.chip.getModeFlags        = TDFX_getModeFlags;
    DV.chip.roundBytesPerLine   = TDFX_roundBytesPerLine;
    DV.chip.setMemoryClock      = TDFX_setMemoryClock;
    DV.chip.postModeSet         = TDFX_postModeSet;
    DV.chip.preModeSet          = TDFX_preModeSet;
    DV.chip.configChipsetFlags  = TDFX_configChipsetFlags;

    //------------------------------------------------------------------
    // Configure the chipset specific API functions
    //------------------------------------------------------------------
    DV.driver.SetBank               = TDFX_SetBank;
    DV.driver.SetDisplayStart       = TDFX_SetDisplayStart;
    DV.driver.SetStereoDisplayStart = CNOT_SUPPORTED;
    DV.driver.GetDisplayStartStatus = VGA_GetDisplayStartStatus;
    DV.driver.EnableStereoMode      = CNOT_SUPPORTED;
    DV.driver.IsVSync               = VGA_IsVSync;
    DV.driver.WaitVSync             = VGA_WaitVSync;
    DV.driver.SetVSyncWidth         = VGA_SetVSyncWidth;
    DV.driver.GetVSyncWidth         = VGA_GetVSyncWidth;
    DV.driver.SetPaletteData        = VGA_SetPaletteData;
    DV.driver.GetPaletteData        = VGA_GetPaletteData;
    DV.driver.SetGammaCorrectData   = VGA_SetGammaCorrectData;
    DV.driver.GetGammaCorrectData   = VGA_GetGammaCorrectData;
    DV.driver.GetCurrentScanLine    = LCNOT_SUPPORTED;//TDFX_GetCurrentScanLine;

    // TODO: GetCurrentScanLine does not work properly on the Voodoo3 for
    //       some stange reason. It is disabled for the time being.

    //------------------------------------------------------------------
    // Configure the clock, RAMDAC, 2D and 3D modules
    //------------------------------------------------------------------
    CLOCK_init();
    RAMDAC_init();
    ACCEL2D_init();
    TDFX_lockRegisters();
    return nOK;
}

/****************************************************************************
REMARKS:
Returns the mode flags for the specified display mode, which includes flags
such as whether CRTC scaling is enabled etc.
****************************************************************************/
ulong TDFX_getTextModeFlags(
    GA_mode *mode)
{
    // 100x?? text modes and above are bad on Banshee
    if (DV.chipID == TDFX_BANSHEE && mode->xRes > 80)
        return 0;
    else
        return gaText | gaVGA | gaBankedOnly;
}

/****************************************************************************
REMARKS:
Returns the mode flags for the specified display mode, which includes flags
such as whether CRTC scaling is enabled etc.
****************************************************************************/
ulong TDFX_getModeFlags(
    GA_mode *mode)
{
    ulong   flags;

    // Handle restrictions on displayable modes
    if (mode->bits == 15)
        return 0;
    if (mode->xRes > 2032 && DV.chipID <= TDFX_VOODOO3)
        return 0;

    // Handle 4bit modes
    if (mode->bits == 4) {
        if (mode->xRes > 800)
            return 0;
        flags = gaSVGA4 | gaBankedOnly;
        }
    else {
        // Handle all other modes (with multiplexing for high res)
        flags = gaSVGA | gaAccVGA | gaHWCursor;
        if (mode->xRes >= 1152) {
            // In multipledx mode we can only handle 16 pixel granularity
            if (mode->xRes & 0xF)
                return 0;
            flags |= gaMultiplex | gaCRTC_Scale05;
            }
        }

    // Enable double scanning for low resolution modes. Note that we do not
    // support hardware cursors in double scanned modes.
    if (mode->yRes <= 300 || (mode->xRes == 512 && mode->yRes == 384))
        flags |= gaIsDoubleScan;
    return flags;
}

/****************************************************************************
REMARKS:
Configure the 2D acceleration specific information in the mode info block.
****************************************************************************/
void TDFX_get2DAccelInfo(
    GA_modeInfo *gaInfo,
    GA_modeEntry *entry,
    N_int32 virtualX,
    N_int32 virtualY)
{
    if (entry->flags & gaAccMode) {
        gaInfo->Attributes              |= gaHave8bppRGBCursor;
        switch (entry->bits) {
            case 8:
                gaInfo->BitmapStartAlign        = 1;
                gaInfo->BitmapStridePad         = 1;
                break;
            case 15:
            case 16:
                gaInfo->BitmapStartAlign        = 2;
                gaInfo->BitmapStridePad         = 2;
                break;
            case 24:
                gaInfo->BitmapStartAlign        = 3;
                gaInfo->BitmapStridePad         = 3;
                break;
            case 32:
                gaInfo->BitmapStartAlign        = 4;
                gaInfo->BitmapStridePad         = 4;
                break;
            }
        gaInfo->MonoBitmapStartAlign    = 1;
        gaInfo->MonoBitmapStridePad     = 1;
        }
}

/****************************************************************************
REMARKS:
Configure the 2D video overlay specific information in the mode info block.
****************************************************************************/
void TDFX_getVideoAccelInfo(
    GA_modeInfo *gaInfo,
    GA_modeEntry *entry)
{
}

/****************************************************************************
REMARKS:
Configure the 3D acceleration specific information in the mode info block.
****************************************************************************/
void TDFX_get3DAccelInfo(
    GA_modeInfo *gaInfo,
    GA_modeEntry *entry)
{
}

