/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  C support routines for hardware cursor functions.
*
****************************************************************************/

#include "internal.h"
#include "accel.h"

/*--------------------------- Global variables ----------------------------*/

TDFX_CursorState TDFX_cursor = {0};

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
PARAMETERS:
cursor      - Pointer to an GA_monoCursor structure for the cursor image
REMARKS:
This function downloads the specified monochrome cursor definition from the
application into the hardware cursor. The cursor data is passed by the
application in the GA_monoCursor format, which is similar to the Windows
3.1 cursor file format.
****************************************************************************/
void NAPI TDFX_SetMonoCursor(
    GA_monoCursor *cursor)
{
    ulong   offset;
    ulong   *videoMem,buffer[128 * 2],*buf;
    int     i,max,x,y;

    // Compute the address of the cursor data in offscreen video memory
    offset = DV.VGAMemory * 1024L;
    videoMem = (ulong*)((ulong)DC.LinearMem + offset);

    // Move the cursor off-screen
    x = TDFX_cursor.curX;
    y = TDFX_cursor.curY;
    TDFX_SetCursorPos(DV.cntModeInfo.XResolution+100,y);

    // Save the cursor hot spot
    TDFX_cursor.hotX = cursor->HotX;
    TDFX_cursor.hotY = cursor->HotY;
    TDFX_cursor.isDoubleScan = ((DV.cntMode.flags & gaIsDoubleScan) != 0);

    // Hide cursor
    clrregd(TDFX_VIDPROCCFG,(1 << 27) | 0x02);

    // Tell the 3dfx where to find the cursor data
    outpd(TDFX_HWCURPATADDR,offset);

    // Convert the cursor data to the hardware format
    if (TDFX_cursor.isDoubleScan)
        max = 64;
    else
        max = 128;
    for (i = 0,buf = buffer; i < max; i += 2, buf += 4) {
        buf[0] = ~GET_LONG_LE(cursor->ANDMask,i);
        buf[1] = ~GET_LONG_LE(cursor->ANDMask,i+1);
        buf[2] = GET_LONG_LE(cursor->XORMask,i);
        buf[3] = GET_LONG_LE(cursor->XORMask,i+1);
        if (TDFX_cursor.isDoubleScan) {
            // Double up the scanlines in double scanned modes
            buf += 4;
            buf[0] = ~GET_LONG_LE(cursor->ANDMask,i);
            buf[1] = ~GET_LONG_LE(cursor->ANDMask,i+1);
            buf[2] = GET_LONG_LE(cursor->XORMask,i);
            buf[3] = GET_LONG_LE(cursor->XORMask,i+1);
            }
        }

    // We have to download the cursor image in DWORD's to avoid hardware
    // bugs, so we create a temporary buffer and then download that to the
    // hardware directly.
    rep_movsd(videoMem,buffer,sizeof(buffer) / 4);

    // We must wait for a vSync here before changing the cursor location
    // back to the new location to make the shape change transition smooth.
    VGA_WaitVSync();

    // Move back to the original location
    TDFX_SetCursorPos(x,y);

    // Restore hide value to previous setting
    setregd(TDFX_VIDPROCCFG,(TDFX_cursor.visible << 27));
}

/****************************************************************************
PARAMETERS:
foreground  - Foreground color for the cursor
background  - Background color for the cursor

REMARKS:
This function sets the hardware cursor foreground and background colors.
****************************************************************************/
void NAPI TDFX_SetMonoCursorColor(
    GA_palette *foreground,
    GA_palette *background)
{
    outpd(TDFX_HWCURC0,*((ulong*)background));
    outpd(TDFX_HWCURC1,*((ulong*)foreground));
}

/****************************************************************************
PARAMETERS:
x   - X coordinate of the cursor position
y   - Y coordinate of the cursor position

REMARKS:
This function sets the location of the hardware cursor in display
coordinates. This function takes the X and Y coordinates of the new cursor
location. This function will place the cursor so that the hotspot of the
cursor image is located at the specified (X,Y) location, and will correctly
handle special cases where the cursor definition needs to be located off the
edges of the display screen (such as when X=0, Y=0 and HotX,HotY > 0).
****************************************************************************/
N_int32 NAPI TDFX_SetCursorPos(
    N_int32 x,
    N_int32 y)
{
    // Save the current cursor position for later
    TDFX_cursor.curX = x;
    TDFX_cursor.curY = y;

    // Program X position, adjusting to go off left edge of screen
    x = x - TDFX_cursor.hotX + 63;
    y = y - TDFX_cursor.hotY + 63;
    outpd(TDFX_HWCURLOC,(y << 16) | x);
    return true;
}

/****************************************************************************
PARAMETERS:
visible - 1 to show the cursor, 0 to hide the cursor

REMARKS:
This function unconditionally either show or hides the hardware cursor.
****************************************************************************/
void NAPI TDFX_ShowCursor(
    N_int32 visible)
{
    TDFX_cursor.visible = visible;
    modregd(TDFX_VIDPROCCFG,(1 << 27),(TDFX_cursor.visible << 27));
}

