/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement mode initialisation and non-drawing
*               runtime related functions.
*
****************************************************************************/

#include "snap/graphics/driver.h"
#include "internal.h"
#include "accel.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Programs the current bank for banked framebuffer modes.
****************************************************************************/
void NAPI TDFX_SetBank(
    N_int32 bank)
{
}

/****************************************************************************
REMARKS:
Programs the display start address
****************************************************************************/
void NAPI TDFX_SetDisplayStart(
    N_int32 offset,
    N_int32 waitVRT)
{
    DV.cntDisplayStart = offset;
    WaitDEVGA();
    ClearVERTLatch();
    if (DV.cntMode.bits >= 8) {
        outpd(TDFX_STARTADDR,offset);
        SetLowStartAddress8(offset);
        }
    else if (DV.cntMode.bits == 4) {
        // Program the display start for all 4 bit modes
        SetLowStartAddress4(offset);
        }
    else {
        // Program the display start for all VGA text modes
        SetLowStartAddress0(offset);
        }
    if (waitVRT)
        WaitVERTVGA();
}

/****************************************************************************
REMARKS:
Programs the display pitch
****************************************************************************/
N_int32 TDFX_setBytesPerLine(
    N_int32 bytesPerLine)
{
    if (DV.cntMode.bits == 0) {
        int offset = (bytesPerLine + 3) >> 2;
        wrinx(0x3D4,0x13,offset);
        return offset << 2;
        }
    else if (DV.cntMode.bits == 4) {
        int offset = (bytesPerLine + 1) >> 1;
        wrinx(0x3D4,0x13,offset);
        return offset << 1;
        }
    else {
        outpd(TDFX_STRIDE,bytesPerLine);
        return bytesPerLine;
        }
}

/****************************************************************************
REMARKS:
Rounds the display pitch to a useable value that the accelerator supports.
****************************************************************************/
N_int32 TDFX_roundBytesPerLine(
    N_int32 bytesPerLine,
    N_int32 bitsPerPixel)
{
    int pixQWidth,bytesPerPixel;

    if (bitsPerPixel >= 8) {
        switch (bitsPerPixel) {
            case 8:  bytesPerPixel = 1; break;
            case 16: bytesPerPixel = 2; break;
            case 24: bytesPerPixel = 3; break;
            case 32: bytesPerPixel = 4; break;
            }

        // Round the pixel width to a 16 pixel boundary
        pixQWidth = ((bytesPerLine/bytesPerPixel)+63)/64;
        bytesPerLine = (pixQWidth * 64) * bytesPerPixel;
        return bytesPerLine;
        }
    return bytesPerLine;
}

/****************************************************************************
REMARKS:
Programs the default memory clock for the display mode
****************************************************************************/
void TDFX_setMemoryClock(void)
{
    // Program the memory clock
    if (DV.cntMode.bits <= 4)
        DV.pll.setPLLMemoryClock(DV.memoryClockVGA);
    else
        DV.pll.setPLLMemoryClock(DV.opt.memoryClock);
}

/****************************************************************************
REMARKS:
Programs the CRTC registers.
****************************************************************************/
void TDFX_setCRTCRegisters(
    GA_CRTCInfo *crtc)
{
    ulong   videoConfig,vShift,width;

    if (DV.cntMode.flags & gaMultiplex) {
        // If multiplexing is one and the resolution is not an exact
        // multiple of 16, we have to round it up.
        if (DV.cntModeInfo.XResolution % 16) {
            DV.cntModeInfo.XResolution += 8;
            DV.cntMode.xRes += 8;
            }
        }
    VGA_setCRTCRegisters(crtc);

    // VIDSCREENSIZE needs the correct xResolution
    if (DV.cntMode.flags & gaMultiplex)
        width = (DV.hCRTC.hDisp+1)<<4;
    else
        width = (DV.hCRTC.hDisp+1)<<3;
    if (DV.cntMode.flags & gaSVGA) {
        if (crtc->Flags & gaDoubleScan) {
            videoConfig = SST_HALF_MODE;
            vShift = 13;
            }
        else {
            videoConfig = 0;
            vShift = 12;
            }
        outpd(TDFX_VIDSCREENSIZE,(DV.cntMode.yRes<<vShift)|width); // DV.cntMode.xRes);
        modregd(TDFX_VIDPROCCFG,SST_HALF_MODE,videoConfig);
        }
}

/****************************************************************************
REMARKS:
Locks down the VGA compatible timings. This does nothing by default, but
some chips have restrictions on the CRTC timings that must be used so this
function is used to override the default values.
****************************************************************************/
void TDFX_lockDownTimings(
    VGA_hCRTC *h,
    VGA_vCRTC *v,
    GA_CRTCInfo *crtc)
{
}

/****************************************************************************
PARAMETERS:
Return the current scanline that the CRTC is displaying
****************************************************************************/
N_int32 NAPI TDFX_GetCurrentScanLine(void)
{
    return inpd(TDFX_VIDCURRENTLINE) & 0x7FF;
}

/****************************************************************************
REMARKS:
Programs the pixel clock
****************************************************************************/
void TDFX_setPixelClock(N_fix32 clock)
{
    int entry = DV.pll.setPLLPixelClock(clock);

    wrinx(0x3C4,0x00,0x01);
    outpb(0x3C2,(inpb(0x3CC) & 0xF3) | (entry << 2));
    wrinx(0x3C4,0x00,0x03);
}

/****************************************************************************
REMARKS:
Enables the hardware linear framebuffer mode (assumes the modes are
initially configured for banked framebuffer operation).
****************************************************************************/
void TDFX_enableLinear(void)
{
}

/****************************************************************************
REMARKS:
Disables the hardware linear framebuffer mode.
****************************************************************************/
void TDFX_disableLinear(void)
{
}

/****************************************************************************
REMARKS:
Does any additional programming of registers to complete a mode set.
****************************************************************************/
void TDFX_preModeSet(void)
{
    VGA_preModeSet();
}

/****************************************************************************
REMARKS:
Does any additional programming of registers to complete a mode set.
****************************************************************************/
void TDFX_postModeSet(void)
{
    VGA_postModeSet();
}

/****************************************************************************
PARAMETERS:
Function called *only* when the driver is initialising a mode with the
gaPartialModeSet flag set, which assumes someone else (ie: an IHV driver) has
set the display mode. What we need to do is determine anything special
about the environment and configure our internal driver as needed.
****************************************************************************/
void TDFX_configChipsetFlags(void)
{
    DV.dac.setDACRegisters();
}

/****************************************************************************
REMARKS:
Program the chipset specific registers
****************************************************************************/
void TDFX_setChipsetRegisters(void)
{
    DRV_setChipsetRegisters();
}

