/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the code to analyse all display
*               modes for existing DOS and Win32 display drivers and
*               dump them to binary files to later be converted into the
*               real register values.
*
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "pmapi.h"
#include "snap/graphics/driver.h"
#include "graphics/pciinit.h"
#include "graphics/analyse.h"
#include "snap/internal.h"

#ifndef __ANALYSE__
#error This module must be compiled with -D__ANALYSE__!!
#endif

/*--------------------------- Global variables ----------------------------*/

#define ga80x50xText    -1

static ulong        *CHIP_profile;
static FileHeader   hdr;
static ModeEntry    modeTable[GA_MAX_MODES+1];
static int          DACSize;
static uchar        *vgaBuffer;
static uchar        *buffer;
static uchar        *profile;
static uchar        *comments;
static uchar        *vgaDACBuffer = NULL;
static uchar        *colorDepth4DACBuffer = NULL;
static uchar        *colorDepth8DACBuffer = NULL;
static uchar        *colorDepth15DACBuffer = NULL;
static uchar        *colorDepth16DACBuffer = NULL;
static uchar        *colorDepth24DACBuffer = NULL;
static uchar        *colorDepth32DACBuffer = NULL;
static char         filename[_MAX_PATH];

#include "profile.c"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
PARAMETERS:
fmt - Format string
... - Remainder of parameters

REMARKS:
Logs a text string entry to the log file with the specified format.
****************************************************************************/
void log(
    const char *fmt,
    ... )
{
    va_list argptr;
    FILE    *f;

    if ((f = fopen("analyse.log","a+t")) == NULL)
        exit(1);
    va_start(argptr, fmt);
    vfprintf(f, fmt, argptr);
    va_end(argptr);
    fclose(f);
}

/****************************************************************************
PARAMETERS:
profile     - Pointer to the chipset profile table to measure
profileSize - Place to store size of profile table
commentSize - Place to store size of comment data

REMARKS:
This function measures the size of the chipset profile table to be written
to disk. The profile table is written as a single table without any
references (references are expanded inline effectively), and the comment
strings are collected into a single large buffer. The address of the
comment strings is converted into offsets into the comment string buffer
when written to disk so they can be accessed effciently when read by the
optimise program.
****************************************************************************/
static void getProfileSize(
    ulong *profile,
    long *profileSize,
    long *commentSize)
{
    while (*profile != REG_TERMINATOR) {
        if (regIsReference(*profile)) {
            // Referencing another table, so measure size of referenced
            // table with a recursive function call.
            getProfileSize((ulong*)(*(profile+1)),profileSize,commentSize);
            }
        else {
            // Defining a register value. Each entry is 8 bytes in size,
            // 4 bytes for the register definition itself and 4 bytes for
            // the offset of the comment in the comment string table.
            char *c = (char*)(*(profile+1));
            *profileSize += 8;
            *commentSize += strlen(c)+1;
            }
        profile += 2;
        }
}

/****************************************************************************
PARAMETERS:
profile     - Pointer to the chipset profile table to measure
buffer      - Buffer to save the profile table into

RETURNS:
Pointer to the next space in the buffer

REMARKS:
This function prepares the profile table to be written to disk by expanding
all table references inline and coverting comment string pointers to offsets
within the comment table.
****************************************************************************/
static void prepareProfileData(
    ulong *profile,
    uchar *buffer,
    uchar *comments,
    int *profileLen,
    int *commentLen)
{
    while (*profile != REG_TERMINATOR) {
        if (regIsReference(*profile)) {
            // Referencing another table, so measure size of referenced
            // table with a recursive function call.
            prepareProfileData((ulong*)(*(profile+1)),buffer,comments,profileLen,commentLen);
            }
        else {
            // Defining a register value. Each entry is 8 bytes in size,
            // 4 bytes for the register definition itself and 4 bytes for
            // the offset of the comment in the comment string table.
            char *c = (char*)(*(profile+1));
            int len = strlen(c)+1;
            *((ulong*)&buffer[*profileLen+0]) = *profile;
            *((ulong*)&buffer[*profileLen+4]) = *commentLen;
            memcpy(comments+*commentLen,c,len);
            *profileLen += 8;
            *commentLen += len;
            }
        profile += 2;
        }
}

/****************************************************************************
REMARKS:
Initialise the internal driver ready for analysing the hardware.
****************************************************************************/
static void InitDriver(void)
{
    FILE            *f;
    int             size;
    PCIDeviceInfo   *pciInfo;
    GA_devCtx       *dc;
    GA_monitor      monitor;

    // Enumerate the PCI bus and find device info for primary display
    _PCI_enumerateDevices();
    pciInfo = _PCI_findDevice(0);

    // Run the device detection code
    if ((dc = GA_initDriver(false,PM_getSNAPPath(),0,pciInfo,&monitor,&_N_imports,&_PM_imports,&_DRV_imports,0,NULL,NULL,NULL,NULL,0)) == NULL)
        PM_fatalError("Graphics hardware not detected!");
    if (_GA_initInternal(dc,0) != nOK)
        PM_fatalError("Unable to initialise driver internals!");

    // Allocate memory to save the hardware state
    CHIP_profile = DRV_regProfile[DV.chipID];
    hdr.size = size = DRV_getStateSize(CHIP_profile,true);
    hdr.numModes = 0;
    hdr.profileSize = 0;
    hdr.commentSize = 0;
    getProfileSize(CHIP_profile,&hdr.profileSize,&hdr.commentSize);
    hdr.profileSize += 4;
    modeTable[0].xRes = -1;
    vgaBuffer = malloc(hdr.size);
    buffer = malloc(hdr.size * GA_MAX_MODES);
    profile = malloc(hdr.profileSize);
    comments = malloc(hdr.commentSize);
    if (!buffer || !profile || !comments)
        PM_fatalError("Out of memory!");
    if (DV.dacProfile)
        DACSize = DRV_getDACStateSize(DV.dacProfile,true);

    // Read the existing data file so we skip modes that have already
    // been analysed.
    strcat(filename,DRV_chipsetFilenames[DV.chipID]);
    strcat(filename,".dac");
    if ((f = fopen(filename,"rb")) != NULL) {
        fread(&hdr,1,sizeof(hdr),f);
        if (hdr.size != size)
            PM_fatalError("Corrupt data file!");
        fread(modeTable,1,hdr.numModes * sizeof(ModeEntry),f);
        fread(buffer,1,hdr.size * hdr.numModes,f);
        fclose(f);
        }
}

/****************************************************************************
PARAMETERS:
xRes    - X resolution for the mode to be analysed
yRes    - Y resolution for the mode to be analysed
bits    - Bits per pixel for the mode to be analysed

RETURNS:
True if the mode already exists, false if not.

REMARKS:
This function searches the table looking to see if a mode has already
been analysed.
****************************************************************************/
static int FindMode(
    int xRes,
    int yRes,
    int bits,
    int flags)
{
    int i;

    // First check if a compatible mode has already been analysed
    for (i = 0; i < hdr.numModes; i++) {
        if (modeTable[i].xRes == xRes &&
            modeTable[i].yRes == yRes &&
            modeTable[i].bits == bits &&
            (modeTable[i].flags & gaVGACore) == flags)
            return true;
        }
    return false;
}

/****************************************************************************
REMARKS:
This function performs the analysis on the VGA text modes to analyse all
the extended register values for when the card is in standard VGA modes.
****************************************************************************/
static void AnalyseVGA(void)
{
    // Analyse the standard VGA register values
    DV.chip.unlockRegisters();
    DV.chip.blankDisplay();

    vgaDACBuffer = malloc(DACSize);
    DRV_saveDACState(DV.dacProfile,vgaDACBuffer,true);

    DV.chip.unblankDisplay();
    DV.chip.lockRegisters();
}

/****************************************************************************
PARAMETERS:
pitch   - Pitch for the display mode
xRes    - X resolution for the mode to be analysed
yRes    - Y resolution for the mode to be analysed
bits    - Bits per pixel for the mode to be analysed
flags   - Flags for the display mode

REMARKS:
This function performs the analysis on the display mode and stores the
analysed data in the analysis tables.
****************************************************************************/
static void AnalyseMode(
    int pitch,
    int xRes,
    int yRes,
    int bits,
    ulong flags)
{

    // Enable double scan for low res modes by default
    if (xRes >= 320 && xRes <= 512)
        flags |= gaIsDoubleScan;

    // Store the information about the mode in the mode table
    modeTable[hdr.numModes].pitch = pitch;
    modeTable[hdr.numModes].xRes = xRes;
    modeTable[hdr.numModes].yRes = yRes;
    modeTable[hdr.numModes].bits = bits;
    modeTable[hdr.numModes].flags = flags;

    // Analyse the mode
    DV.chip.unlockRegisters();
    DV.chip.blankDisplay();
    switch (bits) {
        case 4:
            if (xRes == 800 && yRes == 600) {
                colorDepth4DACBuffer = malloc(DACSize);
                DRV_saveDACState(DV.dacProfile,colorDepth4DACBuffer,true);
                }
            break;
        case 8:
            if (xRes == 640 && yRes == 480) {
                colorDepth8DACBuffer = malloc(DACSize);
                DRV_saveDACState(DV.dacProfile,colorDepth8DACBuffer,true);
                }
            break;
        case 15:
            if (xRes == 640 && yRes == 480) {
                colorDepth15DACBuffer = malloc(DACSize);
                DRV_saveDACState(DV.dacProfile,colorDepth15DACBuffer,true);
                }
            break;
        case 16:
            if (xRes == 640 && yRes == 480) {
                colorDepth16DACBuffer = malloc(DACSize);
                DRV_saveDACState(DV.dacProfile,colorDepth16DACBuffer,true);
                }
            break;
        case 24:
            if (xRes == 640 && yRes == 480) {
                colorDepth24DACBuffer = malloc(DACSize);
                DRV_saveDACState(DV.dacProfile,colorDepth24DACBuffer,true);
                }
            break;
        case 32:
            if (xRes == 640 && yRes == 480) {
                colorDepth32DACBuffer = malloc(DACSize);
                DRV_saveDACState(DV.dacProfile,colorDepth32DACBuffer,true);
                }
            break;
        }
    DV.chip.unblankDisplay();
    DV.chip.lockRegisters();
}

/****************************************************************************
REMARKS:
Writes the analysed DAC data to disk.
****************************************************************************/
uchar *WriteDACData(
    FILE *f,
    ulong *profile,
    uchar *buffer)
{
    while (*profile != DAC_TERMINATOR) {
        char *comment = (uchar*)profile[1];

        if (dacIsReference(*profile)) {
            // Referencing another table so restore the state with a recursive
            // function call.
            buffer = WriteDACData(f,profile+1,buffer);
            profile += 2;
            }
        else {
            fprintf(f,"\t_BYTE(0x%02X),\t\t// %s\n",*buffer++,comment);
            profile += 2;
            }
        }
    return buffer;
}

/****************************************************************************
REMARKS:
Writes the analysed color depth DAC data to disk.
****************************************************************************/
void WriteColorDepthDACData(
    FILE *f,
    uchar *buffer,
    int bits)
{
    if (buffer) {
        fprintf(f,"static uchar RAMDAC_colorDepth%dData[] = {\n", bits);
        WriteDACData(f,DV.dacProfile,buffer);
        fprintf(f,"\t};\n\n");
        }
}

/****************************************************************************
REMARKS:
Writes the analysed data to disk.
****************************************************************************/
static void writeAnalysedData(void)
{
    FILE    *f;
    int     profileLen,commentLen;

    // Prepare the profile table for writing
    profileLen = 0;
    commentLen = 0;
    prepareProfileData(CHIP_profile,profile,comments,&profileLen,&commentLen);
    *((ulong*)&profile[profileLen]) = REG_TERMINATOR;

    // Write the data to disk
    if (vgaDACBuffer) {
        if ((f = fopen(filename, "wt")) == NULL)
            PM_fatalError("Unable to open data file for output!");
        fprintf(f,"static uchar RAMDAC_VGAData[] = {\n");
        WriteDACData(f,DV.dacProfile,vgaDACBuffer);
        fprintf(f,"\t};\n\n");
        WriteColorDepthDACData(f,colorDepth4DACBuffer,4);
        WriteColorDepthDACData(f,colorDepth8DACBuffer,8);
        WriteColorDepthDACData(f,colorDepth15DACBuffer,15);
        WriteColorDepthDACData(f,colorDepth16DACBuffer,16);
        WriteColorDepthDACData(f,colorDepth24DACBuffer,24);
        WriteColorDepthDACData(f,colorDepth32DACBuffer,32);
        fclose(f);
        }
}

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+ DOS32 version of the analysis code. This version will enumerate all
+ available display modes that our existing VBE/AF drivers support, or
+ that the VBE/Core BIOS on the card supports. This will pick up 4bpp
+ SVGA modes which are not picked up by the Win32 version.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#include "vesavbe.h"
#include "vbeaf.h"

extern int getch(void);

/****************************************************************************
PARAMETERS:
mode    - VGA mode to set

REMARKS:
Calls the VGA BIOS to set a standard VGA display mode.
****************************************************************************/
static void VGA_setMode(int mode)
{
    RMREGS  regs;
    regs.x.ax = mode;
    if (mode == ga80x50xText)
        regs.x.ax = 0x03;
    PM_int86(0x10,&regs,&regs);

    /* Enable 80x50 line text mode if desired */
    if (mode == ga80x50xText) {
        regs.x.ax = 0x1112;
        regs.x.bx = 0;
        PM_int86(0x10,&regs,&regs);
        }
}

/****************************************************************************
REMARKS:
Analyses all the VBE/AF display modes
****************************************************************************/
int AnalyseVBEAFModes(int newModes)
{
    AF_int16        *p;
    AF_devCtx       *dc;
    AF_modeInfo     modeInfo;
    AF_int32        bytesPerLine;

    // Analyse all VBE/AF display modes first
    if ((dc = AF_loadDriver("",false)) == NULL)
        return newModes;
    for (p = dc->AvailableModes; *p != -1; p++) {
        AF_DPMI_regs regs;
        AF_int16 mode = *p,bits;
        ulong flags = DV.defaultModeFlags;
        AF_getVideoModeInfo(dc,mode,&modeInfo);

// HACK: Disable 32bpp modes that are broken in VBEAF
//      if (modeInfo.BitsPerPixel == 32)
//          continue;

        if (modeInfo.BitsPerPixel == 4) {
            if (!((modeInfo.XResolution == 800) && (modeInfo.YResolution == 600)))
                continue;
            }
        else {
            if (!((modeInfo.XResolution == 640) && (modeInfo.YResolution == 480)))
                continue;
            }

        // Call our OEM extension to extract the clock and CRTC scaling
        regs.ecx = mode;
        regs.ebx = 0x200;
        AF_OEMExt(dc,&regs);
        flags |= (regs.ebx << gaCLK_ScaleShift);
        flags |= (regs.ecx << gaCRTC_ScaleShift);

        // Determine mode flags from VBE/AF mode information
        if (!(modeInfo.Attributes & afHaveBankedBuffer)) {
            mode |= vbeLinearBuffer;
            flags |= gaLinearOnly;
            }
        if (!(modeInfo.Attributes & afHaveLinearBuffer)) {
            flags |= gaBankedOnly;
            }
        bits = modeInfo.BitsPerPixel;
        if (bits < 8) {
            flags &= ~(gaAccMode | gaStereo | gaHWCursor | gaVideo | ga3D | gaLinearOnly);
            flags |= gaBankedOnly;
            }
        if (modeInfo.Attributes & afHaveStereo)
            flags |= gaStereo;
        if (modeInfo.Attributes & afNonVGAMode)
            flags |= gaNonVGA;

        // Check for existing mode in tables
        if (FindMode(modeInfo.XResolution,modeInfo.YResolution,bits,0))
            continue;

        // Set the mode
        bytesPerLine = -1;
        AF_setVideoMode(dc,mode | afDontClear,-1,-1,&bytesPerLine,1,NULL);

        // Analyse the display mode
        AnalyseMode(bytesPerLine,modeInfo.XResolution,modeInfo.YResolution,bits,flags);
        hdr.numModes++;
        newModes++;

        // Restore text mode
        AF_restoreTextMode(dc);
        }
    return newModes;
}

/****************************************************************************
REMARKS:
Analyses all the VBE/Core display modes
****************************************************************************/
int AnalyseVBEModes(int newModes)
{
    VBE_vgaInfo     vgaInfo;
    VBE_modeInfo    modeInfo;
    ushort          *p;

    // Analyse all VBE display modes
    if (VBE_detect(&vgaInfo) == 0)
        return newModes;
    for (p = vgaInfo.VideoModePtr; *p != 0xFFFF; p++) {
        ushort mode = *p,bits;
        ulong flags = DV.defaultModeFlags;
        if (!VBE_getModeInfo(mode,&modeInfo))
            continue;
        if (!(modeInfo.ModeAttributes & vbeMdAvailable))
            continue;
        if (!(modeInfo.ModeAttributes & vbeMdGraphMode))
            continue;

        if (modeInfo.BitsPerPixel == 4) {
            if (!((modeInfo.XResolution == 800) && (modeInfo.YResolution == 600)))
                continue;
            }
        else {
            if (!((modeInfo.XResolution == 640) && (modeInfo.YResolution == 480)))
                continue;
            }

        if (modeInfo.ModeAttributes & vbeMdNonBanked) {
            mode |= vbeLinearBuffer;
            flags |= gaLinearOnly;
            }
        if (!(modeInfo.ModeAttributes & vbeMdLinear)) {
            flags |= gaBankedOnly;
            }
        bits = (modeInfo.ModeAttributes & vbeMdGraphMode) ? modeInfo.BitsPerPixel : 0;
        if (bits < 8) {
            flags &= ~(gaAccMode | gaStereo | gaHWCursor | gaVideo | ga3D | gaLinearOnly);
            flags |= gaBankedOnly;
            }
        if (modeInfo.ModeAttributes & vbeMdStereo)
            flags |= gaStereo;

        // Check for existing mode in tables
        if (FindMode(modeInfo.XResolution,modeInfo.YResolution,bits,0))
            continue;

        // Set the mode
        VBE_setVideoMode(mode | vbeDontClear);

        // Analyse the display mode
        AnalyseMode(modeInfo.BytesPerScanLine,modeInfo.XResolution,modeInfo.YResolution,bits,flags);
        hdr.numModes++;
        newModes++;
        }
    return newModes;
}

int main(void)
{
    int newModes;

    // Initialise the internal driver
    ___drv_os_type = PM_getOSType();
    VBE_init();
    InitDriver();

    // Display detected hardware information
    printf("Detected the following graphics hardware:\n");
    printf("\n");
    printf("Graphics Chip:  %s %s ",
        DRV_configInfo.ManufacturerName,
        DRV_configInfo.ChipsetName);
    if ((DV.dc.TotalMemory % 1024) == 0)
        printf("with %d MB\n", DV.dc.TotalMemory / 1024);
    else
        printf("with %d KB\n", DV.dc.TotalMemory);
    printf("RAM DAC:        %s\n", DRV_configInfo.DACName);
    printf("Clock Chip:     %s\n", DRV_configInfo.ClockName);
    printf("Driver Version: %s\n", DRV_configInfo.VersionInfo);
    printf("Driver Data:    %s\n", DRV_configInfo.BuildDate);
    printf("\n");
    printf("If the above is correct press any key to analyse, or ESC to exit ...");
    fflush(stdout);
    if (getch() != '\x1B') {
        // Analyse the standard VGA mode extended register values
        AnalyseVGA();

        // Analyse all display modes
        newModes = AnalyseVBEAFModes(0);
        newModes = AnalyseVBEModes(newModes);
        modeTable[hdr.numModes].xRes = -1;

        VGA_setMode(ga80x50xText);
        printf("Analysed the RAMDAC for a total of %d new modes (%d total)\n", newModes, hdr.numModes);

        // Write the data to disk
        writeAnalysedData();
        }
    return 0;
}
