/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the generic SNAP Graphics
*               Architecture API functions on top of DirectDraw for Win32.
*
****************************************************************************/

#include "internal.h"

/*--------------------------- Global variables ----------------------------*/

#define MAX_REFRESH_RATES   20
static N_int32      DRV_refreshList[MAX_REFRESH_RATES+1];
static ibool        DD_cursorHidden = false;

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
PARAMETERS:
mode        - Mode number to find internal mode entry for

RETURNS:
Pointer to the mode entry structure for the mode, NULL if not found.

REMARKS:
This function searches the table of valid mode information for the
specific mode, and returns a pointer to the entry in the table.
****************************************************************************/
GA_mdEntry *DD_validateMode(
    N_int32 mode)
{
    mode &= 0xFF;
    if (mode < DD_numModes)
        return &DD_modeList[mode];
    return NULL;
}

/****************************************************************************
PARAMETERS:
xRes            - Physical pixel X resolution of the display mode
yRes            - Physical line Y resolution of the display mode
virtualX        - Virtual X resolution of the display mode
virtualY        - Virtual Y resolution of the display mode
bitsPerPixel    - Color depth of the display mode
modeInfo        - Structure to fill in with the mode information
outputDevice    - Output device to test mode information for
outputHead      - Output head to test mode information for

RETURNS:
0 on success, -1 on failure

REMARKS:
Obtains information about a specific SNAP display mode.
****************************************************************************/
N_int32 NAPI DD_GetCustomVideoModeInfoExt(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 virtualX,
    N_int32 virtualY,
    N_int32 bitsPerPixel,
    GA_modeInfo *modeInfo,
    N_int32 outputDevice,
    N_int32 outputHead)
{
    // Not supported via DirectDraw
    (void)xRes;
    (void)yRes;
    (void)virtualX;
    (void)virtualY;
    (void)bitsPerPixel;
    (void)modeInfo;
    (void)outputDevice;
    (void)outputHead;
    return -1;
}

/****************************************************************************
PARAMETERS:
xRes            - Physical pixel X resolution of the display mode
yRes            - Physical line Y resolution of the display mode
virtualX        - Virtual X resolution of the display mode
virtualY        - Virtual Y resolution of the display mode
bitsPerPixel    - Color depth of the display mode
modeInfo        - Structure to fill in with the mode information

RETURNS:
0 on success, -1 on failure

REMARKS:
Obtains information about a specific SNAP display mode. We simply call
the loaded DirectDraw driver to determine the mode information, and then
translate this information into the format used by SNAP.
****************************************************************************/
N_int32 NAPI DD_GetCustomVideoModeInfo(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 virtualX,
    N_int32 virtualY,
    N_int32 bitsPerPixel,
    GA_modeInfo *modeInfo)
{
    // Not supported via DirectDraw
    (void)xRes;
    (void)yRes;
    (void)virtualX;
    (void)virtualY;
    (void)bitsPerPixel;
    (void)modeInfo;
    return -1;
}

/****************************************************************************
PARAMETERS:
mode            - SNAP display mode to get info for
modeInfo        - Structure to fill in with the mode information
outputDevice    - Output device to get mode info for
outputHead      - Output head to get mode info for

RETURNS:
0 on success, -1 on failure

REMARKS:
Obtains information about a specific SNAP display mode. We simply call
the loaded DirectDraw driver to determine the mode information, and then
translate this information into the format used by SNAP.
****************************************************************************/
N_int32 NAPI DD_GetVideoModeInfoExt(
    N_uint32 gaMode,
    GA_modeInfo *modeInfo,
    N_int32 outputDevice,
    N_int32 outputHead)
{
    int             i;
    GA_mdEntry      *md;
    GA_modeInfo     gaInfo;
    ulong           memSize = DD_totalMemory;

    // Validate that the mode is actually available
    if (outputDevice != gaOUTPUT_CRT || (outputHead != gaActiveHeadClone && outputHead != gaActiveHeadPrimary))
        return -1;
    if ((md = DD_validateMode(gaMode)) == NULL)
        return -1;

    // Copy the list of available refresh rates. Note that on Windows 9x it
    // appears this list is always empty. Probably some strange issue with
    // DirectX.
    for (i = 0; md->RefreshList[i] != -1; i++)
        DRV_refreshList[i] = md->RefreshList[i];
    DRV_refreshList[i] = -1;

    // Now fill in the mode information block
    memset(&gaInfo,0,sizeof(gaInfo));
    gaInfo.dwSize                   = sizeof(gaInfo);
    if (gaMode == 0x100)
        gaInfo.Attributes           = gaIsGUIDesktop | gaHaveLinearBuffer;
    else
        gaInfo.Attributes           = DC.Attributes;
    gaInfo.XResolution              = md->xres;
    gaInfo.YResolution              = md->yres;
    gaInfo.XCharSize                = 8;
    gaInfo.YCharSize                = 8;
    gaInfo.BytesPerScanLine         = md->bytesPerLine;
    gaInfo.MaxBytesPerScanLine      = md->bytesPerLine;
    gaInfo.MaxScanLineWidth         = md->xres;
    gaInfo.MaxPixelClock            = 0;
    gaInfo.MaxBuffers               = (memSize / gaInfo.BytesPerScanLine) / md->yres;
    gaInfo.MaxScanLines             = gaInfo.MaxBuffers * md->yres;
    gaInfo.MaxLinearOffset          = gaInfo.MaxScanLines * gaInfo.BytesPerScanLine;
    gaInfo.BitsPerPixel             = md->bits;
    gaInfo.PixelFormat              = DD_pixelFormats[md->pixelFormat];
    gaInfo.BitmapStartAlign         = 1;
    gaInfo.BitmapStridePad          = 1;
    gaInfo.RefreshRateList          = DRV_refreshList;
    gaInfo.DefaultRefreshRate       = 60;
    gaInfo.LinearSize               = 0;
    gaInfo.LinearBasePtr            = 0xFFFFFFFF;

    // If DirectDraw reported 0 memory available, this means that for some reason
    // the DirectDraw support is running in fallback mode (badly installed driver,
    // Windows NT 4.0 with no DirectDraw etc). To make this work, we only report
    // a single page of dispaly memory so all buffering will be done in system
    // memory, so at least this configuration will work.
    if (DD_totalMemory == 0) {
        gaInfo.MaxBuffers           = 1;
        gaInfo.MaxScanLines         = md->yres;
        gaInfo.MaxLinearOffset      = gaInfo.MaxScanLines * gaInfo.BytesPerScanLine;
        }

    // Many DirectDraw drivers will mis-report the scanline width for
    // display modes during the enumeration process and we have to 'fix' this
    // later when we set the actual modes. Unfortunately this means that our
    // maximum flip buffers computation can be slightly off. Since more than
    // 3 flip buffers is not much use in DirectDraw anyway we cap the
    // maximum value to 3 to avoid problems.
    gaInfo.MaxBuffers               = MIN(3,gaInfo.MaxBuffers);

    // Return the structure to the caller
    COPY_STRUCTURE(modeInfo,&gaInfo);
    return 0;
}

/****************************************************************************
PARAMETERS:
mode        - SNAP display mode to get info for
modeInfo    - Structure to fill in with the mode information

RETURNS:
0 on success, -1 on failure

REMARKS:
Obtains information about a specific SNAP display mode. We simply call
the loaded DirectDraw driver to determine the mode information, and then
translate this information into the format used by SNAP.
****************************************************************************/
N_int32 NAPI DD_GetVideoModeInfo(
    N_uint32 gaMode,
    GA_modeInfo *modeInfo)
{
    return DD_GetVideoModeInfoExt(gaMode,modeInfo,DV.outputDevice[0],DV.init.GetActiveHead());
}

/****************************************************************************
DESCRIPTION:
Sets a custom display mode.

PARAMETERS:
xRes            - X resolution for custom display mode
yRes            - Y resolution for custom display mode
bitsPerPixel    - Color depth for custom display mode (0 = text mode)
virtualX        - Requested virtual display X resolution (-1 to use default)
virtualY        - Returns the addressable display memory Y resolution
bytesPerLine    - Returns the scanline width for the mode
maxMem          - Returns the maximum addressable display memory limit
crtc            - CRTC information block containing CRTC timings (required)

RETURNS:
0 on success, -1 on failure

REMARKS:
This function is used to initialize a custom display mode using the
specified X and Y resolution, color depth and CRTC timings.
****************************************************************************/
N_int32 NAPI DD_SetCustomVideoMode(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bitsPerPixel,
    N_uint32 flags,
    N_int32 *virtualX,
    N_int32 *virtualY,
    N_int32 *bytesPerLine,
    N_int32 *maxMem,
    GA_CRTCInfo *crtc)
{
    // Not supported by DirectDraw
    (void)xRes;
    (void)yRes;
    (void)bitsPerPixel;
    (void)flags;
    (void)virtualX;
    (void)virtualY;
    (void)bytesPerLine;
    (void)maxMem;
    (void)crtc;
    return -1;
}

/****************************************************************************
DESCRIPTION:
Sets a specified display mode.

PARAMETERS:
mode            - Mode number to set, including flags
virtualX        - Requesyed virtual display X resolution (-1 to use default)
virtualY        - Returns the addressable display memory Y resolution
bytesPerLine    - Returns the scanline width for the mode
maxMem          - Returns the maximum addressable display memory limit
crtc            - CRTC information block if afRefreshCtrl specified

RETURNS:
0 on success, -1 on failure

REMARKS:
This function is used to initialize a specified display mode. We do this
by calling the DirectDraw driver to initialsie the mode, and translating as
necessary any parameters and return values.
****************************************************************************/
N_int32 NAPI DD_SetVideoMode(
    N_uint32 gaMode,
    N_int32 *virtualX,
    N_int32 *virtualY,
    N_int32 *bytesPerLine,
    N_int32 *maxMem,
    N_int32 refreshRate,
    GA_CRTCInfo *crtc)
{
    HRESULT                 ret;
    GA_mdEntry              *md;
    DDSURFACEDESC2          ddsd;
    int                     i,bits;
    HWND                    hwndConsole = PM_getDirectDrawWindow();
    POINT                   pt;

    // Fail for cases not supported by DirectDraw
    if (gaMode & (gaPartialModeSet | gaRefreshCtrl))
        return -1;
    if (*virtualX != -1 || *virtualY != -1 || *bytesPerLine != -1)
        return -1;

    // Validate that the mode is actually available
    if ((md = DD_validateMode(gaMode)) == NULL)
        return -1;

    // Get the current mode information for the mode
    DV.cntModeInfo.dwSize = sizeof(DV.cntModeInfo);
    if (DD_GetVideoModeInfo(gaMode & gaModeMask,&DV.cntModeInfo) != 0)
        return -1;

    // Release extra DD/D3D objects we created with primary
    for (i = 0; i < DD_numPixelFormats; i++)
        SAFE_RELEASE(DD_lpTexture[i]);
    SAFE_RELEASE(DD_lpD3DDevice);

    // Destroy the primary surface, detach and destroy palette. Note that we
    // *must* release the palette or we get crashes in the NVIDIA drivers.
    SAFE_RELEASE(DD_lpPalette);
    if (DD_lpPrimary) {
        IDirectDrawSurface7_SetPalette(DD_lpPrimary,NULL);
        SAFE_RELEASE(DD_lpPrimary);
        DD_activeBuf = NULL;
        DD_lpActive = NULL;
        DD_lpVisible = NULL;
        }

    // Set the display mode
    if (DV.cntModeInfo.Attributes & gaIsGUIDesktop) {
        _CHK(N_log("DD_SetVideoMode: Restoring original desktop mode, hwndConsole = 0x%08X\n", hwndConsole));

        // Restore the original desktop display mode
        IDirectDraw7_SetCooperativeLevel(DD_lpDD,hwndConsole,DDSCL_NORMAL);
        IDirectDraw7_RestoreDisplayMode(DD_lpDD);
        if (DD_cursorHidden) {
            pShowCursor(true);
            DD_cursorHidden = false;
            }
        DD_clearMessageQueue();
        }
    else {
        if (!(gaMode & gaWindowedMode)) {
            _CHK(N_log("DD_SetVideoMode: Setting fullscreen exclusive mode, hwndConsole = 0x%08X\n", hwndConsole));

            // Show the fullscreen window and re-size it to the screen dimensions
            pSetWindowPos(hwndConsole,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE | SWP_NOSIZE);
            pShowWindow(hwndConsole,SW_SHOWMAXIMIZED);

            // Pump messages to ensure that the window re-sizing is completely handled
            DD_clearMessageQueue();

            // Set the DirectDraw Cooperative level
            if ((ret = IDirectDraw7_SetCooperativeLevel(DD_lpDD,hwndConsole,DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN)) != DD_OK) {
                N_log("DD_SetVideoMode: IDirectDraw7_SetCooperativeLevel failed (%s)!\n", DD_errMsg(ret));
                return -1;
                }

            // Set the display mode
            if ((ret = IDirectDraw7_SetDisplayMode(DD_lpDD,DV.cntModeInfo.XResolution,DV.cntModeInfo.YResolution,
                    (DV.cntModeInfo.BitsPerPixel == 15) ? 16 : DV.cntModeInfo.BitsPerPixel,refreshRate,0)) != DD_OK) {
                if ((ret = IDirectDraw7_SetDisplayMode(DD_lpDD,DV.cntModeInfo.XResolution,DV.cntModeInfo.YResolution,
                    (DV.cntModeInfo.BitsPerPixel == 15) ? 16 : DV.cntModeInfo.BitsPerPixel,0,0)) != DD_OK) {
                    N_log("DD_SetVideoMode: IDirectDraw7_SetDisplayMode failed (%s)!\n", DD_errMsg(ret));
                    return -1;
                    }
                }

            // Disable the mouse cursor. Note that we have to forcibly move the cursor to
            // avoid problems bizarre bugs in NVIDIA DirectDraw drivers. Moving the
            // cursor forces it to go away, otherwise is stays on the screen until the
            // user moves the mouse.
            pShowCursor(false);
            pGetCursorPos(&pt);
            pSetCursorPos(0,0);
            pSetCursorPos(pt.x,pt.y);
            DD_cursorHidden = true;
            }
        else {
            _CHK(N_log("DD_SetVideoMode: Setting windowed cooperative mode\n"));
            // Set the DirectDraw Cooperative level
            if ((ret = IDirectDraw7_SetCooperativeLevel(DD_lpDD,NULL,DDSCL_NORMAL)) != DD_OK) {
                N_log("DD_SetVideoMode: IDirectDraw7_SetCooperativeLevel failed (%s)!\n", DD_errMsg(ret));
                return -1;
                }
            }

        // Create a single primary surface that will be the active surface until the
        // buffer manager is initialised.
        memset(&ddsd,0,sizeof(ddsd));
        ddsd.dwSize = sizeof(ddsd);
        if (!(gaMode & gaWindowedMode) && (DV.cntModeInfo.MaxBuffers > 1)) {
            ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
            ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX | DDSCAPS_VIDEOMEMORY;
            ddsd.dwBackBufferCount = 1;
            }
        else {
            // Note that we need to make sure that we do not set the DDSCAPS_VIDEOMEMORY
            // flag here, as if we are not requesting flip buffers the primary buffer is
            // always is video memory. If we do set this flag, the create surface calls
            // will fail for cases where no native DirectDraw support is available.
            ddsd.dwFlags = DDSD_CAPS;
            ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
            }
        // Direct3D HW support is only available in direct color modes, so there is no
        // point in enabling 3D in palette modes. Otherwise this step would fail to
        // create the primary surface in 8bpp modes on some DX drivers.
        if ((DD_caps.dwCaps & DDCAPS_3D) && DV.cntModeInfo.BitsPerPixel > 8)
            ddsd.ddsCaps.dwCaps |= DDSCAPS_3DDEVICE;
        if ((ret = IDirectDraw7_CreateSurface(DD_lpDD,&ddsd,&DD_lpPrimary,NULL)) != DD_OK) {
            N_log("DD_SetVideoMode: IDirectDraw7_CreateSurface failed (%s)!\n", DD_errMsg(ret));
            goto Error;
            }

        // Read the DirectDraw surface description for our primary surface and
        // re-build the pixel format and bits per pixel values.
        // There are some DirectDraw drivers out there that incorrectly return
        // values for EnumDisplayModes, so we fix them up here if the 16bpp
        // modes was incorrectly reported.
        ddsd.dwSize = sizeof(ddsd);
        IDirectDrawSurface7_GetSurfaceDesc(DD_lpPrimary,&ddsd);
        DD_GetPixelFormat(&ddsd.u4.ddpfPixelFormat,&bits,&DD_iPrimaryPF);
        DV.cntModeInfo.BitsPerPixel = bits;
        DV.cntModeInfo.PixelFormat = DD_pixelFormats[DD_iPrimaryPF];
        if ((ret = IDirectDrawSurface7_Lock(DD_lpPrimary,NULL,&ddsd,LOCK_FLAGS,NULL)) != DD_OK) {
            N_log("DD_SetVideoMode: IDirectDrawSurface7_Lock failed (%s)!\n", DD_errMsg(ret));
            goto Error;
            }
        if ((ret = IDirectDrawSurface7_Unlock(DD_lpPrimary,NULL)) != DD_OK) {
            N_log("DD_SetVideoMode: IDirectDrawSurface7_Unlock failed (%s)!\n", DD_errMsg(ret));
            goto Error;
            }
        DC.LinearMem = ddsd.lpSurface;
        DV.cntModeInfo.BytesPerScanLine = ddsd.u1.lPitch;

        // Indicate that the currently active and visible surfaces are the primary
        DD_lpActive = DD_lpPrimary;
        DD_lpVisible = DD_lpPrimary;

        // Create the default palette object for 8 bit displays
        if (DV.cntModeInfo.BitsPerPixel == 8) {
            GA_palette      *pal = DC.VGAPal8;
            PALETTEENTRY    pe[256];
            int             i;

            for (i = 0; i < 256; i++) {
                pe[i].peRed = pal[i].Red;
                pe[i].peGreen = pal[i].Green;
                pe[i].peBlue = pal[i].Blue;
                pe[i].peFlags = PC_RESERVED;
                }
            if ((ret = IDirectDraw7_CreatePalette(DD_lpDD,DDPCAPS_8BIT | DDPCAPS_ALLOW256,pe,&DD_lpPalette,NULL)) != DD_OK) {
                N_log("DD_SetVideoMode: IDirectDraw7_CreatePalette failed (%s)!\n", DD_errMsg(ret));
                goto Error;
                }
            if ((ret = IDirectDrawSurface7_SetPalette(DD_lpPrimary,DD_lpPalette)) != DD_OK) {
                N_log("DD_SetVideoMode: IDirectDrawSurface7_SetPalette failed (%s)!\n", DD_errMsg(ret));
                goto Error;
                }
            }

        // Setup the 2D acceleration module
        DD_setupAccelerator();
        }

    // Return the display mode values
    *virtualX = DV.cntModeInfo.XResolution;
    *virtualY = DV.cntModeInfo.YResolution;
    *bytesPerLine = DV.cntModeInfo.BytesPerScanLine;

    // Now fill in the current mode information for this mode
    DV.cntModeInfo.BytesPerScanLine = DV.cntMode.bytesPerLine = *bytesPerLine;
    DV.cntMode.xRes = *virtualX;
    DV.cntMode.yRes = *virtualY;
    DV.cntMode.bits = DV.cntModeInfo.BitsPerPixel;
    DV.cntMode.flags = gaLinearOnly;
    if (DV.cntModeInfo.Attributes & gaHaveAccel2D)
        DV.cntMode.flags |= gaAccNonVGA;
    if (DV.cntModeInfo.Attributes & gaHaveAccel3D)
        DV.cntMode.flags |= ga3D;
    if (DV.cntModeInfo.Attributes & gaHaveAccelVideo)
        DV.cntMode.flags |= gaVideo;
    DV.cntDACWidth = 8;
    DV.cntModeNum = gaMode;

    // Clear all pending messages before completing the mode set
    DD_clearMessageQueue();
    return 0;

Error:
    for (i = 0; i < DD_numPixelFormats; i++)
        SAFE_RELEASE(DD_lpTexture[i]);
    SAFE_RELEASE(DD_lpD3DDevice);
    if (DD_lpPrimary) {
        IDirectDrawSurface7_SetPalette(DD_lpPrimary,NULL);
        SAFE_RELEASE(DD_lpPrimary);
        DD_activeBuf = NULL;
        DD_lpActive = NULL;
        DD_lpVisible = NULL;
        }
    SAFE_RELEASE(DD_lpPalette);
    IDirectDraw7_RestoreDisplayMode(DD_lpDD);
    return -1;
}

/****************************************************************************
REMARKS:
Returns the current display mode.
****************************************************************************/
N_uint32 NAPI DD_GetVideoMode(void)
{
    return DV.cntModeNum;
}

/****************************************************************************
PARAMETERS:
mode        - Mode number to get closest clock for
pixelClock  - Requested pixel clock in units of Hz.

RETURNS:
Closest pixel clock in units of Hz.

REMARKS:
This function allows an application to determine if a particular pixel clock
is available.
****************************************************************************/
N_uint32 NAPI DD_GetClosestPixelClock(
    N_int32 outputHead,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bitsPerPixel,
    N_uint32 pixelClock)
{
    // Not supported by DirectDraw
    (void)outputHead;
    (void)xRes;
    (void)yRes;
    (void)bitsPerPixel;
    (void)pixelClock;
    return -1;
}

/****************************************************************************
PARAMETERS:
crtc    - Place to store the CRTC timings for the current mode

REMARKS:
This function returns the set of CRTC timings that are currently active
for the current display mode.
****************************************************************************/
void NAPI DD_GetCRTCTimings(
    GA_CRTCInfo *crtc)
{
    // Not supported by DirectDraw
    memset(crtc,0,sizeof(*crtc));
}

/****************************************************************************
PARAMETERS:
crtc    - New CRTC timings to make active

REMARKS:
This function programs the set of CRTC timings passed in to make them the
active CRTC timings for the current display mode. This function is used
by the SNAP centering and refresh control functions.
****************************************************************************/
void NAPI DD_SetCRTCTimings(
    GA_CRTCInfo *crtc)
{
    // Not supported by DirectDraw
    (void)crtc;
}

/****************************************************************************
PARAMETERS:
crtc    - New CRTC timings to make active

REMARKS:
This function programs the set of CRTC timings passed in to make them the
active CRTC timings for the current display mode. This function is used
by the SNAP centering and refresh control functions.
****************************************************************************/
void NAPI DD_SaveCRTCTimings(
    GA_CRTCInfo *crtc)
{
    // Not supported by DirectDraw
    (void)crtc;
}

/****************************************************************************
PARAMETERS:
refresh - New global refresh rate to make active

REMARKS:
This function sets the global refresh rate for all display modes to the
new global refresh rate setting. This does not change the physical refresh
rate for the current display mode, but only changes the database that will
be used by the next mode set operation (assuming the default refresh
rate is selected).
****************************************************************************/
void NAPI DD_SetGlobalRefresh(
    N_int32 refresh,
    N_int32 outputHead)
{
    // Not supported by DirectDraw
    (void)refresh;
    (void)outputHead;
}

/****************************************************************************
PARAMETERS:
filename    - Place to store the unique filename

REMARKS:
This function generates a unique filename for the current chipset
configuration that is used to store POST register information and the
configuration options to disk at runtime. Because the filename is
unique, each card that is plugged into the system will have a different
set of persistent POST register state and configuration options.
****************************************************************************/
void NAPI DD_getUniqueFilename(
    char *filename,
    int type)
{
    strcpy(filename,DRV_path);
    switch (type) {
        case 1:
            strcat(filename,"modes");
            break;
        case 2:
            strcat(filename,"options");
            break;
        }
    PM_backslash(filename);
    strcat(filename,"ddraw.00");
}

/****************************************************************************
PARAMETERS:
m    - Place to store the monitor configuration information

REMARKS:
This function returns a copy of the currently configured monitor in
the driver.
****************************************************************************/
void NAPI DD_getMonitorInfo(
    GA_monitor *m,
    N_int32 outputHead)
{
    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    if (outputHead >= 0 && outputHead < DV.init.GetNumberOfHeads())
        *m = monitor[outputHead];
}

/****************************************************************************
PARAMETERS:
m    - New monitor configuration information

REMARKS:
This function sets the internal monitor configuration information in the
driver to the passed in values. This automatically changes the internal
behaviour of the driver to support the new monitor configuration.
****************************************************************************/
void NAPI DD_setMonitorInfo(
    GA_monitor *m,
    N_int32 outputHead)
{
    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    if (outputHead >= 0 && outputHead < DV.init.GetNumberOfHeads())
        monitor[outputHead] = *m;
}

/****************************************************************************
REMARKS:
Returns the curent video mode information from the driver.
****************************************************************************/
void NAPI DD_GetCurrentVideoModeInfo(
    GA_modeInfo *modeInfo)
{
    COPY_STRUCTURE(modeInfo,&DV.cntModeInfo);
}

/****************************************************************************
HEADER:
graphics/ddc.h

PARAMETERS:
supportedStates - Place to store the supported states

RETURNS:
True if the device supports DPMS functionality, false if not.

REMARKS:
Returns the supported DPMS states for the chipset driver.
****************************************************************************/
int NAPI DD_DPMSdetect(
    N_int16 *supportedStates)
{
    // Not supported by DirectDraw natively
    (void)supportedStates;
    return false;
}

/****************************************************************************
HEADER:
graphics/ddc.h

PARAMETERS:
capabilities    - Place to store the capabilities for the driver
numchannels - Place to store the number of monitor ports supported

RETURNS:
True if the device supports SCI functionality, false if not.

REMARKS:
Returns the supported SCI capabilities for the chipset driver.
****************************************************************************/
int NAPI DD_SCIdetect(
    N_uint8 *capabilities,
    N_int32 *numchannels)
{
    // Not supported by DirectDraw natively
    (void)capabilities;
    (void)numchannels;
    return false;
}

/****************************************************************************
REMARKS:
This function sets the current output display device for subsequent SNAP
functions.
****************************************************************************/
N_int32 NAPI DD_SetDisplayOutput(
    N_int32 device)
{
    // Not supported by DirectDraw with SNAP ring 0 driver.
    return -1;
}

/****************************************************************************
REMARKS:
This function sets the current output display device for subsequent SNAP
functions.
****************************************************************************/
N_int32 NAPI DD_SetDisplayOutputExt(
    N_int32 device,
    N_int32 outputHead)
{
    // Not supported by DirectDraw with SNAP ring 0 driver.
    return -1;
}

/****************************************************************************
REMARKS:
This function returns the current output device flags.
****************************************************************************/
N_int32 NAPI DD_GetDisplayOutput(void)
{
    // Not supported by DirectDraw with SNAP ring 0 driver. There may be
    // a way to determine if an LCD panel is active, but I am not yet aware
    // of a method we can use.
    return gaOUTPUT_CRT;
}

/****************************************************************************
REMARKS:
This function returns the current output device flags.
****************************************************************************/
N_int32 NAPI DD_GetDisplayOutputExt(
    N_int32 outputHead)
{
    // Not supported by DirectDraw with SNAP ring 0 driver. There may be
    // a way to determine if an LCD panel is active, but I am not yet aware
    // of a method we can use.
    return gaOUTPUT_CRT;
}

/****************************************************************************
REMARKS:
This function will reset the current hardware display mode in a
non-destructive manner such that the screen and current hardware
acceleration state are preserved. This function is used to dynamically
switch the active display device and re-program the hardware as necessary
for the new active display device (ie: LCD/CRT/TV switching etc).
****************************************************************************/
N_int32 NAPI DD_SwitchPhysicalResolution(
    N_int32 physicalXResolution,
    N_int32 physicalYResolution,
    N_int32 refreshRate)
{
    // Not supported on DirectDraw
    return -1;
}

/****************************************************************************
REMARKS:
Returns the the currently active refresh rate for the display mode (in 0.01
Hz increments).
****************************************************************************/
N_int32 NAPI DD_GetCurrentRefreshRate(void)
{
    return DV.cntCRTC.RefreshRate;
}

/****************************************************************************
REMARKS:
Default multi-head functions for single head cards.
****************************************************************************/
N_int32 NAPI DRV_GetNumberOfHeads(void)
{
    return 1;
}

/****************************************************************************
REMARKS:
Default multi-head functions for single head cards.
****************************************************************************/
N_int32 NAPI DRV_SetActiveHead(
    N_int32 headIndex)
{
    (void)headIndex;
    return 0;
}

/****************************************************************************
REMARKS:
Default multi-head functions for single head cards.
****************************************************************************/
N_int32 NAPI DRV_GetActiveHead(void)
{
    return 0;
}

