/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the generic SNAP Graphics
*               Architecture API functions on top of DirectDraw for Win32.
*
****************************************************************************/

#include "internal.h"

/*--------------------------- Global variables ----------------------------*/

static DD_bufMgrState   BM = {0};
static DWORD            dwTileSize = 128;    // D3D texture tile size

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Initialise the buffer manager functions to be exported by the driver.
****************************************************************************/
void BUFFER_init(void)
{
    // Configure the buffer manager functions
    DV.buf.SetRef2d                 = DD_SetRef2d;
    DV.buf.InitBuffers              = DD_InitBuffers;
    DV.buf.GetPrimaryBuffer         = DD_GetPrimaryBuffer;
    DV.buf.GetFlippableBuffer       = DD_GetFlippableBuffer;
    DV.buf.AllocBuffer              = DD_AllocBuffer;
    DV.buf.FreeBuffer               = DD_FreeBuffer;
    DV.buf.LockBuffer               = DD_LockBuffer;
    DV.buf.UnlockBuffer             = DD_UnlockBuffer;
    DV.buf.SetActiveBuffer          = DD_SetActiveBuffer;
    DV.buf.FlipToBuffer             = DD_FlipToBuffer;
    DV.buf.GetFlipStatus            = DD_GetFlipStatus;
    DV.buf.WaitTillFlipped          = DD_WaitTillFlipped;
    DV.buf.UpdateCache              = DD_UpdateCache;
    DV.buf.UpdateFromCache          = DD_UpdateFromCache;
    DV.buf.BitBltBuf                = DD_BitBltBuf;
    DV.buf.BitBltPattBuf            = DD_BitBltPattBuf;
    DV.buf.BitBltColorPattBuf       = DD_BitBltColorPattBuf;
    DV.buf.SrcTransBltBuf           = DD_SrcTransBltBuf;
    DV.buf.DstTransBltBuf           = DD_DstTransBltBuf;
    DV.buf.StretchBltBuf            = DD_StretchBltBuf;
    DV.buf.BitBltPlaneMaskedBuf     = DD_BitBltPlaneMaskedBuf;
    DV.buf.BitBltFxBuf              = DD_BitBltFxBuf;
    DV.buf.DrawRectBuf              = DD_DrawRectBuf;
    DV.buf.GetClipper               = DD_GetClipper;
    DV.buf.SetClipper               = DD_SetClipper;
    DV.buf.GetActiveBuffer          = DD_GetActiveBuffer;
    DV.buf.GetVisibleBuffer         = DD_GetVisibleBuffer;
    DV.buf.GetVisibleRightBuffer    = DD_GetVisibleRightBuffer;

    // Configure the clipper functions
    DV.clip.CreateClipper           = DD_CreateClipper;
    DV.clip.IsClipListChanged       = DD_IsClipListChanged;
    DV.clip.GetClipList             = DD_GetClipList;
    DV.clip.DestroyClipper          = DD_DestroyClipper;
}

/****************************************************************************
REMARKS:
Initial entry point called when the buffer manager is initialised to pass
down a copy of the ref2d driver used for drawing. We store a pointer to
the driver and initialise the internals for the buffer manager code in here.
****************************************************************************/
ibool NAPI DD_SetRef2d(
    REF2D_driver *ref2d)
{
    int size;

    // Initialise device driver internals
    memset(&BM,0,sizeof(BM));
    BM.ref2d = ref2d;

    // Initialise copies of the software rendering functions
    // from the 2D reference rasteriser. These functions are
    // obtained *before* the ref2d driver has overridden any
    // internal functions with buffer manager versions. Hence
    // this avoids any recursive calls in the buffer manager.
    BM.state2d.dwSize = sizeof(BM.state2d);
    BM.ref2d->QueryFunctions(GA_GET_2DSTATEFUNCS,&BM.state2d);

    // Create all the block lists
    if (!N_createBlockList(&BM.clipperList.bl,sizeof(GA_clipperInt) + sizeof(N_node)))
        return false;
    if (!N_createBlockList(&BM.bufferList.bl,sizeof(GA_bufInt) + sizeof(N_node)))
        return false;
    BM.clipperList.freeList = N_buildFreeList(&BM.clipperList.bl);
    BM.bufferList.freeList = N_buildFreeList(&BM.bufferList.bl);

    // Now initialise the heaps. If we have linear blits in the hardware,
    // then we setup for a fixed heap to grow from the start of video
    // memory up to the top of video memory
    N_initList(&BM.bufferHeap);

    // Allocate space to store the list of flippable buffers
    size = sizeof(BM.flipBuffers[0]) * DV.cntModeInfo.MaxBuffers;
    if ((BM.flipBuffers = malloc(size)) == NULL)
        return false;
    memset(BM.flipBuffers,0,size);

    // Allocate space for enumerating all attached flipping surfaces
    size = sizeof(BM.lpSurf[0]) * DV.cntModeInfo.MaxBuffers;
    if ((BM.lpSurf = malloc(size)) == NULL)
        return false;
    memset(BM.lpSurf,0,size);
    size = sizeof(BM.lpSurface[0]) * DV.cntModeInfo.MaxBuffers;
    if ((BM.lpSurface = malloc(size)) == NULL)
        return false;
    memset(BM.lpSurface,0,size);

    // Setup pointers to functions to copy buffer data for color depth
    switch (DV.cntModeInfo.BitsPerPixel) {
        case 8:
            BM.bufReadScan  = DD_readScan_8;
            BM.bufWriteScan = DD_writeScan_8;
            break;
        case 15:
        case 16:
            BM.bufReadScan  = DD_readScan_16;
            BM.bufWriteScan = DD_writeScan_16;
            break;
        case 24:
            BM.bufReadScan  = DD_readScan_24;
            BM.bufWriteScan = DD_writeScan_24;
            break;
        case 32:
            BM.bufReadScan  = DD_readScan_32;
            BM.bufWriteScan = DD_writeScan_32;
            break;
        }
    return true;
}

/****************************************************************************
REMARKS:
Cleans up and exits, freely all memory used by the buffer manager.
****************************************************************************/
void NAPI DD_exitBufMgr(void)
{
    if (BM.flipBuffers) {
        free(BM.flipBuffers);
        N_freeBlockList(&BM.clipperList.bl);
        N_freeBlockList(&BM.bufferList.bl);
        }
}

/****************************************************************************
REMARKS:
Resizes the buffer list if full by increasing its size by the original
default size and building a new free list in the resized portion.
****************************************************************************/
static ibool DD_resizeClipperList(void)
{
    if (!N_resizeBlockList(&BM.clipperList.bl))
        return false;
    BM.clipperList.freeList = N_buildFreeList(&BM.clipperList.bl);
    return true;
}

/****************************************************************************
RETURNS:
Pointer to new clipper structure

REMARKS:
Allocates space for a new clipper in the clipper pool from the free list. If
there are no more clippers, we attempt to re-size the clipper pool which will
bomb out if we run out of memory.
****************************************************************************/
static GA_clipperInt * DD_newClipperOnHeap(void)
{
    N_node  *n;

    if (BM.clipperList.freeList == NULL) {
        if (!DD_resizeClipperList())
            return NULL;
        }
    n = BM.clipperList.freeList;
    BM.clipperList.freeList = N_FREELIST_NEXT(n);
    return (GA_clipperInt*)LST_USERSPACE(n);
}

/****************************************************************************
PARAMETERS:
buffer  - Pointer to the clipper to free

REMARKS:
Frees the clipper by putting it onto the clipper free list.
****************************************************************************/
static void DD_freeClipperOnHeap(
    GA_clipperInt *buffer)
{
    N_node *n = LST_HEADER(buffer);
    N_FREELIST_NEXT(n) = BM.clipperList.freeList;
    BM.clipperList.freeList = n;
}

/****************************************************************************
PARAMETERS:
hwnd    - Window handle to create the clipper for

RETURNS:
Pointer to the allocated clipper, NULL on failure.

REMARKS:
Internal function to do the clipper creation from the clipper heap.
****************************************************************************/
GA_clipper * NAPI DD_CreateClipper(
    PM_HWND hwnd)
{
    GA_clipperInt   *clip = DD_newClipperOnHeap();

    if (IDirectDraw7_CreateClipper(DD_lpDD, 0, &clip->lpClipper, NULL) != DD_OK) {
        DD_freeClipperOnHeap(clip);
        return NULL;
        }
    IDirectDrawClipper_SetHWnd(clip->lpClipper, 0, (HWND)hwnd);
    clip->hwnd = hwnd;
    clip->bufSize = clip->bufSizePrev = 0;
    clip->lpClipList = clip->lpClipListPrev = NULL;
    clip->changed = false;
    return clip;
}

/****************************************************************************
PARAMETERS:
clipper - Clipper object to test

RETURNS:
True if the window clip list has changed, false if not.

REMARKS:
Internal function to check if the window clip list has changed for a clipper
object or not and return the status.

Note:   The IDirectDrawClipper_IsClipListChanged() function always returns
        true for some bizarre reason, so for now I have changed the code to
        get the new clip list in here and do a compare with the previous
        clip list to determine if it has really changed.

Note:   This function always returns true the first time it is called.
****************************************************************************/
ibool NAPI DD_IsClipListChanged(
    GA_clipper *clipper)
{
    GA_clipperInt   *clip = clipper;
    DWORD           dwSize,dwSizePrev;
    HRESULT         err;

    if (!clip->changed) {
        // Copy the current clip list to the previous clip list
        if (clip->bufSize > clip->bufSizePrev) {
            if ((clip->lpClipListPrev = realloc(clip->lpClipListPrev, clip->bufSizePrev = clip->bufSize)) == NULL)
                PM_fatalError("Out of memory!");
            }
        if (clip->lpClipList)
            memcpy(clip->lpClipListPrev,clip->lpClipList,clip->bufSize);

        // Get the clip list from DirectDraw. If the region data block is too small,
        // we keep doubling the size of our clip list data block until it is big enough
        // to fit.
        err = DD_OK;
        for (;;) {
            if (!clip->lpClipList || err == DDERR_REGIONTOOSMALL) {
                clip->bufSize = clip->bufSize ? (clip->bufSize * 2) : (sizeof(RGNDATAHEADER) * 2);
                if ((clip->lpClipList = realloc(clip->lpClipList, clip->bufSize)) == NULL)
                    PM_fatalError("Out of memory!");
                }
            dwSize = clip->bufSize;
            if ((err = IDirectDrawClipper_GetClipList(clip->lpClipper, NULL, clip->lpClipList, &dwSize)) == DD_OK)
                break;
            if (err != DDERR_REGIONTOOSMALL)
                PM_fatalError(DD_errMsg(err));
            }

        // Now compare the two clip lists and determine if it has changed
        if (clip->lpClipListPrev) {
            dwSize = sizeof(RGNDATAHEADER) + clip->lpClipList->rdh.nRgnSize;
            dwSizePrev = sizeof(RGNDATAHEADER) + clip->lpClipList->rdh.nRgnSize;
            if (dwSize != dwSizePrev || memcmp(clip->lpClipList,clip->lpClipListPrev,dwSize) != 0)
                clip->changed = true;
            }
        else
            clip->changed = true;
        }
    return clip->changed;
}

/****************************************************************************
PARAMETERS:
clipper - Clipper object to test
count   - Number of rectangles in clip list

RETURNS:
Pointer to a list of rectangles in the clip list.

REMARKS:
This function returns a complex clip list for the associated clipper
if the window requires complex clipping. If the region does not require
complex clipping, this function will return a single rectangle and return
a value of 1 in the count parameter.
****************************************************************************/
GA_rect * NAPI DD_GetClipList(
    GA_clipper *clipper,
    N_int32 *count)
{
    GA_clipperInt   *clip = clipper;

    // DD_IsClipListChanged takes care of reading the clip list. If it has
    // already been called, then it will return immediately as we have
    // already loaded the changed clip list when it was called previously.
    DD_IsClipListChanged(clipper);
    clip->changed = false;
    *count = clip->lpClipList->rdh.nCount;
    return (GA_rect*)clip->lpClipList->Buffer;
}

/****************************************************************************
PARAMETERS:
clipper - Clipper object to free

REMARKS:
This function frees a previously allocated clipper object.
****************************************************************************/
void NAPI DD_DestroyClipper(
    GA_clipper *clipper)
{
    GA_clipperInt   *clip = clipper;

    if (clip->lpClipListPrev)
        free(clip->lpClipListPrev);
    if (clip->lpClipList)
        free(clip->lpClipList);
    SAFE_RELEASE(clip->lpClipper);
    DD_freeClipperOnHeap(clip);
}

/****************************************************************************
REMARKS:
Resizes the buffer list if full by increasing its size by the original
default size and building a new free list in the resized portion.
****************************************************************************/
static ibool DD_resizeBufferList(void)
{
    if (!N_resizeBlockList(&BM.bufferList.bl))
        return false;
    BM.bufferList.freeList = N_buildFreeList(&BM.bufferList.bl);
    return true;
}

/****************************************************************************
RETURNS:
Pointer to new buffer structure

REMARKS:
Allocates space for a new buffer in the buffer pool from the free list. If
there are no more buffers, we attempt to re-size the buffer pool which will
bomb out if we run out of memory.
****************************************************************************/
static GA_bufInt * DD_newBufferOnHeap(void)
{
    N_node  *n;

    if (BM.bufferList.freeList == NULL) {
        if (!DD_resizeBufferList())
            return NULL;
        }
    n = BM.bufferList.freeList;
    BM.bufferList.freeList = N_FREELIST_NEXT(n);
    return (GA_bufInt*)LST_USERSPACE(n);
}

/****************************************************************************
PARAMETERS:
buffer  - Pointer to the buffer to free

REMARKS:
Frees the buffer by putting it onto the buffer free list.
****************************************************************************/
static void DD_freeBufferOnHeap(
    GA_bufInt *buffer)
{
    N_node *n = LST_HEADER(buffer);
    N_FREELIST_NEXT(n) = BM.bufferList.freeList;
    BM.bufferList.freeList = n;
}

/****************************************************************************
PARAMETERS:
width           - Width of the buffer in pixels
height          - Height of the buffer in scanlines
flags           - Flags to create the buffer with

RETURNS:
Pointer to the allocated buffer, NULL on failure.

REMARKS:
Internal function to do the buffer creation from the buffer heap.
****************************************************************************/
static GA_bufInt * DD_createBuffer(
    N_int32 width,
    N_int32 height,
    N_int32 flags,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf,
    N_int32 format,
    LPDIRECTDRAWSURFACE7 lpSurf)
{
    HRESULT         ret;
    int             sysMemSize;
    GA_bufInt       *buf = DD_newBufferOnHeap();

    // Check for flags that are dependant on each other
    if (!buf)
        return NULL;
    if (flags & gaBufferPageable)
        flags |= gaBufferCached;
    if (flags & gaBufferCached)
        flags &= ~gaBufferSysMem;
    if (flags & gaBufferSysMem)
        flags &= ~gaBufferNoSysMem;
    if ((flags & gaBufferSpecial) && (flags & (gaBufferSysMem | gaBufferCached | gaBufferMoveable | gaBufferPageable)))
        return NULL;

    // Now fill in the information for the buffer and create it
    memset(buf,0,sizeof(*buf));
    buf->p.dwSize = sizeof(GA_buf);
    buf->p.Width = width;
    buf->p.Height = height;
    buf->p.CacheStride = BYTESPERLINE(width,bitsPerPixel);
    buf->p.Flags = flags;
    buf->p.Format = format;
    buf->p.BitsPerPixel = bitsPerPixel;
    buf->p.PixelFormat = *pf;
    sysMemSize = height * buf->p.CacheStride;
    if ((flags & gaBufferSysMem) || (lpSurf == NULL)) {
        // Return failure if not in hardware if not cached
        if (!(flags & (gaBufferCached | gaBufferSysMem)) || (flags &  gaBufferNoSysMem))
            goto Error;

        // Buffer creation failed in hardware, so allocate it in system
        // memory.
        if ((buf->p.Surface = malloc(sysMemSize)) == NULL)
            goto Error;
        memset(buf->p.Surface,0,sysMemSize);
        buf->p.Stride = buf->p.CacheStride;
        buf->p.SurfaceCache = buf->p.Surface;

        // Indicate if this is a system memory only buffer
        if (!flags & gaBufferCached)
            buf->p.Flags |= gaBufferSysMem;
        }

    // If the buffer is not in system memory, create the surface cache for
    // the buffer in system memory if caching has been requested.
    if ((flags & gaBufferCached) && !buf->p.SurfaceCache) {
        if ((buf->p.SurfaceCache = malloc(sysMemSize)) == NULL)
            goto Error;
        }

    // Now lock the DirectDraw surface to find the surface location
    // if the surface was created with DirectDraw.
    if (lpSurf) {
        if ((ret = IDirectDrawSurface7_Lock(lpSurf,NULL,&DD_lockBuf,LOCK_FLAGS,NULL)) != DD_OK) {
            N_log("DD_createBuffer: IDirectDrawSurface7_Lock failed (%s)!\n", DD_errMsg(ret));
            goto Error;
            }
        if ((ret = IDirectDrawSurface7_Unlock(lpSurf,NULL)) != DD_OK) {
            N_log("DD_createBuffer: IDirectDrawSurface7_Unlock failed (%s)!\n", DD_errMsg(ret));
            goto Error;
            }
        buf->p.Stride = DD_lockBuf.u1.lPitch;
        buf->p.Offset = (ulong)DD_lockBuf.lpSurface - (ulong)DC.LinearMem;
        buf->lpSurf = lpSurf;
        buf->lpSurface = DD_lockBuf.lpSurface;
        buf->srcColorKey = 0xFFFFFFFF;
        buf->dstColorKey = 0xFFFFFFFF;
        }

    // Add the new buffer to the head of the buffer chain
    N_addToHead(&BM.bufferHeap,buf);
    return buf;

Error:
    DD_freeBufferOnHeap(buf);
    return NULL;
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to destroy

REMARKS:
This function destroys a SNAP buffer, and frees all resources associated
with the buffer.
****************************************************************************/
static ibool NAPI DD_destroyBuffer(
    GA_bufInt *buf)
{
    // Remove the buffer from the list of allocated buffers and destroy it
    if (!buf)
        return false;
    N_remove(&BM.bufferHeap,buf);
    if (buf->p.Flags & gaBufferSysMem) {
        // Buffer was in system memory
        free(buf->p.Surface);
        }

    // Free up all memory occupied by the buffer
    if (buf->p.SurfaceCache)
        free(buf->p.SurfaceCache);

    // Destroy the buffer object from our heap
    DD_freeBufferOnHeap(buf);
    return true;
}

/****************************************************************************
PARAMETERS:
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
bitsPerPixel- Depth of source bitmap
pf          - Pixel format of source bitmap
flags       - Flags for surface creation
lpSurf      - Pointer to DirectDraw surface to be created

RETURNS:
True if DirectDraw surface was created, or false if not.

REMARKS:
This functions attempts to create a DirectDraw surface object in the
specified pixel format. Flags will determine if the surface needs to
be a 3D render target or texture.
****************************************************************************/
ibool DD_createSurface(
    N_int32 *width,
    N_int32 *height,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf,
    N_uint32 flags,
    LPDIRECTDRAWSURFACE7 *lpSurf)
{
    int             i;
    DDSURFACEDESC2  ddsd;
    HRESULT         ret;

    memset(&ddsd,0,sizeof(ddsd));
    ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_VIDEOMEMORY;
    ddsd.dwWidth = *width;
    ddsd.dwHeight = *height;
    if (flags & gaBuffer3D)
        ddsd.ddsCaps.dwCaps |= DDSCAPS_3DDEVICE;

    // Determine pixel format if specified
    if (pf) {
        ddsd.dwFlags |= DDSD_PIXELFORMAT;
        ddsd.u4.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
        ddsd.u4.ddpfPixelFormat.dwFlags = (bitsPerPixel <= 8) ? DDPF_PALETTEINDEXED8 : DDPF_RGB;
        if (flags & gaBufferAlphaPixels)
            ddsd.u4.ddpfPixelFormat.dwFlags |= DDPF_ALPHAPIXELS;
        ddsd.u4.ddpfPixelFormat.u1.dwRGBBitCount = (bitsPerPixel == 15) ? 16 : bitsPerPixel;
        ddsd.u4.ddpfPixelFormat.u2.dwRBitMask = (DWORD)pf->RedMask << pf->RedPosition;
        ddsd.u4.ddpfPixelFormat.u3.dwGBitMask = (DWORD)pf->GreenMask << pf->GreenPosition;
        ddsd.u4.ddpfPixelFormat.u4.dwBBitMask = (DWORD)pf->BlueMask << pf->BluePosition;
        ddsd.u4.ddpfPixelFormat.u5.dwRGBAlphaBitMask = (DWORD)pf->AlphaMask << pf->AlphaPosition;
        }

    // Create a 3D texture surface if specified
    if (flags & gaBufferTexture) {
        ddsd.dwFlags |= DDSD_TEXTURESTAGE;
        ddsd.dwTextureStage = 0;
        ddsd.ddsCaps.dwCaps = DDSCAPS_TEXTURE;
        ddsd.ddsCaps.dwCaps2 = 0;

        // Check for textures which *must* be created as powers of 2
        if ((DD_caps3d.dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_POW2)
                && !(DD_caps3d.dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_NONPOW2CONDITIONAL)) {
            for (i = 2; i < *width; i *= 2)
                ddsd.dwWidth = i;
            for (i = 2; i < *height; i *= 2)
                ddsd.dwHeight = i;
            }

        // Check for other texture constraints
        if (DD_caps3d.dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_SQUAREONLY)
            ddsd.dwWidth = ddsd.dwHeight = MAX(ddsd.dwWidth,ddsd.dwHeight);
        if (DD_caps3d.dwMaxTextureWidth && (ddsd.dwWidth > DD_caps3d.dwMaxTextureWidth))
            return false;
        if (DD_caps3d.dwMaxTextureHeight && (ddsd.dwHeight > DD_caps3d.dwMaxTextureHeight))
            return false;

        *width = ddsd.dwWidth;
        *height = ddsd.dwHeight;
        }

    // Now create the actual surface
    if ((ret = IDirectDraw7_CreateSurface(DD_lpDD,&ddsd,lpSurf,NULL)) != DD_OK) {
        N_log("DD_createSurface: IDirectDraw7_CreateSurface failed (%s)!\n", DD_errMsg(ret));
        return false;
        }
    return true;
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
bitsPerPixel- Depth of source bitmap
pf          - Pixel format of source bitmap
pal         - Palette of source bitmap (optional)
lpSurf      - Pointer to DirectDraw surface to be created

RETURNS:
True if DirectDraw surface was created, or false if not.

REMARKS:
This functions attempts to create a DirectDraw surface object to describe
a system memory buffer that has already been allocated and loaded with a
bitmap image. If successful, the bitmap image will be associated with a
DirectDraw surface with the pixel format that is supported by hardware.
****************************************************************************/
ibool DD_createSurfaceForMemory(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf,
    GA_palette *pal,
    LPDIRECTDRAWSURFACE7 *lpSurf)
{
    N_uint8         *addr = ((N_uint8*)srcAddr + (srcPitch * srcTop)) + (srcLeft);
    DDSURFACEDESC2  ddsd;

    // Setup DirectDraw surface object for system memory buffer
    memset(&ddsd,0,sizeof(ddsd));
    ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_PITCH | DDSD_LPSURFACE
                        | DDSD_PIXELFORMAT | DDSD_CAPS;
    ddsd.dwWidth = srcWidth;
    ddsd.dwHeight = srcHeight;
    ddsd.u1.lPitch = srcPitch;
    ddsd.lpSurface = addr;
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;

    // Determine pixel format of system memory buffer
    ddsd.u4.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
    ddsd.u4.ddpfPixelFormat.dwFlags = (bitsPerPixel <= 8) ? DDPF_PALETTEINDEXED8 : DDPF_RGB;
    ddsd.u4.ddpfPixelFormat.u1.dwRGBBitCount = bitsPerPixel;
    ddsd.u4.ddpfPixelFormat.u2.dwRBitMask = (DWORD)(pf->RedMask)   << (N_uint8)(pf->RedPosition);
    ddsd.u4.ddpfPixelFormat.u3.dwGBitMask = (DWORD)(pf->GreenMask) << (N_uint8)(pf->GreenPosition);
    ddsd.u4.ddpfPixelFormat.u4.dwBBitMask = (DWORD)(pf->BlueMask)  << (N_uint8)(pf->BluePosition);
    ddsd.u4.ddpfPixelFormat.u5.dwRGBAlphaBitMask = (DWORD)(pf->AlphaMask) << (N_uint8)(pf->AlphaPosition);

    // Create DirectDraw surface object for system memory buffer.
    // No memory is allocated by this type of surface object,
    // since DDSD_LPSURFACE indicates that it already exists.
    return (IDirectDraw7_CreateSurface(DD_lpDD,&ddsd,&*lpSurf,NULL) == DD_OK);
}

/****************************************************************************
REMARKS:
Method called to enumerate all of the attached back buffers for the
DirectDraw surface, storing the pointers to the buffers in our global buffers
array. Note that only one back buffer is ever attached to a surface at
a time as they are all attached in a chain.
****************************************************************************/
static HRESULT CALLBACK EnumBackBuffersCallBack(
    LPDIRECTDRAWSURFACE7 lpSurf,
    LPDDSURFACEDESC2 lpDesc,
    LPVOID lpv)
{
    BM.lpSurf[0] = lpSurf;
    return DDENUMRET_OK;
}

/****************************************************************************
REMARKS:
Method called to enumerate all of the available texture formats.
****************************************************************************/
static HRESULT CALLBACK EnumTextureFormatsCallback(
    LPDDPIXELFORMAT lpDDPixFmt,
    LPVOID lpv)
{
    int bits,ipf;

    if (DD_GetPixelFormat(lpDDPixFmt,&bits,&ipf))
        DD_textureFormats[ipf] = DDPF_RGB | (lpDDPixFmt->dwFlags & DDPF_ALPHAPIXELS);
    return DDENUMRET_OK;
}

/****************************************************************************
PARAMETERS:
numBuffers  - Number of flippable buffers to allocate
flags       - Extra flags to pass when allocating the buffers
stereo      - Software stereo functions, or NULL for non-stereo mode

RETURNS:
True on success, false on failure.

REMARKS:
This function initialises the buffer manager, and allocates the primary
buffer and all related flippable buffers (for a total of numBuffers buffers).
If any of the buffers cannot be allocated, this function will return false.
If the stereo parameter is not NULL, stereo mode is enabled using the
passed in software stereo functions.
****************************************************************************/
ibool NAPI DD_InitBuffers(
    N_int32 numBuffers,
    N_uint32 flags,
    GA_softStereoFuncs *stereo)
{
    HRESULT                 ret;
    DDSURFACEDESC2          ddsd;
    GA_buffer               drawBuf;
    int                     i;
    LPDIRECTDRAWSURFACE7    lpPrevSurf;

    // Check that the number of buffers is within the maximum supported
    if (numBuffers > DV.cntModeInfo.MaxBuffers)
        return false;

    // We don't presently support stereo with the DirectDraw buffer manager
    if (stereo)
        return false;

    // Free the current single primary surface so we can create a flipping
    // primary surface for the buffer manager. We also must detach the palette
    // object from it.
    if (DD_lpPrimary) {
        IDirectDrawSurface7_SetPalette(DD_lpPrimary,NULL);
        SAFE_RELEASE(DD_lpPrimary);
        DD_activeBuf = NULL;
        DD_lpActive = NULL;
        DD_lpVisible = NULL;
        }

    // Now allocate the primary buffer and all subsequent back flipping
    // buffers as requested by the application.
    DD_lpPrimary = NULL;
    BM.numFlipBuffers = 0;
    memset(&ddsd,0,sizeof(ddsd));
    ddsd.dwSize = sizeof(ddsd);
    if (numBuffers > 1) {
        ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
        ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX | DDSCAPS_VIDEOMEMORY;
        ddsd.dwBackBufferCount = numBuffers-1;
        }
    else {
        // Note that we need to make sure that we do not set the DDSCAPS_VIDEOMEMORY
        // flag here, as if we are not requesting flip buffers the primary buffer is
        // always is video memory. If we do set this flag, the create surface calls
        // will fail for cases where no native DirectDraw support is available.
        ddsd.dwFlags = DDSD_CAPS;
        ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
        }
    // Note that Direct3D HW support is only available in direct color modes.
    if ((DD_caps.dwCaps & DDCAPS_3D) && DV.cntModeInfo.BitsPerPixel > 8)
        ddsd.ddsCaps.dwCaps |= DDSCAPS_3DDEVICE;
    if ((ret = IDirectDraw7_CreateSurface(DD_lpDD,&ddsd,&DD_lpPrimary,NULL)) != DD_OK) {
        N_log("DD_InitBuffers: IDirectDraw7_CreateSurface failed (%s)!\n", DD_errMsg(ret));
        goto Error;
        }

    // Attach the palette object for 8 bit displays
    if (DV.cntModeInfo.BitsPerPixel == 8) {
        if ((ret = IDirectDrawSurface7_SetPalette(DD_lpPrimary,DD_lpPalette)) != DD_OK) {
            N_log("DD_InitBuffers: IDirectDrawSurface7_SetPalette failed (%s)!\n", DD_errMsg(ret));
            goto Error;
            }
        }

    // Find the pointer to the primary surface
    BM.flipBuffers[0] = DD_createBuffer(DV.cntModeInfo.XResolution,DV.cntModeInfo.YResolution,
        flags | gaBufferFlippable,DV.cntModeInfo.BitsPerPixel,&DV.cntModeInfo.PixelFormat,
        0,DD_lpPrimary);
    if (BM.flipBuffers[0] == NULL) {
        N_log("DD_InitBuffers: Unable to create primary buffer!\n");
        goto Error;
        }
    BM.numFlipBuffers = 1;

    // Clear all pending messages before completing the mode set
    DD_clearMessageQueue();

    // Enumerate all the attached surfaces so that we can get the pointers to
    // all the allocated back buffer surfaces.
    lpPrevSurf = DD_lpPrimary;
    while (BM.numFlipBuffers < numBuffers) {
        // Find the attached surface (note that only one is attached, we must
        // enumerate this attached surface to find the surface attached to it
        // etc until we get all the buffers). Note also that we always pass the surface
        // pointer around in BM.lpSurf[0] to avoid having to pass data into
        // the EnumBackBuffersCallBack function.
        IDirectDrawSurface7_EnumAttachedSurfaces(lpPrevSurf,NULL,EnumBackBuffersCallBack);
        BM.flipBuffers[BM.numFlipBuffers] = DD_createBuffer(DV.cntModeInfo.XResolution,
            DV.cntModeInfo.YResolution,flags | gaBufferFlippable,DV.cntModeInfo.BitsPerPixel,
            &DV.cntModeInfo.PixelFormat,0,BM.lpSurf[0]);
        if (BM.flipBuffers[BM.numFlipBuffers] == NULL) {
            N_log("DD_InitBuffers: Unable to create back flip buffer!\n");
            goto Error;
            }
        BM.numFlipBuffers++;
        lpPrevSurf = BM.lpSurf[0];
        }

    // Create Direct3D device object and texture surface for blit effects option.
    // Note that Direct3D HW support is only available in direct color modes.
    if (DD_lpD3D && DV.cntModeInfo.BitsPerPixel > 8) {
        // Create D3D device object for the primary surface render target
        memset(&DD_caps3d,0,sizeof(DD_caps3d));
        memset(DD_textureFormats,sizeof(DD_textureFormats[0] * DD_numPixelFormats),0);
        if ((ret = IDirect3D7_CreateDevice(DD_lpD3D,&IID_IDirect3DHALDevice,DD_lpPrimary,&DD_lpD3DDevice)) != DD_OK)
            N_log("DD_InitBuffers: IDirect3D7_CreateDevice failed (%s)!\n", DD_errMsg(ret));
        else {
            // Enumerate all available texture formats
            IDirect3DDevice7_GetCaps(DD_lpD3DDevice,&DD_caps3d);
            IDirect3DDevice7_EnumTextureFormats(DD_lpD3DDevice,EnumTextureFormatsCallback,NULL);

            // Create on texture for each format we need that is supported so
            // we can use them in color conversion operations.
            for (i = 0; i < DD_numPixelFormats; i++) {
                if (!DD_textureFormats[i])
                    continue;
                flags = gaBufferTexture;
                if (DD_textureFormats[i] & DDPF_ALPHAPIXELS)
                    flags |= gaBufferAlphaPixels;
                DD_textureSizeX[i] = DD_textureSizeY[i] = dwTileSize;
                if (!DD_createSurface(&DD_textureSizeX[i],&DD_textureSizeY[i],DD_pixelDepths[i],&DD_pixelFormats[i],flags,&DD_lpTexture[i]))
                    N_log("DD_InitBuffers: IDirectDraw7_CreateSurface Texture failed!\n");
                }
            }
        }

    // Now make the primary buffer the active drawing buffer
    drawBuf.dwSize = sizeof(drawBuf);
    drawBuf.Offset = BM.flipBuffers[0]->p.Offset;
    drawBuf.Stride = BM.flipBuffers[0]->p.Stride;
    drawBuf.Width  = BM.flipBuffers[0]->p.Width;
    drawBuf.Height = BM.flipBuffers[0]->p.Height;
    BM.ref2d->SetDrawBuffer(&drawBuf,DC.LinearMem,DV.cntModeInfo.BitsPerPixel,&DV.cntModeInfo.PixelFormat,&DC,false);

    // Make the primary buffer the currently visible and active drawing buffer
    DD_lpVisible = DD_lpPrimary;
    DD_lpActive = DD_lpPrimary;
    DD_activeBuf = (GA_bufInt*)BM.flipBuffers[0];

    // Clear all pending messages before completing the mode set
    DD_clearMessageQueue();
    return true;

Error:
    SAFE_RELEASE(DD_lpD3DDevice);
    for (i = 0; i < BM.numFlipBuffers; i++)
        DD_destroyBuffer(BM.flipBuffers[i]);
    if (DD_lpPrimary) {
        IDirectDrawSurface7_SetPalette(DD_lpPrimary,NULL);
        SAFE_RELEASE(DD_lpPrimary);
        DD_activeBuf = NULL;
        DD_lpActive = NULL;
        DD_lpVisible = NULL;
        }
    return false;
}

/****************************************************************************
PARAMETERS:
index   - Index of the flippable buffer to obtain the address of

RETURNS:
Pointer to specified flippable buffer, or NULL if not such buffer.

REMARKS:
This function returns a pointer to the flippable buffer with the specified
index in the flippable buffer table. These buffer pointers can then be used
with the SetActiveBuffer and FlipToBuffer functions to implement
hardware page flipping.
****************************************************************************/
GA_buf * NAPI DD_GetFlippableBuffer(
    N_int32 index)
{
    if (index >= BM.numFlipBuffers)
        return NULL;
    return &BM.flipBuffers[index]->p;
}

/****************************************************************************
RETURNS:
Pointer to the primary display buffer, NULL if no primary buffer.

REMARKS:
This function returns a pointer to the primary display buffer, which is
always the first buffer on the fixed heap for regular display output.
****************************************************************************/
GA_buf * NAPI DD_GetPrimaryBuffer(void)
{
    return &BM.flipBuffers[0]->p;
}

/****************************************************************************
PARAMETERS:
width   - Width of the buffer in pixels
height  - Height of the buffer in scanlines
flags   - Flags to create the buffer with

RETURNS:
Pointer to the allocated buffer, NULL on failure.

REMARKS:
This function allocates a new buffer and returns it. This only allocates
video memory buffers for drawing and bitmap storage and caching. It is
not used to allocate depth buffers or texture buffers.
****************************************************************************/
GA_buf * NAPI DD_AllocBuffer(
    N_int32 width,
    N_int32 height,
    N_int32 flags)
{
    LPDIRECTDRAWSURFACE7    lpSurf = NULL;
    GA_bufInt               *buf;

    // This function does not support allocating special buffers!
    if (flags & gaBufferSpecial)
        return NULL;

    // Now allocate the requested offscreen surface from DirectDraw
    if (!(flags & gaBufferSysMem)) {
        if (!DD_createSurface(&width,&height,0,NULL,flags,&lpSurf))
            return NULL;
        }

    // Now allocate the internal buffer for the surface
    if ((buf = DD_createBuffer(width,height,flags,DV.cntModeInfo.BitsPerPixel,&DV.cntModeInfo.PixelFormat,0,lpSurf)) == NULL) {
        SAFE_RELEASE(lpSurf);
        return NULL;
        }
    return &buf->p;
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to destroy

REMARKS:
This function destroys a SNAP buffer, and frees all resources associated
with the buffer.
****************************************************************************/
ibool NAPI DD_FreeBuffer(
    GA_buf *buf)
{
    int         i;
    GA_bufInt   *bufInt = (GA_bufInt*)buf;

    // App is never allowed to free the primary flippable buffers!
    for (i = 0; i < BM.numFlipBuffers; i++) {
        if (BM.flipBuffers[i] == bufInt)
            return false;
        }

    // Detach the clipper if there is one attached
    if (bufInt->clipper)
        IDirectDrawSurface7_SetClipper(bufInt->lpSurf,NULL);

    // Destroy the DirectDraw surface for the buffer
    SAFE_RELEASE(bufInt->lpSurf);

    // Finally destroy the internal buffer structures
    return DD_destroyBuffer(bufInt);
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to get attached clipper for

REMARKS:
This function returns the attached clipper object for the SNAP buffer
if one is attached. If no clipper is attached, this function returns NULL.
****************************************************************************/
GA_clipper * NAPI DD_GetClipper(
    GA_buf *buf)
{
    return ((GA_bufInt*)buf)->clipper;
}

/****************************************************************************
PARAMETERS:
buf     - SNAP buffer to set attached clipper for
clipper - New clipper object to attach to buffer

REMARKS:
This function sets the attached clipper object for the SNAP buffer. To
detach a clipper from the buffer, call this function with NULL for the
clipper parameter.
****************************************************************************/
void NAPI DD_SetClipper(
    GA_buf *buf,
    GA_clipper *clipper)
{
    GA_bufInt       *bufInt = (GA_bufInt*)buf;
    GA_clipperInt   *clipInt = (GA_clipperInt*)clipper;

    bufInt->clipper = clipInt;
    IDirectDrawSurface7_SetClipper(bufInt->lpSurf,clipInt ? clipInt->lpClipper : NULL);
}

/****************************************************************************
PARAMETERS:
buf - SNAP buffer to lock

RETURNS:
Physical address of locked buffer in memory, 0 if a system memory buffer.

REMARKS:
This function locks a buffer so that an application can begin drawing
directly on the surface memory. You /must/ call this function before you
draw directly on the bitmap surface!

The return value from this function is the physical start address of the
buffer in memory, which can be used to program DMA operations from other
hardware devices directly into the video memory for the buffer. This is
useful for frame grabber devices so that the resulting frame from the
frame grabber device can be blitted to the visual display as quickly as
possible (much quicker than if it was DMA'ed into system memory).

If the buffer is a system memory buffer, the return value from this
function will be 0, since we cannot obtain the physical starting address
of a system memory buffer.
****************************************************************************/
N_uint32 NAPI DD_LockBuffer(
    GA_buf *buf)
{
    GA_bufInt   *bufInt = (GA_bufInt*)buf;

    if (!(buf->Flags & gaBufferSysMem)) {
        if (bufInt->surfLock++ == 0) {
            while (IDirectDrawSurface7_Lock(bufInt->lpSurf,NULL,&DD_lockBuf,LOCK_FLAGS | DDLOCK_NOSYSLOCK,NULL) == DDERR_SURFACELOST)
                PM_doSuspendApp();
            buf->Surface = (uchar*)DC.LinearMem + buf->Offset;
            }
        return 0xFFFFFFFF;
        }
    return 0;
}

/****************************************************************************
PARAMETERS:
buf - SNAP buffer to unlock

REMARKS:
This function unlocks a buffer after the application has completed direct
surface access on the buffer.
****************************************************************************/
void NAPI DD_UnlockBuffer(
    GA_buf *buf)
{
    GA_bufInt   *bufInt = (GA_bufInt*)buf;

    if (!(buf->Flags & gaBufferSysMem)) {
        if (--bufInt->surfLock == 0) {
            while (IDirectDrawSurface7_Unlock(bufInt->lpSurf,NULL) == DDERR_SURFACELOST)
                PM_doSuspendApp();
            buf->Surface = NULL;
            }
        if (bufInt->surfLock < 0)
            bufInt->surfLock = 0;
        }
}

/****************************************************************************
PARAMETERS:
buf - SNAP buffer to make active for rendering

REMARKS:
This function makes the passed in buffer the active buffer for all
subsequent rendering operations.
****************************************************************************/
N_int32 NAPI DD_SetActiveBuffer(
    GA_buf *buf)
{
    GA_bufInt   *bufInt = (GA_bufInt*)buf;
    GA_buffer   drawBuf;

    DD_activeBuf = bufInt;
    if ((buf->Flags & gaBufferSysMem)) {
        // Switch drawing buffer to system memory buffer. System memory
        // surfaces are never allocated via DirectDraw, but internally
        // managed by our own code.
        DD_lpActive = NULL;
        BM.ref2d->SetDrawSurface(buf->Surface,buf->Width,buf->Height,buf->Stride,
            DV.cntModeInfo.BitsPerPixel,&DV.cntModeInfo.PixelFormat);
        }
    else {
        // Switch drawing buffer to hardware video memory buffer. Note that
        // we ignore the return value from SetDrawBuffer below because our
        // internal function will always fail as the draw buffer *must* be
        // set by calling this function. We still need to call SetDrawBuffer
        // below however to ensure that the ref2d software rendering functions
        // know the correct place to draw for software rendering.
        DD_lpActive = bufInt->lpSurf;
        BM.ref2d->SetDrawSurface(DC.LinearMem,buf->Width,buf->Height,buf->Stride,
            DV.cntModeInfo.BitsPerPixel,&DV.cntModeInfo.PixelFormat);
        drawBuf.dwSize = sizeof(drawBuf);
        drawBuf.Offset = buf->Offset;
        drawBuf.Stride = buf->Stride;
        drawBuf.Width  = buf->Width;
        drawBuf.Height = buf->Height;
        BM.state2d.SetDrawBuffer(&drawBuf);
        if (DD_lpD3DDevice) {
            // Set the render target for all 3D rendering to the new surface
            IDirect3DDevice7_SetRenderTarget(DD_lpD3DDevice, DD_lpActive, 0);
            }
        }
    return 0;
}

/****************************************************************************
RETURNS:
Pointer to the last buffer passed to SetActiveBuffer
****************************************************************************/
GA_buf * NAPI DD_GetActiveBuffer(void)
{
    return &DD_activeBuf->p;
}

/****************************************************************************
PARAMETERS:
buf - SNAP buffer to make visible for display

REMARKS:
This function makes the passed in buffer visible for the active display.
****************************************************************************/
void NAPI DD_FlipToBuffer(
    GA_buf *buf,
    N_int32 waitVRT)
{
    GA_bufInt               *bufInt = (GA_bufInt*)buf;
    LPDIRECTDRAWSURFACE7    lpSurf;
    int                     i,j;

    // Check that the buffer is not moveable or pageable
    if ((buf->Flags & (gaBufferMoveable | gaBufferPageable)))
        return;

    // Fix the triple buffer flag if we only have two buffers
    if (BM.numFlipBuffers <= 1 && waitVRT == gaTripleBuffer)
        waitVRT = gaWaitVRT;

    // Ensure that previous display start has taken hold if we have hardware
    // triple buffering. We do this by polling the GetDisplayStartStatus
    // function until it tells us that the last frame has taken hold. In most
    // cases unless the framerate is running close to the refresh rate of
    // the video mode, this function loop will terminate immediately and we
    // never wait. If we do wait, it is because we are running too fast!
    if ((DV.cntModeInfo.Attributes & gaHaveTripleBuffer) && (waitVRT == gaTripleBuffer))
        DD_WaitTillFlipped();

    // Now flip to the specified buffer
    BM.visibleBuf = buf;
    DD_lpVisible = bufInt->lpSurf;
    while (IDirectDrawSurface7_Flip(DD_lpPrimary,DD_lpVisible,waitVRT == gaWaitVRT ? DDFLIP_WAIT : DDFLIP_NOVSYNC) == DDERR_SURFACELOST)
        PM_doSuspendApp();

    // Now lock all flipping surfaces and store the information about
    // which surface pointer points to which physical memory location.
    for (i = 0; i < BM.numFlipBuffers; i++) {
        lpSurf = BM.flipBuffers[i]->lpSurf;
        while (IDirectDrawSurface7_Lock(lpSurf,NULL,&DD_lockBuf,LOCK_FLAGS,NULL) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        while (IDirectDrawSurface7_Unlock(lpSurf,NULL) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        BM.lpSurf[i] = lpSurf;
        BM.lpSurface[i] = DD_lockBuf.lpSurface;
        }

    // Now re-arrange the DirectDraw surface pointers for all flipping sufaces
    // since DirectDraw re-arranges all the surface pointers behind our back.
    // We do this by matching up the locked surface pointers in video memory
    // with the values previously saved for all the flipping surfaces and
    // set the buf->Drv1 values to the correct surface pointers.
    for (i = 0; i < BM.numFlipBuffers; i++) {
        for (j = 0; j < BM.numFlipBuffers; j++) {
            if (BM.flipBuffers[i]->lpSurface == BM.lpSurface[j]) {
                BM.flipBuffers[i]->lpSurf = BM.lpSurf[j];
                break;
                }
            }
        }

    // Now reset the active drawing surface since the front and back
    // surfaces actually swap internally when DirectDraw does a Flip().
    DD_lpActive = DD_activeBuf->lpSurf;
}

/****************************************************************************
RETURNS:
Pointer to the last buffer passed to FlipToBuffer
****************************************************************************/
GA_buf * NAPI DD_GetVisibleBuffer(void)
{
    return BM.visibleBuf;
}

/****************************************************************************
RETURNS:
Pointer to the last right buffer passed to FlipToStereoBuffer
****************************************************************************/
GA_buf * NAPI DD_GetVisibleRightBuffer(void)
{
    return NULL;
}

/****************************************************************************
REMARKS:
This function updates returns the flip status for the currently pending
flip operation. 1 means it has been flipped, 0 means it has not yet been
flipped.
****************************************************************************/
int NAPI DD_GetFlipStatus(void)
{
    HRESULT ret;

    do {
        if ((ret = IDirectDrawSurface7_GetFlipStatus(DD_lpVisible,DDGFS_ISFLIPDONE)) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        if (ret == DDERR_UNSUPPORTED)
            PM_fatalError("Unable to get flip status!");
        } while (ret != DD_OK && ret != DDERR_WASSTILLDRAWING);
    return (ret == DD_OK);
}

/****************************************************************************
REMARKS:
This function waits until the pending flip operation has completed and the
visible display page has changed.
****************************************************************************/
void NAPI DD_WaitTillFlipped(void)
{
    while (!DD_GetFlipStatus())
        ;
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void DD_readScan_8(
    uchar *d,
    uchar *s,
    int width)
{
    // Force DWORD source alignment
    while (width && ((uint)s & 3) != 0) {
        *d++ = *s++;
        width--;
        }
    if (width)
        memcpy(d,s,width);
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void DD_writeScan_8(
    uchar *d,
    uchar *s,
    int width)
{
    if (width >= 8) {
        // Force DWORD destination alignment
        while (((uint)d & 3) != 0) {
            *d++ = *s++;
            width--;
            }
        memcpy(d,s,width);
        }
    else
        memcpy(d,s,width);
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void DD_readScan_16(
    uchar *d,
    uchar *s,
    int width)
{
    // Force DWORD source alignment
    width *= 2;
    while (width && ((uint)s & 3) != 0) {
        *d++ = *s++;
        width--;
        }
    if (width)
        memcpy(d,s,width);
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void DD_writeScan_16(
    uchar *din,
    uchar *sin,
    int width)
{
    ushort  *d = (ushort*)din;
    ushort  *s = (ushort*)sin;

    if (width >= 4) {
        // Force DWORD destination alignment
        if (((uint)d & 1) != 0) {
            *d++ = *s++;
            width--;
            }
        rep_movsdw(d,s,width);
        }
    else
        rep_movsdw(d,s,width);
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void DD_readScan_24(
    uchar *d,
    uchar *s,
    int width)
{
    // Force DWORD source alignment
    width *= 3;
    while (width && ((uint)s & 3) != 0) {
        *d++ = *s++;
        width--;
        }
    if (width)
        memcpy(d,s,width);
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void DD_writeScan_24(
    uchar *d,
    uchar *s,
    int width)
{
    // Force DWORD source alignment
    width *= 3;
    while (((uint)s & 3) != 0) {
        *d++ = *s++;
        width--;
        }
    if (width)
        memcpy(d,s,width);
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void DD_readScan_32(
    uchar *d,
    uchar *s,
    int width)
{
    rep_movsd(d,s,width);
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void DD_writeScan_32(
    uchar *d,
    uchar *s,
    int width)
{
    rep_movsd(d,s,width);
}

/****************************************************************************
PARAMETERS:
buf - SNAP buffer to copy into system memory cache

REMARKS:
This function updates the buffer cache in system memory by copying the
contents of the video memory buffer into the system memory cache.
****************************************************************************/
void NAPI DD_UpdateCache(
    GA_buf *buf)
{
    if (buf->Flags & gaBufferCached) {
        int     width = buf->Width;
        int     height = buf->Height;
        int     srcStride = buf->Stride;
        int     dstStride = buf->CacheStride;
        uchar   *s,*d;

        DD_LockBuffer(buf);
        s = buf->Surface;
        d = buf->SurfaceCache;
        while (height--) {
            BM.bufReadScan(d,s,width);
            s += srcStride;
            d += dstStride;
            }
        DD_UnlockBuffer(buf);
        }
}

/****************************************************************************
PARAMETERS:
buf - SNAP buffer to copy from system memory cache

REMARKS:
This function updates the video memory buffer by copying the contents of
the system memory cache buffer into the video memory buffer.
****************************************************************************/
void NAPI DD_UpdateFromCache(
    GA_buf *buf)
{
    if (buf->Flags & gaBufferCached) {
        int     width = buf->Width;
        int     height = buf->Height;
        int     srcStride = buf->CacheStride;
        int     dstStride = buf->Stride;
        uchar   *s,*d;

        DD_LockBuffer(buf);
        s = buf->Surface;
        d = buf->SurfaceCache;
        while (height--) {
            BM.bufWriteScan(d,s,width);
            s += srcStride;
            d += dstStride;
            }
        DD_UnlockBuffer(buf);
        }
}

/****************************************************************************
PARAMETERS:
buf     - SNAP buffer to blit to the active surface
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
mix     - Mix code for the copy (GA_mixCodesType)

REMARKS:
This function blits the contents of the buffer to the active drawing
surface.
****************************************************************************/
void NAPI DD_BitBltBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    RECT        dst,src;
    GA_bufInt   *bufInt = (GA_bufInt*)buf;

    // Increment usage counter
    buf->UseageCount++;
    if (buf->Flags & gaBufferSysMem) {
        // Buffer is in system memory (possibly paged out into cache), or
        // the destination buffer is in system memory
        DV.drawSW.BitBltSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,mix,false);
        }
    else {
        // Buffer is a DirectDraw video memory surface
        if (DD_lpActive && DD_validSrcRops[mix]) {
            src.left = srcLeft;
            src.top = srcTop;
            src.right = srcLeft + width;
            src.bottom = srcTop + height;
            UNLOCK_SURFACE();
            if (mix == GA_REPLACE_MIX && !DD_activeBuf->clipper) {
                while (IDirectDrawSurface7_BltFast(DD_lpActive,dstLeft,dstTop,bufInt->lpSurf,
                        &src,DDBLTFAST_NOCOLORKEY | DDBLTFAST_WAIT) == DDERR_SURFACELOST)
                    PM_doSuspendApp();
                }
            else {
                dst.left = dstLeft;
                dst.top = dstTop;
                dst.right = dstLeft + width;
                dst.bottom = dstTop + height;
                DD_bltFx.dwROP = DD_dwSrcRopCodes[mix];
                while (IDirectDrawSurface7_Blt(DD_lpActive,&dst,bufInt->lpSurf,
                        &src,DDBLT_ROP | DDBLT_WAIT,&DD_bltFx) == DDERR_SURFACELOST)
                    PM_doSuspendApp();
                }
            }
        else {
            // Unsupported ROP so do it in software.
            DD_LockBuffer(buf);
            DV.drawSW.BitBltSys(
                buf->Surface,buf->Stride,srcLeft,srcTop,
                width,height,dstLeft,dstTop,mix,false);
            DD_UnlockBuffer(buf);
            }
        }
}

/****************************************************************************
PARAMETERS:
buf     - SNAP buffer to blit to the active surface
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 code for the copy (GA_rop3CodesType)

REMARKS:
This function blits the contents of the buffer to the active drawing
surface with a mono pattern applied.
****************************************************************************/
void NAPI DD_BitBltPattBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    // Increment usage counter
    buf->UseageCount++;
    if (buf->Flags & (gaBufferCached | gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache).
        // We also end up here if the buffer is in video memory but the
        // buffer is cached in system memory. It is faster to do software
        // rendering from the buffer cache memory than doing software blits
        // within video memory.
        DV.drawSW.BitBltPattSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,rop3,false);
        }
    else {
        // Buffer is a DirectDraw buffer. Since we don't support this
        // in hardware with DirectDraw we punt to software.
        DD_LockBuffer(buf);
        DV.drawSW.BitBltPattSys(
            buf->Surface,buf->Stride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,rop3,false);
        DD_UnlockBuffer(buf);
        }
}

/****************************************************************************
PARAMETERS:
buf     - SNAP buffer to blit to the active surface
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 code for the copy (GA_rop3CodesType)

REMARKS:
This function blits the contents of the buffer to the active drawing
surface with a color pattern applied.
****************************************************************************/
void NAPI DD_BitBltColorPattBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    // Increment usage counter
    buf->UseageCount++;
    if (buf->Flags & (gaBufferCached | gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache).
        // We also end up here if the buffer is in video memory but the
        // buffer is cached in system memory. It is faster to do software
        // rendering from the buffer cache memory than doing software blits
        // within video memory.
        DV.drawSW.BitBltColorPattSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,rop3,false);
        }
    else {
        // Buffer is a DirectDraw buffer. Since we don't support this
        // in hardware with DirectDraw we punt to software.
        DD_LockBuffer(buf);
        DV.drawSW.BitBltColorPattSys(
            buf->Surface,buf->Stride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,rop3,false);
        DD_UnlockBuffer(buf);
        }
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to blit to the active surface
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color for the blit

REMARKS:
This function blits the contents of the buffer to the active drawing
surface with source transparency.
****************************************************************************/
void NAPI DD_SrcTransBltBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    RECT        src,dst;
    DDCOLORKEY  ddck;
    GA_bufInt   *bufInt = (GA_bufInt*)buf;

    // Increment usage counter
    buf->UseageCount++;
    if ((!HAVE_HWSRCTRANSBLT() && (buf->Flags & gaBufferCached)) || (buf->Flags & gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache).
        // We also end up here if the buffer is in video memory but the
        // hardware does not support this function and we have a buffer
        // cache. It is faster to do software rendering from the buffer
        // cache memory than doing software blits within video memory.
        DV.drawSW.SrcTransBltSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,mix,transparent,false);
        }
    else {
        // Buffer is a DirectDraw video memory surface
        if (DD_lpActive && (mix == GA_REPLACE_MIX && (DD_caps.dwCKeyCaps & DDCKEYCAPS_SRCBLT))) {
            src.left = srcLeft;
            src.top = srcTop;
            src.right = srcLeft + width;
            src.bottom = srcTop + height;
            UNLOCK_SURFACE();
            if (bufInt->srcColorKey != transparent) {
                // We avoid re-loading the color key whenever possible
                ddck.dwColorSpaceLowValue  = transparent;
                ddck.dwColorSpaceHighValue = transparent;
                while (IDirectDrawSurface7_SetColorKey(bufInt->lpSurf,DDCKEY_SRCBLT,&ddck) == DDERR_SURFACELOST)
                    PM_doSuspendApp();
                bufInt->srcColorKey = transparent;
                }
            if (!DD_activeBuf->clipper) {
                while (IDirectDrawSurface7_BltFast(DD_lpActive,dstLeft,dstTop,bufInt->lpSurf,
                        &src,DDBLTFAST_SRCCOLORKEY | DDBLTFAST_WAIT) == DDERR_SURFACELOST)
                    PM_doSuspendApp();
                }
            else {
                dst.left = dstLeft;
                dst.top = dstTop;
                dst.right = dstLeft + width;
                dst.bottom = dstTop + height;
                while (IDirectDrawSurface7_Blt(DD_lpActive,&dst,bufInt->lpSurf,
                        &src,DDBLT_KEYSRC | DDBLT_WAIT,NULL) == DDERR_SURFACELOST)
                    PM_doSuspendApp();
                }
            }
        else {
            // Unsupported ROP (or not blit hardware) so do it in software.
            DD_LockBuffer(buf);
            DV.drawSW.SrcTransBltSys(
                buf->Surface,buf->Stride,srcLeft,srcTop,
                width,height,dstLeft,dstTop,mix,transparent,false);
            DD_UnlockBuffer(buf);
            }
        }
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to blit to the active surface
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color for the blit

REMARKS:
This function blits the contents of the buffer to the active drawing
surface with destination transparency.
****************************************************************************/
void NAPI DD_DstTransBltBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    RECT        src,dst;
    DDCOLORKEY  ddck;
    GA_bufInt   *bufInt = (GA_bufInt*)buf;

    // Increment usage counter
    buf->UseageCount++;
    if ((!HAVE_HWDSTTRANSBLT() && (buf->Flags & gaBufferCached)) || (buf->Flags & gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache).
        // We also end up here if the buffer is in video memory but the
        // hardware does not support this function and we have a buffer
        // cache. It is faster to do software rendering from the buffer
        // cache memory than doing software blits within video memory.
        DV.drawSW.DstTransBltSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,mix,transparent,false);
        }
    else {
        // Buffer is a DirectDraw video memory surface
        if (DD_lpActive && mix == GA_REPLACE_MIX && (DD_caps.dwCKeyCaps & DDCKEYCAPS_DESTBLT)) {
            src.left = srcLeft;
            src.top = srcTop;
            src.right = srcLeft + width;
            src.bottom = srcTop + height;
            UNLOCK_SURFACE();
            if (bufInt->dstColorKey != transparent) {
                // We avoid re-loading the color key whenever possible
                ddck.dwColorSpaceLowValue  = transparent;
                ddck.dwColorSpaceHighValue = transparent;
                while (IDirectDrawSurface7_SetColorKey(bufInt->lpSurf,DDCKEY_DESTBLT,&ddck) == DDERR_SURFACELOST)
                    PM_doSuspendApp();
                bufInt->dstColorKey = transparent;
                }
            if (!DD_activeBuf->clipper) {
                while (IDirectDrawSurface7_BltFast(DD_lpActive,dstLeft,dstTop,bufInt->lpSurf,
                        &src,DDBLTFAST_DESTCOLORKEY | DDBLTFAST_WAIT) == DDERR_SURFACELOST)
                    PM_doSuspendApp();
                }
            else {
                dst.left = dstLeft;
                dst.top = dstTop;
                dst.right = dstLeft + width;
                dst.bottom = dstTop + height;
                while (IDirectDrawSurface7_Blt(DD_lpActive,&dst,bufInt->lpSurf,
                        &src,DDBLT_KEYDEST | DDBLT_WAIT,NULL) == DDERR_SURFACELOST)
                    PM_doSuspendApp();
                }
            }
        else {
            // Unsupported ROP so do it in software.
            DD_LockBuffer(buf);
            DV.drawSW.DstTransBltSys(
                buf->Surface,buf->Stride,srcLeft,srcTop,
                width,height,dstLeft,dstTop,mix,transparent,false);
            DD_UnlockBuffer(buf);
            }
        }
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to blit to the active surface
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
doClip      - True if the blit should be clipped
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle
mix         - Mix code for the copy (GA_mixCodesType)

REMARKS:
This function blits the contents of the buffer to the active drawing
surface with stretching.
****************************************************************************/
void NAPI DD_StretchBltBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 mix)
{
    // Increment usage counter
    buf->UseageCount++;
    if (buf->Flags & (gaBufferCached | gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache).
        // We also end up here if the buffer is in video memory but the
        // buffer is cached in system memory. It is faster to do software
        // rendering from the buffer cache memory than doing software blits
        // within video memory.
        DV.drawSW.StretchBltSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,
            doClip,clipLeft,clipTop,clipRight,clipBottom,mix,false);
        }
    else {
        // Buffer is a DirectDraw buffer. Since we don't support this
        // in hardware with DirectDraw we punt to software.
        DD_LockBuffer(buf);
        DV.drawSW.StretchBltSys(
            buf->Surface,buf->Stride,srcLeft,srcTop,
            srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,
            doClip,clipLeft,clipTop,clipRight,clipBottom,mix,false);
        DD_UnlockBuffer(buf);
        }
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to blit to the active surface
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
planeMask   - 8-bit plane mask to use for the copy

REMARKS:
This function blits the contents of the buffer to the active drawing
surface.
****************************************************************************/
void NAPI DD_BitBltPlaneMaskedBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_uint32 planeMask)
{
    // Increment usage counter
    buf->UseageCount++;
    if (buf->Flags & (gaBufferCached | gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache).
        // We also end up here if the buffer is in video memory but the
        // buffer is cached in system memory. It is faster to do software
        // rendering from the buffer cache memory than doing software blits
        // within video memory.
        DV.drawSW.BitBltPlaneMaskedSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,planeMask,false);
        }
    else {
        // Buffer is a DirectDraw buffer. Since we don't support this
        // in hardware with DirectDraw we punt to software.
        DD_LockBuffer(buf);
        DV.drawSW.BitBltPlaneMaskedSys(
            buf->Surface,buf->Stride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,planeMask,false);
        DD_UnlockBuffer(buf);
        }
}

/****************************************************************************
PARAMETERS:
blend       - SNAP alpha blend function enum

REMARKS:
This function returns the D3D alpha blend function corresponding to the
SNAP alpha blend function enum.
****************************************************************************/
static DWORD DD_getD3DBlend(
    GA_blendFuncType blend,
    ibool isDstBlend)
{
    switch (blend) {
        case gaBlendZero:                       return D3DBLEND_ZERO;
        case gaBlendOne:                        return D3DBLEND_ONE;
        case gaBlendSrcColor:                   return D3DBLEND_SRCCOLOR;
        case gaBlendOneMinusSrcColor:           return D3DBLEND_INVSRCCOLOR;
        case gaBlendSrcAlpha:                   return D3DBLEND_SRCALPHA;
        case gaBlendOneMinusSrcAlpha:           return D3DBLEND_INVSRCALPHA;
        case gaBlendDstAlpha:                   return D3DBLEND_DESTALPHA;
        case gaBlendOneMinusDstAlpha:           return D3DBLEND_INVDESTALPHA;
        case gaBlendDstColor:                   return D3DBLEND_DESTCOLOR;
        case gaBlendOneMinusDstColor:           return D3DBLEND_INVDESTCOLOR;
        case gaBlendSrcAlphaSaturate:           return D3DBLEND_SRCALPHASAT;
        case gaBlendConstantColor:              return D3DBLEND_SRCCOLOR;
        case gaBlendOneMinusConstantColor:      return D3DBLEND_INVSRCCOLOR;
        case gaBlendConstantAlpha:              return D3DBLEND_SRCALPHA;
        case gaBlendOneMinusConstantAlpha:      return D3DBLEND_INVSRCALPHA;
        case gaBlendSrcAlphaFast:
        case gaBlendConstantAlphaFast:
        default:                                return isDstBlend ? D3DBLEND_INVSRCALPHA : D3DBLEND_SRCALPHA;
        }
}

/****************************************************************************
REMARKS:
This function renders a single textured quad from the input source
bitmap.
****************************************************************************/
static void DD_draw3DTile(
    GA_bufInt *bufInt,
    GA_bltFx *fx,
    int iTexPF,
    RECT *src,
    RECT *tex,
    float dstLeft,
    float dstTop,
    float dstWidth,
    float dstHeight,
    float fullTexWidth,
    float fullTexHeight)
{
    HRESULT                 ret;
    D3D_vertex              v[4];
    LPDIRECTDRAWSURFACE7    lpTexture = DD_lpTexture[iTexPF];

    // Load the texture with the source pixels
    if ((fx->Flags & gaBltBlend) && ((fx->SrcBlendFunc == gaBlendConstantColor)
            || (fx->SrcBlendFunc == gaBlendOneMinusConstantColor))) {
        DD_bltFx.u5.dwFillColor = fx->ConstColor;
        while ((ret = IDirectDrawSurface7_Blt(lpTexture,tex,NULL,NULL,
                DDBLT_WAIT | DDBLT_COLORFILL,&DD_bltFx)) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        }
    else if (!(bufInt->p.Flags & gaBufferSysMem)) {
        // Buffer is in video memory, so use DirectX to blt
        // across to the texture surface
        while ((ret = IDirectDrawSurface7_Blt(lpTexture,tex,bufInt->lpSurf,src,
                DDBLT_WAIT,&DD_bltFx)) == DDERR_SURFACELOST)
            PM_doSuspendApp();

        // Fill the constant alphe values in the texture if necessary
        if (fx->SrcBlendFunc == gaBlendConstantAlpha || fx->SrcBlendFunc == gaBlendOneMinusConstantAlpha) {
            uchar   *d;
            int     alpha = fx->ConstAlpha;
            int     texHeight = fullTexHeight;

            // Compute surface addresses
            while (IDirectDrawSurface7_Lock(lpTexture,NULL,&DD_lockBuf,LOCK_FLAGS | DDLOCK_NOSYSLOCK,NULL) == DDERR_SURFACELOST)
                PM_doSuspendApp();
            d = (uchar*)DD_lockBuf.lpSurface + (DV.cntModeInfo.PixelFormat.AlphaPosition / 8);

            // Now write each alpha byte directly into texture surface
            while (texHeight--) {
                int width = fullTexWidth;
                uchar *dt = d;
                while (width--) {
                    *dt = (uchar)alpha;
                    dt += 4;
                    }
                d += DD_lockBuf.u1.lPitch;
                }
            while (IDirectDrawSurface7_Unlock(lpTexture, NULL) == DDERR_SURFACELOST)
                PM_doSuspendApp();
            }
        }
    else {
        // Buffer is in system memory, so lock the texture surface
        // and manually copy/color convert the pixel data into
        // the texture.
        uchar       *s,*d;
        int         srcBytesPerPixel = (bufInt->p.BitsPerPixel+1) / 8;
        int         srcStride = bufInt->p.Stride;
        int         srcByteWidth = (src->right - src->left) * srcBytesPerPixel;
        int         srcHeight = src->bottom - src->top;
        GA_palette  *pal;
        int         i;
        N_uint32    alpha = fx->ConstAlpha;
        ibool       constblend = false;
        GA_color    transVec[256];
        GA_color    *translate,*t;

        // Pre-determine logic to overwrite alpha channel with constant blend value
        if ((fx->SrcBlendFunc == gaBlendConstantAlpha) ||
                (fx->SrcBlendFunc == gaBlendOneMinusConstantAlpha) ||
                (fx->SrcBlendFunc == gaBlendConstantAlphaFast))
            constblend = true;

        // Build color translation vector if necessary
        if (fx->BitsPerPixel == 8 || ((fx->BitsPerPixel == 16) && (fx->PixelFormat->AlphaPosition == 8))) {
            if ((DD_iPrimaryPF == pfARGB32) && (fx->Flags & gaBltTranslateVec) && !constblend) {
                // Use passed in translate vector
                translate = fx->TranslateVec;
                }
            else {
                // Build translate vector from scratch. First fix up the alpha
                // channel information that will be packed into the translate
                // vector. If not constant alpha, make it zero so we can or in
                // the alpha channel data for each pixel. Otherwise the alpha
                // channel will be in the translate vector automatically.
                if (!constblend)
                    alpha = 0;
                t = translate = transVec;
                pal = fx->SrcPalette;
                for (i = 0; i < 256; i++, pal++)
                    *t++ = PACK_COLOR_ARGB(alpha,pal->Red,pal->Green,pal->Blue);
                }
            }

        // TODO: Add cases to convert from 15/16/24/32 bit with constant alpha to a 32-bit
        //       ARGB texture, which we can then blit with constant alpha applied. If we
        //       are just doing color conversion, the native texture format can be used
        //       to do it directly in HW.

        // Compute surface addresses
        while (IDirectDrawSurface7_Lock(lpTexture,NULL,&DD_lockBuf,LOCK_FLAGS | DDLOCK_NOSYSLOCK,NULL) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        s = (uchar*)bufInt->p.Surface + (src->top * srcStride) + (src->left * srcBytesPerPixel);
        d = (uchar*)DD_lockBuf.lpSurface + (tex->top * DD_lockBuf.u1.lPitch) + (tex->left * DD_lockBuf.u4.ddpfPixelFormat.u1.dwRGBBitCount/8);

        // Do the actual surface copies
        if (!constblend && (!(fx->Flags & gaBltConvert) || memcmp(fx->PixelFormat,&DD_pixelFormats[iTexPF],sizeof(DD_pixelFormats[iTexPF])) == 0)) {
            // Direct copy data to texture surface when the pixel format is the same
            while (srcHeight--) {
                int width = srcByteWidth;
                if (width >= 8) {
                    // Force DWORD destination alignment
                    while (((uint)d & 3) != 0) {
                        *d++ = *s++;
                        width--;
                        }
                    memcpy(d,s,width);
                    }
                else
                    memcpy(d,s,width);
                s += srcStride;
                d += DD_lockBuf.u1.lPitch;
                }
            }
        else if (constblend && (!(fx->Flags & gaBltConvert) || memcmp(fx->PixelFormat,&DD_pixelFormats[iTexPF],sizeof(DD_pixelFormats[iTexPF])) == 0)) {
            // Input pixel format is same as frame buffer with constant alpha added
            while (srcHeight--) {
                int width = srcByteWidth / srcBytesPerPixel;
                ulong alphamask = (fx->Flags & gaBltConvert)
                    ? fx->PixelFormat->AlphaMask << fx->PixelFormat->AlphaPosition
                    : DV.cntModeInfo.PixelFormat.AlphaMask << DV.cntModeInfo.PixelFormat.AlphaPosition;
                ulong alphashifted = (fx->Flags & gaBltConvert)
                    ? alpha << fx->PixelFormat->AlphaPosition
                    : alpha << DV.cntModeInfo.PixelFormat.AlphaPosition;
                if (srcBytesPerPixel == 4) {
                #if 0 // Deliberately omits loading RGB values...
                    uchar *dt8  = (uchar*)d;
                    // Write every 4th BYTE instead of reading and writing DWORDs
                    if (DV.cntModeInfo.PixelFormat.AlphaPosition == 0) {
                        width = srcByteWidth;
                        while (width--) {
                            *dt8 = (uchar)alpha;
                            dt8 += 4;
                            }
                        }
                    else if (DV.cntModeInfo.PixelFormat.AlphaPosition == 24) {
                        width = srcByteWidth;
                        while (width--) {
                            dt8 += 3;
                            *dt8++ = (uchar)alpha;
                            }
                        }
                #else
                    ulong *st32 = (ulong*)s;
                    ulong *dt32 = (ulong*)d;
                    while (width--) {
                        ulong p = *st32++;
                        *dt32++ = (p & ~alphamask) | alphashifted;
                        }
                #endif
                }
                else if (srcBytesPerPixel == 2) {
                    ushort *st16 = (ushort*)s;
                    ushort *dt16 = (ushort*)d;
                    while (width--) {
                        ushort p = *st16++;
                        *dt16++ = (p & ~alphamask) | alphashifted;
                        }
                    }
                else {
                    uchar *st8 = (uchar*)s;
                    uchar *dt8 = (uchar*)d;
                    while (width--) {
                        uchar p = *st8++;
                        *dt8++ = (p & ~alphamask) | alphashifted;
                        }
                    }
                s += srcStride;
                d += DD_lockBuf.u1.lPitch;
                }
            }
        else if (fx->BitsPerPixel == 8) {
            // 8-bpp input pixel format with constant alpha (dest is 32bit ARGB)
            while (srcHeight--) {
                int width = srcByteWidth;
                uchar *st = (uchar*)s;
                ulong *dt = (ulong*)d;
                while (width--)
                    *dt++ = translate[*st++];
                s += srcStride;
                d += DD_lockBuf.u1.lPitch;
                }
            }
        else if (fx->BitsPerPixel == 24) {
            // 24-bpp input pixel format with constant alpha (dest is 32bit ARGB)
            while (srcHeight--) {
                int width = srcByteWidth;
                uchar *st = (uchar*)s;
                ulong *dt = (ulong*)d;
                if (fx->PixelFormat->RedPosition == 0) {
                    while (width--) {
                        int R = *st++;
                        int G = *st++;
                        int B = *st++;
                        *dt++ = PACK_COLOR_ARGB(alpha,R,G,B);
                        }
                    }
                else {
                    while (width--) {
                        int B = *st++;
                        int G = *st++;
                        int R = *st++;
                        *dt++ = PACK_COLOR_ARGB(alpha,R,G,B);
                        }
                    }
                s += srcStride;
                d += DD_lockBuf.u1.lPitch;
                }
            }
        else if (fx->BitsPerPixel == 32) {
            // 32-bpp input pixel format with constant alpha (dest is 32bit ARGB)
            // Write every 4th BYTE instead of reading and writing DWORDs
            while (srcHeight--) {
                int width = srcByteWidth / srcBytesPerPixel;
                uchar *st = (uchar*)s;
                uchar *dt = (uchar*)d;
                if (fx->PixelFormat->AlphaPosition == 0) {
                    while (width--) {
                        *dt = (uchar)alpha;
                        st++;
                        *dt++ = *st++;
                        *dt++ = *st++;
                        *dt++ = *st++;
                        }
                    }
                else {
                    while (width--) {
                        *dt++ = *st++;
                        *dt++ = *st++;
                        *dt++ = *st++;
                        *dt++ = (uchar)alpha;
                        st++;
                        }
                    }
                s += srcStride;
                d += DD_lockBuf.u1.lPitch;
                }
            }
        else if (fx->BitsPerPixel == 16 && fx->PixelFormat->AlphaPosition == 8 && constblend) {
            // 8-bpp+alpha input pixel format with constant alpha (dest is 32bit ARGB)
            while (srcHeight--) {
                int width = srcByteWidth;
                uchar *st = (uchar*)s;
                ulong *dt = (ulong*)d;
                while (width--) {
                    *dt++ = translate[*st++];
                    st++;
                    }
                s += srcStride;
                d += DD_lockBuf.u1.lPitch;
                }
            }
        else if (fx->BitsPerPixel == 16 && fx->PixelFormat->AlphaPosition == 8) {
            // 8-bpp+alpha input pixel format with alpha channel (dest is 32bit ARGB)
            while (srcHeight--) {
                int width = srcByteWidth;
                uchar *st = (uchar*)s;
                ulong *dt = (ulong*)d;
                while (width--) {
                    *dt++ = translate[st[0]] | ((ulong)st[1] << 24);
                    st += 2;
                    }
                s += srcStride;
                d += DD_lockBuf.u1.lPitch;
                }
            }
        else if (fx->BitsPerPixel == 16 && fx->PixelFormat->AlphaMask == 0 && constblend) {
            // 16bpp 565 RGB input pixel format with constant alpha added (dest is 32bit ARGB)
            while (srcHeight--) {
                int width = srcByteWidth;
                ushort *st = (ushort*)s;
                ulong *dt = (ulong*)d;
                while (width--) {
                    ushort p = *st++;
                    ushort R = p & 0x001F;
                    ushort G = (p & 0x07E0) >> 5;
                    ushort B = (p & 0xF800) >> 11;
                    *dt++ = PACK_COLOR_ARGB(alpha,R,G,B);
                    }
                s += srcStride;
                d += DD_lockBuf.u1.lPitch;
                }
            }
        else if ((fx->BitsPerPixel == 16 || fx->BitsPerPixel == 15) && constblend) {
            // 15bpp 555 RGB input pixel format with constant alpha added (dest is 32bit ARGB)
            // 16bpp 1555 ARGB input pixel format with constant alpha added (dest is 32bit ARGB)
            while (srcHeight--) {
                int width = srcByteWidth;
                ushort *st = (ushort*)s;
                ulong *dt = (ulong*)d;
                while (width--) {
                    ushort p = *st++;
                    ushort R = p & 0x001F;
                    ushort G = (p & 0x03E0) >> 5;
                    ushort B = (p & 0x7C00) >> 10;
                    *dt++ = PACK_COLOR_ARGB(alpha,R,G,B);
                    }
                s += srcStride;
                d += DD_lockBuf.u1.lPitch;
                }
            }
        while (IDirectDrawSurface7_Unlock(lpTexture, NULL) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        }

    // Set raster positions
    v[0].x = dstLeft;                   v[0].y = dstTop;
    v[1].x = dstLeft+dstWidth;          v[1].y = dstTop;
    v[2].x = dstLeft+dstWidth;          v[2].y = dstTop+dstHeight;
    v[3].x = dstLeft;                   v[3].y = dstTop+dstHeight;

    // Set depth coords
    v[0].z = v[1].z = v[2].z = v[3].z = 0.0f;
    v[0].rhw = v[1].rhw = v[2].rhw = v[3].rhw = 1.0f;

    // Set texcoords
    v[0].tu = 0.0f;                     v[0].tv = 0.0f;
    v[1].tu = dstWidth/fullTexWidth;    v[1].tv = 0.0f;
    v[2].tu = dstWidth/fullTexWidth;    v[2].tv = dstHeight/fullTexHeight;
    v[3].tu = 0.0f;                     v[3].tv = dstHeight/fullTexHeight;

    // Adjust texcoords if mirror effect enabled
    if (fx->Flags & gaBltFlipX) {
        v[1].tu *= -1.0f;
        v[2].tu *= -1.0f;
        v[0].tu -= 1.0f/fullTexWidth;
        v[1].tu -= 1.0f/fullTexWidth;
        v[2].tu -= 1.0f/fullTexWidth;
        v[3].tu -= 1.0f/fullTexWidth;
        }
    if (fx->Flags & gaBltFlipY) {
        v[2].tv *= -1.0f;
        v[3].tv *= -1.0f;
        v[0].tv -= 1.0f/fullTexHeight;
        v[1].tv -= 1.0f/fullTexHeight;
        v[2].tv -= 1.0f/fullTexHeight;
        v[3].tv -= 1.0f/fullTexHeight;
        }

    // Render source onto destination as textured quad
    IDirect3DDevice7_BeginScene(DD_lpD3DDevice);
    IDirect3DDevice7_DrawPrimitive(DD_lpD3DDevice, D3DPT_TRIANGLEFAN, D3DFVF_XYZRHW | D3DFVF_TEX1, &v, 4, 0);
    while ((ret = IDirect3DDevice7_EndScene(DD_lpD3DDevice)) == DDERR_SURFACELOST)
        PM_doSuspendApp();
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to blit to the active surface
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
fx          - Effects to apply to the blit operation

REMARKS:
This function blits the contents of the buffer to the active drawing
surface with optional effects.
****************************************************************************/
void NAPI DD_BitBltFxBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    GA_bltFx *fx)
{
    GA_bufInt   *bufInt = (GA_bufInt*)buf;
    RECT        src,tex;
    int         x,y,dx,dy,iTexPF,texWidth,texHeight;

    // Increment usage counter
    buf->UseageCount++;

    // TODO: Implement the following:
    //
    // 2. Color keying
    // 3. Clipping

    // Buffer is a DirectDraw buffer. Test if DirectDraw supports the
    // corresponding effects blit caps in hardware, and blit to the
    // active buffer if supported.
    if (DD_lpActive && DD_BitBltFxTestInt(fx,&iTexPF)) {
        UNLOCK_SURFACE();

        // TODO: Re-allocate needed texture if we are trying to do a stretch blit
        //       operation that needs a full texture. If that fails, punt to
        //       software for all future blits.

        // Setup D3D texture state
        IDirect3DDevice7_SetTexture(DD_lpD3DDevice, 0, DD_lpTexture[iTexPF]);
        IDirect3DDevice7_SetTextureStageState(DD_lpD3DDevice, 0, D3DTSS_ADDRESSU, (fx->Flags & gaBltFlipX) ? D3DTADDRESS_MIRROR : D3DTADDRESS_CLAMP);
        IDirect3DDevice7_SetTextureStageState(DD_lpD3DDevice, 0, D3DTSS_ADDRESSV, (fx->Flags & gaBltFlipY) ? D3DTADDRESS_MIRROR : D3DTADDRESS_CLAMP);
        IDirect3DDevice7_SetTextureStageState(DD_lpD3DDevice, 0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
        IDirect3DDevice7_SetTextureStageState(DD_lpD3DDevice, 0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
        IDirect3DDevice7_SetTextureStageState(DD_lpD3DDevice, 0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
        IDirect3DDevice7_SetTextureStageState(DD_lpD3DDevice, 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);

        // Enable interpolative filtering option, if selected
        if (fx->Flags & (gaBltStretchXInterp | gaBltStretchYInterp)) {
            IDirect3DDevice7_SetTextureStageState(DD_lpD3DDevice, 0, D3DTSS_MINFILTER, D3DTFN_LINEAR);
            IDirect3DDevice7_SetTextureStageState(DD_lpD3DDevice, 0, D3DTSS_MIPFILTER, D3DTFG_LINEAR);
            IDirect3DDevice7_SetTextureStageState(DD_lpD3DDevice, 0, D3DTSS_MAGFILTER, D3DTFG_LINEAR);
            }
        else {
            IDirect3DDevice7_SetTextureStageState(DD_lpD3DDevice, 0, D3DTSS_MINFILTER, D3DTFN_POINT);
            IDirect3DDevice7_SetTextureStageState(DD_lpD3DDevice, 0, D3DTSS_MIPFILTER, D3DTFN_POINT);
            IDirect3DDevice7_SetTextureStageState(DD_lpD3DDevice, 0, D3DTSS_MAGFILTER, D3DTFG_POINT);
            }

        // Enable alpha blending effect, if selected
        if ((fx->Flags & gaBltBlend) && (fx->SrcBlendFunc != gaBlendNone) && (fx->DstBlendFunc != gaBlendNone)) {
            IDirect3DDevice7_SetRenderState(DD_lpD3DDevice, D3DRENDERSTATE_SRCBLEND,  DD_getD3DBlend(fx->SrcBlendFunc,false));
            IDirect3DDevice7_SetRenderState(DD_lpD3DDevice, D3DRENDERSTATE_DESTBLEND, DD_getD3DBlend(fx->DstBlendFunc,true));
            IDirect3DDevice7_SetRenderState(DD_lpD3DDevice, D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
            }
        else
            IDirect3DDevice7_SetRenderState(DD_lpD3DDevice, D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);

        // Enable dithering, if selected
        IDirect3DDevice7_SetRenderState(DD_lpD3DDevice, D3DRENDERSTATE_DITHERENABLE,(fx->Flags & gaBltDither) ? TRUE : FALSE);

        // Perform the texture mapping
        texWidth = DD_textureSizeX[iTexPF];
        texHeight = DD_textureSizeY[iTexPF];
        if (srcWidth <= texWidth && srcHeight <= texHeight) {
            // Render the entire source buffer in one shot
            src.left = srcLeft;
            src.right = srcLeft + srcWidth;
            src.top = srcTop;
            src.bottom = srcTop + srcHeight;
            tex.left = tex.top = 0;
            tex.right = srcWidth;
            tex.bottom = srcHeight;
            DD_draw3DTile(bufInt,fx,iTexPF,&src,&tex,
                dstLeft,dstTop,dstWidth,dstHeight,
                texWidth,texHeight);
            }
        else {
            // Perform texturing in tiled blocks
            for (y = 0; y < srcHeight; y += texWidth) {
                for (x = 0; x < srcWidth; x += texHeight) {
                    // Adjust source rect for incremental tile size
                    tex.left = tex.top = 0;
                    tex.right = texWidth;
                    tex.bottom = texHeight;
                    src.left = srcLeft + x;
                    src.right = src.left + texWidth;
                    if (src.right > srcLeft+srcWidth) {
                        src.right = srcLeft+srcWidth;
                        tex.right = src.right - src.left;
                        if (fx->Flags & gaBltFlipX) {
                            tex.left = texWidth - tex.right;
                            tex.right = texWidth;
                            }
                        }
                    src.top = srcTop + y;
                    src.bottom = src.top + texHeight;
                    if (src.bottom > srcTop+srcHeight) {
                        src.bottom = srcTop+srcHeight;
                        tex.bottom = src.bottom - src.top;
                        if (fx->Flags & gaBltFlipY) {
                            tex.top = texHeight - tex.bottom;
                            tex.bottom = texHeight;
                            }
                        }
                    dx = tex.right - tex.left;
                    dy = tex.bottom - tex.top;

                    // Render the tile from the source buffer
                    DD_draw3DTile(bufInt,fx,iTexPF,&src,&tex,
                        (fx->Flags & gaBltFlipX) ? dstLeft+dstWidth-(x+dx) : dstLeft+x,
                        (fx->Flags & gaBltFlipY) ? dstTop+dstHeight-(y+dy) : dstTop+y,
                        dx,dy,texWidth,texHeight);
                    }
                }
            }

        // Unbined texture stage 0
        IDirect3DDevice7_SetTexture(DD_lpD3DDevice, 0, NULL);
        }
    else if (buf->Flags & gaBufferCached) {
        // Buffer is a DirectDraw buffer in video memory, but it has
        // a system memory cache so we can emulate from there faster.
        DV.drawSW.BitBltFxSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
        }
    else {
        // Buffer is a DirectDraw buffer in video memory, so emulate
        // the functionality in software.
        DD_LockBuffer(buf);
        DV.drawSW.BitBltFxSys(
            buf->Surface,buf->Stride,srcLeft,srcTop,
            srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
        DD_UnlockBuffer(buf);
        }
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to draw to
left        - Left coordinate of the rectangle to draw
top         - Top coordinate of the rectangle to draw
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
color       - Color to fill the rectangle with
mix         - Mix to fill the rectangle with

REMARKS:
This function does a solid fill to the specified destination buffer without
requiring the active destination buffer to be changed.
****************************************************************************/
void NAPI DD_DrawRectBuf(
    GA_buf *buf,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    GA_color color,
    N_int32 mix)
{
    RECT        dst;
    GA_bufInt   *bufInt = (GA_bufInt*)buf;

    // Increment usage counter
    buf->UseageCount++;
    if ((!HAVE_HWRECT() && (buf->Flags & gaBufferCached)) || (buf->Flags & gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache).
        // We also end up here if the buffer is in video memory but the
        // hardware does not support this function and we have a buffer
        // cache. It is faster to do software rendering from the buffer
        // cache memory than doing software blits within video memory.
        BM.ref2d->DrawRectExtSW(buf->SurfaceCache,buf->CacheStride,left,top,width,height,color,mix);
        }
    else {
        // Buffer is a DirectDraw video memory surface
        if (DD_lpActive && DD_validPatRops[mix]) {
            dst.left = left;
            dst.top = top;
            dst.right = left + width;
            dst.bottom = top + height;
            DD_bltFx.u5.dwFillColor = color;
            UNLOCK_SURFACE();
            if (mix == GA_REPLACE_MIX) {
                while (IDirectDrawSurface7_Blt(bufInt->lpSurf,&dst,NULL,NULL,
                        DDBLT_COLORFILL | DDBLT_WAIT,&DD_bltFx) == DDERR_SURFACELOST)
                    PM_doSuspendApp();
                }
            else {
                DD_bltFx.dwROP = DD_dwPatRopCodes[mix];
                while (IDirectDrawSurface7_Blt(bufInt->lpSurf,&dst,NULL,NULL,
                        DDBLT_COLORFILL | DDBLT_ROP | DDBLT_WAIT,&DD_bltFx) == DDERR_SURFACELOST)
                    PM_doSuspendApp();
                }
            }
        else {
            // Unsupported ROP so do it in software.
            DD_LockBuffer(buf);
            BM.ref2d->DrawRectExtSW(buf->Surface,buf->Stride,left,top,width,height,color,mix);
            DD_UnlockBuffer(buf);
            }
        }
}

