/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for the internal SciTech SNAP DirectDraw driver.
*
****************************************************************************/

#ifndef __INTERNAL_H
#define __INTERNAL_H

#include "snap/graphics/driver.h"
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#define NONAMELESSUNION
#define DIRECTDRAW_VERSION 0x0700
#include "pm/ddraw.h"
#define DIRECT3D_VERSION   0x0700
#include "pm/d3d.h"
#include "snap/common/list.h"
#include "snap/common/mem.h"

/*---------------------- Macro and type definitions -----------------------*/

#pragma pack(1)

// Pixel address arithmetic macros

#define LIN_X_ADDR(x,s)             ((s) <= 0 ? ((x) >> (-s)) : ((s) == 3 ? ((x) * 3) : ((x) << (s))))
#define BYTESPERLINE(w,p)           ((p) <= 4 ? ((w+7) / (8/(p))) : ((w) * ((p+7)/8)))
#define PIXELSPERLINE(w,p)          ((p) <= 4 ? ((w) * (8/(p))) : ((w) / ((p+7)/8)))

// Flags passed to the IDirectDrawSurface7_Lock() function

#define LOCK_FLAGS                  (DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR)

/****************************************************************************
REMARKS:
Internal version of the GA_clipper structure that contains internal variables
specific to the DirectDraw clip list manager implementation.

MEMBERS:
lpClipper       - Pointer to DirectDraw clipper object
hwnd            - Window handle associated with this clipper object
bufSize         - Size of the clip list buffer currently allocated
lpClipList      - Pointer to the allocated clip list buffer
bufSizePrev     - Previous clip list size
lpClipListPrev  - Pointer to previous clip list
changed         - True if the clip list has changed
****************************************************************************/
typedef struct {
    LPDIRECTDRAWCLIPPER     lpClipper;
    PM_HWND                 hwnd;
    DWORD                   bufSize;
    LPRGNDATA               lpClipList;
    DWORD                   bufSizePrev;
    LPRGNDATA               lpClipListPrev;
    ibool                   changed;
    } GA_clipperInt;

/****************************************************************************
REMARKS:
Internal version of the GA_buf structure that contains internal variables
specific to the DirectDraw buffer manager implementation.

MEMBERS:
p           - Public data structure
lpSurf      - Pointer to the DirectDraw surface
lpSurface   - Pointer to the actual linear surface address
srcColorKey - Source color key for the surface
dstColorKey - Destination color key for the surface
clipper     - Associated surface clipper object (NULL if none attached)
lpClipper   - Pointer to DirectDrawe clipper object
****************************************************************************/
typedef struct {
    GA_buf                  p;
    LPDIRECTDRAWSURFACE7    lpSurf;
    void                    *lpSurface;
    GA_color                srcColorKey;
    GA_color                dstColorKey;
    GA_clipperInt           *clipper;
    int                     surfLock;
    } GA_bufInt;

// The following defines a structure used to cache all necessary mode
// information after the initial detection of DirectDraw, so that
// it can be re-used efficiently to determine the capabilities of all
// available display modes.

#define MAX_REFRESH_RATES   20

typedef struct {
    ushort  xres;                   // Horizontal resolution
    ushort  yres;                   // Vertical resolution
    uchar   bits;                   // Bits per pixel
    ulong   dwCaps;                 // DirectDraw surface caps
    ushort  bytesPerLine;           // Bytes per scanline
    uchar   pixelFormat;            // Pixel format type (see below)
    short   RefreshList[MAX_REFRESH_RATES+1];
    } GA_mdEntry;

// Define the indexes for the pre-defined pixel format types

typedef enum {
    pfRGB555,           // 15 bits per pixel 5:5:5 format
    pfRGB565,           // 16 bits per pixel 5:6:5 format
    pfRGB24,            // 24 bits per pixel RGB format
    pfBGR24,            // 24 bits per pixel BGR format
    pfARGB32,           // 32 bits per pixel ARGB format
    pfABGR32,           // 32 bits per pixel ABGR format
    pfRGBA32,           // 32 bits per pixel RGBA format
    pfBGRA32,           // 32 bits per pixel BGRA format
    } GA_pixelFormatIdx;

/****************************************************************************
REMARKS:
Structure for block list of buffers.

MEMBERS:
bl          - Memory pool containg objects
freeList    - Start of free list for pool
****************************************************************************/
typedef struct {
    N_blockList bl;
    N_node      *freeList;
    } bufferList_t;

// Buffer scanline blitting functions
typedef void (*DD_bufReadScan_t)(uchar *d,uchar *s,int width);
typedef void (*DD_bufWriteScan_t)(uchar *d,uchar *s,int width);

/****************************************************************************
REMARKS:
Structure to store the current state of the buffer manager
****************************************************************************/
typedef struct {
    REF2D_driver            *ref2d;
    GA_2DStateFuncs         state2d;
    bufferList_t            bufferList;
    N_list                  bufferHeap;
    GA_buf                  *visibleBuf;
    GA_bufInt               **flipBuffers;
    int                     numFlipBuffers;
    LPDIRECTDRAWSURFACE7    *lpSurf;
    void                    **lpSurface;
    DD_bufReadScan_t        bufReadScan;
    DD_bufWriteScan_t       bufWriteScan;
    bufferList_t            clipperList;
    } DD_bufMgrState;

/****************************************************************************
REMARKS:
D3D vertex for 3D rendering
****************************************************************************/
typedef struct {
    FLOAT   x, y;
    FLOAT   z;
    FLOAT   rhw;
    FLOAT   tu, tv;
    } D3D_vertex;

// Macros to determine if certain functions are available. Since we set
// all rendering vectors to the same values, a quick way to check this is
// to look at the DV.draw_8bpp structure for the necessary function pointers.

#define HAVE_HWRECT()           (DV.draw_8bpp.DrawRect != NULL)
#define HAVE_HWBLT()            (DV.draw_8bpp.BitBlt != NULL)
#define HAVE_HWSRCTRANSBLT()    (DV.draw_8bpp.SrcTransBlt != NULL)
#define HAVE_HWDSTTRANSBLT()    (DV.draw_8bpp.DstTransBlt != NULL)

// Macro to forcibly unlock the drawing surface before calling DirectDraw

#define UNLOCK_SURFACE()                                    \
    if (DD_activeBuf->surfLock) {                           \
        IDirectDrawSurface7_Unlock(DD_lpActive,NULL);       \
        DD_activeBuf->surfLock = 0;                         \
        DD_activeBuf->p.Surface = NULL;                     \
        }

// Macro to safely release any DirectDraw COM object

#define SAFE_RELEASE(p)                                     \
    if (p) {                                                \
        p->lpVtbl->Release(p);                              \
        p = NULL;                                           \
        }

// Macro to pack a 32-bit ARGB pixel

#define PACK_COLOR_ARGB(A,R,G,B)                    \
    ( ((ulong)(A) << 24)                            \
    | ((ulong)(R) << 16)                            \
    | ((ulong)(G) << 8)                             \
    | ((ulong)(B) << 0))

#pragma pack()

/*---------------------------- Global Variables ---------------------------*/

extern LPDIRECTDRAW7        DD_lpDD;
extern DDCAPS               DD_caps;
extern int                  DD_numModes;
extern GA_mdEntry           DD_modeList[GA_MAX_MODES+1];
extern GA_pixelFormat       DD_pixelFormats[];
extern int                  DD_pixelDepths[];
extern DWORD                DD_textureFormats[];
extern LPDIRECTDRAWSURFACE7 DD_lpTexture[];
extern int                  DD_textureSizeX[];
extern int                  DD_textureSizeY[];
extern int                  DD_numPixelFormats;
extern int                  DD_iPrimaryPF;
extern ulong                DD_totalMemory;
extern ulong                DD_totalAGPMemory;
extern DDSURFACEDESC2       DD_desktop;
extern LPDIRECTDRAWSURFACE7 DD_lpPrimary;
extern LPDIRECTDRAWSURFACE7 DD_lpActive;
extern LPDIRECTDRAWSURFACE7 DD_lpVisible;
extern LPDIRECTDRAWPALETTE  DD_lpPalette;
extern GA_bufInt            *DD_activeBuf;
extern DDSURFACEDESC2       DD_lockBuf;
extern DDBLTFX              DD_bltFx;
extern DWORD                DD_dwSrcRopCodes[];
extern DWORD                DD_dwPatRopCodes[];
extern ibool                DD_validSrcRops[16];
extern ibool                DD_validPatRops[16];
extern GA_exports           gaExp;
extern GA_initFuncs         hwInit;
extern GA_driverFuncs       hwDriver;
extern ibool                DD_isSDDActive;
extern LPDIRECT3D7          DD_lpD3D;
extern LPDIRECT3DDEVICE7    DD_lpD3DDevice;
extern D3DDEVICEDESC7       DD_caps3d;
extern const GUID           IID_IDirect3DHALDevice;
extern const GUID           IID_IDirect3DTnLHalDevice;

// Global function pointers

extern VOID (WINAPI *pSleep)(DWORD dwMilliseconds);
extern int  (WINAPI *pShowCursor)(BOOL bShow);
extern BOOL (WINAPI *pSetCursorPos)(int X,int Y);
extern BOOL (WINAPI *pGetCursorPos)(LPPOINT lpPoint);
extern BOOL (WINAPI *pPeekMessage)(LPMSG lpMsg,HWND hWnd,UINT wMsgFilterMin,UINT wMsgFilterMax,UINT wRemoveMsg);
extern BOOL (WINAPI *pTranslateMessage)(CONST MSG *lpMsg);
extern LONG (WINAPI *pDispatchMessage)(CONST MSG *lpMsg);
extern BOOL (WINAPI *pSetWindowPos)(HWND hWnd,HWND hWndInsertAfter,int X,int Y,int cx,int cy,UINT uFlags);
extern BOOL (WINAPI *pShowWindow)(HWND hWnd,int nCmdShow);
extern LONG (APIENTRY *pRegOpenKeyEx)(HKEY hKey,LPCSTR lpSubKey,DWORD ulOptions,REGSAM samDesired,PHKEY phkResult);
extern LONG (APIENTRY *pRegQueryValueEx)(HKEY hKey,LPCSTR lpValueName,LPDWORD lpReserved,LPDWORD lpType,LPBYTE lpData,LPDWORD lpcbData);
extern LONG (APIENTRY *pRegSetValueEx)(HKEY hKey,LPCSTR lpValueName,DWORD Reserved,DWORD dwType,CONST BYTE* lpData,DWORD cbData);
extern LONG (APIENTRY *pRegCloseKey)(HKEY hKey);

/*-------------------------- Function Prototypes --------------------------*/

//---------------------------------------------------------------------------
// DirectDraw helper functions
//---------------------------------------------------------------------------

const char * DD_errMsg(HRESULT errCode);
ibool       DD_GetPixelFormat(DDPIXELFORMAT *ddpfPixelFormat,int *bits,int *pixelFormat);
void        DD_clearMessageQueue(void);

//---------------------------------------------------------------------------
// Mode init emulation functions
//---------------------------------------------------------------------------

N_int32     NAPI DD_GetVideoModeInfo(N_uint32 mode,GA_modeInfo *modeInfo);
N_int32     NAPI DD_SetVideoMode(N_uint32 mode,N_int32 *virtualX,N_int32 *virtualY,N_int32 *bytesPerLine,N_int32 *maxMem,N_int32 refreshRate,GA_CRTCInfo *crtc);
N_uint32    NAPI DD_GetVideoMode(void);
N_int32     NAPI DD_GetCustomVideoModeInfo(N_int32 xRes,N_int32 yRes,N_int32 virtualX,N_int32 virtualY,N_int32 bitsPerPixel,GA_modeInfo *modeInfo);
N_int32     NAPI DD_SetCustomVideoMode(N_int32 xRes,N_int32 yRes,N_int32 bitsPerPixel,N_uint32 flags,N_int32 *virtualX,N_int32 *virtualY,N_int32 *bytesPerLine,N_int32 *maxMem,GA_CRTCInfo *crtc);
N_uint32    NAPI DD_GetClosestPixelClock(N_int32 outputHead,N_int32 xRes,N_int32 yRes,N_int32 bitsPerPixel,N_uint32 pixelClock);
void        NAPI DD_GetCRTCTimings(GA_CRTCInfo *crtc);
void        NAPI DD_SetCRTCTimings(GA_CRTCInfo *crtc);
void        NAPI DD_SaveCRTCTimings(GA_CRTCInfo *crtc);
void        NAPI DD_SetGlobalRefresh(N_int32 refresh,N_int32 outputHead);
N_int32     NAPI DD_SetDisplayOutput(N_int32 device);
N_int32     NAPI DD_GetDisplayOutput(void);
N_int32     NAPI DD_SetDisplayOutputExt(N_int32 device,N_int32 outputHead);
N_int32     NAPI DD_GetDisplayOutputExt(N_int32 outputHead);
void        NAPI DD_getUniqueFilename(char *filename,int type);
void        NAPI DD_getMonitorInfo(GA_monitor *monitor,N_int32 outputHead);
void        NAPI DD_setMonitorInfo(GA_monitor *monitor,N_int32 outputHead);
void        NAPI DD_GetCurrentVideoModeInfo(GA_modeInfo *modeInfo);
N_int32     NAPI DD_GetCurrentRefreshRate(void);
N_int32     NAPI DD_SwitchPhysicalResolution(N_int32 physicalXResolution,N_int32 physicalYResolution,N_int32 refreshRate);
N_int32     NAPI DD_GetVideoModeInfoExt(N_uint32 mode,GA_modeInfo *modeInfo,N_int32 outputDevice,N_int32 outputHead);
N_int32     NAPI DD_GetCustomVideoModeInfoExt(N_int32 xRes,N_int32 yRes,N_int32 virtualX,N_int32 virtualY,N_int32 bitsPerPixel,GA_modeInfo *modeInfo,N_int32 outputDevice,N_int32 outputHead);
void        NAPI DD_getCertifyInfo(GA_certifyInfo *info);

//---------------------------------------------------------------------------
// DPMS/SCI emulation functions
//---------------------------------------------------------------------------

int         NAPI DD_DPMSdetect(N_int16 *supportedStates);
int         NAPI DD_SCIdetect(N_uint8 *capabilities,N_int32 *numMonitorPorts);

//---------------------------------------------------------------------------
// Non drawing, mode specific emulation functions
//---------------------------------------------------------------------------

void        NAPI DD_SetDisplayStart(N_int32 offset,N_int32 waitVRT);
N_int32     NAPI DD_GetDisplayStartStatus(void);
N_int32     NAPI DD_IsVSync(void);
void        NAPI DD_WaitVSync(void);
void        NAPI DD_SetPaletteData(GA_palette *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI DD_GetPaletteData(GA_palette *pal,N_int32 num,N_int32 index);

//---------------------------------------------------------------------------
// 2D state management function prototypes
//---------------------------------------------------------------------------

N_int32     NAPI DD_SetDrawBuffer(GA_buffer *drawBuf);
N_int32     NAPI DD_IsIdle(void);
void        NAPI DD_WaitTillIdle(void);
void        NAPI DD_EnableDirectAccess(void);
void        NAPI DD_DisableDirectAccess(void);
N_int32     NAPI DD_SetMix(N_int32 mix);
void        NAPI DD_SetForeColor(GA_color color);
void        NAPI DD_SetBackColor(GA_color color);

//---------------------------------------------------------------------------
// 2D drawing function prototypes
//---------------------------------------------------------------------------

void        DD_setupAccelerator(void);
void        NAPI DD_DrawScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI DD_DrawRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI DD_DrawRectExt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI DD_DrawTrap(GA_trap *trap);
void        NAPI DD_BitBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI DD_SrcTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI DD_DstTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
N_int32     DD_BitBltFxTestInt(GA_bltFx *fx,int *ipf);
N_int32     NAPI DD_BitBltFxTest(GA_bltFx *fx);
void        NAPI DD_BitBltFx(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI DD_BitBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
void        NAPI DD_BitBltFxSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);

//---------------------------------------------------------------------------
// Buffer management functions
//---------------------------------------------------------------------------

void        BUFFER_init(void);
ibool       NAPI DD_SetRef2d(REF2D_driver *ref2d);
void        NAPI DD_exitBufMgr(void);
ibool       NAPI DD_InitBuffers(N_int32 numBuffers,N_uint32 flags,GA_softStereoFuncs *stereo);
GA_buf *    NAPI DD_GetPrimaryBuffer(void);
GA_buf *    NAPI DD_GetFlippableBuffer(N_int32 index);
GA_buf *    NAPI DD_AllocBuffer(N_int32 width,N_int32 height,N_int32 flags);
ibool       NAPI DD_FreeBuffer(GA_buf *buf);
N_uint32    NAPI DD_LockBuffer(GA_buf *buf);
void        NAPI DD_UnlockBuffer(GA_buf *buf);
N_int32     NAPI DD_SetActiveBuffer(GA_buf *buf);
void        NAPI DD_FlipToBuffer(GA_buf *buf,N_int32 waitVRT);
void        NAPI DD_FlipToStereoBuffer(GA_buf *left,GA_buf *right,N_int32 waitVRT);
int         NAPI DD_GetFlipStatus(void);
void        NAPI DD_WaitTillFlipped(void);
void        NAPI DD_UpdateCache(GA_buf *buf);
void        NAPI DD_UpdateFromCache(GA_buf *buf);
void        NAPI DD_BitBltBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI DD_BitBltPattBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI DD_BitBltColorPattBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI DD_SrcTransBltBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI DD_DstTransBltBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI DD_StretchBltBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI DD_BitBltPlaneMaskedBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask);
void        NAPI DD_BitBltFxBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI DD_DrawRectBuf(GA_buf *buf,N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
GA_clipper * NAPI DD_GetClipper(GA_buf *buf);
void        NAPI DD_SetClipper(GA_buf *buf,GA_clipper *clipper);
GA_buf *    NAPI DD_GetActiveBuffer(void);
GA_buf *    NAPI DD_GetVisibleBuffer(void);
GA_buf *    NAPI DD_GetVisibleRightBuffer(void);

//---------------------------------------------------------------------------
// Internal buffer management functions
//---------------------------------------------------------------------------

ibool       DD_createSurface(N_int32 *width,N_int32 *height,N_int32 bitsPerPixel,GA_pixelFormat *pf,N_uint32 flags,LPDIRECTDRAWSURFACE7 *lpSurf);
ibool       DD_createSurfaceForMemory(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 bitsPerPixel,GA_pixelFormat *pf,GA_palette *pal,LPDIRECTDRAWSURFACE7 *lpSurf);

//---------------------------------------------------------------------------
// Window clip list management
//---------------------------------------------------------------------------

GA_clipper *    NAPI DD_CreateClipper(PM_HWND hwnd);
ibool           NAPI DD_IsClipListChanged(GA_clipper *clipper);
GA_rect *       NAPI DD_GetClipList(GA_clipper *clipper,N_int32 *count);
void            NAPI DD_DestroyClipper(GA_clipper *clipper);

//---------------------------------------------------------------------------
// Internal buffer scanline reading/writing functions
//---------------------------------------------------------------------------

void        DD_readScan_8(uchar *d,uchar *s,int width);
void        DD_writeScan_8(uchar *d,uchar *s,int width);
void        DD_readScan_16(uchar *d,uchar *s,int width);
void        DD_writeScan_16(uchar *d,uchar *s,int width);
void        DD_readScan_24(uchar *d,uchar *s,int width);
void        DD_writeScan_24(uchar *d,uchar *s,int width);
void        DD_readScan_32(uchar *d,uchar *s,int width);
void        DD_writeScan_32(uchar *d,uchar *s,int width);

#endif  /* __INTERNAL_H */

