/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Generic device detection module.
*
****************************************************************************/

#include "snap/graphics/driver.h"

/*--------------------------- Global variables ----------------------------*/

GA_driver               DV;             // Global device driver structure
PM_imports _VARAPI      _PM_imports;    // Imports to PM library functions
DETECT_imports _VARAPI _DETECT_imports; // Imports from generic code

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler for non-imported GA_imports.
****************************************************************************/
static void _DETECT_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported DETECT_imports import function called! Please re-compile!\n");
}

/****************************************************************************
HEADER:
graphics/driver.h

PARAMETERS:
pmImp       - PM library imports
detImp      - SNAP driver imports
info        - Pointer to the PCI configuration space info for the device
mfrName     - Place to store the chipset manufacturer name
chipName    - Place to store the chipset name

RETURNS:
True if hardware detected, false if not.

REMARKS:
This function detects the hardware chipset family and name for a single
graphics chipset family, built using the same detection code used by the
real drivers.
****************************************************************************/
ibool _CEXPORT _GA_detectChipset(
    PM_imports *pmImp,
    DETECT_imports *detImp,
    PCIDeviceInfo *info,
    char *mfrName,
    char *chipName)
{
    int     i,max;
    ulong   *p;

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_PM_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;
    max = sizeof(_DETECT_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_DETECT_imports; i < max; i++)
        *p++ = (ulong)_DETECT_fatalErrorHandler;

    // Now copy all our imported functions
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    memcpy(&_DETECT_imports,detImp,MIN(sizeof(_DETECT_imports),detImp->dwSize));

    // Initialise default information
    memset(&DV,0,sizeof(DV));
    memcpy(&DV.PCIConfig,info,sizeof(DV.PCIConfig));

    // Now run the chipset detection code
    if (!PNP_detect())
        return false;
    if (!CHIP_detect())
        return false;

    // Return the chipset name information
    strcpy(mfrName,DRV_manufacturerName);
    strcpy(chipName,DRV_chipsetNames[DV.chipID]);
    return true;
}

