/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to find the location of the linear framebuffer on
*               cards where we cannot determine this directly from the
*               chipset register or PCI/AGP bus.
*
****************************************************************************/

#include "snap/graphics/driver.h"

/*--------------------------- Global variables ----------------------------*/

static int oldMode,old50Lines,oldx,oldy;
static void *oldText;
static void *oldScr = NULL;
#define TEXT_SIZE   (80 * 50 * 2)

/*----------------------------- Implementation ----------------------------*/

#define TEST_STR        "Test for linear buffer"
#define TEST_STR_SIZE   23
#define MEM_WRITE       1
#define MEM_READ        0

/****************************************************************************
REMARKS:
Saves the current video mode to be restored at a later date.
****************************************************************************/
void DRV_saveMode(void)
{
    RMREGS  regs;

    regs.h.ah = 0x0F;
    PM_int86(0x10,&regs,&regs);         /* Save previous video mode     */
    oldMode = regs.h.al & 0x7F;
    regs.h.ah = 0x3;
    PM_int86(0x10,&regs,&regs);         /* Save previous cursor pos     */
    oldx = regs.h.dl;
    oldy = regs.h.dh;
    regs.x.ax = 0x1130;
    regs.h.bh = 0x00;
    regs.h.dl = 0x00;
    PM_int86(0x10,&regs,&regs);
    old50Lines = (regs.h.dl == 49);

    /* Save the old text screen values */
    if ((oldText = malloc(TEXT_SIZE)) != NULL) {
        if (!oldScr)
            oldScr = PM_mapPhysicalAddr(0xB8000,0xFFFF,true);
        memcpy(oldText,oldScr,TEXT_SIZE);
        }
}

/****************************************************************************
REMARKS:
Restores the previous video mode, without clearing the display screen.
This routine and the above routine can be used to check for specific
features.
****************************************************************************/
void DRV_restoreMode(void)
{
    RMREGS      regs;

    /* Reset text mode - dont clear display */
    regs.h.ah = 0x0;
    regs.h.al = 0x83;
    PM_int86(0x10,&regs,&regs);
    if (old50Lines) {
        regs.x.ax = 0x1112;
        regs.h.bl = 0x00;
        PM_int86(0x10,&regs,&regs);
        }
    regs.h.ah = 0x02;
    regs.h.dl = oldx;
    regs.h.dh = oldy;
    regs.h.bh = 0x00;
    PM_int86(0x10,&regs,&regs); /* Restore cursor position              */

    /* Restore the old text screen values */
    if (oldText) {
        memcpy(oldScr,oldText,TEXT_SIZE);
        free(oldText);
        oldText = NULL;
        }
}

/****************************************************************************
REMARKS:
Returns the amount of memory installed in the machine in bytes.
****************************************************************************/
ulong DRV_getMainMemSize(void)
{
    static int memSize = 0;

    if (memSize == 0) {
        // Find the amount of main memory on the machine
        memSize = ((DRV_readCMOS(0x18) >> 2) + 1) * MB;
        }
    return memSize;
}

/****************************************************************************
REMARKS:
Writes the test string to the normal VGA framebuffer location, so that we
can begin searching for the linear buffer by reading possible memory
locations (a non-destructive operation that wont hang the machine).
****************************************************************************/
void DRV_writeLinearTestString(void)
{
    void *p = PM_getA0000Pointer();
    memcpy(p,TEST_STR,TEST_STR_SIZE);
}

/****************************************************************************
PARAMETERS:
addr    - Address of buffer to check

RETURNS:
True if buffer is responding, false if not.

REMARKS:
Looks for an active linear framebuffer at the specified location.
****************************************************************************/
ibool DRV_checkLinearBuffer(
    ulong addr)
{
    char    test[TEST_STR_SIZE];
    void    *p;

    if (addr < DRV_getMainMemSize())
        return false;
    p = PM_mapPhysicalAddr(addr,4096,true);
    memset(test,0,sizeof(test));
    memcpy(test, p, sizeof(test));
    if (strcmp(TEST_STR,test) != 0)
        return false;
    return true;
}

/****************************************************************************
RETURNS:
Increment used to step linear framebuffer detection sequence by
****************************************************************************/
static ulong GetIncrement(
    ulong addr)
{
    if (addr < 16 * MB)
        return 1 * MB;
    else if (addr < 128 * MB)
        return 16 * MB;
    else
        return 64 * MB;
}

/****************************************************************************
RETURNS:
Address of linear responding framebuffer

REMARKS:
Begins looking for the linear framebuffer ending with the end of physical
memory on the machine from to the end of physical memory to the specified
maximum address.

In order to check that we really have a linear framebuffer, we check that
we have active memory at each 1Mb boundary.

NOTE: You must *first* call the writeLinearTestString() to write the test
      string to the framebuffer before calling this routine.
****************************************************************************/
ulong DRV_findLinearBuffer(
    ulong maxAddr)
{
    ulong   addr = DRV_getMainMemSize();
    ulong   incr = GetIncrement(addr);

    while (addr < maxAddr) {
        if (DRV_checkLinearBuffer(addr))
            return addr;
        addr += incr;
        incr = GetIncrement(addr);
        }
    return 0;
}

