/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module containing the clock chip detection routines.
*
****************************************************************************/

#include "snap/graphics/driver.h"
#include "snap/graphics/dacid.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Function to detect the type of clock chip depending on the type of
RAMDAC ID. For external programmable clock chips (like the ICS9161 etc)
you will have to write code specifically for your chipset driver to
detect the presence of the clock chip.
****************************************************************************/
int CLK_detect(
    int dacID)
{
    int clock = CLK_HARD;

    switch (dacID) {
        case DAC_ATT08DAC:      clock = CLK_ATT08;      break;
        case DAC_ATT99DAC:      clock = CLK_ATT99;      break;
        case DAC_TV5DAC:        clock = CLK_TVP5;       break;
        case DAC_TV6DAC:        clock = CLK_TVP6;       break;
        case DAC_TV30DAC:       clock = CLK_TVP30;      break;
        case DAC_MU99DAC:       clock = CLK_MU99;       break;
        case DAC_STG3DAC:       clock = CLK_STG3;       break;
        case DAC_ICS00DAC:
        case DAC_ICS01DAC:      clock = CLK_ICS0x;      break;
        case DAC_ICS40DAC:
        case DAC_ICS41DAC:
        case DAC_ICS42DAC:      clock = CLK_ICS4x;      break;
        case DAC_CH91DAC:
        case DAC_CH98DAC:       clock = CLK_CH91;       break;
        case DAC_CH98ADAC:      clock = CLK_CH98A;      break;
        case DAC_ICW524DAC:     clock = CLK_ICW524;     break;
        case DAC_IBM24DAC:      clock = CLK_IBM24;      break;
        case DAC_IBM25DAC:      clock = CLK_IBM25;      break;
        case DAC_IBM26DAC:      clock = CLK_IBM26;      break;
        case DAC_IBM26DBDAC:    clock = CLK_IBM26;      break;
        case DAC_IBM28DAC:      clock = CLK_IBM24;      break;
        case DAC_IBM624DAC:     clock = CLK_IBM24;      break;
        case DAC_ALG13DAC:      clock = CLK_ALG13;      break;
        case DAC_OTI088DAC:     clock = CLK_OTI088;     break;
        case DAC_TL5DAC:
        case DAC_ATI60DAC:      clock = CLK_ATI18_1;    break;
        }
    return clock;
}

#ifdef INCLUDE_DISCRETECLOCKS
/****************************************************************************
REMARKS:
Detects the type of discrete clock present
****************************************************************************/
static ibool DRV_detectICS9161(void)
{
    int     i;
    ulong   count1, count2;
    uchar   save3C2 = inpb(0x3CC);
    GA_PLLParams savedPLLValues;

    // Initialise the ICS clock module, so we can detect the ICS9161
    // and ICD2061 compatible programmable clocks. This *must* be done
    // here first!
    savedPLLValues = DV.pixelClock;
    ICS9161CLK_init();

    // Program ICS9161 pixel clock to 25Mhz and measure refresh rate
    ICS9161_setPLLPixelClock(N_FIXCONST(25.00));
    DV.driver.WaitVSync();
    DV.driver.WaitVSync();
    LZTimerOn();
    for (i = 0; i < 10; i++)
        DV.driver.WaitVSync();
    LZTimerOff();
    count1 = LZTimerCount();

    // Program ICS9161 pixel clock to 40Mhz and measure refresh rate
    ICS9161_setPLLPixelClock(N_FIXCONST(40.00));
    DV.driver.WaitVSync();
    DV.driver.WaitVSync();
    LZTimerOn();
    for (i = 0; i < 10; i++)
        DV.driver.WaitVSync();
    LZTimerOff();
    count2 = LZTimerCount();

    // Restore original register value
    outpb(0x3C2,save3C2);

    // restore PLL structure
    DV.pixelClock = savedPLLValues;

    // Check if the counts indicate the clock was correctly programmed
    return (count1 > ((count2 * 15) / 10));
}

/****************************************************************************
REMARKS:
Log discrete clock information to graphics.log file.
****************************************************************************/
void LogDiscreteClocks(void)
{
    int i;

    N_log("\nList of discrete clock values being used:\n");
    for (i = 0; i < DV.pixelClock.numClocks; i++)
        N_log("  %02d: %3d MHz\n", i, N_FIXROUND(DV.pixelClock.validClocks[i]));
}

/****************************************************************************
REMARKS:
Detects the type of discrete clock present
****************************************************************************/
void DRV_detectDiscreteClocks(
    int maxClocks)
{
    FILE    *f;
    char    filename[PM_MAX_PATH];
    ushort  VendorID,DeviceID;
    long    mfrID,chipID;

    // Read previous clock tables from disk if available
    DRV_getUniqueFilename(filename,4);
    if ((f = fopen(filename,"rb")) != NULL) {
        fread(&VendorID,1,sizeof(VendorID),f);
        fread(&DeviceID,1,sizeof(DeviceID),f);
        fread(&chipID,1,sizeof(chipID),f);
        fread(&mfrID,1,sizeof(mfrID),f);
        if (DV.chipID == chipID && DV.mfrID == mfrID && DV.PCIConfig.VendorID == VendorID && DV.PCIConfig.DeviceID == DeviceID) {
            // Read the parameters from the clocks file and initialise clock modules
            fread(&DV.clockID,1,sizeof(DV.clockID),f);
            fread(&DV.pixelClock,1,sizeof(DV.pixelClock),f);
            fread(&DV.pixelClockMax8,1,sizeof(DV.pixelClockMax8),f);
            DV.pixelClockMax4   = DV.pixelClockMax8;
            DV.pixelClockMax16  = DV.pixelClockMax8;
            DV.pixelClockMax24  = DV.pixelClockMax8;
            DV.pixelClockMax32  = DV.pixelClockMax8;
            fclose(f);

            // Re-initialise the clock module
            if (DV.clockID == CLK_ICS9161)
                ICS9161CLK_init();
            else {
                DISCRETECLK_init();
                LogDiscreteClocks();
                }
            return;
            }
        fclose(f);
        }

    // the clocks.00 file was not detected in the SNAP_PATH\config
    // directory, so we try to detect the ICS9161 and setup discrete
    // clock analysis if it fails.
    if (DRV_detectICS9161()) {
        DV.clockID = CLK_ICS9161;
        ICS9161CLK_init();
        }
    else {
        // Setup values needed for DRV_analyseDiscreteClocks()

        // Indicate that we have 'maxClocks' discrete clock lines
        // available, and that we should analyse them after the
        // driver has been completed initialised.
        DV.pixelClock.numClocks = -maxClocks;
        DV.clockID = CLK_HARD;
        DV.pixelClock.flags = 0;
        DISCRETECLK_init();
        }

    // Save the clock table values for later
    if ((f = fopen(filename,"wb")) != NULL) {
        fwrite(&DV.PCIConfig.VendorID,1,sizeof(DV.PCIConfig.VendorID),f);
        fwrite(&DV.PCIConfig.DeviceID,1,sizeof(DV.PCIConfig.DeviceID),f);
        fwrite(&DV.chipID,1,sizeof(DV.chipID),f);
        fwrite(&DV.mfrID,1,sizeof(DV.mfrID),f);
        fwrite(&DV.clockID,1,sizeof(DV.clockID),f);
        fwrite(&DV.pixelClock,1,sizeof(DV.pixelClock),f);
        fwrite(&DV.pixelClockMax8,1,sizeof(DV.pixelClockMax8),f);
        fclose(f);
        }

}

/****************************************************************************
REMARKS:
Sets the first SVGA graphics mode found.
****************************************************************************/
static void SetGraphicsMode(
    int xRes,
    int yRes,
    int bits)
{
    GA_CRTCInfo         crtc;
    N_uint16            *modes;
    int                 mode = -1;
    N_int32             virtualX,virtualY,bytesPerLine,maxMem;
    static GA_modeInfo  gMi;

    /* Search for the 640x480x256 @ 60Hz graphics mode */
    for (modes = DC.AvailableModes; *modes != 0xFFFF; modes++) {
        gMi.dwSize = sizeof(gMi);
        if (DV.init.GetVideoModeInfo(*modes,&gMi) != 0)
            continue;
        if (gMi.Attributes & gaIsTextMode)
            continue;
        if (gMi.XResolution == xRes && gMi.YResolution == yRes && gMi.BitsPerPixel == bits) {
            mode = *modes;
            if (gMi.Attributes & gaHaveLinearBuffer)
                mode |= gaLinearBuffer;
            break;
            }
        }
    if (mode == -1)
        PM_fatalError("Unable to set graphics mode!");
    virtualX = virtualY = bytesPerLine = -1;
    DV.init.GetVideoMode();
    DV.init.SetVideoMode(mode,&virtualX,&virtualY,&bytesPerLine,&maxMem,60,&crtc);
}

/****************************************************************************
REMARKS:
Restore the old display mode active before graphics mode was started.
****************************************************************************/
static void RestoreTextMode(void)
{
    GA_CRTCInfo     crtc;
    N_int32         virtualX,virtualY,bytesPerLine,maxMem;

    /* Set text mode. Note that mode 0x100 is always 80x25 text mode! */
    virtualX = virtualY = bytesPerLine = -1;
    DV.init.SetVideoMode(0x100,&virtualX,&virtualY,&bytesPerLine,&maxMem,0,&crtc);
}

/****************************************************************************
REMARKS:
Checks to see if the vSync is actually present or not
****************************************************************************/
static ibool CheckForVSync(void)
{
    ibool   active = false;

    // Delay for 50ms before testing for vSync
    LZTimerOn();
    while (LZTimerLap() < 50000UL)
        ;
    LZTimerOff();

    // Check for active vSync signals
    LZTimerOn();
    while (LZTimerLap() < 100000UL) {
        if (DV.driver.IsVSync()) {
            active = true;
            break;
            }
        }
    LZTimerOff();
    return active;
}

/****************************************************************************
REMARKS:
Detects the type of discrete clock present
****************************************************************************/
void DRV_analyseDiscreteClocks(void)
{
    FILE    *f;
    int     i,j;
    char    filename[PM_MAX_PATH];
    ulong   count,clock;

    // Bail if we don't have discrete clocks or they have already been
    // analysed previously and read from disk.
    if (DV.clockID != CLK_HARD || DV.pixelClock.numClocks >= 0)
        return;

    // Set 640x480x8 @60Hz graphics mode and measure the clock frequencies
    SetGraphicsMode(640,480,8);
    DV.pixelClock.numClocks = -DV.pixelClock.numClocks;
    DV.pixelClockMax8 = 0;
    for (i = 0; i < DV.pixelClock.numClocks; i++) {
        // Select the clock entry and measure the frequency
        DV.chip.selectClock(i,0);
        if (CheckForVSync()) {
            DV.driver.WaitVSync();
            LZTimerOn();
            for (j = 0; j < 8; j++) {
                // Dont put any delays in here, or we will measure the
                // clocks incorrectly!
                DV.driver.WaitVSync();
                }
            LZTimerOff();
            count = LZTimerCount();
            clock = FXdiv(N_TOFIX(336000000 / count),N_FIXCONST(100));
            }
        else
            clock = 0;
        DV.pixelClock.validClocks[i] = clock;
        if (clock > DV.pixelClockMax8)
            DV.pixelClockMax8 = clock;
        }

    // Update maximum pixel clock fields
    DV.pixelClockMaxTxt = N_FIXCONST(60);
    DV.pixelClockMax4   = DV.pixelClockMax8;
    DV.pixelClockMax16  = DV.pixelClockMax8;
    DV.pixelClockMax24  = DV.pixelClockMax8;
    DV.pixelClockMax32  = DV.pixelClockMax8;

    // Re-initialise the clock module
    DISCRETECLK_init();

    // Restore text mode
    RestoreTextMode();

    // Save the clock table values for later
    DRV_getUniqueFilename(filename,4);
    if ((f = fopen(filename,"wb")) != NULL) {
        fwrite(&DV.PCIConfig.VendorID,1,sizeof(DV.PCIConfig.VendorID),f);
        fwrite(&DV.PCIConfig.DeviceID,1,sizeof(DV.PCIConfig.DeviceID),f);
        fwrite(&DV.chipID,1,sizeof(DV.chipID),f);
        fwrite(&DV.mfrID,1,sizeof(DV.mfrID),f);
        fwrite(&DV.clockID,1,sizeof(DV.clockID),f);
        fwrite(&DV.pixelClock,1,sizeof(DV.pixelClock),f);
        fwrite(&DV.pixelClockMax8,1,sizeof(DV.pixelClockMax8),f);
        fclose(f);
        }

    // Log the discrete clock information
    if (DV.clockID == CLK_HARD)
        LogDiscreteClocks();
}
#endif
