/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module containing the RAMDAC detection routines.
*
****************************************************************************/

#include "snap/graphics/driver.h"
#include "snap/graphics/dacid.h"
#include "../ramdac/ibmdac.h"
#include "../ramdac/tvpdac.h"
#include "../ramdac/atidac.h"
#include "../ramdac/btdac.h"

/*--------------------- Macros and type definitions -----------------------*/

typedef struct {
    uchar   index;
    uchar   red;
    uchar   green;
    uchar   blue;
    } PEL;

/*----------------------------- Implementation ----------------------------*/

#ifdef  INCLUDE_EARLYDACS
/****************************************************************************
REMARKS:
Sets the command register and then returns the value read from it.
****************************************************************************/
static int setCommand(
    int value)
{
    VGA_dacToCommand();
    outpb(0x3C6,value);
    VGA_dacToCommand();
    return inpb(0x3C6);
}

/****************************************************************************
REMARKS:
Sets the command register specific bits and then returns the value read from
it.
****************************************************************************/
static int setCommandBits(
    int bits)
{
    int tmp;
    VGA_dacToCommand();
    tmp = inpb(0x3C6) | bits;
    return setCommand(tmp);
}

/****************************************************************************
REMARKS:
Reads the value of a VGA palette register from the hardware.
****************************************************************************/
static void readPelReg(
    int index,
    PEL *pel)
{
    pel->index = index;
    outpb(0x3C7, index);
    pel->red = inpb(0x3C9);
    pel->blue = inpb(0x3C9);
    pel->green = inpb(0x3C9);
}

/****************************************************************************
REMARKS:
Writes the value of a VGA palette register to the hardware.
****************************************************************************/
static void writePelReg(
    PEL *pel)
{
    outpb(0x3C8, pel->index);
    outpb(0x3C9, pel->red);
    outpb(0x3C9, pel->blue);
    outpb(0x3C9, pel->green);
}

/****************************************************************************
REMARKS:
Determines if the RAMDAC supports 8 bit per primary mode.
****************************************************************************/
static ibool dacIs8Bit(void)
{
    int old,v,x;
    PEL pel;

    old = inpb(0x3C8);
    readPelReg(255, &pel);
    v = pel.red;
    pel.red = 0xFF;
    writePelReg(&pel);
    readPelReg(255, &pel);
    x = pel.red;
    pel.red = v;
    writePelReg(&pel);
    outpb(0x3C8, old);
    return (x == 0xFF);
}

/****************************************************************************
REMARKS:
Determines if the specific bit is active in the command register. If the
bit is active we are able to read and write values to it properly.
****************************************************************************/
static ibool testDacBit(
    int bit,
    int oldPel,
    int oldCommand)
{
    int v;

    VGA_dacToPel();
    outpb(0x3C6, oldPel & (bit ^ 0xFF));
    VGA_dacToCommand();
    outpb(0x3C6, oldCommand | bit);
    VGA_dacToCommand();
    v = inpb(0x3C6);
    VGA_dacToCommand();
    outpb(0x3C6, v & (bit ^ 0xFF));
    return (v & bit);
}
#endif

/****************************************************************************
REMARKS:
Function to detect the type of external RAMDAC chip. This function uses
the RAMDAC glue functions to gain access to the internal RAMDAC registers,
so those functions must be set up before this code is called.
****************************************************************************/
int DAC_detect(void)
{
    int x,y,z,v,oldPel,oldCommand,dac;
#ifdef  INCLUDE_EARLYDACS
    int notCommand;
#endif

    /* Original detect code */
    dac = DAC_INTERNAL;
#ifdef  INCLUDE_EARLYDACS
    VGA_dacToCommand();
    oldCommand = inpb(0x3C6);
    VGA_dacToPel();
    oldPel = inpb(0x3C6);
    VGA_dacToPel();
    notCommand = oldCommand ^ 0xFF;
    outpb(0x3C6,notCommand);
    VGA_dacToCommand();
    v = inpb(0x3C6);
    if (v != notCommand) {
        /* The command register is present and working, so we must have
         * a HiColor or TrueColor DAC of some type. First we test for
         * those DAC's that have an id register.
         */
        VGA_dacToPel();
        VGA_dacToCommand();
        inpb(0x3C6);                /* Read CR0                         */
        x = inpb(0x3C6);            /* Read 21c4x8 MIR register         */
        v = inpb(0x3C6);            /* Read 21c4x8 DIR register         */
        z = inpb(0x3C6);
        VGA_dacToPel();
        if ((x == 0x84) && (v == 0x98)) {
            if (z == 0x4F) {
                dac = DAC_ICW516DAC;    /* We have an IC Works W30C516 DAC  */
                goto FoundDAC;
                }
            dac = DAC_ATT98DAC;     /* We have an ATT21c498 24 bit DAC  */
            goto FoundDAC;
            }
        if ((x == 0x84) && (v == 0x99)) {
            if (z == 0x4F) {
                dac = DAC_ICW524DAC;    /* We have an IC Works W30C524 DAC  */
                goto FoundDAC;
                }
            dac = DAC_ATT99DAC;     /* We have an ATT21c499 24 bit DAC  */
            goto FoundDAC;
            }
        if ((x == 0x84) && (v == 0x09)) {
            dac = DAC_ATT08DAC;     /* We have an ATT21c408 24 bit DAC  */
            goto FoundDAC;
            }
        setCommandBits(0x10);
        if (rdinx(0x3C7,0x9) == 0x53) {
            x = (rdinx(0x3C7,10) << 8) | rdinx(0x3C7,11);
            switch (x) {
                case 15020:
                case 15021: dac = DAC_SC21DAC;  goto FoundDAC;
                case 15025: dac = DAC_SC25DAC;  goto FoundDAC;
                case 15026: dac = DAC_SC26DAC;  goto FoundDAC;
                default:    dac = DAC_SC20DAC;  goto FoundDAC;
                }
            }
        setCommand(oldCommand);
        setCommandBits(0x10);
        VGA_dacToCommand();
        inpb(0x3C6);
        outpb(0x3C6,0x00);
        outpb(0x3C6,0x00);
        x = inpb(0x3C6);            /* Read STG1700 company ID          */
        v = inpb(0x3C6);            /* Read STG1700 device ID           */
        VGA_dacToPel();
        if (x == 0x44) {
            switch (v) {
                case 0:     dac = DAC_STG0DAC;  goto FoundDAC;  /* STG1700  */
                case 2:     dac = DAC_STG2DAC;  goto FoundDAC;  /* STG1702  */
                case 3:     dac = DAC_STG3DAC;  goto FoundDAC;  /* STG1703  */
                }
            }
        setCommand(oldCommand);
        VGA_dacToPel();
        outpb(0x3C6, oldPel);
        VGA_dacToPel();
        inpb(0x3C6);
        inpb(0x3C6);
        inpb(0x3C6);
        if ((inpb(0x3C6) & 0xF0) == 0x70) { /* Read chipID reg      */
            dac = DAC_ICS00DAC;     /* We have an ICS 5300 GENDAC       */
            goto FoundDAC;
            }
        VGA_dacToPel();
        if (VGA_dacToCommand() == 0x31) {
            dac = DAC_ALG13DAC;     /* We have an ALG1301 DAC           */
            goto FoundDAC;
            }
        VGA_dacToPel();

        /* Check for ICS RAMDAC via the ID registers */
        inpb(0x3C7);
        x = VGA_dacToCommand();
        if (x == 0xB1) {
            dac = DAC_ICS41DAC;
            goto FoundDAC;
            }
        else if (x == 0xB2) {
            dac = DAC_ICS42DAC;
            goto FoundDAC;
            }
        VGA_dacToPel();

        /* The remaining DAC's dont have ID registers, so we poke around
         * in the command register to try and identify which DAC is
         * which (which can sometimes mean a quick reboot for the machine!)
         */
        if ((setCommand(0xE0) & 0xE0) != 0xE0) {
            /* Test for Music 9C1880 and Sierra SC11486 DAC's */
            VGA_dacToPel();
            x = inpb(0x3C6);
            do {
                v = x;
                x = inpb(0x3C6);
                } while (x != v);   /* Wait for the same value twice    */
            if (VGA_dacToCommand() == 0x8E) {
                dac = DAC_MU18DAC;  /* We have a Music 9C1880 DAC       */
                goto FoundDAC;
                }
            for (x = 0; x < 8; x++)
                if (inpb(0x3C6) == 0x8E) {
                    dac = DAC_MU18DAC;/* We have an Music 9C1880 DAC        */
                    goto FoundDAC;
                    }
            dac = DAC_SC6DAC;           /* We have Sierra SC11486 DAC       */
            goto FoundDAC;
            }
        if ((setCommand(0x60) & 0xE0) == 0) {
            if ((setCommand(0x02) & 0x02) != 0) {
                dac = DAC_ATT90DAC; /* We have an AT&T ATT 20c490 DAC   */
                goto FoundDAC;
                }
            dac = DAC_ATT92DAC;     /* We have an AT&T ATT 20c493 DAC   */
            goto FoundDAC;
            }
        if (((v = VGA_triggerDac()) == 0x58) || v == 0xA7) {
            if (VGA_triggerDac() == (v == 0x58) ? 0xA7 : 0x58) {
                dac = DAC_OTI088DAC;
                goto FoundDAC;
                }
            }

        VGA_dacToPel();
        outpb(0x3C6,notCommand);
        setCommand(oldCommand);
        if (inpb(0x3C6) == notCommand) {
            /* Falls back to PEL register after first read */
            v = setCommand(0xFC);
            if (v != 0xFC) {
                if (v == 0xE2) {
                    dac = DAC_ALG12DAC;
                    goto FoundDAC;  /* We have an Avance Logic 1201 DAC */
                    }
                if (v == 0xE0) {
                    dac = DAC_ICS01DAC;
                    goto FoundDAC;  /* We have an ICS5301 GENDAC        */
                    }
                if (v == 0xF0) {
                    dac = DAC_ICS41DAC;
                    goto FoundDAC;  /* We have an ICS5341 GENDAC        */
                    }
                if (v == 0xF4) {
                    dac = DAC_ICS42DAC;
                    goto FoundDAC;  /* We have an ICS5342 GENDAC        */
                    }
                if (testDacBit(0xF0,oldPel,oldCommand)) {
                    dac = DAC_UMC18DAC;
                    goto FoundDAC;  /* We have a UMC UM70c188 DAC       */
                    }
                dac = DAC_ADAC1DAC; /* We have an Acumos ADAC1 DAC      */
                goto FoundDAC;
                }

            /* All 8 bits are fully read/write */
            dacIs8Bit();
            setCommand((oldCommand | 2) & 0xFE);
            if (dacIs8Bit()) {
                dac = DAC_ATT91DAC; /* We have an AT&T 20c491 DAC       */
                goto FoundDAC;
                }
            if (VGA_triggerDac() == 0xB3) {
                dac = DAC_CH91DAC;  /* We have a Chrontel 8391 DAC      */
                goto FoundDAC;
                }
            dac = DAC_ATT92DAC;     /* We have an AT&T 20c492 DAC       */
            goto FoundDAC;
            }
        else if (VGA_triggerDac() == notCommand) {
            dac = DAC_CH98ADAC;
            goto FoundDAC;
            }

        VGA_dacToPel();
        outpb(0x3C6,0xFF);
        switch (VGA_triggerDac()) {
            case 0x44:  dac = DAC_MU99DAC;  goto FoundDAC;
            case 0x88:
            case 0x89:  dac = DAC_MU16DAC;  goto FoundDAC;
            case 0x82:  dac = DAC_MU49DAC;  goto FoundDAC;
            case 0x8E:  dac = DAC_MU18DAC;    goto FoundDAC;
            }
        if (setCommand(0x1) == 0xAA) {
            dac = DAC_ALG12DAC;
            goto FoundDAC;      /* We have an Avance Logic 1201 DAC     */
            }
        if (VGA_dacToCommand() == 0xC0) {
            dac = DAC_CH98DAC;
            goto FoundDAC;
            }
        if (testDacBit(0x10,oldPel,oldCommand)) {
            dac = DAC_TR8DAC;       /* We have a Trident TR8001 DAC         */
            goto FoundDAC;
            }
        VGA_dacToPel();
        outpb(0x3C6,0xFF);
        VGA_dacToCommand();
        outpb(0x3C6,0);
        VGA_dacToPel();
        VGA_dacToCommand();
        outpb(0x3C6,0x7F);
        VGA_dacToCommand();
        x = inpb(0x3C6);
        VGA_dacToCommand();
        VGA_dacToCommand();
        outpb(0x3C6,0xFF);
        VGA_dacToCommand();
        v = inpb(0x3C6);
        VGA_dacToCommand();
        outpb(0x3C6,oldCommand);
        VGA_dacToPel();
        if ((x == 0x60) && (v == 0xE0)) {
            dac = DAC_UMC17DAC; /* We have a UMC UM70c178 16 bit DAC    */
            goto FoundDAC;
            }
        if ((x == 0x7F) && (v == 0xFE)) {
            dac = DAC_SC7DAC;       /* We have a Sierra SC11487 16 bit DAC  */
            goto FoundDAC;
            }
        dac = DAC_SC3DAC;           /* We have a Sierra SC11483 15 bit DAC  */
        goto FoundDAC;
        }

FoundDAC:
    VGA_dacToCommand();
    outpb(0x3C6,oldCommand);    /* Restore old command register         */
    VGA_dacToPel();
    outpb(0x3C6,oldPel);        /* Restore old PEL register             */
#endif
    /* Now we detect for advanced DAC's that require access to the
     * DAC register set through the DAC RS[3:2] page select lines.
     */
    if (dac == DAC_INTERNAL) {
        /* Save old register values */
        oldPel = DV.chip.rdDACReg(dacSTDpelMask);
        oldCommand = DV.chip.rdDACReg(dacSTDpelMask+4);

        /* Check for ATI68860 series DAC's. We must do this before
         * running the following detection code as the following code
         * will hang up the ATI DAC
         */
        x = DV.chip.rdDACReg(dacATI6id);
        DV.chip.wrDACReg(dacATI6id,0xFF);
        y = DV.chip.rdDACReg(dacATI6id);
        switch (y) {
            case 0xE0:              /* Rev B */
            case 0xD0:              /* Rev C */
                dac = DAC_ATI60DAC; /* We have an ATI 68860 DAC     */
                goto Done;
            }
        DV.chip.wrDACReg(dacATI6id,x);

        /* Check for TLC3407x series DAC's */
        DV.chip.wrDACReg(dacSTDpelMask,0);
        x = DV.chip.rdDACReg(dacTLCtest);
        DV.chip.wrDACReg(dacTLCtest,3);
        y = DV.chip.rdDACReg(dacTLCtest);
        DV.chip.wrDACReg(dacTLCtest,x);
        switch (y) {
            case 0x75:
            case 0x76:
            case 0x77:
                dac = DAC_TL5DAC;       /* We have a TLC34075 DAC       */
                goto Done;
            }

        /* Check for TVP3020/5 series DAC's */
        x = DV.chip.rdDACReg(dacTVPindex);
        DV.chip.wrDACReg(dacTVPindex,dacTVPid);
        v = DV.chip.rdDACReg(dacTVPdata);
        DV.chip.wrDACReg(dacTVPindex,x);
        switch (v) {
            case 0x20:
                dac = DAC_TV0DAC;       /* We have a TI TVP3020         */
                goto Done;
            case 0x25:
                dac = DAC_TV5DAC;       /* We have a TI TVP3025         */
                goto Done;
            }

        /* Check for TVP3026/30 series DAC's */
        x = DV.chip.rdDACReg(dacTVP6index);
        DV.chip.wrDACReg(dacTVP6index,dacTVPid);
        v = DV.chip.rdDACReg(dacTVP6data);
        DV.chip.wrDACReg(dacTVP6index,x);
        switch (v) {
            case 0x26:
                dac = DAC_TV6DAC;       /* We have a TI TVP3026         */
                goto Done;
            case 0x30:
                dac = DAC_TV30DAC;  /* We have a TI TVP3030         */
                goto Done;
            }

        /* Check if we have functioning RS[3:2] lines on advanced DAC's */
        x = DV.chip.rdDACReg(dacSTDpelMask+8);
        DV.chip.wrDACReg(dacSTDpelMask,oldPel ^ 0xFF);
        y = DV.chip.rdDACReg(dacSTDpelMask+4);
        z = DV.chip.rdDACReg(dacSTDpelMask+8);
        if ((oldCommand == y) || (x == z)) {
#ifdef  INCLUDE_BT481_DETECT
            /* Check for Bt481/2 DAC's. This code misdetects on some
             * Tridents, so it is not used unless we specifically need
             * it.
             */
            x = DV.chip.rdDACReg(dacBT1cmdA);
            DV.chip.wrDACReg(dacBT1cmdA,x & 0xFE);
            y = DV.chip.rdDACReg(dacSTDpelMask);
            /* Switch to Bt481 indexed */
            DV.chip.wrDACReg(dacBT1cmdA,x | 1);
            DV.chip.wrDACReg(dacSTDwrInx,dacBTIipixm);
            DV.chip.wrDACReg(dacSTDpelMask,y ^ 0x55);
            z = DV.chip.rdDACReg(dacSTDpelMask);
            DV.chip.wrDACReg(dacSTDwrInx,dacBTIcurX);
            DV.chip.wrDACReg(dacSTDpelMask,y ^ 0xAA);
            v = DV.chip.rdDACReg(dacSTDpelMask);
            /* Back to standard registers */
            DV.chip.wrDACReg(dacBT1cmdA,x & 0xFE);
            if ((y == DV.chip.rdDACReg(dacSTDpelMask)) && (z == (y ^ 0x55))) {
                if (v == (y ^ 0xAA)) {
                    dac = DAC_BT2DAC;       /* We have a Bt482 DAC          */
                    goto Done;
                    }
                dac = DAC_BT1DAC;           /* We have a Bt481 DAC          */
                goto Done;
                }
#endif
            /* Check for IBM RGB 514/24/25/28 DAC's */
            x = DV.chip.rdDACReg(dacIBMind0);
            DV.chip.wrDACReg(dacIBMind1,0);
            DV.chip.wrDACReg(dacIBMind0,dacIBMrev);
            y = DV.chip.rdDACReg(dacIBMdata);
            DV.chip.wrDACReg(dacIBMind0,dacIBMid);
            y |= DV.chip.rdDACReg(dacIBMdata) << 8;
            switch (y) {
                case 0x2E0:
                case 0x2F0:
                    dac = DAC_IBM24DAC; /* We have an IBM 524/526/528   */
                    goto Done;
                case 0x2C0:
                    dac = DAC_IBM26DAC; /* We have an IBM 526 DAC       */
                    goto Done;
                case 0x280:
                    dac = DAC_IBM26DBDAC;   /* We have an IBM 526DB DAC     */
                    goto Done;
                case 0x1E0:
                case 0x1F0:
                    dac = DAC_IBM25DAC; /* We have an IBM 514/525 DAC   */
                    goto Done;
                case 0x30C0:
                    dac = DAC_IBM624DAC;    /* We have an IBM 624 DAC       */
                    goto Done;
                }

            /* Force the Bt485 status register */
            VGA_dacToCommand();
            DV.chip.wrDACReg(dacSTDwrInx,0);
            DV.chip.wrDACReg(dacBTcmd0,DV.chip.rdDACReg(dacBTcmd0) & 0x7F);
            x = DV.chip.rdDACReg(dacBTstat);
            switch (x & 0xF0) {
                case 0x40:
                    dac = DAC_ATT4DAC;  /* We have an ATT 20c504 DAC    */
                    goto Done;
                case 0xD0:
                    dac = DAC_ATT5DAC;  /* We have an ATT 20c505 DAC    */
                    goto Done;
                default:
                    if ((0x20 <= x && x <= 0xB0) || x == 0xFF) {
                        y = DV.chip.rdDACReg(dacBTcmd0);
                        /* Enable Stat/Cmd3 */
                        DV.chip.wrDACReg(dacBTcmd0,y | 0x80);
                        /* Force the cmd3 reg */
                        DV.chip.wrDACReg(dacSTDwrInx,1);
                        z = DV.chip.rdDACReg(dacBTstat);
                        if (z == x) {
                            DV.chip.wrDACReg(dacBTstat,x ^ 0x55);
                            /* Force the stat reg */
                            DV.chip.wrDACReg(dacSTDwrInx,0);
                            v = DV.chip.rdDACReg(dacBTstat);
                            /* Force the stat reg */
                            DV.chip.wrDACReg(dacSTDwrInx,1);
                            DV.chip.wrDACReg(dacBTstat,z);
                            if (v == x)
                                z = x + 1;  /* Ie. Bt485    */
                            }
                        DV.chip.wrDACReg(dacBTcmd0,y);
                        if (x == z) {
                            dac = DAC_BT4DAC;   /* We have a Bt484 DAC      */
                            goto Done;
                            }
                        else {
                            dac = DAC_BT5DAC;   /* We have a Bt485 DAC      */
                            goto Done;
                            }
                        }
                    break;
                }
            }
Done:   DV.chip.wrDACReg(dacSTDpelMask,oldPel);
        DV.chip.wrDACReg(dacSTDpelMask+4,oldCommand);
        }
    return dac;
}
