/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Generic device detection module.
*
****************************************************************************/

#include "snap/graphics/driver.h"

/*--------------------------- Global variables ----------------------------*/

static GA_sharedInfo localInfo = {0,-1};
int                 DRV_loadInGUI;          // Amount of memory forced for detect
char                DRV_path[256];
GA_driver           DV;                     // Global device driver structure
GA_monitor          monitor[GA_MAX_HEADS];  // Global monitor records
#ifdef  __DRIVER__
N_imports _VARAPI   _N_imports;             // Imports to OS specific functions
PM_imports _VARAPI  _PM_imports;            // Imports to PM library functions
DRV_imports _VARAPI _DRV_imports;           // Imports to DDC specific functions
#endif
GA_configInfo       DRV_configInfo = {
    sizeof(GA_configInfo),
    CONFIG_SIGNATURE,
    CONFIG_SIGNATURE,
    CONFIG_SIGNATURE,
    CONFIG_SIGNATURE,
    DRIVER_VERSION,
    DRIVER_DATE,
    false,
    "",
    0,
    };

//--------------------------------------------------------------------------
// This mode enumeration table is now longer used (all default modes are
// set up in the generic code now), however we do need at least text
// modes and 640x480 graphics mode for the discrete clock analysis code for
// older graphics cards. Note that we need all text modes so we can
// correctly determine the default text mode is use when the driver is
// first loaded (thankfully we probably won't be added many custom text
// modes!
//--------------------------------------------------------------------------

#ifndef __ANALYSE__
static GA_mode DRV_defaultModeList[] = {
    {80,25,0},
    {80,43,0},
    {80,50,0},
    {80,60,0},
    {100,25,0},
    {100,43,0},
    {100,50,0},
    {100,60,0},
    {132,25,0},
    {132,43,0},
    {132,50,0},
    {132,60,0},
    {640,480,8},
    };
#endif

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
HEADER:
graphics/internal.h

PARAMETERS:
mfrName         - Place to store the manufacturer name
chipsetNames    - Place to store the list of chipset names

REMARKS:
Returns the chipset driver manufacturer name and list of supported chipsets.
****************************************************************************/
void _CEXPORT GA_getChipsetNames(
    char **mfr,
    char ***chipsetNames)
{
    *mfr = DRV_manufacturerName;
    *chipsetNames = DRV_chipsetNames;
}

/****************************************************************************
HEADER:
graphics/internal.h

PARAMETERS:
info    - Place to return the certification information

REMARKS:
Returns the chipset driver certification information structures.
****************************************************************************/
void _CEXPORT GA_getCertifyInfo(
    GA_certifyInfo *info)
{
    COPY_STRUCTURE(info,&DRV_certifyInfo);
}

#ifdef  __DRIVER__
/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler for non-imported N_imports.
****************************************************************************/
static void _N_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported N_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler for non-imported GA_imports.
****************************************************************************/
static void _DRV_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported DRV_imports import function called! Please re-compile!\n");
}
#endif

/****************************************************************************
HEADER:
graphics/driver.h

PARAMETERS:
path        - Path to Binary Portable DLL
deviceIndex - Index of the display device in use
info        - Pointer to the PCI configuration space info for the device
nImp        - Generic SNAP imports
drvImp      - Driver specific imports
af          - Pointer to optional VBE/AF driver

RETURNS:
Pointer to exported function list

REMARKS:
Does generic Plug and Play device detection and fills in the device resource
allocation information for the device so that the driver loader can
allocate the necessary OS resources for the device before calling the
real detection code.
****************************************************************************/
GA_devCtx * _CEXPORT GA_initDriver(
    ibool shared,
    const char *path,
    int deviceIndex,
    PCIDeviceInfo *info,
    GA_monitor *pMonitor,
    N_imports *nImp,
    PM_imports *pmImp,
    DRV_imports *drvImp,
    GA_options *licensedOptions,
    GA_globalOptions *globalOpt,
    ibool sharedIRQ,
    GA_exports *exportTable,
    GA_sharedInfo *sharedInfo,
    int loadInGUI)
{
#ifdef  __DRIVER__
    int     i,max;
    ulong   *p;

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_N_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_N_imports; i < max; i++)
        *p++ = (ulong)_N_fatalErrorHandler;
    max = sizeof(_PM_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;
    max = sizeof(_DRV_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_DRV_imports; i < max; i++)
        *p++ = (ulong)_DRV_fatalErrorHandler;

    // Now copy all our imported functions
    memcpy(&_N_imports,nImp,MIN(sizeof(_N_imports),nImp->dwSize));
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    memcpy(&_DRV_imports,drvImp,MIN(sizeof(_DRV_imports),drvImp->dwSize));
#else
    (void)nImp;
    (void)pmImp;
    (void)drvImp;
#endif
    (void)exportTable;
    monitor[0] = *pMonitor;
    strcpy(DRV_path,path);

    // Save the forced memory size
    DRV_loadInGUI = loadInGUI;

    // Initialise default information
    memset(&DV,0,sizeof(DV));
    if (info)
        memcpy(&DV.PCIConfig,info,sizeof(DV.PCIConfig));
    strcpy(DC.Signature,GA_SIGNATURE);
    DC.Version = GA_VERSION;
    DC.DriverRev = (GA_RELEASE_MAJOR << 8) | GA_RELEASE_MINOR;
    DC.DeviceIndex = deviceIndex;
    strcpy(DC.OemVendorName,SCITECH_STR);
    strcpy(DC.OemCopyright,COPYRIGHT_STR);
    DC.AvailableModes = (void*)&DV.AvailableModes;
    DC.TextSize = 32 * 1024;
    DC.TextBasePtr = 0xB8000;
    DC.BankSize = 64 * 1024;
    DC.BankedBasePtr = 0xA0000;
    DV.sharedIRQ = sharedIRQ;
    DV.s = &localInfo;
    if (sharedInfo)
        DV.s = sharedInfo;

    // Initialise list of licensed options
    if (licensedOptions)
        DV.licensed = *licensedOptions;
    else {
        rep_stosb(&DV.licensed,sizeof(DV.licensed),1);
        DV.licensed.dwSize = sizeof(DV.licensed);
        DV.licensed.resolutions[0] = 0xFFFFFFFFU;
        DV.licensed.colorDepths[0] = 0xFF;
        DV.licensed.maxRefresh = 0xFF;
        }
    if (globalOpt)
        DV.globalOpt = *globalOpt;
    else
        rep_stosb(&DV.globalOpt,sizeof(DV.globalOpt),0);

#ifndef __ANALYSE__
    // Wakeup the graphics controller if we are secondary
    if (!CHIP_wakeup(DRV_loadInGUI))
        return NULL;
#endif

    // Now run the chipset specific PnP detection code
    if (!PNP_detect())
        return NULL;

    // Fill in the driver init function pointers
    DC.loader.InitDriver = DRV_initDriver;
    DC.loader.QueryFunctions = DRV_queryFunctions;
    DC.loader.UnloadDriver = DRV_unloadDriver;
    return &DC;
}

/****************************************************************************
DESCRIPTION:
Detects that the device is installed and configures the driver.

HEADER:
graphics/driver.h

RETURNS:
True if hardware detection correctly, false if not.

REMARKS:
Does generic device detection and fills in the generic components of the
SNAP device context header information after calling the chipset
specific detection code.
****************************************************************************/
N_int32 NAPI DRV_initDriver(void)
{
    int                 status,i;
    N_uint16            *m;
#ifndef __ANALYSE__
    int                 numModes,mode;
    FILE                *f;
    char                filename[256],*chipName;
    GA_certifyChipInfo  *certified;
#endif

#ifdef  TEST_HARNESS
    if (DRV_chipsetNames[(int)DRV_certifyInfo.MaxCertifiedChips] != NULL)
        PM_fatalError("DRV_chipsetNames[] *must* be NULL terminated!");
#endif

#define INIT_STRUCTURE(p)       \
    memset(&p,0xFF,sizeof(p));  \
    p.dwSize = sizeof(p);

    // Set all device specific vectors to NOT_IMPLEMENTED by default
    INIT_STRUCTURE(DV.init);
    INIT_STRUCTURE(DV.driver);
    INIT_STRUCTURE(DV.cursor);
    INIT_STRUCTURE(DV.video);
    INIT_STRUCTURE(DV.dpms);
    INIT_STRUCTURE(DV.sci);
    INIT_STRUCTURE(DV.vbe);
    INIT_STRUCTURE(DV.buf);
    INIT_STRUCTURE(DV.clip);
    INIT_STRUCTURE(DV.state2d);
    INIT_STRUCTURE(DV.draw_8bpp);
    INIT_STRUCTURE(DV.draw_16bpp);
    INIT_STRUCTURE(DV.draw_24bpp);
    INIT_STRUCTURE(DV.draw_32bpp);
    INIT_STRUCTURE(DV.drawSW);

    // Configure all the generic API functions for the driver.
    DV.init.GetConfigInfo               = DRV_getConfigInfo;
    DV.init.SetModeProfile              = DRV_setModeProfile;
    DV.init.GetOptions                  = DRV_getOptions;
    DV.init.SetOptions                  = DRV_setOptions;
    DV.init.GetVideoModeInfo            = DRV_GetVideoModeInfo;
    DV.init.SetVideoMode                = DRV_SetVideoMode;
    DV.init.GetVideoMode                = DRV_GetVideoMode;
    DV.init.GetCustomVideoModeInfo      = DRV_GetCustomVideoModeInfo;
    DV.init.SetCustomVideoMode          = DRV_SetCustomVideoMode;
    DV.init.GetClosestPixelClockV1      = DRV_GetClosestPixelClockV1;
    DV.init.GetClosestPixelClock        = DRV_GetClosestPixelClock;
    DV.init.GetCRTCTimings              = DRV_GetCRTCTimings;
    DV.init.SetCRTCTimings              = DRV_SetCRTCTimings;
    DV.init.SaveCRTCTimings             = DRV_SaveCRTCTimings;
    DV.init.SetGlobalRefresh            = DRV_SetGlobalRefresh;
    DV.init.SaveRestoreState            = NULL;         // TODO!!
    DV.init.SetDisplayOutput            = DRV_SetDisplayOutput;
    DV.init.GetDisplayOutput            = DRV_GetDisplayOutput;
    DV.init.SetSoftwareRenderFuncs      = DRV_SetSoftwareRenderFuncs;
    DV.init.GetUniqueFilename           = DRV_getUniqueFilename;
    DV.init.GetMonitorInfo              = DRV_getMonitorInfo;
    DV.init.SetMonitorInfo              = DRV_setMonitorInfo;
    DV.init.GetCurrentVideoModeInfo     = DRV_GetCurrentVideoModeInfo;
    DV.init.GetCertifyInfo              = DRV_getCertifyInfo;
    DV.init.GetCurrentRefreshRate       = DRV_GetCurrentRefreshRate;
    DV.init.PerformDisplaySwitch        = DRV_PerformDisplaySwitch;
    DV.init.GetVideoModeInfoExt         = DRV_GetVideoModeInfoExt;
    DV.init.GetCustomVideoModeInfoExt   = DRV_GetCustomVideoModeInfoExt;
    DV.init.SwitchPhysicalResolution    = DRV_SwitchPhysicalResolution;
    DV.init.GetNumberOfHeads            = DRV_GetNumberOfHeads;
    DV.init.SetActiveHead               = DRV_SetActiveHead;
    DV.init.GetActiveHead               = DRV_GetActiveHead;
    DV.init.SetRef2dPointer             = DRV_SetRef2dPointer;
    DV.init.SetDisplayOutputExt         = DRV_SetDisplayOutputExt;
    DV.init.GetDisplayOutputExt         = DRV_GetDisplayOutputExt;
    DV.dpms.DPMSdetect                  = DRV_DPMSdetect;
    DV.dpms.DPMSsetState                = DRV_DPMSsetState;
    DV.sci.SCIdetect                    = DRV_SCIdetect;
    DV.chip.getTextModeFlags            = DRV_getTextModeFlags;
    DV.chip.getChipData                 = DRV_getChipData;
    DV.chip.configChipsetFlags          = DRV_configChipsetFlags;
    DV.chip.setChipsetRegistersExt      = DRV_setChipsetRegistersExt;
    DV.pll.adjustMaxPixelClock          = DRV_adjustMaxPixelClock;
    DV.vbe.SetBytesPerLine              = DRVVBE_SetBytesPerLine;
    DV.vbe.Set8BitDAC                   = DRVVBE_Set8BitDAC;
    DV.vbe.SetPaletteData               = DRVVBE_SetPaletteData;
    DV.vbe.GetPaletteData               = DRVVBE_GetPaletteData;
    DV.resScanLines                     = 0;

    // Setup default vectors to VGA compatible functions
    VGA_init();

    // Now detect the install chipset configuration
    if (!CHIP_detect())
        return nNotDetected;

#ifndef __ANALYSE__
    // Check if chipset is certified, fail if not!
    certified = NULL;
    if (DRV_certifyInfo.NumCertifiedChips) {
        certified = DRV_certifyInfo.CertifiedCards;
        chipName = DRV_chipsetNames[DV.chipID];
        for (i = 0; i < DRV_certifyInfo.NumCertifiedChips; i++,certified++) {
            if (strcmp(chipName,certified->ChipsetName) == 0) {
                if (certified->CertifyVersion >= DV.globalOpt.wCertifiedVersion) {
                    // This chipset is certified, so allow driver to load
                    break;
                    }
                }
            }

        // Check if the chipset is certified, and bail if not and the
        // options do not allow certified drivers to run.
        if (i == DRV_certifyInfo.NumCertifiedChips) {
            if (!DV.globalOpt.bAllowNonCertified)
                return nNotCertified;
            certified = NULL;
            }
        }
#endif

    // Save the HW framebuffer location and size. We need to do that
    // here because the memory window may slide around at runtime depending
    // on the display mode, but we need the original values so that we
    // can properly map values and report the 'slid' values to apps for
    // each mode in GetVideoModeInfo.
    DV.HWLinearMem      = DC.LinearMem;
    DV.HWLinearSize     = DC.LinearSize;
    DV.HWLinearBasePtr  = DC.LinearBasePtr;

    // Set the configuration options to all enabled by default.
#ifndef __ANALYSE__
    DV.prof.dwSize = sizeof(DV.prof);
    memcpy(DV.prof.m.modeList,DRV_defaultModeList,sizeof(DRV_defaultModeList));
    DV.prof.m.numModes = ARR_SIZE(DRV_defaultModeList);
#endif
    rep_stosb(&DV.opt,sizeof(DV.opt),1);
    DV.opt.dwSize = sizeof(DV.opt);
    DV.opt.bLoadedFromDisk = 0;
    DV.opt.memoryClock = 0;
    DV.opt.defaultMemoryClock = 0;
    DV.opt.maxMemoryClock = 0;
    DV.opt.outputDevice = DV.outputDevice[0] = gaOUTPUT_CRT;
    DV.opt.bDebugMode = 0;
    DV.opt.bMultiDisplay = false;
    DV.opt.bPortrait = false;
    DV.opt.bFlipped = false;
    DV.opt.bInvertColors = false;
    DV.opt.bZoom = false;
    DV.opt.bMultiHead = false;
    DV.opt.RTCFrequency = 4096;
    DV.opt.RTCAdvanceTicks = 1;
    DV.opt.glassesType = gaGlassesFreeRunning;
    DV.opt.stereoDevice = 0x01010101;
    DV.opt.ioPort = 0;
    DV.opt.ioAndMask = 0;
    DV.opt.ioLeftOrMask = 0;
    DV.opt.ioRightOrMask = 0;
    DV.opt.ioOffOrMask = 0;
    DV.opt.vSyncWidthLeft = 4;
    DV.opt.vSyncWidthRight = 3;
    DV.opt.text80x43Height = 8;
    DV.opt.text80x50Height = 8;
    DV.opt.text80x60Height = 8;
    DV.opt.text100x43Height = 8;
    DV.opt.text100x50Height = 8;
    DV.opt.text100x60Height = 8;
    DV.opt.text132x43Height = 8;
    DV.opt.text132x50Height = 8;
    DV.opt.text132x60Height = 8;
    DV.opt.tripleScanLowRes = 0;
    DV.opt.doubleScan512 = 1;
    DV.opt.stereoRefresh = 120;
    DV.opt.stereoRefreshInterlaced = 120;
    DV.opt.stereoRefreshWindowed = 120;
    DV.opt.stereoRefreshWindowedInterlaced = 120;
    DV.opt.stereoBlankInterval = 0;
    DV.opt.stereoBlankIntervalPercent = 500;
    DV.opt.LCDPanelWidth = -1;
    DV.opt.LCDPanelHeight = -1;
    DV.opt.bLCDExpand = 1;
    DV.opt.bHardwareStereoSync = 0;
    DV.opt.TVMaxWidth = 800;
    DV.opt.TVMaxHeight = 600;
    DV.opt.bCompressedFB = true;

    // Set default linear ramp for gamma correction
    for (i = 0; i < 256; i++) {
        DV.opt.gammaRamp[i].Red = i * 256;
        DV.opt.gammaRamp[i].Green = i * 256;
        DV.opt.gammaRamp[i].Blue = i * 256;
        }

    // Set the default minimum pixel clock frequency
    DV.pixelClock.minFreq = N_FIXCONST(12.58);

    // Set default values for minimum R values for clock formula's
    DV.pixelClock.minR = 0;
    DV.memoryClock.minR = 0;

    // Setup default values for memory overrides. Most drivers wont set these
    // values at all, but we now allow the useable memory to be special
    // cased on a per color depth basis (ie: to allow more memory to be used
    // for internal pattern caches etc based on the color depth).
    DV.VGA8Memory = DV.VGA16Memory = DV.VGA24Memory = DV.VGA32Memory = - 1;
    DV.NV8Memory = DV.NV16Memory = DV.NV24Memory = DV.NV32Memory = - 1;

    // Now configure the driver based on all the detected information and
    // any options information loaded from disk
    if ((status = DRV_configure(DRV_loadInGUI)) != nOK)
        return status;

    // Setup dynamic display switch polling function for laptops
    if (DV.chip.getPendingOptions)
        DV.init.PollForDisplaySwitch = DRV_PollForDisplaySwitch;

    // Setup the default functions for extended palette and gamma correction
    // functions, which simply do translation to the non-extended format.
    if (DV.driver.SetPaletteDataExt == CNOT_IMPLEMENTED) {
        DV.driver.SetPaletteDataExt         = DRV_SetPaletteDataExt;
        DV.driver.GetPaletteDataExt         = DRV_GetPaletteDataExt;
        }
    if (IS_CVALID_PTR(DV.driver.SetGammaCorrectData)) {
        DV.driver.SetGammaCorrectDataExt    = DRV_SetGammaCorrectDataExt;
        DV.driver.GetGammaCorrectDataExt    = DRV_GetGammaCorrectDataExt;
        }
    else {
        DV.driver.SetGammaCorrectDataExt    = CNOT_SUPPORTED;
        DV.driver.GetGammaCorrectDataExt    = CNOT_SUPPORTED;
        }

    // Now get the default chipset specific options values, and override
    // those with values read from disk.
#ifndef __ANALYSE__
    DRV_getUniqueFilename(filename,2);
    if ((f = fopen(filename,"rb")) != NULL) {
        fread(&DV.opt,1,sizeof(DV.opt),f);
        fclose(f);
        DV.opt.bLoadedFromDisk = 1;
        }
#endif
    DV.chip.getDefaultOptions();
    DV.opt.outputDevice = DV.outputDevice[0];

    // Determine the default option for the bLCDUseBIOS flag. If the
    // system boots up in either LCD only mode or in Simultaneous mode,
    // we want to always use the BIOS even if we switch to CRT mode. If
    // the system boot up in CRT only mode, we don't use the BIOS unless
    // we switch to LCD panel mode. This is designed to handle the most
    // common cases when the user needs the 'Alt-Func' type hot key
    // handling on their laptop to properly support switching between the
    // CRT and LCD panel. If we use the BIOS, it always works correctly.
    // If we do a NoBIOS mode set, it usually does not work properly unless
    // the SNAP API's are used to make the switch.
    if (DV.globalOpt.bLCDUseBIOS == gaLCDUseBIOS_Auto) {
        if (DV.outputDevice[0] & gaOUTPUT_LCD)
            DV.LCDUseBIOS = true;
        else
            DV.LCDUseBIOS = false;
        }
    else if (DV.globalOpt.bLCDUseBIOS == gaLCDUseBIOS_On)
        DV.LCDUseBIOS = true;
    else
        DV.LCDUseBIOS = false;

    // Now filter all options based on licensing
    DRV_setOptions(&DV.opt);

    // Disable multi-head features by overridding chip functions
    // with the default single head functions.
    if (!DV.licensed.bMultiHead) {
        DV.init.GetNumberOfHeads    = DRV_GetNumberOfHeads;
        DV.init.SetActiveHead       = DRV_SetActiveHead;
        DV.init.GetActiveHead       = DRV_GetActiveHead;
        }

    // Now construct the list of available modes
#ifndef __ANALYSE__
    numModes = DV.prof.m.numModes;
    mode = 0x100;
    m = DV.AvailableModes;
    while (numModes--) {
        *m++ = mode;
        mode++;
        }
    *m = 0xFFFF;
#endif

    // Construct the detected configuration info block
    DRV_configInfo.dwSize = sizeof(DRV_configInfo);
    strcpy(DRV_configInfo.ManufacturerName,DRV_manufacturerName);
    strcpy(DRV_configInfo.ChipsetName,DRV_chipsetNames[DV.chipID]);
    strcpy(DRV_configInfo.DACName,DRV_DACNames[DV.dacID]);
    strcpy(DRV_configInfo.ClockName,DRV_clockNames[DV.clockID]);
#ifndef __ANALYSE__
    if (certified) {
        DRV_configInfo.Certified = true;
        strcpy(DRV_configInfo.CertifiedDate,certified->CertifiedDate);
        DRV_configInfo.CertifyVersion = certified->CertifyVersion;
        }
#endif

    // Determine if the controller has any NonVGA modes
    for (m = DC.AvailableModes; *m != 0xFFFF; m++) {
        N_uint16        physX,physY;
        GA_modeEntry    *entry;

        if ((entry = DRV_validateMode(*m,0,gaActiveHeadPrimary,&physX,&physY)) == NULL)
            continue;
        if (entry->flags & gaNonVGA) {
            DC.Attributes |= gaHaveNonVGAMode;
            break;
            }
        }

    // Disable global attributes for disabled or unlicensed features
    if (!DV.opt.bDoubleBuffer)
        DC.Attributes &= ~gaHaveDisplayStart;
    if (!DV.opt.bTripleBuffer)
        DC.Attributes &= ~gaHaveTripleBuffer;
    if (!DV.opt.bHardwareAccel2D)
        DC.Attributes &= ~(gaHaveAccel2D | gaHaveAccelVideo | gaHaveAccel3D);
    if (!DV.opt.bStereo)
        DC.Attributes &= ~gaHaveStereo;
    if (DV.opt.bHardwareStereoSync)
        DC.Attributes |= gaHaveHWStereoSync;
    if (!DV.opt.bHardwareCursor)
        DC.Attributes &= ~gaHaveHWCursor;
    if (!DV.opt.bHardwareVideo)
        DC.Attributes &= ~gaHaveAccelVideo;
    if (!DV.opt.bHardwareAccel3D)
        DC.Attributes &= ~gaHaveAccel3D;

    // Analyse the discrete clocks if required
#if !defined(__ANALYSE__) && defined(INCLUDE_DISCRETECLOCKS)
    DRV_analyseDiscreteClocks();
#endif
    return nOK;
}

/****************************************************************************
DESCRIPTION:
Performs any steps necessary to unload the driver from memory.

HEADER:
graphics/driver.h

REMARKS:
Performs any steps necessary to unload the driver from memory. At present
this simply includes calling any chipset specific code to unload the driver
from memory.
****************************************************************************/
void NAPI DRV_unloadDriver(void)
{
    if (DV.chip.unloadDriver)
        DV.chip.unloadDriver();
}

