/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the generic SNAP Graphics
*               Architecture API functions based on chipset specific
*               support modules.
*
****************************************************************************/

#include "snap/graphics/driver.h"

/*--------------------------- Global variables ----------------------------*/

#define MAX_REFRESH_RATES   20

static N_int32          DRV_refreshList[MAX_REFRESH_RATES+1];
static GA_paletteExt    scratchPalExt[256];
#ifdef DEBUG_BLANKS
extern int              hBlankStart,hBlankEnd,vBlankStart,vBlankEnd;
#endif

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
RETURNS:
Pointer to the chip data entry structure for the chipset.

REMARKS:
This function returns the correct chip data tables depending on whether
we are displaying output on the CRT, LCD or TVOut connectors.
****************************************************************************/
GA_chipEntry *DRV_getChipData(void)
{
    return DRV_chipData[DV.chipID];
}

/****************************************************************************
PARAMETERS:
mode        - Mode number to find internal mode entry for

RETURNS:
Pointer to the mode entry structure for the mode, NULL if not found.

REMARKS:
This function searches the table of valid mode information for the
specific mode, and returns a pointer to the entry in the table.
****************************************************************************/
static GA_modeEntry *DRV_validateModeXY(
    GA_mode *e,
    N_int32 outputDevice,
    N_int32 outputHead,
    N_uint16 *physicalXRes,
    N_uint16 *physicalYRes)
{
    static GA_modeEntry entry;

    switch (e->bits) {
        case 0:     entry.bytesPerLine = (e->xRes * 2); break;
        case 4:     entry.bytesPerLine = (e->xRes+7)/8; break;
        case 8:     entry.bytesPerLine = e->xRes;       break;
        case 15:
        case 16:    entry.bytesPerLine = e->xRes * 2;   break;
        case 24:    entry.bytesPerLine = e->xRes * 3;   break;
        case 32:    entry.bytesPerLine = e->xRes * 4;   break;
        }
    *physicalXRes = entry.xRes = e->xRes;
    *physicalYRes = entry.yRes = e->yRes;
    entry.bits = e->bits;
    if (e->bits == 0)
        entry.flags = DV.chip.getTextModeFlags(e);
    else if (DV.chip.getModeFlagsExt)
        entry.flags = DV.chip.getModeFlagsExt(e,outputDevice,outputHead,physicalXRes,physicalYRes);
    else
        entry.flags = DV.chip.getModeFlags(e);
    if (entry.flags == 0)
        return NULL;

    // Disable acceleration if not enabled
    if (!DV.opt.bHardwareAccel2D)
        entry.flags &= ~gaAccMode;

    // Check to see if the mode should be a panning mode or not.
    if (outputDevice & (gaOUTPUT_LCD | gaOUTPUT_DFP)) {
        if ((entry.xRes > DV.opt.LCDPanelWidth) ||
            (entry.yRes > DV.opt.LCDPanelHeight)) {
            // We don't support panning modes for DFP flat panel monitors
            // as this is not necessary. We only support it for LCD
            // monitors and TV's (which are usually used on conjunction
            // with a CRT display).
            if (outputDevice & gaOUTPUT_DFP)
                return NULL;
            entry.flags |= gaPanningMode;
            *physicalXRes = DV.opt.LCDPanelWidth;
            *physicalYRes = DV.opt.LCDPanelHeight;
            }
        }
    if (outputDevice & gaOUTPUT_TV) {
        if ((entry.xRes > DV.opt.TVMaxWidth) ||
            (entry.yRes > DV.opt.TVMaxHeight)) {
            entry.flags |= gaPanningMode;
            *physicalXRes = DV.opt.TVMaxWidth;
            *physicalYRes = DV.opt.TVMaxHeight;
            }
        }
    return &entry;
}

/****************************************************************************
PARAMETERS:
mode        - Mode number to find internal mode entry for

RETURNS:
Pointer to the mode entry structure for the mode, NULL if not found.

REMARKS:
This function searches the table of valid mode information for the
specific mode, and returns a pointer to the entry in the table.
****************************************************************************/
GA_modeEntry *DRV_validateMode(
    N_int32 mode,
    N_int32 outputDevice,
    N_int32 outputHead,
    N_uint16 *physicalXRes,
    N_uint16 *physicalYRes)
{
    GA_mode *e = DV.prof.m.modeList;

    if (!DV.chip.getModeFlags && !DV.chip.getModeFlagsExt)
        PM_fatalError("You need to implement CHIP_getModeFlags or CHIP_getModeFlagsExt!\n");
    mode &= 0xFF;
    if (mode < DV.prof.m.numModes)
        return DRV_validateModeXY(e+mode,outputDevice,outputHead,physicalXRes,physicalYRes);
    return NULL;
}

/****************************************************************************
PARAMETERS:
clock   - Pixel clock to search for
bits    - Bits per pixel for the mode
select  - Place to store the close select index
divide  - Place to store the divide factor (1, 2 or 4)

RETURNS:
Pointer to the mode entry structure for the mode, NULL if not found.

REMARKS:
This function searches the table of valid mode information for the
specific mode, and returns a pointer to the entry in the table.
****************************************************************************/
N_fix32 DRV_findClosestDiscreteClock(
    N_fix32 clock,
    int bits,
    int *select,
    int *divide)
{
    int     i;
    N_fix32 diff,best,*valid = DV.pixelClock.validClocks;

    // First search for a native clock that suits our needs
    diff = 0x7FFFFFFF;
    for (i =  0; i < DV.pixelClock.numClocks; i++) {
        N_fix32 d = ABS(clock - valid[i]);
        if (d < diff) {
            diff = d;
            best = valid[i];
            *select = i;
            *divide = 1;
            }
        }

    // Text modes do not support clock halving
    if (bits == 0)
        return best;

    // Now search for a better halved clock if this is supported
    if (bits == 8 && !(DV.pixelClock.flags & gaPLL_clockHalve8Bit))
        return best;
    if ((bits == 15 || bits == 16) && !(DV.pixelClock.flags & gaPLL_clockHalve16Bit))
        return best;
    if (bits == 24 && !(DV.pixelClock.flags & gaPLL_clockHalve24Bit))
        return best;
    if (bits == 32 && !(DV.pixelClock.flags & gaPLL_clockHalve32Bit))
        return best;
    for (i =  0; i < DV.pixelClock.numClocks; i++) {
        N_fix32 d = ABS(clock - valid[i]/2);
        if (d < diff) {
            diff = d;
            best = valid[i]/2;
            *select = i;
            *divide = 2;
            }
        }

    // Now search for a better quartered clock if this is supported
    if (bits == 8 && !(DV.pixelClock.flags & gaPLL_clockQuarter8Bit))
        return best;
    if ((bits == 15 || bits == 16) && !(DV.pixelClock.flags & gaPLL_clockQuarter16Bit))
        return best;
    if (bits == 24 && !(DV.pixelClock.flags & gaPLL_clockQuarter24Bit))
        return best;
    if (bits == 32 && !(DV.pixelClock.flags & gaPLL_clockQuarter32Bit))
        return best;
    for (i =  0; i < DV.pixelClock.numClocks; i++) {
        N_fix32 d = ABS(clock - valid[i]/4);
        if (d < diff) {
            diff = d;
            best = valid[i]/4;
            *select = i;
            *divide = 4;
            }
        }
    return best;
}

/****************************************************************************
PARAMETERS:
value   - Value to scale
scale   - Scale flag to determine how to scale

RETURNS:
Scaled value
****************************************************************************/
N_uint32 DRV_scaleValue(
    N_uint32 value,
    int scale)
{
    switch (scale) {
        case gaScale05: return value / 2;
        case gaScale15: return (value * 3) / 2;
        case gaScale20: return value * 2;
        case gaScale30: return value * 3;
        case gaScale40: return value * 4;
        }
    return value;
}

/****************************************************************************
PARAMETERS:
value   - Value to unscale
scale   - Scale flag to determine how to unscale

RETURNS:
Scaled value
****************************************************************************/
N_uint32 DRV_unscaleValue(
    N_uint32 value,
    int scale)
{
    switch (scale) {
        case gaScale05: return value * 2;
        case gaScale15: return (value * 2) / 3;
        case gaScale20: return value / 2;
        case gaScale30: return value / 3;
        case gaScale40: return value / 4;
        }
    return value;
}

/****************************************************************************
PARAMETERS:
value   - Value to scale

RETURNS:
Scaled clock value
****************************************************************************/
N_uint32 DRV_scaleClock(
    N_uint32 value,
    GA_modeEntry *entry)
{
    return DRV_scaleValue(value,(entry->flags & gaCLK_ScaleMask) >> gaCLK_ScaleShift);
}

/****************************************************************************
PARAMETERS:
value   - Value to scale

RETURNS:
Scaled clock value
****************************************************************************/
N_uint32 DRV_unscaleClock(
    N_uint32 value,
    GA_modeEntry *entry)
{
    return DRV_unscaleValue(value,(entry->flags & gaCLK_ScaleMask) >> gaCLK_ScaleShift);
}

/****************************************************************************
PARAMETERS:
value   - Value to scale

RETURNS:
Scaled CRTC value
****************************************************************************/
N_uint32 DRV_scaleCRTC(
    N_uint32 value)
{
    return DRV_scaleValue(value,(DV.cntMode.flags & gaCRTC_ScaleMask) >> gaCRTC_ScaleShift);
}

/****************************************************************************
PARAMETERS:
value   - Value to scale

RETURNS:
Scaled CRTC value
****************************************************************************/
N_uint32 DRV_unscaleCRTC(
    N_uint32 value)
{
    return DRV_unscaleValue(value,(DV.cntMode.flags & gaCRTC_ScaleMask) >> gaCRTC_ScaleShift);
}

/****************************************************************************
REMARKS:
Round a value to the specified integer boundary
****************************************************************************/
static ulong RoundUp(
    ulong value,
    ulong boundary)
{
    return ((value + (boundary-1)) / boundary) * boundary;
}

/****************************************************************************
PARAMETERS:
entry   - Entry to check

RETURNS:
Character cell height for the entry

REMARKS:
Determines the character cell height to use for the text mode.
****************************************************************************/
static uint DRV_getCharHeight(
    GA_modeEntry *entry)
{
    if (entry->yRes <= 40)
        return 16;
    if (entry->xRes == 80) {
        if (entry->yRes <= 43)
            return DV.opt.text80x43Height;
        else if (entry->yRes <= 50)
            return DV.opt.text80x50Height;
        else
            return DV.opt.text80x60Height;
        }
    else if (entry->xRes <= 100) {
        if (entry->yRes <= 43)
            return DV.opt.text100x43Height;
        else if (entry->yRes <= 50)
            return DV.opt.text100x50Height;
        else
            return DV.opt.text100x60Height;
        }
    else {
        if (entry->yRes <= 43)
            return DV.opt.text132x43Height;
        else if (entry->yRes <= 50)
            return DV.opt.text132x50Height;
        else
            return DV.opt.text132x60Height;
        }
}

/****************************************************************************
PARAMETERS:
xRes            - Physical pixel X resolution of the display mode
yRes            - Physical line Y resolution of the display mode
virtualX        - Virtual X resolution of the display mode
virtualY        - Virtual Y resolution of the display mode
bitsPerPixel    - Color depth of the display mode
gaInfo          - Structure to fill in with the mode information
outputDevice    - Output device to test mode information for
outputHead      - Output head to test mode information for

RETURNS:
Pointer to GA_modeEntry block on success, false on failure.

REMARKS:
Obtains information about a specific SNAP display mode. Note that this
is an internal function that always assumes the incoming GA_modeInfo
block is the correct size for the version of the driver (ie: the calling
function does any necessary versioning and translation on the user app
mode info block passed in). Note also that this version does not require
the dwSize member to be initialise and the gaInfo block passed in will
be correctly initialised as it should never be directly accessed by
external code.
****************************************************************************/
static GA_modeEntry *GetCustomVideoModeInfo(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 virtualX,
    N_int32 virtualY,
    N_int32 bitsPerPixel,
    GA_modeInfo *gaInfo,
    N_int32 outputDevice,
    N_int32 outputHead)
{
    GA_mode         e;
    GA_modeEntry    *entry;
    N_int32         memSize,bytesPerPixel = 1;

    // Make sure the output head parameter is in range
    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    if (outputHead < 0 || outputHead >= DV.init.GetNumberOfHeads())
        return NULL;

    // Clear the modeInfo block and initialise the size member
    memset(gaInfo,0,sizeof(GA_modeInfo));
    gaInfo->dwSize = sizeof(GA_modeInfo);

    // Find the valid mode entry for the mode
    if (bitsPerPixel > 0)
        e.xRes = RoundUp(xRes,8);
    else
        e.xRes = xRes;
    e.yRes = yRes;
    e.bits = bitsPerPixel;
    if ((entry = DRV_validateModeXY(&e,outputDevice,outputHead,&gaInfo->PhysicalXResolution,
            &gaInfo->PhysicalYResolution)) == NULL)
        return NULL;

    // Validate resolutions based on licensing
    if (DV.licensed.resolutions[0] != 0xFFFFFFFFU) {
        int i,xRes,yRes,found = false;
        for (i = 0; DV.licensed.resolutions[i] != 0xFFFFFFFF; i++) {
            xRes = DV.licensed.resolutions[i] >> 16;
            yRes = DV.licensed.resolutions[i] & 0xFFFF;
            if (xRes == e.xRes && yRes == e.yRes) {
                found = true;
                break;
                }
            }
        if (!found) {
            // Allow for basic 80x25/43/50 text modes through
            if (e.bits > 0 || e.xRes != 80 || e.yRes > 50)
                return NULL;
            }
        }

    // Validate color depths based on licensing
    if (DV.licensed.colorDepths[0] != 0xFF && e.bits > 0) {
        int i,found = false;
        for (i = 0; DV.licensed.colorDepths[i] != 0xFF; i++) {
            if (e.bits == DV.licensed.colorDepths[i]) {
                found = true;
                break;
                }
            }
        if (!found)
            return NULL;
        }

    // Now fill in the information for the mode
    gaInfo->Attributes          = DC.Attributes;
    gaInfo->AttributesExt       = DC.AttributesExt;
    if (!DV.opt.bDoubleBuffer)
        gaInfo->Attributes &= ~gaHaveDisplayStart;
    if (!DV.opt.bTripleBuffer)
        gaInfo->Attributes &= ~gaHaveTripleBuffer;
    if (entry->flags & gaBankedOnly)
        gaInfo->Attributes &= ~gaHaveLinearBuffer;
    if (entry->flags & gaLinearOnly)
        gaInfo->Attributes &= ~gaHaveBankedBuffer;
    if (!(entry->flags & gaAccMode) || !DV.opt.bHardwareAccel2D)
        gaInfo->Attributes &= ~(gaHaveAccel2D | gaHaveAccelVideo | gaHaveAccel3D);
    if (!(entry->flags & gaStereo) || !DV.opt.bStereo)
        gaInfo->Attributes &= ~gaHaveStereo;
    if (DV.opt.bHardwareStereoSync)
        gaInfo->Attributes |= gaHaveHWStereoSync;
    if (!(entry->flags & gaHWCursor) || !DV.opt.bHardwareCursor)
        gaInfo->Attributes &= ~gaHaveHWCursor;
    if (!(entry->flags & gaVideo) || !DV.opt.bHardwareVideo)
        gaInfo->Attributes &= ~gaHaveAccelVideo;
    if (!(entry->flags & ga3D) || !DV.opt.bHardwareAccel3D)
        gaInfo->Attributes &= ~gaHaveAccel3D;
    if (entry->xRes < 640 || (entry->flags & gaNoInterlaced))
        gaInfo->Attributes &= ~gaHaveInterlaced;
    if (entry->flags & gaIsDoubleScan)
        gaInfo->Attributes |= gaHaveDoubleScan;
    if (entry->flags & gaIsTripleScan)
        gaInfo->Attributes |= gaHaveTripleScan;
    if (entry->flags & gaNonVGA)
        gaInfo->Attributes |= gaHaveNonVGAMode;
    else
        gaInfo->Attributes &= ~gaHaveNonVGAMode;
    if (entry->flags & gaPanningMode)
        gaInfo->AttributesExt |= gaIsPanningMode;
    if (entry->flags & gaText) {
        gaInfo->Attributes |= gaIsTextMode;
        gaInfo->Attributes &= ~gaHaveInterlaced;
        gaInfo->XCharSize           = 8;
        gaInfo->YCharSize           = DRV_getCharHeight(entry);
        gaInfo->XResolution         = entry->xRes * gaInfo->XCharSize;
        gaInfo->YResolution         = entry->yRes * gaInfo->YCharSize;
        gaInfo->PhysicalXResolution *= gaInfo->XCharSize;
        gaInfo->PhysicalYResolution *= gaInfo->YCharSize;
        if (gaInfo->YResolution == 344)
            gaInfo->YResolution     = 350;
        if (gaInfo->PhysicalYResolution == 344)
            gaInfo->PhysicalYResolution = 350;
        }
    else {
        gaInfo->XResolution         = entry->xRes;
        gaInfo->YResolution         = entry->yRes;
        gaInfo->XCharSize           = 8;
        gaInfo->YCharSize           = 8;
        }

    // Handle virtual display modes (up to 32,767 bytes wide)
    if (virtualX != -1 && entry->bits >= 4) {
        int bytes;
        switch (entry->bits) {
            case 4:     bytes = virtualX / 8;   break;
            case 8:     bytes = virtualX;       break;
            case 15:
            case 16:    bytes = virtualX * 2;   break;
            case 24:    bytes = virtualX * 3;   break;
            case 32:    bytes = virtualX * 4;   break;
            default:    return NULL;
            }
        if (bytes >= 32768)
            return NULL;
        gaInfo->BytesPerScanLine = bytes;
        gaInfo->AttributesExt |= gaIsPanningMode;
        gaInfo->XResolution = virtualX;
        gaInfo->YResolution = virtualY;
        }
    else {
        virtualY = gaInfo->YResolution;
        gaInfo->BytesPerScanLine = entry->bytesPerLine;
        }
    gaInfo->BytesPerScanLine = DV.chip.roundBytesPerLine(gaInfo->BytesPerScanLine,entry->bits);
    if (gaInfo->BytesPerScanLine == 0xFFFFU)
        return NULL;
    switch (entry->bits) {
        case 0:
            gaInfo->MaxBytesPerScanLine     = DV.maxBytesPerLineTxt;
            gaInfo->MaxScanLineWidth        = DV.maxBytesPerLineTxt;
            gaInfo->MaxPixelClock           = DV.pixelClockMaxTxt;
            memSize                         = 128;
            break;
        case 4:
            gaInfo->MaxBytesPerScanLine     = DV.maxBytesPerLine4;
            gaInfo->MaxScanLineWidth        = DV.maxBytesPerLine4 * 8;
            gaInfo->MaxPixelClock           = DV.pixelClockMax4;
            if (entry->flags & gaVGA)
                memSize                     = 64;
            else if (entry->flags & gaSVGA4)
                memSize                     = DV.VGA4Memory / 4;
            else
                memSize                     = DV.VGAMemory / 4;
            break;
        case 8:
            if (entry->flags & gaNonVGA) {
                gaInfo->MaxBytesPerScanLine = DV.maxBytesPerLineNV8;
                gaInfo->MaxPixelClock       = DV.pixelClockMaxNV8;
                if (DV.NV8Memory != -1)
                    memSize                 = DV.NV8Memory;
                else
                    memSize                 = DV.NVMemory;
                }
            else {
                gaInfo->MaxBytesPerScanLine = DV.maxBytesPerLine8;
                gaInfo->MaxPixelClock       = DV.pixelClockMax8;
                if (entry->flags & gaVGA)
                    memSize                 = 64;
                else if (DV.VGA8Memory != -1)
                    memSize                 = DV.VGA8Memory;
                else
                    memSize                 = DV.VGAMemory;
                }
            gaInfo->MaxScanLineWidth        = gaInfo->MaxBytesPerScanLine;
            bytesPerPixel                   = 1;
            break;
        case 15:
        case 16:
            if (entry->flags & gaNonVGA) {
                gaInfo->MaxBytesPerScanLine = DV.maxBytesPerLineNV16;
                gaInfo->MaxPixelClock       = DV.pixelClockMaxNV16;
                if (DV.NV16Memory != -1)
                    memSize                 = DV.NV16Memory;
                else
                    memSize                 = DV.NVMemory;
                }
            else {
                gaInfo->MaxBytesPerScanLine = DV.maxBytesPerLine8;
                gaInfo->MaxPixelClock       = DV.pixelClockMax16;
                if (DV.VGA16Memory != -1)
                    memSize                 = DV.VGA16Memory;
                else
                    memSize                 = DV.VGAMemory;
                }
            gaInfo->MaxScanLineWidth        = gaInfo->MaxBytesPerScanLine/2;
            if (entry->bits == 15)
                gaInfo->PixelFormat         = DRV_pf15;
            else
                gaInfo->PixelFormat         = DRV_pf16;
            bytesPerPixel                   = 2;
            break;
        case 24:
            if (entry->flags & gaNonVGA) {
                gaInfo->MaxBytesPerScanLine = DV.maxBytesPerLineNV24;
                gaInfo->MaxPixelClock       = DV.pixelClockMaxNV24;
                if (DV.NV24Memory != -1)
                    memSize                 = DV.NV24Memory;
                else
                    memSize                 = DV.NVMemory;
                }
            else {
                gaInfo->MaxBytesPerScanLine = DV.maxBytesPerLine24;
                gaInfo->MaxPixelClock       = DV.pixelClockMax24;
                if (DV.VGA24Memory != -1)
                    memSize                 = DV.VGA24Memory;
                else
                    memSize                 = DV.VGAMemory;
                }
            gaInfo->MaxScanLineWidth        = gaInfo->MaxBytesPerScanLine/3;
            gaInfo->PixelFormat             = *DV.pf24;
            bytesPerPixel                   = 3;
            break;
        case 32:
            if (entry->flags & gaNonVGA) {
                gaInfo->MaxBytesPerScanLine = DV.maxBytesPerLineNV32;
                gaInfo->MaxPixelClock       = DV.pixelClockMaxNV32;
                if (DV.NV32Memory != -1)
                    memSize                 = DV.NV32Memory;
                else
                    memSize                 = DV.NVMemory;
                }
            else {
                gaInfo->MaxBytesPerScanLine = DV.maxBytesPerLine8;
                gaInfo->MaxPixelClock       = DV.pixelClockMax32;
                if (DV.VGA32Memory != -1)
                    memSize                 = DV.VGA32Memory;
                else
                    memSize                 = DV.VGAMemory;
                }
            gaInfo->MaxScanLineWidth        = gaInfo->MaxBytesPerScanLine/4;
            gaInfo->PixelFormat             = *DV.pf32;
            bytesPerPixel                   = 4;
            break;
        }

    // Check that scanline width is within range
    if (gaInfo->BytesPerScanLine > gaInfo->MaxBytesPerScanLine
            || gaInfo->BytesPerScanLine == 0)
        return NULL;

    // Fill in gaInfo fields for use in subsequent function calls
    gaInfo->BitsPerPixel                = entry->bits;
    gaInfo->LinearSize                  = DV.HWLinearSize;
    gaInfo->LinearBasePtr               = DV.HWLinearBasePtr;
    gaInfo->BitmapStartAlign            = 1 * bytesPerPixel;
    gaInfo->BitmapStridePad             = 1 * bytesPerPixel;
    gaInfo->MonoBitmapStartAlign        = 1;
    gaInfo->MonoBitmapStridePad         = 1;

    // Allow driver specific code to override max pixel clock on a per
    // display mode basis
    DV.pll.adjustMaxPixelClock(gaInfo,entry);

    // Convert maximum pixel clock to Hz and unscale the value. We also
    // obtain the list of available refresh rates from our database
    gaInfo->MaxPixelClock = N_FIXROUND(FXmul(gaInfo->MaxPixelClock,N_TOFIX(50)));
    gaInfo->MaxPixelClock = DRV_unscaleValue(gaInfo->MaxPixelClock * 20000,
        (entry->flags & gaCLK_ScaleMask) >> gaCLK_ScaleShift);
    if (!(entry->flags & gaVGA)) {
        gaInfo->MaxScanLines            = (memSize * 1024L) / gaInfo->BytesPerScanLine;
        if ((int)DV.maxAccelY != -1) {
            if (DV.maxAccelY < gaInfo->MaxScanLines)
                gaInfo->MaxScanLines    = DV.maxAccelY-DV.resScanLines;
            else
                gaInfo->MaxScanLines    = gaInfo->MaxScanLines-DV.resScanLines;
            }
        else
            gaInfo->MaxScanLines        = gaInfo->MaxScanLines-DV.resScanLines;
        if ((int)gaInfo->MaxScanLines < virtualY)
            return NULL;
        gaInfo->MaxBuffers              = gaInfo->MaxScanLines / virtualY;
        }
    else if (entry->bits >= 4) {
        gaInfo->MaxScanLines            = (memSize * 1024L) / gaInfo->BytesPerScanLine;
        if ((int)gaInfo->MaxScanLines < virtualY)
            return NULL;
        gaInfo->MaxBuffers              = gaInfo->MaxScanLines / virtualY;
        }
    else {
        gaInfo->MaxScanLines            = virtualY;
        gaInfo->MaxBuffers              = 1;
        }
    DV.chip.get2DAccelInfo(gaInfo,entry,virtualX,virtualY);
    DV.chip.getVideoAccelInfo(gaInfo,entry);
    DV.chip.get3DAccelInfo(gaInfo,entry);
    if (!DV.opt.bDoubleBuffer || !(DC.Attributes & gaHaveDisplayStart))
        gaInfo->MaxBuffers              = 1;

    // We no longer support linear only heaps, so mark the end at MaxScanLines
    gaInfo->LinearHeapStart         = gaInfo->MaxScanLines * gaInfo->BytesPerScanLine;
    gaInfo->MaxLinearOffset         = gaInfo->MaxScanLines * gaInfo->BytesPerScanLine;

    // Account for olders chipset drivers that don't align to a pixel
    // boundary. We should never align bitmaps to non-pixel boundaries
    // for any card.
    if (gaInfo->BitmapStartAlign == 1)
        gaInfo->BitmapStartAlign        = 1 * bytesPerPixel;
    if (gaInfo->BitmapStridePad == 1)
        gaInfo->BitmapStridePad         = 1 * bytesPerPixel;

    // If any of the above chip specific functions fail a mode, they will
    // set the MaxBuffers value to zero. We check for that here and fail
    // the mode when this happens (Cirrus Laguna is one).
    if (gaInfo->MaxBuffers == 0)
        return NULL;
    return entry;
}

/****************************************************************************
PARAMETERS:
mode            - SNAP display mode to get info for
virtualX        - Virtual X dimension for video mode
virtualY        - Virtual Y dimension for video mode
gaInfo          - Structure to fill in with the mode information
outputDevice    - Output device to test mode information for
outputHead      - Output head to test mode information for

RETURNS:
0 on success, -1 on failure

REMARKS:
Obtains information about a specific SNAP display mode.
****************************************************************************/
static GA_modeEntry *GetVideoModeInfo(
    N_uint32 mode,
    N_int32 virtualX,
    N_int32 virtualY,
    GA_modeInfo *gaInfo,
    N_int32 outputDevice,
    N_int32 outputHead)
{
    GA_modeEntry    *entry;
    CDBX_record     crtcRec;
    N_uint32        attributes;
    ibool           isVirtualDesktop = false;
    int             aspect;
    GA_mode         e;

    // Make sure the output head parameter is in range
    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    if (outputHead < 0 || outputHead >= DV.init.GetNumberOfHeads())
        return NULL;
    if (DV.opt.bMultiHead && ((mode & gaModeMask) >= gaVirtualMode8bpp)
            && (DV.init.GetNumberOfHeads() > 1)) {
        // If this is a virtual desktop mode, search for the parent mode for this
        // mode to be used and use that instead.
        isVirtualDesktop = true;
        e.xRes = DV.opt.multiHeadRes[0].right;
        e.yRes = DV.opt.multiHeadRes[0].bottom;
        switch (mode & gaModeMask) {
            case gaVirtualMode8bpp:
                if (DC.WorkArounds & gaNo8bppMultiHead)
                    return NULL;
                e.bits = 8;
                break;
            case gaVirtualMode15bpp:
                e.bits = 15;
                break;
            case gaVirtualMode16bpp:
                e.bits = 16;
                break;
            case gaVirtualMode24bpp:
                if (DC.WorkArounds & gaNo24bppMultiHead)
                    return NULL;
                e.bits = 24;
                break;
            case gaVirtualMode32bpp:
                e.bits = 32;
                break;
            }
        if ((entry = DRV_validateModeXY(&e,outputDevice,outputHead,&gaInfo->PhysicalXResolution,
                &gaInfo->PhysicalYResolution)) == NULL)
            return NULL;
        virtualX = DV.opt.multiHeadSize.right;
        virtualY = DV.opt.multiHeadSize.bottom;
        }
    else {
        // Find the valid mode entry for the mode
        if ((entry = DRV_validateMode(mode,outputDevice,outputHead,&gaInfo->PhysicalXResolution,
                &gaInfo->PhysicalYResolution)) == NULL)
            return NULL;
        }
    if (GetCustomVideoModeInfo(entry->xRes,entry->yRes,virtualX,virtualY,entry->bits,gaInfo,outputDevice,outputHead) == NULL)
        return NULL;

    // Filter out 16:9 and 16:10 aspect ratio modes if not enabled
    if (monitor[outputHead].maxResolution != 0xFF && gaInfo->BitsPerPixel != 0) {
        // Find the aspect ratio of the display mode. For multi-head modes we actually
        // need to use the physical resolution which is always the base mode used on
        // each head.
        if ((mode & gaModeMask) >= gaVirtualMode8bpp)
            aspect = ((gaInfo->PhysicalXResolution * 10) / gaInfo->PhysicalYResolution);
        else
            aspect = ((gaInfo->XResolution * 10) / gaInfo->YResolution);

        // Fix for 320x200 and 640x400 which are actually 16:10 modes!
        if (aspect >= 16 && (gaInfo->XResolution == 640 || gaInfo->XResolution == 320))
            aspect = 13;

        // Now check the aspect ratio against those supported
        if (aspect >= 16) {
            // 16:9 or 16:10 display mode, so check if they are enabled for monitor
            if (!(monitor[outputHead].flags & Monitor_16to9))
                return NULL;
            }
        else {
            // 4:3 or 5:4 display modes. Note that we never allow 640x480
            // to be excluded since this mode is *so* common.
            if (monitor[outputHead].flags & Monitor_Exclude4to3) {
                if (gaInfo->XResolution != 640 || gaInfo->YResolution != 480)
                    return NULL;
                }
            }
        }

    // Allow interlaced modes to be disabled on a mode by mode basis
    attributes = DC.Attributes;
    if (entry->flags & (gaNoInterlaced | gaText))
        attributes &= ~gaHaveInterlaced;

    // Find the list of refresh rates and default CRTC timings
    if (!(entry->flags & gaUseBIOS) && (outputDevice & gaOUTPUT_CRT)) {
        if (DV.pixelClock.flags & gaPLL_tableDriven) {
            if (!_DRV_imports.CDBX_getRefreshList(DC.DeviceIndex,outputHead,attributes,gaInfo->PhysicalXResolution,gaInfo->PhysicalYResolution,gaInfo->BitsPerPixel,DV.opt.tripleScanLowRes,DV.opt.doubleScan512,&monitor,gaInfo->MaxPixelClock,DV.licensed.maxRefresh,MAX_REFRESH_RATES,DRV_refreshList,DV.init.GetClosestPixelClock))
                return NULL;
            }
        else {
            if (!_DRV_imports.CDBX_getRefreshList(DC.DeviceIndex,outputHead,attributes,gaInfo->PhysicalXResolution,gaInfo->PhysicalYResolution,gaInfo->BitsPerPixel,DV.opt.tripleScanLowRes,DV.opt.doubleScan512,&monitor,gaInfo->MaxPixelClock,DV.licensed.maxRefresh,MAX_REFRESH_RATES,DRV_refreshList,NULL))
                return NULL;
            }
        if (!_DRV_imports.CDBX_getCRTCTimings(DC.DeviceIndex,outputHead,attributes,gaInfo->PhysicalXResolution,gaInfo->PhysicalYResolution,DV.opt.tripleScanLowRes,DV.opt.doubleScan512,0,&monitor,gaInfo->MaxPixelClock,DV.licensed.maxRefresh,&crtcRec))
            return NULL;
        }
    else {
        // If we are using the BIOS or we are not driving a CRT, we don't
        // control the refresh rate so we use whatever the default refresh
        // rate for the output device is (ie: LCD panel, DFP monitor, TV etc).
        // We assume 60 for the default refresh rate, but that is just so
        // we have a valid number to put in the field ;-).
        int yRes = gaInfo->YResolution * ((entry->flags & gaIsDoubleScan) ? 2 : 1);
        DRV_refreshList[0] = -1;
        crtcRec.crtc.Flags = 0;
        crtcRec.crtc.RefreshRate = yRes < 480 ? 7000 : 6000;
        }
    gaInfo->RefreshRateList     = DRV_refreshList;
    gaInfo->DefaultRefreshRate  = crtcRec.crtc.RefreshRate / 100;
    if (crtcRec.crtc.Flags & gaInterlaced)
        gaInfo->DefaultRefreshRate = -gaInfo->DefaultRefreshRate;

    // Get the refresh list from the BIOS if we have BIOS refresh control.
    if ((entry->flags & gaUseBIOS) && DV.chip.getBIOSRefreshList)
        DV.chip.getBIOSRefreshList(entry,DRV_refreshList);

    // Now override the settings if this is a virtual desktop mode on multi
    // head controllers
    if (isVirtualDesktop) {
        gaInfo->Attributes              |= gaIsVirtualMode;
        gaInfo->MaxScanLines            /= DV.init.GetNumberOfHeads();
        gaInfo->MaxBuffers              /= DV.init.GetNumberOfHeads();
        }
    return entry;
}

/****************************************************************************
PARAMETERS:
xRes            - Physical pixel X resolution of the display mode
yRes            - Physical line Y resolution of the display mode
virtualX        - Virtual X resolution of the display mode
virtualY        - Virtual Y resolution of the display mode
bitsPerPixel    - Color depth of the display mode
modeInfo        - Structure to fill in with the mode information
outputDevice    - Output device to test mode information for
outputHead      - Output head to test mode information for

RETURNS:
0 on success, -1 on failure

REMARKS:
Obtains information about a specific SNAP display mode.
****************************************************************************/
N_int32 NAPI DRV_GetCustomVideoModeInfoExt(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 virtualX,
    N_int32 virtualY,
    N_int32 bitsPerPixel,
    GA_modeInfo *modeInfo,
    N_int32 outputDevice,
    N_int32 outputHead)
{
    static GA_modeInfo  gaInfo;

    // Make sure the output head parameter is in range
    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    if (outputHead < 0 || outputHead >= DV.init.GetNumberOfHeads())
        return -1;

    // Obtain the mode info for internal mode info block
    if (GetCustomVideoModeInfo(xRes,yRes,virtualX,virtualY,bitsPerPixel,&gaInfo,outputDevice,outputHead) == NULL)
        return -1;

    // Return the structure to caller, taking into account structure size
    COPY_STRUCTURE(modeInfo,&gaInfo);
    return 0;
}

/****************************************************************************
PARAMETERS:
mode            - SNAP display mode to get info for
modeInfo        - Structure to fill in with the mode information
outputDevice    - Output device to test mode information for
outputHead      - Output head to test mode information for

RETURNS:
0 on success, -1 on failure

REMARKS:
Obtains information about a specific SNAP display mode.
****************************************************************************/
N_int32 NAPI DRV_GetVideoModeInfoExt(
    N_uint32 mode,
    GA_modeInfo *modeInfo,
    N_int32 outputDevice,
    N_int32 outputHead)
{
    static GA_modeInfo  gaInfo;

    // Obtain the mode info for internal mode info block
    if (GetVideoModeInfo(mode,-1,-1,&gaInfo,outputDevice,outputHead) == NULL)
        return -1;

    // Return the structure to caller, taking into account structure size
    COPY_STRUCTURE(modeInfo,&gaInfo);
    return 0;
}

/****************************************************************************
PARAMETERS:
xRes            - Physical pixel X resolution of the display mode
yRes            - Physical line Y resolution of the display mode
virtualX        - Virtual X resolution of the display mode
virtualY        - Virtual Y resolution of the display mode
bitsPerPixel    - Color depth of the display mode
modeInfo        - Structure to fill in with the mode information

RETURNS:
0 on success, -1 on failure

REMARKS:
Obtains information about a specific SNAP display mode.
****************************************************************************/
N_int32 NAPI DRV_GetCustomVideoModeInfo(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 virtualX,
    N_int32 virtualY,
    N_int32 bitsPerPixel,
    GA_modeInfo *modeInfo)
{
    static GA_modeInfo  gaInfo;

    // Obtain the mode info for internal mode info block
    if (GetCustomVideoModeInfo(xRes,yRes,virtualX,virtualY,bitsPerPixel,&gaInfo,DV.outputDevice[0],DV.init.GetActiveHead()) == NULL)
        return -1;

    // Return the structure to caller, taking into account structure size
    COPY_STRUCTURE(modeInfo,&gaInfo);
    return 0;
}

/****************************************************************************
PARAMETERS:
mode        - SNAP display mode to get info for
modeInfo    - Structure to fill in with the mode information

RETURNS:
0 on success, -1 on failure

REMARKS:
Obtains information about a specific SNAP display mode.
****************************************************************************/
N_int32 NAPI DRV_GetVideoModeInfo(
    N_uint32 mode,
    GA_modeInfo *modeInfo)
{
    static GA_modeInfo  gaInfo;

    // Obtain the mode info for internal mode info block
    if (GetVideoModeInfo(mode,-1,-1,&gaInfo,DV.outputDevice[0],DV.init.GetActiveHead()) == NULL)
        return -1;

    // Return the structure to caller, taking into account structure size
    COPY_STRUCTURE(modeInfo,&gaInfo);
    return 0;
}

/****************************************************************************
REMARKS:
Returns the curent video mode information from the driver.
****************************************************************************/
void NAPI DRV_GetCurrentVideoModeInfo(
    GA_modeInfo *modeInfo)
{
    int dwSize = modeInfo->dwSize;

    if ((DV.cntModeNum & gaModeMask) == 0) {
        // Mode is a custom video mode
        COPY_STRUCTURE(modeInfo,&DV.cntModeInfo);
        }
    else {
        // Start with the real mode information and fill in the
        // stuff that changes dynamically.
        DRV_GetVideoModeInfoExt(DV.cntModeNum & gaModeMask,modeInfo,DV.outputDevice[0],DV.init.GetActiveHead());
        modeInfo->Attributes = DV.cntModeInfo.Attributes;
        modeInfo->BytesPerScanLine = DV.cntModeInfo.BytesPerScanLine;
        modeInfo->LinearHeapStart = DV.cntModeInfo.LinearHeapStart;
        modeInfo->MaxBuffers = DV.cntModeInfo.MaxBuffers;

        // Fill in the physical resolution values for new apps.
        if (dwSize >= 172) {
            if (DV.cntVirtualX != -1) {
                // The app has requested a virtual mode so we change the
                // reported mode info structure to convert it into a
                // panning mode with the appropriate physical and virtual
                // resolutions.
                modeInfo->AttributesExt |= gaIsPanningMode;
                modeInfo->XResolution = DV.cntVirtualX;
                modeInfo->YResolution = DV.cntVirtualY;
                }
            modeInfo->PhysicalXResolution = DV.cntModeInfo.PhysicalXResolution;
            modeInfo->PhysicalYResolution = DV.cntModeInfo.PhysicalYResolution;
            }
        }
}

/****************************************************************************
PARAMETERS:
pixelClock  - Pixel clock in Hz to convert to fixed point

RETURNS:
Pixel clock value in fixed point.

REMARKS:
This function converts the pixel clock from Hz to a fixed point value in
Mhz. This function also adjusts the pixel clock for 9 character text modes
as necessary at the same time.
****************************************************************************/
static N_fix32 GetFixedPixelClock(
    GA_modeEntry *entry,
    N_int32 pixelClock)
{
    N_fix32 fxClock;

    fxClock = DRV_scaleClock(pixelClock,entry) / 20000;
    fxClock = FXdiv(N_TOFIX(fxClock),N_TOFIX(50));
    if (entry->flags & gaText) {
        // Adjust pixel clock for 9 character clock modes text modes
        fxClock = FXmul(fxClock,N_FIXCONST(1.125));
        }
    return fxClock;
}

/****************************************************************************
REMARKS:
Function to load the correct text mode font tables.
****************************************************************************/
void DRV_LoadTextFontExt(
    int height)
{
    switch (height) {
        case 8:
            VGA_loadFont(DC.TextFont8x8,height);
            break;
        case 14:
            VGA_loadFont(DC.TextFont8x14,height);
            break;
        case 16:
            VGA_loadFont(DC.TextFont8x16,height);
            break;
        }
    if (height >= 16)
        height--;
    wrinx(0x3D4,0x0A,height-2);
    wrinx(0x3D4,0x0B,height-1);
}

/****************************************************************************
REMARKS:
Function to load the correct text mode font tables.
****************************************************************************/
void DRV_LoadTextFont(void)
{
    DRV_LoadTextFontExt(DRV_getCharHeight(&DV.cntMode));
}

/****************************************************************************
DESCRIPTION:
Sets a custom display mode.

PARAMETERS:
xRes            - X resolution for custom display mode
yRes            - Y resolution for custom display mode
bitsPerPixel    - Color depth for custom display mode (0 = text mode)
virtualX        - Requested virtual display X resolution (-1 to use default)
virtualY        - Returns the addressable display memory Y resolution
bytesPerLine    - Returns the scanline width for the mode
maxMem          - Returns the maximum addressable display memory limit
crtc            - CRTC information block containing CRTC timings (required)

RETURNS:
0 on success, -1 on failure

REMARKS:
This function is used to initialize a custom display mode using the
specified X and Y resolution, color depth and CRTC timings.
****************************************************************************/
static N_int32 SetCustomVideoMode(
    GA_modeEntry *entry,
    GA_modeInfo *modeInfo,
    N_uint32 flags,
    N_int32 *virtualX,
    N_int32 *virtualY,
    N_int32 *bytesPerLine,
    N_int32 *maxMem,
    GA_CRTCInfo *crtc)
{
    int                     i;
    ulong                   *vec;

    // Fail the mode set if virtual scrolling is disabled
    if (!DV.opt.bVirtualScroll && (*virtualX != -1 || *virtualY != -1))
        return -1;

    // Find the valid mode entry for the mode
    if (*bytesPerLine == -1)
        *bytesPerLine = modeInfo->BytesPerScanLine;

    // Disable acceleration if this has been requested
    if (flags & gaNoAccel)
        entry->flags &= ~(gaAccMode | gaHWCursor);

    if (!(flags & gaPartialModeSet)) {
        if ((*bytesPerLine = DV.chip.roundBytesPerLine(*bytesPerLine,entry->bits)) == -1)
            return -1;
#ifdef TEST_HARNESS
        if (*bytesPerLine != (int)modeInfo->BytesPerScanLine)
            PM_fatalError("Mismatch in round bytes per line!");
#endif
        }

    // Wait till engine idle if *existing* mode is accelerated (not the one we are
    // setting, but the one that is currently active). We also reset the draw buffer
    // back to the primary display surface before switching display modes. We also
    // make sure we don't do this the first time we set a display mode.
    if (!(flags & gaPartialModeSet)) {
        DV.cntVirtualX = *virtualX;
        DV.cntVirtualY = *virtualY;
        if (*virtualX == -1) {
            *virtualX = modeInfo->XResolution;
            *virtualY = modeInfo->YResolution;
            }
        if (DV.cntMode.flags & gaAccMode) {
#ifndef DEBUG_NEW_DRIVER
            GA_buffer drawBuf;
            drawBuf.dwSize = 0;
            drawBuf.Offset = 0;
            drawBuf.Stride = DV.cntModeInfo.BytesPerScanLine;
            drawBuf.Width = DV.cntModeInfo.XResolution;
            drawBuf.Height = DV.cntModeInfo.YResolution;
            DV.state2d.SetDrawBuffer(&drawBuf);
            DV.state2d.WaitTillIdle();
#endif
            // Hide the cursor
            if ((DV.cntMode.flags & gaHWCursor) && IS_CVALID_PTR(DV.cursor.ShowCursor))
                DV.cursor.ShowCursor(false);
            }

        // Clear the framebuffer for the current display mode
        if (!(flags & gaDontClear))
            DV.chip.clearMemory(false);
        }

    // Save the CRTC timings for the mode
    DV.cntCRTC = *crtc;

    // Store current mode information; needs to be done before clearMemory is called
    DV.cntMode = *entry;
    DV.cntMode.bytesPerLine = *bytesPerLine;
    DV.cntModeInfo = *modeInfo;
    DV.cntModeInfo.BytesPerScanLine = *bytesPerLine;
    DV.cntModeNum = flags & ~gaRefreshCtrl;
    DV.cntMaxMem = modeInfo->MaxScanLines * modeInfo->BytesPerScanLine;
    DV.cntFlags = flags;

    // Handle special case workarounds for current mode info structure
    if ((DV.cntCRTC.Flags & gaInterlaced) && (DC.WorkArounds & gaNoInterlacedCursor))
        DV.cntModeInfo.Attributes &= ~gaHaveHWCursor;
    if ((DC.WorkArounds & gaNoLCDExpandCursor) && DV.opt.bLCDExpand && ((DV.cntMode.xRes < DV.opt.LCDPanelWidth) || (DV.cntMode.yRes < DV.opt.LCDPanelHeight)))
        DV.cntModeInfo.Attributes &= ~gaHaveHWCursor;

    // Set the double scan flag based on the passed in CRTC registers
    if (DV.cntCRTC.Flags & gaDoubleScan)
        DV.cntMode.flags |= gaIsDoubleScan;
    else
        DV.cntMode.flags &= ~gaIsDoubleScan;

    // Reset stored blanks when debugging blank timings
#ifdef DEBUG_BLANKS
    hBlankStart = hBlankEnd = vBlankStart = vBlankEnd = -1;
#endif

    // Now initialise the mode
    if (!(flags & gaPartialModeSet)) {
#ifdef  DEBUG_BIOS
        if (DV.cntMode.flags & gaText) {
            RMREGS regs;
            regs.x.ax = 0x3;
            PM_int86(0x10,&regs,&regs);
            }
        else {
            RMREGS regs;
            regs.x.ax = 0x4F02;
            regs.x.bx = 0x8101;
            PM_int86(0x10,&regs,&regs);
            }
#endif
        // Find fixed point pixel clock for selected refresh rate
        DV.cntPixelClock = GetFixedPixelClock(&DV.cntMode,DV.cntCRTC.PixelClock);
        DV.chip.setOptions();
        DV.chip.preModeSet();
        DV.chip.disableLinear();
        if (DV.cntMode.flags & gaUseBIOS) {
            DV.chip.setBIOSMode();
            VGA_preModeSet();
            }
        if (!(DV.cntMode.flags & gaUseBIOS)) {
            DV.chip.setChipsetRegistersExt(true);
            DV.dac.setDACRegisters();
            DV.chip.setMemoryClock();
            if (DC.Attributes & gaHaveEngineClock)
                DV.chip.setEngineClock();
            DV.chip.setCRTCRegisters(&DV.cntCRTC);
            DV.chip.setPixelClock(DV.cntPixelClock);
            }
        DV.cntMode.bytesPerLine = DV.chip.setBytesPerLine(DV.cntMode.bytesPerLine);
        DV.cntDACWidth = 6;
        DV.dac.set8BitDAC(false);
        if (!(DV.cntMode.flags & gaVGA) && !(flags & ga6BitDAC)) {
            if (DV.dac.set8BitDAC(true))
                DV.cntDACWidth = 8;
            }
        if (flags & gaLinearBuffer)
            DV.chip.enableLinear();
        else {
            if (IS_CVALID_PTR(DV.driver.SetBank))
                DV.driver.SetBank(0);
            if (DV.cntMode.flags & gaText)
                DRV_LoadTextFont();
            }
        DV.driver.SetDisplayStart(0,false);
        if (DV.chip.initCursor)
            DV.chip.initCursor();
        if (DV.cntMode.flags & gaAccMode)
            DV.chip.setupAccelerator();
        if ((DV.cntMode.flags & gaHWCursor) && IS_CVALID_PTR(DV.cursor.ShowCursor)) {
            GA_palette white = {0xFF,0xFF,0xFF};
            GA_palette black = {0x00,0x00,0x00};
            DV.cursor.ShowCursor(false);
            DV.cursor.SetMonoCursorColor(&white,&black);
            }
        DV.chip.postModeSet();
        if (!(flags & gaDontClear))
            DV.chip.clearMemory(true);
        if (DV.cntModeInfo.BitsPerPixel <= 4)
            DV.driver.SetPaletteData(DC.VGAPal4,64,0,false);
        else if (DV.cntModeInfo.BitsPerPixel == 8)
            DV.driver.SetPaletteData(DC.VGAPal8,256,0,false);
        else {
            // Program default gamma ramp for RGB modes
            if (IS_CVALID_PTR(DV.driver.SetGammaCorrectData))
                DV.driver.SetGammaCorrectDataExt(DV.opt.gammaRamp,256,0,false);
            }
        // Reset the software rendering vectors to point to our stub function
        for (i = 1,vec = ((ulong*)&DV.drawSW)+1; i < sizeof(DV.drawSW) / sizeof(void*); i++,vec++) {
            *vec = (ulong)DRV_stub;
            }
        }
    else {
        // Save the hardware palette/gamma ramp so we can restore it
        if (DV.cntMode.bits <= 8)
            DV.driver.GetPaletteDataExt(scratchPalExt,256,0);
        else if (IS_CVALID_PTR(DV.driver.GetGammaCorrectDataExt))
            DV.driver.GetGammaCorrectDataExt(scratchPalExt,256,0);

        // Perform partial mode set
        VGA_preModeSet();
        DV.cntPixelClock = GetFixedPixelClock(&DV.cntMode,DV.cntCRTC.PixelClock);
        // Query output device only on "old style" drivers where BIOS performs the switching.
        // On newer drivers we handle the switching ourselves and BIOS might give us
        // incorrect information.
        if (!DV.chip.getPendingOptions)
            DV.chip.getDefaultOptions();
        DV.chip.configChipsetFlags();
        if ((flags & gaRefreshCtrl) && !(DV.cntMode.flags & gaUseBIOS)) {
            DV.chip.setMemoryClock();
            if (DC.Attributes & gaHaveEngineClock)
                DV.chip.setEngineClock();
            DV.chip.setCRTCRegisters(&DV.cntCRTC);
            DV.chip.setPixelClock(DV.cntPixelClock);
            }
        DV.cntMode.bytesPerLine = DV.chip.setBytesPerLine(DV.cntMode.bytesPerLine);
        if (DV.cntMode.flags & gaText)
            DRV_LoadTextFont();
        VGA_postModeSet();

        // Restore the hardware palette/gamma ramp
        if (DV.cntMode.bits <= 8)
            DV.driver.SetPaletteDataExt(scratchPalExt,256,0,false);
        else if (IS_CVALID_PTR(DV.driver.GetGammaCorrectDataExt))
            DV.driver.SetGammaCorrectDataExt(scratchPalExt,256,0,false);
        }
    *bytesPerLine = DV.cntMode.bytesPerLine;
    *maxMem = DV.cntMaxMem;
    return 0;
}

/****************************************************************************
DESCRIPTION:
Sets a custom display mode.

PARAMETERS:
xRes            - X resolution for custom display mode
yRes            - Y resolution for custom display mode
bitsPerPixel    - Color depth for custom display mode (0 = text mode)
virtualX        - Requested virtual display X resolution (-1 to use default)
virtualY        - Returns the addressable display memory Y resolution
bytesPerLine    - Returns the scanline width for the mode
maxMem          - Returns the maximum addressable display memory limit
crtc            - CRTC information block containing CRTC timings (required)

RETURNS:
0 on success, -1 on failure

REMARKS:
This function is used to initialize a custom display mode using the
specified X and Y resolution, color depth and CRTC timings.
****************************************************************************/
N_int32 NAPI DRV_SetCustomVideoMode(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bitsPerPixel,
    N_uint32 flags,
    N_int32 *virtualX,
    N_int32 *virtualY,
    N_int32 *bytesPerLine,
    N_int32 *maxMem,
    GA_CRTCInfo *crtc)
{
    GA_modeEntry    *entry;
    GA_modeInfo     modeInfo;

    if ((entry = GetCustomVideoModeInfo(xRes,yRes,*virtualX,*virtualY,bitsPerPixel,&modeInfo,DV.outputDevice[0],DV.init.GetActiveHead())) == NULL)
        return -1;
    return SetCustomVideoMode(entry,&modeInfo,flags,virtualX,virtualY,bytesPerLine,maxMem,crtc);
}

/****************************************************************************
DESCRIPTION:
Sets a specified display mode.

PARAMETERS:
mode            - Mode number to set, including flags
virtualX        - Requested virtual display X resolution (-1 to use default)
virtualY        - Returns the addressable display memory Y resolution
bytesPerLine    - Returns the scanline width for the mode
maxMem          - Returns the maximum addressable display memory limit
crtc            - CRTC information block if afRefreshCtrl specified

RETURNS:
0 on success, -1 on failure

REMARKS:
This function is used to initialize a specified display mode.
****************************************************************************/
N_int32 NAPI DRV_SetVideoMode(
    N_uint32 mode,
    N_int32 *virtualX,
    N_int32 *virtualY,
    N_int32 *bytesPerLine,
    N_int32 *maxMem,
    N_int32 refreshRate,
    GA_CRTCInfo *crtc)
{
    GA_modeEntry    *entry;
    CDBX_record     crtcRec;
    GA_modeInfo     modeInfo;
    GA_CRTCInfo     newCRTC = {0};
    N_uint32        attributes;
    int             i;

    // Quick check when debugging drivers to catch old source code that does not
    // properly set the output device to CRTC by default!
    CHECK(DV.outputDevice[0] != 0);

    // Find the valid mode entry for the mode
    if (*virtualX == -1)
        *virtualX = *virtualY = -1;
    if ((entry = GetVideoModeInfo(mode,*virtualX,*virtualY,&modeInfo,DV.outputDevice[0],DV.init.GetActiveHead())) == NULL)
        return -1;

    // Find the CRTC timings for the mode
    if (mode & gaRefreshCtrl) {
        newCRTC = *crtc;
        }
    else {
        // Allow interlaced modes to be disabled on a mode by mode basis
        attributes = DC.Attributes;
        if (entry->flags & gaNoInterlaced)
            attributes &= ~gaHaveInterlaced;
        if (!(entry->flags & gaUseBIOS) && (DV.outputDevice[0] & gaOUTPUT_CRT)) {
            // Turn on refresh rate control if we are driving an external CRT
            // monitor and we are not using the BIOS.
            if (!_DRV_imports.CDBX_getCRTCTimings(DC.DeviceIndex,DV.init.GetActiveHead(),attributes,modeInfo.PhysicalXResolution,modeInfo.PhysicalYResolution,DV.opt.tripleScanLowRes,DV.opt.doubleScan512,refreshRate,&monitor,modeInfo.MaxPixelClock,DV.licensed.maxRefresh,&crtcRec))
                return -1;
            newCRTC = crtcRec.crtc;
            mode |= gaRefreshCtrl;
            }
        }

    // Now set the mode via the custom video mode interface
    if (modeInfo.Attributes & gaIsVirtualMode) {
        // Set the display mode and start address for each head in virtual mode
        for (i = DV.init.GetNumberOfHeads()-1; i >= 0; i--) {
            DV.init.SetActiveHead(i);
            *virtualX = modeInfo.XResolution;
            *virtualY = modeInfo.YResolution;
            if (SetCustomVideoMode(entry,&modeInfo,mode,virtualX,virtualY,bytesPerLine,maxMem,&newCRTC) == -1)
                return -1;
            DV.driver.SetDisplayStartXY(
                DV.opt.multiHeadBounds[i].left,
                DV.opt.multiHeadBounds[i].top,gaDontWait);
            }
        DV.init.SetActiveHead(gaActiveHeadClone);
        return 0;
        }
    else
        return SetCustomVideoMode(entry,&modeInfo,mode,virtualX,virtualY,bytesPerLine,maxMem,&newCRTC);
}

/****************************************************************************
RETURNS:
Current video mode number active in the driver.

REMARKS:
This function returns the currently active video mode.
****************************************************************************/
N_uint32 NAPI DRV_GetVideoMode(void)
{
    // TODO: Need an output head parameter here!!
    return DV.cntModeNum;
}

/****************************************************************************
PARAMETERS:
outputHead      - Display head to get pixel clock for
xRes            - X resolution for the mode
yRes            - Y resolution for the mode
bitsPerPixel    - Color depth for the mode
pixelClock      - Requested pixel clock in units of Hz.

RETURNS:
Closest pixel clock in units of Hz.

REMARKS:
This function allows an application to determine if a particular pixel clock
is available. When this function is called it will run the requested pixel
clock through the internal PLL programming routines and return the actual
pixel clock that will be programmed into the hardware. The process of running
the PLL clock computation routines may cause the returned pixel clock to be
rounded slightly up or down from the requested value, however the driver
should implement the algorithms to attempt to find clocks that are the same
as or higher than the requested value. Note that the calling application must
also pass in the display mode number for the mode that will be using this
pixel clock to this function. The mode number is necessary so that the
driver can determine any necessary hardware limitations internally for the
display mode before looking for the closest physical pixel clock.
****************************************************************************/
N_uint32 NAPI DRV_GetClosestPixelClock(
    N_int32 outputHead,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bitsPerPixel,
    N_uint32 pixelClock)
{
    GA_mode         e;
    int             N,M,R,select,divide;
    N_fix32         clock;
    GA_modeEntry    *entry;
    N_uint16        physXRes,physYRes;

    // Make sure the output head parameter is in range
    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    if (outputHead < 0 || outputHead >= DV.init.GetNumberOfHeads())
        return 0xFFFFFFFF;

    // Find the valid mode entry for the mode
    if (bitsPerPixel == 0) {
        // Use 8x8 character cell to compute pixel clocks
        e.xRes = xRes / 8;
        e.yRes = yRes / 8;
        }
    else {
        e.xRes = xRes;
        e.yRes = yRes;
        }
    e.bits = bitsPerPixel;
    if ((entry = DRV_validateModeXY(&e,DV.outputDevice[0],outputHead,&physXRes,&physYRes)) == NULL)
        return 0xFFFFFFFF;

    // Scale and convert to kHz * 100 and convert to fixed point
    pixelClock = DRV_scaleClock(pixelClock,entry) / 20000;
    clock = FXdiv(N_TOFIX(pixelClock),N_TOFIX(50));

    // Now run it through the pixel clock calculation routines
    if (DV.pixelClock.flags & gaPLL_tableDriven)
        clock = DRV_findClosestDiscreteClock(clock,entry->bits,&select,&divide);
    else if (DV.pll.calculateClockExt)
        clock = DV.pll.calculateClockExt(outputHead,&DV.pixelClock,clock,&M,&N,&R);
    else
        clock = DV.pll.calculateClock(&DV.pixelClock,clock,&M,&N,&R);

    // If the PLL M,N,P value limits do not allow you to reach PLLMaxClock,
    // you get an invalid clock value (gaTest tests this with addMaxMode()).
    if (clock <= 0 || clock >= N_FIXCONST(1000.0))
        return 0xFFFFFFFF;

    // Convert back to Hz and unscale the value
    pixelClock = N_FIXROUND(FXmul(clock,N_TOFIX(50)));
    pixelClock = DRV_unscaleClock(pixelClock * 20000,entry);
    return pixelClock;
}

/****************************************************************************
REMARKS:
Obsolete compatibility function.
****************************************************************************/
N_uint32 NAPI DRV_GetClosestPixelClockV1(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bitsPerPixel,
    N_uint32 pixelClock)
{
    return DV.init.GetClosestPixelClock(gaActiveHeadPrimary,xRes,yRes,
        bitsPerPixel,pixelClock);
}

/****************************************************************************
PARAMETERS:
crtc    - Place to store the CRTC timings for the current mode

REMARKS:
This function returns the set of CRTC timings that are currently active f
for the current display mode.
****************************************************************************/
void NAPI DRV_GetCRTCTimings(
    GA_CRTCInfo *crtc)
{
    *crtc = DV.cntCRTC;
}

/****************************************************************************
PARAMETERS:
crtc    - New CRTC timings to make active

REMARKS:
This function programs the set of CRTC timings passed in to make them the
active CRTC timings for the current display mode. This function is used
by the SNAP centering and refresh control functions.
****************************************************************************/
void NAPI DRV_SetCRTCTimings(
    GA_CRTCInfo *crtc)
{
    N_fix32 oldClock;

    // Save the CRTC timings as the current timings internally
    if (DV.cntMode.flags & gaUseBIOS)
        return;
    DV.cntCRTC = *crtc;

    // Find fixed point pixel clock for selected refresh rate
    oldClock = DV.cntPixelClock;
    DV.cntPixelClock = GetFixedPixelClock(&DV.cntMode,DV.cntCRTC.PixelClock);

    // Now program the CRTC timings into the hardware
    VGA_preModeSet();
    msDelay(200);
    clrinx(0x3D4,0x11,0x80);    // Disable write protect on CRTC registers
    DV.chip.setCRTCRegisters(&DV.cntCRTC);
    if (oldClock != DV.cntPixelClock)
        DV.chip.setPixelClock(DV.cntPixelClock);
    DV.chip.setBytesPerLine(DV.cntMode.bytesPerLine);
    if (DV.cntMode.flags & gaText)
        DRV_LoadTextFont();
    msDelay(200);
    VGA_postModeSet();
}

/****************************************************************************
PARAMETERS:
crtc    - New CRTC timings to make active

REMARKS:
This function programs the set of CRTC timings passed in to make them the
active CRTC timings for the current display mode. This function is used
by the SNAP centering and refresh control functions.
****************************************************************************/
void NAPI DRV_SaveCRTCTimings(
    GA_CRTCInfo *crtc)
{
    // Save the CRTC timings as the current timings in the database
    if (DV.cntMode.flags & gaText)
        _DRV_imports.CDBX_setCRTCTimings(DC.DeviceIndex,DV.init.GetActiveHead(),DV.cntMode.xRes * DV.cntModeInfo.XCharSize,DV.cntMode.yRes * DV.cntModeInfo.YCharSize,crtc,false,true);
    else
        _DRV_imports.CDBX_setCRTCTimings(DC.DeviceIndex,DV.init.GetActiveHead(),DV.cntMode.xRes,DV.cntMode.yRes,crtc,false,true);

    // Now program the timings into the hardware
    DRV_SetCRTCTimings(crtc);
}

/****************************************************************************
PARAMETERS:
refresh - New global refresh rate to make active

REMARKS:
This function sets the global refresh rate for all display modes to the
new global refresh rate setting for all heads on the device. This does not
change the physical refresh rate for the current display mode, but only
changes the database that will be used by the next mode set operation
(assuming the default refresh rate is selected).
****************************************************************************/
void NAPI DRV_SetGlobalRefresh(
    N_int32 refresh,
    N_int32 outputHead)
{
    // Make sure the output head parameter is in range
    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    if (outputHead < 0 || outputHead >= DV.init.GetNumberOfHeads())
        return;
    _DRV_imports.CDBX_setGlobalRefresh(DC.DeviceIndex,outputHead,refresh);
}

/****************************************************************************
PARAMETERS:
x       - X coordinate of first pixel to display
y       - Y coordinate of first pixel to display
waitVRT - Wait for retrace flag

REMARKS:
This function sets the currently visible hardware display start address as
an (x,y) coordinate offset from the start of physical display memory. This
is similar to SetDisplayStart, however the driver does the conversion
from an (x,y) coordinate to the display memory offset for you. The waitVRT
flag determines if the function will wait for the vertical retrace when
programming the hardware display start address and the following are valid
values:

value   - Description
0       - Schedule change for next retrace and return immediately
1       - Wait for vertical retrace during programming
****************************************************************************/
void NAPI DRV_SetDisplayStartXY(
    N_int32 x,
    N_int32 y,
    N_int32 waitVRT)
{
    int offset;

    if (!DV.opt.bVirtualScroll)
        return;
    switch (DV.cntMode.bits) {
        case 4:
            offset = (y * DV.cntMode.bytesPerLine * 8) + x;
            break;
        case 8:
            offset = y * DV.cntMode.bytesPerLine + x;
            break;
        case 15:
        case 16:
            offset = y * DV.cntMode.bytesPerLine + x*2;
            break;
        case 24:
            // Adjust to an 8 pixel boundary for 24bpp modes
            offset = y * DV.cntMode.bytesPerLine + x*3;
            offset = (offset / 24) * 24;
            break;
        case 32:
            offset = y * DV.cntMode.bytesPerLine + x*4;
            break;
        }
    DV.driver.SetDisplayStart(offset,waitVRT);
}

/****************************************************************************
REMARKS:
This function sets the current output display device for subsequent SNAP
functions.
****************************************************************************/
N_int32 NAPI DRV_SetDisplayOutputExt(
    N_int32 device,
    N_int32 outputHead)
{
    N_int32 oldDevice;

    // Make sure the output head parameter is in range
    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    if (outputHead < 0 || outputHead >= DV.init.GetNumberOfHeads())
        return -1;
    oldDevice = DV.outputDevice[outputHead];

    // Check that LCD is supported if it is requested
    if (device & gaOUTPUT_LCD) {
        if (!(DC.Attributes & gaHaveLCDOutput))
            return -1;
        }
    if (device & gaOUTPUT_DFP) {
        if (!(DC.Attributes & gaHaveDFPOutput))
            return -1;
        }

    // Check that TVOut is supported if it is requested
    if (device & gaOUTPUT_TV) {
        if (!DV.opt.bTVOut)
            return -1;
        if (!(DC.Attributes & gaHaveTVOutput))
            return -1;
        }

    // Now set the internal output device value. Note also that
    // we set the value in the device options structure for the
    // primary head, but not for the secondary heads.
    DV.outputDevice[outputHead] = device;
    if (outputHead == gaActiveHeadPrimary)
        DV.opt.outputDevice = DV.outputDevice[outputHead];

    // Now process the screen switch and reset the display mode
    if (oldDevice != device && DV.chip.setOptions) {
        DV.chip.setOptions();
        DV.init.PerformDisplaySwitch();
        }
    return oldDevice;
}

/****************************************************************************
REMARKS:
This function returns the current output device flags. Note that the
DV.outputDevice[0] flag is always used for the
****************************************************************************/
N_int32 NAPI DRV_GetDisplayOutputExt(
    N_int32 outputHead)
{
    // Make sure the output head parameter is in range
    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    if (outputHead < 0 || outputHead >= DV.init.GetNumberOfHeads())
        return 0;
    return DV.outputDevice[outputHead];
}

/****************************************************************************
REMARKS:
This function sets the current output display device for subsequent SNAP
functions.
****************************************************************************/
N_int32 NAPI DRV_SetDisplayOutput(
    N_int32 device)
{
    return DV.init.SetDisplayOutputExt(device,DV.init.GetActiveHead());
}

/****************************************************************************
REMARKS:
This function returns the current output device flags.
****************************************************************************/
N_int32 NAPI DRV_GetDisplayOutput(void)
{
    return DV.init.GetDisplayOutputExt(DV.init.GetActiveHead());
}

/****************************************************************************
REMARKS:
Returns the mode flags for the specified display mode, which includes flags
such as whether CRTC scaling is enabled etc.
****************************************************************************/
ulong DRV_getTextModeFlags(
    GA_mode *mode)
{
    if (mode->xRes > 132)
        return 0;
    return gaText | gaVGA | gaBankedOnly;
}

/****************************************************************************
REMARKS:
Programs the palette registers. This version forces the internal palette
width flag to 8bit mode, so that the hardware palette functions do not
do any conversion on the input data. This is how the VBE/Core interface
works (the apps do all translation).
****************************************************************************/
void NAPI DRV_SetPaletteDataExt(
    GA_paletteExt *palExt,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    int         i;
    GA_palette  pal[256];

    for (i = 0; i < num; i++) {
        pal[i].Red = palExt[i].Red >> 8;
        pal[i].Green = palExt[i].Green >> 8;
        pal[i].Blue = palExt[i].Blue >> 8;
        }
    DV.driver.SetPaletteData(pal,num,index,waitVRT);
}

/****************************************************************************
REMARKS:
Read the palette registers.
****************************************************************************/
void NAPI DRV_GetPaletteDataExt(
    GA_paletteExt *palExt,
    N_int32 num,
    N_int32 index)
{
    int         i;
    GA_palette  pal[256];

    DV.driver.GetPaletteData(pal,num,index);
    for (i = 0; i < num; i++) {
        palExt[i].Red = pal[i].Red << 8;
        palExt[i].Green = pal[i].Green << 8;
        palExt[i].Blue = pal[i].Blue << 8;
        }
}

/****************************************************************************
REMARKS:
Programs the palette registers. This version forces the internal palette
width flag to 8bit mode, so that the hardware palette functions do not
do any conversion on the input data. This is how the VBE/Core interface
works (the apps do all translation).
****************************************************************************/
void NAPI DRV_SetGammaCorrectDataExt(
    GA_paletteExt *palExt,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    int         i;
    GA_palette  pal[256];

    for (i = 0; i < num; i++) {
        pal[i].Red = palExt[i].Red >> 8;
        pal[i].Green = palExt[i].Green >> 8;
        pal[i].Blue = palExt[i].Blue >> 8;
        }
    DV.driver.SetGammaCorrectData(pal,num,index,waitVRT);
}

/****************************************************************************
REMARKS:
Read the palette registers.
****************************************************************************/
void NAPI DRV_GetGammaCorrectDataExt(
    GA_paletteExt *palExt,
    N_int32 num,
    N_int32 index)
{
    int         i;
    GA_palette  pal[256];

    DV.driver.GetGammaCorrectData(pal,num,index);
    for (i = 0; i < num; i++) {
        palExt[i].Red = pal[i].Red << 8;
        palExt[i].Green = pal[i].Green << 8;
        palExt[i].Blue = pal[i].Blue << 8;
        }
}

/****************************************************************************
REMARKS:
Set the bytes per line value for the display mode. Should not be called
by regular applications.
****************************************************************************/
ibool NAPI DRVVBE_SetBytesPerLine(
    int bytesPerLine,
    int *newBytes)
{
    if ((bytesPerLine = DV.chip.roundBytesPerLine(bytesPerLine,DV.cntMode.bits)) == -1)
        return false;
    DV.cntMode.bytesPerLine = DV.chip.setBytesPerLine(bytesPerLine);
    *newBytes = DV.cntMode.bytesPerLine;
    return true;
}

/****************************************************************************
REMARKS:
Set the RAMDAC into and out of 8-bit per primary mode.
****************************************************************************/
ibool NAPI DRVVBE_Set8BitDAC(
    ibool enable)
{
    if (!DV.dac.set8BitDAC(enable))
        return false;
    DV.cntDACWidth = enable ? 8 : 6;
    return true;
}

/****************************************************************************
REMARKS:
Programs the palette registers. This version forces the internal palette
width flag to 8bit mode, so that the hardware palette functions do not
do any conversion on the input data. This is how the VBE/Core interface
works (the apps do all translation).
****************************************************************************/
void NAPI DRVVBE_SetPaletteData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    int oldWidth = DV.cntDACWidth;

    DV.cntDACWidth = 8;
    DV.driver.SetPaletteData(pal,num,index,waitVRT);
    DV.cntDACWidth = oldWidth;
}

/****************************************************************************
REMARKS:
Read the palette registers.
****************************************************************************/
void NAPI DRVVBE_GetPaletteData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index)
{
    int oldWidth = DV.cntDACWidth;

    DV.cntDACWidth = 8;
    DV.driver.GetPaletteData(pal,num,index);
    DV.cntDACWidth = oldWidth;
}

/****************************************************************************
REMARKS:
This function will reset the current hardware display mode in a
non-destructive manner such that the screen and current hardware
acceleration state are preserved. This function is used to dynamically
switch the active display device and re-program the hardware as necessary
for the new active display device (ie: LCD/CRT/TV switching etc).
****************************************************************************/
N_int32 NAPI DRV_SwitchPhysicalResolution(
    N_int32 physicalXResolution,
    N_int32 physicalYResolution,
    N_int32 refreshRate)
{
    GA_mode                 e;
    GA_modeEntry            *entry;
    CDBX_record             crtcRec;
    N_int32                 attributes = DC.Attributes;
    ibool                   prevUseBIOS = DV.cntMode.flags & gaUseBIOS;

    // Allow the chipset driver to save/restore any extra state as necessary
    if (DV.chip.preModeReset)
        DV.chip.preModeReset();

    // Override the physical resolution with passed in values.
    if (physicalXResolution == -1 || physicalYResolution == -1
        || physicalXResolution > DV.cntModeInfo.XResolution
        || physicalYResolution > DV.cntModeInfo.YResolution) {
        DV.cntMode.xRes = DV.cntModeInfo.PhysicalXResolution = DV.cntModeInfo.XResolution;
        DV.cntMode.yRes = DV.cntModeInfo.PhysicalYResolution = DV.cntModeInfo.YResolution;
        }
    else {
        DV.cntMode.xRes = DV.cntModeInfo.PhysicalXResolution = physicalXResolution;
        DV.cntMode.yRes = DV.cntModeInfo.PhysicalYResolution = physicalYResolution;
        }

    // Find the new mode flags for the mode
    e.bits = DV.cntMode.bits;
    if (e.bits == 0) {
        e.xRes = DV.cntMode.xRes / DV.cntModeInfo.XCharSize;
        e.yRes = DV.cntMode.yRes / DV.cntModeInfo.YCharSize;
        }
    else {
        e.xRes = DV.cntMode.xRes;
        e.yRes = DV.cntMode.yRes;
        }
    if ((entry = DRV_validateModeXY(&e,DV.outputDevice[0],
            DV.init.GetActiveHead(),
            &DV.cntModeInfo.PhysicalXResolution,
            &DV.cntModeInfo.PhysicalYResolution)) == NULL)
        return -1;

    // Disable switching between NonVGA and VGA modes with this function. Theoretically
    // we should be able to get this working, but it does not work on the Mach64 chipsets
    // that need this and it is not really important so we have disabled it for now.
    if ((DV.cntMode.flags & gaNonVGA) != (entry->flags & gaNonVGA))
        return -1;
    DV.cntMode.flags = entry->flags;

    // Determine the CRTC timings for the mode if we are driving an external CRT
    // monitor and we are not using the BIOS.
    if (!(entry->flags & gaUseBIOS) && (DV.outputDevice[0] & gaOUTPUT_CRT)) {
        if (entry->flags & gaNoInterlaced)
            attributes &= ~gaHaveInterlaced;
        if (!_DRV_imports.CDBX_getCRTCTimings(DC.DeviceIndex,DV.init.GetActiveHead(),attributes,DV.cntModeInfo.PhysicalXResolution,DV.cntModeInfo.PhysicalYResolution,DV.opt.tripleScanLowRes,DV.opt.doubleScan512,refreshRate,&monitor,DV.cntModeInfo.MaxPixelClock,DV.licensed.maxRefresh,&crtcRec))
            return -1;
        DV.cntCRTC = crtcRec.crtc;
        }

    // Make sure the panning mode flag is reflected in the current
    // mode information block.
    if (DV.cntMode.flags & gaPanningMode)
        DV.cntModeInfo.AttributesExt |= gaIsPanningMode;
    else
        DV.cntModeInfo.AttributesExt &= ~gaIsPanningMode;

    // Update hardware cursor flag for the mode
    if (DV.cntMode.flags & gaHWCursor)
        DV.cntModeInfo.Attributes |= gaHaveHWCursor;
    else
        DV.cntModeInfo.Attributes &= ~gaHaveHWCursor;

    // Wait till engine idle if existing mode is accelerated before continuing.
    if (DV.cntMode.flags & gaAccMode)
        DV.state2d.WaitTillIdle();

    // Save the hardware palette/gamma ramp so we can restore it
    DV.dac.set8BitDAC(DV.cntDACWidth == 8);
    if (DV.cntMode.bits <= 8)
        DV.driver.GetPaletteDataExt(scratchPalExt,256,0);
    else if (IS_CVALID_PTR(DV.driver.GetGammaCorrectDataExt))
        DV.driver.GetGammaCorrectDataExt(scratchPalExt,256,0);

    // Optionally save the hardware cursor state
    if (DV.chip.saveCursorState)
        DV.chip.saveCursorState();

    // Find fixed point pixel clock for selected refresh rate
    DV.cntPixelClock = GetFixedPixelClock(&DV.cntMode,DV.cntCRTC.PixelClock);
    DV.chip.setOptions();
    DV.chip.preModeSet();
    DV.chip.disableLinear();
    if (DV.cntMode.flags & gaUseBIOS) {
        DV.cntModeNum |= gaDontClear;   // Don't clear the screen!
        DV.chip.setBIOSMode();
        DV.cntModeNum &= ~gaDontClear;
        VGA_preModeSet();
        }
    else {
        if (prevUseBIOS)
            DV.chip.setChipsetRegistersExt(false);
        DV.dac.setDACRegisters();
        DV.chip.setMemoryClock();
        if (DC.Attributes & gaHaveEngineClock)
            DV.chip.setEngineClock();
        DV.chip.setCRTCRegisters(&DV.cntCRTC);
        DV.chip.setPixelClock(DV.cntPixelClock);
        }
    DV.dac.set8BitDAC(DV.cntDACWidth == 8);
    DV.chip.setBytesPerLine(DV.cntMode.bytesPerLine);
    if (DV.cntFlags & gaLinearBuffer)
        DV.chip.enableLinear();
    else if (IS_CVALID_PTR(DV.driver.SetBank))
        DV.driver.SetBank(DV.cntBank);
    DV.driver.SetDisplayStart(DV.cntDisplayStart,false);
    DV.chip.postModeSet();

    // Optionall restore the hardware cursor
    if (DV.chip.restoreCursorState)
        DV.chip.restoreCursorState();

    // Restore the hardware palette/gamma ramp
    if (DV.cntMode.bits <= 8)
        DV.driver.SetPaletteDataExt(scratchPalExt,256,0,false);
    else if (IS_CVALID_PTR(DV.driver.GetGammaCorrectDataExt))
        DV.driver.SetGammaCorrectDataExt(scratchPalExt,256,0,false);

    // Allow the chipset driver to save/restore any extra state as necessary
    if (DV.chip.postModeReset)
        DV.chip.postModeReset();

    // Return success
    return 0;
}

/****************************************************************************
REMARKS:
This function will reset the current hardware display mode in a
non-destructive manner such that the screen and current hardware
acceleration state are preserved. This function is used to dynamically
switch the active display device and re-program the hardware as necessary
for the new active display device (ie: LCD/CRT/TV switching etc).
****************************************************************************/
void NAPI DRV_PerformDisplaySwitch(void)
{
    // Note: we call this function directly rather than via the function
    //       pointer because the zoom feature may be disabled via
    //       licensing while the laptop switching needs to continue to
    //       function.
    DRV_SwitchPhysicalResolution(-1,-1,DV.cntCRTC.RefreshRate / 100);
}

/****************************************************************************
RETURNS:
True if a mode change occurred, false if not.

REMARKS:
This function polls the hardware to determine if a mode change should be
processed, and if so it processes the mode change and returns to the
caller. This function should be called about once every 250ms by the
OS shell display driver to ensure that mode changes are handled in a
timely manner.
****************************************************************************/
N_int32 NAPI DRV_PollForDisplaySwitch(void)
{
    static N_int32  cntOutputDevice = -1;
    static N_int32  cntLCDExpand = -1;
    N_int32         screenSwitched = false;

    // Determine if a screen switch has been requested. Note that we
    // call the getPendingOptions() function even in text modes to
    // ensure that any pending flags are cleared as necessary.
    if (!DV.chip.getPendingOptions)
        return false;
    DV.chip.getPendingOptions();

    // We don't do anything special in text modes
    if (DV.cntMode.flags & gaText)
        return false;

    // Check to see if a screen switch event is now pending
    if ((cntOutputDevice != -1 && cntOutputDevice != DV.outputDevice[0]) ||
        (cntLCDExpand != -1 && cntLCDExpand != (int)DV.opt.bLCDExpand)) {
        // A screen switch is pending so this function should now return true.
        screenSwitched = true;
        }

    // Save current values for the next time we poll the hardware
    cntOutputDevice = DV.outputDevice[0];
    cntLCDExpand = DV.opt.bLCDExpand;
    return screenSwitched;
}

