/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the generic SNAP Graphics
*               Architecture API functions based on chipset specific
*               support modules.
*
****************************************************************************/

#include "snap/graphics/driver.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
HEADER:
graphics/ddc.h

PARAMETERS:
supportedStates - Place to store the supported states

RETURNS:
True if the device supports DPMS functionality, false if not.

REMARKS:
Returns the supported DPMS states for the chipset driver.
****************************************************************************/
int NAPI DRV_DPMSdetect(
    N_int16 *supportedStates)
{
    if (DV.DPMSCaps) {
        *supportedStates = DV.DPMSCaps;
        return true;
        }
    return false;
}

/****************************************************************************
HEADER:
graphics/ddc.h

PARAMETERS:
state   - New DPMS state to make active

REMARKS:
Puts the system into the requested DPMS state.
****************************************************************************/
void NAPI DRV_DPMSsetState(
    N_int32 state)
{
    int i3C4,i3CE,i3D4;

    if (state != DV.cntDPMSState && (state == DPMS_on || (DV.DPMSCaps & state))) {
        // Save the state of the VGA index registers before setting the
        // CRTC values so that we can correctly restore them
        i3C4 = inpb(0x3C4);
        i3CE = inpb(0x3CE);
        i3D4 = inpb(0x3D4);

        // Now blank the VGA display output. This is important so that even
        // non-DPMS monitors and those that support power down by detecting
        // no video signal will still work properly.
        if (state == DPMS_on)
            clrinx(0x3C4,0x01,0x20);
        else
            setinx(0x3C4,0x01,0x20);

        // Call the chipset specific function to power down the monitor
        DV.chip.setDPMS(state);
        DV.cntDPMSState = state;

        // Restore the VGA registers to the state they were in before we
        // started re-programming the CRT controller.
        outpb(0x3C4,i3C4);
        outpb(0x3CE,i3CE);
        outpb(0x3D4,i3D4);
        }
}

/****************************************************************************
HEADER:
graphics/ddc.h

PARAMETERS:
capabilities    - Place to store the capabilities for the driver
numchannels - Place to store the number of monitor ports supported

RETURNS:
True if the device supports SCI functionality, false if not.

REMARKS:
Returns the supported SCI capabilities for the chipset driver.
****************************************************************************/
int NAPI DRV_SCIdetect(
    N_uint8 *capabilities,
    N_int32 *numchannels)
{
    if (DV.SCICaps) {
        *capabilities= DV.SCICaps;
        if (DC.Attributes & gaHaveDFPOutput)
            *numchannels = 2;
        else
            *numchannels = 1;
        return true;
        }
    return false;
}

