/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the generic SNAP Graphics
*               Architecture API functions based on chipset specific
*               support modules.
*
****************************************************************************/

#include "snap/graphics/driver.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Stub function that simply does nothing when called.
****************************************************************************/
void DRV_stub(void)
{
}

/****************************************************************************
DESCRIPTION:
Determines the amount of memory on the controller by probing the linear
framebuffer space.

HEADER:
graphics/driver.h

PARAMETERS:
maxMB   - Maximum amount of memory in MB

RETURNS:
Amount of memory installed in kilobytes.

REMARKS:
Determines the amount of memory installed on the video card by checking for
memory access at the linear framebuffer at different locations. We probe
only at megabyte boundaries.

NOTE:   This function does not require a mode set and assumes the linear
        buffer is enabled.
****************************************************************************/
int DRV_findLinearMem(
    int maxMB)
{
    int     memSize = maxMB;
    ulong   i,*saveBuf;
    uchar   *linPtr = DC.LinearMem;

#define FIND_OFFSET (512*1024)

    // Allocate memory for the save buffer based on maximum memory to check
    // for.
    if ((saveBuf = malloc(maxMB * sizeof(saveBuf[0]))) == NULL)
        return 0;

    // Write the megabyte number to each buffer, moving down from the highest
    // value to the lowest.
    for (i = maxMB; i > 0; i--) {
        ulong address = (ulong)linPtr + (i-1)*MB + FIND_OFFSET;
        saveBuf[i-1] = PM_getLong(address);
        PM_setLong(address,i);
        }

    // Now read from the lowest to the highest until we get a value that
    // does not match what we expect.
    memSize = maxMB;
    for (i = 1; i <= (ulong)maxMB; i++) {
        ulong address = (ulong)linPtr + (i-1)*MB + FIND_OFFSET;
        if (PM_getLong(address) != i) {
            memSize = (i-1);
            break;
            }
        }

    // Now restore the memory locations and return the memory size
    for (i = maxMB; i > 0; i--)
        PM_setLong(linPtr + (i-1)*MB + FIND_OFFSET,saveBuf[i-1]);

    // Free the save buffer on return
    free(saveBuf);
    return memSize * 1024;
}

/****************************************************************************
DESCRIPTION:
Determines the size of the buffer to save the state of the hardware

HEADER:
graphics/driver.h

PARAMETERS:
profile     - Pointer to the chipset profile table to measure
comments    - True if the profile table contains comment strings

RETURNS:
Size of the buffer to save the hardware state into.

REMARKS:
Determines the amount of memory required to save and restore the state
of the hardware registers from the chipset profile table.
****************************************************************************/
int DRV_getStateSize(
    ulong *profile,
    ibool comments)
{
    int     len = 0;

    while (*profile != REG_TERMINATOR) {
        if (regIsReference(*profile)) {
            // Referencing another table, so measure size of referenced
            // table with a recursive function call.
            len += DRV_getStateSize((ulong*)(*(profile+1)),comments);
            profile += 2;
            }
        else {
            // Add the width of the register to the buffer length
            switch (regWidth(*profile)) {
                case REG_SZ_BYTE:   len += 1;   break;
                case REG_SZ_WORD:   len += 2;   break;
                case REG_SZ_DWORD:  len += 4;   break;
                }
            if (comments)
                profile += 2;
            else
                profile ++;
            }
        }
    return len;
}

/****************************************************************************
REMARKS:
Delay for a short period of time.
****************************************************************************/
static void vga_delay(void)
{
#ifdef __INTEL__
    int i;

    // For the loop here we program the POST register. The length of this
    // delay is dependant only on ISA bus speed, but it is enough for
    // what we need.
    for (i = 0; i <= 10; i++)
        outpb(0x80, 0);
#else
    usDelay(1);
#endif
}

/****************************************************************************
DESCRIPTION:
Saves the hardware register state to a buffer.

HEADER:
graphics/driver.h

PARAMETERS:
profile     - Pointer to the chipset profile table
buffer      - Buffer to save the hardware state into
comments    - True if the profile table contains comment strings

RETURNS:
Pointer to the next space in the buffer

REMARKS:
Saves the hardware register state to a buffer using the chipset register
profile table to determine which registers to save.

Note:   After this function has been called if the mode is VGA compatible
        the VGA Attribute Controller index register bit 5 will have been
        turned off so the display output will be blanked. It it up to the
        calling code to turn this bit back on as necessary. This can be
        done with calls to VGA_screenOn and VGA_screenOff.
****************************************************************************/
uchar *DRV_saveState(
    ulong *profile,
    uchar *buffer,
    ibool comments)
{
    MMIO_DECLARE;

    while (*profile != REG_TERMINATOR) {
        if (regIsReference(*profile)) {
            // Referencing another table so save the state with a recursive
            // function call.
            buffer = DRV_saveState((ulong*)(*(profile+1)),buffer,comments);
            profile += 2;
            }
        else {
            if (regIsMMIO(*profile) && regIsIndexed(*profile)) {
                // Register is a memory mapped register
                switch (regWidth(*profile)) {
                    case REG_SZ_BYTE:
                        outmb(regIOPort(*profile),regIOIndex(*profile));
                        *buffer++ = inmb(regIOPort(*profile)+1);
                        break;
                    case REG_SZ_WORD:
                        outmb(regIOPort(*profile),regIOIndex(*profile));
                        *((ushort*)buffer) = inmw(regIOPort(*profile)+2);
                        buffer += 2;
                        break;
                    case REG_SZ_DWORD:
                        outmb(regIOPort(*profile),regIOIndex(*profile));
                        *((ulong*)buffer) = inmd(regIOPort(*profile)+4);
                        buffer += 4;
                        break;
                    }
                }
            else if (regIsMMIO(*profile)) {
                // Register is a memory mapped register
                switch (regWidth(*profile)) {
                    case REG_SZ_BYTE:
                        *buffer++ = inmb(regMMIOOffset(*profile));
                        break;
                    case REG_SZ_WORD:
                        *((ushort*)buffer) = inmw(regMMIOOffset(*profile));
                        buffer += 2;
                        break;
                    case REG_SZ_DWORD:
                        *((ulong*)buffer) = inmd(regMMIOOffset(*profile));
                        buffer += 4;
                        break;
                    }
                }
            else if (regIsReloc(*profile) && regIsIndexed(*profile)) {
                // Register is a relocatable I/O indexed register
                switch (regWidth(*profile)) {
                    case REG_SZ_BYTE:
                        *buffer++ = rdinx(DC.IOBase+regIOPort(*profile),regIOIndex(*profile));
                        break;
                    case REG_SZ_WORD:
                        outpb(DC.IOBase+regIOPort(*profile),regIOIndex(*profile));
                        *((ushort*)buffer) = inpw(DC.IOBase+regIOPort(*profile)+2);
                        buffer += 2;
                        break;
                    case REG_SZ_DWORD:
                        outpb(DC.IOBase+regIOPort(*profile),regIOIndex(*profile));
                        *((ulong*)buffer) = inpd(DC.IOBase+regIOPort(*profile)+4);
                        buffer += 4;
                        break;
                    }
                }
            else if (regIsIndexed(*profile)) {
                // Register is a fixed I/O indexed register
                if (regIOPort(*profile) == 0x3C0) {
                    // Special case handling for VGA attribute controller
                    inpb(0x3DA);
                    vga_delay();
                    outpb(0x3C0,regIOIndex(*profile));
                    vga_delay();
                    *buffer++ = inpb(0x3C1);
                    vga_delay();
                    }
                else {
                    switch (regWidth(*profile)) {
                        case REG_SZ_BYTE:
                            *buffer++ = rdinx(regIOPort(*profile),regIOIndex(*profile));
                            break;
                        case REG_SZ_WORD:
                            outpb(regIOPort(*profile),regIOIndex(*profile));
                            *((ushort*)buffer) = inpw(regIOPort(*profile)+2);
                            buffer += 2;
                            break;
                        case REG_SZ_DWORD:
                            outpb(regIOPort(*profile),regIOIndex(*profile));
                            *((ulong*)buffer) = inpd(regIOPort(*profile)+4);
                            buffer += 4;
                            break;
                        }
                    }
                }
            else if (regIsReloc(*profile)) {
                // Register is a relocatable I/O register
                switch (regWidth(*profile)) {
                    case REG_SZ_BYTE:
                        *buffer++ = inpb(DC.IOBase+regIOPort(*profile));
                        break;
                    case REG_SZ_WORD:
                        *((ushort*)buffer) = inpw(DC.IOBase+regIOPort(*profile));
                        buffer += 2;
                        break;
                    case REG_SZ_DWORD:
                        *((ulong*)buffer) = inpd(DC.IOBase+regIOPort(*profile));
                        buffer += 4;
                        break;
                    }
                }
            else {
                // Register is a fixed I/O direct register
                switch (regWidth(*profile)) {
                    case REG_SZ_BYTE:
                        if (regIOPort(*profile) == 0x3C2)
                            *buffer++ = inpb(0x3CC);        // VGA misc output
                        else
                            *buffer++ = inpb(regIOPort(*profile));
                        break;
                    case REG_SZ_WORD:
                        *((ushort*)buffer) = inpw(regIOPort(*profile));
                        buffer += 2;
                        break;
                    case REG_SZ_DWORD:
                        *((ulong*)buffer) = inpd(regIOPort(*profile));
                        buffer += 4;
                        break;
                    }
                }
            if (comments)
                profile += 2;
            else
                profile ++;
            }
        }

    return buffer;
}

/****************************************************************************
DESCRIPTION:
Restores the hardware register state from a buffer.

HEADER:
graphics/driver.h

PARAMETERS:
profile     - Pointer to the chipset profile table
buffer      - Buffer to restore the hardware state from
comments    - True if the profile table contains comment strings

RETURNS:
Pointer to the next space in the buffer

REMARKS:
Restores the hardware register state from a buffer using the chipset register
profile table to determine which registers to restore.

Note:   After this function has been called if the mode is VGA compatible
        the VGA Attribute Controller index register bit 5 will have been
        turned off so the display output will be blanked. It it up to the
        calling code to turn this bit back on as necessary. This can be
        done with calls to VGA_screenOn and VGA_screenOff.

Note:   This function assumes the calling code will have unlocked any
        necessary VGA or extended registers before calling this function.
        It also assumes the VGA controller has been put into synchronous
        reset mode before calling this function, and when the appropriate
        register is programmed it will return back to normal operation.
****************************************************************************/
uchar *DRV_restoreState(
    ulong *profile,
    uchar *buffer,
    ibool comments)
{
    MMIO_DECLARE;

    while (*profile != REG_TERMINATOR) {
        if (regIsReference(*profile)) {
            // Referencing another table so save the state with a recursive
            // function call.
            buffer = DRV_restoreState((ulong*)(*(profile+1)),buffer,comments);
            profile += 2;
            }
        else {
            if (regIsMMIO(*profile) && regIsIndexed(*profile)) {
                // Register is a memory mapped register
                switch (regWidth(*profile)) {
                    case REG_SZ_BYTE:
                        outmb(regIOPort(*profile),regIOIndex(*profile));
                        outmb(regIOPort(*profile)+1,*((uchar*)buffer));
                        buffer++;
                        break;
                    case REG_SZ_WORD:
                        outmb(regIOPort(*profile),regIOIndex(*profile));
                        outmw(regIOPort(*profile)+2,*((ushort*)buffer));
                        buffer += 2;
                        break;
                    case REG_SZ_DWORD:
                        outmb(regIOPort(*profile),regIOIndex(*profile));
                        outmd(regIOPort(*profile)+4,*((ulong*)buffer));
                        buffer += 4;
                        break;
                    }
                }
            else if (regIsMMIO(*profile)) {
                // Register is a memory mapped register
                switch (regWidth(*profile)) {
                    case REG_SZ_BYTE:
                        outmb(regMMIOOffset(*profile),*buffer++);
                        break;
                    case REG_SZ_WORD:
                        outmw(regMMIOOffset(*profile),*((ushort*)buffer));
                        buffer += 2;
                        break;
                    case REG_SZ_DWORD:
                        outmd(regMMIOOffset(*profile),*((ulong*)buffer));
                        buffer += 4;
                        break;
                    }
                }
            else if (regIsReloc(*profile) && regIsIndexed(*profile)) {
                // Register is a relocatable I/O indexed register
                switch (regWidth(*profile)) {
                    case REG_SZ_BYTE:
                        wrinx(DC.IOBase+regIOPort(*profile),regIOIndex(*profile),*buffer++);
                        break;
                    case REG_SZ_WORD:
                        outpb(DC.IOBase+regIOPort(*profile),regIOIndex(*profile));
                        outpw(DC.IOBase+regIOPort(*profile)+2,*((ushort*)buffer));
                        buffer += 2;
                        break;
                    case REG_SZ_DWORD:
                        outpb(DC.IOBase+regIOPort(*profile),regIOIndex(*profile));
                        outpd(DC.IOBase+regIOPort(*profile)+4,*((ulong*)buffer));
                        buffer += 4;
                        break;
                    }
                }
            else if (regIsIndexed(*profile)) {
                // Register is a fixed I/O indexed register
                if (regIOPort(*profile) == 0x3C0) {
                    // Special case handling for VGA attribute controller
                    inpb(0x3DA);
                    vga_delay();
                    outpb(0x3C0,regIOIndex(*profile));
                    vga_delay();
                    outpb(0x3C0,*buffer++);
                    vga_delay();
                    }
                else {
                    switch (regWidth(*profile)) {
                        case REG_SZ_BYTE:
                            wrinx(regIOPort(*profile),regIOIndex(*profile),*buffer++);
                            break;
                        case REG_SZ_WORD:
                            outpb(regIOPort(*profile),regIOIndex(*profile));
                            outpw(regIOPort(*profile)+2,*((ushort*)buffer));
                            buffer += 2;
                            break;
                        case REG_SZ_DWORD:
                            outpb(regIOPort(*profile),regIOIndex(*profile));
                            outpd(regIOPort(*profile)+4,*((ulong*)buffer));
                            buffer += 4;
                            break;
                        }
                    }
                }
            else if (regIsReloc(*profile)) {
                // Register is a relocatable I/O register
                switch (regWidth(*profile)) {
                    case REG_SZ_BYTE:
                        outpb(DC.IOBase+regIOPort(*profile),*buffer++);
                        break;
                    case REG_SZ_WORD:
                        outpw(DC.IOBase+regIOPort(*profile),*((ushort*)buffer));
                        buffer += 2;
                        break;
                    case REG_SZ_DWORD:
                        outpd(DC.IOBase+regIOPort(*profile),*((ulong*)buffer));
                        buffer += 4;
                        break;
                    }
                }
            else {
                // Register is a fixed I/O direct register
                switch (regWidth(*profile)) {
                    case REG_SZ_BYTE:
                        outpb(regIOPort(*profile),*buffer++);
                        break;
                    case REG_SZ_WORD:
                        outpw(regIOPort(*profile),*((ushort*)buffer));
                        buffer += 2;
                        break;
                    case REG_SZ_DWORD:
                        outpd(regIOPort(*profile),*((ulong*)buffer));
                        buffer += 4;
                        break;
                    }
                }
            if (comments)
                profile += 2;
            else
                profile ++;
            }
        }

    return buffer;
}

/****************************************************************************
DESCRIPTION:
Determines the size of the buffer to save the state of the hardware

HEADER:
graphics/driver.h

PARAMETERS:
profile     - Pointer to the chipset profile table to measure

RETURNS:
Size of the buffer to save the hardware state into.

REMARKS:
Determines the amount of memory required to save and restore the state
of the RAMDAC registers from the RAMDAC profile table.
****************************************************************************/
int DRV_getDACStateSize(
    ulong *profile,
    ibool comments)
{
    int     len = 0;

    while (*profile != DAC_TERMINATOR) {
        if (dacIsReference(*profile)) {
            // Referencing another table, so measure size of referenced
            // table with a recursive function call.
            len += DRV_getDACStateSize((ulong*)(*(profile+1)),comments);
            profile += 2;
            }
        else {
            // Add the width of the register to the buffer length
            len += 1;
            if (comments)
                profile += 2;
            else
                profile++;
            }
        }
    return len;
}

/****************************************************************************
DESCRIPTION:
Saves the RAMDAC register state to a buffer.

HEADER:
graphics/driver.h

PARAMETERS:
profile     - Pointer to the chipset profile table
buffer      - Buffer to save the hardware state into

RETURNS:
Pointer to the next space in the buffer

REMARKS:
Saves the hardware RAMDAC state to a buffer using the RAMDAC register
profile table to determine which registers to save.
****************************************************************************/
uchar *DRV_saveDACState(
    ulong *profile,
    uchar *buffer,
    ibool comments)
{
    while (*profile != DAC_TERMINATOR) {
        if (dacIsReference(*profile)) {
            // Referencing another table so save the state with a recursive
            // function call.
            buffer = DRV_saveDACState((ulong*)(*(profile+1)),buffer,comments);
            profile += 2;
            }
        else {
            if (dacIsDbIndex(*profile)) {
                // Register is a double indexed register
                *buffer++ = DV.chip.rdDACIdx(dacIndex(*profile));
                }
            else {
                // Register is a single indexed register
                *buffer++ = DV.chip.rdDACReg(dacIndex(*profile));
                }
            if (comments)
                profile += 2;
            else
                profile++;
            }
        }
    return buffer;
}

/****************************************************************************
DESCRIPTION:
Restores the RAMDAC register state from a buffer.

HEADER:
graphics/driver.h

PARAMETERS:
profile     - Pointer to the chipset profile table
buffer      - Buffer to restore the hardware state from

RETURNS:
Pointer to the next space in the buffer

REMARKS:
Restores the RAMDAC register state from a buffer using the RAMDAC register
profile table to determine which registers to restore.
****************************************************************************/
uchar *DRV_restoreDACState(
    ulong *profile,
    uchar *buffer,
    ibool comments)
{
    while (*profile != DAC_TERMINATOR) {
        if (dacIsReference(*profile)) {
            // Referencing another table so restore the state with a recursive
            // function call.
            buffer = DRV_restoreDACState((ulong*)(*(profile+1)),buffer,comments);
            profile += 2;
            }
        else {
            if (dacIsDbIndex(*profile)) {
                // Register is a double indexed register
                DV.chip.wrDACIdx(dacIndex(*profile),*buffer++);
                }
            else {
                // Register is a single indexed register
                DV.chip.wrDACReg(dacIndex(*profile),*buffer++);
                }
            if (comments)
                profile += 2;
            else
                profile++;
            }
        }
    return buffer;
}

/****************************************************************************
PARAMETERS:
filename    - Place to store the unique filename

REMARKS:
This function generates a unique filename for the current chipset
configuration that is used to store POST register information and the
configuration options to disk at runtime. Because the filename is
unique, each card that is plugged into the system will have a different
set of persistent POST register state and configuration options.
****************************************************************************/
void NAPI DRV_getUniqueFilename(
    char *filename,
    int type)
{
#ifndef __ANALYSE__
    char    buf[20];

    strcpy(filename,DRV_chipsetFilenames[DV.chipID]);
    if (type == 3) {
        strcat(filename,".lcd");
        strcpy(buf,filename);
        }
    else if (type == 4) {
        strcpy(filename,"clocks.%02x");
        sprintf(buf,filename,DC.DeviceIndex);
        }
    else {
        strcat(filename,".%02x");
        sprintf(buf,filename,DV.revID);
        }
    strcpy(filename,DRV_path);
    switch (type) {
        case 1:
            strcat(filename,"modes");
            break;
        case 2:
            strcat(filename,"options");
            break;
        }
    PM_backslash(filename);
    strcat(filename,buf);
#endif
}

/****************************************************************************
PARAMETERS:
m           - Place to store the monitor configuration information
outputHead  - Output device to detect the PnP monitor on

REMARKS:
This function returns a copy of the currently configured monitor in
the driver.
****************************************************************************/
void NAPI DRV_getMonitorInfo(
    GA_monitor *m,
    N_int32 outputHead)
{
    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    if (outputHead >= 0 && outputHead < DV.init.GetNumberOfHeads())
        *m = monitor[outputHead];
}

/****************************************************************************
PARAMETERS:
m           - New monitor configuration information
outputHead  - Output device to detect the PnP monitor on

REMARKS:
This function sets the internal monitor configuration information in the
driver to the passed in values. This automatically changes the internal
behaviour of the driver to support the new monitor configuration.
****************************************************************************/
void NAPI DRV_setMonitorInfo(
    GA_monitor *m,
    N_int32 outputHead)
{
    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    if (outputHead >= 0 && outputHead < DV.init.GetNumberOfHeads())
        monitor[outputHead] = *m;
}

/****************************************************************************
REMARKS:
This function read's the value of the the POST registers from the hardware
and then writes them to disk. The name of the file is passed in via the
name parameter, but the parameter is assumed to be a 'printf' style format
string to format the DV.revID value such that the filename is unique to
each chip revision ID.

NOTE:   *ASSUMES* that the registers are already unlocked!!!
****************************************************************************/
int DRV_savePOSTRegisters(void)
{
#ifndef __ANALYSE__
    GA_chipEntry    *chip = DV.chip.getChipData();
    size_t          size;
    uchar           *data;
    FILE            *f;
    char            filename[256];

    size = DRV_getStateSize(chip->POSTchipsetProfile,false);
    if ((data = malloc(size)) == NULL)
        return nLoadMem;
    DRV_getUniqueFilename(filename,0);
    if ((f = fopen(filename,"wb")) == NULL)
        return nIOError;
    DV.chip.blankDisplay();
    DRV_saveState(chip->POSTchipsetProfile,data,false);
    DV.chip.unblankDisplay();
    if (fwrite(data,1,size,f) != size)
        return nIOError;
    fclose(f);
    free(data);
#endif
    return nOK;
}

/****************************************************************************
REMARKS:
This function read's the value of the the POST registers from disk
and then writes them to the hardware. The name of the file is passed in via
the name parameter, but the parameter is assumed to be a 'printf' style
format string to format the DV.revID value such that the filename is unique
to each chip revision ID.

NOTE:   *ASSUMES* that the registers are already unlocked!!!
****************************************************************************/
int DRV_restorePOSTRegisters(void)
{
#ifndef __ANALYSE__
    GA_chipEntry    *chip = DV.chip.getChipData();
    size_t          size;
    uchar           *data;
    FILE            *f;
    char            filename[256];

    size = DRV_getStateSize(chip->POSTchipsetProfile,false);
    if ((data = malloc(size)) == NULL)
        return nLoadMem;
    DRV_getUniqueFilename(filename,0);
    if ((f = fopen(filename,"rb")) == NULL)
        return nNotPOSTed;
    if (fread(data,1,size,f) != size)
        return nIOError;
    DV.chip.blankDisplay();
    clrinx(0x3D4,0x11,0x80);    // Disable write protect on CRTC registers
    DRV_restoreState(chip->POSTchipsetProfile,data,false);
    DV.chip.unblankDisplay();
    fclose(f);
    free(data);
#endif
    return nOK;
}

/****************************************************************************
REMARKS:
This function takes the description of a set of register bits in a table
and wites the bits to the registers. The table is terminated with
either an RG_NEWBYTE, RG_NEWWORD or RG_FINISH command. When this function
returns it will have advanced the table pointer to point to the command
that terminated the write function.

Between calls to the function, it will cache the currently active port
in the cntRGPort variable.
****************************************************************************/
uchar *DRV_writeRegister(
    ushort value,
    uchar *p)
{
    int         shift,zeros,ones,mask,cntBit = 0;
    static int  cntPort,cntIndex;

    for (;;) {
        int data = PM_getWord(p) & ~RG_CMD_MASK;
        switch (PM_getWord(p) & RG_CMD_MASK) {
            case RG_DPORT_CMD:
                // Change to a new direct I/O port
                cntPort = PM_getWord(p+2);
                p += 4;
                break;
            case RG_INDEX_CMD:
                // Change to a new index with same I/O port
                outpb(cntPort,cntIndex = data);
                p += 2;
                break;
            case RG_IPORT_CMD:
                // Change to a new I/O port and index
                cntPort = PM_getWord(p+2);
                outpb(cntPort,cntIndex = data);
                p += 4;
                break;
            case RG_MASK_CMD:
                // Write masked value to current register
                mask = data;
                zeros = ones = 0;
                while (!(mask & 0x1)) {
                    zeros++;
                    mask >>= 1;
                    }
                while (mask & 0x1) {
                    ones++;
                    mask >>= 1;
                    }
                shift = (zeros - cntBit);
                cntBit += ones;
                if (shift > 0)
                    modreg(cntPort+1,data,value << shift);
                else
                    modreg(cntPort+1,data,value >> -shift);
                p += 2;
                break;
            case RG_STARTBIT_CMD:
                // Change the current source starting bit to a new value
                cntBit = data;
                p += 2;
                break;
            default:
                return p;
            }
        }
}

/****************************************************************************
REMARKS:
Writes multiple VGA compatible registers from the register description
stream. This allow us to program registers such as the VGA CRTC registers
that have bits spead over multiple registers.
****************************************************************************/
void DRV_writeMultiRegisters(
    uchar *data,
    ushort *profile)
{
    uchar *p = (uchar*)profile;

    for (;;) {
        int offset = PM_getWord(p) & ~RG_CMD_MASK;
        switch (PM_getWord(p) & RG_CMD_MASK) {
            case RG_NEWBYTE_CMD:
            case RG_CNTBYTE_CMD:
                p = DRV_writeRegister(PM_getByte(data+offset),p+2);
                break;
            case RG_NEWWORD_CMD:
            case RG_CNTWORD_CMD:
                p = DRV_writeRegister(PM_getWord(data+offset),p+2);
                break;
            default:
                return;
            }
        }
}

/****************************************************************************
REMARKS:
Generic helper function to program the chipset registers for the display
mode from the profile tables.
****************************************************************************/
void DRV_setChipsetRegisters(void)
{
#ifndef __ANALYSE__
    GA_chipEntry    *chip = DV.chip.getChipData();

    if (DV.cntMode.bits == 0) {
        DRV_restoreState(VGA_modeProfile,VGA_TextData,false);
        DRV_restoreState(VGA_chipsetProfile,VGA_chipsetData,false);
        DRV_restoreState(chip->vgaChipsetProfile,chip->vgaChipsetData,false);
        }
    else {
        switch (DV.cntMode.bits) {
            case 4:
                DRV_restoreState(chip->modeProfile,chip->SVGA4Data,false);
                if (chip->colorDepthProfile && chip->colorDepth4Data)
                    DRV_restoreState(chip->colorDepthProfile,chip->colorDepth4Data,false);
                break;
            case 8:
                DRV_restoreState(chip->modeProfile,chip->SVGAData,false);
                if (chip->colorDepthProfile && chip->colorDepth8Data)
                    DRV_restoreState(chip->colorDepthProfile,chip->colorDepth8Data,false);
                break;
            case 15:
                DRV_restoreState(chip->modeProfile,chip->SVGAData,false);
                if (chip->colorDepthProfile && chip->colorDepth15Data)
                    DRV_restoreState(chip->colorDepthProfile,chip->colorDepth15Data,false);
                break;
            case 16:
                DRV_restoreState(chip->modeProfile,chip->SVGAData,false);
                if (chip->colorDepthProfile && chip->colorDepth16Data)
                    DRV_restoreState(chip->colorDepthProfile,chip->colorDepth16Data,false);
                break;
            case 24:
                DRV_restoreState(chip->modeProfile,chip->SVGAData,false);
                if (chip->colorDepthProfile && chip->colorDepth24Data)
                    DRV_restoreState(chip->colorDepthProfile,chip->colorDepth24Data,false);
                break;
            case 32:
                DRV_restoreState(chip->modeProfile,chip->SVGAData,false);
                if (chip->colorDepthProfile && chip->colorDepth32Data)
                    DRV_restoreState(chip->colorDepthProfile,chip->colorDepth32Data,false);
                break;
            }
        DRV_restoreState(chip->chipsetProfile,chip->chipsetData,false);
        }
#endif
}

/****************************************************************************
REMARKS:
Generic helper function to program the RAMDAC registers for the display
mode from the profile tables.
****************************************************************************/
void DRV_setDACRegisters(void)
{
#ifndef __ANALYSE__
    GA_DACEntry *dac = DV.dacModeData;
    ulong       *profile = DV.dacProfile;

    if (DV.cntMode.flags & gaVGA) {
        DRV_restoreDACState(profile,dac->VGAData,false);
        }
    else {
        switch (DV.cntMode.bits) {
            case 4:
                DRV_restoreDACState(profile,dac->colorDepth4Data,false);
                break;
            case 8:
                DRV_restoreDACState(profile,dac->colorDepth8Data,false);
                break;
            case 15:
                DRV_restoreDACState(profile,dac->colorDepth15Data,false);
                break;
            case 16:
                DRV_restoreDACState(profile,dac->colorDepth16Data,false);
                break;
            case 24:
                DRV_restoreDACState(profile,dac->colorDepth24Data,false);
                break;
            case 32:
                DRV_restoreDACState(profile,dac->colorDepth32Data,false);
                break;
            }
        }
#endif
}

/****************************************************************************
REMARKS:
Generic helper function to clear the display memory using either the
banked or linear framebuffer apertures depending on the display mode.
****************************************************************************/
void DRV_clearMemory(
    int entireFramebuffer)
{
    ibool   cleared = false;

    /* Bail if there has not been a mode set */
    if (DV.cntMode.flags == 0)
        return;

    if (DV.cntMode.flags & gaText) {
        // *Always* write in WORD's for VGA modes!
        rep_stosw(DC.TextMem,0x4000,0x0720);
        cleared = true;
        }
    else if (DV.cntMode.bits == 4) {
        int i,maxBank = DV.cntMaxMem >> 16;
        int partialSize = DV.cntMaxMem & 0xFFFF;
        int GC_INDEX = 0x3CE;

        // Setup the VGA hardware (write mode 0, read mode 0)
        outpw(GC_INDEX,0x0000);
        outpw(GC_INDEX,0x0F01);
        outpw(GC_INDEX,0x0003);
        outpw(GC_INDEX,0x0005);
        outpw(GC_INDEX,0x0F07);

        // Clear all full banks
        for (i = 0; i < maxBank; i++) {
            if (IS_CVALID_PTR(DV.driver.SetBank))
                DV.driver.SetBank(i);
            rep_stosb((uchar*)DC.BankedMem+0x0000,0x8000,0);
            rep_stosb((uchar*)DC.BankedMem+0x8000,0x8000,0);
            }
        // Clear last partial bank
        if (partialSize) {
            DV.driver.SetBank(i);
            rep_stosb(DC.BankedMem,partialSize,0);
            }
        if (IS_CVALID_PTR(DV.driver.SetBank))
            DV.driver.SetBank(0);

        // Now set the GC registers back to defaults!!
        outpw(GC_INDEX,0x0001);
        outpw(GC_INDEX,0x1005);
        cleared = true;
        }
#ifndef DEBUG_NEW_DRIVER
    else if (DV.cntMode.flags & gaAccMode) {
        // Clear the framebuffer using the hardware accelerator
        // when available. We do this to ensure that the buffer
        // clear is very fast (important for cards with lots of
        // memory like 64Meg or more!).
        void  (NAPIP DrawRectExt)(N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
        int         xRes = DV.cntVirtualX == -1 ? DV.cntMode.xRes : DV.cntVirtualX;
        int         yRes = DV.cntVirtualX == -1 ? DV.cntMode.yRes : DV.cntVirtualY;
        int         maxBuffers = entireFramebuffer ? DV.cntModeInfo.MaxBuffers : 1;
        int         bufSize = yRes * DV.cntMode.bytesPerLine;
        GA_buffer   buf;

        // Find the hardware rectangle fill function to use
        switch (DV.cntMode.bits) {
            case 8:  DrawRectExt = DV.draw_8bpp.DrawRectExt;  break;
            case 15:
            case 16: DrawRectExt = DV.draw_16bpp.DrawRectExt; break;
            case 24: DrawRectExt = DV.draw_24bpp.DrawRectExt; break;
            case 32: DrawRectExt = DV.draw_32bpp.DrawRectExt; break;
            }

        // Clear each framebuffer page in turn. We must do this a page
        // at a time, as some hardware devices have restrictions on the
        // total coordinate system addressed by the acceleratro.
        if (IS_CVALID_PTR(DrawRectExt)) {
            buf.dwSize = sizeof(buf);
            buf.Offset = 0;
            buf.Stride = DV.cntMode.bytesPerLine;
            buf.Width = xRes;
            buf.Height = yRes;
            while (maxBuffers--) {
                DV.state2d.SetDrawBuffer(&buf);
                DrawRectExt(0,0,xRes,yRes,0,GA_REPLACE_MIX);
                buf.Offset += bufSize;
                }

            // Make sure we wait for engine idle, since this may be called
            // prior to switching to a new display mode and the engine must be
            // idle before we re-program the hardware registers.
            DV.state2d.WaitTillIdle();
            cleared = true;
            }
        }
#endif
    if (!cleared) {
        ulong   maxMem = DV.cntMaxMem;

        if (!entireFramebuffer)
            maxMem = DV.cntMode.yRes * DV.cntMode.bytesPerLine;
        if (DV.cntModeNum & gaLinearBuffer) {
            rep_stosd(DC.LinearMem,maxMem / 4,0);
            }
        else if (IS_CVALID_PTR(DV.driver.SetBank)) {
            int i,maxBank = maxMem >> 16;
            int partialSize = maxMem & 0xFFFF;

            // Clear all full banks
            for (i = 0; i < maxBank; i++) {
                DV.driver.SetBank(i);
                rep_stosd(DC.BankedMem,0x4000,0);
                }
            // Clear last partial bank
            if (partialSize) {
                DV.driver.SetBank(i);
                rep_stosd(DC.BankedMem,partialSize >> 2,0);
                }
            DV.driver.SetBank(0);
            }
        }
}

/****************************************************************************
REMARKS:
Generic helper function to calculate the PLL pixel clock based on the
generic PLL parameters. The following formulas are using to find the PLL
programming parameters that are used to program the pixel clock in the chip:

    Fout = (M+k1) * Fref * k3
           ------------------
              (N+k2) * 2^R

****************************************************************************/
N_fix32 DRV_calculateClock(
    GA_PLLParams *pll,
    N_fix32 clock,
    int *m,
    int *n,
    int *r)
{
    N_fix32 diff,best,fOut,fVCO,fRef,minClock,harmonic;
    int     i,M,N,R;

    minClock = pll->minVCO >> pll->maxR;
    if (minClock < pll->minFreq)
        minClock = pll->minFreq;
    if (clock < minClock)
        clock = minClock;
    if (pll->flags & gaPLL_favorMinVCO) {
        // Find the smallest possible VCO frequency
        R = pll->minR;
        clock <<= pll->minR;
        while (clock < pll->minVCO) {
            R++;
            clock <<= 1;
            }
        }
    else {
        // Find the highest possible VCO frequency
        R = pll->maxR;
        clock <<= pll->maxR;
        while (clock > pll->maxVCO && R > pll->minR) {
            R--;
            clock >>= 1;
            }
        }

    // Now search for the best clock driving M and then solving for N
    fRef = FXmul(pll->fRef,N_TOFIX(pll->k3));
    diff = 0x7FFFFFFF;
    for (M = pll->minM; M <= pll->maxM; M++) {
        N = N_FIXROUND(FXdiv(FXmul(N_TOFIX(M),fRef),clock));
        if (N < pll->minN || N > pll->maxN)
            continue;
        fVCO = FXdiv(FXmul(N_TOFIX(M),fRef),N_TOFIX(N));
        fOut = fVCO >> R;

        // Check that we are not on a hot frequency
        if (pll != &DV.memoryClock) {
            if (pll->flags & gaPLL_hotFrequencies) {
                for (i = 0; i < pll->numClocks; i++) {
                    if (ABS(pll->hotFreq[i] - fOut) < CLOSE_CLOCK)
                        break;
                    }
                if (i < pll->numClocks)
                    continue;
                }

            // Check that we are not a harmonic multiple of the memory clock
            harmonic = DV.cntMemoryClock >> 2;
            for (i = 0; i < 7; i++) {
                if (ABS(harmonic - fOut) < CLOSE_CLOCK)
                    break;
                harmonic <<= 1;
                }
            if (i < 7)
                continue;
            }

        // Check if this is the closest clock and save it if it is
        if (ABS(fVCO - clock) < diff) {
            diff = ABS(fVCO - clock);
            best = fVCO;
            *m = M;
            *n = N;
            }
        }

    // If a 'best' value was never found
    if (diff != 0x7FFFFFFF) {
        *m -= pll->k1;
        *n -= pll->k2;
        *r = R;
        return (best >> R);
        }
    *m = *n = *r = -1;
    return (ulong)-1;
}

/****************************************************************************
REMARKS:
Function called *only* when the driver is initialising a mode with the
gaPartialModeSet flag set, which assumes someone else (ie: an IHV driver) has
set the display mode. What we need to do is determine anything special
about the environment and configure our internal driver as needed.
****************************************************************************/
void DRV_configChipsetFlags(void)
{
}

/****************************************************************************
REMARKS:
Extended chipset register programming function. Normally most drivers
don't need this parameter so we call the regular function instead.
****************************************************************************/
void DRV_setChipsetRegistersExt(
    int initAccel)
{
    (void)initAccel;
    DV.chip.setChipsetRegisters();
}

/****************************************************************************
REMARKS:
Function to allow drivers to override the pixel clock limits based on
the current resolution instead of fixed maximum pixel clocks per
color depth. By default this function does nothing.
****************************************************************************/
void DRV_adjustMaxPixelClock(
    GA_modeInfo *gaInfo,
    GA_modeEntry *entry)
{
    (void)gaInfo;
    (void)entry;
}

/****************************************************************************
REMARKS:
Returns the the currently active refresh rate for the display mode (in 0.01
Hz increments).
****************************************************************************/
N_int32 NAPI DRV_GetCurrentRefreshRate(void)
{
    return DV.cntCRTC.RefreshRate;
}

/****************************************************************************
REMARKS:
Default multi-head functions for single head cards.
****************************************************************************/
N_int32 NAPI DRV_GetNumberOfHeads(void)
{
    return 1;
}

/****************************************************************************
REMARKS:
Default multi-head functions for single head cards.
****************************************************************************/
N_int32 NAPI DRV_SetActiveHead(
    N_int32 headIndex)
{
    (void)headIndex;
    return 0;
}

/****************************************************************************
REMARKS:
Default multi-head functions for single head cards.
****************************************************************************/
N_int32 NAPI DRV_GetActiveHead(void)
{
    return 0;
}

