/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  IBM PC (MS DOS)
*
* Description:  Module to implement the generic SNAP Graphics
*               Architecture API functions based on chipset specific
*               support modules.
*
****************************************************************************/

#include "snap/graphics/driver.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
PARAMETERS:
funcs       - Pointer to the function list to make valid
size        - Number of functions in the function list
safetyLevel - Safety level requested

RETURNS:
True if the functions are valid, false if not.

REMARKS:
This function validates the function list and returns a NULL pointer if
the function list is empty and does not contain any functions. Internally
in our drivers we list functions that have not yet been implemented on
a particular hardware device with a value of 0xFFFFFFFF for the function
pointer. Unless the safetyLevel is set to -1, we convert those pointers
to NULL pointers so the user code will not try to call them.

However our test programs can request the blocks of function pointers
with a safetyLevel of -1 and get a list of all functions that are
implemented, not supported and simply not implemented yet. This will help
our engineers know which functions need to be implemented or at least
mapped our for specific hardware devices.
****************************************************************************/
static ibool MakeValid(
    ulong *funcs,
    int size,
    int safetyLevel)
{
    int     i,empty = true;

    for (i = 0; i < size; i++) {
        if ((funcs[i] == 0xFFFFFFFF || funcs[i] == 0xFFFFFFFE || funcs[i] == 0xFFFFFFFD) && safetyLevel != -1)
            funcs[i] = 0;
        if (funcs[i] != 0)
            empty = false;
        }
    return !empty;
}

#define RETURN_VALID(p) \
{ \
    COPY_STRUCTURE2(funcs,&(p),dwSize); \
    return MakeValid(((ulong*)funcs)+1,(dwSize-sizeof(ulong))/4,safetyLevel); \
}

/****************************************************************************
DESCRIPTION:
Returns the function pointers for the specified function group.

HEADER:
graphics/driver.h

PARAMETERS:
id          - Identifier for the function group to get pointers for
safetyLevel - Safety level requested
funcs       - Pointer to the function block to fill in

RETURNS:
True if the functions are supported, false if not.

REMARKS:
This function is the main function that is used by the graphics application
code to get a block of function pointers for a specified function group.
The function groups are defined GA_funcGroupsType enumeration, and breaks
up the functions in the device driver API into groups of logically similar
functions.
****************************************************************************/
ibool NAPI DRV_queryFunctions(
    N_uint32 id,
    N_int32 safetyLevel,
    void *funcs)
{
    ulong   dwSize = *((ulong*)funcs);

    memset(funcs,0,dwSize);
    switch (id) {
        case GA_GET_INITFUNCS:
            RETURN_VALID(DV.init);
        case GA_GET_DRIVERFUNCS: {
            static GA_driverFuncs   driver;
            memcpy(&driver,&DV.driver,sizeof(driver));
            if (!DV.opt.bGammaCorrect)
                driver.SetGammaCorrectData = NULL;
            RETURN_VALID(driver);
            }
        case GA_GET_CURSORFUNCS:
            if ((DV.cntModeInfo.Attributes & gaHaveHWCursor) && (DV.cntMode.flags & gaHWCursor) && DV.opt.bHardwareCursor)
                RETURN_VALID(DV.cursor);
            break;
        case GA_GET_DPMSFUNCS:
            if (DV.opt.bDPMS)
                RETURN_VALID(DV.dpms);
            break;
        case GA_GET_SCIFUNCS:
            if (DV.opt.bDDC)
                RETURN_VALID(DV.sci);
            break;
        case GA_GET_2DSTATEFUNCS:
            if ((DV.cntModeInfo.Attributes & gaHaveAccel2D) && (DV.cntMode.flags & gaAccMode) && DV.opt.bHardwareAccel2D) {
                static GA_2DStateFuncs state2d;
                memcpy(&state2d,&DV.state2d,sizeof(state2d));
                if (!DV.opt.bMonoPattern && !DV.opt.bTransMonoPattern)
                    state2d.Set8x8MonoPattern = NULL;
                if (!DV.opt.bMonoPattern)
                    state2d.Use8x8MonoPattern = NULL;
                if (!DV.opt.bTransMonoPattern)
                    state2d.Use8x8TransMonoPattern = NULL;
                if (!DV.opt.bColorPattern && !DV.opt.bTransColorPattern)
                    state2d.Set8x8ColorPattern = NULL;
                if (!DV.opt.bColorPattern)
                    state2d.Use8x8ColorPattern = NULL;
                if (!DV.opt.bTransColorPattern)
                    state2d.Use8x8TransColorPattern = NULL;
                RETURN_VALID(state2d);
                }
            break;
        case GA_GET_2DRENDERFUNCS:
            if ((DV.cntModeInfo.Attributes & gaHaveAccel2D) && (DV.cntMode.flags & gaAccMode) && DV.opt.bHardwareAccel2D) {
                static GA_2DRenderFuncs draw;
                switch (DV.cntMode.bits) {
                    case 8:     memcpy(&draw,&DV.draw_8bpp,sizeof(draw)); break;
                    case 15:
                    case 16:    memcpy(&draw,&DV.draw_16bpp,sizeof(draw)); break;
                    case 24:    memcpy(&draw,&DV.draw_24bpp,sizeof(draw)); break;
                    case 32:    memcpy(&draw,&DV.draw_32bpp,sizeof(draw)); break;
                    }
                if (!DV.opt.bMonoPattern && !DV.opt.bTransMonoPattern) {
                    draw.DrawPattScanList = NULL;
                    draw.DrawPattEllipseList = NULL;
                    draw.DrawPattFatEllipseList = NULL;
                    draw.DrawPattRect = NULL;
                    draw.DrawPattTrap = NULL;
                    draw.BitBltPatt = NULL;
                    draw.BitBltPattLin = NULL;
                    draw.BitBltPattSys = NULL;
                    draw.BitBltPattBM = NULL;
                    }
                if (!DV.opt.bColorPattern && !DV.opt.bTransColorPattern) {
                    draw.DrawColorPattScanList = NULL;
                    draw.DrawColorPattEllipseList = NULL;
                    draw.DrawColorPattFatEllipseList = NULL;
                    draw.DrawColorPattRect = NULL;
                    draw.DrawColorPattTrap = NULL;
                    draw.BitBltColorPatt = NULL;
                    draw.BitBltColorPattLin = NULL;
                    draw.BitBltColorPattSys = NULL;
                    draw.BitBltColorPattBM = NULL;
                    }
                if (!DV.opt.bSysMem) {
                    draw.PutMonoImageMSBSys = NULL;
                    draw.PutMonoImageLSBSys = NULL;
                    draw.ClipMonoImageMSBSys = NULL;
                    draw.ClipMonoImageLSBSys = NULL;
                    draw.BitBltSys = NULL;
                    draw.BitBltPattSys = NULL;
                    draw.BitBltColorPattSys = NULL;
                    draw.SrcTransBltSys = NULL;
                    draw.DstTransBltSys = NULL;
                    draw.StretchBltSys = NULL;
                    draw.ConvertBltSys_Obsolete = NULL;
                    draw.StretchConvertBltSys_Obsolete = NULL;
                    draw.BitBltFxSys = NULL;
                    draw.GetBitmapSys = NULL;
                    }
                if (!DV.opt.bLinear) {
                    draw.PutMonoImageMSBLin = NULL;
                    draw.PutMonoImageLSBLin = NULL;
                    draw.ClipMonoImageMSBLin = NULL;
                    draw.ClipMonoImageLSBLin = NULL;
                    draw.BitBltLin = NULL;
                    draw.BitBltPattLin = NULL;
                    draw.BitBltColorPattLin = NULL;
                    draw.SrcTransBltLin = NULL;
                    draw.DstTransBltLin = NULL;
                    draw.StretchBltLin = NULL;
                    draw.BitBltFxLin = NULL;
                    }
                if (!DV.opt.bBusMaster) {
                    draw.PutMonoImageMSBBM = NULL;
                    draw.PutMonoImageLSBBM = NULL;
                    draw.ClipMonoImageMSBBM = NULL;
                    draw.ClipMonoImageLSBBM = NULL;
                    draw.BitBltBM = NULL;
                    draw.BitBltPattBM = NULL;
                    draw.BitBltColorPattBM = NULL;
                    draw.SrcTransBltBM = NULL;
                    draw.DstTransBltBM = NULL;
                    draw.StretchBltBM = NULL;
                    draw.ConvertBltBM_Obsolete = NULL;
                    draw.StretchConvertBltBM_Obsolete = NULL;
                    draw.BitBltFxBM = NULL;
                    draw.GetBitmapBM = NULL;
                    }
                if (!DV.opt.bDrawScanList) {
                    draw.DrawScanList = NULL;
                    draw.DrawPattScanList = NULL;
                    draw.DrawColorPattScanList = NULL;
                    }
                if (!DV.opt.bDrawEllipseList) {
                    draw.DrawEllipseList = NULL;
                    draw.DrawPattEllipseList = NULL;
                    draw.DrawColorPattEllipseList = NULL;
                    }
                if (!DV.opt.bDrawFatEllipseList) {
                    draw.DrawFatEllipseList = NULL;
                    draw.DrawPattFatEllipseList = NULL;
                    draw.DrawColorPattFatEllipseList = NULL;
                    }
                if (!DV.opt.bDrawRect) {
                    draw.DrawRect = NULL;
                    draw.DrawPattRect = NULL;
                    draw.DrawColorPattRect = NULL;
                    }
                if (!DV.opt.bDrawRectLin) {
                    draw.DrawRectLin = NULL;
                    }
                if (!DV.opt.bDrawTrap) {
                    draw.DrawTrap = NULL;
                    draw.DrawPattTrap = NULL;
                    draw.DrawColorPattTrap = NULL;
                    }
                if (!DV.opt.bDrawLine) {
                    draw.DrawLineInt = NULL;
                    draw.DrawBresenhamLine = NULL;
                    draw.DrawClippedLineInt = NULL;
                    draw.DrawClippedBresenhamLine = NULL;
                    }
                if (!DV.opt.bDrawStippleLine) {
                    draw.DrawStippleLineInt = NULL;
                    draw.DrawBresenhamStippleLine = NULL;
                    draw.DrawClippedStippleLineInt = NULL;
                    draw.DrawClippedBresenhamStippleLine = NULL;
                    }
                if (!DV.opt.bPutMonoImage) {
                    draw.PutMonoImageMSBSys = NULL;
                    draw.PutMonoImageMSBLin = NULL;
                    draw.PutMonoImageMSBBM = NULL;
                    draw.PutMonoImageLSBSys = NULL;
                    draw.PutMonoImageLSBLin = NULL;
                    draw.PutMonoImageLSBBM = NULL;
                    }
                if (!DV.opt.bClipMonoImage) {
                    draw.ClipMonoImageMSBSys = NULL;
                    draw.ClipMonoImageMSBLin = NULL;
                    draw.ClipMonoImageMSBBM = NULL;
                    draw.ClipMonoImageLSBSys = NULL;
                    draw.ClipMonoImageLSBLin = NULL;
                    draw.ClipMonoImageLSBBM = NULL;
                    }
                if (!DV.opt.bBitBlt) {
                    draw.BitBlt = NULL;
                    draw.BitBltLin = NULL;
                    draw.BitBltSys = NULL;
                    draw.BitBltBM = NULL;
                    }
                if (!DV.opt.bBitBltPatt) {
                    draw.BitBltPatt = NULL;
                    draw.BitBltPattLin = NULL;
                    draw.BitBltPattSys = NULL;
                    draw.BitBltPattBM = NULL;
                    }
                if (!DV.opt.bBitBltColorPatt) {
                    draw.BitBltPatt = NULL;
                    draw.BitBltPattLin = NULL;
                    draw.BitBltPattSys = NULL;
                    draw.BitBltPattBM = NULL;
                    }
                if (!DV.opt.bSrcTransBlt) {
                    draw.SrcTransBlt = NULL;
                    draw.SrcTransBltLin = NULL;
                    draw.SrcTransBltSys = NULL;
                    draw.SrcTransBltBM = NULL;
                    }
                if (!DV.opt.bDstTransBlt) {
                    draw.DstTransBlt = NULL;
                    draw.DstTransBltLin = NULL;
                    draw.DstTransBltSys = NULL;
                    draw.DstTransBltBM = NULL;
                    }
                if (!DV.opt.bStretchBlt) {
                    draw.StretchBlt = NULL;
                    draw.StretchBltLin = NULL;
                    draw.StretchBltSys = NULL;
                    draw.StretchBltBM = NULL;
                    }
                if (!DV.opt.bConvertBlt) {
                    draw.ConvertBltSys_Obsolete = NULL;
                    draw.ConvertBltBM_Obsolete = NULL;
                    }
                if (!DV.opt.bStretchConvertBlt) {
                    draw.StretchConvertBltSys_Obsolete = NULL;
                    draw.StretchConvertBltBM_Obsolete = NULL;
                    }
                if (!DV.opt.bBitBltFx) {
                    draw.BitBltFxTest = NULL;
                    draw.BitBltFx = NULL;
                    draw.BitBltFxLin = NULL;
                    draw.BitBltFxSys = NULL;
                    draw.BitBltFxBM = NULL;
                    }
                if (!DV.opt.bGetBitmap) {
                    draw.GetBitmapSys = NULL;
                    draw.GetBitmapBM = NULL;
                    }
                RETURN_VALID(draw);
                }
            break;

// TODO: When the hardware video functions are requested, we will
//       dynamically load the video device driver from disk.

        case GA_GET_VIDEOFUNCS:
            if ((DV.cntModeInfo.Attributes & gaHaveAccelVideo) && (DV.cntMode.flags & gaVideo) && DV.opt.bHardwareVideo)
                RETURN_VALID(DV.video);
            break;
        case GA_GET_VBEFUNCS:
            RETURN_VALID(DV.vbe);
        case GA_GET_BUFFERFUNCS:
            RETURN_VALID(DV.buf);
        case GA_GET_CLIPPERFUNCS:
            RETURN_VALID(DV.clip);
        }
    return false;
}

/****************************************************************************
HEADER:
graphics/driver.h

PARAMETERS:
info    - Place to copy the configuration information

REMARKS:
Returns the configuration information for the chipset driver.
****************************************************************************/
void NAPI DRV_getConfigInfo(
    GA_configInfo *info)
{
    COPY_STRUCTURE(info,&DRV_configInfo);
}

/****************************************************************************
HEADER:
graphics/driver.h

PARAMETERS:
info    - Place to copy the configuration information

REMARKS:
Returns the configuration information for the chipset driver.
****************************************************************************/
void NAPI DRV_getCertifyInfo(
    GA_certifyInfo *info)
{
    COPY_STRUCTURE(info,&DRV_certifyInfo);
}

/****************************************************************************
HEADER:
graphics/driver.h

PARAMETERS:
opt - Place to copy the options information

REMARKS:
Returns the current options information to the user application.
****************************************************************************/
void NAPI DRV_getOptions(
    GA_options *opt)
{
    // Note that we copy the primary head output device field into the
    // options structure to make it persistent, but leave the secondary
    // head values as not persisent (ie: always what is detected on
    // bootup).
    COPY_STRUCTURE(opt,&DV.opt);
    opt->outputDevice = DV.outputDevice[0];
}

/****************************************************************************
HEADER:
graphics/driver.h

PARAMETERS:
info    - Place to copy the configuration information

REMARKS:
Sets the current configuration options in the driver from the user
application.
****************************************************************************/
void NAPI DRV_setOptions(
    GA_options *opt)
{
    N_uint8 *o,*l;
    int     count;

    // Copy the set of new options
    COPY_STRUCTURE(&DV.opt,opt);
    DV.opt.dwSize = sizeof(DV.opt);
    DV.outputDevice[0] = DV.opt.outputDevice;

    // Now filter the set of options based on the licensed features
    o = &DV.opt.FIRST_OPTION;
    l = &DV.licensed.FIRST_OPTION;
    count = &DV.opt.LAST_OPTION - &DV.opt.FIRST_OPTION;
    while (count--) {
        if (*l == 0)
            *o = 0;
        l++;
        o++;
        }

    // Disable TVOut mode if it is not licensed
    if (!DV.licensed.bTVOut)
        DV.outputDevice[0] &= ~(gaOUTPUT_TV | gaOUTPUT_TVCOLORMASK);

    // Allow the chipset driver to respond to the changed options (TV out).
    if (DV.chip.setOptions)
        DV.chip.setOptions();

    // Disable triple buffering if we have a shared IRQ line
    if (DV.sharedIRQ)
        DV.opt.bTripleBuffer = false;
}

/****************************************************************************
HEADER:
graphics/driver.h

PARAMETERS:
info    - Place to copy the configuration information

REMARKS:
Sets the current configuration modeProfile in the driver from the user
application.
****************************************************************************/
void NAPI DRV_setModeProfile(
    GA_modeProfile *prof)
{
    COPY_STRUCTURE(&DV.prof,prof);
    DV.prof.dwSize = sizeof(DV.prof);
}

/****************************************************************************
REMARKS:
This function sets the software rendering callbacks for the driver.
****************************************************************************/
void NAPI DRV_SetSoftwareRenderFuncs(
    GA_2DRenderFuncs *softwareFuncs)
{
    COPY_STRUCTURE(&DV.drawSW,softwareFuncs);
    DV.drawSW.dwSize = sizeof(DV.drawSW);
}

/****************************************************************************
REMARKS:
This function sets the internal device driver pointer to the software
reference rasteriser.
****************************************************************************/
void NAPI DRV_SetRef2dPointer(
    REF2D_driver *ref2d)
{
    DV.ref2d = ref2d;
}

