/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  IBM PC (MS DOS)
*
* Description:  Module for performing simplified I/O port manipulations.
*               These routines are not particularly fast, but the serve
*               the purpose that we require since speed is not essential.
*
****************************************************************************/

#include "snap/graphics/driver.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
HEADER:
graphics/ddc.h

PARAMETERS:
index   - Indexed register to read

RETURNS:
Value read from the register

REMARKS:
This function reads the value from a double indexed RAMDAC register, such
as those on the TVP30xx and IBM RGBxxx series of RAMDACs. It assumes the
driver init code has correctly set up the DV.dacIndex and DV.dacData
variables.
****************************************************************************/
uchar DRV_rdDACIdx(
    uchar index)
{
    DV.chip.wrDACReg(DV.dacIndex,index);
    return DV.chip.rdDACReg(DV.dacData);
}

/****************************************************************************
HEADER:
graphics/ddc.h

PARAMETERS:
index   - Indexed register to write
value   - New value to write

REMARKS:
This function writes a value to a double indexed RAMDAC register, such
as those on the TVP30xx and IBM RGBxxx series of RAMDACs. It assumes the
driver init code has correctly set up the DV.dacIndex and DV.dacData
variables.
****************************************************************************/
void DRV_wrDACIdx(
    uchar index,
    uchar value)
{
    DV.chip.wrDACReg(DV.dacIndex,index);
    DV.chip.wrDACReg(DV.dacData,value);
}

uchar _rdinx(ulong port,uchar index)
{ outpb(port,index); return inpb(port+1); }

void _wrinx(ulong port,uchar index,uchar value)
{ outpb(port,index); outpb(port+1,value); }

void modinx(ulong port,uchar index,uchar mask,uchar value)
{ wrinx(port,index,(rdinx(port,index) & ~mask) | (value & mask)); }

void setinx(ulong port,uchar index,uchar mask)
{ wrinx(port,index,rdinx(port,index) | mask); }

void clrinx(ulong port,uchar index,uchar mask)
{ wrinx(port,index,rdinx(port,index) & ~mask); }

void modreg(ulong port,uchar mask,uchar value)
{ outpb(port,(inpb(port) & ~mask) | (value & mask)); }

void setreg(ulong port,uchar mask)
{ outpb(port,inpb(port) | mask); }

void clrreg(ulong port,uchar mask)
{ outpb(port,inpb(port) & ~mask); }

void modregw(ulong port,ushort mask,ushort value)
{ outpw(port,(inpw(port) & ~mask) | (value & mask)); }

void setregw(ulong port,ushort mask)
{ outpw(port,inpw(port) | mask); }

void clrregw(ulong port,ushort mask)
{ outpw(port,inpw(port) & ~mask); }

void modregd(ulong port,ulong mask,ulong value)
{ outpd(port,(inpd(port) & ~mask) | (value & mask)); }

void setregd(ulong port,ulong mask)
{ outpd(port,inpd(port) | mask); }

void clrregd(ulong port,ulong mask)
{ outpd(port,inpd(port) & ~mask); }

uchar rdinx_m(ulong reg,uchar index)
{
    MMIO_DECLARE;
    outmb(reg,index);
    return inmb(reg+1);
}

void wrinx_m(ulong reg,uchar index,uchar value)
{
    MMIO_DECLARE;
    outmb(reg,index);
    outmb(reg+1,value);
}

void modinx_m(ulong reg,uchar index,uchar mask,uchar value)
{ wrinx_m(reg,index,(rdinx_m(reg,index) & ~mask) | (value & mask)); }

void setinx_m(ulong reg,uchar index,uchar mask)
{ wrinx_m(reg,index,rdinx_m(reg,index) | mask); }

void clrinx_m(ulong reg,uchar index,uchar mask)
{ wrinx_m(reg,index,rdinx_m(reg,index) & ~mask); }

/****************************************************************************
HEADER:
graphics/ddc.h

PARAMETERS:
port    - I/O port to test
mask    - Mask of bits to test

RETURNS:
True if the I/O port bits could be modified.

REMARKS:
Checks to see if the bits specified in the mask can be modified correctly
for the specified I/O port. Returns true if successful.
****************************************************************************/
ibool tstreg(
    ulong port,
    ushort mask)
{
    ushort old,zeros,ones;

    old = inpw(port);               // Save old port value
    outpw(port,old & ~mask);        // Clear all bits specified in mask
    zeros = inpw(port) & mask;      // Read back value now set
    outpw(port,old | mask);         // Set all bits specified in mask
    ones = inpw(port) & mask;       // Read back value now set
    outpw(port,old);                // Restore original port value
    return (zeros == 0 && ones == mask);
}

/****************************************************************************
HEADER:
graphics/ddc.h

PARAMETERS:
port    - I/O port to test
index   - Register index to test
mask    - Mask of bits to test

RETURNS:
True if the I/O port bits could be modified.

REMARKS:
Checks to see if the bits specified in the mask can be modified correctly
for the specified port register. Returns true if successful.
****************************************************************************/
ibool tstinx(
    ulong port,
    ushort index,
    ushort mask)
{
    ushort old,zeros,ones;

    old = rdinx(port,index);            // Save old port value
    wrinx(port,index,old & ~mask);      // Clear all bits specified in mask
    zeros = rdinx(port,index) & mask;   // Read back value now set
    wrinx(port,index,old | mask);       // Set all bits specified in mask
    ones = rdinx(port,index) & mask;    // Read back value now set
    wrinx(port,index,old);              // Restore original port value
    return (zeros == 0 && ones == mask);
}
