/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the generic SNAP Graphics
*               Architecture API functions based on a loaded VBE/Core BIOS.
*
****************************************************************************/

#include "snap/graphics/driver.h"
#include "internal.h"
#include "snap/graphics/confver.h"
#include <clib/os/os.h>

/*--------------------------- Global variables ----------------------------*/

static GA_sharedInfo localInfo = {0,-1};
char                DRV_path[256];
VBE_state           VBE_stateBuf;           // VBE state buffer
uint                VBE_version;            // Version number of VBE driver
VBE_vgaInfo         VBE_info;
GA_mdEntry          VBE_modes[NUM_DRIVER_MODES+1];
GA_driver           DV;                     // Global device driver structure
GA_monitor          monitor[GA_MAX_HEADS];  // Global monitor record
#ifdef  __DRIVER__
N_imports _VARAPI   _N_imports;             // Imports to OS specific functions
PM_imports _VARAPI  _PM_imports;            // Imports to PM library functions
DRV_imports _VARAPI _DRV_imports;           // Imports to driver specific functions
#endif

#define CONF_VERSION    ((CONF_RELEASE_MAJOR << 8) | CONF_RELEASE_MINOR)
GA_configInfo       DRV_configInfo = {
    sizeof(GA_configInfo),
    CONFIG_SIGNATURE,
    CONFIG_SIGNATURE,
    CONFIG_SIGNATURE,
    CONFIG_SIGNATURE,
    DRIVER_VERSION,
    DRIVER_DATE,
    true,
    __DATE__,
    CONF_VERSION,
    };

/* Names of manufacturer and chip revision id's.
 *
 * DON'T CHANGE THIS UNLESS YOU CHANGE THE SNAPBUILD PROGRAM!
 */

char *DRV_manufacturerName = "VESA";

char *DRV_chipsetNames[] = {
    "VBE 1.2",
    "VBE 2.0",
    "VBE 3.0",
    NULL,
    };

#define MAX_CHIPSETS    ((sizeof(DRV_chipsetNames) / sizeof(DRV_chipsetNames[0])) - 1)

GA_certifyChipInfo  DRV_certifyChipInfo[MAX_CHIPSETS] = {
    {sizeof(GA_certifyChipInfo),"VBE 1.2",CONF_VERSION,__DATE__,gaCertified},
    {sizeof(GA_certifyChipInfo),"VBE 2.0",CONF_VERSION,__DATE__,gaCertified},
    {sizeof(GA_certifyChipInfo),"VBE 3.0",CONF_VERSION,__DATE__,gaCertified},
    };

GA_certifyInfo      DRV_certifyInfo = {
    sizeof(GA_certifyInfo),
    CERTIFY_SIGNATURE,
    __DATE__ " " __TIME__,
    MAX_CHIPSETS,
    MAX_CHIPSETS,
    &DRV_certifyChipInfo[0],
    };

//--------------------------------------------------------------------------
// This mode enumeration table is now longer used (all default modes are
// set up in the generic code now), however we do need at least text
// modes and 640x480 graphics mode for the discrete clock analysis code for
// older graphics cards. Note that we need all text modes so we can
// correctly determine the default text mode is use when the driver is
// first loaded (thankfully we probably won't be added many custom text
// modes!
//--------------------------------------------------------------------------

static GA_mode DRV_defaultModeList[] = {
    {80,25,0},
    {80,50,0},
    };

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
HEADER:
graphics/driver.h

PARAMETERS:
mfrName         - Place to store the manufacturer name
chipsetNames    - Place to store the list of chipset names

REMARKS:
Returns the chipset driver manufacturer name and list of supported chipsets.
****************************************************************************/
void _CEXPORT GA_getChipsetNames(
    char **mfr,
    char ***chipsetNames)
{
    *mfr = DRV_manufacturerName;
    *chipsetNames = DRV_chipsetNames;
}

#ifdef  __DRIVER__
/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler for non-imported N_imports.
****************************************************************************/
static void _N_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported N_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler for non-imported GA_imports.
****************************************************************************/
static void _DRV_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported DRV_imports import function called! Please re-compile!\n");
}
#endif

/****************************************************************************
HEADER:
graphics/driver.h

PARAMETERS:
path        - Path to Binary Portable DLL
deviceIndex - Index of the display device in use
info        - Pointer to the PCI configuration space info for the device
nImp        - Generic SNAP imports
drvImp      - Driver specific imports
af          - Pointer to optional VBE/AF driver

RETURNS:
Pointer to exported function list

REMARKS:
Does generic Plug and Play device detection and fills in the device resource
allocation information for the device so that the driver loader can
allocate the necessary OS resources for the device before calling the
real detection code.
****************************************************************************/
GA_devCtx * _CEXPORT GA_initDriver(
    ibool shared,
    const char *path,
    int deviceIndex,
    PCIDeviceInfo *info,
    GA_monitor *pMonitor,
    N_imports *nImp,
    PM_imports *pmImp,
    DRV_imports *drvImp,
    GA_options *licensedOptions,
    GA_globalOptions *globalOpt,
    ibool sharedIRQ,
    GA_exports *exportTable,
    GA_sharedInfo *sharedInfo,
    int loadInGUI)
{
    VBE_modeInfo    modeInfo;
    N_uint16        *vm,modes[NUM_DRIVER_MODES+1];
    GA_mdEntry      *md;
    int             i,numModes;
#ifdef  __DRIVER__
    int             max;
    ulong           *p;

    // The VBE/Core driver does not support multi-controller
    (void)exportTable;
    if (deviceIndex > 0)
        return false;

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_N_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_N_imports; i < max; i++)
        *p++ = (ulong)_N_fatalErrorHandler;
    max = sizeof(_PM_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;
    max = sizeof(_DRV_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_DRV_imports; i < max; i++)
        *p++ = (ulong)_DRV_fatalErrorHandler;

    // Now copy all our imported functions
    memcpy(&_N_imports,nImp,MIN(sizeof(_N_imports),nImp->dwSize));
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    memcpy(&_DRV_imports,drvImp,MIN(sizeof(_DRV_imports),drvImp->dwSize));
#endif
    monitor[0] = *pMonitor;
    strcpy(DRV_path,path);

    // Initialise default information
    memset(&DV,0,sizeof(DV));
    if (info)
        memcpy(&DV.PCIConfig,info,sizeof(DV.PCIConfig));
    strcpy(DC.Signature,GA_SIGNATURE);
    DC.Version = GA_VERSION;
    DC.DriverRev = (GA_RELEASE_MAJOR << 8) | GA_RELEASE_MINOR;
    DC.DeviceIndex = deviceIndex;
    strcpy(DC.OemVendorName,SCITECH_STR);
    strcpy(DC.OemCopyright,COPYRIGHT_STR);
    DV.s = &localInfo;
    if (sharedInfo)
        DV.s = sharedInfo;

    // Detect if there is a valid VESA BIOS present
    if (shared)
        VBE_sharePMCode();
    VBE_setStateBuffer(&VBE_stateBuf);
    VBE_init();
    if ((VBE_version = VBE_detect(&VBE_info)) < 0x100)
        return NULL;

    // Copy list of available video modes to local memory buffer
    for (i = 0,vm = (ushort*)VBE_info.VideoModePtr; *vm != 0xFFFF; vm++) {
        if (*vm >= 0x100 && i < NUM_DRIVER_MODES)
            modes[i++] = *vm;
        }
    modes[i] = 0xFFFF;

    // Now configure the global header based on the VBE/Core driver.
    DC.TotalMemory = VBE_info.TotalMemory * 64;
    DC.AvailableModes = (void*)&DV.AvailableModes;
    DC.TextSize = 32 * 1024;
    DC.TextBasePtr = 0xB8000;
    DC.BankSize = 64 * 1024;
    DC.BankedBasePtr = 0xA0000;

    // Initialise the first few entries of the mode table with VGA modes
    md = VBE_modes;
    md->mode = 0x03;    md->xres = 80;  md->yres = 25;  md->bits = 0; md++;
    md->mode = 0x04;    md->xres = 80;  md->yres = 50;  md->bits = 0; md++;
    md->mode = 0x10;    md->xres = 640; md->yres = 350; md->bits = 4; md++;
    md->mode = 0x12;    md->xres = 640; md->yres = 480; md->bits = 4; md++;

    // Fill in the remainder of the mode table with VBE modes
    for (numModes = 0,vm = modes; *vm != 0xFFFF; vm++) {
        if (!VBE_getModeInfo(*vm,&modeInfo))
            continue;
        if (!(modeInfo.ModeAttributes & vbeMdGraphMode))
            continue;
        if (numModes >= NUM_DRIVER_MODES)
            break;
        md->mode                    = *vm;
        md->xres                    = modeInfo.XResolution;
        md->yres                    = modeInfo.YResolution;
        md->bits                    = modeInfo.BitsPerPixel;
        md->dualBanks               = ((modeInfo.WinAAttributes & 0x7) != 0x7);
        md->bankShift = 0;
        while ((64 >> md->bankShift) != modeInfo.WinGranularity)
            md->bankShift++;
        md->Attributes              = gaHaveDisplayStart;
        if (modeInfo.ModeAttributes & vbeMdNonVGA)
            md->Attributes |= gaHaveNonVGAMode;
        if (!(modeInfo.ModeAttributes & vbeMdNonBanked))
            md->Attributes |= gaHaveBankedBuffer;
        if ((modeInfo.ModeAttributes & vbeMdLinear) && modeInfo.BitsPerPixel >= 8)
            md->Attributes |= gaHaveLinearBuffer;
        if (modeInfo.ModeAttributes & vbeMdDoubleScan)
            md->Attributes |= gaHaveDoubleScan;
        if (modeInfo.ModeAttributes & vbeMdInterlaced)
            md->Attributes |= gaHaveInterlaced;
        if (modeInfo.ModeAttributes & vbeMdTripleBuf)
            md->Attributes |= gaHaveTripleBuffer;
        if ((modeInfo.ModeAttributes & vbeMdStereo) &&
            (modeInfo.ModeAttributes & vbeMdDualDisp))
            md->Attributes |= gaHaveStereo;
        if (VBE_info.Capabilities & vbe8BitDAC)
            md->Attributes |= gaHave8BitDAC;
        if (VBE_version >= 0x300 && (modeInfo.ModeAttributes & vbeMdLinear)) {
            // Linear framebuffer modes
            md->BytesPerScanLine     = modeInfo.LinBytesPerScanLine;
            md->NumberOfImagePages   = modeInfo.LinNumberOfImagePages+1;
            md->RedMaskSize          = modeInfo.LinRedMaskSize;
            md->RedFieldPosition     = modeInfo.LinRedFieldPosition;
            md->GreenMaskSize        = modeInfo.LinGreenMaskSize;
            md->GreenFieldPosition   = modeInfo.LinGreenFieldPosition;
            md->BlueMaskSize         = modeInfo.LinBlueMaskSize;
            md->BlueFieldPosition    = modeInfo.LinBlueFieldPosition;
            md->RsvdMaskSize         = modeInfo.LinRsvdMaskSize;
            md->RsvdFieldPosition    = modeInfo.LinRsvdFieldPosition;
            }
        else {
            // Banked framebuffer modes
            md->BytesPerScanLine    = modeInfo.BytesPerScanLine;
            if (VBE_version >= 0x300)
                md->NumberOfImagePages = modeInfo.BnkNumberOfImagePages+1;
            else
                md->NumberOfImagePages = modeInfo.NumberOfImagePages+1;
            md->RedMaskSize         = modeInfo.RedMaskSize;
            md->RedFieldPosition    = modeInfo.RedFieldPosition;
            md->GreenMaskSize       = modeInfo.GreenMaskSize;
            md->GreenFieldPosition  = modeInfo.GreenFieldPosition;
            md->BlueMaskSize        = modeInfo.BlueMaskSize;
            md->BlueFieldPosition   = modeInfo.BlueFieldPosition;
            md->RsvdMaskSize        = modeInfo.RsvdMaskSize;
            md->RsvdFieldPosition   = modeInfo.RsvdFieldPosition;
            }
        if ((modeInfo.ModeAttributes & vbeMdLinear) && !DC.LinearSize) {
            DC.LinearBasePtr = modeInfo.PhysBasePtr;
            DC.LinearSize = VBE_info.TotalMemory * 64 * 1024;

            // Round the linear memory size to a correct round number, otherwise
            // we will have problems with write combining.
#define ROUND_MB(b,m)   ((b) + (((m) * MB)-1)) & ~(((m) * MB)-1)
            if (DC.LinearSize > (64 * MB))
                DC.LinearSize = ROUND_MB(DC.LinearSize,128);
            else if (DC.LinearSize > (32 * MB))
                DC.LinearSize = ROUND_MB(DC.LinearSize,64);
            else if (DC.LinearSize > (16 * MB))
                DC.LinearSize = ROUND_MB(DC.LinearSize,32);
            else if (DC.LinearSize > (8 * MB))
                DC.LinearSize = ROUND_MB(DC.LinearSize,16);
            else if (DC.LinearSize > (4 * MB))
                DC.LinearSize = ROUND_MB(DC.LinearSize,8);
            else if (DC.LinearSize > (2 * MB))
                DC.LinearSize = ROUND_MB(DC.LinearSize,4);
            }
        md->MaxPixelClock = (VBE_version >= 0x300) ? modeInfo.MaxPixelClock : 0;
        DC.Attributes |= md->Attributes;
        numModes++;
        md++;
        }

    // Add 320x200 VGA mode as the last mode, so that the VBE modes above
    // will override the VGA version in the driver if available.
    md->mode = 0x13;    md->xres = 320; md->yres = 200; md->bits = 8; md++;
    md->mode = 0xFFFF;

    // If we don't have linear modes, we need IOPL for rendering so that
    // we can do bank switching for shadow buffer updates.
    if (!(DC.Attributes & gaHaveLinearBuffer))
        DC.Attributes |= gaHaveAccelIOPL;

    // If we do not have VBE 3.0 then we don't have refresh control
    if (VBE_version < 0x300)
        DC.Attributes |= gaNoRefreshCtrl;

    // Initialise list of licensed options
    if (licensedOptions)
        DV.licensed = *licensedOptions;
    else {
        rep_stosb(&DV.licensed,sizeof(DV.licensed),1);
        DV.licensed.dwSize = sizeof(DV.licensed);
        DV.licensed.resolutions[0] = 0xFFFFFFFFU;
        DV.licensed.colorDepths[0] = 0xFF;
        DV.licensed.maxRefresh = 0xFF;
        }
    DV.globalOpt = *globalOpt;

    // Fill in the driver init function pointers
    DC.loader.InitDriver = DRV_initDriver;
    DC.loader.QueryFunctions = DRV_queryFunctions;
    DC.loader.UnloadDriver = DRV_unloadDriver;
    return &DC;
}

/****************************************************************************
RETURNS:
True if the VBE/Core BIOS device supports DPMS services.
****************************************************************************/
static int DetectVBEPM(void)
{
    RMREGS  regs;
    RMSREGS sregs;

    /* Detect if the VBE/PM interface is active */
    regs.x.ax = 0x4F10;
    regs.h.bl = 0;
    regs.x.di = 0;
    sregs.es = 0;
    PM_int86x(0x10, &regs, &regs, &sregs);
    if (regs.x.ax != 0x004F)
        return 0;
    return regs.h.bh;
}

/****************************************************************************
DESCRIPTION:
Detects that the device is installed and configures the driver.

HEADER:
graphics/driver.h

RETURNS:
True if hardware detection correctly, false if not.

REMARKS:
Does generic device detection and fills in the generic components of the
SNAP device context header information after calling the chipset
specific detection code.
****************************************************************************/
N_int32 NAPI DRV_initDriver(void)
{
    N_uint16    *m;
    int         mode,numModes;
    FILE        *f;
    char        filename[256];

#define INIT_STRUCTURE(p)       \
    memset(&p,0xFF,sizeof(p));  \
    p.dwSize = sizeof(p);

    // Set all device specific vectors to NOT_IMPLEMENTED by default
    INIT_STRUCTURE(DV.init);
    INIT_STRUCTURE(DV.driver);
    INIT_STRUCTURE(DV.cursor);
    INIT_STRUCTURE(DV.video);
    INIT_STRUCTURE(DV.dpms);
    INIT_STRUCTURE(DV.sci);
    INIT_STRUCTURE(DV.vbe);
    INIT_STRUCTURE(DV.buf);
    INIT_STRUCTURE(DV.clip);
    INIT_STRUCTURE(DV.state2d);
    INIT_STRUCTURE(DV.draw_8bpp);
    INIT_STRUCTURE(DV.draw_16bpp);
    INIT_STRUCTURE(DV.draw_24bpp);
    INIT_STRUCTURE(DV.draw_32bpp);
    INIT_STRUCTURE(DV.drawSW);

    // Configure all the generic functions for the driver.
    DV.init.GetConfigInfo               = DRV_getConfigInfo;
    DV.init.GetOptions                  = DRV_getOptions;
    DV.init.SetOptions                  = DRV_setOptions;
    DV.init.SetModeProfile              = DRV_setModeProfile;
    DV.init.GetVideoModeInfo            = DRV_GetVideoModeInfo;
    DV.init.SetVideoMode                = DRV_SetVideoMode;
    DV.init.GetVideoMode                = DRV_GetVideoMode;
    DV.init.GetCustomVideoModeInfo      = DRV_GetCustomVideoModeInfo;
    DV.init.SetCustomVideoMode          = DRV_SetCustomVideoMode;
    DV.init.GetClosestPixelClockV1      = DRV_GetClosestPixelClockV1;
    DV.init.GetClosestPixelClock        = DRV_GetClosestPixelClock;
    DV.init.GetCRTCTimings              = DRV_GetCRTCTimings;
    DV.init.SetCRTCTimings              = DRV_SetCRTCTimings;
    DV.init.SaveCRTCTimings             = DRV_SaveCRTCTimings;
    DV.init.SetGlobalRefresh            = DRV_SetGlobalRefresh;
    DV.init.SaveRestoreState            = NULL;
    DV.init.SetDisplayOutput            = DRV_SetDisplayOutput;
    DV.init.GetDisplayOutput            = DRV_GetDisplayOutput;
    DV.init.SetSoftwareRenderFuncs      = DRV_SetSoftwareRenderFuncs;
    DV.init.GetUniqueFilename           = DRV_getUniqueFilename;
    DV.init.GetMonitorInfo              = DRV_getMonitorInfo;
    DV.init.SetMonitorInfo              = DRV_setMonitorInfo;
    DV.init.GetCurrentVideoModeInfo     = DRV_GetCurrentVideoModeInfo;
    DV.init.GetCurrentRefreshRate       = DRV_GetCurrentRefreshRate;
    DV.init.GetCertifyInfo              = DRV_getCertifyInfo;
    DV.init.GetVideoModeInfoExt         = DRV_GetVideoModeInfoExt;
    DV.init.GetCustomVideoModeInfoExt   = DRV_GetCustomVideoModeInfoExt;
    if (VBE_version >= 0x200)
        DV.init.SwitchPhysicalResolution    = DRV_SwitchPhysicalResolution;
    else
        DV.init.SwitchPhysicalResolution    = NULL;
    DV.driver.SetBank                   = DRV_SetBank;
    DV.driver.SetDisplayStart           = DRV_SetDisplayStart;
    DV.driver.SetDisplayStartXY         = DRV_SetDisplayStartXY;
    DV.driver.SetStereoDisplayStart     = DRV_SetStereoDisplayStart;
    DV.driver.GetDisplayStartStatus     = DRV_GetDisplayStartStatus;
    DV.driver.EnableStereoMode          = DRV_EnableStereoMode;
    DV.driver.IsVSync                   = DRV_IsVSync;
    DV.driver.WaitVSync                 = DRV_WaitVSync;
    DV.driver.SetVSyncWidth             = DRV_SetVSyncWidth;
    DV.driver.GetVSyncWidth             = DRV_GetVSyncWidth;
    DV.driver.SetPaletteData            = DRV_SetPaletteData;
    DV.driver.GetPaletteData            = DRV_GetPaletteData;
    DV.init.GetNumberOfHeads            = DRV_GetNumberOfHeads;
    DV.init.SetActiveHead               = DRV_SetActiveHead;
    DV.init.GetActiveHead               = DRV_GetActiveHead;
    DV.init.SetRef2dPointer             = DRV_SetRef2dPointer;
    DV.init.SetDisplayOutputExt         = DRV_SetDisplayOutputExt;
    DV.init.GetDisplayOutputExt         = DRV_GetDisplayOutputExt;
    DV.driver.SetGammaCorrectData       = CNOT_SUPPORTED;
    DV.driver.GetGammaCorrectData       = CNOT_SUPPORTED;
    DV.driver.GetCurrentScanLine        = LCNOT_SUPPORTED;
    DV.dpms.DPMSdetect                  = DRV_DPMSdetect;
    DV.dpms.DPMSsetState                = DRV_DPMSsetState;
    DV.vbe.SetBytesPerLine              = DRVVBE_SetBytesPerLine;
    DV.vbe.Set8BitDAC                   = DRVVBE_Set8BitDAC;
    DV.vbe.SetPaletteData               = DRVVBE_SetPaletteData;
    DV.vbe.GetPaletteData               = DRVVBE_GetPaletteData;
    DV.sci.SCIdetect                    = DRV_SCIdetect;
    if ((DV.DPMSCaps = DetectVBEPM()) != 0)
        DV.chip.setDPMS                 = DRV_setDPMS;

    // Setup the default functions for extended palette and gamma correction
    // functions, which simply do translation to the non-extended format.
    if (DV.driver.SetPaletteDataExt == CNOT_IMPLEMENTED) {
        DV.driver.SetPaletteDataExt         = DRV_SetPaletteDataExt;
        DV.driver.GetPaletteDataExt         = DRV_GetPaletteDataExt;
        }
    if (DV.driver.SetGammaCorrectData != CNOT_IMPLEMENTED && DV.driver.SetGammaCorrectDataExt == CNOT_IMPLEMENTED) {
        DV.driver.SetGammaCorrectDataExt    = DRV_SetGammaCorrectDataExt;
        DV.driver.GetGammaCorrectDataExt    = DRV_GetGammaCorrectDataExt;
        }

    // Set the configuration options to all enabled by default.
    rep_stosb(&DV.opt,sizeof(DV.opt),1);
    DV.prof.dwSize = sizeof(DV.prof);
    memcpy(DV.prof.m.modeList,DRV_defaultModeList,sizeof(DRV_defaultModeList));
    DV.prof.m.numModes = ARR_SIZE(DRV_defaultModeList);
    DV.opt.dwSize = sizeof(DV.opt);
    DV.opt.memoryClock = 0;
    DV.opt.defaultMemoryClock = 0;
    DV.opt.maxMemoryClock = 0;
    DV.outputDevice[0] = gaOUTPUT_CRT;
    DV.opt.bDebugMode = 0;
    DV.opt.bMultiDisplay = false;
    DV.opt.bPortrait = false;
    DV.opt.bFlipped = false;
    DV.opt.bInvertColors = false;
    DV.opt.RTCFrequency = 2048;
    DV.opt.RTCAdvanceTicks = 2;
    DV.opt.glassesType = gaGlassesBlueCode;
    DV.opt.ioPort = 0;
    DV.opt.ioAndMask = 0;
    DV.opt.ioLeftOrMask = 0;
    DV.opt.ioRightOrMask = 0;
    DV.opt.ioOffOrMask = 0;
    DV.opt.vSyncWidthLeft = 4;
    DV.opt.vSyncWidthRight = 3;
    DV.opt.text80x43Height = 8;
    DV.opt.text80x50Height = 8;
    DV.opt.text80x60Height = 8;
    DV.opt.text100x43Height = 8;
    DV.opt.text100x50Height = 8;
    DV.opt.text100x60Height = 8;
    DV.opt.text132x43Height = 8;
    DV.opt.text132x50Height = 8;
    DV.opt.text132x60Height = 8;
    DV.opt.tripleScanLowRes = 0;
    DV.opt.doubleScan512 = 1;
    DV.opt.bHardwareStereoSync = 0;
    DV.opt.LCDPanelWidth = -1;
    DV.opt.LCDPanelHeight = -1;

    // Now get the default chipset specific options values, and override
    // those with values read from disk.
    DRV_getUniqueFilename(filename,2);
    if ((f = fopen(filename,"rb")) != NULL) {
        fread(&DV.opt,1,sizeof(DV.opt),f);
        fclose(f);
        }
    DRV_setOptions(&DV.opt);

    // Save the HW framebuffer location and size. We need to do that
    // here because the memory window may slide around at runtime depending
    // on the display mode, but we need the original values so that we
    // can properly map values and report the 'slid' values to apps for
    // each mode in GetVideoModeInfo.
    DV.HWLinearMem      = DC.LinearMem;
    DV.HWLinearSize     = DC.LinearSize;
    DV.HWLinearBasePtr  = DC.LinearBasePtr;

    // Now construct the list of available modes
    numModes = DV.prof.m.numModes;
    mode = 0x100;
    m = DV.AvailableModes;
    while (numModes--) {
        *m++ = mode;
        mode++;
        }
    *m = 0xFFFF;

    // Construct the detected configuration info block
    DRV_configInfo.dwSize = sizeof(DRV_configInfo);
    sprintf(filename,"VESA VBE %d.%d",VBE_version >> 8,VBE_version & 0xFF);
    strcpy(DRV_configInfo.ManufacturerName,filename);
    if (VBE_version >= 0x200 && VBE_info.OemProductNamePtr[0] != '\0')
        strcpy(DRV_configInfo.ChipsetName,VBE_info.OemProductNamePtr);
    else
        strcpy(DRV_configInfo.ChipsetName,VBE_info.OemStringPtr);
    strcpy(DRV_configInfo.DACName,"Unknown");
    strcpy(DRV_configInfo.ClockName,"Unknown");
    return nOK;
}

/****************************************************************************
DESCRIPTION:
Performs any steps necessary to unload the driver from memory.

HEADER:
graphics/driver.h

REMARKS:
Performs any steps necessary to unload the driver from memory. At present
this simply includes calling any chipset specific code to unload the driver
from memory.
****************************************************************************/
void NAPI DRV_unloadDriver(void)
{
}

