/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the generic SNAP Graphics
*               Architecture API functions based on a VBE/Core BIOS.
*
****************************************************************************/

#include "snap/graphics/driver.h"
#include "internal.h"

/*--------------------------- Global variables ----------------------------*/

#define MAX_REFRESH_RATES   20

static N_int32      DRV_refreshList[MAX_REFRESH_RATES];
static GA_palette   cntPal[256];
static ibool        dualBanks;
static int          bankShift;
static ibool        nonVGAMode;
static void         *VBE_setBank20;
static void         *VBE_setCRT20;
static void         *VBE_setPal20;

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
PARAMETERS:
mode        - Mode number to find internal mode entry for

RETURNS:
Pointer to the mode entry structure for the mode, NULL if not found.

REMARKS:
This function searches the table of valid mode information for the
specific mode, and returns a pointer to the entry in the table.
****************************************************************************/
GA_mdEntry *VBE_validateMode(
    N_int32 mode)
{
    GA_mode     *e;
    GA_mdEntry  *md;

    mode &= 0xFF;
    if (mode < DV.prof.m.numModes) {
        e = &DV.prof.m.modeList[mode];
        for (md = VBE_modes; md->mode != 0xFFFF; md++) {
            if (md->xres == e->xRes &&
                md->yres == e->yRes &&
                md->bits == e->bits)
                return md;
            }
        }
    return NULL;
}

/****************************************************************************
PARAMETERS:
mode        - Mode number to find internal mode entry for

RETURNS:
Pointer to the mode entry structure for the mode, NULL if not found.

REMARKS:
This function searches the table of valid mode information for the
specific mode, and returns a pointer to the entry in the table.
****************************************************************************/
GA_mdEntry *VBE_validateModeXY(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bits)
{
    GA_mdEntry  *md;

    for (md = VBE_modes; md->mode != 0xFFFF; md++) {
        if (md->xres == xRes &&
            md->yres == yRes &&
            md->bits == bits)
            return md;
        }
    return NULL;
}

/****************************************************************************
PARAMETERS:
xRes            - Physical pixel X resolution of the display mode
yRes            - Physical line Y resolution of the display mode
virtualX        - Virtual X resolution of the display mode
virtualY        - Virtual Y resolution of the display mode
bitsPerPixel    - Color depth of the display mode
modeInfo        - Structure to fill in with the mode information

RETURNS:
0 on success, -1 on failure

REMARKS:
Obtains information about a specific SNAP display mode. We simply call
the loaded VBE/Core driver to determine the mode information, and then
translate this information into the format used by SNAP.
****************************************************************************/
N_int32 NAPI DRV_GetCustomVideoModeInfo(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 virtualX,
    N_int32 virtualY,
    N_int32 bitsPerPixel,
    GA_modeInfo *modeInfo)
{
    // Not supported via VBE/Core
    return -1;
}

/****************************************************************************
PARAMETERS:
xRes            - Physical pixel X resolution of the display mode
yRes            - Physical line Y resolution of the display mode
virtualX        - Virtual X resolution of the display mode
virtualY        - Virtual Y resolution of the display mode
bitsPerPixel    - Color depth of the display mode
modeInfo        - Structure to fill in with the mode information

RETURNS:
0 on success, -1 on failure

REMARKS:
Obtains information about a specific SNAP display mode. We simply call
the loaded VBE/Core driver to determine the mode information, and then
translate this information into the format used by SNAP.
****************************************************************************/
N_int32 NAPI DRV_GetCustomVideoModeInfoExt(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 virtualX,
    N_int32 virtualY,
    N_int32 bitsPerPixel,
    GA_modeInfo *modeInfo,
    N_int32 outputDevice,
    N_int32 outputHead)
{
    // Not supported via VBE/Core
    return -1;
}

/****************************************************************************
PARAMETERS:
mode        - SNAP display mode to get info for
modeInfo    - Structure to fill in with the mode information

RETURNS:
0 on success, -1 on failure

REMARKS:
Obtains information about a specific SNAP display mode. We simply call
the loaded VBE/Core driver to determine the mode information, and then
translate this information into the format used by SNAP.
****************************************************************************/
N_int32 NAPI DRV_GetVideoModeInfo(
    N_uint32 gaMode,
    GA_modeInfo *modeInfo)
{
    GA_mdEntry  *md;
    GA_modeInfo gaInfo;
    CDBX_record crtcRec;
    ulong       memSize;

    // Validate that the mode is actually available
    if ((md = VBE_validateMode(gaMode)) == NULL)
        return -1;

    // Handle special case VGA BIOS modes
    if (md->mode < 0x100) {
        memset(&gaInfo,0,sizeof(gaInfo));
        gaInfo.dwSize = sizeof(gaInfo);
        switch (md->mode) {
            case 0x03:
            case 0x04:
                gaInfo.Attributes           = gaHaveBankedBuffer | gaIsTextMode;
                gaInfo.XResolution          = 640;
                gaInfo.YResolution          = 400;
                gaInfo.XCharSize            = 8;
                if (md->mode == 0x3)
                    gaInfo.YCharSize        = 16;
                else
                    gaInfo.YCharSize        = 8;
                gaInfo.BytesPerScanLine     = 160;
                gaInfo.MaxBytesPerScanLine  = 160;
                gaInfo.MaxScanLineWidth     = 80;
                gaInfo.MaxScanLines         = 25;
                gaInfo.MaxLinearOffset      = 4000;
                gaInfo.BitsPerPixel         = 0;
                gaInfo.MaxBuffers           = 1;
                gaInfo.MaxPixelClock        = 0;
                gaInfo.DefaultRefreshRate   = 70;
                DRV_refreshList[0]          = 70;
                DRV_refreshList[1]          = -1;
                break;
            case 0x10:
            case 0x12:
                gaInfo.Attributes           = gaHaveBankedBuffer | gaHaveDisplayStart;
                gaInfo.XCharSize            = 8;
                gaInfo.YCharSize            = 8;
                gaInfo.BytesPerScanLine     = 80;
                gaInfo.MaxBytesPerScanLine  = 0xFF << 1;
                gaInfo.MaxScanLineWidth     = 4080;
                gaInfo.MaxScanLines         = 819;
                gaInfo.MaxLinearOffset      = 0xFFFF;
                gaInfo.BitsPerPixel         = 4;
                gaInfo.MaxPixelClock        = 0;
                if (md->mode == 0x10) {
                    gaInfo.XResolution          = 640;
                    gaInfo.YResolution          = 350;
                    gaInfo.DefaultRefreshRate   = 70;
                    DRV_refreshList[0]          = 70;
                    DRV_refreshList[1]          = -1;
                    gaInfo.MaxBuffers           = 2;
                    }
                else {
                    gaInfo.XResolution          = 640;
                    gaInfo.YResolution          = 480;
                    gaInfo.DefaultRefreshRate   = 60;
                    DRV_refreshList[0]          = 60;
                    DRV_refreshList[1]          = -1;
                    gaInfo.MaxBuffers           = 1;
                    }
                break;
            case 0x13:
                gaInfo.Attributes           = gaHaveBankedBuffer | gaHaveDisplayStart;
                gaInfo.XResolution          = 320;
                gaInfo.YResolution          = 200;
                gaInfo.XCharSize            = 8;
                gaInfo.YCharSize            = 8;
                gaInfo.BytesPerScanLine     = 320;
                gaInfo.MaxBytesPerScanLine  = 0xFF << 3;
                gaInfo.MaxScanLineWidth     = 0xFF << 3;
                gaInfo.MaxScanLines         = 204;
                gaInfo.MaxLinearOffset      = 0xFFFF;
                gaInfo.BitsPerPixel         = 8;
                gaInfo.MaxBuffers           = 1;
                gaInfo.MaxPixelClock        = 0;
                gaInfo.DefaultRefreshRate   = 70;
                DRV_refreshList[0]          = 70;
                DRV_refreshList[1]          = -1;
                break;
            }
        gaInfo.RefreshRateList      = DRV_refreshList;
        COPY_STRUCTURE(modeInfo,&gaInfo);
        return 0;
        }

    // Report the mode information from the VBE/Core BIOS
    if (VBE_version >= 0x300 && md->xres >= 640) {
        if (!_DRV_imports.CDBX_getRefreshList(DC.DeviceIndex,0,DC.Attributes,md->xres,md->yres,md->bits,DV.opt.tripleScanLowRes,DV.opt.doubleScan512,&monitor,md->MaxPixelClock,DV.licensed.maxRefresh,MAX_REFRESH_RATES,DRV_refreshList,DRV_GetClosestPixelClock))
            return -1;
        if (!_DRV_imports.CDBX_getCRTCTimings(DC.DeviceIndex,0,DC.Attributes,md->xres,md->yres,DV.opt.tripleScanLowRes,DV.opt.doubleScan512,0,&monitor,md->MaxPixelClock,DV.licensed.maxRefresh,&crtcRec))
            return -1;
        }
    else {
        crtcRec.crtc.RefreshRate = 60 * 100;
        DRV_refreshList[0] = -1;
        }

    // Validate resolutions based on licensing
    if (DV.licensed.resolutions[0] != 0xFFFFFFFFU) {
        int i,xRes,yRes,found = false;
        for (i = 0; DV.licensed.resolutions[i] != 0xFFFFFFFF; i++) {
            xRes = DV.licensed.resolutions[i] >> 16;
            yRes = DV.licensed.resolutions[i] & 0xFFFF;
            if (xRes == md->xres && yRes == md->yres) {
                found = true;
                break;
                }
            }
        if (!found)
            return -1;
        }

    // Validate color depths based on licensing
    if (DV.licensed.colorDepths[0] != 0xFF) {
        int i,found = false;
        for (i = 0; DV.licensed.colorDepths[i] != 0xFF; i++) {
            if (md->bits == DV.licensed.colorDepths[i]) {
                found = true;
                break;
                }
            }
        if (!found)
            return -1;
        }

    memset(&gaInfo,0,sizeof(gaInfo));
    gaInfo.dwSize                   = sizeof(gaInfo);
    gaInfo.Attributes               = md->Attributes;
    if (!DV.opt.bDoubleBuffer)
        gaInfo.Attributes &= ~gaHaveDisplayStart;
    if (!DV.opt.bTripleBuffer)
        gaInfo.Attributes &= ~gaHaveTripleBuffer;
    if (!DV.licensed.bStereo)
        gaInfo.Attributes &= ~gaHaveStereo;
    if (DV.opt.bHardwareStereoSync)
        gaInfo.Attributes |= gaHaveHWStereoSync;
    if (gaInfo.XResolution < 640)
        gaInfo.Attributes &= ~gaHaveInterlaced;
    gaInfo.XResolution              = md->xres;
    gaInfo.YResolution              = md->yres;
    gaInfo.XCharSize                = 8;
    gaInfo.YCharSize                = 8;
    gaInfo.BytesPerScanLine         = md->BytesPerScanLine;
    gaInfo.MaxBytesPerScanLine      = md->BytesPerScanLine * 2;
    gaInfo.MaxScanLineWidth         = md->xres * 2;
    gaInfo.MaxPixelClock            = 0;
    memSize = DC.TotalMemory;
    if (md->bits == 4 && memSize > 512)
        memSize = 512;
    if (strncmp(VBE_info.OemStringPtr,"Matrox",6) == 0) {
        // Matrox BIOS'es fuck up going past 8192Kb!
        if (memSize > 8192)
            memSize = 8192;
        }
    gaInfo.MaxScanLines             = (memSize * 1024L) / gaInfo.BytesPerScanLine;
    gaInfo.MaxBuffers               = gaInfo.MaxScanLines / md->yres;
    gaInfo.MaxLinearOffset          = gaInfo.MaxScanLines * gaInfo.BytesPerScanLine;
    gaInfo.BitsPerPixel             = md->bits;
    gaInfo.MaxPixelClock            = md->MaxPixelClock;
    gaInfo.PixelFormat.RedAdjust    = 8 - md->RedMaskSize;
    gaInfo.PixelFormat.RedMask      = (0xFF >> gaInfo.PixelFormat.RedAdjust);
    gaInfo.PixelFormat.RedPosition  = md->RedFieldPosition;
    gaInfo.PixelFormat.GreenAdjust  = 8 - md->GreenMaskSize;
    gaInfo.PixelFormat.GreenMask    = (0xFF >> gaInfo.PixelFormat.GreenAdjust);
    gaInfo.PixelFormat.GreenPosition= md->GreenFieldPosition;
    gaInfo.PixelFormat.BlueAdjust   = 8 - md->BlueMaskSize;
    gaInfo.PixelFormat.BlueMask     = (0xFF >> gaInfo.PixelFormat.BlueAdjust);
    gaInfo.PixelFormat.BluePosition = md->BlueFieldPosition;
    gaInfo.PixelFormat.AlphaAdjust  = 8 - md->RsvdMaskSize;
    gaInfo.PixelFormat.AlphaMask    = (0xFF >> gaInfo.PixelFormat.AlphaAdjust);
    gaInfo.PixelFormat.AlphaPosition= md->RsvdFieldPosition;
    if (gaInfo.PixelFormat.AlphaMask == 0) {
        /* Compatibility with the pixel formats returned by native SNAP drivers */
        gaInfo.PixelFormat.AlphaAdjust  = 0;
        gaInfo.PixelFormat.AlphaMask    = 0;
        gaInfo.PixelFormat.AlphaPosition= 0;
        }
    gaInfo.BitmapStartAlign         = 1;
    gaInfo.BitmapStridePad          = 1;
    gaInfo.RefreshRateList          = DRV_refreshList;
    gaInfo.DefaultRefreshRate       = crtcRec.crtc.RefreshRate / 100;
    gaInfo.LinearSize               = DC.LinearSize;
    gaInfo.LinearBasePtr            = DC.LinearBasePtr;
    gaInfo.PhysicalXResolution      = gaInfo.XResolution;
    gaInfo.PhysicalYResolution      = gaInfo.YResolution;
    if (!DV.opt.bDoubleBuffer)
        gaInfo.MaxBuffers           = 1;
    COPY_STRUCTURE(modeInfo,&gaInfo);
    return 0;
}

/****************************************************************************
PARAMETERS:
mode        - SNAP display mode to get info for
modeInfo    - Structure to fill in with the mode information

RETURNS:
0 on success, -1 on failure

REMARKS:
Obtains information about a specific SNAP display mode.
****************************************************************************/
N_int32 NAPI DRV_GetVideoModeInfoExt(
    N_uint32 mode,
    GA_modeInfo *modeInfo,
    N_int32 outputDevice,
    N_int32 outputHead)
{
    // We don't support panning mode etc so we simply call the normal
    // mode info function.
    return DRV_GetVideoModeInfo(mode,modeInfo);
}

/****************************************************************************
PARAMETERS:
gaFlags - SNAP CRTC info structure flags to conver

RETURNS:
Equivalent VBE/Core CRTC info flags.
****************************************************************************/
static uchar DRV_convertCRTCFlags(
    uchar gaFlags)
{
    struct {uchar vbe; uchar ga; } flags[] = {
        {gaInterlaced,  vbeInterlaced},
        {gaDoubleScan,  vbeDoubleScan},
        {gaHSyncNeg,    vbeHSyncNeg},
        {gaVSyncNeg,    vbeVSyncNeg},
        };
    int     i;
    uchar   vbeFlags = 0;

    for (i = 0; i < (sizeof(flags)/sizeof(flags[0])); i++) {
        if (gaFlags & flags[i].ga)
            vbeFlags |= flags[i].vbe;
        }
    return vbeFlags;
}

/****************************************************************************
DESCRIPTION:
Sets a custom display mode.

PARAMETERS:
xRes            - X resolution for custom display mode
yRes            - Y resolution for custom display mode
bitsPerPixel    - Color depth for custom display mode (0 = text mode)
virtualX        - Requested virtual display X resolution (-1 to use default)
virtualY        - Returns the addressable display memory Y resolution
bytesPerLine    - Returns the scanline width for the mode
maxMem          - Returns the maximum addressable display memory limit
crtc            - CRTC information block containing CRTC timings (required)

RETURNS:
0 on success, -1 on failure

REMARKS:
This function is used to initialize a custom display mode using the
specified X and Y resolution, color depth and CRTC timings.
****************************************************************************/
N_int32 NAPI DRV_SetCustomVideoMode(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bitsPerPixel,
    N_uint32 flags,
    N_int32 *virtualX,
    N_int32 *virtualY,
    N_int32 *bytesPerLine,
    N_int32 *maxMem,
    GA_CRTCInfo *crtc)
{
    return -1;  // Not supported by VBE/Core
}

/****************************************************************************
REMARKS:
Downloads the VGA font data to the text mode display memory.
****************************************************************************/
void VGA_loadFont(
    uchar *data,
    int cellHeight)
{
    int     i;
    uchar   *vgaMem = DC.BankedMem;
    uchar   save[7];

    /* Enable access to character generator RAM */
    save[0] = rdinx(0x3C4,0x00);
    save[1] = rdinx(0x3C4,0x02);
    save[2] = rdinx(0x3C4,0x04);
    save[3] = rdinx(0x3C4,0x00);
    save[4] = rdinx(0x3CE,0x04);
    save[5] = rdinx(0x3CE,0x05);
    save[6] = rdinx(0x3CE,0x06);
    wrinx(0x3C4,0x00,0x01);
    wrinx(0x3C4,0x02,0x04);
    wrinx(0x3C4,0x04,0x07);
    wrinx(0x3C4,0x00,0x03);
    wrinx(0x3CE,0x04,0x02);
    wrinx(0x3CE,0x05,0x00);
    wrinx(0x3CE,0x06,0x00);

    /* Zero character generator RAM */
    memset(vgaMem,0,32 * 256);

    /* Download the characters to RAM. Note that we need a delay to avoid
     * problems on some graphics hardware.
     */
    for (i = 0; i < 256; i++) {
        usDelay(10);
        memcpy(vgaMem+(i*32),data+(i*cellHeight),cellHeight);
        }

    /* Set the text character font height */
    modinx(0x3D4,0x9,0x1F,cellHeight-1);

    /* Restore registers */
    wrinx(0x3C4,0x00,save[0]);
    wrinx(0x3C4,0x02,save[1]);
    wrinx(0x3C4,0x04,save[2]);
    wrinx(0x3C4,0x00,save[3]);
    wrinx(0x3CE,0x04,save[4]);
    wrinx(0x3CE,0x05,save[5]);
    wrinx(0x3CE,0x06,save[6]);
}

/****************************************************************************
REMARKS:
Generic helper function to clear the display memory using either the
banked or linear framebuffer apertures depending on the display mode.
****************************************************************************/
void VBE_clearMemory(void)
{
    ulong   maxMem = DV.cntMode.yRes * DV.cntMode.bytesPerLine;

    if (DV.cntMode.flags & gaText) {
        // *Always* write in WORD's for VGA modes!
        rep_stosw(DC.TextMem,0x4000,0x0720);
        }
    else if (DV.cntMode.bits == 4) {
        int i,maxBank = DV.cntMaxMem >> 16;
        int partialSize = DV.cntMaxMem & 0xFFFF;
        int GC_INDEX = 0x3CE;

        // Setup the VGA hardware (write mode 0, read mode 0)
        outpw(GC_INDEX,0x0000);
        outpw(GC_INDEX,0x0F01);
        outpw(GC_INDEX,0x0003);
        outpw(GC_INDEX,0x0005);
        outpw(GC_INDEX,0x0F07);

        // Clear all full banks
        for (i = 0; i < maxBank; i++) {
            DV.driver.SetBank(i);
            rep_stosb((uchar*)DC.BankedMem+0x0000,0x8000,0);
            rep_stosb((uchar*)DC.BankedMem+0x8000,0x8000,0);
            }
        // Clear last partial bank
        if (partialSize) {
            DV.driver.SetBank(i);
            rep_stosb(DC.BankedMem,partialSize,0);
            }
        DV.driver.SetBank(0);

        // Now set the GC registers back to defaults!!
        outpw(GC_INDEX,0x0001);
        outpw(GC_INDEX,0x1005);
        }
    else {
        if (DV.cntModeNum & gaLinearBuffer) {
            rep_stosd(DC.LinearMem,maxMem / 4,0);
            }
        else {
            int i,maxBank = DV.cntMaxMem >> 16;
            int partialSize = DV.cntMaxMem & 0xFFFF;

            // Clear all full banks
            for (i = 0; i < maxBank; i++) {
                DV.driver.SetBank(i);
                rep_stosd(DC.BankedMem,0x4000,0);
                }
            // Clear last partial bank
            if (partialSize) {
                DV.driver.SetBank(i);
                rep_stosd(DC.BankedMem,partialSize >> 2,0);
                }
            DV.driver.SetBank(0);
            }
        }
}

/****************************************************************************
DESCRIPTION:
Sets a specified display mode.

PARAMETERS:
mode            - Mode number to set, including flags
virtualX        - Requesyed virtual display X resolution (-1 to use default)
virtualY        - Returns the addressable display memory Y resolution
bytesPerLine    - Returns the scanline width for the mode
maxMem          - Returns the maximum addressable display memory limit
crtc            - CRTC information block if afRefreshCtrl specified

RETURNS:
0 on success, -1 on failure

REMARKS:
This function is used to initialize a specified display mode. We do this
by calling the VBE/Core driver to initialsie the mode, and translating as
necessary any parameters and return values.
****************************************************************************/
N_int32 NAPI DRV_SetVideoMode(
    N_uint32 gaMode,
    N_int32 *virtualX,
    N_int32 *virtualY,
    N_int32 *bytesPerLine,
    N_int32 *maxMem,
    N_int32 refreshRate,
    GA_CRTCInfo *crtc)
{
    RMREGS          regs;
    ushort          vbeMode;
    VBE_CRTCInfo    vbeCRTC;
    CDBX_record     crtcRec;
    int             maxScan;
    GA_mdEntry      *md;

    // We need to process the gaPartialModeSet flag because it's used for changing the
    // refresh rate
    if (gaMode & gaPartialModeSet) {
        if (VBE_version >= 0x300) {
            // Setting gaDontClear causes the BIOS on i815 to screw up. Because
            // some cards don't handle it at all, we just work around it in the
            // shell drivers.
            DRV_SetVideoMode((gaMode & ~gaPartialModeSet),virtualX,virtualY,bytesPerLine,maxMem,refreshRate,crtc);
            if (DV.cntMode.bits <= 8)
                DRV_SetPaletteData(cntPal,256,0,false);
            return 0;
            }
        return -1;
        }

    // Validate that the mode is actually available
    if ((md = VBE_validateMode(gaMode)) == NULL)
        return -1;

    // Get the current mode information for the mode
    DV.cntModeInfo.dwSize = sizeof(DV.cntModeInfo);
    if (DRV_GetVideoModeInfo(gaMode & gaModeMask,&DV.cntModeInfo) != 0)
        return -1;

    // Translate mode information and set the mode
    if (md->mode < 0x100) {
        switch (md->mode) {
            case 0x03:
            case 0x04:
                *virtualX = 80;
                if (md->mode == 0x04)
                    *virtualY = 50;
                else
                    *virtualY = 25;
                *bytesPerLine = 160;
                regs.x.ax = 0x03;
                DV.cntMode.bits = 0;
                DV.cntMode.flags = gaText;
                break;
            case 0x10:
                *virtualX = 640;
                *virtualY = 350;
                *bytesPerLine = 80;
                regs.x.ax = 0x10;
                DV.cntMode.bits = 4;
                DV.cntMode.flags = 0;
                break;
            case 0x12:
                *virtualX = 640;
                *virtualY = 480;
                *bytesPerLine = 80;
                regs.x.ax = 0x12;
                DV.cntMode.bits = 4;
                DV.cntMode.flags = 0;
                break;
            case 0x13:
                *virtualX = 320;
                *virtualY = 200;
                *bytesPerLine = 320;
                regs.x.ax = 0x13;
                DV.cntMode.bits = 8;
                DV.cntMode.flags = 0;
                break;
            }
        if (gaMode & gaDontClear)
            regs.x.ax |= 0x80;
        PM_int86(0x10,&regs,&regs);
        if (md->mode == 0x4) {
            // Load the VGA text mode font for text modes
            VGA_loadFont(DC.TextFont8x8,8);
            wrinx(0x3D4,0x0A,8-2);
            wrinx(0x3D4,0x0B,8-1);
            }
        nonVGAMode = false;
        dualBanks = false;
        bankShift = 0;
        DV.cntDisplayStart = 0;
        DV.cntModeInfo.BytesPerScanLine = DV.cntMode.bytesPerLine = *bytesPerLine;
        DV.cntMode.xRes = *virtualX;
        DV.cntMode.yRes = *virtualY;
        DV.cntDACWidth = 6;
        }
    else {
        vbeMode = md->mode;
        if (gaMode & gaDontClear)
            vbeMode |= vbeDontClear;
        if (gaMode & gaLinearBuffer)
            vbeMode |= vbeLinearBuffer;
        if (gaMode & gaRefreshCtrl) {
            if (VBE_version < 0x300)
                return -1;
            DV.cntCRTC = *crtc;
            vbeMode |= vbeRefreshCtrl;
            }
        else {
            if (VBE_version >= 0x300 && DV.cntModeInfo.XResolution >= 640) {
                if (!_DRV_imports.CDBX_getCRTCTimings(DC.DeviceIndex,0,DC.Attributes,DV.cntModeInfo.XResolution,DV.cntModeInfo.YResolution,DV.opt.tripleScanLowRes,DV.opt.doubleScan512,refreshRate,&monitor,DV.cntModeInfo.MaxPixelClock,DV.licensed.maxRefresh,&crtcRec))
                    return -1;
                DV.cntCRTC = crtcRec.crtc;
                vbeMode |= vbeRefreshCtrl;
                }
            }
        if (vbeMode & vbeRefreshCtrl) {
            vbeCRTC.HorizontalTotal     = DV.cntCRTC.HorizontalTotal;
            vbeCRTC.HorizontalSyncStart = DV.cntCRTC.HorizontalSyncStart;
            vbeCRTC.HorizontalSyncEnd   = DV.cntCRTC.HorizontalSyncEnd;
            vbeCRTC.VerticalTotal       = DV.cntCRTC.VerticalTotal;
            vbeCRTC.VerticalSyncStart   = DV.cntCRTC.VerticalSyncStart;
            vbeCRTC.VerticalSyncEnd     = DV.cntCRTC.VerticalSyncEnd;
            vbeCRTC.Flags               = DRV_convertCRTCFlags(DV.cntCRTC.Flags);
            vbeCRTC.PixelClock          = DV.cntCRTC.PixelClock;
            vbeCRTC.RefreshRate         = DV.cntCRTC.RefreshRate;
            }
        VBE_setVideoModeExt(vbeMode,&vbeCRTC);

        // Handle virtual scrolling modes
        DV.cntVirtualX = *virtualX;
        DV.cntVirtualY = *virtualY;
        if (*virtualX != -1 && *virtualY != -1) {
            if (!VBE_setPixelsPerLine(*virtualX,(int*)bytesPerLine,(int*)virtualX,&maxScan)) {
                return -1;
                }
            }
        else {
            *virtualX = DV.cntModeInfo.XResolution;
            *virtualY = DV.cntModeInfo.YResolution;
            *bytesPerLine = DV.cntModeInfo.BytesPerScanLine;
            }

        // Initialise the VBE 2.0 protected mode functions
        VBE_freePMCode();
        nonVGAMode = md->Attributes & gaHaveNonVGAMode;
        dualBanks = md->dualBanks;
        bankShift = md->bankShift;
        DV.cntDisplayStart = 0;
        VBE_setBank20 = VBE_setCRT20 = VBE_setPal20 = NULL;
        if (!VBE_stateBuf.MMIOSel) {
            VBE_setBank20 = (void*)VBE_getSetBank();
            VBE_setCRT20 = (void*)VBE_getSetDisplayStart();
            VBE_setPal20 = (void*)VBE_getSetPalette();
            }

        // Now fill in the current mode information for this mode
        DV.cntModeInfo.BytesPerScanLine = DV.cntMode.bytesPerLine = *bytesPerLine;
        DV.cntMode.xRes = *virtualX;
        DV.cntMode.yRes = *virtualY;
        DV.cntMode.bits = DV.cntModeInfo.BitsPerPixel;
        DV.cntMode.flags = 0;
        DV.cntDACWidth = 6;
        if (!(gaMode & ga6BitDAC) && VBE_setDACWidth(8) && VBE_getDACWidth() == 8)
            DV.cntDACWidth = 8;
        else
            VBE_setDACWidth(6);
        }
    DV.cntModeNum = gaMode;
    if (!(gaMode & gaDontClear))
        VBE_clearMemory();
    DRV_SetDisplayStart(0,1);
    return 0;
}

/****************************************************************************
REMARKS:
Returns the current display mode.
****************************************************************************/
N_uint32 NAPI DRV_GetVideoMode(void)
{ return DV.cntModeNum; }

/****************************************************************************
PARAMETERS:
mode        - Mode number to get closest clock for
pixelClock  - Requested pixel clock in units of Hz.

RETURNS:
Closest pixel clock in units of Hz.

REMARKS:
This function allows an application to determine if a particular pixel clock
is available. When this function is called it will run the requested pixel
clock through the internal PLL programming routines and return the actual
pixel clock that will be programmed into the hardware. The process of running
the PLL clock computation routines may cause the returned pixel clock to be
rounded slightly up or down from the requested value, however the driver
should implement the algorithms to attempt to find clocks that are the same
as or higher than the requested value. Note that the calling application must
also pass in the display mode number for the mode that will be using this
pixel clock to this function. The mode number is necessary so that the
driver can determine any necessary hardware limitations internally for the
display mode before looking for the closest physical pixel clock.
****************************************************************************/
N_uint32 NAPI DRV_GetClosestPixelClock(
    N_int32 outputHead,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bitsPerPixel,
    N_uint32 pixelClock)
{
    GA_mdEntry  *md;

    for (md = VBE_modes; md->mode != 0xFFFF; md++) {
        if (md->xres == xRes && md->yres == yRes && md->bits == bitsPerPixel)
            return VBE_getClosestClock(md->mode,pixelClock);
        }
    return 0;
}

/****************************************************************************
PARAMETERS:
mode        - Mode number to get closest clock for
pixelClock  - Requested pixel clock in units of Hz.

RETURNS:
Closest pixel clock in units of Hz.

REMARKS:
This function allows an application to determine if a particular pixel clock
is available. When this function is called it will run the requested pixel
clock through the internal PLL programming routines and return the actual
pixel clock that will be programmed into the hardware. The process of running
the PLL clock computation routines may cause the returned pixel clock to be
rounded slightly up or down from the requested value, however the driver
should implement the algorithms to attempt to find clocks that are the same
as or higher than the requested value. Note that the calling application must
also pass in the display mode number for the mode that will be using this
pixel clock to this function. The mode number is necessary so that the
driver can determine any necessary hardware limitations internally for the
display mode before looking for the closest physical pixel clock.
****************************************************************************/
N_uint32 NAPI DRV_GetClosestPixelClockV1(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bitsPerPixel,
    N_uint32 pixelClock)
{
    return DRV_GetClosestPixelClock(0,xRes,yRes,bitsPerPixel,pixelClock);
}

/****************************************************************************
PARAMETERS:
crtc    - Place to store the CRTC timings for the current mode

REMARKS:
This function returns the set of CRTC timings that are currently active f
for the current display mode.
****************************************************************************/
void NAPI DRV_GetCRTCTimings(
    GA_CRTCInfo *crtc)
{
    *crtc = DV.cntCRTC;
}

/****************************************************************************
PARAMETERS:
crtc    - New CRTC timings to make active

REMARKS:
This function programs the set of CRTC timings passed in to make them the
active CRTC timings for the current display mode. This function is used
by the SNAP centering and refresh control functions.
****************************************************************************/
void NAPI DRV_SetCRTCTimings(
    GA_CRTCInfo *crtc)
{
    N_int32 virtualX,virtualY,bytesPerLine,maxMem;

    virtualX = DV.cntMode.xRes;
    virtualY = DV.cntMode.yRes;
    bytesPerLine = DV.cntMode.bytesPerLine;
    if (VBE_version >= 0x300) {
        // Setting gaDontClear causes the BIOS on i815 to screw up. Hence
        // we can't use it!
        DRV_SetVideoMode(DV.cntModeNum | gaRefreshCtrl /*| gaDontClear*/,
            &virtualX,&virtualY,&bytesPerLine,&maxMem,0,crtc);
        if (DV.cntMode.bits <= 8)
            DRV_SetPaletteData(cntPal,256,0,false);

        // Save the CRTC timings as the current timings
        DV.cntCRTC = *crtc;
        }
}

/****************************************************************************
PARAMETERS:
crtc    - New CRTC timings to make active

REMARKS:
This function programs the set of CRTC timings passed in to make them the
active CRTC timings for the current display mode. This function is used
by the SNAP centering and refresh control functions.
****************************************************************************/
void NAPI DRV_SaveCRTCTimings(
    GA_CRTCInfo *crtc)
{
    // Save the CRTC timings as the current timings in the database
    if (DV.cntMode.flags & gaText)
        _DRV_imports.CDBX_setCRTCTimings(DC.DeviceIndex,0,DV.cntMode.xRes * DV.cntModeInfo.XCharSize,DV.cntMode.yRes * DV.cntModeInfo.YCharSize,crtc,false,true);
    else
        _DRV_imports.CDBX_setCRTCTimings(DC.DeviceIndex,0,DV.cntMode.xRes,DV.cntMode.yRes,crtc,false,true);

    // Now program the timings into the hardware
    DRV_SetCRTCTimings(crtc);
}

/****************************************************************************
PARAMETERS:
refresh - New global refresh rate to make active

REMARKS:
This function sets the global refresh rate for all display modes to the
new global refresh rate setting. This does not change the physical refresh
rate for the current display mode, but only changes the database that will
be used by the next mode set operation (assuming the default refresh
rate is selected).
****************************************************************************/
void NAPI DRV_SetGlobalRefresh(
    N_int32 refresh,
    N_int32 outputHead)
{
    // Make sure the output head parameter is in range
    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    if (outputHead < 0 || outputHead >= DV.init.GetNumberOfHeads())
        return;
    _DRV_imports.CDBX_setGlobalRefresh(DC.DeviceIndex,outputHead,refresh);
}

/****************************************************************************
REMARKS:
Inline assembler function to call the VBE 2.0 bank functions
****************************************************************************/
void NAPI CallPMSetBank(void *ptr,int bank,int window);
#ifdef  __WATCOMC__
#pragma aux CallPMSetBank =         \
    "call   eax"                    \
    parm [eax] [edx] [ebx]          \
    modify exact [eax ebx ecx edx esi edi];
#endif

/****************************************************************************
REMARKS:
Set a new bank for banked framebuffer modes.
****************************************************************************/
void NAPI DRV_SetBank(
    N_int32 bank)
{
    RMREGS  regs;

    DV.cntBank = bank;
    bank <<= bankShift;
    if (dualBanks) {
        if (VBE_setBank20) {
            CallPMSetBank(VBE_setBank20,bank,0);
            CallPMSetBank(VBE_setBank20,bank,1);
            }
        else {
            regs.x.ax = 0x4F05;
            regs.h.bh = 0;
            regs.h.bl = 0;
            regs.x.dx = bank;
            PM_int86(0x10,&regs,&regs);
            regs.x.ax = 0x4F05;
            regs.h.bh = 0;
            regs.h.bl = 1;
            regs.x.dx = bank;
            PM_int86(0x10,&regs,&regs);
            }
        }
    else {
        if (VBE_setBank20) {
            CallPMSetBank(VBE_setBank20,bank,0);
            }
        else {
            regs.x.ax = 0x4F05;
            regs.h.bh = 0;
            regs.h.bl = 0;
            regs.x.dx = bank;
            PM_int86(0x10,&regs,&regs);
            }
        }
}

/****************************************************************************
REMARKS:
Inline assembler function to call the VBE 2.0 bank functions
****************************************************************************/
void NAPI CallPMDisplayStart(void *ptr,ulong offset,uchar waitFlag);
#ifdef  __WATCOMC__
#pragma aux CallPMDisplayStart =    \
    "mov    cx,dx"                  \
    "shr    edx,16"                 \
    "call   eax"                    \
    parm [eax] [edx] [bl]           \
    modify exact [eax ebx ecx edx esi edi];
#endif

/****************************************************************************
REMARKS:
Set a new bank for banked framebuffer modes.
****************************************************************************/
void NAPI DRV_SetDisplayStart(
    N_int32 offset,
    N_int32 waitVRT)
{
    RMREGS  regs;
    int     waitVERT = false;

    DV.cntDisplayStart = offset;
    if (!nonVGAMode && waitVRT) {
        // Bug fix workaround for broken 3Dfx BIOS'es
        WaitDEVGA();
        waitVERT = true;
        waitVRT = false;
        }
    if ((DV.cntModeNum & 0x1FF) < 0x100) {
        if (DV.cntMode.bits >= 8) {
            // Program the display start for all other modes and controllers
            SetLowStartAddress8(offset);
            }
        else if (DV.cntMode.bits > 0) {
            // Program the display start for all VGA modes
            SetLowStartAddress4(offset);
            }
        else {
            // Program the display start for all VGA text modes
            SetLowStartAddress0(offset);
            }
        }
    else {
        if (DV.cntMode.bits == 4)
            offset >>= 3;
        if (VBE_setCRT20) {
            if (DV.cntMode.bits != 4)
                offset >>= 2;
            CallPMDisplayStart(VBE_setCRT20,offset,waitVRT ? 0x80 : 0x00);
            }
        else {
            int startX,startY;

            startX = offset % DV.cntMode.bytesPerLine;
            startY = offset / DV.cntMode.bytesPerLine;
            switch (DV.cntMode.bits) {
                case 0:
                case 4:     startX *= 8;
                case 15:
                case 16:    startX /= 2;
                case 24:    startX /= 3;
                case 32:    startX /= 4;
                    break;
                }
            regs.x.ax = 0x4F07;
            if (VBE_version >= 0x200)
                regs.x.bx = waitVRT ? 0x80 : 0x00;
            else
                regs.x.bx = 0;
            regs.x.cx = startX;
            regs.x.dx = startY;
            PM_int86(0x10,&regs,&regs);
            }
        }
    if (!nonVGAMode && waitVERT) {
        // Bug fix workaround for broken 3Dfx BIOS'es
        WaitVERTVGA();
        }
}

/****************************************************************************
REMARKS:
Set the display start address to a new x,y location
****************************************************************************/
void NAPI DRV_SetDisplayStartXY(
    N_int32 x,
    N_int32 y,
    N_int32 waitVRT)
{
    RMREGS  regs;

    // Bug fix workaround for broken 3Dfx BIOS'es
    if (!nonVGAMode && waitVRT) {
        WaitDEVGA();
        WaitVERTVGA();
        waitVRT = false;
        }
    regs.x.ax = 0x4F07;
    if (VBE_version >= 0x200)
        regs.x.bx = waitVRT ? 0x80 : 0x00;
    else
        regs.x.bx = 0;
    regs.x.cx = x;
    regs.x.dx = y;
    PM_int86(0x10,&regs,&regs);
}

/****************************************************************************
REMARKS:
Set the stereo display start address
****************************************************************************/
void NAPI DRV_SetStereoDisplayStart(
    N_int32 leftOffset,
    N_int32 rightOffset,
    N_int32 waitVRT)
{
    RMREGS  regs;

    // Bug fix workaround for broken 3Dfx BIOS'es
    if (!nonVGAMode && waitVRT) {
        WaitDEVGA();
        WaitVERTVGA();
        waitVRT = false;
        }
    regs.x.ax = 0x4F07;
    regs.x.bx = waitVRT ? 0x83 : 0x03;
    regs.e.ecx = leftOffset;
    regs.e.edx = rightOffset;
    PM_int86(0x10,&regs,&regs);
}

/****************************************************************************
REMARKS:
Return the display start status for triple buffering
****************************************************************************/
N_int32 NAPI DRV_GetDisplayStartStatus(void)
{
    RMREGS  regs;

    regs.x.ax = 0x4F07;
    regs.x.bx = 0x0004;
    PM_int86(0x10,&regs,&regs);
    return (regs.x.cx != 0);
}

/****************************************************************************
REMARKS:
Return true if the device is current in the vsync.
****************************************************************************/
N_int32 NAPI DRV_IsVSync(void)
{
    static int fakeStatus = 0;

    // For NonVGA modes we cannot implement this correctly, so we fake it
    // using a fake status that changes on each call.
    if (nonVGAMode)
        return fakeStatus ^= 1;
    return (inpb(0x3DA) & 8) >> 3;
}

/****************************************************************************
REMARKS:
Wait for the vertical sync to arrive.
****************************************************************************/
void NAPI DRV_WaitVSync(void)
{
    int waitVRT = 1;

    if (nonVGAMode) {
        // For NonVGA modes we call the display start function which will
        // correctly wait for the vertical retrace for us.
        DRV_SetDisplayStart(DV.cntDisplayStart,true);
        }
    else {
        WaitDEVGA();
        WaitVERTVGA();
        }
}

/****************************************************************************
REMARKS:
Set the vertical sync width.
****************************************************************************/
void NAPI DRV_SetVSyncWidth(
    N_int32 width)
{
    if (!nonVGAMode) {
        DV.cntCRTC.VerticalSyncEnd = DV.cntCRTC.VerticalSyncStart + width;
        modinx(0x3D4,0x11,0x0F,DV.cntCRTC.VerticalSyncEnd);
        }
}

/****************************************************************************
REMARKS:
Get the vertical sync width.
****************************************************************************/
N_int32 NAPI DRV_GetVSyncWidth(void)
{
    return DV.cntCRTC.VerticalSyncEnd - DV.cntCRTC.VerticalSyncStart;
}

/****************************************************************************
REMARKS:
Enable or disable stereo mode.
****************************************************************************/
void NAPI DRV_EnableStereoMode(
    N_int32 enable)
{
    RMREGS  regs;

    if (enable) {
        regs.x.ax = 0x4F07;
        regs.x.bx = 0x0005;
        PM_int86(0x10,&regs,&regs);
        }
    else {
        regs.x.ax = 0x4F07;
        regs.x.bx = 0x0006;
        PM_int86(0x10,&regs,&regs);
        }
}

/****************************************************************************
REMARKS:
Program the RAMDAC data register. We need a delay in here to ensure that
the writes take hold, otherwise they may go too fast for older hardware.
****************************************************************************/
static void writeDAC(
    int index,
    uchar v)
{
    usDelay(1);
    outpb(index,v);
}

/****************************************************************************
REMARKS:
Read the RAMDAC data register. We need a delay in here to ensure that
the writes take hold, otherwise they may go too fast for older hardware.
****************************************************************************/
static uchar readDAC(
    int index)
{
    usDelay(1);
    return inpb(index);
}

/****************************************************************************
REMARKS:
Inline assembler function to call the VBE 2.0 bank functions
****************************************************************************/
void NAPI CallPMSetPalette(void *ptr,GA_palette *pal,int num,int index,uchar waitVRT);
#ifdef  __WATCOMC__
#pragma aux CallPMSetPalette =          \
    "call   eax"                        \
    parm [eax] [edi] [ecx] [edx] [bl]   \
    modify exact [eax ebx ecx edx esi edi];
#endif

/****************************************************************************
REMARKS:
Program the hardware palette
****************************************************************************/
void NAPI DRV_SetPaletteData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    RMREGS      regs;
    int         i;
    GA_palette  tmp[256];

    // Save the palette values so we can return them later
    memcpy(&cntPal[index],pal,sizeof(cntPal[0]) * num);

    // Convert the palette values into our temporary palette
    if (DV.cntDACWidth == 8) {
        // DAC is in 8-bit mode, so program values directly
        for (i = 0; i < num; i++) {
            tmp[i].Red = pal->Red;
            tmp[i].Green = pal->Green;
            tmp[i].Blue = pal->Blue;
            pal++;
            }
        }
    else {
        // DAC is in 6-bit mode, so convert to 6-bit values on the fly
        for (i = 0; i < num; i++) {
            tmp[i].Red = pal->Red >> 2;
            tmp[i].Green = pal->Green >> 2;
            tmp[i].Blue = pal->Blue >> 2;
            pal++;
            }
        }

    // Bug fix workaround for broken 3Dfx BIOS'es
    if (!nonVGAMode && waitVRT) {
        WaitDEVGA();
        WaitVERTVGA();
        waitVRT = false;
        }

    // Now program the palette
    pal = tmp;
    if (!nonVGAMode) {
        int pelData = 0x3C9;

        // Program directly for VGA compatible modes
        if (waitVRT) {
            WaitDEVGA();
            WaitVERTVGA();
            }
        while (num--) {
            writeDAC(0x3C8,index++);
            writeDAC(pelData,pal->Red);
            writeDAC(pelData,pal->Green);
            writeDAC(pelData,pal->Blue);
            pal++;
            }
        }
    else {
        // Call the BIOS for non-VGA modes
        if (VBE_setPal20) {
            CallPMSetPalette(VBE_setPal20,pal,num,index,waitVRT ? 0x80 : 0x00);
            }
        else {
            regs.x.ax = 0x4F09;
            regs.h.bl = waitVRT ? 0x80 : 0x00;
            regs.x.cx = num;
            regs.x.dx = index;
            VBE_callESDI(&regs, pal, sizeof(VBE_palette) * num);
            }
        }
}

/****************************************************************************
REMARKS:
Read the palette registers.
****************************************************************************/
void NAPI DRV_GetPaletteData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index)
{
    if (!nonVGAMode) {
        int pelData = 0x3C9;

        if (DV.cntDACWidth == 8) {
            // DAC is in 8-bit mode, so program values directly
            while (num--) {
                writeDAC(0x3C7,index++);
                pal->Red = readDAC(pelData);
                pal->Green = readDAC(pelData);
                pal->Blue = readDAC(pelData);
                pal++;
                }
            }
        else {
            // DAC is in 6-bit mode, so convert to 6-bit values on the fly
            while (num--) {
                writeDAC(0x3C7,index++);
                pal->Red = readDAC(pelData) << 2;
                pal->Green = readDAC(pelData) << 2;
                pal->Blue = readDAC(pelData) << 2;
                pal++;
                }
            }
        }
    else {
        // Simply return it from our cached value, since the VBE BIOS does
        // not support reading the palette.
        memcpy(pal,&cntPal[index],sizeof(cntPal[0]) * num);
        }
}

/****************************************************************************
PARAMETERS:
filename    - Place to store the unique filename

REMARKS:
This function generates a unique filename for the current chipset
configuration that is used to store POST register information and the
configuration options to disk at runtime. Because the filename is
unique, each card that is plugged into the system will have a different
set of persistent POST register state and configuration options.
****************************************************************************/
void DRV_getUniqueFilename(
    char *filename,
    int type)
{
    strcpy(filename,DRV_path);
    switch (type) {
        case 1:
            strcat(filename,"modes");
            break;
        case 2:
            strcat(filename,"options");
            break;
        }
    PM_backslash(filename);
    strcat(filename,"vbecore.00");
}

/****************************************************************************
PARAMETERS:
m    - Place to store the monitor configuration information

REMARKS:
This function returns a copy of the currently configured monitor in
the driver.
****************************************************************************/
void NAPI DRV_getMonitorInfo(
    GA_monitor *m,
    N_int32 outputHead)
{
    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    if (outputHead < DV.init.GetNumberOfHeads())
        *m = monitor[outputHead];
}

/****************************************************************************
PARAMETERS:
m    - New monitor configuration information

REMARKS:
This function sets the internal monitor configuration information in the
driver to the passed in values. This automatically changes the internal
behaviour of the driver to support the new monitor configuration.
****************************************************************************/
void NAPI DRV_setMonitorInfo(
    GA_monitor *m,
    N_int32 outputHead)
{
    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    if (outputHead < DV.init.GetNumberOfHeads())
        monitor[outputHead] = *m;
}

/****************************************************************************
REMARKS:
This function sets the current output display device for subsequent SNAP
functions.
****************************************************************************/
N_int32 NAPI DRV_SetDisplayOutput(
    N_int32 device)
{
    return gaOUTPUT_CRT;    // Not supported by VBE/Core
}

/****************************************************************************
REMARKS:
This function sets the current output display device for subsequent SNAP
functions.
****************************************************************************/
N_int32 NAPI DRV_SetDisplayOutputExt(
    N_int32 device,
    N_int32 outputHead)
{
    return gaOUTPUT_CRT;    // Not supported by VBE/Core
}

/****************************************************************************
REMARKS:
This function returns the current output device flags.
****************************************************************************/
N_int32 NAPI DRV_GetDisplayOutput(void)
{
    return gaOUTPUT_CRT;    // Not supported by VBE/Core
}

/****************************************************************************
REMARKS:
This function returns the current output device flags.
****************************************************************************/
N_int32 NAPI DRV_GetDisplayOutputExt(
    N_int32 outputHead)
{
    return gaOUTPUT_CRT;    // Not supported by VBE/Core
}

/****************************************************************************
REMARKS:
This function sets the current output display device for subsequent SNAP
functions.
****************************************************************************/
void DRV_setDPMS(
    uchar state)
{
    RMREGS  regs;

    regs.x.ax = 0x4F10;
    regs.h.bl = 0x01;
    regs.h.bh = state;
    PM_int86(0x10, &regs, &regs);
}

/****************************************************************************
REMARKS:
Returns the curent video mode information from the driver.
****************************************************************************/
void NAPI DRV_GetCurrentVideoModeInfo(
    GA_modeInfo *modeInfo)
{
    int dwSize = modeInfo->dwSize;

    // Start with the real mode information and fill in the
    // stuff that changes dynamically.
    DRV_GetVideoModeInfo(DV.cntModeNum & gaModeMask,modeInfo);
    modeInfo->Attributes = DV.cntModeInfo.Attributes;
    modeInfo->BytesPerScanLine = DV.cntModeInfo.BytesPerScanLine;
    modeInfo->LinearHeapStart = DV.cntModeInfo.LinearHeapStart;
    modeInfo->MaxBuffers = DV.cntModeInfo.MaxBuffers;

    // Fill in the physical resolution values for new apps.
    if (dwSize >= 172) {
        if (DV.cntVirtualX != -1) {
            // The app has requested a virtual mode so we change the
            // reported mode info structure to convert it into a
            // panning mode with the appropriate physical and virtual
            // resolutions.
            modeInfo->AttributesExt |= gaIsPanningMode;
            modeInfo->PhysicalXResolution = DV.cntModeInfo.XResolution;
            modeInfo->PhysicalYResolution = DV.cntModeInfo.YResolution;
            modeInfo->XResolution = DV.cntVirtualX;
            modeInfo->YResolution = DV.cntVirtualY;
            }
        else {
            modeInfo->PhysicalXResolution = DV.cntModeInfo.PhysicalXResolution;
            modeInfo->PhysicalYResolution = DV.cntModeInfo.PhysicalYResolution;
            }
        }
}

/****************************************************************************
REMARKS:
Returns the the currently active refresh rate for the display mode (in 0.01
Hz increments).
****************************************************************************/
N_int32 NAPI DRV_GetCurrentRefreshRate(void)
{
    return DV.cntCRTC.RefreshRate;
}

/****************************************************************************
REMARKS:
Set the bytes per line value for the display mode. Should not be called
by regular applications.
****************************************************************************/
ibool NAPI DRVVBE_SetBytesPerLine(
    int bytesPerLine,
    int *newBytes)
{
    if ((bytesPerLine = DV.chip.roundBytesPerLine(bytesPerLine,DV.cntMode.bits)) == -1)
        return false;
    DV.cntMode.bytesPerLine = DV.chip.setBytesPerLine(bytesPerLine);
    *newBytes = DV.cntMode.bytesPerLine;
    return true;
}

/****************************************************************************
REMARKS:
Set the RAMDAC into and out of 8-bit per primary mode.
****************************************************************************/
ibool NAPI DRVVBE_Set8BitDAC(
    ibool enable)
{
    int width = enable ? 8 : 6;

    if (VBE_setDACWidth(width) && VBE_getDACWidth() == width) {
        DV.cntDACWidth = width;
        return true;
        }
    VBE_setDACWidth(DV.cntDACWidth = 6);
    return true;
}

/****************************************************************************
REMARKS:
Programs the palette registers. This version forces the internal palette
width flag to 8bit mode, so that the hardware palette functions do not
do any conversion on the input data. This is how the VBE/Core interface
works (the apps do all translation).
****************************************************************************/
void NAPI DRVVBE_SetPaletteData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    int oldWidth = DV.cntDACWidth;

    DV.cntDACWidth = 8;
    DV.driver.SetPaletteData(pal,num,index,waitVRT);
    DV.cntDACWidth = oldWidth;
}

/****************************************************************************
REMARKS:
Read the palette registers.
****************************************************************************/
void NAPI DRVVBE_GetPaletteData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index)
{
    int oldWidth = DV.cntDACWidth;

    DV.cntDACWidth = 8;
    DV.driver.GetPaletteData(pal,num,index);
    DV.cntDACWidth = oldWidth;
}

/****************************************************************************
REMARKS:
Default multi-head functions for single head cards.
****************************************************************************/
N_int32 NAPI DRV_GetNumberOfHeads(void)
{
    return 1;
}

/****************************************************************************
REMARKS:
Default multi-head functions for single head cards.
****************************************************************************/
N_int32 NAPI DRV_SetActiveHead(
    N_int32 headIndex)
{
    (void)headIndex;
    return 0;
}

/****************************************************************************
REMARKS:
Default multi-head functions for single head cards.
****************************************************************************/
N_int32 NAPI DRV_GetActiveHead(void)
{
    return 0;
}

/****************************************************************************
REMARKS:
Programs the palette registers. This version forces the internal palette
width flag to 8bit mode, so that the hardware palette functions do not
do any conversion on the input data. This is how the VBE/Core interface
works (the apps do all translation).
****************************************************************************/
void NAPI DRV_SetPaletteDataExt(
    GA_paletteExt *palExt,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    int         i;
    GA_palette  pal[256];

    for (i = 0; i < num; i++) {
        pal[i].Red = palExt[i].Red >> 8;
        pal[i].Green = palExt[i].Green >> 8;
        pal[i].Blue = palExt[i].Blue >> 8;
        }
    DV.driver.SetPaletteData(pal,num,index,waitVRT);
}

/****************************************************************************
REMARKS:
Read the palette registers.
****************************************************************************/
void NAPI DRV_GetPaletteDataExt(
    GA_paletteExt *palExt,
    N_int32 num,
    N_int32 index)
{
    int         i;
    GA_palette  pal[256];

    DV.driver.GetPaletteData(pal,num,index);
    for (i = 0; i < num; i++) {
        palExt[i].Red = pal[i].Red << 8;
        palExt[i].Green = pal[i].Green << 8;
        palExt[i].Blue = pal[i].Blue << 8;
        }
}

/****************************************************************************
REMARKS:
Programs the palette registers. This version forces the internal palette
width flag to 8bit mode, so that the hardware palette functions do not
do any conversion on the input data. This is how the VBE/Core interface
works (the apps do all translation).
****************************************************************************/
void NAPI DRV_SetGammaCorrectDataExt(
    GA_paletteExt *palExt,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    int         i;
    GA_palette  pal[256];

    for (i = 0; i < num; i++) {
        pal[i].Red = palExt[i].Red >> 8;
        pal[i].Green = palExt[i].Green >> 8;
        pal[i].Blue = palExt[i].Blue >> 8;
        }
    DV.driver.SetGammaCorrectData(pal,num,index,waitVRT);
}

/****************************************************************************
REMARKS:
Read the palette registers.
****************************************************************************/
void NAPI DRV_GetGammaCorrectDataExt(
    GA_paletteExt *palExt,
    N_int32 num,
    N_int32 index)
{
    int         i;
    GA_palette  pal[256];

    DV.driver.GetGammaCorrectData(pal,num,index);
    for (i = 0; i < num; i++) {
        palExt[i].Red = pal[i].Red << 8;
        palExt[i].Green = pal[i].Green << 8;
        palExt[i].Blue = pal[i].Blue << 8;
        }
}

/****************************************************************************
REMARKS:
This function will reset the current hardware display mode in a
non-destructive manner such that the screen and current hardware
acceleration state are preserved. This function is used to dynamically
switch the active display device and re-program the hardware as necessary
for the new active display device (ie: LCD/CRT/TV switching etc).
****************************************************************************/
N_int32 NAPI DRV_SwitchPhysicalResolution(
    N_int32 physicalXResolution,
    N_int32 physicalYResolution,
    N_int32 refreshRate)
{
    GA_mdEntry      *md;
    CDBX_record     crtcRec;
    ushort          vbeMode;
    VBE_CRTCInfo    vbeCRTC;
    int             xRes,yRes,newBytes,newPixels,maxScanlines;

    // Override the physical resolution with passed in values.
    if (physicalXResolution == -1 || physicalYResolution == -1
        || physicalXResolution > DV.cntModeInfo.XResolution
        || physicalYResolution > DV.cntModeInfo.YResolution) {
        DV.cntMode.xRes = DV.cntModeInfo.PhysicalXResolution = DV.cntModeInfo.XResolution;
        DV.cntMode.yRes = DV.cntModeInfo.PhysicalYResolution = DV.cntModeInfo.YResolution;
        }
    else {
        DV.cntMode.xRes = DV.cntModeInfo.PhysicalXResolution = physicalXResolution;
        DV.cntMode.yRes = DV.cntModeInfo.PhysicalYResolution = physicalYResolution;
        }

    // Find the new mode flags for the mode
    if (DV.cntMode.bits == 0) {
        xRes = DV.cntMode.xRes / DV.cntModeInfo.XCharSize;
        yRes = DV.cntMode.yRes / DV.cntModeInfo.YCharSize;
        }
    else {
        xRes = DV.cntMode.xRes;
        yRes = DV.cntMode.yRes;
        }
    if ((md = VBE_validateModeXY(xRes,yRes,DV.cntMode.bits)) == NULL)
        return -1;

    // Now set the new physical resolution via the BIOS
    vbeMode = md->mode | vbeDontClear;
    if (DV.cntModeNum & gaLinearBuffer)
        vbeMode |= vbeLinearBuffer;
    if (VBE_version >= 0x300) {
        if (!_DRV_imports.CDBX_getCRTCTimings(DC.DeviceIndex,0,DC.Attributes,DV.cntModeInfo.PhysicalXResolution,DV.cntModeInfo.PhysicalYResolution,DV.opt.tripleScanLowRes,DV.opt.doubleScan512,refreshRate,&monitor,DV.cntModeInfo.MaxPixelClock,DV.licensed.maxRefresh,&crtcRec))
            return -1;
        vbeMode |= vbeRefreshCtrl;
        DV.cntCRTC = crtcRec.crtc;
        vbeCRTC.HorizontalTotal     = DV.cntCRTC.HorizontalTotal;
        vbeCRTC.HorizontalSyncStart = DV.cntCRTC.HorizontalSyncStart;
        vbeCRTC.HorizontalSyncEnd   = DV.cntCRTC.HorizontalSyncEnd;
        vbeCRTC.VerticalTotal       = DV.cntCRTC.VerticalTotal;
        vbeCRTC.VerticalSyncStart   = DV.cntCRTC.VerticalSyncStart;
        vbeCRTC.VerticalSyncEnd     = DV.cntCRTC.VerticalSyncEnd;
        vbeCRTC.Flags               = DRV_convertCRTCFlags(DV.cntCRTC.Flags);
        vbeCRTC.PixelClock          = DV.cntCRTC.PixelClock;
        vbeCRTC.RefreshRate         = DV.cntCRTC.RefreshRate;
        }
    VBE_setVideoModeExt(vbeMode,&vbeCRTC);
    if (VBE_version >= 0x200)
        VBE_setBytesPerLine(DV.cntMode.bytesPerLine,&newBytes,&newPixels,&maxScanlines);
    else {
        VBE_setPixelsPerLine(DV.cntMode.bytesPerLine / (DV.cntMode.bits / 8),&newBytes,&newPixels,&maxScanlines);
        }
    VBE_setDACWidth(DV.cntDACWidth);
    DV.driver.SetBank(DV.cntBank);
    DV.driver.SetDisplayStart(DV.cntDisplayStart,false);

    // Restore the hardware palette/gamma ramp
    if (DV.cntMode.bits <= 8)
        DV.driver.SetPaletteData(cntPal,256,0,false);

    // Return success
    return 0;
}

