/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Hardware specific device detection code.
*
****************************************************************************/

#include "internal.h"

/*--------------------------- Global variables ----------------------------*/

/* Names of chipset analysed data files on disk */

char *DRV_chipsetFilenames[] = {
    "vga",
    NULL,
    };

/* Names of RAMDAC id's. */

char *DRV_DACNames[] = {
    "VGA Internal 8 bit DAC",
    };

/* Names of chip revision id's. */

char *DRV_clockNames[] = {
    "VGA Discrete Clocks",
    };

#include "crtc.c"
#include "../../vgadata.c"
#include "data/vga.c"

GA_chipEntry *DRV_chipDataLCD[] = {NULL};
GA_chipEntry *DRV_chipDataPAL[] = {NULL};
GA_chipEntry *DRV_chipDataNTSC[]= {NULL};
GA_chipEntry *DRV_chipData[] = {
    IBMVGA_chipData,
    NULL,
    };

static ibool hasBeenPOSTed;

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Wakes up the graphics controller if running as a secondary controller.
If the BIOS has POST'ed the card (which happens under Windows 98 with
multi controllers, and under Win95 and Linux with our multi-controller
support), this code is skipped. Otherwise we need to wake up the chip
and progrm the BIOS default values into a number of registers as defined
in TRIO_POSTchipsetProfile. The data that we need to program is loaded
dynamically from disk
****************************************************************************/
ibool CHIP_wakeup(int loadInGUI)
{
    if (!loadInGUI && VGA_NOT_ACTIVE()) {
        // Must be running as secondary without the chip having already
        // been woken up. So wake up the chip and remember that we need
        // to POST it later on.
        outpb(0x3C3,0x01);      // Enable VGA
        outpb(0x3C2,0x01);      // Decode 0x3D4 and memory
        outpb(0x3C6,0xFF);      // Unmask DAC
        hasBeenPOSTed = false;
        }
    else {
        // Card has alread been POSTed so we can skip writing the POST
        // register values later on.
        hasBeenPOSTed = true;
        }
    return true;
}

/****************************************************************************
REMARKS:
Configure the global driver device context with the appropriate information
for the detected hardware.
****************************************************************************/
N_int32 DRV_configure(int loadInGUI)
{
    //------------------------------------------------------------------
    // Detect the amount of memory on board
    //------------------------------------------------------------------
    DC.TotalMemory = 256;

    //------------------------------------------------------------------
    // Configure the capabilities information
    //------------------------------------------------------------------
    DC.Attributes = gaHaveDisplayStart |
                    gaHaveBankedBuffer |
                    gaHaveDoubleScan |
                    gaHaveTripleScan |
                    gaHaveTripleBuffer |
                    gaHaveAccelIOPL;

    //------------------------------------------------------------------
    // No offscreen memory.
    //------------------------------------------------------------------
    DV.maxAccelLinearMem = DV.NVMemory = DV.VGAMemory = DC.TotalMemory;
    DV.maxAccelY = -1;

    //------------------------------------------------------------------
    // Configure the maximum scanline widths for each color depth
    //------------------------------------------------------------------
    DV.maxBytesPerLineTxt   = VGA_MAX_OFFSET << 1;
    DV.maxBytesPerLine4     = VGA_MAX_OFFSET << 1;
    DV.maxBytesPerLine8     = VGA_MAX_OFFSET << 3;

    //------------------------------------------------------------------
    // Configure VGA blank/sync timing masks and CRTC register tables
    //------------------------------------------------------------------
    DV.HBlankEndMask        = 0x3F;
    DV.HSyncEndMask         = 0x1F;
    DV.VBlankEndMask        = 0xFF;
    DV.VSyncEndMask         = 0x0F;
    DV.HCRTCRegsDesc        = VGA_HCRTCRegsDesc;
    DV.VCRTCRegsDesc        = VGA_VCRTCRegsDesc;

    //------------------------------------------------------------------
    // Configure the chipset specific support functions
    //------------------------------------------------------------------
    DV.chip.getModeFlags        = VGA_getModeFlags;
    DV.chip.setPixelClock       = VGA_setPixelClock;
    DV.chip.get2DAccelInfo      = VGA_get2DAccelInfo;
    DV.chip.getVideoAccelInfo   = VGA_getVideoAccelInfo;
    DV.chip.get3DAccelInfo      = VGA_get3DAccelInfo;

    //------------------------------------------------------------------
    // Configure the chipset specific API functions
    //------------------------------------------------------------------
    DV.driver.SetBank               = VGA_SetBank;
    DV.driver.SetDisplayStart       = VGA_SetDisplayStart;
    DV.driver.SetStereoDisplayStart = NOT_SUPPORTED;
    DV.driver.GetDisplayStartStatus = NOT_SUPPORTED;
    DV.driver.EnableStereoMode      = NOT_SUPPORTED;
    DV.driver.IsVSync               = VGA_IsVSync;
    DV.driver.WaitVSync             = VGA_WaitVSync;
    DV.driver.SetVSyncWidth         = VGA_SetVSyncWidth;
    DV.driver.GetVSyncWidth         = VGA_GetVSyncWidth;
    DV.driver.SetPaletteData        = VGA_SetPaletteData;
    DV.driver.GetPaletteData        = VGA_GetPaletteData;
    DV.driver.SetGammaCorrectData   = NOT_SUPPORTED;
    DV.driver.GetGammaCorrectData   = NOT_SUPPORTED;
    DV.driver.GetCurrentScanLine    = NOT_SUPPORTED;

    //------------------------------------------------------------------
    // Configure the clock, RAMDAC, 2D and 3D modules
    //------------------------------------------------------------------
    DV.dacID = 0;
    DV.clockID = 0;
    ACCEL2D_init();
    CLOCK_init();
    RAMDAC_init();
    return nOK;
}

/****************************************************************************
REMARKS:
Returns the mode flags for the specified display mode, which includes flags
such as whether CRTC scaling is enabled etc.
****************************************************************************/
ulong VGA_getModeFlags(
    GA_mode *mode)
{
    ulong   flags;

    flags = gaVGA | gaBankedOnly;
    if (mode->xRes <= 400)
        flags |= gaCRTC_Scale20;
    if (mode->yRes <= 300 || (mode->xRes == 512 && mode->yRes == 384))
        flags |= gaIsDoubleScan;
    return flags;
}

/****************************************************************************
REMARKS:
Configure the 2D acceleration specific information in the mode info block.
****************************************************************************/
void VGA_get2DAccelInfo(
    GA_modeInfo *gaInfo,
    GA_modeEntry *entry,
    N_int32 virtualX,
    N_int32 virtualY)
{
}

/****************************************************************************
REMARKS:
Configure the 2D video overlay specific information in the mode info block.
****************************************************************************/
void VGA_getVideoAccelInfo(
    GA_modeInfo *gaInfo,
    GA_modeEntry *entry)
{
}

/****************************************************************************
REMARKS:
Configure the 3D acceleration specific information in the mode info block.
****************************************************************************/
void VGA_get3DAccelInfo(
    GA_modeInfo *gaInfo,
    GA_modeEntry *entry)
{
}
