/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  C module for the Graphics Accelerator Driver API. Uses
*               the SciTech PM library for interfacing with DOS
*               extender specific functions.
*
****************************************************************************/

#include "loader.h"
#include "gtf.h"

/*------------------------- Global Variables ------------------------------*/

// The following structure defines all the variables maintained for every
// display device in the system, with one block allocated for every device.
// {secret}
typedef struct {
    PE_MODULE       *hMod;
    int             hModSize;
    GA_devCtx       *dc;
    PCIDeviceInfo   *PCIInfo;
    int             refCount;
    void            *IOMemMaps[4];
    void            *TextMem;
    void            *BankedMem;
    void            *LinearMem;
    void            *ZBufferMem;
    void            *TexBufferMem;
    char            internalName[13];
    REF2D_driver    *ref2d;
    PE_MODULE       *hModRef2d;
    REF2D_driver    *shadowBuf;
    PE_MODULE       *hModShadow;
    REF2D_driver    *ref2dFilters[GA_MAX_FILTERS];
    PE_MODULE       *hModRef2dFilters[GA_MAX_FILTERS];
    N_int32         numRef2dFilters;
    PE_MODULE       *hModFilters[GA_MAX_FILTERS];
    N_int32         numFilters;
    PE_MODULE       *hModGL;
    GA_glCtx        *glCtx;
    GA_sharedInfo   localInfo;
    GA_sharedInfo   *sharedInfo;
    } GA_deviceEntry;

static GA_deviceEntry       *ga;
static int                  NumDevices = -1;
static int                  cntDevice = 0;
static GA_exports           *exportTable = NULL;
ibool                       BE_loaded = false;
char                        SNAPDriverPath[PM_MAX_PATH] = {"\0"};
char                        SNAPConfigPath[PM_MAX_PATH] = {"\0"};
char                        SNAPLogName[] = "graphics.log";
int                         loadInGUI = 0;
int                         allowFallbackDrivers = true;
uint                        minDriverVersion = 0;
static char                 fakePCIID[13];
static GA_globalOptions     globalOpt = {0};
static GA_options           *licensedOptions = NULL;
static ibool                bDialAMode_Licensed = true;
static ibool                bEvalExpired = false;
GA_imports                  _VARAPI _GA_imports;
#if defined(TEST_HARNESS) && defined(DEBUG_X86EMU)
static FILE                 *logfile = NULL;
#endif

#ifdef  __DRIVER__
PM_imports _VARAPI _PM_imports;
#endif
PM_imports _VARAPI _PM_importsOrg = {0};

#ifdef  __DRIVER__
N_imports _VARAPI _N_imports;
#else
N_imports _VARAPI _N_imports = {
    sizeof(N_imports),
    _OS_delay,
    };
#endif

/* Because of stack space limitations for certain OS'es, we put large
 * temporary structures into the global variable space. Since the SNAP
 * device drivers are not re-entrant anyway, this is not a problem (all
 * re-entrancy issues need to be dealt with at the OS shell driver level).
 */

static GA_initFuncs     gInit;
static GA_configInfo    gInfo;
static GA_options       gOpt;
static GA_modeInfo      gMi;
static GA_modeProfile   gModes;
static GA_monitor       gMonitor[GA_MAX_HEADS];
static GA_monitor       gDDCMonitor[GA_MAX_HEADS];
static GA_certifyInfo   gCert;
static EDID_record      edidRec;
static ibool            validEDID = false;
static uchar            edid[128];

/* {secret} */
void _ASMAPI _OS_msDelay(N_uint32 milliSeconds);

DRV_imports _VARAPI _DRV_imports = {
    sizeof(DRV_imports),
    PCI_accessReg,
    CDBX_getRefreshListV1,
    CDBX_getCRTCTimingsV2,
    CDBX_setGlobalRefreshV2,
    CDBX_setCRTCTimingsV1,
    NULL,
    _OS_delay,
    _OS_msDelay,
    _PCI_mapBIOSImage,
    _PCI_unmapBIOSImage,
    _PCI_postController,
    N_log,
    CDBX_getRefreshListV2,
    _PCI_getPCIList,
    CDBX_setCRTCTimingsV2,
    _DDC_quickDetect,
    _DDC_getFlatPanelResolution,
    CDBX_getRefreshList,
    CDBX_getCRTCTimings,
    CDBX_setGlobalRefresh,
    CDBX_setCRTCTimings,
    };

GL3D_imports _VARAPI _GL3D_imports = {
    sizeof(GL3D_imports),
    N_log,
    _OS_delay,
    _OS_msDelay,
    _PCI_getPCIList,
    GA_loadLibrary,
    GA_getProcAddress,
    GA_freeLibrary,
    };

//--------------------------------------------------------------------------
// Main default mode enumeration table for the default list of display modes
//--------------------------------------------------------------------------

GA_modeProfile GA_defaultModeProfile = {
    /*-----------------------------------------------------------------------
     * Header information (numModes is filled in later)
     *---------------------------------------------------------------------*/
    sizeof(GA_modeProfile),
    {0,
   {

    /*-----------------------------------------------------------------------
     * Text modes
     *---------------------------------------------------------------------*/
#ifdef __INTEL__
    {80,25,0},
    {80,43,0},
    {80,50,0},
    {80,60,0},
    {100,25,0},
    {100,43,0},
    {100,50,0},
    {100,60,0},
    {132,25,0},
    {132,43,0},
    {132,50,0},
    {132,60,0},

    /*-----------------------------------------------------------------------
     * 1:1 aspect ratio 4bpp modes
     *---------------------------------------------------------------------*/
    {640,350,4},
    {640,400,4},
    {640,480,4},
    {720,540,4},
    {800,600,4},
    {960,720,4},
    {1024,768,4},
    {1152,864,4},
    {1280,960,4},
    {1280,1024,4},
    {1600,1200,4},
#endif

    /*-----------------------------------------------------------------------
     * 1:1 aspect ratio 8bpp modes
     *---------------------------------------------------------------------*/
    {320,200,8},
    {320,240,8},
    {320,400,8},
    {320,480,8},
    {400,300,8},
    {400,600,8},
    {512,384,8},
    {512,480,8},
    {640,350,8},
    {640,400,8},
    {640,480,8},
    {720,540,8},
    {800,600,8},
    {960,720,8},
    {1024,768,8},
    {1152,864,8},
    {1280,960,8},
    {1280,1024,8},
    {1400,1050,8},
    {1600,1200,8},
    {1600,1280,8},
    {1800,1350,8},
    {1920,1440,8},
    {2048,1536,8},

    /*-----------------------------------------------------------------------
     * 16:9 aspect ratio 8bpp modes (HDTV)
     *---------------------------------------------------------------------*/
    {848,480,8},
    {1072,600,8},
    {1280,720,8},
    {1360,768,8},
    {1600,900,8},
    {1800,1012,8},
    {1920,1080,8},
    {2048,1152,8},

    /*-----------------------------------------------------------------------
     * 1:1 aspect ratio 15bpp modes
     *---------------------------------------------------------------------*/
    {320,200,15},
    {320,240,15},
    {320,400,15},
    {320,480,15},
    {400,300,15},
    {400,600,15},
    {512,384,15},
    {512,480,15},
    {640,350,15},
    {640,400,15},
    {640,480,15},
    {720,540,15},
    {800,600,15},
    {960,720,15},
    {1024,768,15},
    {1152,864,15},
    {1280,960,15},
    {1280,1024,15},
    {1400,1050,15},
    {1600,1200,15},
    {1600,1280,15},
    {1800,1350,15},
    {1920,1440,15},
    {2048,1536,15},

    /*-----------------------------------------------------------------------
     * 16:9 aspect ratio 15bpp modes (HDTV)
     *---------------------------------------------------------------------*/
    {848,480,15},
    {1072,600,15},
    {1280,720,15},
    {1360,768,15},
    {1600,900,15},
    {1800,1012,15},
    {1920,1080,15},
    {2048,1152,15},

    /*-----------------------------------------------------------------------
     * 1:1 aspect ratio 16bpp modes
     *---------------------------------------------------------------------*/
    {320,200,16},
    {320,240,16},
    {320,400,16},
    {320,480,16},
    {400,300,16},
    {400,600,16},
    {512,384,16},
    {512,480,16},
    {640,350,16},
    {640,400,16},
    {640,480,16},
    {720,540,16},
    {800,600,16},
    {960,720,16},
    {1024,768,16},
    {1152,864,16},
    {1280,960,16},
    {1280,1024,16},
    {1400,1050,16},
    {1600,1200,16},
    {1600,1280,16},
    {1800,1350,16},
    {1920,1440,16},
    {2048,1536,16},

    /*-----------------------------------------------------------------------
     * 16:9 aspect ratio 16bpp modes (HDTV)
     *---------------------------------------------------------------------*/
    {848,480,16},
    {1072,600,16},
    {1280,720,16},
    {1360,768,16},
    {1600,900,16},
    {1800,1012,16},
    {1920,1080,16},
    {2048,1152,16},

    /*-----------------------------------------------------------------------
     * 1:1 aspect ratio 24bpp modes
     *---------------------------------------------------------------------*/
    {320,200,24},
    {320,240,24},
    {320,400,24},
    {320,480,24},
    {400,300,24},
    {400,600,24},
    {512,384,24},
    {512,480,24},
    {640,350,24},
    {640,400,24},
    {640,480,24},
    {720,540,24},
    {800,600,24},
    {960,720,24},
    {1024,768,24},
    {1152,864,24},
    {1280,960,24},
    {1280,1024,24},
    {1400,1050,24},
    {1600,1200,24},
    {1600,1280,24},
    {1800,1350,24},
    {1920,1440,24},
    {2048,1536,24},

    /*-----------------------------------------------------------------------
     * 16:9 aspect ratio 24bpp modes (HDTV)
     *---------------------------------------------------------------------*/
    {848,480,24},
    {1072,600,24},
    {1280,720,24},
    {1360,768,24},
    {1600,900,24},
    {1800,1012,24},
    {1920,1080,24},
    {2048,1152,24},

    /*-----------------------------------------------------------------------
     * 1:1 aspect ratio 32bpp modes
     *---------------------------------------------------------------------*/
    {320,200,32},
    {320,240,32},
    {320,400,32},
    {320,480,32},
    {400,300,32},
    {400,600,32},
    {512,384,32},
    {512,480,32},
    {640,350,32},
    {640,400,32},
    {640,480,32},
    {720,540,32},
    {800,600,32},
    {960,720,32},
    {1024,768,32},
    {1152,864,32},
    {1280,960,32},
    {1280,1024,32},
    {1400,1050,32},
    {1600,1200,32},
    {1600,1280,32},
    {1800,1350,32},
    {1920,1440,32},
    {2048,1536,32},

    /*-----------------------------------------------------------------------
     * 16:9 aspect ratio 32bpp modes (HDTV)
     *---------------------------------------------------------------------*/
    {848,480,32},
    {1072,600,32},
    {1280,720,32},
    {1360,768,32},
    {1600,900,32},
    {1800,1012,32},
    {1920,1080,32},
    {2048,1152,32}},
    },

    /*-----------------------------------------------------------------------
     * Table of restricted modes that are reported to VBE applications.
     * Some older apps have problems with a large list of display modes
     * being reported (>100), so we restrict the list to avoid problems.
     *---------------------------------------------------------------------*/
    {0,
   {{80,    25,     0},
    {80,    60,     0},
    {132,   25,     0},
    {132,   43,     0},
    {132,   50,     0},
    {132,   60,     0},

    {800,   600,    4},
    {1024,  768,    4},
    {1280,  1024,   4},
    {1600,  1200,   4},

    {320,   200,    8},
    {320,   240,    8},
    {400,   300,    8},
    {512,   384,    8},
    {640,   400,    8},
    {640,   480,    8},
    {800,   600,    8},
    {1024,  768,    8},
    {1280,  1024,   8},
    {1600,  1200,   8},

    {320,   200,    15},
    {320,   240,    15},
    {400,   300,    15},
    {512,   384,    15},
    {640,   400,    15},
    {640,   480,    15},
    {800,   600,    15},
    {1024,  768,    15},
    {1280,  1024,   15},
    {1600,  1200,   15},

    {320,   200,    16},
    {320,   240,    16},
    {400,   300,    16},
    {512,   384,    16},
    {640,   400,    16},
    {640,   480,    16},
    {800,   600,    16},
    {1024,  768,    16},
    {1280,  1024,   16},
    {1600,  1200,   16},

    {320,   200,    24},
    {320,   240,    24},
    {400,   300,    24},
    {512,   384,    24},
    {640,   400,    24},
    {640,   480,    24},
    {800,   600,    24},
    {1024,  768,    24},
    {1280,  1024,   24},
    {1600,  1200,   24},

    {320,   200,    32},
    {320,   240,    32},
    {400,   300,    32},
    {512,   384,    32},
    {640,   400,    32},
    {640,   480,    32},
    {800,   600,    32},
    {1024,  768,    32},
    {1280,  1024,   32},
    {1600,  1200,   32}},
    }};

#include "snap/graphics/exports.h"
#include "txtfont.c"
#include "vgapal.c"

/*-------------------------- Implementation -------------------------------*/

#define OPTIONS_NAME    "options.dat"
#define REF2D_NAME      "ref2d_%ld.bpd"
#define SHADOW_NAME     "shadow.bpd"
#define PORTRAIT_NAME   "portrait.bpd"
#define FLIPPED_NAME    "flipped.bpd"
#define INVERT_NAME     "invert.bpd"
#define MULDISP_NAME    "muldisp.bpd"
#define REGION_NAME     "region.bpd"
#define DETECT_NAME     "detect.drv"
#define GL3D_NAME       "gl3d.bpd"
#define MONITOR_NAME    "mon%d%d.dat"
#define BUF_SAVE_SIZE   (64 * 1024)

/****************************************************************************
DESCRIPTION:
Registers a linkable library license with the library.

HEADER:
snap/graphics.h

PARAMETERS:
license - Pointer to the binary license to register
shared  - True if the device driver should be loaded into shared memory

RETURNS:
True on success, false on failure.

REMARKS:
This function is used to register a linkable library license with the
graphics library. This is provided for software vendors who have purchased
a linkable library license to use and distribute the graphics device support
library with their custom application software. Simply pass in a pointer
to the binary license structure that was provided to you by a SciTech
Software sales representative.
****************************************************************************/
int NAPI GA_registerLicense(
    uchar *license,
    N_int32 shared)
{
    (void)license;
    (void)shared;
    return 1;
}

/****************************************************************************
PARAMETERS:
milliSeconds    - Number of milliseconds to delay

REMARKS:
Delays for the specified number of milliseconds.
{secret}
****************************************************************************/
void _ASMAPI _OS_msDelay(
    N_uint32 milliSeconds)
{
    while (milliSeconds--)
        _OS_delay(1000);
}

/****************************************************************************
PARAMETERS:
err - Error to log

RETURNS:
True on success, false on failure.

REMARKS:
Function to log an error message if the MTRR write combining attempt failed.
****************************************************************************/
static ibool LogMTRRError(
    int err)
{
    switch (err) {
        case PM_MTRR_ERR_OK:
            _log("Success!\n");
            return true;
        case PM_MTRR_NOT_SUPPORTED:
            _log("\nFailed: MTRR is not supported by host CPU\n");
            break;
        case PM_MTRR_ERR_PARAMS:
            _log("\nFailed: Invalid parameters passed to PM_enableWriteCombined!\n");
            break;
        case PM_MTRR_ERR_NOT_4KB_ALIGNED:
            _log("\nFailed: Address is not 4Kb aligned!\n");
#ifdef  TEST_HARNESS
            PM_fatalError("Failed: Address is not 4Kb aligned!");
#endif
            break;
        case PM_MTRR_ERR_BELOW_1MB:
            _log("\nFailed: Addresses below 1Mb cannot be write combined!\n");
            break;
        case PM_MTRR_ERR_NOT_ALIGNED:
            _log("\nFailed: Address is not correctly aligned for processor!\n");
#ifdef    TEST_HARNESS
          PM_fatalError("Failed: Address is not correctly aligned for processor!");
#endif
            break;
        case PM_MTRR_ERR_OVERLAP:
            _log("\nFailed: Address overlaps an existing region!\n");
            break;
        case PM_MTRR_ERR_TYPE_MISMATCH:
            _log("\nFailed: Adress is contained with existing region, but type is different!\n");
            break;
        case PM_MTRR_ERR_NONE_FREE:
            _log("\nFailed: Out of MTRR registers!\n");
            break;
        case PM_MTRR_ERR_NOWRCOMB:
            _log("\nFailed: This processor does not support write combining!\n");
            break;
        case PM_MTRR_ERR_NO_OS_SUPPORT:
            _log("\nFailed: MTRR is not supported by host OS\n");
            break;
        default:
            _log("\nFailed: UNKNOWN ERROR!\n");
            break;
        }
    return false;
}

/****************************************************************************
DESCRIPTION:
Programs the MTR registers for the specific graphics device context

HEADER:
snap/graphics.h

PARAMETERS:
dc  - Device context to program MTR registers for

REMARKS:
This is an internal function to program the MTR registers that is not
usually called from outside application code. However this function is
exposed so that it can be called from code that resets the machine to
the correct state after coming out of a power management suspend mode.
****************************************************************************/
void NAPI GA_programMTRRegisters(
    GA_devCtx *dc)
{
    ulong   base,length,limit;
    int     i;

    // Never do write combining when we load the drivers under a GUI
    if (loadInGUI)
        return;

    // Skip write combining for compatibility reasons on some chipsets
    if (globalOpt.bNoWriteCombine || (dc->WorkArounds & gaNoWriteCombine)) {
        _log("Write combining skipped to avoid compatibility problems.\n");
        return;
        }

    // Attempt to enable write combining for the linear framebuffer. The
    // length always needs to get rounds up to a 512K boundary to get the
    // proper performance we need.
    if ((base = dc->LinearBasePtr) != 0) {
        length = (dc->LinearSize + 0x7FFFF) & ~0x7FFFF;
        limit = length - 1;
        if (base & limit) {
            // Account for the fact that some drivers adjust the framebuffer
            // starting address by 1-4K in order to avoid chipset problems,
            // or might have a framebuffer starting address that is not properly
            // aligned. We fix this by rounding down to the next 'length'
            // granularity so that the framebuffer length and base address are
            // on the same boundary (required for Intel chipset MTRR's to work).
            base &= ~limit;
            }
        _log("Attempting to enable write combining, base = 0x%08X, length = 0x%08X ... ", base, length);
        if (!LogMTRRError(PM_enableWriteCombine(base,length,PM_MTRR_WRCOMB)))
            return;

        // Now attempt to enable write combining for MMIO regions that can
        // accept that.
        for (i = 0; i < 4; i++) {
            if (dc->MMIOBase[i] && dc->MMIOBase[i] > 0x100000) {
                base = dc->MMIOBase[i];
                length = dc->MMIOLen[i] & ~0x80000000;
                if (base & 0xFFF) {
                    length += (base & 0xFFF);
                    base &= ~0xFFF;
                    }
                if (dc->MMIOLen[i] & 0x80000000) {
                    length = (length + 0x7FFFF) & ~0x7FFFF;
                    _log("Attempting to enable write combining for MMIO, base = 0x%08X, length = 0x%08X ... ", base, length);
                    LogMTRRError(PM_enableWriteCombine(base,length,PM_MTRR_WRCOMB));
                    }
                }
            }

        // Now attempt to disable caching for the MMIO regions that cannot be
        // cached. Note that we have to round the base down to a 4Kb boundary
        // and round up the length to a 4Kb boundary.
        for (i = 0; i < 4; i++) {
            if (dc->MMIOBase[i] && dc->MMIOBase[i] > 0x100000) {
                base = dc->MMIOBase[i];
                length = dc->MMIOLen[i] & ~0x80000000;
                if (base & 0xFFF) {
                    length += (base & 0xFFF);
                    base &= ~0xFFF;
                    }
                if (!(dc->MMIOLen[i] & 0x80000000)) {
                    length = (length + 0xFFF) & ~0xFFF;
                    _log("Attempting to disable caching for MMIO, base = 0x%08X, length = 0x%08X ... ", base, length);
                    LogMTRRError(PM_enableWriteCombine(base,length,PM_MTRR_UNCACHABLE));
                    }
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
dc      - Pointer to device context

REMARKS:
This function counts the number of modes in the mode profile and generates
the correct AvailableModes table to reflect this.
{secret}
****************************************************************************/
static void GA_createModeList(
    GA_devCtx *dc,
    N_int32 numModes)
{
    int         i;
    N_uint16    *m;

    for (i = 0, m = dc->AvailableModes; i < numModes; i++,m++)
        *m = (N_uint16)(0x100 + i);

    // Check for and add multi head virtual display modes
    // if they are present to the end of the mode list. The virtual
    // modes always starts at gaModeMask and work down.
    if (dc->Attributes & gaHaveMultiHead) {
        *m++ = gaVirtualMode8bpp;
        *m++ = gaVirtualMode15bpp;
        *m++ = gaVirtualMode16bpp;
        *m++ = gaVirtualMode24bpp;
        *m++ = gaVirtualMode32bpp;
        }
    *m = 0xFFFF;
}

/****************************************************************************
PARAMETERS:
dc      - Pointer to device context

REMARKS:
Performs internal initialisation to create the default mode profile for
the display driver. This is now done in the generic code so that we can
retrofit new modes to older drivers with a change to the generic code
without recompiling the device specific binaries.
{secret}
****************************************************************************/
static void GA_initModeTable(
    GA_devCtx *dc)
{
#ifndef __ANALYSE__
    GA_mode         *mode;
    FILE            *f;
    char            filename[256];

    // First try reading the default mode table from disk
    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    gInit.GetUniqueFilename(filename,1);
    if ((f = fopen(filename,"rb")) != NULL) {
        fread(&GA_defaultModeProfile,1,sizeof(GA_defaultModeProfile),f);
        fclose(f);
        }

    // Now count the number of modes in the mode list
    GA_defaultModeProfile.m.numModes = 0;
    for (mode = GA_defaultModeProfile.m.modeList; mode->xRes != 0; mode++)
        GA_defaultModeProfile.m.numModes++;
    GA_createModeList(dc,GA_defaultModeProfile.m.numModes);

    // Now count the number of modes in the VBE mode list
    GA_defaultModeProfile.vbe.numModes = 0;
    for (mode = GA_defaultModeProfile.vbe.modeList; mode->xRes != 0; mode++)
        GA_defaultModeProfile.vbe.numModes++;

    // Now set the mode profile in the driver
    gInit.SetModeProfile(&GA_defaultModeProfile);
#endif
}

/****************************************************************************
PARAMETERS:
dc      - Pointer to device context

REMARKS:
Performs detection to find the current display mode in the system. This
will generally only find the current display mode if the system is in
hardware VGA text mode, and we will match it up against one of the internal
SNAP display modes. For NonVGA or graphics modes, we will simply choose
the first mode in the list as the current default mode (usually 80x25
text mode, but could be a graphics mode if the driver does not have any
text modes).

Also if the passed in oldModeNum is not -1, it means the driver is
loaded in an environment where another driver is already loaded and
knows the current mode number, so we just use that.
{secret}
****************************************************************************/
static void GA_detectCurrentMode(
    GA_devCtx *dc,
    int oldModeNum,
    ibool isTextMode,
    int vDisp,
    int cellHeight,
    int width)
{
#ifndef __ANALYSE__
    GA_deviceEntry  *pga = &ga[dc->DeviceIndex];
    N_uint16        *m;
    int             xChars,yChars,height;

    if (oldModeNum == -1) {
        if (dc->DeviceIndex > 0)
            pga->sharedInfo->cntModeNum = 0xFFFF;
        else
            pga->sharedInfo->cntModeNum = 0x100;
        if (isTextMode) {
            height = vDisp / cellHeight;
            for (m = dc->AvailableModes; *m != 0xFFFF; m++) {
                gMi.dwSize = sizeof(gMi);
                if (gInit.GetVideoModeInfo(*m,&gMi) != 0)
                    continue;
                if (!(gMi.Attributes & gaIsTextMode))
                    continue;
                xChars = gMi.XResolution / gMi.XCharSize;
                yChars = gMi.YResolution / gMi.YCharSize;
                if (xChars == width && yChars == height && gMi.YCharSize == cellHeight) {
                    pga->sharedInfo->cntModeNum = *m;
                    break;
                    }
                }
            }
        }
    else
        pga->sharedInfo->cntModeNum = oldModeNum;
#endif
}

/****************************************************************************
PARAMETERS:
dc      - Pointer to device context
device  - Device index of the device we are initialising

RETURNS:
Error code

REMARKS:
Performs internal initialisation on the GA_devCtx driver block, assuming
that it has been loaded correctly.
{secret}
****************************************************************************/
N_int32 _GA_initInternal(
    GA_devCtx *dc,
    int device)
{
    int     i;

    /* Verify that the file is the driver file we are expecting */
    if (strcmp(dc->Signature,GA_SIGNATURE) != 0)
        return status = nCorruptDriver;

    /* Map the memory mapped register locations for the driver. We need to
     * map up to four different locations that may possibly be needed. If
     * the base address is zero then the memory does not need to be mapped.
     */
    if (ga[device].IOMemMaps[0] == NULL) {
        for (i = 0; i < 4; i++) {
            if (dc->MMIOBase[i]) {
                /* If the top bit is set on the length for the buffer,
                 * then the buffer should be mapped with write combining
                 * enabled. If the top bit is not set, we don't want
                 * write combining *or* caching enabled (or it will
                 * screw up register writes). Some drivers will have
                 * a write combined and a non-write combined region
                 * mapped so that the choice can be made an runtime (ie:
                 * use the write combined region for system blit
                 * transfers for performance).
                 */
                if (dc->MMIOLen[i] & 0x80000000)
                    ga[device].IOMemMaps[i] = PM_mapPhysicalAddr(dc->MMIOBase[i],(dc->MMIOLen[i] & ~0x80000000)-1,true);
                else
                    ga[device].IOMemMaps[i] = PM_mapPhysicalAddr(dc->MMIOBase[i],dc->MMIOLen[i]-1,false);
                if (ga[device].IOMemMaps[i] == NULL) {
                    _log("Failed to map IO memory: 0x%08X for 0x%08X bytes\n",dc->MMIOBase[i],dc->MMIOLen[i]);
                    return status = nMemMapError;
                    }
                }
            }
        }
    for (i = 0; i < 4; i++)
        dc->IOMemMaps[i] = ga[device].IOMemMaps[i];

    /* Map the text mode memory area for the driver */
    if (ga[device].TextMem == NULL && dc->TextBasePtr) {
        ga[device].TextMem = PM_mapPhysicalAddr(dc->TextBasePtr,dc->TextSize-1,true);
        if (ga[device].TextMem == NULL) {
            _log("Failed to map Text memory: 0x%08X for 0x%08X bytes\n",dc->TextBasePtr,dc->TextSize);
            return status = nMemMapError;
            }
        }
    dc->TextMem = ga[device].TextMem;

    /* Map the banked video memory area for the driver */
    if (ga[device].BankedMem == NULL && dc->BankedBasePtr) {
        if (dc->BankedBasePtr == 0xA0000)
            ga[device].BankedMem = PM_getA0000Pointer();
        else
            ga[device].BankedMem = PM_mapPhysicalAddr(dc->BankedBasePtr,dc->BankSize-1,true);
        if (ga[device].BankedMem == NULL) {
            _log("Failed to map graphics memory: 0x%08X for 0x%08X bytes\n",dc->BankedBasePtr,dc->BankSize);
            return status = nMemMapError;
            }
        }
    dc->BankedMem = ga[device].BankedMem;

    /* Map the linear video memory area for the driver */
    if (ga[device].LinearMem == NULL && dc->LinearBasePtr) {
        ulong size = dc->LinearSize;
        while (ga[device].LinearMem == NULL) {
            ga[device].LinearMem = PM_mapPhysicalAddr(dc->LinearBasePtr,size-1,true);
            if (ga[device].LinearMem == NULL) {
                if (size == 0) {
                    _log("Failed to map graphics memory: 0x%08X for 0x%08X bytes\n",dc->LinearBasePtr,dc->LinearSize);
                    return status = nMemMapError;
                    }
                /* Try to map 1/2 of the framebuffer memory if the DOS extender fails
                 * to map in all the memory. That way we can at least get the driver to
                 * work properly albiet with less available memory.
                 */
                size /= 2;
                }
            }
        if (size != dc->LinearSize) {
            _log("Partially mapped graphics memory: 0x%08X instead of 0x%08X bytes\n",size, dc->LinearSize);
            _log("You may not see all available memory in your application due to DOS extender limitiations\n");
            _log("\n");
            dc->LinearSize = size;
            }
        }
    dc->LinearMem = ga[device].LinearMem;

    /* Map the local zbuffer memory area for the driver */
    if (ga[device].ZBufferMem == NULL && dc->ZBufferBasePtr) {
        ga[device].ZBufferMem = PM_mapPhysicalAddr(dc->ZBufferBasePtr,dc->ZBufferSize-1,true);
        if (ga[device].ZBufferMem == NULL) {
            _log("Failed to map zbuffer memory: 0x%08X for 0x%08X bytes\n",dc->ZBufferBasePtr,dc->ZBufferSize);
            return status = nMemMapError;
            }
        }
    dc->ZBufferMem = ga[device].ZBufferMem;

    /* Map the local texture memory area for the driver */
    if (ga[device].TexBufferMem == NULL && dc->TexBufferBasePtr) {
        ga[device].TexBufferMem = PM_mapPhysicalAddr(dc->TexBufferBasePtr,dc->TexBufferSize-1,true);
        if (ga[device].TexBufferMem == NULL) {
            _log("Failed to map texture memory: 0x%08X for 0x%08X bytes\n",dc->TexBufferBasePtr,dc->TexBufferSize);
            return status = nMemMapError;
            }
        }
    dc->TexBufferMem = ga[device].TexBufferMem;

    /* Allocate locked memory for the driver if requested */
    if (dc->LockedMemSize > 0) {
        dc->LockedMem = PM_allocLockedMem(dc->LockedMemSize,&dc->LockedMemPhys,true,false);
        if (dc->LockedMem == NULL) {
            _log("Failed to allocate locked memory: 0x%08X bytes\n", dc->LockedMemSize);
            return status = nMemMapError;
            }
        }

    /* Set pointers to the 8x8, 8x14 and 8x16 text mode fonts */
    dc->TextFont8x8 = _NGA_font8x8;
    dc->TextFont8x14 = _NGA_font8x14;
    dc->TextFont8x16 = _NGA_font8x16;

    /* Set pointers to the default CGA and VGA palettes */
    dc->VGAPal4 = _NGA_VGAPal4;
    dc->VGAPal8 = _NGA_VGAPal8;

    /* Enable linear framebuffer write combining. We must do this *before*
     * we call the InitDriver() function, because we need to ensure all
     * MMIO regions are properly uncached before the driver init sequence
     * is exectued.
     */
    GA_programMTRRegisters(dc);

    /* Now call the driver to initialise itself */
    return (status = dc->loader.InitDriver());
}

/****************************************************************************
REMARKS:
Internal function to load the X86 emulator library when it is needed.
{secret}
****************************************************************************/
ibool _GA_loadBIOSEmulator(
    int deviceIndex)
{
    // Load and initialise the X86 emulator if not already loaded. If we
    // no longer need the X86 emulator for the device driver, we will
    // unload it after the device drivers are loaded.
    if (!BE_loaded) {
#if defined(TEST_HARNESS) && defined(DEBUG_X86EMU)
        if ((logfile = fopen("biosemu.log","w")) == NULL)
            exit(1);
        printk("BIOS emulator initialised.\n");
#endif
        if (!BE_init(0,65536,&VGAInfo[0]))
            return false;
        BE_loaded = true;
        }

    // Point the PM library functions at the X86 emulator if we
    // don't have any BIOS access at all.
    if (_PM_importsOrg.dwSize == 0)
        _PM_importsOrg = _PM_imports;
    if (deviceIndex > 0 || !PM_haveBIOSAccess()) {
        _PM_imports.PM_mapRealPointer   = BE_mapRealPointer;
        _PM_imports.PM_getVESABuf       = BE_getVESABuf;
        _PM_imports.PM_callRealMode     = BE_callRealMode;
        _PM_imports.PM_int86            = BE_int86;
        _PM_imports.PM_int86x           = BE_int86x;
        }
    else {
        _PM_imports.PM_mapRealPointer   = _PM_importsOrg.PM_mapRealPointer;
        _PM_imports.PM_getVESABuf       = _PM_importsOrg.PM_getVESABuf;
        _PM_imports.PM_callRealMode     = _PM_importsOrg.PM_callRealMode;
        _PM_imports.PM_int86            = _PM_importsOrg.PM_int86;
        _PM_imports.PM_int86x           = _PM_importsOrg.PM_int86x;
        }
    return true;
}

/****************************************************************************
DESCRIPTION:
Force the amount of display memory configured for the driver.

HEADER:
snap/graphics.h

PARAMETERS:
totalMemory - Amount of memory in Kb on the board
shared      - True if the device driver should be loaded into shared memory

RETURNS:
True on success, false on error.

REMARKS:
This function loads the SNAP driver and forces the memory size for
subsequently loaded drivers to the specified size in Kb. The method
to detect the amount of installed display memory on a graphics card in many
cases requires the state of the hardware to be changed temporarily so that
the memory size can be detected. This will cause corruption if a driver
is being loaded when native IHV drivers are already running (ie: as a DDC
control panel utility, or a stereo game driver utility). This function is
used to force the memory size and avoid the hardware state change so that
loading a driver is completely non-destructive to the hardware state.
The memory size should be obtained using the native OS system services
(such as DirectX).

SEE ALSO:
GA_loadDriver
****************************************************************************/
ibool NAPI GA_loadInGUI(
    N_int32 shared)
{
    (void)shared;
    loadInGUI = 1;
    return true;
}

/****************************************************************************
REMARKS:
Returns the hardware acceleration type.
****************************************************************************/
static const char *GetLCDMode(void)
{
    switch (globalOpt.bLCDUseBIOS) {
        case gaLCDUseBIOS_Off:  return "Off";
        case gaLCDUseBIOS_Auto: return "Auto";
        case gaLCDUseBIOS_On:   return "On";
        }
    return "Unknown!";
}

/****************************************************************************
REMARKS:
Returns the AGP shared memory size
****************************************************************************/
static const char *GetAGPMemSize(void)
{
    static char buf[20];
    sprintf(buf,"%d Kb", globalOpt.dwSharedAGPMemSize);
    return buf;
}

/****************************************************************************
REMARKS:
Returns the hardware acceleration type.
****************************************************************************/
static const char *GetSysMemMode(void)
{
    static char buf[20];

    if (globalOpt.bUseMemoryDriver) {
        sprintf(buf,"On (%d Kb)", globalOpt.wSysMemSize);
        return buf;
        }
    return "Off";
}

/****************************************************************************
REMARKS:
Function to log the currently configured global options to disk
****************************************************************************/
static void DumpGlobalOptions(void)
{
    int i;

    _log(
        "Global options for all devices:\n"
        "  Force VBE Fallback ...... %s\n"
        "  Force VGA Fallback ...... %s\n"
        "  Allow non-certified ..... %s\n"
        "  Disable write combining . %s\n"
        "  Use BIOS for LCD panel... %s\n"
        "  Shared AGP memory size... %s\n"
        "  Use system memory driver. %s\n"
        "  Disable DDC detection.... %s\n",
        globalOpt.bVBEOnly          ? "On" : "Off",
        globalOpt.bVGAOnly           ? "On" : "Off",
        globalOpt.bAllowNonCertified ? "On" : "Off",
        globalOpt.bNoWriteCombine    ? "On" : "Off",
        GetLCDMode(),
        GetAGPMemSize(),
        GetSysMemMode(),
        globalOpt.bNoDDCDetect    ? "On" : "Off");
    if (globalOpt.bVirtualDisplay) {
        _log("  Virtual Display.......... Yes\n");
        for (i = 0; i < GA_MAX_VIRTUAL_DISPLAYS; i++) {
            if (globalOpt.bounds[i].right == 0)
                break;
            _log("    %2d: %4d x %4d (%4d,%4d,%4d,%4d)\n",
                i,
                globalOpt.resolutions[i].right,
                globalOpt.resolutions[i].bottom,
                globalOpt.bounds[i].left,
                globalOpt.bounds[i].top,
                globalOpt.bounds[i].right,
                globalOpt.bounds[i].bottom);
            }
        }
    else
        _log("  Virtual Display.......... Off\n");
    _log("\n");
}

/****************************************************************************
DESCRIPTION:
Enumerates the number of display devices on the PCI bus.

HEADER:
snap/graphics.h

PARAMETERS:
shared  - True if the device driver should be loaded into shared memory

RETURNS:
Number of display devices found.

REMARKS:
This function enumerates the number of available display devices on the
PCI bus, and returns the number found. The number of devices allows you
to load graphics drivers for each device, but calling GA_loadDriver starting
with device 0 and up to the maximum number of devices-1. Note that
even though we may enumarate a valid PCI display device, it is possible
that no graphics driver is available for the device and the GA_loadDriver
function could fail.

Hence this function does not enumerate all the devices and match them up
with valid graphics drivers; the matching to available drivers is done when
you call GA_loadDriver.

SEE ALSO:
GA_loadDriver
****************************************************************************/
int NAPI GA_enumerateDevices(
    N_int32 shared)
{
#ifndef __DRIVER__
    if (NumDevices == -1) {
        PM_init();
        ___drv_os_type = PM_getOSType();
        PM_findBPD("graphics.bpd",SNAPConfigPath);
        PM_backslash(SNAPConfigPath);
        strcat(SNAPConfigPath,"config");
        PM_backslash(SNAPConfigPath);
        }
#endif
    // Load global options first to get the necessary options from disk
    GA_getGlobalOptions(&globalOpt,shared);

    // Check if we have already been here
    status = nOK;
    if (NumDevices != -1)
        return NumDevices;

    // Detect and log the processor information
    N_logCPUInfo();
    if (_IS_OS_CLASS(_OS_CLASS_WIN32)) {
#if defined(__DRIVER__) || (defined(TEST_HARNESS) && defined(__WIN32__))
        NumDevices = (int)PM_loadDirectDraw(-1);
        _log("GA_enumerateDevices: Found %d DirectX display device\n\n", NumDevices);
#else
        NumDevices = 1;
#endif
        }
    else {
        // Now enumerate all devices on the bus
        if ((NumDevices = _PCI_enumerateDevices(globalOpt.bAGPFastWrite,globalOpt.bMaxAGPRate)) > 0) {
            // Check that Multi-Controller is licensed
            if (NumDevices > 1 && licensedOptions && !licensedOptions->bMultiDisplay)
                NumDevices = 1;
            _log("GA_enumerateDevices: Found %d PCI/AGP display devices\n\n", NumDevices);
            }
        else {
            // No PCI display devices, so we must have a VLB or non-PCI device
            _log("GA_enumerateDevices: No PCI display devices found - assuming VLB/ISA\n\n");
            NumDevices = 1;
            }
        }

    // Allocate enough memory to manage all installed devices
    if ((ga = calloc(NumDevices,sizeof(*ga))) == NULL) {
        status = nOutOfMemory;
        return 0;
        }

    // Dump the global options information
    DumpGlobalOptions();
    return NumDevices;
}

/****************************************************************************
PARAMETERS:
szBPDName   - Name of BPD file to load
size        - Place to store the size of the loaded BPD
shared      - True to load into shared memory

REMARKS:
Internal function to load a BPD file from inside the SNAP flat file
for the main device driver. Mainly used by external utility programs and
drivers to get access to packaged BPD files from the core BPD file. The
BPD should be freed with the regular PE_freeLibrary function.
{secret}
****************************************************************************/
PE_MODULE * NAPI GA_loadLibrary(
    const char *szBPDName,
    ulong *size,
    ibool shared)
{
    PE_MODULE   *hMod;
    N_FILE      *f;
#ifdef CHECKED
    char        path[PM_MAX_PATH];

    // First try to load an external binary for debug builds
    strcpy(path,SNAPDriverPath);
    strcat(path,szBPDName);
    if ((hMod = PE_loadLibrary(path,shared)) != NULL)
        return hMod;
#endif

    // Next try in the bound binary if not found externally
    if ((f = FLAT_openFile(szBPDName)) == NULL)
        return NULL;
    hMod = PE_loadLibraryExt(f->f,f->start,size,shared);
    FLAT_closeFile(f);
    return hMod;
}

/****************************************************************************
PARAMETERS:
hModule     - Module to get procedure address from
szProcName  - Name of procedure to retrieve

REMARKS:
Internal function get the address of a named procedure in a BPD file.
{secret}
****************************************************************************/
void * NAPI GA_getProcAddress(
    PE_MODULE *hModule,
    const char *szProcName)
{
    return PE_getProcAddress(hModule,szProcName);
}

/****************************************************************************
PARAMETERS:
hModule     - Pointer to module to unload

REMARKS:
Internal function to unload a BPD.
{secret}
****************************************************************************/
void NAPI GA_freeLibrary(
    PE_MODULE *hModule)
{
    PE_freeLibrary(hModule);
}

/****************************************************************************
REMARKS:
Function to log the currently configured monitor information to disk.
****************************************************************************/
static void DumpMonitorInfo(
    GA_monitor *monitor)
{
    _log("  Manufacturer... %s\n", monitor->mfr);
    _log("  Model.......... %s\n", monitor->model);
    _log("  Max Resolution. ");
    switch (monitor->maxResolution) {
        case MaxRes_640x480:    _log("640x480\n");      break;
        case MaxRes_800x600:    _log("800x600\n");      break;
        case MaxRes_1024x768:   _log("1024x768\n");     break;
        case MaxRes_1152x864:   _log("1152x864\n");     break;
        case MaxRes_1280x1024:  _log("1280x1024\n");    break;
        case MaxRes_1600x1200:  _log("1600x1200\n");    break;
        case MaxRes_1800x1350:  _log("1800x1350\n");    break;
        case MaxRes_1920x1440:  _log("1920x1440\n");    break;
        case MaxRes_2048x1536:  _log("2048x1536\n");    break;
        case 0xFF:              _log("Unrestricted\n"); break;
        default:                _log("Unknown!\n");
        }
    if (monitor->maxResolution != 0xFF) {
        _log("  Max HScan...... %d KHz\n", monitor->maxHScan);
        _log("  Max VScan...... %d Hz\n", monitor->maxVScan);
        _log("  Features....... ");
        if (monitor->flags & Monitor_DPMSEnabled)
            _log("DPMS ");
        if (monitor->flags & Monitor_GTFEnabled)
            _log("GTF ");
        if (monitor->flags & Monitor_FixedFreq)
            _log("FIXED ");
        if (monitor->flags & Monitor_16to9)
            _log("16:9 ");
        }
    _log("\n\n");
}

/****************************************************************************
REMARKS:
Returns the hardware acceleration type.
****************************************************************************/
static const char *GetHardwareAccelType(void)
{
    switch (gOpt.accelType) {
        case gaAccelType_Custom:    return "Custom";
        case gaAccelType_Full:      return "Full";
        case gaAccelType_Most:      return "Most";
        case gaAccelType_Basic:     return "Basic";
        case gaAccelType_None:      return "None";
        }
    return "Unknown!";
}

/****************************************************************************
REMARKS:
Returns the hardware bus type as a string.
****************************************************************************/
static const char *GetBusName(
    N_uint32 busType)
{
    switch (busType) {
        case gaISABus:
            return "ISA bus";
        case gaMCABus:
            return "MCA bus";
        case gaVLBBus:
            return "VLB bus";
        case gaPCIBus:
            return "PCI bus";
        case gaAGPBus:
            return "AGP bus";
        }
    return "Unknown bus";
}

/****************************************************************************
REMARKS:
Function to log the currently configured device information to disk.
****************************************************************************/
static void DumpDisplayInfo(
    GA_devCtx *dc)
{
    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    gInfo.dwSize = sizeof(gInfo);
    gInit.GetConfigInfo(&gInfo);
    gOpt.dwSize = sizeof(gOpt);
    gInit.GetOptions(&gOpt);
    _log("\nGraphics device configuration:\n");
    _log("  Manufacturer......... %s\n"
         "  Chipset.............. %s\n"
         "  Bus Type............. %s\n"
         "  Memory............... %d Kb\n"
         "  DAC.................. %s\n"
         "  Clock................ %s\n"
         "  Memory Clock......... %d MHz\n"
         "  Default Memory Clock. %d MHz\n"
         "  Maximum Memory Clock. %d MHz\n",
        gInfo.ManufacturerName, gInfo.ChipsetName, GetBusName(dc->BusType),
        dc->TotalMemory, gInfo.DACName, gInfo.ClockName,
        N_FIXTOINT(gOpt.memoryClock),
        N_FIXTOINT(gOpt.defaultMemoryClock),
        N_FIXTOINT(gOpt.maxMemoryClock));
    if (dc->Attributes & gaHaveEngineClock) {
        _log("  Engine Clock......... %d MHz\n"
             "  Default Engine Clock. %d MHz\n"
             "  Maximum Engine Clock. %d MHz\n",
            N_FIXTOINT(gOpt.engineClock),
            N_FIXTOINT(gOpt.defaultEngineClock),
            N_FIXTOINT(gOpt.maxEngineClock));
        }
    _log("  Driver Revision...... %s\n"
         "  Driver Build......... %s\n",
        gInfo.VersionInfo, gInfo.BuildDate);
    if (gInfo.Certified) {
        _log("  Certified Version.... %d.%02d\n"
             "  Certified Date....... %s\n",
            gInfo.CertifyVersion >> 8,
            gInfo.CertifyVersion & 0xFF,
            gInfo.CertifiedDate);
        }
    else {
        _log("  Certified Version.... Not Certified\n");
        }

    // Show current device specific settings
    _log(
        "\nGraphics device options:\n"
        "  Invert .................. %s\n"
        "  Rotation ................ %s\n"
        "  Flipped ................. %s\n"
        "  Prefer 16 bit per pixel.. %s\n"
        "  Prefer 32 bit per pixel.. %s\n"
        "  Hardware acceleration.... %s\n",
        globalOpt.bInvertColors && gOpt.bInvertColors   ? "On" : "Off",
        globalOpt.bPortrait && gOpt.bPortrait           ? "On" : "Off",
        globalOpt.bFlipped && gOpt.bFlipped             ? "On" : "Off",
        gOpt.bPrefer16bpp                               ? "On" : "Off",
        gOpt.bPrefer32bpp                               ? "On" : "Off",
        GetHardwareAccelType());
}

/****************************************************************************
REMARKS:
Function to log the BIOS information for the card to the graphics.log file.
This includes the VESA OEM string information if available, along with
the first few hundred bytes of the BIOS itself.
****************************************************************************/
static void DumpBIOSInfo(
    GA_devCtx *dc)
{
    int     i,j;
    uchar   *BIOSImage;
    char    buf[80],buf2[80];

    if (!_IS_OS_CLASS(_OS_CLASS_WIN32)) {
        _log("\nBIOS header information:\n");
        if ((BIOSImage = _PCI_mapBIOSImage(dc->DeviceIndex)) != NULL) {
            for (j = 0; j < 6; j++) {
                strcpy(buf,"  \'");
                for (i = 0; i < 48; i++) {
                    if (isprint(BIOSImage[i])) {
                        sprintf(buf2,"%c",BIOSImage[i]);
                        strcat(buf,buf2);
                        }
                    else
                        strcat(buf,".");
                    }
                strcat(buf,"\'\n");
                _log("%s",buf);
                BIOSImage += 48;
                }
            _PCI_unmapBIOSImage(dc->DeviceIndex);
            }
        }
}

/****************************************************************************
REMARKS:
Function to find the name of a display mode from the mode information.
****************************************************************************/
static int GetModeName(
    char *buf,
    GA_modeInfo *mi)
{
    char    buf1[80];
    int     attr = mi->Attributes;
    int     pages = mi->MaxBuffers;
    int     xChars,yChars;

    /* Build the name of the mode */
    if (attr & gaIsTextMode) {
        xChars = mi->XResolution / mi->XCharSize;
        yChars = mi->YResolution / mi->YCharSize;
        sprintf(buf,"%dx%d Text Mode (%dx%d character cell)",
            xChars,yChars,mi->XCharSize,mi->YCharSize);
        }
    else {
        if (!(attr & gaHaveLinearBuffer))
            strcpy(buf1,", Banked");
        else
            strcpy(buf1,"");
        sprintf(buf,"%dx%dx%d, %d page%s",mi->XResolution,mi->YResolution,
            mi->BitsPerPixel,pages,buf1);
        if (attr & gaHaveAccel2D)
            strcat(buf,", Accel");
        if (attr & gaHaveTripleBuffer)
            strcat(buf,", TripleBuf");
        if (attr & gaHaveStereo)
            strcat(buf,", Stereo");
        if (attr & gaHaveNonVGAMode)
            strcat(buf,", NonVGA");
        }
    return true;
}

/****************************************************************************
REMARKS:
Function to log the currently configured mode information to disk.
****************************************************************************/
static void DumpModeInfo(
    GA_devCtx *dc,
    int head)
{
    N_uint16    *modes;
    int         i,bpp[] = {0,4,8,15,16,24,32,-1};
    char        modeName[256];
    int         outputDevice;

    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    gOpt.dwSize = sizeof(gOpt);
    gInit.GetOptions(&gOpt);
    outputDevice = (head == gaActiveHeadPrimary) ? gOpt.outputDevice : gaOUTPUT_CRT;
    if ((head == gaActiveHeadPrimary) && (outputDevice & gaOUTPUT_LCD) || (outputDevice & gaOUTPUT_DFP)) {
        if (gOpt.LCDPanelWidth <= 0)
            _log("Detected invalid panel size! Probably a bad chipset options file.\n");
        else {
            _log("LCD panel detected:\n");
            _log("  Width:  %d\n", gOpt.LCDPanelWidth);
            _log("  Height: %d\n", gOpt.LCDPanelHeight);
            }
        _log("\n");
        }
    if (outputDevice & gaOUTPUT_LCD)
        _log("Graphics mode information (LCD panel):\n");
    else if (outputDevice & gaOUTPUT_DFP)
        _log("Graphics mode information (DFP):\n");
    else if (outputDevice & gaOUTPUT_TV)
        _log("Graphics mode information (TV):\n");
    else
        _log("Graphics mode information (CRT):\n");
    for (i = 0; bpp[i] != -1; i++) {
        for (modes = dc->AvailableModes; *modes != 0xFFFF; modes++) {
            gMi.dwSize = sizeof(gMi);
            // old drivers do not have the new function
            if (gInit.GetVideoModeInfoExt(*modes,&gMi,outputDevice,head) != 0)
                continue;
            if (gMi.BitsPerPixel != bpp[i])
                continue;
            if (GetModeName(modeName,&gMi))
                _log("  %s\n", modeName);
            }
        }
}

#ifndef DEBUG_NEW_DRIVER
/****************************************************************************
REMARKS:
Sets the first SVGA graphics mode found.
****************************************************************************/
static int SetGraphicsMode(
    GA_devCtx *dc,
    GA_initFuncs *init,
    void *buf,
    int bufSize,
    int xRes,
    int yRes,
    int bits,
    int linear)
{
    GA_CRTCInfo     crtc;
    N_uint16        *modes;
    int             oldMode,firstmode = -1,mode = -1;
    N_int32         virtualX,virtualY,bytesPerLine,maxMem;

    /* Search for the first 8-bit graphics mode */
    for (modes = dc->AvailableModes; *modes != 0xFFFF; modes++) {
        gMi.dwSize = sizeof(gMi);
        if (init->GetVideoModeInfo(*modes,&gMi) != 0)
            continue;
        if (gMi.Attributes & gaIsTextMode)
            continue;
        if (linear && !(gMi.Attributes & gaHaveLinearBuffer))
            continue;
        if (gMi.BitsPerPixel >= 8) {
            if (firstmode == -1)
                firstmode = *modes;
            }
        if (gMi.XResolution == xRes && gMi.YResolution == yRes && gMi.BitsPerPixel == bits) {
            mode = *modes;
            break;
            }
        }
    if (mode == -1) {
        /* Use the first mode found if we can find the requested mode */
        if (firstmode == -1)
            return -1;
        mode = firstmode;
        }
    mode |= gaNoAccel;
    if (gMi.Attributes & gaHaveLinearBuffer)
        mode |= gaLinearBuffer;
    virtualX = virtualY = bytesPerLine = -1;
    oldMode = init->GetVideoMode();
    init->SetVideoMode(mode,&virtualX,&virtualY,&bytesPerLine,&maxMem,0,&crtc);

    /* Now save the first bufSize bytes of video memory so we can restore it
     * later
     */
    if (mode & gaLinearBuffer)
        rep_movsd(buf,dc->LinearMem,bufSize/4);
    else
        rep_movsd(buf,dc->BankedMem,bufSize/4);
    return oldMode;
}

/****************************************************************************
REMARKS:
Restore the old display mode active before graphics mode was started.
****************************************************************************/
static void RestoreMode(
    GA_devCtx *dc,
    GA_initFuncs *init,
    int oldMode,
    void *buf,
    int bufSize)
{
    GA_CRTCInfo     crtc;
    N_int32         virtualX,virtualY,bytesPerLine,maxMem;

    /* First restore the graphics memory we saved off */
    if (init->GetVideoMode() & gaLinearBuffer)
        rep_movsd(dc->LinearMem,buf,bufSize/4);
    else
        rep_movsd(dc->BankedMem,buf,bufSize/4);

    /* Set text mode and initialise console library */
    virtualX = virtualY = bytesPerLine = -1;
    init->SetVideoMode(oldMode,&virtualX,&virtualY,&bytesPerLine,&maxMem,0,&crtc);
}
#endif

/****************************************************************************
REMARKS:
Search for and find a monitor in our monitor database based on it's PnP ID.
****************************************************************************/
static ibool FindMonitor(
    char *PNPID,
    GA_monitor *monitor)
{
    ibool       found = false;
    GA_monitor  rec;

    if (MDBX_first(&rec) == MDBX_ok) {
        do {
            if (strcmp(rec.PNPID,PNPID) == 0) {
                *monitor = rec;
                found = true;
                break;
                }
            } while (MDBX_next(&rec) == MDBX_ok);
        }
    return found;
}

/****************************************************************************
REMARKS:
Sets the Plug and Play monitor information as the active monitor
configuration in the system.
****************************************************************************/
static void SetPnPMonitorInfo(
    GA_monitor *monitor,
    EDID_record *rec,
    ibool copyNames)
{
    // Update the monitor record as necessary
    if (copyNames) {
        strcpy(monitor->mfr,rec->mfrName);
        strcpy(monitor->model,rec->modelName);
        }
    strcpy(monitor->PNPID,rec->PNPID);
    monitor->maxResolution = rec->maxResolution;
    monitor->minHScan = rec->minHScan;
    monitor->maxHScan = rec->maxHScan;
    monitor->minVScan = rec->minVScan;
    monitor->maxVScan = rec->maxVScan;
    monitor->flags = 0;
    if (rec->flags & EDID_DPMSEnabled)
        monitor->flags |= Monitor_DPMSEnabled;
    if (rec->flags & EDID_GTFEnabled)
        monitor->flags |= Monitor_GTFEnabled;

    // Now flush the changed information to the monitor database
    // so it will match up the GUI control panel applets.
    MDBX_insert(monitor);
}

/****************************************************************************
DESCRIPTION:
Detects a Plug and Play monitor attached to the graphics card

HEADER:
snap/graphics.h

PARAMETERS:
dc          - Device context for the device to do detection for
outputHead  - Output device to detect the PnP monitor on
monitor     - Place to store the returned monitor information.
hasChanged  - Place to store whether the monitor has changed

RETURNS:
0 if no DDC monitor, 1 if found, 2 if found without operational limits

REMARKS:
This function performs DDC monitor detection for the passed in device
context. If the monitor is found, it is returned in the monitor parameter.
Note that if we detect an older DDC monitor that does not have operational
limits, we use the PNPID from the monitor EDID to search for the correct
entry in our monitor database and return those values instead. This allows
us to correct for badly formed monitor EDID information present in early
monitors (ie: EDID's prior to 1.1).

Note that this function will also return whether the monitor information
has changed from the monitor record passed into the function. If the values
have changed in anyway the 'changed' parameter will be set to true
otherwise it will be set to false.

Note:   This function assumes that the system is already in graphics mode.
        Some device's will not allow PnP monitor detection unless the
        graphics card is running in hires graphics mode, so you should
        ensure this is the case before calling this function.

SEE ALSO:
GA_saveMonitorInfo, GA_getParsedEDID
****************************************************************************/
int NAPI GA_detectPnPMonitor(
    GA_devCtx *dc,
    N_int32 outputHead,
    GA_monitor *monitor,
    ibool *hasChanged)
{
    char        filename[PM_MAX_PATH];
    int         retVal = 0;

    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    *hasChanged = false;
    if (DDC_init(dc) == ddcOk && DDC_readEDID(0xA0,edid,sizeof(edid),0,outputHead)) {
        strcpy(filename,SNAPConfigPath);
        strcat(filename,"monitor.dbx");
        MDBX_open(filename);
        validEDID = false;
        if (EDID_parse(edid,&edidRec,true) == 1) {
            // Fully compliant EDID with operational limits found. We
            // look up the monitor manufacturer name and model name
            // from our monitor database, but use all the information
            // from the monitor database.
            if (!FindMonitor(edidRec.PNPID,monitor))
                SetPnPMonitorInfo(monitor,&edidRec,true);
            else
                SetPnPMonitorInfo(monitor,&edidRec,false);
            retVal = 1;
            validEDID = true;
            }
        else if (EDID_parse(edid,&edidRec,false)) {
            // Compatible EDID found, but without operational limits.
            // This has caused problems in the past because some EDID's
            // contain bogus information. Hence we use the PNPID for the
            // monitor to search for it in our monitor database and use
            // the values in there instead.
            if (!FindMonitor(edidRec.PNPID,monitor))
                SetPnPMonitorInfo(monitor,&edidRec,true);
            else {
                if (edidRec.flags & EDID_DPMSEnabled)
                    monitor->flags |= Monitor_DPMSEnabled;
                if (edidRec.flags & EDID_GTFEnabled)
                    monitor->flags |= Monitor_GTFEnabled;
                }
            retVal = 2;
            validEDID = true;
            }

        // Clear out the records to ensure that the binary compare operations succeed
        memset(monitor->mfr+strlen(monitor->mfr),0,sizeof(monitor->mfr)-strlen(monitor->mfr));
        memset(monitor->model+strlen(monitor->model),0,sizeof(monitor->model)-strlen(monitor->model));
        memset(monitor->PNPID+strlen(monitor->PNPID),0,sizeof(monitor->PNPID)-strlen(monitor->PNPID));

        // Check if the detected DDC monitor has changed
        if (memcmp(&gDDCMonitor[outputHead],monitor,sizeof(gDDCMonitor[0])) != 0) {
            // Set the monitor information in the driver as necessary
            *hasChanged = true;
            gDDCMonitor[outputHead] = *monitor;
            monitor->flags |= Monitor_Changed;
            gInit.SetMonitorInfo(monitor,outputHead);
            MDBX_flush();
            }
        MDBX_close();
        }
    return retVal;
}

/****************************************************************************
PARAMETERS:
dc          - Device context for the device to do detection for
monitor     - Place to store the returned monitor information.
hasChanged  - Place to store whether the monitor has changed

RETURNS:
0 if no DDC monitor, 1 if found, 2 if found without operational limits

REMARKS:
Legacy function that only works on the primary head.
{secret}
****************************************************************************/
int NAPI GA_detectPnPMonitorV1(
    GA_devCtx *dc,
    GA_monitor *monitor,
    ibool *hasChanged)
{
    return GA_detectPnPMonitor(dc,gaActiveHeadPrimary,monitor,hasChanged);
}

/****************************************************************************
DESCRIPTION:
Returns the parsed EDID record if a DDC monitor was found

HEADER:
snap/graphics.h

PARAMETERS:
dc      - Device context to retrieve the EDID record for
rec     - Place to return the parsed edid record

RETURNS:
True if DDC monitor found and EDID was valid, false if no valid DDC monitor

REMARKS:
This function returns the existing, parsed EDID record for the primary
monitor for the specific device context. If no valid EDID record was
obtained from the monitor when the driver was loaded, this function
returns false.

SEE ALSO:
GA_detectPnPMonitor
****************************************************************************/
int NAPI GA_getParsedEDID(
    GA_devCtx *dc,
    EDID_record *rec)
{
    if (validEDID) {
        memcpy(rec,&edidRec,sizeof(*rec));
        return true;
        }
    (void)dc;
    return false;
}

#ifndef DEBUG_NEW_DRIVER
/****************************************************************************
PARAMETERS:
dc  - Device context to display splash screen on

REMARKS:
This function detects all attached plug and play monitors for the
newly loaded device, and sets up all the moniotr records as necessary.
Note that some graphics cards need to be in graphics mode for this to work,
so if that is the case we temporarily set graphics mode here while we do
the monitor detection.
****************************************************************************/
static void DetectPlugAndPlayMonitors(
    GA_devCtx *dc)
{
    ibool       hasChanged;
    void        *saveBuf;
    int         head,numHeads,oldMode = -1;
    GA_monitor  monitor;

    // Bail if we should not do plug and play detection
    if (globalOpt.bNoDDCDetect || _IS_OS_CLASS(_OS_CLASS_WIN32) || (dc->WorkArounds & gaNoDDCBIOS))
        return;

    // Get init functions
    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    numHeads = gInit.GetNumberOfHeads();

    // Set graphics mode if necessary
    if (!loadInGUI && (dc->WorkArounds & gaDDCNeedsGraphicsMode)) {
        if ((saveBuf = malloc(BUF_SAVE_SIZE)) == NULL)
            PM_fatalError("Out of memory!");
        oldMode = SetGraphicsMode(dc,&gInit,saveBuf,BUF_SAVE_SIZE,640,480,8,false);
        }

    // Loop over each head, doing the Plug and Play monitor detection
    for (head = 0; head < numHeads; head++) {
        // Now detect the Plug and Play monitor if connected
        switch (GA_detectPnPMonitor(dc,head,&monitor,&hasChanged)) {
            case 1:
                if (hasChanged)
                    _log("\nDetected new Plug and Play monitor on head %d\n", head);
                break;
            case 2:
                if (hasChanged)
                    _log("\nDetected new Plug and Play monitor (No Operational Limits) on head %d\n", head);
                break;
            }

        // Check if the monitor has changed and if so update the changed info
        if (hasChanged) {
            // If we have found a valid monitor, try setting all refresh rates
            // across the board to 85Hz for automatic ergonomic refresh.
            gMonitor[head] = monitor;
            if (gMonitor[head].maxResolution != 0xFF) {
                gInit.SetGlobalRefresh(85,head);
                GA_saveCRTCTimings(dc,head);
                }

            // Now write the monitor record to disk to make it permanent.
            GA_saveMonitorInfo(dc,head,&gMonitor[head]);

            // Since the LCD Panel size is stored in the chipset options file, we need to fix up
            // that information (if the file exists) when the monitor changes.
            gOpt.dwSize = sizeof(gOpt);
            gInit.GetOptions(&gOpt);
            if ((head == 0) && (gOpt.outputDevice & gaOUTPUT_DFP)) {
                FILE    *f;
                char    filename[PM_MAX_PATH];

                gInit.GetUniqueFilename(filename,2);
                if ((f = fopen(filename,"rb")) != NULL) {
                    fclose(f);
                    switch (gMonitor[head].maxResolution) {
                        case MaxRes_640x480:    gOpt.LCDPanelWidth = 640,  gOpt.LCDPanelHeight = 480;  break;
                        case MaxRes_800x600:    gOpt.LCDPanelWidth = 800,  gOpt.LCDPanelHeight = 600;  break;
                        case MaxRes_1024x768:   gOpt.LCDPanelWidth = 1024, gOpt.LCDPanelHeight = 768;  break;
                        case MaxRes_1152x864:   gOpt.LCDPanelWidth = 1152, gOpt.LCDPanelHeight = 864;  break;
                        case MaxRes_1280x1024:  gOpt.LCDPanelWidth = 1280, gOpt.LCDPanelHeight = 1024; break;
                        case MaxRes_1600x1200:  gOpt.LCDPanelWidth = 1600, gOpt.LCDPanelHeight = 1200; break;
                        case MaxRes_1800x1350:  gOpt.LCDPanelWidth = 1800, gOpt.LCDPanelHeight = 1350; break;
                        case MaxRes_1920x1440:  gOpt.LCDPanelWidth = 1920, gOpt.LCDPanelHeight = 1440; break;
                        case MaxRes_2048x1536:  gOpt.LCDPanelWidth = 2048, gOpt.LCDPanelHeight = 1536; break;
                        default:                gOpt.LCDPanelWidth = 1024, gOpt.LCDPanelHeight = 768;  break;
                        }
                    _log("Fixing up panel size to %d x %d\n", gOpt.LCDPanelWidth, gOpt.LCDPanelHeight);
                    gInit.SetOptions(&gOpt);
                    GA_saveOptions(dc,&gOpt);
                    }
                }
            }
        }

    // Restore the original display mode
    if (oldMode != -1)
        RestoreMode(dc,&gInit,oldMode,saveBuf,BUF_SAVE_SIZE);
}
#endif

/****************************************************************************
PARAMETERS:
dc  - Device context to load the chipset filter drivers for

RETURNS:
Pointer to the adjusted device driver context

REMARKS:
Internal function to load and initialise all the chipset filter drivers
for the loaded chipset driver.
****************************************************************************/
static GA_devCtx * LoadChipsetFilters(
    GA_devCtx *dc)
{
    GA_deviceEntry  *pga = &ga[dc->DeviceIndex];
    int             numFilters = 0;

    // Now load and initialise any chipset filter drivers as necessary.
    if (dc->Version < 0x202) {
        // Load "bugfix01.drv" for early drivers
        if ((dc = BUGFIX01_loadDriver(dc,SNAPConfigPath,&_N_imports,&_PM_imports,&_DRV_imports,
                licensedOptions,&globalOpt,&pga->hModFilters[numFilters])) == NULL) {
            _log("Unable to load the bugfix driver!\n");
            return NULL;
            }
        numFilters++;
        }
    pga->numFilters = numFilters;
    return dc;
}

/****************************************************************************
PARAMETERS:
deviceIndex - Index of device to load driver for (0 for primary controller)
shared      - True if the device driver should be loaded into shared memory

RETURNS:
Pointer to the loaded graphics driver, or NULL on error.

REMARKS:
Loads and intialises the named driver.
****************************************************************************/
static GA_devCtx * NAPI LoadDriver(
    N_int32 deviceIndex,
    N_int32 shared,
    PCIDeviceInfo *info,
    const char *drivername,
    N_uint32 busType)
{
    GA_devCtx           *dc;
    GA_deviceEntry      *pga = &ga[deviceIndex];
    char                buf[PM_MAX_PATH],filename[PM_MAX_PATH],*q;
    const char          *p;
    FILE                *fm;
    ibool               sharedIRQ;
    int                 head,numHeads,vDisp,cellHeight,width,isTextMode,oldModeNum;
#ifdef  __DRIVER__
    PE_MODULE           *hModule = NULL;
    GA_initDriver_t     GA_initDriver;
    ulong               size;
#endif

    // Check for special case of VBE/Core fallback and fail loading as necessary
    if (strcmp(drivername,"vbecore.drv") == 0) {
        // We can't support VBE/Core in GUI environments or in Windows NT/2K!
        // Eventually we may be able to support this under Windows XP.
        if (loadInGUI || _IS_OS(_OS_WINNTDRV) || !allowFallbackDrivers)
            return NULL;
        }

    // Check for special case of VGA driver and fail loading as necessary
    if (strcmp(drivername,"vga.drv") == 0) {
        if (!allowFallbackDrivers)
            return NULL;
        }

    // Post the secondary controller here, before we try to load it. Note that
    // this function will not actually post the card if the card is already
    // active, but it is necessary in order to set up the BIOS emulator
    // functions for the secondary card for flat panel and TV support etc.
    if (deviceIndex > 0)
        _PCI_postController(deviceIndex);

    // Now try to load the driver
#ifdef  __DRIVER__
    if ((hModule = GA_loadLibrary(drivername,&size,shared)) == NULL) {
        _log("Unable to load driver '%s'!\n", drivername);
        return NULL;
        }
    status = nCorruptDriver;
    if ((GA_initDriver = PE_getProcAddress(hModule,"_GA_initDriver")) == NULL)
        goto Error;
#endif
    status = nNotDetected;

    // Open the monitor records on disk for all heads
    for (head = 0; head < GA_MAX_HEADS; head++) {
        memset(&gMonitor[head],0,sizeof(gMonitor[0]));
        memset(&gDDCMonitor[head],0,sizeof(gDDCMonitor[0]));
        gMonitor[head].maxResolution = 0xFF;
        strcpy(gMonitor[head].mfr,"Unknown");
        strcpy(gMonitor[head].model,"Unknown");
        sprintf(buf,MONITOR_NAME,head,deviceIndex);
        strcpy(filename,SNAPConfigPath);
        strcat(filename,buf);
        if ((fm = fopen(filename,"rb")) != NULL) {
            fread(&gMonitor[head],1,sizeof(gMonitor[0]),fm);
            fread(&gDDCMonitor[head],1,sizeof(gDDCMonitor[0]),fm);
            fclose(fm);
            }
        }

    // Ensure the interrupt we are enabled for is masked to avoid problems
    // with the interrupt not being handled by the system. If our IRQ is
    // shared with another device, we have to disable triple buffering
    // support since we don't install a real interrupt handler.
#ifdef  ENABLE_TRIPLE_BUFFER
    if (!(sharedIRQ = _PCI_isSharedIRQ(info))) {
        if (3 <= info->InterruptLine && info->InterruptLine <= 7) {
            int pic = inpb(0x21);
            int mask = (0x1 << info->InterruptLine);
            outpb(0x21,pic | mask);
            }
        else if (8 <= info->InterruptLine && info->InterruptLine <= 15) {
            int pic2 = inpb(0xA1);
            int mask = (0x1 << (info->InterruptLine-8));
            outpb(0xA1,pic2 | mask);
            }
        }
    else {
        // The IRQ is shared, so see if we can use the BIOS to set the
        // routing for our IRQ to 0 to disable the interrupt.
        sharedIRQ = false;
        if (!PCI_setHardwareIRQ(info,info->InterruptPin+0x9,0)) {
            _log("WARN: Shared interrupt detected so triple buffering disabled!\n");
            sharedIRQ = true;
            }
        }
#else
    // Disable all triple buffering if not enabled
    sharedIRQ = true;
#endif

    if (!_IS_OS_CLASS(_OS_CLASS_WIN32)) {
        // Detect the current VGA CRTC settings for the original text mode. We
        // do it here rather than after initialising the driver because on
        // some devices we need to put the card into graphics mode in order to
        // detect the amount of memory installed. If we detect the text mode
        // settings after calling GA_initDriver(), it screws up the detection
        // of the initial text mode active when the driver was loaded (and
        // hence MGL programs won't restore the text mode correctly).
        vDisp = (rdinx(0x3D4,0x12)
              | ((rdinx(0x3D4,0x7) << 7) & 0x100)
              | ((rdinx(0x3D4,0x7) << 3) & 0x200)) + 1;
        cellHeight = (rdinx(0x3D4,0x9) & 0x1F) + 1;
        width = rdinx(0x3D4,0x01)+1;
        isTextMode = ((rdinx(0x3CE,0x6) & 0x01) == 0);
        }
    else
        isTextMode = false;

    // Perform internal initialisation
#ifdef  __DRIVER__
    pga->sharedInfo = _GA_imports.GA_getSharedInfo(deviceIndex);
#else
    pga->sharedInfo = GA_getSharedInfo(deviceIndex);
#endif
    if (!pga->sharedInfo) {
        memset(&pga->localInfo,0,sizeof(pga->localInfo));
        pga->localInfo.cntModeNum = -1;
        pga->sharedInfo = &pga->localInfo;
        }
    oldModeNum = pga->sharedInfo->cntModeNum;
    if ((dc = GA_initDriver(shared,SNAPConfigPath,deviceIndex,info,&gMonitor[0],&_N_imports,&_PM_imports,&_DRV_imports,licensedOptions,&globalOpt,sharedIRQ,exportTable,pga->sharedInfo,loadInGUI)) == NULL)
        goto Error;
    dc->Shared = shared;
    dc->DeviceIndex = deviceIndex;
    if (dc->Version < minDriverVersion)
        goto Error;
    if (dc->BusType == gaUnknownBus)
        dc->BusType = busType;
    if (_GA_initInternal(dc,deviceIndex) != nOK)
        goto Error;
    GA_initModeTable(dc);

    // Now load and initialise any device driver filter drivers as necessary
    if ((dc = LoadChipsetFilters(dc)) == NULL)
        return NULL;

    // Open the CRTC databases on disk. If this fails, generate the new
    // CRTC database and then re-load it.
    if (!CDBX_init(NumDevices,GA_MAX_HEADS)) {
        status = nOutOfMemory;
        goto Error;
        }
    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        goto Error;
    numHeads = gInit.GetNumberOfHeads();
    if (!CDBX_open(deviceIndex,numHeads))
        PM_fatalError("Unable to open CRTC database!");

    // Detect the current display mode if possible
    GA_detectCurrentMode(dc,oldModeNum,isTextMode,vDisp,cellHeight,width);

#ifndef DEBUG_NEW_DRIVER
    // Do plug and play monitor detection
    DetectPlugAndPlayMonitors(dc);
#endif

    // Log the configured graphics card information!
    DumpDisplayInfo(dc);
    DumpBIOSInfo(dc);

    // Pass all the monitor records for all heads to the driver. The
    // primary monitor is passed down via the GA_initDriver call above,
    // however secondary monitors need to be handled here. We just do
    // all heads again for completeness.
    for (head = 0; head < numHeads; head++) {
        if (numHeads > 1)
            _log("\nConfiguration for head %d:\n", head);
        gInit.SetMonitorInfo(&gMonitor[head],head);
        _log("\nMonitor configuration:\n");
        DumpMonitorInfo(&gMonitor[head]);
        DumpModeInfo(dc,head);
        }

    // Copy internal driver name to driver name list
    p = drivername;
    q = ga[deviceIndex].internalName;
    while (*p && *p != '.')
        *q++ = *p++;
    *q = 0;
    strcat(ga[deviceIndex].internalName,".drv");
    _log("\nDriver for device %d loaded successfully!\n", deviceIndex);
    _log("---------------------------------------------------------\n\n");

    // Success!
#ifdef  __DRIVER__
    dc->DriverStart = hModule;
    dc->DriverSize = size;
    ga[deviceIndex].hMod = hModule;
    ga[deviceIndex].hModSize = size;
#endif
    ga[deviceIndex].dc = dc;
    ga[deviceIndex].refCount = 1;
    status = nOK;
    return dc;

Error:
#ifdef  __DRIVER__
    PE_freeLibrary(hModule);
#endif
    return NULL;
}

#ifndef TEST_HARNESS
/****************************************************************************
DESCRIPTION:
Loads the 2D reference rasteriser device driver from disk.

HEADER:
snap/ref2d.h

PARAMETERS:
hwCtx           - Hardware context to load driver for (if any)
bitsPerPixel    - Pixel depth for driver to load (1,4,8,16,24 or 32)
shared          - True to load into shared memory
drv             - Pointer to the place to store loaded driver address
hModRef         - Pointer to the place to store the load driver module handle
size            - Pointer to the place to store the size of the driver

RETURNS:
True on success, false on failure.

REMARKS:
This function loads a copy of the SNAP 2D Reference Rasteriser for the
specific color depth from disk. This function will always load the SNAP
reference rasteriser driver from the public MGL directories, and should not
be used to load the reference rasteriser driver for use with a hardware
device context. Rather you should use the GA_loadRef2d function which
automatically handles support for loading and initialising the Portait
Display and Multi-Controller SNAP filter drivers.

However if you wish to load and use the SNAP reference rasteriser for
drawing to system memory bitmaps, you can use this function instead.

SEE ALSO:
GA_loadRef2d, REF2D_unloadDriver
****************************************************************************/
ibool NAPI REF2D_loadDriver(
    GA_devCtx _FAR_ *hwCtx,
    N_int32 bitsPerPixel,
    N_int32 shared,
    REF2D_driver _FAR_ *_FAR_ *drv,
    PE_MODULE _FAR_ *_FAR_ *hModRef,
    ulong _FAR_ *size)
{
    REF2D_initLibrary_t REF2D_initLibrary;
    char                bpdname[PM_MAX_PATH];

    if (bitsPerPixel == 15)
        bitsPerPixel = 16;
    sprintf(bpdname,REF2D_NAME,bitsPerPixel);
    if ((*hModRef = GA_loadLibrary(bpdname,size,shared)) == NULL)
        return false;
    if ((REF2D_initLibrary = (REF2D_initLibrary_t)PE_getProcAddress(*hModRef,"_REF2D_initLibrary")) == NULL)
        PM_fatalError("Unable to get BPD file entry point!");
    if ((*drv = REF2D_initLibrary(hwCtx,&globalOpt,&__GA_exports,&_PM_imports)) != NULL) {
        (*drv)->DriverStart = *hModRef;
        (*drv)->DriverSize = *size;
        }
    return (*drv != NULL);
}

/****************************************************************************
PARAMETERS:
modeInfo        - Mode information for the hardware display mode in use
shared          - True if the device driver should be loaded into shared memory
ref2d           - Pointer to the ref2d driver loaded prior in chain
drv             - Pointer to the place to store loaded driver address
hModule         - Pointer to the place to store the load driver module handle
size            - Pointer to the place to store the size of the driver

RETURNS:
True on success, false on failure.

REMARKS:
This internal function loads a copy of the SNAP SHADOW_ Display filter
driver from disk, and intialises it for use.
{secret}
****************************************************************************/
ibool NAPI SHADOW_loadDriver(
    GA_devCtx *hwCtx,
    GA_globalOptions *opt,
    REF2D_driver **drv,
    PE_MODULE **hModule,
    ulong *size)
{
    FILT_initLibrary_t  SHADOW_initLibrary;

    if ((*hModule = GA_loadLibrary(SHADOW_NAME,size,hwCtx->Shared)) == NULL)
        return false;
    if ((SHADOW_initLibrary = (FILT_initLibrary_t)PE_getProcAddress(*hModule,"_SHADOW_initLibrary")) == NULL)
        PM_fatalError("Unable to get BPD file entry point!");
    if ((*drv = SHADOW_initLibrary(hwCtx,opt,&_PM_imports)) != NULL) {
        (*drv)->DriverStart = *hModule;
        (*drv)->DriverSize = *size;
        }
    return (*drv != NULL);
}

/****************************************************************************
PARAMETERS:
modeInfo        - Mode information for the hardware display mode in use
transferStart   - Start of the offscreen transfer buffer area in bytes
shared          - True if the device driver should be loaded into shared memory
ref2d           - Pointer to the ref2d driver loaded prior in chain
drv             - Pointer to the place to store loaded driver address
hModule         - Pointer to the place to store the load driver module handle
size            - Pointer to the place to store the size of the driver

RETURNS:
True on success, false on failure.

REMARKS:
This internal function loads a copy of the SNAP Rotation Display filter
driver from disk, and intialises it for use.
{secret}
****************************************************************************/
ibool NAPI PORTRAIT_loadDriver(
    GA_devCtx *hwCtx,
    GA_globalOptions *opt,
    REF2D_driver **drv,
    PE_MODULE **hModule,
    ulong *size)
{
    FILT_initLibrary_t  PORTRAIT_initLibrary;

    if ((*hModule = GA_loadLibrary(PORTRAIT_NAME,size,hwCtx->Shared)) == NULL)
        return false;
    if ((PORTRAIT_initLibrary = (FILT_initLibrary_t)PE_getProcAddress(*hModule,"_PORTRAIT_initLibrary")) == NULL)
        PM_fatalError("Unable to get BPD file entry point!");
    if ((*drv = PORTRAIT_initLibrary(hwCtx,opt,&_PM_imports)) != NULL) {
        (*drv)->DriverStart = *hModule;
        (*drv)->DriverSize = *size;
        }
    return (*drv != NULL);
}

/****************************************************************************
PARAMETERS:
modeInfo        - Mode information for the hardware display mode in use
transferStart   - Start of the offscreen transfer buffer area in bytes
shared          - True if the device driver should be loaded into shared memory
ref2d           - Pointer to the ref2d driver loaded prior in chain
drv             - Pointer to the place to store loaded driver address
hModule         - Pointer to the place to store the load driver module handle
size            - Pointer to the place to store the size of the driver

RETURNS:
True on success, false on failure.

REMARKS:
This internal function loads a copy of the SNAP Flipped Display filter
driver from disk, and intialises it for use.
{secret}
****************************************************************************/
ibool NAPI FLIPPED_loadDriver(
    GA_devCtx *hwCtx,
    GA_globalOptions *opt,
    REF2D_driver **drv,
    PE_MODULE **hModule,
    ulong *size)
{
    FILT_initLibrary_t  FLIPPED_initLibrary;

    if ((*hModule = GA_loadLibrary(FLIPPED_NAME,size,hwCtx->Shared)) == NULL)
        return false;
    if ((FLIPPED_initLibrary = (FILT_initLibrary_t)PE_getProcAddress(*hModule,"_FLIPPED_initLibrary")) == NULL)
        PM_fatalError("Unable to get BPD file entry point!");
    if ((*drv = FLIPPED_initLibrary(hwCtx,opt,&_PM_imports)) != NULL) {
        (*drv)->DriverStart = *hModule;
        (*drv)->DriverSize = *size;
        }
    return (*drv != NULL);
}

/****************************************************************************
PARAMETERS:
modeInfo        - Mode information for the hardware display mode in use
transferStart   - Start of the offscreen transfer buffer area in bytes
shared          - True if the device driver should be loaded into shared memory
ref2d           - Pointer to the ref2d driver loaded prior in chain
drv             - Pointer to the place to store loaded driver address
hModule         - Pointer to the place to store the load driver module handle
size            - Pointer to the place to store the size of the driver

RETURNS:
True on success, false on failure.

REMARKS:
This internal function loads a copy of the SNAP Flipped Display filter
driver from disk, and intialises it for use.
{secret}
****************************************************************************/
ibool NAPI INVERT_loadDriver(
    GA_devCtx *hwCtx,
    GA_globalOptions *opt,
    REF2D_driver **drv,
    PE_MODULE **hModule,
    ulong *size)
{
    FILT_initLibrary_t  INVERT_initLibrary;

    if ((*hModule = GA_loadLibrary(INVERT_NAME,size,hwCtx->Shared)) == NULL)
        return false;
    if ((INVERT_initLibrary = (FILT_initLibrary_t)PE_getProcAddress(*hModule,"_INVERT_initLibrary")) == NULL)
        PM_fatalError("Unable to get BPD file entry point!");
    if ((*drv = INVERT_initLibrary(hwCtx,opt,&_PM_imports)) != NULL) {
        (*drv)->DriverStart = *hModule;
        (*drv)->DriverSize = *size;
        }
    return (*drv != NULL);
}

/****************************************************************************
PARAMETERS:
modeInfo        - Mode information for the hardware display mode in use
transferStart   - Start of the offscreen transfer buffer area in bytes
shared          - True if the device driver should be loaded into shared memory
ref2d           - Pointer to the ref2d driver loaded prior in chain
drv             - Pointer to the place to store loaded driver address
hModule         - Pointer to the place to store the load driver module handle
size            - Pointer to the place to store the size of the driver

RETURNS:
True on success, false on failure.

REMARKS:
This internal function loads a copy of the SNAP Rotation Display filter
driver from disk, and intialises it for use.
{secret}
****************************************************************************/
ibool NAPI MULDISP_loadDriver(
    GA_devCtx **hwCtx,
    N_int32 numDevices,
    GA_globalOptions *opt,
    REF2D_driver **drv,
    PE_MODULE **hModule,
    ulong *size,
    GA_setActiveDevice_t setActiveDevice)
{
    MULDISP_initLibrary_t   MULDISP_initLibrary;

    if ((*hModule = GA_loadLibrary(MULDISP_NAME,size,(*hwCtx)->Shared)) == NULL)
        return false;
    if ((MULDISP_initLibrary = (MULDISP_initLibrary_t)PE_getProcAddress(*hModule,"_MULDISP_initLibrary")) == NULL)
        PM_fatalError("Unable to get BPD file entry point!");
    if ((*drv = MULDISP_initLibrary(hwCtx,numDevices,opt,&_PM_imports,setActiveDevice)) != NULL) {
        (*drv)->DriverStart = *hModule;
        (*drv)->DriverSize = *size;
        }
    return (*drv != NULL);
}

/****************************************************************************
PARAMETERS:
modeInfo        - Mode information for the hardware display mode in use
transferStart   - Start of the offscreen transfer buffer area in bytes
shared          - True if the device driver should be loaded into shared memory
ref2d           - Pointer to the ref2d driver loaded prior in chain
drv             - Pointer to the place to store loaded driver address
hModule         - Pointer to the place to store the load driver module handle
size            - Pointer to the place to store the size of the driver

RETURNS:
True on success, false on failure.

REMARKS:
This internal function loads a copy of the SNAP Rotation Display filter
driver from disk, and intialises it for use.
{secret}
****************************************************************************/
GA_devCtx * NAPI BUGFIX01_loadDriver(
    GA_devCtx *dc,
    const char *path,
    N_imports *nImp,
    PM_imports *pmImp,
    DRV_imports *drvImp,
    GA_options *licensedOpt,
    GA_globalOptions *globalOpt,
    PE_MODULE **hModule)
{
    GA_initFilter_t GA_initFilter;
    ulong           size;

    if ((*hModule = GA_loadLibrary("bugfix01.drv",&size,dc->Shared)) == NULL)
        return NULL;
    if ((GA_initFilter = (GA_initFilter_t)PE_getProcAddress(*hModule,"_GA_initFilter")) == NULL)
        PM_fatalError("Unable to get BPD file entry point!");
    return GA_initFilter(dc,path,nImp,pmImp,drvImp,licensedOpt,globalOpt);
}
#endif

/****************************************************************************
PARAMETERS:
deviceIndex - Index of device to load driver for (0 for primary controller)
shared      - True if the device driver should be loaded into shared memory

RETURNS:
Pointer to the loaded graphics driver, or NULL on error.

REMARKS:
Internal function to load and initialise a driver for a single device.
****************************************************************************/
static GA_devCtx * __GA_loadDriver(
    N_int32 deviceIndex,
    N_int32 shared)
{
    char                filename[PM_MAX_PATH] = "null.drv";
    PCIDeviceInfo       *info;
    GA_devCtx           *dc;
    ulong               size;
    N_uint32            busType = gaUnknownBus;
    GA_deviceEntry      *pga = &ga[deviceIndex];
    int                 numFilters;
#ifdef  __DRIVER__
    PE_MODULE           *hDetect = NULL;
    GA_detectDevice_t   GA_detectDevice;
#endif

    // See if the driver has already been loaded first, and return it.
    // Note that we keep track of the reference count so we know how many
    // times it has been farmed out for shared environments.
    if (pga->dc) {
        pga->refCount++;
        return pga->dc;
        }

    // Validate the licensing
    status = nNotDetected;

    // Check that Multi-Controller is enabled
    if (deviceIndex > 0 && licensedOptions && !licensedOptions->bMultiDisplay)
        return NULL;

    // Find the device on the PCI bus
    status = nDriverNotFound;
    if (deviceIndex >= GA_enumerateDevices(shared))
        return NULL;
    if (_IS_OS_CLASS(_OS_CLASS_WIN32)) {
        // For Win32 user land apps create a fake PCI entry
        // and always load the ddraw.drv driver.
        info = _PCI_createFakePCIEntry("FFFF:FFFF");
        if (loadInGUI)
            strcpy(filename,"gdi.drv");
        else
            strcpy(filename,"ddraw.drv");
        }
    else {
        if ((info = _PCI_findDevice(deviceIndex)) == NULL) {
#ifdef  __DRIVER__
            status = nCorruptDriver;
            if ((hDetect = GA_loadLibrary(DETECT_NAME,&size,false)) == NULL)
                return NULL;
            if ((GA_detectDevice = PE_getProcAddress(hDetect,"_GA_detectDevice")) == NULL)
                return NULL;
#endif
#if !defined(TEST_VBEAF) && !defined(TEST_VBECORE) && !defined(TEST_DDRAW)
            // We have not found a valid PCI display device, so now do
            // generic ISA/VLB detection to find the fake PCI device ID for
            // the driver based on chipset. We use fake PCI device ID's so
            // that we can actually certify ISA/VLB devices with the
            // certification program.
            _log("Detecting ISA/VLB device ... ");
            if (!GA_detectDevice(fakePCIID,&_N_imports,&_PM_imports)) {
                // Fall back on VBE/Core driver if none is found
                _log("no device found!\n\n");
                strcpy(filename,"vbecore.drv");
                strcpy(fakePCIID,"FFFF:FFFF");
                }
            else {
                // Assume VLB bus type unless changed by the driver itself
                busType = gaVLBBus;
                _log("detected device ID '%s'\n\n", fakePCIID);
                }
#endif
#ifdef  __DRIVER__
            PE_freeLibrary(hDetect);
#endif
            info = _PCI_createFakePCIEntry(fakePCIID);
            }
        else {
            busType = (_PCI_isAGPDevice(deviceIndex) ? gaAGPBus : gaPCIBus);
            }
#ifdef  __DRIVER__
        // Now find the name of the driver to load from the directory table
        // and load it into memory.
        if (!_PCI_findDriverName(info,filename,"certify.tab")) {
            if (!globalOpt.bAllowNonCertified || !_PCI_findDriverName(info,filename,"driver.tab")) {
                // Fall back on VBE/Core driver if none is found
                strcpy(filename,"vbecore.drv");
                }
            }
#endif
#ifdef  TEST_VBECORE
        strcpy(filename,"vbecore.drv");
#endif

        // Allow fallbacks to VBE/Core or IBM VGA drivers based on user preference
        if (globalOpt.bVBEOnly)
            strcpy(filename,"vbecore.drv");
        if (globalOpt.bVGAOnly || bEvalExpired)
            strcpy(filename,"vga.drv");
        }

    // Force the system memory driver if selected
    if (globalOpt.bUseMemoryDriver)
        strcpy(filename,"memory.drv");

    // Log the loading information to the log file
    _log("Loading driver for device %d (%s)\n", deviceIndex, filename);
    _log("---------------------------------------------------------\n\n");

    // Try to load the driver, falling back to VBE/Core or VGA
    if ((dc = LoadDriver(deviceIndex,shared,info,filename,busType)) == NULL) {
        if (!_IS_OS_CLASS(_OS_CLASS_WIN32) && allowFallbackDrivers) {
            // Fall back on VBE driver if driver fails to load
            _log("Driver failed to load - falling back to vbecore.drv...\n");
            if (strcmp(filename,"vbecore.drv") == 0 || (dc = LoadDriver(deviceIndex,shared,info,"vbecore.drv",gaUnknownBus)) == NULL) {
                // Fall back on VGA driver if VBE/Core driver fails to load
                _log("Driver failed to load - falling back to vga.drv...\n");
                dc = LoadDriver(deviceIndex,shared,info,"vga.drv",gaUnknownBus);
                }
            }
        }
    if (!dc)
        return NULL;

    // Now load and initialise any ref2d filter drivers as necessary
    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        return false;
    gOpt.dwSize = sizeof(gOpt);
    gInit.GetOptions(&gOpt);
    numFilters = 0;

    // Multi-controller does not work with other filter drivers yet
    if (!globalOpt.bVirtualDisplay) {
        // Global flags for filters are on by default but can be disabled through licensing;
        // device flags determine whether rotation/flipped/invert is used or not.
        // If rotation is enabled, load and insert the filter driver.
        if (globalOpt.bPortrait && gOpt.bPortrait) {
            if (!PORTRAIT_loadDriver(dc,&globalOpt,&pga->ref2dFilters[numFilters],&pga->hModRef2dFilters[numFilters],&size))
                return NULL;
            numFilters++;
            }

        // If flipped is enabled, load and insert the filter driver.
        if (globalOpt.bFlipped && gOpt.bFlipped) {
            if (!FLIPPED_loadDriver(dc,&globalOpt,&pga->ref2dFilters[numFilters],&pga->hModRef2dFilters[numFilters],&size))
                return NULL;
            numFilters++;
            }

        // If color inversion is enabled, load and insert the filter driver.
        if (globalOpt.bInvertColors && gOpt.bInvertColors) {
            if (!INVERT_loadDriver(dc,&globalOpt,&pga->ref2dFilters[numFilters],&pga->hModRef2dFilters[numFilters],&size))
                return NULL;
            numFilters++;
            }
        }
    pga->dc = dc;
    pga->PCIInfo = info;
    pga->numRef2dFilters = numFilters;
    return dc;
}

/****************************************************************************
DESCRIPTION:
Loads a graphics driver file from disk for the specified device.

HEADER:
snap/graphics.h

PARAMETERS:
deviceIndex - Index of device to load driver for (0 for primary controller)
shared      - True if the device driver should be loaded into shared memory

RETURNS:
Pointer to the loaded graphics driver, or NULL on error.

REMARKS:
Loads the driver file from disk and intialises the device context ready for
use. If the driver file cannot be found, or the driver does not detect the
installed hardware, we return NULL and the application can get the status
code with the N_status function.

Note that for multiple controller support you can specify the device you
want to load the graphics driver for with the deviceIndex parameter. This
parameter is 0 for the primary controller, 1 for the secondary controller
and so on. The ordering of the controllers is determined by their
enumeration order on the PCI bus, however device 0 is always the primary
controller.

SEE ALSO:
GA_setActiveDevice, GA_unloadDriver, N_status, N_errorMsg,
GA_isSharedDriverLoaded
****************************************************************************/
GA_devCtx * NAPI GA_loadDriver(
    N_int32 deviceIndex,
    N_int32 shared)
{
    int             i;
    ulong           size;
    GA_devCtx       **dcList;

    // Enumerate available devices and load driver
    if (deviceIndex < 0 || deviceIndex >= GA_enumerateDevices(shared)) {
        if (status == nOK)
            status = nInvalidParameter;
        return NULL;
        }
    if (deviceIndex != 0 || !globalOpt.bVirtualDisplay || NumDevices <= 1)
        return __GA_loadDriver(deviceIndex,shared);

    // If multi-controller is enabled, load all device drivers and
    // connect up the multi-controller filter driver.
    if ((dcList = calloc(NumDevices,sizeof(*dcList))) == NULL)
        return NULL;
    for (i = NumDevices-1; i >= 0; i--) {
        GA_setActiveDevice(i);
        if ((dcList[i] = __GA_loadDriver(i,shared)) == NULL) {
            // Cap the max device index if the driver failed to load
            NumDevices = i-1;
            }
        }
    GA_setActiveDevice(0);
    if (!MULDISP_loadDriver(dcList,NumDevices,&globalOpt,&ga[0].ref2dFilters[ga[0].numRef2dFilters],&ga[0].hModRef2dFilters[ga[0].numRef2dFilters],&size,&GA_setActiveDevice))
        goto Error;
    ga[0].numRef2dFilters++;
    free(dcList);
    return ga[0].dc;

Error:
    free(dcList);
    return NULL;
}

/****************************************************************************
DESCRIPTION:
Sets the active display device for multiple display devices

HEADER:
snap/graphics.h

PARAMETERS:
deviceIndex - Index of device to make active

RETURNS:
True on success, false on error.

REMARKS:
This function allows the application to switch between the primary and
secondary display controllers, making one of the controllers the active
controller. When a controller is the active controller it will function
as though it is the only display controller in the system. You can also
select a 'mixed' mode of operation by passing in the DEVICE_MIXED_MODED
parameter, which fully enables the primary display controller but only
enables the secondary display controllers memory mappings. If mixed mode
is selected, it is the responsibility of the /calling/ application to
ensure that the secondary controller's VGA memory spaces are disabled
/before/ setting this mode to avoid conflicts between the controllers.

The device numbering starts at 0 for the primary display controller, and
increments by one for each supports display controller. Ie: the second
controller is device 1 while the third controller is device 2 etc. There
is no limit of the number of devices supported, other than the number of
PCI/AGP slots available in a particular machine.

SEE ALSO:
GA_loadDriver
****************************************************************************/
ibool NAPI GA_setActiveDevice(
    N_int32 deviceIndex)
{
    // Do nothing if device index is out of range
    if ((deviceIndex & DEVICE_MASK) >= NumDevices || NumDevices == 1)
        return false;
    if (!_IS_OS_CLASS(_OS_CLASS_WIN32)) {
        // Set the PCI device to be active
        if (!_PCI_setActiveDevice(deviceIndex))
            return false;

        // Set the BIOS emulator for the new device if we need it
        if (BE_loaded && (VGAInfo[cntDevice].BIOSImage || VGAInfo[deviceIndex].BIOSImage)) {
            BE_getVGA(&VGAInfo[cntDevice]);
            BE_setVGA(&VGAInfo[deviceIndex]);
            if (deviceIndex > 0 || !PM_haveBIOSAccess()) {
                _PM_imports.PM_mapRealPointer   = BE_mapRealPointer;
                _PM_imports.PM_getVESABuf       = BE_getVESABuf;
                _PM_imports.PM_callRealMode     = BE_callRealMode;
                _PM_imports.PM_int86            = BE_int86;
                _PM_imports.PM_int86x           = BE_int86x;
                }
            else {
                _PM_imports.PM_mapRealPointer   = _PM_importsOrg.PM_mapRealPointer;
                _PM_imports.PM_getVESABuf       = _PM_importsOrg.PM_getVESABuf;
                _PM_imports.PM_callRealMode     = _PM_importsOrg.PM_callRealMode;
                _PM_imports.PM_int86            = _PM_importsOrg.PM_int86;
                _PM_imports.PM_int86x           = _PM_importsOrg.PM_int86x;
                }
            }
        }
    cntDevice = deviceIndex;
    return true;
}

/****************************************************************************
PARAMETERS:
dc  - Pointer to device context to unload

REMARKS:
Internal function to do the driver unloading.
****************************************************************************/
static void UnloadDriver(
    int deviceIndex)
{
    int             j;
    GA_deviceEntry  *pga = &ga[deviceIndex];

    if (--pga->refCount == 0) {
        if (pga->glCtx)
            GA_glUnloadDriver(pga->glCtx);
        for (j = 0; j < pga->numRef2dFilters; j++) {
            pga->ref2dFilters[j]->UnloadDriver();
#ifdef  __DRIVER__
            PE_freeLibrary(pga->hModRef2dFilters[j]);
#endif
            pga->hModRef2dFilters[j] = NULL;
            }
        pga->numRef2dFilters = 0;
        if (pga->dc->LockedMem)
            PM_freeLockedMem(pga->dc->LockedMem,pga->dc->LockedMemSize,true);
        if (pga->dc->Version >= 0x101)
            pga->dc->loader.UnloadDriver();
        for (j = 0; j < pga->numFilters; j++) {
#ifdef  __DRIVER__
            PE_freeLibrary(pga->hModFilters[j]);
#endif
            pga->hModFilters[j] = NULL;
            }
        pga->numFilters = 0;
#ifdef  __DRIVER__
        PE_freeLibrary(pga->hMod);
#endif
        CDBX_close(deviceIndex);
        pga->dc = NULL;
        }
}

/****************************************************************************
DESCRIPTION:
Unloads the loaded graphics device driver.

HEADER:
snap/graphics.h

PARAMETERS:
dc  - Pointer to device context to unload

REMARKS:
This function simply unloads the graphics driver from memory, and frees
any of the internal resources that have been allocated for it. Specifically
it will free any locked memory and any interrupt handlers hooked for
the driver.

SEE ALSO:
GA_loadDriver
****************************************************************************/
void NAPI GA_unloadDriver(
    GA_devCtx *dc)
{
    int i;

    for (i = 0; i < NumDevices; i++) {
        if (dc->DeviceIndex == 0 && globalOpt.bVirtualDisplay && NumDevices > 1) {
            UnloadDriver(i);
            }
        else if (dc && ga[i].dc == dc) {
            UnloadDriver(i);
            break;
            }
        }
}

/****************************************************************************
DESCRIPTION:
Returns the internal binary driver name for the device

HEADER:
snap/graphics.h

PARAMETERS:
deviceIndex - Index of device to get driver name for

REMARKS:
This internal function returns the internal binary driver name for the
device. This information is only useful for QA purposes, and may change
over time.

NOTE:   The string is returned in a temporary buffer that is reused. Hence
        the calling application should copy the returned string to a
        different buffer immediately after calling this function and before
        calling any other SNAP functions.
****************************************************************************/
const char * NAPI GA_getInternalName(
    N_int32 deviceIndex)
{
    return ga[deviceIndex].internalName;
}

/****************************************************************************
DESCRIPTION:
Returns the internal fake PCI ID for ISA/VLB devices

HEADER:
snap/graphics.h

REMARKS:
This internal function returns the fake PCI device ID as a string from
the library. This value is only valid for ISA/VLB devices, and is used
so that ISA/VLB devices can be cleanly integrated into the PCI detection
and certification system.

NOTE:   The string is returned in a temporary buffer that is reused. Hence
        the calling application should copy the returned string to a
        different buffer immediately after calling this function and before
        calling any other SNAP functions.
****************************************************************************/
const char * NAPI GA_getFakePCIID(void)
{
    return fakePCIID;
}

/****************************************************************************
DESCRIPTION:
Returns the currently active SNAP configuration file path

HEADER:
snap/graphics.h

RETURNS:
Path to currently active SNAP configuration files

REMARKS:
This function returns the path to the currently active SNAP configuration
files on the system. User applications that need to access the monitor
database and other configuration files directly should use this function
to find the location of those files, rather than the PM_getSNAPConfigPath
function. The reason is that the PM library version always returns the
current system global configuration path, while this versions returns the
currently active versions. It is possible for SNAP to be located
relative to the application directory, and this function will return the
correct path if this is the case.

NOTE:   The string is returned in a temporary buffer that is reused. Hence
        the calling application should copy the returned string to a
        different buffer immediately after calling this function and before
        calling any other SNAP functions.
****************************************************************************/
const char * NAPI GA_getSNAPConfigPath(void)
{
    return SNAPConfigPath;
}

/****************************************************************************
PARAMETERS:
dc              - Device to load reference rasteriser for
shared          - True if the device driver should be loaded into shared memory
modeInfo        - Mode information for the hardware display mode in use
transferStart   - Start of the offscreen transfer buffer area in bytes
drv             - Pointer to the place to store loaded driver address

RETURNS:
True on success, false on failure.

REMARKS:
Internal function to load the ref2d driver for a single device. For
multi-controller modes we load a ref2d driver separately for each device,
as each device maintains it's own internal state (they may be running
different color depths).
****************************************************************************/
static ibool LoadRef2d(
    GA_devCtx *dc,
    N_int32 shared,
    GA_modeInfo *modeInfo,
    N_int32 transferStart,
    REF2D_driver **drv)
{
    ulong           size;
    int             i,bitsPerPixel;
    ibool           useShadowBuffer = false;
    GA_deviceEntry  *pga = &ga[dc->DeviceIndex];

    // If ref2d driver is already loaded, fail the function! We need to
    // ensure that OS drivers always unload the ref2d drivers when going
    // to text mode, and reload them later when going back to graphics
    // mode if the SNAP drivers are to be shared.
    if (pga->ref2d)
        return false;

    // Determine if we should use a shadow buffer for the driver
    bitsPerPixel = modeInfo->BitsPerPixel;
    if (!(modeInfo->Attributes & gaHaveLinearBuffer))
        useShadowBuffer = true;

    // Now load the SNAP reference rasteriser
    if (!REF2D_loadDriver(dc,bitsPerPixel,shared,drv,&pga->hModRef2d,&size))
        return false;
    pga->ref2d = *drv;

    // If the shadow buffer is enabled, load and insert the filter driver.
    if (useShadowBuffer) {
        if (!SHADOW_loadDriver(dc,&globalOpt,&pga->shadowBuf,&pga->hModShadow,&size))
            return false;
        if (!pga->shadowBuf->InitFilter(modeInfo,transferStart,*drv,drv))
            return false;
        *drv = pga->shadowBuf;
        }

    // Now initialise all filter drivers in order as necessary
    for (i = 0; i < pga->numRef2dFilters; i++) {
        if (!pga->ref2dFilters[i]->InitFilter(modeInfo,transferStart,*drv,drv))
            return false;
        transferStart = 0;
        }
    return true;
}

/****************************************************************************
DESCRIPTION:
Loads the 2D reference rasteriser device driver chain, including filter drivers.

HEADER:
snap/graphics.h

PARAMETERS:
dc              - Device to load reference rasteriser for
shared          - True if the device driver should be loaded into shared memory
modeInfo        - Mode information for the hardware display mode in use
transferStart   - Start of the offscreen transfer buffer area in bytes
drv             - Pointer to the place to store loaded driver address

RETURNS:
True on success, false on failure.

REMARKS:
This function loads a copy of the SNAP 2D Reference Rasteriser from
disk for use with the specified hardware display mode. If the user has
enabled rotation display support, or multi-controller display support, this
function will load and initialise all the necessary SNAP filter drivers
to enable those functions for the specified display mode.

The transferStart parameter is used to determine where the start of the
offscreen display memory transfer buffer is located. This transfer buffer
is used by the Portrait and Multi-Controller filter drivers for storing
temporary bitmap images in offscreen display memory for enhanced performance.
If the amount of memory from the start of the transfer area to the end of
available offscreen display memory is large enough to hold the transfer
buffer (no more than an entire display screen is ever required), then the
filter drivers will use that memory as necessary. If the amount of available
memory is less than required, the filter drivers will use system memory
buffers instead. To disable the use of the transfer buffer, simply pass in
a value that is equal to the amount of video memory on the graphics adapter
in bytes.

SEE ALSO:
GA_unloadRef2d, GA_unloadDriver, GA_getCurrentRef2d
****************************************************************************/
ibool NAPI GA_loadRef2d(
    GA_devCtx _FAR_ *dc,
    N_int32 shared,
    GA_modeInfo _FAR_ *modeInfo,
    N_int32 transferStart,
    REF2D_driver _FAR_ *_FAR_ *drv)
{
    int                     i;
#ifdef  __DRIVER__
    MULDISP_InitFilterExt_t MULDISP_InitFilterExt;
#endif
    REF2D_driver            **ref2dList;

    // Load the driver directly if not running multi-controller
    if (!globalOpt.bVirtualDisplay || !(modeInfo->Attributes & gaIsVirtualMode))
        return LoadRef2d(dc,shared,modeInfo,transferStart,drv);

    // If multi-controller is enabled, load all ref2d drivers for all
    // secondary drivers and init multi-controller filter driver.
    if ((ref2dList = calloc(NumDevices,sizeof(*ref2dList))) == NULL)
        return false;
    for (i = NumDevices-1; i >= 0; i--) {
        gInit.dwSize = sizeof(gInit);
        if (!GA_queryFunctions(ga[i].dc,GA_GET_INITFUNCS,&gInit))
            goto Error;

        // We pass a value of -1 so that the multi-controller driver
        // will return the actual hardware info instead of the virtual
        // mode information. We also pass the device index in the
        // Attributes parameter (a gross hack I know, but I had to
        // do something!).
        gMi.dwSize = -1;
        gMi.Attributes = i;
        gInit.GetCurrentVideoModeInfo(&gMi);
        if (!LoadRef2d(ga[i].dc,shared,&gMi,0,&ref2dList[i]))
            goto Error;
        }

    // Now initialise the multi-controller filter drivers
#ifdef  __DRIVER__
    if ((MULDISP_InitFilterExt = (MULDISP_InitFilterExt_t)PE_getProcAddress(ga[0].hModRef2dFilters[0],"_MULDISP_InitFilterExt")) == NULL)
        goto Error;
#endif
    if (!MULDISP_InitFilterExt(modeInfo,transferStart,ref2dList,drv))
        goto Error;
    free(ref2dList);
    return true;

Error:
    free(ref2dList);
    return false;
}

/****************************************************************************
DESCRIPTION:
Unloads the 2D reference rasteriser from memory

HEADER:
snap/ref2d.h

PARAMETERS:
drv     - Pointer to driver to unload
hModule - Pointer to the module for the driver to unload

REMARKS:
This function unloads the SNAP 2D Reference Rasteriser from memory. This
is used for when a reference rasteriser is loaded from disk independantly
of the hardware drivers, for use in rendering to system memory buffers
completely in software.

SEE ALSO:
REF2D_loadDriver
****************************************************************************/
void NAPI REF2D_unloadDriver(
    REF2D_driver _FAR_ *drv,
    PE_MODULE _FAR_ *hModule)
{
    if (drv)
        drv->UnloadDriver();
#ifndef TEST_HARNESS
    if (hModule)
        PE_freeLibrary(hModule);
#else
    (void)hModule;
#endif
}

/****************************************************************************
DESCRIPTION:
Unloads the 2D reference rasteriser device driver chain

HEADER:
snap/graphics.h

PARAMETERS:
dc  - Pointer to SNAP device context to unload driver chain for

REMARKS:
This function unloads the SNAP 2D Reference Rasteriser from memory,
without unloading the actual hardware device driver. This can be used to
unload the reference rasteriser and then reload a new copy such as when
changing color depths dynamically within a display driver.

SEE ALSO:
GA_unloadDriver
****************************************************************************/
void NAPI GA_unloadRef2d(
    GA_devCtx _FAR_ *dc)
{
    int             j;
    GA_deviceEntry  *pga = &ga[dc->DeviceIndex];

    if (pga->ref2d) {
        // Only unload the ref2d drivers if they are actually loaded!
        for (j = 0; j < pga->numRef2dFilters; j++)
            pga->ref2dFilters[j]->DeInitFilter();
        if (pga->shadowBuf)
            pga->shadowBuf->DeInitFilter();
        REF2D_unloadDriver(pga->ref2d,pga->hModRef2d);
        REF2D_unloadDriver(pga->shadowBuf,pga->hModShadow);
        }
    pga->ref2d = pga->shadowBuf = NULL;
    pga->hModRef2d = pga->hModShadow = NULL;
}

#ifdef __DRIVER__
/****************************************************************************
DESCRIPTION:
Loads the 2D region manager driver.

HEADER:
snap/graphics.h

PARAMETERS:
funcs   - Pointer to function structure to fill out
hModule - Place to store the module handle for the driver
size    - Place to store the size of the loaded module

RETURNS:
True on success, false on failure.

REMARKS:
This function loads a new copy of the SNAP 2D region manager functions
and exports the function list from the driver. Usually for SNAP
graphics applications, the region functions can simply be obtained from
the SNAP 2D reference rasteriser which also loads a global set of
region functions for internal use. However SNAP applications can use
this function to load a separate copy that is indepedant of the reference
rasteriser for internal use by the application.

SEE ALSO:
GA_unloadRegionMgr
****************************************************************************/
ibool NAPI GA_loadRegionMgr(
    GA_regionFuncs *funcs,
    PE_MODULE **hModule,
    ulong *size)
{
    RGN_initLibrary_t   RGN_initLibrary;

    if ((*hModule = GA_loadLibrary(REGION_NAME,size,false)) == NULL)
        return false;
    if ((RGN_initLibrary = (RGN_initLibrary_t)PE_getProcAddress(*hModule,"_RGN_initLibrary")) == NULL)
        PM_fatalError("Unable to get BPD file entry point!");
    return RGN_initLibrary(funcs);
}

/****************************************************************************
DESCRIPTION:
Unloads the 2D region manager driver.

HEADER:
snap/graphics.h

PARAMETERS:
hModule - Pointer to the module handle for the driver to unload

REMARKS:
This function unloads the SNAP 2D region manager from memory. This
should only be called if the application has loaded a separate set of
SNAP region manager functions with GA_loadRegionMgr.

SEE ALSO:
GA_loadRegionMgr
****************************************************************************/
void NAPI GA_unloadRegionMgr(
    PE_MODULE *hModule)
{
    RGN_exitLibrary_t   RGN_exitLibrary;

    if ((RGN_exitLibrary = (RGN_exitLibrary_t)PE_getProcAddress(hModule,"_RGN_exitLibrary")) == NULL)
        PM_fatalError("Unable to get BPD file entry point!");
    RGN_exitLibrary();
    PE_freeLibrary(hModule);
}
#endif

/****************************************************************************
DESCRIPTION:
Gets a copy of the current SNAP reference rasteriser for the specified
device index.

HEADER:
snap/graphics.h

PARAMETERS:
deviceIndex - Index of the device to obtain the SNAP reference rasteriser for

RETURNS:
Pointer to current SNAP reference rasteriser on success, NULL on failure.

REMARKS:
This returns a copy of the current SNAP reference rasteriser driver for
the specified device index. This function will return NULL if the driver
has not been loaded (ie: SciTech Display Doctor is not the primary display
driver in the system). The primary purpose of this function is to allow
operating system utilities and applications to access a copy of the
existing, shared SNAP driver used by the display driver in the system.

Note:   Do /not/ call GA_unloadRef2d with the value returned from this
        function!!

SEE ALSO:
GA_loadRef2d, GA_getCurrentDriver
****************************************************************************/
REF2D_driver * NAPI GA_getCurrentRef2d(
    N_int32 deviceIndex)
{
    if (deviceIndex >= NumDevices)
        return NULL;
    return ga[deviceIndex].ref2d;
}

/****************************************************************************
REMARKS:
Filters the global options based on licensing information, disabling those
options that are not licensed.
****************************************************************************/
static void FilterGlobalOptions(void)
{
    if (licensedOptions) {
        /* Filter based on licensing information */
        if (!licensedOptions->bVirtualDisplay)
            globalOpt.bVirtualDisplay = false;
        if (!licensedOptions->bPortrait)
            globalOpt.bPortrait = false;
        if (!licensedOptions->bFlipped)
            globalOpt.bFlipped = false;
        if (!licensedOptions->bInvertColors)
            globalOpt.bInvertColors = false;
        if (!licensedOptions->bAGPFastWrite)
            globalOpt.bAGPFastWrite = false;
        }

    /* Set default dwSharedAGPMemSize */
    if (globalOpt.dwSharedAGPMemSize == 0)
        globalOpt.dwSharedAGPMemSize = 4096;
}

/****************************************************************************
DESCRIPTION:
Returns the current global options from the graphics device driver.

HEADER:
snap/graphics.h

PARAMETERS:
options - Place to store the returned options information
shared  - True if the device driver should be loaded into shared memory

REMARKS:
This function returns a structure which contains global configuration
options effecitve for the installed devices.

Note:   The dwSize member of the profile structure is intended for future
        compatibility, and must be set to the size of the structure before
        calling this function. Only the number of bytes set in the
        dwSize member will be copied into the callers structure.

SEE ALSO:
GA_setGlobalOptions, GA_saveGlobalOptions
****************************************************************************/
void NAPI GA_getGlobalOptions(
    GA_globalOptions *options,
    ibool shared)
{
    FILE    *f;
    char    filename[PM_MAX_PATH];

    // Set default options for all global options values
    if (globalOpt.dwSize == 0) {
        globalOpt.dwSize = sizeof(globalOpt);
        globalOpt.bAllowNonCertified = true;
        globalOpt.bVBEUseLinear = true;
        globalOpt.bVBEUsePal = true;
        globalOpt.bVBEUsePM32 = true;
        globalOpt.bVBEUseVBE20 = true;
        globalOpt.bVBEUseVBE30 = true;
        globalOpt.bVBEUsePM = true;
        globalOpt.bVBEUseSCI = true;
        globalOpt.bVBEUseDDC = true;
        globalOpt.bGDIUseAccel = true;
        globalOpt.bGDIUseBrushCache = true;
        globalOpt.bGDIUseBitmapCache = true;
        globalOpt.bDXUseAccel2D = true;
        globalOpt.bDXUseAccel3D = true;
        globalOpt.bDXUseAccelVideo = true;
        globalOpt.bDXWaitRetrace = true;
        globalOpt.bAGPFastWrite = false;
        globalOpt.bMaxAGPRate = 8;
        globalOpt.bPortrait = true;
        globalOpt.bInvertColors = true;
        globalOpt.bFlipped = true;
        globalOpt.bVirtualDisplay = false;

        // Read existing saved options from disk
        strcpy(filename,SNAPConfigPath);
        strcat(filename,OPTIONS_NAME);
        if ((f = fopen(filename,"rb")) != NULL) {
            fread(&globalOpt,1,sizeof(globalOpt),f);
            fclose(f);
            if (globalOpt.dwSize >= sizeof(globalOpt))
                globalOpt.dwSize = sizeof(globalOpt);
            }

        // Filter options based on licensing and return to caller
        FilterGlobalOptions();
        }
    COPY_STRUCTURE(options,&globalOpt);

    // Copy the log disabling option to the necessary global variable.
    disableLogging = globalOpt.bDisableLogFile;
}

/****************************************************************************
DESCRIPTION:
Sets the current global options for the graphics device driver.

HEADER:
snap/graphics.h

PARAMETERS:
options - Global options to make active.

REMARKS:
This function installs a new set of global device options that control the
operation of all the devices at runtime. A default set of options is
always built into the device driver, but the options can be changed at
any time. The options may be made permanent with a call to the
GA_saveGlobalOptions function.

Note:   The dwSize member of the profile structure is intended for future
        compatibility, and must be set to the size of the structure before
        calling this function. Only the number of bytes set in the
        dwSize member will be copied from the callers structure.

SEE ALSO:
GA_getGlobalOptions, GA_saveGlobalOptions
****************************************************************************/
void NAPI GA_setGlobalOptions(
    GA_globalOptions *options)
{
    int i,index;

    globalOpt.dwSize = sizeof(globalOpt);
    COPY_STRUCTURE(&globalOpt,options);
    FilterGlobalOptions();

    // Now inform all filter drivers that the global options have changed
    for (index = 0; index < NumDevices; index++) {
        for (i = 0; i < ga[index].numRef2dFilters; i++) {
            if (ga[index].ref2dFilters[i]->GlobalOptionsUpdated)
                ga[index].ref2dFilters[i]->GlobalOptionsUpdated();
            }
        }

    // Copy the log disabling option to the necessary global variable.
    disableLogging = globalOpt.bDisableLogFile;
}

/****************************************************************************
DESCRIPTION:
Save the global options to disk

HEADER:
snap/graphics.h

PARAMETERS:
options - Global options information block to save

RETURNS:
True on success, false on error (see GA_status for error info)

REMARKS:
This function saves the global options information to disk.

SEE ALSO:
GA_getGlobalOptions, GA_setGlobalOptions
****************************************************************************/
ibool NAPI GA_saveGlobalOptions(
    GA_globalOptions *options)
{
    FILE    *f;
    char    filename[PM_MAX_PATH];

    strcpy(filename,SNAPConfigPath);
    strcat(filename,OPTIONS_NAME);
    if ((f = fopen(filename,"wb")) == NULL) {
        status = nIOError;
        return false;
        }
    fwrite(options,1,options->dwSize,f);
    fclose(f);
    status = nOK;
    return true;
}

/****************************************************************************
DESCRIPTION:
Load the mode profile information from disk for the specified driver

HEADER:
snap/graphics.h

PARAMETERS:
dc      - Device context to load mode profile information for
profile - Mode profile information block to load

REMARKS:
This function loads the mode profile information from disk. The mode profile
is the list of modes that the driver will report to user applications. You
can modify this list to customise the set of display modes that the driver
supports.

NOTE:   This function does not update the mode profile in the driver itself,
        but rather you should call SetModeProfile to update the internal
        mode profile first before saving the options to disk.

SEE ALSO:
GA_saveModeProfile
****************************************************************************/
void NAPI GA_loadModeProfile(
    GA_devCtx *dc,
    GA_modeProfile *profile)
{
    FILE    *f;
    int     size = profile->dwSize;
    char    filename[PM_MAX_PATH];

    // Get init functions
    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    gInit.GetUniqueFilename(filename,1);
    memset(profile,0,size);
    if ((f = fopen(filename,"rb")) != NULL) {
        fread(profile,1,size,f);
        fclose(f);
        }
    else {
        profile->dwSize = size;
        COPY_STRUCTURE(profile,&GA_defaultModeProfile);
        }
}

/****************************************************************************
DESCRIPTION:
Save the mode profile information to disk for the specified driver

HEADER:
snap/graphics.h

PARAMETERS:
dc      - Device context to save mode profile information for
profile - Mode profile information block to save

RETURNS:
True on success, false on error (see GA_status for error info)

REMARKS:
This function saves the mode profile information to disk. The mode profile
is the list of modes that the driver will report to user applications. You
can modify this list to customise the set of display modes that the driver
supports.

NOTE:   This function does not update the mode profile in the driver itself,
        but rather you should call SetModeProfile to update the internal
        mode profile first before saving the options to disk.

SEE ALSO:
GetModeProfile, SetModeProfile, GA_saveOptions, GA_saveCRTCTimings,
GA_loadModeProfile
****************************************************************************/
ibool NAPI GA_saveModeProfile(
    GA_devCtx *dc,
    GA_modeProfile *profile)
{
    FILE    *f;
    char    filename[PM_MAX_PATH];
    char    drive[PM_MAX_DRIVE];
    char    dir[PM_MAX_PATH];
    char    path[PM_MAX_PATH];

    // Fail if not licensed
    if (!bDialAMode_Licensed) {
        status = nNotLicensed;
        return false;
        }

    // Get init functions
    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    gInit.GetUniqueFilename(filename,1);
    if ((f = fopen(filename,"wb")) == NULL) {
        /* Try creating the options directory if not present */
        PM_splitpath(filename,drive,dir,NULL,NULL);
        PM_makepath(path,drive,dir,NULL,NULL);
        if (path[strlen(path)-1] == '\\')
            path[strlen(path)-1] = 0;
        PM_mkdir(path);
        f = fopen(filename,"wb");
        }
    if (f == NULL) {
        status = nIOError;
        return false;
        }
    fwrite(profile,1,profile->dwSize,f);
    fclose(f);
    status = nOK;
    return true;
}

/****************************************************************************
DESCRIPTION:
Save the options to disk for the specified driver

HEADER:
snap/graphics.h

PARAMETERS:
dc      - Device context to save options information for
options - Driver options information block to save

RETURNS:
True on success, false on error (see GA_status for error info)

REMARKS:
This function saves the driver options information to disk.

NOTE:   This function does not update the driver options in the driver
        itself, but rather you should call SetOptions to update the internal
        options first before saving the options to disk.

SEE ALSO:
GetOptions, SetOptions, GA_saveModeProfile, GA_saveCRTCTimings
****************************************************************************/
ibool NAPI GA_saveOptions(
    GA_devCtx *dc,
    GA_options *options)
{
    FILE    *f;
    char    filename[PM_MAX_PATH];
    char    drive[PM_MAX_DRIVE];
    char    dir[PM_MAX_PATH];
    char    path[PM_MAX_PATH];

    // Get init functions
    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    gInit.GetUniqueFilename(filename,2);
    if ((f = fopen(filename,"wb")) == NULL) {
        /* Try creating the options directory if not present */
        PM_splitpath(filename,drive,dir,NULL,NULL);
        PM_makepath(path,drive,dir,NULL,NULL);
        if (path[strlen(path)-1] == '\\')
            path[strlen(path)-1] = 0;
        PM_mkdir(path);
        f = fopen(filename,"wb");
        }
    if (f == NULL) {
        status = nIOError;
        return false;
        }
    fwrite(options,1,options->dwSize,f);
    fclose(f);
    status = nOK;
    return true;
}

/****************************************************************************
DESCRIPTION:
Save the monitor profile to disk for the specified driver

HEADER:
snap/graphics.h

PARAMETERS:
dc          - Device context to save options information for
outputHead  - Display head to save the monitor information for
monitor     - Monitor profile to save to disk

RETURNS:
True on success, false on error (see GA_status for error info)

REMARKS:
This function saves the monitor profile information to disk.

NOTE:   This function does not update the driver monitor stored in the driver
        itself, but rather you should call SetMonitorInfo to update the internal
        options first before saving the options to disk.

SEE ALSO:
GetMonitorInfo, SetMonitorInfo
****************************************************************************/
ibool NAPI GA_saveMonitorInfo(
    GA_devCtx *dc,
    N_int32 outputHead,
    GA_monitor *monitor)
{
    FILE    *f;
    char    buf[PM_MAX_PATH],filename[PM_MAX_PATH];

    /* Open the monitor configuration file for writing */
    if (outputHead == gaActiveHeadClone)
        outputHead = gaActiveHeadPrimary;
    sprintf(buf,MONITOR_NAME, outputHead, dc->DeviceIndex);
    strcpy(filename,SNAPConfigPath);
    strcat(filename,buf);
    if ((f = fopen(filename,"wb")) == NULL) {
        status = nIOError;
        return false;
        }
    fwrite(monitor,1,sizeof(GA_monitor),f);
    fwrite(&gDDCMonitor[outputHead],1,sizeof(gDDCMonitor[0]),f);
    fclose(f);
    status = nOK;
    return true;
}

/****************************************************************************
PARAMETERS:
dc          - Device context to save options information for
monitor     - Monitor profile to save to disk

RETURNS:
True on success, false on error (see GA_status for error info)

REMARKS:
Legacy function that works only on the primary head.
{secret}
****************************************************************************/
ibool NAPI GA_saveMonitorInfoV1(
    GA_devCtx *dc,
    GA_monitor *monitor)
{
    return GA_saveMonitorInfo(dc,gaActiveHeadPrimary,monitor);
}

/****************************************************************************
DESCRIPTION:
Get the CRTC timings for a specific display mode and refresh rate

HEADER:
snap/graphics.h

PARAMETERS:
dc          - Device context to save CRTC timings information for
outputHead  - Display head to retrieve CRTC timings for
modeInfo    - Display mode information block
refreshRate - Refresh rate to retrieve the CRTC timings for
crtc        - Place to store the retrieved CRTC timings

RETURNS:
True on success, false on error.

REMARKS:
This function obtains the CRTC timings for a particular display mode
and refresh rate from the internal CRTC timings database. If the resolution
and/or refresh rate is not supported by the display mode, this function
returns false.

SEE ALSO:
GA_saveCRTCTimings, GA_restoreCRTCTimings, SetCRTCTimings
****************************************************************************/
ibool NAPI GA_getCRTCTimings(
    GA_devCtx *dc,
    N_int32 outputHead,
    GA_modeInfo *modeInfo,
    N_int32 refreshRate,
    GA_CRTCInfo *crtc)
{
    CDBX_record crtcRec;

    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    gOpt.dwSize = sizeof(gOpt);
    gInit.GetOptions(&gOpt);
    gInit.GetMonitorInfo(&gMonitor[outputHead],outputHead);
    if (!CDBX_getCRTCTimings(dc->DeviceIndex,outputHead,dc->Attributes,
            modeInfo->XResolution,modeInfo->YResolution,
            gOpt.tripleScanLowRes,gOpt.doubleScan512,refreshRate,&gMonitor[outputHead],
            modeInfo->MaxPixelClock,(licensedOptions ? licensedOptions->maxRefresh : 0xFF),
            &crtcRec))
        return false;
    memcpy(crtc,&crtcRec.crtc,sizeof(*crtc));
    return true;
}

/****************************************************************************
REMARKS:
Legacy function for backwards compatibility.
{secret}
****************************************************************************/
ibool NAPI GA_getCRTCTimingsV1(
    GA_devCtx *dc,
    GA_modeInfo *modeInfo,
    N_int32 refreshRate,
    GA_CRTCInfo *crtc)
{
    return GA_getCRTCTimings(dc,gaActiveHeadPrimary,modeInfo,refreshRate,crtc);
}

/****************************************************************************
DESCRIPTION:
Set the CRTC timings for a specific display mode and refresh rate

HEADER:
snap/graphics.h

PARAMETERS:
dc          - Device context to save CRTC timings information for
outputHead  - Display head to set CRTC timing for
modeInfo    - Display mode information block
crtc        - CRTC timings to update for the mode
makeDefault - True to make the timings the default for the mode

RETURNS:
True on success, false on error.

REMARKS:
This function updates the CRTC timings for a particular display mode
and refresh rate (refresh rate stored in the crtc parameter) in the
7internal CRTC timings database. If the resolution and/or refresh rate is
not supported by the display mode, this function returns false.

The makeDefault parameter is used to determine whether the CRTC timings
should be made the default timings for the display mode. If this parameter
is true, the passed in timings will become the default timings for that
display mode. If not, the timings will be updated but the default refresh
rate will not be changed for the display mode.

SEE ALSO:
GA_saveCRTCTimings, GA_restoreCRTCTimings, SetCRTCTimings
****************************************************************************/
ibool NAPI GA_setCRTCTimings(
    GA_devCtx *dc,
    N_int32 outputHead,
    GA_modeInfo *modeInfo,
    GA_CRTCInfo *crtc,
    ibool makeDefault)
{
    // Save the CRTC timings as the current timings in the database
    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    if (modeInfo->Attributes & gaIsTextMode) {
        return CDBX_setCRTCTimings(dc->DeviceIndex,outputHead,
            modeInfo->XResolution * modeInfo->XCharSize,
            modeInfo->YResolution * modeInfo->YCharSize,
            crtc,false,makeDefault);
        }
    else {
        return CDBX_setCRTCTimings(dc->DeviceIndex,outputHead,
            modeInfo->XResolution,modeInfo->YResolution,
            crtc,false,makeDefault);
        }
}

/****************************************************************************
REMARKS:
Legacy function for backwards compatibility.
{secret}
****************************************************************************/
ibool NAPI GA_setCRTCTimingsV1(
    GA_devCtx *dc,
    GA_modeInfo *modeInfo,
    GA_CRTCInfo *crtc,
    ibool makeDefault)
{
    return GA_setCRTCTimings(dc,gaActiveHeadPrimary,modeInfo,crtc,makeDefault);
}

/****************************************************************************
DESCRIPTION:
Set the default refresh rate for a specific resolution and color depth

HEADER:
snap/graphics.h

PARAMETERS:
dc          - Device context to save CRTC timings information for
outputHead  - Display head to set default refresh rate for
xRes        - X resolution for the mode to change
yRes        - Y resolution for the mode to change
bits        - Color depth of the mode to change (0 for text modes)
refreshRate - New refresh rate to make the default
saveToDisk  - True to save the changes permanently to disk

RETURNS:
True on success, false on error.

REMARKS:
This function sets the default refresh rate for a specific resolution
and color depth to the passed in refresh rate for the mode.

Note:   Internally the CRTC timings are maintained independantly of
        specific colors depths, so changing the default refresh rate
        for 640x480x8 also changes it for all other 640x480 modes.
        The main reason we have the 'bits' parameter is so we can
        tell when we are dealing with text modes which are handled
        slightly differently.

SEE ALSO:
GA_saveCRTCTimings, GA_restoreCRTCTimings
****************************************************************************/
ibool NAPI GA_setDefaultRefresh(
    GA_devCtx *dc,
    N_int32 outputHead,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bits,
    N_int32 refreshRate,
    ibool saveToDisk)
{
    GA_CRTCInfo     crtc;
    N_uint16        *modes;
    N_int32         xChars,yChars;
    GA_options      opt;

    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    opt.dwSize = sizeof(opt);
    gInit.GetOptions(&opt);
    for (modes = dc->AvailableModes; *modes != 0xFFFF; modes++) {
        gMi.dwSize = sizeof(gMi);
        if (gInit.GetVideoModeInfoExt(*modes,&gMi,gaOUTPUT_CRT,outputHead) != 0)
            continue;
        if (gMi.Attributes & gaIsTextMode) {
            xChars = gMi.XResolution / gMi.XCharSize;
            yChars = gMi.YResolution / gMi.YCharSize;
            if (xRes != xChars || yRes != yChars || gMi.BitsPerPixel != bits)
                continue;
            }
        else if (gMi.XResolution != xRes || gMi.YResolution != yRes || gMi.BitsPerPixel != bits)
            continue;
        // If portrait is enabled, we need to swap the X/Y resolution back!
        if (globalOpt.bPortrait == 1 && opt.bPortrait == 1 && gMi.BitsPerPixel >= 8) {
            N_int32 temp;
            temp = gMi.YResolution;
            gMi.YResolution = gMi.XResolution;
            gMi.XResolution = temp;
            }
        if (!GA_getCRTCTimings(dc,outputHead,&gMi,refreshRate,&crtc))
            return false;
        if (!GA_setCRTCTimings(dc,outputHead,&gMi,&crtc,true))
            return false;
        return GA_saveCRTCTimings(dc,outputHead);
        }
    return false;
}

/****************************************************************************
REMARKS:
Legacy function for backwards compatibility.
{secret}
****************************************************************************/
ibool NAPI GA_setDefaultRefreshV1(
    GA_devCtx *dc,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bits,
    N_int32 refreshRate,
    ibool saveToDisk)
{
    return GA_setDefaultRefresh(dc,gaActiveHeadPrimary,xRes,yRes,bits,
        refreshRate,saveToDisk);
}

/****************************************************************************
DESCRIPTION:
Save the CRTC timings to disk for the specified driver

HEADER:
snap/graphics.h

PARAMETERS:
dc          - Device context to save CRTC timings information for
outputHead  - Display head to save the CRTC timings for

RETURNS:
True on success, false on error.

REMARKS:
This function saves the driver CRTC timings information to disk. If any
of the timings for the driver have been changed via the SetCRTCTimings or
SetGlobalRefresh functions, this will save those timings to disk to be used
as the defaults from then on.

SEE ALSO:
GA_restoreCRTCTimings, GetCRTCTimings, SaveCRTCTimings, SetGlobalRefresh,
GA_saveModeProfile, GA_saveOptions
****************************************************************************/
ibool NAPI GA_saveCRTCTimings(
    GA_devCtx *dc,
    N_int32 outputHead)
{
    char    filename[PM_MAX_PATH],buf[PM_MAX_PATH];
    void    *data;
    int     size;

    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    CDBX_getCRTCTable(dc->DeviceIndex,outputHead,&data,&size);
    sprintf(buf,CDBX_NAME,outputHead,dc->DeviceIndex);
    strcpy(filename,SNAPConfigPath);
    strcat(filename,buf);
    return CDBX_write(filename,data,size);
}

/****************************************************************************
REMARKS:
Legacy function for backwards compatibility.
{secret}
****************************************************************************/
ibool NAPI GA_saveCRTCTimingsV1(
    GA_devCtx *dc)
{
    return GA_saveCRTCTimings(dc,gaActiveHeadPrimary);
}

/****************************************************************************
DESCRIPTION:
Restore the CRTC timings from disk for the specified driver

HEADER:
snap/graphics.h

PARAMETERS:
dc          - Device context to save CRTC timings information for
outputHead  - Display head to restore the CRTC timings for

RETURNS:
True on success, false on error.

REMARKS:
This function restores the driver CRTC timings information by reading the
current values from disk into the version the driver has loaded in
memory. This is used to restore the timings to those present the last time
the timings were saved to disk.

SEE ALSO:
GA_saveCRTCTimings, GetCRTCTimings, SaveCRTCTimings, SetGlobalRefresh,
GA_saveModeProfile, GA_saveOptions
****************************************************************************/
ibool NAPI GA_restoreCRTCTimings(
    GA_devCtx *dc,
    N_int32 outputHead)
{
    char    filename[PM_MAX_PATH],buf[PM_MAX_PATH];
    void    *data;
    int     size;

    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    CDBX_getCRTCTable(dc->DeviceIndex,outputHead,&data,&size);
    sprintf(buf,CDBX_NAME,outputHead,dc->DeviceIndex);
    strcpy(filename,SNAPConfigPath);
    strcat(filename,buf);
    return CDBX_read(filename,data,size);
}

/****************************************************************************
REMARKS:
Legacy function for backwards compatibility.
{secret}
****************************************************************************/
ibool NAPI GA_restoreCRTCTimingsV1(
    GA_devCtx *dc)
{
    return GA_restoreCRTCTimings(dc,gaActiveHeadPrimary);
}

/****************************************************************************
DESCRIPTION:
Determine if a mode should use double scanning or not.

HEADER:
snap/graphics.h

PARAMETERS:
modeInfo    - Mode information for mode to check

RETURNS:
True if double scan should be used, false if not.

REMARKS:
Determines if we should use a double scan mode for the graphics mode. This
is true if we have a low resolution mode and the mode supports double
scanning.
****************************************************************************/
ibool NAPI GA_useDoubleScan(
    GA_modeInfo *modeInfo)
{
    if (modeInfo->Attributes & gaHaveDoubleScan) {
        if (modeInfo->XResolution == 640 && modeInfo->YResolution == 350)
            return false;
        if (modeInfo->XResolution <= 800 && modeInfo->YResolution <= 384)
            return true;
        }
    return false;
}

/****************************************************************************
DESCRIPTION:
Determine the maximum refresh rate for a particular display mode

HEADER:
snap/graphics.h

PARAMETERS:
dc          - SNAP device context to use
outputHead  - Display head to get maximum refresh rate for
modeInfo    - SuperVGA mode information block for the mode
interlaced  - True if the mode should be interlaced or not
maxRefresh  - Place to store maximum refresh rate in Hz.

REMARKS:
This function computes the maximum refresh rate for the graphics modes
specified by the passed in mode information block. The maximum refresh
rate is computed from the information returned by the SNAP driver,
using CRTC values computed using the VESA GTF formulas.

SEE ALSO:
GA_computeCRTCTimings
****************************************************************************/
void NAPI GA_getMaxRefreshRate(
    GA_devCtx *dc,
    N_int32 outputHead,
    GA_modeInfo *modeInfo,
    N_int32 interlaced,
    float *maxRefresh)
{
    GTF_timings     gtf;
    int             xRes = modeInfo->XResolution;
    int             yRes = modeInfo->YResolution;
    ibool           doubleScan;
    GA_monitor      monitor;

    /* Get attached monitor information */
    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    gInit.GetMonitorInfo(&monitor,outputHead);

    /* Compute the maximum GTF timings */
    doubleScan = GA_useDoubleScan(modeInfo);
    if (doubleScan)
        yRes *= 2;
    GTF_calcTimings(xRes,yRes,modeInfo->MaxPixelClock / 1000000.0,
        GTF_lockPF,false,interlaced,&gtf);
    *maxRefresh = gtf.vFreq;

    /* Now filter the timings based on the monitor operational limits */
    if (monitor.maxResolution != 0xFF) {
        /* Filter on vertical frequency first */
        if ((int)(*maxRefresh + 0.5) > monitor.maxVScan) {
            GTF_calcTimings(xRes,yRes,monitor.maxVScan,GTF_lockVF,false,interlaced,&gtf);
            *maxRefresh = gtf.vFreq;
            }

        /* Filter on horizontal frequency second */
        if ((int)(gtf.hFreq + 0.5) > monitor.maxHScan) {
            GTF_calcTimings(xRes,yRes,monitor.maxHScan,GTF_lockHF,false,interlaced,&gtf);
            *maxRefresh = gtf.vFreq;
            }
        }

    /* Now round the frequency down to the nearest Hz */
    *maxRefresh = (int)*maxRefresh;
}

/****************************************************************************
REMARKS:
Legacy function for backwards compatibility.
{secret}
****************************************************************************/
void NAPI GA_getMaxRefreshRateV1(
    GA_devCtx *dc,
    GA_modeInfo *modeInfo,
    N_int32 interlaced,
    float *maxRefresh)
{
    GA_getMaxRefreshRate(dc,gaActiveHeadPrimary,modeInfo,interlaced,maxRefresh);
}

#ifdef  __WIN32_VXD__
/****************************************************************************
REMARKS:
fabs function for Win32 VxD implementation.
****************************************************************************/
static double fabs(double f)
{
    return (f > 0) ? f : -f;
}
#endif

/****************************************************************************
DESCRIPTION:
Compute the VESA GTF timings for a display mode

HEADER:
snap/graphics.h

PARAMETERS:
dc          - Device context for driver to compute timings for
outputHead  - Display head to compute CRTC timings for
modeInfo    - SuperVGA mode information block for the mode
refreshRate - Desired refresh rate to generate timings for
interlaced  - True if the CRTC timings should be interlaced
crtc        - Place to store the computed CRTC timings
pureGTF     - True to use pure GTF formulas

RETURNS:
True on success, or false if refresh rate is out of range.

REMARKS:
This function computes a set of CRTC timings for the specific graphics
mode at the specified refresh rate. The CRTC timings are computed using
the VESA GTF timing formulas, combined with information returned by the
SNAP driver. Note that the exact refresh rate that you get will not
be exactly what you requested, and the exact value is returned in the
CRTC information block block.

If the pureGTF parameter is set to true, we will leave the CRTC timings
as is after the GTF formulas have computed the appropriate timings for
the refresh rate given the closest available pixel clock. This means that
we may not be able to hit a specific refresh rate exactly, but the CRTC
timings will be GTF compliant. In some cases it may be more beneficial to
hit an exact refresh rate on the nail (such as when viewing sterero images)
and in order to do this we will modify the horizontal and vertical totals
slightly to compensate for the granularity in the pixel clock (which in
some cases may be quite large).

SEE ALSO:
GA_getMaxRefreshRate
****************************************************************************/
ibool NAPI GA_computeCRTCTimings(
    GA_devCtx *dc,
    N_int32 outputHead,
    GA_modeInfo *modeInfo,
    N_int32 refreshRate,
    N_int32 interlaced,
    GA_CRTCInfo *crtc,
    N_int32 pureGTF)
{
    GTF_timings     gtf,bestGTF;
    int             xRes = modeInfo->XResolution;
    int             yRes = modeInfo->YResolution;
    float           closestRefresh,refresh,minRefresh,maxRefresh;
    ulong           clock,closestClock;
    ibool           doubleScan;
    GA_monitor      monitor;

    /* Get gInit functions */
    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");

    /* If we have a double scanned mode, then double the vertical
     * timing parameters.
     */
    doubleScan = GA_useDoubleScan(modeInfo);
    if (doubleScan)
        yRes *= 2;

    /* Now search for the closest refresh rate we can get that is
     * within -1Hz or +3Hz of the desired refresh rate in 0.5Hz increments.
     */
    minRefresh = refreshRate-1;
    maxRefresh = refreshRate+3;
    closestRefresh = 0;
    closestClock = -1;
    for (refresh = minRefresh; refresh < maxRefresh; refresh += 0.5) {
        /* First compute the GTF timings given the specified refresh rate */
        GTF_calcTimings(xRes,yRes,refresh,GTF_lockVF,false,interlaced,&gtf);
        if ((clock = (ulong)(gtf.dotClock * 1000000.0)) > modeInfo->MaxPixelClock)
            break;

        /* Now find the closest pixel clock for the computed timings */
        clock = gInit.GetClosestPixelClock(outputHead,modeInfo->XResolution,modeInfo->YResolution,modeInfo->BitsPerPixel,clock);
        if (clock == (ulong)-1)
            continue;

        /* Now run the GTF timings once more given the closest pixel clock */
        GTF_calcTimings(xRes,yRes,clock / 1000000.0,GTF_lockPF,false,interlaced,&gtf);
        if (fabs(gtf.vFreq-refreshRate) < fabs(closestRefresh-refreshRate)) {
            closestRefresh = gtf.vFreq;
            closestClock = clock;
            bestGTF = gtf;
            }
        }
    if (closestClock == (ulong)-1)
        return false;

    /* Validate that the resolution is valid for the current monitor capabilities */
    gInit.GetMonitorInfo(&monitor,outputHead);
    if (monitor.maxResolution != 0xFF) {
        int hFreq = ((closestClock / bestGTF.h.hTotal) + 500) / 1000;
        int vFreq = closestClock / (bestGTF.v.vTotal * bestGTF.h.hTotal);
        if (hFreq < monitor.minHScan || hFreq > monitor.maxHScan)
            return false;
        if (vFreq < monitor.minVScan || vFreq > monitor.maxVScan)
            return false;
        }

    /* Now return the computed timings to the application */
    crtc->HorizontalTotal = bestGTF.h.hTotal;
    crtc->HorizontalSyncStart = bestGTF.h.hSyncStart;
    crtc->HorizontalSyncEnd = bestGTF.h.hSyncEnd;
    crtc->VerticalTotal = bestGTF.v.vTotal;
    crtc->VerticalSyncStart = bestGTF.v.vSyncStart;
    crtc->VerticalSyncEnd = bestGTF.v.vSyncEnd;
    crtc->PixelClock = closestClock;
    crtc->RefreshRate = (ushort)(bestGTF.vFreq * 100.0);
    crtc->Flags = 0;
    if (interlaced)
        crtc->Flags |= gaInterlaced;
    if (doubleScan)
        crtc->Flags |= gaDoubleScan;
    if (xRes <= 640 && yRes <= 400) {
        /* Get the sync polarities for the mode. For low resolution modes
         * below 640x480, we need to use specific sync polarities since
         * they are used to determine the vertical sizing for compatibility
         * with VGA CRTC timings.
         */
        if (yRes == 350)
            crtc->Flags |= gaVSyncNeg;
        else if (yRes == 200 || yRes == 384 || yRes == 400)
            crtc->Flags |= gaHSyncNeg;
        else
            crtc->Flags |= gaHSyncNeg | gaVSyncNeg;
        }
    else {
        /* Use polarities returned from GTF formulas */
        if (bestGTF.hSyncPol == '-')
            crtc->Flags |= gaHSyncNeg;
        if (bestGTF.vSyncPol == '-')
            crtc->Flags |= gaVSyncNeg;
        }
    (void)pureGTF;
    return true;
}

/****************************************************************************
REMARKS:
Legacy function for backwards compatibility.
{secret}
****************************************************************************/
ibool NAPI GA_computeCRTCTimingsV1(
    GA_devCtx *dc,
    GA_modeInfo *modeInfo,
    N_int32 refreshRate,
    N_int32 interlaced,
    GA_CRTCInfo *crtc,
    N_int32 pureGTF)
{
    return GA_computeCRTCTimings(dc,gaActiveHeadPrimary,modeInfo,
        refreshRate,interlaced,crtc,pureGTF);
}

/****************************************************************************
REMARKS:
Create a new refresh rate in the CRTC database for a specific mode.
****************************************************************************/
static void CreateRefresh(
    GA_devCtx *dc,
    GA_modeInfo *mi,
    N_int32 refresh)
{
    GA_CRTCInfo crtc;
    int         head,numHeads;

    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    numHeads = gInit.GetNumberOfHeads();
    for (head = 0; head < numHeads; head++) {
        if (GA_computeCRTCTimings(dc,head,mi,ABS(refresh),refresh<0,&crtc,false))
            CDBX_setCRTCTimings(dc->DeviceIndex,head,mi->XResolution,mi->YResolution,&crtc,true,true);
        }
}

/****************************************************************************
REMARKS:
Updates the mode profile information in the driver and on disk.
****************************************************************************/
static void UpdateModeTables(
    GA_devCtx *dc,
    ibool saveToDisk)
{
    int head,numHeads;

    gInit.SetModeProfile(&gModes);
    if (saveToDisk) {
        GA_saveModeProfile(dc,&gModes);
        numHeads = gInit.GetNumberOfHeads();
        for (head = 0; head < numHeads; head++)
            GA_saveCRTCTimings(dc,head);
        }
}

/****************************************************************************
DESCRIPTION:
Add a custom display mode to the device driver.

HEADER:
snap/graphics.h

PARAMETERS:
dc          - Device context for driver to add mode to
xRes        - X resolution of the mode to add
yRes        - Y resolution of the mode to add
bits        - Bits per pixel for the mode to add
saveToDisk  - True if the mode profile should be saved to disk

RETURNS:
True on success, or false if the mode could not be added.

REMARKS:
This function is used to add a new display mode to a device. By default
a standard table of display modes is available in the device, however this
function allows the user to create custom display modes and add those to the
SNAP drivers.

Unless you pass a value of true to the saveToDisk parameter, the changes
will only affect the currently loaded copy of SNAP and will not affect
new versions of SNAP that are loaded from disk. To make the change
permanent, you need to set this parameter to true.

Note:   In order to enable support for the newly added mode for VESA VBE
        applications, you must also call the GA_enableVBEMode for the
        same resolution and color depth before it will show up.

SEE ALSO:
GA_addRefresh, GA_delMode, GA_enableVBEMode
****************************************************************************/
ibool NAPI GA_addMode(
    GA_devCtx *dc,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bits,
    N_int32 saveToDisk)
{
    int i;

    // Fail if not licensed
    if (!bDialAMode_Licensed)
        return false;

    // Get init functions
    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    gModes.dwSize = sizeof(gModes);
    GA_loadModeProfile(dc,&gModes);
    if (gModes.m.numModes == 255)
        return false;
    gMi.dwSize = sizeof(gMi);
    if (gInit.GetCustomVideoModeInfo(xRes,yRes,-1,-1,bits,&gMi) != 0)
        return false;
    for (i = 0; i < gModes.m.numModes; i++) {
        if (gModes.m.modeList[i].xRes >= xRes
            && gModes.m.modeList[i].yRes >= yRes
            && gModes.m.modeList[i].bits == bits) {
            if (gModes.m.modeList[i].xRes == xRes
                && gModes.m.modeList[i].yRes == yRes
                && gModes.m.modeList[i].bits == bits)
                return false;
            break;
            }
        }
    if (i < gModes.m.numModes) {
        memmove(&gModes.m.modeList[i+1],&gModes.m.modeList[i],
            (gModes.m.numModes-i) * sizeof(gModes.m.modeList[0]));
        gModes.m.modeList[i].xRes = xRes;
        gModes.m.modeList[i].yRes = yRes;
        gModes.m.modeList[i].bits = bits;
        gModes.m.numModes++;
        if (xRes >= 960 && yRes >= 720) {
            CreateRefresh(dc,&gMi,-87);
            CreateRefresh(dc,&gMi,-97);
            CreateRefresh(dc,&gMi,-120);
            }
        CreateRefresh(dc,&gMi,100);
        CreateRefresh(dc,&gMi,95);
        CreateRefresh(dc,&gMi,90);
        CreateRefresh(dc,&gMi,85);
        CreateRefresh(dc,&gMi,80);
        CreateRefresh(dc,&gMi,75);
        CreateRefresh(dc,&gMi,70);
        if (xRes >= 640 && yRes >= 480) {
            CreateRefresh(dc,&gMi,65);
            CreateRefresh(dc,&gMi,60);
            }
        UpdateModeTables(dc,saveToDisk);

        // Re-create the AvailableModes table
        GA_createModeList(dc,gModes.m.numModes);
        return true;
        }
    return false;
}

/****************************************************************************
DESCRIPTION:
Add a custom refresh rate for a mode to the CRTC database.

HEADER:
snap/graphics.h

PARAMETERS:
dc          - Device context for driver to add mode to
xRes        - X resolution of the mode to add
yRes        - Y resolution of the mode to add
refresh     - New refresh rate to add
saveToDisk  - True if the mode profile should be saved to disk

RETURNS:
True on success, or false if the refresh rate could not be added.

REMARKS:
This function is used to add a custom refresh rate to the CRTC database
for a device. By default a standard databse of known display modes is
available, but this function allows the user to create custom refresh rates
and add those to the SNAP drivers.

Unless you pass a value of true to the saveToDisk parameter, the changes
will only affect the currently loaded copy of SNAP and will not affect
new versions of SNAP that are loaded from disk. To make the change
permanent, you need to set this parameter to true.

SEE ALSO:
GA_addMode, GA_delMode
****************************************************************************/
ibool NAPI GA_addRefresh(
    GA_devCtx *dc,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 refresh,
    N_int32 saveToDisk)
{
    // Fail if not licensed
    if (!bDialAMode_Licensed)
        return false;

    // Get init functions
    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    gModes.dwSize = sizeof(gModes);
    GA_loadModeProfile(dc,&gModes);
    gMi.dwSize = sizeof(gMi);
    if (gInit.GetCustomVideoModeInfo(xRes,yRes,-1,-1,8,&gMi) != 0)
        return false;
    CreateRefresh(dc,&gMi,refresh);
    UpdateModeTables(dc,saveToDisk);
    return true;
}

/****************************************************************************
DESCRIPTION:
Delete a display mode for the device.

HEADER:
snap/graphics.h

PARAMETERS:
dc          - Device context for driver to add mode to
xRes        - X resolution of the mode to add
yRes        - Y resolution of the mode to add
bits        - Bits per pixel for the mode to add
saveToDisk  - True if the mode profile should be saved to disk

RETURNS:
True on success, or false if the mode could not be deleted.

REMARKS:
This function is used to add a delete a display mode for a device. By
default a standard table of display modes is available in the device,
however this function allows the user to remove unwanted display modes
from the SNAP drivers.

Unless you pass a value of true to the saveToDisk parameter, the changes
will only affect the currently loaded copy of SNAP and will not affect
new versions of SNAP that are loaded from disk. To make the change
permanent, you need to set this parameter to true.

SEE ALSO:
GA_addMode, GA_addRefresh, GA_disableVBEMode
****************************************************************************/
ibool NAPI GA_delMode(
    GA_devCtx *dc,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bits,
    N_int32 saveToDisk)
{
    int i;

    // Fail if not licensed
    if (!bDialAMode_Licensed)
        return false;

    // Get init functions
    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    gModes.dwSize = sizeof(gModes);
    GA_loadModeProfile(dc,&gModes);
    for (i = 0; i < gModes.m.numModes; i++) {
        if (gModes.m.modeList[i].xRes == xRes
            && gModes.m.modeList[i].yRes == yRes
            && gModes.m.modeList[i].bits == bits) {
            break;
            }
        }
    if (i < gModes.m.numModes) {
        memmove(&gModes.m.modeList[i],&gModes.m.modeList[i+1],
            (gModes.m.numModes-i-1) * sizeof(gModes.m.modeList[0]));
        gModes.m.numModes--;
        gInit.SetModeProfile(&gModes);
        if (saveToDisk)
            GA_saveModeProfile(dc,&gModes);

        // Re-create the AvailableModes table
        GA_createModeList(dc,gModes.m.numModes);
        return true;
        }
    return false;
}

/****************************************************************************
DESCRIPTION:
Enable a display mode for use by VESA VBE applications.

HEADER:
snap/graphics.h

PARAMETERS:
dc          - Device context for driver to add mode to
xRes        - X resolution of the mode to add
yRes        - Y resolution of the mode to add
bits        - Bits per pixel for the mode to add
saveToDisk  - True if the mode profile should be saved to disk

RETURNS:
True on success, or false if the mode could not be added.

REMARKS:
This function is used to enable a specific display mode for use by VESA
VBE applications in a DOS box environment. Many early VESA applications do
not correctly handle large mode lists returned by the VBE driver, so only
a specific subset of modes are actually reported to the VBE application
by the SNAP VBE emulation driver. This function is used to enable
support for specific modes that the user desires on a mode by mode basis.

Unless you pass a value of true to the saveToDisk parameter, the changes
will only affect the currently loaded copy of SNAP and will not affect
new versions of SNAP that are loaded from disk. To make the change
permanent, you need to set this parameter to true.

Note:   Modes added to this list only affect the filtering mechanism for
        VESA VBE applications. A mode will not be available unless it is
        first added to the standard mode list for the SNAP driver using
        GA_addMode (or if the mode is a standard mode already supported).

SEE ALSO:
GA_disableVBEMode, GA_addMode, GA_delMode
****************************************************************************/
ibool NAPI GA_enableVBEMode(
    GA_devCtx *dc,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bits,
    N_int32 saveToDisk)
{
    int i;

    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    gModes.dwSize = sizeof(gModes);
    GA_loadModeProfile(dc,&gModes);
    if (gModes.vbe.numModes == 255)
        return false;
    for (i = 0; i < gModes.vbe.numModes; i++) {
        if (gModes.vbe.modeList[i].xRes >= xRes
            && gModes.vbe.modeList[i].yRes >= yRes
            && gModes.vbe.modeList[i].bits == bits) {
            if (gModes.vbe.modeList[i].xRes == xRes
                && gModes.vbe.modeList[i].yRes == yRes
                && gModes.vbe.modeList[i].bits == bits)
                return false;
            break;
            }
        }
    if (i < gModes.vbe.numModes) {
        memmove(&gModes.vbe.modeList[i+1],&gModes.vbe.modeList[i],
            (gModes.vbe.numModes-i) * sizeof(gModes.vbe.modeList[0]));
        gModes.vbe.modeList[i].xRes = xRes;
        gModes.vbe.modeList[i].yRes = yRes;
        gModes.vbe.modeList[i].bits = bits;
        gModes.vbe.numModes++;
        gInit.SetModeProfile(&gModes);
        if (saveToDisk)
            GA_saveModeProfile(dc,&gModes);
        return true;
        }
    return false;
}

/****************************************************************************
DESCRIPTION:
Disable a display mode for use by VESA VBE applications.

HEADER:
snap/graphics.h

PARAMETERS:
dc          - Device context for driver to add mode to
xRes        - X resolution of the mode to add
yRes        - Y resolution of the mode to add
bits        - Bits per pixel for the mode to add
saveToDisk  - True if the mode profile should be saved to disk

RETURNS:
True on success, or false if the mode could not be deleted.

REMARKS:
This function is used to disable a specific display mode for use by VESA
VBE applications in a DOS box environment.

Unless you pass a value of true to the saveToDisk parameter, the changes
will only affect the currently loaded copy of SNAP and will not affect
new versions of SNAP that are loaded from disk. To make the change
permanent, you need to set this parameter to true.

SEE ALSO:
GA_enableVBEMode, GA_addMode, GA_delMode
****************************************************************************/
ibool NAPI GA_disableVBEMode(
    GA_devCtx *dc,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bits,
    N_int32 saveToDisk)
{
    int i;

    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        PM_fatalError("Unable to get device driver functions!");
    gModes.dwSize = sizeof(gModes);
    GA_loadModeProfile(dc,&gModes);
    for (i = 0; i < gModes.vbe.numModes; i++) {
        if (gModes.vbe.modeList[i].xRes == xRes
            && gModes.vbe.modeList[i].yRes == yRes
            && gModes.vbe.modeList[i].bits == bits) {
            break;
            }
        }
    if (i < gModes.vbe.numModes) {
        memmove(&gModes.vbe.modeList[i],&gModes.vbe.modeList[i+1],
            (gModes.vbe.numModes-i-1) * sizeof(gModes.vbe.modeList[0]));
        gModes.vbe.numModes--;
        gInit.SetModeProfile(&gModes);
        if (saveToDisk)
            GA_saveModeProfile(dc,&gModes);
        return true;
        }
    return false;
}

/****************************************************************************
DESCRIPTION:
Returns the error code for GA_loadDriver if the function failed.

HEADER:
snap/graphics.h

RETURNS:
Status code for the GA_loadDriver function.

REMARKS:
Returns the error code for GA_loadDriver if the function failed.

SEE ALSO:
GA_errorMsg
****************************************************************************/
int NAPI GA_status(void)
{ return status; }

/****************************************************************************
DESCRIPTION:
Returns a descriptive string for a device loader error code.

HEADER:
snap/graphics.h

PARAMETERS:
status  - Status code to get string for

RETURNS:
String describing error condition.

REMARKS:
Returns a descriptive string for a device loader error code.

NOTE:   The string is returned in a temporary buffer that is reused. Hence
        the calling application should copy the returned string to a
        different buffer immediately after calling this function and before
        calling any other SNAP functions.

SEE ALSO:
GA_status
****************************************************************************/
const char * NAPI GA_errorMsg(
    N_int32 status)
{
    const char *msg[] = {
        "No error",
        "Hardware not detected",
        "Hardware has not been POST'ed by the BIOS",
        "Driver file not found",
        "File loaded was not a driver file",
        "Not enough memory to load driver",
        "Driver file is an older version",
        "Could not map physical memory areas",
        "General I/O error",
        "Could not hook requested hardware IRQ",
        "Driver is not certified",
        "Internal device driver error",
        "Not enough memory to complete operation",
        "Not enough spare resources to complete operation",
        "Invalid parameter passed to function",
        "No AGP support services found",
        "Invalid license!",
        "Feature is not licensed",
        };
    if (status >= nOK && status < nMaxError)
        return msg[status];
    return "Unknown error!";
}

/****************************************************************************
DESCRIPTION:
Returns the number of days left in the evaluation license

HEADER:
snap/graphics.h

RETURNS:
Number of days left in evaluation license, -1 if registered, 0 if timed out.

REMARKS:
This function checks the number of days left in the evaluation period. If the
product has been registered this function returns a value of -1, otherwise
it returns the number of days left in the evaluation period.
****************************************************************************/
int NAPI GA_getDaysLeft(
    N_int32 shared)
{
    (void)shared;
    return -1;
}

/****************************************************************************
DESCRIPTION:
Returns true if the product is an OEM licensed product.

HEADER:
snap/graphics.h

PARAMETERS:
shared  - True if the device driver should be loaded into shared memory

RETURNS:
True if an OEM licensed product, false if not.

REMARKS:
This function can be used to determine if the product is an OEM licensed
version of SNAP, or if it is a retail end user version of the
product. Retail end user versions always require a registration code.
****************************************************************************/
int NAPI GA_isOEMVersion(
    ibool shared)
{
    (void)shared;
    return false;
}

/****************************************************************************
DESCRIPTION:
Returns true if the product is the Lite retail product.

HEADER:
snap/graphics.h

PARAMETERS:
shared  - True if the device driver should be loaded into shared memory

RETURNS:
True if the Lite retail product, false if not.

REMARKS:
This function can be used to determine if the product is the Lite version
of the retail product, or the full Profesional version. OEM versions of
the product always return false for this function.
****************************************************************************/
int NAPI GA_isLiteVersion(ibool shared)
{
    (void)shared;
    return false;
}

/****************************************************************************
DESCRIPTION:
Returns the user serial number as an ASCII string

HEADER:
snap/graphics.h

PARAMETERS:
shared  - True if the device driver should be loaded into shared memory

RETURNS:
Serial number as an ASCII string.

REMARKS:
This function can be used to get an ASCII string version of the user serial
number that can be displayed for the user. This is a printable version
of the serial number, and does not contain the actual registration code
information (ie: it is a serial number that can be used to look up the
user in a customer database).

OEM versions of the product always return "N/A" for this function.

NOTE:   The string is returned in a temporary buffer that is reused. Hence
        the calling application should copy the returned string to a
        different buffer immediately after calling this function and before
        calling any other SNAP functions.
****************************************************************************/
const char * NAPI GA_getDisplaySerialNo(ibool shared)
{
    (void)shared;
    return "N/A";
}

/****************************************************************************
DESCRIPTION:
Returns the user name as an ASCII string

HEADER:
snap/graphics.h

PARAMETERS:
shared  - True if the device driver should be loaded into shared memory

RETURNS:
User name or OEM company name as an ASCII string.

REMARKS:
This function can be used to get an ASCII string version of the user name
or OEM company name (for OEM versons) that that can be displayed for the
user.

NOTE:   The string is returned in a temporary buffer that is reused. Hence
        the calling application should copy the returned string to a
        different buffer immediately after calling this function and before
        calling any other SNAP functions.
****************************************************************************/
const char * NAPI GA_getDisplayUserName(ibool shared)
{
    (void)shared;
    return "SciTech Software, Inc.";
}

/****************************************************************************
DESCRIPTION:
Returns the list of licensed device types for the license files

HEADER:
snap/graphics.h

PARAMETERS:
shared  - True if the device driver should be loaded into shared memory

RETURNS:
List of licensed device types, NULL if not an OEM licensed product.

REMARKS:
This function can be used to get a numerical list of all licensed device
types stored in the license file. This is used by control panel applets
like the DDStereo control panel to determine what stereo glasses types
are licensed by the OEM vendor. This list is always terminated with
a value of 0x10101010.

NOTE:   The array returned by this function lives in a temporary buffer that
        is reused. Hence the calling application should copy the returned
        array to a different buffer immediately after calling this function
        and before calling any other SNAP functions.
****************************************************************************/
N_uint32 * NAPI GA_getLicensedDevices(
    ibool shared)
{
    (void)shared;
    return NULL;
}

/****************************************************************************
DESCRIPTION:
Returns the function pointers for the specified function group.

HEADER:
snap/graphics.h

PARAMETERS:
dc          - Device driver to return the functions for
id          - Identifier for the function group to get pointers for
funcs       - Pointer to function block to fill in

RETURNS:
True if the requested function group is available, false if not.

REMARKS:
This function is the main function that is used by the graphics application
code to get a block of function pointers for a specified function group.
The function groups are defined GA_funcGroupsType enumeration, and breaks
up the functions in the device driver API into groups of logically similar
functions. For instance to get the block of functions for the hardware
cursor, you would call this function with the GA_GET_CURSORFUNCS identifier.

All application and device driver code must call this function and /never/
call the GA_devCtx->loader.QueryFunctions member directly. This is important
because in many cases special procedures must be followed to allow ring 3
application code to directly call the function pointers returned by the
internal function. The GA_devCtx version always returns the actual device
driver functions which may require that they be called at ring 0 on some
platforms. This function insulates the developer from worrying about this
as the functions returned by this function are always safe to be called
from the context that the code is compiled (ie: this function should be used
for all ring 0 device driver and ring 3 application level code).

Note:   To allow for future compatibility, all function blocks begin with
        a dwSize member. The caller is expected to fill in the dwSize
        member with the size of the function block being retrieved /before/
        calling GA_queryFunctions. If the driver exports more functions than
        the application knows about, only a subset of the functions are
        copied to the application. If the application expects more functions
        than the driver provides, the non-existant functions are set to NULL
        pointers by GA_queryFunctions, and the remainder copies from the driver.

Note:   This mechanism also provides for a clean and simple upgrade path
        for future drivers, while ensuring maximum compatibility with
        existing specifications. New functions for a particular group can
        simply be added to the end of the function group to extend that
        group. Totally new function groups can be added by defining new
        identifiers for that function group, and older drivers will
        return a NULL if that function group is requested. Finally if a
        function group requires a complete redesign to achieve maximum
        peformance for next generation hardware, a new extended function
        group can be defined (and drivers can continue to export the older
        and slower function group for backwards compatibility).

SEE ALSO:
REF2D_queryFunctions
****************************************************************************/
ibool NAPI GA_queryFunctions(
    GA_devCtx *dc,
    N_uint32 id,
    void _FAR_ *funcs)
{
    // By default we always call the device driver version. Application level
    // loader libraries will override this function as necessary.
    return dc->loader.QueryFunctions(id,0,funcs);
}

/****************************************************************************
DESCRIPTION:
Returns the function pointers for the specified ref2d function group.

HEADER:
snap/graphics.h

PARAMETERS:
ref2d       - Ref2d driver to return the functions for
id          - Identifier for the function group to get pointers for
funcs       - Pointer to function block to fill in

RETURNS:
True if the requested function group is available, false if not.

REMARKS:
This function is the similar to GA_queryFunctions function except that the
functions are queried via the 2d reference rasteriser library. This is the
function that application level code should use to get access to the SNAP
rendering functions that are fleshed out with software rendered functions
as necessary.

All application and device driver code must call this function and *never*
call the REF2D_driver->QueryFunctions member directly. This is important
because in many cases special procedures must followed to allow ring 3
application code to directly call the function pointers returned by this
function. See GA_queryFunctions for more information.

Note:   To allow for future compatibility, all function blocks begin with
        a dwSize member. The caller is expected to fill in the dwSize
        member with the size of the function block being retrieved /before/
        calling REF2D_queryFunctions. If the driver exports more functions than
        the application knows about, only a subset of the functions are
        copied to the application. If the application expects more functions
        than the driver provides, the non-existant functions are set to NULL
        pointers by REF2D_queryFunctions, and the remainder copies from the
        driver.

SEE ALSO:
GA_queryFunctions
****************************************************************************/
ibool NAPI REF2D_queryFunctions(
    REF2D_driver *ref2d,
    N_uint32 id,
    void _FAR_ *funcs)
{
    // By default we always call the device driver version. Application level
    // loader libraries will override this function as necessary.
    return ref2d->QueryFunctions(id,funcs);
}

/****************************************************************************
REMARKS:
Internal function to check for certified SNAP driver for DDC operation
under Windows. For all other OS'es this function bails out with nOK.
{secret}
****************************************************************************/
int _GA_checkDDCCertification(
    GA_devCtx *dc)
{
    GA_certifyChipInfo* pChip;
    int                 i;

    // Always certified for non-Win32 platforms
    if (!_IS_OS_CLASS(_OS_CLASS_WIN32))
        return nOK;

    // Get SNAP init functions
    gInit.dwSize = sizeof(gInit);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&gInit))
        return nCorruptDriver;

    // Get SNAP options and configuration info
    gOpt.dwSize = sizeof(gOpt);
    gInit.GetOptions(&gOpt);
    gInfo.dwSize = sizeof(gInfo);
    gInit.GetConfigInfo(&gInfo);

    // Check for general certification status
    if (!gInfo.Certified) {
        if (globalOpt.bAllowNonCertified)
            return nOK;
        else
            return nNotCertified;
        }
    else {
        // Check for chipset-specific certification status
        gCert.dwSize = sizeof(gCert);
        gInit.GetCertifyInfo(&gCert);
        pChip = gCert.CertifiedCards;
        for (i = 0; i < gCert.NumCertifiedChips; i++, pChip++) {
            if (strnicmp(pChip->ChipsetName, gInfo.ChipsetName, sizeof(pChip->ChipsetName)) == 0) {
                if (pChip->CertifyFlags & gaCertifiedDDC)
                    return nOK;
                if (globalOpt.bAllowNonCertified)
                    return nOK;
                return nNotCertified;
                }
            }
        }

    // If we made it to here, no positive ID was made
    if (globalOpt.bAllowNonCertified)
        return nOK;
    return nDriverNotFound;
}

/****************************************************************************
REMARKS:
Internal function to set a pointer to the external application exports
table so that it can be overriden if necessary. We use this to override
the pointers to the DDC and MCS functions when the GDI and DirectX drivers
are loaded so that they will call directly into the ring 0 drivers.
{secret}
****************************************************************************/
void NAPI GA_setExportTable(
    GA_exports *table)
{
    exportTable = table;
}

/****************************************************************************
DESCRIPTION:
Sets the minimum driver version supported

HEADER:
snap/graphics.h

PARAMETERS:
version         - Minimum driver version to accept
allowFallback   - Allow the fallback drivers to load
shared          - True if the device driver should be loaded into shared memory

REMARKS:
This function can be used to exclude early device driver versions if they
are known not to have the necessary feature support from being loaded.
This function will stop the drivers from loading early in the driver
load process, so as to avoid any problems that might crop up during
loading and initialisation.

The allowFallback parameter is used to allow the loading of the VBE/Core
and VGA fallback drivers. If this parameter is false, then the VBE/Core
and VGA drivers will fail to load. If this parameter is not false then
they will load as normal.

Note:   The default minimum version is 0 and the default value for
        allowFallback is true.

SEE ALSO:
GA_loadDriver
****************************************************************************/
void NAPI GA_setMinimumDriverVersion(
    N_uint32 version,
    N_int32 allowFallback,
    N_int32 shared)
{
    (void)shared;
    allowFallbackDrivers = allowFallback;
    minDriverVersion = version;
}

/****************************************************************************
DESCRIPTION:
Load the OpenGL compatible rendering engine for the hardware device

HEADER:
snap/graphics.h

PARAMETERS:
hwctx   - SNAP hardware device driver context to load OpenGL engine for
shared  - True to load engine into shared memory

REMARKS:
This function loads a copy of the SciTech SNAP OpenGL compatible rendering
engine from disk for use with the specified hardware device driver. If the
hardware device supports 3D rendering and there is an available hardware
3D device driver, the OpenGL engine will support hardware acceleration. If
not the OpenGL engine will be implemented entirely in software.

SEE ALSO:
GA_glQueryFunctions, GA_glUnloadDriver
****************************************************************************/
GA_glCtx * NAPI GA_glLoadDriver(
    GA_devCtx *hwCtx,
    ibool shared)
{
    GA_deviceEntry      *pga = &ga[hwCtx->DeviceIndex];
    ulong               size;
    GA_gl3DInitDriver_t GA_gl3DInitDriver;

    if ((pga->hModGL = GA_loadLibrary(GL3D_NAME,&size,shared)) == NULL)
        return NULL;
    if ((GA_gl3DInitDriver = (GA_gl3DInitDriver_t)PE_getProcAddress(pga->hModGL,"_GA_gl3DInitDriver")) == NULL)
        PM_fatalError("Unable to get BPD file entry point!");
    if ((pga->glCtx = GA_gl3DInitDriver(shared,SNAPConfigPath,hwCtx,pga->PCIInfo,
            &_PM_imports,&_GL3D_imports,licensedOptions,&globalOpt)) == NULL) {
        PE_freeLibrary(pga->hModGL);
        pga->hModGL = NULL;
        return NULL;
        }
    return pga->glCtx;
}

/****************************************************************************
DESCRIPTION:
Returns the function pointers for the specified OpenGL engine function group.

HEADER:
snap/graphics.h

PARAMETERS:
glCtx   - OpenGL rendering engine context to return the functions for
id      - Identifier for the function group to get pointers for
funcs   - Pointer to function block to fill in

RETURNS:
True if the requested function group is available, false if not.

REMARKS:
This function is the main function that is used by the graphics application
code to get a block of function pointers for a specified function group from
the OpenGL compatible rendering engine. The function groups are defined
GA_glFuncGroupsType enumeration.

All application and device driver code must call this function and /never/
call the GA_glCtx->QueryFunctions member directly. This is important because
in some cases special procedures must be followed to allow ring 3
application code to directly call the function pointers returned by the
internal function.

Note:   To allow for future compatibility, all function blocks begin with
        a dwSize member. The caller is expected to fill in the dwSize
        member with the size of the function block being retrieved /before/
        calling GA_glQueryFunctions. If the driver exports more functions than
        the application knows about, only a subset of the functions are
        copied to the application. If the application expects more functions
        than the driver provides, the non-existant functions are set to NULL
        pointers by GA_glQueryFunctions, and the remainder copies from the
        driver.

Note:   This mechanism also provides for a clean and simple upgrade path
        for future drivers, while ensuring maximum compatibility with
        existing specifications. New functions for a particular group can
        simply be added to the end of the function group to extend that
        group. Totally new function groups can be added by defining new
        identifiers for that function group, and older drivers will
        return a NULL if that function group is requested. Finally if a
        function group requires a complete redesign to achieve maximum
        peformance for next generation hardware, a new extended function
        group can be defined (and drivers can continue to export the older
        and slower function group for backwards compatibility).
****************************************************************************/
ibool NAPI GA_glQueryFunctions(
    GA_glCtx *glCtx,
    N_uint32 id,
    void *funcs)
{
    return glCtx->QueryFunctions(id,funcs);
}

/****************************************************************************
DESCRIPTION:
Unload the OpenGL compatible rendering engine

HEADER:
snap/graphics.h

PARAMETERS:
glCtx   - OpenGL rendering engine to unload

REMARKS:
This function is used to unload the OpenGL compatible rendering engine
when the engine is not longer needed.
****************************************************************************/
void NAPI GA_glUnloadDriver(
    GA_glCtx *glCtx)
{
    int i;

    for (i = 0; i < NumDevices; i++) {
        GA_deviceEntry *pga = &ga[i];
        if (pga->glCtx = glCtx) {
            glCtx->Unload();
            if (pga->hModGL)
                PE_freeLibrary(pga->hModGL);
            pga->glCtx = NULL;
            pga->hModGL = NULL;
            }
        }
}

#ifdef __DRIVER__
/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler for non-imported N_imports.
****************************************************************************/
static void _N_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported N_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler for non-imported GA_imports.
****************************************************************************/
static void _GA_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported GA_imports import function called! Please re-compile!\n");
}

/****************************************************************************
PARAMETERS:
shared  - True if the device driver should be loaded into shared memory
path    - Path to Binary Portable DLL
bpdname - Name of the Binary Portable DLL
pmImp   - PM library imports
nImp    - Generic SNAP imports
gaImp   - Graphics Architecture specific imports

RETURNS:
Pointer to exported function list

REMARKS:
This function initialises the SNAP Graphics Architecture loader library
and returns the list of loader library exported functions.
{secret}
****************************************************************************/
GA_exports * _CEXPORT GA_initLibrary(
    N_int32 shared,
    const char *path,
    const char *bpdname,
    PM_imports *pmImp,
    N_imports *nImp,
    GA_imports *gaImp)
{
    int     i,max;
    ulong   *p;
    char    drive[PM_MAX_DRIVE];
    char    dir[PM_MAX_PATH];

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_PM_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;
    max = sizeof(_N_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_N_imports; i < max; i++)
        *p++ = (ulong)_N_fatalErrorHandler;
    max = sizeof(_GA_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_GA_imports; i < max; i++)
        *p++ = (ulong)_GA_fatalErrorHandler;

    // Now copy all our imported functions
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    memcpy(&_N_imports,nImp,MIN(sizeof(_N_imports),nImp->dwSize));
    memcpy(&_GA_imports,gaImp,MIN(sizeof(_GA_imports),gaImp->dwSize));
    PM_splitpath(bpdname,drive,dir,NULL,NULL);
    PM_makepath(SNAPDriverPath,drive,dir,NULL,NULL);
    PM_backslash(SNAPDriverPath);
    strcat(SNAPDriverPath,"graphics");
    PM_backslash(SNAPDriverPath);
    strcpy(SNAPConfigPath,path);
    PM_backslash(SNAPConfigPath);
#ifndef CHECKED
    if (!FLAT_init(bpdname))
        return false;
#endif
    return &__GA_exports;
}
#endif

#if defined(TEST_HARNESS) && defined(DEBUG_X86EMU)
/* {secret} */
void printk(
    const char *fmt,
    ...)
{
    va_list     argptr;

    va_start(argptr, fmt);
    vfprintf(logfile, fmt, argptr);
    va_end(argptr);
    fflush(logfile);
}

#endif

