/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  Main SciTech SNAP Graphics OpenGL compatible 3D
*               driver module.
*
****************************************************************************/

#include "snapmesa.h"
#include "snap\common\io.h"

/*------------------------- Implementation --------------------------------*/

/**********************************************************************/
/*** Private Device Driver Functions                                ***/
/**********************************************************************/

/****************************************************************************
REMARKS:
Returns driver specific strings to the application
****************************************************************************/
static const GLubyte * get_string(
    GLcontext *ctx,
    GLenum name)
{
    static char rendererString[256];

    switch (name) {
        case GL_RENDERER:
            sprintf(rendererString, RENDERER_STR " %s%s%s%s",
                _mesa_x86_cpu_features  ? "/x86"        : "",
                cpu_has_mmx             ? "/MMX"        : "",
                cpu_has_3dnow           ? "/3DNow!"     : "",
                cpu_has_xmm             ? "/SSE"        : "");
            return (const GLubyte *)rendererString;
        case GL_VENDOR:
            return (const GLubyte *)SCITECH_STR;
        default:
            return NULL;
        }
    (void)ctx;
}

/****************************************************************************
REMARKS:
Update the internal state
****************************************************************************/
static void update_state(
    GLcontext *ctx,
    GLuint new_state)
{
    /* Easy - just propogate */
    _swrast_InvalidateState(ctx, new_state);
    _swsetup_InvalidateState(ctx, new_state);
    _ac_InvalidateState(ctx,new_state);
    _tnl_InvalidateState(ctx,new_state);
}

/****************************************************************************
REMARKS:
Specifies the current buffer for span/pixel writing/reading. buffer indicates
which window to write to / read from.  Normally, this'll be the buffer
currently bound to the context, but it doesn't have to be! bufferBit indicates
which color buffer, one of:
   FRONT_LEFT_BIT - this buffer always exists
   BACK_LEFT_BIT - when double buffering
   FRONT_RIGHT_BIT - when using stereo
   BACK_RIGHT_BIT - when using stereo and double buffering
   AUXn_BIT - if aux buffers are implemented
****************************************************************************/
static void set_buffer(
    GLcontext *ctx,
    GLframebuffer *buffer,
    GLuint bufferBit)
{
    GA_glContext    *snap = SNAP_CONTEXT(ctx);

    /* Separate read buffer not supported */
    ASSERT(buffer == ctx->DrawBuffer);
    if (snap->isSysMem)
        ASSERT(bufferBit == FRONT_LEFT_BIT);
    else {
        switch (bufferBit) {
            case FRONT_LEFT_BIT:
                if (snap->drawBuf != snap->frontBuf) {
                    snap->drawBuf = snap->frontBuf;
                    GL.buf.SetActiveBuffer(snap->drawBuf);
                    }
                break;
            case FRONT_RIGHT_BIT:
                if (snap->frontRightBuf && snap->drawBuf != snap->frontRightBuf) {
                    snap->drawBuf = snap->frontRightBuf;
                    GL.buf.SetActiveBuffer(snap->drawBuf);
                    }
                break;
            case BACK_LEFT_BIT:
                if (snap->backBuf && snap->drawBuf != snap->backBuf) {
                    snap->drawBuf = snap->backBuf;
                    GL.buf.SetActiveBuffer(snap->drawBuf);
                    }
                break;
            case BACK_RIGHT_BIT:
                if (snap->backRightBuf && snap->drawBuf != snap->backRightBuf) {
                    snap->drawBuf = snap->backRightBuf;
                    GL.buf.SetActiveBuffer(snap->drawBuf);
                    }
                break;
            }
        }
}

/****************************************************************************
REMARKS:
Return the buffer size for the context
****************************************************************************/
static void get_buffer_size(
    GLframebuffer *buffer,
    GLuint *width,
    GLuint *height)
{
    GLcontext *ctx = (GLcontext*)_glapi_get_context();
    if (ctx) {
        GA_glContext *snap = SNAP_CONTEXT(ctx);
        *width = snap->width;
        *height = snap->height;
        }
    (void)buffer;
}

/****************************************************************************
REMARKS:
Sets up the buffer for software rendering. We lock the buffer and cache
the buffer surface pointer.
****************************************************************************/
static void span_render_start(
    GLcontext *ctx)
{
    GA_glContext *snap = SNAP_CONTEXT(ctx);
    GL.buf.LockBuffer(snap->drawBuf);
    snap->bottom = snap->drawBuf->Height-1;
    snap->surface = snap->drawBuf->Surface;
    snap->stride = snap->drawBuf->Stride;
}

/****************************************************************************
REMARKS:
Finish software rendering for the buffer
****************************************************************************/
static void span_render_finish(
    GLcontext *ctx)
{
    GA_glContext *snap = SNAP_CONTEXT(ctx);
    GL.buf.UnlockBuffer(snap->drawBuf);
}

/****************************************************************************
REMARKS:
Returns the value representing the color. The value is converted from 24 bit
RGB space into the appropriate color for the video mode.
****************************************************************************/
static GA_color pack_rgba_color(
    GLubyte r,
    GLubyte g,
    GLubyte b,
    GLubyte a,
    GA_pixelFormat *pf)
{
    return ((GLuint)((r >> pf->RedAdjust)   & pf->RedMask)   << pf->RedPosition)
         | ((GLuint)((g >> pf->GreenAdjust) & pf->GreenMask) << pf->GreenPosition)
         | ((GLuint)((b >> pf->BlueAdjust)  & pf->BlueMask)  << pf->BluePosition)
         | ((GLuint)((a >> pf->AlphaAdjust) & pf->AlphaMask) << pf->AlphaPosition);
}

/****************************************************************************
REMARKS:
Buffer clearing function. Here we clear the color buffers with the SciTech
SNAP driver functions, which provide much higher performance than the
software Mesa versions which use the generic span functions. If the
buffers are in video memory, we can clear them with the hardware accelerator.
****************************************************************************/
static void clear(
    GLcontext *ctx,
    GLbitfield mask,
    GLboolean all,
    GLint x,
    GLint y,
    GLint width,
    GLint height)
{
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
    GLuint          colorMask,bufferBit;
    GA_color        clearColor;

    /* Sanity check - we only have a front-left buffer for system memory */
    if (snap->isSysMem) {
        ASSERT((mask & (DD_FRONT_RIGHT_BIT |
                        DD_BACK_LEFT_BIT |
                        DD_BACK_RIGHT_BIT)) == 0);
        }

    /* Use optimized clear for common cases */
    if (mask & ctx->Color._DrawDestMask) {
        GLint       x = ctx->DrawBuffer->_Xmin;
        GLint       y = (snap->bottom - ctx->DrawBuffer->_Ymax + 1);
        GLint       height = ctx->DrawBuffer->_Ymax - ctx->DrawBuffer->_Ymin;
        GLint       width  = ctx->DrawBuffer->_Xmax - ctx->DrawBuffer->_Xmin;

        /* Find the color to clear the buffers with */
        if (ctx->Visual.rgbMode) {
            GLchan c[4];
            CLAMPED_FLOAT_TO_CHAN(c[RCOMP], ctx->Color.ClearColor[0]);
            CLAMPED_FLOAT_TO_CHAN(c[GCOMP], ctx->Color.ClearColor[1]);
            CLAMPED_FLOAT_TO_CHAN(c[BCOMP], ctx->Color.ClearColor[2]);
            CLAMPED_FLOAT_TO_CHAN(c[ACOMP], ctx->Color.ClearColor[3]);
            clearColor = pack_rgba_color(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP],&snap->pf);
            colorMask = *((GLuint*)&ctx->Color.ColorMask);
            }
        else {
            clearColor = ctx->Color.ClearIndex;
            colorMask = ctx->Color.IndexMask;
            }

        if (snap->isSysMem) {
            /* Clear whole system memory buffer */
            if (colorMask == 0xFFFFFFFF) {
                if (snap->bitsPerPixel == 8) {
                    uchar *base = snap->surface + (snap->stride * y) + x;
                    clearColor = mkcolor8(clearColor);
                    while (height--) {
                        int     count = width;
                        uchar   *p = base;

                        if (count >= 8) {
                            while (((uint)p & 3) != 0) {
                                *p++ = clearColor;
                                count--;
                                }
                            rep_stosdb(p,count,clearColor);
                            }
                        else
                            rep_stosb(p,count,clearColor);
                        base += snap->stride;
                        }
                    }
                else if (snap->bitsPerPixel <= 16) {
                    uchar *base = snap->surface + (snap->stride * y) + x * 2;
                    clearColor = mkcolor16(clearColor);
                    while (height--) {
                        int     count = width;
                        ushort  *p = (ushort*)base;

                        if (count >= 4) {
                            while (((uint)p & 1) != 0) {
                                *p++ = clearColor;
                                count--;
                                }
                            rep_stosdw(p,count,clearColor);
                            }
                        else
                            rep_stosw(p,count,clearColor);
                        base += snap->stride;
                        }
                    }
                else if (snap->bitsPerPixel == 24) {
                    uchar   *base = snap->surface + (snap->stride * y) + x * 3;

                    if (clearColor == 0) {
                        while (height--) {
                            memset(base,0,width * 3);
                            base += snap->stride;
                            }
                        }
                    else {
                        uchar   c0 = clearColor;
                        uchar   c1 = clearColor >> 8;
                        uchar   c2 = clearColor >> 16;

                        while (height--) {
                            int     count = width;
                            ushort  *p = (ushort*)base;

                            while (count--) {
                                p[0] = c0;
                                p[1] = c1;
                                p[2] = c2;
                                p += 3;
                                }
                            base += snap->stride;
                            }
                        }
                    }
                else if (snap->bitsPerPixel == 32) {
                    uchar *base = snap->surface + (snap->stride * y) + x * 4;
                    while (height--) {
                        rep_stosd(base,width,clearColor);
                        base += snap->stride;
                        }
                    }
                }
            }
        else {
            /* Clear buffer in hardware using SNAP driver */
            GA_buf  *drawBuf = NULL;

            /* Enable plane masking if necessary */
            if (colorMask != 0xFFFFFFFF)
                GL.state2d.SetPlaneMask(colorMask);

            /* Clear all four possible dest color buffers */
            for (bufferBit = 1; bufferBit <= 8; bufferBit = bufferBit << 1) {
                if (bufferBit & ctx->Color._DrawDestMask) {
                    if (bufferBit == DD_FRONT_LEFT_BIT)
                        drawBuf = snap->frontBuf;
                    else if (bufferBit == FRONT_RIGHT_BIT)
                        drawBuf = snap->frontRightBuf;
                    else if (bufferBit == BACK_LEFT_BIT)
                        drawBuf = snap->backBuf;
                    else if (bufferBit == BACK_RIGHT_BIT)
                        drawBuf = snap->backRightBuf;
                    ASSERT(drawBuf);
                    GL.buf.DrawRectBuf(drawBuf,x,y,width,height,clearColor,GA_REPLACE_MIX);
                    }
                }

            /* Restore plane masking if necessary */
            if (colorMask != 0xFFFFFFFF)
                GL.state2d.SetPlaneMask(0xFFFFFFFF);
            }

        /* Clear flags for the color buffers we cleared */
        mask &= ~ctx->Color._DrawDestMask;
        }

    /* Let Mesa clear everything else in software */
    if (mask)
        _swrast_Clear(ctx, mask, all, x, y, width, height);
}

/****************************************************************************
REMARKS:
Driver hook for glEnable() and glDisable() calls. We mostly use this to
check when dithering is enabled or disabled.
****************************************************************************/
static void gl_enable(
    GLcontext *ctx,
    GLenum cap,
    GLboolean state)
{
    GA_glContext    *snap = SNAP_CONTEXT(ctx);

    if ((snap->bitsPerPixel <= 16) && cap == GL_DITHER) {
        if (ctx->Color.DitherFlag)
            snap->pixelFormat = snap->dithered_pf;
        else
            snap->pixelFormat = snap->undithered_pf;
        _snap_hook_span_funcs(ctx);
        }
}

/****************************************************************************
REMARKS:
Hook all of our driver functions into the rendering context
****************************************************************************/
static void hook_in_driver_functions(
    GLcontext *ctx)
{
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
    swrast_dd       *swdd = _swrast_GetDeviceDriverReference(ctx);
    TNLcontext      *tnl = TNL_CONTEXT(ctx);
    SWcontext       *swrast = SWRAST_CONTEXT(ctx);

    ASSERT((void*)snap == (void*)ctx->DriverCtx);

    /* Use default TNL pipeline */
    tnl->Driver.RunPipeline             = _tnl_run_pipeline;

    ctx->Driver.GetString               = get_string;
    ctx->Driver.UpdateState             = update_state;
    ctx->Driver.ResizeBuffers           = _swrast_alloc_buffers;
    ctx->Driver.GetBufferSize           = get_buffer_size;

    ctx->Driver.Enable                  = gl_enable;

    ctx->Driver.Accum                   = _swrast_Accum;
    ctx->Driver.Bitmap                  = _swrast_Bitmap;
    ctx->Driver.Clear                   = clear;
    ctx->Driver.CopyPixels              = _swrast_CopyPixels;
    ctx->Driver.DrawPixels              = _swrast_DrawPixels;
    ctx->Driver.ReadPixels              = _swrast_ReadPixels;
    ctx->Driver.DrawBuffer              = _swrast_DrawBuffer;

    ctx->Driver.ChooseTextureFormat     = _mesa_choose_tex_format;
    ctx->Driver.TexImage1D              = _mesa_store_teximage1d;
    ctx->Driver.TexImage2D              = _mesa_store_teximage2d;
    ctx->Driver.TexImage3D              = _mesa_store_teximage3d;
    ctx->Driver.TexSubImage1D           = _mesa_store_texsubimage1d;
    ctx->Driver.TexSubImage2D           = _mesa_store_texsubimage2d;
    ctx->Driver.TexSubImage3D           = _mesa_store_texsubimage3d;
    ctx->Driver.TestProxyTexImage       = _mesa_test_proxy_teximage;

    ctx->Driver.CompressedTexImage1D    = _mesa_store_compressed_teximage1d;
    ctx->Driver.CompressedTexImage2D    = _mesa_store_compressed_teximage2d;
    ctx->Driver.CompressedTexImage3D    = _mesa_store_compressed_teximage3d;
    ctx->Driver.CompressedTexSubImage1D = _mesa_store_compressed_texsubimage1d;
    ctx->Driver.CompressedTexSubImage2D = _mesa_store_compressed_texsubimage2d;
    ctx->Driver.CompressedTexSubImage3D = _mesa_store_compressed_texsubimage3d;

    ctx->Driver.CopyTexImage1D          = _swrast_copy_teximage1d;
    ctx->Driver.CopyTexImage2D          = _swrast_copy_teximage2d;
    ctx->Driver.CopyTexSubImage1D       = _swrast_copy_texsubimage1d;
    ctx->Driver.CopyTexSubImage2D       = _swrast_copy_texsubimage2d;
    ctx->Driver.CopyTexSubImage3D       = _swrast_copy_texsubimage3d;
    ctx->Driver.CopyColorTable          = _swrast_CopyColorTable;
    ctx->Driver.CopyColorSubTable       = _swrast_CopyColorSubTable;
    ctx->Driver.CopyConvolutionFilter1D = _swrast_CopyConvolutionFilter1D;
    ctx->Driver.CopyConvolutionFilter2D = _swrast_CopyConvolutionFilter2D;

    swdd->SetBuffer                     = set_buffer;

    /* Set up correct framebuffer access arbitration for video memory buffers */
    if (!snap->isSysMem && !(snap->drawBuf->Flags & gaBufferSysMem)) {
        swdd->SpanRenderStart           = span_render_start;
        swdd->SpanRenderFinish          = span_render_finish;
        }

    /* Now hook all of our software rendering functions */
    _snap_hook_span_funcs(ctx);

    /* Extend the software rasterizer with our line/point/triangle
     * functions.
     */
    swrast->choose_line                 = _snap_choose_line;
    swrast->choose_triangle             = _snap_choose_triangle;
    swrast->invalidate_line             |= SNAP_NEW_LINE;
    swrast->invalidate_triangle         |= SNAP_NEW_TRIANGLE;
}

/**********************************************************************/
/*****                    Public Functions                        *****/
/**********************************************************************/

/****************************************************************************
PARAMETERS:
ref2d   - Pointer to 2D reference rasteriser to initialise with

REMARKS:
This function initialises the internals of the SciTech SNAP OpenGL
compatible rendering engine. The passed in 2D referencr rasteriser is used
to get access to the necessary buffer manager functions for the driver,
as well as access to any necessary 2D rendering functions (potentially in
software or hardware).
****************************************************************************/
ibool NAPI GL3D_init(
    REF2D_driver *ref2d)
{
    // TODO: Here is where we can dynamically load the back end driver
    //       module for the chipset, either software, DirectX 7 or
    //       the hardware driver module. We can even have separate
    //       software modules for different color depths if we wanted
    //       although that may be problematic if we wish to render to
    //       a system memory buffer with a different color depth....

    // Request the buffer manager functions, and ensure the buffer
    // manager has already been initialised. GetPrimaryBuffer
    // will return NULL if the buffer manager has not yet been
    // initialised.
    GL.buf.dwSize = sizeof(GL.buf);
    if (!ref2d->QueryFunctions(GA_GET_BUFFERFUNCS,&GL.buf))
        return false;
    GL.state2d.dwSize = sizeof(GL.state2d);
    if (!ref2d->QueryFunctions(GA_GET_2DSTATEFUNCS,&GL.state2d))
        return false;
    GL.draw2d.dwSize = sizeof(GL.draw2d);
    if (!ref2d->QueryFunctions(GA_GET_2DRENDERFUNCS,&GL.draw2d))
        return false;
    if (!GL.buf.GetPrimaryBuffer())
        return false;
    return true;
}

/****************************************************************************
PARAMETERS:
buf     - SNAP buffer to choose the visual for
visual  - Structure containing visual context information

REMARKS:
This function examines the visual passed to use to deteremine if we support
the requested capabilities. If we don't we will modify the structure for
the capabilities that we do support (ie: lowering the depth buffer size
to 16 bits etc).

If you pass the parameter buf as NULL, then this function will only check
the fields assuming the visual will be used for a system memory buffer.
****************************************************************************/
void NAPI GL3D_chooseVisual(
    GA_buf *buf,
    GA_glVisual *visual)
{
    /* Force the depth buffer, stencil buffer and accum buffers to our sizes */
    if (visual->DepthBits)
        visual->DepthBits = DEFAULT_SOFTWARE_DEPTH_BITS;
    if (visual->StencilBits)
        visual->StencilBits = 8 * sizeof(GLstencil);
    if (visual->AccumBits)
        visual->AccumBits = 8 * sizeof(GLaccum);

    /* Disable ColorIndex mode if not available */
    if (buf && buf->BitsPerPixel > 8)
        visual->ColorIndex = GL_FALSE;

    /* Disable stereo and double buffering if this is requested for a non
     * video memory buffer.
     */
    if (!buf || (buf->Flags & gaBufferSysMem)) {
        visual->DoubleBuffer = GL_FALSE;
        visual->Stereo = GL_FALSE;
        }

    /* Disable stereo support for the moment */
    visual->Stereo = GL_FALSE;
}

/****************************************************************************
PARAMETERS:
buf     - SNAP buffer to check the visual for
visual  - Structure containing visual context information

RETURNS:
True on success, false if visual is not supported

REMARKS:
This function checks the passed in visual to make sure that it is supported.

If you pass the parameter buf as NULL, then this function will only check
the fields assuming the visual will be used for a system memory buffer.
****************************************************************************/
ibool NAPI GL3D_checkVisual(
    GA_buf *buf,
    GA_glVisual *visual)
{
    /* Do quick check on incoming visual */
    if (visual->DepthBits <= 8 || visual->DepthBits > DEFAULT_SOFTWARE_DEPTH_BITS)
        return 0;
    if (visual->StencilBits && visual->StencilBits != 8 * sizeof(GLstencil))
        return false;
    if (visual->AccumBits && visual->AccumBits != 8 * sizeof(GLaccum))
        return false;
    if (buf) {
        if (buf->BitsPerPixel > 8 && visual->ColorIndex)
            return false;
        if (buf->BitsPerPixel < 8)
            return false;
        if ((buf->Width > MAX_WIDTH) || (buf->Height > MAX_HEIGHT))
            return false;
        }
    if (visual->DoubleBuffer && (!buf || (buf->Flags & gaBufferSysMem)))
        return false;
    if (visual->Stereo) {
        if (buf || !(buf->Flags & gaBufferFlippable))
            return false;
        if (GL.buf.GetFlippableBuffer(1) == NULL)
            return false;
        }

    /* Disable stereo for now */
    if (visual->Stereo)
        return false;

    /* Success, visual is accepted! */
    return true;
}

/****************************************************************************
PARAMETERS:
mask    - Mask of bits to count

RETURNS:
Number of one bits in the mask.

REMARKS:
Counts the number of set bits in a mask and determines the shift amount.
****************************************************************************/
static int countBits(
    uint mask)
{
    int i = 0;

    if (mask) {
        while (mask & (1 << i))
            i++;
        }
    return i;
}

/****************************************************************************
PARAMETERS:
buf             - SNAP buffer to create the context for (NULL for sysMem)
width           - Width of the rendering context
height          - Height of the rendering context
bitsPerPixel    - Color depth for the rendering context
surface         - Pointer to buffer surface for rendering context
stride          - Stride or scanline pitch for rendering context
pf              - Pixel format for rendering context
visual          - Structure containing visual context information
shareList       - Context to share display lists with, NULL if none

RETURNS:
Handle to rendering context on success, NULL on failure

REMARKS:
Attempts to create a Mesa rendering context for the device context. If
the SoftwareOnly flag is set and double buffering is enabled, the back
buffer will be allocated in system memory and not in video memory. Otherwise
a video memory back buffer will be allocated.
****************************************************************************/
static GA_HGLRC CreateContext(
    GA_buf *buf,
    int width,
    int height,
    int bitsPerPixel,
    void *surface,
    int stride,
    GA_pixelFormat *pf,
    GA_glVisual *visual,
    GA_HGLRC shareList)
{
    GLint           redBits = 0, greenBits = 0, blueBits = 0, alphaBits = 0;
    GLint           indexBits = 0, backBufferFlags;
    int             i;
    GA_glContext    *snap;

    /* First check that the visual is valid */
    if (!GL3D_checkVisual(buf,visual))
        return NULL;

    /* Determine the number of bits per color channel */
    if (bitsPerPixel <= 8 && visual->ColorIndex)
        indexBits = bitsPerPixel;
    else {
        if (bitsPerPixel <= 8) {
            redBits = 8;
            greenBits = 8;
            blueBits = 8;
            alphaBits = 0;
            }
        else {
            redBits = countBits(pf->RedMask);
            greenBits = countBits(pf->GreenMask);
            blueBits = countBits(pf->BlueMask);
            alphaBits = countBits(pf->AlphaMask);
            }
        }

    /* Allocate the SNAP rendering context */
    if ((snap = (GA_glContext*)CALLOC_STRUCT(GA_glContext)) == NULL)
        return NULL;
    snap->isSysMem = (buf == NULL);
    snap->width = width;
    snap->height = height;
    snap->bitsPerPixel = bitsPerPixel;
    snap->pf = *pf;

    /* Allocate back buffers from SNAP as necessary */
    if (!snap->isSysMem) {
        snap->frontBuf = buf;
        snap->backBuf = snap->backRightBuf = snap->frontRightBuf = NULL;
        backBufferFlags = 0;
        if (!(GL.hwCtx->Attributes & gaHaveAccel2D))
            backBufferFlags |= gaBufferSysMem;
        if (visual->SoftwareOnly)
            backBufferFlags |= gaBufferSysMem;
        if (buf->Flags & gaBufferFlippable) {
            /* Allocate hardware front buffer as first flippable buffer */
            if ((snap->frontBuf = GL.buf.GetFlippableBuffer(0)) == NULL) {
                FREE(snap);
                return NULL;
                }
            if (visual->Stereo) {
                /* Allocate hardware front right buffer as second flippable buffer */
                if ((snap->frontRightBuf = GL.buf.GetFlippableBuffer(1)) == NULL) {
                    FREE(snap);
                    return NULL;
                    }
                }
            if (visual->DoubleBuffer) {
                /* Try to allocate the back buffer. This will be in video memory if
                 * we have hardware, or system memory if we do not have hardware or
                 * software has been forced.
                 */
                snap->backFlipBuf = GL.buf.GetFlippableBuffer(visual->Stereo ? 2 : 1);
                snap->backBuf = snap->backFlipBuf;
                if ((backBufferFlags & gaBufferSysMem) || !snap->backBuf) {
                    if ((snap->backBuf = GL.buf.AllocBuffer(buf->Width,buf->Height,backBufferFlags)) == NULL) {
                        FREE(snap);
                        return NULL;
                        }
                    }
                if (visual->Stereo) {
                    /* Allocate back left and right buffers. We make them the third and
                     * fourth flippable buffers if we have hardware, otherwise they end
                     * up in software or offscreen video memory.
                     */
                    snap->backRightFlipBuf = GL.buf.GetFlippableBuffer(3);
                    snap->backRightBuf = snap->backRightFlipBuf;
                    if ((backBufferFlags & gaBufferSysMem) || !snap->frontRightBuf || !snap->backRightBuf) {
                        if ((snap->backRightBuf = GL.buf.AllocBuffer(buf->Width,buf->Height,backBufferFlags)) == NULL) {
                            FREE(snap);
                            return NULL;
                            }
                        }
                    }
                }
            }
        else if (visual->DoubleBuffer) {
            /* Allocate a new back buffer for the context */
            if ((snap->backBuf = GL.buf.AllocBuffer(buf->Width,buf->Height,backBufferFlags)) == NULL) {
                FREE(snap);
                return NULL;
                }
            }
        }

    /* Allocate memory for the color palette and dithering lookup table for
     * 8-bit RGB visuals where we need to do dithering.
     */
    if (!visual->ColorIndex && bitsPerPixel <= 8) {
        snap->cntPal = malloc(sizeof(snap->cntPal[0]) * 256);
        snap->ditherVec = malloc(sizeof(snap->ditherVec[0]) * 256);
        if (!snap->cntPal || !snap->ditherVec) {
            FREE(snap);
            return NULL;
            }

        /* Set the default palette to halfTone palette, and set up the dither
         * vector to be a 1:1 translation
         */
        memcpy(snap->cntPal,_snap_halftonePal,sizeof(snap->cntPal[0]) * 256);
        for (i = 0; i < 256; i++)
            snap->ditherVec[i] = i+20;
        }

    /* Create the Mesa visual context */
    snap->gl_visual = _mesa_create_visual(
            !visual->ColorIndex,
            visual->DoubleBuffer,
            visual->Stereo,
            redBits,
            greenBits,
            blueBits,
            alphaBits,
            indexBits,
            visual->DepthBits,
            visual->StencilBits,
            visual->AccumBits,
            visual->AccumBits,
            visual->AccumBits,
            alphaBits ? visual->AccumBits : 0,
            1);
    if (!snap->gl_visual) {
        free(snap->cntPal);
        FREE(snap);
        return NULL;
        }

    /* Create the rendering context internals */
    if (!_mesa_initialize_context(
            &snap->mesa,
            snap->gl_visual,
            shareList ? &shareList->mesa : NULL,
            snap,
            GL_FALSE)) {
        _mesa_destroy_visual(snap->gl_visual);
        FREE(snap);
        return NULL;
        }
    _mesa_enable_sw_extensions(&snap->mesa);
    _mesa_enable_1_3_extensions(&snap->mesa);
    _mesa_enable_1_4_extensions(&snap->mesa);
    _mesa_enable_1_5_extensions(&snap->mesa);

    /* Create the internal framebuffers */
    snap->gl_buffer = _mesa_create_framebuffer(
            snap->gl_visual,
            snap->gl_visual->depthBits > 0,
            snap->gl_visual->stencilBits > 0,
            snap->gl_visual->accumRedBits > 0,
            GL_FALSE);
    if (!snap->gl_buffer) {
        _mesa_destroy_visual(snap->gl_visual);
        _mesa_free_context_data(&snap->mesa);
        FREE(snap);
        return NULL;
        }

    /* Initialize double buffering */
    if (!snap->isSysMem) {
        snap->drawBuf = snap->backBuf ? snap->backBuf : snap->frontBuf;
        GL.buf.SetActiveBuffer(snap->drawBuf);
        if (visual->Stereo)
            GL.buf.FlipToStereoBuffer(snap->frontBuf,snap->frontRightBuf,gaDontWait);
        else if (visual->DoubleBuffer)
            GL.buf.FlipToBuffer(snap->frontBuf,gaDontWait);
        }

    /* Initialize the private rendering context information */
    if (visual->ColorIndex) {
        snap->dithered_pf = snap->undithered_pf = PF_INDEX;
        }
    else {
        /* DirectColor or TrueColor display */
        if (bitsPerPixel <= 8) {
            /* We always dither in 8bpp modes. Note that since 8bpp dithering is not
             * very useful, we only implement span rendering and none of the
             * optimised line and triangle functions.
             */
            snap->undithered_pf = PF_DITHER332;
            snap->dithered_pf = PF_DITHER332;
            }
        else if (bitsPerPixel == 15) {
            snap->undithered_pf = PF_RGB555;
            snap->dithered_pf = PF_DITHER555;
            }
        else if (bitsPerPixel == 16) {
            snap->undithered_pf = PF_RGB565;
            snap->dithered_pf = PF_DITHER565;
            }
        else if (bitsPerPixel == 24) {
            if (pf->RedPosition == 0)
                snap->undithered_pf = snap->dithered_pf = PF_BGR888;
            else
                snap->undithered_pf = snap->dithered_pf = PF_RGB888;
            }
        else {
            if (pf->RedPosition == 0)
                snap->undithered_pf = snap->dithered_pf = PF_ABGR8888;
            else if (pf->RedPosition == 8)
                snap->undithered_pf = snap->dithered_pf = PF_BGRA8888;
            else if (pf->RedPosition == 16)
                snap->undithered_pf = snap->dithered_pf = PF_ARGB8888;
            else
                snap->undithered_pf = snap->dithered_pf = PF_RGBA8888;
            }
        }
    snap->pixelFormat = snap->dithered_pf;

    /* Initialise the internal surface pointer values */
    if (!snap->isSysMem) {
        span_render_start(&snap->mesa);
        span_render_finish(&snap->mesa);
        }
    else {
        snap->bottom = snap->height-1;
        snap->surface = surface;
        snap->stride = stride;
        }

    /* Initialize the software rasterizer and helper modules */
    _swrast_CreateContext(&snap->mesa);
    _ac_CreateContext(&snap->mesa);
    _tnl_CreateContext(&snap->mesa);
    _swsetup_CreateContext(&snap->mesa);
    _swsetup_Wakeup(&snap->mesa);
    hook_in_driver_functions(&snap->mesa);
    return snap;
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to create the context for
visual      - Structure containing visual context information
shareList   - Context to share display lists with, NULL if none

RETURNS:
Handle to rendering context on success, NULL on failure

REMARKS:
Attempts to create a Mesa rendering context for the device context. If
the SoftwareOnly flag is set and double buffering is enabled, the back
buffer will be allocated in system memory and not in video memory. Otherwise
a video memory back buffer will be allocated.
****************************************************************************/
GA_HGLRC NAPI GL3D_createContext(
    GA_buf *buf,
    GA_glVisual *visual,
    GA_HGLRC shareList)
{
    return CreateContext(buf,buf->Width,buf->Height,buf->BitsPerPixel,
        buf->Surface,buf->Stride,&buf->PixelFormat,visual,shareList);
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to create the context for
visual      - Structure containing visual context information
shareList   - Context to share display lists with, NULL if none

RETURNS:
Handle to rendering context on success, NULL on failure

REMARKS:
Attempts to create a Mesa rendering context for the device context. If
the SoftwareOnly flag is set and double buffering is enabled, the back
buffer will be allocated in system memory and not in video memory. Otherwise
a video memory back buffer will be allocated.
****************************************************************************/
GA_HGLRC NAPI GL3D_createSysMemContext(
    int width,
    int height,
    int bitsPerPixel,
    void *surface,
    int stride,
    GA_pixelFormat *pf,
    GA_glVisual *visual,
    GA_HGLRC shareList)
{
    return CreateContext(NULL,width,height,bitsPerPixel,surface,stride,pf,visual,shareList);
}

/****************************************************************************
PARAMETERS:
ctx - Context to delete

REMARKS:
Deletes the rendering context.
****************************************************************************/
void NAPI GL3D_deleteContext(
    GA_HGLRC ctx)
{
    GA_HGLRC current = GL3D_getCurrentContext();

    /* Check if we're deleting the currently bound context */
    if (ctx == current)
        _mesa_make_current(NULL, NULL);
    if (ctx) {
        /* Destroy any back buffers we allocated */
        if (ctx->backBuf && (!ctx->backBuf->Flags & gaBufferFlippable))
            GL.buf.FreeBuffer(ctx->backBuf);
        if (ctx->backRightBuf && (!ctx->backRightBuf->Flags & gaBufferFlippable))
            GL.buf.FreeBuffer(ctx->backRightBuf);

        /* Destroy internal Mesa state */
        _swsetup_DestroyContext(&ctx->mesa);
        _tnl_DestroyContext(&ctx->mesa);
        _ac_DestroyContext(&ctx->mesa);
        _swrast_DestroyContext(&ctx->mesa);
        _mesa_destroy_visual(ctx->gl_visual);
        _mesa_destroy_framebuffer(ctx->gl_buffer);
        _mesa_free_context_data(&ctx->mesa);
        FREE(ctx);
        }
}

/****************************************************************************
PARAMETERS:
ctx - Context to make current

REMARKS:
Makes the passed in rendering context the current rendering context for
OpenGL functions.

TODO: We may want to eventually pass a GA_buf parameter in here to allow us
      to rebind an existing rendering context to a new buffer. This would
      allow the buffer to be switched from a front buffer in a window to
      a back buffer that is the size of the actual window (and a new buffer
      if the window is resized). To handle rendering into the primary buffer
      when doing front buffer rendering on the desktop, we will need to
      also add a mechanism to set a hardware 'viewport' on the buffer to
      restrict all rendering to that specific rectangle (with clipping
      of course).

      This would of course be mutually exclusive to page flipped stereo
      modes...
****************************************************************************/
ibool NAPI GL3D_makeCurrent(
    GA_HGLRC ctx)
{
   update_state(&ctx->mesa, 0);
   _mesa_make_current(&ctx->mesa, ctx->gl_buffer);
   return GL_TRUE;
}

/****************************************************************************
REMARKS:
Returns the current rendering context
****************************************************************************/
GA_HGLRC NAPI GL3D_getCurrentContext(void)
{
    return SNAP_CONTEXT(_mesa_get_current_context());
}

/****************************************************************************
PARAMETERS:
ctx     - Context to swap buffers for
waitVRT - True to wait for vertical retrace

REMARKS:
Swaps the display buffers for the rendering context. If we have a memory
DC this is blitted to the display, and if we have harware page flipping
we flip hardware display pages.
****************************************************************************/
void NAPI GL3D_swapBuffers(
    GA_HGLRC ctx,
    N_int32 waitVRT)
{
    GA_buf  *temp;

    /* Bail if no back buffers attached or is a system memory buffer */
    if (ctx->isSysMem || !ctx->backBuf)
        return;

    /* Blt the buffers to the front buffer if the back buffers are not
     * flippable. Note that we blit to the back flip buffers if we
     * have enough pages for hardware page flipping.
     */
    glFlush();
    if (!(ctx->backBuf->Flags & gaBufferFlippable)) {
        if (ctx->backRightBuf) {
            GL.buf.SetActiveBuffer(ctx->backRightFlipBuf ? ctx->backRightFlipBuf : ctx->frontRightBuf);
            GL.buf.BitBltBuf(ctx->backRightBuf,0,0,ctx->backRightBuf->Width,ctx->backRightBuf->Height,0,0,GA_REPLACE_MIX);
            }
        GL.buf.SetActiveBuffer(ctx->backFlipBuf ? ctx->backFlipBuf : ctx->frontBuf);
        GL.buf.BitBltBuf(ctx->backBuf,0,0,ctx->backBuf->Width,ctx->backBuf->Height,0,0,GA_REPLACE_MIX);
        GL.buf.SetActiveBuffer(ctx->drawBuf);
        }

    /* Do hardware page flipping if the back buffers are flippable */
    if (ctx->backFlipBuf) {
        SWAPT(ctx->frontBuf,ctx->backFlipBuf,temp);
        if (ctx->backRightFlipBuf) {
            SWAPT(ctx->frontRightBuf,ctx->backRightFlipBuf,temp);
            GL.buf.FlipToStereoBuffer(ctx->frontBuf,ctx->frontRightBuf,waitVRT);
            }
        else
            GL.buf.FlipToBuffer(ctx->frontBuf,waitVRT);
        if (ctx->backBuf->Flags & gaBufferFlippable) {
            ctx->backBuf = ctx->backFlipBuf;
            ctx->backRightBuf = ctx->backRightFlipBuf;
            _swrast_use_draw_buffer(&ctx->mesa);
            }
        }
}

/****************************************************************************
PARAMETERS:
srcPal      - Source bitmap palette to build translation vector for
dstPal      - Destination bitmap palette to build translation vector for
srcColor    - Number of colors in source palette

REMARKS:
This function builds a translation vector for translating 4bpp or 8bpp
color values to the destination device color values, by finding the closest
color in the destination palette that matches the color in the source
palette.
****************************************************************************/
static void BuildTranslateVector(
    N_uint8 *t,
    GA_palette *dstPal,
    GA_palette *srcPal,
    int srcColors)
{
    N_uint8 *p = t;

    while (srcColors--) {
        int         j,closest = 0,d,distance = 0x7FFFL;
        GA_palette  *dp = dstPal;

        for (j = 0; j < 256; j++,dp++) {
            if ((d = ABS((int)srcPal->Red - (int)dp->Red)
                   + ABS((int)srcPal->Green - (int)dp->Green)
                   + ABS((int)srcPal->Blue - (int)dp->Blue)) < distance) {
                distance = d;
                closest = j;
                }
            }
        *p++ = closest;
        srcPal++;
        }
}

/****************************************************************************
REMARKS:
Sets the palette data in the context. This does *not* actually update the
hardware palette, so that must be done separately by the calling code.
****************************************************************************/
void NAPI GL3D_setPaletteData(
    GA_HGLRC ctx,
    GA_palette *pal,
    N_int32 num,
    N_int32 index)
{
    if (ctx->cntPal) {
        /* Copy the palette data */
        memcpy(ctx->cntPal + index,pal,sizeof(GA_palette) * num);

        /* Re-compute the dither translation vector */
        BuildTranslateVector(ctx->ditherVec + index,pal,
            _snap_halftonePal + index,num);
        }
}

/****************************************************************************
REMARKS:
Returns a copy of the half tone palette for optimal 8-bit dithering.
****************************************************************************/
void NAPI GL3D_getHalfTonePalette(
    GA_palette *pal,
    N_int32 num)
{
    memcpy(pal,_snap_halftonePal,sizeof(GA_palette) * num);
}

/****************************************************************************
REMARKS:
Returns the address of extension functions.
****************************************************************************/
void * NAPI GL3D_getProcAddress(
    const char *procName)
{
    return (void*)_glapi_get_proc_address(procName);
}

