/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Internal header file for the Mesa/OpenGL interface bindings
*               for the SciTech SNAP Graphics architecture.
*
****************************************************************************/

#ifndef __SNAP_INTERNAL_H
#define __SNAP_INTERNAL_H

#define GL_GLEXT_PROTOTYPES
#include "snap/graphics/common.h"
#include "snap/common/internal.h"
#include "snap/gl3d.h"
#include "snap/graphics/snapver.h"
#include "snap/copyrigh.h"
#include "glheader.h"
#include "context.h"
#include "buffers.h"
#include "colormac.h"
#include "depth.h"
#include "extensions.h"
#include "imports.h"
#include "macros.h"
#include "matrix.h"
#include "mtypes.h"
#include "texformat.h"
#include "texobj.h"
#include "teximage.h"
#include "texstore.h"
#include "array_cache/acache.h"
#include "swrast/swrast.h"
#include "swrast_setup/swrast_setup.h"
#include "swrast/s_context.h"
#include "swrast/s_depth.h"
#include "swrast/s_lines.h"
#include "swrast/s_triangle.h"
#include "tnl/tnl.h"
#include "tnl/t_context.h"
#include "tnl/t_pipeline.h"
#include "common_x86_asm.h"

/*---------------------- Macros and type definitions ----------------------*/

#define SCITECH_STR         "SciTech Software, Inc."
#define COPYRIGHT_STR       SCITECH_COPYRIGHT_MSG_SHORT
#define RENDERER_STR        GRAPHICS_PRODUCT " " SNAP_RELEASE_MAJOR_STR "." SNAP_RELEASE_MINOR_STR

/****************************************************************************
REMARKS:
This enumeration defines the specific pixel format for the rendering
context. We define this as an enumeration as we only support a specific
set of pixel formats in the software rendering code.

MEMBERS:
PF_INDEX        - Color index mode
PF_DITHER332    - Color index 3:3:2 with dithering
PF_RGB8         - 8bpp RGB without dithering
PF_DITHER555    - 15bpp RGB dithered
PF_RGB565       - 16bpp RGB without dithering
PF_DITHER565    - 16bpp RGB dithered
PF_RGB888       - 24bpp RGB TrueColor
PF_BGR888       - 24bpp BGR TrueColor
PF_ARGB8888     - 32bpp ARGB TrueColor
PF_ABGR8888     - 32bpp ABGR TrueColor
PF_RGBA8888     - 32bpp RGBA TrueColor
PF_BGRA8888     - 32bpp BGRA TrueColor
****************************************************************************/
typedef enum {
    PF_INDEX,
    PF_DITHER332,
    PF_RGB555,
    PF_DITHER555,
    PF_RGB565,
    PF_DITHER565,
    PF_RGB888,
    PF_BGR888,
    PF_ARGB8888,
    PF_ABGR8888,
    PF_RGBA8888,
    PF_BGRA8888
    };

/****************************************************************************
REMARKS:
This is the SNAP internal device driver context structure. Notice how it
includes a GLcontext directly. By doing this we're mimicking C++
inheritance/derivation, so we can simple cast a GLcontext pointer into a
GL_HGLRC pointer and vice versa.

MEMBERS:
mesa                - The core GL/Mesa context
gl_visual           - Describes the buffers
gl_buffer           - Depth, stencil, accum, etc buffers
width               - Width of the rendering context
height              - Height of the rendering context
bitsPerPixel        - Bits per pixel for the rendering context
pixelFormat         - Pixel format for the buffer
drawBuf             - Currently active draw buffer
frontBuf            - Front buffer for context
backBuf             - Back buffer for context
frontRightBuf       - Front right buffer for context
backRightBuf        - Back right buffer for context
backFlipBuf         - Back buffer for page flipping
backRightFlipBuf    - Back right buffer for page flipping
pixelFormat         - Pixel format for the buffer
undithered_pf       - Undithered pixel format
dithered_pf         - Dithered pixel format
surface             - Buffer surface pointer
stride              - Stride for the buffer surface
bottom              - Bottom coordinate for buffer
cntPal              - Current color lookup table
ditherVec           - Current color palette translation vector for dithering
isSysMem            - True if the buffer is a system memory buffer
****************************************************************************/
typedef struct GA_glContext {
    GLcontext       mesa;
    GLvisual        *gl_visual;
    GLframebuffer   *gl_buffer;
    int             width;
    int             height;
    int             bitsPerPixel;
    GA_pixelFormat  pf;
    GA_buf          *drawBuf;
    GA_buf          *frontBuf;
    GA_buf          *backBuf;
    GA_buf          *frontRightBuf;
    GA_buf          *backRightBuf;
    GA_buf          *backFlipBuf;
    GA_buf          *backRightFlipBuf;
    int             pixelFormat;
    int             undithered_pf;
    int             dithered_pf;
    N_uint8         *surface;
    int             stride;
    int             bottom;
    GA_palette      *cntPal;
    N_uint8         *ditherVec;
    int             isSysMem;
    } GA_glContext;

/* Just cast, since we're using structure containment */
#define SNAP_CONTEXT(ctx)  ((GA_glContext*)(ctx->DriverCtx))

/* Convenience typedefs */

typedef struct swrast_device_driver swrast_dd;

/* Macros to swap two values */

#define SWAP(a,b)   { a^=b; b^=a; a^=b; }

#define SWAPT(v1,v2,t) {    \
    t = v1;                 \
    v1 = v2;                \
    v2 = t;                 \
    }

/* Pixel addressing macros */

#define PIXEL_ADDR_8(snap,X,Y) \
    (GLubyte*)((snap)->surface + ((snap)->bottom - (Y)) * (snap)->stride + (X))

#define PIXEL_ADDR_16(snap,X,Y) \
    (GLushort*)((snap)->surface + ((snap)->bottom - (Y)) * (snap)->stride + (X) * 2)

#define PIXEL_ADDR_24(snap,X,Y) \
    (rgb_t*)((snap)->surface + ((snap)->bottom - (Y)) * (snap)->stride + (X) * 3)

#define PIXEL_ADDR_32(snap,X,Y) \
    (GLuint*)((snap)->surface + ((snap)->bottom - (Y)) * (snap)->stride + (X) * 4)

/* Remove pesky macros from other Mesa header files */

#undef PACK_COLOR_565

/* Macros for 8bpp pixel packing/unpacking */

#define PACK_COLOR_DITHER332(P,X,Y,R,G,B)                               \
{                                                                       \
    int _dither = _snap_dither8x8[(((Y) & 7) << 3) + ((X) & 7)];        \
    (P) = ditherVec[                                                    \
        _snap_div51[R]              + (_snap_mod51[R] > _dither) +      \
        _snap_mul6[_snap_div51[G]   + (_snap_mod51[G] > _dither)] +     \
        _snap_mul36[_snap_div51[B]  + (_snap_mod51[B] > _dither)]];     \
}

#define UNPACK_COLOR_332(C,R,G,B,A)                 \
{                                                   \
    GA_palette *p = &snap->cntPal[C];               \
    R = p->Red;                                     \
    G = p->Green;                                   \
    B = p->Blue;                                    \
    A = CHAN_MAX;                                   \
}

/* Macros for 15bpp pixel packing/unpacking */

#define PACK_COLOR_555(R, G, B) \
    ( (((R) & 0xF8) << 7)       \
    | (((G) & 0xF8) << 2)       \
    | ((B) >> 3) )

/* Note that we do the modulus computations here using a table lookup
 * as that allows us to avoid overflow computations by making the modulus
 * always zero at the high end of the color range (ie: white and close to
 * white is never dithered).
 */

#define PACK_COLOR_DITHER555(P,X,Y,R,G,B)                               \
{                                                                       \
    int _dither = _snap_dither4x4[(((Y) & 3) << 2) + ((X) & 3)];        \
    (P) = (N_uint16)                                                    \
       ((((R >> 3) + (_snap_mod8[R] > _dither)) << 10) +                \
        (((G >> 3) + (_snap_mod8[G] > _dither)) << 5) +                 \
        (((B >> 3) + (_snap_mod8[B] > _dither)) << 0));                 \
}

#define UNPACK_COLOR_555(C,R,G,B,A)                 \
{                                                   \
    R = ( ((C >> 7) & 0xF8) | ((C >> 10) & 0x7) );  \
    G = ( ((C >> 2) & 0xF8) | ((C >>  5) & 0x7) );  \
    B = ( ((C << 3) & 0xF8) | ((C      ) & 0x7) );  \
    A = CHAN_MAX;                                   \
}

/* Macros for 16bpp pixel packing/unpacking */

#define PACK_COLOR_565(R,G,B)               \
    ( (((R) & 0xF8) << 8)       \
    | (((G) & 0xFC) << 3)       \
    | ((B) >> 3) )

#define PACK_COLOR_DITHER565(P,X,Y,R,G,B)                               \
{                                                                       \
    int _dither = _snap_dither4x4[(((Y) & 3) << 2) + ((X) & 3)];        \
    (P) = (N_uint16)                                                    \
       ((((R >> 3) + (_snap_mod8[R] > _dither)) << 11) +                \
        (((G >> 2) + (_snap_mod4[G] > _dither)) << 5) +                 \
        (((B >> 3) + (_snap_mod8[B] > _dither)) << 0));                 \
}

#define UNPACK_COLOR_565(C,R,G,B,A)                 \
{                                                   \
    R = ( ((C >> 8) & 0xF8) | ((C >> 11) & 0x7) );  \
    G = ( ((C >> 3) & 0xFC) | ((C >>  5) & 0x3) );  \
    B = ( ((C << 3) & 0xF8) | ((C      ) & 0x7) );  \
    A = CHAN_MAX;                                   \
}

/* Macros for 32bpp direct surface access */

#define PACK_COLOR_ARGB(R,G,B,A)    \
    (  (N_uint32)(A) << 24)         \
    | ((N_uint32)(R) << 16)         \
    | ((N_uint32)(G) << 8)          \
    | ((N_uint32)(B) << 0)

#define PACK_COLOR_ABGR(R,G,B,A)    \
    (  (N_uint32)(A) << 24)         \
    | ((N_uint32)(B) << 16)         \
    | ((N_uint32)(G) << 8)          \
    | ((N_uint32)(R) << 0)

#define PACK_COLOR_RGBA(R,G,B,A)    \
    (  (N_uint32)(R) << 24)         \
    | ((N_uint32)(G) << 16)         \
    | ((N_uint32)(B) << 8)          \
    | ((N_uint32)(A) << 0)

#define PACK_COLOR_BGRA(R,G,B,A)    \
    (  (N_uint32)(B) << 24)         \
    | ((N_uint32)(G) << 16)         \
    | ((N_uint32)(R) << 8)          \
    | ((N_uint32)(A) << 0)

#define UNPACK_COLOR_ARGB(c,R,G,B,A)                \
{                                                   \
 (A) = (N_uint8)( (N_uint32)(c) >> 24);             \
 (R) = (N_uint8)(((N_uint32)(c) >> 16) & 0xFF);     \
 (G) = (N_uint8)(((N_uint32)(c) >> 8) & 0xFF);      \
 (B) = (N_uint8)(((N_uint32)(c) >> 0) & 0xFF);      \
}

#define UNPACK_COLOR_ABGR(c,R,G,B,A)                \
{                                                   \
 (A) = (N_uint8)( (N_uint32)(c) >> 24);             \
 (B) = (N_uint8)(((N_uint32)(c) >> 16) & 0xFF);     \
 (G) = (N_uint8)(((N_uint32)(c) >> 8) & 0xFF);      \
 (R) = (N_uint8)(((N_uint32)(c) >> 0) & 0xFF);      \
}

#define UNPACK_COLOR_RGBA(c,R,G,B,A)                \
{                                                   \
 (R) = (N_uint8)( (N_uint32)(c) >> 24);             \
 (G) = (N_uint8)(((N_uint32)(c) >> 16) & 0xFF);     \
 (B) = (N_uint8)(((N_uint32)(c) >> 8) & 0xFF);      \
 (A) = (N_uint8)(((N_uint32)(c) >> 0) & 0xFF);      \
}

#define UNPACK_COLOR_BGRA(c,R,G,B,A)                \
{                                                   \
 (B) = (N_uint8)( (N_uint32)(c) >> 24);             \
 (G) = (N_uint8)(((N_uint32)(c) >> 16) & 0xFF);     \
 (R) = (N_uint8)(((N_uint32)(c) >> 8) & 0xFF);      \
 (A) = (N_uint8)(((N_uint32)(c) >> 0) & 0xFF);      \
}

/* Macros defining the state flags we need to set for the optimised
 * rendering functions that we hook out.
 */

#define SNAP_NEW_LINE   (_NEW_LINE | \
                         _NEW_TEXTURE | \
                         _NEW_LIGHT | \
                         _NEW_DEPTH | \
                         _NEW_RENDERMODE | \
                         _SWRAST_NEW_RASTERMASK)

#define SNAP_NEW_TRIANGLE (_NEW_POLYGON | \
                           _NEW_TEXTURE | \
                           _NEW_LIGHT | \
                           _NEW_DEPTH | \
                           _NEW_RENDERMODE | \
                           _SWRAST_NEW_RASTERMASK)

#pragma pack(1)

/****************************************************************************
REMARKS:
Structure defining the pixel type that we use for packed 24-bit modes.

MEMBERS:
c   - Array of color components
****************************************************************************/
typedef struct {
    GLubyte c[3];
    } rgb_t;

/****************************************************************************
REMARKS:
Structure defining the main device context structure for the device driver.
This structure contains all variable declarations that are used by the
core generic device driver code. Private chipset specific variables should
not be declared in here by declared in a chipset specific structure.

MEMBERS:
ctx                 - Public device context
shared              - True if loaded into shared memory
hwCtx               - Pointer to 2D hardware driver context
hModGLU             - Handle to GLU module if loaded
PCIConfig           - PCI device configuration space information
licensed            - Licensed options
globalOpt           - Global options
init                - Device driver init functions group
gl                  - OpenGL API functions group
glu                 - OpenGL utility library API functions group
buf                 - SNAP buffer manager functions
state2d             - SNAP 2D state functions
draw2d              - SNAP 2D drawing functions
****************************************************************************/
typedef struct {
    GA_glCtx            ctx;
    ibool               shared;
    GA_devCtx           *hwCtx;
    PE_MODULE           *hModGLU;
    PCIDeviceInfo       PCIConfig;
    GA_options          *licensed;
    GA_globalOptions    *globalOpt;
    GA_glInitFuncs      init;
    GA_glFuncs          gl;
    GA_gluFuncs         glu;
    GA_bufferFuncs      buf;
    GA_2DStateFuncs     state2d;
    GA_2DRenderFuncs    draw2d;
    } GL3D_driver;

#pragma pack()

/*--------------------------- Global Variables ----------------------------*/

extern GL3D_driver  GL;
extern char         DRV_path[PM_MAX_PATH];
extern GA_palette   _snap_halftonePal[256];
extern int          _snap_dither8x8[64];
extern int          _snap_div51[256];
extern int          _snap_mod51[256];
extern int          _snap_mul6[6];
extern int          _snap_mul36[6];
extern int          _snap_dither4x4[16];
extern int          _snap_mod8[256];
extern int          _snap_mod4[256];
extern int          _snap_dither5[2048];
extern int          _snap_dither6[2048];

/*------------------------- Function Prototypes ---------------------------*/

/* Exported context management functions */

ibool       NAPI GL3D_init(REF2D_driver *ref2d);
void        NAPI GL3D_chooseVisual(GA_buf *buf,GA_glVisual *visual);
ibool       NAPI GL3D_checkVisual(GA_buf *buf,GA_glVisual *visual);
GA_HGLRC    NAPI GL3D_createContext(GA_buf *buf,GA_glVisual *visual,GA_HGLRC shareList);
GA_HGLRC    NAPI GL3D_createSysMemContext(int width,int height,int bitsPerPixel,void *surface,int stride,GA_pixelFormat *pf,GA_glVisual *visual,GA_HGLRC shareList);
void        NAPI GL3D_deleteContext(GA_HGLRC ctx);
ibool       NAPI GL3D_makeCurrent(GA_HGLRC ctx);
GA_HGLRC    NAPI GL3D_getCurrentContext(void);
void        NAPI GL3D_swapBuffers(GA_HGLRC ctx,N_int32 waitVRT);
void        NAPI GL3D_setPaletteData(GA_HGLRC ctx,GA_palette *pal,N_int32 num,N_int32 index);
void        NAPI GL3D_getHalfTonePalette(GA_palette *pal,N_int32 num);
void *      NAPI GL3D_getProcAddress(const char *procName);

/* Function to hook our optimised software rendering code */

void        _snap_hook_span_funcs(GLcontext *ctx);
void        _snap_choose_line(GLcontext *ctx);
void        _snap_choose_triangle(GLcontext *ctx);

#endif  /* __SNAP_INTERNAL_H */

