/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Simple command line based configuration program to create
*               the default CRTC database used to store the information
*               for CRTC timings on disk.
*
*               We can also build this as a BPD file to be bound into the
*               SNAP graphics.bpd to generate the default CRTC data
*               file if the data file is out of date, and is not present.
*
****************************************************************************/

#include "snap/graphics/crtc.h"
#include "cmdline.h"
#include "pmapi.h"
#include <stdlib.h>
#include <string.h>

/*--------------------------- Global variables ----------------------------*/

char    SNAPConfigPath[PM_MAX_PATH];

#include "crtc.c"

#ifdef  __DRIVER__
PM_imports  _VARAPI _PM_imports;
#endif

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Insert a new record into the CRTC database.
****************************************************************************/
void Insert(
    int device,
    int head,
    CDBX_record *rec)
{
    int status;

    if ((status = CDBX_insert(device,head,rec,false)) != CDBX_ok) {
        if (status != CDBX_duplicateRecord)
            PM_fatalError("Failure to insert record!");
        }
}

/****************************************************************************
REMARKS:
Create the CRTC database for the device.
****************************************************************************/
void CreateCRTC(
    int device,
    int head)
{
    FILE    *fDB;
    CDBX    dbx;
    int     i,createDB;
    char    filename[PM_MAX_PATH],buf[PM_MAX_PATH];

    sprintf(buf,CDBX_NAME,head,device);
    strcpy(filename,SNAPConfigPath);
    PM_backslash(filename);
    strcat(filename,buf);
    createDB = false;
    if ((fDB = fopen(filename,"rb")) != NULL) {
        fclose(fDB);
        if (!CDBX_openHead(device,head))
            createDB = true;
        }
    else
        createDB = true;
    if (createDB) {
        if ((fDB = fopen(filename,"wb")) == NULL)
            PM_fatalError("Unable to create database!");
        /* Write the default header to disk */
        strcpy(dbx.h.signature,CDBX_SIGNATURE);
        dbx.h.entries = 0;
        dbx.h.majorVersion = 0;
        dbx.h.minorVersion = 0;
        fwrite(&dbx,1,sizeof(CDBX) + (dbx.h.entries-1) * sizeof(CDBX_record),fDB);
        fclose(fDB);
        if (!CDBX_openHead(device,head))
            PM_fatalError("Unable to open database!");
        }

    /* Insert all known refresh table entries */
    for (i = 0; i < sizeof(crtc)/sizeof(crtc[0]); i++) {
        Insert(device,head,&crtc[i]);
        }

    /* Flush database to disk and close */
    CDBX_flush(device,head);
    CDBX_close(device);
}

#ifdef  __DRIVER__
/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
PARAMETERS:
device  - Index of the device to generate timings for
pmImp   - PM library imports

REMARKS:
This function generates the CRTC timings for the specific SNAP device.
If the timings database already exists, it is upgraded to include any
new timings added after the initial database was created.
{secret}
****************************************************************************/
void _CEXPORT GA_createCRTCTimings(
    int device,
    int head,
    const char *configPath,
    PM_imports *pmImp)
{
    int     i,max;
    ulong   *p;

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_PM_imports)/sizeof(GA_createCRTCTimings_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;

    // Now copy all our imported functions
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    strcpy(SNAPConfigPath,configPath);

    // Initialise the CRTC database
    if (!CDBX_init(device+1,GA_MAX_HEADS))
        PM_fatalError("Unable to open database!");

    // Finally generate the CRTC timings
    CreateCRTC(device,head);
}
#else
int main(int argc, char *argv[])
{
    strcpy(SNAPConfigPath,PM_getSNAPConfigPath());
    PM_backslash(SNAPConfigPath);
    CreateCRTC(0,0);
    CreateCRTC(0,1);
    return 0;
}
#endif

