/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  any
*
* Description:  Module implementing all the monitor database access routines
*               used by SciTech Display Doctor.
*
****************************************************************************/

#include "snap/graphics.h"
#include "snap/monitor.h"
#ifndef __WIN32_VXD__
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#endif

/*---------------------- Macros and type definitions ----------------------*/

#pragma pack(1)

#define MFR_MAX             100
#define MDL_MAX             200
#define STR_MAX             ((MFR_MAX + MDL_MAX) * 2)
#define MFR_ID              "[Manufacturer]"
#define MFR_SIZE            13
#define STR_ID              "[Strings]"
#define STR_SIZE            9
#define ADDREG_ID           "AddReg"
#define ADDREG_SIZE         6
#define MFG_STR             "Mfg"
#define MFG_SIZE            3
#define PNPID_STR           "Monitor\\"
#define PNPID_SIZE          8
#define HKR_MODES           "HKR,\"MODES\\"
#define HKR_MODES_SIZE      10

/* {secret} */
typedef struct {
    char        filename[PM_MAX_PATH];  /* Monitor database filename    */
    long        entries;                /* Number of entries in database*/
    GA_monitor  **records;              /* Array of pointers to records */
    long        cursor;                 /* Current record index         */
    } MDBX;

/* {secret} */
typedef struct {
    char    id[80];
    char    name[80];
    } mfrrec;

/* {secret} */
typedef struct {
    char    id[80];
    char    name[80];
    char    PNPID[8];
    } mdlrec;

/* {secret} */
typedef struct {
    char    *data;  /* Buffer start */
    int     size;   /* Buffer size  */
    int     curpos; /* Current position */
    } FBUF;

/* {secret} */
typedef struct {
    char    key[40];
    char    value[80];
    } stringRef;

/* {secret} */
typedef struct {
    char        *str;
    } tokenbuf;

#pragma pack()

/*------------------------- Global Variables ------------------------------*/

static int          MDBX_errCode;
static MDBX         *db = NULL;
static int          dbrefs = 0;
static int          numStringRefs;
static stringRef    *stringRefs;

/*------------------------- Implementation --------------------------------*/

/****************************************************************************
DESCRIPTION:
Return the current monitor database error code.

HEADER:
snap/monitor.h

RETURNS:
Monitor database error code.

REMARKS:
This function return the current monitor database error code for the
previous operation.

SEE ALSO:
MDBX_getErrorMsg
****************************************************************************/
int NAPI MDBX_getErrCode(void)
{
    return MDBX_errCode;
}

/****************************************************************************
DESCRIPTION:
Return the current monitor database error code as a string.

HEADER:
snap/monitor.h

RETURNS:
Monitor database error code as a string.

REMARKS:
This function return the current monitor database error code for the
previous operation as a string.

NOTE:   The string is returned in a temporary buffer that is reused. Hence
        the calling application should copy the returned string to a
        different buffer immediately after calling this function and before
        calling any other SNAP functions.

SEE ALSO:
MDBX_getErrCode
****************************************************************************/
const char * NAPI MDBX_getErrorMsg(void)
{
    static char *msg[] = {
        "Everything is cool",
        "Database file not found",
        "Database is corrupted",
        "Not enough memory to load database",
        "Record was not found in database",
        "Fatal IO error",
        "Error parsing the INF file",
        "Database handle is invalid",
        "Program tried to access a database when none is open",
        "No valid monitor records found to import"
        };
    return msg[MDBX_errCode];
}

/****************************************************************************
DESCRIPTION:
Close the monitor database.

HEADER:
snap/monitor.h

REMARKS:
Closes the monitor database when done.

SEE ALSO:
MDBX_open
****************************************************************************/
void NAPI MDBX_close(void)
{
    int i;

    if (dbrefs <= 1) {
        if (db) {
            if (db->records) {
                for (i = 0; i < db->entries; i++)
                    free(db->records[i]);
                free(db->records);
                }
            free(db);
            }
        dbrefs = 0;
        db = NULL;
        }
    else
        dbrefs--;
}

/****************************************************************************
DESCRIPTION:
Open the monitor database.

HEADER:
snap/monitor.h

PARAMETERS:
filename    - Path to monitor database to open

RETURNS:
Error code for operation.

REMARKS:
Attempts to open the monitor database. This may fail if the file was not
found or the file was not a valid database file.

SEE ALSO:
MDBX_close
****************************************************************************/
ibool NAPI MDBX_open(
    const char *filename)
{
    FILE        *f;
    MDBX_header h;
    int         i;
    size_t      size;

    // Allow multiple calls to MDBX_open
    dbrefs++;
    if (db)
        return true;
    MDBX_errCode = MDBX_ok;
    if ((f = fopen(filename,"rb")) == NULL) {
        MDBX_errCode = MDBX_fileNotFound;
        goto Error;
        }
    fread(&h,1,sizeof(h),f);
    if (strcmp(h.signature,MDBX_SIGNATURE) != 0) {
        MDBX_errCode = MDBX_corrupt;
        goto Error;
        }
    if ((db = malloc(sizeof(MDBX))) == NULL) {
        MDBX_errCode = MDBX_outOfMemory;
        goto Error;
        }
    strncpy(db->filename, filename, PM_MAX_PATH);
    db->entries = h.entries;
    db->cursor = 0;
    if (h.entries > 0) {
        size = h.entries * sizeof(GA_monitor*);
        if ((db->records = calloc(1,size)) == NULL) {
            MDBX_errCode = MDBX_outOfMemory;
            goto Error;
            }
        for (i = 0; i < h.entries; i++) {
            if ((db->records[i] = malloc(sizeof(GA_monitor))) == NULL) {
                MDBX_errCode = MDBX_outOfMemory;
                goto Error;
                }
            if (fread(db->records[i],1,sizeof(GA_monitor),f) != sizeof(GA_monitor)) {
                MDBX_errCode = MDBX_IOError;
                goto Error;
                }
            }
        }
    else
        db->records = NULL;

    fclose(f);
    return true;

Error:
    if (f)
        fclose(f);
    MDBX_close();
    return false;
}

/****************************************************************************
PARAMETERS:
_a  - First parameter to compare
_b  - Second parameter to compare

RETURNS:
Result of the comparison operation

REMARKS:
Compares the two monitor database entries and returns the comparison result.
****************************************************************************/
static int MDBX_cmp(
    const void *_a,
    const void *_b)
{
    int val;

    GA_monitor *a = *(GA_monitor**)_a;
    GA_monitor *b = *(GA_monitor**)_b;
    if ((val = stricmp(a->mfr,b->mfr)) == 0)
        return stricmp(a->model,b->model);
    return val;
}

/****************************************************************************
DESCRIPTION:
Seeks to the first record in the monitor database

HEADER:
snap/monitor.h

PARAMETERS:
rec - Monitor record to store entry in

RETURNS:
Error code for operation.

REMARKS:
Attempts to seek to the first entry in the monitor database. If the
record exists, the parameters rec is filled in with the monitor information.

SEE ALSO:
MDBX_last, MDBX_next, MDBX_prev
****************************************************************************/
int NAPI MDBX_first(
    GA_monitor *rec)
{
    if (!db)
        return (MDBX_errCode = MDBX_dbNotOpen);
    if (db->records) {
        db->cursor = 0;
        memcpy(rec,db->records[db->cursor],sizeof(GA_monitor));
        return (MDBX_errCode = MDBX_ok);
        }
    return (MDBX_errCode = MDBX_notFound);
}

/****************************************************************************
DESCRIPTION:
Seeks to the last record in the monitor database

HEADER:
snap/monitor.h

PARAMETERS:
rec - Monitor record to store entry in

RETURNS:
Error code for operation.

REMARKS:
Attempts to seek to the last entry in the monitor database. If the
record exists, the parameters rec is filled in with the monitor information.

SEE ALSO:
MDBX_first, MDBX_next, MDBX_prev
****************************************************************************/
int NAPI MDBX_last(
    GA_monitor *rec)
{
    if (!db)
        return (MDBX_errCode = MDBX_dbNotOpen);
    if (db->records) {
        db->cursor = (db->entries-1);
        memcpy(rec,db->records[db->cursor],sizeof(GA_monitor));
        return (MDBX_errCode = MDBX_ok);
        }
    return (MDBX_errCode = MDBX_notFound);
}

/****************************************************************************
DESCRIPTION:
Seeks to the next record in the monitor database

HEADER:
snap/monitor.h

PARAMETERS:
rec - Monitor record to store entry in

RETURNS:
Error code for operation.

REMARKS:
Attempts to seek to the next entry in the monitor database after the passed
in record. If the record exists, the parameters rec is filled in with the
monitor information.

SEE ALSO:
MDBX_first, MDBX_last, MDBX_prev
****************************************************************************/
int NAPI MDBX_next(
    GA_monitor *rec)
{
    if (!db)
        return (MDBX_errCode = MDBX_dbNotOpen);
    if (db->cursor < db->entries-1) {
        db->cursor++;
        memcpy(rec,db->records[db->cursor],sizeof(GA_monitor));
        return (MDBX_errCode = MDBX_ok);
        }
    return (MDBX_errCode = MDBX_notFound);
}

/****************************************************************************
DESCRIPTION:
Seeks to the previous record in the monitor database

HEADER:
snap/monitor.h

PARAMETERS:
rec - Monitor record to store entry in

RETURNS:
Error code for operation.

REMARKS:
Attempts to seek to the previous entry in the monitor database after the
passed in record. If the record exists, the parameters rec is filled in with
the monitor information.

SEE ALSO:
MDBX_first, MDBX_last, MDBX_next
****************************************************************************/
int NAPI MDBX_prev(
    GA_monitor *rec)
{
    if (!db)
        return (MDBX_errCode = MDBX_dbNotOpen);
    if (db->cursor > 0) {
        db->cursor--;
        memcpy(rec,db->records[db->cursor],sizeof(GA_monitor));
        return (MDBX_errCode = MDBX_ok);
        }
    return (MDBX_errCode = MDBX_notFound);
}

/****************************************************************************
DESCRIPTION:
Inserts a new record into the monitor database

HEADER:
snap/monitor.h

PARAMETERS:
rec - Monitor record to store entry in

RETURNS:
Error code for operation.

REMARKS:
Inserts the specified record into the database. Because the database's are
always very small, we simply do a complete copy of the record array,
inserting the new record along the way. If the record already exists, we
update the existing record with the new data.

SEE ALSO:
MDBX_first, MDBX_update, MDBX_flush
****************************************************************************/
int NAPI MDBX_insert(
    GA_monitor *rec)
{
    if (db) {
        GA_monitor  **newRecords;
        char        *p,*newp,*maxp,*maxnewp;
        size_t      size;
        int         val;

        /* Allocate space for new records */
        size = (db->entries++) * sizeof(GA_monitor*);
        p = (char*)db->records;
        maxp = p + size;
        size += sizeof(GA_monitor*);
        if ((newRecords = malloc(size)) == NULL)
            return (MDBX_errCode = MDBX_outOfMemory);
        newp = (char*)newRecords;
        maxnewp = newp + size;

        /* Copy all records prior to new record */
        if (p) {
            while ((p < maxp) && ((val = MDBX_cmp(p,&rec)) <= 0)) {
                /* Check if the record already exists and replace if so */
                if (val == 0) {
                    memcpy(*((GA_monitor**)p),rec,sizeof(GA_monitor));
                    db->entries--;
                    free(newRecords);
                    return (MDBX_errCode = MDBX_ok);
                    }
                memcpy(newp,p,sizeof(GA_monitor*));
                p += sizeof(GA_monitor*);
                newp += sizeof(GA_monitor*);
                }
            }

        /* Add new record */
        if (((*(GA_monitor**)newp) = malloc(sizeof(GA_monitor))) == NULL)
            return (MDBX_errCode = MDBX_outOfMemory);
        memcpy(*(GA_monitor**)newp,rec,sizeof(GA_monitor));
        db->cursor = (newp - (char*)newRecords) / sizeof(GA_monitor*);
        newp += sizeof(GA_monitor*);

        /* Copy all remaining records */
        if (p) {
            while (newp < maxnewp) {
                memcpy(newp,p,sizeof(GA_monitor*));
                p += sizeof(GA_monitor*);
                newp += sizeof(GA_monitor*);
                }
            }

        if (db->records)
            free(db->records);
        db->records = newRecords;
        return (MDBX_errCode = MDBX_ok);
        }
    return (MDBX_errCode = MDBX_invalidDB);
}

/****************************************************************************
DESCRIPTION:
Updates the current record into the monitor database

HEADER:
snap/monitor.h

PARAMETERS:
rec - Monitor record to store entry in

RETURNS:
Error code for operation.

REMARKS:
Updates the current record with the new values in the passed in record.

SEE ALSO:
MDBX_first, MDBX_insert, MDBX_flush
****************************************************************************/
int NAPI MDBX_update(
    GA_monitor *rec)
{
    if (!db)
        return (MDBX_errCode = MDBX_dbNotOpen);
    if (db->cursor <= db->entries-1) {
        memcpy(db->records[db->cursor],rec,sizeof(GA_monitor));
        return (MDBX_errCode = MDBX_ok);
        }
    return (MDBX_errCode = MDBX_notFound);
}

/****************************************************************************
DESCRIPTION:
Flushes the monitor database to disk.

HEADER:
snap/monitor.h

PARAMETERS:
rec - Monitor record to store entry in

RETURNS:
Error code for operation.

REMARKS:
Flushes the database records to disk after they have been updated.

SEE ALSO:
MDBX_first, MDBX_insert, MDBX_update
****************************************************************************/
int NAPI MDBX_flush(void)
{
    int         i;
    FILE        *f;
    MDBX_header h;

    if (db && db->records) {
        if ((f = fopen(db->filename,"rb")) == NULL)
            return (MDBX_errCode = MDBX_IOError);
        fread(&h,1,sizeof(h),f);
        fclose(f);
        if ((f = fopen(db->filename,"wb")) == NULL)
            return (MDBX_errCode = MDBX_IOError);
        h.entries = db->entries;
        fwrite(&h,1,sizeof(h),f);
        for (i = 0; i < db->entries; i++) {
            if (fwrite(db->records[i],1,sizeof(GA_monitor),f) != sizeof(GA_monitor))
                return (MDBX_errCode = MDBX_IOError);
            }
        fclose(f);
        return (MDBX_errCode = MDBX_ok);
        }
    return (MDBX_errCode = MDBX_invalidDB);
}

/****************************************************************************
REMARKS:
Function to display an extended error message when an error occurs during
parsing.
****************************************************************************/
static ibool FatalError(
    const char *fmt,
    ...)
{
#if defined(DEBUG) || defined (TEST_HARNESS)
    va_list argptr;
    char    msg[255];

    va_start(argptr, fmt);
    vsprintf(msg, fmt, argptr);
    va_end(argptr);
    printf("\n\n%s\n\n",msg);
#else
    (void)fmt;
#endif
    return MDBX_parseError;
}

/****************************************************************************
PARAMETERS:
rec     - Place to store monitor config record
xRes    - X resolution for the timing
yRes    - Y resolution for the timing
Hz      - Refresh rate for the timing

REMARKS:
Computes the horizontal scan rate required to support the mode at the
specified refresh rate, and determines the maximum resolution from this
scan rate.
****************************************************************************/
static void CheckMaxRes(
    GA_monitor *rec,
    int xRes,
    int yRes,
    int Hz)
{
    int maxRes,hScan;

    /* Find the horizontal and vertical scan rates for this mode */
    hScan = (int)(((yRes * 1.05) * Hz / 1000) + 0.5);
    if (hScan > rec->maxHScan)
        return;

    /* Find the maximum resolution for this mode */
    if (xRes >= 2048)       maxRes = MaxRes_2048x1536;
    else if (xRes >= 1920)  maxRes = MaxRes_1920x1440;
    else if (xRes >= 1800)  maxRes = MaxRes_1800x1350;
    else if (xRes >= 1600)  maxRes = MaxRes_1600x1200;
    else if (xRes >= 1280)  maxRes = MaxRes_1280x1024;
    else if (xRes >= 1152)  maxRes = MaxRes_1152x864;
    else if (xRes >= 1024)  maxRes = MaxRes_1024x768;
    else if (xRes >= 800)   maxRes = MaxRes_800x600;
    else                    maxRes = MaxRes_640x480;
    if (rec->maxResolution < maxRes)
        rec->maxResolution = maxRes;
}

/****************************************************************************
DESCRIPTION:
Helper function - buffered fgets() replacement
****************************************************************************/
static char *bfgets(
    char *string,
    int n,
    FBUF *f,
    ibool substitute)
{
    stringRef   ref;
    char        *p;
    char        *str = string;
    int         i,valid;

    /* Skip all leading spaces */
    do {
        valid = true;
        while ((f->curpos < f->size) && isspace(f->data[f->curpos]) && f->data[f->curpos] != '\n')
            f->curpos++;
        if (f->curpos == f->size) {
            *str = '\0';
            return NULL;
            }

        /* Skip comment lines */
        if (f->data[f->curpos] == ';') {
            while ((f->curpos < f->size) && f->data[f->curpos] != '\n')
                f->curpos++;
            if (f->curpos == f->size) {
                *str = '\0';
                return NULL;
                }
            if (f->data[f->curpos] == '\n')
                f->curpos++;
            valid = false;
            }
        else if (f->data[f->curpos] == '\n') {
            /* Skip blank lines */
            f->curpos++;
            valid = false;
            }
        } while (!valid);

    /* Grab the string from the buffer */
    while ((f->curpos < f->size) && --n) {
        /* Translate ASCII to binary */
        if (f->data[f->curpos] == '\r' && f->data[f->curpos+1] == '\n')
            f->curpos++;
        *str = f->data[f->curpos++];
        if (*str == '%' && substitute) {
            /* Handle string substitution */
            p = ref.key;
            while ((f->curpos < f->size) && f->data[f->curpos] != '%')
                *p++ = f->data[f->curpos++];
            *p = '\0';

            /* Eat the trailing % symbol */
            if (f->data[f->curpos] == '%')
                f->curpos++;
            for (i = 0; i < numStringRefs; i++) {
                if (stricmp(stringRefs[i].key,ref.key) == 0) {
                    p = stringRefs[i].value;
                    while (--n && *p != '\0')
                        *str++ = *p++;
                    break;
                    }
                }
            }
        else if (*str == ';') {
            /* Skip trailing comments */
            while ((f->curpos < f->size) && f->data[f->curpos] != '\n')
                f->curpos++;
            }
        else if (*(str++) == '\n')
            break;
        }
    *str = '\0';

    /* Replace any TAB characters with a single space character */
    for (p = string; *p != '\0'; p++) {
        if (*p == '\x09')
            *p = ' ';
        }
    return string;
}

/****************************************************************************
DESCRIPTION:
Helper function to return the next string token with leading and trailing
spaces removed.
****************************************************************************/
static char * gettoken(
    tokenbuf *tok,
    const char *delimiters)
{
    char    *str,*p;
    int     len;

    /* Get the next token from the string */
    str = tok->str + strspn(tok->str, delimiters);
    if (*str == '\0') {
        /* No more tokens left */
        tok->str = "";
        return (NULL);
        }
    p = str + strcspn(str, delimiters);
    if (*p != '\0')
        *p++ = '\0';
    tok->str = p;

    /* Trim leading whitespace */
    while (isspace(*str) && *str != '\0')
        str++;

    /* Trim trailing whitespace */
    len = strlen(str);
    while (isspace(str[len-1]))
        str[--len] = '\0';
    return str;
}

/****************************************************************************
DESCRIPTION:
Helper function to copy characters from one string into another and placing
a string terminator at the end of the copied characters.
****************************************************************************/
static void copy_chars(
    char *dst,
    const char *src,
    int count)
{
    memcpy(dst,src,count);
    dst[count] = '\0';
}

/****************************************************************************
DESCRIPTION:
Import a Windows INF file into the monitor database

HEADER:
snap/monitor.h

PARAMETERS:
filename    - Path to the Windows INF file to parse and import
mfr         - Place to store first manufacturer string read (NULL to ignore)

RETURNS:
Error code for operation.

REMARKS:
This function is used to parse a Windows INF file from disk and insert
all monitor records into the monitor database. This can be used to implement
a monitor 'Have Disk' option for any supported Operating System.

NOTE: The monitor database is expected to be opened before this function
      is called, and should be flushed and closed when done. Hence this
      function can also import multiple INF files at a time into the
      database.
****************************************************************************/
int NAPI MDBX_importINF(
    const char *filename,
    char *mfr)
{
    FILE        *f;
    GA_monitor  rec;
    char        *p,*q;
    static char buf[256];
    static char idInstall[260];
    static char idInstall2[260];
    static char idInstall3[260];
    static char idAddReg[260];
    int         i,j,scanRate;
    mfrrec      *mfrList = NULL;
    mdlrec      *mdlList = NULL;
    int         mfrCount, mdlCount;
    int         dbxChanged = 0;
    int         fsize = 0;
    FBUF        filebuf = {NULL, 0, 0};
    tokenbuf    token;

    /* Allocate memory */
    stringRefs = NULL;
    numStringRefs = 0;
    if ((mfrList = malloc(MFR_MAX * sizeof(mfrrec))) == NULL) {
        MDBX_errCode = MDBX_outOfMemory;
        goto Exit;
        }
    if ((mdlList = malloc(MDL_MAX * sizeof(mdlrec))) == NULL) {
        MDBX_errCode = MDBX_outOfMemory;
        goto Exit;
        }
    if ((stringRefs = malloc(STR_MAX * sizeof(stringRef))) == NULL) {
        MDBX_errCode = MDBX_outOfMemory;
        goto Exit;
        }

    /* Open the INF file for import */
    if ((f = fopen(filename,"rb")) == NULL) {
        MDBX_errCode = FatalError("Unable to open %s INF!", filename);
        goto Exit;
        }

    /* Read the entire file into memory - this should speed up processing;
     * very significantly if the file is on a floppy.
     */
    fseek(f, 0, SEEK_END);
    fsize = ftell(f);
    if ((filebuf.data = malloc(fsize)) == NULL) {
        MDBX_errCode = MDBX_outOfMemory;
        goto Exit;
        }
    filebuf.size = fsize;
    fseek(f, 0, SEEK_SET);
    fread(filebuf.data, fsize, 1, f);
    fclose(f);

    /* Read string section for substitution. Format is:
     *
     *  [Strings]
     *  MS="Microsoft"
     *  Vista V72="DAYTEK Vista V72"
     */
    filebuf.curpos = 0;
    while (bfgets(buf,sizeof(buf),&filebuf,false)) {
        if (strnicmp(buf,STR_ID,STR_SIZE) == 0) {
            while (bfgets(buf,sizeof(buf),&filebuf,false)) {
                if (buf[0] == '[')
                    break;
                token.str = buf;
                if ((p = gettoken(&token,"=")) == NULL)
                    continue;
                strcpy(stringRefs[numStringRefs].key,p);
                if ((p = gettoken(&token,"")) == NULL)
                    continue;
                strcpy(stringRefs[numStringRefs].value,p);
                if (strlen(stringRefs[numStringRefs].key) >= sizeof(stringRefs[numStringRefs].key)) {
                    MDBX_errCode = FatalError("String key too long!");
                    goto Exit;
                    }
                if (strlen(stringRefs[numStringRefs].value) >= sizeof(stringRefs[numStringRefs].value)) {
                    MDBX_errCode = FatalError("String value too long!");
                    goto Exit;
                    }
                if (++numStringRefs == STR_MAX) {
                    MDBX_errCode = FatalError("Too many strings!");
                    goto Exit;
                    }
                }
            break;
            }
        }

    /* Read list of manufacturers. Format is:
     *
     *  [Manufacturer]
     *  "Microsoft"=MS
     *  "ADI Corporation."=ADI
     *  "SciTech Software, Inc."=SCITECH
     */
    filebuf.curpos = 0;
    mfrCount = 0;
    while (bfgets(buf,sizeof(buf),&filebuf,false)) {
        if (strnicmp(buf,MFR_ID,MFR_SIZE) == 0) {
            while (bfgets(buf,sizeof(buf),&filebuf,true)) {
                if (buf[0] == '[')
                    break;
                token.str = buf;
                if ((p = gettoken(&token,"=")) == NULL)
                    continue;
                copy_chars(mfrList[mfrCount].name,p+1,strlen(p)-2);
                if ((p = gettoken(&token,"")) == NULL)
                    continue;
                strcpy(mfrList[mfrCount].id,p);
                if (++mfrCount == MFR_MAX) {
                    MDBX_errCode = FatalError("Too many manufacturers!");
                    goto Exit;
                    }
                }
            break;
            }
        }

    /* Read model names and info for each manufacturer. Format is:
     *
     *  [ADI]
     *  "ADI Microscan E50"=ADIE50.Install,Monitor\ADI2E30
     *
     *  [Philips]
     *  "Philips CM0700 (21B)"=PH_CM0700_20T, Monitor\PHL0700
     */
    for (i = 0; i < mfrCount; i++) {
        filebuf.curpos = 0;
        sprintf(idInstall,"[%s]",mfrList[i].id);
        mdlCount = 0;
        while (bfgets(buf,sizeof(buf),&filebuf,false)) {
            if (strnicmp(buf,idInstall,strlen(idInstall)) == 0) {
                while (bfgets(buf,sizeof(buf),&filebuf,true)) {
                    if (buf[0] == '[')
                        break;
                    token.str = buf;
                    if ((p = gettoken(&token,"=")) == NULL || (strlen(p) == 0))
                        continue;
                    copy_chars(mdlList[mdlCount].name,p+1,strlen(p)-2);
                    if ((p = gettoken(&token,",")) == NULL || (strlen(p) == 0))
                        continue;
                    strcpy(mdlList[mdlCount].id,p);
                    mdlList[mdlCount].PNPID[0] = 0;
                    if ((p = gettoken(&token,",")) != NULL) {
                        /* Parse the optional PNPID */
                        if (strnicmp(p,PNPID_STR,PNPID_SIZE) == 0) {
                            if (strlen(p) == 15 &&
                                !isdigit(p[PNPID_SIZE+0]) &&
                                !isdigit(p[PNPID_SIZE+1]) &&
                                !isdigit(p[PNPID_SIZE+2]) &&
                                isxdigit(p[PNPID_SIZE+3]) &&
                                isxdigit(p[PNPID_SIZE+4]) &&
                                isxdigit(p[PNPID_SIZE+5]) &&
                                isxdigit(p[PNPID_SIZE+6]))
                                copy_chars(mdlList[mdlCount].PNPID,&p[PNPID_SIZE],sizeof(mdlList[mdlCount].PNPID)-1);
                            }
                        }
                    if (++mdlCount == MDL_MAX) {
                        MDBX_errCode = FatalError("Too many models!");
                        goto Exit;
                        }
                    }
                }
            }

        /* Extract model info and add record to database. Format is:
         *
         *  [ADIE50.Install]
         *  AddReg=ADIE50.AddReg,1024,DPMS
         *  DelReg=DEL_CURRENT_REG
         *
         *  [ADIE50.AddReg]
         *  HKR,"MODES\1024,768",Mode1,0,"30-70,50-120,+,+"
         *
         *  [UNM_1002.AddReg]
         *  HKR,"MODES\800,600",Mode1,,"30.0-54.0,50.0-120.0,+,+"
         *  HKR,"MODES\1024,768",Mode1,,"30.0-50.0,50.0-65.0,-,-"
         *
         *  [PH_CM0700_21B]
         *  DelReg=DCR
         *  AddReg= PH_CM0700_21B.Add
         *
         *  [PH_CM0700_21B.Add]
         *  HKR,"MODES\1600,1200",Mode1,,"30-94,50-160,+,-"
         *  HKR,,MaxResolution,,"1600,1200"
         *  HKR,,DPMS,,1
         *  HKR,,ICMProfile,1,9
         *
         *  [IA_75]
         *  DelReg=DCR
         *  AddReg=IA_75.Add
         *  CopyFiles=ICMFiles
         *
         *  [IA_75.Add]
         *  HKR,,MaxResolution,,"1600,1200"
         *  HKR,,DPMS,,1
         *  HKR,,ICMProfile,1,7
         *  HKR,"MODES\1600,1200",Mode1,,"30.0-95.0,50.0-180.0,+,"
         */
        for (j = 0; j < mdlCount; j++) {
            if (strlen(mdlList[j].name) == 0)
                continue;
            filebuf.curpos = 0;
            rec.maxResolution = 0xFF;
            rec.minHScan = 0;
            rec.maxHScan = 0;
            rec.minVScan = 0;
            rec.maxVScan = 0;
            rec.flags = 0;

            /* Find model install section and optional DPMS flag.
             *
             * NOTE: Some INF files don't properly match up the .Install sections.
             * Some have .Install in the model list but forget it for the install
             * section and others forget it the other way around. Rather than fix
             * the INF files we handle this here.
             */
            sprintf(idInstall,"[%s]", mdlList[j].id);
            sprintf(idInstall2,"[%s.Install]",mdlList[j].id);
            q = mdlList[j].id;
            p = idInstall3;
            *p++ = '[';
            while (*q != '.' && *q)
                *p++ = *q++;
            *p++ = ']';
            *p = '\0';
            idAddReg[0] = '\0';
            while (bfgets(buf,sizeof(buf),&filebuf,false)) {
                if (strnicmp(buf,idInstall,strlen(idInstall)) == 0
                        || strnicmp(buf,idInstall2,strlen(idInstall2)) == 0
                        || strnicmp(buf,idInstall3,strlen(idInstall3)) == 0) {
                    while (bfgets(buf,sizeof(buf),&filebuf,true)) {
                        if (buf[0] == '[')
                            break;
                        if (strnicmp(buf,ADDREG_ID,ADDREG_SIZE) == 0) {
                            token.str = buf;
                            if ((p = gettoken(&token,"=")) == NULL || (strlen(p) == 0))
                                continue;
                            if ((p = gettoken(&token,",")) == NULL || (strlen(p) == 0))
                                continue;
                            sprintf(idAddReg,"[%s]",p);
                            while ((p = gettoken(&token,",")) != NULL) {
                                if (stricmp(p,"DPMS") == 0)
                                    rec.flags |= Monitor_DPMSEnabled;
                                }
                            }
                        }
                    break;
                    }
                }
            if (idAddReg[0] == '\0') {
                FatalError("%s: Unable to find install section for %s", filename, mdlList[j].name);
                continue;
                }

            /* Find the AddReg section to get monitor frequency limits (and optional DPMS flag) */
            do {
                if (strnicmp(buf,idAddReg,strlen(idAddReg)) == 0) {
                    /* Process the AddReg entries for each display mode */
                    while (bfgets(buf,sizeof(buf),&filebuf,true) && (buf[0] != '[')) {
                        if (strncmp(buf,HKR_MODES,HKR_MODES_SIZE) == 0) {
                            token.str = buf;
                            if ((p = gettoken(&token,"\"")) == NULL || (strlen(p) == 0))    /* HKR, */
                                continue;
                            if ((p = gettoken(&token,"\"")) == NULL || (strlen(p) == 0))    /* MODES\1600,1200 */
                                continue;
                            if ((p = gettoken(&token,"\"")) == NULL || (strlen(p) == 0))    /* ,Mode1,, */
                                continue;
                            if ((p = gettoken(&token,"\"")) == NULL || (strlen(p) == 0))    /* 30-94,50-160,+,- */
                                continue;

                            /* Parse maximum horizontal frequency */
                            token.str = p;
                            if ((p = gettoken(&token,",")) == NULL || (strlen(p) == 0))     /* 30-94 */
                                continue;
                            scanRate = atoi(p);
                            while (isdigit(*p) || *p == '.')
                                p++;
                            if (*p == '-') {
                                p++;
                                scanRate = atoi(p);
                                }
                            rec.maxHScan = MAX(rec.maxHScan,scanRate);

                            /* Parse maxium vertical frequency */
                            if ((p = gettoken(&token,",")) == NULL || (strlen(p) == 0))     /* 50-160 */
                                continue;
                            scanRate = atoi(p);
                            while (isdigit(*p) || *p == '.')
                                p++;
                            if (*p == '-') {
                                p++;
                                scanRate = atoi(p);
                                }
                            rec.maxVScan = MAX(rec.maxVScan,scanRate);
                            }
                        else if (!(rec.flags & Monitor_DPMSEnabled)) {
                            if (strstr(buf,"DPMS"))
                                rec.flags |= Monitor_DPMSEnabled;
                            }
                        }
                    break;
                    }
                } while (bfgets(buf,sizeof(buf),&filebuf,false));

            /* Find maximum resolution from horizontal and vertical limits.
             * We do it this way, because it appears the Monitor .inf files
             * have bogus information in them for some monitors (or at
             * least we dont appear to parse the maximum resolution
             * correctly.
             */
            rec.maxResolution = MaxRes_640x480;
            CheckMaxRes(&rec,640,480,60);
            CheckMaxRes(&rec,800,600,60);
            CheckMaxRes(&rec,1024,768,60);
            CheckMaxRes(&rec,1152,864,60);
            CheckMaxRes(&rec,1280,1024,60);
            CheckMaxRes(&rec,1600,1200,60);
            CheckMaxRes(&rec,1800,1350,60);
            CheckMaxRes(&rec,1920,1440,60);
            CheckMaxRes(&rec,2048,1536,60);

            /* Instead of failing long MFR and Model names, truncate them */
            mfrList[i].name[sizeof(rec.mfr) - 1] = '\0';
            mdlList[j].name[sizeof(rec.model) - 1] = '\0';
            strcpy(rec.mfr,mfrList[i].name);
            strcpy(rec.model,mdlList[j].name);
            strcpy(rec.PNPID,mdlList[j].PNPID);

            /* Fill in default minimum frequencies */
            if (rec.minHScan == 0)
                rec.minHScan = 31;
            if (rec.minVScan == 0)
                rec.minVScan = 50;
            if (rec.maxHScan == 0 || rec.maxVScan == 0) {
                FatalError("\n%s: invalid record (%s %s)!\n", filename, rec.mfr, rec.model);
                }
            else if (MDBX_insert(&rec) != MDBX_ok) {
                FatalError("Failure to insert record!");
                }
            else dbxChanged = 1;
            }
        }

    MDBX_errCode = MDBX_ok;
    if(!dbxChanged)
        MDBX_errCode = MDBX_noRecords;
Exit:
    /* Free memory */
    free(mfrList);
    free(mdlList);
    free(filebuf.data);
    free(stringRefs);
    return MDBX_errCode;
}

