/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  Include module to implement the PCI bus scanning code for
*               graphics devices.
*
****************************************************************************/

/*-------------------------- Implementation -------------------------------*/

/****************************************************************************
DESCRIPTION:
Returns the complete list of PCI configuration spaces to the device driver

HEADER:
graphics/pciinit.h

PARAMETERS:
PCIList - Place to return the PCI device configuration space lists
count   - Place to return the count of PCI devices in list

RETURNS:
True on success, false on error.

REMARKS:
This function returns a pointer to the complete list of PCI configuration
spaces for all devices to the device driver. This is useful for device
drivers that need to access the PCI configuration space for devices other
than the graphics controller device (ie: integrated North Bridge chipsets
like the Intel i81x).
{secret}
****************************************************************************/
ibool NAPI _PCI_getPCIList(
    PCIDeviceInfo **PCIList,
    int *count)
{
    if (PCI) {
        *PCIList = PCI;
        *count = NumPCI;
        return true;
        }
    return false;
}

/****************************************************************************
PARAMETERS:
length  - Size of base address to allocate

RETURNS:
Free base address region

REMARKS:
This function finds a free slot to put the PCI base address register
and programs the register to that location. This fixes up bad BIOS BAR
allocation that occurs on some graphics cards and devices.
****************************************************************************/
static ulong PCI_findFreeBAR(
    ulong length)
{
    int     i,j,valid;
    ulong   freeBase,*base,*size,minBase = 1024 * 1024L * 1024L;

    for (freeBase = 0xFFFFFFFF - length + 1; freeBase > minBase; freeBase -= length) {
        valid = true;
        for (i = 0; i < NumPCI; i++) {
            base = (ulong*)&(PCI[i].u.type0.BaseAddress10);
            size = (ulong*)&(PCI[i].u.type0.BaseAddress10Len);
            for (j = 0x10; j <= 0x24; j += 4, base++, size++) {
                if ((*base & 0x1) || *size == 0)
                    continue;
                if (*base < freeBase && (*base + *size) > freeBase) {
                    valid = false;
                    break;
                    }
                if (freeBase < *base && (freeBase + length) > *base) {
                    valid = false;
                    break;
                    }
                }
            }
        if (valid)
            return freeBase;
        }
    return 0;
}

/****************************************************************************
REMARKS:
Fixes up the PCI base address values for the display device, to compensate
for cases where the BAR's are not properly allocated by the BIOS (ie:
a zero base address).
****************************************************************************/
static void PCI_fixupBAR(
    int device,
    int bar,
    u32 *base,
    u32 *size)
{
    /* First measure the size of the BAR */
    *base = PCI_accessReg(bar,0,PCI_READ_DWORD,&PCI[device]);
    if (!(*base & 0x1)) {
        PCI_accessReg(bar,0xFFFFFFFF,PCI_WRITE_DWORD,&PCI[device]);
        *size = PCI_accessReg(bar,0,PCI_READ_DWORD,&PCI[device]) & ~0xFF;
        *size = ~*size+1;
        PCI_accessReg(bar,*base,PCI_WRITE_DWORD,&PCI[device]);
        if (*size > 0 && *base == 0) {
            *base = PCI_findFreeBAR(*size);
            PCI_accessReg(bar,*base,PCI_WRITE_DWORD,&PCI[device]);
            }
        }
}

/****************************************************************************
PARAMETERS:
info    - PCI device information to search for AGP capabilities block

RETURNS:
Offset in config space to start of AGP capabilities block, 0 if none found.
****************************************************************************/
static uchar PCI_findAGPCaps(
    PCIDeviceInfo *info,
    PCIAGPCapability *caps)
{
    uchar capPtr;

    // Check the Status register to see if this device supports capability
    // lists. If not, it is not an AGP-compliant device.
    if ((info->Status & PCI_statusCapabilitiesList) == 0)
        return 0;

    // The device supports capability lists, find the AGP capabilities
    if ((info->HeaderType & (~PCI_multiFunctionType)) == PCI_bridgeType)
        capPtr = info->u.type1.CapabilitiesPointer;
    else
        capPtr = info->u.type0.CapabilitiesPointer;

    // Search for the AGP capabilities list
    while (capPtr != 0) {
        // Read the start of the capabilities block at this offset
        PCI_readRegBlock(info,capPtr,&caps->h,sizeof(caps->h));
        if (caps->h.capsID == PCI_capsAGP) {
            // Found the AGP capabilities block
            return capPtr;
            }
        else {
            // Some other capability, so keep looking
            capPtr = caps->h.next;
            }
        }

    // If we get here, then no AGP capabilities list was found!
    return 0;
}

/****************************************************************************
PARAMETERS:
info    - PCI device information to get for AGP capabilities block for
caps    - Place to store the returned master device capabilities structure

RETURNS:
True on success, false on failure
****************************************************************************/
static ibool PCI_getAGPCaps(
    PCIDeviceInfo *info,
    PCIAGPCapability *caps)
{
    uchar capPtr;

    // Find the header block for the AGP capabilities list
    if ((capPtr = PCI_findAGPCaps(info,caps)) == 0)
        return false;

    // Now read the rest of the AGP capabilities list
    PCI_readRegBlock(info,capPtr+sizeof(caps->h),&caps->h+1,sizeof(*caps)-sizeof(caps->h));
    return true;
}

/****************************************************************************
RETURNS:
Number of display devices found.

REMARKS:
This internal function enumerates the number of available display devices on
the PCI bus, and returns the number found.
****************************************************************************/
static int _PCI_enumerateGraphicsDevices(void)
{
    int                 i,j,size;
    PCIDeviceInfo       *info;
    PCIAGPCapability    agpCaps;

    // First determine how many devices are on the PCI bus, and allocate
    // enough memory to describe all the devices in one memory block.
    if ((NumPCI = PCI_getNumDevices()) > 0) {
        size = NumPCI * (sizeof(PCIDeviceInfo) + sizeof(int) + sizeof(int) + sizeof(int));
        if ((PCI = calloc(1,size)) == NULL)
            return -1;
        BridgeIndex = (int*)(PCI + NumPCI);
        PCIBridgeIndex = (int*)(BridgeIndex + NumPCI);
        DeviceIndex = (int*)(PCIBridgeIndex + NumPCI);
        for (i = 0; i < NumPCI; i++)
            PCI[i].dwSize = sizeof(PCI[i]);

        // Note that we re-assign NumPCI here as it is possible that when
        // we enumerate the actual values that we get a few less real
        // devices as we can do better duplicate testing once we can see
        // existing devices that have already been enumerated on the bus.
        if ((NumPCI = PCI_enumerate(PCI)) == 0)
            return -1;

        // Build a list of all PCI bridge devices
        for (i = 0,NumBridges = 0,BridgeIndex[0] = -1; i < NumPCI; i++) {
            if (PCI[i].BaseClass == PCI_BRIDGE_CLASS) {
                BridgeIndex[NumBridges++] = i;
                if (PCI[i].SubClass == PCI_HOST_BRIDGE_SUBCLASS && !AGPHostBridge) {
                    if (PCI_getAGPCaps(&PCI[i],&agpCaps))
                        AGPHostBridge = &PCI[i];
                    }
                if (PCI[i].SubClass == PCI_PCI_BRIDGE_SUBCLASS) {
                    // Keep an index of all PCI->PCI bridges based on their secondary
                    // bus number. This allows us to quickly access to bridge control
                    // registers for a device based on the secondary bus number.
                    PCIBridgeIndex[PCI[i].u.type1.SecondayBusNumber] = i;
                    }
                }
            }

        // Now build a list of all display class devices
        DeviceIndex[0] = -1;
        for (i = 0,NumDevices = 1,DeviceIndex[0] = -1; i < NumPCI; i++) {
            if (PCI_IS_DISPLAY_CLASS(&PCI[i])) {
                PCI_fixupBAR(i,0x10,&PCI[i].u.type0.BaseAddress10,&PCI[i].u.type0.BaseAddress10Len);
                PCI_fixupBAR(i,0x14,&PCI[i].u.type0.BaseAddress14,&PCI[i].u.type0.BaseAddress14Len);
                if ((PCI[i].Command & 0x3) == 0x3)
                    DeviceIndex[0] = i;
                else
                    DeviceIndex[NumDevices++] = i;
                if (PCI[i].slot.p.Bus != 0 && PCI_getAGPCaps(&PCI[i],&agpCaps)) {
                    // This device is on a different bus than the primary
                    // PCI bus, so it is probably an AGP device. Find the
                    // AGP bus device that controls that bus so we can
                    // control it.
                    for (j = 0; j < NumBridges; j++) {
                        info = &PCI[BridgeIndex[j]];
                        if (info->u.type1.SecondayBusNumber == (int)PCI[i].slot.p.Bus) {
                            AGPDevice = &PCI[i];
                            AGPBridge = info;
                            break;
                            }
                        }
                    }
                }
            }

        // If we did not find the primary PCI device, then we must have a
        // VLB/ISA device.
        if (DeviceIndex[0] == -1)
            NumDevices = 0;
        }
    else {
        // No PCI bus present, so we must be on an ISA/VLB system
        NumDevices = 0;
        }
    return NumDevices;
}

/****************************************************************************
DESCRIPTION:
Find a specific display device

HEADER:
graphics/pciinit.h

RETURNS:
PCI configuration space info for the device, NULL if not found.

REMARKS:
This function searches the PCI configuration space for the specified device.
The primary device is index 0 and secondary devices range from index 1
to the NumDevices-1. Note that we ignore all disabled devices when
searching for the primary device so it is possible for the primary device
to be enumerated on the PCI bus after the first secondary device if the
secondary device is an motherboard PCI display device since the first PCI
add-in card generally takes precedence over the motherboard controller.
{secret}
****************************************************************************/
PCIDeviceInfo *_PCI_findDevice(
    int device)
{
    if (device < NumDevices)
        return &PCI[DeviceIndex[device]];
    return NULL;
}

/****************************************************************************
DESCRIPTION:
Create a fake PCI device entry for ISA/VLB devices.

HEADER:
graphics/pciinit.h

RETURNS:
PCI configuration space info for the fake device.

REMARKS:
This function builds a fake PCI device entry for ISA/VLB devices.
{secret}
****************************************************************************/
PCIDeviceInfo * _PCI_createFakePCIEntry(
    const char *fakePCIID)
{
    static PCIDeviceInfo fakePCI;

    memset(&fakePCI,0,sizeof(fakePCI));
    fakePCI.dwSize = sizeof(fakePCI);
    sscanf(fakePCIID,"%04hX:%04hX",&fakePCI.VendorID,&fakePCI.DeviceID);
    return &fakePCI;
}

