/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  Module to implement PCI functionality specific for display
*               class devices.
*
****************************************************************************/

#include "loader.h"

/*------------------------- Global Variables ------------------------------*/

static int              NumPCI = -1;
static PCIDeviceInfo    *PCI;
static int              *BridgeIndex;
static int              *PCIBridgeIndex;
static int              *DeviceIndex;
static ulong            *BIOSImagePhys;
static int              *BIOSImageBAR;
static void             **BIOSImage;
/* {secret} */
BE_VGAInfo              *VGAInfo;
static int              NumBridges;
static int              NumDevices;
static PCIDeviceInfo    *AGPDevice = NULL;
static PCIDeviceInfo    *AGPBridge = NULL;
static PCIDeviceInfo    *AGPHostBridge = NULL;
static int              ActiveDevice = 0;

/* Length of the BIOS image */

#define MAX_BIOSLEN         (64 * 1024L)
#define FINAL_BIOSLEN       (32 * 1024L)

#ifdef  FORCE_POST
#undef  VGA_NOT_ACTIVE
#define VGA_NOT_ACTIVE()    1
#endif

/*-------------------------- Implementation -------------------------------*/

#include "pcienum.c"

#ifndef __ANALYSE__
/****************************************************************************
RETURNS:
The address to use to map the secondary BIOS (AGP devices)

REMARKS:
Searches all the PCI base address registers for the device looking for a
memory mapping that is large enough to hold our ROM BIOS. We usually end up
finding the framebuffer mapping (usually BAR 0x10), and we use this mapping
to map the BIOS for the device into. We use a mapping that is already
assigned to the device to ensure the memory range will be passed through
by any PCI->PCI or AGP->PCI bridge that may be present.

NOTE: Usually this function is only used for AGP devices, but it may be
      used for PCI devices that have already been POST'ed and the BIOS
      ROM base address has been zero'ed out.

NOTE: This function leaves the original memory aperture disabled by leaving
      it programmed to all 1's. It must be restored to the correct value
      later.
****************************************************************************/
static ulong PCI_findBIOSAddr(
    int device,
    int *bar)
{
    ulong   base,size;

    for (*bar = 0x10; *bar <= 0x14; (*bar) += 4) {
        base = PCI_readPCIRegL(*bar,device) & ~0xFF;
        if (!(base & 0x1)) {
            PCI_writePCIRegL(*bar,0xFFFFFFFF,device);
            size = PCI_readPCIRegL(*bar,device) & ~0xFF;
            size = ~size+1;
            if (size >= MAX_BIOSLEN)
                return base;
            }
        }
    return 0;
}

/****************************************************************************
REMARKS:
Re-writes the PCI base address registers for the secondary PCI controller
with the values from our initial PCI bus enumeration. This fixes up the
values after we have POST'ed the secondary display controller BIOS, which
may have incorrectly re-programmed the base registers the same as the
primary display controller (the case for identical S3 cards).
{secret}
****************************************************************************/
void _PCI_fixupSecondaryBARs(void)
{
    int i;

    for (i = 0; i < NumDevices; i++) {
        PCI_writePCIRegL(0x10,PCI[DeviceIndex[i]].u.type0.BaseAddress10,i);
        PCI_writePCIRegL(0x14,PCI[DeviceIndex[i]].u.type0.BaseAddress14,i);
        PCI_writePCIRegL(0x18,PCI[DeviceIndex[i]].u.type0.BaseAddress18,i);
        PCI_writePCIRegL(0x1C,PCI[DeviceIndex[i]].u.type0.BaseAddress1C,i);
        PCI_writePCIRegL(0x20,PCI[DeviceIndex[i]].u.type0.BaseAddress20,i);
        PCI_writePCIRegL(0x24,PCI[DeviceIndex[i]].u.type0.BaseAddress24,i);
        }
}

/****************************************************************************
RETURNS:
True if successfully initialised, false if not.

REMARKS:
This function executes the BIOS POST code on the controller. We assume that
at this stage the controller has its I/O and memory space enabled and
that all other controllers are in a disabled state.
****************************************************************************/
static ibool PCI_doBIOSPOST(
    int device,
    ulong BIOSPhysAddr,
    void *mappedBIOS,
    ulong BIOSLen)
{
    RMREGS  regs;
    RMSREGS sregs;

    // Setup the X86 emulator for the VGA BIOS
    BE_setVGA(&VGAInfo[device]);

    // Execute the BIOS POST code
    regs.x.ax = (u16)(PCI[DeviceIndex[device]].slot.i >> 8);
    BE_callRealMode(0xC000,0x0003,&regs,&sregs);

    // Re-enable interrupts. For some reason some BIOS'es may leave them
    // disables after the POST process.
    enable();

    // Cleanup and exit
    BE_getVGA(&VGAInfo[device]);
    (void)BIOSPhysAddr;
    (void)mappedBIOS;
    (void)BIOSLen;
    return true;
}
#endif

/****************************************************************************
PARAMETERS:
device  - Index of device to map the BIOS for

RETURNS:
Pointers to the mapped BIOS image

REMARKS:
Maps a pointer to the BIOS image on the graphics card, so that it can
be used
{secret}
****************************************************************************/
void * NAPI _PCI_mapBIOSImage(
    int device)
{
    // TODO: Fix to work correctly on PowerPC and across the AGP bus!!
    if (device > 0) {
        // It is assumed the controller is currently active when this
        // function is called.
        PCI_writePCIRegL(BIOSImageBAR[device],0,device);
        PCI_writePCIRegL(0x30,BIOSImagePhys[device] | 0x1,device);
        }
    return BIOSImage[device];
}

/****************************************************************************
PARAMETERS:
device  - Index of device to unmap the BIOS for

REMARKS:
Unmaps the BIOS image for the device.
{secret}
****************************************************************************/
void NAPI _PCI_unmapBIOSImage(
    int device)
{
    if (device > 0) {
        // Restore device mappings
        PCI_writePCIRegL(0x30,PCI[DeviceIndex[device]].u.type0.ROMBaseAddress,device);
        PCI_writePCIRegL(0x10,PCI[DeviceIndex[device]].u.type0.BaseAddress10,device);
        PCI_writePCIRegL(0x14,PCI[DeviceIndex[device]].u.type0.BaseAddress14,device);
        }
}

/****************************************************************************
REMARKS:
Enables the specified device, taking into account any bridges that the
device accesses must cross.
****************************************************************************/
static void AdjustDeviceAccess(
    int deviceIndex,
    int commandMask)
{
    PCIDeviceInfo *device = &PCI[DeviceIndex[deviceIndex]];

    PCI_accessReg(0x4,(device->Command & ~0x7) | commandMask,PCI_WRITE_WORD,device);
    if (device->slot.p.Bus > 0) {
        // Enable or disable the PCI->PCI bridge VGA access control for this device
        // depending on whether we are enabling I/O access for this device. Only
        // one device can have the VGA I/O accesses routed to it over the PCI bridge.
        PCIDeviceInfo *bridge = &PCI[PCIBridgeIndex[device->slot.p.Bus]];
        if (commandMask & 0x1)
            PCI_accessReg(0x3E,bridge->u.type1.BridgeControl | 0x8,PCI_WRITE_WORD,bridge);
        else
            PCI_accessReg(0x3E,bridge->u.type1.BridgeControl & ~0x8,PCI_WRITE_WORD,bridge);
        }
}

/****************************************************************************
REMARKS:
Disables the specified device, taking into account any bridges that the
device accesses must cross.
****************************************************************************/
static void DisableDevice(
    int deviceIndex)
{
    PCIDeviceInfo *device = &PCI[DeviceIndex[deviceIndex]];

    PCI_accessReg(0x4,0,PCI_WRITE_WORD,device);
    if (device->slot.p.Bus > 0) {
        // Disable the PCI->PCI bridge VGA access control for this device
        PCIDeviceInfo *bridge = &PCI[PCIBridgeIndex[device->slot.p.Bus]];
        PCI_accessReg(0x3E,bridge->u.type1.BridgeControl & ~0x8,PCI_WRITE_WORD,bridge);
        }
}

/****************************************************************************
PARAMETERS:
device  - Index of the device to POST

RETURNS:
True if successfully initialised, false if not.

REMARKS:
Loads and POST's the secondary controllers BIOS, directly from the BIOS
image we can extract over the PCI bus.
{secret}
****************************************************************************/
ibool NAPI _PCI_postController(
    N_int32 device)
{
#ifndef __ANALYSE__
    int             needToPOST;
    ulong           BIOSImageLen,mappedBIOSPhys;
    uchar           *mappedBIOS,*copyOfBIOS;
    char            filename[PM_MAX_PATH];
    FILE            *f;
    PCIDeviceInfo   info;
    ibool           success = true;

    // Load the BIOS emulator to POST the controller
    if (device == 0 || (!BE_loaded && !_GA_loadBIOSEmulator(device)))
        return false;
    if (device < 0 || device > NumDevices)
        return false;

    // Bail early if we have already POST'ed this controller
    if (VGAInfo[device].pciInfo)
        return true;

    // Disable the current display controller and bridge VGA pass-through
    DisableDevice(ActiveDevice);

    // Enable secondary display controller.
    AdjustDeviceAccess(device,0x7);

    // Determine if we need to POST this device or not. If we do not
    // have BIOS access for the primary device and it is active, then
    // we need to POST it to initialise the emulator. If we have already
    // initialised the emulator and the device is POST'ed, we simply
    // load the BIOS image from disk.
    needToPOST = false;
    if (VGA_NOT_ACTIVE()) {
        // Device is not active, so we must POST it
        needToPOST = true;
        }
    else {
        // Load the BIOS and interrupt vector information from disk
        memset(&VGAInfo[device],0,sizeof(VGAInfo[0]));
        sprintf(filename,"%sbios.%02d",SNAPConfigPath,device);
        if ((f = fopen(filename,"rb")) != NULL) {
            // Allocate memory to store copy of BIOS from secondary controllers
            if ((copyOfBIOS = malloc(FINAL_BIOSLEN)) == NULL) {
                success = false;
                goto Cleanup;
                }
            VGAInfo[device].pciInfo = &PCI[DeviceIndex[device]];
            VGAInfo[device].BIOSImage = copyOfBIOS;
            VGAInfo[device].BIOSImageLen = FINAL_BIOSLEN;
            fread(&info,1,sizeof(PCIDeviceInfo),f);
            fread(copyOfBIOS,1,FINAL_BIOSLEN,f);
            fread(VGAInfo[device].LowMem,1,sizeof(VGAInfo[device].LowMem),f);
            fclose(f);
            if (info.VendorID != PCI[DeviceIndex[device]].VendorID ||
                info.DeviceID != PCI[DeviceIndex[device]].DeviceID ||
                info.RevID != PCI[DeviceIndex[device]].RevID ||
                info.u.type0.SubSystemVendorID != PCI[DeviceIndex[device]].u.type0.SubSystemVendorID ||
                info.u.type0.SubSystemID != PCI[DeviceIndex[device]].u.type0.SubSystemID) {
                free(copyOfBIOS);
                memset(&VGAInfo[device],0,sizeof(VGAInfo[0]));
                needToPOST = true;
                }
            }
        else
            needToPOST = true;
        }

    // POST the controller if necessary
    if (needToPOST) {
        // Find a viable place to map the secondary PCI BIOS image and map it
        _log("Device %d not enabled, so attempting warm boot it\n", device);

        // For AGP devices we have to map the BIOS to a location that
        // is passed by the AGP bridge to the bus. Some AGP devices
        // have the ROM base address already set up for us, and some
        // do not (we map to one of the existing BAR locations in
        // this case).
        if (PCI[DeviceIndex[device]].u.type0.ROMBaseAddress != 0) {
            mappedBIOSPhys = PCI[DeviceIndex[device]].u.type0.ROMBaseAddress & ~0xF;
            BIOSImageBAR[device] = 0x30;
            }
        else
            mappedBIOSPhys = PCI_findBIOSAddr(device,&BIOSImageBAR[device]);
        _log("Mapping BIOS image to 0x%08X\n", mappedBIOSPhys);
        mappedBIOS = PM_mapPhysicalAddr(mappedBIOSPhys,MAX_BIOSLEN-1,false);
        PCI_writePCIRegL(0x30,mappedBIOSPhys | 0x1,device);
        BIOSImageLen = mappedBIOS[2] * 512;
        if ((copyOfBIOS = malloc(BIOSImageLen)) == NULL) {
            success = false;
            goto Cleanup;
            }
        memcpy(copyOfBIOS,mappedBIOS,BIOSImageLen);
        BIOSImagePhys[device] = mappedBIOSPhys;
        BIOSImage[device] = mappedBIOS;

        // Allocate memory to store copy of BIOS from secondary controllers
        VGAInfo[device].pciInfo = &PCI[DeviceIndex[device]];
        VGAInfo[device].BIOSImage = copyOfBIOS;
        VGAInfo[device].BIOSImageLen = BIOSImageLen;

        // Restore device mappings
        PCI_writePCIRegL(0x30,PCI[DeviceIndex[device]].u.type0.ROMBaseAddress,device);
        PCI_writePCIRegL(0x10,PCI[DeviceIndex[device]].u.type0.BaseAddress10,device);
        PCI_writePCIRegL(0x14,PCI[DeviceIndex[device]].u.type0.BaseAddress14,device);

        // Now execute the BIOS POST for the device
        if (copyOfBIOS[0] == 0x55 && copyOfBIOS[1] == 0xAA) {
            _log("Executing BIOS POST for controller.\n");
            if (!PCI_doBIOSPOST(device,mappedBIOSPhys,copyOfBIOS,BIOSImageLen)) {
                success = false;
                goto Cleanup;
                }
            }
        else {
            success = false;
            goto Cleanup;
            }

        // Reset the size of the BIOS image to the final size
        VGAInfo[device].BIOSImageLen = FINAL_BIOSLEN;

        // Save the BIOS and interrupt vector information to disk
        sprintf(filename,"%sbios.%02d",SNAPConfigPath,device);
        if ((f = fopen(filename,"wb")) != NULL) {
            fwrite(&PCI[DeviceIndex[device]],1,sizeof(PCIDeviceInfo),f);
            fwrite(copyOfBIOS,1,FINAL_BIOSLEN,f);
            fwrite(VGAInfo[device].LowMem,1,sizeof(VGAInfo[device].LowMem),f);
            fclose(f);
            }
        _log("\n");
        }
     else {
        // Log that the BIOS emulator is enabled
        _log("Enabling BIOS emulator for device %d\n\n", device);
        }

    // Fix up all the secondary PCI base address registers
    // (restores them all from the values we read previously)
    _PCI_fixupSecondaryBARs();

    // Disable the secondary controller
    DisableDevice(device);

    // Reenable original active controller
Cleanup:
    AdjustDeviceAccess(ActiveDevice,0x7);

    // Restore the X86 emulator BIOS info to primary controller if loaded
    if (BE_loaded)
        BE_setVGA(&VGAInfo[ActiveDevice]);
    return success;
#else
    return true;
#endif
}

/****************************************************************************
RETURNS:
True if successfully initialised, false if not.

REMARKS:
Loads the BIOS emulator for the primary controller, which we use if we do
not have native BIOS access.
{secret}
****************************************************************************/
static ibool _PCI_loadBIOSEmulatorForPrimary(void)
{
#if !defined(__ANALYSE___) && defined(__INTEL__)
    N_uint32    *emuLowMem,*realLowMem;

    // Load the BIOS emulator to POST the controller
    if (!BE_loaded && !_GA_loadBIOSEmulator(0))
        return false;

    // Log that the BIOS emulator is enabled
    _log("Enabling BIOS emulator for primary device\n\n");

    // Set up the BIOS emulator internals
    realLowMem = (_PM_importsOrg.PM_mapRealPointer)(0,0);
    VGAInfo[0].pciInfo = &PCI[DeviceIndex[0]];
    VGAInfo[0].BIOSImage = BIOSImage[0];
    VGAInfo[0].BIOSImageLen = FINAL_BIOSLEN;
    BE_setVGA(&VGAInfo[0]);
    BE_getVGA(&VGAInfo[0]);

    // Now copy the important interrupt vectors from the original
    // machine state into the BIOS emulator state. Note that we use
    // Int 0x6D for the main BIOS entry point, as this will not be
    // hooked and will be pointing to the correct location.
    emuLowMem = (N_uint32*)VGAInfo[0].LowMem;
    if ((realLowMem[0x10] >> 16) != 0xC000)
        emuLowMem[0x10] = realLowMem[0x6D];
    else
        emuLowMem[0x10] = realLowMem[0x10];
    emuLowMem[0x1F] = realLowMem[0x1F];
    emuLowMem[0x43] = realLowMem[0x43];
    emuLowMem[0x6D] = realLowMem[0x6D];

    // Now copy the BIOS data area across
    memcpy((char*)emuLowMem + 0x400,(char*)realLowMem + 0x400,sizeof(VGAInfo[0].LowMem) - 0x400);

    // Finally restore the BIOS emulator state
    BE_setVGA(&VGAInfo[0]);
#endif
    return true;
}

/****************************************************************************
REMARKS:
Enumerates the device index information to the end of the config info line.
****************************************************************************/
static void LogDeviceIndex(
    int i)
{
    int index;

    for (index = 0; index < NumDevices; index++) {
        if (DeviceIndex[index] == i)
            break;
        }
    if (index < NumDevices) {
        if (&PCI[i] == AGPDevice)
            _log("<- %d (AGP Card)\n", index);
        else
            _log("<- %d\n", index);
        }
    else
        _log("\n");
}

/****************************************************************************
REMARKS:
Enumerates the PCI bus and dumps the PCI configuration information to the
log file.
****************************************************************************/
static void EnumeratePCI(void)
{
    int             i,index;
    PCIDeviceInfo   *info;

    _log("Displaying enumeration of PCI bus (%d devices, %d display devices)\n",
        NumPCI, NumDevices);
    for (index = 0; index < NumDevices; index++)
        _log("  Display device %d is PCI device %d\n",index,DeviceIndex[index]);
    _log("\n");
    _log("Bus Slot Fnc DeviceID  SubSystem Rev Class IRQ Base10h  Base14h  Cmd\n");
    for (i = 0; i < NumPCI; i++) {
        _log("%2d   %2d  %2d  %04X:%04X %04X:%04X %02X  %02X:%02X %02X  %08X %08X %04X   ",
            PCI[i].slot.p.Bus,
            PCI[i].slot.p.Device,
            PCI[i].slot.p.Function,
            PCI[i].VendorID,
            PCI[i].DeviceID,
            PCI[i].u.type0.SubSystemVendorID,
            PCI[i].u.type0.SubSystemID,
            PCI[i].RevID,
            PCI[i].BaseClass,
            PCI[i].SubClass,
            PCI[i].u.type0.InterruptLine,
            PCI[i].u.type0.BaseAddress10,
            PCI[i].u.type0.BaseAddress14,
            PCI[i].Command);
        LogDeviceIndex(i);
        }
    _log("\n");
    _log("DeviceID  Stat Ifc Cch Lat Hdr BIST Int Base18h  Base1Ch  Base20h  Base24h  \n");
    for (i = 0; i < NumPCI; i++) {
        _log("%04X:%04X %04X  %02X  %02X  %02X  %02X  %02X  %02X  %08X %08X %08X %08X ",
            PCI[i].VendorID,
            PCI[i].DeviceID,
            PCI[i].Status,
            PCI[i].Interface,
            PCI[i].CacheLineSize,
            PCI[i].LatencyTimer,
            PCI[i].HeaderType,
            PCI[i].BIST,
            PCI[i].u.type0.InterruptPin,
            PCI[i].u.type0.BaseAddress18,
            PCI[i].u.type0.BaseAddress1C,
            PCI[i].u.type0.BaseAddress20,
            PCI[i].u.type0.BaseAddress24);
        LogDeviceIndex(i);
        }
    _log("\n");
    _log("DeviceID  BAR10Len BAR14Len BAR18Len BAR1CLen BAR20Len BAR24Len ROMLen\n");
    for (i = 0; i < NumPCI; i++) {
        _log("%04X:%04X %08X %08X %08X %08X %08X %08X %08X ",
            PCI[i].VendorID,
            PCI[i].DeviceID,
            PCI[i].u.type0.BaseAddress10Len,
            PCI[i].u.type0.BaseAddress14Len,
            PCI[i].u.type0.BaseAddress18Len,
            PCI[i].u.type0.BaseAddress1CLen,
            PCI[i].u.type0.BaseAddress20Len,
            PCI[i].u.type0.BaseAddress24Len,
            PCI[i].u.type0.ROMBaseAddressLen);
        LogDeviceIndex(i);
        }
    _log("\n");
    _log("Displaying enumeration of %d bridge devices\n",NumBridges);
    _log("\n");
    _log("DeviceID  P# S# B# IOB  IOL  MemBase  MemLimit PreBase  PreLimit Ctrl\n");
    for (i = 0; i < NumBridges; i++) {
        info = &PCI[BridgeIndex[i]];
        _log("%04X:%04X %02X %02X %02X %04X %04X %08X %08X %08X %08X %04X ",
            info->VendorID,
            info->DeviceID,
            info->u.type1.PrimaryBusNumber,
            info->u.type1.SecondayBusNumber,
            info->u.type1.SubordinateBusNumber,
            ((u16)info->u.type1.IOBase << 8) & 0xF000,
            info->u.type1.IOLimit ?
                ((u16)info->u.type1.IOLimit << 8) | 0xFFF : 0,
            ((u32)info->u.type1.MemoryBase << 16) & 0xFFF00000,
            info->u.type1.MemoryLimit ?
                ((u32)info->u.type1.MemoryLimit << 16) | 0xFFFFF : 0,
            ((u32)info->u.type1.PrefetchableMemoryBase << 16) & 0xFFF00000,
            info->u.type1.PrefetchableMemoryLimit ?
                ((u32)info->u.type1.PrefetchableMemoryLimit << 16) | 0xFFFFF : 0,
            info->u.type1.BridgeControl);
        if (info == AGPBridge)
            _log("<- AGP Bridge\n");
        else if (info == AGPHostBridge)
            _log("<- AGP Host Bridge\n");
        else
            _log("\n");
        }
    _log("\n");
}

/****************************************************************************
PARAMETERS:
info    - PCI device information to get for AGP capabilities block for
caps    - Place to store the returned master device capabilities structure

RETURNS:
True on success, false on failure
****************************************************************************/
static ibool PCI_setAGPCaps(
    PCIDeviceInfo *info,
    PCIAGPCapability *caps)
{
    uchar capPtr;

    // Find the header block for the AGP capabilities list
    if ((capPtr = PCI_findAGPCaps(info,caps)) == 0)
        return false;

    // Now write the AGP command register
    PCI_writeRegBlock(info,
        capPtr+PM_FIELD_OFFSET(PCIAGPCapability,AGPCommand),
        &caps->AGPCommand,sizeof(caps->AGPCommand));
    return true;
}

/****************************************************************************
PARAMETERS:
enableFastWrite - True to enable AGP FastWrite mode
maxAGPRate      - Maximum AGP data rate to enable

REMARKS:
Function to initialise the AGP device capabilities. Specifically we want to
turn on AGP FastWrite mode if possible, as this will give a big performance
boost for AGP graphics cards in AGP 4x capable systems.
****************************************************************************/
static ibool InitAGPCapabilities(
    ibool enableFastWrite,
    int maxAGPRate)
{
    PCIAGPCapability    masterCap;
    PCIAGPCapability    targetCap;
    ulong               dataRate;
    ulong               fastWrite;
    ulong               SBAEnable;

    // First attempt to get the AGP capabilities for each device
    if (!PCI_getAGPCaps(AGPDevice,&masterCap))
        return false;
    if (!PCI_getAGPCaps(AGPHostBridge,&targetCap))
        return false;
    // Some broken cards (Matrox Millenium II "AGP") report no valid
    // supported transfer rates. These are not really AGP cards. They
    // have an AGP Capabilities structure that reports no capabilities.
    if (masterCap.AGPStatus.rate == 0)
        return false;

    // Determine the greatest common denominator for data rate
    dataRate = targetCap.AGPStatus.rate & masterCap.AGPStatus.rate;
    if (targetCap.AGPStatus.agp3mode) {
        // AGP 3.0 and later devices, where maximum data rate is 8x
        if (maxAGPRate < 4)
            maxAGPRate = 4;
        else if (maxAGPRate > 8)
            maxAGPRate = 8;
        if (dataRate & 2)
            dataRate = MIN(8,maxAGPRate) >> 2;
        else if (dataRate & 1)
            dataRate = MIN(4,maxAGPRate) >> 2;
        }
    else {
        // AGP 2.0 and earlier devices, where maximum data rate is 4x
        if (maxAGPRate < 1)
            maxAGPRate = 1;
        else if (maxAGPRate > 4)
            maxAGPRate = 4;
        if (dataRate & 4)
            dataRate = MIN(4,maxAGPRate);
        else if (dataRate & 2)
            dataRate = MIN(2,maxAGPRate);
        else if (dataRate & 1)
            dataRate = MIN(1,maxAGPRate);
        }

    // Enable side band addressing if both master and target support it
    SBAEnable = (targetCap.AGPStatus.sideBandAddressing & masterCap.AGPStatus.sideBandAddressing);

    // Enable FastWrite if both master and target support it
    fastWrite = enableFastWrite && ((targetCap.AGPStatus.fastWrite & masterCap.AGPStatus.fastWrite));

    // Enable the Target device first
    targetCap.AGPCommand.rate = dataRate;
    targetCap.AGPCommand.AGPEnable = 1;
    targetCap.AGPCommand.SBAEnable = SBAEnable;
    targetCap.AGPCommand.fastWriteEnable = fastWrite;
    targetCap.AGPCommand.fourGBEnable = 0;
    if (!PCI_setAGPCaps(AGPDevice,&targetCap))
        return false;

    // Now enable the Master device
    masterCap.AGPCommand.rate = dataRate;
    masterCap.AGPCommand.AGPEnable = 1;
    masterCap.AGPCommand.SBAEnable = SBAEnable;
    masterCap.AGPCommand.fastWriteEnable = fastWrite;
    masterCap.AGPCommand.fourGBEnable = 0;
    masterCap.AGPCommand.requestQueueDepth = targetCap.AGPStatus.requestQueueDepthMaximum;
    if (!PCI_setAGPCaps(AGPHostBridge,&masterCap))
        return false;
    return true;
}

/****************************************************************************
REMARKS:
Function to read the hardware AGP capabilities and report both the
target and master capabilities to the log file.
****************************************************************************/
static void LogAGPCapabilities(void)
{
    PCIAGPCapability    masterCap;
    PCIAGPCapability    targetCap;
    int                 maxMasterRate;
    int                 maxTargetRate;
    int                 masterRate;
    int                 targetRate;

    // First attempt to get the AGP capabilities for each device
    if (!PCI_getAGPCaps(AGPDevice,&masterCap))
        return;
    if (!PCI_getAGPCaps(AGPHostBridge,&targetCap))
        return;
    // Some broken cards (Matrox Millenium II "AGP") report no valid
    // supported transfer rates. These are not really AGP cards. They
    // have an AGP Capabilities structure that reports no capabilities.
    if (masterCap.AGPStatus.rate == 0)
        return;

    // Find the maximum rates supports by the two devices
    maxMasterRate = masterCap.AGPStatus.rate;
    maxTargetRate = targetCap.AGPStatus.rate;
    masterRate = masterCap.AGPCommand.rate;
    targetRate = targetCap.AGPCommand.rate;
    if (masterCap.AGPStatus.agp3mode) {
        // AGP 3.0 and later devices, where maximum data rate is 8x
        if (maxMasterRate & 2)
            maxMasterRate = 8;
        else if (maxMasterRate & 1)
            maxMasterRate = 4;
        if (maxTargetRate & 2)
            maxTargetRate = 8;
        else if (maxTargetRate & 1)
            maxTargetRate = 4;
        if (masterRate & 2)
            masterRate = 8;
        else if (masterRate & 1)
            masterRate = 4;
        if (targetRate & 2)
            targetRate = 8;
        else if (targetRate & 1)
            targetRate = 4;
        }
    else {
        // AGP 2.0 and ealier devices, where maximum data rate is 4x
        if (maxMasterRate & 4)
            maxMasterRate = 4;
        else if (maxMasterRate & 2)
            maxMasterRate = 2;
        else if (maxMasterRate & 1)
            maxMasterRate = 1;
        if (maxTargetRate & 4)
            maxTargetRate = 4;
        else if (maxTargetRate & 2)
            maxTargetRate = 2;
        else if (maxTargetRate & 1)
            maxTargetRate = 1;
        }

    // Log what capabilities are available
    _log("AGP device capabilities:     AGP Bridge   AGP Card\n");
    _log("  Max AGP texturing rate ... AGP %dX       AGP %dX\n",
        maxTargetRate,
        maxMasterRate);
    _log("  AGP side band addressing . %-3s          %-3s\n",
        targetCap.AGPStatus.sideBandAddressing ? "Yes" : "No",
        masterCap.AGPStatus.sideBandAddressing ? "Yes" : "No");
    _log("  AGP fast write ........... %-3s          %-3s\n",
        targetCap.AGPStatus.fastWrite ? "Yes" : "No",
        masterCap.AGPStatus.fastWrite ? "Yes" : "No");
    _log("  AGP 4GB enable ........... %-3s          %-3s\n",
        targetCap.AGPStatus.fourGB ? "Yes" : "No",
        masterCap.AGPStatus.fourGB ? "Yes" : "No");
    _log("  AGP request queue max .... %-3d          %-3d\n",
        targetCap.AGPStatus.requestQueueDepthMaximum,
        masterCap.AGPStatus.requestQueueDepthMaximum);
    _log("\n");

    // Log what capabilities are enabled
    _log("AGP device features:         AGP Bridge   AGP Card\n");
    _log("  AGP enable ............... %-3s          %-3s\n",
        targetCap.AGPCommand.AGPEnable ? "On" : "Off",
        masterCap.AGPCommand.AGPEnable ? "On" : "Off");
    _log("  AGP texturing rate ....... AGP %dX       AGP %dX\n",
        targetRate,
        masterRate);
    _log("  AGP side band addressing . %-3s          %-3s\n",
        targetCap.AGPCommand.SBAEnable ? "On" : "Off",
        masterCap.AGPCommand.SBAEnable ? "On" : "Off");
    _log("  AGP fast write ........... %-3s          %-3s\n",
        targetCap.AGPCommand.fastWriteEnable ? "On" : "Off",
        masterCap.AGPCommand.fastWriteEnable ? "On" : "Off");
    _log("  AGP 4GB enable ........... %-3s          %-3s\n",
        targetCap.AGPCommand.fourGBEnable ? "On" : "Off",
        masterCap.AGPCommand.fourGBEnable ? "On" : "Off");
    _log("  AGP request queue depth .. %-3d          %-3d\n",
        targetCap.AGPCommand.requestQueueDepth,
        masterCap.AGPCommand.requestQueueDepth);
    _log("\n");
}

/****************************************************************************
PARAMETERS:
enableFastWrite - True to enable AGP fast write mode
maxAGPRate      - Maximum AGP data rate

RETURNS:
Number of display devices found.

REMARKS:
This function enumerates the number of available display devices on the
PCI bus, and returns the number found. The number of devices allows you
to load graphics drivers for each device, but calling DDC_loadDriver starting
with device 0 and up to the maximum number of devices-1. Note that
even though we may enumarate a valid PCI display device, it is possible
that no graphics driver is available for the device and the DDC_loadDriver
function could fail.

Hence this function does not enumerate all the devices and match them up
with valid graphics drivers; the matching to available drivers is done when
you call DDC_loadDriver.

Note:   ISA and VLB devices do not support multiple controllers, and will
        always return -1 for systems without a PCI bus. Note however that
        we do initialise internal structures necessary for mapping BIOS
        images and using the BIOS emulator.

Note:   This function is a part of the standard loader library, and
        as such is not a part of the core specification. The loader library
        internals may change over time between official specification
        releases.
{secret}
****************************************************************************/
int _PCI_enumerateDevices(
    ibool enableFastWrite,
    int maxAGPRate)
{
    int size,maxDevices;

    // If this is the first time we have been called, enumerate all
    // devices on the PCI bus.
    if (NumPCI == -1) {
        // Detect the graphics devices on the bus
        if ((NumDevices = _PCI_enumerateGraphicsDevices()) > 0) {
            // Now display the list of PCI devices
            EnumeratePCI();

            // Now initialise the AGP capabilities for the card
            if (AGPDevice && AGPHostBridge) {
                InitAGPCapabilities(enableFastWrite,maxAGPRate);
                LogAGPCapabilities();
                }
            }

        // Allocate memory for our array of objects in a single memory block
        maxDevices = (NumDevices == 0) ? 1 : NumDevices;
        size = maxDevices * (sizeof(ulong) + sizeof(int) + sizeof(void*) + sizeof(BE_VGAInfo));
        if ((BIOSImagePhys = calloc(1,size)) == NULL)
            return -1;
        BIOSImageBAR = (int*)(BIOSImagePhys + maxDevices);
        BIOSImage = (void**)(BIOSImageBAR + maxDevices);
        VGAInfo = (BE_VGAInfo*)(BIOSImage + maxDevices);

        // Do *not* support multi-controller in a Windows DOS box!!!
        if (_IS_OS_CLASS(_OS_CLASS_DOS)) {
            RMREGS regs;
            regs.x.ax = 0x160A;
            PM_int86(0x2F,&regs,&regs);
            if (regs.x.ax == 0)
                NumDevices = 1;
            }

        // TODO: This needs to be fixed for PPC!

        // Map the BIOS image for the primary controller at 0xC0000
        BIOSImagePhys[0] = 0xC0000;
        BIOSImage[0] = PM_mapPhysicalAddr(0xC0000,MAX_BIOSLEN-1,true);

        // If we have no BIOS access, load the BIOS emulator for the
        // primary controller
        if (!_IS_OS_CLASS(_OS_CLASS_WIN32) && !PM_haveBIOSAccess())
            _PCI_loadBIOSEmulatorForPrimary();
        }
    return NumDevices;
}

/****************************************************************************
DESCRIPTION:
Check if the specific IRQ line is shared by any PCI devices

HEADER:
graphics/pciinit.h

PARAMETERS:
interruptLine   - Interrupt line to check for shared IRQ's on

RETURNS:
True if shared, false if not.

REMARKS:
This function scans all devices on the PCI bus and determines if any of the
devices share the same IRQ as the passed in IRQ.
{secret}
****************************************************************************/
ibool _PCI_isSharedIRQ(
    PCIDeviceInfo *info)
{
    int             i;

    // If the IntPin is not enabled, then the device does not support the
    // IRQ line so we are set.
    if (info->u.type0.InterruptPin == 0)
        return false;

    // If the IRQ line is disable, it is never shared (IRQ 0 is not a valid
    // PCI interrupt line, and is used to indicate that no interrupt is
    // routed to that device).
    if (info->u.type0.InterruptLine == 0 || info->u.type0.InterruptLine == 0xFF)
        return false;

    // Now check the list of all PCI devices for a shared interrupt
    for (i = 0; i < NumPCI; i++) {
        if (PCI[i].slot.i == info->slot.i)
            continue;
        if (PCI[i].u.type0.InterruptLine == info->u.type0.InterruptLine)
            return true;
        }
    return false;
}

/****************************************************************************
DESCRIPTION:
Sets the active display device for multiple display devices

HEADER:
graphics/pciinit.h

PARAMETERS:
deviceIndex - Index of device to make active

RETURNS:
True on success, false on error.

REMARKS:
This function allows the application to switch between the primary and
secondary display controllers, making one of the controllers the active
controller. When a controller is the active controller it will function
as though it is the only display controller in the system. You can also
select a 'mixed' mode of operation by passing in the GA_DEVICE_MIXED_MODED
parameter, which fully enables the primary display controller but only
enables the secondary display controllers memory mappings. If mixed mode
is selected, it is the responsibility of the /calling/ application to
ensure that the secondary controller's VGA memory spaces are disabled
/before/ setting this mode to avoid conflicts between the controllers.

The device numbering starts at 0 for the primary display controller, and
increments by one for each supports display controller. Ie: the second
controller is device 1 while the third controller is device 2 etc.
{secret}
****************************************************************************/
ibool _PCI_setActiveDevice(
    int deviceIndex)
{
    int i,dev = deviceIndex & DEVICE_MASK;

    // Enable the appropriate devices
    if (deviceIndex & GA_DEVICE_MIXED_MODE) {
        // Enable mixed mode for all secondary controllers
        for (i = 0; i < NumDevices; i++) {
            if (i != deviceIndex)
                AdjustDeviceAccess(i,0x6);
            }
        }
    else {
        // Disable all secondary devices first
        for (i = 0; i < NumDevices; i++) {
            if (i != deviceIndex)
                DisableDevice(i);
            }
        }

    // Now enable the new device as active
    AdjustDeviceAccess(ActiveDevice = dev,0x7);
    return true;
}

/****************************************************************************
DESCRIPTION:
Determines if the device controlled by this device context is on the AGP bus

HEADER:
graphics/pciinit.h

PARAMETERS:
deviceIndex - Index of the device to check the bus type for

RETURNS:
True if on the AGP bus, false if not.

REMARKS:
This function determines if the specified display device is on the AGP bus
or if it is on the PCI bus.
{secret}
****************************************************************************/
ibool NAPI _PCI_isAGPDevice(
    int deviceIndex)
{
    if (!AGPBridge)
        return false;
    return (AGPBridge->u.type1.SecondayBusNumber == (int)PCI[DeviceIndex[deviceIndex]].slot.p.Bus);
}

