/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  C module to determine the recommended display modes for
*               the SNAP drivers.
*
****************************************************************************/

#include "loader.h"
#include "snap/common/list.h"
#include "snap/common/mem.h"

//------------------------ Global vars -----------------------------------

static N_blockList  bl;
static N_node       *freeList;

/****************************************************************************
REMARKS:
Resizes the block list if full by increasing its size by the original
default size and building a new free list in the resized portion.
****************************************************************************/
static ibool resizeBlockList(void)
{
    if (!N_resizeBlockList(&bl))
        return false;
    freeList = N_buildFreeList(&bl);
    return true;
}

/****************************************************************************
RETURNS:
Pointer to new structure

REMARKS:
Allocates space for a new block in the buffer pool from the free list. If
there are no more buffers, we attempt to re-size the buffer pool which will
bomb out if we run out of memory.
****************************************************************************/
static GA_recMode *newBlock(void)
{
    N_node  *n;

    if (freeList == NULL) {
        if (!resizeBlockList())
            return NULL;
        }
    n = freeList;
    freeList = N_FREELIST_NEXT(n);
    return (GA_recMode*)LST_USERSPACE(n);
}

/****************************************************************************
REMARKS:
Scan the list of resolutions finding the recommended resolution first,
and if that fails the highest resolution available. If that fails no
mode is recommended (ie: no modes available).
****************************************************************************/
static ibool SetRecommended(
    GA_recMode *recommendedMode,
    N_list *recList,
    int idealXRes,
    int idealYRes)
{
    GA_recMode  *node;

    // Find the highest resolution that matches our ideal mode resolution
    for (node = N_firstNode(recList); node; node = N_nextNode(node)) {
        if (node->XResolution <= idealXRes &&
            node->YResolution <= idealYRes) {
            *recommendedMode = *node;
            return true;
            }
        }

    // If we don't find the ideal resolution, then we must use the first
    // mode in the list which is the highest resolution mode available (which
    // will be lower than the ideal resolution).
    if ((node = N_firstNode(recList)) != NULL) {
        *recommendedMode = *node;
        return true;
        }

    // No mode was found, so clear this entry
    memset(recommendedMode,0,sizeof(*recommendedMode));
    return false;
}

/****************************************************************************
DESCRIPTION:
Determine the recommended display modes for the currently active output device.

HEADER:
snap/graphics.h

PARAMETERS:
dc                  - Device context to get recommended modes for
outputDevice        - Output device to get recommended modes for
outputHead          - Output head to get recommended modes for
recommendedMode     - Overall recommended display mode
recommendedMode8    - 8bpp recommended display mode
recommendedMode16   - 16bpp recommended display mode
recommendedMode24   - 24bpp recommended display mode
recommendedMode32   - 32bpp recommended display mode

RETURNS:
True on success, false on error.

REMARKS:
This function queries the SNAP device driver mode list and monitor
configuration information to determine the best recommended display
modes to be used overall and for each available color depth. If a
color depth is not available, the recommended mode information is set
to all zeros.

NOTE: This function is heavy on stack usage, so don't call it from an
      environment which only has a small stack (ie: less than 16K or so)!
****************************************************************************/
ibool NAPI GA_getRecommendedModes(
    GA_devCtx *dc,
    N_int32 outputDevice,
    N_int32 outputHead,
    GA_recMode *recommendedMode,
    GA_recMode *recommendedMode8,
    GA_recMode *recommendedMode16,
    GA_recMode *recommendedMode24,
    GA_recMode *recommendedMode32)
{
    int                 idealXRes,idealYRes,idealPPI = 85;
    GA_initFuncs        init;
    GA_globalOptions    gOptions;
    GA_options          opt;
    N_uint16            *modes;
    GA_modeInfo	        mi;
    EDID_record         rec;
    N_list              recList8,recList15,recList16,recList24,recList32;
    int                 i,refresh,have1280x960 = false;
    int                 currentAspect,modeAspect;
    GA_monitor          monitor;
    GA_recMode          *recMode;

    // Create the block list to efficiently manage small memory blocks
    if (!N_createBlockList(&bl,sizeof(GA_recMode) + sizeof(N_node)))
        return false;
    freeList = N_buildFreeList(&bl);

    // Initialise all the mode lists
    N_initList(&recList8);
    N_initList(&recList15);
    N_initList(&recList16);
    N_initList(&recList24);
    N_initList(&recList32);

    // Get all the info we need from the SNAP drivers
    init.dwSize = sizeof(init);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&init))
        return false;
    opt.dwSize = sizeof(opt);
    init.GetOptions(&opt);
    gOptions.dwSize = sizeof(gOptions);
    GA_getGlobalOptions(&gOptions,false);
    init.GetMonitorInfo(&monitor,outputHead);

    // Calculate the current screen aspect ratio.
    currentAspect = 1333;
    if (monitor.flags & Monitor_16to9)
        currentAspect = 1777;

    // Get all modes and store them into lists of valids modes in lists
    // per color depth. We cull modes less then 640x480, any
    // aspect ratio not equal to the current desktop aspect ratio and
    // modes with a maximum refresh rate < 75Hz.
    for (modes = dc->AvailableModes; *modes != 0xFFFF; modes++) {
        mi.dwSize = sizeof(mi);
        if (init.GetVideoModeInfoExt(*modes,&mi,outputDevice,outputHead) != 0)
            continue;
        if (mi.Attributes & gaIsTextMode)
            continue;
        if (mi.Attributes & gaIsGUIDesktop)
            continue;
        if (mi.AttributesExt & gaIsPanningMode)
            continue;

        // Never recommend a virtual mode!
        if (mi.Attributes & gaIsVirtualMode)
            continue;
        if ((mi.XResolution < 640) || (mi.YResolution < 480) || (mi.BitsPerPixel < 8))
            continue;
        if (mi.XResolution == 1280 && mi.YResolution == 960)
            have1280x960 = true;

        // Check that the aspect ratio matches the current aspect ratio for
        // the attached monitor.
        modeAspect = (mi.XResolution * 1000) / mi.YResolution;
        if (currentAspect != modeAspect) {
            if (mi.XResolution != 1280 || mi.YResolution != 1024 || have1280x960)
                continue;
            }

        // Only check refresh rates if the CRT is the only attached display
        refresh = 0;
        if (!(outputDevice & (gaOUTPUT_LCD | gaOUTPUT_DFP | gaOUTPUT_TV))) {
            // Loop through the refresh list while the refresh is less then
            // 75. If refreshrate is -1 then I got to the end of the list before
            // findind a refresh rate 75 or higher so discard mode and try next;
            for (i = 0; mi.RefreshRateList[i] != -1; i++) {
                if (mi.RefreshRateList[i] > 85)
                    break;
                if (mi.RefreshRateList[i] >= 75)
                    refresh = mi.RefreshRateList[i];
                }

            // If we don't have a refresh rate between 75 and 85 Hz (we cap the
            // max recommended refresh rate at 85Hz) then we skip this mode.
            // Note that if there are *no* refresh rates, we use a value of
            // zero to indicate the 'BIOS default' will be used (ie: VBE/Core
            // 2.0 fallback driver or flat panel etc).
            if (refresh == 0 && mi.RefreshRateList[0] != -1)
                continue;
            }
        recMode = newBlock();
        recMode->XResolution  = mi.XResolution;
        recMode->YResolution  = mi.YResolution;
        recMode->BitsPerPixel = (N_uint8)mi.BitsPerPixel;
        recMode->RefreshRate  = (N_int8)refresh;

        // Now add the mode to the appropriate list of modes for each color depth.
        switch (mi.BitsPerPixel) {
            case 8:
                N_addToHead(&recList8,recMode);
                break;
            case 15:
                N_addToHead(&recList15,recMode);
                break;
            case 16:
                N_addToHead(&recList16,recMode);
                break;
            case 24:
                // We exclude all partially accelerated 24bpp modes
                if (!(dc->WorkArounds & gaSlow24bpp))
                    N_addToHead(&recList24,recMode);
                break;
            case 32:
                // We exclude all partially accelerated 32bpp modes
                if (!(dc->WorkArounds & gaSlow32bpp))
                    N_addToHead(&recList32,recMode);
                break;
            }
        }

    // Find the ideal mode resolution based on the monitor size. We try to get
    // this from the EDID if possible, otherwise we get it from the cookie
    // set by the monitor size applet.
    if (outputDevice & (gaOUTPUT_LCD | gaOUTPUT_DFP)) {
        // LCD or DFP panel is active, so recommend the panel size
        idealXRes = opt.LCDPanelWidth;
        idealYRes = opt.LCDPanelHeight;
        }
    else if (outputDevice & gaOUTPUT_TV) {
        // TV is active, so recommend the maximum TV resolution
        idealXRes = opt.TVMaxWidth;
        idealYRes = opt.TVMaxHeight;
        }
    else if (GA_getParsedEDID(dc,&rec)){
        // This will be in centimeters so it must be coverted.
        idealXRes = (int)(((int)rec.maxHSize / 2.54) * idealPPI);
        idealYRes = (int)(((int)rec.maxVSize / 2.54) * idealPPI);
        }
    else {
        // Must be running only on CRT display with no other devices active.
        if (!gOptions.wMonitorHSize || !gOptions.wMonitorVSize) {
            // No idea, so choose base fallback modes
            idealXRes = 0;
            idealYRes = 0;
            }
        else {
            // This is always in inches. It will be converted as necessary
            // by the monitor size applet.
            idealXRes = (gOptions.wMonitorHSize / 100.0) * idealPPI;
            idealYRes = (gOptions.wMonitorVSize / 100.0) * idealPPI;
            }
        }

    // Now find the recommended modes
    if (!(dc->Attributes & gaHaveAccel2D) || idealXRes == 0) {
        // If the device has no acceleration or we did not know the
        // monitor size, then return default mode of 800x600 for all
        // color depths
        SetRecommended(recommendedMode8,&recList8,800,600);
        if (!SetRecommended(recommendedMode16,&recList16,800,600))
            SetRecommended(recommendedMode16,&recList15,800,600);
        SetRecommended(recommendedMode24,&recList24,800,600);
        SetRecommended(recommendedMode32,&recList32,800,600);
        if (recommendedMode16->BitsPerPixel != 0)
            *recommendedMode = *recommendedMode16;
        else
            *recommendedMode = *recommendedMode8;
        }
    else {
        // Choose a mode from the list of saved modes.
        SetRecommended(recommendedMode8,&recList8,idealXRes,idealYRes);
        if (!SetRecommended(recommendedMode16,&recList16,idealXRes,idealYRes))
            SetRecommended(recommendedMode16,&recList15,idealXRes,idealYRes);
        SetRecommended(recommendedMode24,&recList24,idealXRes,idealYRes);
        SetRecommended(recommendedMode32,&recList32,idealXRes,idealYRes);
        if (recommendedMode16->BitsPerPixel != 0)
            *recommendedMode = *recommendedMode16;
        else
            *recommendedMode = *recommendedMode8;
        if (recommendedMode24->XResolution >= recommendedMode->XResolution
            && recommendedMode24->YResolution >= recommendedMode->YResolution)
            *recommendedMode = *recommendedMode24;
        if (recommendedMode32->XResolution >= recommendedMode->XResolution
            && recommendedMode32->YResolution >= recommendedMode->YResolution)
            *recommendedMode = *recommendedMode32;
        }

    // Kill all the mode lists. Since all the memory is managed by the block
    // list, we simply kill the block list and all the lists are killed
    // at once.
    N_freeBlockList(&bl);
    return true;
}

/****************************************************************************
PARAMETERS:
dc                  - Device context to get recommended modes for
recommendedMode     - Overall recommended display mode
recommendedMode8    - 8bpp recommended display mode
recommendedMode16   - 16bpp recommended display mode
recommendedMode24   - 24bpp recommended display mode
recommendedMode32   - 32bpp recommended display mode

RETURNS:
True on success, false on error.

REMARKS:
Legacy function that only works on the primary head and the currently
active output device.
{secret}
****************************************************************************/
ibool NAPI GA_getRecommendedModesV1(
    GA_devCtx *dc,
    GA_recMode *recommendedMode,
    GA_recMode *recommendedMode8,
    GA_recMode *recommendedMode16,
    GA_recMode *recommendedMode24,
    GA_recMode *recommendedMode32)
{
    GA_initFuncs    init;

    init.dwSize = sizeof(init);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&init))
        return false;
    return GA_getRecommendedModes(dc,init.GetDisplayOutput(),gaActiveHeadPrimary,recommendedMode,
        recommendedMode8,recommendedMode16,recommendedMode24,recommendedMode32);
}

