/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Offscreen buffer management code.
*
****************************************************************************/

#include "internal.h"

// TODO: Fix this for 4bpp shadow buffer!

// TODO: Implement and track LRU scheme for pageable buffers!

// TODO: Implement support for AGP buffers! We can prototype this on the
//       Intel i810/815 machines, since all buffers are AGP buffers on this
//       machine so the DMA functions are a snap to write and test.

/*--------------------------- Global variables ----------------------------*/

BM_state    BM_devCtx;

// Call region functions by pointer reference when compiled as a BPD

#ifdef __DRIVER__
#define RGN_NewRegion()                             BM.rgnFuncs.NewRegion()
#define RGN_NewRectRegion(left,top,right,bottom)    BM.rgnFuncs.NewRectRegion(left,top,right,bottom)
#define RGN_CopyRegion(s)                           BM.rgnFuncs.CopyRegion(s)
#define RGN_CopyIntoRegion(d,s)                     BM.rgnFuncs.CopyIntoRegion(d,s)
#define RGN_ClearRegion(r)                          BM.rgnFuncs.ClearRegion(r)
#define RGN_FreeRegion(r)                           BM.rgnFuncs.FreeRegion(r)
#define RGN_SectRegion(r1,r2)                       BM.rgnFuncs.SectRegion(r1,r2)
#define RGN_SectRegionRect(r1,r)                    BM.rgnFuncs.SectRegionRect(r1,r)
#define RGN_UnionRegion(r1,r2)                      BM.rgnFuncs.UnionRegion(r1,r2)
#define RGN_UnionRegionRect(r1,r)                   BM.rgnFuncs.UnionRegionRect(r1,r)
#define RGN_UnionRegionOfs(r1,r2,dx,dy)             BM.rgnFuncs.UnionRegionOfs(r1,r2,dx,dy)
#define RGN_DiffRegion(r1,r2)                       BM.rgnFuncs.DiffRegion(r1,r2)
#define RGN_DiffRegionRect(r1,r)                    BM.rgnFuncs.DiffRegionRect(r1,r)
#define RGN_OptimizeRegion(r)                       BM.rgnFuncs.OptimizeRegion(r)
#define RGN_OffsetRegion(r,dx,dy)                   BM.rgnFuncs.OffsetRegion(r,dx,dy)
#define RGN_IsEmptyRegion(r)                        BM.rgnFuncs.IsEmptyRegion(r)
#define RGN_IsEqualRegion(r1,r2)                    BM.rgnFuncs.IsEqualRegion(r1,r2)
#define RGN_PtInRegion(r,x,y)                       BM.rgnFuncs.PtInRegion(r,x,y)
#define RGN_TraverseRegion(rgn,doRect,parms)        BM.rgnFuncs.TraverseRegion(rgn,doRect,parms)
#endif

/*------------------------- Implementation --------------------------------*/

/****************************************************************************
REMARKS:
Round a value to the specified integer boundary
****************************************************************************/
static ulong RoundUp(
    ulong value,
    ulong boundary)
{
    return ((value + (boundary-1)) / boundary) * boundary;
}

/****************************************************************************
REMARKS:
Round a value to the specified integer boundary
****************************************************************************/
static ulong RoundDown(
    ulong value,
    ulong boundary)
{
    return (value / boundary) * boundary;
}

/****************************************************************************
PARAMETERS:
height  - Height of the buffer to align
stride  - Stride of the buffer to align (modified)
offset  - Starting offset of the buffer to align (modified)
size    - Place to return the size of the resulting buffer
growUp  - True if the buffer is allocated on a heap that grows up in memory

RETURNS:
True if the buffer was successfully aligned in video memory, false if not.

REMARKS:
This function is used to align a buffer in offscreen video memory as
necessary for the hardware. This version of the function is used if the
hardware driver does not provide it's own version, and uses the alignment
values stored in the mode information block to correctly handle this.

If the growUp flag is true, the buffer is aligned such that the memory
grows up in memory above the initial value passed in the Offset parameter.
If the growUp flag is false, the buffer is aligned such that the memory
grows down from the initial value passed in the Offset parameter. The
value returned in the Offset parameter is always the value at the start
of the buffer in memory (ie: lowest memory address).
****************************************************************************/
static ibool NAPI BM_AlignLinearBuffer(
    N_int32 height,
    N_int32 *stride,
    N_int32 *offset,
    N_int32 *size,
    N_int32 growUp)
{
    *stride = RoundUp(*stride,BM.gaMode.BitmapStridePad);
    *size = (*stride * height);
    if (growUp)
        *offset = RoundUp(*offset,BM.gaMode.BitmapStartAlign);
    else
        *offset = RoundDown(*offset - *size, BM.gaMode.BitmapStartAlign);
    return (*stride <= BM.gaMode.MaxBytesPerScanLine &&
        *offset >= 0 &&
        (*offset + *size) <= BM.gaMode.MaxLinearOffset);
}

/****************************************************************************
REMARKS:
Create the buffer management block lists and initialises all internal
data structures.
****************************************************************************/
ibool BM_init(
    GA_devCtx *hwCtx,
    REF2D_driver *ref2d,
    bufReadScan_t bufReadScan,
    bufBltScan_t bufBltScan,
    GA_bufferFuncs *funcs)
{
    static GA_bufferFuncs exp = {
        sizeof(GA_bufferFuncs),
        BM_InitBuffers,
        BM_GetPrimaryBuffer,
        BM_GetFlippableBuffer,
        BM_AllocBuffer,
        BM_FreeBuffer,
        BM_LockBuffer,
        BM_UnlockBuffer,
        BM_SetActiveBuffer,
        BM_FlipToBuffer,
        BM_FlipToStereoBuffer,
        BM_GetFlipStatus,
        BM_WaitTillFlipped,
        BM_UpdateCache,
        BM_UpdateFromCache,
        BM_BitBltBuf,
        BM_BitBltPattBuf,
        BM_BitBltColorPattBuf,
        BM_SrcTransBltBuf,
        BM_DstTransBltBuf,
        BM_StretchBltBuf,
        BM_BitBltPlaneMaskedBuf,
        BM_BitBltFxBuf,
        BM_DrawRectBuf,
        NULL,               // BM_SetRef2d
        NULL,               // BM_GetClipper
        NULL,               // BM_SetClipper
        NULL,               // BM_AllocDepthBuffer
        NULL,               // BM_AllocStencilBuffer
        NULL,               // BM_AllocTexture
        NULL,               // BM_AllocVideoTexture
        NULL,               // BM_SetTexturePalette
        NULL,               // BM_GetTexturePalette
        NULL,               // BM_SetTexturePaletteExt
        NULL,               // BM_GetTexturePaletteExt
        BM_GetActiveBuffer,
        BM_GetVisibleBuffer,
        BM_GetVisibleRightBuffer,
        BM_AllocBufferExt,
        };
    int size,pixelsPerLine;

    // Initialise device driver internals
    memset(&BM,0,sizeof(BM));
    BM.ref2d = ref2d;
    BM.bufReadScan = bufReadScan;
    BM.bufBltScan = bufBltScan;
    BM.rgnFuncs.dwSize = sizeof(BM.rgnFuncs);
    if (!hwCtx || !ref2d->QueryFunctions(GA_GET_REGIONFUNCS,&BM.rgnFuncs))
        return false;
    BM.hwCtx = hwCtx;
    BM.hwInit.dwSize = sizeof(BM.hwInit);
    hwCtx->loader.QueryFunctions(GA_GET_INITFUNCS,-2,&BM.hwInit);
    BM.hwState.dwSize = sizeof(BM.hwState);
    hwCtx->loader.QueryFunctions(GA_GET_2DSTATEFUNCS,-2,&BM.hwState);
    BM.hwDraw.dwSize = sizeof(BM.hwDraw);
    hwCtx->loader.QueryFunctions(GA_GET_2DRENDERFUNCS,-2,&BM.hwDraw);

    // We query the hardware driver funcs via ref2d as ref2d overrides
    // some of these for software cursor management.
    BM.hwDrv.dwSize = sizeof(BM.hwDrv);
    ref2d->QueryFunctions(GA_GET_DRIVERFUNCS,&BM.hwDrv);

    // Query the current SNAP mode information
    BM.gaModeNum =  BM.hwInit.GetVideoMode();
    BM.gaMode.dwSize = sizeof(BM.gaMode);
    BM.hwInit.GetCurrentVideoModeInfo(&BM.gaMode);

    // Initialise copies of the software rendering functions
    // from the 2D reference rasteriser. These functions are
    // obtained *before* the ref2d driver has overridden any
    // internal functions with buffer manager versions. Hence
    // this avoids any recursive calls in the buffer manager.
    BM.state2d.dwSize = sizeof(BM.state2d);
    BM.ref2d->QueryFunctions(GA_GET_2DSTATEFUNCS,&BM.state2d);
    BM.draw2d.dwSize = sizeof(BM.draw2d);
    BM.ref2d->QueryFunctions(GA_GET_2DRENDERFUNCS,&BM.draw2d);

    // Setup the pointer to the correct buffer alignment function
    // depending on whether the hardware driver provides its own
    // version of this.
    if (BM.hwInit.AlignLinearBuffer)
        BM.AlignLinearBuffer = BM.hwInit.AlignLinearBuffer;
    else
        BM.AlignLinearBuffer = BM_AlignLinearBuffer;

    // Create all the block lists
    if (!N_createBlockList(&BM.bufferList.bl,sizeof(GA_buf) + sizeof(N_node)))
        return false;
    BM.bufferList.freeList = N_buildFreeList(&BM.bufferList.bl);

    // Now initialise the heaps. If we have linear blits in the hardware,
    // then we setup for a fixed heap to grow from the start of video
    // memory up to the top of video memory
    N_initList(&BM.bufferHeap);
    BM.priorityCount = 0;
    BM.surfLock = 0;

    // Allocate space to store the list of flippable buffers
    size = sizeof(*BM.flipBuffers) * BM.gaMode.MaxBuffers;
    if ((BM.flipBuffers = malloc(size)) == NULL)
        return false;
    memset(BM.flipBuffers,0,size);

    // Setup the buffer heap information
    if (BM.hwDraw.BitBltLin) {
        // Set up the linear heap as empty. The fixed heap grows from the
        // top down, and the moveable heap grows from the bottom up. Flip
        // buffers intially come from the moveable heap, but are then locked
        // down so they are always at the start of video memory.
        BM.linearMoveableBottom = 0;
        BM.linearMoveableCur    = 0;
        BM.linearFixedCur       = BM.gaMode.MaxLinearOffset;
        BM.compacted            = true;
        BM.fixedHeapFragmented  = false;
        BM.pagedOut             = false;
        BM.freeSpace            = NULL;
        }
    else {
        // Set up the rectangular heap as empty. Objects on the rectangular
        // heap are never moveable (but may be pageable).
        if (BM.gaMode.BitsPerPixel == 4)
            pixelsPerLine = BM.gaMode.BytesPerScanLine * 8;
        else
            pixelsPerLine = PIXELSPERLINE(BM.gaMode.BytesPerScanLine,BM.gaMode.BitsPerPixel);
        BM.freeSpace = RGN_NewRectRegion(0,0,pixelsPerLine,BM.gaMode.MaxScanLines);
        if (!BM.freeSpace)
            return false;
        }

    // Return the exported function list
    COPY_STRUCTURE(funcs,&exp);
    return true;
}

/****************************************************************************
REMARKS:
Cleans up and exits, freely all memory used by the buffer manager.
****************************************************************************/
void BM_exit(void)
{
    GA_buf  *buf;

    // Disable stereo mode if it is currently active
    if (BM.isStereo != gaNoStereo) {
        if (BM.isStereo == gaHWStereo)
            BM.hwDrv.EnableStereoMode(false);
        else
            BM.stereo.SoftStereoExit();
        BM.isStereo = gaNoStereo;
        }

    // Free memory for all system memory buffer caches
    for (buf = N_firstNode(&BM.bufferHeap); buf; buf = N_nextNode(buf)) {
        // Free up all memory occupied by the buffer
        if (buf->SurfaceCache)
            free(buf->SurfaceCache);
        }

    // Now block free all the buffers in our block list
    N_freeBlockList(&BM.bufferList.bl);
    free(BM.flipBuffers);
}

#ifdef  __DRIVER__
/****************************************************************************
REMARKS:
Library initialisation entry point.
****************************************************************************/
ibool _CEXPORT BM_initLibrary(
    GA_devCtx *hwCtx,
    REF2D_driver *ref2d,
    bufReadScan_t bufReadScan,
    bufBltScan_t bufBltScan,
    GA_bufferFuncs *funcs)
{
    return BM_init(hwCtx,ref2d,bufReadScan,bufBltScan,funcs);
}

/****************************************************************************
REMARKS:
Library exit entry point.
****************************************************************************/
void _CEXPORT BM_exitLibrary(void)
{
    BM_exit();
}
#endif

/****************************************************************************
REMARKS:
Resizes the buffer list if full by increasing its size by the original
default size and building a new free list in the resized portion.
****************************************************************************/
static ibool BM_resizeBufferList(void)
{
    if (!N_resizeBlockList(&BM.bufferList.bl))
        return false;
    BM.bufferList.freeList = N_buildFreeList(&BM.bufferList.bl);
    return true;
}

/****************************************************************************
RETURNS:
Pointer to new buffer structure

REMARKS:
Allocates space for a new buffer in the buffer pool from the free list. If
there are no more buffers, we attempt to re-size the buffer pool which will
bomb out if we run out of memory.
****************************************************************************/
GA_buf * BM_newBufferOnHeap(void)
{
    N_node  *n;

    if (BM.bufferList.freeList == NULL) {
        if (!BM_resizeBufferList())
            return NULL;
        }
    n = BM.bufferList.freeList;
    BM.bufferList.freeList = N_FREELIST_NEXT(n);
    return (GA_buf*)LST_USERSPACE(n);
}

/****************************************************************************
PARAMETERS:
buffer  - Pointer to the buffer to free

REMARKS:
Frees the buffer by putting it onto the buffer free list.
****************************************************************************/
void BM_freeBufferOnHeap(
    GA_buf *buffer)
{
    N_node *n = LST_HEADER(buffer);
    N_FREELIST_NEXT(n) = BM.bufferList.freeList;
    BM.bufferList.freeList = n;
}

/****************************************************************************
PARAMETERS:
r   - Rectangle to test

REMARKS:
Tries to expand the BM.testRect rectangle vertically with all rectangles
vertically adjacent to it.
****************************************************************************/
static void NAPI expandVert(
    const GA_rect *r,
    void *parms)
{
    (void)parms;
    if (r->top == BM.testRect.bottom && r->left <= BM.testRect.left
            && r->right >= BM.testRect.left + BM.bufWidth) {
        BM.testRect.bottom = r->bottom;
        }
}

/****************************************************************************
PARAMETERS:
r   - Rectangle to test

REMARKS:
Checks the test rectangle against the size of our bitmap, and if it is
larger, then we have found some free space. We also check t find the first
rectangle with the best fit for the bitmap. Note that the BM.bestBounds
returned is the bounds of the free space that we found, and will most likely
be larger than the size of our bitmap!
****************************************************************************/
static void NAPI findFreeRect(
    const GA_rect *r,
    void *parms)
{
    (void)parms;
    if ((r->right - r->left) >= BM.bufWidth) {
        BM.testRect = *r;
        if ((BM.testRect.bottom-BM.testRect.top) < BM.bufHeight) {
            // We have found a rectangle big enough in the width to hold
            // our buffer, so we need to search for and coallesce all
            // rectangles below this rectangle that are also wide enough
            // to hold our buffer.
            RGN_TraverseRegion(BM.rgn,expandVert,NULL);
            }
        if ((BM.testRect.bottom-BM.testRect.top) >= BM.bufHeight) {
            // We have found a rectangle big enough to hold our bitmap
            if (BM.foundSpace) {
                if ((BM.testRect.right-BM.testRect.left) < (BM.bestBounds.right-BM.bestBounds.left)
                        && (BM.testRect.bottom-BM.testRect.top) < (BM.bestBounds.bottom-BM.bestBounds.top)) {
                    BM.bestBounds = BM.testRect;
                    }
                }
            else {
                BM.bestBounds = BM.testRect;
                BM.foundSpace = true;
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
buf     - Offscreen buffer to create

RETURNS:
True if the buffer was allocated, false if not.

REMARKS:
Internal function to allocate a linear memory bitmap.
****************************************************************************/
static ibool AllocateLinearMem(
    GA_buf *buf)
{
    // Allocate buffer in offscreen memory on either fixed or moveable heap
    if (buf->Flags & gaBufferMoveable) {
        // Allocate bottom up
        buf->Offset = BM.linearMoveableCur;
        if (!BM.AlignLinearBuffer(buf->Height,&buf->Stride,&buf->Offset,&buf->Size,true))
            return false;
        if ((buf->Offset + buf->Size) > BM.linearFixedCur)
            return false;
        buf->Size = RoundUp(buf->Offset + buf->Size,BM.gaMode.BitmapStartAlign) - buf->Offset;
        BM.linearMoveableCur = buf->Offset + buf->Size;
        }
    else {
        // Allocate top down
        buf->Offset = BM.linearFixedCur;
        if (!BM.AlignLinearBuffer(buf->Height,&buf->Stride,&buf->Offset,&buf->Size,false))
            return false;
        if (buf->Offset < BM.linearMoveableCur)
            return false;
        buf->Size = BM.linearFixedCur - buf->Offset;
        BM.linearFixedCur = buf->Offset;
        }
    return true;
}

/****************************************************************************
REMARKS:
****************************************************************************/
void BM_pageOutBuffers(void)
{
    // TODO: Add code to page out all buffers that are pageable.
    // When buffers are paged out, they get the gaBufferSysMem flag set
    // Note that we should change this to stop as soon as we have enough
    // memory, and also process buffes in LRU format somehow so we can
    // page out the least recently used buffers first.
    BM.pagedOut = true;
}

/****************************************************************************
REMARKS:
****************************************************************************/
static void BM_pageInBuffers(void)
{
    // TODO: Add code to page in all pageable buffers here, in priority
    //       order based on their LRU counts
    BM.pagedOut = false;
}

/****************************************************************************
REMARKS:
Function to try compacting the moveable heap to make room for new bitmaps.
Try compacting the heap. Note that it is important that we process the
bitmaps in FIFO order, so that we will always be moving the bitmaps *UP*
in memory during compaction on the moveable heap (ie: the moveable heap
grows down from the top of video memory). Hence all bitmaps on our list are
added to the head of the list, so we can efficiently traverse it.
****************************************************************************/
static ibool BM_compactHeap(void)
{
    GA_buffer   *hwDrawBuf = BM.ref2d->GetDrawBuffer();
    GA_buf      *buf;
    N_int32     newOffset;
    GA_buffer   drawBuf = {sizeof(GA_buffer)};

    // We can't compact the heap if we don't have acceleration
    if (!(BM.gaMode.Attributes & gaHaveAccel2D))
        return false;

    // Empty the linear heap to start refilling it from the bottom up
    BM.linearMoveableCur = BM.linearMoveableBottom;

    // Now process all buffers in FIFO order
    for (buf = N_lastNode(&BM.bufferHeap); buf; buf = N_prevNode(buf)) {
        // Ignore non moveable or system memory bitmaps
        if ((buf->Flags & gaBufferSysMem) || !(buf->Flags & gaBufferMoveable))
            continue;

        // Find new location for buffer
        newOffset = BM.linearMoveableCur;
        if (!BM.AlignLinearBuffer(buf->Height,&buf->Stride,&newOffset,&buf->Size,true))
            return false;
        buf->Size = RoundUp(buf->Offset + buf->Size,BM.gaMode.BitmapStartAlign) - buf->Offset;

        // If the bitmap has moved, blit it to the new location
        if (buf->Offset != newOffset) {
            drawBuf.Offset = newOffset;
            drawBuf.Stride = buf->Stride;
            drawBuf.Width = buf->Width;
            drawBuf.Height = buf->Height;
            BM.hwState.SetDrawBuffer(&drawBuf);
            BM.hwDraw.BitBltLin(buf->Offset,buf->Stride,0,0,buf->Width,buf->Height,0,0,GA_REPLACE_MIX);
            }

        // Now relocate the bitmap to the new location and adjust the cursor
        buf->Offset = newOffset;
        BM.linearMoveableCur = buf->Offset + buf->Size;
        }
    BM.compacted = true;

    // Restore the current hardware draw buffer
    BM.hwState.SetDrawBuffer(hwDrawBuf);
    return true;
}

/****************************************************************************
PARAMETERS:
buf     - Offscreen buffer to create

RETURNS:
True if the buffer was allocated, false if not.

REMARKS:
Creates a buffer in offscreen display memory. We support both rectangular
and linear offscreen buffers.
****************************************************************************/
static ibool BM_createOffscreenBuffer(
    GA_buf *buf)
{
    GA_rect         r;

    // No offscreen memory unless we have acceleration, except for flip buffers
    if (!(BM.gaMode.Attributes & gaHaveAccel2D) && !(buf->Flags & gaBufferFlippable))
        return false;
    if (BM.freeSpace) {
        // Allocate the buffer from the rectangular heap. To do this we
        // traverse the region of empty space (always traverses left to
        // right, top to bottom) looking for the best fit rectangle for our
        // buffer. If we find some free space, allocate the space and remove
        // it from the freespace region.
        BM.bufWidth = buf->Width;
        BM.bufHeight = buf->Height;
        BM.foundSpace = false;
        RGN_TraverseRegion(BM.rgn = BM.freeSpace,findFreeRect,NULL);
        if (!BM.foundSpace)
            return false;
        buf->StartX = BM.bestBounds.left;
        buf->StartY = BM.bestBounds.top;
        buf->Stride = BM.gaMode.BytesPerScanLine;
        buf->Offset = (buf->StartY * BM.gaMode.BytesPerScanLine) + BYTESPERLINE(buf->StartX,BM.gaMode.BitsPerPixel);
        r.left = buf->StartX;
        r.right = r.left + buf->Width;
        r.top = buf->StartY;
        r.bottom = r.top + buf->Height;
        RGN_DiffRegionRect(BM.freeSpace,&r);
        RGN_OptimizeRegion(BM.freeSpace);
        return true;
        }
    else {
        // Allocate the buffer in linear memory. If we run out of memory
        // on the moveable heap, we try compacting the heap on the fly to
        // make room for any new bitmaps, or return failure if it is
        // already compacted.
        if (AllocateLinearMem(buf))
            return true;
        if (BM.compacted)
            return false;
        if (!BM_compactHeap())
            return false;
        if (AllocateLinearMem(buf))
            return true;

        // If there is still no room after heap compaction, try to
        // page all pageable buffers to system memory to free up space. Note
        // that we only do this if we are allocating a priority, moveable
        // buffer
        if (!BM.pagedOut && (buf->Flags & gaBufferPriority) && (buf->Flags & gaBufferMoveable)) {
            BM_pageOutBuffers();
            if (!BM_compactHeap())
                return false;
            return AllocateLinearMem(buf);
            }
        return false;
        }
}

/****************************************************************************
PARAMETERS:
width           - Width of the buffer in pixels
height          - Height of the buffer in scanlines
flags           - Flags to create the buffer with
bitsPerPixel    - Color depth for the new buffer
pf              - Pixel format for the new buffer
format          - Internal format flags for the new buffer

RETURNS:
Pointer to the allocated buffer, NULL on failure.

REMARKS:
Internal function to do the buffer creation from the buffer heap.
****************************************************************************/
static GA_buf * BM_createBuffer(
    N_int32 width,
    N_int32 height,
    N_int32 flags,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf,
    N_int32 format,
    ibool internalBuffer)
{
    int     sysMemSize;
    GA_buf  *buf = BM_newBufferOnHeap();

    // Check for flags that are dependant on each other
    if (!buf)
        return NULL;
    if (flags & gaBufferPageable)
        flags |= gaBufferCached;
    if (flags & gaBufferCached)
        flags &= ~gaBufferSysMem;
    if (flags & gaBufferSysMem)
        flags &= ~gaBufferNoSysMem;
    if (!internalBuffer && (flags & gaBufferSpecial) && (flags & (gaBufferSysMem | gaBufferCached | gaBufferMoveable | gaBufferPageable)))
        return NULL;

    // We don't yet support video memory buffers with a different pixel format,
    // so force to system memory if we detect this condition (only for non-special
    // buffers, since depth buffers and textures buffers can be different).
    if (!(flags & gaBufferSpecial) && (bitsPerPixel != BM.gaMode.BitsPerPixel
            || (bitsPerPixel >= 15 && !CMP_PF(*pf,BM.gaMode.PixelFormat)))) {
        if (flags & (gaBufferNoSysMem | gaBufferFlippable))
            return NULL;
        flags |= gaBufferSysMem;
        }

    // Now fill in the information for the buffer and create it
    memset(buf,0,sizeof(*buf));
    buf->dwSize = sizeof(*buf);
    if (flags & gaBufferFlippable) {
        // Flippable buffers must always match the physical screen resolution
        buf->Width = BM.gaMode.XResolution;
        buf->Height = BM.gaMode.YResolution;
        buf->Stride = BM.gaMode.BytesPerScanLine;
        buf->CacheStride = 0;
        }
    else {
        buf->Width = width;
        buf->Height = height;
        buf->Stride = buf->CacheStride = BYTESPERLINE(width,bitsPerPixel);
        }
    buf->Flags = flags;
    buf->Format = format;
    buf->BitsPerPixel = bitsPerPixel;
    buf->StartX = -1;
    buf->StartY = -1;
    if (pf)
        buf->PixelFormat = *pf;
    sysMemSize = height * buf->CacheStride;
    if ((flags & gaBufferSysMem) || !BM_createOffscreenBuffer(buf)) {
        // Return failure if not in hardware and we don't want system memory
        if (!(flags & gaBufferSysMem) || (flags & gaBufferNoSysMem))
            goto Error;

        // Buffer creation failed in hardware, so allocate it in system
        // memory.
        if ((buf->Surface = malloc(sysMemSize)) == NULL)
            goto Error;
        memset(buf->Surface,0,sysMemSize);
        buf->Stride = buf->CacheStride;
        buf->SurfaceCache = buf->Surface;

        // Indicate if this is a system memory only buffer
        if (!(flags & gaBufferCached))
            buf->Flags |= gaBufferSysMem;
        }

    // If the buffer is not in system memory, create the surface cache for
    // the buffer in system memory if caching has been requested.
    if ((flags & gaBufferCached) && !buf->SurfaceCache) {
        if ((buf->SurfaceCache = malloc(sysMemSize)) == NULL)
            goto Error;
        }

    // Add the new buffer to the head of the buffer list
    N_addToHead(&BM.bufferHeap,buf);

    // Keep track of how many priority buffers are currently allocated are
    // on the moveable heap. When this count gets to 0, we automatically
    // try and page back in all pageable bitmaps again back on their LRU
    // counts. Hence we don't do this until graphics intensive apps exit.
    if ((flags & gaBufferPriority) && (flags & gaBufferMoveable))
        BM.priorityCount++;
    return buf;

Error:
    BM_freeBufferOnHeap(buf);
    return NULL;
}

/****************************************************************************
PARAMETERS:
buf     - Offscreen buffer to destroy

REMARKS:
Destroys an offscreen buffer, freeing up the memory used by the buffer.
If compactHeap is set to true, we compact the heap to free up any
free memory blocks.
****************************************************************************/
static void BM_destroyOffscreenBuffer(
    GA_buf *buf)
{
    GA_rect r;
    GA_buf  *b;

    if (BM.freeSpace) {
        // Free the region from the rectangular heap
        r.left = buf->StartX;
        r.right = buf->StartX + buf->Width;
        r.top = buf->StartY;
        r.bottom = buf->StartY + buf->Height;
        RGN_UnionRegionRect(BM.freeSpace,&r);
        RGN_OptimizeRegion(BM.freeSpace);
        }
    else {
        if (buf->Flags & gaBufferMoveable) {
            // Free it from the moveable heap, which grows up from the bottom
            // of video memory.
            if ((buf->Offset + buf->Size) == BM.linearMoveableCur) {
                // Bitmap is the last one on the heap, so simply decrement
                // the linear cursor.
                BM.linearMoveableCur = buf->Offset;

                // If this was the last bitmap on the heap and the heap
                // has not been compacted, compact it now as this may well
                // be the last bitmap so compacting the heap will free up
                // the entire heap for the next allocation.
                if (!BM.compacted)
                    BM_compactHeap();
                }
            else {
                // Bitmap is somewhere in the middle of the heap, so we need
                // to do a heap compaction at some later date.
                BM.compacted = false;
                }
            }
        else {
            // Free it from the fixed heap, which grows down from the top
            // of video memory.
            if (buf->Offset == BM.linearFixedCur) {
                // Bitmap is the last one on the heap, so simply increment
                // the linear cursor to the next buffer
                BM.linearFixedCur = buf->Offset + buf->Size;
                if (BM.fixedHeapFragmented) {
                    // If bitmaps have been freed from the middle of the
                    // heap, we need to process the heap and determine the
                    // new bottom of the heap from the lowest objects on the
                    // heap.
                    BM.linearFixedCur = BM.gaMode.MaxLinearOffset;
                    for (b = N_firstNode(&BM.bufferHeap); b; b = N_nextNode(b)) {
                        // Ignore moveable or system memory bitmaps
                        if (b == buf || (b->Flags & gaBufferSysMem) || (b->Flags & gaBufferMoveable))
                            continue;

                        // Find new lowest memory location on heap
                        if (buf->Offset < BM.linearFixedCur)
                            BM.linearFixedCur = buf->Offset;
                        }
                    BM.fixedHeapFragmented = false;
                    }
                }
            else {
                // Bitmap is somewhere in the middle of the heap, so we need
                // to optimise the heap at a later date.
                BM.fixedHeapFragmented = true;
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to destroy

REMARKS:
This function destroys a SNAP buffer, and frees all resources associated
with the buffer.
****************************************************************************/
static ibool NAPI BM_destroyBuffer(
    GA_buf *buf)
{
    // Remove the buffer from the list of allocated buffers and destroy it
    if (!buf)
        return false;
    N_remove(&BM.bufferHeap,buf);
    if (!(buf->Flags & gaBufferSysMem))
        BM_destroyOffscreenBuffer(buf);

    // Free up all memory occupied by the buffer if cached in system memory
    if (buf->SurfaceCache)
        free(buf->SurfaceCache);

    // Page in all pageable buffers if this is the last priority buffer
    if (buf->Flags & gaBufferPriority) {
        if (--BM.priorityCount == 0)
            BM_pageInBuffers();
        }
    BM_freeBufferOnHeap(buf);
    return true;
}

/****************************************************************************
PARAMETERS:
numBuffers  - Number of flippable buffers to allocate
flags       - Extra flags to pass when allocating the buffers
stereo      - Software stereo functions, or NULL for non-stereo mode

RETURNS:
True on success, false on failure.

REMARKS:
This function initialises the buffer manager, and allocates the primary
buffer and all related flippable buffers (for a total of numBuffers buffers).
If any of the buffers cannot be allocated, this function will return false.
If the stereo parameter is not NULL, stereo mode is enabled using the
passed in software stereo functions.
****************************************************************************/
ibool NAPI BM_InitBuffers(
    N_int32 numBuffers,
    N_uint32 flags,
    GA_softStereoFuncs *stereo)
{
    GA_buffer   drawBuf;
    int         i;
    GA_buf      *primaryBuf;

    // Check that the number of buffers is within the maximum supported
    if (numBuffers > BM.gaMode.MaxBuffers)
        return false;

    // Now allocate the primary buffer and all subsequent back flipping
    // buffers as requested by the application. We allocate them from the
    // moveable heap first, so they will be at the bottom of video memory.
    // But we remove the moveable flag when they are allocated so they
    // will in fact be fixed in video memory. We then compute the bottom
    // of the moveable heap as the end of the last flip buffer.
    for (i = 0; i < numBuffers; i++) {
        BM.flipBuffers[i] = BM_createBuffer(BM.gaMode.XResolution,BM.gaMode.YResolution,flags | gaBufferFlippable | gaBufferMoveable,BM.gaMode.BitsPerPixel,&BM.gaMode.PixelFormat,0,true);
        if (!BM.flipBuffers[i]) {
            // Clean up on error and exit
            while (BM.numFlipBuffers)
                BM_destroyBuffer(BM.flipBuffers[--BM.numFlipBuffers]);
            return false;
            }
        BM.flipBuffers[i]->Flags &= ~gaBufferMoveable;
        BM.numFlipBuffers++;
        }
    BM.linearMoveableBottom = BM.linearMoveableCur;

    // Now make the primary buffer the active drawing buffer
    primaryBuf = BM.flipBuffers[0];
    BM.activeY = primaryBuf->StartY;
    BM.activeX = primaryBuf->StartX;
    drawBuf.dwSize = sizeof(drawBuf);
    drawBuf.Offset = primaryBuf->Offset;
    drawBuf.Stride = primaryBuf->Stride;
    drawBuf.Width  = primaryBuf->Width;
    drawBuf.Height = primaryBuf->Height;
    if (BM.ref2d->SetDrawBuffer(&drawBuf,BM.hwCtx->LinearMem,BM.gaMode.BitsPerPixel,&BM.gaMode.PixelFormat,BM.hwCtx,false) != 0)
        return false;

    // Enable stereo mode if requested
    if (stereo) {
        BM.stereo.dwSize = sizeof(BM.stereo);
        COPY_STRUCTURE(&BM.stereo,stereo);
        if (!(BM.gaMode.Attributes & gaHaveStereo)) {
            // Enable software stereo page flipping using a timer interrupt
            if (!BM.stereo.SoftStereoInit(BM.hwCtx))
                return false;
            BM.isStereo = gaSoftStereo;
            }
        else {
            // Enable support for hardware stereo mode
            BM.isStereo = gaHWStereo;
            }

        // Start the stereo mode
        if (BM.isStereo == gaHWStereo)
            BM.hwDrv.EnableStereoMode(true);
        else
            BM.stereo.SoftStereoOn();
        }
    return true;
}

/****************************************************************************
PARAMETERS:
index   - Index of the flippable buffer to obtain the address of

RETURNS:
Pointer to specified flippable buffer, or NULL if not such buffer.

REMARKS:
This function returns a pointer to the flippable buffer with the specified
index in the flippable buffer table. These buffer pointers can then be used
with the SetActiveBuffer and FlipToBuffer functions to implement
hardware page flipping.
****************************************************************************/
GA_buf * NAPI BM_GetFlippableBuffer(
    N_int32 index)
{
    if (index >= BM.numFlipBuffers)
        return NULL;
    return BM.flipBuffers[index];
}

/****************************************************************************
RETURNS:
Pointer to the primary display buffer, NULL if no primary buffer.

REMARKS:
This function returns a pointer to the primary display buffer, which is
always the first buffer on the fixed heap for regular display output.
If the buffer manager has not yet been initialised, this function will
return NULL.
****************************************************************************/
GA_buf * NAPI BM_GetPrimaryBuffer(void)
{
    if (BM.flipBuffers)
        return BM.flipBuffers[0];
    return NULL;
}

/****************************************************************************
PARAMETERS:
width   - Width of the buffer in pixels
height  - Height of the buffer in scanlines
flags   - Flags to create the buffer with

RETURNS:
Pointer to the allocated buffer, NULL on failure.

REMARKS:
This function allocates a new buffer and returns it. This only allocates
video memory buffers for drawing and bitmap storage and caching. It is
not used to allocate depth buffers or texture buffers.
****************************************************************************/
GA_buf * NAPI BM_AllocBuffer(
    N_int32 width,
    N_int32 height,
    N_int32 flags)
{
    // This function does not support allocating special buffers!
    if (flags & gaBufferSpecial)
        return NULL;
    return BM_createBuffer(width,height,flags,BM.gaMode.BitsPerPixel,&BM.gaMode.PixelFormat,0,false);
}

/****************************************************************************
PARAMETERS:
width   - Width of the buffer in pixels
height  - Height of the buffer in scanlines
flags   - Flags to create the buffer with

RETURNS:
Pointer to the allocated buffer, NULL on failure.

REMARKS:
This function allocates a new buffer and returns it. This only allocates
video memory buffers for drawing and bitmap storage and caching. It is
not used to allocate depth buffers or texture buffers.
****************************************************************************/
GA_buf * NAPI BM_AllocBufferExt(
    N_int32 width,
    N_int32 height,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pixelFormat,
    N_int32 flags)
{
    // This function does not support allocating special buffers!
    if (flags & gaBufferSpecial)
        return NULL;
    return BM_createBuffer(width,height,flags,bitsPerPixel,pixelFormat,0,false);
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to destroy

REMARKS:
This function destroys a SNAP buffer, and frees all resources associated
with the buffer.
****************************************************************************/
ibool NAPI BM_FreeBuffer(
    GA_buf *buf)
{
    int i;

    // App is never allowed to free the primary flippable buffers!
    for (i = 0; i < BM.numFlipBuffers; i++) {
        if (BM.flipBuffers[i] == buf)
            return false;
        }
    return BM_destroyBuffer(buf);
}

/****************************************************************************
PARAMETERS:
buf - SNAP buffer to lock

RETURNS:
Physical address of locked buffer in memory, 0 if a system memory buffer.

REMARKS:
This function locks a buffer so that an application can begin drawing
directly on the surface memory. You /must/ call this function before you
draw directly on the bitmap surface!

The return value from this function is the physical start address of the
buffer in memory, which can be used to program DMA operations from other
hardware devices directly into the video memory for the buffer. This is
useful for frame grabber devices so that the resulting frame from the
frame grabber device can be blitted to the visual display as quickly as
possible (much quicker than if it was DMA'ed into system memory).

If the buffer is a system memory buffer, the return value from this
function will be 0, since we cannot obtain the physical starting address
of a system memory buffer.
****************************************************************************/
N_uint32 NAPI BM_LockBuffer(
    GA_buf *buf)
{
    if (!(buf->Flags & gaBufferSysMem)) {
        buf->Surface = (uchar*)BM.hwCtx->LinearMem + buf->Offset;
        if (BM.gaMode.Attributes & gaHaveAccel2D) {
            if (BM.surfLock++ == 0) {
                if (BM.hwState.EnableDirectAccess)
                    BM.hwState.EnableDirectAccess();
                else
                    BM.hwState.WaitTillIdle();
                }
            }
        return BM.hwCtx->LinearBasePtr + buf->Offset;
        }
    return 0;
}

/****************************************************************************
PARAMETERS:
buf - SNAP buffer to unlock

REMARKS:
This function unlocks a buffer after the application has completed direct
surface access on the buffer.
****************************************************************************/
void NAPI BM_UnlockBuffer(
    GA_buf *buf)
{
    if (!(buf->Flags & gaBufferSysMem)) {
        if (BM.gaMode.Attributes & gaHaveAccel2D) {
            if (--BM.surfLock == 0) {
                if (BM.hwState.DisableDirectAccess)
                    BM.hwState.DisableDirectAccess();
                }
            }
        buf->Surface = NULL;
        }
}

/****************************************************************************
PARAMETERS:
buf - SNAP buffer to make active for rendering

REMARKS:
This function makes the passed in buffer the active buffer for all
subsequent rendering operations.
****************************************************************************/
N_int32 NAPI BM_SetActiveBuffer(
    GA_buf *buf)
{
    GA_buffer   drawBuf;

    BM.activeBuf = buf;
    if ((buf->Flags & gaBufferSysMem)) {
        // Switch drawing buffer to system memory buffer
        BM.ref2d->SetDrawSurface(buf->Surface,buf->Width,buf->Height,buf->Stride,
            buf->BitsPerPixel,&buf->PixelFormat);
        }
    else {
        // Switch drawing buffer to hardware video memory buffer
        BM.ref2d->SetDrawSurface(BM.hwCtx->LinearMem,buf->Width,buf->Height,buf->Stride,
            buf->BitsPerPixel,&buf->PixelFormat);
        BM.activeY     = buf->StartY;
        BM.activeX     = buf->StartX;
        drawBuf.dwSize = sizeof(drawBuf);
        drawBuf.Offset = buf->Offset;
        drawBuf.Stride = buf->Stride;
        drawBuf.Width  = buf->Width;
        drawBuf.Height = buf->Height;
        return BM.state2d.SetDrawBuffer(&drawBuf);
        }
    return 0;
}

/****************************************************************************
RETURNS:
Pointer to the last buffer passed to SetActiveBuffer
****************************************************************************/
GA_buf * NAPI BM_GetActiveBuffer(void)
{
    return BM.activeBuf;
}

/****************************************************************************
PARAMETERS:
buf - SNAP buffer to make visible for display

REMARKS:
This function makes the passed in buffer visible for the active display.
****************************************************************************/
void NAPI BM_FlipToBuffer(
    GA_buf *buf,
    N_int32 waitVRT)
{
    // Do a stereo flip to the same buffer in both eyes for a normal flip
    BM_FlipToStereoBuffer(buf,buf,waitVRT);
}

/****************************************************************************
PARAMETERS:
left    - SNAP buffer to make visible for the left eye
right   - SNAP buffer to make visible for the right eye

REMARKS:
This function makes the passed in buffer visible for the active display.
If stereo mode is on, the left and right buffers are enabled for stereo
flipping.
****************************************************************************/
void NAPI BM_FlipToStereoBuffer(
    GA_buf *left,
    GA_buf *right,
    N_int32 waitVRT)
{
    int leftAddr,rightAddr;

    // Check that the buffer is not moveable or pageable
    if ((left->Flags & (gaBufferMoveable | gaBufferPageable)) ||
        (right->Flags & (gaBufferMoveable | gaBufferPageable)))
        return;

    // Make sure the buffers are flippable!
    if (!(left->Flags & gaBufferFlippable) ||
        !(right->Flags & gaBufferFlippable))
        return;

    // Fix the triple buffer flag if we only have two buffers
    BM.visibleBuf = left;
    BM.visibleRightBuf = right;
    if (BM.numFlipBuffers <= 1 && waitVRT == gaTripleBuffer)
        waitVRT = gaWaitVRT;

    // Ensure that previous display start has taken hold if we have hardware
    // triple buffering. We do this by polling the GetDisplayStartStatus
    // function until it tells us that the last frame has taken hold. In most
    // cases unless the framerate is running close to the refresh rate of
    // the video mode, this function loop will terminate immediately and we
    // never wait. If we do wait, it is because we are running too fast!
    if (BM.isStereo == gaSoftStereo) {
        if (waitVRT == gaTripleBuffer)
            BM.stereo.SoftStereoWaitTillFlipped();
        }
    else {
        if ((BM.gaMode.Attributes & gaHaveTripleBuffer) && (waitVRT == gaTripleBuffer)) {
            while (BM.hwDrv.GetDisplayStartStatus() == 0)
                ;
            }
        }

    // Now flip the visible display pages
    leftAddr = left->Offset;
    rightAddr = right->Offset;
    if (BM.gaMode.BitsPerPixel < 8) {
        leftAddr *= 8;
        rightAddr *= 8;
        }
    if (BM.isStereo == gaSoftStereo) {
        BM.stereo.SoftStereoScheduleFlip(leftAddr,rightAddr);
        if (waitVRT == gaWaitVRT)
            BM.stereo.SoftStereoWaitTillFlipped();
        }
    else if (BM.isStereo == gaHWStereo)
        BM.hwDrv.SetStereoDisplayStart(leftAddr,rightAddr,(waitVRT == gaWaitVRT));
    else
        BM.hwDrv.SetDisplayStart(leftAddr,(waitVRT == gaWaitVRT));
}

/****************************************************************************
RETURNS:
Pointer to the last buffer passed to FlipToBuffer
****************************************************************************/
GA_buf * NAPI BM_GetVisibleBuffer(void)
{
    return BM.visibleBuf;
}

/****************************************************************************
RETURNS:
Pointer to the last right buffer passed to FlipToStereoBuffer
****************************************************************************/
GA_buf * NAPI BM_GetVisibleRightBuffer(void)
{
    return BM.visibleRightBuf;
}

/****************************************************************************
REMARKS:
This function updates returns the flip status for the currently pending
flip operation. 1 means it has been flipped, 0 means it has not yet been
flipped.
****************************************************************************/
int NAPI BM_GetFlipStatus(void)
{
    if (BM.hwState.IsIdle && !BM.hwState.IsIdle())
        return 0;
    if (BM.isStereo == gaSoftStereo)
        return BM.stereo.SoftStereoGetFlipStatus();
    else if (BM.hwDrv.GetDisplayStartStatus)
        return BM.hwDrv.GetDisplayStartStatus();
    return 1;
}

/****************************************************************************
REMARKS:
This function waits until the pending flip operation has completed and the
visible display page has changed.
****************************************************************************/
void NAPI BM_WaitTillFlipped(void)
{
    while (!BM_GetFlipStatus())
        ;
}

/****************************************************************************
PARAMETERS:
buf - SNAP buffer to copy into system memory cache

REMARKS:
This function updates the buffer cache in system memory by copying the
contents of the video memory buffer into the system memory cache.
****************************************************************************/
void NAPI BM_UpdateCache(
    GA_buf *buf)
{
    if (buf->Flags & gaBufferCached) {
        uchar   *s = buf->Surface;
        uchar   *d = buf->SurfaceCache;
        int     width = buf->Width;
        int     height = buf->Height;
        int     srcStride = buf->Stride;
        int     dstStride = buf->CacheStride;

        BM_LockBuffer(buf);
        while (height--) {
            BM.bufReadScan(d,s X_ARG2(0,0),width);
            s += srcStride;
            d += dstStride;
            }
        BM_UnlockBuffer(buf);
        }
}

/****************************************************************************
PARAMETERS:
buf - SNAP buffer to copy from system memory cache

REMARKS:
This function updates the video memory buffer by copying the contents of
the system memory cache buffer into the video memory buffer.
****************************************************************************/
void NAPI BM_UpdateFromCache(
    GA_buf *buf)
{
    if (buf->Flags & gaBufferCached) {
        uchar   *s = buf->SurfaceCache;
        uchar   *d = buf->Surface;
        int     width = buf->Width;
        int     height = buf->Height;
        int     srcStride = buf->CacheStride;
        int     dstStride = buf->Stride;

        BM_LockBuffer(buf);
        while (height--) {
            BM.bufBltScan(d,s X_ARG2(0,0),width);
            s += srcStride;
            d += dstStride;
            }
        BM_UnlockBuffer(buf);
        }
}

/****************************************************************************
PARAMETERS:
buf     - SNAP buffer to blit to the active surface
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
mix     - Mix code for the copy (GA_mixCodesType)

REMARKS:
This function blits the contents of the buffer to the active drawing
surface.
****************************************************************************/
void NAPI BM_BitBltBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    // Increment usage counter
    buf->UseageCount++;
    if (BM.activeBuf != buf && (buf->Flags & gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache)
        BM.draw2d.BitBltSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,mix,false);
        }
    else if (buf->StartX == -1) {
        // Buffer is in linear video memory
        BM.draw2d.BitBltLin(
            buf->Offset,buf->Stride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,mix);
        }
    else {
        // Buffer is in (x,y) video memory or source is the active buffer
        BM.draw2d.BitBlt(
            srcLeft+buf->StartX-BM.activeX,
            srcTop+buf->StartY-BM.activeY,
            width,height,dstLeft,dstTop,mix);
        }
}

/****************************************************************************
PARAMETERS:
buf     - SNAP buffer to blit to the active surface
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 code for the copy (GA_rop3CodesType)

REMARKS:
This function blits the contents of the buffer to the active drawing
surface with a mono pattern applied.
****************************************************************************/
void NAPI BM_BitBltPattBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    // Increment usage counter
    buf->UseageCount++;
    if ((!BM.hwDraw.BitBltPatt && (buf->Flags & gaBufferCached))
            || (buf->Flags & gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache).
        // We also end up here if the buffer is in video memory but the
        // hardware does not support this function and we have a buffer
        // cache. It is faster to do software rendering from the buffer
        // cache memory than doing software blits within video memory.
        BM.draw2d.BitBltPattSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,rop3,false);
        }
    else if (buf->StartX == -1) {
        // Buffer is in linear video memory
        BM.draw2d.BitBltPattLin(
            buf->Offset,buf->Stride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,rop3);
        }
    else {
        // Buffer is in (x,y) video memory
        BM.draw2d.BitBltPatt(
            srcLeft+buf->StartX-BM.activeX,
            srcTop+buf->StartY-BM.activeY,
            width,height,dstLeft,dstTop,rop3);
        }
}

/****************************************************************************
PARAMETERS:
buf     - SNAP buffer to blit to the active surface
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 code for the copy (GA_rop3CodesType)

REMARKS:
This function blits the contents of the buffer to the active drawing
surface with a color pattern applied.
****************************************************************************/
void NAPI BM_BitBltColorPattBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    // Increment usage counter
    buf->UseageCount++;
    if ((!BM.hwDraw.BitBltColorPatt && (buf->Flags & gaBufferCached))
            || (buf->Flags & gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache).
        // We also end up here if the buffer is in video memory but the
        // hardware does not support this function and we have a buffer
        // cache. It is faster to do software rendering from the buffer
        // cache memory than doing software blits within video memory.
        BM.draw2d.BitBltColorPattSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,rop3,false);
        }
    else if (buf->StartX == -1) {
        // Buffer is in linear video memory
        BM.draw2d.BitBltColorPattLin(
            buf->Offset,buf->Stride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,rop3);
        }
    else {
        // Buffer is in (x,y) video memory
        BM.draw2d.BitBltColorPatt(
            srcLeft+buf->StartX-BM.activeX,
            srcTop+buf->StartY-BM.activeY,
            width,height,dstLeft,dstTop,rop3);
        }
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to blit to the active surface
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color for the blit

REMARKS:
This function blits the contents of the buffer to the active drawing
surface with source transparency.
****************************************************************************/
void NAPI BM_SrcTransBltBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    // Increment usage counter
    buf->UseageCount++;
    if ((!BM.hwDraw.SrcTransBlt && (buf->Flags & gaBufferCached))
            || (buf->Flags & gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache).
        // We also end up here if the buffer is in video memory but the
        // hardware does not support this function and we have a buffer
        // cache. It is faster to do software rendering from the buffer
        // cache memory than doing software blits within video memory.
        BM.draw2d.SrcTransBltSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,mix,transparent,false);
        }
    else if (buf->StartX == -1) {
        // Buffer is in linear video memory
        BM.draw2d.SrcTransBltLin(
            buf->Offset,buf->Stride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,mix,transparent);
        }
    else {
        // Buffer is in (x,y) video memory
        BM.draw2d.SrcTransBlt(
            srcLeft+buf->StartX-BM.activeX,
            srcTop+buf->StartY-BM.activeY,
            width,height,dstLeft,dstTop,mix,transparent);
        }
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to blit to the active surface
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color for the blit

REMARKS:
This function blits the contents of the buffer to the active drawing
surface with destination transparency.
****************************************************************************/
void NAPI BM_DstTransBltBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    // Increment usage counter
    buf->UseageCount++;
    if ((!BM.hwDraw.DstTransBlt && (buf->Flags & gaBufferCached))
            || (buf->Flags & gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache).
        // We also end up here if the buffer is in video memory but the
        // hardware does not support this function and we have a buffer
        // cache. It is faster to do software rendering from the buffer
        // cache memory than doing software blits within video memory.
        BM.draw2d.DstTransBltSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,mix,transparent,false);
        }
    else if (buf->StartX == -1) {
        // Buffer is in linear video memory
        BM.draw2d.DstTransBltLin(
            buf->Offset,buf->Stride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,mix,transparent);
        }
    else {
        // Buffer is in (x,y) video memory
        BM.draw2d.DstTransBlt(
            srcLeft+buf->StartX-BM.activeX,
            srcTop+buf->StartY-BM.activeY,
            width,height,dstLeft,dstTop,mix,transparent);
        }
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to blit to the active surface
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
doClip      - True if the blit should be clipped
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle
mix         - Mix code for the copy (GA_mixCodesType)

REMARKS:
This function blits the contents of the buffer to the active drawing
surface with stretching.
****************************************************************************/
void NAPI BM_StretchBltBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 mix)
{
    // Increment usage counter
    buf->UseageCount++;
    if ((!BM.hwDraw.StretchBlt && (buf->Flags & gaBufferCached))
            || (buf->Flags & gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache).
        // We also end up here if the buffer is in video memory but the
        // hardware does not support this function and we have a buffer
        // cache. It is faster to do software rendering from the buffer
        // cache memory than doing software blits within video memory.
        BM.draw2d.StretchBltSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,
            doClip,clipLeft,clipTop,clipRight,clipBottom,mix,false);
        }
    else if (buf->StartX == -1) {
        // Buffer is in linear video memory
        BM.draw2d.StretchBltLin(
            buf->Offset,buf->Stride,srcLeft,srcTop,
            srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,
            doClip,clipLeft,clipTop,clipRight,clipBottom,mix);
        }
    else {
        // Buffer is in (x,y) video memory
        BM.draw2d.StretchBlt(
            srcLeft+buf->StartX-BM.activeX,
            srcTop+buf->StartY-BM.activeY,
            srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,
            doClip,clipLeft,clipTop,clipRight,clipBottom,mix);
        }
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to blit to the active surface
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
planeMask   - 8-bit plane mask to use for the copy

REMARKS:
This function blits the contents of the buffer to the active drawing
surface.
****************************************************************************/
void NAPI BM_BitBltPlaneMaskedBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_uint32 planeMask)
{
    // Increment usage counter
    buf->UseageCount++;
    if ((!BM.hwDraw.BitBltPlaneMasked && (buf->Flags & gaBufferCached))
            || (buf->Flags & gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache).
        // We also end up here if the buffer is in video memory but the
        // hardware does not support this function and we have a buffer
        // cache. It is faster to do software rendering from the buffer
        // cache memory than doing software blits within video memory.
        BM.draw2d.BitBltPlaneMaskedSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,planeMask,false);
        }
    else if (buf->StartX == -1) {
        // Buffer is in linear video memory
        BM.draw2d.BitBltPlaneMaskedLin(
            buf->Offset,buf->Stride,srcLeft,srcTop,
            width,height,dstLeft,dstTop,planeMask);
        }
    else {
        // Buffer is in (x,y) video memory
        BM.draw2d.BitBltPlaneMasked(
            srcLeft+buf->StartX-BM.activeX,
            srcTop+buf->StartY-BM.activeY,
            width,height,dstLeft,dstTop,planeMask);
        }
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to blit to the active surface
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
fx          - Effects to apply to the blit operation

REMARKS:
This function blits the contents of the buffer to the active drawing
surface with optional effects.
****************************************************************************/
void NAPI BM_BitBltFxBuf(
    GA_buf *buf,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    GA_bltFx *fx)
{
    // Increment usage counter
    buf->UseageCount++;
    if ((!BM.hwDraw.BitBltFx && (buf->Flags & gaBufferCached))
            || (buf->Flags & gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache).
        // We also end up here if the buffer is in video memory but the
        // hardware does not support this function and we have a buffer
        // cache. It is faster to do software rendering from the buffer
        // cache memory than doing software blits within video memory.
        BM.draw2d.BitBltFxSys(
            buf->SurfaceCache,buf->CacheStride,srcLeft,srcTop,
            srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
        }
    else if (buf->StartX == -1) {
        // Buffer is in linear video memory
        BM.draw2d.BitBltFxLin(
            buf->Offset,buf->Stride,srcLeft,srcTop,
            srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
        }
    else {
        // Buffer is in (x,y) video memory
        BM.draw2d.BitBltFx(
            srcLeft+buf->StartX-BM.activeX,
            srcTop+buf->StartY-BM.activeY,
            srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
        }
}

/****************************************************************************
PARAMETERS:
buf         - SNAP buffer to draw to
left        - Left coordinate of the rectangle to draw
top         - Top coordinate of the rectangle to draw
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
color       - Color to fill the rectangle with
mix         - Mix to fill the rectangle with

REMARKS:
This function does a solid fill to the specified destination buffer without
requiring the active destination buffer to be changed.
****************************************************************************/
void NAPI BM_DrawRectBuf(
    GA_buf *buf,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    GA_color color,
    N_int32 mix)
{
    GA_buffer   drawBuf;

    // Increment usage counter
    buf->UseageCount++;
    if ((!BM.hwDraw.DrawRect && (buf->Flags & gaBufferCached)) || (buf->Flags & gaBufferSysMem)) {
        // Buffer is in system memory (possibly paged out into cache).
        // We also end up here if the buffer is in video memory but the
        // hardware does not support this function and we have a buffer
        // cache. It is faster to do software rendering from the buffer
        // cache memory than doing software blits within video memory.
        BM.ref2d->DrawRectExtSW(buf->SurfaceCache,buf->CacheStride,left,top,width,height,color,mix);
        }
    else if (buf->StartX == -1 && BM.hwDraw.DrawRectLin) {
        // Buffer is in linear hardware video memory
        BM.draw2d.DrawRectLin(buf->Offset,buf->Stride,left,top,width,height,color,mix);
        }
    else {
        // Buffer is in X,Y hardware video memory
        GA_buffer *hwDrawBuf = BM.ref2d->GetDrawBuffer();

        // Switch the active hardware draw buffer
        drawBuf.dwSize = sizeof(drawBuf);
        drawBuf.Offset = buf->Offset;
        drawBuf.Stride = buf->Stride;
        drawBuf.Width  = buf->Width;
        drawBuf.Height = buf->Height;
        if (BM.gaMode.Attributes & gaHaveAccel2D)
            BM.hwState.SetDrawBuffer(&drawBuf);
        else
            BM.state2d.SetDrawBuffer(&drawBuf);

        // Draw the rectangle in hardware
        BM.draw2d.DrawRectExt(left,top,width,height,color,mix);

        // Restore the current hardware draw buffer
        if (BM.gaMode.Attributes & gaHaveAccel2D)
            BM.hwState.SetDrawBuffer(hwDrawBuf);
        else
            BM.state2d.SetDrawBuffer(hwDrawBuf);
        }
}

