/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
PARAMETERS:
red     - Red portion of color to find
green   - Green portion of color to find
blue    - Blue portion of color to find

RETURNS:
Color index of the closest color to the specified RGB color

REMARKS:
This routine searches the destination bitmap palette for the color that is
the closest to the specified RGB color value, and then returns it. This is
used for converting RGB bitmaps to 4bpp and 8bpp color index modes.
****************************************************************************/
static GA_color FindClosestColor(
    uchar red,
    uchar green,
    uchar blue)
{
    GA_palette      *p = DC.cntPal;
    int             palSize = DC.palSize;
    int             i,d,distance = 0x7FFFL;
    static int      closest = 0;
    static uchar    lastR = 0,lastG = 0,lastB = 0;

    // Cache the last pixel lookup for speed
    if (red == lastR && green == lastG && blue == lastB)
        return closest;
    closest = 0;
    for (i = 0; i < palSize; i++,p++) {
        if ((d = ABS(red - (int)p->Red)
               + ABS(green - (int)p->Green)
               + ABS(blue - (int)p->Blue)) < distance) {
            distance = d;
            closest = i;
            }
        }
    lastR = red;
    lastG = green;
    lastB = blue;
    return closest;
}

/****************************************************************************
PARAMETERS:
color   - Packed RGB color value to find closest color for

RETURNS:
Color index of the closest color to the specified RGB color

REMARKS:
This routine searches the destination bitmap palette for the color that is
the closest to the specified RGB color value, and then returns it. This is
used for converting RGB bitmaps to 8bpp color index modes.
****************************************************************************/
GA_color REF2D_findClosest_15(
    GA_color color)
{
    return FindClosestColor(
        (uchar)((color >> 7) & 0xF8),
        (uchar)((color >> 2) & 0xF8),
        (uchar)(color << 3));
}

/****************************************************************************
PARAMETERS:
color   - Packed RGB color value to find closest color for

RETURNS:
Color index of the closest color to the specified RGB color

REMARKS:
This routine searches the destination bitmap palette for the color that is
the closest to the specified RGB color value, and then returns it. This is
used for converting RGB bitmaps to 8bpp color index modes.
****************************************************************************/
GA_color REF2D_findClosest_16(
    GA_color color)
{
    return FindClosestColor(
        (uchar)((color >> 8) & 0xF8),
        (uchar)((color >> 3) & 0xFC),
        (uchar)(color << 3));
}

/****************************************************************************
PARAMETERS:
color   - Packed RGB color value to find closest color for

RETURNS:
Color index of the closest color to the specified RGB color

REMARKS:
This routine searches the destination bitmap palette for the color that is
the closest to the specified RGB color value, and then returns it. This is
used for converting RGB bitmaps to 8bpp color index modes.
****************************************************************************/
GA_color REF2D_findClosest_24RGB(
    GA_color color)
{
    return FindClosestColor(
        (uchar)(color >> 16),
        (uchar)(color >> 8),
        (uchar)(color));
}

/****************************************************************************
PARAMETERS:
color   - Packed RGB color value to find closest color for

RETURNS:
Color index of the closest color to the specified RGB color

REMARKS:
This routine searches the destination bitmap palette for the color that is
the closest to the specified RGB color value, and then returns it. This is
used for converting RGB bitmaps to 8bpp color index modes.
****************************************************************************/
GA_color REF2D_findClosest_24BGR(
    GA_color color)
{
    return FindClosestColor(
        (uchar)(color),
        (uchar)(color >> 8),
        (uchar)(color >> 16));
}

/****************************************************************************
PARAMETERS:
color   - Packed RGB color value to find closest color for

RETURNS:
Color index of the closest color to the specified RGB color

REMARKS:
This routine searches the destination bitmap palette for the color that is
the closest to the specified RGB color value, and then returns it. This is
used for converting RGB bitmaps to 8bpp color index modes.
****************************************************************************/
GA_color REF2D_findClosest_32RGBA(
    GA_color color)
{
    return FindClosestColor(
        (uchar)(color >> 24),
        (uchar)(color >> 16),
        (uchar)(color >> 8));
}

/****************************************************************************
PARAMETERS:
color   - Packed RGB color value to find closest color for

RETURNS:
Color index of the closest color to the specified RGB color

REMARKS:
This routine searches the destination bitmap palette for the color that is
the closest to the specified RGB color value, and then returns it. This is
used for converting RGB bitmaps to 8bpp color index modes.
****************************************************************************/
GA_color REF2D_findClosest_32BGRA(
    GA_color color)
{
    return FindClosestColor(
        (uchar)(color >> 8),
        (uchar)(color >> 16),
        (uchar)(color >> 24));
}

