/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for the SciTech SNAP Graphics 2D Flipped
*               Display filter driver internal functions.
*
****************************************************************************/

#ifndef __FLIPINT_H
#define __FLIPINT_H

#include "ref2dint.h"

/*---------------------- Macros and type definitions ----------------------*/

#pragma pack(4)

/****************************************************************************
REMARKS:
Structure to store the current state of the 2D portrait filter driver.

MEMBERS:
hwCtx               - Hardware device context bound to this filter driver
shared              - True if loaded into shared memory
enabled             - True if portrait mode is enabled
mi                  - Mode information block for the current hardware mode
buf                 - Current draw buffer
bytesPerPixel       - Number of bytes in a single pixel
ref2d               - Pointer to the ref2d driver
refState            - Pointer to the ref2d driver internal state
cursor              - Copy of the ref2d cursor functions
state2d             - Copy of the ref2d state functions
draw2d              - Copy of the ref2d draw functions
bufferSize          - Size of the transfer buffer in bytes
transferStart       - Starting address for transfer buffer
VRAMBufferStart     - Start of buffer in VRAM, or 0 if in system memory
VRAMBufferStartY    - Y coordinate of the start of the VRAM buffer in memory
buffer              - Temporary buffer for rotating all bitmaps
monoBuffer          - Temporary buffer for rotating monochrome bitmaps
****************************************************************************/
typedef struct {
    GA_devCtx           *hwCtx;
    ibool               enabled;
    GA_modeInfo         mi;
    GA_buffer           buf;
    int                 bytesPerPixel;
    int                 maxX,maxY;
    N_fix32             maxXFX,maxYFX;
    REF2D_driver        *ref2d;
    REF2D_state         *refState;
    ulong               bufferSize;
    ulong               transferStart;
    ulong               VRAMBufferStart;
    ulong               VRAMBufferStartY;
    void                *buffer;
    void                *sysBuffer;
    GA_options          opt;
    GA_globalOptions    *gOpt;
    int                 cursorHotX;
    int                 cursorHotY;
    GA_monoCursor       monoCursor;
    GA_colorCursor      colorCursor;
    GA_colorCursor256   colorCursor256;
    GA_colorCursorRGB   colorCursorRGB;
    GA_colorCursorRGBA  colorCursorRGBA;
    ibool               (NAPIP QueryFunctionsDC)(N_uint32 id,N_int32 safetyLevel,void *funcs);
    } FLIPPED_state;

/****************************************************************************
PARAMETERS:
y       - Y coordinate for point
height  - Height of the rectangle

REMARKS:
Macro to flip the Y coordinate for a rectangle
****************************************************************************/
#define FLIP_RECTX(x,width)     (DC.maxX-(x)-(width)+1)
#define FLIP_RECTY(y,height)    (DC.maxY-(y)-(height)+1)

/****************************************************************************
PARAMETERS:
y   - Y coordinate for point

REMARKS:
Macro to flip the Y coordinate for a primitive
****************************************************************************/
#define FLIPX(x)    (DC.maxX-(x))
#define FLIPY(y)    (DC.maxY-(y))

/****************************************************************************
PARAMETERS:
y   - Y coordinate for point (fixed point)

REMARKS:
Macro to flip the Y coordinate for a primitive
****************************************************************************/
#define FLIPFX(x)   (DC.maxYFX-(x))
#define FLIPFY(y)   (DC.maxYFX-(y))

/****************************************************************************
REMARKS:
Macro to determine if invert mode is enabled or not.
****************************************************************************/
#define IS_FLIPPED() \
    (DC.hwCtx != 0 && (DC.gOpt->bFlipped == 1 && DC.opt.bFlipped == 1) && DC.mi.BitsPerPixel >= 8)

/****************************************************************************
REMARKS:
Round a value to the specified integer boundary
****************************************************************************/
#define ROUNDUP(value,boundary) \
    ((((value) + ((boundary)-1)) / (boundary)) * (boundary))

/****************************************************************************
REMARKS:
Macros to set a state 2d vector if not overriden by the filter driver
****************************************************************************/
#define SET_STATE2D_VECTOR(name)    \
    if (!FLIPPED_state2dPT.name) FLIPPED_state2dPT.name = FLIPPED_state2dREF.name

/****************************************************************************
REMARKS:
Macros to set a draw 2d vector if not overriden by the filter driver
****************************************************************************/
#define SET_DRAW2D_VECTOR(name) \
    if (FLIPPED_draw2dStub.name && !FLIPPED_draw2dPT.name) FLIPPED_draw2dPT.name = FLIPPED_draw2dREF.name

/****************************************************************************
REMARKS:
Macros to set a state 2d vector if not overriden by the filter init
****************************************************************************/
#define SET_INIT_VECTOR(name)   \
    if (!FLIPPED_initPT.name) FLIPPED_initPT.name = FLIPPED_initREF.name

/****************************************************************************
REMARKS:
Macros to set a state 2d vector if not overriden by the filter driver
****************************************************************************/
#define SET_CURSOR_VECTOR(name) \
    if (!FLIPPED_cursorPT.name) FLIPPED_cursorPT.name = FLIPPED_cursorREF.name

#pragma pack()

/*---------------------------- Global Variables ---------------------------*/

extern FLIPPED_state    FLIPPED_devCtx;
extern REF2D_driver     FLIPPED_drv;
extern GA_cursorFuncs   FLIPPED_cursorREF;
extern GA_2DStateFuncs  FLIPPED_state2dREF;
extern GA_2DRenderFuncs FLIPPED_draw2dREF;
extern GA_2DRenderFuncs _VARAPI FLIPPED_draw2dCnt;

#undef  DC
#undef  DRV
#define DC      FLIPPED_devCtx
#define DC2D    (*FLIPPED_devCtx.refState)
#define DRV     FLIPPED_drv
#define CURSOR  FLIPPED_cursorREF
#define STATE2D FLIPPED_state2dREF
#define DRAW2D  FLIPPED_draw2dREF

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

/* Library initialisation entry points */

ibool       NAPI FLIPPED_QueryFunctionsDC(N_uint32 id,N_int32 safetyLevel,void *funcs);
ibool       NAPI FLIPPED_QueryFunctions(N_uint32 id,void *funcs);
N_int32     NAPI FLIPPED_SetDrawBufferSW(GA_buffer *drawBuf,void *framebuffer,N_int32 bitsPerPixel,GA_pixelFormat *pf,GA_devCtx *hwCtx,N_int32 softwareOnly);
void        NAPI FLIPPED_SetDrawSurface(void *surface,N_int32 xRes,N_int32 yRes,N_int32 bytesPerLine,N_int32 bitsPerPixel,GA_pixelFormat *pf);
void *      NAPI FLIPPED_RotateBitmap(void *src,N_int32 bitsPerPixel,N_int32 *stride,N_int32 width,N_int32 height);
void        NAPI FLIPPED_UnloadDriver(void);
ibool       NAPI FLIPPED_InitFilter(GA_modeInfo *modeInfo,N_int32 transferStart,struct _REF2D_driver *ref2d,struct _REF2D_driver **drv);
void        NAPI FLIPPED_DeInitFilter(void);
void        NAPI FLIPPED_GlobalOptionsUpdated(void);

/* Flipped init functions */

void        NAPI FLIPPED_SetOptions(GA_options *options);

/* Flipped cursor functions */

void        NAPI FLIPPED_SetMonoCursor(GA_monoCursor *cursor);
void        NAPI FLIPPED_SetColorCursor(GA_colorCursor *cursor);
void        NAPI FLIPPED_SetColorCursor256(GA_colorCursor256 *cursor);
void        NAPI FLIPPED_SetColorCursorRGB(GA_colorCursorRGB *cursor);
void        NAPI FLIPPED_SetColorCursorRGBA(GA_colorCursorRGBA *cursor);
N_int32     NAPI FLIPPED_SetCursorPos(N_int32 x,N_int32 y);
void        NAPI FLIPPED_BeginAccess(N_int32 left,N_int32 top,N_int32 right,N_int32 bottom);

/* 2D state reference functions */

N_int32     NAPI FLIPPED_SetDrawBuffer(GA_buffer *drawBuf);
void        NAPI FLIPPED_Set8x8MonoPattern(N_int32 index,GA_pattern *pattern);
void        NAPI FLIPPED_Set8x8ColorPattern(N_int32 index,GA_colorPattern *pattern);

/* 2D generic reference functions */

void        NAPI FLIPPED_DrawScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI FLIPPED_DrawPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI FLIPPED_DrawColorPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI FLIPPED_DrawScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI FLIPPED_DrawPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI FLIPPED_DrawColorPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI FLIPPED_DrawEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI FLIPPED_DrawPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI FLIPPED_DrawColorPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI FLIPPED_DrawFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI FLIPPED_DrawPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI FLIPPED_DrawColorPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI FLIPPED_DrawRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI FLIPPED_DrawPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI FLIPPED_DrawColorPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI FLIPPED_DrawRectPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI FLIPPED_DrawRectColorPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI FLIPPED_DrawTrap(GA_trap *trap);
void        NAPI FLIPPED_DrawPattTrap(GA_trap *trap);
void        NAPI FLIPPED_DrawColorPattTrap(GA_trap *trap);
void        NAPI FLIPPED_DrawLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void        NAPI FLIPPED_DrawBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
void        NAPI FLIPPED_DrawStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI FLIPPED_DrawBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
void        NAPI FLIPPED_PutMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI FLIPPED_PutMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI FLIPPED_ClipMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI FLIPPED_ClipMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI FLIPPED_BitBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI FLIPPED_BitBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
void        NAPI FLIPPED_BitBltPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI FLIPPED_BitBltPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI FLIPPED_BitBltColorPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI FLIPPED_BitBltColorPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI FLIPPED_SrcTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI FLIPPED_SrcTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI FLIPPED_DstTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI FLIPPED_DstTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI FLIPPED_StretchBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI FLIPPED_StretchBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix,N_int32 flipY);
N_int32     NAPI FLIPPED_ConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32     NAPI FLIPPED_StretchConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
void        NAPI FLIPPED_GetBitmapSys(void *dstAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI FLIPPED_DrawEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B);
void        NAPI FLIPPED_ClipEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
GA_color    NAPI FLIPPED_GetPixel(N_int32 x,N_int32 y);
void        NAPI FLIPPED_PutPixel(N_int32 x,N_int32 y);
void        NAPI FLIPPED_DrawStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI FLIPPED_DrawBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
N_int32     NAPI FLIPPED_DrawClippedStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI FLIPPED_DrawClippedBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);

/* Assembler stubs to call the functions via the global REF_draw2d */

void        NAPI FSTUB_DrawScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI FSTUB_DrawPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI FSTUB_DrawColorPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI FSTUB_DrawScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI FSTUB_DrawPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI FSTUB_DrawColorPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI FSTUB_DrawEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI FSTUB_DrawPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI FSTUB_DrawColorPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI FSTUB_DrawFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI FSTUB_DrawPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI FSTUB_DrawColorPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI FSTUB_DrawRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI FSTUB_DrawPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI FSTUB_DrawColorPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI FSTUB_DrawRectPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI FSTUB_DrawRectColorPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI FSTUB_DrawTrap(GA_trap *trap);
void        NAPI FSTUB_DrawPattTrap(GA_trap *trap);
void        NAPI FSTUB_DrawColorPattTrap(GA_trap *trap);
void        NAPI FSTUB_DrawLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void        NAPI FSTUB_DrawBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
void        NAPI FSTUB_DrawStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI FSTUB_DrawBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
void        NAPI FSTUB_PutMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI FSTUB_PutMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI FSTUB_ClipMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI FSTUB_ClipMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI FSTUB_BitBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI FSTUB_BitBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
void        NAPI FSTUB_BitBltPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI FSTUB_BitBltPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI FSTUB_BitBltColorPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI FSTUB_BitBltColorPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI FSTUB_SrcTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI FSTUB_SrcTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI FSTUB_DstTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI FSTUB_DstTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI FSTUB_StretchBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI FSTUB_StretchBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix,N_int32 flipY);
N_int32     NAPI FSTUB_ConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32     NAPI FSTUB_StretchConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
void        NAPI FSTUB_GetBitmapSys(void *dstAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI FSTUB_DrawEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B);
void        NAPI FSTUB_ClipEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
GA_color    NAPI FSTUB_GetPixel(N_int32 x,N_int32 y);
void        NAPI FSTUB_PutPixel(N_int32 x,N_int32 y);
void        NAPI FSTUB_DrawStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI FSTUB_DrawBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
N_int32     NAPI FSTUB_DrawClippedStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI FSTUB_DrawClippedBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __FLIPINT_H */

