/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Internal header file to bring in most header files needed
*               by ref2d drivers.
*
****************************************************************************/

#ifndef __INTERNAL_H
#define __INTERNAL_H

#include "snap/graphics.h"
#include "snap/common/io.h"
#include "snap/common/list.h"
#include "snap/common/mem.h"
#include "clipline.h"

/*---------------------- Macros and type definitions ----------------------*/

// Define which colors depths are supported. If none are defined, then we
// include them all in a single driver.

#if !defined(SUPPORT_1BPP) && !defined(SUPPORT_4BPP) && \
    !defined(SUPPORT_8BPP) && !defined(SUPPORT_16BPP) && \
    !defined(SUPPORT_24BPP) && !defined(SUPPORT_32BPP)
//#define   SUPPORT_1BPP
#define SUPPORT_4BPP
#define SUPPORT_8BPP
#define SUPPORT_16BPP
#define SUPPORT_24BPP
#define SUPPORT_32BPP
#define SUPPORT_ALLBPP
#endif

// For 1bpp and 4pp packed pixel renderers, we need to pass and extra X
// coordinate parameter to each rendering function for proper address
// calculations. Since we don't want to burden the 8bpp and higher color
// depths with the extra overhead, we only pass this parameter if 1bpp
// or 4bpp packed pixel modes are compiled in (ie: compiling the source
// debug library).

#if defined(SUPPORT_1BPP) || defined(SUPPORT_4BPP)
#define DCL_X_ARG       ,int x
#define DCL_X_ARG2      ,int srcLeft,int dstLeft
#define DCL_X_ARG3      ,int srcLeft
#define X_ARG(x)        ,x
#define X_ARG2(s,d)     ,s,d
#else
#define DCL_X_ARG
#define DCL_X_ARG2
#define DCL_X_ARG3
#define X_ARG(x)
#define X_ARG2(s,d)
#endif

// Macro to determine if the surface pointer is in hardware video memory

#define IS_HARDWARE_SURFACE(surface)    (DC.hwBufferStart <= (uchar*)surface && (uchar*)surface < DC.hwBufferEnd)

// Determine if blending is enabled

#define BLENDING_ENABLED_EX(sf,df)  (sf != gaBlendNone && df != gaBlendNone)
#define BLENDING_ENABLED()          BLENDING_ENABLED_EX(DC.srcBlendFunc,DC.dstBlendFunc)

// Determine if clamping is required

#define NO_CLAMPING_EX(sf,df)   \
    ((sf == gaBlendSrcColor && df == gaBlendOneMinusSrcColor) || \
     (sf == gaBlendSrcAlpha && df == gaBlendOneMinusSrcAlpha) || \
     (sf == gaBlendDstAlpha && df == gaBlendOneMinusDstAlpha) || \
     (sf == gaBlendDstColor && df == gaBlendOneMinusDstColor) || \
     (sf == gaBlendConstantColor && df == gaBlendOneMinusConstantColor) || \
     (sf == gaBlendConstantAlpha && df == gaBlendOneMinusConstantAlpha))
#define NO_CLAMPING()   NO_CLAMPING_EX(DC.srcBlendFunc,DC.dstBlendFunc)

// Determine if we can use optimised solid drawing functions

#define IS_SOLID_MIX(mix)   ((mix) == GA_REPLACE_MIX && DC.planeMask == 0xFFFFFFFF && !BLENDING_ENABLED())

// Determine if the blit should be dithered or not

#define IS_DITHERED(dither,bitsPerPixel,pixelFormat)                                        \
    (dither && bitsPerPixel >= 15 && DC.mi.bitsPerPixel >= 4 && DC.mi.bitsPerPixel <= 16 && \
    (bitsPerPixel != DC.mi.bitsPerPixel || (bitsPerPixel == DC.mi.bitsPerPixel &&           \
    !CMP_PF(DC.mi.pf,*pixelFormat))))

// Macro to get the correct pixel drawing function for the current mode

#define GET_PIXEL_FUNC_EX(sf,df,bi,mix) BLENDING_ENABLED_EX(sf,df) ? DC.blendPixelTable[(bi)] : DC.drawPixelTable[(mix)&0xF]
#define GET_PIXEL_FUNC(mix)             GET_PIXEL_FUNC_EX(DC.srcBlendFunc,DC.dstBlendFunc,DC.blendIndex,mix)

// Indexes of pixel blending functions in the DC.blendPixelTable array

#define BLEND_PIXEL_SRC_ALPHA_FAST      0
#define BLEND_PIXEL_CONST_ALPHA_FAST    1
#define BLEND_PIXEL                     2
#define BLEND_PIXEL_CLAMPED             3

// Function to get the pixel blending function index given the source
// and dest blend functions

#define GET_BLEND_INDEX(sf,df) ((sf == gaBlendSrcAlphaFast) ? BLEND_PIXEL_SRC_ALPHA_FAST : ((sf == gaBlendConstantAlphaFast) ? BLEND_PIXEL_CONST_ALPHA_FAST : (NO_CLAMPING_EX(sf,df) ? BLEND_PIXEL : BLEND_PIXEL_CLAMPED)))

// Pixel address arithmetic macros

#if defined(SUPPORT_ALLBPP)
#define LIN_X_ADDR(x,s)             ((s) <= 0 ? ((x) >> (-s)) : ((s) == 3 ? ((x) * 3) : ((x) << (s))))
#define BYTESPERLINE(w,p)           ((p) <= 4 ? ((w) / (8/(p))) : ((w) * ((p+7)/8)))
#define PIXELSPERLINE(w,p)          ((p) <= 4 ? ((w) * (8/(p))) : ((w) / ((p+7)/8)))

#elif defined(SUPPORT_1BPP)
#define LIN_X_ADDR(x,s)             ((x) >> 3)
#define BYTESPERLINE(w,p)           ((w) >> 3)
#define PIXELSPERLINE(w,p)          ((w) << 3)

#elif defined(SUPPORT_4BPP)
#define LIN_X_ADDR(x,s)             ((x) >> 1)
#define BYTESPERLINE(w,p)           ((w) >> 1)
#define PIXELSPERLINE(w,p)          ((w) << 1)

#elif defined(SUPPORT_8BPP)
#define LIN_X_ADDR(x,s)             (x)
#define BYTESPERLINE(w,p)           (w)
#define PIXELSPERLINE(w,p)          (w)

#elif defined(SUPPORT_16BPP)
#define LIN_X_ADDR(x,s)             ((x) << 1)
#define BYTESPERLINE(w,p)           ((w) << 1)
#define PIXELSPERLINE(w,p)          ((w) >> 1)

#elif defined(SUPPORT_24BPP)
#define LIN_X_ADDR(x,s)             ((x) * 3)
#define BYTESPERLINE(w,p)           ((w) * 3)
#define PIXELSPERLINE(w,p)          ((w) / 3)

#else
#define LIN_X_ADDR(x,s)             ((x) << 2)
#define BYTESPERLINE(w,p)           ((w) << 2)
#define PIXELSPERLINE(w,p)          ((w) >> 2)

#endif
#define LIN_Y_ADDR(y,s,b)           ((uchar*)(s) + (b) * (y))
#define LIN_PIXEL_ADDR(x,y,s,b,sh)  (LIN_Y_ADDR(y,s,b) + LIN_X_ADDR(x,sh))
#define X_ADDR(x)                   LIN_X_ADDR(x,DC.mi.byteShift)
#define Y_ADDR(y)                   LIN_Y_ADDR(y,DC.surface,DC.mi.bytesPerLine)
#define PIXEL_ADDR(x,y)             (Y_ADDR(y) + X_ADDR(x))

// Macro to increment any pointer by a number of bytes

#define BINC(v,inc)     *((char**)&(v)) += (inc)

// Macro to decrement any pointer by a number of bytes

#define BDEC(v,inc)     *((char**)&(v)) -= (inc)

// Macro to swap the value of two integer variables

#define SWAP(a,b)       { a^=b; b^=a; a^=b; }

// Macro to clamp a value to a maximum

#define CLAMP(v,max)    (((v) > (max)) ? (max) : (v))

// Macros to convert between integer and 32 bit fixed point format

#define FIX_1           0x10000L
#define FIX_2           0x20000L
#define FIX_HALF        0x08000L
#define TOFIX(i)        ((long)(i) << 16)
#define FIXTOINT(f)     ((int)((f) >> 16))
#define FIXROUND(f)     ((int)(((f) + FIX_HALF) >> 16))

// Macros to access the RGB components in a packed 24 bit RGB tuple

#define BYTE0(c)        (((uchar*)&(c))[0])
#define BYTE1(c)        (((uchar*)&(c))[1])
#define BYTE2(c)        (((uchar*)&(c))[2])
#define BYTE3(c)        (((uchar*)&(c))[3])

// Macros to access word values in a packed 24 bit RGB tuple

#define WORD0(c)        (((ushort*)&(c))[0])
#define WORD1(c)        ((ushort)((c) >> 8))
#define WORD2(c)        (((ushort*)&(c))[1])

// Macro to test if an X coordinate is inside the clip rectangle

#define XINSIDE(x)      ((x) >= clipLeft)

// Macro to check if an X,Y coordinate is inside the clip rectangle.
// Assume the clip rectangle is in the local variables
// (clipLeft,clipTop,clipRight,clipBottom).

#define INSIDE_CLIPRECT(x,y)        \
    ((x) >= clipLeft &&             \
     (x) < clipRight &&             \
     (y) >= clipTop &&              \
     (y) < clipBottom)

// Macros to plot a pixel for each raster operation type

#define PLOT_DSon(dest,color)   dest = ~(dest | color)
#define PLOT_DSna(dest,color)   dest = (dest & ~color)
#define PLOT_Sn(dest,color)     dest = ~color
#define PLOT_SDna(dest,color)   dest = (~dest & color)
#define PLOT_DSx(dest,color)    dest = (dest ^ color)
#define PLOT_DSan(dest,color)   dest = ~(dest & color)
#define PLOT_DSa(dest,color)    dest = (dest & color)
#define PLOT_DSxn(dest,color)   dest = ~(dest ^ color)
#define PLOT_DSno(dest,color)   dest = (dest | ~color)
#define PLOT_S(dest,color)      dest = color
#define PLOT_SDno(dest,color)   dest = (~dest | color)
#define PLOT_DSo(dest,color)    dest = (dest | color)
#define PLOT_Dn(dest)           dest = ~dest

// Macros to plot a WORD pixel for each raster operation type

#define PLOTW_DSon(dest,color)  (((ushort*)&(dest))[0]) = ~(dest | color)
#define PLOTW_DSna(dest,color)  (((ushort*)&(dest))[0]) = (dest & ~color)
#define PLOTW_Sn(dest,color)    (((ushort*)&(dest))[0]) = ~color
#define PLOTW_SDna(dest,color)  (((ushort*)&(dest))[0]) = (~dest & color)
#define PLOTW_DSx(dest,color)   (((ushort*)&(dest))[0]) = (dest ^ color)
#define PLOTW_DSan(dest,color)  (((ushort*)&(dest))[0]) = ~(dest & color)
#define PLOTW_DSa(dest,color)   (((ushort*)&(dest))[0]) = (dest & color)
#define PLOTW_DSxn(dest,color)  (((ushort*)&(dest))[0]) = ~(dest ^ color)
#define PLOTW_DSno(dest,color)  (((ushort*)&(dest))[0]) = (dest | ~color)
#define PLOTW_S(dest,color)     (((ushort*)&(dest))[0]) = color
#define PLOTW_SDno(dest,color)  (((ushort*)&(dest))[0]) = (~dest | color)
#define PLOTW_DSo(dest,color)   (((ushort*)&(dest))[0]) = (dest | color)
#define PLOTW_Dn(dest)          (((ushort*)&(dest))[0]) = ~dest

// Macros to define the scanline indices for the FatEllipseList functions

#define I_LEFTL     0
#define I_LEFTR     1
#define I_RIGHTL    2
#define I_RIGHTR    3

// Macro to compare two pixel formats structures

#define CMP_PF(pf1,pf2)                                                 \
   (((pf1).RedPosition == (pf2).RedPosition) &&                         \
    ((pf1).GreenPosition == (pf2).GreenPosition) &&                     \
    ((pf1).BluePosition == (pf2).BluePosition) &&                       \
    ((pf1).AlphaPosition == (pf2).AlphaPosition))

// Macros for packing RGB color values for different pixel formats

#define PACK_COLOR_15(R,G,B)                        \
    ( (((ulong)(R) & 0xF8) << 7)                    \
    | (((ulong)(G) & 0xF8) << 2)                    \
    | (((ulong)(B) & 0xF8) >> 3))

#define PACK_COLOR_15_ALPHA(A,R,G,B)                \
    ( (((ulong)(A) & 0x80) << 8)                    \
    | (((ulong)(R) & 0xF8) << 7)                    \
    | (((ulong)(G) & 0xF8) << 2)                    \
    | (((ulong)(B) & 0xF8) >> 3))

#define PACK_COLOR_16(R,G,B)                        \
    ( (((ulong)(R) & 0xF8) << 8)                    \
    | (((ulong)(G) & 0xFC) << 3)                    \
    | (((ulong)(B) & 0xF8) >> 3))

#define PACK_COLOR_16_ALPHA(A,R,G,B)                \
    ( (((ulong)(R) & 0xF8) << 8)                    \
    | (((ulong)(G) & 0xFC) << 3)                    \
    | (((ulong)(B) & 0xF8) >> 3))

#define UNPACK_A_15(c)  (uchar)((c & 0x8000) ? 0xFF : 0x00)
#define UNPACK_R_15(c)  (uchar)(((ulong)(c) >> 7) & 0xF8)
#define UNPACK_G_15(c)  (uchar)(((ulong)(c) >> 2) & 0xF8)
#define UNPACK_B_15(c)  (uchar)(((ulong)(c) << 3) & 0xF8)

#define UNPACK_COLOR_15(c,R,G,B)                    \
{                                                   \
 (R) = UNPACK_R_15(c);                              \
 (G) = UNPACK_G_15(c);                              \
 (B) = UNPACK_B_15(c);                              \
}

#define UNPACK_COLOR_15_ALPHA(c,A,R,G,B)            \
{                                                   \
 (A) = UNPACK_A_15(c);                              \
 (R) = UNPACK_R_15(c);                              \
 (G) = UNPACK_G_15(c);                              \
 (B) = UNPACK_B_15(c);                              \
}

#define UNPACK_A_16(c)  (uchar)(0)
#define UNPACK_R_16(c)  (uchar)(((ulong)(c) >> 8) & 0xF8)
#define UNPACK_G_16(c)  (uchar)(((ulong)(c) >> 3) & 0xFC)
#define UNPACK_B_16(c)  (uchar)(((ulong)(c) << 3) & 0xF8)

#define UNPACK_COLOR_16(c,R,G,B)                    \
{                                                   \
 (R) = UNPACK_R_16(c);                              \
 (G) = UNPACK_G_16(c);                              \
 (B) = UNPACK_B_16(c);                              \
}

#define UNPACK_COLOR_16_ALPHA(c,A,R,G,B)            \
{                                                   \
 (A) = UNPACK_A_16(c);                              \
 (R) = UNPACK_R_16(c);                              \
 (G) = UNPACK_G_16(c);                              \
 (B) = UNPACK_B_16(c);                              \
}

#define PACK_COLOR_RGB(R,G,B)                       \
    ( ((ulong)(R) << 16)                            \
    | ((ulong)(G) << 8)                             \
    | ((ulong)(B) << 0))

#define PACK_COLOR_BGR(R,G,B)                       \
    ( ((ulong)(R) << 0)                             \
    | ((ulong)(G) << 8)                             \
    | ((ulong)(B) << 16))

#define UNPACK_A_RGB(c)     (uchar)(0)
#define UNPACK_R_RGB(c)     (uchar)((ulong)(c) >> 16)
#define UNPACK_G_RGB(c)     (uchar)((ulong)(c) >> 8)
#define UNPACK_B_RGB(c)     (uchar)((ulong)(c) >> 0)

#define UNPACK_A_RGBEX(c)   (uchar)(0)
#define UNPACK_R_RGBEX(c)   ((uchar*)(c))[2]
#define UNPACK_G_RGBEX(c)   ((uchar*)(c))[1]
#define UNPACK_B_RGBEX(c)   ((uchar*)(c))[0]

#define UNPACK_COLOR_RGB(c,R,G,B)                   \
{                                                   \
 (R) = UNPACK_R_RGBEX(c);                           \
 (G) = UNPACK_G_RGBEX(c);                           \
 (B) = UNPACK_B_RGBEX(c);                           \
}

#define UNPACK_A_BGR(c)     (uchar)(0)
#define UNPACK_R_BGR(c)     (uchar)((ulong)(c) >> 0)
#define UNPACK_G_BGR(c)     (uchar)((ulong)(c) >> 8)
#define UNPACK_B_BGR(c)     (uchar)((ulong)(c) >> 16)

#define UNPACK_A_BGREX(c)   (uchar)(0)
#define UNPACK_R_BGREX(c)   ((uchar*)(c))[0]
#define UNPACK_G_BGREX(c)   ((uchar*)(c))[1]
#define UNPACK_B_BGREX(c)   ((uchar*)(c))[2]

#define UNPACK_COLOR_BGR(c,R,G,B)                   \
{                                                   \
 (R) = UNPACK_R_BGREX(c);                           \
 (G) = UNPACK_G_BGREX(c);                           \
 (B) = UNPACK_B_BGREX(c);                           \
}

#define PACK_COLOR_ARGB(R,G,B)                      \
    ( ((ulong)(R) << 16)                            \
    | ((ulong)(G) << 8)                             \
    | ((ulong)(B) << 0))

#define PACK_COLOR_ARGB_ALPHA(A,R,G,B)              \
    ( ((ulong)(A) << 24)                            \
    | ((ulong)(R) << 16)                            \
    | ((ulong)(G) << 8)                             \
    | ((ulong)(B) << 0))

#define PACK_COLOR_ABGR(R,G,B)                      \
    ( ((ulong)(R) << 0)                             \
    | ((ulong)(G) << 8)                             \
    | ((ulong)(B) << 16))

#define PACK_COLOR_ABGR_ALPHA(A,R,G,B)              \
    ( ((ulong)(A) << 24)                            \
    | ((ulong)(R) << 0)                             \
    | ((ulong)(G) << 8)                             \
    | ((ulong)(B) << 16))

#define PACK_COLOR_RGBA(R,G,B)                      \
    ( ((ulong)(R) << 24)                            \
    | ((ulong)(G) << 16)                            \
    | ((ulong)(B) << 8))

#define PACK_COLOR_RGBA_ALPHA(A,R,G,B)              \
    ( ((ulong)(A) << 0)                             \
    | ((ulong)(R) << 24)                            \
    | ((ulong)(G) << 16)                            \
    | ((ulong)(B) << 8))

#define PACK_COLOR_BGRA(R,G,B)                      \
    ( ((ulong)(R) << 8)                             \
    | ((ulong)(G) << 16)                            \
    | ((ulong)(B) << 24))

#define PACK_COLOR_BGRA_ALPHA(A,R,G,B)              \
    ( ((ulong)(A) << 0)                             \
    | ((ulong)(R) << 8)                             \
    | ((ulong)(G) << 16)                            \
    | ((ulong)(B) << 24))

#define UNPACK_A_ARGB(c)    (uchar)((ulong)(c) >> 24)
#define UNPACK_R_ARGB(c)    (uchar)((ulong)(c) >> 16)
#define UNPACK_G_ARGB(c)    (uchar)((ulong)(c) >> 8)
#define UNPACK_B_ARGB(c)    (uchar)((ulong)(c) >> 0)

#define UNPACK_COLOR_ARGB(c,R,G,B)                  \
{                                                   \
 (R) = UNPACK_R_ARGB(c);                            \
 (G) = UNPACK_G_ARGB(c);                            \
 (B) = UNPACK_B_ARGB(c);                            \
}

#define UNPACK_COLOR_ARGB_ALPHA(c,A,R,G,B)          \
{                                                   \
 (A) = UNPACK_A_ARGB(c);                            \
 (R) = UNPACK_R_ARGB(c);                            \
 (G) = UNPACK_G_ARGB(c);                            \
 (B) = UNPACK_B_ARGB(c);                            \
}

#define UNPACK_A_ABGR(c)    (uchar)((ulong)(c) >> 24)
#define UNPACK_R_ABGR(c)    (uchar)((ulong)(c) >> 0)
#define UNPACK_G_ABGR(c)    (uchar)((ulong)(c) >> 8)
#define UNPACK_B_ABGR(c)    (uchar)((ulong)(c) >> 16)

#define UNPACK_COLOR_ABGR(c,R,G,B)                  \
{                                                   \
 (R) = UNPACK_R_ABGR(c);                            \
 (G) = UNPACK_G_ABGR(c);                            \
 (B) = UNPACK_B_ABGR(c);                            \
}

#define UNPACK_COLOR_ABGR_ALPHA(c,A,R,G,B)          \
{                                                   \
 (A) = UNPACK_A_ABGR(c);                            \
 (R) = UNPACK_R_ABGR(c);                            \
 (G) = UNPACK_G_ABGR(c);                            \
 (B) = UNPACK_B_ABGR(c);                            \
}

#define UNPACK_A_RGBA(c)    (uchar)((ulong)(c) >> 0)
#define UNPACK_R_RGBA(c)    (uchar)((ulong)(c) >> 24)
#define UNPACK_G_RGBA(c)    (uchar)((ulong)(c) >> 16)
#define UNPACK_B_RGBA(c)    (uchar)((ulong)(c) >> 8)

#define UNPACK_COLOR_RGBA(c,R,G,B)                  \
{                                                   \
 (R) = UNPACK_R_RGBA(c);                            \
 (G) = UNPACK_G_RGBA(c);                            \
 (B) = UNPACK_B_RGBA(c);                            \
}

#define UNPACK_COLOR_RGBA_ALPHA(c,A,R,G,B)          \
{                                                   \
 (A) = UNPACK_A_RGBA(c);                            \
 (R) = UNPACK_R_RGBA(c);                            \
 (G) = UNPACK_G_RGBA(c);                            \
 (B) = UNPACK_B_RGBA(c);                            \
}

#define UNPACK_A_BGRA(c)    (uchar)((ulong)(c) >> 0)
#define UNPACK_R_BGRA(c)    (uchar)((ulong)(c) >> 8)
#define UNPACK_G_BGRA(c)    (uchar)((ulong)(c) >> 16)
#define UNPACK_B_BGRA(c)    (uchar)((ulong)(c) >> 24)

#define UNPACK_COLOR_BGRA(c,R,G,B)                  \
{                                                   \
 (R) = UNPACK_R_BGRA(c);                            \
 (G) = UNPACK_G_BGRA(c);                            \
 (B) = UNPACK_B_BGRA(c);                            \
}

#define UNPACK_COLOR_BGRA_ALPHA(c,A,R,G,B)          \
{                                                   \
 (A) = UNPACK_A_BGRA(c);                            \
 (R) = UNPACK_R_BGRA(c);                            \
 (G) = UNPACK_G_BGRA(c);                            \
 (B) = UNPACK_B_BGRA(c);                            \
}

// Macros to compute the pixel mask for mono bitmap functions

#define MONO_SHIFT(x)   (((x) & 0x7) ^ 0x7)
#define MONO_MASK(x)    (0x01 << MONO_SHIFT(x))
#define MONO_L_MASK(x)  (0xFF >> ((x) & 0x7))
#define MONO_R_MASK(x)  (0xFF << MONO_SHIFT(x))

// Macro to validate the return structure for QueryFunctions

#define RETURN_VALID(p) \
{ \
    COPY_STRUCTURE2(funcs,&(p),dwSize); \
    return MakeValid(((ulong*)funcs)+1,dwSize-sizeof(ulong)); \
}

// Buffer scanline blitting functions
typedef void        (*bufReadScan_t)(uchar *d,uchar *s DCL_X_ARG2,int width);
typedef void        (*bufBltScan_t)(uchar *d,uchar *s DCL_X_ARG2,int width);

#include "dither.h"
#include "regint.h"
#include "bufint.h"

#endif  /* __INTERNAL_H */

