/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for the SciTech SNAP Graphics 2D Color Inversion
*               filter driver internal functions.
*
****************************************************************************/

#ifndef __INVINT_H
#define __INVINT_H

#include "ref2dint.h"

/*---------------------- Macros and type definitions ----------------------*/

#pragma pack(4)

/****************************************************************************
REMARKS:
Structure to store the current state of the 2D portrait filter driver.

MEMBERS:
hwCtx               - Hardware device context bound to this filter driver
shared              - True if loaded into shared memory
enabled             - True if portrait mode is enabled
mi                  - Mode information block for the current hardware mode
buf                 - Current draw buffer
bytesPerPixel       - Number of bytes in a single pixel
colorMask           - Mask for active bits in color mask
fullColorMask       - Mask for active bits in a full DWORD
ref2d               - Pointer to the ref2d driver
refState            - Pointer to the ref2d driver internal state
cursor              - Copy of the ref2d cursor functions
state2d             - Copy of the ref2d state functions
draw2d              - Copy of the ref2d draw functions
bufferSize          - Size of the transfer buffer in bytes
transferStart       - Starting address for transfer buffer
VRAMBufferStart     - Start of buffer in VRAM, or 0 if in system memory
VRAMBufferStartY    - Y coordinate of the start of the VRAM buffer in memory
buffer              - Temporary buffer for rotating all bitmaps
monoBuffer          - Temporary buffer for rotating monochrome bitmaps
opt                 - Copy of current driver options
****************************************************************************/
typedef struct {
    GA_devCtx           *hwCtx;
    ibool               enabled;
    GA_modeInfo         mi;
    GA_buffer           buf;
    int                 bytesPerPixel;
    ulong               colorMask;
    ulong               fullColorMask;
    REF2D_driver        *ref2d;
    REF2D_state         *refState;
    ulong               bufferSize;
    ulong               transferStart;
    ulong               VRAMBufferStart;
    ulong               VRAMBufferStartY;
    void                *buffer;
    void                *sysBuffer;
    GA_options          opt;
    GA_globalOptions    *gOpt;
    GA_palette          invPal[256];
    GA_paletteExt       invPalExt[256];
    GA_colorCursor      colorCursor;
    GA_colorCursor256   colorCursor256;
    GA_colorCursorRGB   colorCursorRGB;
    GA_colorCursorRGBA  colorCursorRGBA;
    ibool               (NAPIP QueryFunctionsDC)(N_uint32 id,N_int32 safetyLevel,void *funcs);
    } INVERT_state;

/****************************************************************************
PARAMETERS:
left    - Left coordinate for rectangle
top     - Top coordinate for rectangle
width   - Width of the rectangle (pre-rotation)

REMARKS:
Macro to rotate a top left rectangle coordinate 90 degrees to the left
so that it becomes the top left coordinate of the portait rectangle.
****************************************************************************/
#define ROTATE(left,top,width)          \
{                                       \
    int t = top;                        \
    top = DC.buf.Height - left - width; \
    left = t;                           \
}

/****************************************************************************
PARAMETERS:
x   - X coordinate for point
y   - Y coordinate for point

REMARKS:
Macro to rotate a point coordinate 90 degrees to the left.
****************************************************************************/
#define ROTATE_PT(x,y)  \
{                       \
    int t = y;          \
    y = DC.maxY - x;    \
    x = t;              \
}

/****************************************************************************
PARAMETERS:
x   - X coordinate for point (fixed point)
y   - Y coordinate for point (fixed point)

REMARKS:
Macro to rotate a point coordinate 90 degrees to the left.
****************************************************************************/
#define ROTATE_PTFX(x,y)\
{                       \
    N_fix32 t = y;      \
    y = DC.maxYFX - x;  \
    x = t;              \
}

/****************************************************************************
REMARKS:
Round a value to the specified integer boundary
****************************************************************************/
#define ROUNDUP(value,boundary) \
    ((((value) + ((boundary)-1)) / (boundary)) * (boundary))

/****************************************************************************
REMARKS:
Macro to determine if invert mode is enabled or not.
****************************************************************************/
#define IS_INVERTED() \
    (DC.hwCtx != 0 && (DC.gOpt->bInvertColors == 1 && DC.opt.bInvertColors == 1) && DC.mi.BitsPerPixel >= 8)

/****************************************************************************
REMARKS:
Macros to set a state 2d vector if not overriden by the filter driver
****************************************************************************/
#define SET_STATE2D_VECTOR(name)    \
    if (!INVERT_state2dPT.name) INVERT_state2dPT.name = INVERT_state2dREF.name

/****************************************************************************
REMARKS:
Macros to set a draw 2d vector if not overriden by the filter driver
****************************************************************************/
#define SET_DRAW2D_VECTOR(name) \
    if (INVERT_draw2dStub.name && !INVERT_draw2dPT.name) INVERT_draw2dPT.name = INVERT_draw2dREF.name

/****************************************************************************
REMARKS:
Macros to set a state 2d vector if not overriden by the filter init
****************************************************************************/
#define SET_INIT_VECTOR(name)   \
    if (!INVERT_initPT.name) INVERT_initPT.name = INVERT_initREF.name

/****************************************************************************
REMARKS:
Macros to set a state 2d vector if not overriden by the filter driver
****************************************************************************/
#define SET_DRIVER_VECTOR(name) \
    if (!INVERT_driverPT.name) INVERT_driverPT.name = INVERT_driverREF.name

/****************************************************************************
REMARKS:
Macros to set a state 2d vector if not overriden by the filter driver
****************************************************************************/
#define SET_CURSOR_VECTOR(name) \
    if (!INVERT_cursorPT.name) INVERT_cursorPT.name = INVERT_cursorREF.name

#pragma pack()

/*---------------------------- Global Variables ---------------------------*/

extern INVERT_state     INVERT_devCtx;
extern REF2D_driver     INVERT_drv;
extern GA_cursorFuncs   INVERT_cursorREF;
extern GA_2DStateFuncs  INVERT_state2dREF;
extern GA_2DRenderFuncs INVERT_draw2dREF;
extern GA_2DRenderFuncs _VARAPI INVERT_draw2dCnt;

#undef  DC
#undef  DRV
#define DC      INVERT_devCtx
#define DC2D    (*INVERT_devCtx.refState)
#define DRV     INVERT_drv
#define CURSOR  INVERT_cursorREF
#define STATE2D INVERT_state2dREF
#define DRAW2D  INVERT_draw2dREF

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

/* Library initialisation entry points */

ibool       NAPI INVERT_QueryFunctionsDC(N_uint32 id,N_int32 safetyLevel,void *funcs);
ibool       NAPI INVERT_QueryFunctions(N_uint32 id,void *funcs);
N_int32     NAPI INVERT_SetDrawBufferSW(GA_buffer *drawBuf,void *framebuffer,N_int32 bitsPerPixel,GA_pixelFormat *pf,GA_devCtx *hwCtx,N_int32 softwareOnly);
void        NAPI INVERT_SetDrawSurface(void *surface,N_int32 xRes,N_int32 yRes,N_int32 bytesPerLine,N_int32 bitsPerPixel,GA_pixelFormat *pf);
void *      NAPI INVERT_RotateBitmap(void *src,N_int32 bitsPerPixel,N_int32 *stride,N_int32 width,N_int32 height);
void        NAPI INVERT_UnloadDriver(void);
ibool       NAPI INVERT_InitFilter(GA_modeInfo *modeInfo,N_int32 transferStart,struct _REF2D_driver *ref2d,struct _REF2D_driver **drv);
void        NAPI INVERT_DeInitFilter(void);
void        NAPI INVERT_GlobalOptionsUpdated(void);
GA_buffer * NAPI INVERT_GetDrawBuffer(void);
void        NAPI INVERT_DrawRectExtSW(void *buffer,N_int32 dstPitch,N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
N_int32     NAPI INVERT_ForceSoftwareOnly(N_int32 enable);
N_uint32    NAPI INVERT_SetColorCompareMask(N_uint32 mask);
void        NAPI INVERT_PostSwitchPhysicalResolution(void);

/* Invert init functions */

void        NAPI INVERT_SetOptions(GA_options *options);

/* Invert driver functions */

void        NAPI INVERT_SetPaletteData(GA_palette *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI INVERT_GetPaletteData(GA_palette *pal,N_int32 num,N_int32 index);
void        NAPI INVERT_SetGammaCorrectData(GA_palette *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI INVERT_GetGammaCorrectData(GA_palette *pal,N_int32 num,N_int32 index);
void        NAPI INVERT_SetPaletteDataExt(GA_paletteExt *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI INVERT_GetPaletteDataExt(GA_paletteExt *pal,N_int32 num,N_int32 index);
void        NAPI INVERT_SetGammaCorrectDataExt(GA_paletteExt *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI INVERT_GetGammaCorrectDataExt(GA_paletteExt *pal,N_int32 num,N_int32 index);

/* Invert cursor functions */

void        NAPI INVERT_SetMonoCursorColor(GA_palette *foreground,GA_palette *background);
void        NAPI INVERT_SetColorCursor(GA_colorCursor *cursor);
void        NAPI INVERT_SetColorCursor256(GA_colorCursor256 *cursor);
void        NAPI INVERT_SetColorCursorRGB(GA_colorCursorRGB *cursor);
void        NAPI INVERT_SetColorCursorRGBA(GA_colorCursorRGBA *cursor);

/* 2D state reference functions */

N_int32     NAPI INVERT_SetDrawBuffer(GA_buffer *drawBuf);
void        NAPI INVERT_SetForeColor(GA_color color);
void        NAPI INVERT_SetBackColor(GA_color color);
void        NAPI INVERT_Set8x8ColorPattern(N_int32 index,GA_colorPattern *pattern);

/* Assembler stubs for the 2D state functions */

N_int32     NAPI ISTUB_SetDrawBuffer(GA_buffer *drawBuf);
N_int32     NAPI ISTUB_IsIdle(void);
void        NAPI ISTUB_WaitTillIdle(void);
void        NAPI ISTUB_EnableDirectAccess(void);
void        NAPI ISTUB_DisableDirectAccess(void);
N_int32     NAPI ISTUB_SetMix(N_int32 mix);
void        NAPI ISTUB_SetForeColor(GA_color color);
void        NAPI ISTUB_SetBackColor(GA_color color);
void        NAPI ISTUB_Set8x8MonoPattern(N_int32 index,GA_pattern *pattern);
void        NAPI ISTUB_Use8x8MonoPattern(N_int32 index);
void        NAPI ISTUB_Use8x8TransMonoPattern(N_int32 index);
void        NAPI ISTUB_Set8x8ColorPattern(N_int32 index,GA_colorPattern *pattern);
void        NAPI ISTUB_Use8x8ColorPattern(N_int32 index);
void        NAPI ISTUB_Use8x8TransColorPattern(N_int32 index,GA_color transparent);
void        NAPI ISTUB_SetLineStipple(GA_stipple stipple);
void        NAPI ISTUB_SetLineStippleCount(N_uint32 count);
void        NAPI ISTUB_SetPlaneMask(N_uint32 mask);
void        NAPI ISTUB_SetSrcBlendFunc(N_int32 blendFunc);
void        NAPI ISTUB_SetDstBlendFunc(N_int32 blendFunc);
void        NAPI ISTUB_SetAlphaValue(N_uint8 alpha);
void        NAPI ISTUB_SetupForWindows(void);
void        NAPI ISTUB_ResetWindows(void);
void        NAPI ISTUB_SetLineStyle(N_uint32 styleMask,N_uint32 styleStep,N_uint32 styleValue);
void        NAPI ISTUB_BuildTranslateVector(GA_color *translate,GA_palette *dstPal,GA_palette *srcPal,int srcColors);
void        NAPI ISTUB_SetBlendFunc(N_int32 srcBlendFunc,N_int32 dstBlendFunc);

/* 2D generic reference functions */

void        NAPI INVERT_BitBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
void        NAPI INVERT_BitBltPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI INVERT_BitBltColorPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI INVERT_SrcTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI INVERT_SrcTransBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI INVERT_SrcTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI INVERT_SrcTransBltBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI INVERT_DstTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI INVERT_DstTransBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI INVERT_DstTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI INVERT_DstTransBltBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI INVERT_StretchBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix,N_int32 flipY);
N_int32     NAPI INVERT_ConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32     NAPI INVERT_StretchConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
void        NAPI INVERT_BitBltFx(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI INVERT_BitBltFxLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI INVERT_BitBltFxSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI INVERT_BitBltFxBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI INVERT_GetBitmapSys(void *dstAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
GA_color    NAPI INVERT_GetPixel(N_int32 x,N_int32 y);

/* Assembler stubs to call the functions via the global REF_draw2d */

void        NAPI ISTUB_DrawScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI ISTUB_DrawPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI ISTUB_DrawColorPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI ISTUB_DrawScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI ISTUB_DrawPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI ISTUB_DrawColorPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI ISTUB_DrawEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI ISTUB_DrawPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI ISTUB_DrawColorPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI ISTUB_DrawFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI ISTUB_DrawPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI ISTUB_DrawColorPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI ISTUB_DrawRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI ISTUB_DrawRectExt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI ISTUB_DrawRectLin(N_int32 dstOfs,N_int32 dstPitch,N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI ISTUB_DrawPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI ISTUB_DrawColorPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI ISTUB_DrawRectPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI ISTUB_DrawRectColorPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI ISTUB_DrawTrap(GA_trap *trap);
void        NAPI ISTUB_DrawPattTrap(GA_trap *trap);
void        NAPI ISTUB_DrawColorPattTrap(GA_trap *trap);
void        NAPI ISTUB_DrawLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void        NAPI ISTUB_DrawBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
void        NAPI ISTUB_DrawStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI ISTUB_DrawBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
void        NAPI ISTUB_PutMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI ISTUB_PutMonoImageMSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent);
void        NAPI ISTUB_PutMonoImageMSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 imagePhysAddr,N_int32 transparent);
void        NAPI ISTUB_PutMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI ISTUB_PutMonoImageLSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent);
void        NAPI ISTUB_PutMonoImageLSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 imagePhysAddr,N_int32 transparent);
void        NAPI ISTUB_ClipMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void        NAPI ISTUB_ClipMonoImageMSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void        NAPI ISTUB_ClipMonoImageMSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 imagePhysAddr,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void        NAPI ISTUB_ClipMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void        NAPI ISTUB_ClipMonoImageLSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void        NAPI ISTUB_ClipMonoImageLSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 imagePhysAddr,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void        NAPI ISTUB_BitBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI ISTUB_BitBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI ISTUB_BitBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
void        NAPI ISTUB_BitBltBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI ISTUB_BitBltPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI ISTUB_BitBltPattLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI ISTUB_BitBltPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI ISTUB_BitBltPattBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI ISTUB_BitBltColorPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI ISTUB_BitBltColorPattLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI ISTUB_BitBltColorPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI ISTUB_BitBltColorPattBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI ISTUB_SrcTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI ISTUB_SrcTransBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI ISTUB_SrcTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI ISTUB_SrcTransBltBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI ISTUB_DstTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI ISTUB_DstTransBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI ISTUB_DstTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI ISTUB_DstTransBltBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI ISTUB_StretchBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI ISTUB_StretchBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI ISTUB_StretchBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix,N_int32 flipY);
void        NAPI ISTUB_StretchBltBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
N_int32     NAPI ISTUB_ConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32     NAPI ISTUB_ConvertBltBM_Obsolete(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix);
N_int32     NAPI ISTUB_StretchConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32     NAPI ISTUB_StretchConvertBltBM_Obsolete(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix);
N_int32     NAPI ISTUB_BitBltFxTest(GA_bltFx *fx);
void        NAPI ISTUB_BitBltFx(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI ISTUB_BitBltFxLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI ISTUB_BitBltFxSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI ISTUB_BitBltFxBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI ISTUB_GetBitmapSys(void *dstAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI ISTUB_GetBitmapBM(void *dstAddr,N_int32 dstPhysAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI ISTUB_DrawEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B);
void        NAPI ISTUB_ClipEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
GA_color    NAPI ISTUB_GetPixel(N_int32 x,N_int32 y);
void        NAPI ISTUB_PutPixel(N_int32 x,N_int32 y);
void        NAPI ISTUB_BitBltPlaneMasked(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask);
void        NAPI ISTUB_BitBltPlaneMaskedLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask);
void        NAPI ISTUB_BitBltPlaneMaskedSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask,N_int32 flipY);
void        NAPI ISTUB_BitBltPlaneMaskedBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask);
void        NAPI ISTUB_DrawStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI ISTUB_DrawBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
N_int32     NAPI ISTUB_DrawClippedStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI ISTUB_DrawClippedBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __INVINT_H */
