/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser pixel
*               processing functions for the pixel processing stack
*               in the BitBltFx() function.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Pixel processing function to implement source color keying with a
single color key value.

PARAMETERS:
S   - Source pixel value
D   - Destination pixel value

RETURNS:
New destination pixel value after processing.
****************************************************************************/
GA_color REF2D_processSrcColorKey(
    GA_color S,
    GA_color D)
{
    if (S != DC.colorKeyLo)
        return S;
    return D;
}

/****************************************************************************
REMARKS:
Pixel processing function to implement source color keying with a
range of color values.

PARAMETERS:
S   - Source pixel value
D   - Destination pixel value

RETURNS:
New destination pixel value after processing.
****************************************************************************/
GA_color REF2D_processSrcColorKeyRange(
    GA_color S,
    GA_color D)
{
    if (S < DC.colorKeyLo || S > DC.colorKeyHi)
        return S;
    return D;
}

/****************************************************************************
REMARKS:
Pixel processing function to implement destination color keying with a
single color key value.

PARAMETERS:
S   - Source pixel value
D   - Destination pixel value

RETURNS:
New destination pixel value after processing.
****************************************************************************/
GA_color REF2D_processDstColorKey(
    GA_color S,
    GA_color D)
{
    if (D == DC.colorKeyLo)
        return S;
    return D;
}

/****************************************************************************
REMARKS:
Pixel processing function to implement destination color keying with a
range of color values.

PARAMETERS:
S   - Source pixel value
D   - Destination pixel value

RETURNS:
New destination pixel value after processing.
****************************************************************************/
GA_color REF2D_processDstColorKeyRange(
    GA_color S,
    GA_color D)
{
    if (DC.colorKeyLo <= D && D <= DC.colorKeyHi)
        return S;
    return D;
}

/****************************************************************************
REMARKS:
Pixel processing function to implement source color keying with a
single color key value. This version maks the pixel values against the
color key compare mask.

PARAMETERS:
S   - Source pixel value
D   - Destination pixel value

RETURNS:
New destination pixel value after processing.
****************************************************************************/
GA_color REF2D_processSrcColorKeyMasked(
    GA_color S,
    GA_color D)
{
    if ((S & DC.colorCmpMask) != DC.colorKeyLo)
        return S;
    return D;
}

/****************************************************************************
REMARKS:
Pixel processing function to implement source color keying with a
range of color values. This version maks the pixel values against the
color key compare mask.

PARAMETERS:
S   - Source pixel value
D   - Destination pixel value

RETURNS:
New destination pixel value after processing.
****************************************************************************/
GA_color REF2D_processSrcColorKeyRangeMasked(
    GA_color S,
    GA_color D)
{
    S &= DC.colorCmpMask;
    if (S < DC.colorKeyLo || S > DC.colorKeyHi)
        return S;
    return D;
}

/****************************************************************************
REMARKS:
Pixel processing function to implement destination color keying with a
single color key value. This version maks the pixel values against the
color key compare mask.

PARAMETERS:
S   - Source pixel value
D   - Destination pixel value

RETURNS:
New destination pixel value after processing.
****************************************************************************/
GA_color REF2D_processDstColorKeyMasked(
    GA_color S,
    GA_color D)
{
    if ((D & DC.colorCmpMask) == DC.colorKeyLo)
        return S;
    return D;
}

/****************************************************************************
REMARKS:
Pixel processing function to implement destination color keying with a
range of color values. This version maks the pixel values against the
color key compare mask.

PARAMETERS:
S   - Source pixel value
D   - Destination pixel value

RETURNS:
New destination pixel value after processing.
****************************************************************************/
GA_color REF2D_processDstColorKeyRangeMasked(
    GA_color S,
    GA_color D)
{
    D &= DC.colorCmpMask;
    if (DC.colorKeyLo <= D && D <= DC.colorKeyHi)
        return S;
    return D;
}

