/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for the SciTech SNAP Graphics 2D Portrait
*               Display filter driver internal functions.
*
****************************************************************************/

#ifndef __PORTINT_H
#define __PORTINT_H

#include "ref2dint.h"

/*---------------------- Macros and type definitions ----------------------*/

#pragma pack(4)

/****************************************************************************
REMARKS:
Structure to store the current state of the 2D portrait filter driver.

MEMBERS:
hwCtx               - Hardware device context bound to this filter driver
shared              - True if loaded into shared memory
enabled             - True if portrait mode is enabled
mi                  - Mode information block for the current hardware mode
buf                 - Current draw buffer
bytesPerPixel       - Number of bytes in a single pixel
ref2d               - Pointer to the ref2d driver
refState            - Pointer to the ref2d driver internal state
cursor              - Copy of the ref2d cursor functions
state2d             - Copy of the ref2d state functions
draw2d              - Copy of the ref2d draw functions
bufferSize          - Size of the transfer buffer in bytes
transferStart       - Starting address for transfer buffer
VRAMBufferStart     - Start of buffer in VRAM, or 0 if in system memory
VRAMBufferStartY    - Y coordinate of the start of the VRAM buffer in memory
buffer              - Temporary buffer for rotating all bitmaps
monoBuffer          - Temporary buffer for rotating monochrome bitmaps
****************************************************************************/
typedef struct {
    GA_devCtx           *hwCtx;
    ibool               enabled;
    GA_modeInfo         mi;
    GA_buffer           buf;
    int                 bytesPerPixel;
    int                 maxY;
    N_fix32             maxYFX;
    REF2D_driver        *ref2d;
    REF2D_state         *refState;
    ulong               bufferSize;
    ulong               transferStart;
    ulong               VRAMBufferStart;
    ulong               VRAMBufferStartY;
    void                *buffer;
    void                *sysBuffer;
    GA_options          opt;
    GA_globalOptions    *gOpt;
    int                 cursorHotX;
    int                 cursorHotY;
    GA_monoCursor       monoCursor;
    GA_colorCursor      colorCursor;
    GA_colorCursor256   colorCursor256;
    GA_colorCursorRGB   colorCursorRGB;
    GA_colorCursorRGBA  colorCursorRGBA;
    ibool               (NAPIP QueryFunctionsDC)(N_uint32 id,N_int32 safetyLevel,void *funcs);
    } PORTRAIT_state;

/****************************************************************************
PARAMETERS:
left    - Left coordinate for rectangle
top     - Top coordinate for rectangle
width   - Width of the rectangle (pre-rotation)

REMARKS:
Macro to rotate a top left rectangle coordinate 90 degrees to the left
so that it becomes the top left coordinate of the portrait rectangle.
****************************************************************************/
#define ROTATE(left,top,width)          \
{                                       \
    int t = top;                        \
    top = DC.buf.Height - left - width; \
    left = t;                           \
}

/****************************************************************************
PARAMETERS:
x   - X coordinate for point
y   - Y coordinate for point

REMARKS:
Macro to rotate a point coordinate 90 degrees to the left.
****************************************************************************/
#define ROTATE_PT(x,y)  \
{                       \
    int t = y;          \
    y = DC.maxY - x;    \
    x = t;              \
}

/****************************************************************************
PARAMETERS:
x   - X coordinate for point (fixed point)
y   - Y coordinate for point (fixed point)

REMARKS:
Macro to rotate a point coordinate 90 degrees to the left.
****************************************************************************/
#define ROTATE_PTFX(x,y)\
{                       \
    N_fix32 t = y;      \
    y = DC.maxYFX - x;  \
    x = t;              \
}

/****************************************************************************
REMARKS:
Macro to determine if invert mode is enabled or not.
****************************************************************************/
#define IS_PORTRAIT_OPT() \
    (DC.gOpt->bPortrait == 1 && DC.opt.bPortrait == 1)

/****************************************************************************
REMARKS:
Macro to determine if invert mode is enabled or not.
****************************************************************************/
#define IS_PORTRAIT() \
    (DC.hwCtx != 0 && IS_PORTRAIT_OPT() && DC.mi.BitsPerPixel >= 8)

/****************************************************************************
REMARKS:
Round a value to the specified integer boundary
****************************************************************************/
#define ROUNDUP(value,boundary) \
    ((((value) + ((boundary)-1)) / (boundary)) * (boundary))

/****************************************************************************
REMARKS:
Macros to set a state 2d vector if not overriden by the filter driver
****************************************************************************/
#define SET_STATE2D_VECTOR(name)    \
    if (!PORTRAIT_state2dPT.name) PORTRAIT_state2dPT.name = PORTRAIT_state2dREF.name

/****************************************************************************
REMARKS:
Macros to set a draw 2d vector if not overriden by the filter driver
****************************************************************************/
#define SET_DRAW2D_VECTOR(name) \
    if (PORTRAIT_draw2dStub.name && !PORTRAIT_draw2dPT.name) PORTRAIT_draw2dPT.name = PORTRAIT_draw2dREF.name

/****************************************************************************
REMARKS:
Macros to set a state 2d vector if not overriden by the filter init
****************************************************************************/
#define SET_INIT_VECTOR(name)   \
    if (!PORTRAIT_initPT.name) PORTRAIT_initPT.name = PORTRAIT_initREF.name

/****************************************************************************
REMARKS:
Macros to set a state 2d vector if not overriden by the filter driver
****************************************************************************/
#define SET_CURSOR_VECTOR(name) \
    if (!PORTRAIT_cursorPT.name) PORTRAIT_cursorPT.name = PORTRAIT_cursorREF.name

#pragma pack()

/*---------------------------- Global Variables ---------------------------*/

extern PORTRAIT_state   PORTRAIT_devCtx;
extern REF2D_driver     PORTRAIT_drv;
extern GA_cursorFuncs   PORTRAIT_cursorREF;
extern GA_2DStateFuncs  PORTRAIT_state2dREF;
extern GA_2DRenderFuncs PORTRAIT_draw2dREF;
extern GA_2DRenderFuncs _VARAPI PORTRAIT_draw2dCnt;

#undef  DC
#undef  DRV
#define DC      PORTRAIT_devCtx
#define DC2D    (*PORTRAIT_devCtx.refState)
#define DRV     PORTRAIT_drv
#define CURSOR  PORTRAIT_cursorREF
#define STATE2D PORTRAIT_state2dREF
#define DRAW2D  PORTRAIT_draw2dREF

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

/* Library initialisation entry points */

ibool       NAPI PORTRAIT_QueryFunctionsDC(N_uint32 id,N_int32 safetyLevel,void *funcs);
ibool       NAPI PORTRAIT_QueryFunctions(N_uint32 id,void *funcs);
N_int32     NAPI PORTRAIT_SetDrawBufferSW(GA_buffer *drawBuf,void *framebuffer,N_int32 bitsPerPixel,GA_pixelFormat *pf,GA_devCtx *hwCtx,N_int32 softwareOnly);
void        NAPI PORTRAIT_SetDrawSurface(void *surface,N_int32 xRes,N_int32 yRes,N_int32 bytesPerLine,N_int32 bitsPerPixel,GA_pixelFormat *pf);
void *      NAPI PORTRAIT_RotateBitmap(void *src,N_int32 bitsPerPixel,N_int32 *stride,N_int32 width,N_int32 height);
void        NAPI PORTRAIT_UnloadDriver(void);
ibool       NAPI PORTRAIT_InitFilter(GA_modeInfo *modeInfo,N_int32 transferStart,struct _REF2D_driver *ref2d,struct _REF2D_driver **drv);
void        NAPI PORTRAIT_DeInitFilter(void);
void        NAPI PORTRAIT_GlobalOptionsUpdated(void);
GA_buffer * NAPI PORTRAIT_GetDrawBuffer(void);
void        NAPI PORTRAIT_DrawRectExtSW(void *buffer,N_int32 dstPitch,N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
N_int32     NAPI PORTRAIT_ForceSoftwareOnly(N_int32 enable);
N_uint32    NAPI PORTRAIT_SetColorCompareMask(N_uint32 mask);
void        NAPI PORTRAIT_PostSwitchPhysicalResolution(void);

/* Portrait init functions */

void        NAPI PORTRAIT_SetOptions(GA_options *options);
N_int32     NAPI PORTRAIT_GetVideoModeInfo(N_uint32 mode,GA_modeInfo *modeInfo);
void        NAPI PORTRAIT_GetCurrentVideoModeInfo(GA_modeInfo *modeInfo);
N_int32     NAPI PORTRAIT_SetVideoMode(N_uint32 mode,N_int32 *virtualX,N_int32 *virtualY,N_int32 *bytesPerLine,N_int32 *maxMem,N_int32 refreshRate,GA_CRTCInfo *crtc);
N_int32     NAPI PORTRAIT_GetCustomVideoModeInfo(N_int32 xRes,N_int32 yRes,N_int32 virtualX,N_int32 virtualY,N_int32 bitsPerPixel,GA_modeInfo *modeInfo);
N_int32     NAPI PORTRAIT_SetCustomVideoMode(N_int32 xRes,N_int32 yRes,N_int32 bitsPerPixel,N_uint32 flags,N_int32 *virtualX,N_int32 *virtualY,N_int32 *bytesPerLine,N_int32 *maxMem,GA_CRTCInfo *crtc);

/* Portrait cursor functions */

void        NAPI PORTRAIT_SetMonoCursor(GA_monoCursor *cursor);
void        NAPI PORTRAIT_SetColorCursor(GA_colorCursor *cursor);
void        NAPI PORTRAIT_SetColorCursor256(GA_colorCursor256 *cursor);
void        NAPI PORTRAIT_SetColorCursorRGB(GA_colorCursorRGB *cursor);
void        NAPI PORTRAIT_SetColorCursorRGBA(GA_colorCursorRGBA *cursor);
N_int32     NAPI PORTRAIT_SetCursorPos(N_int32 x,N_int32 y);
void        NAPI PORTRAIT_BeginAccess(N_int32 left,N_int32 top,N_int32 right,N_int32 bottom);

/* 2D state reference functions */

N_int32     NAPI PORTRAIT_SetDrawBuffer(GA_buffer *drawBuf);
void        NAPI PORTRAIT_Set8x8MonoPattern(N_int32 index,GA_pattern *pattern);
void        NAPI PORTRAIT_Set8x8ColorPattern(N_int32 index,GA_colorPattern *pattern);

/* 2D generic reference functions */

void        NAPI PORTRAIT_DrawScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI PORTRAIT_DrawPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI PORTRAIT_DrawColorPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI PORTRAIT_DrawScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI PORTRAIT_DrawPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI PORTRAIT_DrawColorPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI PORTRAIT_DrawEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI PORTRAIT_DrawPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI PORTRAIT_DrawColorPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI PORTRAIT_DrawFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI PORTRAIT_DrawPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI PORTRAIT_DrawColorPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI PORTRAIT_DrawRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI PORTRAIT_DrawRectExt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI PORTRAIT_DrawPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI PORTRAIT_DrawColorPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI PORTRAIT_DrawRectPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI PORTRAIT_DrawRectColorPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI PORTRAIT_DrawTrap(GA_trap *trap);
void        NAPI PORTRAIT_DrawPattTrap(GA_trap *trap);
void        NAPI PORTRAIT_DrawColorPattTrap(GA_trap *trap);
void        NAPI PORTRAIT_DrawLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void        NAPI PORTRAIT_DrawBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
void        NAPI PORTRAIT_DrawStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI PORTRAIT_DrawBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
void        NAPI PORTRAIT_PutMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI PORTRAIT_PutMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI PORTRAIT_ClipMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI PORTRAIT_ClipMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI PORTRAIT_BitBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI PORTRAIT_BitBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
void        NAPI PORTRAIT_BitBltPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI PORTRAIT_BitBltPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI PORTRAIT_BitBltColorPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI PORTRAIT_BitBltColorPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI PORTRAIT_SrcTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI PORTRAIT_SrcTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI PORTRAIT_DstTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI PORTRAIT_DstTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI PORTRAIT_StretchBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI PORTRAIT_StretchBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix,N_int32 flipY);
N_int32     NAPI PORTRAIT_ConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32     NAPI PORTRAIT_StretchConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
void        NAPI PORTRAIT_GetBitmapSys(void *dstAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI PORTRAIT_DrawEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B);
void        NAPI PORTRAIT_ClipEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
GA_color    NAPI PORTRAIT_GetPixel(N_int32 x,N_int32 y);
void        NAPI PORTRAIT_PutPixel(N_int32 x,N_int32 y);
void        NAPI PORTRAIT_DrawStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI PORTRAIT_DrawBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
N_int32     NAPI PORTRAIT_DrawClippedStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI PORTRAIT_DrawClippedBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);

/* Assembler stubs to call the functions via the global REF_draw2d */

void        NAPI PTSTUB_DrawScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI PTSTUB_DrawPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI PTSTUB_DrawColorPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI PTSTUB_DrawScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI PTSTUB_DrawPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI PTSTUB_DrawColorPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI PTSTUB_DrawEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI PTSTUB_DrawPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI PTSTUB_DrawColorPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI PTSTUB_DrawFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI PTSTUB_DrawPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI PTSTUB_DrawColorPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI PTSTUB_DrawRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI PTSTUB_DrawRectExt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI PTSTUB_DrawPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI PTSTUB_DrawColorPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI PTSTUB_DrawRectPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI PTSTUB_DrawRectColorPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI PTSTUB_DrawTrap(GA_trap *trap);
void        NAPI PTSTUB_DrawPattTrap(GA_trap *trap);
void        NAPI PTSTUB_DrawColorPattTrap(GA_trap *trap);
void        NAPI PTSTUB_DrawLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void        NAPI PTSTUB_DrawBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
void        NAPI PTSTUB_DrawStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI PTSTUB_DrawBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
void        NAPI PTSTUB_PutMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI PTSTUB_PutMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI PTSTUB_ClipMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI PTSTUB_ClipMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI PTSTUB_BitBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI PTSTUB_BitBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
void        NAPI PTSTUB_BitBltPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI PTSTUB_BitBltPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI PTSTUB_BitBltColorPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI PTSTUB_BitBltColorPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI PTSTUB_SrcTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI PTSTUB_SrcTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI PTSTUB_DstTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI PTSTUB_DstTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI PTSTUB_StretchBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI PTSTUB_StretchBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix,N_int32 flipY);
N_int32     NAPI PTSTUB_ConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32     NAPI PTSTUB_StretchConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
void        NAPI PTSTUB_GetBitmapSys(void *dstAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI PTSTUB_DrawEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B);
void        NAPI PTSTUB_ClipEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
GA_color    NAPI PTSTUB_GetPixel(N_int32 x,N_int32 y);
void        NAPI PTSTUB_PutPixel(N_int32 x,N_int32 y);
void        NAPI PTSTUB_DrawStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI PTSTUB_DrawBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
N_int32     NAPI PTSTUB_DrawClippedStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI PTSTUB_DrawClippedBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __PORTINT_H */

