/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Include module to implement color blending
*
****************************************************************************/

/****************************************************************************
REQUIRED MACROS:
COLOR_TYPE      - Type of color variable (ie: uchar, ushort, ulong)
BLEND_PROTO     - Macro to declare the blending function prototype
BLEND_COLOR     - Name of variable to use as blend color
PACK_COLOR      - Macro to pack a color value
UNPACK_COLOR    - Macro to unpack a color value

PARAMETERS:
S   - Source pixel color value to blend
D   - Destination pixel color value to blend
C   - Constant color value to blend
A   - Constant alpha value to blend

RETURNS:
Value of the blended colors

REMARKS:
This function blends the source, destination and blend colors together
and returns the result of this blending operation. This set of functions
form the core of the generic pixel blending in the SNAP reference
rasteriser. Note that the code is highly optimised, however we don't know
if the programmer needs the destination alpha channel to be preserved or
not. For this reason we have to compute the value of the destination
alpha channel, even though it may be thrown away or not required. Considering
that we have an optimised special case for src alpha and constant alpha
blending that discards the destination alpha channel, this is not considered
a big issue (ie: programmers using the more complex blending functions
probably do want destination alpha to be preserved).
****************************************************************************/

#define UNUSED      (void)S; (void)D; (void)C; (void)A;
#define xA(c)       UNPACK_A_ARGB(c)
#define xR(c)       UNPACK_R_ARGB(c)
#define xG(c)       UNPACK_G_ARGB(c)
#define xB(c)       UNPACK_B_ARGB(c)

BLEND_PROTO(_Zero)
{
    UNUSED;
    return 0;
}

BLEND_PROTO(_One)
{
    UNUSED;
    return BLEND_COLOR;
}

BLEND_PROTO(_SrcColor)
{
    GA_color    cA = xA(S);
    GA_color    cR = xR(S);
    GA_color    cG = xG(S);
    GA_color    cB = xB(S);
    UNUSED;
    if (cA) cA++; if (cR) cR++;
    if (cG) cG++; if (cB) cB++;
    return ((((BLEND_COLOR & 0x000000FF) * cB) / 256) & 0x000000FF)
         + ((((BLEND_COLOR & 0x0000FF00) * cG) / 256) & 0x0000FF00)
         + ((((BLEND_COLOR & 0x00FF0000) * cR) / 256) & 0x00FF0000)
         + ((((BLEND_COLOR >> 24) * cA) / 256) << 24);
}

BLEND_PROTO(_OneMinusSrcColor)
{
    GA_color    cA = xA(S);
    GA_color    cR = xR(S);
    GA_color    cG = xG(S);
    GA_color    cB = xB(S);
    UNUSED;
    if (cA) cA++; if (cR) cR++;
    if (cG) cG++; if (cB) cB++;
    return ((((BLEND_COLOR & 0x000000FF) * (256-cB)) / 256) & 0x000000FF)
         + ((((BLEND_COLOR & 0x0000FF00) * (256-cG)) / 256) & 0x0000FF00)
         + ((((BLEND_COLOR & 0x00FF0000) * (256-cR)) / 256) & 0x00FF0000)
         + ((((BLEND_COLOR >> 24) * (256-cA)) / 256) << 24);
}

BLEND_PROTO(_SrcAlpha)
{
    UNUSED;
    A = xA(S);
    if (A) A++;
    return ((((BLEND_COLOR & 0x00FF00FF) * A) / 256) & 0x00FF00FF)
         + ((((BLEND_COLOR & 0xFF00FF00) >> 8) * A) & 0xFF00FF00);
}

BLEND_PROTO(_OneMinusSrcAlpha)
{
    UNUSED;
    A = xA(S);
    if (A) A++;
    return ((((BLEND_COLOR & 0x00FF00FF) * (256-A)) / 256) & 0x00FF00FF)
         + ((((BLEND_COLOR & 0xFF00FF00) >> 8) * (256-A)) & 0xFF00FF00);
}

BLEND_PROTO(_DstAlpha)
{
    UNUSED;
    A = xA(D);
    if (A) A++;
    return ((((BLEND_COLOR & 0x00FF00FF) * A) / 256) & 0x00FF00FF)
         + ((((BLEND_COLOR & 0xFF00FF00) >> 8) * A) & 0xFF00FF00);
}

BLEND_PROTO(_OneMinusDstAlpha)
{
    UNUSED;
    A = xA(D);
    if (A) A++;
    return ((((BLEND_COLOR & 0x00FF00FF) * (256-A)) / 256) & 0x00FF00FF)
         + ((((BLEND_COLOR & 0xFF00FF00) >> 8) * (256-A)) & 0xFF00FF00);
}

BLEND_PROTO(_DstColor)
{
    GA_color    cA = xA(D);
    GA_color    cR = xR(D);
    GA_color    cG = xG(D);
    GA_color    cB = xB(D);
    UNUSED;
    if (cA) cA++; if (cR) cR++;
    if (cG) cG++; if (cB) cB++;
    return ((((BLEND_COLOR & 0x000000FF) * cB) / 256) & 0x000000FF)
         + ((((BLEND_COLOR & 0x0000FF00) * cG) / 256) & 0x0000FF00)
         + ((((BLEND_COLOR & 0x00FF0000) * cR) / 256) & 0x00FF0000)
         + ((((BLEND_COLOR >> 24) * cA) / 256) << 24);
}

BLEND_PROTO(_OneMinusDstColor)
{
    GA_color    cA = xA(D);
    GA_color    cR = xR(D);
    GA_color    cG = xG(D);
    GA_color    cB = xB(D);
    UNUSED;
    if (cA) cA++; if (cR) cR++;
    if (cG) cG++; if (cB) cB++;
    return ((((BLEND_COLOR & 0x000000FF) * (256-cB)) / 256) & 0x000000FF)
         + ((((BLEND_COLOR & 0x0000FF00) * (256-cG)) / 256) & 0x0000FF00)
         + ((((BLEND_COLOR & 0x00FF0000) * (256-cR)) / 256) & 0x00FF0000)
         + ((((BLEND_COLOR >> 24) * (256-cA)) / 256) << 24);
}

BLEND_PROTO(_SrcAlphaSaturate)
{
    // TODO: Need to figure this one out!
    UNUSED;
    return 0;
}

BLEND_PROTO(_ConstantColor)
{
    GA_color    cA = xA(C);
    GA_color    cR = xR(C);
    GA_color    cG = xG(C);
    GA_color    cB = xB(C);
    UNUSED;
    if (cA) cA++; if (cR) cR++;
    if (cG) cG++; if (cB) cB++;
    return ((((BLEND_COLOR & 0x000000FF) * cB) / 256) & 0x000000FF)
         + ((((BLEND_COLOR & 0x0000FF00) * cG) / 256) & 0x0000FF00)
         + ((((BLEND_COLOR & 0x00FF0000) * cR) / 256) & 0x00FF0000)
         + ((((BLEND_COLOR >> 24) * cA) / 256) << 24);
}

BLEND_PROTO(_OneMinusConstantColor)
{
    GA_color    cA = xA(C);
    GA_color    cR = xR(C);
    GA_color    cG = xG(C);
    GA_color    cB = xB(C);
    UNUSED;
    if (cA) cA++; if (cR) cR++;
    if (cG) cG++; if (cB) cB++;
    return ((((BLEND_COLOR & 0x000000FF) * (256-cB)) / 256) & 0x000000FF)
         + ((((BLEND_COLOR & 0x0000FF00) * (256-cG)) / 256) & 0x0000FF00)
         + ((((BLEND_COLOR & 0x00FF0000) * (256-cR)) / 256) & 0x00FF0000)
         + ((((BLEND_COLOR >> 24) * (256-cA)) / 256) << 24);
}

BLEND_PROTO(_ConstantAlpha)
{
    UNUSED;
    if (A) A++;
    return ((((BLEND_COLOR & 0x00FF00FF) * A) / 256) & 0x00FF00FF)
         + ((((BLEND_COLOR & 0xFF00FF00) >> 8) * A) & 0xFF00FF00);
}

BLEND_PROTO(_OneMinusConstantAlpha)
{
    UNUSED;
    if (A) A++;
    return ((((BLEND_COLOR & 0x00FF00FF) * (256-A)) / 256) & 0x00FF00FF)
         + ((((BLEND_COLOR & 0xFF00FF00) >> 8) * (256-A)) & 0xFF00FF00);
}

#undef  BLEND_PROTO
#undef  BLEND_COLOR
#undef  UNUSED
#undef  xA
#undef  xR
#undef  xG
#undef  xB

