/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
*                   Portions Copyright (c) 1987 X Consortium
*           Portions Copyright 1987 by Digital Equipment Corporation,
*                           Maynard, Massachusetts.
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Bresenham clipped line function internals.
*
****************************************************************************/

/*------------------------- Implementation --------------------------------*/

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

RETURNS:
-1 trivial reject, 0 clipped and drawn, 1 trivial accept,

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
{
    N_int32 absDeltaX,absDeltaY,clipdx,clipdy,x2,y2;
    N_int32 new_x1,new_y1,new_x2,new_y2;
    N_int32 clip1,clip2,outcode1,outcode2;
    N_int32 count2 = count * 2;

    /* Compute values of absDeltaX and absDeltaY and (x2,y2) */
    if (flags & gaLineXMajor) {
        absDeltaY = majorInc / 2;
        absDeltaX = (majorInc - diagInc) / 2;

        /* Adjust delta's if clipping fixed point lines */
        while (absDeltaX > count2) {
            absDeltaX >>= 1;
            absDeltaY >>= 1;
            }
        }
    else {
        absDeltaX = majorInc / 2;
        absDeltaY = (majorInc - diagInc) / 2;

        /* Adjust delta's if clipping fixed point lines */
        while (absDeltaY > count2) {
            absDeltaX >>= 1;
            absDeltaY >>= 1;
            }
        }
    if (flags & gaLineXPositive)
        x2 = x1 + absDeltaX;
    else
        x2 = x1 - absDeltaX;
    if (flags & gaLineYPositive)
        y2 = y1 + absDeltaY;
    else
        y2 = y1 - absDeltaY;

    /* Pre-clip the line and check for trival accept/reject */
    outcode1 = 0;
    outcode2 = 0;
    OUTCODES(outcode1,x1,y1);
    OUTCODES(outcode2,x2,y2);
    if ((outcode1 | outcode2) == 0) {
        /* Trivial accept */
        DRAW_BRESENHAM_LINE(x1,y1,initialError,majorInc,diagInc,count,flags);
        return 1;
        }
    else if (outcode1 & outcode2) {
        /* Trivial reject */
        return -1;
        }

    /* Clip the line and reject it if nothing is to be drawn */
    clip1 = 0;
    clip2 = 0;
    new_x1 = x1;
    new_y1 = y1;
    new_x2 = x2;
    new_y2 = y2;
    if (REF2D_clipLine(clipLeft,clipTop,clipRight-1,clipBottom-1,
            &new_x1, &new_y1, &new_x2, &new_y2,
            absDeltaX, absDeltaY, &clip1, &clip2,
            flags, 0, outcode1, outcode2) == -1) {
        /* Trivial rejection */
        return -1;
        }
    if (flags & gaLineXMajor)
        count = abs(new_x2 - new_x1);
    else
        count = abs(new_y2 - new_y1);

    /* Always draw the last pixel if it has been clipped */
    if (clip2 != 0 || (flags & gaLineDoLastPel))
        count++;

    /* Now draw the clipped line */
    if (count) {
        /* Unwind bresenham error term to first point */
        if (clip1) {
            clipdx = abs(new_x1 - x1);
            clipdy = abs(new_y1 - y1);
            if (flags & gaLineXMajor)
                initialError += ((clipdy*diagInc) + ((clipdx-clipdy)*majorInc));
            else
                initialError += ((clipdx*diagInc) + ((clipdy-clipdx)*majorInc));
            }
        DRAW_BRESENHAM_LINE(new_x1,new_y1,initialError,majorInc,diagInc,count,flags);
        }
    return 0;
}

#undef DRAW_BRESENHAM_LINE
