/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement software mouse cursor functions using
*               the reference rasteriser functions.
*
****************************************************************************/

#include "ref2dint.h"
#include "clib/os/os.h"

/*--------------------------- Global variables ----------------------------*/

static GA_monoCursor        monoCurs;
static GA_colorCursor256    colorCurs;
static GA_color             colorData[4096];
static ibool                numColors;
static ibool                isCursorVisible;
static ibool                isSoftwareCursor;
static ibool                isCursorExcluded;
static ibool                disableHWCursor;
static N_int32              drawOffsetShadow;
static GA_rect              oldPos,cntPos;
static GA_color             saveBuf[CURSOR_SIZE][CURSOR_SIZE];
static GA_color             composeBuf[CURSOR_SIZE][CURSOR_SIZE];
static GA_color             tempBuf[CURSOR_SIZE][CURSOR_SIZE];

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Inititalises the software cursor module. This is called every time that a
new graphics mode is initialised.
****************************************************************************/
void REF2D_cursorInit(void)
{
    memset(&monoCurs,0,sizeof(monoCurs));
    memset(&colorCurs,0,sizeof(colorCurs));
    memset(&oldPos,0,sizeof(oldPos));
    memset(&cntPos,0,sizeof(cntPos));
    numColors = 2;
    isCursorVisible = false;
    isSoftwareCursor = true;
    isCursorExcluded = false;
    disableHWCursor = false;
    drawOffsetShadow = 0;
    if (DC.hwCtx)
        DC.cursorDrawAddr = DC.hwCtx->LinearMem;
    else
        DC.cursorDrawAddr = NULL;
    DC.cntCursorX = 0;
    DC.cntCursorY = 0;
    DC.cntHotX = 0;
    DC.cntHotY = 0;
    DC.cntCursorWidth = 0;
    DC.cntCursorHeight = 0;
}

/****************************************************************************
REMARKS:
Saves the image under the mouse cursor, so that we can display it again at
a later date.
****************************************************************************/
static void saveBuffer(
    GA_rect *pos,
    void *fbAddr,
    void *buf)
{
    int left   = MAX(pos->left,0);
    int right  = MIN(pos->right,DC.mi.xRes);
    int top    = MAX(pos->top,0);
    int bottom = MIN(pos->bottom,DC.mi.yRes);
    int adjX   = left - pos->left;
    int adjY   = top - pos->top;
    int width  = right - left;
    int height = bottom - top;
    uchar *d = LIN_Y_ADDR(adjY,buf,CUR_BUF_PITCH) + X_ADDR(adjX);
    uchar *s = LIN_Y_ADDR(top,fbAddr,DC.mi.bytesPerLine) + X_ADDR(left);

    DC.beginDirectAccess();
    while (height--) {
        DC.readScan(d,s X_ARG2(left,adjX),width);
        d += CUR_BUF_PITCH;
        s += DC.mi.bytesPerLine;
        }
    DC.endDirectAccess();
}

/****************************************************************************
REMARKS:
Restores the image under the mouse cursor, by displaying the image that we
saved previously.
****************************************************************************/
static void restoreBuffer(
    GA_rect *pos,
    void *fbAddr,
    void *buf,
    ibool updateFromShadow)
{
    int left   = MAX(pos->left,0);
    int right  = MIN(pos->right,DC.mi.xRes);
    int top    = MAX(pos->top,0);
    int bottom = MIN(pos->bottom,DC.mi.yRes);
    int adjX   = left - pos->left;
    int adjY   = top - pos->top;
    int width  = right - left;
    int height = bottom - top;
    uchar *s = LIN_Y_ADDR(adjY,buf,CUR_BUF_PITCH) + X_ADDR(adjX);
    uchar *d = LIN_Y_ADDR(top,fbAddr,DC.mi.bytesPerLine) + X_ADDR(left);

    DC.beginDirectAccess();
    while (height--) {
        DC.bltScan_S(d,s X_ARG2(adjX,left),width);
        s += CUR_BUF_PITCH;
        d += DC.mi.bytesPerLine;
        }
    DC.endDirectAccess();

    // Copy from shadow buffer if there is one
    if (DC.updateCursor && updateFromShadow)
        DC.updateCursor(drawOffsetShadow,left,top,right - left,bottom - top);
}

/****************************************************************************
REMARKS:
Copies the combined contents of the tempBuf and composeBuf buffers to the
screen, avoiding duplicate copies for the overlapping region (all
overlapping data comes from the composeBuffer).
****************************************************************************/
static void restoreCombinedBuffers(
    GA_rect *oldPos,
    GA_rect *cntPos)
{
    int     tLeft   = MAX(oldPos->left,0);
    int     tRight  = MIN(oldPos->right,DC.mi.xRes);
    int     tTop    = MAX(oldPos->top,0);
    int     tBottom = MIN(oldPos->bottom,DC.mi.yRes);
    int     cLeft   = MAX(cntPos->left,0);
    int     cRight  = MIN(cntPos->right,DC.mi.xRes);
    int     cTop    = MAX(cntPos->top,0);
    int     cBottom = MIN(cntPos->bottom,DC.mi.yRes);
    int     tSrcX   = tLeft - oldPos->left;
    int     tSrcY   = tTop - oldPos->top;
    int     tDstX   = tLeft;
    int     cSrcX   = cLeft - cntPos->left;
    int     cSrcY   = cTop - cntPos->top;
    int     cDstX   = cLeft;
    int     tWidth  = tRight - tLeft;
    int     cWidth  = cRight - cLeft;
    int     y;
    uchar   *ts = LIN_Y_ADDR(tSrcY,tempBuf,CUR_BUF_PITCH) + X_ADDR(tSrcX);
    uchar   *td = LIN_Y_ADDR(tTop,DC.cursorDrawAddr,DC.mi.bytesPerLine) + X_ADDR(tDstX);
    uchar   *cs = LIN_Y_ADDR(cSrcY,composeBuf,CUR_BUF_PITCH) + X_ADDR(cSrcX);
    uchar   *cd = LIN_Y_ADDR(cTop,DC.cursorDrawAddr,DC.mi.bytesPerLine) + X_ADDR(cDstX);

    if (DC.updateCursor) {
        // If we are running with a shaow buffer, don't bother
        // optimising overlapping region since we won't see the
        // update until the combined result is copied to the screen.
        GA_rect r;
        restoreBuffer(oldPos,DC.cursorDrawAddr,tempBuf,false);
        restoreBuffer(cntPos,DC.cursorDrawAddr,composeBuf,false);
        r.left   = MIN(tLeft,cLeft);
        r.right  = MAX(tRight,cRight);
        r.top    = MIN(tTop,cTop);
        r.bottom = MAX(tBottom,cBottom);
        DC.updateCursor(drawOffsetShadow,r.left,r.top,r.right - r.left,r.bottom - r.top);
        }
    else {
        DC.beginDirectAccess();
        if (oldPos->top < cntPos->top) {
            y = tTop;
            while (y < cTop) {
                DC.bltScan_S(td,ts X_ARG2(tSrcX,tDstX),tWidth);
                ts += CUR_BUF_PITCH;
                td += DC.mi.bytesPerLine;
                y++;
                }
            while (y < tBottom) {
                // TODO: Handle this section without any overlap!
                DC.bltScan_S(td,ts X_ARG2(tSrcX,tDstX),tWidth);
                DC.bltScan_S(cd,cs X_ARG2(cSrcX,cDstX),cWidth);
                ts += CUR_BUF_PITCH;
                cs += CUR_BUF_PITCH;
                td += DC.mi.bytesPerLine;
                cd += DC.mi.bytesPerLine;
                y++;
                }
            while (y < cBottom) {
                DC.bltScan_S(cd,cs X_ARG2(cSrcX,cDstX),cWidth);
                cs += CUR_BUF_PITCH;
                cd += DC.mi.bytesPerLine;
                y++;
                }
            }
        else {
            y = cTop;
            while (y < tTop) {
                DC.bltScan_S(cd,cs X_ARG2(cSrcX,cDstX),cWidth);
                cs += CUR_BUF_PITCH;
                cd += DC.mi.bytesPerLine;
                y++;
                }
            while (y < cBottom) {
                // TODO: Handle this section without any overlap!
                DC.bltScan_S(td,ts X_ARG2(tSrcX,tDstX),tWidth);
                DC.bltScan_S(cd,cs X_ARG2(cSrcX,cDstX),cWidth);
                ts += CUR_BUF_PITCH;
                cs += CUR_BUF_PITCH;
                td += DC.mi.bytesPerLine;
                cd += DC.mi.bytesPerLine;
                y++;
                }
            while (y < tBottom) {
                DC.bltScan_S(td,ts X_ARG2(tSrcX,tDstX),tWidth);
                ts += CUR_BUF_PITCH;
                td += DC.mi.bytesPerLine;
                y++;
                }
            }
        DC.endDirectAccess();
        }
}

/****************************************************************************
REMARKS:
Draws the monochrome mouse cursor image.
****************************************************************************/
static void drawMonoCursor(void)
{
    uchar   *d,*i,*m;
    int     height = DC.cntCursorHeight;

    // Combine AND, XOR mask and data from save buffer into compose buffer
    d = (uchar*)composeBuf;
    m = (uchar*)&monoCurs.ANDMask[0];
    i = (uchar*)&monoCurs.XORMask[0];
    while (height--) {
        DC.drawMonoCursor(d,m,i X_ARG(0),8);
        m += 8;
        i += 8;
        d += CUR_BUF_PITCH;
        }
}

/****************************************************************************
REMARKS:
Draws the colorCurs mouse cursor image.
****************************************************************************/
static void drawColorCursor(void)
{
    uchar   *d,*i,*m;
    int     height = DC.cntCursorHeight;
    int     pitch = X_ADDR(64);

    // Combine AND, cursor data and data from save buffer into compose buffer
    d = (uchar*)composeBuf;
    m = (uchar*)&colorCurs.ANDMask[0];
    i = (uchar*)&colorData[0];
    while (height--) {
        DC.drawColorCursor(d,m,i X_ARG(0),8);
        m += 8;
        i += pitch;
        d += CUR_BUF_PITCH;
        }
}

/****************************************************************************
REMARKS:
Updates the cursor image on the screen. We do this in such a way that the
we only draw the final changes to the screen, to eliminate all flicker
during the cursor update process.
****************************************************************************/
static void updateCursor(void)
{
    int     left   = MAX(oldPos.left,cntPos.left);
    int     right  = MIN(oldPos.right,cntPos.right);
    int     top    = MAX(oldPos.top,cntPos.top);
    int     bottom = MIN(oldPos.bottom,cntPos.bottom);
    int     width  = right - left;
    int     height = bottom - top;
    int     srcX   = left - oldPos.left;
    int     srcY   = top - oldPos.top;
    int     dstX   = left - cntPos.left;
    int     dstY   = top - cntPos.top;
    uchar   *s,*d;

    // Save the current screen buffer into our temporary buffer
    rep_movsd(tempBuf,saveBuf,sizeof(saveBuf)/4);
    saveBuffer(&cntPos,DC.cursorDrawAddr,saveBuf);

    // Copy intersection of tempBuf and saveBuf into saveBuf
    if (width > 0 && height > 0) {
        int count = height;
        s = LIN_Y_ADDR(srcY,tempBuf,CUR_BUF_PITCH) + X_ADDR(srcX);
        d = LIN_Y_ADDR(dstY,saveBuf,CUR_BUF_PITCH) + X_ADDR(dstX);
        while (count--) {
            DC.bltScan_S(d,s X_ARG2(srcX,dstX),width);
            s += CUR_BUF_PITCH;
            d += CUR_BUF_PITCH;
            }
        }

    // Composite the cursor into the compose buffer
    rep_movsd(composeBuf,saveBuf,sizeof(saveBuf)/4);
    if (numColors > 2)
        drawColorCursor();
    else
        drawMonoCursor();

    // Restore the result to the screen
    if (width > 0 && height > 0)
        restoreCombinedBuffers(&oldPos,&cntPos);
    else {
        restoreBuffer(&oldPos,DC.cursorDrawAddr,tempBuf,true);
        restoreBuffer(&cntPos,DC.cursorDrawAddr,composeBuf,true);
        }
}

/****************************************************************************
REMARKS:
Hides the cursor image at the current location.
****************************************************************************/
static void hideCursor(void)
{
    restoreBuffer(&cntPos,DC.cursorDrawAddr,saveBuf,true);
}

/****************************************************************************
REMARKS:
Shows the cursor image at the current location. Assumes the cursor is not
currently visible when this function is called.
****************************************************************************/
static void drawCursor(void)
{
    saveBuffer(&cntPos,DC.cursorDrawAddr,saveBuf);
    rep_movsd(composeBuf,saveBuf,sizeof(saveBuf)/4);
    if (numColors > 2)
        drawColorCursor();
    else
        drawMonoCursor();
    restoreBuffer(&cntPos,DC.cursorDrawAddr,composeBuf,true);
}

/****************************************************************************
REMARKS:
Update the current position internally based on (x,y) coord and mouse cursor.
****************************************************************************/
static void updateCntPos(void)
{
    oldPos = cntPos;
    cntPos.left   = DC.cntCursorX - DC.cntHotX;
    cntPos.right  = cntPos.left + DC.cntCursorWidth;
    cntPos.top    = DC.cntCursorY - DC.cntHotY;
    cntPos.bottom = cntPos.top + DC.cntCursorHeight;
}

/****************************************************************************
REMARKS:
Switches to hardware cursor rendering from software cursor rendering.
****************************************************************************/
static void switchToHardware(
    ibool redraw)
{
    if (isSoftwareCursor && isCursorVisible) {
        DC.hwCursor.SetCursorPos(DC.cntCursorX,DC.cntCursorY);
        DC.hwCursor.ShowCursor(true);
        if (redraw)
            hideCursor();
        }
    isSoftwareCursor = false;
}

/****************************************************************************
REMARKS:
Switches to software cursor rendering from hardware cursor rendering.
****************************************************************************/
static void switchToSoftware(
    ibool redraw)
{
    if (!isSoftwareCursor && isCursorVisible) {
        saveBuffer(&cntPos,DC.cursorDrawAddr,saveBuf);
        DC.hwCursor.ShowCursor(false);
        }
    isSoftwareCursor = true;
    if (isCursorVisible) {
        updateCntPos();
        if (redraw)
            updateCursor();
        else
            saveBuffer(&cntPos,DC.cursorDrawAddr,saveBuf);
        }
}

/****************************************************************************
REMARKS:
Measures the dimensions of a color bitmap. The width is returned in bytes.
****************************************************************************/
static void measureBitmapSize(
    uchar *bmp,
    int *pWidth,
    int *pHeight,
    int pitch)
{
    int x,y,width,height;

    width = height = 0;
    bmp = &bmp[64 * pitch - 1];
    for (y = 64; y > 0; y--) {
        uchar byte = 0;
        for (x = pitch; x > 0; x--) {
            if ((byte |= *bmp--) != 0)
                width = MAX(width,x);
            }
        if (byte != 0)
            height = MAX(height,y);
        }
    *pWidth = width;
    *pHeight = height;
}

/****************************************************************************
PARAMETERS:
offset  - Offset to start of display memory to make visible
waitVRT - Wait for retrace flag

REMARKS:
This function sets the currently visible hardware display start address as
a byte offset from the start of physical display memory. We hook this so
we can keep track of where we are supposed to be displaying the mouse
cursor.

NOTE: This code is intended to handle double buffered page flipping with a
      flicker free mouse cursor. It will not work if used for hardware
      panning, but the should be using SetDisplayStartXY() in that case
      anyway!
****************************************************************************/
void NAPI REF2D_SetDisplayStart(
    N_int32 offset,
    N_int32 waitVRT)
{
    void *oldFrameBufferAddr = DC.cursorDrawAddr;

    DC.cursorDrawAddr = DC.frameBuffer + offset;
    if (DC.updateCursor) {
        drawOffsetShadow = offset;
        DC.cursorDrawAddr = DC.frameBuffer;
        }
    if (isCursorVisible && isSoftwareCursor)
        drawCursor();
    DC.hwDrv.SetDisplayStart(offset,waitVRT);
    if (isCursorVisible && isSoftwareCursor)
        restoreBuffer(&cntPos,oldFrameBufferAddr,saveBuf,true);
}

/****************************************************************************
PARAMETERS:
x       - X coordinate of first pixel to display
y       - Y coordinate of first pixel to display
waitVRT - Wait for retrace flag

REMARKS:
This function sets the currently visible hardware display start address as
an (x,y) coordinate offset from the start of physical display memory. We
hook this so we can keep track of where we are supposed to be displaying
the mouse cursor.
****************************************************************************/
void NAPI REF2D_SetDisplayStartXY(
    N_int32 x,
    N_int32 y,
    N_int32 waitVRT)
{
    if (isCursorVisible && isSoftwareCursor)
        hideCursor();
    DC.hwDrv.SetDisplayStartXY(x,y,waitVRT);
    DC.cursorDrawAddr = LIN_Y_ADDR(y,DC.frameBuffer,DC.mi.bytesPerLine) + X_ADDR(x);
    if (DC.updateCursor) {
        drawOffsetShadow = (y * DC.mi.bytesPerLine) + X_ADDR(x);
        DC.cursorDrawAddr = DC.frameBuffer;
        }
    if (isCursorVisible && isSoftwareCursor)
        drawCursor();
}

/****************************************************************************
PARAMETERS:
buf     - Buffer to flip to
waitVRT - Wait for retrace flag

REMARKS:
This function makes the passed in buffer visible for the active display. We
hook this function here so that we can handle double buffer mouse cursors
correctly with the buffer manage being active, which will work even with
the DirectDraw driver version of the buffer manager.
****************************************************************************/
void NAPI REF2D_FlipToBuffer(
    GA_buf *buf,
    N_int32 waitVRT)
{
    void *oldFrameBufferAddr = DC.cursorDrawAddr;

    DC.BUF_LockBuffer(buf);
    DC.cursorDrawAddr = buf->Surface;
    if (DC.updateCursor) {
        drawOffsetShadow = buf->Offset;
        DC.cursorDrawAddr = DC.frameBuffer;
        }
    if (isCursorVisible && isSoftwareCursor)
        drawCursor();
    DC.BUF_UnlockBuffer(buf);
    DC.BUF_FlipToBufferOrg(buf,waitVRT);
    DC.BUF_LockBuffer(buf);
    if (isCursorVisible && isSoftwareCursor)
        restoreBuffer(&cntPos,oldFrameBufferAddr,saveBuf,true);
    DC.BUF_UnlockBuffer(buf);
}

#ifdef  SUPPORT_8BPP
/****************************************************************************
REMARKS:
Find the closest colorCurs in the current palette to the passed in colorCurs value.
****************************************************************************/
static GA_color findClosest(
    GA_palette *colorCurs)
{
    GA_palette  *p = DC.shadowPal;
    int         i,closest = 0,d,distance = 0x7FFFL;

    for (i = 0; i < 256; i++,p++) {
        if ((d = ABS(colorCurs->Red - (int)p->Red)
               + ABS(colorCurs->Green - (int)p->Green)
               + ABS(colorCurs->Blue - (int)p->Blue)) < distance) {
            distance = d;
            closest = i;
            }
        }
    return closest;
}
#endif

#if defined(SUPPORT_4BPP) || defined(SUPPORT_8BPP)
/****************************************************************************
REMARKS:
Maps the color cursor values based on the new hardware palette.
****************************************************************************/
static void mapColorCursorPalette(void)
{
    GA_color    *translate = DC.buildTranslateVector(DC.translateVec,DC.shadowPal,colorCurs.Palette,numColors);
    uchar       *s = (uchar*)colorCurs.ColorData;
    uchar       *d = (uchar*)colorData;
    int         height = 64;

#ifdef SUPPORT_4BPP
    if (DC.mi.bitsPerPixel == 4) {
        if (numColors > 16) {
            while (height--) {
                DC.convScan8_S(d,s X_ARG(0),64,translate);
                s += 64;
                d += 32;
                }
            }
        else {
            while (height--) {
                DC.convScan4_S(d,s X_ARG(0),64,0,translate);
                s += 32;
                d += 32;
                }
            }
        }
#endif
#ifdef SUPPORT_8BPP
    if (DC.mi.bitsPerPixel == 8) {
        if (numColors > 16) {
            while (height--) {
                DC.convScan8_S(d,s X_ARG(0),64,translate);
                s += 64;
                d += 64;
                }
            }
        else {
            while (height--) {
                DC.convScan4_S(d,s X_ARG(0),64,0,translate);
                s += 32;
                d += 64;
                }
            }
        }
#endif
}

/****************************************************************************
PARAMETERS:
pal     - Pointer to the palette data to program
num     - Number of palette entries to program
index   - Index of first entry to program
waitVRT - Wait for vertical retrace flag

REMARKS:
This function programs the colorCurs palette information for the current graphics
mode, and is only valid in 8-bpp and lower colorCurs index modes. We hook this
so we can keep track of the colorCurs palette for the mouse cursor.
****************************************************************************/
void NAPI REF2D_SetPaletteData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
#ifdef SUPPORT_8BPP
    GA_palette fore = DC.shadowPal[DC.cursorForeColor];
    GA_palette back = DC.shadowPal[DC.cursorBackColor];
#endif

    DC.hwDrv.SetPaletteData(pal,num,index,waitVRT);
    if (DC.mi.bitsPerPixel <= 8) {
        memcpy(&DC.shadowPal[index],pal,num * sizeof(*pal));
        if (numColors > 2)
            mapColorCursorPalette();
#ifdef SUPPORT_8BPP
        else if (DC.gaMode.Attributes & gaHave8bppRGBCursor) {
            DC.cursorForeColor = findClosest(&fore);
            DC.cursorBackColor = findClosest(&back);
            }
#endif
        }
}

/****************************************************************************
PARAMETERS:
pal     - Pointer to the palette data to program
num     - Number of palette entries to program
index   - Index of first entry to program
waitVRT - Wait for vertical retrace flag

REMARKS:
This function programs the colorCurs palette information for the current graphics
mode, and is only valid in 8-bpp and lower colorCurs index modes. We hook this
so we can keep track of the colorCurs palette for the mouse cursor.
****************************************************************************/
void NAPI REF2D_SetPaletteDataExt(
    GA_paletteExt *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
#ifdef SUPPORT_8BPP
    GA_palette  fore = DC.shadowPal[DC.cursorForeColor];
    GA_palette  back = DC.shadowPal[DC.cursorBackColor];
#endif
    int         i;

    DC.hwDrv.SetPaletteDataExt(pal,num,index,waitVRT);
    if (DC.mi.bitsPerPixel <= 8) {
        for (i = 0; i < num; i++) {
            DC.shadowPal[index+i].Red = pal->Red >> 8;
            DC.shadowPal[index+i].Green = pal->Green >> 8;
            DC.shadowPal[index+i].Blue = pal->Blue >> 8;
            }
        if (numColors > 2)
            mapColorCursorPalette();
#ifdef SUPPORT_8BPP
        else if (DC.gaMode.Attributes & gaHave8bppRGBCursor) {
            DC.cursorForeColor = findClosest(&fore);
            DC.cursorBackColor = findClosest(&back);
            }
#endif
        }
}
#endif

/****************************************************************************
PARAMETERS:
foreground  - Foreground colorCurs for the cursor
background  - Background colorCurs for the cursor

REMARKS:
This function sets both the foreground and background colors for the
hardware cursor, which as passed in as GA_palette structures. In 8bpp
display modes the colorCurs index of the cursor colorCurs is passed in the
red member of GA_palette structure. The exception to this is for 8bpp
display modes that have the gaHave8bppRGBCursor flag set in the GA_modeInfo
Attributes field. When this is set, the full RGB colorCurs of the cursor is
passed in the red, green and blue members of the GA_palette structue.

For 15-bpp and above RGB modes, the full RGB colorCurs of the cursor is always
passed in the red, green and blue members of the GA_palette structure.
****************************************************************************/
void NAPI REF2D_SetMonoCursorColor(
    GA_palette *foreground,
    GA_palette *background)
{
    switch (DC.mi.bitsPerPixel) {
#ifdef  SUPPORT_1BPP
        case 1:
            DC.cursorForeColor = foreground->Red;
            DC.cursorBackColor = background->Red;
            break;
#endif
#ifdef  SUPPORT_4BPP
        case 4:
            DC.cursorForeColor = foreground->Red;
            DC.cursorBackColor = background->Red;
            break;
#endif
#ifdef  SUPPORT_8BPP
        case 8:
            if (DC.gaMode.Attributes & gaHave8bppRGBCursor) {
                DC.cursorForeColor = findClosest(foreground);
                DC.cursorBackColor = findClosest(background);
                }
            else {
                DC.cursorForeColor = foreground->Red;
                DC.cursorBackColor = background->Red;
                }
            break;
#endif
#ifdef  SUPPORT_16BPP
        case 15:
            DC.cursorForeColor = PACK_COLOR_15(foreground->Red,foreground->Green,foreground->Blue);
            DC.cursorBackColor = PACK_COLOR_15(background->Red,background->Green,background->Blue);
            break;
        case 16:
            DC.cursorForeColor = PACK_COLOR_16(foreground->Red,foreground->Green,foreground->Blue);
            DC.cursorBackColor = PACK_COLOR_16(background->Red,background->Green,background->Blue);
            break;
#endif
#if defined(SUPPORT_24BPP)
        case 24:
            if (DC.mi.pf.RedPosition == 0) {
                DC.cursorForeColor = PACK_COLOR_BGR(foreground->Red,foreground->Green,foreground->Blue);
                DC.cursorBackColor = PACK_COLOR_BGR(background->Red,background->Green,background->Blue);
                }
            else {
                DC.cursorForeColor = PACK_COLOR_RGB(foreground->Red,foreground->Green,foreground->Blue);
                DC.cursorBackColor = PACK_COLOR_RGB(background->Red,background->Green,background->Blue);
                }
            break;
#endif
#if defined(SUPPORT_32BPP)
        case 32:
            if (DC.mi.pf.RedPosition == 0) {
                DC.cursorForeColor = PACK_COLOR_ABGR(foreground->Red,foreground->Green,foreground->Blue);
                DC.cursorBackColor = PACK_COLOR_ABGR(background->Red,background->Green,background->Blue);
                }
            else if (DC.mi.pf.RedPosition == 8) {
                DC.cursorForeColor = PACK_COLOR_BGRA(foreground->Red,foreground->Green,foreground->Blue);
                DC.cursorBackColor = PACK_COLOR_BGRA(background->Red,background->Green,background->Blue);
                }
            else if (DC.mi.pf.RedPosition == 16) {
                DC.cursorForeColor = PACK_COLOR_ARGB(foreground->Red,foreground->Green,foreground->Blue);
                DC.cursorBackColor = PACK_COLOR_ARGB(background->Red,background->Green,background->Blue);
                }
            else {
                DC.cursorForeColor = PACK_COLOR_RGBA(foreground->Red,foreground->Green,foreground->Blue);
                DC.cursorBackColor = PACK_COLOR_RGBA(background->Red,background->Green,background->Blue);
                }
            break;
#endif
        }
    if (DC.hwCursor.SetMonoCursorColor)
        DC.hwCursor.SetMonoCursorColor(foreground,background);
    if (isSoftwareCursor && isCursorVisible) {
        oldPos = cntPos;
        updateCursor();
        }
}

/****************************************************************************
PARAMETERS:
cursor      - Pointer to an GA_monoCursor structure for the cursor image

REMARKS:
This function downloads the specified monochrome cursor definition from the
application into the hardware cursor. The cursor data is passed by the
application in the GA_monoCursor structure.
****************************************************************************/
void NAPI REF2D_SetMonoCursor(
    GA_monoCursor *cursor)
{
    int     ANDWidth,XORWidth;
    int     ANDHeight,XORHeight;

    numColors = 2;
    monoCurs = *cursor;
    DC.cntHotX = monoCurs.HotX;
    DC.cntHotY = monoCurs.HotY;
    measureBitmapSize(monoCurs.ANDMask,&ANDWidth,&ANDHeight,8);
    measureBitmapSize(monoCurs.XORMask,&XORWidth,&XORHeight,8);
    DC.cntCursorWidth = MAX(ANDWidth,XORWidth) * 8;
    DC.cntCursorHeight = MAX(ANDHeight,XORHeight);

    if (isSoftwareCursor && isCursorVisible)
        hideCursor();
    if (DC.hwCursor.SetMonoCursor && DC.cntCursorWidth <= DC.maxCursorWidth && DC.cntCursorHeight <= DC.maxCursorHeight) {
        switchToHardware(false);
        DC.hwCursor.SetMonoCursor(cursor);
        disableHWCursor = false;
        }
    else {
        switchToSoftware(false);
        disableHWCursor = true;
        }
    if (isSoftwareCursor) {
        if (isCursorVisible)
            drawCursor();
        else
            updateCntPos();
        }
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursor structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor. The cursor data is passed by the
application in the GA_colorCursor format.
****************************************************************************/
void NAPI REF2D_SetColorCursor(
    GA_colorCursor *cursor)
{
    int     ANDWidth,colorWidth;
    int     ANDHeight,colorHeight;

    // TODO: Up convert to 256 or RGB color cursors if 16-color is not
    //       supported by the hardware...

    numColors = 16;
    memcpy(colorCurs.ColorData,cursor->ColorData,sizeof(cursor->ColorData));
    memcpy(colorCurs.ANDMask,cursor->ANDMask,sizeof(cursor->ANDMask));
    memcpy(colorCurs.Palette,cursor->Palette,sizeof(cursor->Palette));
    DC.cntHotX = cursor->HotX;
    DC.cntHotY = cursor->HotY;
    measureBitmapSize(colorCurs.ANDMask,&ANDWidth,&ANDHeight,8);
    measureBitmapSize(colorCurs.ColorData,&colorWidth,&colorHeight,32);
    DC.cntCursorWidth = MAX(ANDWidth * 8,colorWidth * 2);
    DC.cntCursorHeight = MAX(ANDHeight,colorHeight);
    if (DC.mi.bitsPerPixel <= 8) {
#if defined(SUPPORT_4BPP) || defined(SUPPORT_8BPP)
        mapColorCursorPalette();
#endif
        }
    else {
#if defined(SUPPORT_16BPP) || defined(SUPPORT_24BPP) || defined(SUPPORT_32BPP)
        GA_color    *translate = DC.buildTranslateVector(DC.translateVec,DC.shadowPal,colorCurs.Palette,16);
        uchar       *s = (uchar*)colorCurs.ColorData;
        uchar       *d = (uchar*)colorData;
        int         height = 64;
        int         dstPitch = X_ADDR(64);
        while (height--) {
            DC.convScan4_S(d,s X_ARG(0),64,0,translate);
            s += 32;
            d += dstPitch;
            }
#endif
        }
    if (isSoftwareCursor && isCursorVisible)
        hideCursor();
    if (DC.hwCursor.SetColorCursor && DC.cntCursorWidth <= DC.maxCursorWidth && DC.cntCursorHeight <= DC.maxCursorHeight) {
        switchToHardware(false);
        DC.hwCursor.SetColorCursor(cursor);
        disableHWCursor = false;
        }
    else {
        switchToSoftware(false);
        disableHWCursor = true;
        }
    if (isSoftwareCursor) {
        if (isCursorVisible)
            drawCursor();
        else
            updateCntPos();
        }
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursor256 structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor. The cursor data is passed by the
application in the GA_colorCursor256 format.
****************************************************************************/
void NAPI REF2D_SetColorCursor256(
    GA_colorCursor256 *cursor)
{
    int     ANDWidth,colorWidth;
    int     ANDHeight,colorHeight;

    // TODO: Up convert to RGB color cursors if 256-color is not
    //       supported by the hardware...

    // TODO: Down convert to 16-color cursor if 256-color is not supported
    //       and cursor image has less than 16 distinct colors

    numColors = 256;
    memcpy(colorCurs.ColorData,cursor->ColorData,sizeof(cursor->ColorData));
    memcpy(colorCurs.ANDMask,cursor->ANDMask,sizeof(cursor->ANDMask));
    memcpy(colorCurs.Palette,cursor->Palette,sizeof(cursor->Palette));
    DC.cntHotX = cursor->HotX;
    DC.cntHotY = cursor->HotY;
    measureBitmapSize(colorCurs.ANDMask,&ANDWidth,&ANDHeight,8);
    measureBitmapSize(colorCurs.ColorData,&colorWidth,&colorHeight,64);
    DC.cntCursorWidth = MAX(ANDWidth * 8,colorWidth);
    DC.cntCursorHeight = MAX(ANDHeight,colorHeight);
    if (DC.mi.bitsPerPixel <= 8) {
#if defined(SUPPORT_4BPP) || defined(SUPPORT_8BPP)
        mapColorCursorPalette();
#endif
        }
    else {
#if defined(SUPPORT_16BPP) || defined(SUPPORT_24BPP) || defined(SUPPORT_32BPP)
        GA_color    *translate = DC.buildTranslateVector(DC.translateVec,DC.shadowPal,colorCurs.Palette,256);
        uchar       *s = (uchar*)colorCurs.ColorData;
        uchar       *d = (uchar*)colorData;
        int         height = 64;
        int         dstPitch = X_ADDR(64);
        while (height--) {
            DC.convScan8_S(d,s X_ARG(0),64,translate);
            s += 64;
            d += dstPitch;
            }
#endif
        }
    if (isSoftwareCursor && isCursorVisible)
        hideCursor();
    if (DC.hwCursor.SetColorCursor256 && DC.cntCursorWidth <= DC.maxCursorWidth && DC.cntCursorHeight <= DC.maxCursorHeight) {
        switchToHardware(false);
        DC.hwCursor.SetColorCursor256(cursor);
        disableHWCursor = false;
        }
    else {
        switchToSoftware(false);
        disableHWCursor = true;
        }
    if (isSoftwareCursor) {
        if (isCursorVisible)
            drawCursor();
        else
            updateCntPos();
        }
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursorRGB structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor. The cursor data is passed by the
application in the GA_colorCursorRGB format.
****************************************************************************/
void NAPI REF2D_SetColorCursorRGB(
    GA_colorCursorRGB *cursor)
{
    // TODO!!

    // TODO: Down convert to 16-color cursor if 256-color or RGB is not
    //       supported and cursor image has less than 16 distinct colors

    // TODO: Down convert to 256-color cursor if RGB is not
    //       supported and cursor image has less than 256 distinct colors

    (void)cursor;
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursorRGBA structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor. The cursor data is passed by the
application in the GA_colorCursorRGBA format.
****************************************************************************/
void NAPI REF2D_SetColorCursorRGBA(
    GA_colorCursorRGBA *cursor)
{
    // TODO!!
    (void)cursor;
}

/****************************************************************************
PARAMETERS:
x   - X coordinate of the cursor position
y   - Y coordinate of the cursor position

REMARKS:
This function sets the location of the hardware cursor in display
coordinates. This function takes the X and Y coordinates of the new cursor
location. This function will place the cursor so that the hotspot of the
cursor image is located at the specified (X,Y) location, and will correctly
handle special cases where the cursor definition needs to be located off the
edges of the display screen (such as when X=0, Y=0 and HotX,HotY > 0).
****************************************************************************/
N_int32 NAPI REF2D_SetCursorPos(
    N_int32 x,
    N_int32 y)
{
    DC.cntCursorX = x;
    DC.cntCursorY = y;

    if (numColors == 256) {
        if (DC.hwCursor.SetColorCursor256) {
            DC.hwCursor.SetCursorPos(x,y);
            if (disableHWCursor)
                switchToSoftware(true);
            else if (isSoftwareCursor)
                switchToHardware(true);
            }
        updateCntPos();
        if (isCursorVisible && isSoftwareCursor && !isCursorExcluded)
            updateCursor();
        }
    else if (numColors == 16) {
        if (DC.hwCursor.SetColorCursor) {
            DC.hwCursor.SetCursorPos(x,y);
            if (disableHWCursor)
                switchToSoftware(true);
            else if (isSoftwareCursor)
                switchToHardware(true);
            }
        updateCntPos();
        if (isCursorVisible && isSoftwareCursor && !isCursorExcluded)
            updateCursor();
        }
    else {
        if (DC.hwCursor.SetMonoCursor) {
            DC.hwCursor.SetCursorPos(x,y);
            if (disableHWCursor)
                switchToSoftware(true);
            else if (isSoftwareCursor)
                switchToHardware(true);
            }
        updateCntPos();
        if (isCursorVisible && isSoftwareCursor && !isCursorExcluded)
            updateCursor();
        }
    return true;
}

/****************************************************************************
PARAMETERS:
visible - 1 to show the cursor, 0 to hide the cursor

REMARKS:
This function unconditionally either show or hides the hardware cursor.

Note: If this function gets called with a visible flag of -1, this tells
the software rendering code to force the cursor to a non-visible state,
but not actually touch any of the graphics memory. This occurs when a
program is Alt-Tab'ed away from under DirectX, and all the surfaces are
lost. We need to hide the cursor, but we cannot touch the graphics memory
as the surfaces have already been destroyed by DirectX.
****************************************************************************/
void NAPI REF2D_ShowCursor(
    N_int32 visible)
{
    if (visible != isCursorVisible) {
        if (visible == -1)
            isCursorVisible = false;
        else {
            isCursorVisible = visible;
            if (!isSoftwareCursor)
                DC.hwCursor.ShowCursor(visible);
            else if (!isCursorExcluded) {
                if (visible)
                    drawCursor();
                else
                    hideCursor();
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of display affected
top     - Top coordinate of display affected
right   - Right coordinate of display affected (exclusive)
bottom  - Bottom coordinate of display affected (exclusive)

REMARKS:
This function is used when the SNAP driver is using software monoCurs and
colorCurs cursor acceleration. The cursor images are stored in offscreen
memory buffers, and the hardware blitter is used to draw the cursor images.
This function is used to inform the SNAP driver that the shell display
driver is about to draw to a particular location of memory, and to remove
the cursor image from the screen temporarily while the drawing is taking
place if the drawing region and the cursor position overlap.
****************************************************************************/
void NAPI REF2D_BeginAccess(
    N_int32 left,
    N_int32 top,
    N_int32 right,
    N_int32 bottom)
{
    if (isCursorVisible && isSoftwareCursor && !isCursorExcluded) {
        left   = MAX(left,cntPos.left);
        right  = MIN(right,cntPos.right);
        top    = MAX(top,cntPos.top);
        bottom = MIN(bottom,cntPos.bottom);
        if (bottom > top && right > left) {
            hideCursor();
            isCursorExcluded = true;
            }
        }
}

/****************************************************************************
REMARKS:
This function is used when the SNAP driver is using software monoCurs and
colorCurs cursor acceleration. This function informs the display driver that
drawing has completed, and that it may restore the cursor image back
to the display screen at it's original location (if it needed to be hidden).
****************************************************************************/
void NAPI REF2D_EndAccess(void)
{
    if (isCursorVisible && isSoftwareCursor && isCursorExcluded)
        drawCursor();
}

/****************************************************************************
REMARKS:
This function return true if the cursor is currently running in hardware,
or false if it is currently running in software.
****************************************************************************/
N_int32 NAPI REF2D_IsHardwareCursor(void)
{
    return !isSoftwareCursor;
}

