/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

#ifdef SUPPORT_1BPP

#define BUF_INCR            1
#define COLOR_TYPE          N_uint8
#define FULL_COLOR          N_uint8
#define COLOR_WHITE         0x01
#define PATT_TYPE           b1
#define INC_DEST(x,addr)    if (((x) & 0x7) == 0x7) (addr)++
#define INC_X(x)            (x)++
#define DEC_DEST(x,addr)    if (((x) & 0x7) == 0x0) (addr)--
#define DEC_X(x)            (x)--

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_fillScan_1_S(
    uchar *addr,
    int x,
    int width,
    GA_color color)
{
    GA_color    mask;
    int         bytes;
    int         right = x+width-1;

    // Handle first odd pixels if any
    if (x & 7) {
        mask = MONO_L_MASK(x);
        *addr = (*addr & ~mask) | (color ? mask : 0);
        addr++;
        width -= (x & 7);
        }

    // Block fill the middle bytes as fast as possible
    bytes = width/8;
    DWORD_FILL(addr,bytes,color);

    // Handle last odd pixels if any
    if (width & 7) {
        mask = MONO_R_MASK(right);
        *addr = (*addr & ~mask) | (color ? mask : 0);
        }
}

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_fillScan_1##rop (uchar *addr,int x,int width,GA_color color)
#include "raster/fillscan.c"

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_drawScan_1_S(
    uchar *addr,int x,
    int width)
{
    GA_color    color = DC.foreColor;
    GA_color    mask;
    int         bytes;
    int         right = x+width-1;

    // Handle first odd pixels if any
    if (x & 7) {
        mask = MONO_L_MASK(x);
        *addr = (*addr & ~mask) | (color ? mask : 0);
        addr++;
        width -= (x & 7);
        }

    // Block fill the middle bytes as fast as possible
    bytes = width/8;
    DWORD_FILL(addr,bytes,color);

    // Handle last odd pixels if any
    if (width & 7) {
        mask = MONO_R_MASK(right);
        *addr = (*addr & ~mask) | (color ? mask : 0);
        }
}

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawScan_1##rop (uchar *addr,int x,int width)
#include "raster/scan.c"

#if 0

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawPattScan_1##rop (uchar *addr,int x1,int y,int width)
#include "raster_m/monoscan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawColorPattScan_1##rop (uchar *addr,int x1,int y,int width)
#include "raster_m/colrscan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoMSBTrans_1##rop (uchar *addr,uchar *image,int x,int width)
#include "raster_m/msbtran.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoMSBOpaque_1##rop (uchar *addr,uchar *image,int x,int width)
#include "raster_m/msbopaq.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoMSBTrans_1##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster_m/cmsbtran.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoMSBOpaque_1##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster_m/cmsbopaq.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoLSBTrans_1##rop (uchar *addr,uchar *image,int x,int width)
#include "raster_m/lsbtran.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoLSBOpaque_1##rop (uchar *addr,uchar *image,int x,int width)
#include "raster_m/lsbopaq.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoLSBTrans_1##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster_m/clsbtran.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoLSBOpaque_1##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster_m/clsbopaq.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_drawMonoCursor_1(uchar *addr,uchar *maskPtr,uchar *imagePtr,int x,int width)
#include "raster_m/monocurs.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_drawColorCursor_1(uchar *addr,uchar *maskPtr,uchar *imagePtr,int x,int width)
#include "raster_m/colrcurs.c"

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_readScan_1(
    uchar *d,
    uchar *s,int srcLeft,dstLeft,
    int width)
{
    // Force DWORD source alignment
    while (width && ((uint)s & 3) != 0) {
        *d++ = *s++;
        width--;
        }
    if (width)
        memcpy(d,s,width);
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_bltScan_1_S(
    uchar *d,
    uchar *s,int srcLeft,dstLeft,
    int width)
{
    if (width >= 8) {
        // Force DWORD destination alignment
        while (((uint)d & 3) != 0) {
            *d++ = *s++;
            width--;
            }
        memcpy(d,s,width);
        }
    else
        memcpy(d,s,width);
}

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_bltScan_1##rop (uchar *din,uchar *sin,int srcLeft,dstLeft,int width)
#include "raster_m/bltscan.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_bltScanPatt_1 (uchar *din,uchar *sin DCL_X_ARG3,int x1,int y,int width)
#include "raster_m/bltmono.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_bltScanColorPatt_1 (uchar *din,uchar *sin DCL_X_ARG3,int x1,int y,int width)
#include "raster_m/bltcolor.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_srcTransBltScan_1##rop (uchar *din,uchar *sin,int srcLeft,dstLeft,int width,GA_color transparent)
#include "raster_m/transcan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_dstTransBltScan_1##rop (uchar *din,uchar *sin,int srcLeft,dstLeft,int width,GA_color transparent)
#include "raster_m/dstscan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_stretchScan2x_1##rop (uchar *din,uchar *sin,int srcLeft,dstLeft,int width)
#include "raster_m/stretch2.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_stretchScan_1##rop (uchar *din,uchar *sin,int srcLeft,dstLeft,int width,int *repCounts,int skipLeft)
#include "raster_m/stretch.c"

#define DRAW_LINEXPOSLO_PROTO(rop) \
    void REF2D_drawLineXPosLo_1##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXPOSHI_PROTO(rop) \
    void REF2D_drawLineXPosHi_1##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXNEGLO_PROTO(rop) \
    void REF2D_drawLineXNegLo_1##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXNEGHI_PROTO(rop) \
    void REF2D_drawLineXNegHi_1##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr)
#include "raster_m/line.c"

#define DRAW_LINEXPOSLO_PROTO(rop) \
    void REF2D_drawStippleLineXPosLo_1##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#define DRAW_LINEXPOSHI_PROTO(rop) \
    void REF2D_drawStippleLineXPosHi_1##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#define DRAW_LINEXNEGLO_PROTO(rop) \
    void REF2D_drawStippleLineXNegLo_1##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#define DRAW_LINEXNEGHI_PROTO(rop) \
    void REF2D_drawStippleLineXNegHi_1##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#include "raster_m/stipline.c"

#define DRAW_LINEXPOSLO_PROTO(rop) \
    void REF2D_drawStyleLineXPosLo_1##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#define DRAW_LINEXPOSHI_PROTO(rop) \
    void REF2D_drawStyleLineXPosHi_1##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#define DRAW_LINEXNEGLO_PROTO(rop) \
    void REF2D_drawStyleLineXNegLo_1##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#define DRAW_LINEXNEGHI_PROTO(rop) \
    void REF2D_drawStyleLineXNegHi_1##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#include "raster_m/stylline.c"

/****************************************************************************
PARAMETERS:
srcPal      - Source bitmap palette to build translation vector for
dstPal      - Destination bitmap palette to build translation vector for
srcColor    - Number of colors in source palette

REMARKS:
This function builds a translation vector for translating 4bpp or 8bpp
color values to the destination device color values, by finding the closest
color in the destination palette that matches the color in the source
palette.
****************************************************************************/
uchar * REF2D_buildTranslateVector_1(
    GA_palette *dstPal,
    GA_palette *srcPal,
    int srcColors)
{
    static uchar    t[256];
    uchar           *p = t;

    while (srcColors--) {
        int         j,closest = 0,d,distance = 0x7FFFL;
        GA_palette  *dp = dstPal;

        for (j = 0; j < 256; j++,dp++) {
            if ((d = ABS((int)srcPal->Red - (int)dp->Red)
                   + ABS((int)srcPal->Green - (int)dp->Green)
                   + ABS((int)srcPal->Blue - (int)dp->Blue)) < distance) {
                distance = d;
                closest = j;
                }
            }
        *p++ = closest;
        srcPal++;
        }
    return t;
}

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_convScan4_1##rop (uchar *din,uchar *sin,int x,int width,int odd,uchar *translate, int *repCounts, int skipLeft)
#include "raster_m/cvscan4.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_convScan8_1##rop (uchar *din,uchar *sin,int x,int width,uchar *translate, int *repCounts, int skipLeft)
#include "raster_m/cvscan8.c"

void REF2D_convScan16_1_Generic(
    uchar *din,
    uchar *sin,int x,
    int width,
    int *repCounts,
    int skipLeft)
{
    COLOR_TYPE  *d = (COLOR_TYPE*)din;
    ushort      *s = (ushort*)sin;

    if (repCounts) {
        repCounts += skipLeft;
        while (width--) {
            int i = *repCounts++;
            while (i--)
                DC.drawPixel((uchar*)d++ X_ARG(x),DC.findClosest(*s));
            s++;
            }
        }
    else {
        while (width--)
            DC.drawPixel((uchar*)d++ X_ARG(x),DC.findClosest(*s++));
        }
}

void REF2D_convScan24_1_Generic(
    uchar *din,
    uchar *sin,int x,
    int width,
    int *repCounts,
    int skipLeft)
{
    COLOR_TYPE  *d = (COLOR_TYPE*)din;
    uchar       *s = (uchar*)sin;

    if (repCounts) {
        repCounts += skipLeft;
        while (width--) {
            int i = *repCounts++;
            while (i--)
                DC.drawPixel((uchar*)d++ X_ARG(x),DC.findClosest(*((GA_color*)s)));
            s += 3;
            }
        }
    else {
        while (width--) {
            DC.drawPixel((uchar*)d++ X_ARG(x),DC.findClosest(*((GA_color*)s)));
            s += 3;
            }
        }
}

void REF2D_convScan32_1_Generic(
    uchar *din,
    uchar *sin,int x,
    int width,
    int *repCounts,
    int skipLeft)
{
    COLOR_TYPE  *d = (COLOR_TYPE*)din;
    ulong       *s = (ulong*)sin;

    if (repCounts) {
        repCounts += skipLeft;
        while (width--) {
            int i = *repCounts++;
            while (i--)
                DC.drawPixel((uchar*)d++ X_ARG(x),DC.findClosest(*s));
            s++;
            }
        }
    else {
        while (width--)
            DC.drawPixel((uchar*)d++ X_ARG(x),DC.findClosest(*s++));
        }
}

#define FIND_CLOSEST_COLOR()                                \
    GA_palette  *p = DC.cntPal;                             \
    int         i,closest = 0,d,distance = 0x7FFFL;         \
                                                            \
    for (i = 0; i < 256; i++,p++) {                         \
        if ((d = ABS(red - (int)p->Red)                     \
               + ABS(green - (int)p->Green)                 \
               + ABS(blue - (int)p->Blue)) < distance) {    \
            distance = d;                                   \
            closest = i;                                    \
            }                                               \
        }                                                   \
    return closest;

/****************************************************************************
PARAMETERS:
color   - Packed RGB color value to find closest color for

RETURNS:
Color index of the closest color to the specified RGB color

REMARKS:
This routine searches the destination bitmap palette for the color that is
the closest to the specified RGB color value, and then returns it. This is
used for converting RGB bitmaps to 8bpp color index modes.
****************************************************************************/
GA_color REF2D_findClosest_15(
    GA_color color)
{
    uchar   red = (color >> 7) & 0xF8;
    uchar   green = (color >> 2) & 0xF8;
    uchar   blue = (color << 3);
    FIND_CLOSEST_COLOR();
}

/****************************************************************************
PARAMETERS:
color   - Packed RGB color value to find closest color for

RETURNS:
Color index of the closest color to the specified RGB color

REMARKS:
This routine searches the destination bitmap palette for the color that is
the closest to the specified RGB color value, and then returns it. This is
used for converting RGB bitmaps to 8bpp color index modes.
****************************************************************************/
GA_color REF2D_findClosest_16(
    GA_color color)
{
    uchar   red = (color >> 8) & 0xF8;
    uchar   green = (color >> 3) & 0xFC;
    uchar   blue = (color << 3);
    FIND_CLOSEST_COLOR();
}

/****************************************************************************
PARAMETERS:
color   - Packed RGB color value to find closest color for

RETURNS:
Color index of the closest color to the specified RGB color

REMARKS:
This routine searches the destination bitmap palette for the color that is
the closest to the specified RGB color value, and then returns it. This is
used for converting RGB bitmaps to 8bpp color index modes.
****************************************************************************/
GA_color REF2D_findClosest_24RGB(
    GA_color color)
{
    uchar   red = color >> 16;
    uchar   green = color >> 8;
    uchar   blue = color;
    FIND_CLOSEST_COLOR();
}

/****************************************************************************
PARAMETERS:
color   - Packed RGB color value to find closest color for

RETURNS:
Color index of the closest color to the specified RGB color

REMARKS:
This routine searches the destination bitmap palette for the color that is
the closest to the specified RGB color value, and then returns it. This is
used for converting RGB bitmaps to 8bpp color index modes.
****************************************************************************/
GA_color REF2D_findClosest_24BGR(
    GA_color color)
{
    uchar   red = color;
    uchar   green = color >> 8;
    uchar   blue = color >> 16;
    FIND_CLOSEST_COLOR();
}

/****************************************************************************
PARAMETERS:
color   - Packed RGB color value to find closest color for

RETURNS:
Color index of the closest color to the specified RGB color

REMARKS:
This routine searches the destination bitmap palette for the color that is
the closest to the specified RGB color value, and then returns it. This is
used for converting RGB bitmaps to 8bpp color index modes.
****************************************************************************/
GA_color REF2D_findClosest_32RGBA(
    GA_color color)
{
    uchar   red = color >> 24;
    uchar   green = color >> 16;
    uchar   blue = color >> 8;
    FIND_CLOSEST_COLOR();
}

/****************************************************************************
PARAMETERS:
color   - Packed RGB color value to find closest color for

RETURNS:
Color index of the closest color to the specified RGB color

REMARKS:
This routine searches the destination bitmap palette for the color that is
the closest to the specified RGB color value, and then returns it. This is
used for converting RGB bitmaps to 8bpp color index modes.
****************************************************************************/
GA_color REF2D_findClosest_32BGRA(
    GA_color color)
{
    uchar   red = color >> 8;
    uchar   green = color >> 16;
    uchar   blue = color >> 24;
    FIND_CLOSEST_COLOR();
}

#define DITHER_VARS         DITHER_VARS8
#define SETUP_DITHER        SETUP_DITHER8
#define PACK_COLOR_DITHER   PACK_COLOR_DITHER8
#define DITHER_SCAN_PROTO(depth,rop) \
    void REF2D_ditherScan##depth##_1##rop (int x,int y,uchar *din,uchar *sin,int width,int *repCounts,int skipLeft)
#include "raster_m/dither.c"

#define DRAW_ELLIPSE_PROTO(rop) \
    void REF2D_drawEllipse_1##rop(int ileft,int itop,int A,int B)
#include "raster_m/ellipse.c"

#define DRAW_ELLIPSE_PROTO(rop) \
    void REF2D_clipEllipse_1##rop(int ileft,int itop,int A,int B,int clipLeft,int clipTop,int clipRight,int clipBottom)
#include "raster_m/cellipse.c"
#endif

#endif
