/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

#define BUF_INCR            3
#define BUF_INCR2           3
#define COLOR_TYPE          N_uint8
#define FULL_COLOR          GA_color
#define PACKED_3BYTE
#define INC_DEST(x,addr)    (addr) += 3
#define INC_X(x)
#define DEC_DEST(x,addr)    (addr) -= 3
#define DEC_X(x)

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_fillScan_24_S(
    uchar *addr DCL_X_ARG,
    int width,
    GA_color color)
{
    ushort  fw0 = WORD0(color);
    ushort  fw1 = WORD1(color);
    uchar   fb0 = BYTE0(color);
    uchar   fb2 = BYTE2(color);

    // We split up the writes into 16:8 and 8:16 chunks for performance
    // reasons. Hence we have two inner loops depending on whether
    // the initial address is odd or not.
    if ((int)addr & 1) {
        while (width >= 4) {
            PLOT_S (addr[0+0],fb0);
            PLOTW_S(addr[0+1],fw1);
            PLOTW_S(addr[3+0],fw0);
            PLOT_S (addr[3+2],fb2);
            PLOT_S (addr[6+0],fb0);
            PLOTW_S(addr[6+1],fw1);
            PLOTW_S(addr[9+0],fw0);
            PLOT_S (addr[9+2],fb2);
            addr += 3 * 4;
            width -= 4;
            }
        while (width--) {
            PLOT_S(addr[0],fb0);
            PLOTW_S(addr[1],fw1);
            addr += 3;
            if (width) {
                PLOTW_S(addr[0],fw0);
                PLOT_S(addr[2],fb2);
                addr += 3;
                width--;
                }
            }
        }
    else {
        while (width >= 4) {
            PLOTW_S(addr[0+0],fw0);
            PLOT_S (addr[0+2],fb2);
            PLOT_S (addr[3+0],fb0);
            PLOTW_S(addr[3+1],fw1);
            PLOTW_S(addr[6+0],fw0);
            PLOT_S (addr[6+2],fb2);
            PLOT_S (addr[9+0],fb0);
            PLOTW_S(addr[9+1],fw1);
            addr += 3 * 4;
            width -= 4;
            }
        while (width--) {
            PLOTW_S(addr[0],fw0);
            PLOT_S(addr[2],fb2);
            addr += 3;
            if (width) {
                PLOT_S(addr[0],fb0);
                PLOTW_S(addr[1],fw1);
                addr += 3;
                width--;
                }
            }
        }
}

void REF2D_fillScan_24_Generic(
    uchar *addr DCL_X_ARG,
    int width,
    GA_color color)
{
    while (width--) {
        DC.drawPixel(addr X_ARG(x),color);
        addr += 3;
        }
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_drawScan_24_S(
    uchar *addr DCL_X_ARG,
    int width)
{
    ushort  fw0 = WORD0(DC.foreColor);
    ushort  fw1 = WORD1(DC.foreColor);
    uchar   fb0 = BYTE0(DC.foreColor);
    uchar   fb2 = BYTE2(DC.foreColor);

    // We split up the writes into 16:8 and 8:16 chunks for performance
    // reasons. Hence we have two inner loops depending on whether
    // the initial address is odd or not.
    if ((int)addr & 1) {
        while (width >= 4) {
            PLOT_S (addr[0+0],fb0);
            PLOTW_S(addr[0+1],fw1);
            PLOTW_S(addr[3+0],fw0);
            PLOT_S (addr[3+2],fb2);
            PLOT_S (addr[6+0],fb0);
            PLOTW_S(addr[6+1],fw1);
            PLOTW_S(addr[9+0],fw0);
            PLOT_S (addr[9+2],fb2);
            addr += 3 * 4;
            width -= 4;
            }
        while (width--) {
            PLOT_S(addr[0],fb0);
            PLOTW_S(addr[1],fw1);
            addr += 3;
            if (width) {
                PLOTW_S(addr[0],fw0);
                PLOT_S(addr[2],fb2);
                addr += 3;
                width--;
                }
            }
        }
    else {
        while (width >= 4) {
            PLOTW_S(addr[0+0],fw0);
            PLOT_S (addr[0+2],fb2);
            PLOT_S (addr[3+0],fb0);
            PLOTW_S(addr[3+1],fw1);
            PLOTW_S(addr[6+0],fw0);
            PLOT_S (addr[6+2],fb2);
            PLOT_S (addr[9+0],fb0);
            PLOTW_S(addr[9+1],fw1);
            addr += 3 * 4;
            width -= 4;
            }
        while (width--) {
            PLOTW_S(addr[0],fw0);
            PLOT_S(addr[2],fb2);
            addr += 3;
            if (width) {
                PLOT_S(addr[0],fb0);
                PLOTW_S(addr[1],fw1);
                addr += 3;
                width--;
                }
            }
        }
}

void REF2D_drawScan_24_Generic(
    uchar *addr DCL_X_ARG,
    int width)
{
    GA_color color = DC.foreColor;

    while (width--) {
        DC.drawPixel(addr X_ARG(x),color);
        addr += 3;
        }
}

#define PREAMBLE \
    GA_color    foreColor = DC.foreColor; \
    GA_color    backColor = DC.backColor; \
    uchar       pattern = DC.mono8x8Patt[DC.mono8x8Index].p[y & 7]; \
    uchar       p = pattern << (x1 & 7)

void REF2D_drawPattScan_24_Generic(uchar *dest,int x1,int y,int width)
{
    #define PLOT_F()    DC.drawPixel((uchar*)dest X_ARG(x1),foreColor)
    #define PLOT_B()    DC.drawPixel((uchar*)dest X_ARG(x1),backColor)
    #include "raster/inner/monoscan.c"
}

#undef  PREAMBLE

#define PREAMBLE \
    COLOR_TYPE  *pattern = &DC.color8x8Patt[DC.color8x8Index].b24.p[y & 7][0][0]; \
    COLOR_TYPE  *p = &pattern[(x1 & 7) * 3]; \
    COLOR_TYPE  trans0 = BYTE0(DC.color8x8TransColor); \
    COLOR_TYPE  trans1 = BYTE1(DC.color8x8TransColor); \
    COLOR_TYPE  trans2 = BYTE2(DC.color8x8TransColor)

void REF2D_drawColorPattScan_24_Generic(uchar *dest,int x1,int y,int width)
{
    #define PLOT()                                                      \
    {                                                                   \
        DC.drawPixel((uchar*)dest X_ARG(x1),*((ulong*)p));              \
        dest += 3;                                                      \
        p += 3;                                                         \
    }
    #define PLOT_TRANS()                                                \
    {                                                                   \
        if (p[0] != trans0 || p[1] != trans1 || p[2] != trans2)         \
            DC.drawPixel((uchar*)dest X_ARG(x1),*((ulong*)p));          \
        dest += 3;                                                      \
        p += 3;                                                         \
    }
    #include "raster/inner/colrscan.c"
}

#undef  PREAMBLE

#define PIXEL0_MASK 0x80
#define PIXEL1_MASK 0x40
#define PIXEL2_MASK 0x20
#define PIXEL3_MASK 0x10
#define PIXEL4_MASK 0x08
#define PIXEL5_MASK 0x04
#define PIXEL6_MASK 0x02
#define PIXEL7_MASK 0x01
#define PIXEL_SHIFT <<=
#define PIXEL_MASK  0x1000000
#define MASK_SHIFT  >>

void REF2D_drawMonoMSBTrans_24_Generic(
    uchar *addr,
    uchar *image DCL_X_ARG,
    int width)
{
    #define PLOT(d,c) DC.drawPixel((uchar*)&(d) X_ARG(x),c)
    #include "raster/inner/mtran.c"
}

void REF2D_drawMonoMSBOpaque_24_Generic(
    uchar *addr,
    uchar *image DCL_X_ARG,
    int width)
{
    #define PLOT(d,c) DC.drawPixel((uchar*)&(d) X_ARG(x),c)
    #include "raster/inner/mopaq.c"
}

void REF2D_drawClipMonoMSBTrans_24_Generic(
    int x,
    uchar *addr,
    uchar *image,
    int width,
    int clipLeft)
{
    #define PLOT(d,c) DC.drawPixel((uchar*)&(d) X_ARG(x),c)
    #include "raster/inner/cmtran.c"
}

void REF2D_drawClipMonoMSBOpaque_24_Generic(
    int x,
    uchar *addr,
    uchar *image,
    int width,
    int clipLeft)
{
    #define PLOT(d,c) DC.drawPixel((uchar*)&(d) X_ARG(x),c)
    #include "raster/inner/cmopaq.c"
}

#undef  PIXEL0_MASK
#undef  PIXEL1_MASK
#undef  PIXEL2_MASK
#undef  PIXEL3_MASK
#undef  PIXEL4_MASK
#undef  PIXEL5_MASK
#undef  PIXEL6_MASK
#undef  PIXEL7_MASK
#undef  PIXEL_SHIFT
#undef  PIXEL_MASK
#undef  MASK_SHIFT

#define PIXEL0_MASK 0x01
#define PIXEL1_MASK 0x02
#define PIXEL2_MASK 0x04
#define PIXEL3_MASK 0x08
#define PIXEL4_MASK 0x10
#define PIXEL5_MASK 0x20
#define PIXEL6_MASK 0x40
#define PIXEL7_MASK 0x80
#define PIXEL_SHIFT >>=
#define PIXEL_MASK  0x100
#define MASK_SHIFT  <<

void REF2D_drawMonoLSBTrans_24_Generic(
    uchar *addr,
    uchar *image DCL_X_ARG,
    int width)
{
    #define PLOT(d,c) DC.drawPixel((uchar*)&(d) X_ARG(x),c)
    #include "raster/inner/mtran.c"
}

void REF2D_drawMonoLSBOpaque_24_Generic(
    uchar *addr,
    uchar *image DCL_X_ARG,
    int width)
{
    #define PLOT(d,c) DC.drawPixel((uchar*)&(d) X_ARG(x),c)
    #include "raster/inner/mopaq.c"
}

void REF2D_drawClipMonoLSBTrans_24_Generic(
    int x,
    uchar *addr,
    uchar *image,
    int width,
    int clipLeft)
{
    #define PLOT(d,c) DC.drawPixel((uchar*)&(d) X_ARG(x),c)
    #include "raster/inner/cmtran.c"
}

void REF2D_drawClipMonoLSBOpaque_24_Generic(
    int x,
    uchar *addr,
    uchar *image,
    int width,
    int clipLeft)
{
    #define PLOT(d,c) DC.drawPixel((uchar*)&(d) X_ARG(x),c)
    #include "raster/inner/cmopaq.c"
}

#undef  DRAW_SCAN_PROTO
#undef  PREAMBLE

void REF2D_drawMonoCursor_24(
    uchar *d,
    uchar *maskPtr,
    uchar *imagePtr DCL_X_ARG,
    int width)
{
    int     i;
    uchar   foreColor0 = BYTE0(DC.cursorForeColor);
    uchar   foreColor1 = BYTE1(DC.cursorForeColor);
    uchar   foreColor2 = BYTE2(DC.cursorForeColor);
    uchar   backColor0 = BYTE0(DC.cursorBackColor);
    uchar   backColor1 = BYTE1(DC.cursorBackColor);
    uchar   backColor2 = BYTE2(DC.cursorBackColor);

    while (width--) {
        uchar mask = *maskPtr++;
        uchar image = *imagePtr++;
        for (i = 0; i < 8; i++) {
            if (mask & 0x80) {
                if (image & 0x80) {
                    PLOT_S(d[0],foreColor0);
                    PLOT_S(d[1],foreColor1);
                    PLOT_S(d[2],foreColor2);
                    }
                else {
                    PLOT_S(d[0],backColor0);
                    PLOT_S(d[1],backColor1);
                    PLOT_S(d[2],backColor2);
                    }
                }
            else if (image & 0x80) {
                PLOT_Dn(d[0]);
                PLOT_Dn(d[1]);
                PLOT_Dn(d[2]);
                }
            mask <<= 1;
            image <<= 1;
            d += 3;
            }
        }
}

void REF2D_drawColorCursor_24(
    uchar *d,
    uchar *maskPtr,
    uchar *imagePtr DCL_X_ARG,
    int width)
{
    int     i;

    while (width--) {
        uchar mask = *maskPtr++;
        for (i = 0; i < 8; i++) {
            uchar image0 = *imagePtr++;
            uchar image1 = *imagePtr++;
            uchar image2 = *imagePtr++;
            if (mask & 0x80) {
                PLOT_S(d[0],image0);
                PLOT_S(d[1],image1);
                PLOT_S(d[2],image2);
                }
            else {
                PLOT_DSx(d[0],image0);
                PLOT_DSx(d[1],image1);
                PLOT_DSx(d[2],image2);
                }
            mask <<= 1;
            d += 3;
            }
        }
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_readScan_24(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width)
{
    // Force DWORD source alignment
    width *= 3;
    while (width && ((uint)s & 3) != 0) {
        *d++ = *s++;
        width--;
        }
    if (width)
        memcpy(d,s,width);
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_bltScan_24_S(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width)
{
    // Force DWORD source alignment
    width *= 3;
    while (((uint)s & 3) != 0) {
        *d++ = *s++;
        width--;
        }
    if (width)
        memcpy(d,s,width);
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_bltScan_24_Generic(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width)
{
    GA_color    c;

    while (width--) {
        BYTE0(c) = s[0];
        BYTE1(c) = s[1];
        BYTE2(c) = s[2];
        DC.drawPixel(d X_ARG(dstLeft),c);
        d += 3;
        s += 3;
        }
}

#undef  PREAMBLE

#define PREAMBLE \
    GA_color    foreColor = DC.foreColor; \
    GA_color    backColor = DC.backColor; \
    uchar       pattern = DC.mono8x8Patt[DC.mono8x8Index].p[y & 7]; \
    uchar       p = pattern << (x1 & 7)

void REF2D_bltScanPatt_24(
    uchar *d,
    uchar *s DCL_X_ARG3,
    int x1,
    int y,
    int width)
{
    #define PLOT_F()    DC.drawPixelR3((uchar*)d X_ARG(x1),*((ulong*)s),foreColor)
    #define PLOT_B()    DC.drawPixelR3((uchar*)d X_ARG(x1),*((ulong*)s),backColor)
    #include "raster/inner/bltmono.c"
}

#undef  PREAMBLE

#define PREAMBLE \
    COLOR_TYPE  *pattern = &DC.color8x8Patt[DC.color8x8Index].b24.p[y & 7][0][0]; \
    COLOR_TYPE  *p = &pattern[(x1 & 7) * 3]; \

void REF2D_bltScanColorPatt_24(
    uchar *din,
    uchar *sin DCL_X_ARG3,
    int x1,
    int y,
    int width)
{
    #define PLOT()                                                  \
    {                                                               \
        DC.drawPixelR3((uchar*)din X_ARG(x1),*((ulong*)sin),*((ulong*)p));\
        din += 3;                                                   \
        sin += 3;                                                   \
        p += 3;                                                     \
    }
    #include "raster/inner/bltcolor.c"
}

#undef  PREAMBLE

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_srcTransBltScan_24_S(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width)
{
    GA_color    c = 0;
    GA_color    transparent = DC.transparent;

    while (width--) {
        BYTE0(c) = s[0];
        BYTE1(c) = s[1];
        BYTE2(c) = s[2];
        if (c != transparent) {
            d[0] = BYTE0(c);
            d[1] = BYTE1(c);
            d[2] = BYTE2(c);
            }
        d += 3;
        s += 3;
        }
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_srcTransBltScan_24_Generic(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width)
{
    GA_color    c = 0;
    GA_color    transparent = DC.transparent;

    while (width--) {
        BYTE0(c) = s[0];
        BYTE1(c) = s[1];
        BYTE2(c) = s[2];
        if (c != transparent)
            DC.drawPixel(d X_ARG(dstLeft),c);
        d += 3;
        s += 3;
        }
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_dstTransBltScan_24_S(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width)
{
    GA_color    c = 0;
    GA_color    transparent = DC.transparent;

    while (width--) {
        BYTE0(c) = d[0];
        BYTE1(c) = d[1];
        BYTE2(c) = d[2];
        if (c == transparent) {
            d[0] = s[0];
            d[1] = s[1];
            d[2] = s[2];
            }
        d += 3;
        s += 3;
        }
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_dstTransBltScan_24_Generic(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width)
{
    GA_color    c = 0;
    GA_color    transparent = DC.transparent;

    while (width--) {
        BYTE0(c) = d[0];
        BYTE1(c) = d[1];
        BYTE2(c) = d[2];
        if (c == transparent) {
            BYTE0(c) = s[0];
            BYTE1(c) = s[1];
            BYTE2(c) = s[2];
            DC.drawPixel(d X_ARG(dstLeft),c);
            }
        d += 3;
        s += 3;
        }
}

/****************************************************************************
PARAMETERS:
din     - Destination memory address
sin     - Source memory address
width   - Number of pixels to copy

REMARKS:
This function stretchs a scanline by a factor of 2 in the X direction.
****************************************************************************/
void REF2D_stretchScan2x_24_S(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width)
{
    while (width--) {
        d[0] = s[0];
        d[1] = s[1];
        d[2] = s[2];
        d[3] = s[0];
        d[4] = s[1];
        d[5] = s[2];
        d += 6;
        s += 3;
        }
}

/****************************************************************************
PARAMETERS:
din     - Destination memory address
sin     - Source memory address
width   - Number of pixels to copy

REMARKS:
This function stretchs a scanline by a factor of 2 in the X direction.
****************************************************************************/
void REF2D_stretchScan2x_24_Generic(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width)
{
    GA_color    c;

    while (width--) {
        BYTE0(c) = s[0];
        BYTE1(c) = s[1];
        BYTE2(c) = s[2];
        DC.drawPixel(d+0 X_ARG(dstLeft),c);
        DC.drawPixel(d+3 X_ARG(dstLeft),c);
        d += 6;
        s += 3;
        }
}

/****************************************************************************
PARAMETERS:
din         - Destination memory address
sin         - Source memory address
width       - Number of pixels to copy
repCounts   - Array of pixel replication counts

REMARKS:
This function stretchs a scanline by a generic factor in the X direction.
****************************************************************************/
void REF2D_stretchScan_24_S(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width,
    int *repCounts)
{
    while (width--) {
        int i = *repCounts++;
        while (i--) {
            d[0] = s[0];
            d[1] = s[1];
            d[2] = s[2];
            d += 3;
            }
        s += 3;
        }
}

/****************************************************************************
PARAMETERS:
din         - Destination memory address
sin         - Source memory address
width       - Number of pixels to copy
repCounts   - Array of pixel replication counts

REMARKS:
This function stretchs a scanline by a generic factor in the X direction.
****************************************************************************/
void REF2D_stretchScan_24_Generic(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width,
    int *repCounts)
{
    GA_color    c;

    while (width--) {
        int i = *repCounts++;
        BYTE0(c) = s[0];
        BYTE1(c) = s[1];
        BYTE2(c) = s[2];
        while (i--) {
            DC.drawPixel(d X_ARG(dstLeft),c);
            d += 3;
            }
        s += 3;
        }
}

void REF2D_drawLineXPosLo_24_S(
    uchar *addr DCL_X_ARG,
    int d,
    int Eincr,
    int NEincr,
    int count,
    int fincr)
{
    uchar   c0 = BYTE0(DC.foreColor);
    uchar   c1 = BYTE1(DC.foreColor);
    uchar   c2 = BYTE2(DC.foreColor);

    while (count--) {
        addr[0] = c0;
        addr[1] = c1;
        addr[2] = c2;
        addr += 3;
        if (d <= 0) {
            d += Eincr;         // Choose the Eastern Pixel
            }
        else {
            d += NEincr;        // Choose the North Eastern Pixel
            BINC(addr,fincr);   // Increment framebuffer
            }
        }
}

void REF2D_drawLineXPosHi_24_S(
    uchar *addr DCL_X_ARG,
    int d,
    int Eincr,
    int NEincr,
    int count,
    int fincr)
{
    uchar   c0 = BYTE0(DC.foreColor);
    uchar   c1 = BYTE1(DC.foreColor);
    uchar   c2 = BYTE2(DC.foreColor);

    while (count--) {
        addr[0] = c0;
        addr[1] = c1;
        addr[2] = c2;
        BINC(addr,fincr);       // Increment framebuffer
        if (d <= 0) {
            d += Eincr;         // Choose the Eastern Pixel
            }
        else {
            d += NEincr;        // Choose the North Eastern Pixel
            addr += 3;
            }
        }
}

void REF2D_drawLineXNegLo_24_S(
    uchar *addr DCL_X_ARG,
    int d,
    int Eincr,
    int NEincr,
    int count,
    int fincr)
{
    uchar   c0 = BYTE0(DC.foreColor);
    uchar   c1 = BYTE1(DC.foreColor);
    uchar   c2 = BYTE2(DC.foreColor);

    while (count--) {
        addr[0] = c0;
        addr[1] = c1;
        addr[2] = c2;
        addr -= 3;
        if (d <= 0) {
            d += Eincr;         // Choose the Eastern Pixel
            }
        else {
            d += NEincr;        // Choose the North Eastern Pixel
            BINC(addr,fincr);   // Increment framebuffer
            }
        }
}

void REF2D_drawLineXNegHi_24_S(
    uchar *addr DCL_X_ARG,
    int d,
    int Eincr,
    int NEincr,
    int count,
    int fincr)
{
    uchar   c0 = BYTE0(DC.foreColor);
    uchar   c1 = BYTE1(DC.foreColor);
    uchar   c2 = BYTE2(DC.foreColor);

    while (count--) {
        addr[0] = c0;
        addr[1] = c1;
        addr[2] = c2;
        BINC(addr,fincr);       // Increment framebuffer
        if (d <= 0) {
            d += Eincr;         // Choose the Eastern Pixel
            }
        else {
            d += NEincr;        // Choose the North Eastern Pixel
            addr -= 3;
            }
        }
}

#define NO_SOLID_LINE
#define DRAW_LINEXPOSLO_PROTO(rop) \
    void REF2D_drawLineXPosLo_24##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXPOSHI_PROTO(rop) \
    void REF2D_drawLineXPosHi_24##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXNEGLO_PROTO(rop) \
    void REF2D_drawLineXNegLo_24##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXNEGHI_PROTO(rop) \
    void REF2D_drawLineXNegHi_24##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr)
#include "raster/line.c"

#define DRAW_LINEXPOSLO_PROTO(rop) \
    void REF2D_drawStippleLineXPosLo_24##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#define DRAW_LINEXPOSHI_PROTO(rop) \
    void REF2D_drawStippleLineXPosHi_24##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#define DRAW_LINEXNEGLO_PROTO(rop) \
    void REF2D_drawStippleLineXNegLo_24##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#define DRAW_LINEXNEGHI_PROTO(rop) \
    void REF2D_drawStippleLineXNegHi_24##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#include "raster/stipline.c"

#define DRAW_LINEXPOSLO_PROTO(rop) \
    void REF2D_drawStyleLineXPosLo_24##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#define DRAW_LINEXPOSHI_PROTO(rop) \
    void REF2D_drawStyleLineXPosHi_24##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#define DRAW_LINEXNEGLO_PROTO(rop) \
    void REF2D_drawStyleLineXNegLo_24##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#define DRAW_LINEXNEGHI_PROTO(rop) \
    void REF2D_drawStyleLineXNegHi_24##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#include "raster/stylline.c"

#define BUILD_TRANSLATE_PROTO() \
    GA_color *  REF2D_buildTranslateVector_24RGB(GA_color *t,GA_palette *dstPal,GA_palette *srcPal,int srcColors)
#define PACK_COLOR  PACK_COLOR_RGB
#include "raster/mktrans.c"

#define BUILD_TRANSLATE_PROTO() \
    GA_color *  REF2D_buildTranslateVector_24BGR(GA_color *t,GA_palette *dstPal,GA_palette *srcPal,int srcColors)
#define PACK_COLOR  PACK_COLOR_BGR
#include "raster/mktrans.c"

void REF2D_convScan4_24_S(
    uchar *din,
    uchar *sin DCL_X_ARG,
    int width,
    int odd,
    GA_color *t)
{
    uchar       *d = (uchar*)din;
    uchar       *s = (uchar*)sin;

    if (odd) {
        REF2D_drawPixel_24_S(d X_ARG(x),t[*s++ & 0xF]);  // Handle first odd pixel
        d += 3;
        width--;
        }
    while (width--) {
        REF2D_drawPixel_24_S(d X_ARG(x),t[*s >> 4]);
        d += 3;
        if (width-- == 0) break;
        REF2D_drawPixel_24_S(d X_ARG(x),t[*s++ & 0xF]);
        d += 3;
        }
}

void REF2D_convScan8_24_S(
    uchar *din,
    uchar *sin DCL_X_ARG,
    int width,
    GA_color *t)
{
    uchar       *d = (uchar*)din;
    uchar       *s = (uchar*)sin;

    while (width--) {
        REF2D_drawPixel_24_S(d X_ARG(x),t[*s++]);
        d += 3;
        }
}

#define DRAW_ELLIPSE_PROTO(rop) \
    void REF2D_drawEllipse_24##rop(int ileft,int itop,int A,int B)
#include "raster/ellipse.c"

#define DRAW_ELLIPSE_PROTO(rop) \
    void REF2D_clipEllipse_24##rop(int ileft,int itop,int A,int B,int clipLeft,int clipTop,int clipRight,int clipBottom)
#include "raster/cellipse.c"

