/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

#define BUF_INCR                    4
#define COLOR_TYPE                  N_uint32
#define FULL_COLOR                  N_uint32
#define COLOR_WHITE                 0xFFFFFFFF
#define PATT_TYPE                   b32
#define INC_DEST(x,addr)            (addr)++
#define INC_X(x)
#define DEC_DEST(x,addr)            (addr)--
#define DEC_X(x)
#define REPLACE_PIXEL(dest,x,color) PLOT_S(dest,color)
#define READ_PIXEL(src,x)           (src)

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_fillScan_32_S(
    uchar *addr DCL_X_ARG,
    int width,
    GA_color color)
{
    rep_stosd(addr,width,color);
}

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_fillScan_32##rop (uchar *addr DCL_X_ARG,int width,GA_color color)
#include "raster/fillscan.c"

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_drawScan_32_S(
    uchar *addr DCL_X_ARG,
    int width)
{
    rep_stosd(addr,width,DC.foreColor);
}

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawScan_32##rop (uchar *addr DCL_X_ARG,int width)
#include "raster/scan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawPattScan_32##rop (uchar *addr,int x1,int y,int width)
#include "raster/monoscan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawColorPattScan_32##rop (uchar *addr,int x1,int y,int width)
#include "raster/colrscan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoMSBTrans_32##rop (uchar *addr,uchar *image DCL_X_ARG,int width)
#include "raster/msbtran.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoMSBOpaque_32##rop (uchar *addr,uchar *image DCL_X_ARG,int width)
#include "raster/msbopaq.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoMSBTrans_32##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster/cmsbtran.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoMSBOpaque_32##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster/cmsbopaq.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoLSBTrans_32##rop (uchar *addr,uchar *image DCL_X_ARG,int width)
#include "raster/lsbtran.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoLSBOpaque_32##rop (uchar *addr,uchar *image DCL_X_ARG,int width)
#include "raster/lsbopaq.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoLSBTrans_32##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster/clsbtran.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoLSBOpaque_32##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster/clsbopaq.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_drawMonoCursor_32(uchar *addr,uchar *maskPtr,uchar *imagePtr DCL_X_ARG,int width)
#include "raster/monocurs.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_drawColorCursor_32(uchar *addr,uchar *maskPtr,uchar *imagePtr DCL_X_ARG,int width)
#include "raster/colrcurs.c"

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_readScan_32(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width)
{
    rep_movsd(d,s,width);
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_bltScan_32_S(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width)
{
    rep_movsd(d,s,width);
}

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_bltScan_32##rop (uchar *din,uchar *sin DCL_X_ARG2,int width)
#include "raster/bltscan.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_bltScanPatt_32 (uchar *din,uchar *sin DCL_X_ARG3,int x1,int y,int width)
#include "raster/bltmono.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_bltScanColorPatt_32 (uchar *din,uchar *sin DCL_X_ARG3,int x1,int y,int width)
#include "raster/bltcolor.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_srcTransBltScan_32##rop (uchar *din,uchar *sin DCL_X_ARG2,int width)
#include "raster/transcan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_dstTransBltScan_32##rop (uchar *din,uchar *sin DCL_X_ARG2,int width)
#include "raster/dstscan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_stretchScan2x_32##rop (uchar *din,uchar *sin DCL_X_ARG2,int width)
#include "raster/stretch2.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_stretchScan_32##rop (uchar *din,uchar *sin DCL_X_ARG2,int width,int *repCounts)
#include "raster/stretch.c"

#define DRAW_LINEXPOSLO_PROTO(rop) \
    void REF2D_drawLineXPosLo_32##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXPOSHI_PROTO(rop) \
    void REF2D_drawLineXPosHi_32##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXNEGLO_PROTO(rop) \
    void REF2D_drawLineXNegLo_32##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXNEGHI_PROTO(rop) \
    void REF2D_drawLineXNegHi_32##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr)
#include "raster/line.c"

#define DRAW_LINEXPOSLO_PROTO(rop) \
    void REF2D_drawStippleLineXPosLo_32##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#define DRAW_LINEXPOSHI_PROTO(rop) \
    void REF2D_drawStippleLineXPosHi_32##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#define DRAW_LINEXNEGLO_PROTO(rop) \
    void REF2D_drawStippleLineXNegLo_32##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#define DRAW_LINEXNEGHI_PROTO(rop) \
    void REF2D_drawStippleLineXNegHi_32##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#include "raster/stipline.c"

#define DRAW_LINEXPOSLO_PROTO(rop) \
    void REF2D_drawStyleLineXPosLo_32##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#define DRAW_LINEXPOSHI_PROTO(rop) \
    void REF2D_drawStyleLineXPosHi_32##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#define DRAW_LINEXNEGLO_PROTO(rop) \
    void REF2D_drawStyleLineXNegLo_32##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#define DRAW_LINEXNEGHI_PROTO(rop) \
    void REF2D_drawStyleLineXNegHi_32##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#include "raster/stylline.c"

#define BUILD_TRANSLATE_PROTO() \
    GA_color *  REF2D_buildTranslateVector_32ARGB(GA_color *t,GA_palette *dstPal,GA_palette *srcPal,int srcColors)
#define PACK_COLOR  PACK_COLOR_ARGB
#include "raster/mktrans.c"

#define BUILD_TRANSLATE_PROTO() \
    GA_color *  REF2D_buildTranslateVector_32ABGR(GA_color *t,GA_palette *dstPal,GA_palette *srcPal,int srcColors)
#define PACK_COLOR  PACK_COLOR_ABGR
#include "raster/mktrans.c"

#define BUILD_TRANSLATE_PROTO() \
    GA_color *  REF2D_buildTranslateVector_32RGBA(GA_color *t,GA_palette *dstPal,GA_palette *srcPal,int srcColors)
#define PACK_COLOR  PACK_COLOR_RGBA
#include "raster/mktrans.c"

#define BUILD_TRANSLATE_PROTO() \
    GA_color *  REF2D_buildTranslateVector_32BGRA(GA_color *t,GA_palette *dstPal,GA_palette *srcPal,int srcColors)
#define PACK_COLOR  PACK_COLOR_BGRA
#include "raster/mktrans.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_convScan4_32##rop (uchar *din,uchar *sin DCL_X_ARG,int width,int odd,GA_color *t)
#include "raster/cvscan4.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_convScan8_32##rop (uchar *din,uchar *sin DCL_X_ARG,int width,GA_color *t)
#include "raster/cvscan8.c"

#define DRAW_ELLIPSE_PROTO(rop) \
    void REF2D_drawEllipse_32##rop(int ileft,int itop,int A,int B)
#include "raster/ellipse.c"

#define DRAW_ELLIPSE_PROTO(rop) \
    void REF2D_clipEllipse_32##rop(int ileft,int itop,int A,int B,int clipLeft,int clipTop,int clipRight,int clipBottom)
#include "raster/cellipse.c"

