/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

#ifdef SUPPORT_4BPP

#define NO_UNROLLED_PIXELS
#define MULTIPLE_PIXELS_PER_BYTE
#define BUF_INCR            1
#define COLOR_TYPE          N_uint8
#define FULL_COLOR          N_uint8
#define COLOR_WHITE         0x0F
#define PATT_TYPE           b8              // Use 8bpp patterns internally
#define INC_DEST_2(x,op)    if ((x) & 1) (op)
#define INC_DEST(x,addr)    INC_DEST_2(x,(addr)++)
#define INC_X(x)            (x)++
#define DEC_DEST_2(x,op)    if (!((x) & 1)) (op)
#define DEC_DEST(x,addr)    DEC_DEST_2(x,(addr)--)
#define DEC_X(x)            (x)--
#undef PLOT_S

// Macro to replace a pixel in the framebuffer
#define REPLACE_PIXEL(dest,x,color)                                     \
{                                                                       \
    if ((x) & 1)                                                        \
        (dest) = (uchar)(((dest) & 0xF0) | ((color) & 0x0F));           \
    else                                                                \
        (dest) = (uchar)(((dest) & 0x0F) | ((color) << 4));             \
}

// Macro to invert a pixel in the framebuffer
#define INVERT_PIXEL(dest,x)                                            \
{                                                                       \
    if ((x) & 1)                                                        \
        (dest) = (uchar)((dest) ^ 0x0F);                                \
    else                                                                \
        (dest) = (uchar)((dest) ^ 0xF0);                                \
}

// Macro to XOR a pixel in the framebuffer
#define XOR_PIXEL(dest,x,color)                                         \
{                                                                       \
    GA_color D = dest;                                                  \
    if (x & 1)                                                          \
        (dest) = (uchar)((D & 0xF0) | (((D & 0x0F) ^ (color)) & 0x0F)); \
    else                                                                \
        (dest) = (uchar)((D & 0x0F) | (((D >> 4) ^ (color)) << 4));     \
}

// Macro to read a pixel from the framebuffer
#define READ_PIXEL(src,x)   (((x) & 1) ? ((src) & 0x0F) : ((src) >> 4))

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_fillScan_4_S(
    uchar *addr,
    int x,
    int width,
    GA_color color)
{
    int bytes;

    // Handle first odd pixel if any
    if (x & 1) {
        *addr = (uchar)((*addr & 0xF0) | (color & 0x0F));
        addr++; width--;
        }

    // Block fill the middle bytes as fast as possible
    bytes = width/2;
    DWORD_FILL(addr,bytes,color);

    // Handle last odd pixel if any
    if (width & 1)
        *addr = (uchar)((*addr & 0x0F) | (color << 4));
}

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_fillScan_4##rop (uchar *addr,int x,int width,GA_color color)
#include "raster/fillscan.c"

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_drawScan_4_S(
    uchar *addr,
    int x,
    int width)
{
    GA_color    color = DC.foreColor;
    int         bytes;

    // Handle first odd pixel if any
    if (x & 1) {
        *addr = (uchar)((*addr & 0xF0) | (color & 0x0F));
        addr++; width--;
        }

    // Block fill the middle bytes as fast as possible
    bytes = width/2;
    DWORD_FILL(addr,bytes,color);

    // Handle last odd pixel if any
    if (width & 1)
        *addr = (uchar)((*addr & 0x0F) | (color << 4));
}

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawScan_4##rop (uchar *addr,int x,int width)
#include "raster/scan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawPattScan_4##rop (uchar *addr,int x1,int y,int width)
#include "raster/monoscan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawColorPattScan_4##rop (uchar *addr,int x1,int y,int width)
#include "raster/colrscan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoMSBTrans_4##rop (uchar *addr,uchar *image,int x,int width)
#include "raster/msbtran4.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoMSBOpaque_4##rop (uchar *addr,uchar *image,int x,int width)
#include "raster/msbopaq4.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoMSBTrans_4##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster/cmsbtrn4.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoMSBOpaque_4##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster/cmsbopq4.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoLSBTrans_4##rop (uchar *addr,uchar *image,int x,int width)
#include "raster/lsbtran4.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoLSBOpaque_4##rop (uchar *addr,uchar *image,int x,int width)
#include "raster/lsbopaq4.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoLSBTrans_4##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster/clsbtrn4.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoLSBOpaque_4##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster/clsbopq4.c"

/****************************************************************************
PARAMETERS:
d           - Memory address to draw pixel at
maskPtr     - Pointer to cursor mask image
imagePtr    - Pointer to cursor XOR image
x           - X coordinate to begin drawing at
width       - Width of the cursor image

REMARKS:
This function draws the mono cursor image into the packed pixel memory buffer.
****************************************************************************/
void REF2D_drawMonoCursor_4(
    uchar *d,
    uchar *maskPtr,
    uchar *imagePtr,
    int x,
    int width)
{
    uchar   color;
    int     i;

    while (width--) {
        uchar mask = *maskPtr++;
        uchar image = *imagePtr++;
        for (i = 0; i < 8; i++) {
            if (mask & 0x80) {
                color = (image & 0x80) ? DC.cursorForeColor : DC.cursorBackColor;
                REPLACE_PIXEL(d[0],x,color);
                }
            else if (image & 0x80)
                INVERT_PIXEL(d[0],x);
            mask <<= 1;
            image <<= 1;
            if (x++ & 1) d++;
            }
        }
}

/****************************************************************************
PARAMETERS:
d           - Memory address to draw pixel at
maskPtr     - Pointer to cursor mask image
imagePtr    - Pointer to cursor XOR image
x           - X coordinate to begin drawing at
width       - Width of the cursor image

REMARKS:
This function draws the color cursor image into the packed pixel memory buffer.
****************************************************************************/
void REF2D_drawColorCursor_4(
    uchar *d,
    uchar *maskPtr,
    uchar *imagePtr,
    int x,
    int width)
{
    int     i;
    uchar   *ip = (uchar*)imagePtr;

    // TODO: This is wrong, since the color cursor data is in packed
    //       4bpp format!? Or perhaps it is expanded in the rcursor code.
    while (width--) {
        uchar mask = *maskPtr++;
        for (i = 0; i < 8; i++) {
            uchar image = *ip++;
            if (mask & 0x80) {
                REPLACE_PIXEL(d[0],x,image);
                }
            else {
                XOR_PIXEL(d[0],x,image);
                }
            mask <<= 1;
            if (x++ & 1) d++;
            }
        }
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_readScan_4(
    uchar *d,
    uchar *s,
    int srcLeft,
    int dstLeft,
    int width)
{
    uchar   residual;
    int     bytes;

    if ((srcLeft & 1) == (dstLeft & 1)) {
        // Handle first odd pixel if any
        if (srcLeft & 1) {
            *d = (uchar)((*d & 0xF0) | (*s++ & 0x0F));
            d++;
            width--;
            }

        // Block copy the middle bytes DWORD's at a time.
        // Make sure we force DWORD source alignment
        bytes = width/2;
        while (bytes && ((uint)s & 3) != 0) {
            *d++ = *s++;
            bytes--;
            }
        if (bytes) {
            memcpy(d,s,bytes);
            d += bytes;
            s += bytes;
            }

        // Handle last odd pixel if any
        if (width & 1)
            *d = (uchar)((*d & 0x0F) | (*s & 0xF0));
        }
    else {
        // Handle odd source or destination pixels
        if (dstLeft & 1) {
            *d = (*d & 0xF0) | (*s >> 4);
            d++;
            width--;
            }
        residual = (*s++ << 4);

        // Copy pixels for all middle bytes
        bytes = width/2;
        while (bytes--) {
            *d++ = (*s >> 4) | residual;
            residual = (*s++ << 4);
            }

        // Handle last odd source pixel
        if (width & 1)
            *d = (*d & 0x0F) | residual;
        }
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_bltScan_4_S(
    uchar *d,
    uchar *s,
    int srcLeft,
    int dstLeft,
    int width)
{
    uchar   residual;
    int     bytes;

    if ((srcLeft & 1) == (dstLeft & 1)) {
        // Handle first odd pixel if any
        if (srcLeft & 1) {
            *d = (uchar)((*d & 0xF0) | (*s++ & 0x0F));
            d++;
            width--;
            }

        // Block copy the middle bytes DWORD's at a time.
        // Make sure we force DWORD source alignment
        bytes = width/2;
        while (bytes && ((uint)s & 3) != 0) {
            *d++ = *s++;
            bytes--;
            }
        if (bytes) {
            memcpy(d,s,bytes);
            d += bytes;
            s += bytes;
            }

        // Handle last odd pixel if any
        if (width & 1)
            *d = (uchar)((*d & 0x0F) | (*s & 0xF0));
        }
    else {
        // Handle odd source or destination pixels
        if (dstLeft & 1) {
            *d = (*d & 0xF0) | (*s >> 4);
            d++;
            width--;
            }
        residual = (*s++ << 4);

        // Copy pixels for all middle bytes
        bytes = width/2;
        while (bytes--) {
            *d++ = (*s >> 4) | residual;
            residual = (*s++ << 4);
            }

        // Handle last odd source pixel
        if (width & 1)
            *d = (*d & 0x0F) | residual;
        }
}

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_bltScan_4##rop (uchar *d,uchar *s,int srcLeft,int dstLeft,int width)
#include "raster/bltscan4.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_bltScanPatt_4 (uchar *din,uchar *sin,int srcLeft,int x1,int y,int width)
#include "raster/bltmono.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_bltScanColorPatt_4 (uchar *din,uchar *sin,int srcLeft,int x1,int y,int width)
#include "raster/bltcolor.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_srcTransBltScan_4##rop (uchar *din,uchar *sin,int srcLeft,int dstLeft,int width)
#include "raster/transcan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_dstTransBltScan_4##rop (uchar *din,uchar *sin,int srcLeft,int dstLeft,int width)
#include "raster/dstscan.c"

/****************************************************************************
PARAMETERS:
din     - Destination memory address
sin     - Source memory address
width   - Number of pixels to copy

REMARKS:
This function stretchs a scanline by a factor of 2 in the X direction.
****************************************************************************/
void REF2D_stretchScan2x_4_S(
    uchar *d,
    uchar *s,
    int srcLeft,
    int dstLeft,
    int width)
{
    while (width--) {
        uchar c = READ_PIXEL(*s,srcLeft);
        REPLACE_PIXEL(*d,dstLeft,c);
        INC_DEST(dstLeft,d);
        dstLeft++;
        REPLACE_PIXEL(*d,dstLeft,c);
        INC_DEST(dstLeft,d);
        dstLeft++;
        INC_DEST(srcLeft,s);
        srcLeft++;
        }
}

/****************************************************************************
PARAMETERS:
din     - Destination memory address
sin     - Source memory address
width   - Number of pixels to copy

REMARKS:
This function stretchs a scanline by a factor of 2 in the X direction.
****************************************************************************/
void REF2D_stretchScan2x_4_Generic(
    uchar *d,
    uchar *s,
    int srcLeft,
    int dstLeft,
    int width)
{
    while (width--) {
        uchar c = READ_PIXEL(*s,srcLeft);
        DC.drawPixel(d,dstLeft,c);
        INC_DEST(dstLeft,d);
        dstLeft++;
        DC.drawPixel(d,dstLeft,c);
        INC_DEST(dstLeft,d);
        dstLeft++;
        INC_DEST(srcLeft,s);
        srcLeft++;
        }
}

/****************************************************************************
PARAMETERS:
din         - Destination memory address
sin         - Source memory address
width       - Number of pixels to copy
repCounts   - Array of pixel replication counts

REMARKS:
This function stretchs a scanline by a generic factor in the X direction.
****************************************************************************/
void REF2D_stretchScan_4_S(
    uchar *d,
    uchar *s,
    int srcLeft,
    int dstLeft,
    int width,
    int *repCounts)
{
    while (width--) {
        int i = *repCounts++;
        uchar c = READ_PIXEL(*s,srcLeft);
        while (i--) {
            REPLACE_PIXEL(*d,dstLeft,c);
            INC_DEST(dstLeft,d);
            dstLeft++;
            }
        INC_DEST(srcLeft,s);
        srcLeft++;
        }
}

/****************************************************************************
PARAMETERS:
din         - Destination memory address
sin         - Source memory address
width       - Number of pixels to copy
repCounts   - Array of pixel replication counts

REMARKS:
This function stretchs a scanline by a generic factor in the X direction.
****************************************************************************/
void REF2D_stretchScan_4_Generic(
    uchar *d,
    uchar *s,
    int srcLeft,
    int dstLeft,
    int width,
    int *repCounts)
{
    while (width--) {
        int i = *repCounts++;
        uchar c = READ_PIXEL(*s,srcLeft);
        while (i--) {
            DC.drawPixel(d,dstLeft,c);
            INC_DEST(dstLeft,d);
            dstLeft++;
            }
        INC_DEST(srcLeft,s);
        srcLeft++;
        }
}

#define DRAW_LINEXPOSLO_PROTO(rop) \
    void REF2D_drawLineXPosLo_4##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXPOSHI_PROTO(rop) \
    void REF2D_drawLineXPosHi_4##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXNEGLO_PROTO(rop) \
    void REF2D_drawLineXNegLo_4##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXNEGHI_PROTO(rop) \
    void REF2D_drawLineXNegHi_4##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr)
#include "raster/line.c"

#define DRAW_LINEXPOSLO_PROTO(rop) \
    void REF2D_drawStippleLineXPosLo_4##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#define DRAW_LINEXPOSHI_PROTO(rop) \
    void REF2D_drawStippleLineXPosHi_4##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#define DRAW_LINEXNEGLO_PROTO(rop) \
    void REF2D_drawStippleLineXNegLo_4##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#define DRAW_LINEXNEGHI_PROTO(rop) \
    void REF2D_drawStippleLineXNegHi_4##rop (uchar *addr,int x,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#include "raster/stipline.c"

#define DRAW_LINEXPOSLO_PROTO(rop) \
    void REF2D_drawStyleLineXPosLo_4##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#define DRAW_LINEXPOSHI_PROTO(rop) \
    void REF2D_drawStyleLineXPosHi_4##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#define DRAW_LINEXNEGLO_PROTO(rop) \
    void REF2D_drawStyleLineXNegLo_4##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#define DRAW_LINEXNEGHI_PROTO(rop) \
    void REF2D_drawStyleLineXNegHi_4##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#include "raster/stylline.c"

/****************************************************************************
PARAMETERS:
srcPal      - Source bitmap palette to build translation vector for
dstPal      - Destination bitmap palette to build translation vector for
srcColor    - Number of colors in source palette

REMARKS:
This function builds a translation vector for translating 4bpp or 8bpp
color values to the destination device color values, by finding the closest
color in the destination palette that matches the color in the source
palette.
****************************************************************************/
GA_color * REF2D_buildTranslateVector_4(
    GA_color *t,
    GA_palette *dstPal,
    GA_palette *srcPal,
    int srcColors)
{
    GA_color *p = t;

    while (srcColors--) {
        int         j,closest = 0,d,distance = 0x7FFFL;
        GA_palette  *dp = dstPal;

        for (j = 0; j < 16; j++,dp++) {
            if ((d = ABS((int)srcPal->Red - (int)dp->Red)
                   + ABS((int)srcPal->Green - (int)dp->Green)
                   + ABS((int)srcPal->Blue - (int)dp->Blue)) < distance) {
                distance = d;
                closest = j;
                }
            }
        *p++ = closest;
        srcPal++;
        }
    return t;
}

/****************************************************************************
PARAMETERS:
d           - Pointer to destination bitmap scanline
s           - Pointer to source bitmap scanline
x           - X coordinate of first destination pixel
width       - Number of pixels to convert
odd         - True if the first source pixel is odd
t           - Palette translation vector for destination bitmap
repCounts   - replication counts if this is to be stretched NULL if not
****************************************************************************/
void REF2D_convScan4_4_S(
    uchar *d,
    uchar *s,
    int x,
    int width,
    int odd,
    GA_color *t)
{
    uchar   c;

    if (odd) {
        c = t[*s++ & 0xF];          // Handle first odd pixel
        REPLACE_PIXEL(*d,x,c);
        INC_DEST(x,d);
        x++;
        width--;
        }
    while (width--) {
        c = t[*s >> 4];
        REPLACE_PIXEL(*d,x,c);
        INC_DEST(x,d);
        x++;
        if (width-- == 0) break;
        c = t[*s++ & 0xF];
        REPLACE_PIXEL(*d,x,c);
        INC_DEST(x,d);
        x++;
        }
}

/****************************************************************************
PARAMETERS:
d           - Pointer to destination bitmap scanline
s           - Pointer to source bitmap scanline
width       - Number of pixels to convert
t           - Palette translation vector for destination bitmap
repCounts   - replication counts if this is to be stretched NULL if not
****************************************************************************/
void REF2D_convScan8_4_S(
    uchar *d,
    uchar *s,
    int x,
    int width,
    GA_color *t)
{
    uchar   c;

    while (width--) {
        c = t[*s++];
        REPLACE_PIXEL(*d,x,c);
        INC_DEST(x,d);
        x++;
        }
}

/****************************************************************************
PARAMETERS:
ileft   - Left coorindate of ellipse rectangle
itop    - Top coodinate of ellipse rectangle
A       - Major axis
B       - Minor axis
****************************************************************************/
void REF2D_drawEllipse_4_Generic(
    int ileft,
    int itop,
    int A,
    int B)
{
    #define SET4PIXELS()                                                \
        {                                                               \
            if (inc_x) {                                                \
                if (right++ & 1) {                                      \
                    pTopRight++;                                        \
                    pBotRight++;                                        \
                    }                                                   \
                if (!(left-- & 1)) {                                    \
                    pTopLeft--;                                         \
                    pBotLeft--;                                         \
                    }                                                   \
                }                                                       \
            if (inc_y) {                                                \
                BINC(pTopLeft,DC.mi.bytesPerLine);                      \
                BINC(pTopRight,DC.mi.bytesPerLine);                     \
                top++;                                                  \
                BDEC(pBotLeft,DC.mi.bytesPerLine);                      \
                BDEC(pBotRight,DC.mi.bytesPerLine);                     \
                bottom--;                                               \
                }                                                       \
            DC.drawPixel((uchar*)pTopLeft X_ARG(left),color);           \
            if (left != right)                                          \
                DC.drawPixel((uchar*)pTopRight X_ARG(right),color);     \
            if (bottom != top) {                                        \
                DC.drawPixel((uchar*)pBotLeft X_ARG(left),color);       \
                if (left != right)                                      \
                    DC.drawPixel((uchar*)pBotRight X_ARG(right),color); \
                }                                                       \
        }
    #include "raster/inner/ellipse.c"
}

/****************************************************************************
PARAMETERS:
ileft   - Left coorindate of ellipse rectangle
itop    - Top coodinate of ellipse rectangle
A       - Major axis
B       - Minor axis
****************************************************************************/
void REF2D_drawEllipse_4_S(
    int ileft,
    int itop,
    int A,
    int B)
{
    #define SET4PIXELS()                                                \
        {                                                               \
            if (inc_x) {                                                \
                if (right++ & 1) {                                      \
                    pTopRight++;                                        \
                    pBotRight++;                                        \
                    }                                                   \
                if (!(left-- & 1)) {                                    \
                    pTopLeft--;                                         \
                    pBotLeft--;                                         \
                    }                                                   \
                }                                                       \
            if (inc_y) {                                                \
                BINC(pTopLeft,DC.mi.bytesPerLine);                      \
                BINC(pTopRight,DC.mi.bytesPerLine);                     \
                top++;                                                  \
                BDEC(pBotLeft,DC.mi.bytesPerLine);                      \
                BDEC(pBotRight,DC.mi.bytesPerLine);                     \
                bottom--;                                               \
                }                                                       \
            REPLACE_PIXEL(*pTopLeft,left,(COLOR_TYPE)color);            \
            REPLACE_PIXEL(*pTopRight,right,(COLOR_TYPE)color);          \
            REPLACE_PIXEL(*pBotLeft,left,(COLOR_TYPE)color);            \
            REPLACE_PIXEL(*pBotRight,right,(COLOR_TYPE)color);          \
        }
    #include "raster/inner/ellipse.c"
}

/****************************************************************************
PARAMETERS:
ileft       - Left coorindate of ellipse rectangle
itop        - Top coodinate of ellipse rectangle
A           - Major axis
B           - Minor axis
clipLeft    - Left coordinate for clip rectangle
clipTop     - Top coordinate for clip rectangle
clipRight   - Right coordinate for clip rectangle
clipBottom  - Bottom coordinate for clip rectangle
****************************************************************************/
void REF2D_clipEllipse_4_Generic(
    int ileft,
    int itop,
    int A,
    int B,
    int clipLeft,
    int clipTop,
    int clipRight,
    int clipBottom)
{
    #define SET4PIXELS()                                                    \
        {                                                                   \
            if (inc_x) {                                                    \
                if (right++ & 1) {                                          \
                    pTopRight++;                                            \
                    pBotRight++;                                            \
                    }                                                       \
                if (!(left-- & 1)) {                                        \
                    pTopLeft--;                                             \
                    pBotLeft--;                                             \
                    }                                                       \
                }                                                           \
            if (inc_y) {                                                    \
                BINC(pTopLeft,DC.mi.bytesPerLine);                          \
                BINC(pTopRight,DC.mi.bytesPerLine);                         \
                top++;                                                      \
                BDEC(pBotLeft,DC.mi.bytesPerLine);                          \
                BDEC(pBotRight,DC.mi.bytesPerLine);                         \
                bottom--;                                                   \
                }                                                           \
            if (INSIDE_CLIPRECT(left,top))                                  \
                DC.drawPixel((uchar*)pTopLeft X_ARG(ileft),color);          \
            if (INSIDE_CLIPRECT(right,top) && pTopLeft != pTopRight)        \
                DC.drawPixel((uchar*)pTopRight X_ARG(ileft),color);         \
            if (pTopLeft != pBotLeft) {                                     \
                if (INSIDE_CLIPRECT(left,bottom))                           \
                    DC.drawPixel((uchar*)pBotLeft X_ARG(ileft),color);      \
                if (INSIDE_CLIPRECT(right,bottom) && pBotLeft != pBotRight) \
                    DC.drawPixel((uchar*)pBotRight X_ARG(ileft),color);     \
                }                                                           \
            if (INSIDE_CLIPRECT(left,top))                                  \
                DC.drawPixel((uchar*)pTopLeft X_ARG(left),color);           \
            if (INSIDE_CLIPRECT(right,top) && left != right)                \
                DC.drawPixel((uchar*)pTopRight X_ARG(right),color);         \
            if (bottom != top) {                                            \
                if (INSIDE_CLIPRECT(left,bottom))                           \
                    DC.drawPixel((uchar*)pBotLeft X_ARG(left),color);       \
                if (INSIDE_CLIPRECT(right,bottom) && left != right)         \
                    DC.drawPixel((uchar*)pBotRight X_ARG(right),color);     \
                }                                                           \
        }
    #include "raster/inner/ellipse.c"
}

#endif

