/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Region memory management routines. Our shape algebra is
*               implemented using linked lists of spans and segments with
*               the structures allocated dymanically from large arrays
*               managed with a free list.
*
*               If we run out of local memory in the span and segment pools,
*               we attempt to allocate more memory to the region pools on
*               the fly. This allows us to start off with a reasonably small
*               amount of memory, and expand if the application demand for
*               complex regions is higher than normal.
*
****************************************************************************/

#include "regint.h"

/*--------------------------- Global variables ----------------------------*/

RGN_state   RGN_devCtx;

/*------------------------- Implementation --------------------------------*/

/****************************************************************************
REMARKS:
Create the region management block lists, initialises all internal
data structures and returns the exports function list.
****************************************************************************/
static ibool RGN_init(
    GA_regionFuncs *funcs)
{
    GA_regionFuncs exp = {
        sizeof(GA_regionFuncs),
        RGN_NewRegion,
        RGN_NewRectRegion,
        RGN_CopyRegion,
        RGN_CopyIntoRegion,
        RGN_ClearRegion,
        RGN_FreeRegion,
        RGN_SectRegion,
        RGN_SectRegionRect,
        RGN_UnionRegion,
        RGN_UnionRegionRect,
        RGN_UnionRegionOfs,
        RGN_DiffRegion,
        RGN_DiffRegionRect,
        RGN_OptimizeRegion,
        RGN_OffsetRegion,
        RGN_IsEmptyRegion,
        RGN_IsEqualRegion,
        RGN_PtInRegion,
        RGN_TraverseRegion,
        };

    // Initialise device driver internals
    memset(&RG,0,sizeof(RG));

    // Create all the block lists
    if (!N_createBlockList(&RG.regionList.bl,sizeof(GA_region)))
        return false;
    if (!N_createBlockList(&RG.spanList.bl,sizeof(GA_span)))
        return false;
    if (!N_createBlockList(&RG.segList.bl,sizeof(GA_segment)))
        return false;
    RG.regionList.freeList = N_buildFreeList(&RG.regionList.bl);
    RG.spanList.freeList = N_buildFreeList(&RG.spanList.bl);
    RG.segList.freeList = N_buildFreeList(&RG.segList.bl);

    // Initialise internal global variables
    RG.tmpX1.next = &RG.tmpX2;
    RG.tmpY1.next = &RG.tmpY2;
    RG.tmpY1.seg  = &RG.tmpX1;
    RG.rtmp.spans = &RG.tmpY1;
    COPY_STRUCTURE(funcs,&exp);
    return true;
}

/****************************************************************************
REMARKS:
Cleans up and exits, freely all memory used by the region manager.
****************************************************************************/
static void RGN_exit(void)
{
    N_freeBlockList(&RG.segList.bl);
    N_freeBlockList(&RG.spanList.bl);
    N_freeBlockList(&RG.regionList.bl);
}

#ifdef  __DRIVER__
/****************************************************************************
REMARKS:
Library initialisation entry point.
****************************************************************************/
ibool _CEXPORT RGN_initLibrary(
    GA_regionFuncs *funcs)
{
    return RGN_init(funcs);
}

/****************************************************************************
REMARKS:
Library exit entry point.
****************************************************************************/
void _CEXPORT RGN_exitLibrary(void)
{
    RGN_exit();
}
#else
/****************************************************************************
REMARKS:
Function entry point to fake loading of a driver when linked into app.
****************************************************************************/
ibool NAPI GA_loadRegionMgr(
    GA_regionFuncs *funcs,
    PE_MODULE **hModule,
    ulong *size)
{
    (void)hModule;
    (void)size;
    return RGN_init(funcs);
}

/****************************************************************************
REMARKS:
Function entry point to fake unloading of a driver when linked into app.
****************************************************************************/
void NAPI GA_unloadRegionMgr(
    PE_MODULE *hModule)
{
    (void)hModule;
    RGN_exit();
}
#endif

/****************************************************************************
REMARKS:
Resizes the region list if full by increasing its size by the original
default size and building a new free list in the resized portion.
****************************************************************************/
ibool RGN_resizeRegionList(void)
{
    if (!N_resizeBlockList(&RG.regionList.bl))
        return false;
    RG.regionList.freeList = N_buildFreeList(&RG.regionList.bl);
    return true;
}

/****************************************************************************
REMARKS:
Resizes the span list if full by increasing its size by the original
default size and building a new free list in the resized portion.
****************************************************************************/
ibool RGN_resizeSpanList(void)
{
    if (!N_resizeBlockList(&RG.spanList.bl))
        return false;
    RG.spanList.freeList = N_buildFreeList(&RG.spanList.bl);
    return true;
}

/****************************************************************************
REMARKS:
Resizes the segment list if full by increasing its size by the original
default size and building a new free list in the resized portion.
****************************************************************************/
ibool RGN_resizeSegmentList(void)
{
    if (!N_resizeBlockList(&RG.segList.bl))
        return false;
    RG.segList.freeList = N_buildFreeList(&RG.segList.bl);
    return true;
}

/****************************************************************************
PARAMETERS:
r   - Region to expand

RETURNS:
Pointer to the first span in the region

REMARKS:
Expands the regions from a special case rectangle into a proper rectangular
region.
****************************************************************************/
GA_span * RGN_rectRegion(
    GA_region *r)
{
    GA_span     *s1,*s2;
    GA_segment  *seg;

    r->spans = s1 = RGN_newSpan();
    if (!s1) return NULL;
    s1->y = r->rect.top;
    seg = (s1->seg = RGN_newSegment());
    if (!seg) return NULL;
    seg->x = r->rect.left;
    seg = (seg->next = RGN_newSegment());
    if (!seg) return NULL;
    seg->x = r->rect.right;
    seg->next = NULL;
    s1->next = s2 = RGN_newSpan();
    if (!s2) return NULL;
    s2->y = r->rect.bottom;
    s2->seg = NULL;
    s2->next = NULL;
    return s1;
}

/****************************************************************************
RETURNS:
Pointer to the new region, NULL if out of memory.

REMARKS:
Allocates a new complex region. The new region is empty when first created.
Note that MGL maintains a local memory pool for all region allocations in
order to provide the maximum speed and minimum memory overheads for region
allocations.
****************************************************************************/
GA_region * NAPI RGN_NewRegion(void)
{
    GA_region   *p;

    if (RG.regionList.freeList == NULL) {
        if (!RGN_resizeRegionList())
            return NULL;
        }
    p = RG.regionList.freeList;
    RG.regionList.freeList = N_FREELIST_NEXT(p);
    memset(p,0,sizeof(GA_region));
    return p;
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle
top     - Top coordinate of the rectangle
right   - Right coordinate of the rectangle
bottom  - Bottom coordinate of the rectangle

RETURNS:
New region generated, NULL if out of memory.

REMARKS:
This function generates a solid rectangle as a region.
****************************************************************************/
GA_region * NAPI RGN_NewRectRegion(
    N_int32 left,
    N_int32 top,
    N_int32 right,
    N_int32 bottom)
{
    GA_region *r = RGN_NewRegion();
    if (!r) return NULL;
    r->rect.left = left;
    r->rect.top = top;
    r->rect.right = right;
    r->rect.bottom = bottom;
    r->spans = NULL;
    return r;
}

/****************************************************************************
PARAMETERS:
s   - Pointer to source region

RETURNS:
Pointer to the copied region, or NULL if out of memory.

REMARKS:
Copies the definition for an entire region and returns a pointer to the
newly created region. The space for the copied region is allocated from
the region memory pool, which MGL uses to maintain a local memory
allocation scheme for regions to increase performance.

If there is not enough memory to copy the region, this routine will
return NULL.
****************************************************************************/
GA_region * NAPI RGN_CopyRegion(
    const GA_region *s)
{
    GA_region   *region;

    if ((region = RGN_NewRegion()) == NULL)
        return NULL;
    *region = *s;
    region->spans = RGN_copySpans(s->spans);
    if (!region->spans) {
        RGN_FreeRegion(region);
        return NULL;
        }
    return region;
}

/****************************************************************************
PARAMETERS:
d   - Pointer to destination region
s   - Pointer to source region

REMARKS:
Copies the definition for an entire region into the destination region,
clearing any region information already present in the destination. This
function is similar to RGN_copyRegion, however it does not allocate a new
region but rather copies the data into an existing region.
****************************************************************************/
void NAPI RGN_CopyIntoRegion(
    GA_region *d,
    const GA_region *s)
{
    RGN_ClearRegion(d);
    *d = *s;
    d->spans = RGN_copySpans(s->spans);
}

/****************************************************************************
PARAMETERS:
r   - Pointer to the region to free

REMARKS:
Frees all the memory allocated by the complex region. When you are finished
with a complex region you must free it to free up the memory used to
represent the union of rectangles.
****************************************************************************/
void NAPI RGN_FreeRegion(
    GA_region *r)
{
    GA_span *s,*p;

    if (!r)
        return;
    for (s = r->spans; s; s = p) {
        p = s->next;
        RGN_freeSpan(s);
        }
    N_FREELIST_NEXT(r) = RG.regionList.freeList;
    RG.regionList.freeList = r;
}

/****************************************************************************
PARAMETERS:
r   - region to be cleared

REMARKS:
This function clears the specified region to an empty region, freeing up
all the memory used to store the region data.
****************************************************************************/
void NAPI RGN_ClearRegion(
    GA_region *r)
{
    GA_span *s,*p;

    for (s = r->spans; s; s = p) {
        p = s->next;
        RGN_freeSpan(s);
        }
    memset(r,0,sizeof(GA_region));
}

/****************************************************************************
PARAMETERS:
r   - Region to optimize

REMARKS:
This function optimizes the specified region by traversing the region
structure looking for identical spans in the region. The region algebra
functions (RGN_unionRegion, RGN_diffRegion, RGN_sectRegion etc.) do not
fully optimize the resulting region to save time, so after you have
created a complex region you may wish to call this routine to optimize it.

Optimizing the region will find the minimum number of rectangles required
to represent that region, and will result in faster drawing and traversing
of the resulting region.
****************************************************************************/
void NAPI RGN_OptimizeRegion(
    GA_region *r)
{
    GA_span *s = r->spans,*prev,*next;

    if (!s)
        return;

    prev = s;
    for (s = s->next; s; s = next) {
        GA_segment *seg = s->seg;
        GA_segment *segPrev = prev->seg;
        ibool identical = (seg && segPrev);

        /* Check if spans are identical */
        while (seg && segPrev) {
            if (seg->x != segPrev->x) {
                identical = false;
                break;
                }
            seg = seg->next;
            segPrev = segPrev->next;
            }

        /* Unlink the identical span from the list */
        if (identical && !seg && !segPrev) {
            next = prev->next = s->next;
            RGN_freeSpan(s);
            }
        else {
            next = s->next;
            prev = s;
            }
        }
}

