/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Region memory management routines. Our shape algebra is
*               implemented using linked lists of spans and segments with
*               the structures allocated dymanically from large arrays
*               managed with a free list.
*
*               If we run out of local memory in the span and segment pools,
*               we attempt to allocate more memory to the region pools on
*               the fly. This allows us to start off with a reasonably small
*               amount of memory, and expand if the application demand for
*               complex regions is higher than normal.
*
****************************************************************************/

#include "ref2dint.h"

/*------------------------- Implementation --------------------------------*/

/****************************************************************************
RETURNS:
Pointer to new segment

REMARKS:
Allocates space for a new segment in the segment pool from the free list.
If there are no more segments, we attempt to re-size the segment pool
which will bomb out if we run out of memory.
****************************************************************************/
GA_segment *RGN_newSegment(void)
{
    GA_segment  *p;

    if (RG.segList.freeList == NULL) {
        if (!RGN_resizeSegmentList())
            return NULL;
        }
    p = RG.segList.freeList;
    RG.segList.freeList = N_FREELIST_NEXT(p);
    return p;
}

/****************************************************************************
PARAMETERS:
s   - Pointer to source segment

RETURNS:
Pointer to copied segment

REMARKS:
Copies the definitions for an entire segment list and returns the pointer
of the first segment in the list. The space for the copied segment is
allocated from the memory pool.
****************************************************************************/
GA_segment * RGN_copySegment(
    GA_segment *s)
{
    GA_segment  *head,*p,**prev;

    for (prev = &head; s; s = s->next) {
        *prev = p = RGN_newSegment();
        if (!p) return NULL;
        p->x = s->x;
        prev = &p->next;
        }
    *prev = NULL;
    return head;
}

/****************************************************************************
PARAMETERS:
s       - Pointer to source segment
xOffset - X coordinate offset value

RETURNS:
Pointer to copied segment

REMARKS:
Copies the definitions for an entire segment list and returns the pointer
of the first segment in the list. The space for the copied segment is
allocated from the memory pool. This version offsets all the X coordinates
in the segment list by the specified X offset value.
****************************************************************************/
GA_segment * RGN_copySegmentOfs(
    GA_segment *s,
    int xOffset)
{
    GA_segment  *head,*p,**prev;

    for (prev = &head; s; s = s->next) {
        *prev = p = RGN_newSegment();
        if (!p) return NULL;
        p->x = xOffset + s->x;
        prev = &p->next;
        }
    *prev = NULL;
    return head;
}

