/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Region memory management routines. Our shape algebra is
*               implemented using linked lists of spans and segments with
*               the structures allocated dymanically from large arrays
*               managed with a free list.
*
*               If we run out of local memory in the span and segment pools,
*               we attempt to allocate more memory to the region pools on
*               the fly. This allows us to start off with a reasonably small
*               amount of memory, and expand if the application demand for
*               complex regions is higher than normal.
*
****************************************************************************/

#include "ref2dint.h"

/*------------------------- Implementation --------------------------------*/

/****************************************************************************
RETURNS:
Pointer to new span

REMARKS:
Allocates space for a new span in the span pool from the free list. If
there are no more spans, we attempt to re-size the span pool which will
bomb out if we run out of memory.
****************************************************************************/
GA_span * RGN_newSpan(void)
{
    GA_span *p;

    if (RG.spanList.freeList == NULL) {
        if (!RGN_resizeSpanList())
            return NULL;
        }
    p = RG.spanList.freeList;
    RG.spanList.freeList = N_FREELIST_NEXT(p);
    return p;
}

/****************************************************************************
PARAMETERS:
s   - Pointer to source span

RETURNS:
Pointer to copied span

REMARKS:
Copies the definitions for a since span and returns the pointer to
the new span. The space for the copied span is allocated from the
memory pool.
****************************************************************************/
GA_span * RGN_copySpan(
    GA_span *s)
{
    GA_span *span = RGN_newSpan();
    if (!span) return NULL;
    span->y = s->y;
    span->seg = RGN_copySegment(s->seg);
    if (!span->seg) return NULL;
    span->next = NULL;
    return span;
}

/****************************************************************************
PARAMETERS:
s       - Pointer to source span
xOffset - X coordinate offset value
yOffset - Y coordinate offset value

RETURNS:
Pointer to copied span

REMARKS:
Copies the definitions for a since span and returns the pointer to
the new span. The space for the copied span is allocated from the
memory pool. This version offsets all the X and Y coordinates
in the span and segment lists by the specified offset values.
****************************************************************************/
GA_span * RGN_copySpanOfs(
    GA_span *s,
    int xOffset,
    int yOffset)
{
    GA_span *span = RGN_newSpan();
    if (!span) return NULL;
    span->y = yOffset + s->y;
    span->seg = RGN_copySegmentOfs(s->seg,xOffset);
    if (!span->seg) return NULL;
    span->next = NULL;
    return span;
}

/****************************************************************************
PARAMETERS:
s   - Pointer to source spans

RETURNS:
Pointer to copied spans

REMARKS:
Copies the definitions for an entire span list and returns the pointer to
the first span in the list. The space for the copied spans is allocated
from the memory pool.
****************************************************************************/
GA_span * RGN_copySpans(
    GA_span *s)
{
    GA_span *p,**prev,*head;

    for (head = NULL,prev = &head; s; s = s->next) {
        *prev = p = RGN_newSpan();
        if (!p) return NULL;
        p->y = s->y;
        p->seg = RGN_copySegment(s->seg);
        if (!p->seg) return NULL;
        prev = &p->next;
        }
    *prev = NULL;
    return head;
}

/****************************************************************************
PARAMETERS:
s       - Pointer to source spans
xOffset - X coordinate offset value
yOffset - Y coordinate offset value

RETURNS:
Pointer to copied spans

REMARKS:
Copies the definitions for an entire span list and returns the pointer to
the first span in the list. The space for the copied spans is allocated
from the memory pool. This version offsets all the X and Y coordinates
in the span and segment lists by the specified offset values.
****************************************************************************/
GA_span * RGN_copySpansOfs(
    GA_span *s,
    int xOffset,
    int yOffset)
{
    GA_span *p,**prev,*head;

    for (head = NULL,prev = &head; s; s = s->next) {
        *prev = p = RGN_newSpan();
        if (!p) return NULL;
        p->y = yOffset + s->y;
        p->seg = RGN_copySegmentOfs(s->seg,xOffset);
        if (!p->seg) return NULL;
        prev = &p->next;
        }
    *prev = NULL;
    return head;
}

/****************************************************************************
PARAMETERS:
span    - Pointer to the span to free

REMARKS:
Frees all the memory allocated by the span by moving though the list of
segments freeing each one.
****************************************************************************/
void RGN_freeSpan(
    GA_span *span)
{
    GA_segment *s,*p;

    for (s = span->seg; s; s = p) {
        p = s->next;
        N_FREELIST_NEXT(s) = RG.segList.freeList;
        RG.segList.freeList = s;
        }
    N_FREELIST_NEXT(span) = RG.spanList.freeList;
    RG.spanList.freeList = span;
}

/****************************************************************************
PARAMETERS:
y   - Y coordinate for span
x1  - X1 coordinate for span
x2  - X2 coordinate for span

RETURNS:
Pointer to the new span

REMARKS:
Function to create a single scanline span. This should be turned into a
macro for speed.
****************************************************************************/
GA_span * RGN_createSpan(
    int y,
    int x1,
    int x2)
{
    GA_segment  *seg;
    GA_span         *st;

    st = RGN_newSpan();
    if (!st) return NULL;
    st->y = y;
    seg = (st->seg = RGN_newSegment());
    if (!seg) return NULL;
    seg->x = x1;
    seg = (seg->next = RGN_newSegment());
    if (!seg) return NULL;
    seg->x = x2;
    seg->next = NULL;
    return st;
}

/****************************************************************************
PARAMETERS:
y   - Y coordinate for span
x1  - X1 coordinate for first segment
x2  - X2 coordinate for first segment
x3  - X1 coordinate for second segment
x4  - X2 coordinate for second segment

RETURNS:
Pointer to the new span

REMARKS:
Function to create a double scanline span.
****************************************************************************/
GA_span * RGN_createSpan2(
    int y,
    int x1,
    int x2,
    int x3,
    int x4)
{
    GA_segment  *seg;
    GA_span         *st;

    st = RGN_newSpan();
    if (!st) return NULL;
    st->y = y;
    seg = (st->seg = RGN_newSegment());
    if (!seg) return NULL;
    seg->x = x1;
    seg = (seg->next = RGN_newSegment());
    if (!seg) return NULL;
    seg->x = x2;
    seg = (seg->next = RGN_newSegment());
    if (!seg) return NULL;
    seg->x = x3;
    seg = (seg->next = RGN_newSegment());
    if (!seg) return NULL;
    seg->x = x4;
    seg->next = NULL;
    return st;
}

