/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the shadow buffer filter driver support
*               routines in the SNAP Reference Rasteriser.
*
****************************************************************************/

#include "shadint.h"

/*----------------------------- Implementation ----------------------------*/

#define PickBit(a,i)    (((a) >> (i)) & 1)

/****************************************************************************
REMARKS:
Write a scanline of bytes to the hardware.
****************************************************************************/
static void WriteBytes(
    uchar *d,
    uchar *s,
    int bytes)
{
    int plane,SEQ_DATA = 0x3C5;

    // Process scanline a plane at a time, writing 32 bits at a time
    for (plane = 0; plane < 4; plane++) {
        GA_color    s0,s1,s2,s3;
        ulong       *sha = (ulong*)s;
        ulong       *dst = (ulong*)d;
        int         count = (bytes+3)/4;

        outpb(SEQ_DATA,0x1 << plane);
        while (count--) {
            s0 = *sha++ >> plane;
            s0 = ((PickBit(s0,4)  << 7) |
                  (PickBit(s0,0)  << 6) |
                  (PickBit(s0,12) << 5) |
                  (PickBit(s0,8)  << 4) |
                  (PickBit(s0,20) << 3) |
                  (PickBit(s0,16) << 2) |
                  (PickBit(s0,28) << 1) |
                  (PickBit(s0,24) << 0));
            s1 = *sha++ >> plane;
            s1 = ((PickBit(s1,4)  << 7) |
                  (PickBit(s1,0)  << 6) |
                  (PickBit(s1,12) << 5) |
                  (PickBit(s1,8)  << 4) |
                  (PickBit(s1,20) << 3) |
                  (PickBit(s1,16) << 2) |
                  (PickBit(s1,28) << 1) |
                  (PickBit(s1,24) << 0));
            s2 = *sha++ >> plane;
            s2 = ((PickBit(s2,4)  << 7) |
                  (PickBit(s2,0)  << 6) |
                  (PickBit(s2,12) << 5) |
                  (PickBit(s2,8)  << 4) |
                  (PickBit(s2,20) << 3) |
                  (PickBit(s2,16) << 2) |
                  (PickBit(s2,28) << 1) |
                  (PickBit(s2,24) << 0));
            s3 = *sha++ >> plane;
            s3 = ((PickBit(s3,4)  << 7) |
                  (PickBit(s3,0)  << 6) |
                  (PickBit(s3,12) << 5) |
                  (PickBit(s3,8)  << 4) |
                  (PickBit(s3,20) << 3) |
                  (PickBit(s3,16) << 2) |
                  (PickBit(s3,28) << 1) |
                  (PickBit(s3,24) << 0));
            *dst++ = s0 | (s1 << 8) | (s2 << 16) | (s3 << 24);
            }
        }
}

/****************************************************************************
REMARKS:
Updates the screen by copying a portion of the shadow buffer the screen.
****************************************************************************/
void SHADOW_UpdateScreen4(
    int offset,
    int left,
    int top,
    int width,
    int height)
{
    int     bytes,startAddr,endAddr,newAddr,bank,bytesStart,bytesEnd,right,srcStride;
    int     GC_INDEX = 0x3CE;
    uchar   *s,*d,*bankedMem;
    int     SEQ_INDEX = 0x3C4;
    int     SEQ_DATA = 0x3C5;

    // Setup the VGA hardware (write mode 0, read mode 0, replace mix)
    outpw(GC_INDEX,0x0001);
    outpw(GC_INDEX,0x0003);
    outpw(GC_INDEX,0x0005);
    outpw(GC_INDEX,0x0F07);
    outpw(GC_INDEX,0xFF08);

    // Compute framebuffer address (aligned to DWORD boundaries for speed!)
    right = (left + width + 31) & ~31;
    left = left & ~31;
    bytes = (right - left)/8;
    s = (uchar*)DC.buffer + ((DC.mi.XResolution * top) + left)/2;
    startAddr = (DC.mi.BytesPerScanLine * top) + left/8;
    bank = (startAddr += offset) >> 16;
    bankedMem = DC.hwCtx->BankedMem;
    d = bankedMem + (startAddr & 0xFFFF);
    srcStride = DC.mi.XResolution/2;

    // Now copy the data
    if (DC.cntBank != bank)
        DC.SetBank(DC.cntBank = bank);
    outpb(SEQ_INDEX,0x02);
    while (height--) {
        newAddr = startAddr + DC.mi.BytesPerScanLine;
        if ((bank = newAddr >> 16) != DC.cntBank) {
            // Bank boundary is present before next scanline
            if ((endAddr = startAddr + bytes) >> 16 != DC.cntBank) {
                // Data to copy is split by bank boundary
                bytesStart = (endAddr & ~0xFFFF) - startAddr;
                bytesEnd = endAddr & 0xFFFF;
                WriteBytes(d,s,bytesStart);
                DC.SetBank(DC.cntBank = bank);
                outpb(SEQ_INDEX,0x02);
                WriteBytes(bankedMem,s+(bytesStart*4),bytesEnd);
                }
            else {
                // Data to copy is contained within the current bank
                WriteBytes(d,s,bytes);
                DC.SetBank(DC.cntBank = bank);
                outpb(SEQ_INDEX,0x02);
                }
            d = bankedMem + (newAddr & 0xFFFF);
            }
        else {
            // Scanline is contained within the current bank
            WriteBytes(d,s,bytes);
            d += DC.mi.BytesPerScanLine;
            }
        startAddr += DC.mi.BytesPerScanLine;
        s += srcStride;
        }
    outpb(SEQ_DATA,0xF);
}

/****************************************************************************
REMARKS:
Updates the screen by copying a portion of the shadow buffer the screen.
****************************************************************************/
void SHADOW_UpdateCursor4(
    int offset,
    int left,
    int top,
    int width,
    int height)
{
    SHADOW_UpdateScreen4(offset/8,left,top,width,height);
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI SHADOW_DrawRectExt_4_Solid(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    GA_color color,
    N_int32 mix)
{
    int     bytes,startAddr,endAddr,newAddr,bank,bytesStart,bytesEnd;
    int     right,GC_INDEX = 0x3CE,GC_DATA = 0x3CF;
    uchar   maskLeft,maskRight,*d,*bankedMem,*dd;

    // If this is not a supported mix, copy from the shadow buffer
    if (mix != GA_REPLACE_MIX && mix != GA_AND_MIX && mix != GA_OR_MIX && mix != GA_XOR_MIX) {
        DC.UpdateScreen(DC.buf.Offset,left,top,width,height);
        return;
        }

    // Setup the VGA hardware (write mode 0, read mode 0, passed in mix)
    wrinx(GC_INDEX,0,(uchar)color);
    outpw(GC_INDEX,0x0F01);
    if (mix == GA_REPLACE_MIX)
        outpw(GC_INDEX,0x0003);
    else if (mix == GA_AND_MIX)
        outpw(GC_INDEX,0x0803);
    else if (mix == GA_OR_MIX)
        outpw(GC_INDEX,0x1003);
    else
        outpw(GC_INDEX,0x1803);
    outpw(GC_INDEX,0x0005);
    outpw(GC_INDEX,0x0F07);

    // Compute framebuffer address
    right = left+width-1;
    bytes = (right/8) - (left/8) + 1;
    startAddr = (DC.mi.BytesPerScanLine * top) + left/8;
    bank = (startAddr += DC.buf.Offset) >> 16;
    bankedMem = DC.hwCtx->BankedMem;
    d = bankedMem + (startAddr & 0xFFFF);
    maskLeft = 0xFF >> (left & 7);
    maskRight = 0xFF << ((right & 7) ^ 7);

    // Now fill the rectangle
    if (DC.cntBank != bank)
        DC.SetBank(DC.cntBank = bank);
    outpb(GC_INDEX,0x08);
    if (bytes == 1) {
        // Single byte wide fill. Combine left and right masks and draw
        // vertically a single byte strip
        outpb(GC_DATA,maskLeft & maskRight);
        while (height--) {
            newAddr = startAddr + DC.mi.BytesPerScanLine;
            if ((bank = newAddr >> 16) != DC.cntBank) {
                DC.SetBank(DC.cntBank = bank);
                outpb(GC_INDEX,0x08);
                d = bankedMem + (newAddr & 0xFFFF);
                }
            rwb(d);
            d += DC.mi.BytesPerScanLine;
            startAddr += DC.mi.BytesPerScanLine;
            }
        }
    else {
        while (height--) {
            newAddr = startAddr + DC.mi.BytesPerScanLine;
            if ((bank = newAddr >> 16) != DC.cntBank) {
                // Bank boundary is present before next scanline
                if ((endAddr = startAddr + bytes) >> 16 != DC.cntBank) {
                    // Data to copy is split by bank boundary
                    bytesStart = (endAddr & ~0xFFFF) - startAddr;
                    bytesEnd = endAddr & 0xFFFF;
                    outpb(GC_DATA,maskLeft);
                    rwb(d);
                    outpb(GC_DATA,0xFF);
                    if (bytesStart > 1)
                        rep_stosb(d+1,bytesStart-1,0);
                    DC.SetBank(DC.cntBank = bank);
                    outpb(GC_INDEX,0x08);
                    if (bytesEnd > 1)
                        dd = rep_stosb(bankedMem,bytesEnd-1,0);
                    else
                        dd = bankedMem;
                    outpb(GC_DATA,maskRight);
                    rwb(dd);
                    }
                else {
                    // Data to copy is contained within the current bank
                    outpb(GC_DATA,maskLeft);
                    rwb(d);
                    if (bytes > 2) {
                        outpb(GC_DATA,0xFF);
                        dd = rep_stosb(d+1,bytes-2,0);
                        }
                    else
                        dd = d+1;
                    outpb(GC_DATA,maskRight);
                    rwb(dd);
                    DC.SetBank(DC.cntBank = bank);
                    outpb(GC_INDEX,0x08);
                    }
                d = bankedMem + (newAddr & 0xFFFF);
                }
            else {
                // Scanline is contained within the current bank
                outpb(GC_DATA,maskLeft);
                rwb(d);
                if (bytes > 2) {
                    outpb(GC_DATA,0xFF);
                    dd = rep_stosb(d+1,bytes-2,0);
                    }
                else
                    dd = d+1;
                outpb(GC_DATA,maskRight);
                rwb(dd);
                d += DC.mi.BytesPerScanLine;
                }
            startAddr += DC.mi.BytesPerScanLine;
            }
        }
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI SHADOW_DrawRect_4_Solid(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    SHADOW_DrawRectExt_4_Solid(left,top,width,height,DC.foreColor,DC.mix);
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
****************************************************************************/
void NAPI SHADOW_DrawLineInt_4_Solid(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast)
{
    int absDeltaX,absDeltaY,majorInc,diagInc;
    int initialError,flags;

    flags = gaLineXPositive | gaLineYPositive | gaLineXMajor;
    if (drawLast)
        flags |= gaLineDoLastPel;
    if ((absDeltaX = x2 - x1) < 0) {
        absDeltaX = -absDeltaX;
        flags &= ~gaLineXPositive;
        }
    if ((absDeltaY = y2 - y1) < 0) {
        absDeltaY = -absDeltaY;
        flags &= ~gaLineYPositive;
        }
    if (absDeltaY > absDeltaX) {
        SWAP(absDeltaX,absDeltaY);
        flags &= ~gaLineXMajor;
        }
    majorInc = 2 * absDeltaY;               // 2 * dy
    initialError = majorInc - absDeltaX;    // 2 * dy - dx
    diagInc = initialError - absDeltaX;     // 2 * (dy - dx)
    SHADOW_DrawBresenhamLine_4_Solid(x1,y1,initialError,majorInc,diagInc,absDeltaX+1,flags);
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
****************************************************************************/
void NAPI SHADOW_DrawLineInt_4_Generic(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast)
{
    int absDeltaX,absDeltaY,majorInc,diagInc;
    int initialError,flags;

    flags = gaLineXPositive | gaLineYPositive | gaLineXMajor;
    if (drawLast)
        flags |= gaLineDoLastPel;
    if ((absDeltaX = x2 - x1) < 0) {
        absDeltaX = -absDeltaX;
        flags &= ~gaLineXPositive;
        }
    if ((absDeltaY = y2 - y1) < 0) {
        absDeltaY = -absDeltaY;
        flags &= ~gaLineYPositive;
        }
    if (absDeltaY > absDeltaX) {
        SWAP(absDeltaX,absDeltaY);
        flags &= ~gaLineXMajor;
        }
    majorInc = 2 * absDeltaY;               // 2 * dy
    initialError = majorInc - absDeltaX;    // 2 * dy - dx
    diagInc = initialError - absDeltaX;     // 2 * (dy - dx)
    SHADOW_DrawBresenhamLine_4_Generic(x1,y1,initialError,majorInc,diagInc,absDeltaX+1,flags);
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
****************************************************************************/
void NAPI SHADOW_DrawBresenhamLine_4_Solid(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags)
{
    int     fincr,dstAddr,bank,GC_INDEX = 0x3CE;
    uchar   *dstBase,*d,mask;
    int     mix = DC.mix;

    // If this is not a supported mix, copy from the shadow buffer
    if (mix != GA_REPLACE_MIX && mix != GA_AND_MIX && mix != GA_OR_MIX && mix != GA_XOR_MIX) {
        SHADOW_DrawBresenhamLine_4_Generic(x1,y1,initialError,majorInc,diagInc,count,flags);
        return;
        }

    // Compute framebuffer address
    if (flags & gaLineYPositive)
        fincr = (int)DC.mi.BytesPerScanLine;
    else
        fincr = -((int)DC.mi.BytesPerScanLine);
    dstBase = DC.hwCtx->BankedMem;
    dstAddr = (DC.mi.BytesPerScanLine * y1) + x1/8 + DC.buf.Offset;
    d = dstBase + (dstAddr & 0xFFFF);
    mask = 0x1 << ((x1 & 7) ^ 7);
    if (!(flags & gaLineDoLastPel))
        count--;

    // Setup the VGA hardware (write mode 3, passed in mix)
    wrinx(GC_INDEX,0,(uchar)DC.foreColor);
    outpw(GC_INDEX,0x0F01);
    if (mix == GA_REPLACE_MIX)
        outpw(GC_INDEX,0x0003);
    else if (mix == GA_AND_MIX)
        outpw(GC_INDEX,0x0803);
    else if (mix == GA_OR_MIX)
        outpw(GC_INDEX,0x1003);
    else
        outpw(GC_INDEX,0x1803);
    outpw(GC_INDEX,0x0B05);
    outpw(GC_INDEX,0x0007);
    outpw(GC_INDEX,0xFF08);

    // Now draw the line
    if (flags & gaLineXMajor) {
        if (flags & gaLineXPositive) {
            while (count--) {
                if ((bank = dstAddr >> 16) != DC.cntBank) {
                    DC.SetBank(DC.cntBank = bank);
                    d = dstBase + (dstAddr & 0xFFFF);
                    }
                *d &= mask;
                if (rorb1(&mask)) {
                    d++;
                    dstAddr++;
                    }
                if (initialError < 0)
                    initialError += majorInc;
                else {
                    initialError += diagInc;
                    d += fincr;
                    dstAddr += fincr;
                    }
                }
            }
        else {
            while (count--) {
                if ((bank = dstAddr >> 16) != DC.cntBank) {
                    DC.SetBank(DC.cntBank = bank);
                    d = dstBase + (dstAddr & 0xFFFF);
                    }
                *d &= mask;
                if (rolb1(&mask)) {
                    d--;
                    dstAddr--;
                    }
                if (initialError < 0)
                    initialError += majorInc;
                else {
                    initialError += diagInc;
                    d += fincr;
                    dstAddr += fincr;
                    }
                }
            }
        }
    else {
        if (flags & gaLineXPositive) {
            while (count--) {
                if ((bank = dstAddr >> 16) != DC.cntBank) {
                    DC.SetBank(DC.cntBank = bank);
                    d = dstBase + (dstAddr & 0xFFFF);
                    }
                *d &= mask;
                d += fincr;
                dstAddr += fincr;
                if (initialError < 0)
                    initialError += majorInc;
                else {
                    initialError += diagInc;
                    if (rorb1(&mask)) {
                        d++;
                        dstAddr++;
                        }
                    }
                }
            }
        else {
            while (count--) {
                if ((bank = dstAddr >> 16) != DC.cntBank) {
                    DC.SetBank(DC.cntBank = bank);
                    d = dstBase + (dstAddr & 0xFFFF);
                    }
                *d &= mask;
                d += fincr;
                dstAddr += fincr;
                if (initialError < 0)
                    initialError += majorInc;
                else {
                    initialError += diagInc;
                    if (rolb1(&mask)) {
                        d--;
                        dstAddr--;
                        }
                    }
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
****************************************************************************/
void NAPI SHADOW_DrawBresenhamLine_4_Generic(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags)
{
    int     fincr,sincr,dstAddr,bank,GC_INDEX = 0x3CE,GC_DATA = 0x3CF;
    uchar   *dstBase,*d,*s,mask;

    // Compute framebuffer address
    if (flags & gaLineYPositive) {
        sincr = (int)DC.mi.XResolution/2;
        fincr = (int)DC.mi.BytesPerScanLine;
        }
    else {
        sincr = -((int)DC.mi.XResolution/2);
        fincr = -((int)DC.mi.BytesPerScanLine);
        }
    dstBase = DC.hwCtx->BankedMem;
    dstAddr = (DC.mi.BytesPerScanLine * y1) + x1/8 + DC.buf.Offset;
    d = dstBase + (dstAddr & 0xFFFF);
    s = (uchar*)DC.buffer + ((DC.mi.XResolution * y1) + x1)/2;
    mask = 0x1 << ((x1 & 7) ^ 7);
    if (!(flags & gaLineDoLastPel))
        count--;

    // Setup the VGA hardware (write mode 2, read mode 0, replace mix)
    outpw(GC_INDEX,0x0000);
    outpw(GC_INDEX,0x0F01);
    outpw(GC_INDEX,0x0003);
    outpw(GC_INDEX,0x0205);
    outpw(GC_INDEX,0x0F07);
    outpb(GC_INDEX,0x08);

    // Now draw the line
    if (flags & gaLineXMajor) {
        if (flags & gaLineXPositive) {
            while (count--) {
                if ((bank = dstAddr >> 16) != DC.cntBank) {
                    DC.SetBank(DC.cntBank = bank);
                    outpb(GC_INDEX,0x08);
                    d = dstBase + (dstAddr & 0xFFFF);
                    }
                outpb(GC_DATA,mask);
                rwbv(d,(x1 & 1) ? *s : (*s >> 4));
                if (x1 & 1) s++;
                x1++;
                if (rorb1(&mask)) {
                    d++;
                    dstAddr++;
                    }
                if (initialError < 0)
                    initialError += majorInc;
                else {
                    initialError += diagInc;
                    s += sincr;
                    d += fincr;
                    dstAddr += fincr;
                    }
                }
            }
        else {
            while (count--) {
                if ((bank = dstAddr >> 16) != DC.cntBank) {
                    DC.SetBank(DC.cntBank = bank);
                    outpb(GC_INDEX,0x08);
                    d = dstBase + (dstAddr & 0xFFFF);
                    }
                outpb(GC_DATA,mask);
                rwbv(d,(x1 & 1) ? *s : (*s >> 4));
                if (!(x1 & 1)) s--;
                x1--;
                if (rolb1(&mask)) {
                    d--;
                    dstAddr--;
                    }
                if (initialError < 0)
                    initialError += majorInc;
                else {
                    initialError += diagInc;
                    s += sincr;
                    d += fincr;
                    dstAddr += fincr;
                    }
                }
            }
        }
    else {
        if (flags & gaLineXPositive) {
            while (count--) {
                if ((bank = dstAddr >> 16) != DC.cntBank) {
                    DC.SetBank(DC.cntBank = bank);
                    outpb(GC_INDEX,0x08);
                    d = dstBase + (dstAddr & 0xFFFF);
                    }
                outpb(GC_DATA,mask);
                rwbv(d,(x1 & 1) ? *s : (*s >> 4));
                s += sincr;
                d += fincr;
                dstAddr += fincr;
                if (initialError < 0)
                    initialError += majorInc;
                else {
                    initialError += diagInc;
                    if (rorb1(&mask)) {
                        d++;
                        dstAddr++;
                        }
                    if (x1 & 1) s++;
                    x1++;
                    }
                }
            }
        else {
            while (count--) {
                if ((bank = dstAddr >> 16) != DC.cntBank) {
                    DC.SetBank(DC.cntBank = bank);
                    outpb(GC_INDEX,0x08);
                    d = dstBase + (dstAddr & 0xFFFF);
                    }
                outpb(GC_DATA,mask);
                rwbv(d,(x1 & 1) ? *s : (*s >> 4));
                s += sincr;
                d += fincr;
                dstAddr += fincr;
                if (initialError < 0)
                    initialError += majorInc;
                else {
                    initialError += diagInc;
                    if (rolb1(&mask)) {
                        d--;
                        dstAddr--;
                        }
                    if (!(x1 & 1)) s--;
                    x1--;
                    }
                }
            }
        }
}

