/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  C module to implement the support for software stereo
*               page flipping in the SNAP drivers. The software stereo
*               support requires an interrupt handler to be installed,
*               which for OS'es other than DOS requires calls from the
*               application down to a ring 0 helper device driver.
*
****************************************************************************/

#ifndef __DRIVER__
#include "snap/graphics.h"          // Allows to be compiled into debug test code
#define inpb    PM_inpb
#define outpb   PM_outpb
#else
#include "loader.h"
#endif

/*------------------------- Global Variables ------------------------------*/

#define CAL_LENGTH          30      // Calibration length in refresh cycles
#define TIMEOUT_TICKS       10000   // Ticks for initial timeout (10ms)
#define FULLTIMEOUT_TICKS   1000000 // Ticks for full timeout (1s)
#define TIMEOUT_LOOPS       4       // Number of loops to try starting RTC

/* {secret} */
typedef enum {
    gaTimerOff,
    gaTimerStereo,
    } TimerModeFlags;

/* {secret} */
typedef enum {
    gaLeftEye,
    gaRightEye,
    } EyeTypeFlags;

static volatile N_int32     interruptCount;
static volatile N_int32     timerMode;
static volatile N_int32     flipHasPassed;
static volatile N_int32     flipHasPassedLeft;
static volatile N_int32     flipHasPassedRight;
static GA_options           opt;
static N_int32              cntEye;
static ulong                tickCount;
static ulong                fullFrameCount;
static ulong                fullFrameTicks;
static GA_largeInteger      lastRetraceCount;
static N_uint32             frameCheckCount;
static N_uint32             leftEyeAddr;
static N_uint32             rightEyeAddr;
static N_uint32             leftEyeAddrNext;
static N_uint32             rightEyeAddrNext;
static N_int32              oldVSyncWidth;
static GA_driverFuncs       driver;
static GA_SCIFuncs          sci;
static ibool                bGlassesIOPort;
static ibool                bFlipperActive = false;
static int                  waitVRTFlag = 1;
static PM_lockHandle        codeLockHandle,dataLockHandle;

#ifdef  DEBUG_STEREO_FLIPS
N_uint32        minSyncTime = 0x7FFFFFFFUL;
N_uint32        maxSyncTime = 0;
float           minSyncPercent = 0;
float           maxSyncPercent = 0;
int             missedFlips = 0;
#endif

#ifdef DEBUG_STEREO
#define STEREO_ENTRY    static
#else
#define STEREO_ENTRY
#endif

/*-------------------------- Implementation -------------------------------*/

#ifdef  __WATCOMC__
#pragma off (check_stack)           /* No stack checking under Watcom   */
#endif

/****************************************************************************
REMARKS:
This is the workhorse Real Time Clock interrupt handler function, that
implements our timer driver real time clock support.
****************************************************************************/
static void PMAPI RTCHandler(void)
{
    GA_largeInteger count;
    N_uint32        diff;
#ifdef  DEBUG_STEREO_FLIPS
    GA_largeInteger dbgStartCount,dbgEndCount;
#endif

    interruptCount++;
    if (timerMode != gaTimerOff) {
        GA_TimerRead(&count);
        diff = GA_TimerDifference(&count,&lastRetraceCount);
        if (diff >= frameCheckCount) {
#ifdef  DEBUG_STEREO_FLIPS
            GA_TimerRead(&dbgStartCount);
#endif
            /* Stop the clock so we can restart it again immediately after
             * we finish our busy wait loop.
             */
            PM_stopRealTimeClock();

            /* Insure that *both* left and right buffers get updated
             * as a pair, in sync with left re-load.
             */
            if (flipHasPassedRight && (cntEye == gaLeftEye)) {
                leftEyeAddr = leftEyeAddrNext;
                rightEyeAddr = rightEyeAddrNext;
                }
            if (cntEye == gaLeftEye) {
                if (opt.glassesType == gaGlassesVSync)
                    driver.SetVSyncWidth(opt.vSyncWidthLeft);
                driver.SetDisplayStart(leftEyeAddr,waitVRTFlag);
                GA_TimerRead(&lastRetraceCount);
                PM_restartRealTimeClock(opt.RTCFrequency);
                if (bGlassesIOPort)
                    outpb(opt.ioPort,(uchar)((inpb(opt.ioPort) & opt.ioAndMask) | opt.ioLeftOrMask));
                if (opt.glassesType == gaGlassesH3DPlus)
                    sci.SCIwriteSDA(0,1);
                cntEye = gaRightEye;
                flipHasPassedLeft = 1;
                }
            else {
                if (opt.glassesType == gaGlassesVSync)
                    driver.SetVSyncWidth(opt.vSyncWidthRight);
                driver.SetDisplayStart(rightEyeAddr,waitVRTFlag);
                GA_TimerRead(&lastRetraceCount);
                PM_restartRealTimeClock(opt.RTCFrequency);
                if (bGlassesIOPort)
                    outpb(opt.ioPort,(uchar)((inpb(opt.ioPort) & opt.ioAndMask) | opt.ioRightOrMask));
                if (opt.glassesType == gaGlassesH3DPlus)
                    sci.SCIwriteSDA(0,0);
                cntEye = gaLeftEye;
                flipHasPassedRight = 1;
                }
            /* Set flipHasPassed *after* the left/right pair have been
             * loaded.
             */
            if (flipHasPassedLeft && flipHasPassedRight && (cntEye == gaRightEye))
                flipHasPassed = 1;

#ifdef  DEBUG_STEREO_FLIPS
            if (interruptCount > opt.RTCFrequency) {
                GA_TimerRead(&dbgEndCount);
                diff = GA_TimerDifference(&dbgEndCount,&dbgStartCount);
                if (diff < minSyncTime)
                    minSyncTime = diff;
                if (diff > maxSyncTime)
                    maxSyncTime = diff;
                if (diff > fullFrameCount) {
                    missedFlips++;
                    DebugVxD();
                    }
                }
#endif
            }
        }
}

/****************************************************************************
DESCRIPTION:
Initialises the software stereo page flipping module.

HEADER:
snap/graphics.h

PARAMETERS:
dc  - SNAP device context to use to access the hardware
opt - Options structure to determine how to control the glasses

REMARKS:
This function initialises the software stereo page flip handler, and does
one time initialisation and calibration. The calibration will take about
1 second for most resolutions and refresh rates.

NOTE:   This function *must* be called after the system has already been
        put into graphics mode, since it does calibration based on the
        currently active display mode and refresh rate.

NOTE:   Software stereo page flipping is off by default when this function
        is called, and you must call GA_softStereoOn to activate it.

SEE ALSO:
GA_softStereoExit, GA_softStereoOn
****************************************************************************/
STEREO_ENTRY ibool NAPI GA_softStereoInit(
    GA_devCtx *dc)
{
    int             loopCount,rtcTickCount;
    GA_largeInteger start,count;
    GA_initFuncs    init;
    N_uint8         caps;
    N_int32         ports;
    ulong           oldPriority;

    /* Check if flipper is already running in case another instance
     * of stereo app is launched, or another stereo window is opened.
     */
    if (bFlipperActive)
        return false;

    /* Copy options structure and get driver functions */
    init.dwSize = sizeof(init);
    if (!GA_queryFunctions(dc,GA_GET_INITFUNCS,&init))
        return false;
    driver.dwSize = sizeof(driver);
    if (!GA_queryFunctions(dc,GA_GET_DRIVERFUNCS,&driver))
        return false;
    opt.dwSize = sizeof(opt);
    init.GetOptions(&opt);
    if (!opt.bStereo)
        return false;

    /* Check if we need to use DDC SCI functions for special signalling */
    if (opt.glassesType == gaGlassesH3DPlus) {
        sci.dwSize = sizeof(sci);
        if (!GA_queryFunctions(dc,GA_GET_SCIFUNCS,&sci))
            return false;
        if (!sci.SCIdetect(&caps,&ports))
            return false;
        }

    /* Obtain I/O port values for LPT and COM ports as necessary */
    switch ((opt.glassesType & gaGlassesMask)) {
        case gaGlassesLPTPort:
            opt.ioPort = PM_getLPTPort(((opt.glassesType & gaGlassesPortMask) - gaGlassesLPT1) >> 12);
            bGlassesIOPort = true;
            break;
        case gaGlassesCOMPort:
            opt.ioPort = PM_getCOMPort(((opt.glassesType & gaGlassesPortMask) - gaGlassesCOM1) >> 12) + 0x4;
            bGlassesIOPort = true;
            break;
        case gaGlassesIOPort:
            bGlassesIOPort = true;
            break;
        default:
            bGlassesIOPort = false;
            break;
        }

    /* Check that we have an operational high resolution timer */
    if (!GA_TimerInit())
        return false;

    /* Attempt to install the real time clock interrupt handler */
    timerMode = gaTimerOff;
    PM_lockCodePages((__codePtr)RTCHandler,4096,&codeLockHandle);
    PM_lockDataPages((void*)&interruptCount,4096,&dataLockHandle);
    if (!PM_setRealTimeClockHandler(RTCHandler,0))
        return false;

    /* If we are using a newer driver (V1.04 or later) then we can set the
     * waitVRT flag to -1 which avoids the WaitDEVGA() calls internally
     * in the driver. This is better since we can come in later in RTC
     * tick counts as this code will wait for too long if the tick
     * comes along when the vertical blank interval has just started.
     */
    opt.RTCFrequency = 4096;
    opt.RTCAdvanceTicks = 2;
    if (dc->Version >= 0x104)
        waitVRTFlag = -1;
    else
        waitVRTFlag = 1;

    /* Increase the thread priority for the following important time
     * critical calibration code.
     */
    oldPriority = PM_setMaxThreadPriority();

    /* Now time how long the real time clock tick takes, averaged over
     * 1/2 a second of real time clock ticks. We try this four times
     * because sometimes the real time clock can stop for no reason
     * the first few times it is started after booting the system.
     */
    rtcTickCount = opt.RTCFrequency / 2;
    ZTimerInit();
    for (loopCount = 0; loopCount < TIMEOUT_LOOPS; loopCount++) {
        PM_setRealTimeClockFrequency(opt.RTCFrequency);
        LZTimerOn();
        GA_TimerRead(&start);
        interruptCount = -1;
        while (interruptCount == -1) {
            if (LZTimerLap() >= TIMEOUT_TICKS)
                break;
            }
        LZTimerOff();
        if (LZTimerCount() >= TIMEOUT_TICKS) {
            PM_setRealTimeClockFrequency(0);
            continue;
            }
        LZTimerOn();
        GA_TimerRead(&start);
        while (interruptCount < rtcTickCount) {
            if (LZTimerLap() >= FULLTIMEOUT_TICKS)
                break;
            }
        GA_TimerRead(&count);
        LZTimerOff();
        PM_setRealTimeClockFrequency(0);
        tickCount = GA_TimerDifference(&count,&start) / rtcTickCount;
        if (LZTimerCount() < FULLTIMEOUT_TICKS)
            break;
        }
    if (loopCount == TIMEOUT_LOOPS) {
        PM_restoreRealTimeClockHandler();
        return false;
        }

    /* Calibrate length of 1 refresh, averaged over CAL_LENGTH
     * refreshes.
     */
    loopCount = CAL_LENGTH;
    driver.WaitVSync();
    GA_TimerRead(&start);
    while (loopCount--)
        driver.WaitVSync();
    GA_TimerRead(&count);
    fullFrameCount = GA_TimerDifference(&count,&start) / CAL_LENGTH;

    /* Restore the original thread priority */
    PM_restoreThreadPriority(oldPriority);

    /* Now determine the point that we need to handle the flip inside our
     * RTC interrupt handler. This should be equal to the number of whole
     * ticks in a single frame minus the 'advance' tick count that we come
     * in early. We also take off a third of a tick count to ensure that we
     * come in early.
     */
    fullFrameTicks = fullFrameCount / tickCount;
    frameCheckCount = (fullFrameTicks - opt.RTCAdvanceTicks) * tickCount - (tickCount/3);

    /* Finally initialise the rest of the variables for the stereo flipper */
    cntEye = gaLeftEye;
    leftEyeAddr = leftEyeAddrNext = 0;
    rightEyeAddr = rightEyeAddrNext = 0;
    oldVSyncWidth = driver.GetVSyncWidth();

    /* We are now active and ready to run */
    bFlipperActive = true;
    return true;
}

/****************************************************************************
DESCRIPTION:
Enables software stereo page flipping.

HEADER:
snap/graphics.h

REMARKS:
This function enables the software stereo page flipping, and starts the
real time clock counter running in the background.

SEE ALSO:
GA_softStereoScheduleFlip, GA_softStereoGetFlipStatus, GA_softStereoOff
****************************************************************************/
STEREO_ENTRY void NAPI GA_softStereoOn(void)
{
    ulong   oldPriority;

    /* Don't do anything unless flipper has been properly setup. */
    if (!bFlipperActive)
        return;

    if (timerMode == gaTimerOff) {
        leftEyeAddr = leftEyeAddrNext;
        rightEyeAddr = rightEyeAddrNext;
        flipHasPassed = flipHasPassedLeft = flipHasPassedRight = 0;

        /* Increase the thread priority and calibrate and start the
         * flipper
         */
        oldPriority = PM_setMaxThreadPriority();
        driver.WaitVSync();
        PM_setRealTimeClockFrequency(opt.RTCFrequency);
        GA_TimerRead(&lastRetraceCount);
        timerMode = gaTimerStereo;
        if (opt.glassesType == gaGlassesH3DPlus)
            sci.SCIbegin(0);

        /* Restore the original thread priority */
        PM_restoreThreadPriority(oldPriority);
#ifdef DEBUG_STEREO_FLIPS
        interruptCount = 0;
        minSyncTime = 0x7FFFFFFFUL;
        maxSyncTime = 0;
        missedFlips = 0;
#endif
        }
}

/****************************************************************************
DESCRIPTION:
Schedule a new stereo display start address change.

HEADER:
snap/graphics.h

PARAMETERS:
leftAddr    - Left display start address to make active
rightAddr   - Right display start address to make active

REMARKS:
This function schedules a new left and right start address to take hold on
the next vertical retrace for the LC shutter glasses. This is used for
double and triple buffering when stereo mode is active.

SEE ALSO:
GA_softStereoGetFlipStatus, GA_softStereoWaitTillFlipped
****************************************************************************/
STEREO_ENTRY void NAPI GA_softStereoScheduleFlip(
    N_uint32 leftAddr,
    N_uint32 rightAddr)
{
    /* Don't do anything unless flipper has been properly setup. */
    if (!bFlipperActive)
        return;

    /* Buffer the next left and right start addresses for loading in sync.
     * Note that we *cannot* disable interrupts in here. To ensure that
     * we don't update these values while the vertical retrace is being
     * serviced, we block on updating these values until we are not in
     * the vertical retrace period.
     */
    while (driver.IsVSync())
        ;
    leftEyeAddrNext = leftAddr;
    rightEyeAddrNext = rightAddr;
    flipHasPassed = flipHasPassedLeft = flipHasPassedRight = 0;
}

/****************************************************************************
DESCRIPTION:
Returns the status of the last scheduled stereo display start change.

HEADER:
snap/graphics.h

RETURNS:
-1 if flipper is not running at all;
0 if flip has not occured;
1 if flip has occured and currently on left eye;
2 if flip has occured and currently on right eye.

REMARKS:
This function returns the status of the last scheduled software stereo
display start change. This bit is sticky and is set to 0 when the
GA_softStereoScheduleFlip function is called, and reset to 1 when the
flip actually occurs.

This function also serves to inform the calling application whether the
flipper is already running or not. This will help insure that the flipper
will only be used one instance at a time.

SEE ALSO:
GA_softStereoScheduleFlip, GA_softStereoWaitTillFlipped
****************************************************************************/
STEREO_ENTRY N_int32 NAPI GA_softStereoGetFlipStatus(void)
{
    /* Return status if the flipper is not running at all. */
    if (!bFlipperActive)
        return -1;
    return (flipHasPassed) ? ((cntEye == gaRightEye) ? 2 : 1) : 0;
}

/****************************************************************************
DESCRIPTION:
Waits until the last scheduled flip operation has occurred.

HEADER:
snap/graphics.h

REMARKS:
This function polls the status of the last scheduled flip operation, and
waits until it has completed. However this function also does the important
job of checking the RTC clock to ensure that it continues to run, and will
re-start it if it has stopped for some reason. Hence you should call
this function when you need to spin until the flip has occurred to catch
problems when the RTC clock stops running.

SEE ALSO:
GA_softStereoScheduleFlip, GA_softStereoGetFlipStatus
****************************************************************************/
STEREO_ENTRY void NAPI GA_softStereoWaitTillFlipped(void)
{
    /* Don't do anything unless flipper has been properly setup. */
    if (!bFlipperActive)
        return;

    while (!GA_softStereoGetFlipStatus())
        ;
}

/****************************************************************************
DESCRIPTION:
Turns off the automatic software stereo page flipping.

HEADER:
snap/graphics.h

REMARKS:
This function turns off the software stereo page flipping, putting the
system back into 2D mode and turning off the LC shutter glasses.

SEE ALSO:
GA_softStereoExit, GA_softStereoOn
****************************************************************************/
STEREO_ENTRY void NAPI GA_softStereoOff(void)
{
    /* Don't do anything unless flipper has been properly setup. */
    if (!bFlipperActive)
        return;

    if (timerMode != gaTimerOff) {
        PM_setRealTimeClockFrequency(0);
        timerMode = gaTimerOff;
        driver.SetDisplayStart(leftEyeAddr,0);
        if (opt.glassesType == gaGlassesVSync)
            driver.SetVSyncWidth(oldVSyncWidth);
        if (bGlassesIOPort)
            outpb(opt.ioPort,(uchar)((inpb(opt.ioPort) & opt.ioAndMask) | opt.ioOffOrMask));
        if (opt.glassesType == gaGlassesH3DPlus)
            sci.SCIend(0);
        cntEye = gaLeftEye;
        }

#ifdef  DEBUG_STEREO_FLIPS
    minSyncPercent = (minSyncTime * 100.0) / fullFrameCount;
    maxSyncPercent = (maxSyncTime * 100.0) / fullFrameCount;
#endif
}

/****************************************************************************
DESCRIPTION:
Cleans up and exits the software stereo module.

HEADER:
snap/graphics.h

REMARKS:
This function cleans up and exits the software stereo page flipping module,
and restores all interrupt handlers hooked by the driver.

SEE ALSO:
GA_softStereoInit
****************************************************************************/
STEREO_ENTRY void NAPI GA_softStereoExit(void)
{
    GA_softStereoOff();
    PM_restoreRealTimeClockHandler();
    PM_unlockCodePages((__codePtr)RTCHandler,4096,&codeLockHandle);
    PM_unlockDataPages((void*)&interruptCount,4096,&dataLockHandle);
    bFlipperActive = false;
}

