/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

// TODO: Optimised ROP blits for software rendering! We can do this by
//       adding two combined methods:
//
//  1. For all color depths < 32bpp, we destination align to 32bpp and
//     then do process 32-bit pixels at a time with each raster OP
//     using 32-bit pixel functions.
//
//  2. If we are drawing to a hardware buffer, read the scanline to
//     system memory, process the ROP to system memory buffer and
//     write the result back to the frameBuffer.
//
//  3. If we are drawing to a hardware buffer, it may be faster if we
//     do complex rendering a scanline at a time to the system buffer
//     and then blit this to the screen 32-bits at a time (ie: replace
//     mix pattern fills, text rendering etc).

// TODO: We need to add the following to eliminate SOFTDRAW from our
//       OS/2 drivers:
//
//  1. Support for arbitrary mono and color patterns, perhaps with a
//     caching scheme. We may want to make sure things can be fast for
//     the case where we need to set the pattern and use it immediately
//     in a temporary pattern rather than doing a real cached pattern
//     system (ie: allocate the temp pattern to fit the largest pattern
//     and keep re-using it). Or perhaps the auto-reuse could be for
//     the entire set of 8 patterns cached in video memory?

// TODO: Parameterise all the internal functions into separate BPD files
//       in the drivers. We can do this to cut down on code space for
//       embedded systems developers who don't need all features, such
//       NOP'ing out all ROP3 pixel functions, stretch blts etc if the
//       drivers are not bound in the BPD file!!

/*--------------------------- Global variables ----------------------------*/

REF2D_state REF2D_devCtx;

REF2D_driver REF2D_drv = {
    sizeof(REF2D_drv),
    REF2D_QueryFunctionsSW,
    REF2D_SetDrawBufferSW,
    REF2D_SetDrawSurface,
    NULL,
    REF2D_UnloadDriverInt,
    NULL,
    NULL,
    NULL,
    REF2D_GetDrawBuffer,
    REF2D_DrawRectExtSW,
    REF2D_ForceSoftwareOnly,
    REF2D_SetColorCompareMask,
    REF2D_PostSwitchPhysicalResolution,
    };

static GA_cursorFuncs REF_cursor = {
    sizeof(GA_cursorFuncs),
    REF2D_SetMonoCursor,
    REF2D_SetMonoCursorColor,
    REF2D_SetColorCursor,
    REF2D_SetCursorPos,
    REF2D_ShowCursor,
    REF2D_BeginAccess,
    REF2D_EndAccess,
    REF2D_SetColorCursor256,
    REF2D_SetColorCursorRGBA,
    REF2D_IsHardwareCursor,
    REF2D_SetColorCursorRGB,
    };

static GA_2DStateFuncs REF_state2d = {
    sizeof(GA_2DStateFuncs),
    REF2D_SetDrawBuffer,
    REF2D_IsIdle,
    REF2D_WaitTillIdle,
    NULL,                       // DRIVER_EnableDirectAccess,
    NULL,                       // DRIVER_DisableDirectAccess,
    REF2D_SetMix,
    REF2D_SetForeColor,
    REF2D_SetBackColor,
    REF2D_Set8x8MonoPattern,
    REF2D_Use8x8MonoPattern,
    REF2D_Use8x8TransMonoPattern,
    REF2D_Set8x8ColorPattern,
    REF2D_Use8x8ColorPattern,
    REF2D_Use8x8TransColorPattern,
    REF2D_SetLineStipple,
    REF2D_SetLineStippleCount,
    REF2D_SetPlaneMask,
    REF2D_SetSrcBlendFunc,
    REF2D_SetDstBlendFunc,
    REF2D_SetAlphaValue,
    NULL,                       // DRIVER_SetupForWindows
    NULL,                       // DRIVER_ResetWindows
    REF2D_SetLineStyle,
    REF2D_BuildTranslateVector,
    REF2D_SetBlendFunc,
    };

GA_2DRenderFuncs REF_draw2dSW = {
    sizeof(GA_2DRenderFuncs),
    REF2D_GetPixel,
    REF2D_PutPixel,
    REF2D_DrawScanList,
    REF2D_DrawPattScanList,
    REF2D_DrawColorPattScanList,
    REF2D_DrawEllipseList,
    REF2D_DrawPattEllipseList,
    REF2D_DrawColorPattEllipseList,
    REF2D_DrawFatEllipseList,
    REF2D_DrawPattFatEllipseList,
    REF2D_DrawColorPattFatEllipseList,
    REF2D_DrawRect,
    NULL,
    REF2D_DrawPattRect,
    REF2D_DrawColorPattRect,
    REF2D_DrawTrap,
    REF2D_DrawPattTrap,
    REF2D_DrawColorPattTrap,
    REF2D_DrawLineInt,
    REF2D_DrawBresenhamLine,
    REF2D_DrawStippleLineInt,
    REF2D_DrawBresenhamStippleLine,
    REF2D_DrawEllipse,
    REF2D_ClipEllipse,
    REF2D_PutMonoImageMSBSys,
    REF2D_PutMonoImageMSBLin,
    REF2D_PutMonoImageMSBBM,
    REF2D_PutMonoImageLSBSys,
    REF2D_PutMonoImageLSBLin,
    REF2D_PutMonoImageLSBBM,
    REF2D_ClipMonoImageMSBSys,
    REF2D_ClipMonoImageMSBLin,
    REF2D_ClipMonoImageMSBBM,
    REF2D_ClipMonoImageLSBSys,
    REF2D_ClipMonoImageLSBLin,
    REF2D_ClipMonoImageLSBBM,
    REF2D_BitBlt,
    REF2D_BitBltLin,
    REF2D_BitBltSys,
    REF2D_BitBltBM,
    REF2D_BitBltPatt,
    REF2D_BitBltPattLin,
    REF2D_BitBltPattSys,
    REF2D_BitBltPattBM,
    REF2D_BitBltColorPatt,
    REF2D_BitBltColorPattLin,
    REF2D_BitBltColorPattSys,
    REF2D_BitBltColorPattBM,
    REF2D_SrcTransBlt,
    REF2D_SrcTransBltLin,
    REF2D_SrcTransBltSys,
    REF2D_SrcTransBltBM,
    REF2D_DstTransBlt,
    REF2D_DstTransBltLin,
    REF2D_DstTransBltSys,
    REF2D_DstTransBltBM,
    REF2D_StretchBlt,
    REF2D_StretchBltLin,
    REF2D_StretchBltSys,
    REF2D_StretchBltBM,
    REF2D_ConvertBltSys_Obsolete,
    REF2D_ConvertBltBM_Obsolete,
    REF2D_StretchConvertBltSys_Obsolete,
    REF2D_StretchConvertBltBM_Obsolete,
    REF2D_BitBltFxTest,
    REF2D_BitBltFx,
    REF2D_BitBltFxLin,
    REF2D_BitBltFxSys,
    REF2D_BitBltFxBM,
    REF2D_GetBitmapSys,
    REF2D_GetBitmapBM,
    NULL,
    REF2D_DrawClippedLineInt,
    REF2D_DrawClippedBresenhamLine,
    REF2D_DrawClippedStippleLineInt,
    REF2D_DrawClippedBresenhamStippleLine,
    REF2D_BitBltPlaneMasked,
    REF2D_BitBltPlaneMaskedLin,
    REF2D_BitBltPlaneMaskedSys,
    REF2D_BitBltPlaneMaskedBM,
    REF2D_DrawRectLin,
    REF2D_DrawRectExt,
    REF2D_DrawStyleLineInt,
    REF2D_DrawBresenhamStyleLine,
    REF2D_DrawClippedStyleLineInt,
    REF2D_DrawClippedBresenhamStyleLine,
    };

static GA_2DRenderFuncs REF_draw2dStub = {
    sizeof(GA_2DRenderFuncs),
    STUB_GetPixel,
    STUB_PutPixel,
    STUB_DrawScanList,
    STUB_DrawPattScanList,
    STUB_DrawColorPattScanList,
    STUB_DrawEllipseList,
    STUB_DrawPattEllipseList,
    STUB_DrawColorPattEllipseList,
    STUB_DrawFatEllipseList,
    STUB_DrawPattFatEllipseList,
    STUB_DrawColorPattFatEllipseList,
    STUB_DrawRect,
    NULL,
    STUB_DrawPattRect,
    STUB_DrawColorPattRect,
    STUB_DrawTrap,
    STUB_DrawPattTrap,
    STUB_DrawColorPattTrap,
    STUB_DrawLineInt,
    STUB_DrawBresenhamLine,
    STUB_DrawStippleLineInt,
    STUB_DrawBresenhamStippleLine,
    STUB_DrawEllipse,
    STUB_ClipEllipse,
    STUB_PutMonoImageMSBSys,
    STUB_PutMonoImageMSBLin,
    STUB_PutMonoImageMSBBM,
    STUB_PutMonoImageLSBSys,
    STUB_PutMonoImageLSBLin,
    STUB_PutMonoImageLSBBM,
    STUB_ClipMonoImageMSBSys,
    STUB_ClipMonoImageMSBLin,
    STUB_ClipMonoImageMSBBM,
    STUB_ClipMonoImageLSBSys,
    STUB_ClipMonoImageLSBLin,
    STUB_ClipMonoImageLSBBM,
    STUB_BitBlt,
    STUB_BitBltLin,
    STUB_BitBltSys,
    STUB_BitBltBM,
    STUB_BitBltPatt,
    STUB_BitBltPattLin,
    STUB_BitBltPattSys,
    STUB_BitBltPattBM,
    STUB_BitBltColorPatt,
    STUB_BitBltColorPattLin,
    STUB_BitBltColorPattSys,
    STUB_BitBltColorPattBM,
    STUB_SrcTransBlt,
    STUB_SrcTransBltLin,
    STUB_SrcTransBltSys,
    STUB_SrcTransBltBM,
    STUB_DstTransBlt,
    STUB_DstTransBltLin,
    STUB_DstTransBltSys,
    STUB_DstTransBltBM,
    STUB_StretchBlt,
    STUB_StretchBltLin,
    STUB_StretchBltSys,
    STUB_StretchBltBM,
    STUB_ConvertBltSys_Obsolete,
    STUB_ConvertBltBM_Obsolete,
    STUB_StretchConvertBltSys_Obsolete,
    STUB_StretchConvertBltBM_Obsolete,
    STUB_BitBltFxTest,
    STUB_BitBltFx,
    STUB_BitBltFxLin,
    STUB_BitBltFxSys,
    STUB_BitBltFxBM,
    STUB_GetBitmapSys,
    STUB_GetBitmapBM,
    NULL,
    STUB_DrawClippedLineInt,
    STUB_DrawClippedBresenhamLine,
    STUB_DrawClippedStippleLineInt,
    STUB_DrawClippedBresenhamStippleLine,
    STUB_BitBltPlaneMasked,
    STUB_BitBltPlaneMaskedLin,
    STUB_BitBltPlaneMaskedSys,
    STUB_BitBltPlaneMaskedBM,
    STUB_DrawRectLin,
    STUB_DrawRectExt,
    STUB_DrawStyleLineInt,
    STUB_DrawBresenhamStyleLine,
    STUB_DrawClippedStyleLineInt,
    STUB_DrawClippedBresenhamStyleLine,
    };

/* Pixel blending function tables */

#if defined(SUPPORT_16BPP) || defined(SUPPORT_24BPP) || defined(SUPPORT_32BPP)
REF2D_blendColor_t  REF2D_blendSrcTable[16] = {
    ULNOP_VECTOR,
    REF2D_blendSrc_Zero,
    REF2D_blendSrc_One,
    REF2D_blendSrc_SrcColor,
    REF2D_blendSrc_OneMinusSrcColor,
    REF2D_blendSrc_SrcAlpha,
    REF2D_blendSrc_OneMinusSrcAlpha,
    REF2D_blendSrc_DstAlpha,
    REF2D_blendSrc_OneMinusDstAlpha,
    REF2D_blendSrc_DstColor,
    REF2D_blendSrc_OneMinusDstColor,
    REF2D_blendSrc_SrcAlphaSaturate,
    REF2D_blendSrc_ConstantColor,
    REF2D_blendSrc_OneMinusConstantColor,
    REF2D_blendSrc_ConstantAlpha,
    REF2D_blendSrc_OneMinusConstantAlpha
    };

REF2D_blendColor_t  REF2D_blendDstTable[16] = {
    ULNOP_VECTOR,
    REF2D_blendDst_Zero,
    REF2D_blendDst_One,
    REF2D_blendDst_SrcColor,
    REF2D_blendDst_OneMinusSrcColor,
    REF2D_blendDst_SrcAlpha,
    REF2D_blendDst_OneMinusSrcAlpha,
    REF2D_blendDst_DstAlpha,
    REF2D_blendDst_OneMinusDstAlpha,
    REF2D_blendDst_DstColor,
    REF2D_blendDst_OneMinusDstColor,
    REF2D_blendDst_SrcAlphaSaturate,
    REF2D_blendDst_ConstantColor,
    REF2D_blendDst_OneMinusConstantColor,
    REF2D_blendDst_ConstantAlpha,
    REF2D_blendDst_OneMinusConstantAlpha
    };
#endif

static int              refCount = 0;
static int              softwareOnly;
static int              softwareOnlyForced;
static int              setHardwareState;
static int              transMonoPattSW;
static int              transColorPattSW;
static int              REF_bugFixHWROP3;
ibool                   REF_slowBltSys;
ibool                   REF_slowBltSysOrg;
ibool                   REF_noHWPlaneMask;
int                     REF_mix;
GA_2DRenderFuncs        REF_draw2dCnt;
GA_2DRenderFuncs        REF_draw2dHW;
GA_2DRenderFuncs        REF_draw2dHWSW;
static GA_regionFuncs   REF_region = {0};
static GA_bufferFuncs   REF_buffer = {0};
#ifdef  __DRIVER__
PM_imports _VARAPI      _PM_imports;
static GA_exports       _GA_exports;
#define GA_loadRegionMgr(funcs,hModule,size)    _GA_exports.GA_loadRegionMgr(funcs,hModule,size)
#define GA_unloadRegionMgr(hModule)             _GA_exports.GA_unloadRegionMgr(hModule)
#define GA_loadLibrary(szBPDName,size,shared)   _GA_exports.GA_loadLibrary(szBPDName,size,shared)
#define GA_getProcAddress(hModule,szProcName)   _GA_exports.GA_getProcAddress(hModule,szProcName)
#endif

// Names of all internal BPD's

#define BUFFER_NAME "buffer.bpd"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Function entry point to fake loading of a driver when linked into app.
****************************************************************************/
static ibool REF2D_loadBufferMgr(
    GA_devCtx *hwCtx,
    REF2D_driver *ref2d,
    GA_bufferFuncs *funcs,
    PE_MODULE **hModule,
    ulong *size)
{
#ifdef __DRIVER__
    BM_initLibrary_t    BM_init;

    if ((*hModule = GA_loadLibrary(BUFFER_NAME,size,false)) == NULL)
        return false;
    if ((BM_init = (BM_initLibrary_t)GA_getProcAddress(*hModule,"_BM_initLibrary")) == NULL)
        PM_fatalError("Unable to get buffer BPD entry point!");
#else
    (void)hModule;
    (void)size;
#endif
    return BM_init(hwCtx,ref2d,DC.bufReadScan,DC.bufBltScan,funcs);
}

/****************************************************************************
REMARKS:
Function entry point to fake unloading of a driver when linked into app.
****************************************************************************/
static void REF2D_unloadBufferMgr(
    PE_MODULE *hModule)
{
#ifdef __DRIVER__
    BM_exitLibrary_t BM_exit;
    if ((BM_exit = (BM_exitLibrary_t)GA_getProcAddress(hModule,"_BM_exitLibrary")) == NULL)
        PM_fatalError("Unable to get buffer BPD entry point!");
#else
    (void)hModule;
#endif
    BM_exit();
}

/****************************************************************************
REMARKS:
Function to initialise the internal mode and cursor information for the
ref2d library. We do it here because we also need to redo this if the
physical screen resolution has changed when doing on the fly zooming.
****************************************************************************/
static void InitModeAndCursorInfo(void)
{
    GA_CRTCInfo     cntCRTC;

    // Query the current SNAP mode information
    DC.gaMode.dwSize = sizeof(DC.gaMode);
    DC.hwInit.GetCurrentVideoModeInfo(&DC.gaMode);
    DC.hwInit.GetCRTCTimings(&cntCRTC);

    // Determine maximum hardware cursor dimensions. We need to halve
    // the maximum hardware cursor dimensions vertically in double
    // scan modes.
    DC.maxCursorWidth = DC.maxCursorHeight = 64;
    if (DC.hwCtx->WorkArounds & gaHWCursor32x32)
        DC.maxCursorWidth = DC.maxCursorHeight = 32;
    if (cntCRTC.Flags & gaDoubleScan || !(DC.gaMode.Attributes & gaHaveHWCursor)) {
        // Hardware cursors are currently incorrectly handled in all
        // SNAP drivers at the moment, as they fail to correctly
        // take into account the 'doubled' hotSpot. If we fix this
        // at some stage we can re-enabled it, but for now the cursor
        // is always done in software
        DC.maxCursorWidth = 0;
        DC.maxCursorHeight = 0;
        }
}

/****************************************************************************
DESCRIPTION:
Fix up the software rasteriser after a mode switch has occured

HEADER:
snap/ref2d.h

REMARKS:
This function is used to to re-initialise the internals for the reference
rasteriser library after a physical mode switch has occured (ie:
GA_initFuncs::PerformDisplaySwitch after GA_initFuncs::PollForDisplaySwitch
returned true. Mostly this entails fixing up the internal software renderer
to handle changes in hardware functions, such as hardware cursor support
etc for the new mode. Application and shell drivers must /always/ call
this function after calling GA_initFuncs::PerformDisplaySwitch.

SEE ALSO:
GA_initFuncs::PerformDisplaySwitch
****************************************************************************/
void NAPI REF2D_PostSwitchPhysicalResolution(void)
{
    InitModeAndCursorInfo();
}

/****************************************************************************
REMARKS:
Stub to do nothing when no hardware acceleration is present.
****************************************************************************/
N_int32 NAPI REF2D_IsIdle(void)
{
    return 1;
}

/****************************************************************************
REMARKS:
Stub to do nothing when no hardware acceleration is present.
****************************************************************************/
void NAPI REF2D_WaitTillIdle(void)
{
}

/****************************************************************************
REMARKS:
Function to initialise the filter driver. The ref2d driver is always
loaded *after* a mode has been initialised, and hence we know all the
drawing vectors etc are correctly set up at this stage. This is unlike
a filter driver which gets loaded when the device driver is first loaded,
and then the InitFilter functions get called *after* this function has been
called when the 2D reference rasteriser library is actually loaded.
****************************************************************************/
static ibool REF2D_init(
    GA_devCtx *hwCtx,
    GA_exports *gaExp,
    GA_globalOptions *gOpt)
{
    GA_configInfo   configInfo;

    // Initialise device driver internals
    memset(&DC,0,sizeof(DC));
    DV.DriverState = &DC;
    DC.hwCtx                = hwCtx;
    DC.gOpt                 = gOpt;
    DC.colorCmpMask         = 0xFFFFFFFF;
    DC.planeMask            = 0xFFFFFFFF;
    DC.srcBlendFunc         = gaBlendNone;
    DC.dstBlendFunc         = gaBlendNone;
    DC.alphaValue           = 0xFF;
    DC.beginDirectAccess    = CNOP_VECTOR;
    DC.endDirectAccess      = CNOP_VECTOR;
    DC.LoadBufferMgr        = REF2D_loadBufferMgr;
    DC.UnloadBufferMgr      = REF2D_unloadBufferMgr;
    REF_noHWPlaneMask       = true;
    REF_bugFixHWROP3        = false;
    transMonoPattSW         = false;
    transColorPattSW        = false;
    memset(&REF_draw2dHW,0,sizeof(REF_draw2dHW));
    memset(&DC.hwCursor,0,sizeof(DC.hwCursor));
    memset(&DC.hwDrv,0,sizeof(DC.hwDrv));
#ifdef __DRIVER__
    _GA_exports.dwSize = sizeof(_GA_exports);
    COPY_STRUCTURE(&_GA_exports,gaExp);
#endif

    // Initialise the hardware rendering information if available
    if (hwCtx) {
        DC.hwInit.dwSize = sizeof(DC.hwInit);
        hwCtx->loader.QueryFunctions(GA_GET_INITFUNCS,-2,&DC.hwInit);
        DC.hwState.dwSize = sizeof(DC.hwState);
        if (hwCtx->loader.QueryFunctions(GA_GET_2DSTATEFUNCS,-2,&DC.hwState)) {
            REF_state2d.WaitTillIdle = DC.hwState.WaitTillIdle;
            if (DC.hwState.EnableDirectAccess) {
                REF_state2d.EnableDirectAccess  = DC.hwState.EnableDirectAccess;
                REF_state2d.DisableDirectAccess = DC.hwState.DisableDirectAccess;
                DC.beginDirectAccess            = DC.hwState.EnableDirectAccess;
                DC.endDirectAccess              = DC.hwState.DisableDirectAccess;
                }
            else {
                DC.beginDirectAccess            = DC.hwState.WaitTillIdle;
                }
            REF_noHWPlaneMask = (DC.hwState.SetPlaneMask == NULL);
            }
        DC.hwDraw.dwSize = sizeof(DC.hwDraw);
        DC.haveHWDraw = false;
        if (hwCtx->loader.QueryFunctions(GA_GET_2DRENDERFUNCS,-2,&DC.hwDraw)) {
            REF_draw2dHW = DC.hwDraw;
            DC.haveHWDraw = true;
            }
        DC.hwCursor.dwSize = sizeof(DC.hwCursor);
        hwCtx->loader.QueryFunctions(GA_GET_CURSORFUNCS,-2,&DC.hwCursor);
        DC.hwDrv.dwSize = sizeof(DC.hwDrv);
        hwCtx->loader.QueryFunctions(GA_GET_DRIVERFUNCS,-2,&DC.hwDrv);
        InitModeAndCursorInfo();

        // Set the slow blit sys flag if CPU blits are slower in replace mode
        REF_slowBltSysOrg = REF_slowBltSys = ((hwCtx->WorkArounds & gaSlowBltSys) != 0);

        // Check the version that the driver was certified with, and if
        // it is less than version 2.0, then the ROP3 functions are not
        // guaranteed to properly save/restore the hardware mix. Hence we
        // implement these functions via small stubs that call the
        // hardware function and then reset the correct ROP2 mix on exit.
        configInfo.dwSize = sizeof(configInfo);
        DC.hwInit.GetConfigInfo(&configInfo);
        if (configInfo.Certified && configInfo.CertifyVersion < 0x200)
            REF_bugFixHWROP3 = true;

        // Setup pointers to functions to copy buffer data for color depth
        switch (DC.gaMode.BitsPerPixel) {
#ifdef  SUPPORT_8BPP
            case 8:
                DC.bufReadScan  = REF2D_readScan_8;
                DC.bufBltScan   = REF2D_bltScan_8_S;
                break;
#endif
#ifdef  SUPPORT_16BPP
            case 15:
            case 16:
                DC.bufReadScan  = REF2D_readScan_16;
                DC.bufBltScan   = REF2D_bltScan_16_S;
                break;
#endif
#ifdef  SUPPORT_24BPP
            case 24:
                DC.bufReadScan  = REF2D_readScan_24;
                DC.bufBltScan   = REF2D_bltScan_24_S;
                break;
#endif
#ifdef  SUPPORT_32BPP
            case 32:
                DC.bufReadScan  = REF2D_readScan_32;
                DC.bufBltScan   = REF2D_bltScan_32_S;
                break;
#endif
            }

        // Now set the software rendering vectors in the driver to point
        // back to us so it can handle hardware workarounds.
        if (DC.hwInit.SetRef2dPointer)
            DC.hwInit.SetRef2dPointer(&REF2D_drv);
        if (DC.hwInit.SetSoftwareRenderFuncs)
            DC.hwInit.SetSoftwareRenderFuncs(&REF_draw2dSW);
        }
    REF2D_cursorInit();

    // Initialise default blit buffer sizes to the full screen resolution.
    // Note that when initialising a system memory buffer driver, DC.gaMode
    // is all zeros so we initialise the blt buffers to be empty forcing
    // them to be allocated on the first blit.
    if (DC.hwCtx)
        REF2D_resizeBltBuffers(DC.gaMode.XResolution);
    return true;
}

#ifdef  __DRIVER__
/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Function entry point for the reference rasteriser library.
****************************************************************************/
REF2D_driver * _CEXPORT REF2D_initLibrary(
    GA_devCtx *hwCtx,
    GA_globalOptions *gOpt,
    GA_exports *gaExp,
    PM_imports *pmImp)
{
    int     i,max;
    ulong   *p;

    // Bail if we have been here before
    refCount++;
    if (DC.hwCtx)
        return &DV;

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_PM_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    if (!REF2D_init(hwCtx,gaExp,gOpt))
        return NULL;
    (void)gOpt;
    return &DV;
}
#else
/****************************************************************************
REMARKS:
Function entry point to fake loading of a driver when linked into app.
****************************************************************************/
ibool NAPI REF2D_loadDriver(
    GA_devCtx *hwCtx,
    N_int32 bitsPerPixel,
    N_int32 shared,
    REF2D_driver **drv,
    PE_MODULE **hMod,
    ulong *size)
{
    *drv = &DV;
    *hMod = NULL;
    *size = 0;
    refCount++;
    if (DC.hwCtx)
        return true;
    (void)bitsPerPixel;
    (void)shared;
    return REF2D_init(hwCtx,NULL,NULL);
}
#endif

/****************************************************************************
REMARKS:
Function to clean up when unloading the driver.
****************************************************************************/
void NAPI REF2D_UnloadDriverInt(void)
{
    if (--refCount == 0) {
        free(DC.scanBuf);
#ifdef TEST_HARNESS
        if (REF_buffer.dwSize)
            REF2D_unloadBufferMgr(DC.hModBuffer);
#else
        if (DC.hModBuffer)
            REF2D_unloadBufferMgr(DC.hModBuffer);
#endif
        memset(&REF_buffer,0,sizeof(REF_buffer));
        if (DC.hModRegion)
            GA_unloadRegionMgr(DC.hModRegion);
        memset(&REF_region,0,sizeof(REF_region));
        memset(&DC,0,sizeof(DC));
        }
}

/****************************************************************************
REMARKS:
This function resizes the internal scanline blit buffers to be large
enough to handle the incoming blit width. Because we do this dynamically,
we can essentially handle any sized system memory buffers dependant on
the amount of available video memory.
****************************************************************************/
void REF2D_resizeBltBuffers(
    N_int32 width)
{
    N_int32 totalSize;

    // Determine width in pixels and bytes
#ifdef SUPPORT_ALLBPP
    DC.maxBltWidth = width;
    DC.maxBltBytes = DC.maxBltWidth * 4;
#elif defined(SUPPORT_1BPP)
    DC.maxBltWidth = ((width + 0x7) & ~0x7);
    DC.maxBltBytes = DC.maxBltWidth / 8;
#elif defined(SUPPORT_4BPP)
    DC.maxBltWidth = ((width + 0x1) & ~0x1);
    DC.maxBltBytes = DC.maxBltWidth / 2;
#elif defined(SUPPORT_8BPP)
    DC.maxBltWidth = width;
    DC.maxBltBytes = DC.maxBltWidth;
#elif defined(SUPPORT_16BPP)
    DC.maxBltWidth = width;
    DC.maxBltBytes = DC.maxBltWidth * 2;
#elif defined(SUPPORT_24BPP)
    DC.maxBltWidth = width;
    DC.maxBltBytes = DC.maxBltWidth * 3;
#elif defined(SUPPORT_32BPP)
    DC.maxBltWidth = width;
    DC.maxBltBytes = DC.maxBltWidth * 4;
#endif

    // Free the old buffers if they were already allocated
    if (DC.scanBuf) free(DC.scanBuf);

    // Allocate memory for the temporary scanline buffers
    totalSize = (DC.maxBltBytes * 2) + (DC.maxBltWidth * sizeof(int));
    if ((DC.scanBuf = malloc(totalSize)) == NULL)
        PM_fatalError("Out of memory!");
    DC.scanBuf2 = DC.scanBuf + DC.maxBltBytes;
    DC.repCounts = (int*)(DC.scanBuf2 + DC.maxBltBytes);
}

#ifdef  SUPPORT_1BPP
/****************************************************************************
REMARKS:
This function sets the pixel drawing functions for the color depth
****************************************************************************/
static void REF2D_SetPixelFuncs1bpp(void)
{
    static REF2D_drawPixel_t pixelFuncs[] = {
        DEFINE_ROP2_PIXELS(1),
        };
    static REF2D_drawPixelR3_t pixelFuncsR3[] = {
        DEFINE_ROP3_PIXELS(1),
        };
    DC.drawPixelTable = pixelFuncs;
    DC.drawPixelR3Table = pixelFuncsR3;
    DC.drawPixelSrcColorKey         = REF2D_drawPixelSrcColorKey_1;
    DC.drawPixelSrcColorKeyRange    = REF2D_drawPixelSrcColorKeyRange_1;
}

/****************************************************************************
REMARKS:
This function sets the hardware mix operation for subsequent drawing for
a specific color depth.
****************************************************************************/
static void REF2D_SetDrawBuffer1bpp(void)
{
    REF2D_SetPixelFuncs1bpp();
    DC.readPixel                = REF2D_readPixel_1;
    DC.fillScan_S               = REF2D_fillScan_1_S;
    DC.fillScan_Generic         = REF2D_fillScan_1_Generic;
    DC.drawMonoCursor           = REF2D_drawMonoCursor_1;
    DC.drawColorCursor          = REF2D_drawColorCursor_1;
    DC.readScan                 = REF2D_readScan_1;
    DC.bltScan_S                = REF2D_bltScan_1_S;
    DC.bltScan_Generic          = REF2D_bltScan_1_Generic;
    DC.bltScanPatt              = REF2D_bltScanPatt_1;
    DC.bltScanColorPatt         = REF2D_bltScanColorPatt_1;
    DC.srcTransBltScan_S        = REF2D_srcTransBltScan_1_S;
    DC.srcTransBltScan_Generic  = REF2D_srcTransBltScan_1_Generic;
    DC.srcTransBltScan_Masked   = REF2D_srcTransBltScan_1_Masked;
    DC.dstTransBltScan_S        = REF2D_dstTransBltScan_1_S;
    DC.dstTransBltScan_Generic  = REF2D_dstTransBltScan_1_Generic;
    DC.dstTransBltScan_Masked   = REF2D_dstTransBltScan_1_Masked;
    DC.stretchScan2x_S          = REF2D_stretchScan2x_1_S;
    DC.stretchScan2x_Generic    = REF2D_stretchScan2x_1_Generic;
    DC.stretchScan_S            = REF2D_stretchScan_1_S;
    DC.stretchScan_Generic      = REF2D_stretchScan_1_Generic;
    DC.buildTranslateVector     = REF2D_buildTranslateVector_1;
    DC.convScan4_S              = REF2D_convScan4_1_S;
    DC.convScan4_Generic        = REF2D_convScan4_1_Generic;
    DC.convScan8_S              = REF2D_convScan8_1_S;
    DC.convScan8_Generic        = REF2D_convScan8_1_Generic;
    DC.convScan15_S             = REF2D_convScan16_1_Generic;
    DC.convScan15_Generic       = REF2D_convScan16_1_Generic;
    DC.convScan16_S             = REF2D_convScan16_1_Generic;
    DC.convScan16_Generic       = REF2D_convScan16_1_Generic;
    DC.convScan24RGB_S          = REF2D_convScan24_1_Generic;
    DC.convScan24RGB_Generic    = REF2D_convScan24_1_Generic;
    DC.convScan24BGR_S          = REF2D_convScan24_1_Generic;
    DC.convScan24BGR_Generic    = REF2D_convScan24_1_Generic;
    DC.convScan32ARGB_S         = REF2D_convScan32_1_Generic;
    DC.convScan32ARGB_Generic   = REF2D_convScan32_1_Generic;
    DC.convScan32ABGR_S         = REF2D_convScan32_1_Generic;
    DC.convScan32ABGR_Generic   = REF2D_convScan32_1_Generic;
    DC.convScan32RGBA_S         = REF2D_convScan32_1_Generic;
    DC.convScan32RGBA_Generic   = REF2D_convScan32_1_Generic;
    DC.convScan32BGRA_S         = REF2D_convScan32_1_Generic;
    DC.convScan32BGRA_Generic   = REF2D_convScan32_1_Generic;
    DC.ditherScan15_S           = REF2D_ditherScan15_1_S;
    DC.ditherScan15_Generic     = REF2D_ditherScan15_1_Generic;
    DC.ditherScan16_S           = REF2D_ditherScan16_1_S;
    DC.ditherScan16_Generic     = REF2D_ditherScan16_1_Generic;
    DC.ditherScanRGB_S          = REF2D_ditherScanRGB_1_S;
    DC.ditherScanRGB_Generic    = REF2D_ditherScanRGB_1_Generic;
    DC.ditherScanBGR_S          = REF2D_ditherScanBGR_1_S;
    DC.ditherScanBGR_Generic    = REF2D_ditherScanBGR_1_Generic;
    DC.ditherScanARGB_S         = REF2D_ditherScanARGB_1_S;
    DC.ditherScanARGB_Generic   = REF2D_ditherScanARGB_1_Generic;
    DC.ditherScanABGR_S         = REF2D_ditherScanABGR_1_S;
    DC.ditherScanABGR_Generic   = REF2D_ditherScanABGR_1_Generic;
    DC.ditherScanRGBA_S         = REF2D_ditherScanRGBA_1_S;
    DC.ditherScanRGBA_Generic   = REF2D_ditherScanRGBA_1_Generic;
    DC.ditherScanBGRA_S         = REF2D_ditherScanBGRA_1_S;
    DC.ditherScanBGRA_Generic   = REF2D_ditherScanBGRA_1_Generic;
    DC.clipEllipse              = REF2D_clipEllipse_1_Generic;
}

/****************************************************************************
PARAMETERS:
mix     - New mix code to set (GA_mixCodesType)

REMARKS:
This function sets the hardware mix operation for subsequent drawing for
a specific color depth.
****************************************************************************/
static void REF2D_SetMix1bpp(
    int mix)
{
    DC.drawPixel = GET_PIXEL_FUNC(mix);
    if (IS_SOLID_MIX(mix)) {
        DC.drawScan                 = REF2D_drawScan_1_S;
        DC.drawPattScan             = REF2D_drawPattScan_1_S;
        DC.drawColorPattScan        = REF2D_drawColorPattScan_1_S;
        DC.drawLineXPosLo           = REF2D_drawLineXPosLo_1_S;
        DC.drawLineXPosHi           = REF2D_drawLineXPosHi_1_S;
        DC.drawLineXNegLo           = REF2D_drawLineXNegLo_1_S;
        DC.drawLineXNegHi           = REF2D_drawLineXNegHi_1_S;
        DC.drawStippleLineXPosLo    = REF2D_drawStippleLineXPosLo_1_S;
        DC.drawStippleLineXPosHi    = REF2D_drawStippleLineXPosHi_1_S;
        DC.drawStippleLineXNegLo    = REF2D_drawStippleLineXNegLo_1_S;
        DC.drawStippleLineXNegHi    = REF2D_drawStippleLineXNegHi_1_S;
        DC.drawStyleLineXPosLo      = REF2D_drawStyleLineXPosLo_1_S;
        DC.drawStyleLineXPosHi      = REF2D_drawStyleLineXPosHi_1_S;
        DC.drawStyleLineXNegLo      = REF2D_drawStyleLineXNegLo_1_S;
        DC.drawStyleLineXNegHi      = REF2D_drawStyleLineXNegHi_1_S;
        DC.drawMonoMSBTrans         = REF2D_drawMonoMSBTrans_1_S;
        DC.drawMonoMSBOpaque        = REF2D_drawMonoMSBOpaque_1_S;
        DC.drawClipMonoMSBTrans     = REF2D_drawClipMonoMSBTrans_1_S;
        DC.drawClipMonoMSBOpaque    = REF2D_drawClipMonoMSBOpaque_1_S;
        DC.drawMonoLSBTrans         = REF2D_drawMonoLSBTrans_1_S;
        DC.drawMonoLSBOpaque        = REF2D_drawMonoLSBOpaque_1_S;
        DC.drawClipMonoLSBTrans     = REF2D_drawClipMonoLSBTrans_1_S;
        DC.drawClipMonoLSBOpaque    = REF2D_drawClipMonoLSBOpaque_1_S;
        DC.drawEllipse              = REF2D_drawEllipse_1_S;
        }
    else {
        DC.drawScan                 = REF2D_drawScan_1_Generic;
        DC.drawPattScan             = REF2D_drawPattScan_1_Generic;
        DC.drawColorPattScan        = REF2D_drawColorPattScan_1_Generic;
        DC.drawLineXPosLo           = REF2D_drawLineXPosLo_1_Generic;
        DC.drawLineXPosHi           = REF2D_drawLineXPosHi_1_Generic;
        DC.drawLineXNegLo           = REF2D_drawLineXNegLo_1_Generic;
        DC.drawLineXNegHi           = REF2D_drawLineXNegHi_1_Generic;
        DC.drawStippleLineXPosLo    = REF2D_drawStippleLineXPosLo_1_Generic;
        DC.drawStippleLineXPosHi    = REF2D_drawStippleLineXPosHi_1_Generic;
        DC.drawStippleLineXNegLo    = REF2D_drawStippleLineXNegLo_1_Generic;
        DC.drawStippleLineXNegHi    = REF2D_drawStippleLineXNegHi_1_Generic;
        DC.drawStyleLineXPosLo      = REF2D_drawStyleLineXPosLo_1_Generic;
        DC.drawStyleLineXPosHi      = REF2D_drawStyleLineXPosHi_1_Generic;
        DC.drawStyleLineXNegLo      = REF2D_drawStyleLineXNegLo_1_Generic;
        DC.drawStyleLineXNegHi      = REF2D_drawStyleLineXNegHi_1_Generic;
        DC.drawMonoMSBTrans         = REF2D_drawMonoMSBTrans_1_Generic;
        DC.drawMonoMSBOpaque        = REF2D_drawMonoMSBOpaque_1_Generic;
        DC.drawClipMonoMSBTrans     = REF2D_drawClipMonoMSBTrans_1_Generic;
        DC.drawClipMonoMSBOpaque    = REF2D_drawClipMonoMSBOpaque_1_Generic;
        DC.drawMonoLSBTrans         = REF2D_drawMonoLSBTrans_1_Generic;
        DC.drawMonoLSBOpaque        = REF2D_drawMonoLSBOpaque_1_Generic;
        DC.drawClipMonoLSBTrans     = REF2D_drawClipMonoLSBTrans_1_Generic;
        DC.drawClipMonoLSBOpaque    = REF2D_drawClipMonoLSBOpaque_1_Generic;
        DC.drawEllipse              = REF2D_drawEllipse_1_Generic;
        }
}
#endif

#ifdef  SUPPORT_4BPP
/****************************************************************************
REMARKS:
This function sets the pixel drawing functions for the color depth
****************************************************************************/
static void REF2D_SetPixelFuncs4bpp(void)
{
    static REF2D_drawPixel_t pixelFuncs[] = {
        DEFINE_ROP2_PIXELS(4),
        };
    static REF2D_drawPixel_t pixelFuncsMasked[] = {
        DEFINE_ROP2_MASKED_PIXELS(4),
        };
    static REF2D_drawPixelR3_t pixelFuncsR3[] = {
        DEFINE_ROP3_PIXELS(4),
        };
    static REF2D_drawPixelR3_t pixelFuncsR3Masked[] = {
        DEFINE_ROP3_MASKED_PIXELS(4),
        };

    if (DC.planeMask != 0xFFFFFFFF) {
        DC.drawPixelTable = pixelFuncsMasked;
        DC.drawPixelR3Table = pixelFuncsR3Masked;
        }
    else {
        DC.drawPixelTable = pixelFuncs;
        DC.drawPixelR3Table = pixelFuncsR3;
        }
    DC.drawPixelSrcColorKey         = REF2D_drawPixelSrcColorKey_4;
    DC.drawPixelSrcColorKeyRange    = REF2D_drawPixelSrcColorKeyRange_4;
}

/****************************************************************************
REMARKS:
This function sets the hardware mix operation for subsequent drawing for
a specific color depth.
****************************************************************************/
static void REF2D_SetDrawBuffer4bpp(void)
{
    REF2D_SetPixelFuncs4bpp();
    DC.readPixel                = REF2D_readPixel_4;
    DC.fillScan_S               = REF2D_fillScan_4_S;
    DC.fillScan_Generic         = REF2D_fillScan_4_Generic;
    DC.drawMonoCursor           = REF2D_drawMonoCursor_4;
    DC.drawColorCursor          = REF2D_drawColorCursor_4;
    DC.readScan                 = REF2D_readScan_4;
    DC.bltScan_S                = REF2D_bltScan_4_S;
    DC.bltScan_Generic          = REF2D_bltScan_4_Generic;
    DC.bltScanPatt              = REF2D_bltScanPatt_4;
    DC.bltScanColorPatt         = REF2D_bltScanColorPatt_4;
    DC.srcTransBltScan_S        = REF2D_srcTransBltScan_4_S;
    DC.srcTransBltScan_Generic  = REF2D_srcTransBltScan_4_Generic;
    DC.srcTransBltScan_Masked   = REF2D_srcTransBltScan_4_Masked;
    DC.dstTransBltScan_S        = REF2D_dstTransBltScan_4_S;
    DC.dstTransBltScan_Generic  = REF2D_dstTransBltScan_4_Generic;
    DC.dstTransBltScan_Masked   = REF2D_dstTransBltScan_4_Masked;
    DC.stretchScan2x_S          = REF2D_stretchScan2x_4_S;
    DC.stretchScan2x_Generic    = REF2D_stretchScan2x_4_Generic;
    DC.stretchScan_S            = REF2D_stretchScan_4_S;
    DC.stretchScan_Generic      = REF2D_stretchScan_4_Generic;
    DC.buildTranslateVector     = REF2D_buildTranslateVector_4;
    DC.convScan4_S              = REF2D_convScan4_4_S;
    DC.convScan8_S              = REF2D_convScan8_4_S;
    DC.clipEllipse              = REF2D_clipEllipse_4_Generic;
}

/****************************************************************************
PARAMETERS:
mix     - New mix code to set (GA_mixCodesType)

REMARKS:
This function sets the hardware mix operation for subsequent drawing for
a specific color depth.
****************************************************************************/
static void REF2D_SetMix4bpp(
    int mix)
{
    DC.drawPixel = GET_PIXEL_FUNC(mix);
    if (IS_SOLID_MIX(mix)) {
        DC.drawScan                 = REF2D_drawScan_4_S;
        DC.drawPattScan             = REF2D_drawPattScan_4_S;
        DC.drawColorPattScan        = REF2D_drawColorPattScan_4_S;
        DC.drawLineXPosLo           = REF2D_drawLineXPosLo_4_S;
        DC.drawLineXPosHi           = REF2D_drawLineXPosHi_4_S;
        DC.drawLineXNegLo           = REF2D_drawLineXNegLo_4_S;
        DC.drawLineXNegHi           = REF2D_drawLineXNegHi_4_S;
        DC.drawStippleLineXPosLo    = REF2D_drawStippleLineXPosLo_4_S;
        DC.drawStippleLineXPosHi    = REF2D_drawStippleLineXPosHi_4_S;
        DC.drawStippleLineXNegLo    = REF2D_drawStippleLineXNegLo_4_S;
        DC.drawStippleLineXNegHi    = REF2D_drawStippleLineXNegHi_4_S;
        DC.drawStyleLineXPosLo      = REF2D_drawStyleLineXPosLo_4_S;
        DC.drawStyleLineXPosHi      = REF2D_drawStyleLineXPosHi_4_S;
        DC.drawStyleLineXNegLo      = REF2D_drawStyleLineXNegLo_4_S;
        DC.drawStyleLineXNegHi      = REF2D_drawStyleLineXNegHi_4_S;
        DC.drawMonoMSBTrans         = REF2D_drawMonoMSBTrans_4_S;
        DC.drawMonoMSBOpaque        = REF2D_drawMonoMSBOpaque_4_S;
        DC.drawClipMonoMSBTrans     = REF2D_drawClipMonoMSBTrans_4_S;
        DC.drawClipMonoMSBOpaque    = REF2D_drawClipMonoMSBOpaque_4_S;
        DC.drawMonoLSBTrans         = REF2D_drawMonoLSBTrans_4_S;
        DC.drawMonoLSBOpaque        = REF2D_drawMonoLSBOpaque_4_S;
        DC.drawClipMonoLSBTrans     = REF2D_drawClipMonoLSBTrans_4_S;
        DC.drawClipMonoLSBOpaque    = REF2D_drawClipMonoLSBOpaque_4_S;
        DC.drawEllipse              = REF2D_drawEllipse_4_S;
        }
    else {
        DC.drawScan                 = REF2D_drawScan_4_Generic;
        DC.drawPattScan             = REF2D_drawPattScan_4_Generic;
        DC.drawColorPattScan        = REF2D_drawColorPattScan_4_Generic;
        DC.drawLineXPosLo           = REF2D_drawLineXPosLo_4_Generic;
        DC.drawLineXPosHi           = REF2D_drawLineXPosHi_4_Generic;
        DC.drawLineXNegLo           = REF2D_drawLineXNegLo_4_Generic;
        DC.drawLineXNegHi           = REF2D_drawLineXNegHi_4_Generic;
        DC.drawStippleLineXPosLo    = REF2D_drawStippleLineXPosLo_4_Generic;
        DC.drawStippleLineXPosHi    = REF2D_drawStippleLineXPosHi_4_Generic;
        DC.drawStippleLineXNegLo    = REF2D_drawStippleLineXNegLo_4_Generic;
        DC.drawStippleLineXNegHi    = REF2D_drawStippleLineXNegHi_4_Generic;
        DC.drawStyleLineXPosLo      = REF2D_drawStyleLineXPosLo_4_Generic;
        DC.drawStyleLineXPosHi      = REF2D_drawStyleLineXPosHi_4_Generic;
        DC.drawStyleLineXNegLo      = REF2D_drawStyleLineXNegLo_4_Generic;
        DC.drawStyleLineXNegHi      = REF2D_drawStyleLineXNegHi_4_Generic;
        DC.drawMonoMSBTrans         = REF2D_drawMonoMSBTrans_4_Generic;
        DC.drawMonoMSBOpaque        = REF2D_drawMonoMSBOpaque_4_Generic;
        DC.drawClipMonoMSBTrans     = REF2D_drawClipMonoMSBTrans_4_Generic;
        DC.drawClipMonoMSBOpaque    = REF2D_drawClipMonoMSBOpaque_4_Generic;
        DC.drawMonoLSBTrans         = REF2D_drawMonoLSBTrans_4_Generic;
        DC.drawMonoLSBOpaque        = REF2D_drawMonoLSBOpaque_4_Generic;
        DC.drawClipMonoLSBTrans     = REF2D_drawClipMonoLSBTrans_4_Generic;
        DC.drawClipMonoLSBOpaque    = REF2D_drawClipMonoLSBOpaque_4_Generic;
        DC.drawEllipse              = REF2D_drawEllipse_4_Generic;
        }
}
#endif

#ifdef  SUPPORT_8BPP
/****************************************************************************
REMARKS:
This function sets the pixel drawing functions for the color depth
****************************************************************************/
static void REF2D_SetPixelFuncs8bpp(void)
{
    static REF2D_drawPixel_t pixelFuncs[] = {
        DEFINE_ROP2_PIXELS(8),
        };
    static REF2D_drawPixel_t pixelFuncsMasked[] = {
        DEFINE_ROP2_MASKED_PIXELS(8),
        };
    static REF2D_drawPixelR3_t pixelFuncsR3[] = {
        DEFINE_ROP3_PIXELS(8),
        };
    static REF2D_drawPixelR3_t pixelFuncsR3Masked[] = {
        DEFINE_ROP3_MASKED_PIXELS(8),
        };

    if (DC.planeMask != 0xFFFFFFFF) {
        DC.drawPixelTable = pixelFuncsMasked;
        DC.drawPixelR3Table = pixelFuncsR3Masked;
        }
    else {
        DC.drawPixelTable = pixelFuncs;
        DC.drawPixelR3Table = pixelFuncsR3;
        }
    DC.drawPixelSrcColorKey         = REF2D_drawPixelSrcColorKey_8;
    DC.drawPixelSrcColorKeyRange    = REF2D_drawPixelSrcColorKeyRange_8;
}

/****************************************************************************
REMARKS:
This function sets the hardware mix operation for subsequent drawing for
a specific color depth.
****************************************************************************/
static void REF2D_SetDrawBuffer8bpp(void)
{
    REF2D_SetPixelFuncs8bpp();
    DC.readPixel                            = REF2D_readPixel_8;
    DC.fillScan_S                           = REF2D_fillScan_8_S;
    DC.fillScan_Generic                     = REF2D_fillScan_8_Generic;
    DC.drawMonoCursor                       = REF2D_drawMonoCursor_8;
    DC.drawColorCursor                      = REF2D_drawColorCursor_8;
    DC.readScan                             = REF2D_readScan_8;
    DC.bltScan_S                            = REF2D_bltScan_8_S;
    DC.bltScan_Generic                      = REF2D_bltScan_8_Generic;
    DC.bltScanPatt                          = REF2D_bltScanPatt_8;
    DC.bltScanColorPatt                     = REF2D_bltScanColorPatt_8;
    DC.srcTransBltScan_S                    = REF2D_srcTransBltScan_8_S;
    DC.srcTransBltScan_Generic              = REF2D_srcTransBltScan_8_Generic;
    DC.srcTransBltScan_Masked               = REF2D_srcTransBltScan_8_Masked;
    DC.dstTransBltScan_S                    = REF2D_dstTransBltScan_8_S;
    DC.dstTransBltScan_Generic              = REF2D_dstTransBltScan_8_Generic;
    DC.dstTransBltScan_Masked               = REF2D_dstTransBltScan_8_Masked;
    DC.stretchScan2x_S                      = REF2D_stretchScan2x_8_S;
    DC.stretchScan2x_Generic                = REF2D_stretchScan2x_8_Generic;
    DC.stretchScan_S                        = REF2D_stretchScan_8_S;
    DC.stretchScan_Generic                  = REF2D_stretchScan_8_Generic;
    DC.buildTranslateVector                 = REF2D_buildTranslateVector_8;
    DC.convScan4_S                          = REF2D_convScan4_8_S;
    DC.convScan8_S                          = REF2D_convScan8_8_S;
    DC.clipEllipse                          = REF2D_clipEllipse_8_Generic;
}

/****************************************************************************
PARAMETERS:
mix     - New mix code to set (GA_mixCodesType)

REMARKS:
This function sets the hardware mix operation for subsequent drawing for
a specific color depth.
****************************************************************************/
static void REF2D_SetMix8bpp(
    int mix)
{
    DC.drawPixel = GET_PIXEL_FUNC(mix);
    if (IS_SOLID_MIX(mix)) {
        DC.drawScan                 = REF2D_drawScan_8_S;
        DC.drawPattScan             = REF2D_drawPattScan_8_S;
        DC.drawColorPattScan        = REF2D_drawColorPattScan_8_S;
        DC.drawLineXPosLo           = REF2D_drawLineXPosLo_8_S;
        DC.drawLineXPosHi           = REF2D_drawLineXPosHi_8_S;
        DC.drawLineXNegLo           = REF2D_drawLineXNegLo_8_S;
        DC.drawLineXNegHi           = REF2D_drawLineXNegHi_8_S;
        DC.drawStippleLineXPosLo    = REF2D_drawStippleLineXPosLo_8_S;
        DC.drawStippleLineXPosHi    = REF2D_drawStippleLineXPosHi_8_S;
        DC.drawStippleLineXNegLo    = REF2D_drawStippleLineXNegLo_8_S;
        DC.drawStippleLineXNegHi    = REF2D_drawStippleLineXNegHi_8_S;
        DC.drawStyleLineXPosLo      = REF2D_drawStyleLineXPosLo_8_S;
        DC.drawStyleLineXPosHi      = REF2D_drawStyleLineXPosHi_8_S;
        DC.drawStyleLineXNegLo      = REF2D_drawStyleLineXNegLo_8_S;
        DC.drawStyleLineXNegHi      = REF2D_drawStyleLineXNegHi_8_S;
        DC.drawMonoMSBTrans         = REF2D_drawMonoMSBTrans_8_S;
        DC.drawMonoMSBOpaque        = REF2D_drawMonoMSBOpaque_8_S;
        DC.drawClipMonoMSBTrans     = REF2D_drawClipMonoMSBTrans_8_S;
        DC.drawClipMonoMSBOpaque    = REF2D_drawClipMonoMSBOpaque_8_S;
        DC.drawMonoLSBTrans         = REF2D_drawMonoLSBTrans_8_S;
        DC.drawMonoLSBOpaque        = REF2D_drawMonoLSBOpaque_8_S;
        DC.drawClipMonoLSBTrans     = REF2D_drawClipMonoLSBTrans_8_S;
        DC.drawClipMonoLSBOpaque    = REF2D_drawClipMonoLSBOpaque_8_S;
        DC.drawEllipse              = REF2D_drawEllipse_8_S;
        }
    else {
        DC.drawScan                 = REF2D_drawScan_8_Generic;
        DC.drawPattScan             = REF2D_drawPattScan_8_Generic;
        DC.drawColorPattScan        = REF2D_drawColorPattScan_8_Generic;
        DC.drawLineXPosLo           = REF2D_drawLineXPosLo_8_Generic;
        DC.drawLineXPosHi           = REF2D_drawLineXPosHi_8_Generic;
        DC.drawLineXNegLo           = REF2D_drawLineXNegLo_8_Generic;
        DC.drawLineXNegHi           = REF2D_drawLineXNegHi_8_Generic;
        DC.drawStippleLineXPosLo    = REF2D_drawStippleLineXPosLo_8_Generic;
        DC.drawStippleLineXPosHi    = REF2D_drawStippleLineXPosHi_8_Generic;
        DC.drawStippleLineXNegLo    = REF2D_drawStippleLineXNegLo_8_Generic;
        DC.drawStippleLineXNegHi    = REF2D_drawStippleLineXNegHi_8_Generic;
        DC.drawStyleLineXPosLo      = REF2D_drawStyleLineXPosLo_8_Generic;
        DC.drawStyleLineXPosHi      = REF2D_drawStyleLineXPosHi_8_Generic;
        DC.drawStyleLineXNegLo      = REF2D_drawStyleLineXNegLo_8_Generic;
        DC.drawStyleLineXNegHi      = REF2D_drawStyleLineXNegHi_8_Generic;
        DC.drawMonoMSBTrans         = REF2D_drawMonoMSBTrans_8_Generic;
        DC.drawMonoMSBOpaque        = REF2D_drawMonoMSBOpaque_8_Generic;
        DC.drawClipMonoMSBTrans     = REF2D_drawClipMonoMSBTrans_8_Generic;
        DC.drawClipMonoMSBOpaque    = REF2D_drawClipMonoMSBOpaque_8_Generic;
        DC.drawMonoLSBTrans         = REF2D_drawMonoLSBTrans_8_Generic;
        DC.drawMonoLSBOpaque        = REF2D_drawMonoLSBOpaque_8_Generic;
        DC.drawClipMonoLSBTrans     = REF2D_drawClipMonoLSBTrans_8_Generic;
        DC.drawClipMonoLSBOpaque    = REF2D_drawClipMonoLSBOpaque_8_Generic;
        DC.drawEllipse              = REF2D_drawEllipse_8_Generic;
        }
}
#endif

#ifdef  SUPPORT_16BPP
/****************************************************************************
REMARKS:
This function sets the pixel drawing functions for the color depth
****************************************************************************/
static void REF2D_SetPixelFuncs16bpp(void)
{
    static REF2D_drawPixel_t pixelFuncs[] = {
        DEFINE_ROP2_PIXELS(16),
        };
    static REF2D_drawPixel_t pixelFuncsMasked[] = {
        DEFINE_ROP2_MASKED_PIXELS(16),
        };
    static REF2D_drawPixelR3_t pixelFuncsR3[] = {
        DEFINE_ROP3_PIXELS(16),
        };
    static REF2D_drawPixelR3_t pixelFuncsR3Masked[] = {
        DEFINE_ROP3_MASKED_PIXELS(16),
        };

    if (DC.planeMask != 0xFFFFFFFF) {
        DC.drawPixelTable = pixelFuncsMasked;
        DC.drawPixelR3Table = pixelFuncsR3Masked;
        }
    else {
        DC.drawPixelTable = pixelFuncs;
        DC.drawPixelR3Table = pixelFuncsR3;
        }
    DC.drawPixelSrcColorKey         = REF2D_drawPixelSrcColorKey_16;
    DC.drawPixelSrcColorKeyRange    = REF2D_drawPixelSrcColorKeyRange_16;
}

/****************************************************************************
REMARKS:
This function sets the hardware mix operation for subsequent drawing for
a specific color depth.
****************************************************************************/
static void REF2D_SetDrawBuffer16bpp(void)
{
    DEFINE_BLENDED_PIXELS(15);
    DEFINE_BLENDED_PIXELS(16);

    REF2D_SetPixelFuncs16bpp();
    DC.readPixel                            = REF2D_readPixel_16;
    DC.fillScan_S                           = REF2D_fillScan_16_S;
    DC.fillScan_Generic                     = REF2D_fillScan_16_Generic;
    DC.drawMonoCursor                       = REF2D_drawMonoCursor_16;
    DC.drawColorCursor                      = REF2D_drawColorCursor_16;
    DC.readScan                             = REF2D_readScan_16;
    DC.bltScan_S                            = REF2D_bltScan_16_S;
    DC.bltScan_Generic                      = REF2D_bltScan_16_Generic;
    DC.bltScanPatt                          = REF2D_bltScanPatt_16;
    DC.bltScanColorPatt                     = REF2D_bltScanColorPatt_16;
    DC.srcTransBltScan_S                    = REF2D_srcTransBltScan_16_S;
    DC.srcTransBltScan_Generic              = REF2D_srcTransBltScan_16_Generic;
    DC.srcTransBltScan_Masked               = REF2D_srcTransBltScan_16_Masked;
    DC.dstTransBltScan_S                    = REF2D_dstTransBltScan_16_S;
    DC.dstTransBltScan_Generic              = REF2D_dstTransBltScan_16_Generic;
    DC.dstTransBltScan_Masked               = REF2D_dstTransBltScan_16_Masked;
    DC.stretchScan2x_S                      = REF2D_stretchScan2x_16_S;
    DC.stretchScan2x_Generic                = REF2D_stretchScan2x_16_Generic;
    DC.stretchScan_S                        = REF2D_stretchScan_16_S;
    DC.stretchScan_Generic                  = REF2D_stretchScan_16_Generic;
    DC.clipEllipse                          = REF2D_clipEllipse_16_Generic;
    DC.convScan4_S                          = REF2D_convScan4_16_S;
    DC.convScan8_S                          = REF2D_convScan8_16_S;
    if (DC.mi.bitsPerPixel == 15) {
        DC.blendPixelTable                  = blendPixelFuncs15;
        DC.blendPixelFast                   = REF2D_drawPixelBlendedFast_15;
        DC.buildTranslateVector             = REF2D_buildTranslateVector_15;
        }
    else {
        DC.blendPixelTable                  = blendPixelFuncs16;
        DC.blendPixelFast                   = REF2D_drawPixelBlendedFast_16;
        DC.buildTranslateVector             = REF2D_buildTranslateVector_16;
        }
}

/****************************************************************************
PARAMETERS:
mix     - New mix code to set (GA_mixCodesType)

REMARKS:
This function sets the hardware mix operation for subsequent drawing for
a specific color depth.
****************************************************************************/
static void REF2D_SetMix16bpp(
    int mix)
{
    DC.drawPixel = GET_PIXEL_FUNC(mix);
    if (IS_SOLID_MIX(mix)) {
        DC.drawScan                 = REF2D_drawScan_16_S;
        DC.drawPattScan             = REF2D_drawPattScan_16_S;
        DC.drawColorPattScan        = REF2D_drawColorPattScan_16_S;
        DC.drawLineXPosLo           = REF2D_drawLineXPosLo_16_S;
        DC.drawLineXPosHi           = REF2D_drawLineXPosHi_16_S;
        DC.drawLineXNegLo           = REF2D_drawLineXNegLo_16_S;
        DC.drawLineXNegHi           = REF2D_drawLineXNegHi_16_S;
        DC.drawStippleLineXPosLo    = REF2D_drawStippleLineXPosLo_16_S;
        DC.drawStippleLineXPosHi    = REF2D_drawStippleLineXPosHi_16_S;
        DC.drawStippleLineXNegLo    = REF2D_drawStippleLineXNegLo_16_S;
        DC.drawStippleLineXNegHi    = REF2D_drawStippleLineXNegHi_16_S;
        DC.drawStyleLineXPosLo      = REF2D_drawStyleLineXPosLo_16_S;
        DC.drawStyleLineXPosHi      = REF2D_drawStyleLineXPosHi_16_S;
        DC.drawStyleLineXNegLo      = REF2D_drawStyleLineXNegLo_16_S;
        DC.drawStyleLineXNegHi      = REF2D_drawStyleLineXNegHi_16_S;
        DC.drawMonoMSBTrans         = REF2D_drawMonoMSBTrans_16_S;
        DC.drawMonoMSBOpaque        = REF2D_drawMonoMSBOpaque_16_S;
        DC.drawClipMonoMSBTrans     = REF2D_drawClipMonoMSBTrans_16_S;
        DC.drawClipMonoMSBOpaque    = REF2D_drawClipMonoMSBOpaque_16_S;
        DC.drawMonoLSBTrans         = REF2D_drawMonoLSBTrans_16_S;
        DC.drawMonoLSBOpaque        = REF2D_drawMonoLSBOpaque_16_S;
        DC.drawClipMonoLSBTrans     = REF2D_drawClipMonoLSBTrans_16_S;
        DC.drawClipMonoLSBOpaque    = REF2D_drawClipMonoLSBOpaque_16_S;
        DC.drawEllipse              = REF2D_drawEllipse_16_S;
        }
    else {
        DC.drawScan                 = REF2D_drawScan_16_Generic;
        DC.drawPattScan             = REF2D_drawPattScan_16_Generic;
        DC.drawColorPattScan        = REF2D_drawColorPattScan_16_Generic;
        DC.drawLineXPosLo           = REF2D_drawLineXPosLo_16_Generic;
        DC.drawLineXPosHi           = REF2D_drawLineXPosHi_16_Generic;
        DC.drawLineXNegLo           = REF2D_drawLineXNegLo_16_Generic;
        DC.drawLineXNegHi           = REF2D_drawLineXNegHi_16_Generic;
        DC.drawStippleLineXPosLo    = REF2D_drawStippleLineXPosLo_16_Generic;
        DC.drawStippleLineXPosHi    = REF2D_drawStippleLineXPosHi_16_Generic;
        DC.drawStippleLineXNegLo    = REF2D_drawStippleLineXNegLo_16_Generic;
        DC.drawStippleLineXNegHi    = REF2D_drawStippleLineXNegHi_16_Generic;
        DC.drawStyleLineXPosLo      = REF2D_drawStyleLineXPosLo_16_Generic;
        DC.drawStyleLineXPosHi      = REF2D_drawStyleLineXPosHi_16_Generic;
        DC.drawStyleLineXNegLo      = REF2D_drawStyleLineXNegLo_16_Generic;
        DC.drawStyleLineXNegHi      = REF2D_drawStyleLineXNegHi_16_Generic;
        DC.drawMonoMSBTrans         = REF2D_drawMonoMSBTrans_16_Generic;
        DC.drawMonoMSBOpaque        = REF2D_drawMonoMSBOpaque_16_Generic;
        DC.drawClipMonoMSBTrans     = REF2D_drawClipMonoMSBTrans_16_Generic;
        DC.drawClipMonoMSBOpaque    = REF2D_drawClipMonoMSBOpaque_16_Generic;
        DC.drawMonoLSBTrans         = REF2D_drawMonoLSBTrans_16_Generic;
        DC.drawMonoLSBOpaque        = REF2D_drawMonoLSBOpaque_16_Generic;
        DC.drawClipMonoLSBTrans     = REF2D_drawClipMonoLSBTrans_16_Generic;
        DC.drawClipMonoLSBOpaque    = REF2D_drawClipMonoLSBOpaque_16_Generic;
        DC.drawEllipse              = REF2D_drawEllipse_16_Generic;
        }
}
#endif

#ifdef  SUPPORT_24BPP
/****************************************************************************
REMARKS:
This function sets the pixel drawing functions for the color depth
****************************************************************************/
static void REF2D_SetPixelFuncs24bpp(void)
{
    static REF2D_drawPixel_t pixelFuncs[] = {
        DEFINE_ROP2_PIXELS(24),
        };
    static REF2D_drawPixel_t pixelFuncsMasked[] = {
        DEFINE_ROP2_MASKED_PIXELS(24),
        };
    static REF2D_drawPixelR3_t pixelFuncsR3[] = {
        DEFINE_ROP3_PIXELS(24),
        };
    static REF2D_drawPixelR3_t pixelFuncsR3Masked[] = {
        DEFINE_ROP3_MASKED_PIXELS(24),
        };

    if (DC.planeMask != 0xFFFFFFFF) {
        DC.drawPixelTable = pixelFuncsMasked;
        DC.drawPixelR3Table = pixelFuncsR3Masked;
        }
    else {
        DC.drawPixelTable = pixelFuncs;
        DC.drawPixelR3Table = pixelFuncsR3;
        }
    DC.drawPixelSrcColorKey         = REF2D_drawPixelSrcColorKey_24;
    DC.drawPixelSrcColorKeyRange    = REF2D_drawPixelSrcColorKeyRange_24;
}

/****************************************************************************
REMARKS:
This function sets the hardware mix operation for subsequent drawing for
a specific color depth.
****************************************************************************/
static void REF2D_SetDrawBuffer24bpp(void)
{
    DEFINE_BLENDED_PIXELS(24);

    REF2D_SetPixelFuncs24bpp();
    DC.readPixel                            = REF2D_readPixel_24;
    DC.fillScan_S                           = REF2D_fillScan_24_S;
    DC.fillScan_Generic                     = REF2D_fillScan_24_Generic;
    DC.drawMonoCursor                       = REF2D_drawMonoCursor_24;
    DC.drawColorCursor                      = REF2D_drawColorCursor_24;
    DC.readScan                             = REF2D_readScan_24;
    DC.bltScan_S                            = REF2D_bltScan_24_S;
    DC.bltScan_Generic                      = REF2D_bltScan_24_Generic;
    DC.bltScanPatt                          = REF2D_bltScanPatt_24;
    DC.bltScanColorPatt                     = REF2D_bltScanColorPatt_24;
    DC.srcTransBltScan_S                    = REF2D_srcTransBltScan_24_S;
    DC.srcTransBltScan_Generic              = REF2D_srcTransBltScan_24_Generic;
    DC.srcTransBltScan_Masked               = REF2D_srcTransBltScan_24_Generic;
    DC.dstTransBltScan_S                    = REF2D_dstTransBltScan_24_S;
    DC.dstTransBltScan_Generic              = REF2D_dstTransBltScan_24_Generic;
    DC.dstTransBltScan_Masked               = REF2D_dstTransBltScan_24_Generic;
    DC.dstTransBltScan_S                    = REF2D_dstTransBltScan_24_S;
    DC.dstTransBltScan_Generic              = REF2D_dstTransBltScan_24_Generic;
    DC.dstTransBltScan_Generic              = REF2D_dstTransBltScan_24_Generic;
    DC.stretchScan2x_S                      = REF2D_stretchScan2x_24_S;
    DC.stretchScan2x_Generic                = REF2D_stretchScan2x_24_Generic;
    DC.stretchScan_S                        = REF2D_stretchScan_24_S;
    DC.stretchScan_Generic                  = REF2D_stretchScan_24_Generic;
    DC.drawStippleLineXPosLo                = REF2D_drawStippleLineXPosLo_24_Generic;
    DC.drawStippleLineXPosHi                = REF2D_drawStippleLineXPosHi_24_Generic;
    DC.drawStippleLineXNegLo                = REF2D_drawStippleLineXNegLo_24_Generic;
    DC.drawStippleLineXNegHi                = REF2D_drawStippleLineXNegHi_24_Generic;
    DC.drawStyleLineXPosLo                  = REF2D_drawStyleLineXPosLo_24_Generic;
    DC.drawStyleLineXPosHi                  = REF2D_drawStyleLineXPosHi_24_Generic;
    DC.drawStyleLineXNegLo                  = REF2D_drawStyleLineXNegLo_24_Generic;
    DC.drawStyleLineXNegHi                  = REF2D_drawStyleLineXNegHi_24_Generic;
    DC.drawPattScan                         = REF2D_drawPattScan_24_Generic;
    DC.drawColorPattScan                    = REF2D_drawColorPattScan_24_Generic;
    DC.drawMonoMSBTrans                     = REF2D_drawMonoMSBTrans_24_Generic;
    DC.drawMonoMSBOpaque                    = REF2D_drawMonoMSBOpaque_24_Generic;
    DC.drawClipMonoMSBTrans                 = REF2D_drawClipMonoMSBTrans_24_Generic;
    DC.drawClipMonoMSBOpaque                = REF2D_drawClipMonoMSBOpaque_24_Generic;
    DC.drawMonoLSBTrans                     = REF2D_drawMonoLSBTrans_24_Generic;
    DC.drawMonoLSBOpaque                    = REF2D_drawMonoLSBOpaque_24_Generic;
    DC.drawClipMonoLSBTrans                 = REF2D_drawClipMonoLSBTrans_24_Generic;
    DC.drawClipMonoLSBOpaque                = REF2D_drawClipMonoLSBOpaque_24_Generic;
    DC.clipEllipse                          = REF2D_clipEllipse_24_Generic;
    DC.convScan4_S                          = REF2D_convScan4_24_S;
    DC.convScan8_S                          = REF2D_convScan8_24_S;
    if (DC.mi.pf.RedPosition == 0) {
        DC.blendPixelTable                  = blendPixelFuncs24;
        DC.blendPixelFast                   = REF2D_drawPixelBlendedFast_24;
        DC.buildTranslateVector             = REF2D_buildTranslateVector_24BGR;
        }
    else {
        DC.blendPixelTable                  = blendPixelFuncs24;
        DC.blendPixelFast                   = REF2D_drawPixelBlendedFast_24;
        DC.buildTranslateVector             = REF2D_buildTranslateVector_24RGB;
        }
}

/****************************************************************************
PARAMETERS:
mix     - New mix code to set (GA_mixCodesType)

REMARKS:
This function sets the hardware mix operation for subsequent drawing for
a specific color depth.
****************************************************************************/
static void REF2D_SetMix24bpp(
    int mix)
{
    DC.drawPixel = GET_PIXEL_FUNC(mix);
    if (IS_SOLID_MIX(mix)) {
        DC.drawScan             = REF2D_drawScan_24_S;
        DC.drawLineXPosLo       = REF2D_drawLineXPosLo_24_S;
        DC.drawLineXPosHi       = REF2D_drawLineXPosHi_24_S;
        DC.drawLineXNegLo       = REF2D_drawLineXNegLo_24_S;
        DC.drawLineXNegHi       = REF2D_drawLineXNegHi_24_S;
        DC.drawEllipse          = REF2D_drawEllipse_24_S;
        }
    else {
        DC.drawScan             = REF2D_drawScan_24_Generic;
        DC.drawLineXPosLo       = REF2D_drawLineXPosLo_24_Generic;
        DC.drawLineXPosHi       = REF2D_drawLineXPosHi_24_Generic;
        DC.drawLineXNegLo       = REF2D_drawLineXNegLo_24_Generic;
        DC.drawLineXNegHi       = REF2D_drawLineXNegHi_24_Generic;
        DC.drawEllipse          = REF2D_drawEllipse_24_Generic;
        }
}
#endif

#ifdef  SUPPORT_32BPP
/****************************************************************************
REMARKS:
This function sets the pixel drawing functions for the color depth
****************************************************************************/
static void REF2D_SetPixelFuncs32bpp(void)
{
    static REF2D_drawPixel_t pixelFuncs[] = {
        DEFINE_ROP2_PIXELS(32),
        };
    static REF2D_drawPixel_t pixelFuncsMasked[] = {
        DEFINE_ROP2_MASKED_PIXELS(32),
        };
    static REF2D_drawPixelR3_t pixelFuncsR3[] = {
        DEFINE_ROP3_PIXELS(32),
        };
    static REF2D_drawPixelR3_t pixelFuncsR3Masked[] = {
        DEFINE_ROP3_MASKED_PIXELS(32),
        };

    if (DC.planeMask != 0xFFFFFFFF) {
        DC.drawPixelTable = pixelFuncsMasked;
        DC.drawPixelR3Table = pixelFuncsR3Masked;
        }
    else {
        DC.drawPixelTable = pixelFuncs;
        DC.drawPixelR3Table = pixelFuncsR3;
        }
    DC.drawPixelSrcColorKey         = REF2D_drawPixelSrcColorKey_32;
    DC.drawPixelSrcColorKeyRange    = REF2D_drawPixelSrcColorKeyRange_32;
}

/****************************************************************************
REMARKS:
This function sets the hardware mix operation for subsequent drawing for
a specific color depth.
****************************************************************************/
static void REF2D_SetDrawBuffer32bpp(void)
{
    DEFINE_BLENDED_PIXELS(Axxx32);
    DEFINE_BLENDED_PIXELS(xxxA32);

    REF2D_SetPixelFuncs32bpp();
    DC.readPixel                            = REF2D_readPixel_32;
    DC.fillScan_S                           = REF2D_fillScan_32_S;
    DC.fillScan_Generic                     = REF2D_fillScan_32_Generic;
    DC.drawMonoCursor                       = REF2D_drawMonoCursor_32;
    DC.drawColorCursor                      = REF2D_drawColorCursor_32;
    DC.readScan                             = REF2D_readScan_32;
    DC.bltScan_S                            = REF2D_bltScan_32_S;
    DC.bltScan_Generic                      = REF2D_bltScan_32_Generic;
    DC.bltScanPatt                          = REF2D_bltScanPatt_32;
    DC.bltScanColorPatt                     = REF2D_bltScanColorPatt_32;
    DC.srcTransBltScan_S                    = REF2D_srcTransBltScan_32_S;
    DC.srcTransBltScan_Generic              = REF2D_srcTransBltScan_32_Generic;
    DC.srcTransBltScan_Masked               = REF2D_srcTransBltScan_32_Masked;
    DC.dstTransBltScan_S                    = REF2D_dstTransBltScan_32_S;
    DC.dstTransBltScan_Generic              = REF2D_dstTransBltScan_32_Generic;
    DC.dstTransBltScan_Masked               = REF2D_dstTransBltScan_32_Masked;
    DC.stretchScan2x_S                      = REF2D_stretchScan2x_32_S;
    DC.stretchScan2x_Generic                = REF2D_stretchScan2x_32_Generic;
    DC.stretchScan_S                        = REF2D_stretchScan_32_S;
    DC.stretchScan_Generic                  = REF2D_stretchScan_32_Generic;
    DC.clipEllipse                          = REF2D_clipEllipse_32_Generic;
    DC.convScan4_S                          = REF2D_convScan4_32_S;
    DC.convScan8_S                          = REF2D_convScan8_32_S;
    if (DC.mi.pf.RedPosition == 0) {
        DC.blendPixelTable                  = blendPixelFuncsAxxx32;
        DC.blendPixelFast                   = REF2D_drawPixelBlendedFast_Axxx32;
        DC.buildTranslateVector             = REF2D_buildTranslateVector_32ABGR;
        }
    else if (DC.mi.pf.RedPosition == 8) {
        DC.blendPixelTable                  = blendPixelFuncsxxxA32;
        DC.blendPixelFast                   = REF2D_drawPixelBlendedFast_xxxA32;
        DC.buildTranslateVector             = REF2D_buildTranslateVector_32BGRA;
        }
    else if (DC.mi.pf.RedPosition == 16) {
        DC.blendPixelTable                  = blendPixelFuncsAxxx32;
        DC.blendPixelFast                   = REF2D_drawPixelBlendedFast_Axxx32;
        DC.buildTranslateVector             = REF2D_buildTranslateVector_32ARGB;
        }
    else {
        DC.blendPixelTable                  = blendPixelFuncsxxxA32;
        DC.blendPixelFast                   = REF2D_drawPixelBlendedFast_xxxA32;
        DC.buildTranslateVector             = REF2D_buildTranslateVector_32RGBA;
        }
}

/****************************************************************************
PARAMETERS:
mix     - New mix code to set (GA_mixCodesType)

REMARKS:
This function sets the hardware mix operation for subsequent drawing for
a specific color depth.
****************************************************************************/
static void REF2D_SetMix32bpp(
    int mix)
{
    DC.drawPixel = GET_PIXEL_FUNC(mix);
    if (IS_SOLID_MIX(mix)) {
        DC.drawScan                 = REF2D_drawScan_32_S;
        DC.drawPattScan             = REF2D_drawPattScan_32_S;
        DC.drawColorPattScan        = REF2D_drawColorPattScan_32_S;
        DC.drawLineXPosLo           = REF2D_drawLineXPosLo_32_S;
        DC.drawLineXPosHi           = REF2D_drawLineXPosHi_32_S;
        DC.drawLineXNegLo           = REF2D_drawLineXNegLo_32_S;
        DC.drawLineXNegHi           = REF2D_drawLineXNegHi_32_S;
        DC.drawStippleLineXPosLo    = REF2D_drawStippleLineXPosLo_32_S;
        DC.drawStippleLineXPosHi    = REF2D_drawStippleLineXPosHi_32_S;
        DC.drawStippleLineXNegLo    = REF2D_drawStippleLineXNegLo_32_S;
        DC.drawStippleLineXNegHi    = REF2D_drawStippleLineXNegHi_32_S;
        DC.drawStyleLineXPosLo      = REF2D_drawStyleLineXPosLo_32_S;
        DC.drawStyleLineXPosHi      = REF2D_drawStyleLineXPosHi_32_S;
        DC.drawStyleLineXNegLo      = REF2D_drawStyleLineXNegLo_32_S;
        DC.drawStyleLineXNegHi      = REF2D_drawStyleLineXNegHi_32_S;
        DC.drawMonoMSBTrans         = REF2D_drawMonoMSBTrans_32_S;
        DC.drawMonoMSBOpaque        = REF2D_drawMonoMSBOpaque_32_S;
        DC.drawClipMonoMSBTrans     = REF2D_drawClipMonoMSBTrans_32_S;
        DC.drawClipMonoMSBOpaque    = REF2D_drawClipMonoMSBOpaque_32_S;
        DC.drawMonoLSBTrans         = REF2D_drawMonoLSBTrans_32_S;
        DC.drawMonoLSBOpaque        = REF2D_drawMonoLSBOpaque_32_S;
        DC.drawClipMonoLSBTrans     = REF2D_drawClipMonoLSBTrans_32_S;
        DC.drawClipMonoLSBOpaque    = REF2D_drawClipMonoLSBOpaque_32_S;
        DC.drawEllipse              = REF2D_drawEllipse_32_S;
        }
    else {
        DC.drawScan                 = REF2D_drawScan_32_Generic;
        DC.drawPattScan             = REF2D_drawPattScan_32_Generic;
        DC.drawColorPattScan        = REF2D_drawColorPattScan_32_Generic;
        DC.drawLineXPosLo           = REF2D_drawLineXPosLo_32_Generic;
        DC.drawLineXPosHi           = REF2D_drawLineXPosHi_32_Generic;
        DC.drawLineXNegLo           = REF2D_drawLineXNegLo_32_Generic;
        DC.drawLineXNegHi           = REF2D_drawLineXNegHi_32_Generic;
        DC.drawStippleLineXPosLo    = REF2D_drawStippleLineXPosLo_32_Generic;
        DC.drawStippleLineXPosHi    = REF2D_drawStippleLineXPosHi_32_Generic;
        DC.drawStippleLineXNegLo    = REF2D_drawStippleLineXNegLo_32_Generic;
        DC.drawStippleLineXNegHi    = REF2D_drawStippleLineXNegHi_32_Generic;
        DC.drawStyleLineXPosLo      = REF2D_drawStyleLineXPosLo_32_Generic;
        DC.drawStyleLineXPosHi      = REF2D_drawStyleLineXPosHi_32_Generic;
        DC.drawStyleLineXNegLo      = REF2D_drawStyleLineXNegLo_32_Generic;
        DC.drawStyleLineXNegHi      = REF2D_drawStyleLineXNegHi_32_Generic;
        DC.drawMonoMSBTrans         = REF2D_drawMonoMSBTrans_32_Generic;
        DC.drawMonoMSBOpaque        = REF2D_drawMonoMSBOpaque_32_Generic;
        DC.drawClipMonoMSBTrans     = REF2D_drawClipMonoMSBTrans_32_Generic;
        DC.drawClipMonoMSBOpaque    = REF2D_drawClipMonoMSBOpaque_32_Generic;
        DC.drawMonoLSBTrans         = REF2D_drawMonoLSBTrans_32_Generic;
        DC.drawMonoLSBOpaque        = REF2D_drawMonoLSBOpaque_32_Generic;
        DC.drawClipMonoLSBTrans     = REF2D_drawClipMonoLSBTrans_32_Generic;
        DC.drawClipMonoLSBOpaque    = REF2D_drawClipMonoLSBOpaque_32_Generic;
        DC.drawEllipse              = REF2D_drawEllipse_32_Generic;
        }
}
#endif

/****************************************************************************
PARAMETERS:
funcs       - Pointer to the function list to make valid
size        - Size of the function list in bytes

RETURNS:
True if valid, false if not.
****************************************************************************/
static ibool MakeValid(
    ulong *funcs,
    int size)
{
    int i,empty = true;

    for (i = 0; i < size/4; i++) {
        if (funcs[i] != 0)
            empty = false;
        }
    return !empty;
}

/****************************************************************************
DESCRIPTION:
Returns the reference rasterisers for the specified function group.

PARAMETERS:
id      - Identifier for the function group to get pointers for
funcs   - Place to store the function block

RETURNS:
True if functions available, false if not.

REMARKS:
This function is the main function that is used by the graphics application
code to get a block of function pointers for a specified function group.
The function groups are defined GA_funcGroupsType enumeration, and breaks
up the functions in the device driver API into groups of logically similar
functions.
****************************************************************************/
ibool NAPI REF2D_QueryFunctionsSW(
    N_uint32 id,
    void *funcs)
{
    ulong   dwSize = *((ulong*)funcs);
    ulong   size;

    memset(funcs,0,dwSize);
    switch (id) {
        case GA_GET_CURSORFUNCS:        RETURN_VALID(REF_cursor);
        case GA_GET_2DSTATEFUNCS:       RETURN_VALID(REF_state2d);
        case GA_GET_2DRENDERFUNCS:      RETURN_VALID(REF_draw2dStub);
        case GA_GET_DRIVERFUNCS:
            if (DC.hwCtx) {
                GA_driverFuncs *drv = funcs;
                COPY_STRUCTURE2(funcs,&DC.hwDrv,dwSize);

                // Override functions we hook for our software mouse cursor
                drv->SetDisplayStart   = REF2D_SetDisplayStart;
                drv->SetDisplayStartXY = REF2D_SetDisplayStartXY;
#if defined(SUPPORT_4BPP) || defined(SUPPORT_8BPP)
                drv->SetPaletteData    = REF2D_SetPaletteData;
                drv->SetPaletteDataExt = REF2D_SetPaletteDataExt;
#endif
                return true;
                }
            return false;
        case GA_GET_REGIONFUNCS:
            if (REF_region.dwSize == 0) {
                // Dynamically load and initialise the region manager
                REF_region.dwSize = sizeof(REF_region);
                if (!GA_loadRegionMgr(&REF_region,&DC.hModRegion,&size)) {
                    REF_region.dwSize = 0;
                    return false;
                    }
                }
            RETURN_VALID(REF_region);
        case GA_GET_BUFFERFUNCS: {
            // The buffer manager functions are only relevant to managing
            // offscreen video memory so are not supported for system
            // memory buffer drivers.
            if (!DC.hwCtx)
                return false;
            if (REF_buffer.dwSize == 0) {
                // Dynamically load and initialise the buffer manager
                REF_buffer.dwSize = sizeof(REF_buffer);
                DC.hwBuf.dwSize = sizeof(DC.hwBuf);
                if (DC.hwCtx->loader.QueryFunctions(GA_GET_BUFFERFUNCS,-2,&DC.hwBuf)) {
                    // The hardware driver has provided it's own buffer manager
                    // implementation. Most likely the DirectDraw SNAP driver.
                    COPY_STRUCTURE(&REF_buffer,&DC.hwBuf);
                    }
                else {
                    // Initialise the internal buffer manager
                    if (!REF2D_loadBufferMgr(DC.hwCtx,&DV,&REF_buffer,&DC.hModBuffer,&size))
                        return false;
                    }

                // Once the buffer manager is loaded, override the InitBuffers
                // function so that we can take control first. This will allow
                // the ref2d driver to optimise lots of functions by using
                // offscreen video memory managed by the buffer manager.
                DC.BUF_InitBuffersOrg = REF_buffer.InitBuffers;
                DC.BUF_FlipToBufferOrg = REF_buffer.FlipToBuffer;
                DC.BUF_SetRef2d = REF_buffer.SetRef2d;
                DC.BUF_LockBuffer = REF_buffer.LockBuffer;
                DC.BUF_UnlockBuffer = REF_buffer.UnlockBuffer;
                REF_buffer.InitBuffers = REF2D_InitBuffers;
                REF_buffer.FlipToBuffer = REF2D_FlipToBuffer;
                }
            RETURN_VALID(REF_buffer);
            }
        }
    if (DC.hwCtx) {
        *((ulong*)funcs) = dwSize;
        return DC.hwCtx->loader.QueryFunctions(id,-2,funcs);
        }
    return false;
}

/****************************************************************************
DESCRIPTION:
Switches to the current set of rendering vectors

REMARKS:
This function is called in multiple places in the reference rasteriser
where the rendering vectors may change, and quickly switches to the new
set of rendering vectors depending on whether we require software or
hardware rendering for the current rasteriser state.
****************************************************************************/
void SetupRenderVectors(
    int bitsPerPixel)
{
    switch (bitsPerPixel) {
#ifdef  SUPPORT_1BPP
        case 1:
            DC.mi.byteShift = -3;
            DC.colorMask = 0x01;
            DC.palSize = 2;
            REF2D_SetDrawBuffer1bpp();
            break;
#endif
#ifdef  SUPPORT_4BPP
        case 4:
            DC.mi.byteShift = -1;
            DC.colorMask = 0x0F;
            DC.palSize = 16;
            REF2D_SetDrawBuffer4bpp();
            break;
#endif
#ifdef  SUPPORT_8BPP
        case 8:
            DC.mi.byteShift = 0;
            DC.colorMask = 0xFF;
            DC.palSize = 256;
            REF2D_SetDrawBuffer8bpp();
            break;
#endif
#ifdef  SUPPORT_16BPP
        case 15:
        case 16:
            DC.mi.byteShift = 1;
            DC.colorMask = (bitsPerPixel == 15) ? 0x7FFF : 0xFFFF;
            DC.palSize = 256;
            REF2D_SetDrawBuffer16bpp();
            break;
#endif
#ifdef  SUPPORT_24BPP
        case 24:
            DC.mi.byteShift = 3;
            DC.colorMask = 0xFFFFFF;
            DC.palSize = 256;
            REF2D_SetDrawBuffer24bpp();
            break;
#endif
#ifdef  SUPPORT_32BPP
        case 32:
            DC.mi.byteShift = 2;
            if (DC.mi.pf.AlphaPosition == 0)
                DC.colorMask = 0xFFFFFF00;
            else
                DC.colorMask = 0x00FFFFFF;
            DC.palSize = 256;
            REF2D_SetDrawBuffer32bpp();
            break;
#endif
        }
}

/****************************************************************************
DESCRIPTION:
Switches to the current set of rendering vectors

REMARKS:
This function is called in multiple places in the reference rasteriser
where the rendering vectors may change, and quickly switches to the new
set of rendering vectors depending on whether we require software or
hardware rendering for the current rasteriser state.
****************************************************************************/
static void SwitchVectors(void)
{
    if (softwareOnly || (REF_noHWPlaneMask && DC.planeMask != 0xFFFFFFFF) || BLENDING_ENABLED())
        REF_draw2dCnt = REF_draw2dSW;
    else {
        REF_draw2dCnt = REF_draw2dHWSW;
        if (transMonoPattSW) {
            REF_draw2dCnt.DrawPattScanList          = REF_draw2dSW.DrawPattScanList;
            REF_draw2dCnt.DrawPattEllipseList       = REF_draw2dSW.DrawPattEllipseList;
            REF_draw2dCnt.DrawPattFatEllipseList    = REF_draw2dSW.DrawPattFatEllipseList;
            REF_draw2dCnt.DrawPattRect              = REF_draw2dSW.DrawPattRect;
            REF_draw2dCnt.DrawPattTrap              = REF_draw2dSW.DrawPattTrap;
            }
        if (transColorPattSW) {
            REF_draw2dCnt.DrawColorPattScanList     = REF_draw2dSW.DrawColorPattScanList;
            REF_draw2dCnt.DrawColorPattEllipseList  = REF_draw2dSW.DrawColorPattEllipseList;
            REF_draw2dCnt.DrawPattFatEllipseList    = REF_draw2dSW.DrawColorPattFatEllipseList;
            REF_draw2dCnt.DrawColorPattRect         = REF_draw2dSW.DrawColorPattRect;
            REF_draw2dCnt.DrawColorPattTrap         = REF_draw2dSW.DrawColorPattTrap;
            }
        if (DC.mono8x8Transparent) {
            REF_draw2dCnt.BitBltPatt                = REF_draw2dSW.BitBltPatt;
            REF_draw2dCnt.BitBltPattLin             = REF_draw2dSW.BitBltPattLin;
            REF_draw2dCnt.BitBltPattSys             = REF_draw2dSW.BitBltPattSys;
            REF_draw2dCnt.BitBltPattBM              = REF_draw2dSW.BitBltPattBM;
            }
        if (DC.color8x8Transparent) {
            REF_draw2dCnt.BitBltColorPatt           = REF_draw2dSW.BitBltColorPatt;
            REF_draw2dCnt.BitBltColorPattLin        = REF_draw2dSW.BitBltColorPattLin;
            REF_draw2dCnt.BitBltColorPattSys        = REF_draw2dSW.BitBltColorPattSys;
            REF_draw2dCnt.BitBltColorPattBM         = REF_draw2dSW.BitBltColorPattBM;
            }
        }
}

/****************************************************************************
DESCRIPTION:
Switches to the current set of rendering vectors

REMARKS:
This function is called in multiple places in the reference rasteriser
where the rendering vectors may change, and quickly switches to the new
set of rendering vectors depending on whether we require software or
hardware rendering for the current rasteriser state.
****************************************************************************/
static void SwitchBltPatVectors(void)
{
    if (DC.mono8x8Transparent) {
        REF_draw2dCnt.BitBltPatt            = REF_draw2dSW.BitBltPatt;
        REF_draw2dCnt.BitBltPattLin         = REF_draw2dSW.BitBltPattLin;
        REF_draw2dCnt.BitBltPattSys         = REF_draw2dSW.BitBltPattSys;
        REF_draw2dCnt.BitBltPattBM          = REF_draw2dSW.BitBltPattBM;
        }
    else {
        REF_draw2dCnt.BitBltPatt            = REF_draw2dHWSW.BitBltPatt;
        REF_draw2dCnt.BitBltPattLin         = REF_draw2dHWSW.BitBltPattLin;
        REF_draw2dCnt.BitBltPattSys         = REF_draw2dHWSW.BitBltPattSys;
        REF_draw2dCnt.BitBltPattBM          = REF_draw2dHWSW.BitBltPattBM;
        }
}

/****************************************************************************
DESCRIPTION:
Switches to the current set of rendering vectors

REMARKS:
This function is called in multiple places in the reference rasteriser
where the rendering vectors may change, and quickly switches to the new
set of rendering vectors depending on whether we require software or
hardware rendering for the current rasteriser state.
****************************************************************************/
static void SwitchBltColorPatVectors(void)
{
    if (DC.color8x8Transparent) {
        REF_draw2dCnt.BitBltColorPatt           = REF_draw2dSW.BitBltColorPatt;
        REF_draw2dCnt.BitBltColorPattLin        = REF_draw2dSW.BitBltColorPattLin;
        REF_draw2dCnt.BitBltColorPattSys        = REF_draw2dSW.BitBltColorPattSys;
        REF_draw2dCnt.BitBltColorPattBM         = REF_draw2dSW.BitBltColorPattBM;
        }
    else {
        REF_draw2dCnt.BitBltColorPatt           = REF_draw2dHWSW.BitBltColorPatt;
        REF_draw2dCnt.BitBltColorPattLin        = REF_draw2dHWSW.BitBltColorPattLin;
        REF_draw2dCnt.BitBltColorPattSys        = REF_draw2dHWSW.BitBltColorPattSys;
        REF_draw2dCnt.BitBltColorPattBM         = REF_draw2dHWSW.BitBltColorPattBM;
        }
}

/****************************************************************************
DESCRIPTION:
Forces software only rendering on or off.

PARAMETERS:
enable  - True to force software rendering, false to disable it

REMARKS:
This function is primarily intended for conformance testing to allow all
the drawing functions to be forcibly switched to software rendering or to
switch back to hardware rendering.
****************************************************************************/
N_int32 NAPI REF2D_ForceSoftwareOnly(
    N_int32 enable)
{
    N_int32 oldEnable = softwareOnlyForced;

    // Compute the set of combined hardware/software rendering vectors
    memset(&REF_draw2dHWSW,0,sizeof(REF_draw2dHWSW));
    if (DC.haveHWDraw && !enable) {
        REF_draw2dHWSW = REF_draw2dHW;
        if (REF_bugFixHWROP3) {
            // Enable ROP'3 bug fix workaround for older drivers
            if (REF_draw2dHW.BitBltPatt)            REF_draw2dHWSW.BitBltPatt           = REF2D_BugFixBitBltPatt;
            if (REF_draw2dHW.BitBltPattLin)         REF_draw2dHWSW.BitBltPattLin        = REF2D_BugFixBitBltPattLin;
            if (REF_draw2dHW.BitBltPattSys)         REF_draw2dHWSW.BitBltPattSys        = REF2D_BugFixBitBltPattSys;
            if (REF_draw2dHW.BitBltPattBM)          REF_draw2dHWSW.BitBltPattBM         = REF2D_BugFixBitBltPattBM;
            if (REF_draw2dHW.BitBltColorPatt)       REF_draw2dHWSW.BitBltColorPatt      = REF2D_BugFixBitBltColorPatt;
            if (REF_draw2dHW.BitBltColorPattLin)    REF_draw2dHWSW.BitBltColorPattLin   = REF2D_BugFixBitBltColorPattLin;
            if (REF_draw2dHW.BitBltColorPattSys)    REF_draw2dHWSW.BitBltColorPattSys   = REF2D_BugFixBitBltColorPattSys;
            if (REF_draw2dHW.BitBltColorPattBM)     REF_draw2dHWSW.BitBltColorPattBM    = REF2D_BugFixBitBltColorPattBM;
            }
        }
    if (!REF_draw2dHWSW.GetPixel)                       REF_draw2dHWSW.GetPixel                 = REF_draw2dSW.GetPixel;
    if (!REF_draw2dHWSW.PutPixel)                       REF_draw2dHWSW.PutPixel                 = REF_draw2dSW.PutPixel;
    if (!REF_draw2dHWSW.DrawScanList)                   REF_draw2dHWSW.DrawScanList             = REF_draw2dSW.DrawScanList;
    if (!REF_draw2dHWSW.DrawPattScanList)               REF_draw2dHWSW.DrawPattScanList         = REF_draw2dSW.DrawPattScanList;
    if (!REF_draw2dHWSW.DrawColorPattScanList)          REF_draw2dHWSW.DrawColorPattScanList    = REF_draw2dSW.DrawColorPattScanList;
    if (!REF_draw2dHWSW.DrawEllipseList)                REF_draw2dHWSW.DrawEllipseList          = REF_draw2dSW.DrawEllipseList;
    if (!REF_draw2dHWSW.DrawPattEllipseList)            REF_draw2dHWSW.DrawPattEllipseList      = REF_draw2dSW.DrawPattEllipseList;
    if (!REF_draw2dHWSW.DrawColorPattEllipseList)       REF_draw2dHWSW.DrawColorPattEllipseList = REF_draw2dSW.DrawColorPattEllipseList;
    if (!REF_draw2dHWSW.DrawFatEllipseList)             REF_draw2dHWSW.DrawFatEllipseList       = REF_draw2dSW.DrawFatEllipseList;
    if (!REF_draw2dHWSW.DrawPattFatEllipseList)         REF_draw2dHWSW.DrawPattFatEllipseList   = REF_draw2dSW.DrawPattFatEllipseList;
    if (!REF_draw2dHWSW.DrawColorPattFatEllipseList)    REF_draw2dHWSW.DrawPattFatEllipseList   = REF_draw2dSW.DrawColorPattFatEllipseList;
    if (!REF_draw2dHWSW.DrawRect)                       REF_draw2dHWSW.DrawRect                 = REF_draw2dSW.DrawRect;
    if (!REF_draw2dHWSW.DrawRectExt)                    REF_draw2dHWSW.DrawRectExt              = REF_draw2dSW.DrawRectExt;
    if (!REF_draw2dHWSW.DrawRectLin)                    REF_draw2dHWSW.DrawRectLin              = REF_draw2dSW.DrawRectLin;
    if (!REF_draw2dHWSW.DrawPattRect)                   REF_draw2dHWSW.DrawPattRect             = REF_draw2dSW.DrawPattRect;
    if (!REF_draw2dHWSW.DrawColorPattRect)              REF_draw2dHWSW.DrawColorPattRect        = REF_draw2dSW.DrawColorPattRect;
    if (!REF_draw2dHWSW.DrawTrap)                       REF_draw2dHWSW.DrawTrap                 = REF_draw2dSW.DrawTrap;
    if (!REF_draw2dHWSW.DrawPattTrap)                   REF_draw2dHWSW.DrawPattTrap             = REF_draw2dSW.DrawPattTrap;
    if (!REF_draw2dHWSW.DrawColorPattTrap)              REF_draw2dHWSW.DrawColorPattTrap        = REF_draw2dSW.DrawColorPattTrap;
    if (!REF_draw2dHWSW.DrawLineInt)                    REF_draw2dHWSW.DrawLineInt              = REF_draw2dSW.DrawLineInt;
    if (!REF_draw2dHWSW.DrawBresenhamLine)              REF_draw2dHWSW.DrawBresenhamLine        = REF_draw2dSW.DrawBresenhamLine;
    if (!REF_draw2dHWSW.DrawStippleLineInt)             REF_draw2dHWSW.DrawStippleLineInt       = REF_draw2dSW.DrawStippleLineInt;
    if (!REF_draw2dHWSW.DrawBresenhamStippleLine)       REF_draw2dHWSW.DrawBresenhamStippleLine = REF_draw2dSW.DrawBresenhamStippleLine;
    if (!REF_draw2dHWSW.DrawEllipse)                    REF_draw2dHWSW.DrawEllipse              = REF_draw2dSW.DrawEllipse;
    if (!REF_draw2dHWSW.ClipEllipse)                    REF_draw2dHWSW.ClipEllipse              = REF_draw2dSW.ClipEllipse;
    if (!REF_draw2dHWSW.PutMonoImageMSBSys)             REF_draw2dHWSW.PutMonoImageMSBSys       = REF_draw2dSW.PutMonoImageMSBSys;
    if (!REF_draw2dHWSW.PutMonoImageMSBLin)             REF_draw2dHWSW.PutMonoImageMSBLin       = REF_draw2dSW.PutMonoImageMSBLin;
    if (!REF_draw2dHWSW.PutMonoImageMSBBM)              REF_draw2dHWSW.PutMonoImageMSBBM        = REF_draw2dSW.PutMonoImageMSBBM;
    if (!REF_draw2dHWSW.PutMonoImageLSBSys)             REF_draw2dHWSW.PutMonoImageLSBSys       = REF_draw2dSW.PutMonoImageLSBSys;
    if (!REF_draw2dHWSW.PutMonoImageLSBLin)             REF_draw2dHWSW.PutMonoImageLSBLin       = REF_draw2dSW.PutMonoImageLSBLin;
    if (!REF_draw2dHWSW.PutMonoImageLSBBM)              REF_draw2dHWSW.PutMonoImageLSBBM        = REF_draw2dSW.PutMonoImageLSBBM;
    if (!REF_draw2dHWSW.ClipMonoImageMSBSys)            REF_draw2dHWSW.ClipMonoImageMSBSys      = REF_draw2dSW.ClipMonoImageMSBSys;
    if (!REF_draw2dHWSW.ClipMonoImageMSBLin)            REF_draw2dHWSW.ClipMonoImageMSBLin      = REF_draw2dSW.ClipMonoImageMSBLin;
    if (!REF_draw2dHWSW.ClipMonoImageMSBBM)             REF_draw2dHWSW.ClipMonoImageMSBBM       = REF_draw2dSW.ClipMonoImageMSBBM;
    if (!REF_draw2dHWSW.ClipMonoImageLSBSys)            REF_draw2dHWSW.ClipMonoImageLSBSys      = REF_draw2dSW.ClipMonoImageLSBSys;
    if (!REF_draw2dHWSW.ClipMonoImageLSBLin)            REF_draw2dHWSW.ClipMonoImageLSBLin      = REF_draw2dSW.ClipMonoImageLSBLin;
    if (!REF_draw2dHWSW.ClipMonoImageLSBBM)             REF_draw2dHWSW.ClipMonoImageLSBBM       = REF_draw2dSW.ClipMonoImageLSBBM;
    if (!REF_draw2dHWSW.BitBlt)                         REF_draw2dHWSW.BitBlt                   = REF_draw2dSW.BitBlt;
    if (!REF_draw2dHWSW.BitBltLin)                      REF_draw2dHWSW.BitBltLin                = REF_draw2dSW.BitBltLin;
    if (!REF_draw2dHWSW.BitBltSys)                      REF_draw2dHWSW.BitBltSys                = REF_draw2dSW.BitBltSys;
    if (!REF_draw2dHWSW.BitBltBM)                       REF_draw2dHWSW.BitBltBM                 = REF_draw2dSW.BitBltBM;
    if (!REF_draw2dHWSW.BitBltPatt)                     REF_draw2dHWSW.BitBltPatt               = REF_draw2dSW.BitBltPatt;
    if (!REF_draw2dHWSW.BitBltPattLin)                  REF_draw2dHWSW.BitBltPattLin            = REF_draw2dSW.BitBltPattLin;
    if (!REF_draw2dHWSW.BitBltPattSys)                  REF_draw2dHWSW.BitBltPattSys            = REF_draw2dSW.BitBltPattSys;
    if (!REF_draw2dHWSW.BitBltPattBM)                   REF_draw2dHWSW.BitBltPattBM             = REF_draw2dSW.BitBltPattBM;
    if (!REF_draw2dHWSW.BitBltColorPatt)                REF_draw2dHWSW.BitBltColorPatt          = REF_draw2dSW.BitBltColorPatt;
    if (!REF_draw2dHWSW.BitBltColorPattLin)             REF_draw2dHWSW.BitBltColorPattLin       = REF_draw2dSW.BitBltColorPattLin;
    if (!REF_draw2dHWSW.BitBltColorPattSys)             REF_draw2dHWSW.BitBltColorPattSys       = REF_draw2dSW.BitBltColorPattSys;
    if (!REF_draw2dHWSW.BitBltColorPattBM)              REF_draw2dHWSW.BitBltColorPattBM        = REF_draw2dSW.BitBltColorPattBM;
    if (!REF_draw2dHWSW.SrcTransBlt)                    REF_draw2dHWSW.SrcTransBlt              = REF_draw2dSW.SrcTransBlt;
    if (!REF_draw2dHWSW.SrcTransBltLin)                 REF_draw2dHWSW.SrcTransBltLin           = REF_draw2dSW.SrcTransBltLin;
    if (!REF_draw2dHWSW.SrcTransBltSys)                 REF_draw2dHWSW.SrcTransBltSys           = REF_draw2dSW.SrcTransBltSys;
    if (!REF_draw2dHWSW.SrcTransBltBM)                  REF_draw2dHWSW.SrcTransBltBM            = REF_draw2dSW.SrcTransBltBM;
    if (!REF_draw2dHWSW.DstTransBlt)                    REF_draw2dHWSW.DstTransBlt              = REF_draw2dSW.DstTransBlt;
    if (!REF_draw2dHWSW.DstTransBltLin)                 REF_draw2dHWSW.DstTransBltLin           = REF_draw2dSW.DstTransBltLin;
    if (!REF_draw2dHWSW.DstTransBltSys)                 REF_draw2dHWSW.DstTransBltSys           = REF_draw2dSW.DstTransBltSys;
    if (!REF_draw2dHWSW.DstTransBltBM)                  REF_draw2dHWSW.DstTransBltBM            = REF_draw2dSW.DstTransBltBM;
    if (!REF_draw2dHWSW.StretchBlt)                     REF_draw2dHWSW.StretchBlt               = REF_draw2dSW.StretchBlt;
    if (!REF_draw2dHWSW.StretchBltLin)                  REF_draw2dHWSW.StretchBltLin            = REF_draw2dSW.StretchBltLin;
    if (!REF_draw2dHWSW.StretchBltSys)                  REF_draw2dHWSW.StretchBltSys            = REF_draw2dSW.StretchBltSys;
    if (!REF_draw2dHWSW.StretchBltBM)                   REF_draw2dHWSW.StretchBltBM             = REF_draw2dSW.StretchBltBM;
    if (!REF_draw2dHWSW.ConvertBltSys_Obsolete)         REF_draw2dHWSW.ConvertBltSys_Obsolete           = REF_draw2dSW.ConvertBltSys_Obsolete;
    if (!REF_draw2dHWSW.ConvertBltBM_Obsolete)          REF_draw2dHWSW.ConvertBltBM_Obsolete            = REF_draw2dSW.ConvertBltBM_Obsolete;
    if (!REF_draw2dHWSW.StretchConvertBltSys_Obsolete)  REF_draw2dHWSW.StretchConvertBltSys_Obsolete    = REF_draw2dSW.StretchConvertBltSys_Obsolete;
    if (!REF_draw2dHWSW.StretchConvertBltBM_Obsolete)   REF_draw2dHWSW.StretchConvertBltBM_Obsolete     = REF_draw2dSW.StretchConvertBltBM_Obsolete;
    if (!REF_draw2dHWSW.BitBltFx)                       REF_draw2dHWSW.BitBltFx                 = REF_draw2dSW.BitBltFx;
    if (!REF_draw2dHWSW.BitBltFxLin)                    REF_draw2dHWSW.BitBltFxLin              = REF_draw2dSW.BitBltFxLin;
    if (!REF_draw2dHWSW.BitBltFxSys)                    REF_draw2dHWSW.BitBltFxSys              = REF_draw2dSW.BitBltFxSys;
    if (!REF_draw2dHWSW.BitBltFxBM)                     REF_draw2dHWSW.BitBltFxBM               = REF_draw2dSW.BitBltFxBM;
    if (!REF_draw2dHWSW.GetBitmapSys)                   REF_draw2dHWSW.GetBitmapSys             = REF_draw2dSW.GetBitmapSys;
    if (!REF_draw2dHWSW.GetBitmapBM)                    REF_draw2dHWSW.GetBitmapBM              = REF_draw2dSW.GetBitmapBM;
    if (!REF_draw2dHWSW.DrawClippedLineInt)             REF_draw2dHWSW.DrawClippedLineInt       = REF_draw2dSW.DrawClippedLineInt;
    if (!REF_draw2dHWSW.DrawClippedBresenhamLine)       REF_draw2dHWSW.DrawClippedBresenhamLine = REF_draw2dSW.DrawClippedBresenhamLine;
    if (!REF_draw2dHWSW.DrawClippedStippleLineInt)      REF_draw2dHWSW.DrawClippedStippleLineInt= REF_draw2dSW.DrawClippedStippleLineInt;
    if (!REF_draw2dHWSW.DrawClippedBresenhamStippleLine)REF_draw2dHWSW.DrawClippedBresenhamStippleLine= REF_draw2dSW.DrawClippedBresenhamStippleLine;
    if (!REF_draw2dHWSW.DrawStyleLineInt)               REF_draw2dHWSW.DrawStyleLineInt         = REF_draw2dSW.DrawStyleLineInt;
    if (!REF_draw2dHWSW.DrawBresenhamStyleLine)         REF_draw2dHWSW.DrawBresenhamStyleLine   = REF_draw2dSW.DrawBresenhamStyleLine;
    if (!REF_draw2dHWSW.DrawClippedStyleLineInt)        REF_draw2dHWSW.DrawClippedStyleLineInt  = REF_draw2dSW.DrawClippedStyleLineInt;
    if (!REF_draw2dHWSW.DrawClippedBresenhamStyleLine)  REF_draw2dHWSW.DrawClippedBresenhamStyleLine = REF_draw2dSW.DrawClippedBresenhamStyleLine;

    // To make sure the ref2d software driver can be called for cases that the
    // hardware driver cannot handle, we always set the BitBltFxTest exported
    // from ref2d to point to the software version.
    REF_draw2dHWSW.BitBltFxTest = REF_draw2dSW.BitBltFxTest;

    // For older drivers that do not implement the new DrawRectExt function, we
    // override this function with an internal version that implements it on top
    // of the existing hardware DrawRect function.
    if (REF_draw2dHWSW.DrawRectExt == REF_draw2dSW.DrawRectExt && REF_draw2dHW.DrawRect)
        REF_draw2dHWSW.DrawRectExt = REF2D_DrawRectExtHW;

    // Configure the color depth dependent internal rendering functions
    SetupRenderVectors(DC.mi.bitsPerPixel);

    // Switch to the default set of rendering vectors
    softwareOnlyForced = softwareOnly = enable;
    setHardwareState = !enable;
    SwitchVectors();
    return oldEnable;
}

/****************************************************************************
DESCRIPTION:
Sets the software rasteriser active drawing buffer.

HEADER:
snap/graphics.h

PARAMETERS:
drawBuf         - Buffer to make the active drawing buffer
frameBuffer     - Pointer to the start of frameBuffer memory
bitsPerPixel    - Color depth for the buffer
pf              - Pixel format for the buffer
hwCtx           - SNAP driver to use (NULL if none)
softwareOnly    - True to force software only mode

RETURNS:
0 on success, -1 on failure.

REMARKS:
This function allows the application to make a video memory or system
memory buffer the active rendering buffer for all subsequent drawing
commands for the 2d reference rasteriser. This function mimics
GA_2DStateFuncs::SetDrawBuffer, however it also allows the software
rasteriser to be pointed at a system memory buffer for drawing without
any hardware acceleration at all.

If the hwCtx parameter is set to NULL, this function will enable only
software rendering to a system memory buffer pointed to by the frameBuffer
pointer. If the hwCtx member is not NULL and the frameBuffer pointer points
to a located in video memory, hardware acceleration will be enabled when
possible (provided softwareOnly is also set to false). If the softwareOnly
flag is true, no hardware rendering will be used at all.

SEE ALSO:
GA_2DStateFuncs::SetDrawBuffer
****************************************************************************/
N_int32 NAPI REF2D_SetDrawBufferSW(
    GA_buffer *drawBuf,
    void *frameBuffer,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf,
    GA_devCtx *hwCtx,
    N_int32 forceSoftwareOnly)
{
    // Initialise internal rendering functions
    DC.hwCtx                = hwCtx;
    if (hwCtx) {
        DC.hwBufferStart    = hwCtx->LinearMem;
        DC.hwBufferEnd      = DC.hwBufferStart + hwCtx->TotalMemory * 1024;
        DC.isHWBuffer       = true;
        }
    else {
        DC.hwBufferStart    = frameBuffer;
        DC.hwBufferEnd      = frameBuffer;
        DC.isHWBuffer       = false;
        }
    DC.frameBuffer          = frameBuffer;
    DC.drawOffset           = drawBuf->Offset;
    DC.surface              = DC.frameBuffer + DC.drawOffset;
    DC.mi.xRes              = drawBuf->Width;
    DC.mi.yRes              = drawBuf->Height;
    DC.mi.bitsPerPixel      = bitsPerPixel;
    DC.mi.bytesPerLine      = drawBuf->Stride;
    if (pf) DC.mi.pf        = *pf;

    // Set the hardware draw buffer
    DC.cntDrawBuf = *drawBuf;
    if (hwCtx && DC.hwState.SetDrawBuffer) {
        DC.hwState.SetDrawBuffer(drawBuf);

        // Now set the software rendering vectors in the driver to point
        // back to us so it can handle hardware workarounds.
        DC.hwInit.SetSoftwareRenderFuncs(&REF_draw2dSW);
        }

    // Now set up the rendering state
    REF2D_ForceSoftwareOnly(forceSoftwareOnly);
    return 0;
}

/****************************************************************************
PARAMETERS:
mask    - New color compare mask

REMARKS:
This function is used to override the internal software rendering color
compare mask used for source and destination transparent blit functions.
If this mask is set to 0xFFFFFFFF, compare compare masking is disabled
(the default state). Otherwise it is enabled. This is primarily used by
the conformance tests to ensure that color compares are done with the
appropriate color mask, while regular code (MGL etc) does not do this
for speed reasons.
****************************************************************************/
N_uint32 NAPI REF2D_SetColorCompareMask(
    N_uint32 mask)
{
    N_uint32 oldMask = DC.colorCmpMask;
    if (DC.gaMode.BitsPerPixel > 8)
        DC.colorCmpMask = mask;
    return oldMask;
}

/****************************************************************************
PARAMETERS:
surface         - Pointer to the start of the memory buffer for drawing
xRes            - X resolution for the surface
yRes            - Y resolution for the surface
bytesPerLine    - Scanline width for the memory buffer for drawing
bitsPerPixel    - New color depth for the surface
pf              - New pixel format for the surface

REMARKS:
This function changes the software draw buffer quickly, to allow the
software renderer to render to bitmaps in different locations in memory.
We do in here for speed, since we want to avoid the overhead of setting
the full draw buffer for the reference rasteriser code. If the color
depth or pixel format changes, this function will switch all the internal
rendering vectors to put to the new vectors for the appropriate functions.
This allows the reference rasteriser to work with all color depths active
in a single driver (and for our debug library to work properly).

This function is also used by the buffer manager code to quickly change
the draw buffer surface.
****************************************************************************/
void NAPI REF2D_SetDrawSurface(
    void *surface,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bytesPerLine,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf)
{
    ibool   forceSoftwareOnly,setupVectors;

    // Determine if we need to run in software only mode
    if (DC.hwCtx && DC.haveHWDraw && IS_HARDWARE_SURFACE(surface)) {
        forceSoftwareOnly = false;
        DC.isHWBuffer = true;
        }
    else {
        forceSoftwareOnly = true;
        DC.isHWBuffer = false;
        }

    // Determine if we need to swap the rendering vectors
    if (DC.mi.bitsPerPixel != bitsPerPixel || (bitsPerPixel > 8 && !CMP_PF(DC.mi.pf,*pf)))
        setupVectors = true;
    else
        setupVectors = false;

    // Update internal surface information
    DC.surface          = surface;
    DC.mi.xRes          = xRes;
    DC.mi.yRes          = yRes;
    DC.mi.bitsPerPixel  = bitsPerPixel;
    DC.mi.bytesPerLine  = bytesPerLine;
    DC.mi.pf            = *pf;

    // Switch rendering vectors if the pixel format is not compatible
    if (setupVectors)
        SetupRenderVectors(bitsPerPixel);

    // Switch the default set of rendering vectors if software only changed
    if (forceSoftwareOnly != softwareOnly) {
        softwareOnly = forceSoftwareOnly;
        setHardwareState = !forceSoftwareOnly;
        SwitchVectors();
        }
}

/****************************************************************************
PARAMETERS:
drawBuf - New hardware draw buffer to make active

RETURNS:
0 on success, -1 on failure.

REMARKS:
This function changes the hardware draw buffer for both the software
rasteriser and the hardware device. This can also be used to change the
draw buffer for a software only buffer, but is generally only used when
using hardware.
****************************************************************************/
N_int32 NAPI REF2D_SetDrawBuffer(
    GA_buffer *drawBuf)
{
    // Change the parameters for the software rendering engine
    DC.drawOffset       = drawBuf->Offset;
    DC.surface          = DC.frameBuffer + DC.drawOffset;
    DC.mi.xRes          = drawBuf->Width;
    DC.mi.yRes          = drawBuf->Height;
    DC.mi.bytesPerLine  = drawBuf->Stride;

    // Initialise the hardware rendering information if available
    DC.cntDrawBuf = *drawBuf;
    if (DC.hwCtx && DC.haveHWDraw && setHardwareState && DC.frameBuffer == DC.hwCtx->LinearMem)
        return DC.hwState.SetDrawBuffer(drawBuf);
    return 0;
}

/****************************************************************************
REMARKS:
This returns the address of the currently active draw buffer
****************************************************************************/
GA_buffer * NAPI REF2D_GetDrawBuffer(void)
{
    return &DC.cntDrawBuf;
}

/****************************************************************************
PARAMETERS:
numBuffers  - Number of primary flip buffers to allocate
flags       - Flags to pass for all primary buffers
stereo      - Pointer to software stereo functions, NULL for non stereo

REMARKS:
This function is an override that allows the ref2d driver to hook into
the buffer manager functions and provided optimised rendering functions
using offscreen video memory managed by the buffer manager.
****************************************************************************/
ibool NAPI REF2D_InitBuffers(
    N_int32 numBuffers,
    N_uint32 flags,
    GA_softStereoFuncs *stereo)
{
    // TODO: Add code in here to optimise blitting using the buffer
    //       manager!

    if (DC.BUF_SetRef2d) {
        if (!DC.BUF_SetRef2d(&DV))
            return false;
        }
    return DC.BUF_InitBuffersOrg(numBuffers,flags,stereo);
}

void REF2D_nop(void) {}

/****************************************************************************
PARAMETERS:
color   - New foreground color to set

REMARKS:
This function sets the hardware foreground color for subsequent accelerated
rendering primitives. For solid primitives such as DrawRect only the
foreground color is used. For patterns primitives such as DrawPattRect and
DrawStippleLine both the foreground and background colors are used.
****************************************************************************/
void NAPI REF2D_SetForeColor(
    GA_color color)
{
    switch (DC.mi.bitsPerPixel) {
#ifdef  SUPPORT_1BPP
        case 1:
            DC.foreColor = mkcolor1(color);
            break;
#endif
#ifdef  SUPPORT_4BPP
        case 4:
            DC.foreColor = mkcolor4(color);
            break;
#endif
#ifdef  SUPPORT_8BPP
        case 8:
            DC.foreColor = mkcolor8(color);
            break;
#endif
#ifdef  SUPPORT_16BPP
        case 15:
        case 16:
            DC.foreColor = mkcolor16(color);
            break;
#endif
#if defined(SUPPORT_24BPP) || defined(SUPPORT_32BPP)
        case 24:
        case 32:
            DC.foreColor = color;
            break;
#endif
        }
    if (DC.haveHWDraw && setHardwareState)
        DC.hwState.SetForeColor(color);
}

/****************************************************************************
PARAMETERS:
color   - New background color to set

REMARKS:
This function sets the hardware background color for subsequent accelerated
rendering primitives. For solid primitives such as DrawRect only the
foreground color is used. For patterns primitives such as DrawPattRect and
DrawStippleLine both the foreground and background colors are used.
****************************************************************************/
void NAPI REF2D_SetBackColor(
    GA_color color)
{
    switch (DC.mi.bitsPerPixel) {
#ifdef  SUPPORT_1BPP
        case 1:
            DC.backColor = mkcolor1(color);
            break;
#endif
#ifdef  SUPPORT_4BPP
        case 4:
            DC.backColor = mkcolor4(color);
            break;
#endif
#ifdef  SUPPORT_8BPP
        case 8:
            DC.backColor = mkcolor8(color);
            break;
#endif
#ifdef  SUPPORT_16BPP
        case 15:
        case 16:
            DC.backColor = mkcolor16(color);
            break;
#endif
#if defined(SUPPORT_24BPP) || defined(SUPPORT_32BPP)
        case 24:
        case 32:
            DC.backColor = color;
            break;
#endif
        }
    if (DC.haveHWDraw && setHardwareState)
        DC.hwState.SetBackColor(color);
}

/****************************************************************************
PARAMETERS:
mix     - New mix code to set (GA_mixCodesType)

REMARKS:
This function sets the hardware mix operation for subsequent accelerated
rendering primitives. The mix does not change that often, and is usually
only set once for a range of rendering primitives, so it is set here via
a state function rather than being passed to each rendering functions
(unlike the colors which change constantly). The mix modes are defined
in the GA_mixCodesType enumeration.
****************************************************************************/
N_int32 NAPI REF2D_SetMix(
    N_int32 mix)
{
    DC.mix = REF_mix = mix;
    switch (DC.mi.bitsPerPixel) {
#ifdef  SUPPORT_1BPP
        case 1:     REF2D_SetMix1bpp(mix);  break;
#endif
#ifdef  SUPPORT_4BPP
        case 4:     REF2D_SetMix4bpp(mix);  break;
#endif
#ifdef  SUPPORT_8BPP
        case 8:     REF2D_SetMix8bpp(mix);  break;
#endif
#ifdef  SUPPORT_16BPP
        case 15:
        case 16:    REF2D_SetMix16bpp(mix); break;
#endif
#ifdef  SUPPORT_24BPP
        case 24:    REF2D_SetMix24bpp(mix); break;
#endif
#ifdef  SUPPORT_32BPP
        case 32:    REF2D_SetMix32bpp(mix); break;
#endif
        }
    if (DC.haveHWDraw && setHardwareState) {
        int newSoftwareOnly = !DC.hwState.SetMix(mix);
        if (softwareOnly != newSoftwareOnly) {
            softwareOnly = newSoftwareOnly;
            SwitchVectors();
            }
        }
    return true;
}

/****************************************************************************
PARAMETERS:
index   - Index of the pattern to download
pattern - Pointer to pattern data to download

REMARKS:
This downloads one of 8, 8x8 monochrome patterns for all subsequent monchrome
pattern filled functions. The 8x8 monochrome fill pattern is used for
rectangle and scanline filling, where the pattern is X and Y coordinate
aligned with the left edge and top edge of the display. In the bitmap
pattern, pixel 0 corresponds to bit 7 in byte 0, pixel 1 = bit 6 in byte 0,
... pixel 8 = bit 7 in byte 1 etc. It is the responsibility of the calling
application to rotate the pattern before calling this routine if it is
desired that the pattern be aligned to a different starting coordinate
(such as with Windows Bitmaps and setting the bitmap origin). The bitmap
pattern is passed as a packed array of 8 bytes.
****************************************************************************/
void NAPI REF2D_Set8x8MonoPattern(
    N_int32 index,
    GA_pattern *pattern)
{
    if (index >= 0 && index < 8) {
        memcpy(&DC.mono8x8Patt[index],pattern,sizeof(DC.mono8x8Patt[index]));
        if (DC.haveHWDraw && setHardwareState && DC.hwState.Set8x8MonoPattern)
            DC.hwState.Set8x8MonoPattern(index,pattern);
        }
}

/****************************************************************************
PARAMETERS:
index       - Index of the 8x8 mono pattern to make active
transparent - 1 if the pattern is trasparent, 0 for opaque

REMARKS:
This function selects one of the 8, 8x8 monochrome patterns to be used for
all subsequent monochrome pattern filled functions. If the transparent
parameter is set to 1, where a bit is 0 in the bitmap pattern the destination
pixel remains untouched. If the transparent parameter is set to 0, where a
bit is 0 in the bitmap pattern the destination pixel is drawn in the
background color. In all cases where a bit in the pattern is 1, the
pixel is drawn in the foreground color.
****************************************************************************/
void NAPI REF2D_Use8x8MonoPattern(
    N_int32 index)
{
    int old8x8Trans = DC.mono8x8Transparent;

    if (index >= 0 && index < 8) {
        DC.mono8x8Index         = index;
        DC.mono8x8Transparent   = false;
        if (DC.haveHWDraw && setHardwareState && DC.hwState.Use8x8MonoPattern) {
            DC.hwState.Use8x8MonoPattern(index);
            if (old8x8Trans)
                SwitchBltPatVectors();
            if (transMonoPattSW != false) {
                transMonoPattSW = false;
                SwitchVectors();
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
index       - Index of the 8x8 mono pattern to make active
transparent - 1 if the pattern is trasparent, 0 for opaque

REMARKS:
This function selects one of the 8, 8x8 monochrome patterns to be used for
all subsequent monochrome pattern filled functions. If the transparent
parameter is set to 1, where a bit is 0 in the bitmap pattern the destination
pixel remains untouched. If the transparent parameter is set to 0, where a
bit is 0 in the bitmap pattern the destination pixel is drawn in the
background color. In all cases where a bit in the pattern is 1, the
pixel is drawn in the foreground color.
****************************************************************************/
void NAPI REF2D_Use8x8TransMonoPattern(
    N_int32 index)
{
    int old8x8Trans = DC.mono8x8Transparent;

    if (index >= 0 && index < 8) {
        DC.mono8x8Index         = index;
        DC.mono8x8Transparent   = true;
        if (DC.haveHWDraw && setHardwareState) {
            if (DC.hwState.Use8x8TransMonoPattern) {
                DC.hwState.Use8x8TransMonoPattern(index);
                if (!old8x8Trans)
                    SwitchBltPatVectors();
                }
            else if (transMonoPattSW != true) {
                transMonoPattSW = true;
                SwitchVectors();
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
index   - Index of the pattern to download
pattern - Pointer to pattern data to download

REMARKS:
This downloads one of 8, 8x8 color patterns for all subsequent color
pattern filled functions. The 8x8 color fill pattern is used for rectangle
and scanline filling, where the pattern is X and Y coordinate aligned with
the left edge and top edge of the display. Thus the colors in the pattern
that applies to a specific pixel in the scanline is determine by the
pixel's X starting at the left. Hence pixel 0 corresponds to color 0,
pixel 1 = color 1 etc. It is the responsibility of the calling application
to rotate the pattern before calling this routine if it is desired that the
pattern be aligned to a different starting coordinate (such as with Windows
Bitmaps and setting the bitmap origin). The color pattern is represented
as an 8x8 array of packed pixel data. In 8bpp modes there is 8 bytes per
line, for 16bpp modes there are 16bytes per line, for 24bpp modes there are
24bytes per line and for 32bpp modes there are 32 bytes per line. Hence the
size of the pattern data is different depending on the color depth currently
active. Each pixel color value is packed for the appropriate display mode.

Note:   8 cached patterns are supported because some hardware supports
        caching multiple patterns in offscreen video memory for maximum
        performance. In cases where the hardware only supports a single
        hardware pattern, the driver is responsible for caching the
        pattern data internally and downloading it as efficiently as
        possible to the display hardware.
****************************************************************************/
void NAPI REF2D_Set8x8ColorPattern(
    N_int32 index,
    GA_colorPattern *pattern)
{
    if (index >= 0 && index < 8) {
        switch (DC.mi.bitsPerPixel) {
#ifdef  SUPPORT_1BPP
            case 1:
                memcpy(DC.color8x8Patt[index].b1.p,pattern,sizeof(DC.color8x8Patt[index].b1.p));
                break;
#endif
#ifdef  SUPPORT_4BPP
            case 4: {
                // Internally 4bpp color patterns are expanded to 8bpp patterns using only
                // the four bottom bits of each 8bpp pattern pixel. This speeds up and
                // simplifies the internal pattern fill code considerably.
                int     y;
                uchar   *d = (uchar*)DC.color8x8Patt[index].b8.p;
                uchar   *s = (uchar*)pattern->b4.p;
                for (y = 0; y < 8; y++) {
                    d[0] = s[0] >> 4;
                    d[1] = s[0] & 0x0F;
                    d[2] = s[1] >> 4;
                    d[3] = s[1] & 0x0F;
                    d[4] = s[2] >> 4;
                    d[5] = s[2] & 0x0F;
                    d[6] = s[3] >> 4;
                    d[7] = s[3] & 0x0F;
                    d += 8;
                    s += 4;
                    }
                } break;
#endif
#ifdef  SUPPORT_8BPP
            case 8:
                memcpy(DC.color8x8Patt[index].b8.p,pattern,sizeof(DC.color8x8Patt[index].b8.p));
                break;
#endif
#ifdef  SUPPORT_16BPP
            case 15:
            case 16:
                memcpy(DC.color8x8Patt[index].b16.p,pattern,sizeof(DC.color8x8Patt[index].b16.p));
                break;
#endif
#ifdef  SUPPORT_24BPP
            case 24:
                memcpy(DC.color8x8Patt[index].b24.p,pattern,sizeof(DC.color8x8Patt[index].b24.p));
                break;
#endif
#ifdef  SUPPORT_32BPP
            case 32:
                memcpy(DC.color8x8Patt[index].b32.p,pattern,sizeof(DC.color8x8Patt[index].b32.p));
                break;
#endif
            }
        if (DC.haveHWDraw && setHardwareState && DC.hwState.Set8x8ColorPattern)
            DC.hwState.Set8x8ColorPattern(index,pattern);
        }
}

/****************************************************************************
PARAMETERS:
index       - Index of the 8x8 color pattern to make active

REMARKS:
This function selects one of the 8, 8x8 color patterns to be used for
all subsequent color pattern filled functions.
****************************************************************************/
void NAPI REF2D_Use8x8ColorPattern(
    N_int32 index)
{
    int old8x8Trans = DC.color8x8Transparent;

    if (index >= 0 && index < 8) {
        DC.color8x8Index        = index;
        DC.color8x8Transparent  = false;
        if (DC.haveHWDraw && setHardwareState && DC.hwState.Use8x8ColorPattern) {
            DC.hwState.Use8x8ColorPattern(index);
            if (old8x8Trans)
                SwitchBltColorPatVectors();
            if (transColorPattSW != false) {
                transColorPattSW = false;
                SwitchVectors();
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
index       - Index of the 8x8 color pattern to make active

REMARKS:
This function selects one of the 8, 8x8 color patterns to be used for
all subsequent color pattern filled functions.
****************************************************************************/
void NAPI REF2D_Use8x8TransColorPattern(
    N_int32 index,
    GA_color transparent)
{
    int old8x8Trans = DC.color8x8Transparent;

    if (index >= 0 && index < 8) {
        DC.color8x8Index        = index;
        DC.color8x8Transparent  = true;
        DC.color8x8TransColor   = transparent;
        if (DC.haveHWDraw && setHardwareState) {
            if (DC.hwState.Use8x8TransColorPattern) {
                DC.hwState.Use8x8TransColorPattern(index,transparent);
                if (!old8x8Trans)
                    SwitchBltColorPatVectors();
                }
            else if (transColorPattSW != true) {
                transColorPattSW = true;
                SwitchVectors();
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
stipple - New 16-bit line stipple pattern to set

REMARKS:
This function sets up a 16-bit line stipple for all subsequent stippled line
drawing functions. In the stipple, pixel 0 corresponds to bit 0, pixel 1 =
bit 1, ... pixel 15 = bit 15 etc. If the stippled line is drawn in
transparent mode, where a bit is 0 in the stipple pattern the destination
pixel remains untouched. If the stipple line is drawin in opaque mode,
where a bit is 0 in the stipple pattern the destination pixel is drawn in
the background color. In all cases where a bit in the stipple pattern is 1,
the pixel is drawn in the foreground color.

Note:   When a new stipple pattern is downloaded, the line stipple count is
        reset back to 0.
****************************************************************************/
void NAPI REF2D_SetLineStipple(
    GA_stipple stipple)
{
    DC.stipple      = stipple;
    DC.stippleCount = 0;
    if (DC.haveHWDraw && setHardwareState && DC.hwState.SetLineStipple)
        DC.hwState.SetLineStipple(stipple);
}

/****************************************************************************
PARAMETERS:
count   - New 32-bit line stipple count

REMARKS:
This function sets the line stipple count for stippled line drawing. When
a line is drawn, every pixel that is drawn in the line increments the
stipple counter by 1. The stipple counter modules 16 is used to determine
which bit in the stipple pattern should be used for the next pixel drawn
in the line, and because the driver maintains this stipple count across
stippled line drawing functions, it allows a single stipple pattern to
be correctly applied to a number of connected line segments. This function
allows the user application to preset the stipple count to a specified
value before drawing the next stippled line.
****************************************************************************/
void NAPI REF2D_SetLineStippleCount(
    N_uint32 count)
{
    DC.stippleCount = count;
    if (DC.haveHWDraw && setHardwareState && DC.hwState.SetLineStippleCount)
        DC.hwState.SetLineStippleCount(count);
}

/****************************************************************************
PARAMETERS:
styleMask   - 32-bit style mask for styled lines
styleStep   - 8-bit value added to styleValue for each major pixel
styleValue  - The style value for the first pixel in the line

REMARKS:
This function sets up the parameters for drawing 32-bit OS/2 style styled
lines for all subsequent styled line drawing functions. In the styleMask,
pixel 0 corresponds to bit 0, pixel 1 = bit 1, ... pixel 31 = bit 31 etc. If
the styled line is drawn in transparent mode, where a bit is 0 in the
styleMask the destination pixel remains untouched. If the styled line is
drawn in opaque mode, where a bit is 0 in the styleMask the destination
pixel is drawn in the background color. In all cases where a bit in the
styleMask pattern is 1, the pixel is drawn in the foreground color.

The styleValue passed in is composed of an error value and a mask position
as follows:

    +----------------+----------+----------+-------------+
    |    high word   |  3 bits  |  5 bits  |   8 bits    |
    +----------------+----------+----------+-------------+
    |    not used    | not used | mask pos | error value |
    +----------------+----------+----------+-------------+

The error value determines the error value at the first pixel in the line.
The mask position is an index into the styleMask to determine how the
pixel should be drawn as outlined above.
****************************************************************************/
void NAPI REF2D_SetLineStyle(
    N_uint32 styleMask,
    N_uint32 styleStep,
    N_uint32 styleValue)
{
    DC.styleMask    = styleMask;
    DC.styleStep    = styleStep;
    DC.styleValue   = styleValue;
    if (DC.haveHWDraw && setHardwareState && DC.hwState.SetLineStyle)
        DC.hwState.SetLineStyle(styleMask,styleStep,styleValue);

    // TODO: Add code in here to detect special cases that can be implemented
    //       on top of the old 16-bit fixed stipple hardware line drawing
    //       code!!
}

/****************************************************************************
PARAMETERS:
dstPal      - Destination palette to map to
srcPal      - Source palette to map to
srcColors   - Number of colors in the source palette

REMARKS:
This function builds a color palette translation vector and returns it.
****************************************************************************/
void NAPI REF2D_BuildTranslateVector(
    GA_color *translate,
    GA_palette *dstPal,
    GA_palette *srcPal,
    int srcColors)
{
    DC.buildTranslateVector(translate,dstPal,srcPal,srcColors);
}

/****************************************************************************
PARAMETERS:
mask    - New 32-bit plane mask to set

REMARKS:
This function sets the currently active plane mask for the device driver.
If the plane mask is 0xFFFFFFFF, then plane masking is disabled and we
run at maximum performance. If the plane mask is some other value, plane
masking is enabled at the pixel level.
****************************************************************************/
void NAPI REF2D_SetPlaneMask(
    N_uint32 mask)
{
    if (DC.planeMask != mask) {
        DC.planeMask = mask;
        DC.planeMaskInv = ~mask;
        switch (DC.mi.bitsPerPixel) {
#ifdef  SUPPORT_1BPP
            case 1:
                REF2D_SetPixelFuncs1bpp();
                break;
#endif
#ifdef  SUPPORT_4BPP
            case 4:
                REF2D_SetPixelFuncs4bpp();
                break;
#endif
#ifdef  SUPPORT_8BPP
            case 8:
                REF2D_SetPixelFuncs8bpp();
                break;
#endif
#ifdef  SUPPORT_16BPP
            case 15:
            case 16:
                REF2D_SetPixelFuncs16bpp();
                break;
#endif
#ifdef  SUPPORT_24BPP
            case 24:
                REF2D_SetPixelFuncs24bpp();
                break;
#endif
#ifdef  SUPPORT_32BPP
            case 32:
                REF2D_SetPixelFuncs32bpp();
                break;
#endif
            }

        if (DC.haveHWDraw && setHardwareState && DC.hwState.SetPlaneMask) {
            /* If we are setting the hardware plane mask to something other
             * than 0xFFFFFFFF, then we want to ensure that all system blits
             * go via the accelerator for maximum performance.
             */
            if (mask != 0xFFFFFFFF)
                REF_slowBltSys = true;
            else
                REF_slowBltSys = REF_slowBltSysOrg;
            DC.hwState.SetPlaneMask(mask);
            }
        else
            SwitchVectors();
        REF2D_SetMix(DC.mix);
        }
}

/****************************************************************************
PARAMETERS:
blendFunc   - New source blending function to make active

REMARKS:
Obsolete, use SetBlendFunc instead.
****************************************************************************/
void NAPI REF2D_SetSrcBlendFunc(
    N_int32 blendFunc)
{
    REF2D_SetBlendFunc(blendFunc,DC.dstBlendFunc);
}

/****************************************************************************
PARAMETERS:
blendFunc   - New destination blending function to make active

REMARKS:
Obsolete, use SetBlendFunc instead.
****************************************************************************/
void NAPI REF2D_SetDstBlendFunc(
    N_int32 blendFunc)
{
    REF2D_SetBlendFunc(DC.srcBlendFunc,blendFunc);
}

/****************************************************************************
PARAMETERS:
srcBlendFunc    - New source blending function to make active
dstBlendFunc    - New destination blending function to make active
****************************************************************************/
void NAPI REF2D_SetBlendFunc(
    N_int32 srcBlendFunc,
    N_int32 dstBlendFunc)
{
#if defined(SUPPORT_16BPP) || defined(SUPPORT_24BPP) || defined(SUPPORT_32BPP)
    if (DC.srcBlendFunc != srcBlendFunc || DC.dstBlendFunc != dstBlendFunc) {
        DC.srcBlendFunc = srcBlendFunc;
        DC.dstBlendFunc = dstBlendFunc;
        DC.blendIndex = GET_BLEND_INDEX(srcBlendFunc,dstBlendFunc);
        DC.blendSrc = REF2D_blendSrcTable[srcBlendFunc & 0xF];
        DC.blendDst = REF2D_blendDstTable[dstBlendFunc & 0xF];
        switch (DC.mi.bitsPerPixel) {
#ifdef  SUPPORT_1BPP
            case 1:
                DC.srcBlendFunc = gaBlendNone;
                DC.dstBlendFunc = gaBlendNone;
                return;
#endif
#ifdef  SUPPORT_4BPP
            case 4:
                DC.srcBlendFunc = gaBlendNone;
                DC.dstBlendFunc = gaBlendNone;
                return;
#endif
#ifdef  SUPPORT_8BPP
            case 8:
                DC.srcBlendFunc = gaBlendNone;
                DC.dstBlendFunc = gaBlendNone;
                return;
#endif
#ifdef  SUPPORT_16BPP
            case 15:
            case 16:
                REF2D_SetMix16bpp(DC.mix);
                break;
#endif
#ifdef  SUPPORT_24BPP
            case 24:
                REF2D_SetMix24bpp(DC.mix);
                break;
#endif
#ifdef  SUPPORT_32BPP
            case 32:
                REF2D_SetMix32bpp(DC.mix);
                break;
#endif
            }
        SwitchVectors();
        }
#endif
}

/****************************************************************************
PARAMETERS:
blendFunc   - New destination blending function to make active
****************************************************************************/
void NAPI REF2D_SetAlphaValue(
    N_uint8 alpha)
{
    DC.alphaValue = alpha;
}

/****************************************************************************
PARAMETERS:
x   - X coordinate to read pixel from
y   - Y coordinate to read pixel from

RETURNS:
Color read from the pixel on the screen

REMARKS:
Reads a pixel from the frameBuffer.

NOTE:   This function *assumes* the caller has already enabled direct
        frameBuffer access for performance!!
****************************************************************************/
GA_color NAPI REF2D_GetPixel(
    N_int32 x,
    N_int32 y)
{
    uchar *d = PIXEL_ADDR(x,y);
    return DC.readPixel(d X_ARG(x));
}

/****************************************************************************
PARAMETERS:
x   - X coordinate to plot pixel at
y   - Y coordinate to plot pixel at

REMARKS:
Writes a pixel to the frameBuffer.

NOTE:   This function *assumes* the caller has already enabled direct
        frameBuffer access for performance!!
****************************************************************************/
void NAPI REF2D_PutPixel(
    N_int32 x,
    N_int32 y)
{
    uchar *d = PIXEL_ADDR(x,y);
    DC.drawPixel(d X_ARG(x),DC.foreColor);
}

/****************************************************************************
PARAMETERS:
ileft   - Left coordinate of ellipse rectangle
itop    - Top coordinate of ellipse rectangle
A       - Major axis dimension
B       - Minor axis dimension
****************************************************************************/
void NAPI REF2D_DrawEllipse(
    N_int32 left,
    N_int32 top,
    N_int32 A,
    N_int32 B)
{
    DC.drawEllipse(left,top,A,B);
}

/****************************************************************************
PARAMETERS:
left        - Left coordinate of ellipse rectangle
top         - Top coordinate of ellipse rectangle
A           - Major axis dimension
B           - Minor axis dimension
clipLeft    - Left coordinate for clip rectangle
clipTop     - Top coordinate for clip rectangle
clipRight   - Right coordinate for clip rectangle
clipBottom  - Bottom coordinate for clip rectangle
****************************************************************************/
void NAPI REF2D_ClipEllipse(
    N_int32 left,
    N_int32 top,
    N_int32 A,
    N_int32 B,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    DC.clipEllipse(left,top,A,B,clipLeft,clipTop,clipRight,clipBottom);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI REF2D_DrawScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    uchar *base = Y_ADDR(y);

    DC.beginDirectAccess();
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            DC.drawScan(base + X_ADDR(x1) X_ARG(x1),width);
            }
        base += DC.mi.bytesPerLine;
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI REF2D_DrawPattScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    uchar *base = Y_ADDR(y);

    DC.beginDirectAccess();
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            DC.drawPattScan(base + X_ADDR(x1),x1,y,width);
            }
        base += DC.mi.bytesPerLine;
        y++;
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI REF2D_DrawColorPattScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    uchar *base = Y_ADDR(y);

    DC.beginDirectAccess();
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            DC.drawColorPattScan(base + X_ADDR(x1),x1,y,width);
            }
        base += DC.mi.bytesPerLine;
        y++;
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI REF2D_DrawEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int     i,j,maxIndex;
    uchar   *basei,*basej;

    DC.beginDirectAccess();
    maxIndex = length-1;
    basei = Y_ADDR(y);
    basej = Y_ADDR(y+height);
    for (i = 0,j = height; i < maxIndex; i++,j--,scans += 2) {
        int width = scans[1]-scans[0];
        int xaddr = X_ADDR(scans[0]);
        DC.drawScan(basei + xaddr X_ARG(scans[0]),width);
        DC.drawScan(basej + xaddr X_ARG(scans[0]),width);
        basei += DC.mi.bytesPerLine;
        basej -= DC.mi.bytesPerLine;
        }
    if (!(height & 1))
        DC.drawScan(basei + X_ADDR(scans[0]) X_ARG(scans[0]),scans[1]-scans[0]);
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI REF2D_DrawPattEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int     i,j,maxIndex;
    uchar   *basei,*basej;

    DC.beginDirectAccess();
    maxIndex = length-1;
    basei = Y_ADDR(y);
    basej = Y_ADDR(y+height);
    for (i = 0,j = height; i < maxIndex; i++,j--,scans += 2) {
        int width = scans[1]-scans[0];
        int xaddr = X_ADDR(scans[0]);
        DC.drawPattScan(basei + xaddr,scans[0],y+i,width);
        DC.drawPattScan(basej + xaddr,scans[0],y+j,width);
        basei += DC.mi.bytesPerLine;
        basej -= DC.mi.bytesPerLine;
        }
    if (!(height & 1))
        DC.drawPattScan(basei + X_ADDR(scans[0]),scans[0],y+i,scans[1]-scans[0]);
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI REF2D_DrawColorPattEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int     i,j,maxIndex;
    uchar   *basei,*basej;

    DC.beginDirectAccess();
    maxIndex = length-1;
    basei = Y_ADDR(y);
    basej = Y_ADDR(y+height);
    for (i = 0,j = height; i < maxIndex; i++,j--,scans += 2) {
        int width = scans[1]-scans[0];
        int xaddr = X_ADDR(scans[0]);
        DC.drawColorPattScan(basei + xaddr,scans[0],y+i,width);
        DC.drawColorPattScan(basej + xaddr,scans[0],y+j,width);
        basei += DC.mi.bytesPerLine;
        basej -= DC.mi.bytesPerLine;
        }
    if (!(height & 1))
        DC.drawColorPattScan(basei + X_ADDR(scans[0]),scans[0],y+i,scans[1]-scans[0]);
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI REF2D_DrawFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int     i,j;
    uchar   *basei,*basej;

    DC.beginDirectAccess();
    basei = Y_ADDR(y);
    basej = Y_ADDR(y+height);
    for (i = 0,j = height; i < length; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            int widthL = scans[I_LEFTR]-scans[I_LEFTL];
            int widthR = scans[I_RIGHTR]-scans[I_RIGHTL];
            int xaddrL = X_ADDR(scans[I_LEFTL]);
            int xaddrR = X_ADDR(scans[I_RIGHTL]);
            DC.drawScan(basei + xaddrL X_ARG(scans[I_LEFTL]),widthL);
            DC.drawScan(basei + xaddrR X_ARG(scans[I_RIGHTL]),widthR);
            DC.drawScan(basej + xaddrL X_ARG(scans[I_LEFTL]),widthL);
            DC.drawScan(basej + xaddrR X_ARG(scans[I_RIGHTL]),widthR);
            }
        else {
            int width = scans[I_RIGHTR]-scans[I_LEFTL];
            int xaddr = X_ADDR(scans[I_LEFTL]);
            DC.drawScan(basei + xaddr X_ARG(scans[I_LEFTL]),width);
            DC.drawScan(basej + xaddr X_ARG(scans[I_LEFTL]),width);
            }
        basei += DC.mi.bytesPerLine;
        basej -= DC.mi.bytesPerLine;
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            int widthL = scans[I_LEFTR]-scans[I_LEFTL];
            int widthR = scans[I_RIGHTR]-scans[I_RIGHTL];
            int xaddrL = X_ADDR(scans[I_LEFTL]);
            int xaddrR = X_ADDR(scans[I_RIGHTL]);
            DC.drawScan(basei + xaddrL X_ARG(scans[I_LEFTL]),widthL);
            DC.drawScan(basei + xaddrR X_ARG(scans[I_RIGHTL]),widthR);
            }
        else {
            int width = scans[I_RIGHTR]-scans[I_LEFTL];
            int xaddr = X_ADDR(scans[I_LEFTL]);
            DC.drawScan(basei + xaddr X_ARG(scans[I_LEFTL]),width);
            }
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI REF2D_DrawPattFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int     i,j;
    uchar   *basei,*basej;

    DC.beginDirectAccess();
    basei = Y_ADDR(y);
    basej = Y_ADDR(y+height);
    for (i = 0,j = height; i < length; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            int widthL = scans[I_LEFTR]-scans[I_LEFTL];
            int widthR = scans[I_RIGHTR]-scans[I_RIGHTL];
            int xaddrL = X_ADDR(scans[I_LEFTL]);
            int xaddrR = X_ADDR(scans[I_RIGHTL]);
            DC.drawPattScan(basei + xaddrL,scans[I_LEFTL],y+i,widthL);
            DC.drawPattScan(basei + xaddrR,scans[I_RIGHTL],y+i,widthR);
            DC.drawPattScan(basej + xaddrL,scans[I_LEFTL],y+j,widthL);
            DC.drawPattScan(basej + xaddrR,scans[I_RIGHTL],y+j,widthR);
            }
        else {
            int width = scans[I_RIGHTR]-scans[I_LEFTL];
            int xaddr = X_ADDR(scans[I_LEFTL]);
            DC.drawPattScan(basei + xaddr,scans[I_LEFTL],y+i,width);
            DC.drawPattScan(basej + xaddr,scans[I_LEFTL],y+j,width);
            }
        basei += DC.mi.bytesPerLine;
        basej -= DC.mi.bytesPerLine;
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            int widthL = scans[I_LEFTR]-scans[I_LEFTL];
            int widthR = scans[I_RIGHTR]-scans[I_RIGHTL];
            int xaddrL = X_ADDR(scans[I_LEFTL]);
            int xaddrR = X_ADDR(scans[I_RIGHTL]);
            DC.drawPattScan(basei + xaddrL,scans[I_LEFTL],y+i,widthL);
            DC.drawPattScan(basei + xaddrR,scans[I_RIGHTL],y+i,widthR);
            }
        else {
            int width = scans[I_RIGHTR]-scans[I_LEFTL];
            int xaddr = X_ADDR(scans[I_LEFTL]);
            DC.drawPattScan(basei + xaddr,scans[I_LEFTL],y+i,width);
            }
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI REF2D_DrawColorPattFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int     i,j;
    uchar   *basei,*basej;

    DC.beginDirectAccess();
    basei = Y_ADDR(y);
    basej = Y_ADDR(y+height);
    for (i = 0,j = height; i < length; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            int widthL = scans[I_LEFTR]-scans[I_LEFTL];
            int widthR = scans[I_RIGHTR]-scans[I_RIGHTL];
            int xaddrL = X_ADDR(scans[I_LEFTL]);
            int xaddrR = X_ADDR(scans[I_RIGHTL]);
            DC.drawColorPattScan(basei + xaddrL,scans[I_LEFTL],y+i,widthL);
            DC.drawColorPattScan(basei + xaddrR,scans[I_RIGHTL],y+i,widthR);
            DC.drawColorPattScan(basej + xaddrL,scans[I_LEFTL],y+j,widthL);
            DC.drawColorPattScan(basej + xaddrR,scans[I_RIGHTL],y+j,widthR);
            }
        else {
            int width = scans[I_RIGHTR]-scans[I_LEFTL];
            int xaddr = X_ADDR(scans[I_LEFTL]);
            DC.drawColorPattScan(basei + xaddr,scans[I_LEFTL],y+i,width);
            DC.drawColorPattScan(basej + xaddr,scans[I_LEFTL],y+j,width);
            }
        basei += DC.mi.bytesPerLine;
        basej -= DC.mi.bytesPerLine;
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            int widthL = scans[I_LEFTR]-scans[I_LEFTL];
            int widthR = scans[I_RIGHTR]-scans[I_RIGHTL];
            int xaddrL = X_ADDR(scans[I_LEFTL]);
            int xaddrR = X_ADDR(scans[I_RIGHTL]);
            DC.drawColorPattScan(basei + xaddrL,scans[I_LEFTL],y+i,widthL);
            DC.drawColorPattScan(basei + xaddrR,scans[I_RIGHTL],y+i,widthR);
            }
        else {
            int width = scans[I_RIGHTR]-scans[I_LEFTL];
            int xaddr = X_ADDR(scans[I_LEFTL]);
            DC.drawColorPattScan(basei + xaddr,scans[I_LEFTL],y+i,width);
            }
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI REF2D_DrawRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    uchar *base = PIXEL_ADDR(left,top);

    DC.beginDirectAccess();
    while (height--) {
        DC.drawScan(base X_ARG(left),width);
        base += DC.mi.bytesPerLine;
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
dstOfs      - Offset of destination rectangle in video memory
dstPitch    - Pitch of destination rectangle in bytes
left        - Left coordinate of the rectangle to draw
top         - Top coordinate of the rectangle to draw
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
color       - Color to fill the rectangle with
mix         - Mix to use for drawing the rectangle
****************************************************************************/
void NAPI REF2D_DrawRectLin(
    N_int32 dstOfs,
    N_int32 dstPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    GA_color color,
    N_int32 mix)
{
    DC.beginDirectAccess();
    REF2D_DrawRectExtSW(DC.frameBuffer + dstOfs,DC.mi.bytesPerLine,left,top,width,height,color,mix);
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
left        - Left coordinate of the rectangle to draw
top         - Top coordinate of the rectangle to draw
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
color       - Color to fill the rectangle with
mix         - Mix to use for drawing the rectangle
****************************************************************************/
void NAPI REF2D_DrawRectExt(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    GA_color color,
    N_int32 mix)
{
    DC.beginDirectAccess();
    REF2D_DrawRectExtSW(DC.surface,DC.mi.bytesPerLine,left,top,width,height,color,mix);
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
left        - Left coordinate of the rectangle to draw
top         - Top coordinate of the rectangle to draw
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
color       - Color to fill the rectangle with
mix         - Mix to use for drawing the rectangle
****************************************************************************/
void NAPI REF2D_DrawRectExtSW(
    void *buffer,
    N_int32 dstPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    GA_color color,
    N_int32 mix)
{
    uchar *base = (uchar*)buffer + dstPitch * top + X_ADDR(left);

#if defined(SUPPORT_1BPP) || defined(SUPPORT_4BPP) || defined(SUPPORT_8BPP) || defined(SUPPORT_16BPP)
    switch (DC.mi.bitsPerPixel) {
#ifdef  SUPPORT_1BPP
        case 1:
            color = mkcolor1(color);
            break;
#endif
#ifdef  SUPPORT_4BPP
        case 4:
            color = mkcolor4(color);
            break;
#endif
#ifdef  SUPPORT_8BPP
        case 8:
            color = mkcolor8(color);
            break;
#endif
#ifdef  SUPPORT_16BPP
        case 15:
        case 16:
            color = mkcolor16(color);
            break;
#endif
        }
#endif
    if (IS_SOLID_MIX(mix)) {
        while (height--) {
            DC.fillScan_S(base X_ARG(left),width,color);
            base += dstPitch;
            }
        }
    else {
        REF2D_drawPixel_t   oldDrawPixel = DC.drawPixel;
        DC.drawPixel = GET_PIXEL_FUNC(mix);
        while (height--) {
            DC.fillScan_Generic(base X_ARG(left),width,color);
            base += dstPitch;
            }
        DC.drawPixel = oldDrawPixel;
        }
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI REF2D_DrawPattRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    uchar *base = PIXEL_ADDR(left,top);

    DC.beginDirectAccess();
    while (height--) {
        DC.drawPattScan(base,left,top,width);
        base += DC.mi.bytesPerLine;
        top++;
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI REF2D_DrawColorPattRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    uchar *base = PIXEL_ADDR(left,top);

    DC.beginDirectAccess();
    while (height--) {
        DC.drawColorPattScan(base,left,top,width);
        base += DC.mi.bytesPerLine;
        top++;
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid
****************************************************************************/
void NAPI REF2D_DrawTrap(
    GA_trap *trap)
{
    uchar   *base = Y_ADDR(trap->y);
    int     x1,x2,count = trap->count;

    if (!count)
        return;
    x1 = trap->x1;
    x2 = trap->x2;
    trap->y += count;
    DC.beginDirectAccess();
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            DC.drawScan(base + X_ADDR(ix1) X_ARG(ix1),width);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        base += DC.mi.bytesPerLine;
        }
    DC.endDirectAccess();
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid
****************************************************************************/
void NAPI REF2D_DrawPattTrap(
    GA_trap *trap)
{
    uchar   *base = Y_ADDR(trap->y);
    int     y,x1,x2,count = trap->count;

    if (!count)
        return;
    x1 = trap->x1;
    x2 = trap->x2;
    y = trap->y;
    trap->y += count;
    DC.beginDirectAccess();
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            DC.drawPattScan(base + X_ADDR(ix1),ix1,y,width);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        base += DC.mi.bytesPerLine;
        y++;
        }
    DC.endDirectAccess();
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid
****************************************************************************/
void NAPI REF2D_DrawColorPattTrap(
    GA_trap *trap)
{
    uchar   *base = Y_ADDR(trap->y);
    int     y,x1,x2,count = trap->count;

    if (!count)
        return;
    x1 = trap->x1;
    x2 = trap->x2;
    y = trap->y;
    trap->y += count;
    DC.beginDirectAccess();
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            DC.drawColorPattScan(base + X_ADDR(ix1),ix1,y,width);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        base += DC.mi.bytesPerLine;
        y++;
        }
    DC.endDirectAccess();
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
****************************************************************************/
void NAPI REF2D_DrawLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast)
{
    int     absDeltaX,absDeltaY,majorInc,diagInc;
    int     initialError,flags,fincr,count;
    uchar   *addr;

    flags = gaLineXPositive | gaLineYPositive | gaLineXMajor;
    if ((absDeltaX = x2 - x1) < 0) {
        absDeltaX = -absDeltaX;
        flags &= ~gaLineXPositive;
        }
    if ((absDeltaY = y2 - y1) < 0) {
        fincr = -DC.mi.bytesPerLine;
        absDeltaY = -absDeltaY;
        flags &= ~gaLineYPositive;
        }
    else
        fincr = DC.mi.bytesPerLine;
    if (absDeltaY > absDeltaX) {
        SWAP(absDeltaX,absDeltaY);
        flags &= ~gaLineXMajor;
        }
    majorInc = 2 * absDeltaY;               // 2 * dy
    initialError = majorInc - absDeltaX;    // 2 * dy - dx
    diagInc = initialError - absDeltaX;     // 2 * (dy - dx)
    addr = PIXEL_ADDR(x1,y1);
    if (drawLast)
        count = absDeltaX+1;
    else
        count = absDeltaX;
    DC.beginDirectAccess();
    if (flags & gaLineXMajor) {
        if (flags & gaLineXPositive)
            DC.drawLineXPosLo(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr);
        else
            DC.drawLineXNegLo(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr);
        }
    else {
        if (flags & gaLineXPositive)
            DC.drawLineXPosHi(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr);
        else
            DC.drawLineXNegHi(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr);
        }
    DC.endDirectAccess();
}

/****************************************************************************
DESCRIPTION:
Draws a solid, single pixel wide line with bresenham parameters.

HEADER:
snap/graphics.h

PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing

REMARKS:
This function is similar to the regular DrawLineInt function,
except that the bresenham parameters for the line are pre-computed and
passed to this function. Allowing the application to pre-compute the
bresenham parameters allows for accurate clipping as the bresenham
parameters can be computed for the unclipped line, and initialised to the
start of the clipped line segement. This function is used by OS/2 display
drivers to get accurate line clipping. The values passed to the function
for a regular integer line can be computed as follows:

    flags = gaLineXPositive | gaLineYPositive | gaLineXMajor
        | gaLineDoLastPel;
    if ((absDeltaX = x2 - x1) < 0) {
        absDeltaX = -absDeltaX;
        flags &= ~gaLineXPositive;
        }
    if (absDeltaX == 0)
        flags |= gaLineVertical;
    if ((absDeltaY = y2 - y1) < 0) {
        absDeltaY = -absDeltaY;
        flags &= ~gaLineYPositive;
        }
    if (absDeltaY == 0)
        flags |= gaLineHorizontal;
    if (absDeltaY > absDeltaX) {
        majorInc = 2 * absDeltaY;               // 2 * dy
        initialError = majorInc - absDeltaX;    // 2 * dy - dx
        diagInc = initialError - absDeltaX;     // 2 * (dy - dx)
        count = absDeltaX+1;
        flags &= ~gaLineXMajor;
        }
    else {
        majorInc = 2 * absDeltaX;               // 2 * dx
        initialError = majorInc - absDeltaY;    // 2 * dx - dy
        diagInc = initialError - absDeltaY;     // 2 * (dx - dy)
        count = absDeltaY+1;
        }

****************************************************************************/
void NAPI REF2D_DrawBresenhamLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags)
{
    int     fincr;              // Increment for values
    uchar   *addr;              // Framebuffer address

    if (flags & gaLineYPositive)
        fincr = DC.mi.bytesPerLine;
    else
        fincr = -DC.mi.bytesPerLine;
    addr = PIXEL_ADDR(x1,y1);
    if (!(flags & gaLineDoLastPel))
        count--;
    DC.beginDirectAccess();
    if (flags & gaLineXMajor) {
        if (flags & gaLineXPositive)
            DC.drawLineXPosLo(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr);
        else
            DC.drawLineXNegLo(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr);
        }
    else {
        if (flags & gaLineXPositive)
            DC.drawLineXPosHi(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr);
        else
            DC.drawLineXNegHi(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr);
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
transparent - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI REF2D_DrawStippleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent)
{
    int     absDeltaX,absDeltaY,majorInc,diagInc;
    int     initialError,flags,fincr,count;
    ulong   mask,stipple;
    uchar   *addr;

    flags = gaLineXPositive | gaLineYPositive | gaLineXMajor;
    if ((absDeltaX = x2 - x1) < 0) {
        absDeltaX = -absDeltaX;
        flags &= ~gaLineXPositive;
        }
    if ((absDeltaY = y2 - y1) < 0) {
        fincr = -DC.mi.bytesPerLine;
        absDeltaY = -absDeltaY;
        flags &= ~gaLineYPositive;
        }
    else
        fincr = DC.mi.bytesPerLine;
    if (absDeltaY > absDeltaX) {
        SWAP(absDeltaX,absDeltaY);
        flags &= ~gaLineXMajor;
        }
    majorInc = 2 * absDeltaY;               // 2 * dy
    initialError = majorInc - absDeltaX;    // 2 * dy - dx
    diagInc = initialError - absDeltaX;     // 2 * (dy - dx)
    addr = PIXEL_ADDR(x1,y1);
    stipple = DC.stipple | 0x10000L;
    mask = (stipple >> (DC.stippleCount & 0xF));
    if (drawLast)
        count = absDeltaX+1;
    else
        count = absDeltaX;
    DC.stippleCount += count;
    DC.beginDirectAccess();
    if (flags & gaLineXMajor) {
        if (flags & gaLineXPositive)
            DC.drawStippleLineXPosLo(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,mask,stipple,transparent);
        else
            DC.drawStippleLineXNegLo(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,mask,stipple,transparent);
        }
    else {
        if (flags & gaLineXPositive)
            DC.drawStippleLineXPosHi(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,mask,stipple,transparent);
        else
            DC.drawStippleLineXNegHi(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,mask,stipple,transparent);
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
transparent     - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI REF2D_DrawBresenhamStippleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent)
{
    ulong   mask,stipple;
    int     fincr;
    uchar   *addr;

    if (flags & gaLineYPositive)
        fincr = DC.mi.bytesPerLine;
    else
        fincr = -DC.mi.bytesPerLine;
    addr = PIXEL_ADDR(x1,y1);
    stipple = DC.stipple | 0x10000L;
    mask = (stipple >> (DC.stippleCount & 0xF));
    if (!(flags & gaLineDoLastPel))
        count--;
    DC.stippleCount += count;
    DC.beginDirectAccess();
    if (flags & gaLineXMajor) {
        if (flags & gaLineXPositive)
            DC.drawStippleLineXPosLo(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,mask,stipple,transparent);
        else
            DC.drawStippleLineXNegLo(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,mask,stipple,transparent);
        }
    else {
        if (flags & gaLineXPositive)
            DC.drawStippleLineXPosHi(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,mask,stipple,transparent);
        else
            DC.drawStippleLineXNegHi(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,mask,stipple,transparent);
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
transparent - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI REF2D_DrawStyleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent)
{
    int     absDeltaX,absDeltaY,majorInc,diagInc;
    int     initialError,flags,fincr,count;
    ulong   mask,stipple;
    uchar   *addr;

    flags = gaLineXPositive | gaLineYPositive | gaLineXMajor;
    if ((absDeltaX = x2 - x1) < 0) {
        absDeltaX = -absDeltaX;
        flags &= ~gaLineXPositive;
        }
    if ((absDeltaY = y2 - y1) < 0) {
        fincr = -DC.mi.bytesPerLine;
        absDeltaY = -absDeltaY;
        flags &= ~gaLineYPositive;
        }
    else
        fincr = DC.mi.bytesPerLine;
    if (absDeltaY > absDeltaX) {
        SWAP(absDeltaX,absDeltaY);
        flags &= ~gaLineXMajor;
        }
    majorInc = 2 * absDeltaY;               // 2 * dy
    initialError = majorInc - absDeltaX;    // 2 * dy - dx
    diagInc = initialError - absDeltaX;     // 2 * (dy - dx)
    addr = PIXEL_ADDR(x1,y1);
    stipple = DC.stipple | 0x10000L;
    mask = (stipple >> (DC.stippleCount & 0xF));
    if (drawLast)
        count = absDeltaX+1;
    else
        count = absDeltaX;
    DC.beginDirectAccess();
    if (flags & gaLineXMajor) {
        if (flags & gaLineXPositive)
            DC.drawStyleLineXPosLo(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,DC.styleMask,DC.styleStep,DC.styleValue,transparent);
        else
            DC.drawStyleLineXNegLo(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,DC.styleMask,DC.styleStep,DC.styleValue,transparent);
        }
    else {
        if (flags & gaLineXPositive)
            DC.drawStyleLineXPosHi(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,DC.styleMask,DC.styleStep,DC.styleValue,transparent);
        else
            DC.drawStyleLineXNegHi(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,DC.styleMask,DC.styleStep,DC.styleValue,transparent);
        }
    DC.endDirectAccess();

    // Adjust the internal style values for the next line
    DC.styleValue += DC.styleStep * count;
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
transparent     - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI REF2D_DrawBresenhamStyleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent)
{
    int     fincr;
    uchar   *addr;

    if (flags & gaLineYPositive)
        fincr = DC.mi.bytesPerLine;
    else
        fincr = -DC.mi.bytesPerLine;
    addr = PIXEL_ADDR(x1,y1);
    if (!(flags & gaLineDoLastPel))
        count--;
    DC.beginDirectAccess();
    if (flags & gaLineXMajor) {
        if (flags & gaLineXPositive)
            DC.drawStyleLineXPosLo(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,DC.styleMask,DC.styleStep,DC.styleValue,transparent);
        else
            DC.drawStyleLineXNegLo(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,DC.styleMask,DC.styleStep,DC.styleValue,transparent);
        }
    else {
        if (flags & gaLineXPositive)
            DC.drawStyleLineXPosHi(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,DC.styleMask,DC.styleStep,DC.styleValue,transparent);
        else
            DC.drawStyleLineXNegHi(addr X_ARG(x1),initialError,majorInc,diagInc,count,fincr,DC.styleMask,DC.styleStep,DC.styleValue,transparent);
        }
    DC.endDirectAccess();

    // Adjust the internal style values for the next line
    DC.styleValue += DC.styleStep * count;
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI REF2D_DrawClippedLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{

// TODO: We could add some new functions to handle hardware clipping for
//       line drawing and we could automatically use this if we know the
//       range of values that the hardware can accept and clip against.
//
//       Then we can check for this particular case directly and use the
//       hardware as necessary to clip. Although we may just want to have
//       the hardware driver itself check for this condition and handle
//       it internally.
//
//       Note that I was thinking we could use some decent 'defaults' for
//       the range of values the hardware can support, and then we can
//       extend this as necessary for new hardware devices as we move
//       forward.

#define DRAW_LINE_INT           STUB_DrawLineInt
#define DRAW_BRESENHAM_LINE     STUB_DrawBresenhamLine
#include "raster/lineclip.c"
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI REF2D_DrawClippedBresenhamLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
#define DRAW_BRESENHAM_LINE     STUB_DrawBresenhamLine
#include "raster/blineclp.c"
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI REF2D_DrawClippedStippleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
#define DRAW_LINE_INT(a,b,c,d,e)            STUB_DrawStippleLineInt(a,b,c,d,e,transparent)
#define DRAW_BRESENHAM_LINE(a,b,c,d,e,f,g)  STUB_DrawBresenhamStippleLine(a,b,c,d,e,f,g,transparent)
#include "raster/lineclip.c"
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI REF2D_DrawClippedBresenhamStippleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
#define DRAW_BRESENHAM_LINE(a,b,c,d,e,f,g)  STUB_DrawBresenhamStippleLine(a,b,c,d,e,f,g,transparent)
#include "raster/blineclp.c"
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI REF2D_DrawClippedStyleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
#define DRAW_LINE_INT(a,b,c,d,e)            STUB_DrawStyleLineInt(a,b,c,d,e,transparent)
#define DRAW_BRESENHAM_LINE(a,b,c,d,e,f,g)  STUB_DrawBresenhamStyleLine(a,b,c,d,e,f,g,transparent)
#include "raster/lineclip.c"
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI REF2D_DrawClippedBresenhamStyleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
#define DRAW_BRESENHAM_LINE(a,b,c,d,e,f,g)  STUB_DrawBresenhamStyleLine(a,b,c,d,e,f,g,transparent)
#include "raster/blineclp.c"
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque

REMARKS:
If the transparent parameter is set to 1, where a bit is 0 in the bitmap
image the destination pixel remains untouched. If the transparent
parameter is set to 0, where a bit is 0 in the bitmap image the
destination pixel is drawn in the background color. In all cases where a
bit in the bitmap image is 1, the pixel is drawn in the foreground color.

Note:   This function processes the bitmap data in MSB fashion, in that the
        first pixel drawn corrsponds to bit 7 of the first byte. The second
        pixel is bit 6 of the first byte, ...,the 8th pixel is bit 7 of
        the second byte etc. Both LSB and MSB versions are provided for
        performance.
****************************************************************************/
void NAPI REF2D_PutMonoImageMSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent)
{
#define MONO_TRANS          drawMonoMSBTrans
#define MONO_OPAQUE         drawMonoMSBOpaque
#include "raster/monoimg.c"
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
imageOfs    - Offset of bitmap image data in video memory
transparent - 1 for transparent, 0 for opaque
****************************************************************************/
void NAPI REF2D_PutMonoImageMSBLin(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_int32 imageOfs,
    N_int32 transparent)
{
    // Call the system memory version above.
    REF2D_PutMonoImageMSBSys(x,y,width,height,byteWidth,DC.hwBufferStart+imageOfs,transparent);
}

/****************************************************************************
PARAMETERS:
x               - Destination X coordinate to draw the bitmap at
y               - Destination Y coordinate to draw the bitmap at
width           - Width of the bitmap in pixels
height          - Height of the bitmap in pixels
byteWidth       - Width of the bitmap in bytes
image           - Pointer to the bitmap image data to draw
imagePhysAddr   - Offset of bitmap image data in video memory
transparent     - 1 for transparent, 0 for opaque
****************************************************************************/
void NAPI REF2D_PutMonoImageMSBBM(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 imagePhysAddr,
    N_int32 transparent)
{
    // Call the system memory version above
    (void)imagePhysAddr;
    REF2D_PutMonoImageMSBSys(x,y,width,height,byteWidth,image,transparent);
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque

REMARKS:
This function is identical to the PutMonoImageMSBSys function, except
that it processes the bitmap data in LSB fashion. This means that the
first pixel drawn corrsponds to bit 0 of the first byte. The second
pixel is bit 1 of the first byte, ...,the 8th pixel is bit 0 of the second
byte etc. Both LSB and MSB versions are provided for performance.
****************************************************************************/
void NAPI REF2D_PutMonoImageLSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent)
{
#define MONO_TRANS          drawMonoLSBTrans
#define MONO_OPAQUE         drawMonoLSBOpaque
#include "raster/monoimg.c"
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
imageOfs    - Offset of bitmap image data in video memory
transparent - 1 for transparent, 0 for opaque
****************************************************************************/
void NAPI REF2D_PutMonoImageLSBLin(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_int32 imageOfs,
    N_int32 transparent)
{
    // Call the system memory version above.
    REF2D_PutMonoImageLSBSys(x,y,width,height,byteWidth,DC.hwBufferStart+imageOfs,transparent);
}

/****************************************************************************
PARAMETERS:
x               - Destination X coordinate to draw the bitmap at
y               - Destination Y coordinate to draw the bitmap at
width           - Width of the bitmap in pixels
height          - Height of the bitmap in pixels
byteWidth       - Width of the bitmap in bytes
image           - Pointer to the bitmap image data to draw
imagePhysAddr   - Offset of bitmap image data in video memory
transparent     - 1 for transparent, 0 for opaque
****************************************************************************/
void NAPI REF2D_PutMonoImageLSBBM(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 imagePhysAddr,
    N_int32 transparent)
{
    // Call the system memory version above
    (void)imagePhysAddr;
    REF2D_PutMonoImageLSBSys(x,y,width,height,byteWidth,image,transparent);
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque
clipLeft    - Left edge of clip rectangle (inclusive)
clipTop     - Top edge of clip rectangle (inclusive)
clipRight   - Right edge of clip rectangle (exclusive)
clipBottom  - Bottom edge of clip rectangle (exclusive)

REMARKS:
If the transparent parameter is set to 1, where a bit is 0 in the bitmap
image the destination pixel remains untouched. If the transparent
parameter is set to 0, where a bit is 0 in the bitmap image the
destination pixel is drawn in the background color. In all cases where a
bit in the bitmap image is 1, the pixel is drawn in the foreground color.

Note:   This function processes the bitmap data in MSB fashion, in that the
        first pixel drawn corrsponds to bit 7 of the first byte. The second
        pixel is bit 6 of the first byte, ...,the 8th pixel is bit 7 of
        the second byte etc. Both LSB and MSB versions are provided for
        performance.
****************************************************************************/
void NAPI REF2D_ClipMonoImageMSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
#define CLIP_MONO_TRANS     drawClipMonoMSBTrans
#define CLIP_MONO_OPAQUE    drawClipMonoMSBOpaque
#define MONO_TRANS          drawMonoMSBTrans
#define MONO_OPAQUE         drawMonoMSBOpaque
#include "raster/clipmono.c"
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
imageOfs    - Offset of bitmap image data in video memory
transparent - 1 for transparent, 0 for opaque
clipLeft    - Left edge of clip rectangle (inclusive)
clipTop     - Top edge of clip rectangle (inclusive)
clipRight   - Right edge of clip rectangle (exclusive)
clipBottom  - Bottom edge of clip rectangle (exclusive)
****************************************************************************/
void NAPI REF2D_ClipMonoImageMSBLin(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_int32 imageOfs,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    // Call the system memory version above.
    REF2D_ClipMonoImageMSBSys(x,y,width,height,byteWidth,DC.hwBufferStart+imageOfs,transparent,clipLeft,clipTop,clipRight,clipBottom);
}

/****************************************************************************
PARAMETERS:
x               - Destination X coordinate to draw the bitmap at
y               - Destination Y coordinate to draw the bitmap at
width           - Width of the bitmap in pixels
height          - Height of the bitmap in pixels
byteWidth       - Width of the bitmap in bytes
image           - Pointer to the bitmap image data to draw
imagePhysAddr   - Offset of bitmap image data in video memory
transparent     - 1 for transparent, 0 for opaque
clipLeft    - Left edge of clip rectangle (inclusive)
clipTop     - Top edge of clip rectangle (inclusive)
clipRight   - Right edge of clip rectangle (exclusive)
clipBottom  - Bottom edge of clip rectangle (exclusive)
****************************************************************************/
void NAPI REF2D_ClipMonoImageMSBBM(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 imagePhysAddr,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    // Call the system memory version above
    (void)imagePhysAddr;
    REF2D_ClipMonoImageMSBSys(x,y,width,height,byteWidth,image,transparent,clipLeft,clipTop,clipRight,clipBottom);
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque
clipLeft    - Left edge of clip rectangle (inclusive)
clipTop     - Top edge of clip rectangle (inclusive)
clipRight   - Right edge of clip rectangle (exclusive)
clipBottom  - Bottom edge of clip rectangle (exclusive)

REMARKS:
This function is identical to the ClipMonoImageMSBSys function, except
that it processes the bitmap data in LSB fashion. This means that the
first pixel drawn corrsponds to bit 0 of the first byte. The second
pixel is bit 1 of the first byte, ...,the 8th pixel is bit 0 of the second
byte etc. Both LSB and MSB versions are provided for performance.
****************************************************************************/
void NAPI REF2D_ClipMonoImageLSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
#define CLIP_MONO_TRANS     drawClipMonoLSBTrans
#define CLIP_MONO_OPAQUE    drawClipMonoLSBOpaque
#define MONO_TRANS          drawMonoLSBTrans
#define MONO_OPAQUE         drawMonoLSBOpaque
#include "raster/clipmono.c"
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
imageOfs    - Offset of bitmap image data in video memory
transparent - 1 for transparent, 0 for opaque
clipLeft    - Left edge of clip rectangle (inclusive)
clipTop     - Top edge of clip rectangle (inclusive)
clipRight   - Right edge of clip rectangle (exclusive)
clipBottom  - Bottom edge of clip rectangle (exclusive)
****************************************************************************/
void NAPI REF2D_ClipMonoImageLSBLin(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_int32 imageOfs,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    // Call the system memory version above.
    REF2D_ClipMonoImageLSBSys(x,y,width,height,byteWidth,DC.hwBufferStart+imageOfs,transparent,clipLeft,clipTop,clipRight,clipBottom);
}

/****************************************************************************
PARAMETERS:
x               - Destination X coordinate to draw the bitmap at
y               - Destination Y coordinate to draw the bitmap at
width           - Width of the bitmap in pixels
height          - Height of the bitmap in pixels
byteWidth       - Width of the bitmap in bytes
image           - Pointer to the bitmap image data to draw
imagePhysAddr   - Offset of bitmap image data in video memory
transparent     - 1 for transparent, 0 for opaque
clipLeft    - Left edge of clip rectangle (inclusive)
clipTop     - Top edge of clip rectangle (inclusive)
clipRight   - Right edge of clip rectangle (exclusive)
clipBottom  - Bottom edge of clip rectangle (exclusive)
****************************************************************************/
void NAPI REF2D_ClipMonoImageLSBBM(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 imagePhysAddr,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    // Call the system memory version above
    (void)imagePhysAddr;
    REF2D_ClipMonoImageLSBSys(x,y,width,height,byteWidth,image,transparent,clipLeft,clipTop,clipRight,clipBottom);
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
mix     - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_BitBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    uchar               *s,*d;
    int                 incy = (dstTop <= srcTop),dstPitch;
    REF2D_drawPixel_t   oldDrawPixel = DC.drawPixel;
    bltScan_t           bltScan;

    // Ensure our blit buffers are large enough
    if (width > DC.maxBltWidth)
        REF2D_resizeBltBuffers(width);

    // Determine if the mix is a solid mix, and setup appropriate pixel
    // function if not.
    if (IS_SOLID_MIX(mix))
        bltScan = DC.bltScan_S;
    else {
        DC.drawPixel = GET_PIXEL_FUNC(mix);
        bltScan = DC.bltScan_Generic;
        }
    DC.beginDirectAccess();
    if (incy) {
        s = PIXEL_ADDR(srcLeft,srcTop);
        d = PIXEL_ADDR(dstLeft,dstTop);
        dstPitch = DC.mi.bytesPerLine;
        }
    else {
        s = PIXEL_ADDR(srcLeft,srcTop+height-1);
        d = PIXEL_ADDR(dstLeft,dstTop+height-1);
        dstPitch = -DC.mi.bytesPerLine;
        }
    while (height--) {
        DC.readScan(DC.scanBuf,s X_ARG2(srcLeft,0),width);
        bltScan(d,DC.scanBuf X_ARG2(0,dstLeft),width);
        s += dstPitch;
        d += dstPitch;
        }
    DC.endDirectAccess();
    DC.drawPixel = oldDrawPixel;
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_BitBltLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    uchar               *s = (uchar*)DC.hwBufferStart + srcOfs + (srcPitch * srcTop) + X_ADDR(srcLeft);
    uchar               *d = PIXEL_ADDR(dstLeft,dstTop);
    REF2D_drawPixel_t   oldDrawPixel = DC.drawPixel;
    bltScan_t           bltScan;

    // Ensure our blit buffers are large enough
    if (width > DC.maxBltWidth)
        REF2D_resizeBltBuffers(width);

    // Determine if the mix is a solid mix, and setup appropriate pixel
    // function if not.
    if (IS_SOLID_MIX(mix))
        bltScan = DC.bltScan_S;
    else {
        DC.drawPixel = GET_PIXEL_FUNC(mix);
        bltScan = DC.bltScan_Generic;
        }
    DC.beginDirectAccess();
    while (height--) {
        DC.readScan(DC.scanBuf,s X_ARG2(srcLeft,0),width);
        bltScan(d,DC.scanBuf X_ARG2(0,dstLeft),width);
        s += srcPitch;
        d += DC.mi.bytesPerLine;
        }
    DC.drawPixel = oldDrawPixel;
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI REF2D_BitBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    N_int32 flipY)
{
    uchar               *s = ((uchar*)srcAddr + (srcPitch * srcTop)) + X_ADDR(srcLeft);
    uchar               *d = PIXEL_ADDR(dstLeft,dstTop);
    int                 dstPitch;
    REF2D_drawPixel_t   oldDrawPixel = DC.drawPixel;
    bltScan_t           bltScan;

    // Determine if the mix is a solid mix, and setup appropriate pixel
    // function if not.
    if (IS_SOLID_MIX(mix))
        bltScan = DC.bltScan_S;
    else {
        DC.drawPixel = GET_PIXEL_FUNC(mix);
        bltScan = DC.bltScan_Generic;
        }
    if (flipY) {
        d += (height-1) * DC.mi.bytesPerLine;
        dstPitch = -DC.mi.bytesPerLine;
        }
    else
        dstPitch = DC.mi.bytesPerLine;
    DC.beginDirectAccess();
    while (height--) {
        bltScan(d,s X_ARG2(srcLeft,dstLeft),width);
        s += srcPitch;
        d += dstPitch;
        }
    DC.drawPixel = oldDrawPixel;
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPhysAddr - Physical address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_BitBltBM(
    void *srcAddr,
    N_int32 srcPhysAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    (void)srcPhysAddr;
    REF2D_BitBltSys(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,mix,false);
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 raster operation code for the copy (GA_ROP3CodesType)
****************************************************************************/
void NAPI REF2D_BitBltPatt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    int     yinc = (dstTop <= srcTop) ? 1 : -1,dstPitch;
    uchar   *s,*d;

    if (width > DC.maxBltWidth)
        REF2D_resizeBltBuffers(width);
    DC.drawPixelR3 = DC.drawPixelR3Table[rop3&0xFF];
    if (yinc < 0) {
        srcTop += height-1;
        dstTop += height-1;
        s = PIXEL_ADDR(srcLeft,srcTop);
        d = PIXEL_ADDR(dstLeft,dstTop);
        dstPitch = -DC.mi.bytesPerLine;
        }
    else {
        s = PIXEL_ADDR(srcLeft,srcTop);
        d = PIXEL_ADDR(dstLeft,dstTop);
        dstPitch = DC.mi.bytesPerLine;
        }
    DC.beginDirectAccess();
    if (DC.isHWBuffer) {
        while (height--) {
            DC.readScan(DC.scanBuf,s X_ARG2(srcLeft,dstLeft),width);
            DC.readScan(DC.scanBuf2,d X_ARG2(dstLeft,dstLeft),width);
            DC.bltScanPatt(DC.scanBuf2,DC.scanBuf X_ARG(dstLeft),dstLeft,dstTop,width);
            DC.bltScan_S(d,DC.scanBuf2 X_ARG2(dstLeft,dstLeft),width);
            s += dstPitch;
            d += dstPitch;
            dstTop += yinc;
            }
        }
    else {
        while (height--) {
            DC.bltScanPatt(d,s X_ARG(srcLeft),dstLeft,dstTop,width);
            s += dstPitch;
            d += dstPitch;
            dstTop += yinc;
            }
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_BitBltPattLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    uchar *s = (uchar*)DC.hwBufferStart + srcOfs + (srcPitch * srcTop) + X_ADDR(srcLeft);
    uchar *d = PIXEL_ADDR(dstLeft,dstTop);

    DC.drawPixelR3 = DC.drawPixelR3Table[rop3&0xFF];
    DC.beginDirectAccess();
    while (height--) {
        DC.bltScanPatt(d,s X_ARG(srcLeft),dstLeft,dstTop,width);
        s += srcPitch;
        d += DC.mi.bytesPerLine;
        dstTop++;
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - ROP3 raster operation code for the copy (GA_ROP3CodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI REF2D_BitBltPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    uchar   *s = ((uchar*)srcAddr + (srcPitch * srcTop)) + X_ADDR(srcLeft);
    uchar   *d = PIXEL_ADDR(dstLeft,dstTop);
    int     dstPitch,yinc;

    if (width > DC.maxBltWidth)
        REF2D_resizeBltBuffers(width);
    DC.drawPixelR3 = DC.drawPixelR3Table[rop3&0xFF];
    if (flipY) {
        d += (height-1) * DC.mi.bytesPerLine;
        dstTop += (height-1);
        dstPitch = -DC.mi.bytesPerLine;
        yinc = -1;
        }
    else {
        dstPitch = DC.mi.bytesPerLine;
        yinc = 1;
        }
    DC.beginDirectAccess();
    if (DC.isHWBuffer) {
        while (height--) {
            DC.readScan(DC.scanBuf,d X_ARG2(dstLeft,dstLeft),width);
            DC.bltScanPatt(DC.scanBuf,s X_ARG(srcLeft),dstLeft,dstTop,width);
            DC.bltScan_S(d,DC.scanBuf X_ARG2(dstLeft,dstLeft),width);
            s += srcPitch;
            d += dstPitch;
            dstTop += yinc;
            }
        }
    else {
        while (height--) {
            DC.bltScanPatt(d,s X_ARG(srcLeft),dstLeft,dstTop,width);
            s += srcPitch;
            d += dstPitch;
            dstTop += yinc;
            }
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPhysAddr - Physical address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_BitBltPattBM(
    void *srcAddr,
    N_int32 srcPhysAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    (void)srcPhysAddr;
    REF2D_BitBltPattSys(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,rop3,false);
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 raster operation code for the copy (GA_ROP3CodesType)
****************************************************************************/
void NAPI REF2D_BitBltColorPatt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    int     yinc = (dstTop <= srcTop) ? 1 : -1,dstPitch;
    uchar   *s,*d;

    if (width > DC.maxBltWidth)
        REF2D_resizeBltBuffers(width);
    DC.drawPixelR3 = DC.drawPixelR3Table[rop3&0xFF];
    if (yinc < 0) {
        srcTop += height-1;
        dstTop += height-1;
        s = PIXEL_ADDR(srcLeft,srcTop);
        d = PIXEL_ADDR(dstLeft,dstTop);
        dstPitch = -DC.mi.bytesPerLine;
        }
    else {
        s = PIXEL_ADDR(srcLeft,srcTop);
        d = PIXEL_ADDR(dstLeft,dstTop);
        dstPitch = DC.mi.bytesPerLine;
        }
    DC.beginDirectAccess();
    if (DC.isHWBuffer) {
        while (height--) {
            DC.readScan(DC.scanBuf,s X_ARG2(srcLeft,dstLeft),width);
            DC.readScan(DC.scanBuf2,d X_ARG2(dstLeft,dstLeft),width);
            DC.bltScanColorPatt(DC.scanBuf2,DC.scanBuf X_ARG(srcLeft),dstLeft,dstTop,width);
            DC.bltScan_S(d,DC.scanBuf2 X_ARG2(dstLeft,dstLeft),width);
            s += dstPitch;
            d += dstPitch;
            dstTop += yinc;
            }
        }
    else {
        while (height--) {
            DC.bltScanColorPatt(d,s X_ARG(srcLeft),dstLeft,dstTop,width);
            s += dstPitch;
            d += dstPitch;
            dstTop += yinc;
            }
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_BitBltColorPattLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    uchar *s = (uchar*)DC.hwBufferStart + srcOfs + (srcPitch * srcTop) + X_ADDR(srcLeft);
    uchar *d = PIXEL_ADDR(dstLeft,dstTop);

    DC.drawPixelR3 = DC.drawPixelR3Table[rop3&0xFF];
    DC.beginDirectAccess();
    while (height--) {
        DC.bltScanColorPatt(d,s X_ARG(srcLeft),dstLeft,dstTop,width);
        s += srcPitch;
        d += DC.mi.bytesPerLine;
        dstTop++;
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - ROP3 raster operation code for the copy (GA_ROP3CodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI REF2D_BitBltColorPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    uchar   *s = ((uchar*)srcAddr + (srcPitch * srcTop)) + X_ADDR(srcLeft);
    uchar   *d = PIXEL_ADDR(dstLeft,dstTop);
    int     dstPitch,yinc;

    if (width > DC.maxBltWidth)
        REF2D_resizeBltBuffers(width);
    DC.drawPixelR3 = DC.drawPixelR3Table[rop3&0xFF];
    if (flipY) {
        d += (height-1) * DC.mi.bytesPerLine;
        dstTop += (height-1);
        dstPitch = -DC.mi.bytesPerLine;
        yinc = -1;
        }
    else {
        dstPitch = DC.mi.bytesPerLine;
        yinc = 1;
        }
    DC.beginDirectAccess();
    if (DC.isHWBuffer) {
        while (height--) {
            DC.readScan(DC.scanBuf,d X_ARG2(dstLeft,dstLeft),width);
            DC.bltScanColorPatt(DC.scanBuf,s X_ARG(srcLeft),dstLeft,dstTop,width);
            DC.bltScan_S(d,DC.scanBuf X_ARG2(dstLeft,dstLeft),width);
            s += srcPitch;
            d += dstPitch;
            dstTop += yinc;
            }
        }
    else {
        while (height--) {
            DC.bltScanColorPatt(d,s X_ARG(srcLeft),dstLeft,dstTop,width);
            s += srcPitch;
            d += dstPitch;
            dstTop += yinc;
            }
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPhysAddr - Physical address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_BitBltColorPattBM(
    void *srcAddr,
    N_int32 srcPhysAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    (void)srcPhysAddr;
    REF2D_BitBltColorPattSys(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,rop3,false);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI REF2D_SrcTransBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    REF2D_SrcTransBltSys(DC.surface,DC.mi.bytesPerLine,srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent,false);
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI REF2D_SrcTransBltLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    REF2D_SrcTransBltSys(DC.hwBufferStart+srcOfs,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent,false);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI REF2D_SrcTransBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent,
    N_int32 flipY)
{
    uchar               *s = ((uchar*)srcAddr + (srcPitch * srcTop)) + X_ADDR(srcLeft);
    uchar               *d = PIXEL_ADDR(dstLeft,dstTop);
    int                 dstPitch;
    REF2D_drawPixel_t   oldDrawPixel = DC.drawPixel;
    bltScan_t           srcTransBltScan;

    // Determine if the mix is a solid mix, and setup appropriate pixel
    // function if not.
    if (DC.colorCmpMask != 0xFFFFFFFF) {
        // Special case only for software rendering to handle a
        // color compare mask.
        DC.drawPixel = GET_PIXEL_FUNC(mix);
        srcTransBltScan = DC.srcTransBltScan_Masked;
        }
    else if (IS_SOLID_MIX(mix))
        srcTransBltScan = DC.srcTransBltScan_S;
    else {
        DC.drawPixel = GET_PIXEL_FUNC(mix);
        srcTransBltScan = DC.srcTransBltScan_Generic;
        }
    if (flipY) {
        d += (height-1) * DC.mi.bytesPerLine;
        dstPitch = -DC.mi.bytesPerLine;
        }
    else
        dstPitch = DC.mi.bytesPerLine;
    DC.transparent = transparent & DC.colorMask & DC.colorCmpMask;
    DC.beginDirectAccess();
    while (height--) {
        srcTransBltScan(d,s X_ARG2(srcLeft,dstLeft),width);
        s += srcPitch;
        d += dstPitch;
        }
    DC.drawPixel = oldDrawPixel;
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPhysAddr - Physical address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI REF2D_SrcTransBltBM(
    void *srcAddr,
    N_int32 srcPhysAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    (void)srcPhysAddr;
    REF2D_SrcTransBltSys(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent,false);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI REF2D_DstTransBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    REF2D_DstTransBltSys(DC.surface,DC.mi.bytesPerLine,srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent,false);
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI REF2D_DstTransBltLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    REF2D_DstTransBltSys(DC.hwBufferStart+srcOfs,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent,false);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI REF2D_DstTransBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent,
    N_int32 flipY)
{
    uchar               *s = ((uchar*)srcAddr + (srcPitch * srcTop)) + X_ADDR(srcLeft);
    uchar               *d = PIXEL_ADDR(dstLeft,dstTop);
    int                 dstPitch;
    REF2D_drawPixel_t   oldDrawPixel = DC.drawPixel;
    bltScan_t           dstTransBltScan;

    // Determine if the mix is a solid mix, and setup appropriate pixel
    // function if not.
    if (DC.colorCmpMask != 0xFFFFFFFF) {
        // Special case only for software rendering to handle a
        // color compare mask.
        DC.drawPixel = GET_PIXEL_FUNC(mix);
        dstTransBltScan = DC.dstTransBltScan_Masked;
        }
    else if (IS_SOLID_MIX(mix))
        dstTransBltScan = DC.dstTransBltScan_S;
    else {
        DC.drawPixel = GET_PIXEL_FUNC(mix);
        dstTransBltScan = DC.dstTransBltScan_Generic;
        }
    if (flipY) {
        d += (height-1) * DC.mi.bytesPerLine;
        dstPitch = -DC.mi.bytesPerLine;
        }
    else
        dstPitch = DC.mi.bytesPerLine;
    DC.transparent = transparent & DC.colorMask & DC.colorCmpMask;
    DC.beginDirectAccess();
    while (height--) {
        dstTransBltScan(d,s X_ARG2(srcLeft,dstLeft),width);
        s += srcPitch;
        d += dstPitch;
        }
    DC.drawPixel = oldDrawPixel;
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPhysAddr - Physical address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI REF2D_DstTransBltBM(
    void *srcAddr,
    N_int32 srcPhysAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    (void)srcPhysAddr;
    REF2D_DstTransBltSys(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent,false);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
doClip      - True if the blit should be clipped
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle
mix         - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_StretchBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 mix)
{
    REF2D_StretchBltSys(DC.surface,DC.mi.bytesPerLine,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,doClip,clipLeft,clipTop,clipRight,clipBottom,mix,false);
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
doClip      - True if the blit should be clipped
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle
mix         - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_StretchBltLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 mix)
{
    REF2D_StretchBltSys(DC.hwBufferStart+srcOfs,srcPitch,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,doClip,clipLeft,clipTop,clipRight,clipBottom,mix,false);
}

/****************************************************************************
PARAMETERS:
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
doClip      - True if the blit should be clipped
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle
zoomx       - Place to store computed X zoom factor
zoomy       - Place to store computed Y zoom factor
clippedx    - Contains true if clipped in X coordinate, false otherwise
clippedy    - Contains true if clipped in Y coordinate, false otherwise
clipped     - Contains true if clipped, false otherwise

REMARKS:
Function to pre-clip and setup values before performing a stretch blit
operation.
****************************************************************************/
static ibool PreClipStretchBlit(
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_fix32 *zoomx,
    N_fix32 *zoomy,
    ibool *clippedx,
    ibool *clippedy,
    ibool *clipped)
{
    int dstRight = (dstLeft + dstWidth), dstBottom = (dstTop + dstHeight);

    // Determine if we need to clip the image based on input coordinates
    *clippedx = *clippedy = *clipped = false;
    if (doClip) {
        // Check for trivial rejection
        if (dstLeft >= clipRight || dstRight <= clipLeft || dstTop >= clipBottom || dstBottom <= clipTop)
            return false;

        // Check for X coordinate clipping
        if (dstLeft < clipLeft || dstRight > clipRight)
            *clipped = *clippedx = true;

        // Check for Y coordinate clipping
        if (dstTop < clipTop || dstBottom > clipBottom)
            *clipped = *clippedy = true;
        }

    // Ensure the blit is not too wide for our software rasteriser
    if (MAX(srcWidth,dstWidth) > DC.maxBltWidth)
        REF2D_resizeBltBuffers(MAX(srcWidth,dstWidth));

    // Calculate the x zoom factor
    if (dstWidth == srcWidth)
        *zoomx = N_FIX_1;
    else if (dstWidth == (srcWidth * 2))
        *zoomx = N_FIX_2;
    else
        *zoomx = FXdiv(N_TOFIX(dstWidth),N_TOFIX(srcWidth));

    // Calculate the y zoom factor
    if (dstHeight == srcHeight)
        *zoomy = N_FIX_1;
    else if (dstHeight == (srcHeight * 2))
        *zoomy = N_FIX_2;
    else
        *zoomy = FXdiv(N_TOFIX(dstHeight),N_TOFIX(srcHeight));
    return true;
}

/****************************************************************************
PARAMETERS:
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstHeight   - Height of the destination rectangle in scanlines
mix         - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
isSolidMix  - Returns true if the mix is solid
dstPitch    - Returns the pitch increment for destination buffer
y           - Returns the starting destination Y coordinate
yinc        - Returns the Y increment value for each scanline

REMARKS:
Function to setup for rendering a system blit function by determining the
destination buffer address and the values used to traverse the buffer
memory based on the blit vertical direction.
****************************************************************************/
uchar *SetupForSysBlt(
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstHeight,
    N_int32 mix,
    N_int32 flipY,
    ibool *isSolidMix,
    int *dstPitch,
    N_int32 *y,
    int *yinc)
{
    uchar *d;

    // Determine if the mix is a solid mix, and setup appropriate pixel
    // function if not.
    DC.drawPixel = GET_PIXEL_FUNC(mix);
    if (IS_SOLID_MIX(mix))
        *isSolidMix = true;
    else {
        *isSolidMix = false;
        }

    // Determine starting pixel coordinates and blit vertical direction
    d = PIXEL_ADDR(dstLeft,dstTop);
    if (flipY) {
        d += (dstHeight-1) * DC.mi.bytesPerLine;
        *dstPitch = -DC.mi.bytesPerLine;
        *y = (dstTop + dstHeight) - 1;
        *yinc = -1;
        }
    else {
        *dstPitch = DC.mi.bytesPerLine;
        *y = dstTop;
        *yinc = 1;
        }
    return d;
}

/****************************************************************************
PARAMETERS:
zoomx       - Place to store computed X zoom factor
zoomy       - Place to store computed Y zoom factor
psrcLeft    - Source pixel coordinate for bitmap (modified)
psrcWidth   - Width of the source rectangle in pixels (modified)
pdstLeft    - Left coordinate of destination (modified)
clipLeft    - Left coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clippedx    - Contains true if clipped in X coordinate, false otherwise

REMARKS:
Function to setup and pre-render the pixel replication count array to be
used for stretch blit code. Properly handles X coordinate clipping for the
pixel replication count array. The pixel replication array is computed
into DC.repCounts.
****************************************************************************/
int *SetupPixelReplicationCounts(
    N_fix32 zoomx,
    N_fix32 zoomy,
    N_int32 *psrcLeft,
    N_int32 *psrcWidth,
    N_int32 *pdstLeft,
    N_int32 clipLeft,
    N_int32 clipRight,
    ibool clippedx)
{
    N_fix32 fx;
    int     srcLeft,srcWidth,dstLeft;
    int     x,next,prev,count;
    int     *rc;

    // Pre-render the pixel replication counts with clipping. These
    // values are identical for each scanline so we save time by
    // doing this one for all scanlines.
    fx = zoomx;
    prev = 0;
    count = *psrcWidth;
    rc = DC.repCounts;
    if (clippedx) {
        // Obtain initial values from passed in pointers
        srcLeft = *psrcLeft;
        srcWidth = *psrcWidth;
        dstLeft = *pdstLeft;

        // Clip left edge up to initial left clip coordinate
        x = dstLeft;
        while ((x < clipLeft) && count--) {
            if ((x += (next = N_FIXROUND(fx)) - prev) >= clipLeft) {
                *rc++ = (x - clipLeft);
                srcLeft += (srcWidth-count-1);
                srcWidth = count+1;
                dstLeft = clipLeft;
                }
            else
                dstLeft = x;
            prev = next;
            fx += zoomx;
            }

        // Now generate pixel replication counts clipped to right edge
        while (count-- && (x < clipRight)) {
            x += (*rc++ = ((next = N_FIXROUND(fx)) - prev));
            if (x > clipRight)
                *(rc-1) -= (x - clipRight);
            prev = next;
            fx += zoomx;
            }
        srcWidth -= (count+1);

        // Return values via passed in pointers
        *psrcLeft = srcLeft;
        *psrcWidth = srcWidth;
        *pdstLeft = dstLeft;
        }
    else {
        // Generate pixel replication counts without clipping
        while (count--) {
            *rc++ = (next = N_FIXROUND(fx)) - prev;
            prev = next;
            fx += zoomx;
            }
        }
    return DC.repCounts;
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
doClip      - True if the blit should be clipped
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle
mix         - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI REF2D_StretchBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 mix,
    N_int32 flipY)
{
    N_fix32             zoomx,zoomy,fx;
    ibool               clippedx,clippedy,clipped,isSolidMix = IS_SOLID_MIX(mix);
    int                 next,prev,dstPitch,yinc;
    uchar               *s,*d;
    REF2D_drawPixel_t   oldDrawPixel = DC.drawPixel;

    // Do pre-clipping and setup for stretch blit operation
    if (!PreClipStretchBlit(
            srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,
            doClip,clipLeft,clipTop,clipRight,clipBottom,
            &zoomx,&zoomy,&clippedx,&clippedy,&clipped))
        return;

    // Handle special case of 1:1 stretch blit (with clipping)
    if (zoomx == N_FIX_1 && zoomy == N_FIX_1) {
        if (!clipped)
            REF2D_BitBltSys(srcAddr,srcPitch,srcLeft,srcTop,dstWidth,dstHeight,dstLeft,dstTop,mix,flipY);
        else {
            int dstLeft2 = MAX(dstLeft, clipLeft);
            int dstTop2 = MAX(dstTop, clipTop);
            int dstRight = MIN((dstLeft + dstWidth), clipRight);
            int dstBottom = MIN((dstTop + dstHeight), clipBottom);

            REF2D_BitBltSys(srcAddr,srcPitch,
                srcLeft+(dstLeft2-dstLeft), srcTop+(dstTop2-dstTop),
                dstRight-dstLeft2,dstBottom-dstTop2,
                dstLeft2,dstTop2,mix,flipY);
            }
        return;
        }

    if (!clipped && zoomx == N_FIX_1 && zoomy == N_FIX_2) {
        // Handle special case of 1x2 stretch blit (no clipping!). We do this
        // by doing a regular blit operation but with duplication of each
        // scanline on the destination for a 2x stretch in Y.
        bltScan_t bltScan = isSolidMix ? DC.bltScan_S : DC.bltScan_Generic;

        // Determine starting pixel coordinates and setup for blit operation
        s = ((uchar*)srcAddr + (srcPitch * srcTop)) + X_ADDR(srcLeft);
        d = SetupForSysBlt(dstLeft,dstTop,dstHeight,mix,flipY,&isSolidMix,&dstPitch,&dstTop,&yinc);
        DC.beginDirectAccess();
        while (srcHeight--) {
            bltScan(d,s X_ARG2(srcLeft,dstLeft),dstWidth);
            d += dstPitch;
            bltScan(d,s X_ARG2(srcLeft,dstLeft),dstWidth);
            d += dstPitch;
            s += srcPitch;
            }
        DC.endDirectAccess();
        }
    else if (!clipped && zoomx == N_FIX_2 && zoomy == N_FIX_2) {
        // Handle special case of 2x2 stretch blit (no clipping!). We do this
        // by duplicating each line vertically, and using a high performance
        // 2x stretch blitter function for each scanline.
        void (*stretchScan2x)(uchar *d,uchar *s DCL_X_ARG2,int width);
        stretchScan2x = isSolidMix ? DC.stretchScan2x_S : DC.stretchScan2x_Generic;

        // Determine starting pixel coordinates and setup for blit operation
        s = ((uchar*)srcAddr + (srcPitch * srcTop)) + X_ADDR(srcLeft);
        d = SetupForSysBlt(dstLeft,dstTop,dstHeight,mix,flipY,&isSolidMix,&dstPitch,&dstTop,&yinc);
        DC.beginDirectAccess();
        while (srcHeight--) {
            stretchScan2x(d,s X_ARG2(srcLeft,dstLeft),srcWidth);
            d += dstPitch;
            stretchScan2x(d,s X_ARG2(srcLeft,dstLeft),srcWidth);
            d += dstPitch;
            s += srcPitch;
            }
        DC.endDirectAccess();
        }
    else {
        // Handle generic stretch blit (with optional clipping)
        stretchScan_t stretchScan = isSolidMix ? DC.stretchScan_S : DC.stretchScan_Generic;

        // Pre-render the pixel replication counts with clipping
        SetupPixelReplicationCounts(zoomx,zoomy,&srcLeft,&srcWidth,&dstLeft,clipLeft,clipRight,clippedx);

        // Determine starting pixel coordinates and setup for blit operation
        s = ((uchar*)srcAddr + (srcPitch * srcTop)) + X_ADDR(srcLeft);
        d = SetupForSysBlt(dstLeft,dstTop,dstHeight,mix,flipY,&isSolidMix,&dstPitch,&dstTop,&yinc);

        // Now perform the stretch blit
        fx = zoomy;
        prev = 0;
        DC.beginDirectAccess();
        while (srcHeight--) {
            int dy = (next = N_FIXROUND(fx)) - prev;
            prev = next;
            fx += zoomy;
            while (dy--) {
                if (!clippedy || (dstTop >= clipTop && dstTop < clipBottom))
                    stretchScan(d,s X_ARG2(srcLeft,dstLeft),srcWidth,DC.repCounts);
                d += dstPitch;
                dstTop += yinc;
                }
            s += srcPitch;
            }
        DC.endDirectAccess();
        }

    // Restore pixel drawing function is modified internally
    DC.drawPixel = oldDrawPixel;
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPhysAddr - Physical address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
doClip      - True if the blit should be clipped
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle
mix         - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_StretchBltBM(
    void *srcAddr,
    N_int32 srcPhysAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 mix)
{
    (void)srcPhysAddr;
    REF2D_StretchBltSys(srcAddr,srcPitch,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,doClip,clipLeft,clipTop,clipRight,clipBottom,mix,false);
}

/****************************************************************************
PARAMETERS:
srcAddr         - Address of source bitmap in system memory
srcPitch        - Pitch of source bitmap in bytes
srcLeft         - Left coordinate within source bitmap to copy
srcTop          - Top coordinate within source bitmap to copy
width           - Width of the rectangle in pixels
height          - Height of the rectangle in scanlines
dstLeft         - Left coordinate of destination
dstTop          - Top coordinate of destination
bitsPerPixel    - Pixel depth for the bitmap to copy
pixelFormat     - Pixel format for the bitmap to copy
dstPal          - Color palette for the destination device (if dest is 8bpp)
srcPal          - Color palette for the bitmap to copy (if source is 8bpp)
dither          - True if the output should be dithered
mix             - Mix code for the copy (GA_mixCodesType)
flipY           - True if the bitmap should be flipped in Y

RETURNS:
0 on success, -1 on failure.
****************************************************************************/
N_int32 NAPI REF2D_ConvertBltSys_Obsolete(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pixelFormat,
    GA_palette *dstPal,
    GA_palette *srcPal,
    N_int32 dither,
    N_int32 mix,
    N_int32 flipY)
{
    GA_bltFx    fx;

    // This function has been depracted and replaced by BitBltFx. Hence to ensure
    // existing code still runs, this function calls BitBltFx() internally. New
    // code should always call BitBltFx for better performance.
    fx.dwSize = sizeof(fx);
    fx.Flags = gaBltConvert | gaBltMixEnable;
    if (flipY)
        fx.Flags |= gaBltFlipY;
    if (dither == -1)
        fx.Flags |= gaBltTranslateVec | gaBltDither;
    else if (dither)
        fx.Flags |= gaBltDither;
    fx.Mix = mix;
    fx.BitsPerPixel = bitsPerPixel;
    fx.PixelFormat = pixelFormat;
    fx.DstPalette = dstPal;
    fx.SrcPalette = srcPal;
    fx.TranslateVec = (GA_color*)dstPal;
    REF2D_BitBltFxSys(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,width,height,&fx);
    return 0;
}

/****************************************************************************
PARAMETERS:
srcAddr         - Address of source bitmap in system memory
srcPhysAddr     - Physical address of source bitmap in system memory
srcPitch        - Pitch of source bitmap in bytes
srcLeft         - Left coordinate within source bitmap to copy
srcTop          - Top coordinate within source bitmap to copy
width           - Width of the rectangle in pixels
height          - Height of the rectangle in scanlines
dstLeft         - Left coordinate of destination
dstTop          - Top coordinate of destination
bitsPerPixel    - Pixel depth for the bitmap to copy
pixelFormat     - Pixel format for the bitmap to copy
dstPal          - Color palette for the destination device (if dest is 8bpp)
srcPal          - Color palette for the bitmap to copy (if source is 8bpp)
dither          - True if the output should be dithered
mix             - Mix code for the copy (GA_mixCodesType)

RETURNS:
0 on success, -1 on failure.
****************************************************************************/
N_int32 NAPI REF2D_ConvertBltBM_Obsolete(
    void *srcAddr,
    N_int32 srcPhysAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pixelFormat,
    GA_palette *dstPal,
    GA_palette *srcPal,
    N_int32 dither,
    N_int32 mix)
{
    (void)srcPhysAddr;
    return REF2D_ConvertBltSys_Obsolete(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,bitsPerPixel,pixelFormat,dstPal,srcPal,dither,mix,false);
}

/****************************************************************************
PARAMETERS:
srcAddr         - Address of source bitmap in system memory
srcPitch        - Pitch of source bitmap in bytes
srcLeft         - Left coordinate of the source rectangle to copy
srcTop          - Top coordinate of the source rectangle to copy
srcWidth        - Width of the source rectangle in pixels
srcHeight       - Height of the source rectangle in scanlines
dstLeft         - Left coordinate of destination
dstTop          - Top coordinate of destination
dstWidth        - Width of the destination rectangle in pixels
dstHeight       - Height of the destination rectangle in scanlines
doClip          - True if the blit should be clipped
clipLeft        - Left coordinate of clip rectangle
clipTop         - Top coordinate of clip rectangle
clipRight       - Right coordinate of clip rectangle
clipBottom      - Bottom coordinate of clip rectangle
bitsPerPixel    - Pixel depth for the bitmap to copy
pixelFormat     - Pixel format for the bitmap to copy
dstPal          - Color palette for the destination device (if dest is 8bpp)
srcPal          - Color palette for the bitmap to copy (if source is 8bpp)
dither          - True if the output should be dithered
mix             - Mix code for the copy (GA_mixCodesType)
flipY           - True if the bitmap should be flipped in Y

RETURNS:
0 on success, -1 on failure.
****************************************************************************/
N_int32 NAPI REF2D_StretchConvertBltSys_Obsolete(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pixelFormat,
    GA_palette *dstPal,
    GA_palette *srcPal,
    N_int32 dither,
    N_int32 mix,
    N_int32 flipY)
{
    GA_bltFx    fx;

    // This function has been depracted and replaced by BitBltFx. Hence to ensure
    // existing code still runs, this function calls BitBltFx() internally. New
    // code should always call BitBltFx for better performance.
    fx.dwSize = sizeof(fx);
    fx.Flags = gaBltConvert | gaBltMixEnable | gaBltStretchNearest;
    if (flipY)
        fx.Flags |= gaBltFlipY;
    if (dither == -1)
        fx.Flags |= gaBltTranslateVec | gaBltDither;
    else if (dither)
        fx.Flags |= gaBltDither;
    if (doClip) {
        fx.Flags |= gaBltClip;
        fx.ClipLeft = clipLeft;
        fx.ClipTop = clipTop;
        fx.ClipRight = clipRight;
        fx.ClipBottom = clipBottom;
        }
    fx.Mix = mix;
    fx.BitsPerPixel = bitsPerPixel;
    fx.PixelFormat = pixelFormat;
    fx.DstPalette = dstPal;
    fx.SrcPalette = srcPal;
    fx.TranslateVec = (GA_color*)dstPal;
    REF2D_BitBltFxSys(srcAddr,srcPitch,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,&fx);
    return 0;
}

/****************************************************************************
PARAMETERS:
srcAddr         - Address of source bitmap in system memory
srcPhysAddr     - Physical address of source bitmap in system memory
srcPitch        - Pitch of source bitmap in bytes
srcLeft         - Left coordinate of the source rectangle to copy
srcTop          - Top coordinate of the source rectangle to copy
srcWidth        - Width of the source rectangle in pixels
srcHeight       - Height of the source rectangle in scanlines
dstLeft         - Left coordinate of destination
dstTop          - Top coordinate of destination
dstWidth        - Width of the destination rectangle in pixels
dstHeight       - Height of the destination rectangle in scanlines
doClip          - True if the blit should be clipped
clipLeft        - Left coordinate of clip rectangle
clipTop         - Top coordinate of clip rectangle
clipRight       - Right coordinate of clip rectangle
clipBottom      - Bottom coordinate of clip rectangle
bitsPerPixel    - Pixel depth for the bitmap to copy
pixelFormat     - Pixel format for the bitmap to copy
dstPal          - Color palette for the destination device (if dest is 8bpp)
srcPal          - Color palette for the bitmap to copy (if source is 8bpp)
dither          - True if the output should be dithered
mix             - Mix code for the copy (GA_mixCodesType)

RETURNS:
0 on success, -1 on failure.
****************************************************************************/
N_int32 NAPI REF2D_StretchConvertBltBM_Obsolete(
    void *srcAddr,
    N_int32 srcPhysAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pixelFormat,
    GA_palette *dstPal,
    GA_palette *srcPal,
    N_int32 dither,
    N_int32 mix)
{
    (void)srcPhysAddr;
    return REF2D_StretchConvertBltSys_Obsolete(srcAddr,srcPitch,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,doClip,clipLeft,clipTop,clipRight,clipBottom,bitsPerPixel,pixelFormat,dstPal,srcPal,dither,mix,false);
}

/****************************************************************************
PARAMETERS:
fx      - GA_bltFx structure to check support for

RETURNS:
1 if the features are supported, 0 if not.
****************************************************************************/
N_int32 NAPI REF2D_BitBltFxTest(
    GA_bltFx *fx)
{
    // We only support this for destination color depths of 8bpp and greater!
    if (DC.mi.bitsPerPixel < 8)
        return false;

    // Blending is only available in 15bpp and higher color depths
    if (fx->Flags & gaBltBlend) {
        if (DC.mi.bitsPerPixel < 15)
            return false;
        }

    // Check for unsupported color conversion operations. We only support
    // 8bpp and higher color depths for source bitmaps.
    if (fx->Flags & gaBltConvert) {
        if (fx->BitsPerPixel < 8)
            return false;
        }

    // We don't presently support stretching for effects blits
    if (fx->Flags & gaBltAnyStretch)
        return false;

    // We don't presently support clipping for effects blits (only needed for stretching)
    if (fx->Flags & gaBltClip)
        return false;

    // We support this type of blit!
    return true;
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
fx          - GA_bltFx structure describing the requested effects
****************************************************************************/
void NAPI REF2D_BitBltFx(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    GA_bltFx *fx)
{
    REF2D_BitBltFxSys(DC.surface,DC.mi.bytesPerLine,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
fx          - GA_bltFx structure describing the requested effects
****************************************************************************/
void NAPI REF2D_BitBltFxLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    GA_bltFx *fx)
{
    REF2D_BitBltFxSys(DC.hwBufferStart+srcOfs,srcPitch,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
}

/****************************************************************************
REMARKS:
Function to convert 8bit pixels via the color translation vector. Used
for all color depths.
****************************************************************************/
static GA_color REF2D_convertPixel8(
    uchar *addr DCL_X_ARG)
{
    return DC.translate[*((uchar*)addr)];
}

/****************************************************************************
REMARKS:
Function to convert 8bit pixels via the color translation vector. Used
for all color depths. This verision handles stored the second alpha byte
in the global alpha value variable for source alpha bitmaps.
****************************************************************************/
static GA_color REF2D_convertPixel8_Alpha(
    uchar *addr DCL_X_ARG)
{
    DC.alphaValue = ((uchar*)addr)[1];
    return DC.translate[((uchar*)addr)[0]];
}

/****************************************************************************
REMARKS:
Function to convert 8bit pixels via the color translation vector. Used
for all color depths. This verision handles stored the second alpha byte
in the global alpha value variable for source alpha bitmaps.
****************************************************************************/
static GA_color REF2D_convertPixel8_Alpha_32Axxx(
    uchar *addr DCL_X_ARG)
{
    return DC.translate[((uchar*)addr)[0]] | (((ulong)((uchar*)addr)[1]) << 24);
}

/****************************************************************************
REMARKS:
Function to convert 8bit pixels via the color translation vector. Used
for all color depths. This verision handles stored the second alpha byte
in the global alpha value variable for source alpha bitmaps.
****************************************************************************/
static GA_color REF2D_convertPixel8_Alpha_32xxxA(
    uchar *addr DCL_X_ARG)
{
    return DC.translate[((uchar*)addr)[0]] | (((ulong)((uchar*)addr)[1]) << 0);
}

/****************************************************************************
REMARKS:
Function to pre-expand 4bpp scanlines to 8bpp data to simplify inner loops.
****************************************************************************/
static void REF2D_expandScan4_8(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width)
{
    if (DC.srcX & 1) {
        // Handle first odd pixel
        *d++ = (*s++ & 0xF);
        width--;
        }
    while (width--) {
        *d++ = (*s >> 4);
        if (width-- == 0) break;
        *d++ = (*s++ & 0xF);
        }
}

/****************************************************************************
PARAMETERS:
fx          - GA_bltFx structure describing the requested effects

RETURNS:
If we are drawing an 8-bit alpha bitmap or we are drawing a 32-bit
bitmap to a 15-bit or 16-bit destination, then the pixel reading
function will store the pixel alpha value in the DC.alphaValue
variable. This is done as there is no other way to pass a full
8-bit alpha value from the source pixel to the destination blending
functions. So to get the DC.alphaValue value used as though it is
the source alpha value, we convert the blending functions as
necessary to use constant alpha instead.
****************************************************************************/
static ibool AdjustBlendFunc(
    N_int32 *srcBlendFunc,
    N_int32 *dstBlendFunc)
{
    N_int32 oldSrcBlendFunc = *srcBlendFunc;
    N_int32 oldDstBlendFunc = *dstBlendFunc;

    // Adjust source blending function
    if (*srcBlendFunc == gaBlendSrcAlphaFast)
        *srcBlendFunc = gaBlendConstantAlphaFast;
    else if (*srcBlendFunc == gaBlendSrcAlpha)
        *srcBlendFunc = gaBlendConstantAlpha;
    else if (*srcBlendFunc == gaBlendOneMinusSrcAlpha)
        *srcBlendFunc = gaBlendOneMinusConstantAlpha;

    // Adjust destination blending function
    if (*dstBlendFunc == gaBlendSrcAlphaFast)
        *dstBlendFunc = gaBlendConstantAlphaFast;
    else if (*dstBlendFunc == gaBlendSrcAlpha)
        *dstBlendFunc = gaBlendConstantAlpha;
    else if (*dstBlendFunc == gaBlendOneMinusSrcAlpha)
        *dstBlendFunc = gaBlendOneMinusConstantAlpha;

    // Return true if the blending functions were adjusted
    return (oldSrcBlendFunc != *srcBlendFunc) || (oldDstBlendFunc != *dstBlendFunc);
}

/****************************************************************************
PARAMETERS:
fx          - GA_bltFx structure describing the requested effects

RETURNS:
Pointer to the pixel reading function to handle the color conversion
operation.
****************************************************************************/
static REF2D_readPixel_t REF2D_InitPixelReader(
    GA_bltFx *fx,
    N_int32 *srcBlendFunc,
    N_int32 *dstBlendFunc,
    bufReadScan_t *readScan,
    ibool *preReadBuffer,
    REF2D_setupDither_t *setupDither)
{
    REF2D_readPixel_t   readPixel;

    switch (fx->BitsPerPixel) {
        case 4:
            // Handle color converting 4bpp bitmaps to other pixel formats.
            // When converting 4bpp source bitmaps, we pre-expand to 8bpp
            // with a fast scanline function before processing the pixels
            // as 8bpp values in the inner loops. This simplifies the inner
            // loop pixel processing considerably.
            if (fx->Flags & gaBltTranslateVec) {
                // Pixel translation vector is pre-computed
                DC.translate = fx->TranslateVec;
                }
            else {
                // Compute translation vector between source and
                // destination color palettes.
                DC.translate = DC.buildTranslateVector(DC.translateVec,fx->DstPalette,fx->SrcPalette,16);
                }
            readPixel = REF2D_convertPixel8;
            *readScan = REF2D_expandScan4_8;
            *preReadBuffer = true;
            break;
        case 8:
            // Handle color converting 8bpp bitmaps to other pixel formats.
            if (fx->Flags & gaBltTranslateVec) {
                // Pixel translation vector is pre-computed
                DC.translate = fx->TranslateVec;
                }
            else {
                // Compute translation vector between source and
                // destination color palettes.
                DC.translate = DC.buildTranslateVector(DC.translateVec,fx->DstPalette,fx->SrcPalette,256);
                }
            readPixel = REF2D_convertPixel8;
            break;
        case 15:
            // Handle color converting 15bpp bitmaps to other
            // pixel formats.
            switch (DC.mi.bitsPerPixel) {
#ifdef SUPPORT_8BPP
                case 8:
                    DC.cntPal = fx->DstPalette;
                    if (fx->Flags & gaBltDither) {
                        // Setup for real time dithering
                        readPixel = REF2D_convertPixel15_8_Dithered;
                        if (fx->Flags & gaBltTranslateVec)
                            _REF2D_ditherVec = fx->TranslateVec;
                        else
                            _REF2D_ditherVec = DC.buildTranslateVector(DC.translateVec,fx->DstPalette,_REF2D_halftonePal,256);
                        *setupDither = REF2D_setupDither_8;
                        }
                    else {
                        // Setup for closest color matching (slow)
                        readPixel = REF2D_convertPixel16_8;
                        DC.findClosest = REF2D_findClosest_15;
                        }
                    break;
#endif
#ifdef SUPPORT_16BPP
                case 15:
                    readPixel = DC.readPixel;
                    break;
                case 16:
                    readPixel = REF2D_convertPixel15_16;
                    break;
#endif
#ifdef SUPPORT_24BPP
                case 24:
                    if (DC.mi.pf.RedPosition == 0)
                        readPixel = REF2D_convertPixel15_24BGR;
                    else
                        readPixel = REF2D_convertPixel15_24RGB;
                    break;
#endif
#ifdef SUPPORT_32BPP
                case 32:
                    if (fx->Flags & gaBltBlend) {
                        if (DC.mi.pf.RedPosition == 0)
                            readPixel = REF2D_convertPixel15_32ABGR_Blend;
                        else if (DC.mi.pf.RedPosition == 8)
                            readPixel = REF2D_convertPixel15_32BGRA_Blend;
                        else if (DC.mi.pf.RedPosition == 16)
                            readPixel = REF2D_convertPixel15_32ARGB_Blend;
                        else
                            readPixel = REF2D_convertPixel15_32RGBA_Blend;
                        }
                    else {
                        if (DC.mi.pf.RedPosition == 0)
                            readPixel = REF2D_convertPixel15_32ABGR;
                        else if (DC.mi.pf.RedPosition == 8)
                            readPixel = REF2D_convertPixel15_32BGRA;
                        else if (DC.mi.pf.RedPosition == 16)
                            readPixel = REF2D_convertPixel15_32ARGB;
                        else
                            readPixel = REF2D_convertPixel15_32RGBA;
                        }
                    break;
#endif
                }
            break;
        case 16:
            if (fx->PixelFormat->AlphaPosition == 8) {
                // Special case of 16bpp bitmaps that are actually an
                // 8-bit color index bitmap with an 8-bit alpha channel
                // stores as a single 16-bit pixel.
                if (DC.mi.bitsPerPixel == 32) {
                    if (DC.mi.pf.AlphaPosition == 0)
                        readPixel = REF2D_convertPixel8_Alpha_32xxxA;
                    else
                        readPixel = REF2D_convertPixel8_Alpha_32Axxx;
                    }
                else if ((fx->Flags & gaBltBlend) && AdjustBlendFunc(srcBlendFunc,dstBlendFunc))
                    readPixel = REF2D_convertPixel8_Alpha;
                else
                    readPixel = REF2D_convertPixel8;
                if (fx->Flags & gaBltTranslateVec)
                    DC.translate = fx->TranslateVec;
                else
                    DC.translate = DC.buildTranslateVector(DC.translateVec,fx->DstPalette,fx->SrcPalette,256);
                }
            else {
                // Handle color converting 16bpp bitmaps to other
                // pixel formats.
                switch (DC.mi.bitsPerPixel) {
#ifdef SUPPORT_8BPP
                    case 8:
                        DC.cntPal = fx->DstPalette;
                        if (fx->Flags & gaBltDither) {
                            // Setup for real time dithering
                            readPixel = REF2D_convertPixel16_8_Dithered;
                            if (fx->Flags & gaBltTranslateVec)
                                _REF2D_ditherVec = fx->TranslateVec;
                            else
                                _REF2D_ditherVec = DC.buildTranslateVector(DC.translateVec,fx->DstPalette,_REF2D_halftonePal,256);
                            *setupDither = REF2D_setupDither_8;
                            }
                        else {
                            // Setup for closest color matching (slow)
                            readPixel = REF2D_convertPixel16_8;
                            DC.findClosest = REF2D_findClosest_16;
                            }
                        break;
#endif
#ifdef SUPPORT_16BPP
                    case 15:
                        readPixel = REF2D_convertPixel16_15;
                        break;
                    case 16:
                        readPixel = DC.readPixel;
                        break;
#endif
#ifdef SUPPORT_24BPP
                    case 24:
                        if (DC.mi.pf.RedPosition == 0)
                            readPixel = REF2D_convertPixel16_24BGR;
                        else
                            readPixel = REF2D_convertPixel16_24RGB;
                        break;
#endif
#ifdef SUPPORT_32BPP
                    case 32:
                        if (DC.mi.pf.RedPosition == 0)
                            readPixel = REF2D_convertPixel16_32ABGR;
                        else if (DC.mi.pf.RedPosition == 8)
                            readPixel = REF2D_convertPixel16_32BGRA;
                        else if (DC.mi.pf.RedPosition == 16)
                            readPixel = REF2D_convertPixel16_32ARGB;
                        else
                            readPixel = REF2D_convertPixel16_32RGBA;
                        break;
#endif
                    }
                }
            break;
        case 24:
            // Handle color converting 24bpp bitmaps to other
            // pixel formats.
            switch (DC.mi.bitsPerPixel) {
#ifdef SUPPORT_8BPP
                case 8:
                    DC.cntPal = fx->DstPalette;
                    if (fx->Flags & gaBltDither) {
                        // Setup for real time dithering
                        if (fx->PixelFormat->RedPosition == 0)
                            readPixel = REF2D_convertPixel24BGR_8_Dithered;
                        else
                            readPixel = REF2D_convertPixel24RGB_8_Dithered;
                        if (fx->Flags & gaBltTranslateVec)
                            _REF2D_ditherVec = fx->TranslateVec;
                        else
                            _REF2D_ditherVec = DC.buildTranslateVector(DC.translateVec,fx->DstPalette,_REF2D_halftonePal,256);
                        *setupDither = REF2D_setupDither_8;
                        }
                    else {
                        // Setup for closest color matching (slow)
                        readPixel = REF2D_convertPixel24_8;
                        if (fx->PixelFormat->RedPosition == 0)
                            DC.findClosest = REF2D_findClosest_24BGR;
                        else
                            DC.findClosest = REF2D_findClosest_24RGB;
                        }
                    break;
#endif
#ifdef SUPPORT_16BPP
                case 15:
                    if (fx->Flags & gaBltDither) {
                        // Setup for real time dithering
                        if (fx->PixelFormat->RedPosition == 0)
                            readPixel = REF2D_convertPixel24BGR_15_Dithered;
                        else
                            readPixel = REF2D_convertPixel24RGB_15_Dithered;
                        *setupDither = REF2D_setupDither_16;
                        }
                    else {
                        // Setup for simple pixel conversion
                        if (fx->PixelFormat->RedPosition == 0)
                            readPixel = REF2D_convertPixel24BGR_15;
                        else
                            readPixel = REF2D_convertPixel24RGB_15;
                        }
                    break;
                case 16:
                    if (fx->Flags & gaBltDither) {
                        // Setup for real time dithering
                        if (fx->PixelFormat->RedPosition == 0)
                            readPixel = REF2D_convertPixel24BGR_16_Dithered;
                        else
                            readPixel = REF2D_convertPixel24RGB_16_Dithered;
                        *setupDither = REF2D_setupDither_16;
                        }
                    else {
                        // Setup for simple pixel conversion
                        if (fx->PixelFormat->RedPosition == 0)
                            readPixel = REF2D_convertPixel24BGR_16;
                        else
                            readPixel = REF2D_convertPixel24RGB_16;
                        }
                    break;
#endif
#ifdef SUPPORT_24BPP
                case 24:
                    if (DC.mi.pf.RedPosition != fx->PixelFormat->RedPosition)
                        readPixel = REF2D_convertPixel24RGB_24BGR;
                    else
                        readPixel = DC.readPixel;
                    break;
#endif
#ifdef SUPPORT_32BPP
                case 32:
                    if (fx->PixelFormat->RedPosition == 0) {
                        if (DC.mi.pf.RedPosition == 0)
                            readPixel = REF2D_convertPixel24BGR_32ABGR;
                        else if (DC.mi.pf.RedPosition == 8)
                            readPixel = REF2D_convertPixel24BGR_32BGRA;
                        else if (DC.mi.pf.RedPosition == 16)
                            readPixel = REF2D_convertPixel24BGR_32ARGB;
                        else
                            readPixel = REF2D_convertPixel24BGR_32RGBA;
                        }
                    else {
                        if (DC.mi.pf.RedPosition == 0)
                            readPixel = REF2D_convertPixel24RGB_32ABGR;
                        else if (DC.mi.pf.RedPosition == 8)
                            readPixel = REF2D_convertPixel24RGB_32BGRA;
                        else if (DC.mi.pf.RedPosition == 16)
                            readPixel = REF2D_convertPixel24RGB_32ARGB;
                        else
                            readPixel = REF2D_convertPixel24RGB_32RGBA;
                        }
                    break;
#endif
                }
            break;
        case 32:
            // Handle color converting 32bpp bitmaps to other
            // pixel formats.
            switch (DC.mi.bitsPerPixel) {
#ifdef SUPPORT_8BPP
                case 8:
                    DC.cntPal = fx->DstPalette;
                    if (fx->Flags & gaBltDither) {
                        // Setup for real time dithering
                        if (fx->PixelFormat->RedPosition == 0)
                            readPixel = REF2D_convertPixel24BGR_8_Dithered;
                        else if (fx->PixelFormat->RedPosition == 8)
                            readPixel = REF2D_convertPixel32BGRA_8_Dithered;
                        else if (fx->PixelFormat->RedPosition == 16)
                            readPixel = REF2D_convertPixel24RGB_8_Dithered;
                        else
                            readPixel = REF2D_convertPixel32RGBA_8_Dithered;
                        if (fx->Flags & gaBltTranslateVec)
                            _REF2D_ditherVec = fx->TranslateVec;
                        else
                            _REF2D_ditherVec = DC.buildTranslateVector(DC.translateVec,fx->DstPalette,_REF2D_halftonePal,256);
                        *setupDither = REF2D_setupDither_8;
                        }
                    else {
                        // Setup for closest color matching (slow)
                        readPixel = REF2D_convertPixel32_8;
                        if (fx->PixelFormat->RedPosition == 0)
                            DC.findClosest = REF2D_findClosest_24BGR;
                        else if (fx->PixelFormat->RedPosition == 8)
                            DC.findClosest = REF2D_findClosest_32BGRA;
                        else if (fx->PixelFormat->RedPosition == 16)
                            DC.findClosest = REF2D_findClosest_24RGB;
                        else
                            DC.findClosest = REF2D_findClosest_32RGBA;
                        }
                    break;
#endif
#ifdef SUPPORT_16BPP
                case 15:
                    if ((fx->Flags & gaBltBlend) && AdjustBlendFunc(srcBlendFunc,dstBlendFunc)) {
                        // Pixel functions read the alpha value and return
                        // it so it can be used for source alpha blending
                        if (fx->Flags & gaBltDither) {
                            // Setup for real time dithering
                            if (fx->PixelFormat->RedPosition == 0)
                                readPixel = REF2D_convertPixel32ABGR_15_Dithered_Blend;
                            else if (fx->PixelFormat->RedPosition == 8)
                                readPixel = REF2D_convertPixel32BGRA_15_Dithered_Blend;
                            else if (fx->PixelFormat->RedPosition == 16)
                                readPixel = REF2D_convertPixel32ARGB_15_Dithered_Blend;
                            else
                                readPixel = REF2D_convertPixel32RGBA_15_Dithered_Blend;
                            }
                        else {
                            if (fx->PixelFormat->RedPosition == 0)
                                readPixel = REF2D_convertPixel32ABGR_15_Blend;
                            else if (fx->PixelFormat->RedPosition == 8)
                                readPixel = REF2D_convertPixel32BGRA_15_Blend;
                            else if (fx->PixelFormat->RedPosition == 16)
                                readPixel = REF2D_convertPixel32ARGB_15_Blend;
                            else
                                readPixel = REF2D_convertPixel32RGBA_15_Blend;
                            }
                        }
                    else {
                        if (fx->Flags & gaBltDither) {
                            // Setup for real time dithering
                            if (fx->PixelFormat->RedPosition == 0)
                                readPixel = REF2D_convertPixel24BGR_15_Dithered;
                            else if (fx->PixelFormat->RedPosition == 8)
                                readPixel = REF2D_convertPixel32BGRA_15_Dithered;
                            else if (fx->PixelFormat->RedPosition == 16)
                                readPixel = REF2D_convertPixel24RGB_15_Dithered;
                            else
                                readPixel = REF2D_convertPixel32RGBA_15_Dithered;
                            *setupDither = REF2D_setupDither_16;
                            }
                        else {
                            if (fx->PixelFormat->RedPosition == 0)
                                readPixel = REF2D_convertPixel24BGR_15;
                            else if (fx->PixelFormat->RedPosition == 8)
                                readPixel = REF2D_convertPixel32BGRA_15;
                            else if (fx->PixelFormat->RedPosition == 16)
                                readPixel = REF2D_convertPixel24RGB_15;
                            else
                                readPixel = REF2D_convertPixel32RGBA_15;
                            }
                        }
                    break;
                case 16:
                    if (DC.mi.pf.AlphaPosition == 8) {
                        // Special case of writing to 16bpp bitmaps that are actually an
                        // 8-bit color index bitmap with an 8-bit alpha channel
                        // stores as a single 16-bit pixel.
                        DC.cntPal = fx->DstPalette;
                        if (fx->Flags & gaBltDither) {
                            // Setup for real time dithering
                            if (fx->PixelFormat->RedPosition == 0)
                                readPixel = REF2D_convertPixel32ABGR_8A_Dithered;
                            else if (fx->PixelFormat->RedPosition == 8)
                                readPixel = REF2D_convertPixel32BGRA_8A_Dithered;
                            else if (fx->PixelFormat->RedPosition == 16)
                                readPixel = REF2D_convertPixel32ARGB_8A_Dithered;
                            else
                                readPixel = REF2D_convertPixel32RGBA_8A_Dithered;
                            if (fx->Flags & gaBltTranslateVec)
                                _REF2D_ditherVec = fx->TranslateVec;
                            else
                                _REF2D_ditherVec = DC.buildTranslateVector(DC.translateVec,fx->DstPalette,_REF2D_halftonePal,256);
                            *setupDither = REF2D_setupDither_8A;
                            }
                        else {
                            // Setup for closest color matching (slow)
                            if (fx->PixelFormat->RedPosition == 0)
                                readPixel = REF2D_convertPixel32ABGR_8A;
                            else if (fx->PixelFormat->RedPosition == 8)
                                readPixel = REF2D_convertPixel32BGRA_8A;
                            else if (fx->PixelFormat->RedPosition == 16)
                                readPixel = REF2D_convertPixel32ARGB_8A;
                            else
                                readPixel = REF2D_convertPixel32RGBA_8A;
                            }
                        }
                    else if ((fx->Flags & gaBltBlend) && AdjustBlendFunc(srcBlendFunc,dstBlendFunc)) {
                        // Pixel functions read the alpha value and return
                        // it so it can be used for source alpha blending
                        if (fx->Flags & gaBltDither) {
                            // Setup for real time dithering
                            if (fx->PixelFormat->RedPosition == 0)
                                readPixel = REF2D_convertPixel32ABGR_16_Dithered_Blend;
                            else if (fx->PixelFormat->RedPosition == 8)
                                readPixel = REF2D_convertPixel32BGRA_16_Dithered_Blend;
                            else if (fx->PixelFormat->RedPosition == 16)
                                readPixel = REF2D_convertPixel32ARGB_16_Dithered_Blend;
                            else
                                readPixel = REF2D_convertPixel32RGBA_16_Dithered_Blend;
                            }
                        else {
                            if (fx->PixelFormat->RedPosition == 0)
                                readPixel = REF2D_convertPixel32ABGR_16_Blend;
                            else if (fx->PixelFormat->RedPosition == 8)
                                readPixel = REF2D_convertPixel32BGRA_16_Blend;
                            else if (fx->PixelFormat->RedPosition == 16)
                                readPixel = REF2D_convertPixel32ARGB_16_Blend;
                            else
                                readPixel = REF2D_convertPixel32RGBA_16_Blend;
                            }
                        }
                    else {
                        if (fx->Flags & gaBltDither) {
                            // Setup for real time dithering
                            if (fx->PixelFormat->RedPosition == 0)
                                readPixel = REF2D_convertPixel24BGR_16_Dithered;
                            else if (fx->PixelFormat->RedPosition == 8)
                                readPixel = REF2D_convertPixel32BGRA_16_Dithered;
                            else if (fx->PixelFormat->RedPosition == 16)
                                readPixel = REF2D_convertPixel24RGB_16_Dithered;
                            else
                                readPixel = REF2D_convertPixel32RGBA_16_Dithered;
                            *setupDither = REF2D_setupDither_16;
                            }
                        else {
                            if (fx->PixelFormat->RedPosition == 0)
                                readPixel = REF2D_convertPixel24BGR_16;
                            else if (fx->PixelFormat->RedPosition == 8)
                                readPixel = REF2D_convertPixel32BGRA_16;
                            else if (fx->PixelFormat->RedPosition == 16)
                                readPixel = REF2D_convertPixel24RGB_16;
                            else
                                readPixel = REF2D_convertPixel32RGBA_16;
                            }
                        }
                    break;
#endif
#ifdef SUPPORT_24BPP
                case 24:
                    if (fx->Flags & gaBltBlend) {
                        // Pixel functions read the alpha value and return
                        // it so it can be used for source alpha blending
                        if (fx->PixelFormat->RedPosition == 0) {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = REF2D_convertPixel32ABGR_24BGR_Blend;
                            else
                                readPixel = REF2D_convertPixel32ABGR_24RGB_Blend;
                            }
                        else if (fx->PixelFormat->RedPosition == 8) {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = REF2D_convertPixel32BGRA_24BGR_Blend;
                            else
                                readPixel = REF2D_convertPixel32BGRA_24RGB_Blend;
                            }
                        else if (fx->PixelFormat->RedPosition == 16) {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = REF2D_convertPixel32ARGB_24BGR_Blend;
                            else
                                readPixel = REF2D_convertPixel32ARGB_24RGB_Blend;
                            }
                        else {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = REF2D_convertPixel32RGBA_24BGR_Blend;
                            else
                                readPixel = REF2D_convertPixel32RGBA_24RGB_Blend;
                            }
                        }
                    else {
                        // Pixel functions that ignore alpha channel
                        if (fx->PixelFormat->RedPosition == 0) {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = REF2D_convertPixel32ABGR_24BGR;
                            else
                                readPixel = REF2D_convertPixel32ABGR_24RGB;
                            }
                        else if (fx->PixelFormat->RedPosition == 8) {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = REF2D_convertPixel32BGRA_24BGR;
                            else
                                readPixel = REF2D_convertPixel32BGRA_24RGB;
                            }
                        else if (fx->PixelFormat->RedPosition == 16) {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = REF2D_convertPixel32ARGB_24BGR;
                            else
                                readPixel = REF2D_convertPixel32ARGB_24RGB;
                            }
                        else {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = REF2D_convertPixel32RGBA_24BGR;
                            else
                                readPixel = REF2D_convertPixel32RGBA_24RGB;
                            }
                        }
                    break;
#endif
#ifdef SUPPORT_32BPP
                case 32:
                    if (fx->Flags & gaBltBlend) {
                        // Pixel functions read the alpha value and return
                        // it so it can be used for source alpha blending
                        if (fx->PixelFormat->RedPosition == 0) {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = DC.readPixel;
                            else if (DC.mi.pf.RedPosition == 8)
                                readPixel = REF2D_convertPixel32ABGR_32BGRA_Blend;
                            else if (DC.mi.pf.RedPosition == 16)
                                readPixel = REF2D_convertPixel32ABGR_32ARGB_Blend;
                            else
                                readPixel = REF2D_convertPixel32ABGR_32RGBA_Blend;
                            }
                        else if (fx->PixelFormat->RedPosition == 8) {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = REF2D_convertPixel32BGRA_32ABGR_Blend;
                            else if (DC.mi.pf.RedPosition == 8)
                                readPixel = DC.readPixel;
                            else if (DC.mi.pf.RedPosition == 16)
                                readPixel = REF2D_convertPixel32BGRA_32ARGB_Blend;
                            else
                                readPixel = REF2D_convertPixel32BGRA_32RGBA_Blend;
                            }
                        else if (fx->PixelFormat->RedPosition == 16) {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = REF2D_convertPixel32ARGB_32ABGR_Blend;
                            else if (DC.mi.pf.RedPosition == 8)
                                readPixel = REF2D_convertPixel32ARGB_32BGRA_Blend;
                            else if (DC.mi.pf.RedPosition == 16)
                                readPixel = DC.readPixel;
                            else
                                readPixel = REF2D_convertPixel32ARGB_32RGBA_Blend;
                            }
                        else {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = REF2D_convertPixel32RGBA_32ABGR_Blend;
                            else if (DC.mi.pf.RedPosition == 8)
                                readPixel = REF2D_convertPixel32RGBA_32BGRA_Blend;
                            else if (DC.mi.pf.RedPosition == 16)
                                readPixel = REF2D_convertPixel32RGBA_32ARGB_Blend;
                            else
                                readPixel = DC.readPixel;
                            }
                        }
                    else {
                        if (fx->PixelFormat->RedPosition == 0) {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = DC.readPixel;
                            else if (DC.mi.pf.RedPosition == 8)
                                readPixel = REF2D_convertPixel32ABGR_32BGRA;
                            else if (DC.mi.pf.RedPosition == 16)
                                readPixel = REF2D_convertPixel32ABGR_32ARGB;
                            else
                                readPixel = REF2D_convertPixel32ABGR_32RGBA;
                            }
                        else if (fx->PixelFormat->RedPosition == 8) {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = REF2D_convertPixel32BGRA_32ABGR;
                            else if (DC.mi.pf.RedPosition == 8)
                                readPixel = DC.readPixel;
                            else if (DC.mi.pf.RedPosition == 16)
                                readPixel = REF2D_convertPixel32BGRA_32ARGB;
                            else
                                readPixel = REF2D_convertPixel32BGRA_32RGBA;
                            }
                        else if (fx->PixelFormat->RedPosition == 16) {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = REF2D_convertPixel32ARGB_32ABGR;
                            else if (DC.mi.pf.RedPosition == 8)
                                readPixel = REF2D_convertPixel32ARGB_32BGRA;
                            else if (DC.mi.pf.RedPosition == 16)
                                readPixel = DC.readPixel;
                            else
                                readPixel = REF2D_convertPixel32ARGB_32RGBA;
                            }
                        else {
                            if (DC.mi.pf.RedPosition == 0)
                                readPixel = REF2D_convertPixel32RGBA_32ABGR;
                            else if (DC.mi.pf.RedPosition == 8)
                                readPixel = REF2D_convertPixel32RGBA_32BGRA;
                            else if (DC.mi.pf.RedPosition == 16)
                                readPixel = REF2D_convertPixel32RGBA_32ARGB;
                            else
                                readPixel = DC.readPixel;
                            }
                        }
                    break;
#endif
                }
            break;
        }
    return readPixel;
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
fx          - GA_bltFx structure describing the requested effects
****************************************************************************/
void NAPI REF2D_BitBltFxSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    GA_bltFx *fx)
{
    uchar                   *src,*srcStart,*dst;
    uchar                   *scanBuf = DC.scanBuf;
    int                     srcInc,dstInc,yInc,dstPitch,mix;
    REF2D_blendColor_t      oldBlendSrc;
    REF2D_blendColor_t      oldBlendDst;
    N_uint8                 oldAlphaValue;
    GA_color                oldForeColor;
    REF2D_readPixel_t       readSrcPixel;
    REF2D_readPixel_t       readDstPixel;
    REF2D_drawPixel_t       drawPixel;
    bufReadScan_t           readScan = DC.readScan;
    ibool                   preReadBuffer,readDest;
    REF2D_processPixel_t    processPixel;
    int                     *repCounts;
    N_int32                 srcBlendFunc = fx->SrcBlendFunc;
    N_int32                 dstBlendFunc = fx->DstBlendFunc;
    REF2D_setupDither_t     setupDither = NULL;
    N_fix32                 zoomx,zoomy,zoom;
    ibool                   clippedx,clippedy,clipped;
    int                     next,prev;

    // Ensure our blit buffers are large enough
    if (dstWidth > DC.maxBltWidth)
        REF2D_resizeBltBuffers(dstWidth);

    // Save internal variables that need to be restored later
    oldBlendSrc = DC.blendSrc;
    oldBlendDst = DC.blendDst;
    oldForeColor = DC.foreColor;
    oldAlphaValue = DC.alphaValue;

    // Setup and handle bitmap stretching
    repCounts = NULL;
    if (fx->Flags & gaBltAnyStretch) {
        // Do pre-clipping and setup for stretch blit operation
        if (!PreClipStretchBlit(
                srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,
                (fx->Flags & gaBltClip),fx->ClipLeft,fx->ClipTop,
                fx->ClipRight,fx->ClipBottom,&zoomx,&zoomy,
                &clippedx,&clippedy,&clipped))
            return;

        // Pre-render the pixel replication counts with clipping
        if (zoomx != N_FIX_1 || zoomy != N_FIX_1) {
            repCounts = SetupPixelReplicationCounts(zoomx,zoomy,&srcLeft,&srcWidth,&dstLeft,
                fx->ClipLeft,fx->ClipRight,clippedx);
            }
        }
    else if (fx->Flags & gaBltClip) {
        // Clip coordinates when not doing stretching if desired
        int dstLeft2 = MAX(dstLeft, fx->ClipLeft);
        int dstTop2 = MAX(dstTop, fx->ClipTop);
        int dstRight = MIN((dstLeft + dstWidth), fx->ClipRight);
        int dstBottom = MIN((dstTop + dstHeight), fx->ClipBottom);
        srcLeft = srcLeft + (dstLeft2 - dstLeft);
        srcTop = srcTop + (dstTop2 - dstTop);
        srcWidth = dstWidth = dstRight - dstLeft2;
        srcHeight = dstHeight = dstBottom - dstTop2;
        dstLeft = dstLeft2;
        dstTop = dstTop2;
        }

    // Setup for color conversion of source bitmap
    preReadBuffer = (DC.isHWBuffer && IS_HARDWARE_SURFACE(srcAddr));
    readSrcPixel = DC.readPixel;
    readDstPixel = DC.readPixel;
    srcInc = dstInc = (DC.mi.bitsPerPixel+7)/8;
    if (fx->Flags & gaBltConvert) {
        srcInc = (fx->BitsPerPixel+7)/8;
        DC.srcX = srcLeft;
        DC.dstX = dstLeft;
        readSrcPixel = REF2D_InitPixelReader(fx,&srcBlendFunc,&dstBlendFunc,&readScan,&preReadBuffer,&setupDither);
        }

    // Setup source and destination bitmap pointers
    src = ((uchar*)srcAddr + (srcPitch * srcTop)) + (srcLeft * srcInc);
    dst = PIXEL_ADDR(dstLeft,dstTop);

    // Setup flipping bitmap in X direction. When we do X flipping
    // we always process the source bitmap in reverse order so
    // that the destination is always written in ascending addresses.
    srcStart = src;
    if (fx->Flags & gaBltFlipX) {
        int adjust = (dstWidth - 1) * srcInc;
        src += adjust;
        scanBuf += adjust;
        srcInc = -srcInc;
        }

    // Setup flipping in the Y direction. For Y flipping we always
    // process the source bitmap in top to bottom fashion but
    // write the scanlines to the destination bitmap in bottom up
    // fashion.
    if (fx->Flags & gaBltFlipY) {
        dst += (dstHeight-1) * DC.mi.bytesPerLine;
        dstPitch = -DC.mi.bytesPerLine;
        yInc = -1;
        }
    else {
        dstPitch = DC.mi.bytesPerLine;
        yInc = 1;
        }

    // Find the mix to use for the blit operation and the
    // default pixel drawing function. This will be overrriden
    // later if we are doing source transparency with a solid mix.
    if (fx->Flags & gaBltMixEnable) {
        mix = fx->Mix;
        }
    else
        mix = GA_REPLACE_MIX;
    drawPixel = DC.drawPixelTable[mix & 0xF];
    readDest = false;

    // Initialise the pixel processing function to empty
    processPixel = NULL;

    // Setup transparent color key operations first
    if (fx->Flags & gaBltAnyColorKey) {
        DC.colorKeyLo = fx->ColorKeyLo & DC.colorMask;
        DC.colorKeyHi = fx->ColorKeyHi & DC.colorMask;
        if (DC.colorCmpMask != 0xFFFFFFFF) {
            // Setup special case only for software rendering to handle a
            // color compare mask and blending. This is mostly used by the
            // conformance tests to ensure correct WHQL operation.
            if (fx->Flags & gaBltAnySingleColorKey) {
                if (fx->Flags & gaBltColorKeySrcSingle)
                    processPixel = REF2D_processSrcColorKeyMasked;
                else
                    processPixel = REF2D_processDstColorKeyMasked;
                }
            else {
                if (fx->Flags & gaBltColorKeySrcRange)
                    processPixel = REF2D_processSrcColorKeyRangeMasked;
                else
                    processPixel = REF2D_processDstColorKeyRangeMasked;
                }
            readDest = true;
            }
        else {
            // No color compare masking so use optimal pixel processing functions
            if (fx->Flags & gaBltAnySingleColorKey) {
                if (fx->Flags & gaBltColorKeySrcSingle) {
                    if ((mix == GA_REPLACE_MIX) && !(fx->Flags & gaBltBlend)) {
                        // When doing source transparency in
                        // with no blending, we can use a faster source
                        // transparent pixel writing function for speed.
                        drawPixel = DC.drawPixelSrcColorKey;
                        }
                    else {
                        processPixel = REF2D_processSrcColorKey;
                        readDest = true;
                        }
                    }
                else {
                    processPixel = REF2D_processDstColorKey;
                    readDest = true;
                    }
                }
            else {
                if (fx->Flags & gaBltColorKeySrcRange) {
                    if ((mix == GA_REPLACE_MIX) && !(fx->Flags & gaBltBlend)) {
                        // When doing source transparency in GA_REPLACE_MIX
                        // with no blending, we can use a faster source
                        // transparent pixel writing function for speed.
                        drawPixel = DC.drawPixelSrcColorKeyRange;
                        }
                    else {
                        processPixel = REF2D_processSrcColorKeyRange;
                        readDest = true;
                        }
                    }
                else {
                    processPixel = REF2D_processDstColorKeyRange;
                    readDest = true;
                    }
                }
            }
        }

    // Setup for alpha blending last after transparency
#if defined(SUPPORT_16BPP) || defined(SUPPORT_24BPP) || defined(SUPPORT_32BPP)
    if ((fx->Flags & gaBltBlend) && DC.mi.bitsPerPixel >= 15) {
        DC.blendSrc = REF2D_blendSrcTable[srcBlendFunc & 0xF];
        DC.blendDst = REF2D_blendDstTable[dstBlendFunc & 0xF];
        DC.foreColor = fx->ConstColor;
        DC.alphaValue = fx->ConstAlpha;
        drawPixel = DC.blendPixelTable[GET_BLEND_INDEX(srcBlendFunc,dstBlendFunc)];
        }
#endif

    // Do the actual blit operation
    DC.beginDirectAccess();
    if (repCounts) {
        int clipTop = fx->ClipTop;
        int clipBottom = fx->ClipBottom;

        zoom = zoomy;
        prev = 0;
        while (srcHeight--) {
            int     dy = (next = N_FIXROUND(zoom)) - prev;

            prev = next;
            zoom += zoomy;
            if (preReadBuffer) {
                if (dy)
                    readScan(DC.scanBuf,srcStart X_ARG2(srcLeft,srcLeft),srcWidth);
                srcStart += srcPitch;
                }
            while (dy--) {
                uchar   *s = preReadBuffer ? scanBuf : src;

                if (!clippedy || (dstTop >= clipTop && dstTop < clipBottom)) {
                    uchar       *d = dst;
                    GA_color    S,D;
                    int         count = srcWidth;
                    int         *r = repCounts;

                    if (setupDither)
                        setupDither(dstTop);
                    while (count--) {
                        int i = *r++;
                        if (i) {
                            S = readSrcPixel(s X_ARG(dstLeft));
                            if (readDest)
                                D = readDstPixel(d X_ARG(dstLeft));
                            D = (processPixel) ? processPixel(S,D) : S;
                            }
                        while (i--) {
                            drawPixel(d X_ARG(dstLeft),D);
                            d += dstInc;
                            }
                        s += srcInc;
                        }
                    }
                dst += dstPitch;
                dstTop += yInc;
                }
            src += srcPitch;
            }
        }
    else {
        while (dstHeight--) {
            uchar       *s = src,*d = dst;
            GA_color    S,D;
            int         count = dstWidth;

            if (setupDither) {
                // Setup scanline for fast dithering
                setupDither(dstTop);
                dstTop += yInc;
                }
            if (preReadBuffer) {
                // Pre-read the entire scanline from hardware surfaces
                // before it is processed for efficiency. We also pre-expand
                // 1bpp and 4bpp input bitmaps to 8bpp for efficiency in the
                // internal pixel processing routines here also.
                readScan(DC.scanBuf,srcStart X_ARG2(srcLeft,srcLeft),dstWidth);
                srcStart += srcPitch;
                s = scanBuf;
                }
            if (readDest) {
                // Draw the scanline with pixel reads from the destination
                // if this is needed by the pixel processing stack. Note
                // that we always have a pixel processing function if we
                // need to read the destination pixel value, so we can
                // avoid the check to see if the processing function is active.
                while (count--) {
                    S = readSrcPixel(s X_ARG(dstLeft));
                    D = readDstPixel(d X_ARG(dstLeft));
                    D = processPixel(S,D);
                    drawPixel(d X_ARG(dstLeft),D);
                    s += srcInc;
                    d += dstInc;
                    }
                }
            else {
                // Draw the scanline as fast as possible for the most
                // common case of no pixel replication and no destination
                // pixel reads. Each pixel is passed through the pixel
                // processing stack in order before being written to the
                // destination.
                while (count--) {
                    S = readSrcPixel(s X_ARG(dstLeft));
                    D = (processPixel) ? processPixel(S,D) : S;
                    drawPixel(d X_ARG(dstLeft),D);
                    s += srcInc;
                    d += dstInc;
                    }
                }
            src += srcPitch;
            dst += dstPitch;
            }
        }
    DC.blendSrc = oldBlendSrc;
    DC.blendDst = oldBlendDst;
    DC.foreColor = oldForeColor;
    DC.alphaValue = oldAlphaValue;
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPhysAddr - Physical address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
fx          - GA_bltFx structure describing the requested effects
****************************************************************************/
void NAPI REF2D_BitBltFxBM(
    void *srcAddr,
    N_int32 srcPhysAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    GA_bltFx *fx)
{
    // Pass through to hardware if hardware can handle the blit directly
    if (REF_draw2dHW.BitBltFxTest && REF_draw2dHW.BitBltFxBM && REF_draw2dHW.BitBltFxTest(fx))
        REF_draw2dHW.BitBltFxBM(srcAddr,srcPhysAddr,srcPitch,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
    else
        REF2D_BitBltFxSys(srcAddr,srcPitch,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
}

/****************************************************************************
PARAMETERS:
dstAddr     - Address of destination bitmap in system memory
dstPhysAddr - Physical address of destination bitmap in system memory
dstPitch    - Pitch of destination bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_GetBitmapSys(
    void *dstAddr,
    N_int32 dstPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    uchar *s = PIXEL_ADDR(srcLeft,srcTop);
    uchar *d = ((uchar*)dstAddr + (dstPitch * dstTop)) + X_ADDR(dstLeft);

    // Ensure our blit buffers are large enough
    if (width > DC.maxBltWidth)
        REF2D_resizeBltBuffers(width);

    DC.beginDirectAccess();
    if (IS_SOLID_MIX(mix)) {
        while (height--) {
            DC.readScan(d,s X_ARG2(srcLeft,dstLeft),width);
            s += DC.mi.bytesPerLine;
            d += dstPitch;
            }
        }
    else {
        REF2D_drawPixel_t   oldDrawPixel = DC.drawPixel;
        DC.drawPixel = GET_PIXEL_FUNC(mix);
        while (height--) {
            DC.readScan(DC.scanBuf,s X_ARG2(srcLeft,0),width);
            DC.bltScan_Generic(d,DC.scanBuf X_ARG2(0,dstLeft),width);
            s += DC.mi.bytesPerLine;
            d += dstPitch;
            }
        DC.drawPixel = oldDrawPixel;
        }
    DC.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
dstAddr     - Address of destination bitmap in system memory
dstPhysAddr - Physical address of destination bitmap in system memory
dstPitch    - Pitch of destination bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_GetBitmapBM(
    void *dstAddr,
    N_int32 dstPhysAddr,
    N_int32 dstPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    (void)dstPhysAddr;
    REF2D_GetBitmapSys(dstAddr,dstPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,mix);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
planeMask   - Plane mask to use for blit
****************************************************************************/
void NAPI REF2D_BitBltPlaneMasked(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_uint32 planeMask)
{
    N_uint32 oldMask = DC.planeMask;

    // TODO: Optimise this to use native hardware plane masks, ROP3's or
    //       multi-passes using a temporary scanline.
    REF2D_SetPlaneMask(planeMask);
    STUB_BitBlt(srcLeft,srcTop,width,height,dstLeft,dstTop,GA_REPLACE_MIX);
    REF2D_SetPlaneMask(oldMask);
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
planeMask       - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_BitBltPlaneMaskedLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_uint32 planeMask)
{
    N_uint32 oldMask = DC.planeMask;

    // TODO: Optimise this to use native hardware plane masks, ROP3's or
    //       multi-passes using a temporary scanline.
    REF2D_SetPlaneMask(planeMask);
    STUB_BitBltLin(srcOfs,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,GA_REPLACE_MIX);
    REF2D_SetPlaneMask(oldMask);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
planeMask       - planeMask raster operation code for the copy (GA_planeMaskCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI REF2D_BitBltPlaneMaskedSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_uint32 planeMask,
    N_int32 flipY)
{
    N_uint32 oldMask = DC.planeMask;

    // TODO: Optimise this to use native hardware plane masks, ROP3's or
    //       multi-passes using a temporary scanline.
    REF2D_SetPlaneMask(planeMask);
    STUB_BitBltSys(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,GA_REPLACE_MIX,flipY);
    REF2D_SetPlaneMask(oldMask);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPhysAddr - Physical address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
planeMask       - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_BitBltPlaneMaskedBM(
    void *srcAddr,
    N_int32 srcPhysAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_uint32 planeMask)
{
    (void)srcPhysAddr;
    REF2D_BitBltPlaneMaskedSys(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,planeMask,false);
}

