/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for the SciTech SNAP Graphics 2D Reference
*               Rasteriser internal functions.
*
****************************************************************************/

#ifndef __REF2DINT_H
#define __REF2DINT_H

#include "internal.h"

/*---------------------- Macros and type definitions ----------------------*/

#pragma pack(1)

/****************************************************************************
REMARKS:
Structure to store the information about the drawing device for the
reference rasteriser functions. We have a structure that is similar to the
MGL to make it easy to re-use MGL drawing functions for the reference
rasteriser.

MEMBERS:
xRes                - Device X resolution
yRes                - Device Y resolution
bitsPerPixel        - Pixel depth
bytesPerLine        - Number of bytes in a single device scanline
byteShift           - Bits to shift left to convert to bytes
****************************************************************************/
typedef struct {
    int             xRes;
    int             yRes;
    int             bitsPerPixel;
    int             bytesPerLine;
    signed char     byteShift;
    GA_pixelFormat  pf;
    } REF2D_modeInfo;

#pragma pack(4)

// Size of internal cursor structures

#define CURSOR_SIZE     64
#define CUR_BUF_PITCH   (sizeof(GA_color) * CURSOR_SIZE)

// Function pointer prototypes for pixel functions
typedef GA_color    (*REF2D_readPixel_t)(uchar *addr DCL_X_ARG);
typedef void        (*REF2D_drawPixel_t)(uchar *addr DCL_X_ARG,GA_color color);
typedef void        (*REF2D_drawPixelR3_t)(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern);
typedef GA_color    (*REF2D_blendColor_t)(GA_color S,GA_color D,GA_color C,GA_color A);
typedef void        (*REF2D_blendPixel_t)(uchar *addr DCL_X_ARG,GA_color color,GA_color alpha);
typedef GA_color    (*REF2D_processPixel_t)(GA_color S,GA_color D);
typedef void        (*REF2D_setupDither_t)(N_int32 y);

// Function pointer prototypes for scanline functions
typedef void        (*bltScan_t)(uchar *d,uchar *s DCL_X_ARG2,int width);
typedef void        (*bltScanPatt_t)(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
typedef void        (*bltScanColorPatt_t)(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
typedef void        (*stretchScan2x_t)(uchar *d,uchar *s DCL_X_ARG2,int width);
typedef void        (*stretchScan_t)(uchar *d,uchar *s DCL_X_ARG2,int width,int *repCounts);
typedef void        (*convScan4_t)(uchar *d,uchar *s DCL_X_ARG,int width,int odd,GA_color *translate);
typedef void        (*convScan8_t)(uchar *d,uchar *s DCL_X_ARG,int width,GA_color *translate);

/****************************************************************************
REMARKS:
Structure to store the current state of the 2D reference rasteriser
functions.

MEMBERS:
hwCtx               - Pointer to hardware device context (if any)
hwBufferStart       - Start of hardware framebuffer
hwBufferEnd         - End of hardware framebuffer
frameBuffer         - Pointer to the start of physical framebuffer memory
cursorDrawAddr      - Pointer to the framebuffer to use for cursor drawing
drawOffset          - Offset within framebuffer for current drawing surface
surface             - Pointer to the start of drawing memory buffer
mi                  - Mode information describing the buffer
mix                 - Currently active mix
foreColor           - Currently active foreground color
backColor           - Currently active background color
colorMask           - Mask for active color bits for color depth
planeMask           - Plane mask for current drawing operations
planeMaskInv        - Invert of the plane mask
srcBlendFunc        - Source blend function
dstBlendFunc        - Destination blend function
alphaValue          - Alpha value for contsant alpha blending
mono8x8Index        - Currently active 8x8 pattern index
mono8x8Transparent  - True if mono patterns should be transparent
mono8x8Patt         - Cache of 8x8 mono pattern data
color8x8Index       - Currently active 8x8 color pattern index
color8x8Patt        - Cache of 8x8 color pattern data
stipple             - Current line stipple pattern
stippleCount        - Current line stipple count
cntPal              - Current palette used for pixel format translations
cursorForeColor     - Current color for cursor foreground
cursorBackColor     - Current color for cursor background
****************************************************************************/
typedef struct {
    GA_devCtx           *hwCtx;
    GA_initFuncs        hwInit;
    GA_2DStateFuncs     hwState;
    GA_2DRenderFuncs    hwDraw;
    GA_cursorFuncs      hwCursor;
    GA_driverFuncs      hwDrv;
    GA_bufferFuncs      hwBuf;
    GA_globalOptions    *gOpt;
    ibool               haveHWDraw;
    ibool               isHWBuffer;
    uchar               *hwBufferStart;
    uchar               *hwBufferEnd;
    uchar               *frameBuffer;
    void                *cursorDrawAddr;
    N_uint32            drawOffset;
    GA_buffer           cntDrawBuf;
    uchar               *surface;
    REF2D_modeInfo      mi;
    GA_modeInfo         gaMode;
    N_int32             mix;
    GA_color            foreColor;
    GA_color            backColor;
    GA_color            transparent;
    GA_color            colorKeyLo;
    GA_color            colorKeyHi;
    N_int32             palSize;
    N_uint32            colorMask;
    N_uint32            colorCmpMask;
    N_uint32            planeMask;
    N_uint32            planeMaskInv;
    N_int32             srcBlendFunc;
    N_int32             dstBlendFunc;
    N_int32             blendIndex;
    N_uint32            alphaValue;
    N_int32             mono8x8Index;
    N_int32             mono8x8Transparent;
    GA_pattern          mono8x8Patt[8];
    N_int32             color8x8Index;
    N_int32             color8x8Transparent;
    GA_color            color8x8TransColor;
    GA_colorPattern     color8x8Patt[8];
    GA_stipple          stipple;
    N_uint32            stippleCount;
    N_uint32            styleMask;
    N_uint32            styleStep;
    N_uint32            styleValue;
    GA_color            *translate;
    N_int32             srcX;
    N_int32             dstX;
    GA_color            translateVec[256];
    GA_palette          shadowPal[256];
    GA_palette          *cntPal;
    int                 *ditherRow;
    GA_color            cursorForeColor;
    GA_color            cursorBackColor;
    N_int32             cntCursorX;
    N_int32             cntCursorY;
    N_int32             cntHotX;
    N_int32             cntHotY;
    N_int32             cntCursorWidth;
    N_int32             cntCursorHeight;
    int                 maxCursorWidth;
    int                 maxCursorHeight;
    N_uint32            maxBltBytes;
    N_uint32            maxBltWidth;
    uchar               *scanBuf;
    uchar               *scanBuf2;
    int                 *repCounts;
    PE_MODULE           *hModRegion;
    PE_MODULE           *hModBuffer;

    // Functions to load and unload the buffer manager
    ibool               (*LoadBufferMgr)(GA_devCtx *hwCtx,REF2D_driver *ref2d,GA_bufferFuncs *funcs,PE_MODULE **hModule,ulong *size);
    void                (*UnloadBufferMgr)(PE_MODULE *hModule);

    // Callback to original buffer manager init function
    ibool               (NAPIP BUF_InitBuffersOrg)(N_int32 numBuffers,N_uint32 flags,GA_softStereoFuncs *stereo);
    void                (NAPIP BUF_FlipToBufferOrg)(GA_buf *buf,N_int32 waitVRT);
    ibool               (NAPIP BUF_SetRef2d)(struct _REF2D_driver *ref2d);
    N_uint32            (NAPIP BUF_LockBuffer)(GA_buf *buf);
    void                (NAPIP BUF_UnlockBuffer)(GA_buf *buf);

    // Active drawing functions for current display mode (never changes)
    bufReadScan_t       bufReadScan;
    bufBltScan_t        bufBltScan;

    // Currently active blending functions
    REF2D_blendColor_t  blendSrc;
    REF2D_blendColor_t  blendDst;

    // Currently active pixel functions
    REF2D_readPixel_t   readPixel;
    REF2D_drawPixel_t   drawPixel;
    REF2D_drawPixel_t   drawPixelSrcColorKey;
    REF2D_drawPixel_t   drawPixelSrcColorKeyRange;
    REF2D_drawPixel_t   *drawPixelTable;
    REF2D_drawPixelR3_t drawPixelR3;
    REF2D_drawPixelR3_t *drawPixelR3Table;
    REF2D_drawPixel_t   *blendPixelTable;
    REF2D_blendPixel_t  blendPixelFast;

    // Currently active drawing functions
    void                (NAPIP beginDirectAccess)(void);
    void                (NAPIP endDirectAccess)(void);
    void                (*fillScan_S)(uchar *addr DCL_X_ARG,int width,GA_color color);
    void                (*fillScan_Generic)(uchar *addr DCL_X_ARG,int width,GA_color color);
    void                (*drawScan)(uchar *addr DCL_X_ARG,int width);
    void                (*drawPattScan)(uchar *addr,int x1,int y,int width);
    void                (*drawColorPattScan)(uchar *addr,int x1,int y,int width);
    void                (*drawLineXPosLo)(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
    void                (*drawLineXPosHi)(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
    void                (*drawLineXNegLo)(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
    void                (*drawLineXNegHi)(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
    void                (*drawStippleLineXPosLo)(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
    void                (*drawStippleLineXPosHi)(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
    void                (*drawStippleLineXNegLo)(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
    void                (*drawStippleLineXNegHi)(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
    void                (*drawStyleLineXPosLo)(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
    void                (*drawStyleLineXPosHi)(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
    void                (*drawStyleLineXNegLo)(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
    void                (*drawStyleLineXNegHi)(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
    void                (*drawMonoMSBTrans)(uchar *d,uchar *image DCL_X_ARG,int width);
    void                (*drawMonoMSBOpaque)(uchar *d,uchar *image DCL_X_ARG,int width);
    void                (*drawClipMonoMSBTrans)(int x,uchar *d,uchar *image,int width,int clipLeft);
    void                (*drawClipMonoMSBOpaque)(int x,uchar *d,uchar *image,int width,int clipLeft);
    void                (*drawMonoLSBTrans)(uchar *d,uchar *image DCL_X_ARG,int width);
    void                (*drawMonoLSBOpaque)(uchar *d,uchar *image DCL_X_ARG,int width);
    void                (*drawClipMonoLSBTrans)(int x,uchar *d,uchar *image,int width,int clipLeft);
    void                (*drawClipMonoLSBOpaque)(int x,uchar *d,uchar *image,int width,int clipLeft);
    void                (*drawMonoCursor)(uchar *d,uchar *mask,uchar *image DCL_X_ARG,int width);
    void                (*drawColorCursor)(uchar *d,uchar *mask,uchar *image DCL_X_ARG,int width);
    void                (*readScan)(uchar *d,uchar *s DCL_X_ARG2,int width);
    void                (*bltScan_S)(uchar *d,uchar *s DCL_X_ARG2,int width);
    void                (*bltScan_Generic)(uchar *d,uchar *s DCL_X_ARG2,int width);
    void                (*bltScanPatt)(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
    void                (*bltScanColorPatt)(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
    void                (*srcTransBltScan_S)(uchar *d,uchar *s DCL_X_ARG2,int width);
    void                (*srcTransBltScan_Generic)(uchar *d,uchar *s DCL_X_ARG2,int width);
    void                (*srcTransBltScan_Masked)(uchar *d,uchar *s DCL_X_ARG2,int width);
    void                (*dstTransBltScan_S)(uchar *d,uchar *s DCL_X_ARG2,int width);
    void                (*dstTransBltScan_Generic)(uchar *d,uchar *s DCL_X_ARG2,int width);
    void                (*dstTransBltScan_Masked)(uchar *d,uchar *s DCL_X_ARG2,int width);
    void                (*stretchScan2x_S)(uchar *d,uchar *s DCL_X_ARG2,int width);
    void                (*stretchScan2x_Generic)(uchar *d,uchar *s DCL_X_ARG2,int width);
    void                (*stretchScan_S)(uchar *d,uchar *s DCL_X_ARG2,int width,int *repCounts);
    void                (*stretchScan_Generic)(uchar *d,uchar *s DCL_X_ARG2,int width,int *repCounts);
    GA_color *          (*buildTranslateVector)(GA_color *translate,GA_palette *dstPal,GA_palette *srcPal,int srcColors);
    GA_color            (*findClosest)(GA_color color);
    void                (*convScan4_S)(uchar *d,uchar *s DCL_X_ARG,int width,int odd,GA_color *translate);
    void                (*convScan8_S)(uchar *d,uchar *s DCL_X_ARG,int width,GA_color *translate);
    void                (*drawEllipse)(int ileft,int itop,int A,int B);
    void                (*clipEllipse)(int ileft,int itop,int A,int B,int clipLeft,int clipTop,int clipRight,int clipBottom);
    void                (*updateCursor)(int dispOffset,int left,int top,int width,int height);
    } REF2D_state;

#pragma pack()

#if defined(__IBMC__) || defined(__IBMCPP__)
#define     CNOP_VECTOR         (__c_code_ptr_t)REF2D_nop
#define     NOP_VECTOR          (__code_ptr_t)REF2D_nop
#define     ULNOP_VECTOR        (__ulcode_ptr_t)REF2D_nop
#else
#define     CNOP_VECTOR         (void*)REF2D_nop
#define     NOP_VECTOR          (void*)REF2D_nop
#define     ULNOP_VECTOR        (void*)REF2D_nop
#endif

#define DECLARE_ROP2_PIXELS(bpp)    \
    void REF2D_drawPixel_##bpp##_0(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixel_##bpp##_DSon(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixel_##bpp##_DSna(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixel_##bpp##_Sn(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixel_##bpp##_SDna(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixel_##bpp##_Dn(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixel_##bpp##_DSx(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixel_##bpp##_DSan(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixel_##bpp##_DSa(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixel_##bpp##_DSxn(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixel_##bpp##_DSno(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixel_##bpp##_S(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixel_##bpp##_SDno(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixel_##bpp##_DSo(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixel_##bpp##_1(uchar *addr DCL_X_ARG,GA_color color)

#define DEFINE_ROP2_PIXELS(bpp) \
    REF2D_drawPixel_##bpp##_0, \
    REF2D_drawPixel_##bpp##_DSon, \
    REF2D_drawPixel_##bpp##_DSna, \
    REF2D_drawPixel_##bpp##_Sn, \
    REF2D_drawPixel_##bpp##_SDna, \
    REF2D_drawPixel_##bpp##_Dn, \
    REF2D_drawPixel_##bpp##_DSx, \
    REF2D_drawPixel_##bpp##_DSan, \
    REF2D_drawPixel_##bpp##_DSa, \
    REF2D_drawPixel_##bpp##_DSxn, \
    NOP_VECTOR, \
    REF2D_drawPixel_##bpp##_DSno, \
    REF2D_drawPixel_##bpp##_S, \
    REF2D_drawPixel_##bpp##_SDno, \
    REF2D_drawPixel_##bpp##_DSo, \
    REF2D_drawPixel_##bpp##_1

#define DECLARE_ROP2_MASKED_PIXELS(bpp) \
    void REF2D_drawPixelMasked_##bpp##_0(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelMasked_##bpp##_DSon(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelMasked_##bpp##_DSna(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelMasked_##bpp##_Sn(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelMasked_##bpp##_SDna(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelMasked_##bpp##_Dn(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelMasked_##bpp##_DSx(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelMasked_##bpp##_DSan(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelMasked_##bpp##_DSa(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelMasked_##bpp##_DSxn(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelMasked_##bpp##_DSno(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelMasked_##bpp##_S(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelMasked_##bpp##_SDno(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelMasked_##bpp##_DSo(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelMasked_##bpp##_1(uchar *addr DCL_X_ARG,GA_color color)

#define DEFINE_ROP2_MASKED_PIXELS(bpp)  \
    REF2D_drawPixelMasked_##bpp##_0, \
    REF2D_drawPixelMasked_##bpp##_DSon, \
    REF2D_drawPixelMasked_##bpp##_DSna, \
    REF2D_drawPixelMasked_##bpp##_Sn, \
    REF2D_drawPixelMasked_##bpp##_SDna, \
    REF2D_drawPixelMasked_##bpp##_Dn, \
    REF2D_drawPixelMasked_##bpp##_DSx, \
    REF2D_drawPixelMasked_##bpp##_DSan, \
    REF2D_drawPixelMasked_##bpp##_DSa, \
    REF2D_drawPixelMasked_##bpp##_DSxn, \
    NOP_VECTOR, \
    REF2D_drawPixelMasked_##bpp##_DSno, \
    REF2D_drawPixelMasked_##bpp##_S, \
    REF2D_drawPixelMasked_##bpp##_SDno, \
    REF2D_drawPixelMasked_##bpp##_DSo, \
    REF2D_drawPixelMasked_##bpp##_1

#define DECLARE_ROP3_PIXELS_1(bpp)  \
    void REF2D_drawPixelR3_##bpp##_0(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSoon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSona(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPona(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSxnon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPnaa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSxon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDnaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSnaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSonon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_Pn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSona(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPxnon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSxnon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDPSanaxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SSPxDSxaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPxPDxa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSanaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSPaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSxaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDPaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPDxaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSnaa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPxon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDnaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPxDSxa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSPanaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSxnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSxa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDPSaoxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSana(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SSPxPDxaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDPxox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDnoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPnaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSoox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_Sn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSxnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDPoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSxox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDnoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSonox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSnaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDnaa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSxon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDxPDxa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSanaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSnaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPDaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDPxaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPxa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSPDaoxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSDoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPana(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SSPxDSxoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSPxox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSnoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPnaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSDaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSxaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSonon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_Dn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSPoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSDonox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSDxox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSnoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSDnaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSxa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPDSaoxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPDoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSonox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPDSonoxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDPSoaxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSPDoaxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSnoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSxnan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSana(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SSDxPDxaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSxox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPnoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPDxox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPnoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSnaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPDSoaxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSDnoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPxnan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSnoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSxnan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPxDSxo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSaan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern);

#define DECLARE_ROP3_PIXELS_2(bpp)  \
    void REF2D_drawPixelR3_##bpp##_DPSaa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPxDSxon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSxna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSnoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPxna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSPnoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPDSoaxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSnaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPnoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPDxoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPnoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSxoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SSDxPDxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSanan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSxna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSnoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSDPoaxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDPSoaxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDPSonoxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSonoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSnax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDnax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPDoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPDSaoxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSxan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSPnaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSnoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSDxoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSPonoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPnax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSPoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_D(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSono(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSDaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPnao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSnoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSPxoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SSPxDSxox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPanan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDnax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSDoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSDPaoxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPxan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDPxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPDaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSnao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSanax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDxPDxan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSxo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSano(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSnaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSonoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDnoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSxoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPnax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDPoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSDxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_S(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPono(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPnao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDnoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDPxoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSnax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDSoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SSPxPDxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSanan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDPSaoxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSxan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSPxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSDanax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPxDSxan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPDnao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPxo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPano(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPDxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDPaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSPaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPSanax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SPxPDxan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SSPxDSxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSPDSanaxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSxno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPxno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DSo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_SDPnoo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_P(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSono(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSnao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDnao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSxo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSano(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PDSxno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSnoo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_PSDnoo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_DPSoo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3_##bpp##_1(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern)

/* We split up the above macros to avoid compile errors with some compilers */

#define DECLARE_ROP3_PIXELS(bpp)    \
    DECLARE_ROP3_PIXELS_1(bpp)      \
    DECLARE_ROP3_PIXELS_2(bpp)

#define DEFINE_ROP3_PIXELS_1(bpp)   \
    REF2D_drawPixelR3_##bpp##_0, \
    REF2D_drawPixelR3_##bpp##_DPSoon, \
    REF2D_drawPixelR3_##bpp##_DPSona, \
    REF2D_drawPixelR3_##bpp##_PSon, \
    REF2D_drawPixelR3_##bpp##_SDPona, \
    REF2D_drawPixelR3_##bpp##_DPon, \
    REF2D_drawPixelR3_##bpp##_PDSxnon, \
    REF2D_drawPixelR3_##bpp##_PDSaon, \
    REF2D_drawPixelR3_##bpp##_SDPnaa, \
    REF2D_drawPixelR3_##bpp##_PDSxon, \
    REF2D_drawPixelR3_##bpp##_DPna, \
    REF2D_drawPixelR3_##bpp##_PSDnaon, \
    REF2D_drawPixelR3_##bpp##_SPna, \
    REF2D_drawPixelR3_##bpp##_PDSnaon, \
    REF2D_drawPixelR3_##bpp##_PDSonon, \
    REF2D_drawPixelR3_##bpp##_Pn, \
    REF2D_drawPixelR3_##bpp##_PDSona, \
    REF2D_drawPixelR3_##bpp##_DSon, \
    REF2D_drawPixelR3_##bpp##_SDPxnon, \
    REF2D_drawPixelR3_##bpp##_SDPaon, \
    REF2D_drawPixelR3_##bpp##_DPSxnon, \
    REF2D_drawPixelR3_##bpp##_DPSaon, \
    REF2D_drawPixelR3_##bpp##_PSDPSanaxx, \
    REF2D_drawPixelR3_##bpp##_SSPxDSxaxn, \
    REF2D_drawPixelR3_##bpp##_SPxPDxa, \
    REF2D_drawPixelR3_##bpp##_SDPSanaxn, \
    REF2D_drawPixelR3_##bpp##_PDSPaox, \
    REF2D_drawPixelR3_##bpp##_SDPSxaxn, \
    REF2D_drawPixelR3_##bpp##_PSDPaox, \
    REF2D_drawPixelR3_##bpp##_DSPDxaxn, \
    REF2D_drawPixelR3_##bpp##_PDSox, \
    REF2D_drawPixelR3_##bpp##_PDSoan, \
    REF2D_drawPixelR3_##bpp##_DPSnaa, \
    REF2D_drawPixelR3_##bpp##_SDPxon, \
    REF2D_drawPixelR3_##bpp##_DSna, \
    REF2D_drawPixelR3_##bpp##_SPDnaon, \
    REF2D_drawPixelR3_##bpp##_SPxDSxa, \
    REF2D_drawPixelR3_##bpp##_PDSPanaxn, \
    REF2D_drawPixelR3_##bpp##_SDPSaox, \
    REF2D_drawPixelR3_##bpp##_SDPSxnox, \
    REF2D_drawPixelR3_##bpp##_DPSxa, \
    REF2D_drawPixelR3_##bpp##_PSDPSaoxxn, \
    REF2D_drawPixelR3_##bpp##_DPSana, \
    REF2D_drawPixelR3_##bpp##_SSPxPDxaxn, \
    REF2D_drawPixelR3_##bpp##_SPDSoax, \
    REF2D_drawPixelR3_##bpp##_PSDnox, \
    REF2D_drawPixelR3_##bpp##_PSDPxox, \
    REF2D_drawPixelR3_##bpp##_PSDnoan, \
    REF2D_drawPixelR3_##bpp##_PSna, \
    REF2D_drawPixelR3_##bpp##_SDPnaon, \
    REF2D_drawPixelR3_##bpp##_SDPSoox, \
    REF2D_drawPixelR3_##bpp##_Sn, \
    REF2D_drawPixelR3_##bpp##_SPDSaox, \
    REF2D_drawPixelR3_##bpp##_SPDSxnox, \
    REF2D_drawPixelR3_##bpp##_SDPox, \
    REF2D_drawPixelR3_##bpp##_SDPoan, \
    REF2D_drawPixelR3_##bpp##_PSDPoax, \
    REF2D_drawPixelR3_##bpp##_SPDnox, \
    REF2D_drawPixelR3_##bpp##_SPDSxox, \
    REF2D_drawPixelR3_##bpp##_SPDnoan, \
    REF2D_drawPixelR3_##bpp##_PSx, \
    REF2D_drawPixelR3_##bpp##_SPDSonox, \
    REF2D_drawPixelR3_##bpp##_SPDSnaox, \
    REF2D_drawPixelR3_##bpp##_PSan, \
    REF2D_drawPixelR3_##bpp##_PSDnaa, \
    REF2D_drawPixelR3_##bpp##_DPSxon, \
    REF2D_drawPixelR3_##bpp##_SDxPDxa, \
    REF2D_drawPixelR3_##bpp##_SPDSanaxn, \
    REF2D_drawPixelR3_##bpp##_SDna, \
    REF2D_drawPixelR3_##bpp##_DPSnaon, \
    REF2D_drawPixelR3_##bpp##_DSPDaox, \
    REF2D_drawPixelR3_##bpp##_PSDPxaxn, \
    REF2D_drawPixelR3_##bpp##_SDPxa, \
    REF2D_drawPixelR3_##bpp##_PDSPDaoxxn, \
    REF2D_drawPixelR3_##bpp##_DPSDoax, \
    REF2D_drawPixelR3_##bpp##_PDSnox, \
    REF2D_drawPixelR3_##bpp##_SDPana, \
    REF2D_drawPixelR3_##bpp##_SSPxDSxoxn, \
    REF2D_drawPixelR3_##bpp##_PDSPxox, \
    REF2D_drawPixelR3_##bpp##_PDSnoan, \
    REF2D_drawPixelR3_##bpp##_PDna, \
    REF2D_drawPixelR3_##bpp##_DSPnaon, \
    REF2D_drawPixelR3_##bpp##_DPSDaox, \
    REF2D_drawPixelR3_##bpp##_SPDSxaxn, \
    REF2D_drawPixelR3_##bpp##_DPSonon, \
    REF2D_drawPixelR3_##bpp##_Dn, \
    REF2D_drawPixelR3_##bpp##_DPSox, \
    REF2D_drawPixelR3_##bpp##_DPSoan, \
    REF2D_drawPixelR3_##bpp##_PDSPoax, \
    REF2D_drawPixelR3_##bpp##_DPSnox, \
    REF2D_drawPixelR3_##bpp##_DPx, \
    REF2D_drawPixelR3_##bpp##_DPSDonox, \
    REF2D_drawPixelR3_##bpp##_DPSDxox, \
    REF2D_drawPixelR3_##bpp##_DPSnoan, \
    REF2D_drawPixelR3_##bpp##_DPSDnaox, \
    REF2D_drawPixelR3_##bpp##_DPan, \
    REF2D_drawPixelR3_##bpp##_PDSxa, \
    REF2D_drawPixelR3_##bpp##_DSPDSaoxxn, \
    REF2D_drawPixelR3_##bpp##_DSPDoax, \
    REF2D_drawPixelR3_##bpp##_SDPnox, \
    REF2D_drawPixelR3_##bpp##_SDPSoax, \
    REF2D_drawPixelR3_##bpp##_DSPnox, \
    REF2D_drawPixelR3_##bpp##_DSx, \
    REF2D_drawPixelR3_##bpp##_SDPSonox, \
    REF2D_drawPixelR3_##bpp##_DSPDSonoxxn, \
    REF2D_drawPixelR3_##bpp##_PDSxxn, \
    REF2D_drawPixelR3_##bpp##_DPSax, \
    REF2D_drawPixelR3_##bpp##_PSDPSoaxxn, \
    REF2D_drawPixelR3_##bpp##_SDPax, \
    REF2D_drawPixelR3_##bpp##_PDSPDoaxxn, \
    REF2D_drawPixelR3_##bpp##_SDPSnoax, \
    REF2D_drawPixelR3_##bpp##_PDSxnan, \
    REF2D_drawPixelR3_##bpp##_PDSana, \
    REF2D_drawPixelR3_##bpp##_SSDxPDxaxn, \
    REF2D_drawPixelR3_##bpp##_SDPSxox, \
    REF2D_drawPixelR3_##bpp##_SDPnoan, \
    REF2D_drawPixelR3_##bpp##_DSPDxox, \
    REF2D_drawPixelR3_##bpp##_DSPnoan, \
    REF2D_drawPixelR3_##bpp##_SDPSnaox, \
    REF2D_drawPixelR3_##bpp##_DSan, \
    REF2D_drawPixelR3_##bpp##_PDSax, \
    REF2D_drawPixelR3_##bpp##_DSPDSoaxxn, \
    REF2D_drawPixelR3_##bpp##_DPSDnoax, \
    REF2D_drawPixelR3_##bpp##_SDPxnan, \
    REF2D_drawPixelR3_##bpp##_SPDSnoax, \
    REF2D_drawPixelR3_##bpp##_DPSxnan, \
    REF2D_drawPixelR3_##bpp##_SPxDSxo, \
    REF2D_drawPixelR3_##bpp##_DPSaan,

#define DEFINE_ROP3_PIXELS_2(bpp)   \
    REF2D_drawPixelR3_##bpp##_DPSaa, \
    REF2D_drawPixelR3_##bpp##_SPxDSxon, \
    REF2D_drawPixelR3_##bpp##_DPSxna, \
    REF2D_drawPixelR3_##bpp##_SPDSnoaxn, \
    REF2D_drawPixelR3_##bpp##_SDPxna, \
    REF2D_drawPixelR3_##bpp##_PDSPnoaxn, \
    REF2D_drawPixelR3_##bpp##_DSPDSoaxx, \
    REF2D_drawPixelR3_##bpp##_PDSaxn, \
    REF2D_drawPixelR3_##bpp##_DSa, \
    REF2D_drawPixelR3_##bpp##_SDPSnaoxn, \
    REF2D_drawPixelR3_##bpp##_DSPnoa, \
    REF2D_drawPixelR3_##bpp##_DSPDxoxn, \
    REF2D_drawPixelR3_##bpp##_SDPnoa, \
    REF2D_drawPixelR3_##bpp##_SDPSxoxn, \
    REF2D_drawPixelR3_##bpp##_SSDxPDxax, \
    REF2D_drawPixelR3_##bpp##_PDSanan, \
    REF2D_drawPixelR3_##bpp##_PDSxna, \
    REF2D_drawPixelR3_##bpp##_SDPSnoaxn, \
    REF2D_drawPixelR3_##bpp##_DPSDPoaxx, \
    REF2D_drawPixelR3_##bpp##_SPDaxn, \
    REF2D_drawPixelR3_##bpp##_PSDPSoaxx, \
    REF2D_drawPixelR3_##bpp##_DPSaxn, \
    REF2D_drawPixelR3_##bpp##_DPSxx, \
    REF2D_drawPixelR3_##bpp##_PSDPSonoxx, \
    REF2D_drawPixelR3_##bpp##_SDPSonoxn, \
    REF2D_drawPixelR3_##bpp##_DSxn, \
    REF2D_drawPixelR3_##bpp##_DPSnax, \
    REF2D_drawPixelR3_##bpp##_SDPSoaxn, \
    REF2D_drawPixelR3_##bpp##_SPDnax, \
    REF2D_drawPixelR3_##bpp##_DSPDoaxn, \
    REF2D_drawPixelR3_##bpp##_DSPDSaoxx, \
    REF2D_drawPixelR3_##bpp##_PDSxan, \
    REF2D_drawPixelR3_##bpp##_DPa, \
    REF2D_drawPixelR3_##bpp##_PDSPnaoxn, \
    REF2D_drawPixelR3_##bpp##_DPSnoa, \
    REF2D_drawPixelR3_##bpp##_DPSDxoxn, \
    REF2D_drawPixelR3_##bpp##_PDSPonoxn, \
    REF2D_drawPixelR3_##bpp##_PDxn, \
    REF2D_drawPixelR3_##bpp##_DSPnax, \
    REF2D_drawPixelR3_##bpp##_PDSPoaxn, \
    REF2D_drawPixelR3_##bpp##_DPSoa, \
    REF2D_drawPixelR3_##bpp##_DPSoxn, \
    REF2D_drawPixelR3_##bpp##_D, \
    REF2D_drawPixelR3_##bpp##_DPSono, \
    REF2D_drawPixelR3_##bpp##_SPDSxax, \
    REF2D_drawPixelR3_##bpp##_DPSDaoxn, \
    REF2D_drawPixelR3_##bpp##_DSPnao, \
    REF2D_drawPixelR3_##bpp##_DPno, \
    REF2D_drawPixelR3_##bpp##_PDSnoa, \
    REF2D_drawPixelR3_##bpp##_PDSPxoxn, \
    REF2D_drawPixelR3_##bpp##_SSPxDSxox, \
    REF2D_drawPixelR3_##bpp##_SDPanan, \
    REF2D_drawPixelR3_##bpp##_PSDnax, \
    REF2D_drawPixelR3_##bpp##_DPSDoaxn, \
    REF2D_drawPixelR3_##bpp##_DPSDPaoxx, \
    REF2D_drawPixelR3_##bpp##_SDPxan, \
    REF2D_drawPixelR3_##bpp##_PSDPxax, \
    REF2D_drawPixelR3_##bpp##_DSPDaoxn, \
    REF2D_drawPixelR3_##bpp##_DPSnao, \
    REF2D_drawPixelR3_##bpp##_DSno, \
    REF2D_drawPixelR3_##bpp##_SPDSanax, \
    REF2D_drawPixelR3_##bpp##_SDxPDxan, \
    REF2D_drawPixelR3_##bpp##_DPSxo, \
    REF2D_drawPixelR3_##bpp##_DPSano, \
    REF2D_drawPixelR3_##bpp##_PSa, \
    REF2D_drawPixelR3_##bpp##_SPDSnaoxn, \
    REF2D_drawPixelR3_##bpp##_SPDSonoxn, \
    REF2D_drawPixelR3_##bpp##_PSxn, \
    REF2D_drawPixelR3_##bpp##_SPDnoa, \
    REF2D_drawPixelR3_##bpp##_SPDSxoxn, \
    REF2D_drawPixelR3_##bpp##_SDPnax, \
    REF2D_drawPixelR3_##bpp##_PSDPoaxn, \
    REF2D_drawPixelR3_##bpp##_SDPoa, \
    REF2D_drawPixelR3_##bpp##_SPDoxn, \
    REF2D_drawPixelR3_##bpp##_DPSDxax, \
    REF2D_drawPixelR3_##bpp##_SPDSaoxn, \
    REF2D_drawPixelR3_##bpp##_S, \
    REF2D_drawPixelR3_##bpp##_SDPono, \
    REF2D_drawPixelR3_##bpp##_SDPnao, \
    REF2D_drawPixelR3_##bpp##_SPno, \
    REF2D_drawPixelR3_##bpp##_PSDnoa, \
    REF2D_drawPixelR3_##bpp##_PSDPxoxn, \
    REF2D_drawPixelR3_##bpp##_PDSnax, \
    REF2D_drawPixelR3_##bpp##_SPDSoaxn, \
    REF2D_drawPixelR3_##bpp##_SSPxPDxax, \
    REF2D_drawPixelR3_##bpp##_DPSanan, \
    REF2D_drawPixelR3_##bpp##_PSDPSaoxx, \
    REF2D_drawPixelR3_##bpp##_DPSxan, \
    REF2D_drawPixelR3_##bpp##_PDSPxax, \
    REF2D_drawPixelR3_##bpp##_SDPSaoxn, \
    REF2D_drawPixelR3_##bpp##_DPSDanax, \
    REF2D_drawPixelR3_##bpp##_SPxDSxan, \
    REF2D_drawPixelR3_##bpp##_SPDnao, \
    REF2D_drawPixelR3_##bpp##_SDno, \
    REF2D_drawPixelR3_##bpp##_SDPxo, \
    REF2D_drawPixelR3_##bpp##_SDPano, \
    REF2D_drawPixelR3_##bpp##_PDSoa, \
    REF2D_drawPixelR3_##bpp##_PDSoxn, \
    REF2D_drawPixelR3_##bpp##_DSPDxax, \
    REF2D_drawPixelR3_##bpp##_PSDPaoxn, \
    REF2D_drawPixelR3_##bpp##_SDPSxax, \
    REF2D_drawPixelR3_##bpp##_PDSPaoxn, \
    REF2D_drawPixelR3_##bpp##_SDPSanax, \
    REF2D_drawPixelR3_##bpp##_SPxPDxan, \
    REF2D_drawPixelR3_##bpp##_SSPxDSxax, \
    REF2D_drawPixelR3_##bpp##_DSPDSanaxxn, \
    REF2D_drawPixelR3_##bpp##_DPSao, \
    REF2D_drawPixelR3_##bpp##_DPSxno, \
    REF2D_drawPixelR3_##bpp##_SDPao, \
    REF2D_drawPixelR3_##bpp##_SDPxno, \
    REF2D_drawPixelR3_##bpp##_DSo, \
    REF2D_drawPixelR3_##bpp##_SDPnoo, \
    REF2D_drawPixelR3_##bpp##_P, \
    REF2D_drawPixelR3_##bpp##_PDSono, \
    REF2D_drawPixelR3_##bpp##_PDSnao, \
    REF2D_drawPixelR3_##bpp##_PSno, \
    REF2D_drawPixelR3_##bpp##_PSDnao, \
    REF2D_drawPixelR3_##bpp##_PDno, \
    REF2D_drawPixelR3_##bpp##_PDSxo, \
    REF2D_drawPixelR3_##bpp##_PDSano, \
    REF2D_drawPixelR3_##bpp##_PDSao, \
    REF2D_drawPixelR3_##bpp##_PDSxno, \
    REF2D_drawPixelR3_##bpp##_DPo, \
    REF2D_drawPixelR3_##bpp##_DPSnoo, \
    REF2D_drawPixelR3_##bpp##_PSo, \
    REF2D_drawPixelR3_##bpp##_PSDnoo, \
    REF2D_drawPixelR3_##bpp##_DPSoo, \
    REF2D_drawPixelR3_##bpp##_1

#define DEFINE_ROP3_PIXELS(bpp) \
    DEFINE_ROP3_PIXELS_1(bpp)   \
    DEFINE_ROP3_PIXELS_2(bpp)   \

#define DECLARE_ROP3_MASKED_PIXELS_1(bpp)   \
    void REF2D_drawPixelR3Masked_##bpp##_0(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSoon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSona(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPona(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSxnon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPnaa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSxon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDnaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSnaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSonon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_Pn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSona(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPxnon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSxnon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDPSanaxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SSPxDSxaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPxPDxa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSanaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSPaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSxaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDPaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPDxaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSnaa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPxon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDnaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPxDSxa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSPanaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSxnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSxa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDPSaoxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSana(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SSPxPDxaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDPxox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDnoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPnaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSoox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_Sn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSxnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDPoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSxox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDnoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSonox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSnaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDnaa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSxon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDxPDxa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSanaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSnaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPDaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDPxaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPxa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSPDaoxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSDoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPana(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SSPxDSxoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSPxox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSnoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPnaon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSDaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSxaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSonon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_Dn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSPoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSDonox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSDxox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSnoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSDnaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSxa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPDSaoxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPDoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPnox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSonox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPDSonoxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDPSoaxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSPDoaxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSnoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSxnan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSana(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SSDxPDxaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSxox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPnoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPDxox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPnoan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSnaox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPDSoaxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSDnoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPxnan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSnoax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSxnan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPxDSxo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSaan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern);

#define DECLARE_ROP3_MASKED_PIXELS_2(bpp)   \
    void REF2D_drawPixelR3Masked_##bpp##_DPSaa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPxDSxon(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSxna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSnoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPxna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSPnoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPDSoaxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSnaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPnoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPDxoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPnoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSxoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SSDxPDxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSanan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSxna(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSnoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSDPoaxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDPSoaxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDPSonoxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSonoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSnax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDnax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPDoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPDSaoxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSxan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSPnaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSnoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSDxoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSPonoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPnax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSPoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_D(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSono(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSDaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPnao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSnoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSPxoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SSPxDSxox(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPanan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDnax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSDoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSDPaoxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPxan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDPxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPDaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSnao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSanax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDxPDxan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSxo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSano(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSnaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSonoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDnoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSxoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPnax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDPoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSDxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_S(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPono(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPnao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDnoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDPxoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSnax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDSoaxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SSPxPDxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSanan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDPSaoxx(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSxan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSPxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSDanax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPxDSxan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPDnao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPxo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPano(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSoa(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPDxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDPaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSPaoxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPSanax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SPxPDxan(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SSPxDSxax(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSPDSanaxxn(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSxno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPxno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DSo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_SDPnoo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_P(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSono(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSnao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDnao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSxo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSano(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSao(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PDSxno(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSnoo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_PSDnoo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_DPSoo(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern); \
    void REF2D_drawPixelR3Masked_##bpp##_1(uchar *addr DCL_X_ARG,GA_color color,GA_color pattern)

#define DECLARE_ROP3_MASKED_PIXELS(bpp) \
    DECLARE_ROP3_MASKED_PIXELS_1(bpp)   \
    DECLARE_ROP3_MASKED_PIXELS_2(bpp)   \

#define DEFINE_ROP3_MASKED_PIXELS_1(bpp)    \
    REF2D_drawPixelR3Masked_##bpp##_0, \
    REF2D_drawPixelR3Masked_##bpp##_DPSoon, \
    REF2D_drawPixelR3Masked_##bpp##_DPSona, \
    REF2D_drawPixelR3Masked_##bpp##_PSon, \
    REF2D_drawPixelR3Masked_##bpp##_SDPona, \
    REF2D_drawPixelR3Masked_##bpp##_DPon, \
    REF2D_drawPixelR3Masked_##bpp##_PDSxnon, \
    REF2D_drawPixelR3Masked_##bpp##_PDSaon, \
    REF2D_drawPixelR3Masked_##bpp##_SDPnaa, \
    REF2D_drawPixelR3Masked_##bpp##_PDSxon, \
    REF2D_drawPixelR3Masked_##bpp##_DPna, \
    REF2D_drawPixelR3Masked_##bpp##_PSDnaon, \
    REF2D_drawPixelR3Masked_##bpp##_SPna, \
    REF2D_drawPixelR3Masked_##bpp##_PDSnaon, \
    REF2D_drawPixelR3Masked_##bpp##_PDSonon, \
    REF2D_drawPixelR3Masked_##bpp##_Pn, \
    REF2D_drawPixelR3Masked_##bpp##_PDSona, \
    REF2D_drawPixelR3Masked_##bpp##_DSon, \
    REF2D_drawPixelR3Masked_##bpp##_SDPxnon, \
    REF2D_drawPixelR3Masked_##bpp##_SDPaon, \
    REF2D_drawPixelR3Masked_##bpp##_DPSxnon, \
    REF2D_drawPixelR3Masked_##bpp##_DPSaon, \
    REF2D_drawPixelR3Masked_##bpp##_PSDPSanaxx, \
    REF2D_drawPixelR3Masked_##bpp##_SSPxDSxaxn, \
    REF2D_drawPixelR3Masked_##bpp##_SPxPDxa, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSanaxn, \
    REF2D_drawPixelR3Masked_##bpp##_PDSPaox, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSxaxn, \
    REF2D_drawPixelR3Masked_##bpp##_PSDPaox, \
    REF2D_drawPixelR3Masked_##bpp##_DSPDxaxn, \
    REF2D_drawPixelR3Masked_##bpp##_PDSox, \
    REF2D_drawPixelR3Masked_##bpp##_PDSoan, \
    REF2D_drawPixelR3Masked_##bpp##_DPSnaa, \
    REF2D_drawPixelR3Masked_##bpp##_SDPxon, \
    REF2D_drawPixelR3Masked_##bpp##_DSna, \
    REF2D_drawPixelR3Masked_##bpp##_SPDnaon, \
    REF2D_drawPixelR3Masked_##bpp##_SPxDSxa, \
    REF2D_drawPixelR3Masked_##bpp##_PDSPanaxn, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSaox, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSxnox, \
    REF2D_drawPixelR3Masked_##bpp##_DPSxa, \
    REF2D_drawPixelR3Masked_##bpp##_PSDPSaoxxn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSana, \
    REF2D_drawPixelR3Masked_##bpp##_SSPxPDxaxn, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSoax, \
    REF2D_drawPixelR3Masked_##bpp##_PSDnox, \
    REF2D_drawPixelR3Masked_##bpp##_PSDPxox, \
    REF2D_drawPixelR3Masked_##bpp##_PSDnoan, \
    REF2D_drawPixelR3Masked_##bpp##_PSna, \
    REF2D_drawPixelR3Masked_##bpp##_SDPnaon, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSoox, \
    REF2D_drawPixelR3Masked_##bpp##_Sn, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSaox, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSxnox, \
    REF2D_drawPixelR3Masked_##bpp##_SDPox, \
    REF2D_drawPixelR3Masked_##bpp##_SDPoan, \
    REF2D_drawPixelR3Masked_##bpp##_PSDPoax, \
    REF2D_drawPixelR3Masked_##bpp##_SPDnox, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSxox, \
    REF2D_drawPixelR3Masked_##bpp##_SPDnoan, \
    REF2D_drawPixelR3Masked_##bpp##_PSx, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSonox, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSnaox, \
    REF2D_drawPixelR3Masked_##bpp##_PSan, \
    REF2D_drawPixelR3Masked_##bpp##_PSDnaa, \
    REF2D_drawPixelR3Masked_##bpp##_DPSxon, \
    REF2D_drawPixelR3Masked_##bpp##_SDxPDxa, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSanaxn, \
    REF2D_drawPixelR3Masked_##bpp##_SDna, \
    REF2D_drawPixelR3Masked_##bpp##_DPSnaon, \
    REF2D_drawPixelR3Masked_##bpp##_DSPDaox, \
    REF2D_drawPixelR3Masked_##bpp##_PSDPxaxn, \
    REF2D_drawPixelR3Masked_##bpp##_SDPxa, \
    REF2D_drawPixelR3Masked_##bpp##_PDSPDaoxxn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSDoax, \
    REF2D_drawPixelR3Masked_##bpp##_PDSnox, \
    REF2D_drawPixelR3Masked_##bpp##_SDPana, \
    REF2D_drawPixelR3Masked_##bpp##_SSPxDSxoxn, \
    REF2D_drawPixelR3Masked_##bpp##_PDSPxox, \
    REF2D_drawPixelR3Masked_##bpp##_PDSnoan, \
    REF2D_drawPixelR3Masked_##bpp##_PDna, \
    REF2D_drawPixelR3Masked_##bpp##_DSPnaon, \
    REF2D_drawPixelR3Masked_##bpp##_DPSDaox, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSxaxn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSonon, \
    REF2D_drawPixelR3Masked_##bpp##_Dn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSox, \
    REF2D_drawPixelR3Masked_##bpp##_DPSoan, \
    REF2D_drawPixelR3Masked_##bpp##_PDSPoax, \
    REF2D_drawPixelR3Masked_##bpp##_DPSnox, \
    REF2D_drawPixelR3Masked_##bpp##_DPx, \
    REF2D_drawPixelR3Masked_##bpp##_DPSDonox, \
    REF2D_drawPixelR3Masked_##bpp##_DPSDxox, \
    REF2D_drawPixelR3Masked_##bpp##_DPSnoan, \
    REF2D_drawPixelR3Masked_##bpp##_DPSDnaox, \
    REF2D_drawPixelR3Masked_##bpp##_DPan, \
    REF2D_drawPixelR3Masked_##bpp##_PDSxa, \
    REF2D_drawPixelR3Masked_##bpp##_DSPDSaoxxn, \
    REF2D_drawPixelR3Masked_##bpp##_DSPDoax, \
    REF2D_drawPixelR3Masked_##bpp##_SDPnox, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSoax, \
    REF2D_drawPixelR3Masked_##bpp##_DSPnox, \
    REF2D_drawPixelR3Masked_##bpp##_DSx, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSonox, \
    REF2D_drawPixelR3Masked_##bpp##_DSPDSonoxxn, \
    REF2D_drawPixelR3Masked_##bpp##_PDSxxn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSax, \
    REF2D_drawPixelR3Masked_##bpp##_PSDPSoaxxn, \
    REF2D_drawPixelR3Masked_##bpp##_SDPax, \
    REF2D_drawPixelR3Masked_##bpp##_PDSPDoaxxn, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSnoax, \
    REF2D_drawPixelR3Masked_##bpp##_PDSxnan, \
    REF2D_drawPixelR3Masked_##bpp##_PDSana, \
    REF2D_drawPixelR3Masked_##bpp##_SSDxPDxaxn, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSxox, \
    REF2D_drawPixelR3Masked_##bpp##_SDPnoan, \
    REF2D_drawPixelR3Masked_##bpp##_DSPDxox, \
    REF2D_drawPixelR3Masked_##bpp##_DSPnoan, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSnaox, \
    REF2D_drawPixelR3Masked_##bpp##_DSan, \
    REF2D_drawPixelR3Masked_##bpp##_PDSax, \
    REF2D_drawPixelR3Masked_##bpp##_DSPDSoaxxn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSDnoax, \
    REF2D_drawPixelR3Masked_##bpp##_SDPxnan, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSnoax, \
    REF2D_drawPixelR3Masked_##bpp##_DPSxnan, \
    REF2D_drawPixelR3Masked_##bpp##_SPxDSxo, \
    REF2D_drawPixelR3Masked_##bpp##_DPSaan,

#define DEFINE_ROP3_MASKED_PIXELS_2(bpp)    \
    REF2D_drawPixelR3Masked_##bpp##_DPSaa, \
    REF2D_drawPixelR3Masked_##bpp##_SPxDSxon, \
    REF2D_drawPixelR3Masked_##bpp##_DPSxna, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSnoaxn, \
    REF2D_drawPixelR3Masked_##bpp##_SDPxna, \
    REF2D_drawPixelR3Masked_##bpp##_PDSPnoaxn, \
    REF2D_drawPixelR3Masked_##bpp##_DSPDSoaxx, \
    REF2D_drawPixelR3Masked_##bpp##_PDSaxn, \
    REF2D_drawPixelR3Masked_##bpp##_DSa, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSnaoxn, \
    REF2D_drawPixelR3Masked_##bpp##_DSPnoa, \
    REF2D_drawPixelR3Masked_##bpp##_DSPDxoxn, \
    REF2D_drawPixelR3Masked_##bpp##_SDPnoa, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSxoxn, \
    REF2D_drawPixelR3Masked_##bpp##_SSDxPDxax, \
    REF2D_drawPixelR3Masked_##bpp##_PDSanan, \
    REF2D_drawPixelR3Masked_##bpp##_PDSxna, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSnoaxn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSDPoaxx, \
    REF2D_drawPixelR3Masked_##bpp##_SPDaxn, \
    REF2D_drawPixelR3Masked_##bpp##_PSDPSoaxx, \
    REF2D_drawPixelR3Masked_##bpp##_DPSaxn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSxx, \
    REF2D_drawPixelR3Masked_##bpp##_PSDPSonoxx, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSonoxn, \
    REF2D_drawPixelR3Masked_##bpp##_DSxn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSnax, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSoaxn, \
    REF2D_drawPixelR3Masked_##bpp##_SPDnax, \
    REF2D_drawPixelR3Masked_##bpp##_DSPDoaxn, \
    REF2D_drawPixelR3Masked_##bpp##_DSPDSaoxx, \
    REF2D_drawPixelR3Masked_##bpp##_PDSxan, \
    REF2D_drawPixelR3Masked_##bpp##_DPa, \
    REF2D_drawPixelR3Masked_##bpp##_PDSPnaoxn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSnoa, \
    REF2D_drawPixelR3Masked_##bpp##_DPSDxoxn, \
    REF2D_drawPixelR3Masked_##bpp##_PDSPonoxn, \
    REF2D_drawPixelR3Masked_##bpp##_PDxn, \
    REF2D_drawPixelR3Masked_##bpp##_DSPnax, \
    REF2D_drawPixelR3Masked_##bpp##_PDSPoaxn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSoa, \
    REF2D_drawPixelR3Masked_##bpp##_DPSoxn, \
    REF2D_drawPixelR3Masked_##bpp##_D, \
    REF2D_drawPixelR3Masked_##bpp##_DPSono, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSxax, \
    REF2D_drawPixelR3Masked_##bpp##_DPSDaoxn, \
    REF2D_drawPixelR3Masked_##bpp##_DSPnao, \
    REF2D_drawPixelR3Masked_##bpp##_DPno, \
    REF2D_drawPixelR3Masked_##bpp##_PDSnoa, \
    REF2D_drawPixelR3Masked_##bpp##_PDSPxoxn, \
    REF2D_drawPixelR3Masked_##bpp##_SSPxDSxox, \
    REF2D_drawPixelR3Masked_##bpp##_SDPanan, \
    REF2D_drawPixelR3Masked_##bpp##_PSDnax, \
    REF2D_drawPixelR3Masked_##bpp##_DPSDoaxn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSDPaoxx, \
    REF2D_drawPixelR3Masked_##bpp##_SDPxan, \
    REF2D_drawPixelR3Masked_##bpp##_PSDPxax, \
    REF2D_drawPixelR3Masked_##bpp##_DSPDaoxn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSnao, \
    REF2D_drawPixelR3Masked_##bpp##_DSno, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSanax, \
    REF2D_drawPixelR3Masked_##bpp##_SDxPDxan, \
    REF2D_drawPixelR3Masked_##bpp##_DPSxo, \
    REF2D_drawPixelR3Masked_##bpp##_DPSano, \
    REF2D_drawPixelR3Masked_##bpp##_PSa, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSnaoxn, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSonoxn, \
    REF2D_drawPixelR3Masked_##bpp##_PSxn, \
    REF2D_drawPixelR3Masked_##bpp##_SPDnoa, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSxoxn, \
    REF2D_drawPixelR3Masked_##bpp##_SDPnax, \
    REF2D_drawPixelR3Masked_##bpp##_PSDPoaxn, \
    REF2D_drawPixelR3Masked_##bpp##_SDPoa, \
    REF2D_drawPixelR3Masked_##bpp##_SPDoxn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSDxax, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSaoxn, \
    REF2D_drawPixelR3Masked_##bpp##_S, \
    REF2D_drawPixelR3Masked_##bpp##_SDPono, \
    REF2D_drawPixelR3Masked_##bpp##_SDPnao, \
    REF2D_drawPixelR3Masked_##bpp##_SPno, \
    REF2D_drawPixelR3Masked_##bpp##_PSDnoa, \
    REF2D_drawPixelR3Masked_##bpp##_PSDPxoxn, \
    REF2D_drawPixelR3Masked_##bpp##_PDSnax, \
    REF2D_drawPixelR3Masked_##bpp##_SPDSoaxn, \
    REF2D_drawPixelR3Masked_##bpp##_SSPxPDxax, \
    REF2D_drawPixelR3Masked_##bpp##_DPSanan, \
    REF2D_drawPixelR3Masked_##bpp##_PSDPSaoxx, \
    REF2D_drawPixelR3Masked_##bpp##_DPSxan, \
    REF2D_drawPixelR3Masked_##bpp##_PDSPxax, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSaoxn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSDanax, \
    REF2D_drawPixelR3Masked_##bpp##_SPxDSxan, \
    REF2D_drawPixelR3Masked_##bpp##_SPDnao, \
    REF2D_drawPixelR3Masked_##bpp##_SDno, \
    REF2D_drawPixelR3Masked_##bpp##_SDPxo, \
    REF2D_drawPixelR3Masked_##bpp##_SDPano, \
    REF2D_drawPixelR3Masked_##bpp##_PDSoa, \
    REF2D_drawPixelR3Masked_##bpp##_PDSoxn, \
    REF2D_drawPixelR3Masked_##bpp##_DSPDxax, \
    REF2D_drawPixelR3Masked_##bpp##_PSDPaoxn, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSxax, \
    REF2D_drawPixelR3Masked_##bpp##_PDSPaoxn, \
    REF2D_drawPixelR3Masked_##bpp##_SDPSanax, \
    REF2D_drawPixelR3Masked_##bpp##_SPxPDxan, \
    REF2D_drawPixelR3Masked_##bpp##_SSPxDSxax, \
    REF2D_drawPixelR3Masked_##bpp##_DSPDSanaxxn, \
    REF2D_drawPixelR3Masked_##bpp##_DPSao, \
    REF2D_drawPixelR3Masked_##bpp##_DPSxno, \
    REF2D_drawPixelR3Masked_##bpp##_SDPao, \
    REF2D_drawPixelR3Masked_##bpp##_SDPxno, \
    REF2D_drawPixelR3Masked_##bpp##_DSo, \
    REF2D_drawPixelR3Masked_##bpp##_SDPnoo, \
    REF2D_drawPixelR3Masked_##bpp##_P, \
    REF2D_drawPixelR3Masked_##bpp##_PDSono, \
    REF2D_drawPixelR3Masked_##bpp##_PDSnao, \
    REF2D_drawPixelR3Masked_##bpp##_PSno, \
    REF2D_drawPixelR3Masked_##bpp##_PSDnao, \
    REF2D_drawPixelR3Masked_##bpp##_PDno, \
    REF2D_drawPixelR3Masked_##bpp##_PDSxo, \
    REF2D_drawPixelR3Masked_##bpp##_PDSano, \
    REF2D_drawPixelR3Masked_##bpp##_PDSao, \
    REF2D_drawPixelR3Masked_##bpp##_PDSxno, \
    REF2D_drawPixelR3Masked_##bpp##_DPo, \
    REF2D_drawPixelR3Masked_##bpp##_DPSnoo, \
    REF2D_drawPixelR3Masked_##bpp##_PSo, \
    REF2D_drawPixelR3Masked_##bpp##_PSDnoo, \
    REF2D_drawPixelR3Masked_##bpp##_DPSoo, \
    REF2D_drawPixelR3Masked_##bpp##_1

#define DEFINE_ROP3_MASKED_PIXELS(bpp)  \
    DEFINE_ROP3_MASKED_PIXELS_1(bpp)    \
    DEFINE_ROP3_MASKED_PIXELS_2(bpp)    \

#define DECLARE_BLENDED_PIXELS(bpp) \
    void REF2D_drawPixelBlendedFast_##bpp(uchar *addr DCL_X_ARG,GA_color color,GA_color alpha); \
    void REF2D_drawPixelBlendedSrcAlphaFast_##bpp(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelBlendedConstAlphaFast_##bpp(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelBlended_##bpp(uchar *addr DCL_X_ARG,GA_color color); \
    void REF2D_drawPixelBlendedClamped_##bpp(uchar *addr DCL_X_ARG,GA_color color)

#define DEFINE_BLENDED_PIXELS(bpp) \
static REF2D_drawPixel_t blendPixelFuncs##bpp[] = { \
    REF2D_drawPixelBlendedSrcAlphaFast_##bpp, \
    REF2D_drawPixelBlendedConstAlphaFast_##bpp, \
    REF2D_drawPixelBlended_##bpp, \
    REF2D_drawPixelBlendedClamped_##bpp, \
    }

// Macro to setup the internal state and drawing vectors to point to the
// REF2D functions directly where we don't need to do anything special for
// filter drivers.

#define INIT_FILTER_VECTORS                                             \
    SET_STATE2D_VECTOR(IsIdle);                                         \
    SET_STATE2D_VECTOR(WaitTillIdle);                                   \
    SET_STATE2D_VECTOR(EnableDirectAccess);                             \
    SET_STATE2D_VECTOR(DisableDirectAccess);                            \
    SET_STATE2D_VECTOR(SetMix);                                         \
    SET_STATE2D_VECTOR(SetForeColor);                                   \
    SET_STATE2D_VECTOR(SetBackColor);                                   \
    SET_STATE2D_VECTOR(Set8x8MonoPattern);                              \
    SET_STATE2D_VECTOR(Use8x8MonoPattern);                              \
    SET_STATE2D_VECTOR(Use8x8TransMonoPattern);                         \
    SET_STATE2D_VECTOR(Set8x8ColorPattern);                             \
    SET_STATE2D_VECTOR(Use8x8ColorPattern);                             \
    SET_STATE2D_VECTOR(Use8x8TransColorPattern);                        \
    SET_STATE2D_VECTOR(SetLineStipple);                                 \
    SET_STATE2D_VECTOR(SetLineStippleCount);                            \
    SET_STATE2D_VECTOR(SetPlaneMask);                                   \
    SET_STATE2D_VECTOR(SetBlendFunc);                                   \
    SET_STATE2D_VECTOR(SetAlphaValue);                                  \
    SET_STATE2D_VECTOR(SetLineStyle);                                   \
                                                                        \
    SET_DRAW2D_VECTOR(GetPixel);                                        \
    SET_DRAW2D_VECTOR(PutPixel);                                        \
    SET_DRAW2D_VECTOR(DrawEllipse);                                     \
    SET_DRAW2D_VECTOR(ClipEllipse);                                     \
    SET_DRAW2D_VECTOR(DrawScanList);                                    \
    SET_DRAW2D_VECTOR(DrawPattScanList);                                \
    SET_DRAW2D_VECTOR(DrawColorPattScanList);                           \
    SET_DRAW2D_VECTOR(DrawEllipseList);                                 \
    SET_DRAW2D_VECTOR(DrawPattEllipseList);                             \
    SET_DRAW2D_VECTOR(DrawColorPattEllipseList);                        \
    SET_DRAW2D_VECTOR(DrawFatEllipseList);                              \
    SET_DRAW2D_VECTOR(DrawPattFatEllipseList);                          \
    SET_DRAW2D_VECTOR(DrawColorPattFatEllipseList);                     \
    SET_DRAW2D_VECTOR(DrawRect);                                        \
    SET_DRAW2D_VECTOR(DrawPattRect);                                    \
    SET_DRAW2D_VECTOR(DrawColorPattRect);                               \
    SET_DRAW2D_VECTOR(DrawTrap);                                        \
    SET_DRAW2D_VECTOR(DrawPattTrap);                                    \
    SET_DRAW2D_VECTOR(DrawColorPattTrap);                               \
    SET_DRAW2D_VECTOR(DrawLineInt);                                     \
    SET_DRAW2D_VECTOR(DrawBresenhamLine);                               \
    SET_DRAW2D_VECTOR(DrawStippleLineInt);                              \
    SET_DRAW2D_VECTOR(DrawBresenhamStippleLine);                        \
    SET_DRAW2D_VECTOR(PutMonoImageMSBSys);                              \
    SET_DRAW2D_VECTOR(PutMonoImageMSBLin);                              \
    SET_DRAW2D_VECTOR(PutMonoImageMSBBM);                               \
    SET_DRAW2D_VECTOR(PutMonoImageLSBSys);                              \
    SET_DRAW2D_VECTOR(PutMonoImageLSBLin);                              \
    SET_DRAW2D_VECTOR(PutMonoImageLSBBM);                               \
    SET_DRAW2D_VECTOR(ClipMonoImageMSBSys);                             \
    SET_DRAW2D_VECTOR(ClipMonoImageMSBLin);                             \
    SET_DRAW2D_VECTOR(ClipMonoImageMSBBM);                              \
    SET_DRAW2D_VECTOR(ClipMonoImageLSBSys);                             \
    SET_DRAW2D_VECTOR(ClipMonoImageLSBLin);                             \
    SET_DRAW2D_VECTOR(ClipMonoImageLSBBM);                              \
    SET_DRAW2D_VECTOR(BitBlt);                                          \
    SET_DRAW2D_VECTOR(BitBltLin);                                       \
    SET_DRAW2D_VECTOR(BitBltSys);                                       \
    SET_DRAW2D_VECTOR(BitBltBM);                                        \
    SET_DRAW2D_VECTOR(BitBltPatt);                                      \
    SET_DRAW2D_VECTOR(BitBltPattLin);                                   \
    SET_DRAW2D_VECTOR(BitBltPattSys);                                   \
    SET_DRAW2D_VECTOR(BitBltPattBM);                                    \
    SET_DRAW2D_VECTOR(BitBltColorPatt);                                 \
    SET_DRAW2D_VECTOR(BitBltColorPattLin);                              \
    SET_DRAW2D_VECTOR(BitBltColorPattSys);                              \
    SET_DRAW2D_VECTOR(BitBltColorPattBM);                               \
    SET_DRAW2D_VECTOR(SrcTransBlt);                                     \
    SET_DRAW2D_VECTOR(SrcTransBltLin);                                  \
    SET_DRAW2D_VECTOR(SrcTransBltSys);                                  \
    SET_DRAW2D_VECTOR(SrcTransBltBM);                                   \
    SET_DRAW2D_VECTOR(DstTransBlt);                                     \
    SET_DRAW2D_VECTOR(DstTransBltLin);                                  \
    SET_DRAW2D_VECTOR(DstTransBltSys);                                  \
    SET_DRAW2D_VECTOR(DstTransBltBM);                                   \
    SET_DRAW2D_VECTOR(StretchBlt);                                      \
    SET_DRAW2D_VECTOR(StretchBltLin);                                   \
    SET_DRAW2D_VECTOR(StretchBltSys);                                   \
    SET_DRAW2D_VECTOR(StretchBltBM);                                    \
    SET_DRAW2D_VECTOR(ConvertBltSys_Obsolete);                                   \
    SET_DRAW2D_VECTOR(ConvertBltBM_Obsolete);                                    \
    SET_DRAW2D_VECTOR(StretchConvertBltSys_Obsolete);                            \
    SET_DRAW2D_VECTOR(StretchConvertBltBM_Obsolete);                             \
    SET_DRAW2D_VECTOR(BitBltFxTest);                                    \
    SET_DRAW2D_VECTOR(BitBltFx);                                        \
    SET_DRAW2D_VECTOR(BitBltFxLin);                                     \
    SET_DRAW2D_VECTOR(BitBltFxSys);                                     \
    SET_DRAW2D_VECTOR(BitBltFxBM);                                      \
    SET_DRAW2D_VECTOR(GetBitmapSys);                                    \
    SET_DRAW2D_VECTOR(GetBitmapBM);                                     \
    SET_DRAW2D_VECTOR(DrawClippedLineInt);                              \
    SET_DRAW2D_VECTOR(DrawClippedBresenhamLine);                        \
    SET_DRAW2D_VECTOR(DrawClippedStippleLineInt);                       \
    SET_DRAW2D_VECTOR(DrawClippedBresenhamStippleLine);                 \
    SET_DRAW2D_VECTOR(BitBltPlaneMasked);                               \
    SET_DRAW2D_VECTOR(BitBltPlaneMaskedLin);                            \
    SET_DRAW2D_VECTOR(BitBltPlaneMaskedSys);                            \
    SET_DRAW2D_VECTOR(BitBltPlaneMaskedBM);                             \
    SET_DRAW2D_VECTOR(DrawRectLin);                                     \
    SET_DRAW2D_VECTOR(DrawRectExt);                                     \
    SET_DRAW2D_VECTOR(DrawStyleLineInt);                                \
    SET_DRAW2D_VECTOR(DrawBresenhamStyleLine);                          \
    SET_DRAW2D_VECTOR(DrawClippedStyleLineInt);                         \
    SET_DRAW2D_VECTOR(DrawClippedBresenhamStyleLine);

// Macro to setup the internal cursor vectors to point to the REF2D
// functions directly where we don't need to do anything special for
// filter drivers.

#define INIT_CURSOR_VECTORS                                             \
    SET_CURSOR_VECTOR(SetMonoCursor);                                   \
    SET_CURSOR_VECTOR(SetMonoCursorColor);                              \
    SET_CURSOR_VECTOR(SetColorCursor);                                  \
    SET_CURSOR_VECTOR(SetCursorPos);                                    \
    SET_CURSOR_VECTOR(ShowCursor);                                      \
    SET_CURSOR_VECTOR(BeginAccess);                                     \
    SET_CURSOR_VECTOR(EndAccess);                                       \
    SET_CURSOR_VECTOR(SetColorCursor256);                               \
    SET_CURSOR_VECTOR(SetColorCursorRGBA);                              \
    SET_CURSOR_VECTOR(IsHardwareCursor);

// Macro to setup the internal driver vectors to point to the REF2D
// functions directly where we don't need to do anything special for
// filter drivers.

#define INIT_INIT_VECTORS                                               \
    SET_INIT_VECTOR(GetConfigInfo);                                     \
    SET_INIT_VECTOR(SetModeProfile);                                    \
    SET_INIT_VECTOR(GetOptions);                                        \
    SET_INIT_VECTOR(SetOptions);                                        \
    SET_INIT_VECTOR(GetVideoModeInfo);                                  \
    SET_INIT_VECTOR(SetVideoMode);                                      \
    SET_INIT_VECTOR(GetVideoMode);                                      \
    SET_INIT_VECTOR(GetCustomVideoModeInfo);                            \
    SET_INIT_VECTOR(SetCustomVideoMode);                                \
    SET_INIT_VECTOR(GetClosestPixelClock);                              \
    SET_INIT_VECTOR(GetCRTCTimings);                                    \
    SET_INIT_VECTOR(SetCRTCTimings);                                    \
    SET_INIT_VECTOR(SetGlobalRefresh);                                  \
    SET_INIT_VECTOR(SaveRestoreState);                                  \
    SET_INIT_VECTOR(SetDisplayOutput);                                  \
    SET_INIT_VECTOR(GetDisplayOutput);                                  \
    SET_INIT_VECTOR(SetSoftwareRenderFuncs);                            \
    SET_INIT_VECTOR(GetUniqueFilename);                                 \
    SET_INIT_VECTOR(GetMonitorInfo);                                    \
    SET_INIT_VECTOR(SetMonitorInfo);                                    \
    SET_INIT_VECTOR(GetCurrentVideoModeInfo);                           \
    SET_INIT_VECTOR(GetCertifyInfo);                                    \
    SET_INIT_VECTOR(SetCRTCTimings);                                    \
    SET_INIT_VECTOR(AlignLinearBuffer);                                 \
    SET_INIT_VECTOR(GetCurrentRefreshRate);                             \
    SET_INIT_VECTOR(PollForDisplaySwitch);                              \
    SET_INIT_VECTOR(PerformDisplaySwitch);                              \
    SET_INIT_VECTOR(GetVideoModeInfoExt);                               \
    SET_INIT_VECTOR(GetCustomVideoModeInfoExt);                         \
    SET_INIT_VECTOR(SwitchPhysicalResolution);                          \
    SET_INIT_VECTOR(GetNumberOfHeads);                                  \
    SET_INIT_VECTOR(SetActiveHead);                                     \
    SET_INIT_VECTOR(GetActiveHead);                                     \
    SET_INIT_VECTOR(SetRef2dPointer);

// Macro to setup the internal driver vectors to point to the REF2D
// functions directly where we don't need to do anything special for
// filter drivers.

#define INIT_DRIVER_VECTORS                                             \
    SET_DRIVER_VECTOR(SetBank);                                         \
    SET_DRIVER_VECTOR(SetDisplayStart);                                 \
    SET_DRIVER_VECTOR(SetDisplayStartXY);                               \
    SET_DRIVER_VECTOR(SetStereoDisplayStart);                           \
    SET_DRIVER_VECTOR(GetDisplayStartStatus);                           \
    SET_DRIVER_VECTOR(EnableStereoMode);                                \
    SET_DRIVER_VECTOR(IsVSync);                                         \
    SET_DRIVER_VECTOR(WaitVSync);                                       \
    SET_DRIVER_VECTOR(SetVSyncWidth);                                   \
    SET_DRIVER_VECTOR(GetVSyncWidth);                                   \
    SET_DRIVER_VECTOR(SetPaletteData);                                  \
    SET_DRIVER_VECTOR(GetPaletteData);                                  \
    SET_DRIVER_VECTOR(GetGammaCorrectData);                             \
    SET_DRIVER_VECTOR(GetGammaCorrectData);                             \
    SET_DRIVER_VECTOR(GetCurrentScanLine);                              \
    SET_DRIVER_VECTOR(SetPaletteDataExt);                               \
    SET_DRIVER_VECTOR(GetPaletteDataExt);                               \
    SET_DRIVER_VECTOR(SetGammaCorrectDataExt);                          \
    SET_DRIVER_VECTOR(GetGammaCorrectDataExt);

/*---------------------------- Global Variables ---------------------------*/

extern ibool            REF_slowBltSys;
extern ibool            REF_slowBltSysOrg;
extern ibool            REF_noHWPlaneMask;
extern int              REF_mix;
extern GA_2DRenderFuncs REF_draw2dCnt;
extern GA_2DRenderFuncs REF_draw2dHW;
extern GA_2DRenderFuncs REF_draw2dHWSW;
extern REF2D_state      REF2D_devCtx;
extern REF2D_driver     REF2D_drv;

#define DC  REF2D_devCtx
#define DV  REF2D_drv

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

/* Library initialisation entry points */

ibool       NAPI REF2D_QueryFunctionsSW(N_uint32 id,void *funcs);
N_int32     NAPI REF2D_SetDrawBufferSW(GA_buffer *drawBuf,void *framebuffer,N_int32 bitsPerPixel,GA_pixelFormat *pf,GA_devCtx *hwCtx,N_int32 softwareOnly);
void        NAPI REF2D_SetDrawSurface(void *surface,N_int32 xRes,N_int32 yRes,N_int32 bytesPerLine,N_int32 bitsPerPixel,GA_pixelFormat *pf);
void        NAPI REF2D_UnloadDriverInt(void);
void        REF2D_resizeBltBuffers(N_int32 width);
ibool       NAPI REF2D_InitBuffers(N_int32 numBuffers,N_uint32 flags,GA_softStereoFuncs *stereo);
void        NAPI REF2D_FlipToBuffer(GA_buf *buf,N_int32 waitVRT);
N_int32     NAPI REF2D_ForceSoftwareOnly(N_int32 enable);
N_uint32    NAPI REF2D_SetColorCompareMask(N_uint32 mask);
void        NAPI REF2D_PostSwitchPhysicalResolution(void);

/* Generic software cursor functions in cursor.c */

void        REF2D_cursorInit(void);
void        NAPI REF2D_SetMonoCursor(GA_monoCursor *cursor);
void        NAPI REF2D_SetMonoCursorColor(GA_palette *foreground,GA_palette *background);
void        NAPI REF2D_SetColorCursor(GA_colorCursor *cursor);
void        NAPI REF2D_SetColorCursor256(GA_colorCursor256 *cursor);
void        NAPI REF2D_SetColorCursorRGB(GA_colorCursorRGB *cursor);
void        NAPI REF2D_SetColorCursorRGBA(GA_colorCursorRGBA *cursor);
N_int32     NAPI REF2D_SetCursorPos(N_int32 x,N_int32 y);
void        NAPI REF2D_ShowCursor(N_int32 visible);
void        NAPI REF2D_BeginAccess(N_int32 left,N_int32 top,N_int32 right,N_int32 bottom);
void        NAPI REF2D_EndAccess(void);
N_int32     NAPI REF2D_IsHardwareCursor(void);

/* Software line clipper */

int         REF2D_clipLine(N_int32 xmin,N_int32 ymin,N_int32 xmax,N_int32 ymax,N_int32 *new_x1,N_int32 *new_y1,N_int32 *new_x2,N_int32 *new_y2,N_int32 absDeltaX,N_int32 absDeltaY,N_int32 *pt1_clipped,N_int32 *pt2_clipped,N_int32 flags,N_int32 bias,N_int32 outcode1,N_int32 outcode2);

/* Driver function hooks in cursor.c */

void        NAPI REF2D_SetDisplayStart(N_int32 offset,N_int32 waitVRT);
void        NAPI REF2D_SetDisplayStartXY(N_int32 x,N_int32 y,N_int32 waitVRT);
void        NAPI REF2D_SetPaletteData(GA_palette *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI REF2D_SetPaletteDataExt(GA_paletteExt *pal,N_int32 num,N_int32 index,N_int32 waitVRT);

/* 2D state reference functions */

N_int32     NAPI REF2D_SetDrawBuffer(GA_buffer *drawBuf);
N_int32     NAPI REF2D_IsIdle(void);
void        NAPI REF2D_WaitTillIdle(void);
GA_buffer * NAPI REF2D_GetDrawBuffer(void);
N_int32     NAPI REF2D_SetMix(N_int32 mix);
void        NAPI REF2D_SetForeColor(GA_color color);
void        NAPI REF2D_SetBackColor(GA_color color);
void        NAPI REF2D_Set8x8MonoPattern(N_int32 index,GA_pattern *pattern);
void        NAPI REF2D_Use8x8MonoPattern(N_int32 index);
void        NAPI REF2D_Use8x8TransMonoPattern(N_int32 index);
void        NAPI REF2D_Set8x8ColorPattern(N_int32 index,GA_colorPattern *pattern);
void        NAPI REF2D_Use8x8ColorPattern(N_int32 index);
void        NAPI REF2D_Use8x8TransColorPattern(N_int32 index,GA_color transparent);
void        NAPI REF2D_SetLineStipple(GA_stipple stipple);
void        NAPI REF2D_SetLineStippleCount(N_uint32 count);
void        NAPI REF2D_SetPlaneMask(N_uint32 mask);
void        NAPI REF2D_SetSrcBlendFunc(N_int32 blendFunc);
void        NAPI REF2D_SetDstBlendFunc(N_int32 blendFunc);
void        NAPI REF2D_SetBlendFunc(N_int32 srcBlendFunc,N_int32 dstBlendFunc);
void        NAPI REF2D_SetAlphaValue(N_uint8 alpha);
void        NAPI REF2D_SetLineStyle(N_uint32 styleMask,N_uint32 styleStep,N_uint32 styleValue);
void        NAPI REF2D_BuildTranslateVector(GA_color *translate,GA_palette *dstPal,GA_palette *srcPal,int srcColors);

/* 2D generic reference functions */

void        NAPI REF2D_DrawScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI REF2D_DrawPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI REF2D_DrawColorPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI REF2D_DrawEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI REF2D_DrawPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI REF2D_DrawColorPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI REF2D_DrawFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI REF2D_DrawPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI REF2D_DrawColorPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI REF2D_DrawRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI REF2D_DrawRectExt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI REF2D_DrawRectExtHW(N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI REF2D_DrawRectExtSW(void *buffer,N_int32 dstPitch,N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI REF2D_DrawRectLin(N_int32 dstOfs,N_int32 dstPitch,N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI REF2D_DrawPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI REF2D_DrawColorPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI REF2D_DrawRectPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI REF2D_DrawRectColorPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI REF2D_DrawTrap(GA_trap *trap);
void        NAPI REF2D_DrawPattTrap(GA_trap *trap);
void        NAPI REF2D_DrawColorPattTrap(GA_trap *trap);
void        NAPI REF2D_DrawLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void        NAPI REF2D_DrawBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
void        NAPI REF2D_DrawStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI REF2D_DrawBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
void        NAPI REF2D_PutMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI REF2D_PutMonoImageMSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent);
void        NAPI REF2D_PutMonoImageMSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 imagePhysAddr,N_int32 transparent);
void        NAPI REF2D_PutMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI REF2D_PutMonoImageLSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent);
void        NAPI REF2D_PutMonoImageLSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 imagePhysAddr,N_int32 transparent);
void        NAPI REF2D_ClipMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI REF2D_ClipMonoImageMSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI REF2D_ClipMonoImageMSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 imagePhysAddr,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI REF2D_ClipMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI REF2D_ClipMonoImageLSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI REF2D_ClipMonoImageLSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 imagePhysAddr,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI REF2D_BitBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI REF2D_BitBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI REF2D_BitBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
void        NAPI REF2D_BitBltBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI REF2D_BitBltPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI REF2D_BitBltPattLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI REF2D_BitBltPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI REF2D_BitBltPattBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI REF2D_BitBltColorPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI REF2D_BitBltColorPattLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI REF2D_BitBltColorPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI REF2D_BitBltColorPattBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI REF2D_SrcTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI REF2D_SrcTransBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI REF2D_SrcTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI REF2D_SrcTransBltBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI REF2D_DstTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI REF2D_DstTransBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI REF2D_DstTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI REF2D_DstTransBltBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI REF2D_StretchBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI REF2D_StretchBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI REF2D_StretchBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix,N_int32 flipY);
void        NAPI REF2D_StretchBltBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
N_int32     NAPI REF2D_ConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32     NAPI REF2D_ConvertBltBM_Obsolete(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix);
N_int32     NAPI REF2D_StretchConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32     NAPI REF2D_StretchConvertBltBM_Obsolete(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix);
N_int32     NAPI REF2D_BitBltFxTest(GA_bltFx *fx);
void        NAPI REF2D_BitBltFx(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI REF2D_BitBltFxLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI REF2D_BitBltFxSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI REF2D_BitBltFxBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI REF2D_GetBitmapSys(void *dstAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI REF2D_GetBitmapBM(void *dstAddr,N_int32 dstPhysAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI REF2D_DrawEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B);
void        NAPI REF2D_ClipEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
GA_color    NAPI REF2D_GetPixel(N_int32 x,N_int32 y);
void        NAPI REF2D_PutPixel(N_int32 x,N_int32 y);
N_int32     NAPI REF2D_DrawClippedLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI REF2D_DrawClippedBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI REF2D_DrawClippedStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI REF2D_DrawClippedBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void        NAPI REF2D_BitBltPlaneMasked(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask);
void        NAPI REF2D_BitBltPlaneMaskedLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask);
void        NAPI REF2D_BitBltPlaneMaskedSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask,N_int32 flipY);
void        NAPI REF2D_BitBltPlaneMaskedBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask);
void        NAPI REF2D_DrawStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI REF2D_DrawBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
N_int32     NAPI REF2D_DrawClippedStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI REF2D_DrawClippedBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);

/* Bug fix workarounds for ROP3 functions */

void        NAPI REF2D_BugFixBitBltPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI REF2D_BugFixBitBltPattLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI REF2D_BugFixBitBltPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI REF2D_BugFixBitBltPattBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI REF2D_BugFixBitBltColorPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI REF2D_BugFixBitBltColorPattLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI REF2D_BugFixBitBltColorPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI REF2D_BugFixBitBltColorPattBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);

/* Assembler stubs to call the functions via the global REF_draw2d */

void        NAPI STUB_DrawScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI STUB_DrawPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI STUB_DrawColorPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI STUB_DrawEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI STUB_DrawPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI STUB_DrawColorPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI STUB_DrawFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI STUB_DrawPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI STUB_DrawColorPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI STUB_DrawRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI STUB_DrawRectExt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI STUB_DrawRectLin(N_int32 dstOfs,N_int32 dstPitch,N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI STUB_DrawPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI STUB_DrawColorPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI STUB_DrawRectPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI STUB_DrawRectColorPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI STUB_DrawTrap(GA_trap *trap);
void        NAPI STUB_DrawPattTrap(GA_trap *trap);
void        NAPI STUB_DrawColorPattTrap(GA_trap *trap);
void        NAPI STUB_DrawLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void        NAPI STUB_DrawBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
void        NAPI STUB_DrawStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI STUB_DrawBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
void        NAPI STUB_PutMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI STUB_PutMonoImageMSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent);
void        NAPI STUB_PutMonoImageMSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 imagePhysAddr,N_int32 transparent);
void        NAPI STUB_PutMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI STUB_PutMonoImageLSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent);
void        NAPI STUB_PutMonoImageLSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 imagePhysAddr,N_int32 transparent);
void        NAPI STUB_ClipMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI STUB_ClipMonoImageMSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI STUB_ClipMonoImageMSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 imagePhysAddr,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI STUB_ClipMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI STUB_ClipMonoImageLSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI STUB_ClipMonoImageLSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 imagePhysAddr,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI STUB_BitBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI STUB_BitBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI STUB_BitBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
void        NAPI STUB_BitBltBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI STUB_BitBltPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI STUB_BitBltPattLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI STUB_BitBltPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI STUB_BitBltPattBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI STUB_BitBltColorPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI STUB_BitBltColorPattLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI STUB_BitBltColorPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI STUB_BitBltColorPattBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI STUB_SrcTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI STUB_SrcTransBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI STUB_SrcTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI STUB_SrcTransBltBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI STUB_DstTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI STUB_DstTransBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI STUB_DstTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI STUB_DstTransBltBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI STUB_StretchBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI STUB_StretchBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI STUB_StretchBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix,N_int32 flipY);
void        NAPI STUB_StretchBltBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
N_int32     NAPI STUB_ConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32     NAPI STUB_ConvertBltBM_Obsolete(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix);
N_int32     NAPI STUB_StretchConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32     NAPI STUB_StretchConvertBltBM_Obsolete(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix);
N_int32     NAPI STUB_BitBltFxTest(GA_bltFx *fx);
void        NAPI STUB_BitBltFx(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI STUB_BitBltFxLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI STUB_BitBltFxSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI STUB_BitBltFxBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI STUB_GetBitmapSys(void *dstAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI STUB_GetBitmapBM(void *dstAddr,N_int32 dstPhysAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI STUB_DrawEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B);
void        NAPI STUB_ClipEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
GA_color    NAPI STUB_GetPixel(N_int32 x,N_int32 y);
void        NAPI STUB_PutPixel(N_int32 x,N_int32 y);
N_int32     NAPI STUB_DrawClippedLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI STUB_DrawClippedBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI STUB_DrawClippedStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI STUB_DrawClippedBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void        NAPI STUB_BitBltPlaneMasked(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask);
void        NAPI STUB_BitBltPlaneMaskedLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask);
void        NAPI STUB_BitBltPlaneMaskedSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask,N_int32 flipY);
void        NAPI STUB_BitBltPlaneMaskedBM(void *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask);
void        NAPI STUB_DrawStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI STUB_DrawBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
N_int32     NAPI STUB_DrawClippedStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI STUB_DrawClippedBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);

/* No-op rendering function for GA_R2_NOP (D) raster ops */

void        REF2D_nop(void);

/* Pixel blending functions (ARGB format) */

GA_color    REF2D_blendSrc_Zero(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendSrc_One(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendSrc_SrcColor(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendSrc_OneMinusSrcColor(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendSrc_SrcAlpha(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendSrc_OneMinusSrcAlpha(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendSrc_DstAlpha(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendSrc_OneMinusDstAlpha(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendSrc_DstColor(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendSrc_OneMinusDstColor(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendSrc_SrcAlphaSaturate(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendSrc_ConstantColor(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendSrc_OneMinusConstantColor(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendSrc_ConstantAlpha(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendSrc_OneMinusConstantAlpha(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendDst_Zero(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendDst_One(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendDst_SrcColor(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendDst_OneMinusSrcColor(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendDst_SrcAlpha(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendDst_OneMinusSrcAlpha(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendDst_DstAlpha(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendDst_OneMinusDstAlpha(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendDst_DstColor(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendDst_OneMinusDstColor(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendDst_SrcAlphaSaturate(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendDst_ConstantColor(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendDst_OneMinusConstantColor(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendDst_ConstantAlpha(GA_color S,GA_color D,GA_color C,GA_color A);
GA_color    REF2D_blendDst_OneMinusConstantAlpha(GA_color S,GA_color D,GA_color C,GA_color A);

/* Pixel processing functions for BitBltFx */

GA_color    REF2D_processSrcColorKey(GA_color S,GA_color D);
GA_color    REF2D_processSrcColorKeyRange(GA_color S,GA_color D);
GA_color    REF2D_processDstColorKey(GA_color S,GA_color D);
GA_color    REF2D_processDstColorKeyRange(GA_color S,GA_color D);
GA_color    REF2D_processSrcColorKeyMasked(GA_color S,GA_color D);
GA_color    REF2D_processSrcColorKeyRangeMasked(GA_color S,GA_color D);
GA_color    REF2D_processDstColorKeyMasked(GA_color S,GA_color D);
GA_color    REF2D_processDstColorKeyRangeMasked(GA_color S,GA_color D);

/* 2D 1bpp reference functions */

DECLARE_ROP2_PIXELS(1);
DECLARE_ROP3_PIXELS(1);
GA_color    REF2D_readPixel_1(uchar *addr DCL_X_ARG);
void        REF2D_drawPixelSrcColorKey_1(uchar *addr DCL_X_ARG,GA_color color);
void        REF2D_drawPixelSrcColorKeyRange_1(uchar *addr DCL_X_ARG,GA_color color);
void        REF2D_fillScan_1_Generic(uchar *addr DCL_X_ARG,int width,GA_color);
void        REF2D_fillScan_1_S(uchar *addr DCL_X_ARG,int width,GA_color);
void        REF2D_drawScan_1_Generic(uchar *addr DCL_X_ARG,int width);
void        REF2D_drawScan_1_S(uchar *addr DCL_X_ARG,int width);
void        REF2D_drawPattScan_1_Generic(uchar *addr,int x1,int y,int width);
void        REF2D_drawPattScan_1_S(uchar *addr,int x1,int y,int width);
void        REF2D_drawColorPattScan_1_Generic(uchar *addr,int x1,int y,int width);
void        REF2D_drawColorPattScan_1_S(uchar *addr,int x1,int y,int width);
void        REF2D_drawLineXPosLo_1_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosLo_1_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosHi_1_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosHi_1_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegLo_1_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegLo_1_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegHi_1_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegHi_1_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawStippleLineXPosLo_1_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosLo_1_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosHi_1_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosHi_1_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegLo_1_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegLo_1_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegHi_1_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegHi_1_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStyleLineXPosLo_1_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosLo_1_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosHi_1_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosHi_1_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegLo_1_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegLo_1_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegHi_1_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegHi_1_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawMonoMSBTrans_1_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBTrans_1_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBOpaque_1_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBOpaque_1_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawClipMonoMSBTrans_1_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBTrans_1_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBOpaque_1_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBOpaque_1_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawMonoLSBTrans_1_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBTrans_1_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBOpaque_1_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBOpaque_1_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawClipMonoLSBTrans_1_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBTrans_1_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBOpaque_1_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBOpaque_1_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawMonoCursor_1(uchar *d,uchar *mask,uchar *image DCL_X_ARG,int width);
void        REF2D_drawColorCursor_1(uchar *d,uchar *mask,uchar *image DCL_X_ARG,int width);
void        REF2D_readScan_1(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScan_1_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScan_1_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScanPatt_1(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
void        REF2D_bltScanColorPatt_1(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
void        REF2D_srcTransBltScan_1_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_srcTransBltScan_1_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_srcTransBltScan_1_Masked(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_1_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_1_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_1_Masked(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan2x_1_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan2x_1_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan_1_S(uchar *d,uchar *s,int width DCL_X_ARG2,int *repCounts);
void        REF2D_stretchScan_1_Generic(uchar *d,uchar *s DCL_X_ARG2,int width,int *repCounts);
GA_color *  REF2D_buildTranslateVector_1(GA_color *translate,GA_palette *dstPal,GA_palette *srcPal,int srcColors);
void        REF2D_drawEllipse_1_S(int ileft,int itop,int A,int B);
void        REF2D_drawEllipse_1_Generic(int ileft,int itop,int A,int B);
void        REF2D_clipEllipse_1_Generic(int ileft,int itop,int A,int B,int clipLeft,int clipTop,int clipRight,int clipBottom);

/* 2D 4bpp reference functions */

DECLARE_ROP2_PIXELS(4);
DECLARE_ROP3_PIXELS(4);
DECLARE_ROP2_MASKED_PIXELS(4);
DECLARE_ROP3_MASKED_PIXELS(4);
GA_color    REF2D_readPixel_4(uchar *addr DCL_X_ARG);
void        REF2D_drawPixelSrcColorKey_4(uchar *addr DCL_X_ARG,GA_color color);
void        REF2D_drawPixelSrcColorKeyRange_4(uchar *addr DCL_X_ARG,GA_color color);
void        REF2D_fillScan_4_Generic(uchar *addr DCL_X_ARG,int width,GA_color);
void        REF2D_fillScan_4_S(uchar *addr DCL_X_ARG,int width,GA_color);
void        REF2D_drawScan_4_Generic(uchar *addr DCL_X_ARG,int width);
void        REF2D_drawScan_4_S(uchar *addr DCL_X_ARG,int width);
void        REF2D_drawPattScan_4_Generic(uchar *addr,int x1,int y,int width);
void        REF2D_drawPattScan_4_S(uchar *addr,int x1,int y,int width);
void        REF2D_drawColorPattScan_4_Generic(uchar *addr,int x1,int y,int width);
void        REF2D_drawColorPattScan_4_S(uchar *addr,int x1,int y,int width);
void        REF2D_drawLineXPosLo_4_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosLo_4_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosHi_4_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosHi_4_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegLo_4_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegLo_4_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegHi_4_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegHi_4_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawStippleLineXPosLo_4_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosLo_4_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosHi_4_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosHi_4_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegLo_4_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegLo_4_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegHi_4_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegHi_4_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStyleLineXPosLo_4_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosLo_4_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosHi_4_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosHi_4_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegLo_4_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegLo_4_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegHi_4_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegHi_4_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawMonoMSBTrans_4_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBTrans_4_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBOpaque_4_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBOpaque_4_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawClipMonoMSBTrans_4_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBTrans_4_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBOpaque_4_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBOpaque_4_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawMonoLSBTrans_4_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBTrans_4_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBOpaque_4_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBOpaque_4_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawClipMonoLSBTrans_4_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBTrans_4_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBOpaque_4_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBOpaque_4_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawMonoCursor_4(uchar *d,uchar *mask,uchar *image DCL_X_ARG,int width);
void        REF2D_drawColorCursor_4(uchar *d,uchar *mask,uchar *image DCL_X_ARG,int width);
void        REF2D_readScan_4(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScan_4_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScan_4_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScanPatt_4(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
void        REF2D_bltScanColorPatt_4(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
void        REF2D_srcTransBltScan_4_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_srcTransBltScan_4_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_srcTransBltScan_4_Masked(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_4_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_4_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_4_Masked(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan2x_4_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan2x_4_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan_4_S(uchar *d,uchar *s,int width DCL_X_ARG2,int *repCounts);
void        REF2D_stretchScan_4_Generic(uchar *d,uchar *s DCL_X_ARG2,int width,int *repCounts);
GA_color *  REF2D_buildTranslateVector_4(GA_color *translate,GA_palette *dstPal,GA_palette *srcPal,int srcColors);
void        REF2D_convScan4_4_S(uchar *d,uchar *s DCL_X_ARG,int width,int odd,GA_color *translate);
void        REF2D_convScan8_4_S(uchar *d,uchar *s DCL_X_ARG,int width,GA_color *translate);
void        REF2D_drawEllipse_4_S(int ileft,int itop,int A,int B);
void        REF2D_drawEllipse_4_Generic(int ileft,int itop,int A,int B);
void        REF2D_clipEllipse_4_Generic(int ileft,int itop,int A,int B,int clipLeft,int clipTop,int clipRight,int clipBottom);

/* 2D 8bpp reference functions */

DECLARE_ROP2_PIXELS(8);
DECLARE_ROP3_PIXELS(8);
DECLARE_ROP2_MASKED_PIXELS(8);
DECLARE_ROP3_MASKED_PIXELS(8);
void        REF2D_setupDither_8(N_int32 y);
GA_color    REF2D_readPixel_8(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel16_8(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24_8(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32_8(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel15_8_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel16_8_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24RGB_8_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24BGR_8_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_8_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_8_Dithered(uchar *addr DCL_X_ARG);
void        REF2D_drawPixelSrcColorKey_8(uchar *addr DCL_X_ARG,GA_color color);
void        REF2D_drawPixelSrcColorKeyRange_8(uchar *addr DCL_X_ARG,GA_color color);
void        REF2D_fillScan_8_Generic(uchar *addr DCL_X_ARG,int width,GA_color);
void        REF2D_fillScan_8_S(uchar *addr DCL_X_ARG,int width,GA_color);
void        REF2D_drawScan_8_Generic(uchar *addr DCL_X_ARG,int width);
void        REF2D_drawScan_8_S(uchar *addr DCL_X_ARG,int width);
void        REF2D_drawPattScan_8_Generic(uchar *addr,int x1,int y,int width);
void        REF2D_drawPattScan_8_S(uchar *addr,int x1,int y,int width);
void        REF2D_drawColorPattScan_8_Generic(uchar *addr,int x1,int y,int width);
void        REF2D_drawColorPattScan_8_S(uchar *addr,int x1,int y,int width);
void        REF2D_drawLineXPosLo_8_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosLo_8_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosHi_8_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosHi_8_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegLo_8_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegLo_8_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegHi_8_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegHi_8_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawStippleLineXPosLo_8_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosLo_8_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosHi_8_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosHi_8_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegLo_8_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegLo_8_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegHi_8_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegHi_8_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStyleLineXPosLo_8_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosLo_8_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosHi_8_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosHi_8_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegLo_8_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegLo_8_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegHi_8_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegHi_8_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawMonoMSBTrans_8_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBTrans_8_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBOpaque_8_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBOpaque_8_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawClipMonoMSBTrans_8_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBTrans_8_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBOpaque_8_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBOpaque_8_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawMonoLSBTrans_8_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBTrans_8_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBOpaque_8_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBOpaque_8_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawClipMonoLSBTrans_8_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBTrans_8_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBOpaque_8_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBOpaque_8_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawMonoCursor_8(uchar *d,uchar *mask,uchar *image DCL_X_ARG,int width);
void        REF2D_drawColorCursor_8(uchar *d,uchar *mask,uchar *image DCL_X_ARG,int width);
void        REF2D_readScan_8(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScan_8_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScan_8_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScanPatt_8(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
void        REF2D_bltScanColorPatt_8(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
void        REF2D_srcTransBltScan_8_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_srcTransBltScan_8_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_srcTransBltScan_8_Masked(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_8_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_8_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_8_Masked(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan2x_8_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan2x_8_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan_8_S(uchar *d,uchar *s,int width DCL_X_ARG2,int *repCounts);
void        REF2D_stretchScan_8_Generic(uchar *d,uchar *s DCL_X_ARG2,int width,int *repCounts);
GA_color *  REF2D_buildTranslateVector_8(GA_color *translate,GA_palette *dstPal,GA_palette *srcPal,int srcColors);
void        REF2D_convScan4_8_S(uchar *d,uchar *s DCL_X_ARG,int width,int odd,GA_color *translate);
void        REF2D_convScan8_8_S(uchar *d,uchar *s DCL_X_ARG,int width,GA_color *translate);
void        REF2D_drawEllipse_8_S(int ileft,int itop,int A,int B);
void        REF2D_drawEllipse_8_Generic(int ileft,int itop,int A,int B);
void        REF2D_clipEllipse_8_Generic(int ileft,int itop,int A,int B,int clipLeft,int clipTop,int clipRight,int clipBottom);

/* 2D 16bpp reference functions */

DECLARE_ROP2_PIXELS(16);
DECLARE_ROP3_PIXELS(16);
DECLARE_ROP2_MASKED_PIXELS(16);
DECLARE_ROP3_MASKED_PIXELS(16);
DECLARE_BLENDED_PIXELS(15);
DECLARE_BLENDED_PIXELS(16);
void        REF2D_setupDither_8A(N_int32 y);
void        REF2D_setupDither_16(N_int32 y);
GA_color    REF2D_readPixel_16(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel16_15(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel15_16(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24RGB_15(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24BGR_15(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24RGB_16(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24BGR_16(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_15(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_15(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_16(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_16(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_15_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_15_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_15_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_15_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_16_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_16_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_16_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_16_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24RGB_15_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24BGR_15_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24RGB_16_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24BGR_16_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_15_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_15_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_16_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_16_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_15_Dithered_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_15_Dithered_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_15_Dithered_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_15_Dithered_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_16_Dithered_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_16_Dithered_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_16_Dithered_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_16_Dithered_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_8A(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_8A(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_8A(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_8A(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_8A_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_8A_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_8A_Dithered(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_8A_Dithered(uchar *addr DCL_X_ARG);
void        REF2D_drawPixelSrcColorKey_16(uchar *addr DCL_X_ARG,GA_color color);
void        REF2D_drawPixelSrcColorKeyRange_16(uchar *addr DCL_X_ARG,GA_color color);
void        REF2D_fillScan_16_Generic(uchar *addr DCL_X_ARG,int width,GA_color);
void        REF2D_fillScan_16_S(uchar *addr DCL_X_ARG,int width,GA_color);
void        REF2D_drawScan_16_Generic(uchar *addr DCL_X_ARG,int width);
void        REF2D_drawScan_16_S(uchar *addr DCL_X_ARG,int width);
void        REF2D_drawPattScan_16_Generic(uchar *addr,int x1,int y,int width);
void        REF2D_drawPattScan_16_S(uchar *addr,int x1,int y,int width);
void        REF2D_drawColorPattScan_16_Generic(uchar *addr,int x1,int y,int width);
void        REF2D_drawColorPattScan_16_S(uchar *addr,int x1,int y,int width);
void        REF2D_drawLineXPosLo_16_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosLo_16_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosHi_16_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosHi_16_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegLo_16_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegLo_16_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegHi_16_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegHi_16_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawStippleLineXPosLo_16_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosLo_16_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosHi_16_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosHi_16_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegLo_16_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegLo_16_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegHi_16_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegHi_16_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStyleLineXPosLo_16_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosLo_16_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosHi_16_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosHi_16_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegLo_16_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegLo_16_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegHi_16_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegHi_16_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawMonoMSBTrans_16_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBTrans_16_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBOpaque_16_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBOpaque_16_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawClipMonoMSBTrans_16_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBTrans_16_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBOpaque_16_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBOpaque_16_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawMonoLSBTrans_16_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBTrans_16_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBOpaque_16_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBOpaque_16_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawClipMonoLSBTrans_16_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBTrans_16_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBOpaque_16_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBOpaque_16_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawMonoCursor_16(uchar *d,uchar *mask,uchar *image DCL_X_ARG,int width);
void        REF2D_drawColorCursor_16(uchar *d,uchar *mask,uchar *image DCL_X_ARG,int width);
void        REF2D_readScan_16(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScan_16_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScan_16_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScanPatt_16(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
void        REF2D_bltScanColorPatt_16(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
void        REF2D_srcTransBltScan_16_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_srcTransBltScan_16_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_srcTransBltScan_16_Masked(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_16_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_16_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_16_Masked(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan2x_16_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan2x_16_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan_16_S(uchar *d,uchar *s,int width DCL_X_ARG2,int *repCounts);
void        REF2D_stretchScan_16_Generic(uchar *d,uchar *s DCL_X_ARG2,int width,int *repCounts);
GA_color *  REF2D_buildTranslateVector_15(GA_color *translate,GA_palette *dstPal,GA_palette *srcPal,int srcColors);
GA_color *  REF2D_buildTranslateVector_16(GA_color *translate,GA_palette *dstPal,GA_palette *srcPal,int srcColors);
GA_color    REF2D_findClosest_15(GA_color color);
GA_color    REF2D_findClosest_16(GA_color color);
void        REF2D_convScan4_16_S(uchar *d,uchar *s DCL_X_ARG,int width,int odd,GA_color *translate);
void        REF2D_convScan8_16_S(uchar *d,uchar *s DCL_X_ARG,int width,GA_color *translate);
void        REF2D_drawEllipse_16_S(int ileft,int itop,int A,int B);
void        REF2D_drawEllipse_16_Generic(int ileft,int itop,int A,int B);
void        REF2D_clipEllipse_16_Generic(int ileft,int itop,int A,int B,int clipLeft,int clipTop,int clipRight,int clipBottom);

/* 2D 24bpp reference functions */

DECLARE_ROP2_PIXELS(24);
DECLARE_ROP3_PIXELS(24);
DECLARE_ROP2_MASKED_PIXELS(24);
DECLARE_ROP3_MASKED_PIXELS(24);
DECLARE_BLENDED_PIXELS(24);
GA_color    REF2D_readPixel_24(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel15_24RGB(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel15_24BGR(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel16_24RGB(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel16_24BGR(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24RGB_24BGR(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_24RGB(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_24RGB(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_24RGB(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_24RGB(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_24BGR(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_24BGR(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_24BGR(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_24BGR(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_24RGB_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_24RGB_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_24RGB_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_24RGB_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_24BGR_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_24BGR_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_24BGR_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_24BGR_Blend(uchar *addr DCL_X_ARG);
void        REF2D_drawPixelSrcColorKey_24(uchar *addr DCL_X_ARG,GA_color color);
void        REF2D_drawPixelSrcColorKeyRange_24(uchar *addr DCL_X_ARG,GA_color color);
void        REF2D_fillScan_24_Generic(uchar *addr DCL_X_ARG,int width,GA_color);
void        REF2D_fillScan_24_S(uchar *addr DCL_X_ARG,int width,GA_color);
void        REF2D_drawScan_24_Generic(uchar *addr DCL_X_ARG,int width);
void        REF2D_drawScan_24_S(uchar *addr DCL_X_ARG,int width);
void        REF2D_drawPattScan_24_Generic(uchar *addr,int x1,int y,int width);
void        REF2D_drawColorPattScan_24_Generic(uchar *addr,int x1,int y,int width);
void        REF2D_drawLineXPosLo_24_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosLo_24_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosHi_24_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosHi_24_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegLo_24_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegLo_24_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegHi_24_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegHi_24_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawStippleLineXPosLo_24_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosHi_24_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegLo_24_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegHi_24_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStyleLineXPosLo_24_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosHi_24_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegLo_24_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegHi_24_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawMonoMSBTrans_24_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBOpaque_24_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawClipMonoMSBTrans_24_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBOpaque_24_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawMonoLSBTrans_24_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBOpaque_24_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawClipMonoLSBTrans_24_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBOpaque_24_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawMonoCursor_24(uchar *d,uchar *mask,uchar *image DCL_X_ARG,int width);
void        REF2D_drawColorCursor_24(uchar *d,uchar *mask,uchar *image DCL_X_ARG,int width);
void        REF2D_readScan_24(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScan_24_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScan_24_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScanPatt_24(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
void        REF2D_bltScanColorPatt_24(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
void        REF2D_srcTransBltScan_24_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_srcTransBltScan_24_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_24_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_24_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan2x_24_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan2x_24_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan_24_S(uchar *d,uchar *s DCL_X_ARG2,int width,int *repCounts);
void        REF2D_stretchScan_24_Generic(uchar *d,uchar *s DCL_X_ARG2,int width,int *repCounts);
GA_color *  REF2D_buildTranslateVector_24RGB(GA_color *translate,GA_palette *dstPal,GA_palette *srcPal,int srcColors);
GA_color *  REF2D_buildTranslateVector_24BGR(GA_color *translate,GA_palette *dstPal,GA_palette *srcPal,int srcColors);
GA_color    REF2D_findClosest_24RGB(GA_color color);
GA_color    REF2D_findClosest_24BGR(GA_color color);
void        REF2D_convScan4_24_S(uchar *d,uchar *s DCL_X_ARG,int width,int odd,GA_color *translate);
void        REF2D_convScan8_24_S(uchar *d,uchar *s DCL_X_ARG,int width,GA_color *translate);
void        REF2D_drawEllipse_24_S(int ileft,int itop,int A,int B);
void        REF2D_drawEllipse_24_Generic(int ileft,int itop,int A,int B);
void        REF2D_clipEllipse_24_Generic(int ileft,int itop,int A,int B,int clipLeft,int clipTop,int clipRight,int clipBottom);

/* 2D 32bpp reference functions */

DECLARE_ROP2_PIXELS(32);
DECLARE_ROP3_PIXELS(32);
DECLARE_ROP2_MASKED_PIXELS(32);
DECLARE_ROP3_MASKED_PIXELS(32);
DECLARE_BLENDED_PIXELS(Axxx32);
DECLARE_BLENDED_PIXELS(xxxA32);
GA_color    REF2D_readPixel_32(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel15_32ARGB(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel15_32ABGR(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel15_32RGBA(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel15_32BGRA(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel16_32ARGB(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel16_32ABGR(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel16_32RGBA(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel16_32BGRA(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel15_32ARGB_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel15_32ABGR_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel15_32RGBA_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel15_32BGRA_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24RGB_32ARGB(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24RGB_32ABGR(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24RGB_32RGBA(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24RGB_32BGRA(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24BGR_32ARGB(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24BGR_32ABGR(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24BGR_32RGBA(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel24BGR_32BGRA(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_32ARGB(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_32ARGB(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_32ARGB(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_32ABGR(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_32ABGR(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_32ABGR(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_32RGBA(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_32RGBA(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_32RGBA(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_32BGRA(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_32BGRA(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_32BGRA(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_32ARGB_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_32ARGB_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_32ARGB_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_32ABGR_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_32ABGR_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_32ABGR_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_32RGBA_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_32RGBA_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32BGRA_32RGBA_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ARGB_32BGRA_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32ABGR_32BGRA_Blend(uchar *addr DCL_X_ARG);
GA_color    REF2D_convertPixel32RGBA_32BGRA_Blend(uchar *addr DCL_X_ARG);
void        REF2D_drawPixelSrcColorKey_32(uchar *addr DCL_X_ARG,GA_color color);
void        REF2D_drawPixelSrcColorKeyRange_32(uchar *addr DCL_X_ARG,GA_color color);
void        REF2D_fillScan_32_Generic(uchar *addr DCL_X_ARG,int width,GA_color);
void        REF2D_fillScan_32_S(uchar *addr DCL_X_ARG,int width,GA_color);
void        REF2D_drawScan_32_Generic(uchar *addr DCL_X_ARG,int width);
void        REF2D_drawScan_32_S(uchar *addr DCL_X_ARG,int width);
void        REF2D_drawPattScan_32_Generic(uchar *addr,int x1,int y,int width);
void        REF2D_drawPattScan_32_S(uchar *addr,int x1,int y,int width);
void        REF2D_drawColorPattScan_32_Generic(uchar *addr,int x1,int y,int width);
void        REF2D_drawColorPattScan_32_S(uchar *addr,int x1,int y,int width);
void        REF2D_drawLineXPosLo_32_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosLo_32_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosHi_32_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXPosHi_32_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegLo_32_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegLo_32_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegHi_32_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawLineXNegHi_32_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr);
void        REF2D_drawStippleLineXPosLo_32_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosLo_32_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosHi_32_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXPosHi_32_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegLo_32_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegLo_32_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegHi_32_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStippleLineXNegHi_32_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent);
void        REF2D_drawStyleLineXPosLo_32_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosLo_32_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosHi_32_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXPosHi_32_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegLo_32_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegLo_32_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegHi_32_S(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawStyleLineXNegHi_32_Generic(uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent);
void        REF2D_drawMonoMSBTrans_32_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBTrans_32_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBOpaque_32_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoMSBOpaque_32_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawClipMonoMSBTrans_32_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBTrans_32_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBOpaque_32_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoMSBOpaque_32_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawMonoLSBTrans_32_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBTrans_32_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBOpaque_32_Generic(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawMonoLSBOpaque_32_S(uchar *addr,uchar *image DCL_X_ARG,int width);
void        REF2D_drawClipMonoLSBTrans_32_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBTrans_32_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBOpaque_32_Generic(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawClipMonoLSBOpaque_32_S(int x,uchar *addr,uchar *image,int width,int clipLeft);
void        REF2D_drawMonoCursor_32(uchar *d,uchar *mask,uchar *image DCL_X_ARG,int width);
void        REF2D_drawColorCursor_32(uchar *d,uchar *mask,uchar *image DCL_X_ARG,int width);
void        REF2D_readScan_32(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScan_32_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScan_32_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_bltScanPatt_32(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
void        REF2D_bltScanColorPatt_32(uchar *d,uchar *s DCL_X_ARG3,int x1,int y,int width);
void        REF2D_srcTransBltScan_32_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_srcTransBltScan_32_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_srcTransBltScan_32_Masked(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_32_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_32_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_dstTransBltScan_32_Masked(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan2x_32_S(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan2x_32_Generic(uchar *d,uchar *s DCL_X_ARG2,int width);
void        REF2D_stretchScan_32_S(uchar *d,uchar *s,int width DCL_X_ARG2,int *repCounts);
void        REF2D_stretchScan_32_Generic(uchar *d,uchar *s DCL_X_ARG2,int width,int *repCounts);
GA_color *  REF2D_buildTranslateVector_32ARGB(GA_color *translate,GA_palette *dstPal,GA_palette *srcPal,int srcColors);
GA_color *  REF2D_buildTranslateVector_32ABGR(GA_color *translate,GA_palette *dstPal,GA_palette *srcPal,int srcColors);
GA_color *  REF2D_buildTranslateVector_32RGBA(GA_color *translate,GA_palette *dstPal,GA_palette *srcPal,int srcColors);
GA_color *  REF2D_buildTranslateVector_32BGRA(GA_color *translate,GA_palette *dstPal,GA_palette *srcPal,int srcColors);
GA_color    REF2D_findClosest_32RGBA(GA_color color);
GA_color    REF2D_findClosest_32BGRA(GA_color color);
void        REF2D_convScan4_32_S(uchar *d,uchar *s DCL_X_ARG,int width,int odd,GA_color *translate);
void        REF2D_convScan8_32_S(uchar *d,uchar *s DCL_X_ARG,int width,GA_color *translate);
void        REF2D_drawEllipse_32_S(int ileft,int itop,int A,int B);
void        REF2D_drawEllipse_32_Generic(int ileft,int itop,int A,int B);
void        REF2D_clipEllipse_32_Generic(int ileft,int itop,int A,int B,int clipLeft,int clipTop,int clipRight,int clipBottom);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __REF2DINT_H */

